// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver

import com.intellij.codeInsight.hint.DeclarationRangeUtil
import com.intellij.codeWithMe.ClientId
import com.intellij.ide.structureView.StructureViewBuilder
import com.intellij.ide.structureView.StructureViewModel
import com.intellij.ide.structureView.StructureViewTreeElement
import com.intellij.ide.structureView.newStructureView.StructureViewComponent
import com.intellij.ide.util.treeView.smartTree.TreeElement
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.fileEditor.FileEditor
import com.intellij.openapi.fileEditor.FileEditorManagerEvent
import com.intellij.openapi.fileEditor.FileEditorManagerListener
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.pom.Navigatable
import com.intellij.psi.PsiElement
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.hasTrueValue
import com.jetbrains.rd.util.reactive.whenTrue
import com.jetbrains.rdserver.vfs.createDocumentIdFromFile
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import java.util.concurrent.atomic.AtomicInteger

data class LastStructureViewData(val model: StructureViewModel, val file: VirtualFile)

@Suppress("NonDefaultConstructor")
class BackendFileStructureHost(
  lifetime: Lifetime,
  private val clientId: ClientId,
  private val project: Project,
  private val solution: Solution
) {

  companion object {
    fun getInstance(project: Project): BackendFileStructureHost = RdServerPerClientComponents.getService(project)
    private val LOG = logger<BackendFileStructureHost>()
  }

  private val idCounter = AtomicInteger()
  private val idToElementMap = Int2ObjectOpenHashMap<TreeElement>()
  private var lastSelectedFile: VirtualFile? = null
  private var lastSelectedEditor: FileEditor? = null
  private var lastStructureViewData: LastStructureViewData? = null

  init {
    val model = solution.codeStructureTreeModel

    model.navigateTo.advise(lifetime) { navigateTo ->
      (idToElementMap.get(navigateTo.node) as? Navigatable)?.navigate(navigateTo.requestFocus)
    }

    model.toolWindowVisible.whenTrue(lifetime) {
      val file = lastSelectedFile ?: return@whenTrue
      val editor = lastSelectedEditor ?: return@whenTrue
      updateFileStructure(solution, file, editor)
    }

    StructureViewComponent.registerPsiListener(project, lifetime.createNestedDisposable(), Runnable {
      application.invokeLater {
        ClientId.withClientId(clientId) {
          if (lifetime.isNotAlive) return@invokeLater
          lastStructureViewData?.let { sendFileStructureToFrontend(solution, it.model, it.file) }
        }
      }
    })
  }

  class MyListener : FileEditorManagerListener {
    override fun selectionChanged(event: FileEditorManagerEvent) {
      if (ClientId.isCurrentlyUnderLocalId) return

      getInstance(event.manager.project).selectionChanged(event)
    }
  }

  private fun selectionChanged(event: FileEditorManagerEvent) {
    lastSelectedFile = event.newFile
    lastSelectedEditor = event.newEditor

    val newFile = event.newFile ?: return
    val newEditor = event.newEditor ?: return

    if (solution.codeStructureTreeModel.toolWindowVisible.hasTrueValue) {
      updateFileStructure(solution, newFile, newEditor)
    }
  }

  private fun updateFileStructure(solution: Solution, file: VirtualFile, editor: FileEditor) = ClientId.withClientId(clientId) {
    idToElementMap.clear()
    val structureViewBuilder = StructureViewBuilder.PROVIDER.getStructureViewBuilder(file.fileType, file, project)
    if (structureViewBuilder == null) {
      // todo use RdFileId
      val documentId = createDocumentIdFromFile(project, file)
      solution.codeStructureTreeModel.treeNodes.set(CodeStructureTreeNodesList(documentId, emptyList()))
      lastStructureViewData = null
      return
    }
    val structureView = structureViewBuilder.createStructureView(editor, project)
    val model = structureView.treeModel
    lastStructureViewData = LastStructureViewData(model, file)
    sendFileStructureToFrontend(solution, model, file)
  }

  private fun sendFileStructureToFrontend(solution: Solution, model: StructureViewModel, file: VirtualFile) {
    idToElementMap.clear()
    val nodeList = mutableListOf<CodeStructureTreeNode?>()
    convertStructureTree(model.root, nodeList)
    val finalNodeList = nodeList.filterNotNull()
    // todo use RdFileId
    val documentId = createDocumentIdFromFile(project, file)
    if (finalNodeList.size != nodeList.size) {
      LOG.error("Error in tree construction: node list size differs")
      solution.codeStructureTreeModel.treeNodes.set(CodeStructureTreeNodesList(documentId, emptyList()))
      return
    }
    LOG.info("Sending file structure to frontend")
    solution.codeStructureTreeModel.treeNodes.set(CodeStructureTreeNodesList(documentId, finalNodeList))
  }

  private fun convertStructureTree(node: TreeElement, nodeList: MutableList<CodeStructureTreeNode?>): Int {
    val index = nodeList.size
    nodeList.add(null)
    val childIndexList = mutableListOf<Int>()
    for (child in node.children) {
      childIndexList.add(convertStructureTree(child, nodeList))
    }
    val psiElement = (node as? StructureViewTreeElement)?.value as? PsiElement
    val declarationRange = psiElement?.let { DeclarationRangeUtil.getPossibleDeclarationAtRange(it) }?.toRdTextRange()
    val codeStructureTreeNode = CodeStructureTreeNode(
      idCounter.incrementAndGet(),
      childIndexList,
      psiElement != null,
      psiElement?.textRange?.startOffset ?: 0,
      psiElement?.textRange?.startOffset ?: 0,
      node.presentation.presentableText ?: "",
      node.presentation.getIcon(false)?.toProtocolModel(),
      declarationRange
    )
    nodeList[index] = codeStructureTreeNode
    idToElementMap.put(codeStructureTreeNode.id, node)
    return index
  }
}
