// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.actions.SearchEverywhereAction
import com.intellij.ide.actions.searcheverywhere.*
import com.intellij.navigation.NavigationItem
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.application.runReadAction
import com.intellij.openapi.module.ModuleUtilCore
import com.intellij.openapi.progress.EmptyProgressIndicator
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.psi.PsiElement
import com.intellij.ui.JBColor
import com.intellij.ui.popup.AbstractPopup
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.getOrCreateUserData
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.vfs.createDocumentIdFromFile
import it.unimi.dsi.fastutil.ints.Int2ObjectOpenHashMap
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.atomic.AtomicInteger

@Suppress("NonDefaultConstructor")
class BackendGotoHost(
  lifetime: Lifetime,
  project: Project,
  solution: Solution
) {

  private val idCounter = AtomicInteger()

  init {
    val gotoModel = solution.gotoModel
    gotoModel.searchPopups.view(lifetime) { popupLifetime, _, popupInstance ->
      subscribeForThisPopup(popupLifetime, popupInstance, project)
    }
  }

  private fun subscribeForThisPopup(lifetime: Lifetime, popupInstance: SearchPopup, project: Project) {
    val popupDisposable = object : AbstractPopup() {}
    project.getOrCreateUserData(SearchEverywhereAction.SEARCH_EVERYWHERE_POPUP) { ConcurrentHashMap() }[ClientId.current] = popupDisposable
    lifetime.onTermination {
      project.getOrCreateUserData(SearchEverywhereAction.SEARCH_EVERYWHERE_POPUP) { ConcurrentHashMap() }.remove(ClientId.current)
      Disposer.dispose(popupDisposable)
    }
    val event = AnActionEvent.createFromDataContext("", null) { if (CommonDataKeys.PROJECT.`is`(it)) project else null } // todo: better datacontext
    val model = when (popupInstance.currentPopupKind) {
      GotoKind.GotoType -> ClassSearchEverywhereContributor(event)
      GotoKind.GotoFile -> FileSearchEverywhereContributor(event)
      GotoKind.GotoSymbol -> SymbolSearchEverywhereContributor(event)
      else -> {
        LOG.info("Unsupported popup kind ${popupInstance.currentPopupKind}")
        return
      }
    } as WeightedSearchEverywhereContributor<*>
    val everywhereToggleAction = model.getActions {}.filterIsInstance<SearchEverywhereToggleAction>().single()

    val itemsMap = Int2ObjectOpenHashMap<NavigationItem>()
    popupInstance.input.view(lifetime) { inputLt, gotoInput ->
      LOG.info("Received goto input: $gotoInput")
      val results = mutableListOf<SearchItemData>()
      application.executeOnPooledThread {
        everywhereToggleAction.isEverywhere = gotoInput.scope
        val progress = EmptyProgressIndicator()
        if (!inputLt.onTerminationIfAlive { progress.cancel() }) return@executeOnPooledThread
        model.fetchWeightedElements(gotoInput.text, progress) { descriptor ->
          return@fetchWeightedElements runReadAction {
            val it = descriptor.item
            if (it !is NavigationItem) {
              LOG.info("Received unknown element from goto model: $it")
              return@runReadAction true
            }

            convertItem(it, descriptor.weight)?.let { searchItemData ->
              ProgressManager.checkCanceled()
              itemsMap[searchItemData.id] = it
              results.add(searchItemData)
            }
            results.size < 100
          }
        }
        popupInstance.getItemsTask.fire(GotoResultList(gotoInput.text, gotoInput.scope, results))
      }
    }

    popupInstance.itemExecutedSignal.advise(lifetime) { executedItem ->
      itemsMap.get(executedItem.id)?.navigate(true)
    }
  }

  private fun convertItem(item: NavigationItem, weight: Int): SearchItemData? {
    val icon = item.presentation?.getIcon(false)?.toProtocolModel() ?: run {
      LOG.info("Failed to convert icon to IDEA for item $item")
      return null
    }
    val name = item.name ?: run {
      LOG.info("Failed to convert item with null name: $item")
      return null
    }

    val presentableRichText = richText {
      append(item.presentation?.presentableText ?: name)
      item.presentation?.locationString?.let {
        append(" ", JBColor.GRAY) // unstick location from main text
        append(it, JBColor.GRAY)
      }
    }

    val psiElement = item as? PsiElement
    val moduleName = psiElement?.let { ModuleUtilCore.findModuleForPsiElement(it)?.name }
    val virtualFile = psiElement?.containingFile?.virtualFile
    // todo use fileId
    val documentId = psiElement?.let { createDocumentIdFromFile(it.project, virtualFile) }
    if (virtualFile != null && documentId == null) {
      // not enough permissions
      return null
    }

    return SearchItemData(
      icon,
      idCounter.incrementAndGet(),
      presentableRichText.toString(),
      documentId,
      presentableRichText,
      moduleName,
      null,
      weight,
      emptyList()
    )
  }

  companion object {
    val LOG = getLogger<BackendGotoHost>()
  }
}
