package com.jetbrains.rdserver

import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isLocal
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.components.service
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.ProgressManagerListener
import com.intellij.openapi.progress.Task
import com.intellij.openapi.project.Project
import com.intellij.openapi.wm.ex.ProgressIndicatorEx
import com.intellij.util.ui.update.MergingUpdateQueue
import com.jetbrains.ide.model.notifications.BackgroundTaskModel
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.idea.LifetimedProjectService
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.IScheduler
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.startup.RdServerPerClientSolutions
import com.jetbrains.rdserver.ui.BackendNotificationsHost
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.atomic.AtomicInteger

class BackendProgressManagerService {
  private var taskToId = ConcurrentHashMap<Task, MutableList<Pair<Int, ClientId>>>()
  private var id = AtomicInteger()
  private val originalProgressToTaskMap = ConcurrentHashMap<ProgressIndicator, Task>()
  private val originalProgressToDelegateMap = ConcurrentHashMap<ProgressIndicator, RdProgressIndicator>()

  private fun bindTaskWithModel(clientId: ClientId,
                                task: Task,
                                taskProject: Project): Triple<BackgroundTaskModel, ClientId, IScheduler>? {
    val service = service<BackendProgressManagerService>()
    return ClientId.withClientId(clientId) {
      val notificationsHost = getNotificationsHost(taskProject) ?: return null
      val rdTask = BackgroundTaskModel(task.title, task.isCancellable)
      val currentId = service.id.incrementAndGet()
      service.taskToId.getOrPut(task) { mutableListOf() }.add(currentId to clientId)
      var scheduler : IScheduler? = null
      application.invokeAndWait({
        notificationsHost.notificationsModel.tasks[currentId] = rdTask
        try {
          scheduler = rdTask.protocol.scheduler
        } catch (e: IllegalStateException) { /* not bound */ }
      }, ModalityState.any())
      scheduler?.let { Triple(rdTask, clientId, it) }
    }
  }

  private fun isGlobalProgress(task: Task): Boolean {
    if (task.javaClass.`package`.name.contains("index", true)
        || task.javaClass.`package`.name.contains("vcs", true)
        || task.javaClass.`package`.name.contains("platform.impl", true)
        || !task.isCancellable)
      return true

    return false
  }

  private fun getNotificationsHost(project: Project): BackendNotificationsHost? =
    try {
      RdServerPerClientComponents.getService(project, BackendNotificationsHost::class.java)
    } catch (e: Throwable) {
      null
    }

  private class Listener : ProgressManagerListener {
    private val queue = MergingUpdateQueue("CodeWithMe progress indicators update queue", 1000, true, null)

    override fun onTaskRunnableCreated(task: Task, indicator: ProgressIndicator, continuation: Runnable?) {
      val service = service<BackendProgressManagerService>()
      service.originalProgressToTaskMap[indicator] = task

      val taskProject = task.project ?: return
      val currentClientIdList = if (!ClientId.currentOrNull.isLocal) listOf(ClientId.current) else emptyList()
      val allClientIds = if (service.isGlobalProgress(task))
        taskProject.getUserData(RdServerPerClientSolutions.SOLUTION_KEY)?.keys ?: currentClientIdList
      else
        currentClientIdList

      if (currentClientIdList.size == 1 && currentClientIdList.single() == ClientId.localId)
        return

      val taskList = allClientIds.mapNotNullTo(mutableListOf()) { clientId ->
        service.bindTaskWithModel(clientId, task, taskProject)
      }
      if (indicator is ProgressIndicatorEx) {
        val rdIndicator = RdProgressIndicator(indicator, taskList, queue)
        indicator.addStateDelegate(rdIndicator)
        service.originalProgressToDelegateMap[indicator] = rdIndicator
      }
    }

    override fun onTaskFinished(task: Task, canceled: Boolean, error: Throwable?) {
      val service = service<BackendProgressManagerService>()
      try {
        val idList = service.taskToId.remove(task)
        idList?.forEach { (id, clientId) ->
          ClientId.withClientId(clientId) {
            service.getNotificationsHost(task.project ?: return@forEach)?.notificationsModel?.tasks?.remove(id)
          }
        }
      } finally {
        if (service.originalProgressToTaskMap.size > 100) {
          logger.warn("Progress to task map seems to be too big, it contains ${service.originalProgressToTaskMap.size} elements. " +
                      "Iteration over the whole map for each task completion can drop performance.")
        }

        service.originalProgressToTaskMap.filter { it.value == task }.forEach { (indicator, _) ->
          service.originalProgressToDelegateMap.remove(indicator)
        }
        service.originalProgressToTaskMap.values.removeIf { it == task }
      }
    }
  }

  class Updater(project: Project): LifetimedProjectService(project) {
    init {
      application.invokeLater l1@{
        if (projectServiceLifetime.isNotAlive) return@l1
        FollowMeManager.getInstance(project).users.view(projectServiceLifetime) { _, clientId, _ ->
          if (clientId.isLocal) return@view

          application.invokeLater l2@{
            if (projectServiceLifetime.isNotAlive) return@l2
            val service = service<BackendProgressManagerService>()
            for ((progress, task) in service.originalProgressToTaskMap) {
              if (task.project != project) continue

              val rdIndicator = service.originalProgressToDelegateMap[progress]
              if (rdIndicator == null) {
                logger.warn("Found no delegating indicators for $progress")
                continue
              }

              service.bindTaskWithModel(clientId, task, project)?.let { result ->
                rdIndicator.taskList.add(result)
              }
            }
          }
        }
      }
    }
  }

  companion object {
    private val logger = getLogger<BackendProgressManagerService>()
  }
}