package com.jetbrains.rdserver.terminal

import com.intellij.codeWithMe.ClientId
import com.intellij.icons.AllIcons
import com.intellij.openapi.project.Project
import com.intellij.openapi.wm.ToolWindow
import com.jetbrains.codeWithMe.model.RdTerminalSession
import com.jetbrains.codeWithMe.model.RdTerminalSessionOrigin
import com.jetbrains.codeWithMe.model.TerminalModel
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.bindDisplayName
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.whenTrue
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.permissions.TerminalAccessLevel
import com.intellij.codeWithMe.ui.AccessNotification
import org.jetbrains.plugins.terminal.ShellTerminalWidget
import org.jetbrains.plugins.terminal.TerminalView

@Suppress("NonDefaultConstructor")
class BackendTerminalHost(private val lifetime: Lifetime, val project: Project, protocol: IProtocol) {

  private val model = TerminalModel.create(lifetime, protocol)
  private val clientId = ClientId.current

  init {
    // Subscribe for new client sessions
    model.sessions.view(lifetime) { _, sessionId, session ->
      if (session.origin == RdTerminalSessionOrigin.Host) return@view
      val accessLevel = PermissionManager.getInstance(project).getTerminalAccess(clientId)

      if (accessLevel == TerminalAccessLevel.Disabled || accessLevel == TerminalAccessLevel.Readonly) {
        session.write(CodeWithMePluginBundle.message("terminal.message.access.denied"))
        return@view
      }

      BackendTerminalManager.getInstance(project).createNewSharedTerminal(sessionId, session.name.valueOrNull)
    }

    // Sync all existing shared terminals with this client
    application.invokeLater {
      BackendTerminalManager.getInstance(project).connectClient(clientId)
    }
  }

  fun bind(widget: ShellTerminalWidget, connector: BackendTtyConnector) {
    val sessionId = connector.sessionId ?: return
    if (PermissionManager.getInstance(project).getTerminalAccess(clientId) == TerminalAccessLevel.Disabled) return

    val tabContent = TerminalView.getInstance(project).getContainer(widget).content
    tabContent.putUserData(ToolWindow.SHOW_CONTENT_ICON, java.lang.Boolean.TRUE)
    tabContent.icon = AllIcons.Actions.Share

    // Get an existing session from protocol model or create a new one
    //   if the session was not requested by this client
    val sessionModel = model.sessions[sessionId] ?: RdTerminalSession(RdTerminalSessionOrigin.Host).apply {
      name.set(tabContent.displayName)
      model.sessions[sessionId] = this
    }

    // Bind protocol session name to local tab name
    tabContent.bindDisplayName(connector.lifetime, sessionModel.name)

    sessionModel.input.advise(connector.lifetime) { bytes ->
      val accessLevel = PermissionManager.getInstance(project).getTerminalAccess(clientId)
      if (accessLevel == TerminalAccessLevel.Readonly) return@advise

      val byteArray = bytes.toByteArray()
      val enterPressed = bytes.any { it.toInt() == 13 } && !widget.hasRunningCommands()

      if (accessLevel == TerminalAccessLevel.FullAccess) {
        connector.write(byteArray)
        if (enterPressed) widget.handleEnterPressed()
        else widget.handleAnyKeyPressed()
        return@advise
      }

      if (enterPressed) {
        val userName = FollowMeManager.getInstance(project).getUser(clientId).name
        val message = CodeWithMePluginBundle.message("terminal.access.command.notification", userName, widget.typedShellCommand)
        val accepted = AccessNotification.showStrictAccessNotification(project, lifetime, message, userName)
        accepted.whenComplete { accessAllowed, _ ->
          if (accessAllowed) {
            connector.write(byteArray)
            widget.handleEnterPressed()
          }
        }
      }
      else {
        connector.write(byteArray)
        widget.handleAnyKeyPressed()
      }
    }

    // Sync output
    sessionModel.outputConnected.whenTrue(connector.lifetime) {
      val currentOutput = connector.getCurrentOutput()
      if (currentOutput.any()) {
        sessionModel.output.fire(currentOutput)
      }

      connector.output.advise(connector.lifetime) {
        sessionModel.output.fire(it)
      }
    }
  }
}