package com.jetbrains.rdserver.terminal

import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.util.io.toByteArray
import com.jediterm.terminal.Questioner
import com.jediterm.terminal.TtyConnector
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.Signal
import com.jetbrains.rdserver.CodeWithMePluginBundle
import org.jetbrains.plugins.terminal.ProxyTtyConnector
import java.awt.Dimension

class BackendTtyConnector(project: Project, override val connector: TtyConnector) : ProxyTtyConnector {

  companion object {
    private const val CACHE_SIZE = 1024
  }

  private val lifetimeDef = project.defineNestedLifetime()
  private val outputCache = ArrayDeque<Byte>(CACHE_SIZE)

  val output = Signal<List<Byte>>()
  val lifetime = lifetimeDef.lifetime

  /**
   * Protocol ID to share this terminal across clients
   * Null if the terminal is private to host
   */
  var sessionId: String? = null
    private set

  fun setSessionId(sessionId: String) {
    assert(this.sessionId == null) { "Terminal is already shared" }
    this.sessionId = sessionId
  }

  /**
   * Returns cached output for this terminal process (if any)
   * Used to sync state for new clients
   */
  fun getCurrentOutput() = outputCache.toList()

  override fun init(q: Questioner) = connector.init(q)
  override fun close() {
    connector.close()
    application.invokeLater {
      if (lifetime.isNotAlive) return@invokeLater
      output.fire(("\n" + CodeWithMePluginBundle.message("terminal.message.terminal.closed")).encode())
      lifetimeDef.terminate()
    }
  }

  override fun resize(termSize: Dimension) = connector.resize(termSize)
  override fun getName() = connector.name
  override fun isConnected() = connector.isConnected
  override fun waitFor() = connector.waitFor()
  override fun write(bytes: ByteArray) = connector.write(bytes)
  override fun write(string: String) = connector.write(string)

  override fun read(buf: CharArray, offset: Int, length: Int): Int {
    val res = connector.read(buf, offset, length)
    if (res > 0) {
      // Charset should be the same as in ThinClientTtyConnector
      val bytes = Charsets.UTF_8.encode(buf.concatToString(offset, res)).toByteArray().toList()
      output.fire(bytes)

      for (byte in bytes) {
        if (outputCache.size == CACHE_SIZE - 1) outputCache.removeFirst()
        outputCache.addLast(byte)
      }
    }
    return res
  }

}