// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver

import com.intellij.openapi.util.TextRange
import com.intellij.ui.ColorUtil
import com.intellij.ui.SimpleColoredComponent
import com.intellij.ui.SimpleTextAttributes
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rdserver.ui.BackendIconHost
import java.awt.Color
import java.io.StringReader
import javax.swing.Icon
import javax.swing.text.MutableAttributeSet
import javax.swing.text.html.HTML
import javax.swing.text.html.HTMLEditorKit
import javax.swing.text.html.parser.ParserDelegator

// todo revisit and move to rd-platform

class RichTextBuilder {
  private val fragments = mutableListOf<RichStringModel>()

  fun append(text: RichStringModel) {
    fragments.add(text)
  }

  fun append(text: String, fgColor: Color? = null) {
    fragments.add(RichStringModel(text, fgColor.toRgbColor(), null, null, EffectStyle.None, FontStyle.Regular))
  }

  fun append(text: String, attributes: SimpleTextAttributes) {
    fragments.add(text.toRichStringModel(attributes))
  }

  fun build() = RichTextModel(fragments)
}

class CompletionItemTextBuilder {
  private val fragments = mutableListOf<CompletionPresentationPart>()

  fun append(text: CompletionPresentationPart) {
    fragments.add(text)
  }

  fun append(text: String, isGrayed: Boolean) {
    fragments.add(CompletionPresentationPart(text, isGrayed, EffectStyle.None, FontStyle.Regular))
  }

  fun build() = CompletionPresentationModel(fragments)
}

fun String.toRichStringModel(attributes: SimpleTextAttributes): RichStringModel {
  return RichStringModel(this,
                         attributes.fgColor.toRgbColor(),
                         attributes.bgColor.toRgbColor(),
                         attributes.waveColor?.toRgbColor(),
                         attributes.style.toEffectStyle(),
                         attributes.style.toFontStyle())
}

fun Int.toEffectStyle(): EffectStyle = when {
  this and SimpleTextAttributes.STYLE_UNDERLINE != 0 -> EffectStyle.StraightUnderline
  this and SimpleTextAttributes.STYLE_WAVED != 0 -> EffectStyle.WeavyUnderline
  this and SimpleTextAttributes.STYLE_STRIKEOUT != 0 -> EffectStyle.StrikeOut
  this and SimpleTextAttributes.STYLE_SEARCH_MATCH != 0 -> EffectStyle.SearchMatch_IDEA
  else -> EffectStyle.None
}

fun Int.toFontStyle(): FontStyle = when {
  this and SimpleTextAttributes.STYLE_BOLD != 0 -> FontStyle.Bold
  this and SimpleTextAttributes.STYLE_ITALIC != 0 -> FontStyle.Italic
  this and SimpleTextAttributes.STYLE_UNDERLINE != 0 -> FontStyle.Underline
  this and SimpleTextAttributes.STYLE_STRIKEOUT != 0 -> FontStyle.Strikeout
  else -> FontStyle.Regular
}

fun richText(body: RichTextBuilder.() -> Unit): RichTextModel {
  return RichTextBuilder().apply { body() }.build()
}

fun completionItemText(body: CompletionItemTextBuilder.() -> Unit): CompletionPresentationModel {
  return CompletionItemTextBuilder().apply { body() }.build()
}

fun String.toRichText() = richText { append(this@toRichText) }

fun String.htmlToRichText(): RichTextModel {
  if (!startsWith("<html>", ignoreCase = true)) {
    return toRichText()
  }

  val parts = mutableListOf<RichStringModel>()
  val parser = ParserDelegator()
  var foregroundColor: AbstractColor? = null
  var bold = false
  var styleTagDepth = 0

  parser.parse(StringReader(this), object : HTMLEditorKit.ParserCallback() {
    override fun handleStartTag(t: HTML.Tag, a: MutableAttributeSet, pos: Int) {
      // TODO support more tags
      when (t) {
        HTML.Tag.FONT -> {
          (a.getAttribute(HTML.Attribute.COLOR) as? String)?.let {
            foregroundColor = parseHtmlColor(it)
          }
        }
        HTML.Tag.B -> bold = true
        HTML.Tag.STYLE -> styleTagDepth++
      }
    }

    override fun handleEndTag(t: HTML.Tag, pos: Int) {
      when (t) {
        HTML.Tag.FONT -> foregroundColor = null
        HTML.Tag.B -> bold = false
        HTML.Tag.STYLE -> styleTagDepth--
      }
    }

    override fun handleText(data: CharArray, pos: Int) {
      if (styleTagDepth > 0) return
      parts.add(RichStringModel(String(data), foregroundColor, fontStyle = if (bold) FontStyle.Bold else FontStyle.Regular))
    }
  }, false)
  return RichTextModel(parts)
}

fun parseHtmlColor(color: String): RgbColor? {
  try {
    return ColorUtil.fromHex(color).toRgbColor()
  }
  catch (e: IllegalArgumentException) {
    return null
  }
}

fun Color?.toRgbColor(): RgbColor? {
  if (this == null) return null
  return RgbColor(red, green, blue)
}

fun Color?.toRgbaColor(): RgbaColor? {
  if (this == null) return null
  return RgbaColor(red, green, blue, alpha)
}

fun RgbaColor?.toColor(): Color? {
  if (this == null) return null
  return Color(red, green, blue, alpha)
}

fun Icon.toProtocolModel(): IconModel? = BackendIconHost.getInstance().createModel(this)

fun SimpleColoredComponent.toRichText(): RichTextModel {
  return richText {
    val iterator = iterator()
    while (iterator.hasNext()) {
      val string = iterator.next()
      append(string, iterator.textAttributes)
    }
  }
}

fun Pair<Int, Int>.toSelectionRange() = SelectionRange(
  DocOffsetAndVirtualModel(first, 0),
  DocOffsetAndVirtualModel(second, 0)
)

fun TextRange.toSelectionRange() = SelectionRange(
  DocOffsetAndVirtualModel(startOffset, 0),
  DocOffsetAndVirtualModel(endOffset, 0)
)

fun SelectionRange.withVersion(version: SelectionVersion) = SelectionRangeWithVersion(this, version)

fun TextRange.toRdTextRange() = RdTextRange(startOffset, endOffset)