package com.jetbrains.rdserver

import com.intellij.openapi.components.Service
import com.intellij.openapi.components.ServiceManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ProjectManager
import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isValid
import com.intellij.openapi.progress.ProcessCanceledException
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import org.picocontainer.PicoContainer
import java.util.*

@Service
class RdServerPerClientComponents(val project: Project) {
  private val clientIdToServiceLists = HashMap<ClientId, PicoContainer>()

  companion object {
    fun getInstance(project: Project): RdServerPerClientComponents {
      return ServiceManager.getService(project, RdServerPerClientComponents::class.java)
    }

    fun <T : Any> getService(project: Project, clazz: Class<T>): T {
      return getInstance(project).getService(clazz)
    }

    fun <T : Any> getService(project: Project, clientId: ClientId, clazz: Class<T>): T {
      if (!clientId.isValid) throw ProcessCanceledException()
      return ClientId.withClientId(clientId) {
        getInstance(project).getService(clazz)
      }
    }

    fun <T : Any> tryGetService(project: Project, clientId: ClientId, clazz: Class<T>): T? {
      return ClientId.withClientId(clientId) {
        getInstance(project).tryGetService(clazz)
      }
    }

    inline fun <reified TService : Any> getService(project: Project): TService {
      return getService(project, TService::class.java)
    }

    fun <T : Any> getServiceInRandomProject(clazz: Class<T>): T {
      ProjectManager.getInstance().openProjects.forEach {
        val service = getInstance(it).tryGetService(clazz)
        if (service != null)
          return service
      }

      error("No projects contain service $clazz for ClientId ${ClientId.current}, IsValid ${ClientId.current.isValid}")
    }
  }

  fun registerServices(lifetime: Lifetime, clientId: ClientId, container: PicoContainer) {
    clientIdToServiceLists.addUnique(lifetime, clientId, container)
  }

  private fun <T : Any> tryGetService(clazz: Class<T>): T? {
    val currentClientId = ClientId.currentOrNull ?: error("No ClientId on getting '${clazz.name} service'")
    if (ClientId.isLocalId(currentClientId)) error("Can't get RdServer component '${clazz.name}' for local ClientId")
    return clientIdToServiceLists[currentClientId]?.getComponentInstanceOfType(clazz) as T?
  }

  private fun <T : Any> getService(clazz: Class<T>): T {
    if (!ClientId.current.isValid) throw ProcessCanceledException()
    return tryGetService(clazz) ?: error("No service ${clazz.name} for client ${ClientId.current} in project $project")
  }
}