package com.jetbrains.cwm.unattended.common

import com.intellij.openapi.ui.DialogWrapper
import com.intellij.ui.ColoredTreeCellRenderer
import com.intellij.ui.components.JBScrollPane
import com.intellij.ui.treeStructure.Tree
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.intellij.util.ui.tree.AbstractTreeModel
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.reactive.adviseOnce
import com.jetbrains.codeWithMe.model.RemoteControlFsEntry
import com.jetbrains.codeWithMe.model.RemoteControlSession
import java.awt.BorderLayout
import java.util.*
import javax.swing.JComponent
import javax.swing.JPanel
import javax.swing.JTree
import javax.swing.tree.TreePath

class UnattendedHostBrowserDialog(val controlSession: RemoteControlSession) : DialogWrapper(true) {
  private val myLifetime = LifetimeDefinition()
  private lateinit var myTree: Tree

  init {
    isModal = false
    init()
  }

  override fun createCenterPanel(): JComponent {
    val panel = JPanel(BorderLayout())
    myTree = Tree(MyModel())

    panel.add(JBScrollPane(myTree), BorderLayout.CENTER)
    myTree.addTreeSelectionListener {
      okAction.isEnabled = it.paths != null
    }

    myTree.cellRenderer = MyRenderer()
    myTree.isRootVisible = false

    panel.minimumSize = JBUI.size(400, 400)

    return panel
  }

  override fun doOKAction() {
    super.doOKAction()
    val entry = myTree.selectionPath?.lastPathComponent as? RemoteControlFsEntry
    if (entry != null) {
      controlSession.openNewProject.start(entry.fullPath)
    }
  }

  override fun dispose() {
    super.dispose()
    myLifetime.terminate()
  }

  private class MyRenderer : ColoredTreeCellRenderer() {
    override fun customizeCellRenderer(tree: JTree, value: Any?, selected: Boolean, expanded: Boolean, leaf: Boolean, row: Int, hasFocus: Boolean) {
      background = UIUtil.getListBackground(selected, hasFocus)

      if (value !is RemoteControlFsEntry) return

      append(value.name)
    }
  }

  private inner class MyModel : AbstractTreeModel() {
    private val root = Any()
    private val fsRoots = ArrayList<RemoteControlFsEntry>()
    private val nodeChildren = HashMap<RemoteControlFsEntry, List<RemoteControlFsEntry>>()
    private val nodeParents = HashMap<RemoteControlFsEntry, RemoteControlFsEntry>()

    init {
      controlSession.getFsRoots.start(Unit).result.adviseOnce(Lifetime.Eternal) {
        fsRoots.addAll(it.unwrap())
        treeNodesInserted(TreePath(root), IntArray(fsRoots.size) { it }, fsRoots.toArray())
      }
    }

    override fun getRoot(): Any {
      return root
    }

    fun getNodeChildrenList(node: RemoteControlFsEntry): List<RemoteControlFsEntry>? {
      if (!node.isDirectory)
        return null

      if (!nodeChildren.contains(node)) {
        val list = ArrayList<RemoteControlFsEntry>()
        nodeChildren[node] = list
        controlSession.listFsEntries.start(node.fullPath).result.adviseOnce(Lifetime.Eternal) { taskResult ->
          list.addAll(taskResult.unwrap())
          list.sortBy { it.name }
          list.forEach {
            nodeParents[it] = node
          }
          treeNodesInserted(getPath(node), IntArray(list.size) {it}, list.toArray())
        }
      }

      return nodeChildren[node]
    }

    private fun getPath(node: RemoteControlFsEntry): TreePath {
      val parent = nodeParents[node]
      if (parent == null) return TreePath(arrayOf(root, node))
      val parentPath = getPath(parent)
      return parentPath.pathByAddingChild(node)
    }

    override fun getChild(parent: Any?, index: Int): Any? {
      if (parent == root)
        return fsRoots[index]

      val node = parent as RemoteControlFsEntry
      return getNodeChildrenList(node)?.get(index)
    }

    override fun getChildCount(parent: Any?): Int {
      if (parent == root)
        return fsRoots.size

      val node = parent as RemoteControlFsEntry
      return getNodeChildrenList(node)?.size ?: 0
    }

    override fun isLeaf(node: Any?): Boolean {
      return (node as? RemoteControlFsEntry)?.isDirectory == false
    }

    override fun getIndexOfChild(parent: Any?, child: Any?): Int {
      if (parent == root)
        return fsRoots.indexOf(child)

      val node = parent as RemoteControlFsEntry
      return getNodeChildrenList(node)?.indexOf(child) ?: -1
    }
  }
}