package com.jetbrains.rdserver.actions

import com.intellij.openapi.actionSystem.DataContext
import com.jetbrains.rd.ide.model.ActionTimestampModel
import com.jetbrains.rd.ide.model.UpgradableActionTimestampModel

/**
 * Type safe key for getting timestamps both platform-aware [ActionTimestamp] and protocol [ActionTimestampModel]
 */
class ActionTimestampKey<T : ActionTimestamp<P>, P : ActionTimestampModel>(val name: String) {
  companion object {
    fun <T : ActionTimestamp<P>, P : ActionTimestampModel> create(name: String) = ActionTimestampKey<T, P>(name)
  }
}

/**
 * Base class for action timestamps which are like [com.intellij.openapi.actionSystem.DataKey]
 * but protocol-friendly and have built-in support for validation and recovery
 */
abstract class ActionTimestamp<T : ActionTimestampModel>(val key: ActionTimestampKey<*, T>, val model: T) {
  /**
   * Actual data associated with a timestamp which can be consumed by the platform (e.g. project, editor, caret)
   * similar to [DataContext.getData]
   */
  open fun getData(): Any? = null
}

/**
 * Timestamp which is not supposed to be transferred through the protocol
 */
abstract class TransparentTimestamp<T : ActionTimestampModel>(name: ActionTimestampKey<*, T>, model: T) : ActionTimestamp<T>(name, model)

/**
 * Timestamp which provides a way to be upgraded to a new value according to changes made in the IDE
 * rather than just returning null from [getData] when it's not valid
 */
abstract class UpgradableTimestamp<T : UpgradableActionTimestampModel>(name: ActionTimestampKey<*, T>, model: T) : ActionTimestamp<T>(name, model) {
  abstract fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult
}

sealed class TimestampUpgradeResult {
  class Failure(val message: String) : TimestampUpgradeResult()

  class RetryRequest(val message: String) : TimestampUpgradeResult()

  class Success<T : ActionTimestampModel>(val timestamp: ActionTimestamp<T>) : TimestampUpgradeResult()
}

interface TimestampUpgradeResultHandler {
  fun handleFailure(actionId: String, failure: TimestampUpgradeResult.Failure)
  fun handleRetryRequest(actionId: String, request: TimestampUpgradeResult.RetryRequest)
}
