package com.jetbrains.rdserver.actions

import com.intellij.openapi.diagnostic.logger
import com.jetbrains.rd.ide.model.ActionTimestampModel
import com.jetbrains.rd.ide.model.ActionTimestampSetModel

class ActionTimestampSet private constructor(private val timestamps: Map<String, ActionTimestamp<*>>) {
  constructor(vararg timestamp: ActionTimestamp<*>) : this(timestamp.map { it.key.name to it }.toMap())

  companion object {
    val EMPTY = ActionTimestampSet()
    private val LOG = logger<ActionTimestampSet>()
  }

  fun <T : ActionTimestamp<*>> getTimestamp(key: ActionTimestampKey<T, *>): T? {
    @Suppress("UNCHECKED_CAST")
    return timestamps[key.name] as T?
  }

  /**
   * [com.intellij.openapi.actionSystem.DataContext] interop
   */
  fun getTimestamp(dataId: String): ActionTimestamp<*>? {
    return timestamps[dataId]
  }

  /**
   * [com.intellij.openapi.actionSystem.DataContext] interop
   */
  fun getAllTimestamps(): Map<String, ActionTimestamp<*>> {
    return timestamps
  }

  /**
   * Used to combine sets from different ActionTimestampProviders in a single update/execute session
   */
  fun merge(other: ActionTimestampSet): ActionTimestampSet {
    if (this == EMPTY) return other
    if (other == EMPTY) return this

    val result = HashMap(other.timestamps)
    for ((name, timestamp) in this.timestamps) {
      val otherTimestamp = other.timestamps[name]
      if (otherTimestamp != null) {
        if (timestamp.model != otherTimestamp.model) {
          LOG.error("Timestamp data mismatched. Name:\"${timestamp.key}\", Data: \"${timestamp.model}\" vs \"${otherTimestamp.model}")
        }
      }
      result[name] = timestamp
    }

    return ActionTimestampSet(result)
  }

  fun createModel(): ActionTimestampSetModel {
    val models = timestamps.values.filter { it !is TransparentTimestamp<*> }.map { it.model }
    return ActionTimestampSetModel(models)
  }
}

inline fun <reified P : ActionTimestampModel> ActionTimestampSetModel.getTimestamp(key: ActionTimestampKey<*, P>): P? {
  @Suppress("UNCHECKED_CAST")
  return models.find { it is P } as? P
}
