package com.jetbrains.rdserver.actions

import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.actionSystem.impl.SimpleDataContext
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.fileEditor.impl.text.TextEditorPsiDataProvider
import com.intellij.openapi.project.Project
import com.jetbrains.rdserver.actions.timestamps.EditorTimestamp
import com.jetbrains.rd.ide.model.ActionTimestampSetModel
import com.jetbrains.rd.ide.model.InputEventModel
import com.jetbrains.rd.ide.model.MouseEventModel
import java.awt.event.InputEvent
import java.awt.event.MouseEvent

private val logger = Logger.getInstance("#com.jetbrains.rdserver.actions.ActionUtilsKt")

fun createInputEvent(inputEventModel: InputEventModel?, dataContext: DataContext): InputEvent? {
  if (inputEventModel !is MouseEventModel) {
    return null
  }
  val editor = dataContext.getData(CommonDataKeys.EDITOR) ?: return null
  return MouseEvent(editor.component, inputEventModel.id, System.currentTimeMillis(), inputEventModel.modifiers, 0, 0,
                    inputEventModel.clickCount, false, inputEventModel.button)
}

fun createDataContext(actionId: String, protocolTimestampSet: ActionTimestampSetModel,
                      handler: TimestampUpgradeResultHandler, project: Project): DataContext {
  val requestTimestampSet = createSetFromModel(actionId, protocolTimestampSet, project)

  val editor = requestTimestampSet.getTimestamp(EditorTimestamp.NAME)?.getData()
  val editorDataContext = editor?.let {
    val textEditorDataProvider = TextEditorPsiDataProvider()
    DataContext { dataId -> textEditorDataProvider.getData(dataId, editor, editor.caretModel.currentCaret) }
  }

  val requestDataMap = requestTimestampSet.getAllTimestamps()
    .map { it.key to it.value.getData() }
    .filter { it.second != null }
    .toMap()
  val contextForProviders = SimpleDataContext.getSimpleContext(requestDataMap, editorDataContext)

  // DataContext is needed for constructing the actual set
  val actualTimestampSet = createSetFromDataContext(actionId, contextForProviders, project)

  validateTimestampSet(actionId, requestTimestampSet, actualTimestampSet, handler,
                                                                         project)

  return BackendActionDataContext(actionId, requestTimestampSet, actualTimestampSet, editorDataContext, handler)
}

private fun validateTimestampSet(actionId: String, requestTimestampSet: ActionTimestampSet,
                                 actualTimestampSet: ActionTimestampSet, handler: TimestampUpgradeResultHandler, project: Project) {
  for (validator in BackendActionTimestampValidator.EP.extensions(project)) {
    val result = validator.validate(actionId, requestTimestampSet, actualTimestampSet) ?: continue
    when (result) {
      is TimestampUpgradeResult.Failure -> {
        logger.trace { "Timestamp upgrade failed: ${result.message}" }
        handler.handleFailure(actionId, result)
        throw TimestampUpgradeFailedException(result.message)
      }
      is TimestampUpgradeResult.RetryRequest -> {
        logger.trace { "Timestamp upgrade required a retry: ${result.message}" }
        handler.handleRetryRequest(actionId, result)
        throw TimestampUpgradeFailedException(result.message)
      }
      is TimestampUpgradeResult.Success<*> -> Unit
    }
  }
}

private fun createSetFromModel(actionId: String, protocolTimestampSet: ActionTimestampSetModel, project: Project): ActionTimestampSet {
  var resultSet = ActionTimestampSet.EMPTY
  for (provider in BackendActionTimestampProvider.EP.extensions(project)) {
    val set = provider.createSet(actionId, protocolTimestampSet)
    resultSet = resultSet.merge(set)
  }
  return resultSet
}

private fun createSetFromDataContext(actionId: String, dataContext: DataContext, project: Project): ActionTimestampSet {
  var resultSet = ActionTimestampSet.EMPTY
  for (provider in BackendActionTimestampProvider.EP.extensions(project)) {
    val set = provider.createSet(actionId, dataContext)
    resultSet = resultSet.merge(set)
  }
  return resultSet
}