package com.jetbrains.rdserver.actions

import com.intellij.ide.DataManager
import com.intellij.ide.impl.DataManagerImpl
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.actionSystem.DataProvider
import com.intellij.openapi.actionSystem.PlatformDataKeys
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.runAndLogException
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.wm.IdeFocusManager

class BackendActionDataContext(
  private val actionId: String,
  private val requestTimestampSet: ActionTimestampSet,
  private val actualTimestampSet: ActionTimestampSet,
  private val parentContext: DataContext?,
  private val handler: TimestampUpgradeResultHandler
) : DataContext {
  companion object {
    private val LOG = logger<BackendActionDataContext>()
  }

  private val myCache = hashMapOf<String, Any?>()
  private var failed = false
  private val dataProvider = DataProvider { dataId -> getDataFromSelfOrParent(dataId) }

  // todo unify with SimpleDataContext
  override fun getData(dataId: String): Any? {
    getDataFromSelfOrParent(dataId)?.let { return it }

    for (provider in BackendActionDataProvider.EP_NAME.extensions) {
      val data = LOG.runAndLogException { provider.getData(dataId, dataProvider) }
      if (data != null) return data
    }

    val rule = (DataManager.getInstance() as DataManagerImpl).getDataRule(dataId)
    if (rule != null) {
      return rule.getData(dataProvider)
    }

    if (PlatformDataKeys.CONTEXT_COMPONENT.`is`(dataId)) {
      return IdeFocusManager.getGlobalInstance().focusOwner
    }

    return null
  }

  private fun getDataFromSelfOrParent(dataId: String): Any? {
    if (failed) {
      LOG.warn("Requesting $dataId for already failed DataContext")
      throw TimestampUpgradeFailedException("Requesting $dataId for already failed DataContext")
    }

    if (myCache.containsKey(dataId)) return myCache[dataId]

    val requestTimestamp = requestTimestampSet.getTimestamp(dataId)
    if (requestTimestamp == null) {
      return parentContext?.getData(dataId)
    }

    if (requestTimestamp is UpgradableTimestamp<*>) {
      when (val result = requestTimestamp.upgrade(actualTimestampSet)) {
        is TimestampUpgradeResult.Failure -> {
          failed = true
          handler.handleFailure(actionId, result)
          LOG.trace("Timestamp upgrade for $dataId failed")
          throw TimestampUpgradeFailedException(result.message)
        }
        is TimestampUpgradeResult.RetryRequest -> {
          failed = true
          handler.handleRetryRequest(actionId, result)
          LOG.trace("Timestamp upgrade for $dataId required a retry")
          throw TimestampUpgradeFailedException(result.message)
        }
        is TimestampUpgradeResult.Success<*> -> {
          val upgradedTimestamp = result.timestamp
          val data = upgradedTimestamp.getData()

          myCache[dataId] = data
          return data
        }
      }
    }
    else {
      val data = requestTimestamp.getData()
      myCache[dataId] = data
      return data
    }
  }
}