package com.jetbrains.rdserver.actions

import com.intellij.codeWithMe.ClientId
import com.intellij.core.CoreBundle
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.actionSystem.ex.ActionManagerEx
import com.intellij.openapi.actionSystem.ex.ActionUtil
import com.intellij.openapi.application.impl.LaterInvocator
import com.intellij.openapi.command.executeCommand
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.editor.ReadOnlyModificationException
import com.intellij.openapi.project.DumbService
import com.intellij.openapi.project.Project
import com.jetbrains.ide.model.uiautomation.BeBackendIdeAction
import com.jetbrains.rd.actions.ActionProvider
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.ui.bedsl.actionButton
import com.jetbrains.rd.ui.bindable.ActionFilterProvider
import com.jetbrains.rd.ui.bindable.ActionFilterResult
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.daemon.BackendMarkupSuppressors
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.editors.BackendEditorHost.Companion.synchronizer
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.permissions.ToolWindowAccessLevel
import com.jetbrains.rdserver.requests.BackendAsyncRequestDelegate
import com.jetbrains.rdserver.requests.BackendIdeGateway
import com.jetbrains.rdserver.toolWindow.BackendToolWindowModeProvider
import com.jetbrains.rdserver.ui.BackendNotificationsHost
import com.jetbrains.rdserver.ui.feelLuckUi.BackendFeelLuckyUiHelper

private val logger = logger<BackendActionExecutorDelegate>()

class BackendActionExecutorDelegate : BackendAsyncRequestDelegate<RdCallActionRequest> {
  companion object {
    private fun createActionForbiddenMessage(actionText: String?): String {
      return if (actionText != null) {
        CodeWithMePlatformBundle.message("notification.action.forbidden", actionText)
      } else {
        CodeWithMePlatformBundle.message("notification.action.forbidden.generic")
      }
    }

    private fun createReadOnlyModificationMessage(actionText: String?): String {
      return if (actionText != null) {
        CodeWithMePlatformBundle.message("notification.action.read.only.modification", actionText)
      } else {
        CoreBundle.message("attempt.to.modify.read.only.document.error.message")
      }
    }

    private fun ActionFilterProvider.isActionForbidden(action: AnAction, actionId: String, project: Project): Boolean {
      return filterAction(action, actionId, project, ClientId.current) == ActionFilterResult.FORBIDDEN
    }
  }

  override suspend fun execute(request: RdCallActionRequest, project: Project, consumer: (RdResponse) -> Unit) {
    val actionId = request.actionId
    logger.trace { "Received execute request for action $actionId" }
    val requestId = request.id

    var action: AnAction? = null
    val handler = ActionExecutionResultHandler(requestId)
    try {
      val dataContext = createDataContext(actionId, request.timestampSetModel, handler, project)
      val inputEvent = createInputEvent(request.inputEventModel, dataContext)
      val place = request.actionPlace
      val presentation = Presentation()
      val actionEvent = AnActionEvent.createFromInputEvent(inputEvent, place, presentation, dataContext)
      action = ActionProvider.EP_NAME.extensions.mapNotNull { it.getAction(actionId, actionEvent) }.firstOrNull() ?: run {
        logger.error("`$actionId` action not found")
        return consumer(RdFailureResponse("`$actionId` action not found", requestId))
      }
      if (ActionFilterProvider.EP_NAME.findFirstSafe { it.isActionForbidden(action, actionId, project) } != null) {
        notifyClient(createActionForbiddenMessage(action.templateText), project)
        return consumer(RdSuccessResponse(requestId))
      }
      val actionManager = ActionManagerEx.getInstanceEx()
      return tryInvokeAction(action, actionId, actionEvent, actionManager, project, requestId, consumer)
    }
    catch (e: ReadOnlyModificationException) {
      logger.info(e)
      notifyClient(createReadOnlyModificationMessage(action?.templateText), project)
      return consumer(RdSuccessResponse(requestId))
    }
    catch (_: TimestampUpgradeFailedException) {
      return consumer(handler.response)
    }
  }

  private fun notifyClient(message: String, project: Project) {
    val groupId = PermissionManager.NOTIFICATION_GROUP_ID
    val title = CodeWithMePlatformBundle.message("product.name")
    val notification = Notification(groupId, title, message, NotificationType.INFORMATION)
    BackendNotificationsHost.getInstance(project).notifyClient(notification)
  }

  private suspend fun tryInvokeAction(action: AnAction,
                                      actionId: String,
                                      actionEvent: AnActionEvent,
                                      actionManager: ActionManagerEx,
                                      project: Project,
                                      requestId: Int,
                                      consumer: (RdResponse) -> Unit) {
    ActionUtil.performDumbAwareUpdate(LaterInvocator.isInModalContext(), action, actionEvent, true)
    if (!actionEvent.presentation.isEnabled) {
      return consumer(RdSuccessResponse(requestId)) // todo return not available status
    }

    if (DumbService.getInstance(project).isDumb && !action.isDumbAware) {
      if (false == actionEvent.presentation.getClientProperty(ActionUtil.WOULD_BE_ENABLED_IF_NOT_DUMB_MODE)) {
        return consumer(RdSuccessResponse(requestId))
      }
      val unavailableMessage = ActionUtil.getUnavailableMessage(action.templatePresentation.text, false)
      TODO("return dumb tooltip request")
    }

    if (PermissionManager.getInstance(project).getToolWindowAccess(ClientId.current) == ToolWindowAccessLevel.Readonly) {
      val toolWindow = actionEvent.dataContext.getData(PlatformDataKeys.TOOL_WINDOW)
      if (toolWindow != null && !BackendToolWindowModeProvider.isSafeToolWindow(toolWindow.id)) {
        logger.trace { "Cancelled execute request for action $actionId in toolwindow ${toolWindow.id} by ${ClientId.current}" }
        return
      }
    }

    BackendMarkupSuppressors.fullSuppressMarkupSynchronization(project, "Executing $actionId action request#$requestId") {
      val ideGateway = RdServerPerClientComponents.getService<BackendIdeGateway>(project)
      val patch = ideGateway.executeBulkUpdate {
        if (action is DefaultActionGroup) {
          sendActionGroupStep(action, actionEvent, ideGateway, project)
          return@executeBulkUpdate
        }
        val feelLuckyUiHelper = RdServerPerClientComponents.getService<BackendFeelLuckyUiHelper>(project)
        feelLuckyUiHelper.executeWithUiInteractionMarker(AnActionFeelLuckyUiMarker(actionId, action)) {
          actionManager.fireBeforeActionPerformed(action, actionEvent.dataContext, actionEvent)
          executeCommand(project) {
            ActionUtil.performActionDumbAware(action, actionEvent)
          }
          actionManager.fireAfterActionPerformed(action, actionEvent.dataContext, actionEvent)
        }
      } ?: return@fullSuppressMarkupSynchronization consumer(RdSuccessResponse(requestId))

      consumer(RdPatchResponse(patch, requestId))
    }
  }

  private fun sendActionGroupStep(actionGroup: ActionGroup, actionEvent: AnActionEvent, ideGateway: BackendIdeGateway, project: Project) {
    val lifetime = BackendActionGroupLifetimeHost.getInstance(project).nextGroupLifetime
    val beAction = actionButton(actionGroup, lifetime) as? BeBackendIdeAction ?: return
    val actionModel = beAction.presentation ?: return
    val editor = actionEvent.dataContext.getData(CommonDataKeys.EDITOR) ?: return
    val editorId = editor.getTextControlId() ?: return
    val step = RdActionGroupStep(actionModel, editorId)
    val editorVersion = editor.synchronizer?.version ?: return
    val documentVersion = BackendDocumentHost.getInstance(project).tryGetSynchronizer(editorId.documentId)?.version ?: return
    ideGateway.sendEditorChange(step, editorVersion, editorVersion, documentVersion)
  }
}