package com.jetbrains.rdserver.actions

import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.runAndLogException
import com.intellij.openapi.editor.actionSystem.EditorAction
import com.intellij.openapi.project.Project
import com.jetbrains.rd.platform.util.idea.LifetimedProjectService
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rdserver.RdServerPerClientComponents
import org.jetbrains.annotations.NonNls

@Suppress("NonDefaultConstructor")
class BackendActionsRegistrationHost(project: Project, actionHost: BackendAsyncActionHost) : LifetimedProjectService(project) {
  companion object {
    private val logger = logger<BackendActionsRegistrationHost>()
    fun getInstance(project: Project): BackendActionsRegistrationHost {
      return RdServerPerClientComponents.getService(project)
    }
  }

  private val actionsModel = actionHost.actionsModel

  private val actionManager get() = ActionManager.getInstance()

  private val id2GroupIds by lazy { BackendDelegatingActionIdsProvider.getHintParentGroupIds(project) }
  private val id2ActionModels = mutableMapOf<String, ActionRegistrationModel>()
  private val editorActionIdWithImmediateResults by lazy { BackendDelegatingActionIdsProvider.getEditorActionWithImmediateResultIds(project).toHashSet() }

  init {
    sendActionRegistry()
    // todo listen dynamic plugins
  }

  private fun sendActionRegistry() {
    val models = mutableListOf<ActionRegistrationModel>()

    val backendDelegatingActionIds = BackendDelegatingActionIdsProvider.getActionIds(project) +
                                     BackendDelegatingActionIdsProvider.getEditorActionWithImmediateResultIds(project)

    for (actionId in backendDelegatingActionIds) {
      val action = actionManager.getAction(actionId)
      if (action == null) {
        logger.warn("Not found action with id: $actionId")
        continue
      }
      val model = logger.runAndLogException { createModel(action) } ?: continue
      models.add(model)
    }

    actionsModel.registerActions.fire(models)

    id2ActionModels.clear()
    id2GroupIds.clear()
  }

  fun createModel(action: AnAction): ActionRegistrationModel? {
    return if (action is Separator) {
      SeparatorModel(action.text)
    }
    else {
      val id = actionManager.getId(action)
      if (id == null) {
        logger.warn("$action with NULL id") // possibly we try to transfer custom action group with dynamic children
        return null
      }
      val model = id2ActionModels.getOrPut(id) {
        if (action is ActionGroup)
          createGroupModel(action)
        else
          createActionModel(action)
      }
      model
    }
  }

  private fun createGroupModel(action: ActionGroup): GroupModel {
    val actionId = actionManager.getId(action)
    val children = action.getChildren(null)
    val childModels = children.mapNotNull {
      createModel(it)
    }
    val addToGroupRules = createAddToGroupRules(actionId)
    return with(action.templatePresentation) {
      GroupModel(actionId, text, description, icon?.toProtocolModel(), action.isPopup, childModels, addToGroupRules)
    }
  }

  private fun createActionModel(action: AnAction): ActionModel {
    val actionId = actionManager.getId(action)
    val addToGroupRules = createAddToGroupRules(actionId)
    return createActionModel(action, actionId, addToGroupRules)
  }

  fun createActionModel(action: AnAction,
                        actionId: @NonNls String,
                        addToGroupRules: List<AddToGroupRuleModel>): ActionModel {
    return when (action) {
      is EditorAction -> {
        with(action.templatePresentation) {
          EditorActionModel(editorActionIdWithImmediateResults.contains(actionId), actionId, text, description, icon?.toProtocolModel(),
                            addToGroupRules)
        }
      }
      is ToggleAction -> {
        with(action.templatePresentation) {
          ToggleActionModel(Toggleable.isSelected(this), actionId, text, description, icon?.toProtocolModel(), addToGroupRules)
        }
      }
      else -> {
        if (editorActionIdWithImmediateResults.contains(actionId)) {
          logger.error("Register `$actionId` as action without immediate result")
        }
        with(action.templatePresentation) {
          ActionModel(actionId, text, description, icon?.toProtocolModel(), addToGroupRules)
        }
      }
    }
  }

  private fun createAddToGroupRules(actionId: String): List<AddToGroupRuleModel> {
    // todo use ActionManager here to get info
    val rules = id2GroupIds[actionId] ?: return emptyList()
    if (rules.isEmpty()) return emptyList()

    return rules
  }
}

fun Constraints.toModel(): ActionConstraintModel {
  return ActionConstraintModel(this.myRelativeToActionId, this.myAnchor.toModel())
}

fun Anchor.toModel(): AnchorModel {
  return when (this) {
    Anchor.FIRST -> AnchorModel.First
    Anchor.LAST -> AnchorModel.Last
    Anchor.BEFORE -> AnchorModel.Before
    Anchor.AFTER -> AnchorModel.After
    else -> throw IllegalArgumentException("Unknown anchor: $this")
  }
}