package com.jetbrains.rdserver.actions

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.runAndLogException
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ProjectManager
import com.intellij.util.containers.CollectionFactory
import com.jetbrains.ide.model.uiautomation.BeBackendIdeAction
import com.jetbrains.ide.model.uiautomation.BeIdeAction
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ui.bedsl.dsl.beInit
import com.jetbrains.rd.ui.bedsl.extensions.extend
import com.jetbrains.rd.ui.bedsl.extensions.getExtended
import com.jetbrains.rd.ui.bindable.ActionFilterProvider
import com.jetbrains.rd.ui.bindable.ActionFilterResult
import com.jetbrains.rd.ui.bindable.BeActionManager
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.toProtocolModel
import org.jetbrains.annotations.NonNls
import org.jetbrains.annotations.NotNull
import java.util.*

class BeHostActionManager : BeActionManager {
  private val actions = CollectionFactory.createSmallMemoryFootprintSet<BeIdeAction>()
  companion object {
    private val logger = logger<BeHostActionManager>()
  }

  //if exists?
  override fun getBeAction(action: AnAction, lifetime: Lifetime): BeBackendIdeAction? {
    val (save, id) = getId(action)
    val project = ProjectManager.getInstance().openProjects.firstOrNull() ?: return null

    var result = ActionFilterResult.OK
    ActionFilterProvider.EP_NAME.forEachExtensionSafe {
      //TODO: May be action place should be added, or some additional context
      when (val filterAction = logger.runAndLogException {it.filterAction(action, id, project, ClientId.current)}) {
        ActionFilterResult.REMOVE -> {
          result = filterAction
          logger.trace("Action (id:${id}) is filtered in ${it.javaClass.simpleName} and will be removed")
          return@forEachExtensionSafe
        }
        ActionFilterResult.FORBIDDEN -> {
          logger.trace("Action (id:${id}) is forbidden in ${it.javaClass.simpleName} and has to be disabled or removed")
          //todo: removed until presentation is desided
          result = ActionFilterResult.REMOVE
          return@forEachExtensionSafe
        }
        ActionFilterResult.DISABLE -> {
          logger.trace("Action (id:${id}) is filtered in ${it.javaClass.simpleName} and has to be disabled or removed")
          result = filterAction
        }

        else -> { }
      }
    }
    return when (result) {
      ActionFilterResult.REMOVE -> null
      ActionFilterResult.DISABLE -> BeBackendIdeAction(createDisabledActionModel(action, id), id).apply { this.beInit(enabled = false) }
      ActionFilterResult.FORBIDDEN -> BeBackendIdeAction(createForbiddenActionModel(action, id), id).apply { this.beInit(enabled = false) }
      ActionFilterResult.OK -> when (action) {
        is ActionGroup -> createBeGroupAction(action, id, lifetime, project)
        is Separator -> BeBackendIdeAction(SeparatorModel(action.text), id)
        else -> createActionButton(action, id, project)
      }.apply {
        if (save) {
          this.extend(action, id)
          actions.add(this)
          lifetime.onTermination {
            actions.remove(this)
          }
        }
        this.beInit()
      }

    }
  }

  private fun createActionButton(action: AnAction,
                                 id: @NonNls String,
                                 project: @NotNull Project?): BeBackendIdeAction {
    val actionHost = if (project != null && !ClientId.isCurrentlyUnderLocalId) BackendActionsRegistrationHost.getInstance(project) else null
    val actionRegistrationModel = if (actionHost?.createModel(action) != null)
      actionHost.createModel(action)
    else
      when (action) {
        is ToggleAction -> createActionModel(action, id, emptyList(), project)
        else -> actionHost?.createActionModel(action, id, emptyList())
      } ?: createActionModel(action, id, emptyList(), project)
    return BeBackendIdeAction(actionRegistrationModel, id)
  }

  private fun createActionModel(action: AnAction,
                                actionId: @NonNls String,
                                addToGroupRules: List<AddToGroupRuleModel>,
                                project: @NotNull Project?): ActionRegistrationModel {
    return when (action) {
      is ToggleAction -> with(action.templatePresentation) {
        ToggleActionModel(action.isSelected(createEvent(action, project)), actionId, text, description, icon?.toProtocolModel(),
                          addToGroupRules)
      }
      else -> with(action.templatePresentation) { ActionModel(actionId, text, description, icon?.toProtocolModel(), addToGroupRules) }
    }
  }

  private fun createDisabledActionModel(action: AnAction, actionId: @NonNls String): ActionRegistrationModel {
    //todo: add text with explanation
    return with(action.templatePresentation) { ActionModel(actionId, text, description, icon?.toProtocolModel(), emptyList()) }
  }

  private fun createForbiddenActionModel(action: AnAction, actionId: @NonNls String): ActionRegistrationModel {
    //todo: add text with explanation
    return with(action.templatePresentation) { ActionModel(actionId, text, description, icon?.toProtocolModel(), emptyList()) }
  }
  private fun getId(action: AnAction): Pair<Boolean, @NonNls String> {
    var save = false
    var id = ActionManager.getInstance().getId(action)
    if (id.isNullOrEmpty()) {
      id = "${UUID.randomUUID()}"
      save = true
    }
    return Pair(save, id)
  }

  private fun createBeGroupAction(action: ActionGroup,
                                  actionId: String,
                                  lifetime: Lifetime,
                                  project: @NotNull Project?): BeBackendIdeAction {
    val children = action.getChildren(createEvent(action, project))
    val childModels = children.mapNotNull {
      getBeAction(it, lifetime)
    }
    //TODO : val addToGroupRules = createAddToGroupRules(actionId)
    val groupModelPresentation = with(action.templatePresentation) {
      GroupModel(actionId, text, description, icon?.toProtocolModel(), action.isPopup, childModels.mapNotNull { it.presentation },
                 emptyList())
    }
    return BeBackendIdeAction(groupModelPresentation, actionId)
  }

  private fun createEvent(
    action: AnAction,
    project: Project?): @NotNull AnActionEvent {
    return AnActionEvent.createFromDataContext(ActionPlaces.UNKNOWN,
                                               Presentation().apply { copyFrom(action.templatePresentation) },
                                               createDataContext(project))
  }

  private fun createDataContext(project: Project?): DataContext {
    return DataContext {
      when (it) {
        CommonDataKeys.PROJECT.name -> project
        /*CommonDataKeys.EDITOR.name -> editor
        CommonDataKeys.PSI_FILE.name -> file
        CommonDataKeys.VIRTUAL_FILE.name -> file.virtualFile*/
        else -> null //component.toolbarDataContext.getData(it)
      }
    }
  }

  override fun getAction(backendActionId: String): AnAction? {
    return actions.firstOrNull { it.actionId == backendActionId }?.getExtended(backendActionId)
  }
}