package com.jetbrains.rdserver.actions.timestamps

import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.project.Project
import com.intellij.xdebugger.XDebuggerManager
import com.intellij.xdebugger.impl.actions.XDebuggerActions
import com.jetbrains.rd.ide.model.ActionTimestampSetModel
import com.jetbrains.rd.ide.model.DebuggerExecutionLineTimestampModel
import com.jetbrains.rdserver.actions.*
import com.jetbrains.rdserver.document.BackendDocumentHost


class DebuggerExecutionLineTimestamp(private val offset: OffsetTimestamp) :
  UpgradableTimestamp<DebuggerExecutionLineTimestampModel>(NAME, DebuggerExecutionLineTimestampModel(offset.model)) {
  companion object {
    val NAME = ActionTimestampKey<DebuggerExecutionLineTimestamp, DebuggerExecutionLineTimestampModel>("EXECUTION_LINE")
  }

  override fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult {
    val project = actualSet.getTimestamp(ProjectTimestamp.PROJECT)?.getData() ?: return TimestampUpgradeResult.Failure("No Project")
    val session = XDebuggerManager.getInstance(project).currentSession ?: return TimestampUpgradeResult.Failure("Debug session is terminated")
    val currentPosition = session.currentPosition ?: return TimestampUpgradeResult.Failure("Execution position is not available")

    val offsetUpgradeResult = offset.upgrade(actualSet)
    if (offsetUpgradeResult !is TimestampUpgradeResult.Success<*>) return offsetUpgradeResult
    val upgradedOffset = offsetUpgradeResult.timestamp as OffsetTimestamp
    val offset = upgradedOffset.model.offset

    if (offset != currentPosition.offset) return TimestampUpgradeResult.RetryRequest("Execution position was changed")

    return TimestampUpgradeResult.Success(DebuggerExecutionLineTimestamp(upgradedOffset))
  }
}

class BackendExecutionLineTimestampProvider(project: Project) : BackendActionTimestampProvider(project) {
  override fun isApplicable(actionId: String): Boolean {
    return ActionManager.getInstance().getAction(actionId) is XDebuggerActions // todo more fine grained control
  }

  override fun createSet(actionId: String, dataContext: DataContext): ActionTimestampSet {
    val emptySet = ActionTimestampSet.EMPTY
    val project = CommonDataKeys.PROJECT.getData(dataContext) ?: return emptySet

    val session = XDebuggerManager.getInstance(project).currentSession ?: return emptySet
    val currentPosition = session.currentPosition ?: return emptySet

    val document = FileDocumentManager.getInstance().getDocument(currentPosition.file) ?: return emptySet
    val documentSynchronizer = BackendDocumentHost.getInstance(project).tryGetSynchronizer(document) ?: return emptySet

    val offset = OffsetTimestamp.create(documentSynchronizer, currentPosition.offset)
    return ActionTimestampSet(DebuggerExecutionLineTimestamp(offset))
  }

  override fun createSet(actionId: String, setModel: ActionTimestampSetModel): ActionTimestampSet {
    val model = setModel.getTimestamp(DebuggerExecutionLineTimestamp.NAME)?.offset ?: return ActionTimestampSet.EMPTY
    val offsetTimestamp = OffsetTimestamp(model.offset, model.isStickingToRight, model.document)
    return ActionTimestampSet(DebuggerExecutionLineTimestamp(offsetTimestamp))
  }
}

