package com.jetbrains.rdserver.actions.timestamps

import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.jetbrains.rd.platform.Range
import com.jetbrains.rdserver.actions.*
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rd.ide.model.ActionTimestampSetModel
import com.jetbrains.rd.ide.model.EditorTimestampModel
import com.jetbrains.rd.ide.model.TextControlId

class EditorTimestamp(
  textControlId: TextControlId,
  private val project: Project,
  private val caretTimestamps: List<CaretTimestamp>
) : UpgradableTimestamp<EditorTimestampModel>(NAME, EditorTimestampModel(textControlId, caretTimestamps.map { it.model })) {
  companion object {
    val NAME = ActionTimestampKey<EditorTimestamp, EditorTimestampModel>(CommonDataKeys.EDITOR.name)
  }

  override fun getData(): Editor? {
    val editorId = model.editorId
    val editor = BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(editorId)
    if (editor != null) return editor
    return null
  }

  override fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult {
    val editor = getData() ?: return TimestampUpgradeResult.Failure("Editor not found")
    if (caretTimestamps.size != editor.caretModel.caretCount) return TimestampUpgradeResult.Failure("Caret count was changed")

    val allCarets = editor.caretModel.allCarets
    val upgradedCarets = mutableListOf<CaretTimestamp>()
    for ((idx, caretTimestamp) in caretTimestamps.withIndex()) {
      val upgradeResult = caretTimestamp.upgrade(actualSet)
      if (upgradeResult !is TimestampUpgradeResult.Success<*>) return upgradeResult
      val upgradedCaret = upgradeResult.timestamp as CaretTimestamp

      val caret = allCarets[idx]
      val offset = upgradedCaret.model.offset.offset
      if (caret.offset != offset) return TimestampUpgradeResult.RetryRequest("Caret position was changed")

      val (selectionStart, selectionEnd) = upgradedCaret.model.selection.range
      if (caret.selectionStart != selectionStart || caret.selectionEnd != selectionEnd) return TimestampUpgradeResult.RetryRequest("Selection was changed")

      upgradedCarets.add(upgradedCaret)
    }

    return TimestampUpgradeResult.Success(EditorTimestamp(model.editorId, project, upgradedCarets))
  }
}

class BackendEditorTimestampProvider(project: Project) : BackendActionTimestampProvider(project) {
  override fun createSet(actionId: String, dataContext: DataContext): ActionTimestampSet {
    val editor = CommonDataKeys.EDITOR.getData(dataContext) ?: return ActionTimestampSet.EMPTY
    val textControlId = editor.getTextControlId() ?: return ActionTimestampSet.EMPTY

    val caretTimestamps = editor.caretModel.allCarets.mapNotNull { caret ->
      CaretTimestamp.create(project, caret)
    }
    return ActionTimestampSet(EditorTimestamp(textControlId, project, caretTimestamps))
  }

  override fun createSet(actionId: String, setModel: ActionTimestampSetModel): ActionTimestampSet {
    val editorTimestampModel = setModel.getTimestamp(EditorTimestamp.NAME) ?: return ActionTimestampSet.EMPTY

    val textControlId = editorTimestampModel.editorId

    val caretTimestamps = editorTimestampModel.carets.map {
      val offsetTimestampModel = it.offset
      val offsetTimestamp = OffsetTimestamp(offsetTimestampModel.offset, offsetTimestampModel.isStickingToRight, offsetTimestampModel.document)

      val rangeMarkerTimestampModel = it.selection
      val range = rangeMarkerTimestampModel.range.let { r -> Range(r.startOffset, r.endOffset) }
      val allowInnerModifications = rangeMarkerTimestampModel.allowInnerModifications
      val rangeMarkerTimestamp = RangeMarkerTimestamp(range, allowInnerModifications, rangeMarkerTimestampModel.document)
      CaretTimestamp(offsetTimestamp, rangeMarkerTimestamp)
    }

    val editorTimestamp = EditorTimestamp(textControlId, project, caretTimestamps)
    return ActionTimestampSet(editorTimestamp)
  }
}