package com.jetbrains.rdserver.actions.timestamps

import com.intellij.openapi.diagnostic.logger
import com.jetbrains.rd.ide.document.DocumentLogRangeShifter
import com.jetbrains.rd.ide.document.ProtocolDocumentSynchronizer
import com.jetbrains.rd.ide.model.DocumentTimestampModel
import com.jetbrains.rd.ide.model.OffsetTimestampModel
import com.jetbrains.rd.ide.model.RangeMarkerTimestampModel
import com.jetbrains.rd.platform.Range
import com.jetbrains.rd.platform.fromModel
import com.jetbrains.rd.platform.toModel
import com.jetbrains.rd.platform.util.IDocumentRangeShifter
import com.jetbrains.rdserver.actions.ActionTimestampKey
import com.jetbrains.rdserver.actions.ActionTimestampSet
import com.jetbrains.rdserver.actions.TimestampUpgradeResult
import com.jetbrains.rdserver.actions.UpgradableTimestamp
import com.jetbrains.rdserver.document.BackendDocumentHost

/**
 * Base timestamp which doesn't get transferred on its own
 * but as a part of other data which has a range in it (possible data could be selected text in the editor).
 * Using this timestamp turns a range into some sort of [com.intellij.openapi.editor.RangeMarker]
 * which gets updated according to document changes happened after protocol counter part was created.
 * @see OffsetTimestamp for transferring arbitrary data with one offset
 */
class RangeMarkerTimestamp(range: Range, allowInnerModifications: Boolean, documentTimestampModel: DocumentTimestampModel) :
  UpgradableTimestamp<RangeMarkerTimestampModel>(NAME, RangeMarkerTimestampModel(range.toModel(), allowInnerModifications, documentTimestampModel)) {
  companion object {
    val NAME = ActionTimestampKey<RangeMarkerTimestamp, RangeMarkerTimestampModel>("DOCUMENT_RANGE")
    val LOG = logger<RangeMarkerTimestamp>()

    fun create(documentSynchronizer: ProtocolDocumentSynchronizer, range: Range, allowInnerModifications: Boolean = true): RangeMarkerTimestamp {
      val newDocumentTimestamp = DocumentTimestampModel(documentSynchronizer.documentId, documentSynchronizer.version)
      return RangeMarkerTimestamp(range, allowInnerModifications, newDocumentTimestamp)
    }
  }

  override fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult {
    val documentTimestampModel = model.document
    val documentId = documentTimestampModel.documentId // maybe upgrade to handle file renames and such

    val project = actualSet.getTimestamp(ProjectTimestamp.PROJECT)?.getData() ?: return TimestampUpgradeResult.Failure("No Project")
    val documentHost = BackendDocumentHost.getInstance(project)
    val synchronizer = documentHost.tryGetSynchronizer(documentId) ?: return TimestampUpgradeResult.Failure("No Document Synchronizer")
    val shifter = createRangeShifter(synchronizer) ?: return TimestampUpgradeResult.Failure("No Document Log")

    val repairedRange = shifter.shiftRange(
      documentTimestampModel.version,
      model.range.fromModel(),
      model.allowInnerModifications,
      false, // todo support greediness for timestamp ranges
      false
    )
    if (!repairedRange.isValid) return TimestampUpgradeResult.RetryRequest("Range was changed beyond recovering")

    val rangeMarkerTimestamp = create(synchronizer, repairedRange, model.allowInnerModifications)
    return TimestampUpgradeResult.Success(rangeMarkerTimestamp)
  }
}


/**
 * Base timestamp which doesn't get transferred on its own
 * but as a part of other data which has an offset in it (e.g. a caret position, execution point).
 * Using this timestamp turns an offset into some sort of [com.intellij.openapi.editor.RangeMarker]
 * which gets updated according to document changes happened after protocol counter part was created.
 * @see RangeMarkerTimestamp for transferring arbitrary data with two offsets
 */
class OffsetTimestamp(offset: Int, isStickingToRight: Boolean, documentTimestampModel: DocumentTimestampModel) :
  UpgradableTimestamp<OffsetTimestampModel>(NAME, OffsetTimestampModel(offset, isStickingToRight, documentTimestampModel)) {
  companion object {
    val NAME = ActionTimestampKey<OffsetTimestamp, OffsetTimestampModel>("DOCUMENT_OFFSET")
    val LOG = logger<OffsetTimestamp>()

    fun create(documentSynchronizer: ProtocolDocumentSynchronizer, offset: Int, isStickingToRight: Boolean = false): OffsetTimestamp {
      val newDocumentTimestamp = DocumentTimestampModel(documentSynchronizer.documentId, documentSynchronizer.version)
      return OffsetTimestamp(offset, isStickingToRight, newDocumentTimestamp)
    }
  }

  override fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult {
    val documentTimestampModel = model.document
    val documentId = documentTimestampModel.documentId // maybe upgrade to handle file renames and such

    val project = actualSet.getTimestamp(ProjectTimestamp.PROJECT)?.getData() ?: return TimestampUpgradeResult.Failure("No Project")
    val documentHost = BackendDocumentHost.getInstance(project)
    val synchronizer = documentHost.tryGetSynchronizer(documentId) ?: return TimestampUpgradeResult.Failure("No Document Synchronizer")
    val shifter = createRangeShifter(synchronizer) ?: return TimestampUpgradeResult.Failure("No Document")

    val repairedOffset = shifter.shiftOnePoint(documentTimestampModel.version, model.offset, model.isStickingToRight)
    if (repairedOffset == -1) return TimestampUpgradeResult.RetryRequest("The offset was changed beyond recovering")

    return TimestampUpgradeResult.Success(create(synchronizer, repairedOffset, model.isStickingToRight))
  }
}

private fun createRangeShifter(documentSynchronizer: ProtocolDocumentSynchronizer): IDocumentRangeShifter? {
  return DocumentLogRangeShifter(documentSynchronizer)
}
