package com.jetbrains.rdserver.build

import com.intellij.build.BuildDescriptor
import com.intellij.build.DefaultBuildDescriptor
import com.intellij.build.FileNavigatable
import com.intellij.build.events.*
import com.intellij.build.events.OutputBuildEvent
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.intellij.pom.Navigatable
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.ThreadSafeIdHolder

fun BuildEvent.toProtocolModel(project: Project, buildId: Any): BuildEventModel = when (this) {
  is StartBuildEvent -> StartBuildEventModel(
    buildDescriptor.toProtocolModel(project),
    project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) }, eventTime, message
  )
  is StartEvent -> StartEventModel(
    project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) }, eventTime, message
  )
  is FinishBuildEvent -> FinishBuildEventModel(
    result.toProtocolModel(),
    project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) }, eventTime, message
  )
  is FinishEvent -> FinishEventModel(
    result.toProtocolModel(),
    project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) }, eventTime, message
  )
  is MessageEvent -> MessageEventModel(
    kind.toProtocolModel(), group, getNavigatable(project)?.toProtocolModel(), result.details,
    project.getIdFrom(buildId), project.getIdFrom(id), project.getIdFrom(logIfNull(parentId, 0)), eventTime, message
  )
  is OutputBuildEvent -> OutputBuildEventModel(
    isStdOut, project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) },
    eventTime, (this as OutputBuildEvent).message
  )
  is ProgressBuildEvent -> ProgressBuildEventModel(
    total, progress, unit,
    project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) }, eventTime, message
  )
  else -> PlainBuildEventModel(
    project.getIdFrom(buildId), project.getIdFrom(id), parentId?.let { project.getIdFrom(it) }, eventTime, message
  )
}

private fun Project.getIdFrom(value: Any): Long {
  return this.service<ThreadSafeIdHolder>().getId(value)
}

private fun Navigatable.toProtocolModel() = when (this) {
  is FileNavigatable -> fileDescriptor?.let {
    FilePositionModel(it.file.path, it.line, it.column)
  }
  else -> null
}

private fun MessageEvent.Kind.toProtocolModel() = when (this) {
  MessageEvent.Kind.ERROR -> MessageEventKind.Error
  MessageEvent.Kind.WARNING -> MessageEventKind.Warning
  MessageEvent.Kind.INFO -> MessageEventKind.Info
  MessageEvent.Kind.STATISTICS -> MessageEventKind.Statistics
  MessageEvent.Kind.SIMPLE -> MessageEventKind.Simple
}

private fun EventResult.toProtocolModel() = when (this) {
  is SuccessResult -> BuildResultKindModel.Success
  is FailureResult -> BuildResultKindModel.Failure
  is SkippedResult -> BuildResultKindModel.Skipped
  else -> BuildResultKindModel.Other
}

private fun BuildDescriptor.toProtocolModel(project: Project): BuildDescriptorModel {
  val defaultBuildDescriptor = this as? DefaultBuildDescriptor
  val isActivateToolWindowWhenAdded = defaultBuildDescriptor?.isActivateToolWindowWhenAdded
  val isActivateToolWindowWhenFailed = defaultBuildDescriptor?.isActivateToolWindowWhenFailed
  val isAutoFocusContent = defaultBuildDescriptor?.isAutoFocusContent
  return BuildDescriptorModel(project.getIdFrom(id), title, workingDir, startTime, 
                              isActivateToolWindowWhenAdded, isActivateToolWindowWhenFailed, isAutoFocusContent)
}

private fun <T> logIfNull(value: T?, default: T): T {
  if (value != null)
    return value

  logger<BuildHost>().error("Value must be not null")

  return default
}