// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.codeInsight

import com.intellij.codeInsight.hint.HintManager
import com.intellij.codeInsight.hint.HintManagerImpl
import com.intellij.codeInsight.hint.HintUtil
import com.intellij.codeInsight.hint.QuestionAction
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.Disposable
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.rd.createLifetime
import com.intellij.ui.LightweightHint
import com.intellij.ui.awt.RelativePoint
import com.jetbrains.ide.model.uicontext.ContextKey
import com.jetbrains.rd.util.reactive.IProperty
import com.jetbrains.rd.util.reactive.adviseOnce
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.startup.serverSolution
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rdserver.ui.BackendUiContextHost
import com.jetbrains.rd.ide.model.PopupActionModel
import java.awt.Point
import javax.swing.JComponent

/**
 * @author yole
 */
class BackendHintManager : HintManagerImpl(), Disposable {
  private var currentQuestionAction: QuestionAction? = null
  private var currentQuestionHint: LightweightHint? = null
  private var currentPopupActionProperty: IProperty<PopupActionModel?>? = null

  override fun showHint(component: JComponent, p: RelativePoint, flags: Int, timeout: Int) {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.showHint(component, p, flags, timeout)
    }
    //TODO
  }


  override fun showHint(component: JComponent, p: RelativePoint, flags: Int, timeout: Int, onHintHidden: Runnable?) {
      if (ClientId.isCurrentlyUnderLocalId) {
        return super.showHint(component, p, flags, timeout, onHintHidden)
      }
      onHintHidden?.run()
  }

  override fun showQuestionHint(editor: Editor,
                                p: Point,
                                offset1: Int,
                                offset2: Int,
                                hint: LightweightHint,
                                action: QuestionAction,
                                constraint: Short) {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.showQuestionHint(editor, p, offset1, offset2, hint, action, constraint)
    }

    val textControlId = editor.getTextControlId() ?: return
    val project = editor.project ?: return
    val solution = project.serverSolution
    val text = HintUtil.getHintLabel(hint.component) ?: return
    updateLastEditor(editor)
    val icon = HintUtil.getHintIcon(hint.component)
    val popupActionModel = PopupActionModel(
      offset1,
      offset2,
      text,
      icon?.toProtocolModel(),
      BackendUiContextHost.getInstance(project).get(ContextKey.Focus, ContextKey.TextControl, ContextKey.Document),
      textControlId
    )
    currentQuestionAction = action
    currentQuestionHint = hint
    currentPopupActionProperty = solution.actionPopupSession.popupAction
    val lifetime = this@BackendHintManager.createLifetime()

    popupActionModel.actionExecuted.adviseOnce(lifetime) {
      action.execute()
      solution.actionPopupSession.popupAction.set(null)
    }

    solution.actionPopupSession.popupAction.set(popupActionModel)
    solution.actionPopupSession.popupAction.change.adviseOnce(lifetime) {
      if (it == null) {
        hint.hide(false)
        currentQuestionAction = null
        currentPopupActionProperty = null
      }
    }
  }

  override fun hideHints(mask: Int, onlyOne: Boolean, editorChanged: Boolean): Boolean {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.hideHints(mask, onlyOne, editorChanged)
    }

    if (mask and (HintManager.HIDE_BY_ANY_KEY or HintManager.HIDE_BY_TEXT_CHANGE or HintManager.UPDATE_BY_SCROLLING or HintManager.HIDE_IF_OUT_OF_EDITOR or HintManager.DONT_CONSUME_ESCAPE) != 0) {
      currentPopupActionProperty?.set(null)
    }

    return false
  }

  override fun performCurrentQuestionAction(): Boolean {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.performCurrentQuestionAction()
    }

    currentQuestionAction?.let {
      if (it.execute()) {
        currentQuestionHint?.hide()
        currentPopupActionProperty?.set(null)
      }
      return true
    }
    currentQuestionHint = null
    currentQuestionAction = null
    return false
  }

  override fun dispose() {
  }
}
