// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.codeInsight

import com.intellij.codeInsight.template.TemplateManagerListener
import com.intellij.codeInsight.template.impl.*
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.TextRange
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.rdtext.intrinsics.TextBufferVersion
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.requests.BackendIdeGateway

@Suppress("NonDefaultConstructor")
class BackendLiveTemplateHost(
  lifetime: Lifetime,
  project: Project,
  private val clientId: ClientId
  ){
  companion object {
    private val LOG = logger<BackendLiveTemplateHost>()
  }

  init {
    project.messageBus.connect(lifetime.createNestedDisposable()).subscribe(TemplateManagerImpl.TEMPLATE_STARTED_TOPIC, TemplateManagerListener {
      if (clientId != ClientId.current) return@TemplateManagerListener
      if (it.isFinished) return@TemplateManagerListener

      val editor = it.editor ?: run {
        LOG.error("editor is null")
        return@TemplateManagerListener
      }

      val textControlId = editor.getTextControlId() ?: run {
        LOG.error("textControlId is null")
        return@TemplateManagerListener
      }

      val liveTemplateDescription = convertTemplateData(it, textControlId)
      val templateExpansion = RdLiveTemplateExpansion(liveTemplateDescription, textControlId)

      saveSelectionAndCaretState(editor) {
        it.gotoEnd(false)
      }
      val editorHost = BackendEditorHost.getInstance(project)
      val editorVersion = editorHost.tryGetSynchronizer(textControlId)?.version ?: error("editor synchronizer is null")
      val documentHost = BackendDocumentHost.getInstance(project)
      val documentVersion = documentHost.tryGetSynchronizer(textControlId.documentId)?.version ?: error("document synchronizer is null")
      val ideGateway = BackendIdeGateway.getInstance(project)
      ideGateway.sendEditorChange(templateExpansion, editorVersion, editorVersion, documentVersion)
    })
  }

  private fun saveSelectionAndCaretState(editor: Editor, block: () -> Unit) {
    val beforeCaretOffset = editor.caretModel.offset
    val beforeSelectionStart = editor.selectionModel.selectionStart
    val beforeSelectionEnd = editor.selectionModel.selectionEnd
    try {
      block()
    }
    finally {
      editor.selectionModel.setSelection(beforeSelectionStart, beforeSelectionEnd)
      editor.caretModel.moveToOffset(beforeCaretOffset)
    }
  }

  private data class TemplateSegment(val name: String, val textRange: TextRange)

  private fun convertTemplateData(templateState: TemplateState,
                                  textControlId: TextControlId): LiveTemplateInitialDataModel {
    val template = templateState.template
    val segments = (0 until template.segmentsCount).map {
      TemplateSegment(template.getSegmentName(it), templateState.getSegmentRange(it))
    }

    // If variable is tracked using VariableNode, additional range will be added to this variable
    val additionalRanges = mutableSetOf<Pair<String, LiveTemplateVariableRangeModel>>()
    val variables = template.variables.mapNotNull { variable ->
      createVariableModel(variable, templateState, segments, additionalRanges)
    }.toMutableList()

    additionalRanges.forEach { nameToRange ->
      variables.forEachIndexed { index, variable ->
        if (variable.name == nameToRange.first) {
          variables[index] = variable withRange nameToRange.second
        }
      }
    }

    return LiveTemplateInitialDataModel(textControlId, TextBufferVersion.INIT_VERSION, variables,
                                        segments.firstOrNull { it.name == "END" }?.textRange?.startOffset ?: -1, TemplateHint.Other)
  }

  private fun createVariableModel(
    variable: Variable,
    templateState: TemplateState,
    segments: List<TemplateSegment>,
    additionalRanges: MutableSet<Pair<String, LiveTemplateVariableRangeModel>>
  ): LiveTemplateVariableModel? {
    val variableName = variable.name
    val value = templateState.getVariableValue(variableName)?.text ?: run {
      LOG.error("`$variableName` variable value is null")
      return@run ""
    }
    val expression = variable.expression

    if (expression is VariableNode) {
      segments.filter { it.name == variableName }.map {
        additionalRanges += expression.name to LiveTemplateVariableRangeModel(it.textRange.startOffset, it.textRange.endOffset)
      }
      return null
    }

    val liveTemplateVariableRange = segments.filter { it.name == variableName }.map {
      LiveTemplateVariableRangeModel(it.textRange.startOffset, it.textRange.endOffset)
    }

    val segmentIndex = segments.withIndex().first { it.value.name == variableName }.index
    val lookupItems = expression.calculateLookupItems(templateState.getExpressionContextForSegment(segmentIndex))?.map { it.lookupString } ?: emptyList()

    return LiveTemplateVariableModel(
      variableName,
      value,
      variable.isAlwaysStopAt,
      variable.skipOnStart(),
      false,
      liveTemplateVariableRange,
      listOf(),
      convertMacro(variable),
      lookupItems,
      FocusBehaviour.Hard
    )
  }

  private infix fun LiveTemplateVariableModel.withRange(newRange: LiveTemplateVariableRangeModel) =
    LiveTemplateVariableModel(name, currentValue, isStopAt, isSkipOnStart, isDiscardCompletion,ranges + newRange, listOf(), macro, lookupItems, focusBehaviour)

  private fun convertMacro(variable: Variable): LiveTemplateMacroModel {
    if (variable.expressionString.isNotEmpty()) {
      return LiveTemplateMacroModel(variable.expressionString, emptyList())
    }
    val expression = variable.expression
    if (expression is MacroCallNode) {
      return LiveTemplateMacroModel(expression.macro.name, emptyList())
    }
    return LiveTemplateMacroModel("", emptyList())
  }
}
