// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.codeInsight.hint

import com.intellij.codeInsight.AutoPopupController
import com.intellij.codeInsight.hint.*
import com.intellij.codeWithMe.ClientId
import com.intellij.lang.parameterInfo.ParameterInfoHandler
import com.intellij.lang.parameterInfo.ParameterInfoUIContext
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.application.ReadAction
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.NlsContexts
import com.intellij.openapi.util.TextRange
import com.intellij.openapi.util.text.StringUtil
import com.intellij.psi.PsiElement
import com.intellij.psi.util.PsiUtilBase
import com.intellij.ui.JBColor
import com.intellij.util.indexing.DumbModeAccessType
import com.intellij.util.indexing.FileBasedIndex
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ide.parameterInfo.RichTextParameterInfoUIContext
import com.jetbrains.rd.util.getLogger
import com.jetbrains.rd.util.trace
import com.jetbrains.rdserver.editors.parameterInfoModel
import java.awt.Color
import java.util.*
import java.util.function.Consumer

class BackendParameterInfoController(project: Project,
                                     editor: Editor,
                                     lbraceOffset: Int,
                                     descriptors: Array<Any?>?,
                                     highlighted: Any?,
                                     parameterOwner: PsiElement?,
                                     handler: ParameterInfoHandler<out PsiElement, out Any>,
                                     showHint: Boolean,
                                     requestFocus: Boolean) : ParameterInfoControllerBase(project, editor, lbraceOffset, descriptors,
                                                                                                                                           highlighted, parameterOwner, handler, showHint) {
  companion object {
    private val logger = getLogger<BackendParameterInfoController>()
    // TODO unify with ParameterInfoComponent.BACKGROUND
    private val BACKGROUND: Color = JBColor.namedColor("ParameterInfo.background", HintUtil.getInformationColor())
  }

  private val parameterInfoModel: ParameterInfoModel = editor.parameterInfoModel
    ?: throw IllegalStateException("There is no ParameterInfoModel attached to editor $editor")

  init {
    setupListeners()
    showHint(requestFocus, mySingleParameterInfo)
  }

  override fun canBeDisposed(): Boolean {
    return !isHintShown(true)
  }

  override fun isHintShown(anyType: Boolean): Boolean {
    return parameterInfoModel.parameterInfo.value != null
  }

  override fun showHint(requestFocus: Boolean, singleParameterInfo: Boolean) {
    val context = ParameterInfoContext(listOf(), 0, Range(0, 0))
    val model = ParameterInfoContextsModel(context)
    parameterInfoModel.parameterInfo.value = model
    updateComponent()
  }

  override fun hideHint() {
    parameterInfoModel.parameterInfo.value = null
  }

  override fun moveToParameterAtOffset(offset: Int) {
    // todo
  }

  override fun setPreservedOnHintHidden(value: Boolean) {
    // do nothing
  }

  override fun isPreservedOnHintHidden(): Boolean {
    return false
  }


  private fun updateUi(singleParameterInfo: Boolean): Model {
    val context = MyParameterContext(myParameterInfoControllerData, singleParameterInfo)
    for (i in myParameterInfoControllerData.descriptors.indices) {
      context.i = i
      val o = myParameterInfoControllerData.descriptors[i]
      val isHighlighted = myParameterInfoControllerData.descriptors[i] == myParameterInfoControllerData.highlighted
      if (isHighlighted) {
        context.result.highlightedSignature = i
      }
      val handler = myParameterInfoControllerData.handler
      DumbModeAccessType.RELIABLE_DATA_ONLY.ignoreDumbMode { handler.updateUI(o, context) }
    }
    return context.result
  }

  override fun updateComponent() {
    if (canBeDisposed()) {
      Disposer.dispose(this)
      return
    }
    val file = PsiUtilBase.getPsiFileInEditor(myEditor, myProject)
    val offset: Int = currentOffset
    val context = UpdateParameterInfoContextBase(offset, file)
    executeFindElementForUpdatingParameterInfo(context, Consumer { elementForUpdating: PsiElement? ->
      myParameterInfoControllerData.handler.processFoundElementForUpdatingParameterInfo(elementForUpdating, context)
      if (elementForUpdating != null) {
        executeUpdateParameterInfo(elementForUpdating, context, Runnable {
          val knownParameter = (myParameterInfoControllerData.descriptors.size == 1 ||
                                myParameterInfoControllerData.highlighted != null) &&
                               myParameterInfoControllerData.currentParameterIndex != -1
          if (mySingleParameterInfo && !knownParameter && isHintShown(true)) {
            hideHint()
          }
          if (isPreservedOnHintHidden && knownParameter && !isHintShown(true)) {
            AutoPopupController.getInstance(myProject).autoPopupParameterInfo(myEditor, null)
          }
          if (!myDisposed && (isHintShown(true) && !myEditor.isDisposed
                              || ApplicationManager.getApplication().isHeadlessEnvironment)) {
            val result: Model = updateUi(mySingleParameterInfo)
            result.project = myProject
            result.range = myParameterInfoControllerData.parameterOwner.textRange
            result.editor = myEditor
            for (listener in ParameterInfoListener.EP_NAME.extensionList) {
              listener.hintUpdated(result)
            }
            syncHintWithThinClient(result)
          }
        })
      }
      else {
        hideHint()
        if (!isPreservedOnHintHidden) {
          Disposer.dispose(this)
        }
      }
    })
  }

  private fun executeUpdateParameterInfo(elementForUpdating: PsiElement,
                                         context: UpdateParameterInfoContextBase,
                                         continuation: Runnable?) {
    val parameterOwner = context.parameterOwner
    if (parameterOwner != null && parameterOwner != elementForUpdating) {
      context.removeHint()
      return
    }
    ParameterInfoTaskRunnerUtil.runTask(
      myProject,
      ReadAction.nonBlocking<PsiElement> {
        DumbModeAccessType.RELIABLE_DATA_ONLY.ignoreDumbMode {
          myParameterInfoControllerData.handler.updateParameterInfo(elementForUpdating, context)
        }
        elementForUpdating
      }
        .withDocumentsCommitted(myProject)
        .expireWhen {
          canBeDisposed() || currentOffset != context.offset ||
          !elementForUpdating.isValid
        }
        .expireWith(this),
      Consumer { element: PsiElement? ->
        if (element != null && continuation != null) {
          context.applyUIChanges()
          continuation.run()
        }
      },
      null,
      myEditor)
  }

  private fun syncHintWithThinClient(model: Model) {
    val signatures = model.signatures.mapIndexedNotNull { _, it ->
      when (it) {
        is SignatureItem -> {
          SignatureItem(
            it.text,
            // TODO try don't use empty list for true-condition-case and fix rendering of Black Bold text in disabled lines in Idea.
            if (it.disabled) listOf() else it.startOffsets,
            if (it.disabled) listOf() else it.endOffsets,
            model.current,
            it.deprecated,
            it.disabled
          )
        }
        is RawSignatureItem -> {
          HtmlRawSignatureItem(it.htmlText, model.current, isDeprecated = false, isDisabled = false)
        }
        // in case of CWM and Rider we can intercept parameter info model with rich text using MyParameterContext.setupRawUIComponentPresentation(RichTextRawSignatureItem)
        // and pass the rich text item further to thin client instead of rendering into html here. This prevents rendering of wrong colors into HTML
        // because it must be rendered on client side with proper colors of user's theme
        is RichTextSignatureItem -> {
          it.richTextRawSignatureItem
        }
        else -> {
          logger.trace { "Unknown signature instance $it" }
          null
        }
      }
    }
    val context = ParameterInfoContext(signatures, null,
                                       Range(model.range.startOffset, model.range.endOffset))
    val parameterInfoContextsModel = ParameterInfoContextsModel(context)
    parameterInfoModel.parameterInfo.value = parameterInfoContextsModel
    parameterInfoContextsModel.context.set(context)
  }

  internal class MyParameterContext(private val myParameterInfoControllerData: ParameterInfoControllerData,
                                    private val mySingleParameterInfo: Boolean) : ParameterInfoUIContext, RichTextParameterInfoUIContext {
    internal var i = 0
    internal val result = Model()

    override fun setupUIComponentPresentation(text: @NlsContexts.Label String?,
                                              highlightStartOffset: Int,
                                              highlightEndOffset: Int,
                                              isDisabled: Boolean,
                                              strikeout: Boolean,
                                              isDisabledBeforeHighlight: Boolean,
                                              background: Color?): String? {
      val split = StringUtil.split(text!!, ",", false)
      val plainLine = StringBuilder()
      val startOffsets: MutableList<Int> = ArrayList()
      val endOffsets: MutableList<Int> = ArrayList()
      val highlightRange = if (0 <= highlightStartOffset && highlightStartOffset <= highlightEndOffset) TextRange(highlightStartOffset,
                                                                                                                  highlightEndOffset)
      else null
      for (j in split.indices) {
        val line = split[j]
        val startOffset = plainLine.length
        startOffsets.add(startOffset)
        plainLine.append(line)
        val endOffset = plainLine.length
        endOffsets.add(endOffset)
        if (highlightRange != null && highlightRange.intersects(TextRange(startOffset, endOffset))) {
          result.current = j
        }
      }
      val item = SignatureItem(plainLine.toString(), strikeout, isDisabled,
                               startOffsets, endOffsets)
      result.signatures.add(item)
      return null
    }

    override fun setupRawUIComponentPresentation(htmlText: @NlsContexts.Label String?) {
      val item = RawSignatureItem(htmlText)
      result.current = currentParameterIndex
      result.signatures.add(item)
    }

    override val needForwardRichTextSignature: Boolean
      get() = !ClientId.isCurrentlyUnderLocalId

    override fun setupRawUIComponentPresentation(richTextSignatureModel: RichTextRawSignatureItem) {
      result.current = currentParameterIndex
      result.signatures.add(RichTextSignatureItem(richTextSignatureModel))
    }

    override fun isUIComponentEnabled(): Boolean {
      return myParameterInfoControllerData.isDescriptorEnabled(i)
    }

    override fun setUIComponentEnabled(enabled: Boolean) {
      myParameterInfoControllerData.setDescriptorEnabled(i, enabled)
    }

    override fun getCurrentParameterIndex(): Int {
      return myParameterInfoControllerData.currentParameterIndex
    }

    override fun getParameterOwner(): PsiElement {
      return myParameterInfoControllerData.parameterOwner
    }

    override fun isSingleOverload(): Boolean {
      return myParameterInfoControllerData.descriptors.size == 1
    }

    override fun isSingleParameterInfo(): Boolean {
      return mySingleParameterInfo
    }

    override fun getDefaultParameterColor(): Color {
      return BACKGROUND
    }
  }

  class RichTextSignatureItem(val richTextRawSignatureItem: RichTextRawSignatureItem) : SignatureItemModel
}