package com.intellij.codeWithMe.ui

import com.intellij.codeWithMe.ClientId
import com.intellij.icons.AllIcons
import com.intellij.notification.Notification
import com.intellij.notification.NotificationListener
import com.intellij.notification.NotificationType
import com.intellij.notification.impl.NotificationsManagerImpl
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.intellij.openapi.wm.WindowManager
import com.intellij.ui.BalloonLayoutData
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.CwmNotifications.CWM_NOTIFICATION_GROUP_ID
import com.jetbrains.rd.platform.codeWithMe.ui.AcceptUserDialogService
import com.jetbrains.rd.platform.codeWithMe.DistributedTestIndicator.isInDistributedTest
import com.jetbrains.rd.platform.codeWithMe.ui.CWMBaseAcceptUserDialog
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import org.jetbrains.annotations.Nls
import java.util.concurrent.CompletableFuture
import javax.swing.event.HyperlinkEvent

object AccessNotification {

  private val LOG = Logger.getInstance(javaClass)

  /**
   * Shows notification with two actions: 'accept' and 'decline'
   * Can be uses to request access to some feature \ resource from remote client
   * Can be executed from any thread
   */
  fun showNewAccessNotification(project: Project,
                                lifetime: Lifetime,
                                userName: String,
                                clientId: ClientId,
                                fingerprint: String,
                                dialogFactory: () -> CWMBaseAcceptUserDialog = { AcceptUserDialogService.getInstance().createAcceptUserDialog(project, userName, clientId) }): CompletableFuture<Boolean> {
    val future = CompletableFuture<Boolean>()

    application.invokeLater {
      val dialog = dialogFactory()
      dialog.setFingerprintString(fingerprint)

      if (application.isInDistributedTest) {
        LOG.warn("Host was launched as a part of a distributed test, automatically providing access to userName=$userName.")
        future.complete(true)
        return@invokeLater
      }

      dialog.show()

      dialog.result.whenComplete { r, ex ->
        if (ex != null)
          future.completeExceptionally(ex)
        else
          future.complete(r)
      }

      lifetime.onTermination {
        if (dialog.isShowing) {
          future.complete(false)
          dialog.close(1, false)
        }
      }
    }

    return future
  }

  /**
   * Shows notification with two actions: 'accept' and 'decline'
   * Can be uses to request access to some feature \ resource from remote client
   * Can be executed from any thread
   */
  @Deprecated("Use showNewAccessNotification")
  fun showAccessNotification(project: Project, lifetime: Lifetime, @Nls message: String): CompletableFuture<Boolean> {
    val accepted = CompletableFuture<Boolean>()
    val clientId = ClientId.currentOrNull

    application.invokeLater {
      var result = -1
      if (application.isInDistributedTest) {
        LOG.warn("Host was launched as a part of a distributed test, automatically providing an access to client request=$message.")
        result = Messages.YES
      }
      else {
        ClientId.withClientId(ClientId.localId) {
          result = Messages.showYesNoCancelDialog(
            project, message,
            CodeWithMePlatformBundle.message("product.name"),
            CodeWithMePlatformBundle.message("notification.access.accept"),
            CodeWithMePlatformBundle.message("notification.access.decline"),
            CodeWithMePlatformBundle.message("notification.access.skip"),
            AllIcons.Plugins.PluginLogo)
        }
      }

      ClientId.withClientId(clientId) {
        when (result) {
          Messages.YES -> {
            accepted.complete(true)
          }
          Messages.NO -> {
            accepted.complete(false)
          }
          else -> {
            showBallon(message, clientId, accepted, project, lifetime)
          }
        }
      }
    }

    return accepted
  }

  fun showStrictAccessNotification(project: Project, lifetime: Lifetime, @Nls message: String, userName: String): CompletableFuture<Boolean> {
    val accepted = CompletableFuture<Boolean>()
    val clientId = ClientId.currentOrNull

    application.invokeLater {
      var i = -1
      ClientId.withClientId(ClientId.localId) {
        val dialog =
          ConfirmationWithCheckboxDialog(message,
                                             CodeWithMePlatformBundle.message("notification.terminal.security.caption"),
                                             arrayOf(
                                               CodeWithMePlatformBundle.message("notification.access.decline"),
                                               CodeWithMePlatformBundle.message("notification.access.grant.access")
                                             ),
                                             CodeWithMePlatformBundle.message("notification.i.fully.trust", userName),
                                             false,
                                             1,
                                             0,
                                             Messages.getWarningIcon(),
                                             null)

        dialog.show()
        i = dialog.exitCode
      }
      ClientId.withClientId(clientId) {
        when (i) {
          Messages.YES -> {
            accepted.complete(true)
          }
          Messages.NO -> {
            accepted.complete(false)
          }
          else -> {
            showBallon(message, clientId, accepted, project, lifetime)
          }
        }
      }
    }

    return accepted
  }

  private fun showBallon(message: String,
                         clientId: ClientId?,
                         accepted: CompletableFuture<Boolean>,
                         project: Project,
                         lifetime: Lifetime) {
    val notification = Notification(
      CWM_NOTIFICATION_GROUP_ID,
      CodeWithMePlatformBundle.message("product.name"),
      "$message<br><a href=\"/accept\">${CodeWithMePlatformBundle.message("notification.access.accept")}</a> <a href=\"/decline\">${
        CodeWithMePlatformBundle.message("notification.access.decline")
      }</a>",
      NotificationType.INFORMATION,
      object : NotificationListener.Adapter() {
        override fun hyperlinkActivated(notificationFromListener: Notification, e: HyperlinkEvent) {
          notificationFromListener.hideBalloon()
          notificationFromListener.expire()

          ClientId.withClientId(clientId) {
            when (e.description) {
              "/accept" -> accepted.complete(true)
              "/decline" -> accepted.complete(false)
              else -> LOG.error("Unknown notification button: ${e.description}")
            }
          }
        }
      }
    ).also { it.isImportant = false }

    ClientId.withClientId(ClientId.localId) {
      val frame = WindowManager.getInstance().getIdeFrame(project)!!
      val balloon = NotificationsManagerImpl.createBalloon(
        frame, notification, true, false,
        BalloonLayoutData.fullContent(), project)

      val layout = frame.balloonLayout!!
      layout.add(balloon)

      lifetime.onTermination {
        application.invokeLater {
          balloon.dispose()
        }
      }
    }
  }
}