package com.intellij.ide.license;

import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.ide.plugins.PluginManagerCore;
import com.intellij.openapi.extensions.PluginId;
import com.intellij.openapi.util.NlsSafe;
import com.jetbrains.licenseServer.Products;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.UUID;

@SuppressWarnings("unused")
public final class ProductDescriptor {
  public static final UUID UNDEFINED = UUID.fromString("c3ea6721-9669-4af3-895f-75302648ff5e");

  public static final ProductDescriptor IDEA = new ProductDescriptor("idea", "II", Products.IDEA);
  public static final ProductDescriptor RUBYMINE = new ProductDescriptor("RubyMine", "RM", Products.RUBYMINE);
  public static final ProductDescriptor PYCHARM  = new ProductDescriptor("PyCharm", "PC", Products.PYCHARM);
  public static final ProductDescriptor PYCHARM_ANACONDA  = new ProductDescriptor("PyCharm (Anaconda Edition)", "PCA", Products.PYCHARM);
  public static final ProductDescriptor PYCHARM_DS  = new ProductDescriptor("JetBrains DataSpell", "PD", Products.PYCHARM);
  public static final ProductDescriptor WEBSTORM = new ProductDescriptor("WebStorm", "WS", Products.WEBSTORM);
  public static final ProductDescriptor PHPSTORM = new ProductDescriptor("PhpStorm", "PS", Products.PHPSTORM);
  public static final ProductDescriptor APPCODE = new ProductDescriptor("AppCode", "AC", Products.APPCODE);
  public static final ProductDescriptor DATAGRIP = new ProductDescriptor("DataGrip", "DB", Products.DBIDE);
  public static final ProductDescriptor CLION = new ProductDescriptor("CLion", "CL", Products.CLION);
  public static final ProductDescriptor RIDER = new ProductDescriptor("Rider", "RD", Products.RIDER);
  public static final ProductDescriptor GOIDE = new ProductDescriptor("GoLand", "GO", UUID.fromString("6ca374ac-f547-4984-be94-adb3e47b580c"));
  public static final ProductDescriptor DATALORE = new ProductDescriptor("Datalore", "DL", UNDEFINED);
  public static final ProductDescriptor DOTCOVER = new ProductDescriptor("dotCover", "DC", Products.DOTCOVER);
  public static final ProductDescriptor DOTMEMORY = new ProductDescriptor("dotMemory", "DM", Products.DOTMEMORY);
  public static final ProductDescriptor DOTTRACE = new ProductDescriptor("dotTrace", "DPN", Products.DOTTRACE);
  public static final ProductDescriptor RESHARPER = new ProductDescriptor("ReSharper", "RS", Products.RESHARPER);
  public static final ProductDescriptor RESHARPER_CPP = new ProductDescriptor("ReSharper C++", "RC", Products.RESHARPER_CPP);
  public static final ProductDescriptor UPSOURCE = new ProductDescriptor("Upsource", "US", UNDEFINED);
  public static final ProductDescriptor RIDER_CPP = new ProductDescriptor("Rider C++", "RDCPPP", UUID.fromString("eb4a28b4-a30f-41ee-9df5-6a42f2a9a58a"));
  public static final ProductDescriptor CWM = new ProductDescriptor("CodeWithMe", "PCWMP", UNDEFINED, false);
  public static final ProductDescriptor CWMRD = new ProductDescriptor("CodeWithMe RD", "PRDEV", UNDEFINED, false);

  private final @NotNull String shortName;
  private final @NotNull String productCode;
  private final @NotNull UUID productFamily; // older identifier, left for compatibility; used by license server
  private final boolean evalSupported;

  private static final ProductDescriptor[] VALUES;
  static {
    final List<ProductDescriptor> result = new ArrayList<>();
    for (Field field : ProductDescriptor.class.getDeclaredFields()) {
      final int modifiers = field.getModifiers();
      if (Modifier.isStatic(modifiers) && Modifier.isFinal(modifiers) && ProductDescriptor.class.equals(field.getType())) {
        try {
          result.add((ProductDescriptor)field.get(null));
        }
        catch (Throwable ignored) {
        }
      }
    }
    VALUES = result.toArray(new ProductDescriptor[0]);
  }

  //public static void main(String[] args) {
  //  System.out.println(UUID.randomUUID());
  //}

  private ProductDescriptor(@NonNls @NotNull String shortName, @NonNls @NotNull String productCode, @NotNull UUID productFamily) {
    this(shortName, productCode, productFamily, true);
  }

  private ProductDescriptor(@NonNls @NotNull String shortName, @NonNls @NotNull String productCode, @NotNull UUID productFamily, boolean evalSupported) {
    this.shortName = shortName;
    this.productCode = productCode;
    this.productFamily = productFamily;
    this.evalSupported = evalSupported;
  }

  public boolean isPluginDisabled() {
    for (IdeaPluginDescriptor pl : PluginManagerCore.getPlugins()) {
      if (productCode.equals(pl.getProductCode())) {
        PluginId id;
        return !pl.isEnabled() || (id = pl.getPluginId()) != null && PluginManagerCore.isDisabled(id);
      }
    }
    return false;
  }

  @NotNull
  public @NlsSafe String getShortName() {
    return shortName;
  }

  @NotNull
  public String getProductCode() {
    return productCode;
  }

  @NotNull
  public UUID getProductFamily() {
    return productFamily;
  }

  public boolean isEvalSupported() {
    return evalSupported;
  }

  @Nullable
  public static ProductDescriptor findByProductName(String productName) {
    if (productName != null) {
      for (ProductDescriptor descriptor : values()) {
        if (descriptor.getShortName().equalsIgnoreCase(productName)) {
          return descriptor;
        }
      }
    }
    return null;
  }

  @Nullable
  public static ProductDescriptor findByProductCode(String productCode) {
    if (productCode != null) {
      for (ProductDescriptor descriptor : values()) {
        if (descriptor.getProductCode().equals(productCode)) {
          return descriptor;
        }
      }
    }
    return null;
  }

  @NotNull
  public static ProductDescriptor get(String productName, String productCode) {
    final ProductDescriptor known = findByProductCode(productCode);
    return known != null? known : new ProductDescriptor(productName, productCode, UNDEFINED);
  }

  public static ProductDescriptor[] values() {
    return VALUES;
  }
}