package com.intellij.ide.license.common;

import com.intellij.ide.Prefs;
import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.license.LicenseValidator;
import com.intellij.ide.license.PluginLicenseFactory;
import com.intellij.ide.license.ProductDescriptor;
import com.intellij.ide.license.impl.*;
import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.licensecommon.util.SubscriptionVersion;
import com.intellij.openapi.application.PathManager;
import com.intellij.util.text.DateFormatUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.io.*;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Locale;

/**
 * @author Eugene Zhuravlev
 * Date: 02-Mar-18
 */
public final class PluginLicenseFactoryImpl implements PluginLicenseFactory {
  private static final int EVAL_PERIOD_DAYS = 30;
  private final ProductDescriptor myProductDescriptor;
  private final Date myPluginReleaseDate;
  private final int myPluginReleaseVersion;
  private final LicenseStore myLicenseStore;
  private final String myEvalSelector;
  private final boolean myIsOptional;

  public PluginLicenseFactoryImpl(IdeaPluginDescriptor descriptor) {
    myProductDescriptor = ProductDescriptor.get(descriptor.getName(), descriptor.getProductCode());
    final Date releaseDate = descriptor.getReleaseDate();
    myPluginReleaseDate = releaseDate != null? releaseDate : new Date();
    myEvalSelector = releaseDate != null? new SimpleDateFormat("yyyyMMdd", Locale.US).format(releaseDate) : "";
    myIsOptional = descriptor.isLicenseOptional();
    int ver = descriptor.getReleaseVersion();
    if (ver == 0 && releaseDate != null) {
      // build the version from release date
      try {
        ver = SubscriptionVersion.createVersion(releaseDate);
      }
      catch (Throwable ignored) {
      }
    }
    myPluginReleaseVersion = ver;
    myLicenseStore = new PluginLicenseFile(myProductDescriptor, myPluginReleaseDate);
  }

  @Override
  public Date getPluginReleaseDate() {
    return myPluginReleaseDate;
  }

  @Override
  public boolean isLicenseOptional() {
    return myIsOptional;
  }

  @Override
  public LicenseData createFromUser(String key, String userName) {
    try {
      return new CertKeyLicenseData(key, myProductDescriptor.getProductCode(), myPluginReleaseDate);
    }
    catch (Throwable ignored) {
    }
    return null;
  }

  @Override
  public ServerLicenseData createServerLicense(String ticketId, long prolongationPeriod, String serverUrl, String licensee,
                                               int licenseType, String confirmationStamp, String metadata) {
    return new ServerLicenseData(ticketId, prolongationPeriod, serverUrl, licensee, licenseType, confirmationStamp, metadata);
  }

  @Override
  public LicenseValidator getLicenseValidator() {
    return licenseData -> {
      if (licenseData instanceof EvaluationLicenseData) {
        if (licenseData.isExpired()) {
          throw new EvaluationExpiredException();
        }
        return true;
      }
      if (licenseData instanceof CertKeyLicense) {
        final CertKeyLicense certKeyLicense = (CertKeyLicense)licenseData;
        if (!Boolean.TRUE.equals(certKeyLicense.getCertKey().eval(new IsApplicableTo(certKeyLicense.getProductCode())))) {
          throw new InvalidLicenseException("Specified activation code is not applicable to this product");
        }
        if (licenseData.isExpired()) {
          throw new LicenseExpiredException(licenseData.getExpirationDate());
        }
        return true;
      }
      return false;
    };
  }

  @Override
  public LicenseStore getLicenseStore() {
    return myLicenseStore;
  }

  @Override
  public int getCurrentMajorVersion() {
    return myPluginReleaseVersion;
  }

  @NotNull
  @Override
  public ProductDescriptor getProductDescriptor() {
    return myProductDescriptor;
  }

  @Override
  public String getCurrentEvaluationSelector() {
    return myEvalSelector;
  }

  @Override
  public boolean isForceReleaseEvaluationSelector() {
    return false;
  }

  @Override
  @Nullable
  public EvaluationLicenseData findExistingEvaluationLicense() {
    if (!myProductDescriptor.isEvalSupported()) {
      return null;
    }
    final File evalLicenseFile = getEvalLicenseFile();
    if (evalLicenseFile.exists()) {
      try (DataInputStream stream = new DataInputStream(new FileInputStream(evalLicenseFile))){
        final EvaluationLicenseData eval = createEval(new Date(~stream.readLong()));
        final long expirationDate = eval.getExpirationDate().getTime();
        final String encodedExpirationDate = readPluginPropertyValue(getPluginEvalExpirationRegKey());
        if (encodedExpirationDate != null) {
          if (isCheating(encodedExpirationDate, expirationDate)) {
            // return the license that is for sure expired
            return createEval(new Date(System.currentTimeMillis() - (EVAL_PERIOD_DAYS + 1) * DateFormatUtil.DAY));
          }
        }
        else {
          final long threshold = (11 + (int) (Math.random() * 10)) * DateFormatUtil.DAY; // random days before eval expires
          if (expirationDate - System.currentTimeMillis() < threshold) {
            encodePluginExpirationDate(expirationDate);
          }
        }
        return eval;
      }
      catch (Throwable e) {
        return null;
      }
    }
    else {
      final String encodedExpirationDate = readPluginPropertyValue(getPluginEvalExpirationRegKey());
      if (encodedExpirationDate != null) {
        try {
          return createEval(new Date(
            Long.parseLong(encodedExpirationDate, 16) - EVAL_PERIOD_DAYS * DateFormatUtil.DAY
          ));
        }
        catch (NumberFormatException ignored) {
        }
      }
    }

    return null;
  }

  @Override
  @NotNull
  public EvaluationLicenseData createEvaluationLicense() {
    return createEval(new Date());
  }

  @NotNull
  private EvaluationLicenseData createEval(final Date issueDate) {
    return new EvaluationLicenseData(
      issueDate, getCurrentMajorVersion(), "Plugin " + myProductDescriptor.getShortName() + " evaluator", EVAL_PERIOD_DAYS
    );
  }

  @Override
  public boolean storeEvaluationLicense(EvaluationLicenseData license) {
    if (!myProductDescriptor.isEvalSupported()) {
      return false;
    }
    try {
      final File file = getEvalLicenseFile();
      try (DataOutputStream stream = new DataOutputStream(openFileStream(file))) {
        stream.writeLong(~license.getGenerationDate().getTime());
      }
      return true;
    }
    catch (IOException e) {
      JOptionPane.showMessageDialog(null, LicenseBundle.message("error.message.error.saving.evaluation.license.0", e.getMessage()));
      return false;
    }
  }

  @NotNull
  private static FileOutputStream openFileStream(File file) throws IOException {
    try {
      return new FileOutputStream(file);
    }
    catch (IOException e) {
      final File parentFile = file.getParentFile();
      if (parentFile != null && parentFile.mkdirs()) {
        return new FileOutputStream(file);
      }
      throw e;
    }
  }

  private File getEvalLicenseFile() {
    return new File(
      PathManager.getConfigPath() + File.separator + "eval" + File.separator + "plg_" + getProductDescriptor().getProductCode() + "_" + getCurrentEvaluationSelector() + ".evaluation.key"
    );
  }

  private boolean isCheating(final String encodedOriginalExpirationDate, final long expirationDate) {
    final String encodedTwoWeeksStamp = readPluginPropertyValue(getPluginTimestampRegKey());
    if (encodedOriginalExpirationDate != null || encodedTwoWeeksStamp != null) {
      boolean updateSavedData = false;
      long parsedExpirationDate;
      try {
        parsedExpirationDate = Long.parseLong(encodedOriginalExpirationDate != null ? encodedOriginalExpirationDate : "", 16);
      }
      catch (NumberFormatException e) {
        parsedExpirationDate = expirationDate; // Corrupted. Let's believe this is not a cheater.
        updateSavedData = true;
      }
      long twoWeeksStamp;
      try {
        twoWeeksStamp = Long.parseLong(encodedTwoWeeksStamp != null? encodedTwoWeeksStamp : "", 16);
      }
      catch (NumberFormatException e) {
        twoWeeksStamp = System.currentTimeMillis();
        updateSavedData = true;
      }
      if (updateSavedData) {
        encodePluginExpirationDate(parsedExpirationDate);
      }
      if (expirationDate != parsedExpirationDate || System.currentTimeMillis() - twoWeeksStamp < -DateFormatUtil.HOUR) {
        // Cheater. Original evaluation marker have either been deleted or modified or current timestamp once saved appears in the future.
        return true;
      }
    }
    return false;
  }

  private void encodePluginExpirationDate(final long expirationDate) {
    final long stamp = System.currentTimeMillis();
    try {
      Prefs.putLong(getPluginEvalExpirationRegKey(), ~expirationDate);
      Prefs.putLong(getPluginTimestampRegKey(), ~stamp);
    }
    catch (Throwable ignored) {
    }
  }

  private static String readPluginPropertyValue(String key) {
    try {
      long longValue = ~Prefs.getLong(key, 0L);
      return longValue > 0L ? Long.toHexString(longValue) : null;
    }
    catch (Throwable ignored) {
    }
    return null;
  }

  private String getPluginEvalExpirationRegKey() {
    return getPluginRegKey("evlsprt");
  }

  private String getPluginTimestampRegKey() {
    return getPluginRegKey("evlsprt2");
  }

  private String getPluginRegKey(String keyName) {
    final String evalSelector = getCurrentEvaluationSelector();
    return "JetBrains." + myProductDescriptor.getProductCode() + "." + (evalSelector.isEmpty()? "0" : evalSelector) + "." + keyName;
  }

}
