package com.intellij.ide.license.common;

import com.intellij.ide.BrowserUtil;
import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.license.LicenseFactory;
import com.intellij.ide.license.LicenseInfo;
import com.intellij.ide.license.LicenseMessages;
import com.intellij.ide.license.impl.*;
import com.intellij.idea.Main;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.util.text.DateFormatUtil;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;

import javax.swing.*;
import java.awt.*;
import java.lang.reflect.Proxy;
import java.util.List;
import java.util.*;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

@SuppressWarnings("Java9CollectionFactory")
public abstract class StandardEAPLicenseManager extends PluginsAwareLicenseManager {
  private LicenseFactory factory;
  private final String myProductName;
  private final int myExpirationPeriod;
  private final String myDownloadUrl;
  private final LicenseMessages myMessages;
  private static final @NonNls Set<String> EAP_ENABLED_PLUGINS = Set.of(
    // register here all JB plugins that should support EAP
    "DPN", // dotTrace
    "DC",   // dotCover
    "DPA", // Dynamic Performance Analysis
    "PDB", // Database plugin
    "PWS", // JavaScript Plugin
    "PGO", // Go Plugin
    "PPS", // PHP Plugin
    "PPC", // Python Pro Plugin
    "PRB", // Ruby Plugin
    "PSW", // Swift Plugin
    "PSI",  // SharedIndexes Plugin
    "PCWMP",  // CWM Plugin
    "PRDEV"  // CWM RD Plugin
  );

  protected StandardEAPLicenseManager(final @NonNls String productName, final int expirationPeriod, final @NonNls String downloadUrl, LicenseMessages messages) {
    super(new ExpirationDateCheck(false));
    myProductName = productName;
    myExpirationPeriod = expirationPeriod;
    myDownloadUrl = downloadUrl;
    myMessages = messages;
  }

  @NotNull
  @Override
  public LicenseInfo getPlatformLicenseInfo() {
    final LicenseFactory factory = getLicenseFactory();
    final EvaluationLicenseData license = new EvaluationLicenseData(
      ApplicationInfoImpl.getShadowInstance().getBuildDate().getTime(),
      factory.getCurrentMajorVersion(),
      licensedToMessage(),
      myExpirationPeriod
    );
    return new LicenseInfo(factory.getProductDescriptor(), license, false, false);
  }

  public Date getExpirationDate() {
    final LicenseData license = getPlatformLicenseInfo().getLicense();
    assert license != null;
    return license.getExpirationDate();
  }

  @Override
  public LicenseMessages getMessages() {
    return myMessages;
  }

  @Override
  public void startUp(@NotNull StartupAction action) {
    final Date now = new Date();
    if (now.before(getExpirationDate()) ||
        ((Calendar)ApplicationInfoImpl.getShadowInstance().getBuildDate().clone()).get(Calendar.YEAR) == 2 /* built from sources */) {
      try {
        // deliberately used reflection to obscure field initialization
        Class.forName("com.intellij.ultimate.UltimateVerifier")
          .getField("VERIFIER")
          .set(null, Proxy.newProxyInstance(null, new Class[]{Function.class}, new VerifyUltimate()));
      }
      catch (Throwable ignored) {
      }
      finally {
        super.startUp(action);
      }
    }
    else {
      String message = LicenseBundle.message("error.message.0.eap.build.expired", myProductName, myDownloadUrl);
      if (GraphicsEnvironment.isHeadless()) {
        //noinspection UseOfSystemOutOrSystemErr
        System.out.println(message);
      }
      else {
        int rc = JOptionPane.showOptionDialog(
          JOptionPane.getRootFrame(),
          message, LicenseBundle.message("title.0.eap.build.expired", myProductName),
          JOptionPane.DEFAULT_OPTION, JOptionPane.INFORMATION_MESSAGE, null,
          new Object[] {LicenseBundle.message("button.download"), LicenseBundle.message("button.shutdown") }, LicenseBundle.message("button.download")
        );
        if (rc == 0) {
          BrowserUtil.browse(myDownloadUrl);
        }
      }
      System.exit(Main.LICENSE_ERROR);
    }
  }

  @Override
  protected Collection<PluginProductKey> revalidatePluginLicenses(boolean reuseExistingLicensesIfPossible, final Predicate<? super LicenseData> licenseDataValidator) {
    // force EAP license for EAP plugins
    final Date buildTime = ApplicationInfoImpl.getShadowInstance().getBuildDate().getTime();
    for (PluginProductKey key : myPluginKeyData) {
      final LicenseFactory factory = key.getLicenseFactory();
      if (EAP_ENABLED_PLUGINS.contains(factory.getProductDescriptor().getProductCode())) {
        final BuiltInEvaluationSource evalSource = key.findSource(BuiltInEvaluationSource.class);
        if (evalSource != null) {
          final LicenseData lic = evalSource.getLicense();
          if (lic == null || !buildTime.equals(lic.getGenerationDate())) {
            evalSource.setLicense(new EvaluationLicenseData(
              buildTime, factory.getCurrentMajorVersion(), licensedToMessage(), myExpirationPeriod
            ));
            key.setActiveSource(evalSource, false);
          }
        }
      }
    }
    return super.revalidatePluginLicenses(reuseExistingLicensesIfPossible, licenseDataValidator);
  }

  @Override
  public String licensedToMessage() {
    return myProductName + " EAP User";
  }

  @Override
  public @NotNull Supplier<List<String>> licensedRestrictionsMessages() {
    return () -> Collections.singletonList(LicenseBundle.message("message.expiration.date", DateFormatUtil.formatAboutDialogDate(getExpirationDate())));
  }

  @Override
  public LicenseFactory getLicenseFactory() {
    if (factory == null) {
      factory = createLicenseFactory();
    }
    return factory;
  }

  protected abstract LicenseFactory createLicenseFactory();

  @Override
  public boolean needEvaluationFollowUp() {
    return false;
  }

  @Override
  public boolean needDetectCheater() {
    return false;
  }
}