package com.intellij.ide.license.common;

import com.intellij.ide.license.LicenseManager;
import com.intellij.ide.license.LicenseValidator;
import com.intellij.ide.license.ProductDescriptor;
import com.intellij.ide.license.impl.*;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.openapi.application.PathManager;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.util.text.StringUtil;
import org.jetbrains.annotations.NotNull;

import java.io.File;
import java.util.Date;

/**
 * @author yole
 */
public class StandardSubscriptionLicenseFactory extends LicenseFactoryImpl {
  @NotNull
  private final ProductDescriptor myProductDescriptor;
  private final int myCurrentMajorVersion;
  private final int myEvaluationPeriod;
  private final String myCurrentEvaluationSelector;
  private final boolean myForceReleaseEvaluationSelector;
  private String myEvaluatorName;

  public StandardSubscriptionLicenseFactory(ProductDescriptor pd, int evaluationPeriod, LicenseManager licenseManager) {
    this(pd,
         ApplicationInfoImpl.getShadowInstance().getBuild().getBaselineVersion(),
         evaluationPeriod,
         String.valueOf(ApplicationInfoImpl.getShadowInstance().getBuild().getBaselineVersion()),
         licenseManager
    );
  }

  public StandardSubscriptionLicenseFactory(ProductDescriptor pd,
                                            int baseLineVersion,
                                            int evaluationPeriod,
                                            String currentEvaluationSelector,
                                            LicenseManager licenseManager) {
    myProductDescriptor = pd;
    myCurrentMajorVersion = getInternalVersionNumber(baseLineVersion);
    myEvaluationPeriod = evaluationPeriod;
    myCurrentEvaluationSelector = currentEvaluationSelector;
    myForceReleaseEvaluationSelector = !(licenseManager instanceof StandardEAPLicenseManager);
  }

  public void setEvaluatorName(String evaluatorName) {
    myEvaluatorName = evaluatorName;
  }

  @Override
  public LicenseValidator getLicenseValidator() {
    return new GeneralLicenseValidator();
  }

  @Override
  public LicenseData createFromUser(String key, String userName) {
    try {
      return new CertKeyLicenseData(key, getProductCode(), getProductBuildDate());
    }
    catch (Throwable ignored) {
    }
    return null;
  }

  @Override
  @NotNull
  public ProductDescriptor getProductDescriptor() {
    return myProductDescriptor;
  }

  @NotNull
  public String getProductCode() {
    return getProductDescriptor().getProductCode();
  }

  public String getProductShortName() {
    return getProductDescriptor().getShortName();
  }

  @Override
  public int getCurrentMajorVersion() {
    return myCurrentMajorVersion;
  }

  @Override
  protected EvaluationLicenseData createEvaluationLicenseData(long date) {
    return new EvaluationLicenseData(new Date(date),
                                     getCurrentMajorVersion(),
                                     getEvaluatorName(),
                                     myEvaluationPeriod);
  }

  private String getEvaluatorName() {
    if (myEvaluatorName != null) {
      return myEvaluatorName;
    }
    return getProductShortName() + " Evaluator";
  }

  @Override
  public LicenseStore getLicenseStore() {
    final String fileName = StringUtil.toLowerCase(myProductDescriptor.getShortName());
    final String licensePath = PathManager.getConfigPath() + File.separatorChar + fileName + ".key";
    final String licensePathSystem = PathManager.getSystemPath() + File.separatorChar + fileName + ".license";
    return new StandardLicenseFile(getProductDescriptor(), licensePath, licensePathSystem, licensePath);
  }

  @Override
  public String getCurrentEvaluationSelector() {
    return myCurrentEvaluationSelector;
  }

  @Override
  public boolean isForceReleaseEvaluationSelector() {
    return myForceReleaseEvaluationSelector;
  }
}
