package com.intellij.ide.license.impl;

import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.Disposer;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.jetbrains.license.util.RpcUtil;
import com.jetbrains.ls.responses.HubAuthTokenResponse;
import com.jetbrains.ls.responses.ResponseCode;
import com.sun.net.httpserver.HttpExchange;
import com.sun.net.httpserver.HttpServer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.net.InetSocketAddress;
import java.net.MalformedURLException;
import java.net.URL;
import java.nio.charset.StandardCharsets;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.SecureRandom;
import java.util.*;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.function.BiConsumer;

public class AuthTokenManager implements Disposable{
  protected static final Logger LOG = Logger.getInstance("#LicenseManager");
  private static final int PORT_RANGE_START = 62345;
  private static final int PORT_RANGE_END = 62364;
  private static final BiConsumer<StringBuilder, String> PARAM_STATE = urlParam("state");
  private static final BiConsumer<StringBuilder, String> PARAM_CLIENT_ID = urlParam("client_id");
  private static final BiConsumer<StringBuilder, String> PARAM_CODE_CHALLENGE = urlParam("code_challenge");
  private static final BiConsumer<StringBuilder, String> PARAM_REDIRECT_URL = urlParam("redirect_url");
  private static final BiConsumer<StringBuilder, String> PARAM_CODE_CHALLENGE_METHOD = urlParam("code_challenge_method");

  private final AtomicBoolean myIsDisposed = new AtomicBoolean(false);
  private final Map<String, AuthData> myAuthData = Collections.synchronizedMap(new HashMap<>());
  private final Map<String, Token> myTokenData = Collections.synchronizedMap(new HashMap<>());

  private final AtomicReference<Future<?>> myRefreshTask = new AtomicReference<>(null);
  private volatile long myUpdatePeriodSec;

  private HttpServer myHttpServer;

  public AuthTokenManager() {
    myUpdatePeriodSec = calcTokenRefreshPeriodSec();
    LicenseEntryUtil.executeOnApplicationExit(() -> Disposer.dispose(this));
  }

  @NotNull
  public String registerAuthorizationAttempt(String serviceUrl, String authUrl, String exchangeUrl, String clientId) {
    final String serviceId = getServiceId(serviceUrl);
    
    final AuthData request;
    synchronized (myAuthData) {
      request = new AuthData(serviceId, exchangeUrl, clientId, ensureCallbackServerListening());
      myAuthData.put(request.state, request);
    }

    final StringBuilder auth = new StringBuilder(authUrl);
    if (!authUrl.contains("?")) {
      auth.append("?");
    }
    PARAM_CLIENT_ID.accept(auth, request.clientId);
    PARAM_STATE.accept(auth, request.state);
    PARAM_CODE_CHALLENGE_METHOD.accept(auth, "S256");
    PARAM_CODE_CHALLENGE.accept(auth, request.codeChallenge);
    PARAM_REDIRECT_URL.accept(auth, request.redirectUrl);
    return auth.toString();
  }

  public void clearAuthorizationAttempts() {
    synchronized (myAuthData) {
      myAuthData.clear();
      shutdownServer();
    }
  }

  private void clearAuthorizationAttempts(String serviceId) {
    synchronized (myAuthData) {
      for (Iterator<Map.Entry<String, AuthData>> it = myAuthData.entrySet().iterator(); it.hasNext(); ) {
        final Map.Entry<String, AuthData> entry = it.next();
        if (serviceId.equals(entry.getValue().serviceId)) {
          it.remove();
        }
      }
      if (myAuthData.isEmpty()) {
        shutdownServer();
      }
    }
  }

  // an already existing or refreshed access token
  public TokenFuture<Token> addTokenIfAbsent(String serviceUrl, Token token) {
    final String serviceId = getServiceId(serviceUrl);
    final Token currentToken = myTokenData.putIfAbsent(serviceId, token);
    if (currentToken == null) {
      // token has just been added
      return TokenFuture.create(AppExecutorUtil.getAppExecutorService().submit(() -> {
        try {
          return refreshToken(serviceId, token);
        }
        finally {
          rescheduleRefreshTask();
        }
      }), () -> null);
    }
    return TokenFuture.createCompleted(currentToken);
  }

  @Nullable
  public Token getToken(String serviceUrl) {
    return myTokenData.get(getServiceId(serviceUrl));
  }

  public void removeToken(String serviceUrl) {
    myTokenData.remove(getServiceId(serviceUrl));
  }

  public static String getServiceId(String serviceUrl) {
    try {
      final URL url = new URL(serviceUrl);
      final String host = url.getHost();
      final int port = url.getPort();
      return port < 0 || port == url.getDefaultPort()? host : host + ":" + port;
    }
    catch (MalformedURLException e) {
      return serviceUrl;
    }
  }

  // returns server host:port
  private String ensureCallbackServerListening() {
    if (myIsDisposed.get()) {
      return null;
    }
    HttpServer server = myHttpServer;
    if (server == null) {
      for (int port = PORT_RANGE_START; port < PORT_RANGE_END; port++) {
        try {
          server = HttpServer.create(new InetSocketAddress(port), 0);
          server.setExecutor(AppExecutorUtil.getAppExecutorService());
          server.createContext("/", this::handleHttpRequest);
          server.start();
          break;
        }
        catch (IOException ignored) {
          // try another port
        }
        catch (Throwable e) {
          LOG.info("Authorization ", e);
          break;
        }
      }
      if (server == null) {
        return null;
      }
      myHttpServer = server;
    }

    final InetSocketAddress addr = server.getAddress();
    return addr.getHostString() + ":" + addr.getPort();
  }
  
  private void shutdownServer() {
    final HttpServer server = myHttpServer;
    if (server != null) {
      myHttpServer = null;
      server.stop(0);
    }
  }

  private void handleHttpRequest(HttpExchange exchange) {
    try {
      if (!"GET".equalsIgnoreCase(exchange.getRequestMethod())) {
        reply(exchange, 405, ""); // Method Not Allowed
        return;
      }
      final QueryParams queryParams = new QueryParams(exchange.getRequestURI());
      final String code = queryParams.get("code");
      final String state = queryParams.get("state");
      if (code == null) {
        String msg = queryParams.getOrDefault("error", "Code is missing from reply");
        final String descr = queryParams.get("error_description");
        if (descr != null) {
          msg = msg + ": " + descr;
        }
        reply(exchange, 404, getFailurePage(msg));
      }
      else if (state != null) {
        final AuthData authData = myAuthData.get(state);
        if (authData != null && state.equals(authData.state)) {
          try {
            final HubAuthTokenResponse response = RpcUtil.obtainHubAuthentication(
              authData.exchangeUrl, code, authData.codeVerifier, authData.clientId, authData.redirectUrl
            );
            if (response.getResponseCode() == ResponseCode.OK) {
              final Token token = Token.fromJson(response.getData(), authData.exchangeUrl, authData.clientId);
              if (token != null) {
                myTokenData.put(authData.serviceId, token);
                reply(exchange, response.getHttpResponseCode(), getSuccessPage());
                rescheduleRefreshTask();
                //clearAuthorizationAttempts(authData.serviceId);
              }
              else {
                reply(exchange, 401, getFailurePage("Authorization data is incomplete."));
              }
            }
            else {
              LOG.info("Authorization code: " + code);
              reply(exchange, response.getHttpResponseCode(), getFailurePage(response.getMessage()));
            }
          }
          catch (IOException e) {
            reply(exchange, 404, getFailurePage(e.getMessage()));
          }
        }
        else {
          reply(exchange, 404, getFailurePage("Unknown state " + state + "."));
        }
      }
      else {
        reply(exchange, 404, getFailurePage("State parameter missing."));
      }
    }
    catch (Throwable e) {
      LOG.info("Authorization redirect error: ", e);
    }
    finally {
      exchange.close();
    }
  }

  private void rescheduleRefreshTask() {
    final long newPeriod = calcTokenRefreshPeriodSec();
    final boolean needReschedule = newPeriod < myUpdatePeriodSec || myRefreshTask.get() == null;
    myUpdatePeriodSec = newPeriod;
    if (needReschedule) {
      scheduleTokenRefreshTask();
    }
  }

  private void scheduleTokenRefreshTask() {
    final Future<?> prev = myRefreshTask.getAndSet(AppExecutorUtil.getAppScheduledExecutorService().schedule(() -> {
      try {
        for (Map.Entry<String, Token> entry : new HashMap<>(myTokenData).entrySet()) {
          refreshToken(entry.getKey(), entry.getValue());
        }
      }
      finally {
        scheduleTokenRefreshTask();
      }
    }, myUpdatePeriodSec, TimeUnit.SECONDS));
    
    if (prev != null) {
      prev.cancel(false);
    }
  }

  private long calcTokenRefreshPeriodSec() {
    synchronized (myTokenData) {
      long minPeriod;
      if (!myTokenData.isEmpty()) {
        minPeriod = Long.MAX_VALUE;
        for (Token token : myTokenData.values()) {
          minPeriod = Math.min(minPeriod, token.getExpiresInSec());
        }
      }
      else {
        minPeriod = Token.DEFAULT_EXPIRATION_PERIOD_SEC;
      }
      return Math.max(60, minPeriod / 2);
    }
  }

  private Token refreshToken(String serviceId, Token token) {
    try {
      final HubAuthTokenResponse response = RpcUtil.refreshHubAuthentication(token.getExchangeUrl(), token.getRefreshToken(), token.getClientId());
      if (response.getResponseCode() == ResponseCode.OK) {
        myTokenData.put(serviceId, token = Token.updateFromJson(token, response.getData()));
      }
    }
    catch (IOException ignored) {
    }
    return token;
  }

  private static void reply(HttpExchange exchange, int responseCode, String body) throws IOException {
    final byte[] bytes = body.getBytes(StandardCharsets.UTF_8);
    exchange.sendResponseHeaders(responseCode, bytes.length);
    exchange.getResponseBody().write(bytes);
  }

  @Override
  public void dispose() {
    if (!myIsDisposed.getAndSet(true)) {
      try {
        synchronized (myAuthData) {
          shutdownServer();
        }
        final Future<?> refreshTask = myRefreshTask.getAndSet(null);
        if (refreshTask != null) {
          refreshTask.cancel(true);
        }
      }
      catch (Throwable ignored) {
      }
    }
  }

  private static class AuthData {
    private static final Random ourRandom = new SecureRandom();
    final String serviceId;
    final String exchangeUrl;
    final String clientId;
    final String redirectUrl;
    final String state;
    final String codeVerifier;
    final String codeChallenge;

    AuthData(String serviceId, String exchangeUrl, String clientId, String redirectUrl) {
      this.serviceId = serviceId;
      this.exchangeUrl = exchangeUrl;
      this.clientId = clientId;
      this.redirectUrl = redirectUrl;
      try {
        state = UUID.randomUUID().toString();

        final byte[] bytes = new byte[32];
        ourRandom.nextBytes(bytes);
        final Base64.Encoder base64 = Base64.getUrlEncoder().withoutPadding();
        final byte[] verifierBytes = base64.encode(bytes);
        codeVerifier = new String(verifierBytes, StandardCharsets.UTF_8);
        codeChallenge = new String(base64.encode(MessageDigest.getInstance("SHA-256").digest(verifierBytes)), StandardCharsets.UTF_8);
      }
      catch (NoSuchAlgorithmException e) {
        throw new RuntimeException(e); // should not happen
      }
    }
  }

  private static BiConsumer<StringBuilder, String> urlParam(String name) {
    return (builder, value) -> {
      if (value != null) {
        final int len = builder.length();
        if (len > 0 && builder.charAt(len - 1) != '?') {
          builder.append("&");
        }
        builder.append(name).append("=").append(value);
      }
    };
  }
  
  private static String getSuccessPage() {
    final String title = LicenseBundle.message("message.fls.authorization.successful.title");
    final String text = LicenseBundle.message("message.fls.authorization.successful.text");

    return "<html>\n" +
      "<head>\n" +
      "  <title>" + title + "</title>\n" +
      "  <style>\n" +
      "    body {\n" +
      "      margin: 0;\n" +
      "    }\n" +
      "    .header {\n" +
      "      height: 72px;\n" +
      "      background-color: #27282c;\n" +
      "    }\n" +
      "    .container {\n" +
      "      max-width: 996px;\n" +
      "      margin: 0 auto;\n" +
      "      padding: 0 22px;\n" +
      "    }\n" +
      "    #jetbrains-simple {\n" +
      "      height: 100%;\n" +
      "    }\n" +
      "    .center {\n" +
      "      position: absolute;\n" +
      "      top: 50%;\n" +
      "      left: 50%;\n" +
      "      margin-right: -50%;\n" +
      "      transform: translate(-50%, -50%);\n" +
      "    }\n" +
      "    p {\n" +
      "      text-align: center;\n" +
      "      font-size: 150%;\n" +
      "    }\n" +
      "    h1 {\n" +
      "      text-align: center;\n" +
      "      font-size: 200%;\n" +
      "    }\n" +
      "  </style>\n" +
      "</head>\n" +
      "<body>\n" +
      "  <header class=\"header\">\n" +
      "    <div class=\"container\">\n" +
      "      <svg xmlns=\"http://www.w3.org/2000/svg\" viewBox=\"0 0 700 700\" id=\"jetbrains-simple\">\n" +
      "        <path d=\"m0 0h700v700h-700z\" fill=\"#000\"></path>\n" +
      "        <path d=\"m60.379 568.75h262.5v43.75h-262.5z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m57.428 184.315 20.372-19.232c5.542 6.682 10.758 10.594 17.929 10.594 7.823 0 12.877-5.378 12.877-15.972v-72.205h31.457v72.367c0 14.343-3.586 24.448-11.246 32.109-7.5 7.5-18.254 11.572-31.294 11.572-19.885 0-31.946-8.312-40.095-19.233z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m147.394 87.5h91.762v26.73h-60.468v17.44h54.763v24.937h-54.763v18.092h61.283v26.893h-92.577z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m280.491 115.208h-34.064v-27.708h99.911v27.708h-34.227v86.384h-31.62z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m139.736 282.7c10.106-4.4 17.6-12.224 17.6-25.426v-.326a25.675 25.675 0 0 0 -7.336-18.584c-6.682-6.52-16.788-10.106-31.131-10.106h-58.507v114.092h58.838c27.218 0 43.191-11.9 43.191-31.457v-.326c.001-15.484-8.8-23.307-22.655-27.867zm-48.57-29.011h20.7c9.29 0 14.343 3.422 14.343 9.779v.326c0 6.682-5.542 9.942-15.158 9.942h-19.885v-20.051zm39.607 52.808c0 6.682-5.379 10.431-15.158 10.431h-24.449v-21.028h24.123c10.594 0 15.484 4.075 15.484 10.269v.326z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m335.8 227.444h-30.475l-42.63 101.193-17.833-26.056c14.18-6.031 23.469-17.6 23.469-35.205v-.326c0-11.246-3.422-19.885-10.1-26.567-7.661-7.661-19.722-12.224-37.162-12.224h-53.953v114.091h31.619v-34.55h14.017l22.981 34.553h54.267l8.15-20.536h44.169l8.149 20.536h33.9zm-99.093 42.05c0 8.312-6.357 13.529-16.951 13.529h-21.02v-27.546h20.864c10.432 0 17.114 4.564 17.114 13.692v.325zm70.737 27.706 12.877-32.271 12.712 32.271z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m388.119 228.258h31.619v114.092h-31.619z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m427.56 228.258h29.501l46.94 60.306v-60.306h31.294v114.092h-27.545l-48.896-62.587v62.587h-31.294z\" fill=\"#fff\"></path>\n" +
      "        <path d=\"m537.277 325.4 17.6-21.025c11.409 8.964 23.8 13.691 37 13.691 8.638 0 13.2-2.934 13.2-7.824v-.325c0-4.89-3.749-7.335-19.4-11.084-24.286-5.541-43.03-12.387-43.03-35.694v-.326c0-21.188 16.788-36.509 44.17-36.509 19.4 0 34.553 5.216 46.94 15.158l-15.801 22.328c-10.431-7.5-21.84-11.246-31.946-11.246-7.66 0-11.409 3.1-11.409 7.334v.322c0 5.216 3.912 7.5 19.885 11.083 26.078 5.7 42.377 14.18 42.377 35.531v.326c0 23.307-18.418 37.161-46.126 37.161-20.211.005-39.28-6.351-53.46-18.901z\" fill=\"#fff\"></path>\n" +
      "      </svg>\n" +
      "    </div>\n" +
      "  </header>\n" +
      "  <div class=\"center\">\n" +
      "    <h1>" + title + "</h1>\n" +
      "    <p>"  + text + "</p>\n" +
      "    </div>\n" +
      "</body>\n" +
      "</html>";
  }

  private static String getFailurePage(String message) {
    LOG.info("Authorization failure: " + message);

    final String title = LicenseBundle.message("message.fls.authorization.failed.title");
    final String text = LicenseBundle.message("message.fls.authorization.failed.text");

    return "<html>\n" +
      "<head>\n" +
      "    <title>"+title+"</title>\n" +
      "    <style>\n" +
      "        body {\n" +
      "            margin: 0;\n" +
      "        }\n" +
      "\n" +
      "        .header {\n" +
      "            height: 72px;\n" +
      "            background-color: #27282c;\n" +
      "        }\n" +
      "\n" +
      "        .container {\n" +
      "            max-width: 996px;\n" +
      "            margin: 0 auto;\n" +
      "            padding: 0 22px;\n" +
      "        }\n" +
      "\n" +
      "        #jetbrains-simple {\n" +
      "            height: 100%;\n" +
      "        }\n" +
      "\n" +
      "        .center {\n" +
      "            position: absolute;\n" +
      "            top: 50%;\n" +
      "            left: 50%;\n" +
      "            margin-right: -50%;\n" +
      "            transform: translate(-50%, -50%);\n" +
      "        }\n" +
      "\n" +
      "        p {\n" +
      "            text-align: center;\n" +
      "            font-size: 150%;\n" +
      "        }\n" +
      "\n" +
      "        h1 {\n" +
      "            text-align: center;\n" +
      "            font-size: 200%;\n" +
      "        }\n" +
      "    </style>\n" +
      "</head>\n" +
      "<body>\n" +
      "    <header class=\"header\">\n" +
      "        <div class=\"container\">\n" +
      "            <svg xmlns=\"http://www.w3.org/2000/svg\" viewBox=\"0 0 700 700\" id=\"jetbrains-simple\">\n" +
      "                <path d=\"m0 0h700v700h-700z\" fill=\"#000\"></path>\n" +
      "                <path d=\"m60.379 568.75h262.5v43.75h-262.5z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m57.428 184.315 20.372-19.232c5.542 6.682 10.758 10.594 17.929 10.594 7.823 0 12.877-5.378 12.877-15.972v-72.205h31.457v72.367c0 14.343-3.586 24.448-11.246 32.109-7.5 7.5-18.254 11.572-31.294 11.572-19.885 0-31.946-8.312-40.095-19.233z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m147.394 87.5h91.762v26.73h-60.468v17.44h54.763v24.937h-54.763v18.092h61.283v26.893h-92.577z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m280.491 115.208h-34.064v-27.708h99.911v27.708h-34.227v86.384h-31.62z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m139.736 282.7c10.106-4.4 17.6-12.224 17.6-25.426v-.326a25.675 25.675 0 0 0 -7.336-18.584c-6.682-6.52-16.788-10.106-31.131-10.106h-58.507v114.092h58.838c27.218 0 43.191-11.9 43.191-31.457v-.326c.001-15.484-8.8-23.307-22.655-27.867zm-48.57-29.011h20.7c9.29 0 14.343 3.422 14.343 9.779v.326c0 6.682-5.542 9.942-15.158 9.942h-19.885v-20.051zm39.607 52.808c0 6.682-5.379 10.431-15.158 10.431h-24.449v-21.028h24.123c10.594 0 15.484 4.075 15.484 10.269v.326z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m335.8 227.444h-30.475l-42.63 101.193-17.833-26.056c14.18-6.031 23.469-17.6 23.469-35.205v-.326c0-11.246-3.422-19.885-10.1-26.567-7.661-7.661-19.722-12.224-37.162-12.224h-53.953v114.091h31.619v-34.55h14.017l22.981 34.553h54.267l8.15-20.536h44.169l8.149 20.536h33.9zm-99.093 42.05c0 8.312-6.357 13.529-16.951 13.529h-21.02v-27.546h20.864c10.432 0 17.114 4.564 17.114 13.692v.325zm70.737 27.706 12.877-32.271 12.712 32.271z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m388.119 228.258h31.619v114.092h-31.619z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m427.56 228.258h29.501l46.94 60.306v-60.306h31.294v114.092h-27.545l-48.896-62.587v62.587h-31.294z\" fill=\"#fff\"></path>\n" +
      "                <path d=\"m537.277 325.4 17.6-21.025c11.409 8.964 23.8 13.691 37 13.691 8.638 0 13.2-2.934 13.2-7.824v-.325c0-4.89-3.749-7.335-19.4-11.084-24.286-5.541-43.03-12.387-43.03-35.694v-.326c0-21.188 16.788-36.509 44.17-36.509 19.4 0 34.553 5.216 46.94 15.158l-15.801 22.328c-10.431-7.5-21.84-11.246-31.946-11.246-7.66 0-11.409 3.1-11.409 7.334v.322c0 5.216 3.912 7.5 19.885 11.083 26.078 5.7 42.377 14.18 42.377 35.531v.326c0 23.307-18.418 37.161-46.126 37.161-20.211.005-39.28-6.351-53.46-18.901z\" fill=\"#fff\"></path>\n" +
      "            </svg>\n" +
      "        </div>\n" +
      "    </header>\n" +
      "    <div class=\"center\">\n" +
      "        <h1>" + title + "</h1>\n" +
      "\t<p>" + message + "</p>\n" +
      "        <p>" + text + "</p>\n" +
      "    </div>\n" +
      "</body>\n" +
      "</html>";
  }
}
