package com.intellij.ide.license.impl;

import com.google.gson.*;
import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.license.LicenseCreationException;
import com.intellij.ide.misc.CrlResource;
import com.intellij.openapi.application.ApplicationNamesInfo;
import com.intellij.util.text.DateFormatUtil;
import com.jetbrains.license.newLicenses.CertLicenseDecoder;
import com.jetbrains.license.util.Base64Support;
import com.jetbrains.license.util.Ref;
import com.jetbrains.ls.Constants;
import com.jetbrains.ls.data.ProductData;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.lang.reflect.Array;
import java.lang.reflect.Method;
import java.lang.reflect.Type;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.sql.Timestamp;
import java.text.DateFormat;
import java.text.ParseException;
import java.text.SimpleDateFormat;
import java.util.*;
import java.util.concurrent.Callable;

/**
 * @author Eugene Zhuravlev
 * @noinspection TrivialFunctionalExpressionUsage, unchecked
 */
public final class CertKey {
  //private static final int AUTO_PROLONGATION_GRACE_PERIOD_DAYS = 3;

  private final Object mySignatureImpl; // deliberately declared Object to obscure references
  private final Object mySignatureBytes; // deliberately declared Object to obscure references
  private final Object myKeyData; // deliberately declared Object to obscure references to LicenseData
  private final Object myGson; // deliberately declared Object to obscure references
  private final Object myCrlResource; // deliberately declared Object to obscure references

  public CertKey(Object /*@NotNull String*/ key) throws LicenseCreationException{
    try {
      myGson = ((Callable)()->{
        final Class<?> builderClass = Class.forName("com.google.gson.GsonBuilder");
        final Method disableHtmlEscaping = builderClass.getMethod("disableHtmlEscaping");
        final Method registerTypeAdapter = builderClass.getMethod("registerTypeAdapter", Type.class, Object.class);
        final Method create = builderClass.getMethod("create");
        return new Ref(create.invoke(registerTypeAdapter.invoke(disableHtmlEscaping.invoke(builderClass.newInstance()), Date.class, new JetProfileDateAdapter())));
      }).call();

      myCrlResource = ((Callable)()-> new Ref(CrlResource.createDefault())).call();

      // expected key format: licenseId-licenseData-signatureBase64-certificatesBase64
      final Object licenseParts = ((Callable)()-> {
        final Class stringClass = key.getClass();
        // avoid direct use of '-' symbol in the code: existing cracks use this as a marker
        return stringClass.getDeclaredMethod("split", stringClass).invoke(key, stringClass.getDeclaredMethod("substring", int.class, int.class).invoke("a string with - in it", 14, 15));
      }).call();

      final Ref decodedData = ((Callable<Ref>)() -> {
        try {
          final Object arr = Array.newInstance(CharSequence.class, 3);
          Array.set(arr, 0, Array.get(licenseParts, 1));
          Array.set(arr, 1, Array.get(licenseParts, 2));
          Array.set(arr, 2, Array.get(licenseParts, 3));
          final Object keyToDecode = key.getClass().getDeclaredMethod("join", CharSequence.class, CharSequence[].class).invoke(
            null, key.getClass().getDeclaredMethod("substring", int.class, int.class).invoke("another string with - in it", 20, 21), arr
          );
          Object licenseStruct = ((Gson)Ref.deref((Ref)myGson)).fromJson((String)Ref.deref(
            (Ref)CertLicenseDecoder.create(((CrlResource)Ref.deref((Ref)myCrlResource)).getData()).decodeLicense(Ref.create(keyToDecode))
          ), Class.forName("com.jetbrains.ls.data.LicenseData"));
          //final String jsonLicenseId = ((LicenseData) licenseKeyData).getLicenseId();
          if (Array.get(licenseParts, 0).equals(licenseStruct.getClass().getDeclaredMethod("getLicenseId").invoke(licenseStruct))) {
            return Ref.create(licenseStruct);
          }
        }
        catch (Throwable ignored) {
        }
        return null;
      }).call();
      if (Ref.deref(decodedData) == null) {
        throw new Exception() {
          @Override
          public synchronized Throwable fillInStackTrace() {
            return this;
          }
        };
      }
      final Object signatureBase64 = Array.get(licenseParts, 2);
      mySignatureBytes = new Ref(Base64Support.decode((byte[])signatureBase64.getClass().getDeclaredMethod("getBytes", Charset.class).invoke(signatureBase64, StandardCharsets.UTF_8)));
      final Object sig = Class.forName("java.security.Signature").getDeclaredMethod("getInstance", signatureBase64.getClass()).invoke(null, "SHA1withRSA");
      final Object certBase64 = Array.get(licenseParts, 3);
      sig.getClass().getMethod("initVerify", Class.forName("java.security.cert.Certificate")).invoke(sig, Ref.deref((Ref)(
        (Callable)()-> CertLicenseDecoder.create(((CrlResource)Ref.deref((Ref)myCrlResource)).getData()).createCertificate(Base64Support.decode((byte[])certBase64.getClass().getDeclaredMethod("getBytes", Charset.class).invoke(certBase64, StandardCharsets.UTF_8)), Collections.emptySet(), false)
      ).call()));

      // check signature on serialized license data to make sure it matches the data encoded into the key
      //noinspection PrimitiveArrayArgumentToVarargsMethod
      sig.getClass().getMethod("update", byte[].class).invoke(
        sig, ((Gson)Ref.deref((Ref)myGson)).toJson(Ref.deref(decodedData), Ref.deref(decodedData).getClass()).getBytes(StandardCharsets.UTF_8)
      );
      if ((Boolean)sig.getClass().getMethod("verify", byte[].class).invoke(sig, Ref.deref((Ref)mySignatureBytes))) {
        mySignatureImpl = new Ref(sig);
        myKeyData = decodedData;
      }
      else {
        throw new Exception() {
          @Override
          public synchronized Throwable fillInStackTrace() {
            return this;
          }
        };
      }
    }
    catch (LicenseCreationException e) {
      throw e;
    }
    catch (Throwable e) {
      throw new LicenseCreationException(LicenseBundle.message("error.message.license.key.verification.error"), e.getMessage()) {
        @Override
        public synchronized Throwable fillInStackTrace() {
          return this;
        }
      };
    }
  }

  public synchronized String getLicenseId() {
    try {
      return (String)invokeGetter("getLicenseId");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized String getLicenseeName() {
    try {
      return (String)invokeGetter("getLicenseeName");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized String getAssigneeName() {
    try {
      return (String)invokeGetter("getAssigneeName");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized String getAssigneeEmail() {
    try {
      return (String)invokeGetter("getAssigneeEmail");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized String getLicenseRestriction() {
    try {
      return (String)invokeGetter("getLicenseRestriction");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized boolean isCheckConcurrentUse() {
    try {
      return Boolean.TRUE.equals(invokeGetter("getCheckConcurrentUse"));
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized boolean isAutoProlongated() {
    try {
      return Boolean.TRUE.equals(invokeGetter("getAutoProlongated"));
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized String getHash() {
    try {
      return (String)invokeGetter("getHash");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  public synchronized int getGracePeriodDays() {
    try {
      return (Integer)invokeGetter("getGracePeriodDays");
    }
    catch (RuntimeException e) {
      scheduleExit();
      throw e;
    }
  }

  @Nullable
  public Date getExpirationDate(@NotNull String productCode) {
    final Object pd = findProductData(productCode);
    return pd == null ? null : ((ProductData)pd).getPaidUpTo();
  }

  @Nullable
  public Date getFirstExpiredDate(@NotNull String productCode) {
    final Date expirationDate = getExpirationDate(productCode);
    return expirationDate == null ? null : plusDays(expirationDate, 1);
  }

  @Nullable
  public Date getPerpetualFallbackDate(@NotNull String productCode) {
    final Object pd = findProductData(productCode);
    return pd == null ? null : ((ProductData)pd).getFallbackDate();
  }

  public boolean isPerpetualForProduct(@NotNull String productCode, @NotNull Date productBuildDate) {
    final Object pd = findProductData(productCode);
    return pd != null && isPerpetual(pd, productBuildDate);
  }

  public boolean isApplicableTo(@NotNull String productCode) {
    return findProductData(productCode) != null;
  }

  public boolean isApplicableTo(@NotNull Date now, @NotNull String productCode, @NotNull Date productBuildDate) {
    final Object pd = findProductData(productCode);
    if (pd == null) {
      return false;
    }
    Date subEnd = ((ProductData)pd).getPaidUpTo();
    if (subEnd == null) {
      return true; // 'endless' subscription
    }

    subEnd = plusDays(subEnd, 1); // normalize the end date to point at the very end of subscription
    final int grace = getGracePeriodDays();

    if (now.before(subEnd)) {
      // preventing dates cheating ('now' in the past):
      // builds released after grace period end, are guaranteed not covered
      return productBuildDate.before(grace> 0? plusDays(subEnd, grace) : subEnd);
    }

    if (grace > 0) {
      final Date backdate = plusDays(now, -grace);
      if (backdate.before(subEnd)) {
        // preventing dates cheating ('now' in the past):
        // builds released after grace period end, are guaranteed not covered
        return productBuildDate.before(plusDays(subEnd, grace));
      }
    }

    // check perpetual fallback license case, if present
    if (isPerpetual(pd, productBuildDate)) {
      return true;
    }

    return false;
  }

  private static boolean isPerpetual(Object pd, @NotNull Date productBuildDate) {
    final Date fallback = ((ProductData)pd).getFallbackDate();
    return fallback != null && productBuildDate.before(plusDays(fallback, 1));
  }

  @Nullable
  public String getApplicableWarningMessage(@NotNull Date now, @NotNull String productCode, @NotNull Date productBuildDate) {
    final Object pd = findProductData(productCode);
    if (pd == null) {
      return null;
    }

    if (isPerpetual(pd, productBuildDate)) {
      return null;
    }

    final Date subEnd = ((ProductData)pd).getPaidUpTo(); // aligned on the beginning of the last subscription date
    if (subEnd == null) {
      return null;
    }

    final Date subEndExclusive = plusDays(subEnd, 1);
    if (now.before(subEndExclusive)) {
      // subscription is still valid, but some warnings may apply
      if (isAutoProlongated()) {
        return null;
      }
      final Date weekBackdate = plusDays(subEnd, -7);
      if (now.before(weekBackdate)) {
        return null;
      }
      return "Your " + ApplicationNamesInfo.getInstance().getFullProductName() + " subscription expires on " +
             DateFormatUtil.formatDate(subEnd) + ".\nAfter this date you will no longer be able to use the product";
    }

    // subscription is over
    final int grace = getGracePeriodDays();
    if (grace > 0) {
      final Date backdate = plusDays(now, -grace);
      if (backdate.before(subEndExclusive)) {
        // current date falls within the grace period
        final Date graceEnd = plusDays(subEnd, grace); // end of grace period, inclusively
        if (isAutoProlongated()) {
          // notify only after AUTO_PROLONGATION_GRACE_PERIOD_DAYS passes, or at least AUTO_PROLONGATION_GRACE_PERIOD_DAYS before grace period ends
          // if grace <= AUTO_PROLONGATION_GRACE_PERIOD_DAYS, notify immediately
          final int prolongationGrace = Math.max(0, Math.min(3/*AUTO_PROLONGATION_GRACE_PERIOD_DAYS*/, grace - 3/*AUTO_PROLONGATION_GRACE_PERIOD_DAYS*/));
          final Date autoProlongationGraceEnd = plusDays(subEndExclusive, prolongationGrace);
          if (!now.before(autoProlongationGraceEnd)) {
            return "Something went wrong with license auto-prolongation.\n" +
                   "Please make sure the contract is prolonged until " + DateFormatUtil.formatDate(graceEnd) + ",\n" +
                   "otherwise usage of " + ApplicationNamesInfo.getInstance().getFullProductName() + " will no longer be possible.";
          }
        }
        else {
          return "Your " + ApplicationNamesInfo.getInstance().getFullProductName() + " subscription expired on " + DateFormatUtil.formatDate(subEnd) + ".\n" +
                 "If you don't renew it before " + DateFormatUtil.formatDate(graceEnd) + ", you will no longer be able to use the product";
        }
      }
    }

    return null;
  }

  @NotNull
  private static Date plusDays(@NotNull Date date, int dayCount) {
    final Calendar calendar = new GregorianCalendar();
    calendar.setTime(date);
    calendar.add(Calendar.DAY_OF_MONTH, dayCount);
    return calendar.getTime();
  }

  @Nullable
  private synchronized Object findProductData(@NotNull String productCode) {
    try {
      final List products = (List)invokeGetter("getProducts");
      if (products != null) {
        for (Object pd : products) {
          if (productCode.equals(((ProductData)pd).getCode())) {
            return pd;
          }
        }
      }
    }
    catch (RuntimeException e) {
      scheduleExit();
    }
    return null;
  }

  private static final class JetProfileDateAdapter implements JsonSerializer<Date>, JsonDeserializer<Date> {
    private static DateFormat getFormat() {
      return new SimpleDateFormat(Constants.DATE_FORMAT_STRING, Locale.US);
    }

    @Override
    public JsonElement serialize(Date src, Type typeOfSrc, JsonSerializationContext context) {
      String dateFormatAsString = getFormat().format(src);
      return new JsonPrimitive(dateFormatAsString);
    }

    @Override
    public Date deserialize(JsonElement json, Type typeOfT, JsonDeserializationContext context)
      throws JsonParseException {
      if (!(json instanceof JsonPrimitive)) {
        throw new JsonParseException("The date should be a string value");
      }
      Date result;
      try {
        result = getFormat().parse(json.getAsString());
      }
      catch (ParseException e) {
        throw new JsonSyntaxException(json.getAsString(), e);
      }
      if (typeOfT == Date.class) {
        return result;
      }
      else if (typeOfT == Timestamp.class) {
        return new Timestamp(result.getTime());
      }
      else if (typeOfT == java.sql.Date.class) {
        return new java.sql.Date(result.getTime());
      }
      else {
        throw new IllegalArgumentException(getClass() + " cannot deserialize to " + typeOfT);
      }
    }
  }

  private Object invokeGetter(Object getterName) {
    try {
      final Object sig = Ref.deref((Ref)mySignatureImpl);
      final Object keyData = Ref.deref((Ref)myKeyData);
      //noinspection PrimitiveArrayArgumentToVarargsMethod
      sig.getClass().getMethod("update", byte[].class).invoke(sig, ((Gson)Ref.deref((Ref)myGson)).toJson(keyData, keyData.getClass()).getBytes(StandardCharsets.UTF_8));
      if ((Boolean)sig.getClass().getMethod("verify", byte[].class).invoke(sig, Ref.deref((Ref)mySignatureBytes))) {
        return keyData.getClass().getDeclaredMethod((String)getterName).invoke(keyData);
      }
    }
    catch (Throwable ignored) {
    }
    throw new RuntimeException() {
      @Override
      public synchronized Throwable fillInStackTrace() {
        return this;
      }
    };
  }

  private static void scheduleExit() {
    new Thread(()->{
      try {
        Thread.sleep(Math.max(100, new Random().nextInt(10000)));
        System.exit(-1);
      }
      catch (Throwable ignored) {
      }
    }).start();
  }
}
