package com.intellij.ide.license.impl;

import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.misc.CrlResource;
import com.intellij.openapi.util.NlsContexts;
import com.jetbrains.license.newLicenses.CertLicenseDecoder;
import com.jetbrains.license.util.*;
import com.jetbrains.license.util.org.bouncycastle.asn1.ASN1ObjectIdentifier;
import com.jetbrains.license.util.org.bouncycastle.crypto.Digest;
import com.jetbrains.license.util.org.bouncycastle.crypto.params.RSAKeyParameters;
import com.jetbrains.ls.Constants;
import org.jetbrains.annotations.NotNull;

import java.lang.invoke.MethodHandles;
import java.lang.invoke.MethodType;
import java.lang.reflect.Array;
import java.lang.reflect.Proxy;
import java.nio.charset.StandardCharsets;
import java.security.Signature;
import java.security.cert.Certificate;
import java.security.interfaces.RSAPublicKey;
import java.text.SimpleDateFormat;
import java.time.temporal.ChronoUnit;
import java.util.*;
import java.util.function.Function;
import java.util.function.Predicate;

/**
 * @author Eugene Zhuravlev
 * @noinspection unchecked, RedundantClassCall
 *
 * Important! do not reference ClientUtil from this class to avoid possibly slow static initialization
 */
@SuppressWarnings("override")
public final class CertKey2 {
  // expected root certificate fingerprints  to detect JCA classes tampering
  private static final String ROOT_CERTIFICATE_FINGERPRINT = "75C73E17ED54BCBA105BDA0B63AEE6A372263B67440EFEADD1A3112E6BB76D431927973973C10058E199EF97A4F34F498962DDD64E4872065A5DC71220A19A01";
  private static final String ROOT_PUBLIC_KEY_FINGERPRINT = "0178A57CF9E50368EC9C1708B8F8FE48ECF8C81BE754AB5B45C6B0E5A3CC706BC8A3ECB506B6E20E585578BCE938E71CA7A8282A8DE3F366CBAC405FE6D706EF";
  private static final int AUTO_PROLONGATION_GRACE_PERIOD_DAYS = 3;
  private final Object myRawKey; // a ref to a raw key

  public CertKey2(Object /*@NotNull String*/ key) {
    myRawKey = new Ref(key);
  }

  public Object eval(Object param) {
    try {
      // expected key format: licenseId-licenseData-signatureBase64-certificatesBase64
      final Object key = Ref.deref((Ref)myRawKey);
      if (param instanceof GetRawKey) {
        return key;
      }
      final Object licenseParts = ((Function)Proxy.newProxyInstance(null, new Class[]{Function.class}, SplitCertKey.INSTANCE)).apply(key);
      final byte[] jsonBytes = Base64Support.decode(String.valueOf(Array.get(licenseParts, 1)).getBytes(StandardCharsets.UTF_8));
      final byte[] signatureBytes = Base64Support.decode(String.valueOf(Array.get(licenseParts, 2)).getBytes(StandardCharsets.UTF_8));

      final CertLicenseDecoder certDecoder = CertLicenseDecoder.create(CrlResource.createDefault().getData());
      final Certificate cert = Certificate.class.cast(Ref.deref(Ref.class.cast(certDecoder.createCertificate(
        Base64Support.decode(String.valueOf(Array.get(licenseParts, 3)).getBytes(StandardCharsets.UTF_8)), Collections.emptySet(), false
      ))));
      final boolean singatureCheckPassed;
      final Pair<Digest, ASN1ObjectIdentifier> digest = VerifyingSignatureFactory.createBCDigest("SHA1withRSA");
      if (digest != null && cert.getPublicKey() instanceof RSAPublicKey) {
        final RSAPublicKey pubKey = RSAPublicKey.class.cast(cert.getPublicKey());
        RSAKeyParameters params = new RSAKeyParameters(false, pubKey.getModulus(), pubKey.getPublicExponent());
        singatureCheckPassed = VerifyingSignatureFactory.checkSignatureBC(digest, params, jsonBytes, signatureBytes);
      }
      else {
        final Signature sig = Signature.getInstance("SHA1withRSA");
        sig.initVerify(cert);
        sig.update(jsonBytes);
        singatureCheckPassed = sig.verify(signatureBytes);
      }
      if (singatureCheckPassed) {
        final MethodHandles.Lookup lookup = MethodHandles.lookup();
        final Class<?> builderClass = Class.forName("com.google.gson.GsonBuilder");
        final Object gson = lookup.findVirtual(builderClass, "create", MethodType.methodType(Class.forName("com.google.gson.Gson"))).invoke(
          lookup.findVirtual(builderClass, "disableHtmlEscaping", MethodType.methodType(builderClass)).invoke(
            lookup.findConstructor(builderClass, MethodType.methodType(void.class)).invoke()
          )
        );
        //final Map keyData = gson.fromJson(new String(jsonBytes, StandardCharsets.UTF_8), Map.class);
        final Object keyData = lookup.findVirtual(gson.getClass(), "fromJson", MethodType.methodType(Object.class, String.class, Class.class)).invoke(gson, new String(jsonBytes, StandardCharsets.UTF_8), Map.class);
        final Function keyDataGetter = mapDataGetter(keyData);
        // Additional check for the case when JCA classes were tampered.
        // It is expected the certificate fingerprint on tampered JCA classes will be different
        boolean fingerprintCheckOk = true;
        try {
          final Certificate rootCert = Certificate.class.cast(Ref.deref(Ref.class.cast(certDecoder.getRootCertificate())));
          fingerprintCheckOk = Arrays.equals(decodeHex(ROOT_CERTIFICATE_FINGERPRINT), Blake2b.digest(rootCert.getEncoded()));
          if (fingerprintCheckOk) {
            // check public key fingerprint
            fingerprintCheckOk = Arrays.equals(decodeHex(ROOT_PUBLIC_KEY_FINGERPRINT), Blake2b.digest(rootCert.getPublicKey().getEncoded()));
          }
        }
        catch (Throwable ignored) {
        }
        if (fingerprintCheckOk && Array.get(licenseParts, 0).equals(keyDataGetter.apply("licenseId"))) {
          Object rv = null;
          if (param instanceof IsValidKey) {
            rv = Boolean.TRUE;
          }
          else if (param instanceof GetLicenseId) {
            rv = keyDataGetter.apply("licenseId");
          }
          else if (param instanceof GetLicenseeName) {
            rv = keyDataGetter.apply("licenseeName");
          }
          else if (param instanceof GetAssigneeName) {
            rv = keyDataGetter.apply("assigneeName");
          }
          else if (param instanceof GetAssigneeEmail) {
            rv = keyDataGetter.apply("assigneeEmail");
          }
          else if (param instanceof GetLicenseRestriction) {
            rv = keyDataGetter.apply("licenseRestriction");
          }
          else if (param instanceof IsCheckConcurrentUse) {
            rv = Boolean.valueOf(String.valueOf(keyDataGetter.apply("checkConcurrentUse")));
          }
          else {
            final boolean isAutoProlongated = Boolean.parseBoolean(String.valueOf(keyDataGetter.apply("isAutoProlongated")));
            if (param instanceof IsAutoProlongated) {
              rv = isAutoProlongated;
            }
            else if (param instanceof GetHash) {
              rv = keyDataGetter.apply("hash");
            }
            else if (param instanceof GetMetadata) {
              rv = keyDataGetter.apply("metadata");
            }
            else {
              final Object days = keyDataGetter.apply("gracePeriodDays");
              final int grace = days == null ? 0 : (int)Float.parseFloat(String.valueOf(days));
              if (param instanceof GetGracePeriodDays) {
                rv = grace;
              }
              else if (param instanceof GetExpirationDate) {
                rv = getDate(findProductData(keyDataGetter, ((GetExpirationDate)param).getProductCode()), "paidUpTo");
              }
              else if (param instanceof GetFirstExpiredDate) {
                final Object expirationDate = getDate(findProductData(keyDataGetter, ((GetFirstExpiredDate)param).getProductCode()), "paidUpTo");
                rv = expirationDate == null ? null : plusDays(expirationDate, 1);
              }
              else if (param instanceof GetPerpetualFallbackDate) {
                rv = getDate(findProductData(keyDataGetter, ((GetPerpetualFallbackDate)param).getProductCode()), "fallbackDate");
              }
              else if (param instanceof IsPerpetualForProduct) {
                final IsPerpetualForProduct p = (IsPerpetualForProduct)param;
                rv = isPerpetual(findProductData(keyDataGetter, p.getProductCode()), p.getProductBuildDate());
              }
              else if (param instanceof GetProductProperty) {
                final GetProductProperty p = (GetProductProperty)param;
                final Function pd = findProductData(keyDataGetter, p.getProductCode());
                if (pd != null) {
                  rv = mapDataGetter(pd.apply("properties")).apply(p.getPropertyName());
                }
              }
              else if (param instanceof IsApplicableTo) {
                rv = findProductData(keyDataGetter, ((IsApplicableTo)param).getProductCode()) != null;
              }
              else if (param instanceof IsApplicableTo2) {
                final IsApplicableTo2 p = (IsApplicableTo2)param;
                final Function pd = findProductData(keyDataGetter, p.getProductCode());
                if (pd == null) {
                  return false;
                }
                Object subEnd = getDate(pd, "paidUpTo");
                if (subEnd == null) {
                  return true; // 'endless' subscription
                }
                subEnd = plusDays(subEnd, 1); // normalize the end date to point at the very end of subscription
                if (before(new Date()).test(subEnd)) {
                  // preventing dates cheating ('now' in the past):
                  // builds released after grace period end, are guaranteed not covered
                  return before(p.getProductBuildDate()).test(grace> 0? plusDays(subEnd, grace) : subEnd);
                }

                if (grace > 0) {
                  final Object backdate = plusDays(new Date(), -grace);
                  if (before(backdate).test(subEnd)) {
                    // preventing dates cheating ('now' in the past):
                    // builds released after grace period end, are guaranteed not covered
                    return before(p.getProductBuildDate()).test(plusDays(subEnd, grace));
                  }
                }
                // check perpetual fallback license case, if present
                return isPerpetual(pd, p.getProductBuildDate());
              }
              else if (param instanceof GetApplicableWarningMessage) {
                final GetApplicableWarningMessage p = (GetApplicableWarningMessage)param;
                final Function pd = findProductData(keyDataGetter, p.getProductCode());
                if (pd == null) {
                  return null;
                }
                if (isPerpetual(pd, p.getProductBuildDate())) {
                  return null;
                }
                final Object subEnd = getDate(pd, "paidUpTo"); // aligned on the beginning of the last subscription date
                if (subEnd == null) {
                  return null;
                }
                final Object subEndExclusive = plusDays(subEnd, 1);
                final Date now = new Date();
                final Predicate currentDateBefore = before(now);
                if (currentDateBefore.test(subEndExclusive)) {
                  // subscription is still valid, but some warnings may apply
                  if (isAutoProlongated) {
                    return null;
                  }
                  final Object backdate14 = plusDays(subEnd, -14);
                  if (currentDateBefore.test(backdate14)) {
                    return null;
                  }
                  final Object backdate12 = plusDays(subEnd, -12);
                  if (currentDateBefore.test(backdate12)) {
                    return createExpirationNotification(LicenseNotification.Type.INFO, periodDays(now, subEndExclusive),
                                                        LicenseBundle.message("notification.content.subscription.expires.on", p.getProductName(),
                                                                              Function.class.cast(Proxy.newProxyInstance(null, new Class[]{Function.class}, FormatDate.INSTANCE)).apply(subEnd))
                    );
                  }
                  final Object weekBackdate = plusDays(subEnd, -7);
                  if (currentDateBefore.test(weekBackdate)) {
                    return null;
                  }
                  return createExpirationNotification(LicenseNotification.Type.WARN, periodDays(now, subEndExclusive),
                                                      LicenseBundle.message("notification.content.subscription.expires.on.subscription.renewal.will.be.required",
                                                                            p.getProductName(),
                                                                            Function.class.cast(Proxy.newProxyInstance(null, new Class[]{Function.class}, FormatDate.INSTANCE)).apply(subEnd))
                  );
                }

                // subscription is over
                if (grace > 0) {
                  final Object backdate = plusDays(now, -grace);
                  if (before(backdate).test(subEndExclusive)) {
                    // current date falls within the grace period
                    final Object graceEnd = plusDays(subEnd, grace); // end of grace period, inclusively
                    if (isAutoProlongated) {
                      // notify only after AUTO_PROLONGATION_GRACE_PERIOD_DAYS passes, or at least AUTO_PROLONGATION_GRACE_PERIOD_DAYS before grace period ends
                      // if grace <= AUTO_PROLONGATION_GRACE_PERIOD_DAYS, notify immediately
                      final int prolongationGrace = Math.max(0, Math.min(AUTO_PROLONGATION_GRACE_PERIOD_DAYS, grace - AUTO_PROLONGATION_GRACE_PERIOD_DAYS));
                      if (!currentDateBefore.test(plusDays(subEndExclusive, prolongationGrace) /*auto prolongation grace end*/)) {
                        return createLicenseExpiredWarning(
                          LicenseBundle.message("notification.content.we.were.unable.to.autorenew.your.license", p.getProductName(),
                                                ((Function)Proxy.newProxyInstance(null, new Class[]{Function.class}, FormatDate.INSTANCE))
                                                  .apply(graceEnd))
                        );
                      }
                    }
                    else {
                      return createLicenseExpiredWarning(
                        LicenseBundle.message("notification.content.your.subscription.expired", p.getProductName(),
                                              ((Function)Proxy.newProxyInstance(null, new Class[]{Function.class}, FormatDate.INSTANCE))
                                                .apply(subEnd),
                                              ((Function)Proxy.newProxyInstance(null, new Class[]{Function.class}, FormatDate.INSTANCE))
                                                .apply(graceEnd))
                      );
                    }
                  }
                }
                return null;
              }
            }
          }
          return rv;
        }
      }
    }
    catch (Throwable ignored) {
    }
    return null;
  }

  private static byte[] decodeHex(CharSequence seq) {
    final byte[] sig = new byte[seq.length() / 2];
    int charIndex = 0;
    for (int i = 0; i < sig.length; i++) {
      final int high = Character.digit(seq.charAt(charIndex++), 16);
      final int low = Character.digit(seq.charAt(charIndex++), 16);
      sig[i] = (byte)((high << 4) + low);
    }
    return sig;
  }

  private static boolean isPerpetual(Function pd, @NotNull Object productBuildDate) {
    try {
      final Object fallback = getDate(pd, "fallbackDate");
      return fallback != null && before(productBuildDate).test(plusDays(fallback, 1));
    }
    catch (Throwable e) {
      return false;
    }
  }

  private static Object getDate(Function pd, String attributeName){
    try {
      final Object attributeValue = pd == null || attributeName == null? null : pd.apply(attributeName);
      return attributeValue == null? null : new SimpleDateFormat(Constants.DATE_FORMAT_STRING, Locale.US).parse(String.valueOf(attributeValue));
    }
    catch (Throwable e) {
      throw new RuntimeException(e);
    }
  }

  private static LicenseNotification createExpirationNotification(LicenseNotification.Type type, int expiresInDays, @NlsContexts.NotificationContent String text) {
    return new LicenseNotification() {
      public Type getType() {
        return type;
      }

      public String getTitle() {
        return expiresInDays >= 0 ? LicenseBundle.message("notification.title.license.expires.in.days", expiresInDays)
                                  : LicenseBundle.message("notification.title.license.expired");
      }

      public String getMessage() {
        return text;
      }

      public Collection<Action> getActions() {
        return Arrays.asList(Action.RENEW, Action.FALLBACK_OPTIONS);
      }
    };
  }

  private static LicenseNotification createLicenseExpiredWarning(@NlsContexts.NotificationContent String text) {
    return new LicenseNotification() {
      public Type getType() {
        return Type.WARN;
      }

      public String getTitle() {
        return LicenseBundle.message("notification.title.license.expired");
      }

      public String getMessage() {
        return text;
      }

      public Collection<Action> getActions() {
        return Arrays.asList(Action.RENEW, Action.FALLBACK_OPTIONS);
      }
    };
  }

  /** @noinspection JavaReflectionInvocation*/
  private static Predicate before(Object date) {
    return Predicate.class.cast(Proxy.newProxyInstance(
      null, new Class[]{Predicate.class}, (proxy, method, args) -> Date.class.getMethod("before", Date.class).invoke(date, args[0])
    ));
  }

  private static Object plusDays(Object date, int dayCount) {
    final Calendar calendar = new GregorianCalendar();
    calendar.setTime(Date.class.cast(date));
    calendar.add(Calendar.DAY_OF_MONTH, dayCount);
    return calendar.getTime();
  }

  private static int periodDays(Object date1, Object date2) {
    return (int)ChronoUnit.DAYS.between(Date.class.cast(date1).toInstant(), Date.class.cast(date2).toInstant());
  }

  private static Function findProductData(final Function getter, Object productCode) {
    final Iterable products = Iterable.class.cast(getter.apply("products"));
    if (products != null) {
      for (final Object pd : products) {
        if (productCode.equals((Map.class.cast(pd)).get("code"))) {
          return mapDataGetter(pd);
        }
      }
    }
    return null;
  }

  private static Function mapDataGetter(Object mapObject) {
    return Function.class.cast(Proxy.newProxyInstance(null, new Class[]{Function.class}, (proxy, method, args) -> mapObject != null? (Map.class.cast(mapObject)).get(args[0]) : null));
  }
}
