package com.intellij.ide.license.impl;

import com.intellij.ide.BrowserUtil;
import com.intellij.ide.license.LicenseInfo;
import com.intellij.ide.license.LicenseManager;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.notification.NotificationAction;
import com.intellij.openapi.application.ApplicationNamesInfo;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.text.DateFormatUtil;
import org.jetbrains.annotations.NotNull;

import java.util.Date;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;

/**
 * @author Eugene Zhuravlev
 */
@SuppressWarnings("RedundantClassCall")
public final class ExpirationDateCheck {
  public static final long DEFAULT_CHECK_PERIOD = DateFormatUtil.DAY;
  public static final long FREE_RUN_PERIOD = 30 * DateFormatUtil.MINUTE;

  private final boolean myCheckPlatformLicense;
  private final AtomicBoolean myValidationInProgress = new AtomicBoolean(false);
  private final Object myLock = new Object();
  private final Map<String, String> myLastCheckedLicenseId = new HashMap<>(); // productCode -> licenseId
  private final Map<String, Date> myFreeRunStart = new HashMap<>();   // productCode -> freeRunStart
  private final AtomicReference<ScheduledFuture<?>> myPendingTask = new AtomicReference<>(null);

  public ExpirationDateCheck(boolean checkPlatformLicense) {
    myCheckPlatformLicense = checkPlatformLicense;
  }

  public void scheduleValidation(long delay, final PluginsAwareLicenseManager licenseManager) {
    final ScheduledFuture<?> previous = myPendingTask.getAndSet(AppExecutorUtil.getAppScheduledExecutorService().schedule(asExclusiveTask(()->{
      long nextCheckDelay = DEFAULT_CHECK_PERIOD;
      try {
        if (myCheckPlatformLicense) {
          nextCheckDelay = Math.min(nextCheckDelay, checkLicense(licenseManager, licenseManager.getPlatformLicenseInfo()));
        }
        for (LicenseInfo info : licenseManager.getPluginLicenseInfo()) {
          if (!info.getProductDescriptor().isPluginDisabled() && !info.isOptional()) {
            nextCheckDelay = Math.min(nextCheckDelay, checkLicense(licenseManager, info));
          }
        }
      }
      finally {
        scheduleValidation(nextCheckDelay, licenseManager);
      }
    }), delay, TimeUnit.MILLISECONDS));

    if (previous != null) {
      previous.cancel(false);
    }
  }

  private Runnable asExclusiveTask(Runnable r) {
    return () -> {
      if (!myValidationInProgress.getAndSet(true)) {
        try {
          r.run();
        }
        finally {
          myValidationInProgress.set(false);
        }
      }
    };
  }

  public boolean isFreeRunActive(String productCode, @NotNull LicenseData license) {
    String id = getLicenseId(license);
    synchronized (myLock) {
      return id.equals(myLastCheckedLicenseId.get(productCode)) && myFreeRunStart.containsKey(productCode);
    }
  }

  public void startFreeRun(String productCode, @NotNull LicenseData license, @NotNull Date freeRunStart) {
    synchronized (myLock) {
      myLastCheckedLicenseId.put(productCode, getLicenseId(license));
      myFreeRunStart.put(productCode, freeRunStart);
    }
  }

  private long checkLicense(PluginsAwareLicenseManager licenseManager, LicenseInfo info) {
    final LicenseData license = info.getLicense();
    final String productCode = info.getProductDescriptor().getProductCode();
    if (license == null) {
      synchronized (myLock) {
        myLastCheckedLicenseId.remove(productCode);
        myFreeRunStart.remove(productCode);
      }
      return 15 * DateFormatUtil.MINUTE;
    }

    long nextCheckDelay = DEFAULT_CHECK_PERIOD;
    final String currentLicenseId = getLicenseId(license);
    final Date now = new Date();
    final boolean isExpired = license.isExpired();
    final Date freeRunStart;
    boolean isFreeRun;
    synchronized (myLock) {
      freeRunStart = myFreeRunStart.get(productCode);
      isFreeRun = isExpired && currentLicenseId.equals(myLastCheckedLicenseId.get(productCode)) && freeRunStart != null;
    }
    if (isFreeRun) {
      // if checking the same license, it is still expired, and free run period has already been started
      final long elapsed = Math.abs(now.getTime() - freeRunStart.getTime());
      if (elapsed >= FREE_RUN_PERIOD) {
        if (LicenseEntryUtil.invokeAndWaitInNonModalContext(() -> {
          licenseManager.getMessages().showLicenseExpiredBeforeShutdownNotification(info.getProductDescriptor(), info.isPlugin(), license instanceof EvaluationLicenseData);
        })) {
          if (info.isPlugin()) {
            licenseManager.disablePlugin(info.getProductDescriptor());
          }
          else {
            licenseManager.shutdownApplication();
          }
        }
        else {
          // was not able to inform the user because another modal dialog was showing
          nextCheckDelay = 2 * DateFormatUtil.MINUTE;
        }
      }
      else {
        nextCheckDelay = FREE_RUN_PERIOD - elapsed;
      }
    }
    else {
      synchronized (myLock) {
        // in non-freeRun-mode perform routine expiration check
        myFreeRunStart.remove(productCode); // reset shutdown flag
        // this license was never checked
        myLastCheckedLicenseId.put(productCode, currentLicenseId);
      }
      if (isExpired) {
        // show expiration warning
        if (LicenseEntryUtil.invokeAndWaitInNonModalContext(() -> {
          final boolean register = licenseManager.getMessages().askEnterKeyBecauseLicenseExpired(info.getProductDescriptor(), info.isPlugin(), license instanceof EvaluationLicenseData);
          if (register) {
            licenseManager.demandLicense(info.getProductDescriptor(), LicenseManager.StartupAction.EMPTY, null);
          }
        })) {
          // start the free run even if the user has chosen to register:
          // in case the license was not updated, we'll shut down after the free run.
          // in case if the license has been updated to a non-expired one, we'll remove the freeRun state on the next check
          startFreeRun(productCode, license, new Date());
          nextCheckDelay = FREE_RUN_PERIOD + 15 * DateFormatUtil.SECOND;
        }
        else {
          // was not able to ask the user because another modal dialog was showing
          nextCheckDelay = 5 * DateFormatUtil.MINUTE;
        }
      }
      else {
        // not expired or within grace period, just show notifications if any
        if (license instanceof CertKeyLicense) {
          // for CertKey-based licenses there might be warning messages available
          final CertKeyLicense certKeyLicense = (CertKeyLicense)license;
          final CertKey2 certKey = certKeyLicense.getCertKey();
          final Object msg = certKey.eval(new GetApplicableWarningMessage(
            certKeyLicense.getProductCode(), certKeyLicense.getProductBuildDate(), info.isPlugin()? info.getProductDescriptor().getShortName() : ApplicationNamesInfo.getInstance().getFullProductName()
          ));
          if (msg instanceof LicenseNotification) {
            LicenseNotification message = (LicenseNotification)msg;
            LicenseEntryUtil.showLicenseNotificationLater(message.getTitle(), message.getMessage(), message.getType().toNotificationType(), notification -> {
              for (LicenseNotification.Action action : message.getActions()) {
                notification.addAction(
                  NotificationAction.createSimple(() -> action.getText(), () -> BrowserUtil.browse(action.getLink()))
                );
              }
              return notification;
            });
          }
          final Date expiration = Date.class.cast(certKey.eval(new GetFirstExpiredDate(certKeyLicense.getProductCode())));
          if (expiration != null) {
            final long timeLeftBeforeExpiration = expiration.getTime() - now.getTime();
            if (timeLeftBeforeExpiration > 0) { // not expired yet
              nextCheckDelay = Math.min(DEFAULT_CHECK_PERIOD, timeLeftBeforeExpiration + 15 * DateFormatUtil.SECOND);
            }
          }
        }
      }
    }
    return nextCheckDelay;
  }

  @NotNull
  private static String getLicenseId(LicenseData license) {
    if (license instanceof CertKeyLicense) {
      final String licenseId = (String)((CertKeyLicense)license).getCertKey().eval(new GetLicenseId());
      if (licenseId == null) { // this means the key has been hacked
        //noinspection RedundantClassCall
        AppExecutorUtil.getAppScheduledExecutorService().execute(() -> System.exit(-1));
        return "";
      }
      return licenseId;
    }
    return license instanceof EvaluationLicenseData ? "_evlsprt_" : "_other_";
  }
}
