package com.intellij.ide.license.impl;

import com.intellij.diagnostic.LoadingState;
import com.intellij.ide.license.*;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.licensecommon.util.SubscriptionVersion;
import com.intellij.notification.NotificationType;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.openapi.util.NlsSafe;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.jetbrains.license.util.JBAClientUtil;
import com.jetbrains.license.util.Pair;
import com.jetbrains.license.util.ResponseMarshaller;
import com.jetbrains.license.util.VerifyingSignatureFactory;
import com.jetbrains.ls.responses.Action;
import com.jetbrains.ls.responses.ResponseCode;
import com.jetbrains.ls.responses.ValidateKeyResponse;
import org.jetbrains.annotations.NotNull;

import java.nio.charset.StandardCharsets;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Lock;
import java.util.concurrent.locks.ReentrantLock;


public class FileLicenseSource extends LicenseKeySource<LicenseData> implements LicenseTarget {
  private final LicenseStore myLicenseStore;
  @NotNull
  private final PluginsAwareLicenseManager myLicenseManager;
  private final Lock myProgressLock = new ReentrantLock();
  private volatile Pair<String, @NlsContexts.DetailedDescription String> myKeyToRevoke = null; // pair: [key;message]
  private volatile ScheduledFuture<?> myNextScheduledServerValidation;
  private final long myBuildDate;

  @Override
  public void create(LicenseData licenseData) {
    setLicense(licenseData);
  }

  @Override
  public void restore() {
    LicenseData license = myLicenseStore.getLicense();
    setLicense(license instanceof JBALicenseData ? null : license); // licenses received from jba are managed by a different source
  }

  @Override
  public void clear() {
    myLicenseStore.delete();
    setLicense(null);
  }

  public FileLicenseSource(@NotNull PluginsAwareLicenseManager manager, LicenseFactory factory) {
    super(factory);
    myLicenseManager = manager;
    myLicenseStore = factory.getLicenseStore();
    myBuildDate = SubscriptionVersion.createVersion(
      factory instanceof PluginLicenseFactory?
      ((PluginLicenseFactory)factory).getPluginReleaseDate() :
      ApplicationInfoImpl.getShadowInstance().getMajorReleaseBuildDate().getTime()
    );
  }


  @Override
  public void save(LicenseData license) throws TargetSaveException {
    myLicenseStore.save(license);
  }

  public LicenseFactory getFactory(){
    return factory;
  }

  private static final long SECOND = 1000L;
  private static final long MINUTE = 60 * SECOND;
  private static final long HOUR = 60 * MINUTE;
  private static final long DAY = 24 * HOUR;

  @Override
  public void scheduleRemoteValidation() {
    AppExecutorUtil.getAppExecutorService().execute(() -> {
      myProgressLock.lock();
      long nextValidation = DAY; // default if validation request failed
      try {
        final LicenseData licenseData = getLicense();
        final CertKeyLicense currentCertKeyLicense = licenseData instanceof CertKeyLicense ? (CertKeyLicense)licenseData : null;
        final String key = currentCertKeyLicense != null ? (String)currentCertKeyLicense.getCertKey().eval(new GetLicenseId()) :
                            licenseData != null? licenseData.getKey() : null;
        final Pair<String, @NlsContexts.DetailedDescription String> knownBadKey = myKeyToRevoke;
        if (key == null || StringUtil.isEmptyOrSpaces(key)) {
          myKeyToRevoke = null;
          nextValidation = HOUR;
        }
        else if (knownBadKey != null && knownBadKey.fst.equals(licenseData.getKey())) {
          if (!revokeKey(knownBadKey.snd)) {
            nextValidation = 5 * MINUTE;
          }
        }
        else {
          myKeyToRevoke = null;
          try {
            final String licenseHash = currentCertKeyLicense != null ? String.valueOf(currentCertKeyLicense.getCertKey().eval(new GetHash())) : null;
            final ValidateKeyResponse response = JBAClientUtil.validateLicenseKey(factory.getProductDescriptor().getProductCode(), key, licenseHash != null? licenseHash : "", factory.getCurrentMajorVersion(), null, myBuildDate);
            // this signature-verification code is actually inlined call to
            //Signatures.verify(response.getSignedContent(), response.getSignature(), VerifyingSignatureFactory.X509_CERTIFICATE_GENERIC_DATA_SIGNING);
            boolean verified = false;
            try {
              final String sigString = response.getSignature();
              final String content = ResponseMarshaller.marshall(response);
              final Pair<byte[], VerifyingSignatureFactory.SignatureVerifier> sig = VerifyingSignatureFactory.X509_CERTIFICATE_GENERIC_DATA_SIGNING.create(sigString, content);
              verified = sig.snd.verify(content.getBytes(StandardCharsets.UTF_8), sig.fst);
            }
            catch (Throwable ignored) {
              verified = false;
            }

            Action action = null;
            try {
              if (!verified) {
                action = Action.REVOKE;
              }
              else if (response.getResponseCode() == ResponseCode.SERVER_INTERNAL_ERROR) {
                action = Action.NONE;
              }
              else {
                final long period = response.getValidationPeriod();
                if (period > 0L) {
                  nextValidation = period;
                }
                if (response.getResponseCode() == ResponseCode.OK) {
                  // in case of ResponseCode.OK only WARN action is supported
                  action = response.getAction() == Action.WARN? Action.WARN : Action.NONE;
                  final String newKey = response.getNewLicenseKey();
                  if (newKey != null && !newKey.isEmpty() && currentCertKeyLicense != null) {
                    final CertKeyLicenseData updatedLicenseData = new CertKeyLicenseData(
                      newKey, currentCertKeyLicense.getProductCode(), currentCertKeyLicense.getProductBuildDate()
                    );
                    save(updatedLicenseData);
                    setLicense(updatedLicenseData);
                    myLicenseManager.updateLicenseAbout();
                    myLicenseManager.scheduleLicenseExpirationCheckNow();
                    checkSameAssignee(String.valueOf(currentCertKeyLicense.getCertKey().eval(new GetAssigneeName())), String.valueOf(updatedLicenseData.getCertKey().eval(new GetAssigneeName())));
                  }
                }
                else {
                  // In case of ResponseCode.ERROR can be anything of: REVOKE, WARN or NONE
                  action = response.getAction();
                  if (action == null || action == Action.NONE) {
                    // REVOKE and NONE should be treated both as REVOKE in this case.
                    action = Action.REVOKE;
                  }
                }
              }
            }
            finally {
              @NlsSafe
              final String responseMessage = response.getMessage();
              if (!verified || action == Action.REVOKE) {
                if (!revokeKey(responseMessage)) {
                  myKeyToRevoke = new Pair<>(licenseData.getKey(), responseMessage);
                  nextValidation = Math.min(nextValidation, 5 * MINUTE);
                }
              }
              else if (action == Action.WARN){
                if (responseMessage != null && !responseMessage.trim().isEmpty()) {
                  LicenseEntryUtil.showLicenseNotificationLater(LicenseBundle.message("notification.title.license.validation"), responseMessage, NotificationType.WARNING);
                }
              }
            }
          }
          catch (Throwable ignored) {
          }
        }

        final ScheduledFuture<?> previousValidation = myNextScheduledServerValidation;
        if (previousValidation != null && !previousValidation.isDone()) {
          previousValidation.cancel(false);
        }
      }
      finally {
        try {
          myNextScheduledServerValidation = AppExecutorUtil.getAppScheduledExecutorService().schedule(() -> {
            scheduleRemoteValidation();
          }, nextValidation, TimeUnit.MILLISECONDS);
        }
        finally {
          myProgressLock.unlock();
        }
      }
    });
  }

  private boolean revokeKey(@NlsContexts.DetailedDescription String message) {
    return ApplicationManager.getApplication() != null && LoadingState.APP_STARTED.isOccurred() && LicenseEntryUtil.invokeAndWaitInNonModalContext(() -> {
      clear();
      myKeyToRevoke = null;
      myLicenseManager.updateLicenseAbout();
      myLicenseManager.demandLicense(factory.getProductDescriptor(), LicenseManager.StartupAction.EMPTY, message);
    });
  }

  private void checkSameAssignee(final String wasAssignedTo, final @NlsSafe String nowAssignedTo) {
    if (wasAssignedTo == null || wasAssignedTo.isEmpty() || nowAssignedTo == null || nowAssignedTo.isEmpty() || wasAssignedTo.equals(nowAssignedTo)) {
      return;
    }
    final ProductDescriptor pd = factory.getProductDescriptor();
    @NlsContexts.NotificationContent String message =
      LicenseBundle.message("notification.content.license.you.have.been.using.has.been.reassigned", nowAssignedTo);
    LicenseEntryUtil.invokeOnEdtWhenApplicationLoaded(new LicenseEntryUtil.ApplicationRunnable() {
      @Override
      public void run(Application app) {
        myLicenseManager.demandLicenseAsyncFromBackgroundTask(app, pd, new Pair<>(null, message));
      }
    });
  }
}
