package com.intellij.ide.license.impl;

import com.intellij.ide.license.*;
import com.intellij.ide.license.ui.LicenseEditingModel;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.idea.Main;
import com.intellij.licensecommon.client.impl.ClientImpl;
import com.intellij.licensecommon.license.LicenseBroadcastData;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.licensecommon.log.Log;
import com.intellij.licensecommon.message.impl.ResharperMessage;
import com.intellij.licensecommon.net.Address;
import com.intellij.licensecommon.net.impl.NetworkImpl;
import com.intellij.licensecommon.net.impl.NetworkSocketFactoryImpl;
import com.intellij.licensecommon.protocol.impl.ResharperProtocol;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.updateSettings.impl.UpdateRequestParameters;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.openapi.util.NlsContexts.DialogMessage;
import com.intellij.openapi.util.text.StringUtilRt;
import com.intellij.util.SmartList;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.text.DateFormatUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.awt.*;
import java.lang.reflect.Proxy;
import java.util.List;
import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

@SuppressWarnings("SSBasedInspection")
public abstract class GeneralLicenseManager extends PluginsAwareLicenseManager {
  private static final long WEEK_PERIOD = 7 * 24 * 60 * 60 * 1000L;
  protected final ClientImpl myClient;
  private LicenseFactory myLicenseFactory;
  private final AtomicBoolean myIsLicenseCollisionDialogShown = new AtomicBoolean(false);
  private ProductKey myPlatformKeyData;

  public GeneralLicenseManager() {
    super(new ExpirationDateCheck(true));
    //noinspection AssignmentToStaticFieldFromInstanceMethod
    Log.FACTORY = new IdeaLoggerFactory();
    myClient = new ClientImpl(
      (from, data) -> duplicateLicenseDetected(from, data),
      new NetworkImpl(new NetworkSocketFactoryImpl(), new ResharperProtocol(), AppExecutorUtil.getAppExecutorService()),
      data -> new ResharperMessage(data),
      AppExecutorUtil.getAppScheduledExecutorService()
    );
  }

  @NotNull
  @Override
  public LicenseInfo getPlatformLicenseInfo() {
    return new LicenseInfo(getLicenseFactory().getProductDescriptor(), getCurrentSource(), false, false);
  }

  private LicenseSource<?> getCurrentSource() {
    final ProductKey mainData = myPlatformKeyData;
    return mainData != null? mainData.getActiveSource() : null;
  }

  @Nullable
  private LicenseData getPlatformLicenseData() {
    final LicenseSource<?> source = getCurrentSource();
    return source != null? source.getLicense() : null;
  }

  @Override
  public void startUp(@NotNull StartupAction action) {
    myPlatformKeyData = new ProductKey(getLicenseFactory());
    initPluginDescriptors();

    proceedWithLicenseOrExit(new StartupAction() {
      @Override
      public void proceed() {
        try {
          // deliberately used reflection to obscure field initialization
          Class.forName("com.intellij.ultimate.UltimateVerifier").getField("VERIFIER").set(
            null, Proxy.newProxyInstance(null, new Class[]{Function.class}, new VerifyUltimate())
          );
        }
        catch (Throwable ignored) {
        }
        try {
          action.proceed();
        }
        finally {
          // register periodic task cleanup code once
          LicenseEntryUtil.executeOnApplicationExit(()-> {
            myClient.broadcastLicenseData(null);

            final List<ProductKey> allKeys = new SmartList<>();
            allKeys.add(myPlatformKeyData);
            allKeys.addAll(myPluginKeyData);
            for (ProductKey pk : allKeys) {
              final ServerSource serverSource = pk.findSource(ServerSource.class);
              if (serverSource != null) {
                serverSource.stopProlongTask();
              }
            }
          });
        }
      }
    });
  }

  private void proceedWithLicenseOrExit(StartupAction action) {  // initially on start
    Status platformStatus = Status.ASK;
    final boolean headless = GraphicsEnvironment.isHeadless();
    // restore platform license
    for (LicenseSource<?> source : getSortedLicenseSources(myPlatformKeyData.getSources())) {
      if (source.getLicense() == null) {
        source.restore(); // try restoring, but only once and only if not restored already!
      }
      final LicenseData licenseFromSource = source.getLicense();
      if (licenseFromSource == null) {
        continue;
      }
      platformStatus = validatePlatformLicenseSource(getLicenseFactory(), source);
      if (platformStatus.isLicensed()) {
        myPlatformKeyData.setActiveSource(source, false);
      }
      UpdateRequestParameters.addParameter("license", licenseInfo(licenseFromSource));
      break;
    }
    // restore plugin licenses
    for (PluginProductKey ppk : myPluginKeyData) {
      if (!ppk.getPluginDescriptor().isEnabled() || ppk.getLicense() != null) {
        continue;
      }
      for (LicenseSource<?> src : getSortedLicenseSources(ppk.getSources())) {
        if (src.getLicense() == null) {
          src.restore();
        }
        final LicenseData licenseData = src.getLicense();
        if (licenseData != null) {
          ppk.setActiveSource(src, false);
          break;
        }
      }
    }
    // validate and, if possible, reuse licenses for plugins
    final Collection<PluginProductKey> unlicensedPlugins = revalidatePluginLicenses(true, null);

    final LicenseData platformLic = myPlatformKeyData.getLicense();
    boolean platformLicensed;
    final boolean canProceed = headless ?
                               platformLicensed = platformStatus.isLicensed() :
                               (platformLicensed = platformStatus.isLicensed() &&
                               !showExpirationWarning(platformLic) /*additionally check expiration date for the platform license if it is valid*/) &&
                               unlicensedPlugins.isEmpty();

    if (platformStatus == Status.FREE_RUN && platformLic != null) {
      myExpirationDateCheck.startFreeRun(myPlatformKeyData.getProductCode(), platformLic, new Date());
    }
    if (canProceed) {
      for (PluginProductKey plugin : unlicensedPlugins) {
        plugin.disablePlugin();
        LOG.info("No valid license found for plugin " + plugin.getProductDescriptor().getShortName() + "; The plugin will be disabled.");
      }
      proceed(action);
    }
    else {
      if (headless) {
        //noinspection UseOfSystemOutOrSystemErr
        System.err.println("No valid license found");
        LOG.error("No valid license found");
        System.exit(Main.LICENSE_ERROR);
      }
      myExpirationDateCheck.scheduleValidation(15 * DateFormatUtil.MINUTE, this);

      ProductDescriptor toSelect =
        platformLicensed && (!(platformLic instanceof ServerLicenseData) || ((ServerLicenseData)platformLic).isTicketObtained()) && !unlicensedPlugins.isEmpty() ? unlicensedPlugins.iterator().next().getProductDescriptor() : null;
      if (!demandLicense(toSelect, action, null)) {
        System.exit(Main.LICENSE_ERROR);
      }
    }
  }

  private static String licenseInfo(LicenseData licenseData) {
    if (licenseData instanceof EvaluationLicenseData) {
      return "eval";
    }
    else {
      StringBuilder licenseInfo = new StringBuilder();
      if (licenseData instanceof ServerLicenseData) {
        licenseInfo.append("server,");
      }
      if (licenseData.isYearAcademic()) {
        licenseInfo.append("classroom");
      }
      else if (licenseData.isPersonal()) {
        licenseInfo.append("personal");
      }
      else if (licenseData.isOpenSource()) {
        licenseInfo.append("opensource");
      }
      else if (licenseData.isNonCommercial()) {
        licenseInfo.append("academic");
      }
      else {
        licenseInfo.append("commercial");
      }
      return licenseInfo.toString();
    }
  }

  protected boolean showExpirationWarning(@Nullable final LicenseData licenseData) {
    if (licenseData == null || licenseData instanceof CertKeyLicense) {
      return false;
    }
    final Date expirationDate = licenseData.getExpirationDate();
    if (expirationDate == null) {
      return false;
    }
    final long timeBeforeExpiration = expirationDate.getTime() - System.currentTimeMillis();
    if (timeBeforeExpiration >= 0 && timeBeforeExpiration <= WEEK_PERIOD) {
      final long daysBeforeExpiration = BuiltInEvaluationSource.remainsDays(licenseData.getExpirationDate(), new Date());
      return getMessages().showWillExpireWarning(myPlatformKeyData.getProductDescriptor(), false, daysBeforeExpiration, licenseData instanceof EvaluationLicenseData);
    }
    return false;
  }

  @NotNull
  private Status validatePlatformLicenseSource(final LicenseFactory factory, final LicenseSource<?> source) {
    final boolean isInteractiveMode = !GraphicsEnvironment.isHeadless();
    final LicenseData license = source.getLicense();
    try {
      source.validate();
    }
    catch (EvaluationExpiredException e) {
      if (isInteractiveMode) {
        final boolean enterLicense = getMessages().askEnterKeyBecauseLicenseExpired(factory.getProductDescriptor(), false, true);
        //activate 30 minutes session if not "Enter Key" selected
        return enterLicense ? Status.ASK : Status.FREE_RUN;
      }
      LOG.info("Evaluation license expired");
      return Status.ASK;
    }
    catch (InvalidLicenseException e) {
      if (isInteractiveMode) {
        getMessages().showLicenseInvalidMessage(factory.getProductDescriptor());
      }
      else {
        LOG.info(LicenseBundle.message("message.invalid.license.data"));
      }
      return Status.ASK;
    }
    catch (LicenseExpiredException e) {
      if (isInteractiveMode) {
        getMessages().showLicenseExpired(factory.getProductDescriptor(), false);
      }
      else {
        LOG.info(LicenseBundle.message("message.license.expired"));
      }
      return Status.ASK;
    }
    catch (ValidationFailedException e) {
      LOG.info("License validation failed " + e.getMessage());
      return Status.ASK;
    }
    return license != null ? Status.OK : Status.ASK;
  }

  @Override
  protected void proceed(@NotNull StartupAction action) {
    try {
      super.proceed(action);
    }
    finally {
      final LicenseSource<?> src = myPlatformKeyData.getActiveSource();
      if (src != null) {
        src.scheduleRemoteValidation();
        myClient.broadcastLicenseData(src.needNetworkLicenseCheck()? convertToBroadcastLicense(src.getLicense()): null);
      }
    }
  }

  private static LicenseBroadcastData convertToBroadcastLicense(LicenseData license) {
    if (license instanceof CertKeyLicense) {
      final CertKey2 certKey = ((CertKeyLicense)license).getCertKey();
      final Object checkConcUse = certKey.eval(new IsCheckConcurrentUse());
      if (checkConcUse == null) {
        // the license is hacked
        AppExecutorUtil.getAppScheduledExecutorService().submit(() -> System.exit(-1));
      }
      if (Boolean.TRUE.equals(checkConcUse)) {
        final Object id = certKey.eval(new GetLicenseId());
        final Object name = certKey.eval(new GetLicenseeName());
        if (id != null && name != null) {
          return new LicenseBroadcastData(0, id.toString(), name.toString(), System.getProperty("user.name", ""));
        }
        // the license is hacked
        AppExecutorUtil.getAppScheduledExecutorService().submit(() -> System.exit(-1));
        return null;
      }
    }
    return null;
  }

  @Override
  public boolean supportsRegistration() {
    return true;
  }

  @NotNull
  @Override
  protected LicenseEditingModel createLicenseEditingModel(ProductDescriptor pd, StartupAction action, @Nullable String statusMessage) {
    return new LicenseEntryModel(pd, action, statusMessage);
  }

  @Override
  public String licensedToMessage() {
    final LicenseData licenseData = getPlatformLicenseData();
    final String licensedToName = licenseData != null ? licenseData.getUserName() : null;
    return licensedToName != null ? "Licensed to "+licensedToName : "";
  }

  @Override
  public @NotNull Supplier<List<String>> licensedRestrictionsMessages() {
    LicenseData licenseData = getPlatformLicenseData();
    return licenseData == null ? () -> Collections.emptyList() : () -> getMessages().getUsageRestrictionsMessages(myPlatformKeyData.getProductDescriptor(), licenseData);
  }

  @Override
  public LicenseFactory getLicenseFactory() {
    if (myLicenseFactory == null) {
      myLicenseFactory = createLicenseFactory();
    }
    return myLicenseFactory;
  }
  protected abstract LicenseFactory createLicenseFactory();

  @Override
  public boolean needEvaluationFollowUp() {
    return getPlatformLicenseInfo().isEvaluation() && !isEap();
  }

  private void duplicateLicenseDetected(Address from, LicenseBroadcastData conflictingData) {
    final LicenseData licenseData = getPlatformLicenseData();
    if (licenseData == null || !licenseData.shouldDetectDuplicates() || from.isAnyLocalAddress()) {
      return;
    }
    if (!myIsLicenseCollisionDialogShown.getAndSet(true)) {
      final String fromText = conflictingData.getUserLogin() + " @ " + from.getDisplay();
      LicenseBroadcastData broadcastData = myClient.getBroadcastData();
      final String selfUsername = broadcastData != null ? broadcastData.getUserLogin() : null;
      SwingUtilities.invokeLater(() -> {
        try {
          myClient.cancelBroadcastTask(); // pause broadcasting while showing the dialog
          final String message = StringUtilRt.isEmpty(selfUsername)?
                                 LicenseBundle.message("error.message.duplicate.license", fromText):
                                 LicenseBundle.message("error.message.duplicate.license.ext", fromText, selfUsername);
          if (reactivateOnLicenseCollision(message)) {
            myClient.broadcastLicenseData(convertToBroadcastLicense(getPlatformLicenseData()));
          }
          else {
            shutdownApplication();
          }
        }
        finally {
          myIsLicenseCollisionDialogShown.set(false);
        }
      });
    }
  }

  private static boolean reactivateOnLicenseCollision(@DialogMessage String message) {
    final String title = LicenseBundle.message("dialog.title.license.collision.detected");
    if (ApplicationManager.getApplication() == null) {
      return JOptionPane.showOptionDialog(
        JOptionPane.getRootFrame(), message, title, JOptionPane.YES_NO_OPTION, JOptionPane.ERROR_MESSAGE, null, new String[]{
          LicenseBundle.message("button.reactivate"), LicenseBundle.message("button.shutdown")
        }, LicenseBundle.message("button.reactivate")
      ) == JOptionPane.YES_OPTION;
    }
    return Messages.showYesNoDialog(message, title, LicenseBundle.message("button.reactivate"), LicenseBundle.message("button.shutdown"), Messages.getErrorIcon()) == Messages.YES;
  }

  private class LicenseEntryModel implements LicenseEditingModel {

    private final ProductDescriptor myProductToSelect;
    private final StartupAction myAction;
    private final @NlsContexts.DetailedDescription String myInitialMessage;
    private final ProductDescriptor myPlatformProduct;
    private final Map<ProductDescriptor, PluginProductKey> myPluginProducts = new HashMap<>();

    LicenseEntryModel(ProductDescriptor productToSelect, StartupAction action, @NlsContexts.DetailedDescription String initialMessage) {
      myProductToSelect = productToSelect;
      myAction = action;
      myInitialMessage = initialMessage;
      myPlatformProduct = getPlatformLicenseInfo().getProductDescriptor();
      for (PluginProductKey key : myPluginKeyData) {
        myPluginProducts.put(key.getProductDescriptor(), key);
      }
    }

    @Override
    public ProductDescriptor getInitialProductToSelect() {
      final ProductDescriptor toSelect = myProductToSelect;
      return toSelect != null ? toSelect : getPlatformProduct();
    }

    @Nullable
    @Override
    public String getInitialProblemDescription() {
      return myInitialMessage;
    }

    @Override
    public ProductDescriptor getPlatformProduct() {
      return myPlatformProduct;
    }

    @Override
    public Collection<ProductDescriptor> getPluginProducts() {
      return Collections.unmodifiableCollection(myPluginProducts.keySet());
    }

    @Override
    public LicenseInfo getLicenseInfo(ProductDescriptor pd) {
      if (pd.equals(getPlatformProduct())) {
        return getPlatformLicenseInfo();
      }
      final PluginProductKey key = myPluginProducts.get(pd);
      return key != null? new LicenseInfo(pd, key.getActiveSource(), true, key.getPluginDescriptor().isLicenseOptional()) : null;
    }

    @NotNull
    @Override
    public Collection<LicenseSource<?>> getAvailableLicenseSources(ProductDescriptor pd) {
      final ProductKey key = pd.equals(getPlatformProduct())? myPlatformKeyData : myPluginProducts.get(pd);
      return key != null? key.getSources() : Collections.emptyList();
    }

    @Override
    public LicenseInfo activateLicense(ProductDescriptor pd, LicenseSource source) {
      final boolean isPlatformProduct = pd.equals(getPlatformProduct());
      final ProductKey key = isPlatformProduct ? myPlatformKeyData : myPluginProducts.get(pd);
      if (key != null) {
        final Status status = key instanceof PluginProductKey? validatePluginLicenseSource((PluginProductKey)key, source, null) : validatePlatformLicenseSource(key.getLicenseFactory(), source);
        if (status.isLicensed()) {
          final LicenseData lic = source.getLicense();
          if (isPlatformProduct && status == Status.FREE_RUN && lic != null) {
            myExpirationDateCheck.startFreeRun(pd.getProductCode(), lic, new Date());
          }
          key.setActiveSource(source, true);
          if (lic != null && isPlatformProduct) {
            UpdateRequestParameters.addParameter("license", licenseInfo(lic));
          }
          revalidatePluginLicenses(true, null);
          updateLicenseAbout();
        }
      }
      return getLicenseInfo(pd);
    }

    @Override
    public String getDisplayName(ProductDescriptor pd) {
      String name = null;
      if (pd.equals(getPlatformProduct())) {
        name = ApplicationInfoImpl.getShadowInstance().getVersionName();
      }
      else {
        final PluginProductKey key = myPluginProducts.get(pd);
        if (key != null) {
          name = key.getPluginDescriptor().getName();
        }
      }
      return name != null ? name : pd.getShortName();
    }

    @Override
    public void forgetLicense(ProductDescriptor pd) {
      try {
        Collection<LicenseSource<?>> availableLicenseSources = getAvailableLicenseSources(pd);
        for (LicenseSource<?> source : availableLicenseSources) {
          try {
            source.clear();
          }
          catch (Throwable ignored) {
          }
        }
        // if there is a non-clearable valid license left (e.g. eval license), return to it
        for (LicenseSource<?> source : availableLicenseSources) {
          source.restore();
          final LicenseData license = source.getLicense();
          if (license != null && !license.isExpired()) {
            activateLicense(pd, source);
            break;
          }
        }
      }
      finally {
        // todo: do we need this notification?
        //LicenseEntryUtil.showLicenseNotificationLater(
        //  "Registration",
        //  ApplicationNamesInfo.getInstance().getFullProductName() + " license information was removed",
        //  NotificationType.INFORMATION
        //);
        updateLicenseAbout();
      }
    }

    @Override
    public void close(boolean forceExit) {
      AppExecutorUtil.getAppExecutorService().submit(() -> {
        getTokenManager().clearAuthorizationAttempts();
        if (forceExit) {
          shutdownApplication();
        }
        else {
          final Predicate<LicenseData> licPredicate = lic -> !(lic instanceof ServerLicenseData) || ((ServerLicenseData)lic).isTicketObtained();
          final LicenseInfo info = getPlatformLicenseInfo();
          final LicenseData platformLicense = info.getLicense();
          if (platformLicense == null ||
              !licPredicate.test(platformLicense) ||
              (platformLicense.isExpired() && !myExpirationDateCheck.isFreeRunActive(info.getProductDescriptor().getProductCode(), platformLicense))) {
            shutdownApplication();
          }
          else {
            try {
              for (PluginProductKey key : revalidatePluginLicenses(false, licPredicate)) {
                key.disablePlugin();
              }
            }
            finally {
              proceed(myAction);
            }
          }
        }
      });
    }
  }
}
