package com.intellij.ide.license.impl;

import com.intellij.diagnostic.LoadingState;
import com.intellij.ide.Prefs;
import com.intellij.ide.license.*;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.licensecommon.util.SubscriptionVersion;
import com.intellij.notification.NotificationType;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.NlsSafe;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.SmartList;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.text.DateFormatUtil;
import com.jetbrains.license.util.*;
import com.jetbrains.ls.responses.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.*;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicReference;
import java.util.zip.InflaterInputStream;

public class JetprofileSource extends GeneralLicenseSource<Pair<String, char[]>> implements LicenseTarget{
  private static final Logger LOG = Logger.getInstance("#JetprofileSource");
  private static final long DEFAULT_CHECK_PERIOD = DateFormatUtil.HOUR;
  private static final long OFFLINE_WARNING_PERIOD = 2 * DateFormatUtil.DAY;
  private static final Charset CHARSET = StandardCharsets.UTF_8;
  private static final byte[] SECRET = "The Drive to Develop.".getBytes(CHARSET);
  private static final int LOAD_FROM_SHARED_ID_TIMEOUT = 5000; // 5 seconds

  private final AtomicReference<JBALicenseData> myLicenseRef = new AtomicReference<>(null);
  private final long myBuildDate;
  private final PluginsAwareLicenseManager myLicenseManager;
  @Nullable
  private final String myEdition;
  private final LicenseStore myLicenseFile;
  private final AtomicBoolean myAllowLoadingFromSharedID;

  // stored in memory only, not persisted!
  private String myLogin;
  private volatile ScheduledFuture<?> myPeriodicCheckFuture;
  private volatile List<JBALicenseData> myAvailableLicenses = Collections.emptyList();
  private volatile boolean myFirstCheckPerformed = false;  // needed to force validation check the very first time after IDE started
  private volatile long myLastOfflineWarningStamp = -1L;

  public JetprofileSource(@NotNull LicenseFactory factory, PluginsAwareLicenseManager licenseManager, final @Nullable String edition) {
    super(factory);
    myLicenseManager = licenseManager;
    myBuildDate = SubscriptionVersion.createVersion(getBuildDate(factory));
    myEdition = edition;
    myLicenseFile = factory.getLicenseStore();
    // attempt to fetch from shared user_id only for platform license
    myAllowLoadingFromSharedID = new AtomicBoolean(!(factory instanceof PluginLicenseFactory));
  }

  @NotNull
  private String getProductCode() {
    return factory.getProductDescriptor().getProductCode();
  }

  private static Date getBuildDate(@Nullable final LicenseFactory factory) {
    if (factory instanceof PluginLicenseFactory) {
      return ((PluginLicenseFactory)factory).getPluginReleaseDate();
    }
    return ApplicationInfoImpl.getShadowInstance().getMajorReleaseBuildDate().getTime();
  }

  @Override
  public void save(LicenseData license) throws TargetSaveException {
    myLicenseFile.save(license);
  }

  @Override
  public void create(Pair<String, char[]> userInput) throws LicenseCreationException{
    try {
      setLicense(obtainRemoteLicense(userInput.fst, new String(userInput.snd)));
    }
    catch (LicenseCreationException e) {
      setLicense(null);
      throw e;
    }
    catch (Throwable e) {
      setLicense(null);
      myLogin = null;
      final String reason = e.getMessage();
      LOG.info("Unexpected JBA connection error: " + (reason != null? reason : "") + " (" +e.getClass().getName() + ")");
      final LicenseCreationException creationException = new LicenseCreationException(
        LicenseBundle.message("error.message.text.jetbrains.account.error"), reason);
      if (e instanceof IOException) {
        creationException.initCause(e);
      }
      throw creationException;
    }
  }

  @Nullable
  private JBALicenseData obtainRemoteLicense(@NlsSafe String login, String password) throws Exception {
    myAvailableLicenses = Collections.emptyList();
    final String userId = obtainUserId(login, password);
    final ObtainLicenseResponse response = obtainLicense(userId);
    final String responseMessage = response.getMessage();

    if (response.getResponseCode() != ResponseCode.OK) {
      final StringBuilder message = new StringBuilder();
      message.append("Error fetching licenses for ").append(login);
      if (!StringUtil.isEmptyOrSpaces(responseMessage)) {
        message.append(": ").append(responseMessage);
      }
      throw new Exception(message.toString());
    }

    final List<JBALicenseData> allLicenses = createLicenses(login, userId, response.getAssets());
    if (allLicenses.isEmpty()) {
      throw new LicenseCreationException(LicenseBundle.message("error.message.text.license.not.found"),
                                         LicenseBundle.message("error.text.no.suitable.licenses.associated.with.account", login));
    }
    myLogin = login;
    updateSharedUserId(userId, login);
    myAvailableLicenses = Collections.unmodifiableList(allLicenses);
    updateValidationStamp(System.currentTimeMillis());
    updateValidationPeriod(response.getValidationPeriod(), response.getValidationDeadlinePeriod());
    return allLicenses.size() == 1? allLicenses.iterator().next() : null;
  }

  @NotNull
  private List<JBALicenseData> createLicenses(String login, String userId, Collection<EncodedAsset> assets) {
    if (assets.isEmpty()) {
      return Collections.emptyList();
    }
    final List<JBALicenseData> allLicenses = new SmartList<>();
    for (EncodedAsset asset : assets) {
      try {
        final JBALicenseData lic = createLicenseData(login, userId, factory.getProductDescriptor(), getBuildDate(factory), asset);
        if (!lic.isExpired()) {
          allLicenses.add(lic);
        }
      }
      catch (LicenseCreationException e) {
        final StringBuilder message = new StringBuilder(e.getMessage());
        final String details = e.getDetails();
        if (details != null) {
          message.append(": ").append(details);
        }
        LOG.info(message.toString());
      }
      catch (Throwable e) {
        LOG.info(e.getMessage());
      }
    }
    return allLicenses;
  }

  static JBALicenseData createLicenseData(String login, String userId, final ProductDescriptor productDescriptor, final Date buildDate, EncodedAsset asset) throws LicenseCreationException, DecodeException {
    final License decoded = asset.decode();
    if (decoded.getLicenseKey() != null) {
      return new JetProfileCertLicenseData(userId, login, productDescriptor.getProductCode(), buildDate, asset, decoded);
    }
    if (decoded.getUpgradeDueDate() == null) {
      return new JetProfileLicenseData(userId, login, asset, decoded);
    }
    return new JetProfileSubscriptionLicenseData(userId, login, asset, decoded);
  }

  @Override
  public void restore() {
    LicenseData lic = myLicenseFile.getLicense();
    if (lic != null && !(lic instanceof JBALicenseData)) {
      return;
    }
    JBALicenseData restored = (JBALicenseData)lic;
    if (restored == null) {
      // check old location in Preferences: compatibility with previous versions
      restored = loadFromPreferences(factory.getProductDescriptor(), getBuildDate(factory));
    }

    if ((restored == null || restored.isExpired()) && myLicenseFile.readServerLicenseData() == null && myAllowLoadingFromSharedID.getAndSet(false)) {
      // attempt to fetch from shared user_id only once
      try {
        final JBALicenseData fromSharedId = AppExecutorUtil.getAppExecutorService().submit(
          () -> loadFromSharedUserId()
        ).get(LOAD_FROM_SHARED_ID_TIMEOUT, TimeUnit.MILLISECONDS);
        if (fromSharedId != null) {
          // save the asset just obtained from remote server
          save(restored = fromSharedId);
        }
      }
      catch (Throwable ignored) {
      }
    }

    setLicense(restored);
    if (restored == null) {
      myLogin = xorDecode(Prefs.get(getSharedUserLoginStorageKey(), null));
    }
    else {
      myLogin = restored.getUserLogin();
    }
  }

  @Override
  public void clear() {
    clearStoredLicenseInfo(factory.getProductDescriptor());
    Prefs.remove(getSharedUserIdStorageKey());
    Prefs.remove(getSharedUserLoginStorageKey());
    setLicense(null);
  }

  @Override
  public boolean needNetworkLicenseCheck() {
    return true;
  }

  private ObtainLicenseResponse obtainLicense(String userId) throws IOException, VerificationFailedException {
    //noinspection ResultOfMethodCallIgnored
    LicensingProxyService.getInstance();
    //final long start = System.currentTimeMillis();
    //try {
    final ObtainLicenseResponse response = JBAClientUtil.obtainLicenses(userId, getProductCode(), factory.getCurrentMajorVersion(), myEdition, myBuildDate);
    final String signature = response.getSignature();
    Signatures.verify(ResponseMarshaller.marshall(response), signature == null ? "" : signature, VerifyingSignatureFactory.X509_CERTIFICATE_GENERIC_DATA_SIGNING);
    return response;
    //}
    //finally {
    //  System.out.println("Obtain license in " + (System.currentTimeMillis() - start) + " ms");
    //}
  }

  private ValidateLicenseResponse validateLicense(String userId, String assetId, String licenseHash) throws IOException, VerificationFailedException {
    //final long start = System.currentTimeMillis();
    //try {
    final ValidateLicenseResponse response = JBAClientUtil.validateLicense(userId, assetId, licenseHash, getProductCode(), factory.getCurrentMajorVersion(), myEdition, myBuildDate);
    final String signature = response.getSignature();
    Signatures.verify(ResponseMarshaller.marshall(response), signature == null ? "" : signature, VerifyingSignatureFactory.X509_CERTIFICATE_GENERIC_DATA_SIGNING);
    return response;
    //}
    //finally {
    //  System.out.println("Validate license in " + (System.currentTimeMillis() - start) + " ms");
    //}
  }

  @NotNull
  private static String obtainUserId(final String login, final String password) throws Exception {
    //noinspection ResultOfMethodCallIgnored
    LicensingProxyService.getInstance();
    String errorMessage;
    IOException cause = null;
    if (login == null) {
      errorMessage = "JetBrains Account login name not specified";
    }
    else {
      try {
        final ObtainUserIdResponse userIdResponse;
        //final long start = System.currentTimeMillis();
        //try {
          userIdResponse = JBAClientUtil.obtainUserId(login, password == null ? "" : password);
        //}
        //finally {
        //  System.out.println("Obtain userId in " + (System.currentTimeMillis() - start) + " ms");
        //}
        if (userIdResponse.getResponseCode() == ResponseCode.OK) {
          return userIdResponse.getUserId();
        }
        errorMessage = "Authorization failed: " + userIdResponse.getMessage();
      }
      catch (IOException e) {
        cause = e;
        errorMessage = "JetBrains Account connection error: " + e.getMessage();
      }
      catch (VerificationFailedException e) {
        errorMessage = "JetBrains Account communication error: " + e.getMessage();
      }
    }
    throw cause == null? new Exception(errorMessage) : new IOException(errorMessage, cause);
  }

  @Nullable
  public String getLogin() {
    return myLogin;
  }

  private static void updateSharedUserId(String userId, String login) {
    Prefs.put(getSharedUserIdStorageKey(), xorEncode(userId));
    if (login != null) {
      Prefs.put(getSharedUserLoginStorageKey(), xorEncode(login));
    }
  }

  //@Nullable
  private JBALicenseData loadFromSharedUserId() {
    String userId = getSharedUserId();
    if (userId != null) {
      try {
        final ObtainLicenseResponse response = obtainLicense(userId);
        if (response.getResponseCode() == ResponseCode.OK) {
          final String login = Prefs.get(getSharedUserLoginStorageKey(), null);
          final List<JBALicenseData> licenses = createLicenses(xorDecode(login), userId, response.getAssets());
          if (licenses.size() == 1) {
            // when multiple assets for this product are available, let the user choose interactively
            return licenses.iterator().next();
          }
        }
      }
      catch (Throwable e) {
        LOG.info(e.getMessage());
      }
    }
    return null;
  }

  @Nullable
  public static String getSharedUserId() {
    String userId = Prefs.get(getSharedUserIdStorageKey(), null);
    return userId == null ? null : xorDecode(userId);
  }

  @Nullable
  public static JBALicenseData loadFromPreferences(final ProductDescriptor pd, final Date buildDate) {
    try {
      final byte[] data = Prefs.getByteArray(getStorageKey(pd), null);
      if (data != null) {
        final Properties properties = new Properties();
        properties.load(new InflaterInputStream(new ByteArrayInputStream(data)));

        final String userId = properties.getProperty("user.id", null);
        if (userId == null) {
          LOG.info("Corrupted JetProfile license data for "+ pd.getProductCode() + ": user id missing");
          return null;
        }
        final String assetData = properties.getProperty("asset.data", null);
        if (assetData == null) {
          LOG.info("Corrupted JetProfile license data for " + pd.getProductCode()+ ": asset data missing");
          return null;
        }
        return createLicenseData(
          properties.getProperty("user.login", null), userId, pd, buildDate, new EncodedAsset(assetData)
        );
      }
    }
    catch (Throwable e) {
      LOG.info("Error loading JetProfile license data for " + pd.getProductCode() + ": " + e.getMessage());
    }
    return null;
  }

  private synchronized long getValidationStamp() {
    return Prefs.getLong(getStampStorageKey(factory.getProductDescriptor()), 0L);
  }

  private synchronized long getValidationPeriod() {
    return Prefs.getLong(getPeriodStorageKey(factory.getProductDescriptor()), DEFAULT_CHECK_PERIOD);
  }

  private synchronized long getValidationDeadlinePeriod() {
    return Prefs.getLong(getDeadlineStorageKey(factory.getProductDescriptor()), -1L);
  }

  private synchronized void updateValidationPeriod(long period, long deadlinePeriod) {
    Prefs.putLong(getPeriodStorageKey(factory.getProductDescriptor()), period);
    if (deadlinePeriod > 0L) {
      Prefs.putLong(getDeadlineStorageKey(factory.getProductDescriptor()), deadlinePeriod);
    }
    else {
      Prefs.remove(getDeadlineStorageKey(factory.getProductDescriptor()));
    }
  }

  private synchronized void updateValidationStamp(long stamp) {
    Prefs.putLong(getStampStorageKey(factory.getProductDescriptor()), stamp);
  }

  private void clearStoredLicenseInfo(final ProductDescriptor pd) {
    Prefs.remove(getStampStorageKey(pd));
    Prefs.remove(getStorageKey(pd));
    Prefs.remove(getPeriodStorageKey(pd));
    myLicenseFile.delete();
  }

  private static String getStorageKey(final ProductDescriptor pd) {
    return "JetBrains.JetProfile.Asset." + pd.getProductCode();
  }

  private static String getSharedUserIdStorageKey() {
    return "JetBrains.JetProfile.UserId";
  }

  private static String getSharedUserLoginStorageKey() {
    return "JetBrains.JetProfile.UserLogin";
  }

  private static String getStampStorageKey(final ProductDescriptor pd) {
    return getStorageKey(pd) + ".Stamp";
  }

  private static String getPeriodStorageKey(final ProductDescriptor pd) {
    return getStorageKey(pd) + ".Period";
  }

  private static String getDeadlineStorageKey(final ProductDescriptor pd) {
    return getStorageKey(pd) + ".Deadline";
  }

  @Override
  public JBALicenseData getLicense() {
    return myLicenseRef.get();
  }

  @Override
  public void setLicense(LicenseData license) {
    myLicenseRef.set((JBALicenseData)license);
      if (license == null) {
        cancelRemoteValidation();
      }
  }

  public List<JBALicenseData> getAvailableLicenses() {
    return myAvailableLicenses;
  }

  private static String xorEncode(String s) {
    return s != null? new String(Base64.getMimeEncoder().encode(xor(s.getBytes(CHARSET))), CHARSET) : null;
  }

  private static String xorDecode(String s) {
    return s != null? new String(xor(Base64.getMimeDecoder().decode(s.getBytes(CHARSET))), CHARSET) : null;
  }

  private static byte[] xor(byte[] str) {
    final byte[] secret = SECRET;
    for (int i = 0; i < str.length; i++) {
      str[i] ^= secret[i % secret.length];
    }
    return str;
  }

  private void cancelRemoteValidation() {
    final ScheduledFuture<?> future = myPeriodicCheckFuture;
    if (future != null) {
      myPeriodicCheckFuture = null;
      future.cancel(true);
    }
  }

  @Override
  public void scheduleRemoteValidation() {
    scheduleRemoteValidation(getValidationPeriod());
  }

  private void scheduleRemoteValidation(final long validationPeriod) {
    cancelRemoteValidation();
    final long effectivePeriod = validationPeriod <= 0L? DEFAULT_CHECK_PERIOD : validationPeriod;
    final long initialDelay;
    if (myFirstCheckPerformed) {
      initialDelay = effectivePeriod;
    }
    else {
      long val = 5 * DateFormatUtil.MINUTE;
      try {
        val = Math.min(Long.parseLong(System.getProperty("jba.remote.validation.initial.delay")), val);
      }
      catch (Throwable ignored) {
      }
      initialDelay = val;
    }
    myPeriodicCheckFuture = AppExecutorUtil.getAppScheduledExecutorService().scheduleWithFixedDelay(
      () -> {
        try {
          final JBALicenseData license = getLicense();
          if (license != null) {
            final Application app = ApplicationManager.getApplication();
            if (app != null) { // app can be null on early stages
              app.executeOnPooledThread(() -> new ValidateLicenseTask(license, validationPeriod).run());
            }
          }
        }
        catch (Throwable ignored) {
        }
      }, initialDelay /*first check*/, effectivePeriod, TimeUnit.MILLISECONDS
    );
  }

  private static final AtomicBoolean myOnlineValidationInProgress = new AtomicBoolean(false);

  private class ValidateLicenseTask {
    @NotNull
    private final JBALicenseData myLicenseData;
    private final long myValidationPeriod;

    ValidateLicenseTask(@NotNull JBALicenseData licenseData, long validationPeriod) {
      myLicenseData = licenseData;
      myValidationPeriod = validationPeriod;
    }

    void run() {
      if (myOnlineValidationInProgress.getAndSet(true)) {
        return;
      }
      try {
        // if app is not initialized or already disposed, the validation does not make sense
        if (!LoadingState.COMPONENTS_LOADED.isOccurred()) {
          return;
        }

        Application app = ApplicationManager.getApplication();
        if (app == null || app.isDisposed()) {
          return;
        }

        final JBALicenseData currentLicenseData = getLicense();
        if (currentLicenseData == null || !Objects.equals(currentLicenseData.getLicenseId(), myLicenseData.getLicenseId())) {
          return;
        }
        myFirstCheckPerformed = true;
        Action action = Action.NONE;
        @NlsSafe
        String problemReason = "";
        final long lastTimeValidated = getValidationStamp();
        final long timePassedSinceLastCheck = System.currentTimeMillis() - lastTimeValidated;
        final long currentDeadlinePeriod = getValidationDeadlinePeriod();
        try {
          final String licenseHash = myLicenseData instanceof CertKeyLicense ? String.valueOf(((CertKeyLicense)myLicenseData).getCertKey().eval(new GetHash())) : "";
          final ValidateLicenseResponse response = validateLicense(myLicenseData.getUserId(), myLicenseData.getLicense().getAssetId(), licenseHash);
          long newValidationPeriod = myValidationPeriod;
          long newDeadlinePeriod = currentDeadlinePeriod;
          if (response.getResponseCode() == ResponseCode.OK) {
            updateValidationStamp(System.currentTimeMillis());
            newValidationPeriod = response.getValidationPeriod();
            newDeadlinePeriod = response.getValidationDeadlinePeriod();
            if (newValidationPeriod < 0L) {
              newValidationPeriod = DEFAULT_CHECK_PERIOD;
            }
            final String newAsset = response.getNewEncodedAsset();
            if (newAsset != null) {
              try {
                final JBALicenseData newLicenseData = createLicenseData(
                  myLicenseData.getUserLogin(), myLicenseData.getUserId(), factory.getProductDescriptor(), getBuildDate(factory), new EncodedAsset(newAsset)
                );
                if (myLicenseRef.compareAndSet(myLicenseData, newLicenseData)) {
                  myLicenseManager.updateLicenseAbout();
                  save(newLicenseData);
                  myLicenseManager.scheduleLicenseExpirationCheckNow();
                }
              }
              catch (Throwable e) {
                LOG.info(e.getMessage());
              }
            }
            final Action _action = response.getAction();
            if (_action == Action.WARN) { // for OK responses support only warn action
              final String msg = response.getMessage();
              if (msg != null && !msg.trim().isEmpty()) {
                action = _action;
                problemReason = msg;
              }
            }
          }
          else {
            final boolean isError = response.getResponseCode() == ResponseCode.ERROR;
            problemReason = response.getMessage();
            if (response.getResponseCode() == ResponseCode.SERVER_INTERNAL_ERROR) {
              newValidationPeriod = DEFAULT_CHECK_PERIOD;
            }
            if (isError) {
              final Action _action = response.getAction();
              action = _action != null && _action != Action.NONE ? _action : Action.REVOKE;
            }
          }
          final boolean validationPeriodUpdated = myValidationPeriod != newValidationPeriod;
          if (validationPeriodUpdated || currentDeadlinePeriod != newDeadlinePeriod) {
            updateValidationPeriod(newValidationPeriod, newDeadlinePeriod);
            if (validationPeriodUpdated) {
              scheduleRemoteValidation(newValidationPeriod);
            }
          }
        }
        catch (IOException ignored) {
          // is offline or communication error
          if (currentDeadlinePeriod > 0L) {
            if (timePassedSinceLastCheck >= currentDeadlinePeriod) {
              action = Action.REVOKE;
              problemReason = LicenseBundle.message("notification.content.unable.to.contact.your.jetbrains.account");
            }
            else {
              final long timeLeftUntilDeadline = currentDeadlinePeriod - timePassedSinceLastCheck;
              if (timeLeftUntilDeadline <= DateFormatUtil.WEEK) {
                final long lastTimeWarned = myLastOfflineWarningStamp;
                final long now = System.currentTimeMillis();
                // warn about close deadline on every start and every OFFLINE_WARNING_PERIOD milliseconds
                final boolean shouldWarn = lastTimeWarned < 0L || (now - lastTimeWarned) >= OFFLINE_WARNING_PERIOD;
                if (shouldWarn) {
                  LicenseEntryUtil.showLicenseNotificationLater(
                    LicenseBundle.message("notification.title.license.validation"),
                    LicenseBundle.message("notification.content.unable.to.contact.your.jetbrains.account.enable.internet",
                                          DateFormatUtil.formatDate(lastTimeValidated + currentDeadlinePeriod)),
                    NotificationType.WARNING
                  );
                  myLastOfflineWarningStamp = now;
                }
              }
            }
          }
        }
        catch (VerificationFailedException e) {
          // protocol hijacked?
          action = Action.REVOKE;
          problemReason = e.getMessage();
        }
        if (action == Action.REVOKE) {
          final String message = problemReason;
          app.invokeLater(() -> {
            if (myLicenseRef.compareAndSet(myLicenseData, null)) {
              myLicenseManager.updateLicenseAbout();
              // makes sense only if checked license is current effective license
              cancelRemoteValidation();
              final ProductDescriptor pd = JetprofileSource.this.factory.getProductDescriptor();
              clearStoredLicenseInfo(pd);
              myLicenseManager.demandLicense(pd, LicenseManager.StartupAction.EMPTY, message);
            }
          }, ModalityState.NON_MODAL);
        }
        else if (action == Action.WARN) {
          LicenseEntryUtil.showLicenseNotificationLater(LicenseBundle.message("notification.title.license.validation"), problemReason, NotificationType.WARNING);
        }
      }
      catch (Throwable e) {
        LOG.debug("Unexpected error checking online license: ", e);
      }
      finally {
        myOnlineValidationInProgress.set(false);
      }
    }
  }
}
