package com.intellij.ide.license.impl;

import com.intellij.ide.license.LicenseFactory;
import com.intellij.openapi.application.PathManager;
import com.intellij.openapi.application.ex.ApplicationInfoEx;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.util.io.FileUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.io.*;
import java.util.Calendar;
import java.util.Date;

@SuppressWarnings({"HardCodedStringLiteral"})
public abstract class LicenseFactoryImpl implements LicenseFactory {

  public LicenseFactoryImpl() {
  }

  /**
   * Returns version number in the form, suitable for sending to JBA or floating license server
   * @param baseLineVersion
   * assumes that baseline version adheres to format yyN, where
   * 'yy' is short year number, and N is version nuber within the year.
   *
   * Examples: baseline version 163 corresponds to internal version 2016300
   * 170 -> 2017000
   */
  public static int getInternalVersionNumber(final int baseLineVersion) {
    return 2000000 + baseLineVersion * 100;
  }


  @Override
  public ServerLicenseData createServerLicense(String ticketId,
                                               long prolongationPeriod,
                                               String serverUrl,
                                               String licensee,
                                               int licenseType,
                                               String confirmationStamp, String metadata) {
    return new ServerLicenseData(ticketId, prolongationPeriod, serverUrl, licensee, licenseType, confirmationStamp, metadata);
  }


  @Override
  @SuppressWarnings({"IOResourceOpenedButNotSafelyClosed"})
  @Nullable
  public EvaluationLicenseData findExistingEvaluationLicense() {
    if (getProductDescriptor().isEvalSupported()) {
      final File evalLicenseFile = evalLicenseFile();
      if (evalLicenseFile.exists()) {
        try {
          DataInputStream stream = new DataInputStream(new FileInputStream(evalLicenseFile));
          final long date = ~stream.readLong();
          stream.close();
          return createEvaluationLicenseData(date);
        }
        catch (IOException e) {
          return null;
        }
      }
    }

    return null;
  }

  protected abstract EvaluationLicenseData createEvaluationLicenseData(long date);


  @Override
  @NotNull
  public EvaluationLicenseData createEvaluationLicense() {
    final long currentTime = new Date().getTime();
    return createEvaluationLicenseData(currentTime);
  }

  @Override
  public boolean storeEvaluationLicense(EvaluationLicenseData license) {
    if (!getProductDescriptor().isEvalSupported()) {
      return false;
    }
    try {
      final File file = evalLicenseFile();
      FileUtil.createIfDoesntExist(file);
      try (DataOutputStream stream = new DataOutputStream(new FileOutputStream(file))) {
        stream.writeLong(~license.getGenerationDate().getTime());
      }
      return true;
    }
    catch (IOException e) {
      JOptionPane.showMessageDialog(null, "Error saving evaluation license: " + e.getMessage());
      return false;
    }
  }

  protected File evalLicenseFile() {
    final ApplicationInfoEx app_info = getAppInfo();
    final String eval_id = app_info.isEAP() && !isForceReleaseEvaluationSelector()?
                           app_info.getBuild().asString():
                           getCurrentEvaluationSelector();
    return new File(
      PathManager.getConfigPath() + File.separator + "eval" + File.separator + getProductDescriptor().getShortName() + eval_id + ".evaluation.key"
    );
  }

  protected static ApplicationInfoEx getAppInfo() {
    return ApplicationInfoImpl.getShadowInstance();
  }

  @NotNull
  protected Date getProductBuildDate() {
    final Calendar date = getAppInfo().getMajorReleaseBuildDate();
    return date.getTime();
  }
}
