package com.intellij.ide.license.impl;

import com.intellij.ide.license.LicenseCreationException;
import com.intellij.ide.license.ProductDescriptor;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.EnvironmentUtil;
import com.jetbrains.ls.responses.EncodedAsset;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.*;
import java.util.Base64;
import java.util.Date;
import java.util.StringTokenizer;

public abstract class LicenseFile implements LicenseStore {
  private static final Logger LOG = Logger.getInstance("#LicenseFile");
  private static final String LS_URL_ENVIRONMENT_VAR = "JETBRAINS_LICENSE_SERVER";
  protected static final String JET_PROFILE_DATA_MARKER = "JetProfile\n";
  protected static final String CERT_KEY_SIGNATURE = "<certificate-key>\n";

  @NotNull
  private final ProductDescriptor myProductDescriptor;
  protected final String myReadPath;
  private final String myWritePath;

  private static final String URL_PREFIX = "URL:";

  protected LicenseFile(@NotNull ProductDescriptor productDescriptor, @NonNls @NotNull String storePath) {
    this(productDescriptor, storePath, storePath);
  }
  
  protected LicenseFile(@NotNull ProductDescriptor productDescriptor, @NonNls String readPath, @NonNls String writePath) {
    myProductDescriptor = productDescriptor;
    myReadPath = readPath;
    myWritePath = writePath;
  }

  @Override
  public boolean exists() {
    return new File(myWritePath).exists();
  }

  @Override
  public boolean delete() {
    return FileUtil.delete(new File(myWritePath));
  }

  @Override
  public final void save(LicenseData license) throws TargetSaveException {
    try (OutputStream out = new BufferedOutputStream(new FileOutputStream(myWritePath))){
      out.write(255);
      out.write(255);
      if (license instanceof ServerLicenseData) {
        writeString(out, URL_PREFIX + ((ServerLicenseData)license).getServerUrl());
      }
      else if (license instanceof JBALicenseData){
        writeString(out, JET_PROFILE_DATA_MARKER);
        writeString(out, Base64.getMimeEncoder().encodeToString(getAssetBytes(((JBALicenseData)license))));
      }
      else if (license instanceof CertKeyLicense) {
        writeString(out, CERT_KEY_SIGNATURE);
        writeString(out, license.getKey());
      }
    }
    catch (Throwable e) {
      LOG.info("Error saving license data: " + e.getMessage());
      throw new TargetSaveException(e) {
        @Override
        public synchronized Throwable fillInStackTrace() {
          return this;
        }
      };
    }
  }

  private static byte[] getAssetBytes(JBALicenseData licenseData) throws IOException {
    final ByteArrayOutputStream out = new ByteArrayOutputStream();
    DataOutputStream dataOut = new DataOutputStream(out);
    try {
      dataOut.writeUTF(licenseData.getUserId());
      String login = licenseData.getUserLogin();
      dataOut.writeUTF(login != null? login: "");
      String assetData = licenseData.getEncodedAsset().getData();
      dataOut.writeUTF(assetData != null? assetData : "");
    }
    finally {
      dataOut.close();
    }
    return out.toByteArray();
  }

  @Nullable
  protected LicenseData parseKeyLicenseData(String text) {
    if (text.startsWith(JET_PROFILE_DATA_MARKER)) {
      final String data = text.substring(JET_PROFILE_DATA_MARKER.length());
      return readJBALicense(data, ApplicationInfoImpl.getShadowInstance().getMajorReleaseBuildDate().getTime());
    }
    if (text.startsWith(CERT_KEY_SIGNATURE)) {
      final String key = text.substring(CERT_KEY_SIGNATURE.length());
      return readCertKeyLicense(key, ApplicationInfoImpl.getShadowInstance().getMajorReleaseBuildDate().getTime());
    }
    return null;
  }

  protected static void writeString(OutputStream out, String s) throws IOException {
    for (int i = 0; i < s.length(); i++) {
      char c = s.charAt(i);
      byte b0 = (byte)(c & 0xFF);
      byte b1 = (byte)((c >> 8) & 0xFF);
      out.write(b0);
      out.write(b1);
    }
  }

  @Override
  @Nullable
  public final LicenseData getLicense() {
    try {
      return parseKeyLicenseData(readFile());
    }
    catch (FileNotFoundException ignored) {
    }
    catch (Throwable e) {
      LOG.info(e.getMessage());
    }
    return null;
  }

  @Override
  @Nullable
  public final ServerLicenseData readServerLicenseData() {
    try {
      return parseServerLicenseData(readFile());
    }
    catch (IOException ignored) {
    }
    catch (Throwable e) {
      // not trying fallback logic for unexpected exceptions
      return null;
    }
    // fallback: 'silent-install' support
    // only check environment variable or a system property when license file does not exist or is empty
    String fromEnv = EnvironmentUtil.getValue(LS_URL_ENVIRONMENT_VAR);
    if (fromEnv == null || fromEnv.isEmpty()) {
      fromEnv = System.getProperty(LS_URL_ENVIRONMENT_VAR, null);
    }
    if (fromEnv != null) {
      fromEnv = fromEnv.trim();
      if (!fromEnv.isEmpty()) {
        final String lowercased = StringUtil.toLowerCase(fromEnv);
        if (lowercased.startsWith("http://") || lowercased.startsWith("https://")) {
          return ServerLicenseData.restoreFromFile(fromEnv);
        }
      }
    }
    return null;
  }

  @Nullable
  protected ServerLicenseData parseServerLicenseData(String text) {
    final StringTokenizer tokenizer = new StringTokenizer(text, "\r\n");
    if (!tokenizer.hasMoreTokens()) {
      return null;
    }
    final String serverUrl = tokenizer.nextToken();
    if (serverUrl != null && serverUrl.startsWith(URL_PREFIX)) {
      return ServerLicenseData.restoreFromFile(serverUrl.substring(URL_PREFIX.length()));
    }
    return null;
  }

  private String readFile() throws IOException {
    try (final InputStream in = new BufferedInputStream(new FileInputStream(myReadPath))) {
      int b0 = in.read();
      int b1 = in.read();
      StringBuilder buffer = new StringBuilder();
      while (true) {
        b0 = in.read();
        if (b0 == -1) break;
        b1 = in.read();
        if (b1 == -1) break;
        buffer.append((char)((b1 << 8) + b0));
      }
      return buffer.toString();
    }
  }

  protected CertKeyLicenseData readCertKeyLicense(String text, Date buildDate) {
    try {
      return new CertKeyLicenseData(text, myProductDescriptor.getProductCode(), buildDate);
    }
    catch (LicenseCreationException e) {
      final String details = e.getDetails();
      final StringBuilder msg = new StringBuilder(e.getMessage());
      if (details != null) {
        msg.append(": ").append(details);
      }
      LOG.info(msg.toString());
    }
    return null;
  }

  protected JBALicenseData readJBALicense(String text, Date buildDate) {
    try {
      DataInputStream in = new DataInputStream(new ByteArrayInputStream(Base64.getMimeDecoder().decode(text)));
      try {
        final String userId = in.readUTF();
        final String userLogin = in.readUTF();
        final String assetData = in.readUTF();
        return JetprofileSource.createLicenseData(
          userLogin, userId, myProductDescriptor, buildDate, new EncodedAsset(assetData)
        );
      }
      finally {
        in.close();
      }
    }
    catch (LicenseCreationException e) {
      final String details = e.getDetails();
      final StringBuilder msg = new StringBuilder(e.getMessage());
      if (details != null) {
        msg.append(": ").append(details);
      }
      LOG.info(msg.toString());
    }
    catch (Throwable e) {
      LOG.info("License data corrupted");
    }
    return null;
  }
}
