package com.intellij.ide.license.impl;

import com.intellij.ide.license.LicenseFactory;
import com.intellij.ide.license.LicenseMessages;
import com.intellij.ide.license.LicenseTarget;
import com.intellij.ide.license.PluginLicenseFactory;
import com.intellij.ide.license.ui.HtmlHelp;
import com.intellij.ide.misc.CrlResource;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.licensecommon.util.SubscriptionVersion;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.diagnostic.Logger;
import com.jetbrains.license.util.*;
import com.jetbrains.ls.responses.ObtainTicketResponse;
import com.jetbrains.ls.responses.ProlongTicketResponse;
import com.jetbrains.ls.responses.ReleaseTicketResponse;
import com.jetbrains.ls.responses.ResponseCode;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.util.concurrent.TimeUnit;


public class LicenseServerSource extends ServerSource implements LicenseTarget{
  private static final Logger LOG = Logger.getInstance("#ServerLicenseSource");
  private final int myEncodedBuildDate;
  private final PluginsAwareLicenseManager myManager;
  private final boolean mySendBuildDateAsVersionToOldServers;
  private final LicenseStore myLicenseStore;

  public LicenseServerSource(@NotNull LicenseFactory factory, PluginsAwareLicenseManager manager) {
    this(factory, manager, factory.getCurrentMajorVersion(), false);
  }

  private LicenseServerSource(@NotNull LicenseFactory factory, PluginsAwareLicenseManager manager, int productVersion, boolean sendBuildDateAsVersionToOldServers) {
    super(factory, productVersion, manager);
    myManager = manager;
    mySendBuildDateAsVersionToOldServers = sendBuildDateAsVersionToOldServers;
    myEncodedBuildDate = SubscriptionVersion.createVersion(
      factory instanceof PluginLicenseFactory?
        ((PluginLicenseFactory)factory).getPluginReleaseDate() :
        ApplicationInfoImpl.getShadowInstance().getMajorReleaseBuildDate().getTime()
    );
    myLicenseStore = factory.getLicenseStore();
    VerifyingSignatureFactory.setCrl(CrlResource.createDefault().getData()); // this will initialize additional CRL
  }

  @Override
  protected String getMachineId() {
    return ClientUtil.getMachineId();
  }

  public String getLastUsedServerUrl() {
    final String url = getServerUrl(getLicense());
    return url != null? url : getServerUrl(myLicenseStore.readServerLicenseData());
  }

  private static String getServerUrl(ServerLicenseData lic) {
    String url = lic != null ? lic.getServerUrl() : null;
    return url != null && !url.isEmpty() ? url : null;
  }

  @Override
  public void restore() {
    final ServerLicenseData restored = myLicenseStore.readServerLicenseData();
    if (restored == null) {
      return;
    }
    final String url = restored.getServerUrl();
    if (url == null || url.isEmpty()) {
      return;
    }
    // Use restored ServerLicenseData and avoid obtaining a ticket from network at this point.
    // The restored data structure will cause the first prolong fail and effectively obtain the ticket from FLS.
    // Potentially time-consuming network operation will happen in a background thread and not on the main thread on startup
    setLicense(restored);
  }

  @Override
  public void clear() {
    final ServerLicenseData license = getLicense();
    String url = null;
    String ticketId = null;
    if (license != null) {
      url = license.getServerUrl();
      ticketId = license.getTicketId();
    }
    releaseServerLicense(url, ticketId);
    deleteStoredTicketInfo();
    setLicense(null);
    if (url != null) {
      myManager.getTokenManager().removeToken(url);
      deleteStoredTokenInfo(url);
    }
  }

  @Override
  public void save(LicenseData license) throws TargetSaveException {
    myLicenseStore.save(license);
  }

  @Override
  @NotNull
  protected ProlongationCode prolongServerLicense(final ServerLicenseData serverLicense) {
    final String serverUrl = serverLicense.getServerUrl();
    try {
      checkUrlAllowed(serverUrl);
      if (!serverLicense.isTicketObtained()) {
        return ProlongationCode.REFUSED;
      }
      final ProlongTicketResponse response = FloatingClientUtil.prolongTicket(serverUrl, serverLicense.getTicketId(), getProductFamilyUUID(), lookupAccessToken(serverUrl).get());
      if (response.getResponseCode() != ResponseCode.OK) {
        return ProlongationCode.REFUSED;
      }
      // additional signature validation as a protection from 'artificially' constructed Response objects
      // The main problem now is that the number of fields used in marshalling/unmarshalling of response objects may be different on server and in the IDE, as they may use different
      // versions of the client_api library. For successful signature check the received xml string and the sting marshalled from the Response object must be identical
      final String signature = response.getSignature();
      Signatures.verify(ResponseMarshaller.marshall(response), signature == null ? "" : signature, VerifyingSignatureFactory.X509_CERTIFICATE_FLOATING_LS);
      serverLicense.setConfirmationStamp(response.getConfirmationStamp());
      return ProlongationCode.OK;
    }
    catch (VerificationFailedException e) {
      return ProlongationCode.VERIFICATION_FAILED;
    }
    catch (Throwable e) {
      final StringBuilder logMessage = new StringBuilder();
      logMessage.append("Error prolonging ticket ").append(serverLicense.getTicketId()).append(" with ").append(serverUrl).append(": ");
      logMessage.append(e.getClass().getName());
      final String message = e.getMessage();
      if (message != null) {
        logMessage.append(": ").append(message);
      }
      LOG.info(logMessage.toString());
      return ProlongationCode.DISCONNECTED;
    }
  }

  @Override
  protected boolean releaseServerLicense(final String serverUrl, final String ticketId) {
    if (serverUrl == null || ticketId == null || ticketId.isEmpty()) {
      return false;
    }
    try {
      final ReleaseTicketResponse response = FloatingClientUtil.releaseTicket(
        serverUrl, ticketId, getProductFamilyUUID(), lookupAccessToken(serverUrl).get(1000, TimeUnit.MILLISECONDS)
      );
      return response.getResponseCode() == ResponseCode.OK;
    }
    catch (VerificationFailedException e) {
      return false;
    }
    catch (Throwable e) {
      return false;
    }
  }

  public boolean ping(String serverUrl) {
    try {
      checkUrlAllowed(serverUrl);
      return FloatingClientUtil.pingFloating(
        serverUrl, getProductFamilyUUID(), lookupAccessToken(serverUrl).get(1000, TimeUnit.MILLISECONDS)
      ).getResponseCode() == ResponseCode.OK;
    }
    catch (VerificationFailedException e) {
      return false;
    }
    catch (IOException e) {
      return false;
    }
  }

  @Override
  @NotNull
  protected ServerLicenseData obtainTicketFromServer(String serverUrl) throws NoFreeTicketException, AuthRequiredException, ServerConnectionException, ServerProtocolHijacked {
    try {
      checkUrlAllowed(serverUrl);
      final ObtainTicketResponse response = FloatingClientUtil.obtainTicket(
        serverUrl, getProductFamilyUUID(), getProductVersion(), myEncodedBuildDate, mySendBuildDateAsVersionToOldServers, lookupAccessToken(serverUrl).get()
      );
      if (response.getResponseCode() == ResponseCode.OK) {
        // additional signature validation as a protection from 'artificially' constructed Response objects
        // The main problem now is that the number of fields used in marshalling/unmarshalling of response objects may be different on server and in the IDE, as they may use different
        // versions of the client_api library. For successful signature check the received xml string and the sting marshalled from the Response object must be identical
        final String signature = response.getSignature();
        Signatures.verify(ResponseMarshaller.marshall(response), signature == null ? "" : signature, VerifyingSignatureFactory.X509_CERTIFICATE_FLOATING_LS);
        return createLicenseData(
          serverUrl, response.getTicketId(), response.getProlongationPeriod(), response.getConfirmationStamp(), response.parseTicketProperties()
        );
      }
      else if (response.getResponseCode() == ResponseCode.ERROR){
        final String authorizationUrl = response.getAuthorizationUrl();
        final String codeExchangeUrl = response.getCodeExchangeUrl();
        final String clientId = response.getAuthorizationClientId();
        if (authorizationUrl != null && codeExchangeUrl != null && clientId != null) {
          throw new AuthRequiredException(response.getMessage(), authorizationUrl, clientId, codeExchangeUrl);
        }
      }
      throw new NoFreeTicketException(response.getMessage());
    }
    catch (NoFreeTicketException e) {
      throw new NoFreeTicketException(e.getMessage());
    }
    catch (AuthRequiredException e) {
      throw new AuthRequiredException(
        e.getMessage(),
        myManager.getTokenManager().registerAuthorizationAttempt(serverUrl, e.getAuthUrl(), e.getCodeExchangeUrl(), e.getClientId()),
        e.getClientId(),
        e.getCodeExchangeUrl()
      );
    }
    catch (IncorrectLSException e) {
      throw new ServerConnectionException(e.getMessage());
    }
    catch (OutdatedLicenseServerException e){
      final LicenseMessages messages = myManager.getMessages();
      final String msg = "You are using an outdated version of the license server";
      final String url = messages != null ? messages.getPurchaseUrl(factory.getProductDescriptor(), factory instanceof PluginLicenseFactory) : null;
      throw new ServerConnectionException(
        url == null? msg : msg + ", please visit " + HtmlHelp.makeUrl(url + "&outdated-license-server", "this page") + " to learn more."
      );
    }
    catch (VerificationFailedException e) {
      final String message = e.getMessage();
      throw message != null? new ServerProtocolHijacked(message) : new ServerProtocolHijacked();
    }
    catch (IOException e) {
      throw (ServerConnectionException)new ServerConnectionException(
        "Cannot obtain ticket from " + serverUrl + " due to connectivity problem: " + e.getMessage(), e.getClass().getName()
      ).initCause(e);
    }
    catch (Throwable e) {
      // might be thrown from inside jdk's networking code, e.g. when URL to license server is incorrect
      throw new ServerConnectionException("Cannot obtain ticket from " + serverUrl + ": " + e.getMessage(), e.getClass().getName());
    }
  }

}
