package com.intellij.ide.license.impl;

import com.github.markusbernhardt.proxy.ProxySearch;
import com.github.markusbernhardt.proxy.selector.fixed.FixedProxySelector;
import com.github.markusbernhardt.proxy.util.PlatformUtil;
import com.github.markusbernhardt.proxy.util.ProxyUtil;
import com.intellij.diagnostic.LoadingState;
import com.intellij.ide.license.LicenseBundle;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.ui.DialogWrapper;
import com.intellij.openapi.util.Comparing;
import com.intellij.openapi.util.NotNullLazyValue;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.Trinity;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.SmartList;
import com.intellij.util.net.AuthenticationDialog;
import com.intellij.util.net.AuthenticationPanel;
import com.intellij.util.proxy.CommonProxy;
import com.intellij.util.proxy.NonStaticAuthenticator;
import com.intellij.util.proxy.SharedProxyConfig;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.awt.*;
import java.io.IOException;
import java.net.*;
import java.util.List;
import java.util.*;

public final class LicensingProxyService {
  private static final Logger LOG = Logger.getInstance("#InitialProxyService");
  private static final String SERVICE_ID = "jba.proxy.detect.service";

  private static final String PROXY_PROPERTY = "jba.http.proxy";
  // the value to be set to PROXY_PROPERTY to disable proxy discovery and force direct connections to JBA and FLS
  private static final String DIRECT_CONN = "direct";

  @Nullable
  private Pair<String, String> myProxyConfig; // host & port
  @Nullable
  private Pair<String, String> myDiscoveredConfig; // host & port
  @Nullable
  private Pair<String, char[]> myProxyAuth; // login & password

  private volatile boolean myUseAutodiscoveredConfig = true;

  private static final class InstanceHolder {
    static final LicensingProxyService ourInstance = new LicensingProxyService();
  }

  private final NotNullLazyValue<ProxySelector> mySelectorRef = NotNullLazyValue.atomicLazy(() -> {
    final String fixedProxy = System.getProperty(PROXY_PROPERTY, null);
    final FixedProxySelector fixedProxySelector =
      DIRECT_CONN.equalsIgnoreCase(fixedProxy) ? new FixedProxySelector(Proxy.NO_PROXY) : ProxyUtil.parseProxySettings(fixedProxy);
    if (fixedProxySelector != null) {
      // if proxy for jba is directly set, just use it
      return new CompoundProxySelector(Collections.singletonList(fixedProxySelector));
    }

    final List<ProxySearch.Strategy> availableStrategies = new ArrayList<>(5);

    // Browser settings strategy
    //proxySearch.addStrategy(ProxySearch.Strategy.BROWSER);
    // the PlatformUtil.getDefaultBrowser() implementation is not that smart:
    // for WIN platform it will always suggest IE, even if default is FireFox, so we have to add FireFox strategy manually
    if (PlatformUtil.getCurrentPlattform() == PlatformUtil.Platform.WIN) {
      availableStrategies.add(ProxySearch.Strategy.IE);
      availableStrategies.add(ProxySearch.Strategy.FIREFOX);
    }

    availableStrategies.add(ProxySearch.Strategy.OS_DEFAULT);
    availableStrategies.add(ProxySearch.Strategy.ENV_VAR);

    final List<ProxySelector> selectors = new SmartList<>();
    for (ProxySearch.Strategy strategy : availableStrategies) {
      // search one-by-one, as ProxySearch implementation stops at first found selector,
      // but we should examine all possible cases
      try {
        final ProxySearch proxySearch = new ProxySearch();
        proxySearch.addStrategy(strategy);
        final ProxySelector selector = proxySearch.getProxySelector();
        if (selector != null) {
          selectors.add(selector);
        }
      }
      catch (Throwable e) {
        LOG.info(e.getClass().getName() + ": " + e.getMessage());
      }
    }

    return new CompoundProxySelector(selectors);
  });

  private LicensingProxyService() {
    CommonProxy.isInstalledAssertion();
    CommonProxy.getInstance().setCustom(SERVICE_ID, new ProxySelector() {
      @Override
      public List<Proxy> select(URI uri) {
        if (isActive()) {
          return mySelectorRef.getValue().select(uri);
        }
        unregisterCustomSelector();
        return Collections.emptyList();
      }

      @Override
      public void connectFailed(URI uri, SocketAddress sa, IOException ioe) {
        if (isActive()) {
          mySelectorRef.getValue().connectFailed(uri, sa, ioe);
        }
        else {
          unregisterCustomSelector();
        }
      }
    });
    CommonProxy.getInstance().setCustomAuth(SERVICE_ID, new LicensingProxyAuthenticator());
  }

  public static boolean isActive() {
    return !LoadingState.COMPONENTS_LOADED.isOccurred();
  }

  private void unregisterCustomSelector() {
    final CommonProxy commonProxy = CommonProxy.getInstance();
    commonProxy.removeCustom(SERVICE_ID);
    commonProxy.removeCustomAuth(SERVICE_ID);
    synchronized (this) {
      myProxyAuth = null;
      myProxyConfig = null;
      myDiscoveredConfig = null;
    }
  }

  public static LicensingProxyService getInstance() {
    return InstanceHolder.ourInstance;
  }

  public void setUserDefinedProxy(String host, String port) {
    setUserDefinedProxy(host, port, null);
  }

  public synchronized void setUserDefinedProxy(String host, String port, @Nullable Pair<String, char[]> auth) {
    final Pair<String, String> cfg = getPair(host, port);
    if (!Comparing.equal(myProxyConfig, cfg) || !Comparing.equal(myProxyAuth, auth)) {
      myUseAutodiscoveredConfig = false; // once user configured proxy explicitly, ignore auto-discovered data
      myProxyConfig = cfg;
      myDiscoveredConfig = null;
      myProxyAuth = auth;
      updateSharedProxyData();
    }
  }

  @Nullable
  private static Pair<String, String> getPair(@Nullable String host, @Nullable String port) {
    return StringUtil.isEmptyOrSpaces(host) || StringUtil.isEmptyOrSpaces(port)? null : Pair.create(host.trim(), port.trim());
  }

  @Nullable
  private static Pair<String, String> getPair(@Nullable String host, int port) {
    return getPair(host, String.valueOf(port));
  }

  @Nullable
  private static Pair<String, char[]> getPair(@Nullable String login, char @NotNull [] password) {
    return StringUtil.isEmptyOrSpaces(login)? null : Pair.create(login.trim(), password);
  }

  private void updateSharedProxyData() {
    try {
      final Pair<String, String> hostPort = getProxyConfig();
      if (hostPort != null) {
        final Pair<String, char[]> auth = getProxyAuth();
        final String login = Pair.getFirst(auth);
        final char[] password = auth != null? auth.second : new char[0];
        SharedProxyConfig.store(new SharedProxyConfig.ProxyParameters(hostPort.first, Integer.parseInt(hostPort.second), login, password));
      }
      else {
        SharedProxyConfig.clear();
      }
    }
    catch (NumberFormatException ignored) {
    }
  }

  @Nullable
  public synchronized Pair<String, String> getProxyConfig() {
    return myProxyConfig;
  }

  @Nullable
  public synchronized Pair<String, char[]> getProxyAuth() {
    return myProxyAuth;
  }

  @Nullable
  public synchronized Pair<String, String> getDiscoveredConfig() {
    return myDiscoveredConfig;
  }

  private synchronized void proxyDiscovered(Proxy proxy) {
    if (myDiscoveredConfig == null) {
      final SocketAddress address = proxy.address();
      if (address instanceof InetSocketAddress) {
        final InetSocketAddress inetAddress = (InetSocketAddress)address;
        myDiscoveredConfig = getPair(inetAddress.getHostName(), inetAddress.getPort());
      }
    }
  }

  @Nullable
  private synchronized Proxy getUserDefinedProxy() {
    final Pair<String, String> hostPort = myProxyConfig;
    if (hostPort != null) {
      try {
        return new Proxy(Proxy.Type.HTTP, InetSocketAddress.createUnresolved(hostPort.first, Integer.parseInt(hostPort.second)));
      }
      catch (NumberFormatException ignored) {
      }
      catch (Throwable e) {
        LOG.info(e.getClass().getName() + ": " + e.getMessage());
      }
    }
    return null;
  }

  private final class CompoundProxySelector extends ProxySelector {
    private final List<ProxySelector> myDelegates;

    CompoundProxySelector(@NotNull List<ProxySelector> delegates) {
      myDelegates = delegates;
    }

    @Override
    public List<Proxy> select(URI uri) {
      final List<Proxy> proxyList = new SmartList<>();
      final Set<Proxy> foundProxies = new HashSet<>();

      final String protocol = uri.getScheme();
      if ("https".equalsIgnoreCase(protocol) || "http".equalsIgnoreCase(protocol)) {
        final Proxy userDefined = getUserDefinedProxy();
        if (userDefined != null && foundProxies.add(userDefined)) {
          proxyList.add(userDefined);
        }
      }

      if (myUseAutodiscoveredConfig) {
        for (ProxySelector delegate : myDelegates) {
          try {
            final List<Proxy> proxies = delegate.select(uri);
            if (proxies != null) {
              for (Proxy proxy : proxies) {
                if (proxy.type() == Proxy.Type.HTTP && foundProxies.add(proxy)) {
                  proxyDiscovered(proxy);
                  proxyList.add(proxy);
                }
              }
            }
          }
          catch (Throwable e) {
            LOG.info(e.getClass().getName() + ": " + e.getMessage());
          }
        }
      }

      if (proxyList.isEmpty()) {
        proxyList.add(Proxy.NO_PROXY);
      }
      return proxyList;
    }

    @Override
    public void connectFailed(URI uri, SocketAddress sa, IOException ioe) {
      final Pair<String, String> config = getProxyConfig();
      if (config != null && sa instanceof InetSocketAddress) {
        final InetSocketAddress addr = (InetSocketAddress)sa;
        if (StringUtil.equalsIgnoreCase(config.first, addr.getHostName())) {
          SharedProxyConfig.clear();
        }
      }
      for (ProxySelector delegate : myDelegates) {
        try {
          delegate.connectFailed(uri, sa, ioe);
        }
        catch (Throwable e) {
          LOG.info(e.getClass().getName() + ": " + e.getMessage());
        }
      }
    }
  }

  private class LicensingProxyAuthenticator extends NonStaticAuthenticator {
    @Override
    public PasswordAuthentication getPasswordAuthentication() {
      if (!isActive()) {
        unregisterCustomSelector();
        return null;
      }
      try {
        if (!Authenticator.RequestorType.PROXY.equals(getRequestorType())) {
          return null;
        }
        Pair<String, String> hostPort = getProxyConfig();
        if (hostPort == null) {
          hostPort = getDiscoveredConfig();
        }
        final String requestHost = CommonProxy.getHostNameReliably(getRequestingHost(), getRequestingSite(), getRequestingURL());
        final int requestPort = getRequestingPort();
        if (hostPort == null || !StringUtil.equalsIgnoreCase(hostPort.first, requestHost) || Integer.parseInt(hostPort.second) != requestPort) {
          return null;
        }
        Pair<String, char[]> auth = getProxyAuth();
        if (auth == null && !GraphicsEnvironment.isHeadless()) {
          final Trinity<String, char[], Boolean> answer = promptPassword(requestHost + ":" + requestPort);
          if (answer != null) {
            auth = getPair(answer.getFirst(), answer.getSecond());
            setUserDefinedProxy(requestHost, String.valueOf(requestPort), answer.getThird()/*should remember*/? auth : null);
          }
        }
        return auth != null? new PasswordAuthentication(auth.first, auth.second) : null;
      }
      catch (NumberFormatException ignored) {
      }
      return null;
    }

    @Nullable
    private Trinity<String, char[], Boolean> promptPassword(final String requestAddress) {
      return UIUtil.invokeAndWaitIfNeeded(() -> {
        Component parent = null;
        for (Window win : Window.getWindows()) {
          if (win.isActive() && win.isShowing()) {
            parent = win;
            break;
          }
          if (parent == null && win.isShowing()) {
            parent = win;
          }
        }
        if (parent == null) {
          return null; // unable to show prompt dialog
        }
        AuthenticationDialog dialog = new AuthenticationDialog(
          parent,
          LicenseBundle.message("dialog.title.proxy.authentication"),
          LicenseBundle.message("dialog.message.please.enter.credentials.for", requestAddress), "", "", false
        );
        dialog.show();
        if (dialog.getExitCode() == DialogWrapper.OK_EXIT_CODE) {
          final AuthenticationPanel panel = dialog.getPanel();
          return Trinity.create(panel.getLogin(), panel.getPassword(), panel.isRememberPassword());
        }
        return null;
      });
    }

  }
}
