package com.intellij.ide.license.impl;

import com.intellij.ide.license.*;
import com.intellij.ide.license.common.PluginLicenseFactoryImpl;
import com.intellij.ide.license.ui.LicenseEditingModel;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.ide.license.ui.LicenseUtils;
import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.ide.plugins.PluginManagerCore;
import com.intellij.idea.Main;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.notification.NotificationType;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.application.ex.ApplicationInfoEx;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.extensions.PluginId;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.text.DateFormatUtil;
import com.jetbrains.license.util.IteratorUtil;
import com.jetbrains.license.util.Pair;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Predicate;

/**
 * @author Eugene Zhuravlev
 * Date: 02-Apr-19
 */
public abstract class PluginsAwareLicenseManager extends LicenseManager {
  protected static final Logger LOG = Logger.getInstance("#LicenseManager");

  protected final ExpirationDateCheck myExpirationDateCheck;
  protected final List<PluginProductKey> myPluginKeyData = new ArrayList<>();
  private volatile boolean myShutdownStarted = false;
  private final AtomicBoolean myDialogRequested = new AtomicBoolean(false);
  private final AuthTokenManager myTokenManager;
  private static final Set<Integer> ourJBFatPlugins;

  static {
    // avoid storing plugin product codes as plain strings
    Set<Integer> result = new HashSet<>();
    for (String code : Arrays.asList("PDB", "PWS", "PGO", "PPS", "PPC", "PRB", "PSW")) {
      result.add(code.hashCode());
    }
    ourJBFatPlugins = result;
  }

  protected PluginsAwareLicenseManager(ExpirationDateCheck expCheck) {
    myExpirationDateCheck = expCheck;
    myTokenManager = new AuthTokenManager();
  }

  protected void shutdownApplication() {
    myShutdownStarted = true;
    Application app = ApplicationManager.getApplication();
    if (app == null) {
      System.exit(Main.LICENSE_ERROR);
    }
    else {
      app.invokeLater(() -> app.exit(true, true, false), ModalityState.NON_MODAL);
    }
  }

  public void scheduleLicenseExpirationCheckNow() {
    myExpirationDateCheck.scheduleValidation(DateFormatUtil.MINUTE, this);
  }

  @Override
  public final @NotNull Iterable<LicenseInfo> getPluginLicenseInfo() {
    if (myPluginKeyData.isEmpty()) {
      return Collections.emptyList();
    }
    return () -> IteratorUtil.map(
      myPluginKeyData.iterator(), pk -> new LicenseInfo(pk.getProductDescriptor(), pk.getActiveSource(), true, pk.getPluginDescriptor().isLicenseOptional())
    );
  }

  @Override
  public void startUp(@NotNull StartupAction action) {
    initPluginDescriptors();
    // initially on start
    for (PluginProductKey ppk : myPluginKeyData) {
      if (!ppk.getPluginDescriptor().isEnabled() || ppk.getLicense() != null) {
        continue;
      }
      for (LicenseSource<?> src : getSortedLicenseSources(ppk.getSources())) {
        if (src.getLicense() == null) {
          src.restore();
        }
        final LicenseData licenseData = src.getLicense();
        if (licenseData != null) {
          ppk.setActiveSource(src, false);
          break;
        }
      }
    }

    final StartupAction wrapped = new StartupAction() {
      @Override
      public void proceed() {
        try {
          action.proceed();
        }
        finally {
          // register periodic task cleanup code once
          LicenseEntryUtil.executeOnApplicationExit(()-> {
            for (PluginProductKey ppk : myPluginKeyData) {
              final ServerSource serverSource = ppk.findSource(ServerSource.class);
              if (serverSource != null) {
                serverSource.stopProlongTask();
              }
            }
          });
        }
      }
    };

    // validate and, if possible, reuse licenses for plugins
    final Collection<PluginProductKey> unlicensedPlugins = revalidatePluginLicenses(true, null);

    if (!unlicensedPlugins.isEmpty()) {
      myExpirationDateCheck.scheduleValidation(15 * DateFormatUtil.MINUTE, this);
      demandLicense(unlicensedPlugins.iterator().next().getProductDescriptor(), wrapped, null);
    }
    else {
      proceed(wrapped);
    }
  }

  protected final void initPluginDescriptors() {
    ApplicationInfoEx appInfo = ApplicationInfoImpl.getShadowInstance();
    for (IdeaPluginDescriptor descriptor : PluginManagerCore.getEnabledPluginRawList()) {
      if (descriptor.getProductCode() != null && !(ourJBFatPlugins.contains(descriptor.getProductCode().hashCode()) && descriptor.getPluginId() != null && appInfo.isEssentialPlugin(descriptor.getPluginId()))) {
        myPluginKeyData.add(new PluginProductKey(descriptor));
      }
    }
  }


  // returns plugins still requiring licensing
  protected Collection<PluginProductKey> revalidatePluginLicenses(boolean reuseExistingLicensesIfPossible, final Predicate<? super LicenseData> licenseDataValidator) {
    if (myPluginKeyData.isEmpty()) {
      return Collections.emptyList();
    }

    if (reuseExistingLicensesIfPossible) {
      final List<PluginProductKey> missingLicense = new ArrayList<>();
      final List<LicenseData> availableLicenses = new ArrayList<>();
      final LicenseData mainLicense = getPlatformLicenseInfo().getLicense();
      if (mainLicense instanceof CertKeyLicense) {
        availableLicenses.add(mainLicense);
      }
      // first, collect available licenses and determine plugins with missing licenses
      for (PluginProductKey ppk : myPluginKeyData) {
        if (ppk.getPluginDescriptor().isEnabled()) {
          final LicenseData lic = ppk.getLicense();
          if (lic == null) {
            missingLicense.add(ppk);
          }
          else {
            if (lic instanceof CertKeyLicense) {
              availableLicenses.add(lic);
            }
          }
        }
      }
      // for all plugins with missing licenses check if any of available licenses can be used
      for (final PluginProductKey ppk : missingLicense) {
        for (LicenseData lic : availableLicenses) {
          if (reuseLicenseIfApplicable(ppk, lic)) {
            break;
          }
        }
      }
    }
    // last, validate licenses for all plugins
    final List<PluginProductKey> unlicensed = new ArrayList<>();
    for (PluginProductKey ppk : myPluginKeyData) {
      if (!ppk.getPluginDescriptor().isEnabled()) {
        continue;
      }
      final LicenseSource<?> src = ppk.getActiveSource();
      final Status status = src == null ? Status.ASK : validatePluginLicenseSource(ppk, src, licenseDataValidator);
      if (!status.isLicensed()) {
        if (!ppk.getPluginDescriptor().isLicenseOptional()) {
          unlicensed.add(ppk);
        }
      }
    }
    return unlicensed;
  }

  @NotNull
  protected Status validatePluginLicenseSource(PluginProductKey ppk, final LicenseSource<?> source, Predicate<? super LicenseData> licValidator) {
    LicenseData license = null;
    try {
      license = source.getLicense();
      if (license == null || (licValidator != null && !licValidator.test(license))) {
        return Status.ASK;
      }
      source.validate();
    }
    catch (EvaluationExpiredException e) {
      LOG.info(ppk.getProductName() + ": " + e.getMessage());
      return myExpirationDateCheck.isFreeRunActive(ppk.getProductCode(), license)? Status.FREE_RUN : Status.ASK;
    }
    catch (InvalidLicenseException e) {
      LOG.info(ppk.getProductName() + ": Invalid license data " + e.getMessage());
      return Status.ASK;
    }
    catch (LicenseExpiredException e) {
      LOG.info(ppk.getProductName() + ": License expired " + e.getMessage());
      return myExpirationDateCheck.isFreeRunActive(ppk.getProductCode(), license)? Status.FREE_RUN : Status.ASK;
    }
    catch (ValidationFailedException e) {
      LOG.info(ppk.getProductName() + ": License validation failed " + e.getMessage());
      return Status.ASK;
    }
    return Status.OK;
  }

  private boolean reuseLicenseIfApplicable(@NotNull PluginProductKey ppk, @NotNull LicenseData license) {
    if (license instanceof CertKeyLicense) {
      final CertKey2 key = ((CertKeyLicense)license).getCertKey();
      if (key != null && Boolean.TRUE.equals(key.eval(new IsApplicableTo(ppk.getProductCode())))) {
        if (license instanceof JetProfileCertLicenseData) {
          final JetprofileSource jbaSource = ppk.findSource(JetprofileSource.class);
          if (jbaSource != null && jbaSource.getLicense() == null) {
            try {
              final JetProfileCertLicenseData data = (JetProfileCertLicenseData)license;
              jbaSource.setLicense(new JetProfileCertLicenseData(
                data.getUserId(), data.getUserLogin(), ppk.getProductCode(), ppk.getReleaseDate(), data.getEncodedAsset(), data.getLicense())
              );
              ppk.setActiveSource(jbaSource, true);
              return true;
            }
            catch (Throwable ignored) {
            }
          }
        }
        else if (license instanceof CertKeyLicenseData){
          // offline license case
          final FileLicenseSource fileSource = ppk.findSource(FileLicenseSource.class);
          if (fileSource != null && fileSource.getLicense() == null) {
            try {
              final CertKeyLicenseData data = (CertKeyLicenseData)license;
              fileSource.setLicense(new CertKeyLicenseData(data.getKey(), ppk.getProductCode(), ppk.getReleaseDate()));
              ppk.setActiveSource(fileSource, true);
              return true;
            }
            catch (Throwable ignored) {
            }
          }
        }
      }
    }
    return false;
  }

  public AuthTokenManager getTokenManager() {
    return myTokenManager;
  }

  @Override
  public boolean supportsRegistration() {
    return !myPluginKeyData.isEmpty();
  }

  protected void proceed(StartupAction action) {
    updateLicenseAbout();
    try {
      action.proceed();
    }
    finally {
      for (ProductKey pk : myPluginKeyData) {
        final LicenseSource<?> src = pk.getActiveSource();
        if (src != null) {
          src.scheduleRemoteValidation();
        }
      }
      myExpirationDateCheck.scheduleValidation(20 * DateFormatUtil.MINUTE, this);
    }
  }

  final void demandLicenseAsyncFromBackgroundTask(Application app, ProductDescriptor pd, @Nullable Pair<@NlsContexts.NotificationTitle String, @NlsContexts.NotificationContent String> titleErrorPair) {
    final String errorMessage = titleErrorPair != null? titleErrorPair.snd : null;
    if (!myDialogRequested.getAndSet(true)) {
      // the invokeLater with NON_MODAL state guarantees that the showLicenseDialog will be executed
      // synchronously from that same Runnable and demandLicense will always return 'true'
      app.invokeLater(() -> {
        demandLicense(pd, StartupAction.EMPTY, errorMessage);
      }, ModalityState.NON_MODAL);
    }
    else if (errorMessage != null){
      final String title = titleErrorPair.fst;
      LicenseEntryUtil.showLicenseNotificationLater(title != null && !title.isEmpty() ? title :
                                                    LicenseBundle.message("notification.title.license.validation"), errorMessage, NotificationType.ERROR);
    }
  }

  /**
   * @return true if the dialog was shown to the user
   *         false would mean that it failed to demand the license and the calling code should try later, if required
   */
  protected final boolean demandLicense(ProductDescriptor pd, StartupAction action, @Nullable @NlsContexts.DetailedDescription String statusMessage) {
    if (pd != null) {
      final String logMessage = "License required for " + pd.getShortName() + "; details: " + (statusMessage == null ? "N/A" : statusMessage);
      AppExecutorUtil.getAppExecutorService().execute(() -> LOG.info(logMessage));
    }
    return LicenseEntryUtil.invokeAndWaitInNonModalContext(
      () -> {
        if (!myShutdownStarted) {
          myDialogRequested.set(true);
          try {
            LicenseUtils.showLicenseDialog(createLicenseEditingModel(pd, action, statusMessage));
          }
          finally {
            myDialogRequested.set(false);
          }
        }
        else {
          shutdownApplication();
        }
      }
    );
  }

  @NotNull
  protected LicenseEditingModel createLicenseEditingModel(ProductDescriptor pd, StartupAction action, @NlsContexts.DetailedDescription @Nullable String statusMessage) {
    return new LicenseEntryModel(pd, action, statusMessage);
  }

  protected static Collection<LicenseSource<?>> getSortedLicenseSources(final Collection<LicenseSource<?>> sources) {
    LicenseSource<?> knownGoodSource = null;
    for (LicenseSource<?> source : sources) {
      try {
        source.restore();
        if (source.getLicense() != null) {
          source.validate();
        }
        else {
          continue;
        }
      }
      catch (Exception e) {
        continue;
      }
      knownGoodSource = source;
      break;
    }

    if (knownGoodSource == null) {
      return Collections.unmodifiableCollection(sources);
    }

    final List<LicenseSource<?>> sorted = new ArrayList<>();
    sorted.add(knownGoodSource);
    for (LicenseSource<?> source : sources) {
      if (source != knownGoodSource) {
        sorted.add(source);
      }
    }
    return Collections.unmodifiableCollection(sorted);
  }

  @Override
  public void register(String productCode, String message) {
    ProductDescriptor pd = ProductDescriptor.findByProductCode(productCode);
    if (pd == null && productCode != null) {
      for (PluginProductKey ppk : myPluginKeyData) {
        if (productCode.equals(ppk.getProductCode())) {
          pd = ppk.getProductDescriptor();
        }
      }
    }
    demandLicense(pd, LicenseManager.StartupAction.EMPTY, message);
  }
  /*
   * @return true if the action succeeded and the dialog was shown to the user
   *         false would mean that it failed to ask the user the license and the calling code should try later
   */


  public int getEvaluationPeriod() {
    return 30;
  }

  protected boolean disablePlugin(ProductDescriptor pd) {
    for (PluginProductKey ppk : myPluginKeyData) {
      if (ppk.getProductDescriptor().equals(pd)) {
        ppk.disablePlugin();
        return true;
      }
    }
    return false;
  }

  protected enum Status {
    ASK, OK, FREE_RUN;

    boolean isLicensed() {
      return this == OK || this == FREE_RUN;
    }
  }

  protected class ProductKey {
    @NotNull
    private final LicenseFactory myLicenseFactory;
    @NotNull
    private final Collection<LicenseSource<?>> mySources;
    @Nullable
    private volatile LicenseSource<?> myActiveSource;

    ProductKey(@NotNull LicenseFactory factory) {
      myLicenseFactory = factory;
      final List<LicenseSource<?>> sources = new ArrayList<>();
      sources.add(new FileLicenseSource(PluginsAwareLicenseManager.this, factory));
      sources.add(new JetprofileSource(factory, PluginsAwareLicenseManager.this, null));
      sources.add(new LicenseServerSource(factory, PluginsAwareLicenseManager.this));
      // important: the eval source should be the last one in the list
      sources.add(new BuiltInEvaluationSource(factory, getEvaluationPeriod()));
      mySources = Collections.unmodifiableList(sources);
    }

    @NotNull
    public ProductDescriptor getProductDescriptor() {
      return myLicenseFactory.getProductDescriptor();
    }

    @NotNull
    public String getProductName() {
      return getProductDescriptor().getShortName();
    }

    @NotNull
    public String getProductCode() {
      return getProductDescriptor().getProductCode();
    }

    @NotNull
    public Date getReleaseDate() {
      return ApplicationInfoImpl.getShadowInstance().getMajorReleaseBuildDate().getTime();
    }

    @NotNull
    public LicenseFactory getLicenseFactory() {
      return myLicenseFactory;
    }

    @Nullable
    public LicenseData getLicense() {
      final LicenseSource<?> source = getActiveSource();
      return source != null? source.getLicense() : null;
    }

    @Nullable
    public LicenseSource<?> getActiveSource() {
      return myActiveSource;
    }

    public void setActiveSource(@Nullable LicenseSource<?> source, boolean saveLicenseData) {
      final LicenseSource<?> previous = myActiveSource;
      myActiveSource = source;
      if (saveLicenseData && source instanceof LicenseTarget) {
        final LicenseData lic = source.getLicense();
        try {
          ((LicenseTarget)source).save(lic);
        }
        catch (TargetSaveException e) {
          LOG.error("License could not be saved due to following error:" + (e.getMessage() != null ? e.getMessage() : "NPE"), e);
          getMessages().showSaveLicenseWarning(getProductDescriptor(), e);
        }
      }
      if (previous != source) {
        if (previous instanceof ServerSource) {
          ((ServerSource)previous).releaseLicense();   // this will return the ticket to the pool of available tickets on the LS
        }
        else if (previous instanceof JetprofileSource) {
          ((JetprofileSource)previous).setLicense(null); // this will prevent sending ValidateLicense requests
        }
        else if (previous instanceof FileLicenseSource) {
          ((FileLicenseSource)previous).setLicense(null); // this will prevent sending ValidateKey requests
        }
      }
    }

    @NotNull
    public Collection<LicenseSource<?>> getSources() {
      return mySources;
    }

    public <T extends LicenseSource<?>> T findSource(Class<T> srcClass) {
      for (LicenseSource<?> source : getSources()) {
        if (srcClass.isAssignableFrom(source.getClass())) {
          return srcClass.cast(source);
        }
      }
      return null;
    }
  }

  protected final class PluginProductKey extends ProductKey {
    private final IdeaPluginDescriptor myPluginDescriptor;

    PluginProductKey(@NotNull IdeaPluginDescriptor descriptor) {
      super(new PluginLicenseFactoryImpl(descriptor));
      myPluginDescriptor = descriptor;
    }

    @NotNull
    public IdeaPluginDescriptor getPluginDescriptor() {
      return myPluginDescriptor;
    }

    @NotNull
    @Override
    public Date getReleaseDate() {
      return ((PluginLicenseFactory)getLicenseFactory()).getPluginReleaseDate();
    }

    void disablePlugin() {
      IdeaPluginDescriptor pd = getPluginDescriptor();
      pd.setEnabled(false);
      PluginId id = pd.getPluginId();
      if (id != null) {
        PluginManagerCore.disablePlugin(id);
      }
    }

    void enablePlugin() {
      final IdeaPluginDescriptor pd = getPluginDescriptor();
      pd.setEnabled(true);
      final PluginId id = pd.getPluginId();
      if (id != null) {
        PluginManagerCore.enablePlugin(id);
      }
    }

    @Override
    public void setActiveSource(@Nullable LicenseSource<?> source, boolean saveLicenseData) {
      try {
        super.setActiveSource(source, saveLicenseData);
      }
      finally {
        if (getLicense() != null) {
          enablePlugin(); // this will just switch plugin state back to enabled, if if was disabled previously in this session
        }
      }
    }
  }

  private class LicenseEntryModel implements LicenseEditingModel {
    private final ProductDescriptor myProductToSelect;
    private final StartupAction myAction;
    private final @NlsContexts.DetailedDescription String myInitialMessage;
    private final Map<ProductDescriptor, PluginProductKey> myPluginProducts = new HashMap<>();

    LicenseEntryModel(ProductDescriptor productToSelect, StartupAction action, @NlsContexts.DetailedDescription String initialMessage) {
      myProductToSelect = productToSelect;
      myAction = action;
      myInitialMessage = initialMessage;
      for (PluginProductKey key : myPluginKeyData) {
        myPluginProducts.put(key.getProductDescriptor(), key);
      }
    }

    @Override
    public ProductDescriptor getInitialProductToSelect() {
      final ProductDescriptor toSelect = myProductToSelect;
      return toSelect != null ? toSelect : getPlatformProduct();
    }

    @Nullable
    @Override
    public String getInitialProblemDescription() {
      return myInitialMessage;
    }

    @Override
    public ProductDescriptor getPlatformProduct() {
      return null;
    }

    @Override
    public Collection<ProductDescriptor> getPluginProducts() {
      return Collections.unmodifiableCollection(myPluginProducts.keySet());
    }

    @Override
    public LicenseInfo getLicenseInfo(ProductDescriptor pd) {
      final PluginProductKey key = myPluginProducts.get(pd);
      return key != null? new LicenseInfo(pd, key.getActiveSource(), true, key.getPluginDescriptor().isLicenseOptional()) : null;
    }

    @NotNull
    @Override
    public Collection<LicenseSource<?>> getAvailableLicenseSources(ProductDescriptor pd) {
      final ProductKey key = myPluginProducts.get(pd);
      return key != null? key.getSources() : Collections.emptyList();
    }

    @Override
    public String getDisplayName(ProductDescriptor pd) {
      String name = null;
      final PluginProductKey key = myPluginProducts.get(pd);
      if (key != null) {
        name = key.getPluginDescriptor().getName();
      }
      return name != null? name : pd.getShortName();
    }

    @Override
    public LicenseInfo activateLicense(ProductDescriptor pd, LicenseSource source) {
      final PluginProductKey key = myPluginProducts.get(pd);
      if (key != null) {
        final Status status = validatePluginLicenseSource(key, source, null);
        if (status.isLicensed()) {
          key.setActiveSource(source, true);
          revalidatePluginLicenses(true, null);
          updateLicenseAbout();
        }
      }
      return getLicenseInfo(pd);
    }

    @Override
    public void forgetLicense(ProductDescriptor pd) {
      Collection<LicenseSource<?>> availableLicenseSources = getAvailableLicenseSources(pd);
      try {
        for (LicenseSource<?> source : availableLicenseSources) {
          try {
            source.clear();
          }
          catch (Throwable ignored) {
          }
        }
        // if there is a non-clearable valid license left (e.g. eval license), return to it
        for (LicenseSource<?> source : availableLicenseSources) {
          source.restore();
          final LicenseData license = source.getLicense();
          if (license != null && !license.isExpired()) {
            activateLicense(pd, source);
            break;
          }
        }
      }
      finally {
        updateLicenseAbout();
      }
    }

    @Override
    public void close(final boolean forceExit) {
      AppExecutorUtil.getAppExecutorService().execute(() -> {
        getTokenManager().clearAuthorizationAttempts();
        if (forceExit) {
          shutdownApplication();
        }
        else {
          try {
            for (PluginProductKey key : revalidatePluginLicenses(false, lic -> !(lic instanceof ServerLicenseData) || ((ServerLicenseData)lic).isTicketObtained())) {
              key.disablePlugin();
            }
          }
          finally {
            proceed(myAction);
          }
        }
      });
    }
  }
}
