package com.intellij.ide.license.impl;

import com.intellij.ide.Prefs;
import com.intellij.ide.license.*;
import com.intellij.ide.license.ui.LicenseEntryUtil;
import com.intellij.ide.misc.BlackListResource;
import com.intellij.licensecommon.util.LicenseType;
import com.intellij.notification.NotificationType;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.PermanentInstallationID;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.BuildNumber;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.openapi.util.NlsSafe;
import com.intellij.openapi.util.Ref;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.containers.ContainerUtil;
import com.jetbrains.license.util.Pair;
import com.jetbrains.license.util.VerificationFailedException;
import com.jetbrains.ls.responses.TicketProperties;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.crypto.Cipher;
import javax.crypto.SecretKeyFactory;
import javax.crypto.spec.IvParameterSpec;
import javax.crypto.spec.PBEKeySpec;
import javax.crypto.spec.SecretKeySpec;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.nio.charset.StandardCharsets;
import java.security.Key;
import java.security.SecureRandom;
import java.util.*;
import java.util.concurrent.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.regex.Pattern;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.InflaterInputStream;

/**
 * @author Eugene Zhuravlev
 */
public abstract class ServerSource extends GeneralLicenseSource<String> {
  private static final Logger LOG = Logger.getInstance("#ServerLicenseSource");

  private static final long HOUR = 60 * 60 * 1000L;
  private static final long SERVER_ABSENSE_TOLERANCE_TIMEOUT = 48L * HOUR;
  private static final long OUTDATED_CERTIFICATE_TOLERANCE_TIMEOUT = 12L * HOUR;
  private static final int DEFAULT_PROLONGATION_PERIOD = 600000; // 10 minutes
  private static final int MAX_PROLONGATION_PERIOD = 30 * 60 * 1000; // 30 minutes
  private static final String TICKET_ID = "ticket.id";
  private static final String SIGNATURE = "signature";
  private static final String MACHINE_ID = "machine.id";
  private static final String TIMESTAMP = "timestamp";
  private static final String PROLONGATION_PERIOD = "prolongation.period";
  private static final String SERVER_URL = "server.url";
  private static final String CONFIRMATION_STAMP = "confirmation.stamp";

  private static final String ACCESS_TOKEN = "access.token";
  private static final String REFRESH_TOKEN = "refresh.token";
  private static final String TOKEN_EXPIRATION = "token.expiration";
  private static final String TOKEN_EXCHANGE_URL = "token.exchange.url";
  private static final String TOKEN_CLIENT_ID = "token.client.id";

  private final AtomicBoolean myReobtainScheduled = new AtomicBoolean(false);
  private final String myProductUUID;
  private final int myProductVersion;
  private final PluginsAwareLicenseManager myManager;

  private volatile long myLastTimeSeenServer = 0;
  private volatile long myVerificationFailureStart = 0;
  private volatile ProlongTaskFuture myProlongTaskFuture;

  private static final BlackListResource ourBlResource = BlackListResource.createDefault();
  private static final String[] URL_BLACKLIST_PATTERNS = {};

  private static final Object ourEncryptionKey;
  private static final Object ourTokensKey;
  static {
    Object key;
    Object tokensKey;
    try {
      final BuildNumber buildInfo = ApplicationInfoImpl.getShadowInstance().getBuild();
      final String platformBuild = buildInfo.asString();
      final int versionSalt = ~buildInfo.getBaselineVersion();
      final byte[] salt = new byte[8];
      for (int i = 0; i < 4; i++ ) {
        salt[i] = (byte)((versionSalt >> 8 * i) & 0xFF);
      }
      for (int i = 4, c = platformBuild.length() - 1; i < 8; i++, c--) {
        if (c < 0) {
          break;
        }
        salt[i] = (byte)(platformBuild.charAt(c) & 0xFF);
      }
      final SecretKeyFactory skf = SecretKeyFactory.getInstance("PBKDF2WithHmacSHA256");
      final String basePass = PermanentInstallationID.get() + System.getProperty("user.name", "unknown-user");
      key = new SecretKeySpec(skf.generateSecret(new PBEKeySpec((platformBuild + basePass).toCharArray(), salt, 10, 128)).getEncoded(), "AES");
      tokensKey = new SecretKeySpec(skf.generateSecret(new PBEKeySpec(basePass.toCharArray(), "JB-Auth-Token-Store".getBytes(StandardCharsets.UTF_8), 10, 128)).getEncoded(), "AES");
    }
    catch (Throwable ignored) {
      // fallback to a random-generated key that the key can be used within this session only
      final byte[] bytes = new byte[16];
      new SecureRandom().nextBytes(bytes);
      key = tokensKey = new SecretKeySpec(bytes, "AES");
    }
    ourEncryptionKey = key;
    ourTokensKey = tokensKey;
  }

  public ServerSource(LicenseFactory factory, int productVersion, PluginsAwareLicenseManager manager) {
    super(factory);
    myProductVersion = productVersion;
    myManager = manager;
    final ProductDescriptor pd = factory.getProductDescriptor();
    myProductUUID = pd.getProductFamily() == ProductDescriptor.UNDEFINED ? pd.getProductCode() : pd.getProductFamily().toString();
  }

  private static String calcSignature(final String serverUrl,
                                      final String machineId,
                                      final long now,
                                      final long prolongationPeriod,
                                      final String userName,
                                      final String ticketId,
                                      final String licenseType) {
    StringBuilder builder = new StringBuilder();
    builder.append(serverUrl);
    builder.append(prolongationPeriod);
    if (now != 0) {
      builder.append(now);
    }
    builder.append(userName);
    builder.append(machineId);
    builder.append(ticketId);
    builder.append(licenseType);

    int signature = builder.toString().hashCode();
    return Integer.toHexString(signature);
  }

  private static String stripTrailingSlash(String url) {
    if (url != null && url.endsWith("/")) {
      return url.substring(0, url.length() - 1);
    }
    return url;
  }

  @Override
  public void create(String serverUrl) throws LicenseCreationException{
    try {
      setLicense(obtainFromServerOrRestore(serverUrl));
    }
    catch (Throwable e) {
      releaseLicense();
      throw e;
    }
  }

  @Nullable
  protected ServerLicenseData obtainFromServerOrRestore(final String url) throws LicenseCreationException {
    if (url != null) {
      try {
        final ServerLicenseData created = obtainTicketFromServer(url);
        updateLicenseStamp();
        return created;
      }
      catch (ServerProtocolHijacked serverProtocolHijacked) {
        final String reason = serverProtocolHijacked.getMessage();
        final String message = LicenseBundle.message("message.server.protocol.hijacked", StringUtil.isEmptyOrSpaces(reason) ? "Server hijacked?" : reason);
        LOG.info(message);
        throw new LicenseCreationException(LicenseBundle.message("error.message.server.protocol.error"), message);
      }
      catch (NoFreeTicketException e) {
        String msg = e.getMessage();
        if (StringUtil.isEmptyOrSpaces(msg)) {
          msg = LicenseBundle.message("error.message.no.free.ticket.available");
        }
        LOG.info(msg);
        throw new LicenseCreationException(msg);
      }
      catch (AuthRequiredException e) {
        LOG.info(e.getMessage());
        throw new AuthorizationRequiredException(e.getMessage(), e.getAuthUrl());
      }
      catch (ServerConnectionException e) {
        // fallback, attempt to survive server absence if a ticket was obtained before
        String additionalErrorDescription = null;
        try {
          final ServerLicenseData license = readStoredTicketInfo(url);
          if (license != null) {
            return license;
          }
        }
        catch (TicketRestoreException ex) {
          additionalErrorDescription = ex.getMessage();
        }
        final String message = getConnectionErrorMessage(e, additionalErrorDescription);
        String logMessage = factory.getProductDescriptor().getShortName() + ": " + message;
        if (e.getDetails() != null) {
          logMessage = logMessage + " (" + e.getDetails() + ")";
        }
        LOG.info(logMessage);
        final LicenseCreationException lce = new LicenseCreationException(message);
        final Throwable cause = e.getCause();
        if (cause instanceof IOException) {
          lce.initCause(cause);
        }
        throw lce;
      }
    }
    return null;
  }

  @NlsContexts.NotificationContent
  private static String getConnectionErrorMessage(Exception e, @NlsContexts.NotificationContent String additionalErrorDescription) {
    final StringBuilder message = new StringBuilder();
    final String reason = e.getMessage();
    if (!StringUtil.isEmptyOrSpaces(reason)) {
      message.append(reason);
    }
    if (additionalErrorDescription != null) {
      if (message.length() > 0) {
        message.append("; ");
      }
      message.append(additionalErrorDescription);
    }
    if (message.length() == 0) {
      message.append(LicenseBundle.message("error.message.unable.to.contact.license.server"));
    }
    @NlsSafe
    final String text = message.toString();
    return text;
  }

  protected void updateLicenseStamp() {
    myLastTimeSeenServer = System.currentTimeMillis();
    myVerificationFailureStart = 0L;
  }

  @Override
  public void validate() {
  }

  protected String getLicenseServerTicketLocation() {
    return getLicenseServerTicketLocation(getProductFamilyUUID());
  }

  @NotNull
  private static String getLicenseServerTicketLocation(final String uuid) {
    return "JetBrains." + uuid + ".LicenseServerTicket";
  }

  @NotNull
  private static String getTokensLocation(String serverUrl) {
    return "JetBrains.fls-tokens." + AuthTokenManager.getServiceId(serverUrl.trim()).replace('.', '_').replace(':', '_');
  }

  protected ServerLicenseData createLicenseData(final String serverUrl,
                                                final String ticketId,
                                                final long prolongationPeriod,
                                                final String confirmationStamp, final Map<String, String> properties) {
    int type = LicenseType.COMMERCIAL;
    try {
      type = Integer.parseInt(properties.get(TicketProperties.LICENSE_TYPE));
    }
    catch (NumberFormatException ignored) {
    }
    String licensee = properties.get(TicketProperties.LICENSEE);
    if (licensee == null) {
      licensee = "N/A";
    }
    String metadata = properties.get(TicketProperties.METADATA);
    return factory.createServerLicense(
      ticketId, prolongationPeriod, serverUrl, licensee, type, confirmationStamp, metadata
    );
  }

  protected abstract String getMachineId();

  protected final String getProductFamilyUUID() {
    return myProductUUID;
  }

  protected final int getProductVersion() {
    return myProductVersion;
  }

  protected enum ProlongationCode {
    OK,
    REFUSED,
    DISCONNECTED,
    VERIFICATION_FAILED
  }

  @NotNull
  protected abstract ServerLicenseData obtainTicketFromServer(final String serverUrl) throws NoFreeTicketException, AuthRequiredException, ServerConnectionException, ServerProtocolHijacked;

  @NotNull
  protected abstract ProlongationCode prolongServerLicense(final ServerLicenseData serverLicense);

  protected abstract boolean releaseServerLicense(final String serverUrl, final String ticketId);

  protected void storeTicketInfo(final ServerLicenseData license) {
    if (!license.isTicketObtained()) {
      return;
    }
    final String serverUrl = license.getServerUrl();
    final long now = myLastTimeSeenServer;
    final String signatureAsString =
      calcSignature(serverUrl, getMachineId(), now, license.getProlongationPeriod(), license.getUserName(),
                    license.getTicketId(), String.valueOf(license.getLicenseType()));

    final Properties properties = new Properties();
    setProperty(properties, SERVER_URL, serverUrl);
    setProperty(properties, PROLONGATION_PERIOD, String.valueOf(license.getProlongationPeriod()));
    setProperty(properties, MACHINE_ID, getMachineId());
    setProperty(properties, TICKET_ID, license.getTicketId());
    setProperty(properties, TicketProperties.LICENSEE, license.getUserName());
    setProperty(properties, SIGNATURE, signatureAsString);
    setProperty(properties, TicketProperties.LICENSE_TYPE, String.valueOf(license.getLicenseType()));
    final String metadata = license.getMetadata();
    if (metadata != null && !metadata.isEmpty()) {
      setProperty(properties, TicketProperties.METADATA, metadata);
    }
    setProperty(properties, CONFIRMATION_STAMP, license.getConfirmationStamp());
    setProperty(properties, TIMESTAMP, String.valueOf(now));

    saveProperties(properties, getLicenseServerTicketLocation(), "License server ticket information. Please do not alter this data", ourEncryptionKey);
    storeTokenInfo(serverUrl);
  }

  protected void storeTokenInfo(final String serverUrl) {
    final Token token = myManager.getTokenManager().getToken(serverUrl);
    if (token != null) {
      final Properties properties = new Properties();
      setProperty(properties, ACCESS_TOKEN, token.getAccessToken());
      setProperty(properties, REFRESH_TOKEN, token.getRefreshToken());
      setProperty(properties, TOKEN_EXPIRATION, String.valueOf(token.getExpiresInSec()));
      setProperty(properties, TOKEN_EXCHANGE_URL, token.getExchangeUrl());
      setProperty(properties, TOKEN_CLIENT_ID, token.getClientId());
      saveProperties(properties, getTokensLocation(serverUrl), null, ourTokensKey);
    }
  }

  private static void saveProperties(Properties properties, final String prefsKey, String comment, final Object encryptKey) {
    final ByteArrayOutputStream out = new ByteArrayOutputStream();
    final OutputStream zip = new DeflaterOutputStream(out);
    try {
      properties.store(zip, comment);
    }
    catch (IOException e) {
      LOG.error(e); // Can't be
    }
    finally {
      try {
        zip.close();
        final Cipher ciph = Cipher.getInstance("AES/CBC/PKCS5Padding");
        ciph.init(Cipher.ENCRYPT_MODE, (Key)encryptKey);
        final byte[] body = ciph.doFinal(out.toByteArray());
        final byte[] iv = ciph.getIV();

        final byte[] data = new byte[4 + iv.length + body.length];

        final int length = body.length;
        data[0] = (byte)((length >> 24) & 0xFF);
        data[1] = (byte)((length >> 16) & 0xFF);
        data[2] = (byte)((length >> 8) & 0xFF);
        data[3] = (byte)(length & 0xFF);

        System.arraycopy(iv, 0, data, 4, iv.length);
        System.arraycopy(body, 0, data, 4 + iv.length, body.length);
        Prefs.putByteArray(prefsKey, data);
      }
      catch (Throwable ignored) {
        //LOG.error(e); // Can't be
      }
    }
  }

  private static void setProperty(Properties properties, final String key, final String value) {
    if (key != null) {
      properties.setProperty(key, value == null? "" : value);
    }
  }

  private static class TicketRestoreException extends Exception {
    @Override
    public Throwable fillInStackTrace() {
      return this;
    }
  }
  private static class TooLongServerAbsence extends TicketRestoreException {
    @Override
    public String getMessage() {
      return "No response from license server for more than " + (SERVER_ABSENSE_TOLERANCE_TIMEOUT / 3600000L)+ " hours";
    }
  }
  private static class CorruptedLicenseData extends TicketRestoreException {
    @Override
    public String getMessage() {
      return "Stored license data is corrupt. License Revoked";
    }
  }

  @Nullable
  private ServerLicenseData readStoredTicketInfo(final String serverUrl) throws TicketRestoreException{
    if (serverUrl == null) {
      return null;
    }
    final Properties properties = loadProperties(getLicenseServerTicketLocation(), ourEncryptionKey);
    if (properties == null) {
      return null;
    }

    final String url = properties.getProperty(SERVER_URL);
    if (!Objects.equals(stripTrailingSlash(serverUrl), stripTrailingSlash(url))) {
      return null;
    }
    try {
      checkUrlAllowed(url);
    }
    catch (IncorrectLSException e) {
      deleteStoredTicketInfo(); // pass null, because there is no need to release ticket on non-authentic server
      return null;
    }
    final String machineId = properties.getProperty(MACHINE_ID);
    final String encodedSignature = properties.getProperty(SIGNATURE);
    final String ticketId = properties.getProperty(TICKET_ID);
    final String licensee = properties.getProperty(TicketProperties.LICENSEE);
    final String licenseType = properties.getProperty(TicketProperties.LICENSE_TYPE);
    final String metadata = properties.getProperty(TicketProperties.METADATA);
    final String confirmationStamp = properties.getProperty(CONFIRMATION_STAMP);
    final String generationStamp = properties.getProperty(TIMESTAMP);

    long prolongation;
    try {
      prolongation = Math.min(Long.parseLong(properties.getProperty(PROLONGATION_PERIOD)), MAX_PROLONGATION_PERIOD);
    }
    catch (NumberFormatException e) {
      prolongation = DEFAULT_PROLONGATION_PERIOD;
    }

    long timestamp = 0;
    if (generationStamp != null) {
      try {
        timestamp = Long.parseLong(generationStamp);
        myLastTimeSeenServer = Math.max(timestamp, myLastTimeSeenServer);
      }
      catch (NumberFormatException ignored) {
        // ignore, it's 0 now.
      }
    }

    if (Math.abs(System.currentTimeMillis() - timestamp) > SERVER_ABSENSE_TOLERANCE_TIMEOUT) {
      deleteStoredTicketInfo();
      throw new TooLongServerAbsence();
    }

    final String signature = calcSignature(url, machineId, timestamp, prolongation, licensee, ticketId, licenseType);

    if (!signature.equals(encodedSignature) || !machineId.equals(getMachineId())) {
      deleteStoredTicketInfo();
      throw new CorruptedLicenseData();
    }

    final Map<String, String> licenseProperties = new HashMap<>();
    licenseProperties.put(TicketProperties.LICENSEE, licensee);
    licenseProperties.put(TicketProperties.LICENSE_TYPE, licenseType);
    licenseProperties.put(TicketProperties.METADATA, metadata);

    return createLicenseData(url, ticketId, prolongation, confirmationStamp, licenseProperties);
  }

  protected TokenFuture<String> lookupAccessToken(String serverUrl) {
    Token token = null;
    if (serverUrl != null) {
      final AuthTokenManager tokenManager = myManager.getTokenManager();
      token = tokenManager.getToken(serverUrl);
      if (token == null) {
        token = readStoredTokenInfo(serverUrl);
        if (token != null) {
          return TokenFuture.create(tokenManager.addTokenIfAbsent(serverUrl, token), t -> t != null? t.getAccessToken() : null, () -> null);
        }
      }
    }
    return TokenFuture.createCompleted(token != null? token.getAccessToken() : null);
  }

  @Nullable
  private static Token readStoredTokenInfo(final String serverUrl) {
    if (serverUrl == null) {
      return null;
    }
    final Properties properties = loadProperties(getTokensLocation(serverUrl), ourTokensKey);
    if (properties == null) {
      return null;
    }
    final String accessToken = properties.getProperty(ACCESS_TOKEN);
    final String refreshToken = properties.getProperty(REFRESH_TOKEN);
    final String exchangeUrl = properties.getProperty(TOKEN_EXCHANGE_URL);
    final String clientId = properties.getProperty(TOKEN_CLIENT_ID);
    if (accessToken != null && refreshToken != null && exchangeUrl != null && clientId != null) {
      Token tokens;
      try {
        tokens = new Token(accessToken, refreshToken, exchangeUrl, clientId, Long.parseLong(properties.getProperty(TOKEN_EXPIRATION)));
      }
      catch (Throwable e) {
        tokens = new Token(accessToken, refreshToken, exchangeUrl, clientId);
      }
      return tokens;
    }
    return null;
  }

  @Nullable
  private static Properties loadProperties(final String key, final Object encryptKey) {
    final byte[] props = Prefs.getByteArray(key, null);
    if (props != null) {
      try {
        final int bodyLength = ((props[0] & 0xFF) << 24) |
                               ((props[1] & 0xFF) << 16) |
                               ((props[2] & 0xFF) << 8) |
                               ((props[3] & 0xFF)      );

        final int ivlength = props.length - 4 - bodyLength;

        final Cipher ciph = Cipher.getInstance("AES/CBC/PKCS5Padding");
        ciph.init(Cipher.DECRYPT_MODE, (Key)encryptKey, new IvParameterSpec(props, 4, ivlength));
        final Properties properties = new Properties();
        properties.load(new InflaterInputStream(new ByteArrayInputStream(ciph.doFinal(props, 4 + ivlength, bodyLength))));
        return properties;
      }
      catch (Throwable ignored) {
        // Can't be
      }
    }
    return null;
  }

  protected void deleteStoredTicketInfo() {
    Prefs.remove(getLicenseServerTicketLocation());
  }

  protected void deleteStoredTokenInfo(String serverUrl) {
    Prefs.remove(getTokensLocation(serverUrl));
  }

  protected void startProlongTask(final ServerLicenseData license) {
    storeTicketInfo(license);

    final ProlongPeriodicTask task = new ProlongPeriodicTask(license);
    long period = Math.min(license.getProlongationPeriod(), MAX_PROLONGATION_PERIOD);
    if (period <= 0) {
      period = DEFAULT_PROLONGATION_PERIOD;
    }
    final ScheduledFuture<?> future = AppExecutorUtil.getAppScheduledExecutorService().scheduleWithFixedDelay(task, period, period, TimeUnit.MILLISECONDS);

    myProlongTaskFuture = new ProlongTaskFuture(future, task);
  }

  @Override
  public ServerLicenseData getLicense() {
    return (ServerLicenseData)super.getLicense();
  }

  public void releaseLicense() {
    try {
      if (!stopProlongTask()) {
        // prolog task was not running, so release the ticket explicitly
        final ServerLicenseData license = getLicense();
        if (license != null) {
          if (releaseServerLicense(license.getServerUrl(), license.getTicketId())) {
            deleteStoredTicketInfo();
          }
        }
      }
    }
    finally {
      setLicense(null);
    }
  }

  protected boolean stopProlongTask() {
    final ProlongTaskFuture future = myProlongTaskFuture;
    if (future != null) {
      myProlongTaskFuture = null;
      future.cancel(false);
      return true;
    }
    return false;
  }

  void reobtainFloatingTicket(final String serverUrl, final boolean delayed, final boolean allowOldTicketReuse) {
    if (myReobtainScheduled.getAndSet(true)) {
      return;
    }
    final Runnable restart = () -> {
      final Ref<ServerLicenseData> newLicenseRef = new Ref<>(null);
      final Ref<ValidationFailedException> errorRef = new Ref<>(null);
      try {
        final ServerLicenseData newLicense = obtainTicketFromServer(serverUrl);
        updateLicenseStamp();
        newLicenseRef.set(newLicense);
      }
      catch (ServerConnectionException | ServerProtocolHijacked | NoFreeTicketException | AuthRequiredException e) {
        errorRef.set(e);
      }

      LicenseEntryUtil.invokeOnEdtWhenApplicationLoaded(new LicenseEntryUtil.ApplicationRunnable() {
        @Override
        public void run(Application app) {
          myReobtainScheduled.set(false);
          final ValidationFailedException error = errorRef.get();
          @NlsContexts.NotificationContent String errorMessage = null;
          if (error instanceof ServerConnectionException) {
            if (allowOldTicketReuse) {
              ServerLicenseData restored = null;
              String errorReason = null;
              try {
                restored = readStoredTicketInfo(serverUrl);
              }
              catch (TooLongServerAbsence ex) {
                errorReason = ex.getMessage();
              }
              catch (TicketRestoreException e) {
                errorReason = e.getMessage();
                //LOG.info("Error reading saved license server ticket: " + e.getMessage());
              }
              newLicenseRef.set(restored);
              if (restored == null) {
                errorMessage = getConnectionErrorMessage(error, errorReason);
              }
            }
            else {
              errorMessage = getConnectionErrorMessage(error, null);
            }
          }
          else if (error instanceof ServerProtocolHijacked) {
            errorMessage = getConnectionErrorMessage(error, LicenseBundle.message("notification.content.license.server.response.has.not.passed.data.integrity.check"));
          }
          else if (error instanceof NoFreeTicketException) {
            errorMessage = getConnectionErrorMessage(error, LicenseBundle.message("notification.content.no.suitable.licenses.left.on.license.server"));
          }
          else if (error instanceof AuthRequiredException) {
            final AuthRequiredException err = (AuthRequiredException)error;
            errorMessage = AuthorizationRequiredException.createAuthorizationPromptMessage(err.getMessage(), err.getAuthUrl());
          }

          final ServerLicenseData newLicense = newLicenseRef.get();
          if (isActive()) { // additionally check if server source is still an active license source for the product
            setLicense(newLicense);
            myManager.updateLicenseAbout();
            @NlsContexts.NotificationTitle String title = LicenseBundle.message("notification.title.floating.license.server");
            if (newLicense != null) {
              startProlongTask(newLicense);
              if (errorMessage != null) {
                LicenseEntryUtil.showLicenseNotificationLater(title, errorMessage, NotificationType.ERROR);
              }
            }
            else {
              if (!(factory instanceof PluginLicenseFactory) || !((PluginLicenseFactory)factory).isLicenseOptional()) {
                myManager.demandLicenseAsyncFromBackgroundTask(app, factory.getProductDescriptor(), errorMessage != null? new Pair<>(title, errorMessage) : null);
              }
            }
          }
          else {
            if (newLicense != null) {
              releaseServerLicense(newLicense.getServerUrl(), newLicense.getTicketId());
            }
          }
        }

        @Override
        public void runFailed() {
          myReobtainScheduled.set(false);
          // if ticket has been obtained, release it
          final ServerLicenseData newLicense = newLicenseRef.get();
          if (newLicense != null) {
            releaseServerLicense(newLicense.getServerUrl(), newLicense.getTicketId());
          }
        }
      });
    };

    AppExecutorUtil.getAppScheduledExecutorService().schedule(restart, delayed ? 15 * 60 : 0, TimeUnit.SECONDS);
  }

  private boolean isActive() {
    final ProductDescriptor selfPd = factory.getProductDescriptor();
    LicenseInfo licenseInfo = myManager.getPlatformLicenseInfo();
    if (!selfPd.equals(licenseInfo.getProductDescriptor())) {
      for (LicenseInfo info : myManager.getPluginLicenseInfo()) {
        if (selfPd.equals(info.getProductDescriptor())) {
          licenseInfo = info;
          break;
        }
      }
    }
    return this.equals(licenseInfo.getLicenseSource());
  }

  @Override
  public void scheduleRemoteValidation() {
    final ServerLicenseData currentLicense = getLicense();
    if (currentLicense != null) {
      final ProlongTaskFuture future = myProlongTaskFuture;
      if (future != null && !future.isDone()) {
        final ServerLicenseData prolonged = future.getProlongedLicense();
        if (prolonged == null || !Objects.equals(prolonged.getTicketId(), currentLicense.getTicketId())) {
          stopProlongTask();
          startProlongTask(currentLicense);
        }
      }
      else {
        startProlongTask(currentLicense);
      }
    }
  }

  @Override
  public final boolean needNetworkLicenseCheck() {
    return false;
  }

  protected static void checkUrlAllowed(String url) throws IncorrectLSException{
    //noinspection ResultOfMethodCallIgnored -- initialize class
    LicensingProxyService.getInstance();
    //noinspection RedundantOperationOnEmptyContainer
    for (String pattern : URL_BLACKLIST_PATTERNS) {
      tryPattern(url, pattern);
    }
    final Collection<String> blPatterns = ourBlResource.getData();
    if (blPatterns != null) {
      final Set<String> processed = ContainerUtil.set(URL_BLACKLIST_PATTERNS);
      for (String pattern : blPatterns) {
        if (processed.add(pattern)) {
          tryPattern(url, pattern);
        }
      }
    }
  }

  private static void tryPattern(String url, String pattern) throws IncorrectLSException {
    try {
      if (Pattern.compile(pattern, Pattern.CASE_INSENSITIVE).matcher(url).matches()) {
        throw new IncorrectLSException();
      }
    }
    catch (VerificationFailedException e) {
      throw e;
    }
    catch (Throwable ignored) {
    }
  }

  private class ProlongPeriodicTask implements Runnable {
    @NotNull
    private final ServerLicenseData myLicense;
    private final AtomicBoolean myInProgress = new AtomicBoolean(false);
    private final AtomicBoolean myCanceled = new AtomicBoolean(false);

    ProlongPeriodicTask(@NotNull ServerLicenseData license) {
      myLicense = license;
    }

    public void cancel() {
      if (!myCanceled.getAndSet(true)) {
        if (releaseServerLicense(myLicense.getServerUrl(), myLicense.getTicketId())) {
          // only delete cached ticket info if we successfully returned the ticket to the pool
          deleteStoredTicketInfo();
        }
      }
    }

    @Override
    public void run() {
      AppExecutorUtil.getAppScheduledExecutorService().execute(() -> {
        if (!myInProgress.getAndSet(true)) {
          try {
            if (!myCanceled.get()) {
              final ProlongationCode rc = prolongServerLicense(myLicense);
              if (rc == ProlongationCode.OK) {
                myManager.updateLicenseAbout(); // this will update license confirmationStamp
                updateLicenseStamp();
                storeTicketInfo(myLicense);
              }
              else {
                final boolean shouldReobtain;
                if (rc == ProlongationCode.VERIFICATION_FAILED) {
                  if (myVerificationFailureStart <= 0) {
                    // the very first verification failure on ticket prolongation
                    myVerificationFailureStart = System.currentTimeMillis();
                  }
                  shouldReobtain = Math.abs(System.currentTimeMillis() - myVerificationFailureStart) > OUTDATED_CERTIFICATE_TOLERANCE_TIMEOUT;
                }
                else {
                  shouldReobtain = true;
                }
                if (shouldReobtain) {
                  final boolean isInitial = !myLicense.isTicketObtained() && myLicense.equals(getLicense());
                  if (isInitial) {
                    // if this is a "temporary" initial license created on startup, it should be removed completely
                    releaseLicense();
                  }
                  else {
                    stopProlongTask();
                  }
                  reobtainFloatingTicket(myLicense.getServerUrl(), rc == ProlongationCode.DISCONNECTED, isInitial || rc == ProlongationCode.DISCONNECTED);
                }
              }
            }
          }
          finally {
            myInProgress.set(false);
          }
        }
      });
    }
  }

  private class ProlongTaskFuture implements Future {
    private final Future myScheduledFuture;
    private final ProlongPeriodicTask myTask;

    ProlongTaskFuture(Future scheduledFuture, ProlongPeriodicTask task) {
      myScheduledFuture = scheduledFuture;
      myTask = task;
    }

    public ServerLicenseData getProlongedLicense() {
      return myTask.myLicense;
    }

    @Override
    public boolean cancel(boolean mayInterruptIfRunning) {
      try {
        return myScheduledFuture.cancel(mayInterruptIfRunning);
      }
      finally {
        myTask.cancel();
      }
    }

    @Override
    public boolean isCancelled() {
      return myScheduledFuture.isCancelled();
    }

    @Override
    public boolean isDone() {
      return myScheduledFuture.isDone();
    }

    @Override
    public Object get() throws InterruptedException, ExecutionException {
      return myScheduledFuture.get();
    }

    @Override
    public Object get(long timeout, TimeUnit unit) throws InterruptedException, ExecutionException, TimeoutException {
      return myScheduledFuture.get(timeout, unit);
    }
  }
}
