package com.intellij.ide.license.impl;

import com.google.gson.Gson;
import com.google.gson.reflect.TypeToken;

import java.lang.reflect.Type;
import java.util.Map;
import java.util.Objects;
import java.util.function.Function;

public class Token {
  public static final int DEFAULT_EXPIRATION_PERIOD_SEC = 3600;
  private static final Gson ourGson = new Gson();
  private static final Type JSON_READ_TARGET_TYPE = new TypeToken<Map<String, Object>>(){}.getType();
  private static final Function<Map<String, ?>, String> ACCESS_TOKEN_PROPERTY = typedGetter("access_token", String.class);
  private static final Function<Map<String, ?>, String> REFRESH_TOKEN_PROPERTY = typedGetter("refresh_token", String.class);
  private static final Function<Map<String, ?>, Number> EXPIRES_IN_PROPERTY = typedGetter("expires_in", Number.class);

  private final String myAccessToken;
  private final String myRefreshToken;
  private final String myExchangeUrl;
  private final String myClientId;
  private final long myExpiresInSec;

  public Token(String accessToken, String refreshToken, String exchangeUrl, final String clientId) {
    this(accessToken, refreshToken, exchangeUrl, clientId, DEFAULT_EXPIRATION_PERIOD_SEC);
  }
  
  public Token(String accessToken, String refreshToken, String exchangeUrl, String clientId, long expiresInSec) {
    myAccessToken = accessToken;
    myRefreshToken = refreshToken;
    myExchangeUrl = exchangeUrl;
    myClientId = clientId;
    myExpiresInSec = expiresInSec;
  }

  public String getAccessToken() {
    return myAccessToken;
  }

  public String getRefreshToken() {
    return myRefreshToken;
  }

  public String getExchangeUrl() {
    return myExchangeUrl;
  }

  public String getClientId() {
    return myClientId;
  }

  public long getExpiresInSec() {
    return myExpiresInSec;
  }

  public Token derive(String accessToken, String refreshToken, Number expiresInSec) {
    return new Token(
      accessToken != null? accessToken : myAccessToken,
      refreshToken != null? refreshToken : myRefreshToken,
      myExchangeUrl,
      myClientId,
      expiresInSec != null? expiresInSec.longValue() : myExpiresInSec
    );
  }

  public static Token updateFromJson(Token tokens, String json) {
    final Map<String, Object> updated = ourGson.fromJson(json, JSON_READ_TARGET_TYPE);
    return tokens.derive(ACCESS_TOKEN_PROPERTY.apply(updated), REFRESH_TOKEN_PROPERTY.apply(updated), EXPIRES_IN_PROPERTY.apply(updated));
  }

  public static Token fromJson(String json, String exchangeUrl, final String clientId) {
    final Map<String, Object> tokens = ourGson.fromJson(json, JSON_READ_TARGET_TYPE);
    final String accessToken = ACCESS_TOKEN_PROPERTY.apply(tokens);
    if (accessToken == null) {
      return null;
    }
    final String refreshToken = REFRESH_TOKEN_PROPERTY.apply(tokens);
    if (refreshToken == null) {
      return null;
    }
    final Number expiration = EXPIRES_IN_PROPERTY.apply(tokens);
    return expiration != null? new Token(accessToken, refreshToken, exchangeUrl, clientId, expiration.longValue()) : new Token(accessToken, refreshToken, exchangeUrl, clientId);
  }

  @Override
  public String toString() {
    return myAccessToken + ":" + myRefreshToken + ":" + myExchangeUrl + ":" + myClientId + ":" + myExpiresInSec;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) {
      return true;
    }
    if (o == null || getClass() != o.getClass()) {
      return false;
    }
    final Token token = (Token)o;
    return myExpiresInSec == token.myExpiresInSec && myAccessToken.equals(token.myAccessToken) && myRefreshToken.equals(token.myRefreshToken) && myExchangeUrl.equals(token.myExchangeUrl) && myClientId.equals(token.myClientId);
  }

  @Override
  public int hashCode() {
    return Objects.hash(myAccessToken, myRefreshToken, myExchangeUrl, myClientId, myExpiresInSec);
  }

  private static <T> Function<Map<String, ?>, T> typedGetter(String propName, Class<T> dataType) {
    return map -> {
      Object value = map.get(propName);
      return dataType.isInstance(value)? dataType.cast(value) : null;
    };
  }

}
