package com.intellij.ide.license.ui;

import com.intellij.icons.AllIcons;
import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.license.LicenseInfo;
import com.intellij.ide.license.ProductDescriptor;
import com.intellij.ide.plugins.IdeaPluginDescriptor;
import com.intellij.ide.plugins.PluginManagerCore;
import com.intellij.ide.plugins.newui.PluginLogoApi;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.openapi.util.SystemInfo;
import com.intellij.ui.*;
import com.intellij.ui.components.JBList;
import com.intellij.ui.components.JBScrollPane;
import com.intellij.ui.scale.ScaleContext;
import com.intellij.util.ui.JBUI;
import com.intellij.util.ui.UIUtil;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import java.awt.*;
import java.awt.event.MouseEvent;
import java.awt.event.MouseMotionAdapter;
import java.text.SimpleDateFormat;
import java.util.*;

public final class FlatMultiLicensePanel extends JPanel implements Disposable {
  private static final Color EXPIRED_LICENSE_COLOR = JBColor.namedColor("Label.errorForeground", new JBColor(new Color(0xC7222D), DarculaColors.RED));
  private static final Color EXPIRED_LICENSE_SELECTED_COLOR = JBColor.PINK;
  public static final int PRODUCT_ICON_SIZE = 40;
  public static final int MIN_PRODUCTS_LIST_WIDTH = JBUI.scale(250);

  private final Map<ProductDescriptor, JPanel> myProductLicensePanels;
  private final LicenseItemsListModel myItemsListModel;
  private final JList<ProductDescriptor> myItemsList;
  private final LicenseEditingModel myModel;
  private final LicenseEntryUI myUI;
  private final Box myLicensePanelPlaceholder = Box.createHorizontalBox();

  public FlatMultiLicensePanel(LicenseEditingModel model, LicenseEntryUI ui) {
    super(new BorderLayout());
    myModel = model;
    myUI = ui;

    myItemsListModel = new LicenseItemsListModel(model);
    myItemsList = createItemsList(myItemsListModel, new ListItemRenderer(model));

    myProductLicensePanels = createLicensePanelsMap(model);

    JScrollPane leftScrollPanel = new JBScrollPane(myItemsList);
    leftScrollPanel.setBorder(null);
    leftScrollPanel.setHorizontalScrollBarPolicy(ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER);

    myItemsList.addListSelectionListener(e -> {
      ProductDescriptor selected = myItemsList.getSelectedValue();
      if (selected == null) {
        myItemsList.setSelectedIndex(0);
      }
      else {
        showLicensePanelForProduct(selected);
      }
    });

    myLicensePanelPlaceholder.setBorder(JBUI.Borders.customLine(JBColor.border(), 0, 1, 0, 0));

    add(leftScrollPanel, BorderLayout.WEST);
    add(myLicensePanelPlaceholder, BorderLayout.CENTER);

    final ProductDescriptor toSelect = model.getInitialProductToSelect();
    if (toSelect == null) {
      myItemsList.setSelectedIndex(0);
    }
    else {
      myItemsList.setSelectedValue(toSelect, true);
    }

    updateLicensingState();
  }

  private JList<ProductDescriptor> createItemsList(LicenseItemsListModel model,
                                                   ListItemRenderer renderer) {
    boolean appIsLoaded = ApplicationManager.getApplication() != null;

    JList<ProductDescriptor> res = appIsLoaded ? new JBList<>(model) : new JList<>(model);
    res.setCellRenderer(renderer);
    res.setSelectionMode(ListSelectionModel.SINGLE_SELECTION);

    int width = MIN_PRODUCTS_LIST_WIDTH;
    if (model.getSize() > 0) {
      ProductDescriptor product = model.getElementAt(0);
      Component renderedProductItem = renderer.getListCellRendererComponent(res, product, 0, false, false);
      width = Integer.max(MIN_PRODUCTS_LIST_WIDTH, renderedProductItem.getPreferredSize().width);
    }

    Dimension maxSize = res.getMaximumSize();
    maxSize.width = width;
    res.setMaximumSize(maxSize);
    Dimension prefSize = res.getPreferredSize();
    prefSize.width = width;
    res.setPreferredSize(prefSize);

    if (appIsLoaded) {
      ((JBList<ProductDescriptor>) res).setExpandableItemsEnabled(true);
    }
    else {
      res.addMouseMotionListener(new MouseMotionAdapter() {
        @Override
        public void mouseMoved(MouseEvent e) {
          int index = res.locationToIndex(e.getPoint());
          if( index > -1 ) {
            res.setToolTipText(myModel.getDisplayName(model.getElementAt(index)));
          }
        }
      });
    }

    return res;
  }

  @Override
  public void dispose() {}

  @Nullable
  public JComponent getPreferredFocusedComponent() {
    if (myLicensePanelPlaceholder.getComponentCount() <= 0) return null;

    Component comp = myLicensePanelPlaceholder.getComponent(0);
    return comp instanceof LicensePanel
           ? ((LicensePanel)comp).getPreferredFocusedComponent()
           : null;
  }

  private Map<ProductDescriptor, JPanel> createLicensePanelsMap(LicenseEditingModel model) {
    Map<ProductDescriptor, JPanel> res = new HashMap<>();
    final ProductDescriptor platform = model.getPlatformProduct();
    if (platform != null) {
      res.put(platform, createDetailsPanelForItem(platform, myModel));
    }
    model.getPluginProducts().forEach(d -> res.put(d, createDetailsPanelForItem(d, myModel)));
    return res;
  }

  private JPanel createDetailsPanelForItem(ProductDescriptor item, LicenseEditingModel model) {
    String error = model.getInitialProductToSelect() == item ? model.getInitialProblemDescription() : null;
    LicensePanel panel = new LicensePanel(item, model, error);
    Disposer.register(this, panel);
    panel.addLicenseUpdateListener(myItemsListModel);
    panel.addLicenseUpdateListener(d -> updateLicensingState());
    panel.addLicenseUpdateListener(new DefaultButtonUpdater());
    panel.setMinimumSize(panel.getPreferredSize());
    return panel;
  }

  private void showLicensePanelForProduct(ProductDescriptor selected) {
    JPanel newPanel = myProductLicensePanels.get(selected);
    myLicensePanelPlaceholder.removeAll();
    myLicensePanelPlaceholder.add(newPanel);
    myLicensePanelPlaceholder.repaint();
  }

  private void updateLicensingState() {
    final ProductDescriptor pp = myModel.getPlatformProduct();
    final boolean mainProductIsLicensed = pp == null || !getProductsWithoutLicense(false).contains(pp);
    myUI.setCanBeClosed(mainProductIsLicensed);

    Collection<ProductDescriptor> pluginsWithoutLicense = getProductsWithoutLicense(true);
    if (pp != null) {
      pluginsWithoutLicense.remove(pp);
    }

    int nonOptional = 0;
    int optional = 0;
    for (ProductDescriptor descriptor : pluginsWithoutLicense) {
      if (myModel.getLicenseInfo(descriptor).isOptional()) optional++;
      else nonOptional++;
    }

    String warning = "";
    if (mainProductIsLicensed && !pluginsWithoutLicense.isEmpty()) {
      warning = nonOptional > 0
                ? LicenseBundle.message("warning.text.0.plugins.without.license.will.be.disabled", nonOptional)
                : LicenseBundle.message("warning.text.0.activate.optional.plugins", optional);
    }

    myUI.setWarning(warning);
  }

  private Set<ProductDescriptor> getProductsWithoutLicense(boolean includeExpired) {
    Set<ProductDescriptor> result = null;
    final int size = myItemsListModel.getSize();
    for (int idx = 0; idx < size; idx++) {
      final ProductDescriptor pd = myItemsListModel.getElementAt(idx);
      final LicenseInfo info = myModel.getLicenseInfo(pd);
      final LicenseData lic = info.getLicense();
      if (lic == null || (includeExpired && lic.isExpired())) {
        if (result == null) {
          result = new HashSet<>();
        }
        result.add(pd);
      }
    }
    return result != null? result : Collections.emptySet();
  }

  private static final class ListItemRenderer implements ListCellRenderer<ProductDescriptor> {
    private final Map<String, Icon> myPluginIconsMap = new HashMap<>();
    private final LicenseEditingModel myModel;

    {
      for (IdeaPluginDescriptor descriptor : PluginManagerCore.getPlugins()) {
        if (descriptor.getProductCode() != null) {
          Icon icon = PluginLogoApi.getIcon(descriptor, PRODUCT_ICON_SIZE, PRODUCT_ICON_SIZE, null);
          myPluginIconsMap.put(descriptor.getProductCode(), icon);
        }
      }
    }

    ListItemRenderer(LicenseEditingModel model) {
      myModel = model;
    }

    @Override
    public Component getListCellRendererComponent(JList<? extends ProductDescriptor> list,
                                                  ProductDescriptor pd,
                                                  int index,
                                                  boolean isSelected,
                                                  boolean cellHasFocus) {
      JPanel panel = new JPanel(new BorderLayout());
      //boolean hasFocus = list.hasFocus();
      boolean hasFocus = true;
      panel.setBackground(UIUtil.getListBackground(isSelected, hasFocus));
      Color foreground = UIUtil.getListForeground(isSelected, hasFocus);

      final LicenseData license = myModel.getLicenseInfo(pd).getLicense();

      Icon icon = getIcon(pd, license == null || license.isExpired());
      JLabel iconLabel = new JLabel(icon);
      iconLabel.setBorder(JBUI.Borders.empty(6, 10));
      panel.add(iconLabel, BorderLayout.WEST);

      Box textPanel = new Box(BoxLayout.Y_AXIS);
      textPanel.add(Box.createVerticalGlue());
      textPanel.setBorder(JBUI.Borders.emptyRight(10));

      JLabel nameLabel = new JLabel(myModel.getDisplayName(pd));
      nameLabel.setFont(nameLabel.getFont().deriveFont(13.f).deriveFont(Font.BOLD));
      nameLabel.setForeground(foreground);
      textPanel.add(nameLabel);

      textPanel.add(Box.createVerticalStrut(JBUI.scale(3)));

      JLabel licenseLabel = new JLabel(getLicenseExpiration(pd));
      Color fg = license != null && license.isExpired()
                 ? isSelected ? EXPIRED_LICENSE_SELECTED_COLOR : EXPIRED_LICENSE_COLOR
                 : foreground;
      licenseLabel.setForeground(fg);
      if (!SystemInfo.isWindows) {
        Font font = licenseLabel.getFont();
        licenseLabel.setFont(font.deriveFont(font.getSize() - 1f));
      }
      textPanel.add(licenseLabel);

      textPanel.add(Box.createVerticalGlue());
      panel.add(textPanel, BorderLayout.CENTER);

      if (isPlugin(pd) && (index == 0 || !isPlugin(list.getModel().getElementAt(index - 1)))) {
        Box separatorBox = new Box(BoxLayout.X_AXIS);
        separatorBox.setOpaque(true);
        separatorBox.setBackground(list.getBackground());

        SeparatorWithText separator = new SeparatorWithText();
        separator.setCaption(LicenseBundle.message("separator.caption.plugins"));
        separator.setCaptionCentered(true);
        separatorBox.add(separator);
        panel.add(separatorBox, BorderLayout.NORTH);
      }
      return panel;
    }

    private boolean isPlugin(ProductDescriptor pd) {
      return !pd.equals(myModel.getPlatformProduct());
    }

    private @NlsContexts.Label String getLicenseExpiration(ProductDescriptor pd) {
      final LicenseData lic = myModel.getLicenseInfo(pd).getLicense();
      if (lic == null) {
        return isPlugin(pd) ? LicenseBundle.message("label.license.status.activate.to.enable") : LicenseBundle.message("label.license.status.activate");
      }
      if (lic.isExpired()) {
        return LicenseBundle.message("label.license.status.expired");
      }
      Date expirationDate = lic.getExpirationDate();
      if (expirationDate == null) {
        return LicenseBundle.message("label.license.status.active");
      }
      SimpleDateFormat format = new SimpleDateFormat("MMM d, yyyy");
      return LicenseBundle.message("label.license.status.active.until.0", format.format(expirationDate));
    }

    private Icon getIcon(final ProductDescriptor pd, boolean disabled) {
      Icon icon = isPlugin(pd)
                  ? myPluginIconsMap.get(pd.getProductCode())
                  : AppUIUtil.loadApplicationIcon(ScaleContext.create(), PRODUCT_ICON_SIZE);

      if (icon == null) return null;

      if (disabled) {
        //icon = IconManager.getInstance().createLayered(icon, AllIcons.Plugins.Disabled);
        LayeredIcon layeredIcon = new LayeredIcon(2);
        layeredIcon.setIcon(icon, 0, SwingConstants.CENTER);
        layeredIcon.setIcon(AllIcons.Plugins.Disabled, 1, SwingConstants.SOUTH_WEST);

        icon = layeredIcon;
      }

      return icon;
    }
  }

  private class DefaultButtonUpdater implements LicensePanel.LicenseUpdateListener {
    @Override
    public void licenseUpdated(ProductDescriptor pd) {
      if (!isLicensed(pd)) return;

      JRootPane rootPane = getRootPane();
      if (rootPane == null) return;

      if (getProductsWithoutLicense(true).isEmpty()) {
        JButton button = myUI.getOkButton();
        rootPane.setDefaultButton(button);
        button.requestFocus();
      }
      else {
        rootPane.setDefaultButton(null);
        myItemsList.requestFocus();
      }
    }

    private boolean isLicensed(ProductDescriptor pd) {
      LicenseInfo info = myModel.getLicenseInfo(pd);
      return info != null && info.getLicense() != null;
    }
  }
}
