package com.intellij.ide.license.ui;

import com.intellij.diagnostic.LoadingState;
import com.intellij.ide.AppLifecycleListener;
import com.intellij.ide.BrowserUtil;
import com.intellij.ide.license.impl.LicenseNotification;
import com.intellij.notification.*;
import com.intellij.notification.impl.NotificationsConfigurationImpl;
import com.intellij.openapi.application.Application;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ModalityState;
import com.intellij.openapi.util.JDOMUtil;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.openapi.util.NlsSafe;
import com.intellij.util.concurrency.EdtExecutorService;
import com.intellij.util.ui.EdtInvocationManager;
import com.jetbrains.license.util.Ref;
import org.jdom.Element;
import org.jdom.Namespace;

import javax.swing.*;
import java.awt.*;
import java.util.List;
import java.util.*;
import java.util.concurrent.TimeUnit;
import java.util.function.Function;

public final class LicenseEntryUtil {
  public static void fixFocusTraversal(final JTextArea field) {
    Set<KeyStroke> strokes = new HashSet<>(Collections.singletonList(KeyStroke.getKeyStroke("pressed TAB")));
    field.setFocusTraversalKeys(KeyboardFocusManager.FORWARD_TRAVERSAL_KEYS, strokes);
    strokes = new HashSet<>(Collections.singletonList(KeyStroke.getKeyStroke("shift pressed TAB")));
    field.setFocusTraversalKeys(KeyboardFocusManager.BACKWARD_TRAVERSAL_KEYS, strokes);
  }

  public static void showLicenseNotificationLater(final @NlsContexts.NotificationTitle String title, final @NlsContexts.NotificationContent String text, final NotificationType notificationType) {
    LicenseNotification message = createFormattedMessage(title, text);
    if (message != null) {
      final LicenseNotification.Type msgType = message.getType();
      showLicenseNotificationLater(message.getTitle(), message.getMessage(), msgType == null ? notificationType : msgType.toNotificationType(), notification -> {
        for (LicenseNotification.Action action : message.getActions()) {
          notification.addAction(
            NotificationAction.createSimple(() -> action.getText(), () -> BrowserUtil.browse(action.getLink()))
          );
        }
        return notification;
      });
    }
    else {
      showLicenseNotificationLater(title, text, notificationType, notification -> notification);
    }
  }

  private static LicenseNotification createFormattedMessage(@NlsContexts.NotificationTitle String ttl, String text) {
    try {
      final Element element = JDOMUtil.load(text);
      if ("msg".equals(element.getName())) {
        final Namespace ns = element.getNamespace();
        @NlsSafe
        final String message = element.getChildText("message", ns);
        if (message == null) {
          return null;
        }
        @NlsSafe
        final String title = element.getChildText("title", ns);
        final LicenseNotification.Type type;
        String typeStr = element.getChildText("type", ns);
        if ("ERROR".equals(typeStr)) {
          type = LicenseNotification.Type.ERR;
        }
        else if ("WARNING".equals(typeStr)) {
          type = LicenseNotification.Type.WARN;
        }
        else if ("INFORMATION".equals(typeStr)) {
          type = LicenseNotification.Type.INFO;
        }
        else {
          type = null;
        }

        List<LicenseNotification.Action> actions;
        Element actionsElem = element.getChild("actions", ns);
        if (actionsElem != null) {
          actions = new ArrayList<>();
          for (Element action : actionsElem.getChildren("action", ns)) {
            final String actionText = action.getChildText("text", ns);
            if (actionText != null) {
              final String actionLink = action.getChildText("url", ns);
              if (actionLink != null) {
                actions.add(new LicenseNotification.Action() {
                  @Override
                  public String getText() {
                    return actionText;
                  }

                  @Override
                  public String getLink() {
                    return actionLink;
                  }
                });
              }
            }
          }
        }
        else {
          actions = Collections.emptyList();
        }
        return new LicenseNotification() {
          @Override
          public Type getType() {
            return type;
          }

          @Override
          public String getTitle() {
            return title == null? ttl : title;
          }

          @Override
          public String getMessage() {
            return message;
          }

          @Override
          public Collection<Action> getActions() {
            return actions;
          }
        };
      }
    }
    catch (Throwable ignored) {
    }
    return null;
  }

  public static void showLicenseNotificationLater(final @NlsContexts.NotificationTitle String title, final @NlsContexts.NotificationContent String text, final NotificationType notificationType, Function<? super Notification, ? extends Notification> setup) {
    invokeOnEdtWhenApplicationLoaded(new ApplicationRunnable() {
      @Override
      public void run(Application app) {
        final String groupID = NotificationsConfiguration.LIGHTWEIGHT_PREFIX + "_LICENSE_NOTIFICATION";
        NotificationsConfiguration.getNotificationsConfiguration().register(groupID, NotificationDisplayType.STICKY_BALLOON);
        Notifications.Bus.notify(setup.apply(new Notification(groupID, title, text, notificationType) {
          @Override
          public void hideBalloon() {
            try {
              super.hideBalloon();
            }
            finally {
              NotificationsConfigurationImpl.remove(groupID);
            }
          }
        }));
      }
    });
  }

  public interface ApplicationRunnable {
    void run(Application app);

    default void runFailed() {
    }
  }

  public static void invokeOnEdtWhenApplicationLoaded(final ApplicationRunnable action) {
    SwingUtilities.invokeLater(new Runnable() {
      @Override
      public void run() {
        final Application app = ApplicationManager.getApplication();
        if (app == null || (!LoadingState.APP_STARTED.isOccurred())) {
          // schedule with some delay to avoid too frequent event posting to EDT
          EdtExecutorService.getScheduledExecutorInstance().schedule(this, 3000L, TimeUnit.MILLISECONDS);
        }
        else {
          if (!app.isDisposed()) {
            action.run(app);
          }
          else {
            action.runFailed();
          }
        }
      }
    });
  }

  public static void executeOnApplicationExit(final Runnable action) {
    invokeOnEdtWhenApplicationLoaded(new ApplicationRunnable() {
      @Override
      public void run(Application app) {
        app.getMessageBus().connect().subscribe(AppLifecycleListener.TOPIC, new AppLifecycleListener() {
          @Override
          public void appWillBeClosed(boolean isRestart) {
            action.run();
          }
        });
      }

      @Override
      public void runFailed() {
        action.run();
      }
    });
  }

  /**
   * @return true if the action succeeded and action has been executed, false otherwise
   */
  public static boolean invokeAndWaitInNonModalContext(final Runnable action) {
    final Ref<Boolean> res = new Ref<>(Boolean.TRUE);
    EdtInvocationManager.invokeAndWaitIfNeeded(() -> {
      final Application app = ApplicationManager.getApplication();
      if (app != null && !ModalityState.NON_MODAL.equals(app.getCurrentModalityState())) {
        res.set(Boolean.FALSE);
      }
      else {
        action.run();
      }
    });
    return res.get();
  }
}
