package com.intellij.ide.license.ui;

import com.intellij.ide.BrowserUtil;
import com.intellij.ide.license.LicenseBundle;
import com.intellij.ide.license.LicenseMessages;
import com.intellij.ide.license.ProductDescriptor;
import com.intellij.ide.license.impl.*;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.util.Ref;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.text.DateFormatUtil;
import com.intellij.util.ui.UIUtil;

import javax.swing.*;
import java.text.DateFormat;
import java.util.*;

public class LicenseMessagesImpl implements LicenseMessages {
  protected LicenseUrls myLicenseUrls;

  public LicenseMessagesImpl(LicenseUrls urls) {
    myLicenseUrls = urls;
  }

  @Override
  public boolean askEnterKeyBecauseLicenseExpired(ProductDescriptor pd, boolean isPlugin, final boolean isEvaluation) {
    final Ref<Integer> rc = new Ref<>();
    UIUtil.invokeAndWaitIfNeeded((Runnable)() -> {

      final String notificationText = isEvaluation?
        LicenseBundle.message("message.evaluation.has.expired", pd.getShortName()) :
        LicenseBundle.message("message.license.has.expired", pd.getShortName());

      rc.set(Messages.showDialog(
        JOptionPane.getRootFrame(), notificationText,
        LicenseBundle.message(isEvaluation ? "title.evaluation.license.expired" : "title.license.expired"),
        new String[]{LicenseBundle.message("button.enter.key"),
          LicenseBundle.message("button.upgrade.online.now"),
          LicenseBundle.message("button.ok")},
        0, Messages.getErrorIcon()
      ));
      if (rc.get() == 1) {
        BrowserUtil.browse(isPlugin? myLicenseUrls.getPluginUpgradeUrl(pd.getProductCode()) : myLicenseUrls.getUpgradeUrl());
      }
    });
    return rc.get() == 0;
  }

  @Override
  public void showLicenseInvalidMessage(ProductDescriptor pd) {
    UIUtil.invokeAndWaitIfNeeded((Runnable)() -> Messages.showMessageDialog(JOptionPane.getRootFrame(),
                                                                            LicenseBundle.message("message.invalid.license.data"),
                                                                            LicenseBundle.message("title.invalid.license.data"),
                                                                            Messages.getErrorIcon()));
  }

  @Override
  public boolean showWillExpireWarning(ProductDescriptor pd, final boolean isPlugin, final long daysBeforeExpiration, final boolean evaluation) {
    final Ref<Integer> rc = new Ref<>();
    UIUtil.invokeAndWaitIfNeeded((Runnable)() -> {
      String days = LicenseBundle.message("message.evaluation.N.days", daysBeforeExpiration);
      if (daysBeforeExpiration == 1) {
        days = LicenseBundle.message("message.evaluation.one.day");
      }
      else if (daysBeforeExpiration == 0) {
        days = LicenseBundle.message("message.evaluation.less.than.one.day");
      }
      final String title = evaluation ? LicenseBundle.message("title.product.evaluation", pd.getShortName()) : LicenseBundle.message(
        "title.license.will.expire", pd.getShortName());
      final String content = evaluation
                             ? LicenseBundle.message("message.evaluation.will.expire", pd.getShortName(), days)
                             : LicenseBundle.message("message.license.will.expire", pd.getShortName(), days);
      rc.set(
        Messages.showDialog(
          JOptionPane.getRootFrame(),
          content,
          title,
          new String[]{LicenseBundle.message("button.enter.license"),
            LicenseBundle.message("button.upgrade.online.now"), LicenseBundle.message("button.continue.evaluation")},
          0,
          Messages.getWarningIcon())
      );
    });
    if (rc.get() == 1) {
      BrowserUtil.browse(isPlugin ? myLicenseUrls.getPluginUpgradeUrl(pd.getProductCode()) : myLicenseUrls.getUpgradeUrl());
    }
    return (rc.get() == 0);
  }

  @Override
  public void showSaveLicenseWarning(ProductDescriptor pd, final TargetSaveException exception) {
    UIUtil.invokeAndWaitIfNeeded((Runnable)() -> {
      Throwable cause = exception.getCause();
      String message = cause == null ? exception.getMessage() : cause.getMessage();
      Messages.showMessageDialog(JOptionPane.getRootFrame(),
                                    LicenseBundle.message("error.saving.license.data", message),
                                    LicenseBundle.message("title.unable.to.save.data"),
                                    Messages.getErrorIcon());
    });
  }

  @Override
  public List<String> getUsageRestrictionsMessages(ProductDescriptor pd, LicenseData license) {
    List<String> result = new ArrayList<>();

    if (license instanceof CertKeyLicense) {
      final CertKeyLicense certKeyLicense = (CertKeyLicense)license;
      final CertKey2 certKey = certKeyLicense.getCertKey();
      if (Boolean.TRUE.equals(certKey.eval(new IsPerpetualForProduct(certKeyLicense.getProductCode(), certKeyLicense.getProductBuildDate())))) {
        result.add(LicenseBundle.message("label.you.have.a.perpetual.fallback.license.for.this.version"));
      }

      final Date expirationDate = license.getExpirationDate();
      final Date firstExpiredDate = (Date)certKey.eval(new GetFirstExpiredDate(certKeyLicense.getProductCode()));
      if (firstExpiredDate != null && expirationDate != null) {
        result.add(
          new Date().before(firstExpiredDate)
          ? LicenseBundle.message("label.subscription.is.active.until.0", DateFormatUtil.formatAboutDialogDate(expirationDate))
          : LicenseBundle.message("label.subscription.expired.on.0", DateFormatUtil.formatAboutDialogDate(firstExpiredDate))
        );
      }
      final String restrictionsText = license.getRestrictionsText();
      if (!StringUtil.isEmpty(restrictionsText)) {
        result.add(restrictionsText.endsWith(".") ? restrictionsText : restrictionsText + ".");
      }
      return result;
    }

    final StringBuilder builder = new StringBuilder();
    final Date expirationDate = license.getExpirationDate();
    if (expirationDate != null) {
      String expirationText = LicenseBundle.message("message.expiration.date", dateString(expirationDate));
      if (license.isYearAcademic()) {
        builder.append(LicenseBundle.message("message.educational.license", expirationText));
      }
      else if (license.isOpenSource()) {
        builder.append(LicenseBundle.message("message.open.source.project.license", expirationText));
      }
      else {
        builder.append(expirationText);
      }
    }
    else if (license.isNonCommercial()) {
      builder.append(LicenseBundle.message("message.non.commercial.use.only"));
    }
    else if (license.isPersonal()) {
      builder.append(LicenseBundle.message("message.personal.license"));
    }

    final String restrictionsText = license.getRestrictionsText();
    if (restrictionsText != null && restrictionsText.length() > 0) {
      final boolean firstMessagePartExists = builder.length() > 0;
      if (firstMessagePartExists) {
        builder.append(". ");
      }
      builder.append(restrictionsText);
      if (firstMessagePartExists && !restrictionsText.endsWith(".")) {
        builder.append(".");
      }
    }
    return Collections.singletonList(builder.toString());
  }


  @Override
  public boolean showLicenseExpired(ProductDescriptor pd, boolean isPlugin) {
    final Ref<Integer> rc = new Ref<>();
    UIUtil.invokeAndWaitIfNeeded((Runnable)() -> {
      final String title = LicenseBundle.message("title.license.expired", pd.getShortName());
      final String content = LicenseBundle.message("message.license.expired");
      final String[] options = {LicenseBundle.message("button.enter.license"), LicenseBundle.message("button.upgrade.online.now")};
      rc.set(Messages.showDialog(JOptionPane.getRootFrame(), content, title, options, 0, Messages.getErrorIcon()));
    });
    if (rc.get() == 1) {
      BrowserUtil.browse(isPlugin? myLicenseUrls.getPluginPurchaseUrl(pd.getProductCode()) : myLicenseUrls.getPurchaseUrl());
    }
    return (rc.get() == 0);
  }

   @Override
   public void showLicenseExpiredBeforeShutdownNotification(ProductDescriptor pd, boolean isPlugin, boolean isEval) {
    final String title = isPlugin?
      LicenseBundle.message(isEval? "title.plugin.evaluation.license.expired" : "title.plugin.license.expired", pd.getShortName()):
      LicenseBundle.message(isEval? "title.evaluation.license.expired" : "title.license.expired");
    final String content = isPlugin?
      LicenseBundle.message(isEval? "message.plugin.evaluation.license.expired.shutdown" : "message.plugin.license.expired.shutdown", pd.getShortName()):
      LicenseBundle.message(isEval? "message.evaluation.license.expired.shutdown" : "message.license.expired.shutdown", pd.getShortName());
    int rc = Messages.showDialog(JOptionPane.getRootFrame(), content, title,
                                 new String[]{LicenseBundle.message("button.upgrade.online.now"), LicenseBundle.message("button.ok")},
                                 0, Messages.getErrorIcon());
    if (rc == 0) {
      BrowserUtil.browse(isPlugin? myLicenseUrls.getPluginPurchaseUrl(pd.getProductCode()) : myLicenseUrls.getPurchaseUrl());
    }
  }

  private static String dateString(Date date) {
    return DateFormat.getDateInstance(DateFormat.LONG, Locale.ENGLISH).format(date);
  }

  @Override
  public String getPurchaseUrl(ProductDescriptor pd, boolean isPlugin) {
    return isPlugin? myLicenseUrls.getPluginPurchaseUrl(pd.getProductCode()) : myLicenseUrls.getPurchaseUrl();
  }

  @Override
  public LicenseUrls getLicenseUrls() {
    return myLicenseUrls;
  }
}
