package com.intellij.ide.license.ui;

import com.intellij.icons.AllIcons;
import com.intellij.ide.BrowserUtil;
import com.intellij.ide.dnd.FileCopyPasteUtil;
import com.intellij.ide.license.*;
import com.intellij.ide.license.impl.*;
import com.intellij.licensecommon.license.LicenseData;
import com.intellij.licensecommon.license.MaintenanceDueAware;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.ApplicationNamesInfo;
import com.intellij.openapi.application.ex.ApplicationInfoEx;
import com.intellij.openapi.application.impl.ApplicationInfoImpl;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.ui.LoadingDecorator;
import com.intellij.openapi.ui.MessageDialogBuilder;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.ui.ex.MultiLineLabel;
import com.intellij.openapi.util.Computable;
import com.intellij.openapi.util.NlsContexts;
import com.intellij.openapi.util.NlsSafe;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.text.HtmlBuilder;
import com.intellij.openapi.util.text.HtmlChunk;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.util.text.TextWithMnemonic;
import com.intellij.openapi.wm.IdeFocusManager;
import com.intellij.ui.*;
import com.intellij.ui.components.JBScrollPane;
import com.intellij.ui.components.JBTextField;
import com.intellij.util.concurrency.AppExecutorUtil;
import com.intellij.util.io.HttpRequests;
import com.intellij.util.text.DateFormatUtil;
import com.intellij.util.ui.*;
import com.jetbrains.license.util.DnsUtil;
import com.jetbrains.license.util.ServiceConfig;
import org.intellij.lang.annotations.MagicConstant;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.border.Border;
import javax.swing.border.LineBorder;
import javax.swing.event.DocumentEvent;
import javax.swing.event.DocumentListener;
import javax.swing.event.HyperlinkEvent;
import javax.swing.text.DefaultEditorKit;
import javax.swing.text.JTextComponent;
import java.awt.*;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.*;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.lang.reflect.Method;
import java.net.MalformedURLException;
import java.net.URL;
import java.net.URLEncoder;
import java.nio.charset.StandardCharsets;
import java.util.List;
import java.util.*;
import java.util.concurrent.Future;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Consumer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;
import java.util.stream.Collectors;
import java.util.stream.Stream;

import static com.intellij.openapi.ui.DialogWrapper.ERROR_FOREGROUND_COLOR;
import static java.awt.GridBagConstraints.*;
import static javax.swing.ScrollPaneConstants.HORIZONTAL_SCROLLBAR_NEVER;
import static javax.swing.ScrollPaneConstants.VERTICAL_SCROLLBAR_AS_NEEDED;

@SuppressWarnings("UseJBColor")
final class LicensePanel extends JPanel implements LicenseEditor, ActionListener, Disposable {
  private static final Logger LOG = Logger.getInstance("#LicensePanel");
  private static final boolean allowReturnToEvaluation = false;
  private static final int ourLicenseKeyLines = 7;
  private static final int GAP = 15;
  private static final Pattern EMAIL_PATTERN = Pattern.compile("^[A-z0-9._%+-]+@[A-z0-9.-]+\\.[A-z]+$");

  private static final Pattern LICENSE_START_PATTERN1 = Pattern.compile("LICENSE BEGIN\\s={3,}[\\n\\r]+");
  private static final Pattern LICENSE_END_PATTERN1 = Pattern.compile("={3,}\\sLICENSE END");
  private static final Pattern LICENSE_PATTERN2 = Pattern.compile("License Key\\s?[\\d]*:\\s(.+)[\\n\\r]+");
  private static final Color MY_LINK_COLOR = JBColor.namedColor("Link.activeForeground", new JBColor(new Color(0x2470b3), DarculaColors.BLUE));

  private static final String LICENSED_CARD = "licensed";
  private static final String NOT_LICENSED_CARD = "not licensed";

  private static final Color myErrorColor = ERROR_FOREGROUND_COLOR;
  private final Color myOkColor = new Color(0x4b9649);
  private final Font myLabelFont = StartupUiUtil.getLabelFont();
  private final LoadingDecorator myDecorator;
  private boolean myIgnoreDocumentChanges = false;
  private final JLabel myMyKeyValidationLabel = new JLabel("", SwingConstants.RIGHT);
  private final SeparatorComponent mySeparator;

  private final CardLayout myMainCard = new CardLayout();
  private final JPanel myMainPanel = new JPanel(myMainCard);

  private enum State {
    JETPROFILE_INPUT, JETPROFILE_CHOOSE, KEY_INPUT, SERVER_INPUT, EVALUATE, LICENSED
  }

  private static final State DEFAULT = State.JETPROFILE_INPUT;

  private BuiltInEvaluationSource myEvalSource;
  private FileLicenseSource myKeySource;
  private JetprofileSource myAccountSource;
  private JBALicenseData myChosenAccountLicenseData = null;
  private LicenseServerSource myServerSource;
  @NotNull private State myState = DEFAULT;
  @NotNull private State myLastInputState = DEFAULT;

  private final JRadioButton myJetProfileButton;
  private final JRadioButton myKeyButton;
  private final JRadioButton myServerButton;

  private final JPanel myLicenseInfoPanel;
  private final JComponent myActivationTypePanel;

  private final JRadioButton myActivateButton;
  private final JRadioButton myEvaluateButton;
  private final Link myBuyLink;

  private final JButton myRemoveLicenseButton;
  private final JButton myNewLicenseButton;
  private final JPanel myRadioPanel;

  private final JPanel myFieldsPanel;
  private final JPanel myProfileChoosePanel;
  private final JTextField myAccountNameField;
  private final JPasswordField myAccountPasswordField;
  private final Link myForgotAccountPasswordLink;
  private final JTextArea myKeyArea;
  private final JTextField myServerAddressField;

  private final JButton myActivateJBALicenseButton;
  private final JButton myActivateJBAChooseLicenseButton;
  private final JButton myActivateKeyLicenseButton;
  private final JButton myActivateServerLicenseButton;
  private final JButton myActivateEvaluationPeriodButton;

  private final JButton myCancelJBALicenseButton;
  private final JButton myCancelKeyLicenseButton;
  private final JButton myCancelServerLicenseButton;
  private final JButton myCancelEvaluationPeriodButton;

  private final Map<State, JButton> defaultButtons = new HashMap<>();

  private final ErrorInfoBox myErrorInfo = new ErrorInfoBox();
  private final Map<State, Container> myErrorPlaceholders = new HashMap<>();

  private final JButton myDiscoverButton;
  private final JButton myTestConnButton;
  private final Link myServerMoreInfoLink;

  private final JPanel myReturnToEvaluationPanel;
  private final JButton myReturnToEvaluationButton;
  private final JLabel myEvalLabel;
  private final JCheckBox myNewsCb;
  private final JCheckBox myTipsCb;
  private final JBTextField mySubscriptionEmailField;

  private final AtomicBoolean myServerDiscoverLock = new AtomicBoolean(false);
  private final AtomicBoolean myJetProfileLicenseLock = new AtomicBoolean(false);
  private final AtomicBoolean myServerLicenseLock = new AtomicBoolean(false);
  private final AtomicBoolean myEmailSubscriptionLock = new AtomicBoolean(false);

  private static final long SECOND = 1000L;
  private static final long DEFAULT_TIMEOUT = 30 * SECOND;
  private static final long SMART_DISCOVER_TIMEOUT = 5 * SECOND;

  private static final Color SEPARATOR_COLOR = JBUI.CurrentTheme.CustomFrameDecorations.separatorForeground();

  private static final long UI_ASYNC_TASK_TIMEOUT;
  static {
    long timeoutBase;
    try {
      timeoutBase = Long.parseLong(System.getProperty("license.request.timeout", String.valueOf(DEFAULT_TIMEOUT)));
    }
    catch (Throwable ignored) {
      timeoutBase = DEFAULT_TIMEOUT;
    }
    UI_ASYNC_TASK_TIMEOUT = timeoutBase;
  }

  private final ProductDescriptor myProductDescriptor;
  private final LicenseEditingModel myModel;
  private LicenseData myLicense;

  private volatile boolean myDisposed = false;

  private final Collection<LicenseUpdateListener> myLicenseUpdateListeners = new ArrayList<>();

  LicensePanel(@NotNull ProductDescriptor pd, @NotNull LicenseEditingModel model, @Nullable @NlsContexts.DetailedDescription String initialProblem) {
    super(new GridLayout(1, 1));
    myProductDescriptor = pd;
    myModel = model;
    myLicense = getSelectedLicenseData();
    LicenseUrls licenseUrls = LicenseManager.getInstance().getMessages().getLicenseUrls();
    String activatedProductName = myModel.getDisplayName(myProductDescriptor);

    DocumentListener listener = new DocumentAdapter() {
      @Override
      protected void textChanged(@NotNull DocumentEvent e) {
        if (myIgnoreDocumentChanges) return;
        updateOKButton();
      }
    };
    myKeyArea = new HintTextArea(Math.max(1, ourLicenseKeyLines), 31, "Paste or drop activation code here");
    myKeyArea.setTransferHandler(new MyTransferHandler());
    myKeyArea.setFont(new JTextPane().getFont());
    myKeyArea.setLineWrap(true);
    LicenseEntryUtil.fixFocusTraversal(myKeyArea);
    myKeyArea.getDocument().addDocumentListener(listener);
    myServerAddressField = new JTextField(31);
    myServerAddressField.getDocument().addDocumentListener(listener);

    myRemoveLicenseButton = new JButton(new RemoveLicenseAction());
    myNewLicenseButton = new JButton(new NewLicenseAction());

    myAccountNameField = new JTextField(18);
    myAccountNameField.getDocument().addDocumentListener(listener);
    myAccountPasswordField = new JPasswordField(18);
    myAccountPasswordField.getDocument().addDocumentListener(listener);
    addPasteAndUndoActions(myAccountNameField);
    addPasteAndUndoActions(myAccountPasswordField);
    addPasteAndUndoActions(myKeyArea);
    addPasteAndUndoActions(myServerAddressField);
    myForgotAccountPasswordLink = new Link(LicenseBundle.message("link.forgot"), "https://account.jetbrains.com/secured/ForgotPasswordForm");
    myDiscoverButton = createButton(LicenseBundle.message("button.discover.server"));
    myTestConnButton = createButton(LicenseBundle.message("button.test.connection"));
    myServerMoreInfoLink = new Link(LicenseBundle.message("link.about"), licenseUrls.getLicenseServerInfoUrl());
    myEvalLabel = new JLabel(LicenseBundle.message("label.evaluation.is.free.for.0.days", 30));
    myTipsCb = new JCheckBox(LicenseBundle.message("checkbox.send.me.tips.to.learn.0.faster", activatedProductName));
    myNewsCb = new JCheckBox(LicenseBundle.message("checkbox.tell.me.about.new.product.features.as.they.come.out"));
    ActionListener checkBoxListener = new ActionListener() {
      @Override
      public void actionPerformed(ActionEvent e) {
        mySubscriptionEmailField.setEnabled(!areTipsAvailable() || myNewsCb.isSelected() || myTipsCb.isSelected());
        updateOKButton();
      }
    };
    myTipsCb.addActionListener(checkBoxListener);
    myNewsCb.addActionListener(checkBoxListener);
    mySubscriptionEmailField = new JBTextField();
    mySubscriptionEmailField.setEnabled(!areTipsAvailable());
    mySubscriptionEmailField.getDocument().addDocumentListener(listener);

    myActivateJBALicenseButton = createButton(LicenseBundle.message("button.activate"));
    myActivateJBAChooseLicenseButton = createButton(LicenseBundle.message("button.activate"));
    myActivateKeyLicenseButton = createButton(LicenseBundle.message("button.activate"));
    myActivateServerLicenseButton = createButton(LicenseBundle.message("button.activate"));
    myActivateEvaluationPeriodButton = createButton(LicenseBundle.message("button.evaluate"));

    BackToLicenseInfoAction cancelAction = new BackToLicenseInfoAction();
    myCancelJBALicenseButton = new JButton(cancelAction);
    myCancelKeyLicenseButton = new JButton(cancelAction);
    myCancelServerLicenseButton = new JButton(cancelAction);
    myCancelEvaluationPeriodButton = new JButton(cancelAction);
    setBackToLicenseInfoPossible(false);

    defaultButtons.put(State.JETPROFILE_INPUT, myActivateJBALicenseButton);
    defaultButtons.put(State.JETPROFILE_CHOOSE, myActivateJBAChooseLicenseButton);
    defaultButtons.put(State.KEY_INPUT, myActivateKeyLicenseButton);
    defaultButtons.put(State.SERVER_INPUT, myActivateServerLicenseButton);
    defaultButtons.put(State.EVALUATE, myActivateEvaluationPeriodButton);

    myLicenseInfoPanel = new JPanel(new GridBagLayout()) {
      @Override
      public Dimension getMinimumSize() {
        Dimension size = super.getMinimumSize();
        size.height = Math.max(getHeight(), size.height);
        return size;
      }
    };
    myMainPanel.add(LICENSED_CARD, myLicenseInfoPanel);

    if (activatedProductName.length() > 25) {
      activatedProductName = activatedProductName.substring(0, 24) + "...";
    }

    myActivateButton = createRadioButton(LicenseBundle.message("radio.button.activate.0", activatedProductName), null);
    myEvaluateButton = createRadioButton(LicenseBundle.message("radio.button.evaluate.for.free"), State.EVALUATE);
    myActivateButton.setOpaque(false);
    myEvaluateButton.setOpaque(false);

    myBuyLink = createBuyLink(licenseUrls);
    myActivationTypePanel = createActivationTypePanel();
    JPanel newLicensePanel = new JPanel(new GridBagLayout());
    myMainPanel.add(NOT_LICENSED_CARD, newLicensePanel);

    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = HORIZONTAL;
    gbc.gridwidth = REMAINDER;
    gbc.anchor = WEST;
    gbc.insets = JBUI.insets(10, GAP, 0, GAP);
    gbc.weightx = 1;
    gbc.weighty = 0;
    newLicensePanel.add(myActivationTypePanel, gbc);

    mySeparator = new SeparatorComponent(SEPARATOR_COLOR, SeparatorOrientation.HORIZONTAL);
    gbc.anchor = CENTER;
    gbc.insets = JBUI.insets(10, GAP);
    newLicensePanel.add(mySeparator, gbc);

    gbc.anchor = WEST;
    gbc.insets = JBUI.insets(0, GAP);
    newLicensePanel.add(myRadioPanel = new JPanel(new GridBagLayout()), gbc);

    myFieldsPanel = new JPanel(new CardLayout());
    myFieldsPanel.add(State.JETPROFILE_INPUT.name(), createProfileInputPanel());
    myFieldsPanel.add(State.JETPROFILE_CHOOSE.name(), myProfileChoosePanel = new JPanel(new GridBagLayout()));
    myFieldsPanel.add(State.KEY_INPUT.name(), createKeyInputPanel());
    myFieldsPanel.add(State.SERVER_INPUT.name(), createServerInputPanel());
    myFieldsPanel.add(State.EVALUATE.name(), createEvaluateInputPanel());

    gbc.weighty = 1;
    gbc.fill = BOTH;
    gbc.anchor = NORTH;
    newLicensePanel.add(myFieldsPanel, gbc);

    gbc.weighty = 0;
    newLicensePanel.add(myReturnToEvaluationPanel = new JPanel(new GridBagLayout()), gbc);

    myDecorator = new LoadingDecorator(myMainPanel, this, 300);
    MouseAdapter mouseAdapter = new MouseAdapter() {
    };
    Component[] components = ((JLayeredPane)myDecorator.getComponent()).getComponentsInLayer(JLayeredPane.DRAG_LAYER);
    if (components.length > 0) {
      Component layer = components[0];
      layer.addMouseListener(mouseAdapter);
      layer.addMouseMotionListener(mouseAdapter);
      layer.addMouseWheelListener(mouseAdapter);
      layer.addKeyListener(new KeyAdapter() {
        @Override
        public void keyPressed(KeyEvent e) {
          if (e.getKeyCode() == KeyEvent.VK_TAB) e.consume();
        }
      });
    }
    add(myDecorator.getComponent());

    ButtonGroup topGroup = new ButtonGroup();
    topGroup.add(myActivateButton);
    topGroup.add(myEvaluateButton);
    myJetProfileButton = createRadioButton(LicenseBundle.message("radio.button.jb.account"), State.JETPROFILE_INPUT);
    myKeyButton = createRadioButton(LicenseBundle.message("radio.button.activation.code"), State.KEY_INPUT);
    myServerButton = createRadioButton(LicenseBundle.message("radio.button.license.server"), State.SERVER_INPUT);
    myReturnToEvaluationButton = createButton(LicenseBundle.message("button.return.to.evaluation"));

    updateState(guessState());
    myActivateButton.setSelected(true);
    doAsyncTask(new AtomicBoolean(false), () -> {
      try {
        return ServiceConfig.getResetPasswordUrl();
      }
      catch (IOException ignored) {
      }
      return null;
    }, resetPassUrl -> {
      if (resetPassUrl != null) {
        myForgotAccountPasswordLink.setUrl(resetPassUrl);
      }
    }, SMART_DISCOVER_TIMEOUT, false
    );

    if (initialProblem != null) {
      myErrorInfo.showInitialError(initialProblem, ErrorInfoBox.ERROR_TYPE);
    }
  }

  @NotNull
  private JPanel createActivationTypePanel() {
    JPanel res = new JPanel(new GridBagLayout());

    GridBagConstraints gbc = new GridBagConstraints();
    gbc.anchor = WEST;
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.weightx = 0;
    gbc.fill = NONE;
    gbc.insets = JBUI.insets(5, 0, 5, 10);
    res.add(myActivateButton, gbc);

    gbc.gridx++;
    res.add(myEvaluateButton, gbc);

    gbc.gridx++;
    gbc.weightx = 1;
    gbc.fill = HORIZONTAL;
    res.add(Box.createHorizontalGlue(), gbc);

    gbc.gridx++;
    gbc.weightx = 0;
    gbc.insets = JBUI.insets(5, 20, 5, 5);
    res.add(myBuyLink, gbc);

    return res;
  }

  @NotNull
  private Link createBuyLink(LicenseUrls licenseUrls) {
    return new Link(LicenseBundle.message("link.buy.license"), myProductDescriptor.equals(myModel.getPlatformProduct())
                                                               ? licenseUrls.getPurchaseUrl()
                                                               : licenseUrls.getPluginPurchaseUrl(myProductDescriptor.getProductCode()));
  }

  @Override
  public void dispose() {
    myDisposed = true;
  }

  @Override
  public void proceedSelectedLicenseSource() {
    myErrorInfo.clear(true);
    if (myState == State.JETPROFILE_INPUT) {
      final String login = myAccountNameField.getText().trim();
      if (StringUtil.isEmpty(login) || myAccountPasswordField.getPassword().length == 0) {
        return;
      }
      doAsyncTask(myJetProfileLicenseLock, () -> {
        try {
          myAccountSource.create(new com.jetbrains.license.util.Pair<>(login, myAccountPasswordField.getPassword()));
        }
        catch (LicenseCreationException e) {
          showErrorMessage(e.getMessage(), e.getDetails(), e.getCause() instanceof IOException && LicensingProxyService.isActive());
        }
        return myAccountSource.getLicense();
      }, data -> {
        if (data != null) {
          proceedLicenseAndUpdateState(myAccountSource, State.JETPROFILE_INPUT);
        }
        else if (!myAccountSource.getAvailableLicenses().isEmpty()) {
          updateState(State.JETPROFILE_CHOOSE);
        }
        else {
          setBackToLicenseInfoPossible(false);
          notifyLicenseListenersIfNeeded();
        }
      }, UI_ASYNC_TASK_TIMEOUT, true
      );
      return;
    }

    if (myState == State.JETPROFILE_CHOOSE) {
      if (myChosenAccountLicenseData != null) {
        myAccountSource.setLicense(myChosenAccountLicenseData);
        proceedLicenseAndUpdateState(myAccountSource, State.JETPROFILE_INPUT);
      }
      return;
    }

    if (myState == State.KEY_INPUT) {
      try {
        LicenseData licenseData = tryCreateKeyLicense(myKeyArea.getText());
        if (licenseData == null) {
          setBackToLicenseInfoPossible(false);
          return;
        }
        myKeySource.create(licenseData);
        proceedLicenseAndUpdateState(myKeySource, State.KEY_INPUT);
      }
      catch (ValidationFailedException e) {
        String error = e.getMessage() != null ? e.getMessage() : LicenseBundle.message("message.text.key.invalid");
        myErrorInfo.showError(error, ErrorInfoBox.ERROR_TYPE, "https://intellij-support.jetbrains.com/hc/en-us/articles/207241025");
      }
      return;
    }

    if (myState == State.SERVER_INPUT) {
      doAsyncTask(myServerLicenseLock, (Computable<LicenseData>)() -> {
        myServerSource.releaseLicense();
        try {
          myServerSource.create(myServerAddressField.getText().trim());
        }
        catch (AuthorizationRequiredException e) {
          showErrorMessage(e.getMessage(), e.getDetails(), false, true);
        }
        catch (LicenseCreationException e) {
          showErrorMessage(e.getMessage(), e.getDetails(), e.getCause() instanceof IOException && LicensingProxyService.isActive());
        }
        return myServerSource.getLicense();
      }, data -> {
        if (data != null) {
          proceedLicenseAndUpdateState(myServerSource, State.SERVER_INPUT);
        }
        else {
          myServerSource.restore();
        }
      }, UI_ASYNC_TASK_TIMEOUT, true
      );
      return;
    }
    if (myState == State.EVALUATE) {
      try {
        myEvalSource.create(null);
        String email = mySubscriptionEmailField.getText();
        String formData = getFormData();
        if (formData != null && isValidEmail(email)) {
          doSubscribeWithEmail(email, formData);
        }
        proceedLicenseAndUpdateState(myEvalSource, State.JETPROFILE_INPUT);
      }
      catch (LicenseCreationException e) {
        showErrorMessage(e.getMessage(), e.getDetails(), false);
      }
    }
  }

  void addLicenseUpdateListener(LicenseUpdateListener licenseUpdateListener) {
    myLicenseUpdateListeners.add(licenseUpdateListener);
  }

  private void proceedLicense(LicenseSource source) {
    try {
      myModel.activateLicense(myProductDescriptor, source);
      notifyLicenseListenersIfNeeded();
    }
    catch (Exception e) {
      throw new RuntimeException(e);  // todo
    }
  }

  private void proceedLicenseAndUpdateState(LicenseSource source, @NotNull State state) {
    try {
      proceedLicense(source);
    }
    finally {
      updateState(getSelectedLicenseData() == null ? state : State.LICENSED);
    }
  }

  private void notifyLicenseListenersIfNeeded() {
    LicenseData newLicense = getSelectedLicenseData();
    if (!Objects.equals(newLicense, myLicense)) {
      myLicenseUpdateListeners.forEach(l -> l.licenseUpdated(myProductDescriptor));
    }
    myLicense = newLicense;
  }

  private static boolean isValidEmail(String s) {
    return s != null && EMAIL_PATTERN.matcher(s).matches();
  }


  private void doSubscribeWithEmail(@NotNull String email, @NotNull String formData) {
    doAsyncTask(myEmailSubscriptionLock, () -> {
      try {
        String[] mx = {"QVD", "" + 114, "" + 426, "munc", "hkinId"};
        HttpRequests
          .post("https://app-lon02.marketo.com/index.php/leadCapture/save", "application/x-www-form-urlencoded; charset=UTF-8")
          .tuner(connection -> connection.setRequestProperty("Origin", "https://www.jetbrains.com"))
          .write(("Email=" + URLEncoder.encode(email, StandardCharsets.UTF_8.name()) + "&" + formData +
                       "&" + mx[3] + mx[4] + "=" + mx[2] + "-" + mx[0] + "-" + mx[1] + "&_mkt_trk=" +
                       "&_mktoReferrer=" + URLEncoder.encode("https://www.jetbrains.com/", StandardCharsets.UTF_8.name()))
                        .getBytes(StandardCharsets.UTF_8));
        return Boolean.TRUE;
      }
      catch (IOException e) {
        LOG.info("Email subscription failed");
        return Boolean.FALSE;
      }
    }, result -> {}, UI_ASYNC_TASK_TIMEOUT, false);
  }

  @Nullable
  private String getFormData() {
    if (!myProductDescriptor.equals(myModel.getPlatformProduct())/*is plugin*/) {
      return null;
    }
    ApplicationInfoEx appInfo = (ApplicationInfoEx)getAppInfo();
    String formId = appInfo.getSubscriptionFormId();
    if (formId == null) {
      return null;
    }

    StringBuilder form = new StringBuilder("formid=").append(formId).append("&formVid=").append(formId);
    String additionalFormData = appInfo.getSubscriptionAdditionalFormData();
    if (!StringUtil.isEmpty(additionalFormData)) {
      form.append('&').append(additionalFormData);
    }
    if (!appInfo.areSubscriptionTipsAvailable() || myNewsCb.isSelected()) {
      form.append('&').append(appInfo.getSubscriptionNewsKey()).append('=').append(appInfo.getSubscriptionNewsValue());
    }
    if (appInfo.areSubscriptionTipsAvailable() && myTipsCb.isSelected()) {
      form.append('&').append(appInfo.getSubscriptionTipsKey()).append("=yes");
    }
    return form.toString();
  }

  private static boolean areTipsAvailable() {
    return ((ApplicationInfoEx)getAppInfo()).areSubscriptionTipsAvailable();
  }

  @Override
  public void addNotify() {
    super.addNotify();
    Window window = SwingUtilities.getWindowAncestor(this);
    if (window != null) {
      FocusTraversalPolicy policy = window.getFocusTraversalPolicy();
      final List<? extends Component> order = Arrays.asList(myAccountNameField, myAccountPasswordField, myForgotAccountPasswordLink);
      if (policy instanceof SortingFocusTraversalPolicy) {
        try {
          Method method = SortingFocusTraversalPolicy.class.getDeclaredMethod("getComparator");
          method.setAccessible(true);
          final Comparator<? super Component> comparator = (Comparator<? super Component>)method.invoke(policy);
          method = SortingFocusTraversalPolicy.class.getDeclaredMethod("setComparator", Comparator.class);
          method.setAccessible(true);
          method.invoke(policy, (Comparator<Component>)(o1, o2) -> {
            int index1 = order.indexOf(o1);
            int index2 = order.indexOf(o2);
            if (index1 == -1 || index2 ==-1) return comparator.compare(o1, o2);
            return index1 - index2;
          });
        }
        catch (Exception ignored) {
        }
      }
      updateState(myState);
    }
  }

  private void showErrorMessage(@NlsContexts.DetailedDescription @NotNull final String message, @NlsContexts.DetailedDescription @Nullable final String details, final boolean suggestProxyConfig) {
    showErrorMessage(message, details, suggestProxyConfig, false);
  }

  private void showErrorMessage(@NlsContexts.DetailedDescription @NotNull final String message, @NlsContexts.DetailedDescription @Nullable final String details, final boolean suggestProxyConfig, boolean clearErrorOnLinkPress) {
    if (myDisposed) {
      return;
    }
    UIUtil.invokeLaterIfNeeded(new Runnable() {
      @Override
      public void run() {
        final String detailsMessage = StringUtil.isEmpty(details)? message : details;

        if (!suggestProxyConfig) {
          myErrorInfo.showError(detailsMessage, ErrorInfoBox.ERROR_TYPE);
          myErrorInfo.setClearOnLinkPress(clearErrorOnLinkPress);
          return;
        }

        final JComponent detailsComponent = createErrorMessageComponent(detailsMessage);

        final LicensingProxyService proxyService = LicensingProxyService.getInstance();
        Pair<String, String> prevHostPort = proxyService.getProxyConfig();
        if (prevHostPort == null) {
          prevHostPort = proxyService.getDiscoveredConfig();
        }

        JPanel panel = new JPanel(new GridBagLayout());
        panel.add(detailsComponent, new GridBagConstraints(0, 0, 2, 1, 1.0, 0.0, NORTHWEST, HORIZONTAL, JBUI.emptyInsets(), 0, 0));
        String text;
        if (prevHostPort == null) {
          text = LicenseBundle.message("warning.text.your.host.may.be.behind.a.proxy.0.was.not.able.to.detect.your.proxy.configuration",
                                       ApplicationNamesInfo.getInstance().getFullProductName());
        }
        else {
          text = LicenseBundle.message("lable.0.was.not.able.to.connect.using.specified.proxy", ApplicationNamesInfo.getInstance().getFullProductName());
        }
        final JLabel proxyHeader = new MultiLineLabel(text);
        panel.add(proxyHeader, new GridBagConstraints(0, RELATIVE, 2, 1, 1.0, 0.0, NORTHWEST, HORIZONTAL, JBUI.insets(5, 0, 10, 0), 0, 0));

        panel.add(new JLabel(LicenseBundle.message("label.proxy.host")), new GridBagConstraints(0, RELATIVE, 1, 1, 0.0, 0.0, NORTHWEST, HORIZONTAL, JBUI.emptyInsets(), 0, 0));
        final JTextField hostField = new JTextField();
        panel.add(hostField, new GridBagConstraints(1, RELATIVE, 1, 1, 1.0, 0.0, NORTHWEST, HORIZONTAL, JBUI.insetsLeft(5), 0, 0));

        panel.add(new JLabel(LicenseBundle.message("label.proxy.port")), new GridBagConstraints(0, RELATIVE, 1, 1, 0.0, 1.0, NORTHWEST, HORIZONTAL, JBUI.insetsTop(5), 0, 0));
        final JTextField portField = new JTextField();
        panel.add(portField, new GridBagConstraints(1, RELATIVE, 1, 1, 1.0, 1.0, NORTHWEST, HORIZONTAL, JBUI.insets(5, 5, 0, 0), 0, 0));

        if (prevHostPort != null) {
          hostField.setText(prevHostPort.first);
          portField.setText(prevHostPort.second);
        }

        JOptionPane.showMessageDialog(LicensePanel.this, panel, StringUtil.isEmpty(details) ? LicenseBundle.message("message.text.error")
                                                                                            : message, JOptionPane.ERROR_MESSAGE);

        proxyService.setUserDefinedProxy(hostField.getText(), portField.getText());
      }

      @NotNull
      private JComponent createErrorMessageComponent(@NlsContexts.DetailedDescription String detailsMessage) {
        if (detailsMessage.length() < 40) {
          return new JLabel(detailsMessage);
        }
        JTextArea textArea = new JTextArea(detailsMessage, 3, 64);
        textArea.setFont(UIManager.getFont("Label.font"));
        textArea.setLineWrap(true);
        textArea.setWrapStyleWord(true);
        textArea.setOpaque(false);
        textArea.setEditable(false);
        textArea.setBorder(JBUI.Borders.empty());
        //noinspection UndesirableClassUsage
        final JScrollPane scrollPane = new JScrollPane(textArea, VERTICAL_SCROLLBAR_AS_NEEDED, HORIZONTAL_SCROLLBAR_NEVER);
        scrollPane.setOpaque(false);
        scrollPane.getViewport().setOpaque(false);
        scrollPane.setBorder(new LineBorder(JBColor.border(), 1) {
          @Override
          public void paintBorder(Component c, Graphics g, int x, int y, int width, int height) {
            if (!scrollPane.getVerticalScrollBar().isVisible()) {
              return;
            }
            super.paintBorder(c, g, x, y, width, height);
          }
        });
        return scrollPane;
      }
    });
  }

  @Override
  public void createUI() {
  }

  @NotNull
  @Override
  public JPanel getPanel() {
    return this;
  }

  @Override
  public Dimension getPreferredSize() {
    return myMainPanel.getPreferredSize();
  }

  private static void addPasteAndUndoActions(JTextComponent textComponent) {
    JPopupMenu popupMenu = new JPopupMenu();
    JMenuItem pasteItem = new JMenuItem(new DefaultEditorKit.PasteAction());
    pasteItem.setText(LicenseBundle.message("menu.item.paste"));
    popupMenu.add(pasteItem);
    JMenuItem copyItem = new JMenuItem(new DefaultEditorKit.CopyAction());
    copyItem.setText(LicenseBundle.message("menu.item.copy"));
    popupMenu.add(copyItem);
    textComponent.setComponentPopupMenu(popupMenu);
    UIUtil.addUndoRedoActions(textComponent);
  }

  private JPanel createProfileInputPanel() {
    JPanel profileInputPanel = new JPanel(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints();

    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.anchor = WEST;
    gbc.fill = HORIZONTAL;
    gbc.weightx = 0;
    gbc.insets = JBUI.insets(5, 0, 0, 5);
    profileInputPanel.add(new JLabel(LicenseBundle.message("label.username.email")), gbc);

    gbc.gridx++;
    gbc.weightx = 1;
    gbc.gridwidth = 2;
    gbc.insets = JBUI.insets(5, 5, 0, 5);
    profileInputPanel.add(myAccountNameField, gbc);

    gbc.gridx = 0;
    gbc.gridy++;
    gbc.weightx = 0;
    gbc.gridwidth = 1;
    gbc.insets = JBUI.insets(5, 0, 0, 5);
    profileInputPanel.add(new JLabel(LicenseBundle.message("label.password")), gbc);

    gbc.gridx++;
    gbc.weightx = 1;
    gbc.insets = JBUI.insets(5, 5, 0, 5);
    profileInputPanel.add(myAccountPasswordField, gbc);

    gbc.gridx++;
    gbc.weightx = 0;
    profileInputPanel.add(myForgotAccountPasswordLink, gbc);

    gbc.gridx = 1;
    gbc.gridy++;
    gbc.fill = NONE;
    gbc.insets = JBUI.insets(15, 5, 0, 5);
    profileInputPanel.add(createButtonsPanel(myActivateJBALicenseButton, myCancelJBALicenseButton), gbc);

    gbc.gridy++;
    gbc.gridwidth = 2;
    gbc.fill = HORIZONTAL;
    gbc.insets = JBUI.insets(5, 5, 0, 5);
    profileInputPanel.add(createAndRegisterErrorPlaceholder(State.JETPROFILE_INPUT), gbc);

    gbc.gridy++;
    gbc.gridx = 0;
    gbc.gridwidth = 3;
    gbc.weighty = 1;
    profileInputPanel.add(Box.createVerticalGlue(), gbc);

    return profileInputPanel;
  }

  private JPanel createKeyInputPanel() {
    JPanel keyInputPanel = new JPanel(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints();

    gbc.gridx = 0;
    gbc.gridy++;
    gbc.anchor = NORTHWEST;
    gbc.weightx = 0;
    gbc.insets = JBUI.insets(5, 0, 0, 5);
    keyInputPanel.add(new JLabel(LicenseBundle.message("label.activation.code")), gbc);

    gbc.gridx++;
    gbc.weightx = 1;
    gbc.weighty = 1;
    gbc.fill = BOTH;
    gbc.insets = JBUI.insets(5, 5, 0, 5);
    final JScrollPane scrollPane = new JScrollPane(myKeyArea, VERTICAL_SCROLLBAR_AS_NEEDED, HORIZONTAL_SCROLLBAR_NEVER) {
      @Override
      public Dimension getMinimumSize() {
        return getPreferredSize();
      }

      @Override
      public Dimension getPreferredSize() {
        Dimension size = super.getPreferredSize();
        size.height+=2;
        return size;
      }
    };
    keyInputPanel.add(scrollPane, gbc);

    gbc.gridy++;
    gbc.weighty = 0;
    gbc.fill = NONE;
    gbc.anchor = WEST;
    keyInputPanel.add(createButtonsPanel(myActivateKeyLicenseButton, myCancelKeyLicenseButton), gbc);

    gbc.gridx = 0;
    gbc.gridy++;
    gbc.weightx = 0;
    @NonNls String invisibleText = "Size Holder";
    final JLabel invisibleSizeBase = new JLabel(invisibleText);
    keyInputPanel.add(new JLabel(){
      @Override
      public Dimension getPreferredSize() {
        return invisibleSizeBase.getPreferredSize();
      }
    }, gbc);

    gbc.gridx = 1;
    gbc.fill = HORIZONTAL;
    gbc.weighty = 0;
    gbc.weightx = 1;
    keyInputPanel.add(createAndRegisterErrorPlaceholder(State.KEY_INPUT), gbc);

    return keyInputPanel;
  }

  private JPanel createServerInputPanel() {
    JPanel serverInputPanel = new JPanel(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.insets = JBUI.insets(5, 0, 5, 5);
    gbc.fill = BOTH;
    gbc.anchor = NORTHWEST;
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.weightx = 0;
    serverInputPanel.add(new JLabel(LicenseBundle.message("label.server.address")), gbc);

    gbc.gridx++;
    gbc.weightx = 1;
    gbc.gridwidth = 3;
    gbc.insets = JBUI.insets(5);
    serverInputPanel.add(myServerAddressField, gbc);

    gbc.gridx = 1;
    gbc.gridy++;
    gbc.weightx = 0;
    gbc.gridwidth = 1;
    gbc.fill = NONE;
    serverInputPanel.add(createButtonsPanel(myActivateServerLicenseButton, myCancelServerLicenseButton), gbc);

    gbc.gridx = 1;
    gbc.gridy++;
    gbc.gridwidth = 3;
    gbc.weightx = 1;
    gbc.fill = HORIZONTAL;
    serverInputPanel.add(new SeparatorComponent(SEPARATOR_COLOR, SeparatorOrientation.HORIZONTAL), gbc);

    gbc.gridy++;
    gbc.anchor = WEST;
    gbc.insets.top = 8;
    gbc.weightx = 0;
    gbc.gridwidth = 2;
    serverInputPanel.add(createButtonsPanel(myTestConnButton, myDiscoverButton), gbc);

    gbc.gridx++;
    gbc.weightx = 1;
    gbc.fill = NONE;
    gbc.anchor = EAST;
    serverInputPanel.add(myServerMoreInfoLink, gbc);

    gbc.gridy++;
    gbc.gridx = 1;
    gbc.gridwidth = 3;
    gbc.weighty = 1;
    gbc.anchor = NORTHWEST;
    gbc.fill = HORIZONTAL;
    serverInputPanel.add(createAndRegisterErrorPlaceholder(State.SERVER_INPUT), gbc);

    return serverInputPanel;
  }

  private JPanel createEvaluateInputPanel() {
    JPanel evaluatePanel = new JPanel(new GridBagLayout());
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.insets = JBUI.insets(10, 5, 15, 5);
    gbc.fill = BOTH;
    gbc.anchor = NORTHWEST;
    gbc.gridx = 0;
    gbc.gridy =0;
    gbc.gridwidth = 4;
    gbc.weightx = 1;
    evaluatePanel.add(myEvalLabel, gbc);
    gbc.gridy++;
    if (getFormData() != null) {
      gbc.insets = JBUI.insets(10, 5, 0, 5);
      mySubscriptionEmailField.getEmptyText().clear();
      if (areTipsAvailable()) {
        mySubscriptionEmailField.getEmptyText().appendText(LicenseBundle.message("empty.text.email.address"));
        evaluatePanel.add(myTipsCb, gbc);
        gbc.gridy++;
        evaluatePanel.add(myNewsCb, gbc);
      } else {
        mySubscriptionEmailField.getEmptyText().appendText(LicenseBundle.message("empty.text.email.address.optional"));
        gbc.insets = JBUI.insets(5, 5, 0, 0);
        evaluatePanel.add(new JLabel(LicenseBundle.message("label.tell.me.about.new.product.features")), gbc);
      }
      gbc.gridy++;
      gbc.insets = JBUI.insets(10, 5, 0, 5);
      evaluatePanel.add(mySubscriptionEmailField, gbc);
      gbc.gridy++;
      gbc.gridwidth = 1;
      gbc.weightx = 0;
      gbc.insets = JBUI.insets(15, 5, 0, 0);
      JEditorPane bigView = createViewerWithSmallFont(
        LicenseBundle.message("html.text.i.agree.to.my.email.address.being.used.by.jetbrains"));
      bigView.addHyperlinkListener(new BrowserHyperlinkListener());
      bigView.setVisible(false);
      AtomicBoolean expanded = new AtomicBoolean(false);

      JEditorPane firstLabel = createViewerWithSmallFont(
        LicenseBundle.message("html.text.by.subscribing.to.these.newsletters.you.agree.to.the.privacy.policy"));
      firstLabel.addHyperlinkListener(new HyperlinkAdapter() {
        @Override
        protected void hyperlinkActivated(HyperlinkEvent e) {
          expanded.set(!expanded.get());
          bigView.setVisible(expanded.get());
        }
      });
      firstLabel.setFont(JBUI.Fonts.smallFont());
      evaluatePanel.add(firstLabel, gbc);
      gbc.gridx++;
      gbc.insets = JBUI.insets(15, 0, 0, 0);
      evaluatePanel.add(new Link(AllIcons.General.ArrowDown) {
        @Override
        protected void doAction() {
          expanded.set(!expanded.get());
          bigView.setVisible(expanded.get());

        }
      }, gbc);
      gbc.insets = JBUI.insets(5, 5, 10, 0);
      gbc.gridx = 0;
      gbc.gridy++;
      gbc.weightx = 1;
      gbc.gridwidth = 2;
      evaluatePanel.add(bigView, gbc);
    }

    gbc.gridx = 0;
    gbc.gridy++;
    gbc.gridwidth = 1;
    gbc.weightx = 1;
    gbc.weighty = 0;
    gbc.fill = VERTICAL;
    evaluatePanel.add(createButtonsPanel(myActivateEvaluationPeriodButton, myCancelEvaluationPeriodButton), gbc);

    gbc.weighty = 1;
    gbc.gridy++;
    evaluatePanel.add(Box.createGlue(), gbc);
    return evaluatePanel;
  }

  private static JEditorPane createViewerWithSmallFont(@NlsContexts.DetailedDescription String text) {
    JEditorPane pane = new SwingHelper.HtmlViewerBuilder().create();
    HtmlBuilder builder = new HtmlBuilder();
    builder.append(HtmlChunk.tag("head").addRaw(
      UIUtil.getCssFontDeclaration(JBUI.Fonts.smallFont(), UIUtil.getLabelDisabledForeground(), MY_LINK_COLOR, null))
    );
    builder.append(HtmlChunk.raw(text).wrapWith(HtmlChunk.body()));
    pane.setText(builder.wrapWith(HtmlChunk.html()).toString());
    return pane;
  }

  private void initLicenses() {
    myIgnoreDocumentChanges = true;
    try {
      for (LicenseSource source : myModel.getAvailableLicenseSources(myProductDescriptor)) {
        if (source instanceof BuiltInEvaluationSource) {
          myEvalSource = (BuiltInEvaluationSource)source;
          if (source.getLicense() == null) {
            myEvalSource.restore();
            myEvalLabel.setText(LicenseBundle.message("label.evaluation.is.free.for.0.days", myEvalSource.getEvalPeriod()));
          }
        }
        else if (source instanceof FileLicenseSource){
          myKeySource = (FileLicenseSource)source;
          final LicenseData keyLicense = source.getLicense();
          if (keyLicense != null) {
            myKeyArea.setText(keyLicense.getKey());
            myKeyArea.setCaretPosition(0);
          }
        }
        else if (source instanceof JetprofileSource) {
          myAccountSource = (JetprofileSource)source;
          myAccountNameField.setText(myAccountSource.getLogin());
        }
        else if (source instanceof LicenseServerSource) {
          myServerSource = (LicenseServerSource)source;
          myServerAddressField.setText(myServerSource.getLastUsedServerUrl());
        }
      }
    }
    finally {
      myIgnoreDocumentChanges = false;
    }
  }

  private JButton createButton(@NlsContexts.Button String text) {
    TextWithMnemonic textWithMnemonic = TextWithMnemonic.parse(text);

    JButton button = new JButton(textWithMnemonic.getText());
    button.setMnemonic(textWithMnemonic.getMnemonic());
    if (StartupUiUtil.isUnderDarcula() || UIUtil.isUnderIntelliJLaF()) {
      button.setOpaque(false);
    }
    button.addActionListener(this);
    return button;
  }

  private State guessState() {
    initLicenses();
    LicenseData licenseData = getSelectedLicenseData();
    if (licenseData != null && !licenseData.isExpired()) return State.LICENSED;

    final LicenseSource currentSource = getSelectedLicenseSource();
    if (currentSource == null || currentSource instanceof BuiltInEvaluationSource) return DEFAULT;
    if (currentSource instanceof FileLicenseSource) return State.KEY_INPUT;
    if (currentSource instanceof JetprofileSource) return State.JETPROFILE_INPUT;
    if (currentSource instanceof LicenseServerSource) return State.SERVER_INPUT;
    return DEFAULT;
  }

  private void setBackToLicenseInfoPossible(boolean val) {
    myCancelJBALicenseButton.setEnabled(val);
    myCancelKeyLicenseButton.setEnabled(val);
    myCancelServerLicenseButton.setEnabled(val);
    myCancelEvaluationPeriodButton.setEnabled(val);

    myCancelJBALicenseButton.setVisible(val);
    myCancelKeyLicenseButton.setVisible(val);
    myCancelServerLicenseButton.setVisible(val);
    myCancelEvaluationPeriodButton.setVisible(val);

    KeyStroke keyStroke = KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0);
    if (val) {
      registerKeyboardAction(new BackToLicenseInfoAction(), keyStroke, WHEN_ANCESTOR_OF_FOCUSED_COMPONENT);
    }
    else {
      unregisterKeyboardAction(keyStroke);
    }
  }

  @Override
  public void actionPerformed(ActionEvent e) {
    if (e.getSource() == myReturnToEvaluationButton) {
      int answer = MessageDialogBuilder.yesNo(LicenseBundle.message("dialog.title.confirmation"),
                                              LicenseBundle.message("message.are.you.sure.you.want.to.return.to.evaluation.license")).show();
      if (answer == Messages.NO) {
        return;
      }
      proceedLicense(myEvalSource);
      return;
    }
    if (e.getSource() == myDiscoverButton) {
      doServerDiscover(false);
      return;
    }
    if (e.getSource() == myTestConnButton) {
      final String url = myServerAddressField.getText().trim();
      if (!StringUtil.isEmpty(url)) {
        testServerConnection(url);
      }
      return;
    }

    if (e.getSource() == myActivateJBALicenseButton || e.getSource() == myActivateJBAChooseLicenseButton
        || e.getSource() == myActivateKeyLicenseButton || e.getSource() == myActivateServerLicenseButton
        || e.getSource() == myActivateEvaluationPeriodButton) {
      proceedSelectedLicenseSource();
    }
  }

  private void updateState(@NotNull State state) {
    Component owner = KeyboardFocusManager.getCurrentKeyboardFocusManager().getFocusOwner();
    myState = state;
    if (state != State.EVALUATE) {
      myLastInputState = state;
    }
    myMainCard.show(myMainPanel, state == State.LICENSED ? LICENSED_CARD : NOT_LICENSED_CARD);

    updateLicenseInfoPanel();
    if (state == State.LICENSED) return;

    updateActivationTypePanel();
    updateRadioPanel();

    ((CardLayout)myFieldsPanel.getLayout()).show(myFieldsPanel, myState.name());
    Container placeholder = myErrorPlaceholders.get(state);
    if (placeholder != null && myErrorInfo.getParent() != placeholder) {
      placeholder.removeAll();
      placeholder.add(myErrorInfo);
    }

    SwingUtilities.invokeLater(() -> {
      JRootPane rootPane = getRootPane();
      if (rootPane != null) {
        rootPane.setDefaultButton(defaultButtons.get(state));
      }
    });

    if (state == State.JETPROFILE_CHOOSE) {
      updateProfileChoosePanel();
    }

    if (state == State.JETPROFILE_INPUT) {
      myAccountPasswordField.setText("");
    }

    updateOKButton();
    updateReturnToEvaluationPanel();
    if (owner != null && owner.isShowing()) {
      IdeFocusManager.getGlobalInstance().doWhenFocusSettlesDown(() -> IdeFocusManager.getGlobalInstance().requestFocus(owner, true));
    }

    if (state == State.SERVER_INPUT) {
      doServerDiscover(true);
    }

     JComponent focusField = getPreferredFocusedComponent();

    if (focusField != null) {
      IdeFocusManager.getGlobalInstance().doWhenFocusSettlesDown(() -> IdeFocusManager.getGlobalInstance().requestFocus(focusField, true));
    }
  }

  @Nullable JComponent getPreferredFocusedComponent() {
    switch (myState) {
      case JETPROFILE_INPUT:
        return myAccountNameField;
      case KEY_INPUT:
        return myKeyArea;
      case SERVER_INPUT:
        return myServerAddressField;
      case JETPROFILE_CHOOSE:
        return myActivateJBAChooseLicenseButton;
      default:
        return null;
    }
  }


  private void updateLicenseInfoPanel() {

    myLicenseInfoPanel.removeAll();
    myRemoveLicenseButton.setEnabled(true);

    GridBagConstraints gbc = new GridBagConstraints();
    gbc.anchor = WEST;
    gbc.fill = BOTH;
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.weightx = 1;
    gbc.insets = JBUI.insets(15, GAP, 5, GAP);
    String biggerStyle = String.format("font-size:%dpt; font-family:%s; font-weight:bold; width:350px;",
                                       myLabelFont.getSize() + 1,
                                       myLabelFont.getFamily());
    HtmlChunk.Element statusDiv = HtmlChunk.div(String.format("margin-top:%dpt; width:350px;", (int) (myLabelFont.getSize() * 0.5)));
    JLabel statusLabel = new JLabel();

    final LicenseData currentLicense = getSelectedLicenseData();
    if (currentLicense instanceof CertKeyLicense) {
      CertKeyLicense license = (CertKeyLicense)currentLicense;
      final String restrictions = currentLicense.getRestrictionsText();
      final CertKey2 certKey = license.getCertKey();
      final String assigneeName = String.valueOf(certKey.eval(new GetAssigneeName()));
      final Date expirationDate = currentLicense.getExpirationDate();
      final Date firstExpiredDate = (Date)certKey.eval(new GetFirstExpiredDate(license.getProductCode()));
      String licensedTo = LicenseBundle.message("label.licensed.to.0.1", certKey.eval(new GetLicenseeName())
                                                                         + (!StringUtil.isEmpty(assigneeName) ? " / " + assigneeName : ""));
      HtmlBuilder builder = new HtmlBuilder();
      builder.append(HtmlChunk.text(licensedTo).wrapWith(HtmlChunk.div("width:350px;")));
      if (expirationDate != null && firstExpiredDate != null) {
        String subscriptionState = new Date().before(firstExpiredDate)
                      ? LicenseBundle.message("label.subscription.is.active.until.0", DateFormatUtil.formatAboutDialogDate(expirationDate))
                      : LicenseBundle.message("label.subscription.expired.on.0", DateFormatUtil.formatAboutDialogDate(firstExpiredDate));
        builder.append(HtmlChunk.text(subscriptionState).wrapWith(statusDiv));
      }
      if (Boolean.TRUE.equals(certKey.eval(new IsPerpetualForProduct(license.getProductCode(), license.getProductBuildDate())))) {
        String perpetualText = LicenseBundle.message("label.you.have.a.perpetual.fallback.license.for.this.version");
        builder.append(HtmlChunk.text(perpetualText).wrapWith(statusDiv));
      }
      if (restrictions != null) {
        builder.append(HtmlChunk.text(restrictions).wrapWith(statusDiv));
      }
      statusLabel.setText(builder.wrapWithHtmlBody().toString());
      gbc.gridwidth = 3;
      myLicenseInfoPanel.add(statusLabel, gbc);
    }
    else if (currentLicense != null && !(currentLicense instanceof EvaluationLicenseData)) {
      final String restrictions = currentLicense.getRestrictionsText();
      HtmlBuilder builder = new HtmlBuilder();
      builder.append(HtmlChunk.text(LicenseBundle.message("label.licensed.to.0.1", currentLicense.getUserName())).wrapWith(HtmlChunk.div()));
      if (restrictions != null) {
        builder.append(HtmlChunk.text(restrictions).wrapWith(statusDiv));
      }
      statusLabel.setText(builder.wrapWithHtmlBody().toString());
      gbc.gridwidth = 3;
      myLicenseInfoPanel.add(statusLabel, gbc);
    }
    else if (currentLicense != null) {
      String daysString;
      if (currentLicense.isExpired()) {
        daysString = LicenseBundle.message("label.evaluation.expired");
      }
      else {
        Date expirationDate = currentLicense.getExpirationDate();
        long days = Math.min(
          myEvalSource.getEvalPeriod(),
          BuiltInEvaluationSource.remainsDays(expirationDate, new Date())
        );
        daysString = LicenseBundle.message("label.evaluation.0.days.left", days);
      }
      HtmlBuilder builder = new HtmlBuilder();
      builder.append(LicenseBundle.message("label.free.evaluation"));
      builder.append(" ");
      builder.append(HtmlChunk.text(daysString).wrapWith(HtmlChunk.span().attr("color", ColorUtil.toHtmlColor(myErrorColor))));
      statusLabel.setText(builder.wrapWithHtmlBody().toString());
      gbc.gridwidth = 2;
      myLicenseInfoPanel.add(statusLabel, gbc);
      gbc.weightx = 0;
      gbc.gridx = 2;
      myLicenseInfoPanel.add(createBuyLink(LicenseManager.getInstance().getMessages().getLicenseUrls()), gbc);
      myRemoveLicenseButton.setEnabled(false);
    }

    gbc.gridx = 0;
    gbc.gridy++;
    gbc.fill = NONE;
    gbc.gridwidth = 1;
    gbc.weightx = 0;
    gbc.insets = JBUI.insets(5, GAP, 5, GAP);
    myLicenseInfoPanel.add(myRemoveLicenseButton, gbc);
    gbc.gridx++;
    myLicenseInfoPanel.add(myNewLicenseButton, gbc);
    gbc.gridx++;
    gbc.weightx = 1;
    myLicenseInfoPanel.add(Box.createHorizontalGlue(), gbc);

    gbc.gridy++;
    gbc.weighty = 1;
    gbc.fill = VERTICAL;
    myLicenseInfoPanel.add(Box.createVerticalGlue(), gbc);

    myLicenseInfoPanel.revalidate();
    myLicenseInfoPanel.repaint();
  }

  private void updateActivationTypePanel() {
    myActivationTypePanel.setVisible(myState != State.JETPROFILE_CHOOSE);
    mySeparator.setVisible(myState != State.JETPROFILE_CHOOSE);
    myActivateButton.setSelected(myState != State.EVALUATE);
    myEvaluateButton.setSelected(myState == State.EVALUATE);

    if (myProductDescriptor.isEvalSupported()) {
      boolean expired = (myEvalSource != null
                         && myEvalSource.getLicense() != null
                         && myEvalSource.getLicense().isExpired());
      myEvaluateButton.setEnabled(!expired);
    }
    else {
      myEvaluateButton.setEnabled(false);
      myEvaluateButton.setVisible(false);
    }
  }

  private void updateProfileChoosePanel() {
    myProfileChoosePanel.removeAll();
    List<JBALicenseData> availableLicenses = myAccountSource.getAvailableLicenses();
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = BOTH;
    gbc.anchor = NORTHWEST;
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.weightx = 1;

    gbc.insets = JBUI.insets(15, 5, 5, 5);
    myProfileChoosePanel.add(new JLabel(LicenseBundle.message("label.you.have.several.licenses.select.license")), gbc);
    gbc.insets = JBUI.insets(5);

    Box licensesPanel = Box.createVerticalBox();
    ButtonGroup group = new ButtonGroup();
    myChosenAccountLicenseData = myAccountSource.getLicense();
    for (final JBALicenseData licenseData : availableLicenses) {
      JRadioButton button = new JRadioButton();
      if (myChosenAccountLicenseData == null) {
        myChosenAccountLicenseData = licenseData; // select the first one, if nothing is chosen
      }
      button.setSelected(myChosenAccountLicenseData == licenseData);
      button.setBorder(null);
      button.addItemListener(new ItemListener() {
        @Override
        public void itemStateChanged(ItemEvent e) {
          if (e.getStateChange() == ItemEvent.SELECTED) {
            myChosenAccountLicenseData = licenseData;
          }
        }
      });
      group.add(button);
      JPanel panel = createPanelForRadioButton(button, licenseData);
      if (licensesPanel.getComponentCount() > 0) licensesPanel.add(Box.createVerticalStrut(10));
      licensesPanel.add(panel);
    }

    JPanel verticalWrapper = new JPanel(new BorderLayout());
    verticalWrapper.add(licensesPanel, BorderLayout.NORTH);
    JPanel horizontalWrapper = new JPanel(new BorderLayout());
    horizontalWrapper.add(verticalWrapper, BorderLayout.WEST);

    JScrollPane licensesScrollBar = new JBScrollPane(horizontalWrapper);
    licensesScrollBar.setBorder(JBUI.Borders.empty());

    gbc.gridy++;
    gbc.weighty = 1;
    gbc.fill = BOTH;
    myProfileChoosePanel.add(licensesScrollBar, gbc);

    gbc.gridy++;
    gbc.fill = NONE;
    gbc.weighty = 0;
    myProfileChoosePanel.add(myActivateJBAChooseLicenseButton, gbc);
  }

  @NotNull
  private static JPanel createPanelForRadioButton(JRadioButton button, JBALicenseData licenseData) {
    HtmlBuilder builder = new HtmlBuilder();
    builder.append(LicenseBundle.message("label.licensed.to.0.1", licenseData.getUserName()));

    final String restrictions = licenseData.getRestrictionsText();
    if (!StringUtil.isEmptyOrSpaces(restrictions)) {
      builder.append(", ");
      builder.append(restrictions);
    }
    builder.append(HtmlChunk.br());

    final Date expiration = licenseData.getExpirationDate();
    if (expiration == null) {
      builder.append(LicenseBundle.message("label.description.perpetual.license"));
    }
    else {
      builder.append(LicenseBundle.message("label.description.active.until.0", DateFormatUtil.formatDate(expiration)));
    }

    if (licenseData instanceof CertKeyLicense) {
      @NonNls String style = "color:" + ColorUtil.toHtmlColor(JBColor.GRAY);
      builder.append(" ");
      builder.append(HtmlChunk.text(LicenseBundle.message("label.license.id.0", ((CertKeyLicense)licenseData).getCertKey()
        .eval(new GetLicenseId()))).wrapWith(HtmlChunk.span(style)));
    }

    JLabel label = new JLabel(builder.wrapWithHtmlBody().toString());
    label.setLabelFor(button);
    label.addMouseListener(new MouseAdapter() {
      @Override
      public void mouseClicked(MouseEvent e) {
        button.requestFocus();
        button.setSelected(true);
      }
    });

    JPanel panel = new JPanel();
    panel.add(button);
    panel.add(label);
    return panel;
  }

  private void updateRadioPanel() {
    myRadioPanel.removeAll();
    if (myState == State.JETPROFILE_CHOOSE) {
      return;
    }
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = BOTH;
    gbc.anchor = WEST;
    gbc.gridx = 0;
    gbc.gridy = 0;
    gbc.insets = JBUI.insets(5, 0, 5, 15);
    String getLicenseText = LicenseBundle.message("label.get.license.from");
    myRadioPanel.add(new JLabel(getLicenseText), gbc);
    gbc.gridx++;
    gbc.insets = JBUI.insets(5, 10, 5, 5);

    ButtonGroup group = new ButtonGroup();
    if (myAccountSource != null) {
      group.add(myJetProfileButton);
      gbc.gridx++;
      myRadioPanel.add(myJetProfileButton, gbc);
    }
    if (myKeySource != null) {
      group.add(myKeyButton);
      gbc.gridx++;
      myRadioPanel.add(myKeyButton, gbc);
    }
    if (myServerSource != null) {
      group.add(myServerButton);
      gbc.gridx++;
      myRadioPanel.add(myServerButton, gbc);
    }
    gbc.insets = JBUI.insets(5);
    gbc.gridx++;
    gbc.weightx = 1;
    myRadioPanel.add(Box.createHorizontalGlue(), gbc);

    myJetProfileButton.setSelected(myState == State.JETPROFILE_INPUT);
    myKeyButton.setSelected(myState == State.KEY_INPUT);
    myServerButton.setSelected(myState == State.SERVER_INPUT);
    myRadioPanel.setVisible(myState != State.JETPROFILE_CHOOSE && myState != State.EVALUATE);
  }

  private void updateOKButton() {
    disableActivateButtons();

    if (myState == State.JETPROFILE_INPUT) {
      myActivateJBALicenseButton.setEnabled(!StringUtil.isEmptyOrSpaces(myAccountNameField.getText()) && myAccountPasswordField.getDocument().getLength() > 0);
      return;
    }
    if (myState == State.JETPROFILE_CHOOSE) {
      myActivateJBAChooseLicenseButton.setEnabled(true);
      return;
    }

    if (myState == State.KEY_INPUT) {
      updateKeyFields(myKeyArea.getText(), false);
      return;
    }

    if (myState == State.SERVER_INPUT) {
      updateServerStatusText(null);
      try {
        final String urlStr = myServerAddressField.getText().trim();
        if (!StringUtil.isEmpty(urlStr)) {
          URL url = new URL(urlStr);
          if (!StringUtil.isEmpty(url.getProtocol()) && !StringUtil.isEmpty(url.getHost())) {
            myTestConnButton.setEnabled(true);
            myActivateServerLicenseButton.setEnabled(true);
            return;
          }
        }
      }
      catch (MalformedURLException ignored) {
      }
      myTestConnButton.setEnabled(false);
      return;
    }
    if (myState == State.EVALUATE) {
      boolean validEmail = isValidEmail(mySubscriptionEmailField.getText());
      boolean emptyEmail = StringUtil.isEmpty(mySubscriptionEmailField.getText());
      boolean enabled = validEmail || emptyEmail;
      if (areTipsAvailable()) {
        enabled = (myNewsCb.isSelected() || myTipsCb.isSelected()) && validEmail;
        enabled |= (!myNewsCb.isSelected() && !myTipsCb.isSelected());
      }

      myActivateEvaluationPeriodButton.setEnabled(enabled);
    }
  }

  private void updateKeyFields(final String key, boolean smartDiscover) {
    myKeyArea.setForeground(UIManager.getColor("TextArea.foreground"));
    myMyKeyValidationLabel.setText("");
    LicenseData license = null;
    String message = null;

    try {
      license = tryCreateKeyLicense(key);
    }
    catch (ValidationFailedException e) {
      message = e.getMessage();
    }
    if (license == null) {
      if (!StringUtil.isEmpty(key)) {
        String error = message != null ? message : LicenseBundle.message("message.text.key.invalid");
        myErrorInfo.showError(error, ErrorInfoBox.ERROR_TYPE, "https://intellij-support.jetbrains.com/hc/en-us/articles/207241025");
      }
      else {
        myErrorInfo.clear(false);
      }
      return;
    }
    myErrorInfo.clear(true);

    if (myState == State.KEY_INPUT) {
      myActivateKeyLicenseButton.setEnabled(true);
    }
    //noinspection SSBasedInspection
    SwingUtilities.invokeLater(() -> {
      myIgnoreDocumentChanges = true;
      try {
        if (!StringUtil.equals(key, myKeyArea.getText())) {
          myKeyArea.setText(key);
          myKeyArea.repaint();
        }
      }
      finally {
        myIgnoreDocumentChanges = false;
      }
    });
    myKeyArea.setForeground(myOkColor);
    final Date expirationDate = license.getExpirationDate();
    if (expirationDate != null && !license.isExpired()) {
      myMyKeyValidationLabel.setText(LicenseBundle.message("license.panel.expirable.license.description", expirationDate));
    }
    else if (license instanceof MaintenanceDueAware) {
      final Date maintenanceDueDate = ((MaintenanceDueAware)license).getMaintenanceDueDate();
      myMyKeyValidationLabel.setText(
        LicenseBundle.message("license.panel.maintenance.aware.license.description", maintenanceDueDate)
      );
    }
    else {
      final int version = license.getMajorVersion();
      final String versionStr;
      if (version > 2000000) {
        versionStr = version / 1000 + "." + (version % 1000) / 100;
      }
      else {
        versionStr = String.valueOf(version);
      }
      myMyKeyValidationLabel.setText(
        LicenseBundle.message("license.panel.perpetual.license.description", myProductDescriptor.getShortName(), versionStr)
      );
    }
  }

  private static ApplicationInfo getAppInfo() {
    ApplicationInfo impl = null;
    if (ApplicationManager.getApplication() != null) {
      impl = ApplicationInfo.getInstance();
    }
    if (impl == null) {
      impl = ApplicationInfoImpl.getShadowInstance();
    }
    return impl;
  }

  private LicenseData tryCreateKeyLicense(String key) throws ValidationFailedException {
    final LicenseFactory factory = myKeySource.getFactory();
    if (StringUtil.isEmptyOrSpaces(key)) {
      return null;
    }
    final LicenseData lic = factory.createFromUser(key.trim(), null);
    if (lic != null) {
      factory.getLicenseValidator().validate(lic);
      return lic;
    }
    throw new ValidationFailedException();
  }

  private void updateReturnToEvaluationPanel() {
    myReturnToEvaluationPanel.removeAll();
    boolean hideAndReturn = !allowReturnToEvaluation;
    LicenseData licenseData = myEvalSource == null ? null : myEvalSource.getLicense();
    if (getSelectedLicenseSource() instanceof BuiltInEvaluationSource) {
      hideAndReturn = true;
    }
    else {
      if (licenseData == null || licenseData.isExpired()) {
        hideAndReturn = true;
      }
    }
    myReturnToEvaluationPanel.setVisible(!hideAndReturn);
    if (hideAndReturn) return;
    GridBagConstraints gbc = new GridBagConstraints();
    gbc.fill = HORIZONTAL;
    gbc.gridx = 0;
    gbc.gridy = 0;
    myReturnToEvaluationPanel.add(myReturnToEvaluationButton, gbc);
    int daysLeft = BuiltInEvaluationSource.remainsDays(licenseData.getExpirationDate(), new Date());
    if (daysLeft != -1) {
      gbc.gridx++;
      gbc.insets.left = 5;
      String daysLeftMessage = daysLeft > 1 ? LicenseBundle.message("license.panel.buildit.evaluation.expires.in.many", daysLeft) :
                               LicenseBundle.message("license.panel.buildit.evaluation.expires.in.one");
      myReturnToEvaluationPanel.add(new JLabel(daysLeftMessage), gbc);
    }
    gbc.gridx++;
    gbc.weightx = 1;
    myReturnToEvaluationPanel.add(Box.createHorizontalGlue(), gbc);
  }

  /**
   *
   * null means that the panel would be dynamically switch to myLastInputState
   */
  private JRadioButton createRadioButton(@NlsContexts.RadioButton String text, @Nullable final State state) {
    TextWithMnemonic textWithMnemonic = TextWithMnemonic.parse(text);

    JRadioButton button = new JRadioButton(textWithMnemonic.getText());
    button.setMnemonic(textWithMnemonic.getMnemonic());
    button.setBorder(null);
    button.addItemListener(e -> {
      State stateToSwitch = state != null ? state : myLastInputState;
      if (e.getStateChange() == ItemEvent.SELECTED && myState != stateToSwitch) {
        myErrorInfo.clear(false);
        updateState(stateToSwitch);
      }
    });

    return button;
  }

  @Override
  public void doSmartDiscover() {
    doServerDiscover(true);
    doClipboardDiscover();
  }

  private void doServerDiscover(boolean isSmartDiscover) {
    if (!myServerButton.isSelected() || (isSmartDiscover && !StringUtil.isEmpty(myServerAddressField.getText())) ) {
      return;
    }
    doAsyncTask(myServerDiscoverLock, () -> {
      try {
        return DnsUtil.detectServerUrl(AppExecutorUtil.getAppExecutorService());
      }
      catch (Throwable e) {
        // Ignore. An unpredictable error happened detecting server, like http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6520665.
        // No luck then.
      }
      return null;
    }, s -> {
      if (s != null) {
        myServerAddressField.setText(s);
      }
      else {
        updateServerStatusText(null);
      }
    }, isSmartDiscover ? SMART_DISCOVER_TIMEOUT : UI_ASYNC_TASK_TIMEOUT, !isSmartDiscover
    );
  }

  private void updateServerStatusText(@Nullable @NlsContexts.DetailedDescription String status) {
    if (status != null) {
      myErrorInfo.showError(status, ErrorInfoBox.INFO_TYPE);
    }
    else if (getSelectedLicenseData() instanceof ServerLicenseData) {
      myErrorInfo.showError(LicenseBundle.message("message.text.floating.ticket.obtained"), ErrorInfoBox.INFO_TYPE);
    }
  }

  @Nullable
  private LicenseSource getSelectedLicenseSource() {
    final LicenseInfo info = myModel.getLicenseInfo(myProductDescriptor);
    return info != null ? info.getLicenseSource() : null;
  }

  @Nullable
  private LicenseData getSelectedLicenseData() {
    final LicenseInfo info = myModel.getLicenseInfo(myProductDescriptor);
    return info != null ? info.getLicense() : null;
  }

  private void testServerConnection(final String url) {
    if (!myServerButton.isSelected() || StringUtil.isEmpty(url) ) {
      return;
    }
    doAsyncTask(myServerDiscoverLock, () -> {
                  try {
                    final LicenseServerSource serverSource = myServerSource;
                    if (serverSource != null) {
                      return serverSource.ping(url);
                    }
                  }
                  catch (Throwable e) {
                    // Ignore. An unpredictable error happened detecting server, like http://bugs.sun.com/bugdatabase/view_bug.do?bug_id=6520665.
                    // No luck then.
                  }
                  return Boolean.FALSE;
                },
                connectionOk -> {
                  String status;
                  if (connectionOk == null) {
                    status = LicenseBundle.message("message.text.timed.out");
                  }
                  else if (connectionOk) {
                    status = LicenseBundle.message("message.text.connection.successful");
                  }
                  else {
                    status = HtmlChunk.text(LicenseBundle.message("message.text.no.response"))
                      .wrapWith(HtmlChunk.tag("font").attr("color", ColorUtil.toHtmlColor(JBColor.RED)))
                      .wrapWith(HtmlChunk.body()).wrapWith(HtmlChunk.html()).toString();
                  }
                  updateServerStatusText(status);
                },
                UI_ASYNC_TASK_TIMEOUT, true
    );
  }

  private <T> void doAsyncTask(final AtomicBoolean inProgressFlag, final Computable<? extends T> computable, final Consumer<? super T> edtRunnable,
                               final long timeout, final boolean useDecorator) {
    if (myDisposed || !inProgressFlag.compareAndSet(false, true)) {
      return;
    }
    Map<JComponent, Boolean> state = null;
    if (useDecorator) {
      myDecorator.setLoadingText(LicenseBundle.message("loading.text.please.wait"));
      myDecorator.startLoading(false);
      state = getControlsState();
      disableControls();
      disableActivateButtons();
    }
    Map<JComponent, Boolean> controlsState = state;
    final Future<T> task = AppExecutorUtil.getAppExecutorService().submit(() -> computable.compute());

    new Thread(() -> {
      T t = null;
      try {
        t = task.get(timeout, TimeUnit.MILLISECONDS);
      }
      catch (Throwable ignored) {
      }

      final T asyncResult = t;
      //noinspection SSBasedInspection
      SwingUtilities.invokeLater(() -> {
        try {
          if (useDecorator) {
            myDecorator.stopLoading();
            restoreControlsState(controlsState);
            updateOKButton();
          }
          if (!myDisposed) {
            edtRunnable.accept(asyncResult);
          }
        }
        finally {
          inProgressFlag.set(false);
        }
      });
    }, "server discover").start();
  }

  private Map<JComponent, Boolean> getControlsState() {
    return Stream.of(myBuyLink, myActivateButton, myEvaluateButton, myRadioPanel, myFieldsPanel, myReturnToEvaluationButton)
      .collect(Collectors.toMap(component -> component, component -> component.isEnabled()));
  }

  private void restoreControlsState(Map<JComponent, Boolean> state) {
    state.forEach((component, enabled) -> UIUtil.setEnabled(component, enabled, true));
  }

  private void disableControls() {
    UIUtil.setEnabled(myBuyLink, false, true);
    UIUtil.setEnabled(myActivateButton, false, true);
    UIUtil.setEnabled(myEvaluateButton, false, true);
    UIUtil.setEnabled(myRadioPanel, false, true);
    UIUtil.setEnabled(myFieldsPanel, false, true);
    UIUtil.setEnabled(myReturnToEvaluationButton, false, true);
  }

  private void disableActivateButtons() {
    myActivateJBALicenseButton.setEnabled(false);
    myActivateJBAChooseLicenseButton.setEnabled(false);
    myActivateKeyLicenseButton.setEnabled(false);
    myActivateServerLicenseButton.setEnabled(false);
    myActivateEvaluationPeriodButton.setEnabled(false);
  }

  private void doClipboardDiscover() {
    if (!StringUtil.isEmpty(myKeyArea.getText()) || myState != State.KEY_INPUT) return;
    try {
      Object data = Toolkit.getDefaultToolkit().getSystemClipboard().getData(DataFlavor.stringFlavor);
      if (data instanceof String) {
        String s = ((String)data);
        if (s.length() > 10000) return;
        s = s.trim();
        Matcher licenseStartMatcher = LICENSE_START_PATTERN1.matcher(s);
        Matcher licenseEndMatcher = LICENSE_END_PATTERN1.matcher(s);
        boolean keyFound =
          licenseStartMatcher.find() && licenseEndMatcher.find() && (licenseStartMatcher.end() < licenseEndMatcher.start());
        if (keyFound) {
          String licenseKey = s.substring(licenseStartMatcher.end(), licenseEndMatcher.start());
          licenseKey = licenseKey.replace(' ', '\n').replace('\r', '\n').replace("\n\n", "\n");
          while (licenseKey.endsWith("\n")) {
            licenseKey = licenseKey.substring(0, licenseKey.length() - 1);
          }
          updateKeyFields(licenseKey, true);
          return;
        }
        Matcher licenseMatcher = LICENSE_PATTERN2.matcher(s);
        //Let's use the first key only
        if (licenseMatcher.find()) {
          String licenseKey = licenseMatcher.group(1);
          updateKeyFields(licenseKey, true);
          return;
        }
        try {
          LicenseData license = tryCreateKeyLicense(s);
          if (license != null) {
            updateKeyFields(s, true);
          }
        }
        catch (ValidationFailedException vfe) {
          //ignored
        }
      }
    }
    catch (Exception e) {
      //ignore
    }
  }

  private Container createAndRegisterErrorPlaceholder(State state) {
    Box res = Box.createHorizontalBox();
    res.setBorder(JBUI.Borders.empty());
    myErrorPlaceholders.put(state, res);
    return res;
  }

  private static Box createButtonsPanel(JButton... buttons) {
    Box res = Box.createHorizontalBox();
    for (JButton button : buttons) {
      res.add(button);
    }
    res.add(Box.createHorizontalGlue());
    return res;
  }

  private static class Link extends JLabel implements Border {
    private String myUrl;
    private Link(Icon icon) {
      setIcon(icon);
      init();
    }

    private Link(@NlsContexts.LinkLabel String text, final String url) {
      setText(text);
      setIcon(AllIcons.Ide.External_link_arrow);
      setHorizontalTextPosition(LEFT);
      setIconTextGap(0);
      myUrl = url;
      init();
    }

    private void init() {
      super.setForeground(MY_LINK_COLOR);
      setCursor(Cursor.getPredefinedCursor(Cursor.HAND_CURSOR));
      addMouseListener(new MouseAdapter() {
        @Override
        public void mousePressed(MouseEvent e) {
          if (SwingUtilities.isLeftMouseButton(e)) {
            requestFocus();
          }
        }

        @Override
        public void mouseClicked(MouseEvent e) {
          if (SwingUtilities.isLeftMouseButton(e)) {
            doAction();
          }
        }
      });
      addKeyListener(new KeyAdapter() {
        @Override
        public void keyPressed(KeyEvent e) {
          if (e.getKeyCode() == KeyEvent.VK_ENTER) doAction();
        }
      });
      addFocusListener(new FocusListener() {
        @Override
        public void focusGained(FocusEvent e) {
          repaint();
        }

        @Override
        public void focusLost(FocusEvent e) {
          repaint();
        }
      });
      setBorder(this);
    }

    @Override
    public void setForeground(Color fg) {
      //ignore
    }

    @Override
    public String toString() {
      return "Link:" + myUrl;
    }

    public void setUrl(String url) {
      myUrl = url;
    }

    protected void doAction() {
      BrowserUtil.browse(myUrl);
    }

    @Override
    public boolean isFocusable() {
      return !StringUtil.isEmpty(getText());
    }

    @Override
    public void paintBorder(Component c, Graphics g, int x, int y, int width, int height) {
      if (!c.isFocusOwner() || StringUtil.isEmpty(getText())) return;
      Graphics2D g2d = (Graphics2D)g.create();
      try {
        g2d.setStroke(new BasicStroke(.5F, BasicStroke.CAP_BUTT, BasicStroke.JOIN_MITER, 1, new float[]{3, 1}, 0));
        g2d.setColor(ColorUtil.withAlpha(MY_LINK_COLOR, .5D));
        g2d.drawRect(x, y, width - 1, height - 1);
      }
      finally {
        g2d.dispose();
      }
    }

    @Override
    public Insets getBorderInsets(Component c) {
      return JBInsets.create(0, 2);
    }

    @Override
    public boolean isBorderOpaque() {
      return false;
    }
  }

  private static class HintTextArea extends JTextArea {
    private final char[] myHint;

    HintTextArea(int rows, int columns, String hint) {
      super(rows, columns);
      myHint = hint.toCharArray();
    }
    @Override
    protected void paintComponent(Graphics g) {
      super.paintComponent(g);
      if (getText().isEmpty()) {
        final Color oldColor = g.getColor();
        final Font oldFont = g.getFont();
        try {
          g.setColor(JBColor.GRAY);
          ((Graphics2D)g).setRenderingHint(RenderingHints.KEY_ANTIALIASING, RenderingHints.VALUE_ANTIALIAS_ON);
          final FontMetrics metrics = g.getFontMetrics();
          int x = Math.abs(getWidth() - metrics.charsWidth(myHint, 0, myHint.length)) / 2;
          int y = Math.abs(getHeight() - metrics.getHeight()) / 2 + metrics.getAscent();
          g.drawChars(myHint, 0, myHint.length, x, y);
        }
        finally {
          g.setColor(oldColor);
          g.setFont(oldFont);
        }
      }
    }
  }

  private class MyTransferHandler extends TransferHandler {
    private static final int maxLength = 15000;
    @Override
    public boolean canImport(TransferSupport support) {
      return true;
    }

    @Override
    public void exportToClipboard(JComponent comp, Clipboard clip, int action) throws IllegalStateException {
      if (action == COPY) {
        clip.setContents(new StringSelection(myKeyArea.getSelectedText()), null);
      }
    }

    @Override
    public boolean importData(TransferSupport support) {
      Transferable transferable = support.getTransferable();
      try {
        List<File> list = FileCopyPasteUtil.getFileList(support.getTransferable());
        if (list != null && list.size() == 1) {
          File toRead = list.get(0);
          FileInputStream stream = null;
          try {
            stream = new FileInputStream(toRead);
            byte[] bytes = new byte[16384];
            ByteArrayOutputStream baos = new ByteArrayOutputStream();
            for (int l = stream.read(bytes); l > 0; l = stream.read(bytes)) {
              baos.write(bytes, 0, l);
              if (baos.size() > maxLength) return false;
            }
            myKeyArea.setText(new String(baos.toByteArray(), StandardCharsets.UTF_8));
            return true;
          } catch (IOException e) {
            //ignored
          } finally {
            if (stream != null) {
              try {
                stream.close();
              }
              catch (IOException e) {
                //ignored
              }
            }
          }
        }
        Object data = transferable.getTransferData(DataFlavor.selectBestTextFlavor(transferable.getTransferDataFlavors()));
        if (data instanceof String && ((String)data).length() <= maxLength) {
          myKeyArea.setText((String)data);
          return true;
        }
        data = transferable.getTransferData(DataFlavor.stringFlavor);
        if (data instanceof String && ((String)data).length() <= maxLength) {
          myKeyArea.setText((String)data);
          return true;
        }
      }
      catch (Exception e) {
        //ignore
      }
      return false;
    }
  }

  private static class ErrorInfoBox extends JEditorPane {

    static final int ERROR_TYPE = 0;
    static final int INFO_TYPE = 1;

    private static final Color ERROR_COLOR = JBColor.namedColor("Label.errorForeground", new JBColor(new Color(0xC7222D), DarculaColors.RED));
    private static final Color INFO_COLOR = JBColor.foreground();

    private boolean myInitialError;
    private boolean myClearOnNavigate;

    ErrorInfoBox() {
      setEditable(false);
      setContentType("text/html");
      setOpaque(false);
      addHyperlinkListener(new BrowserHyperlinkListener() {
        @Override
        protected void hyperlinkActivated(@NotNull HyperlinkEvent e) {
          super.hyperlinkActivated(e);
          if (myClearOnNavigate) clear(false);
        }
      });
      setBorder(JBUI.Borders.empty(1));
    }

    public void showInitialError(@NlsContexts.DetailedDescription String errorText, @MagicConstant(intValues = {ERROR_TYPE, INFO_TYPE}) int type) {
      showError(errorText, type, null, true);
    }

    void showError(@NlsContexts.DetailedDescription String errorText, @MagicConstant(intValues = {ERROR_TYPE, INFO_TYPE}) int type) {
      showError(errorText, type, null);
    }

    void showError(@NlsContexts.DetailedDescription String errorText, @MagicConstant(intValues = {ERROR_TYPE, INFO_TYPE}) int type, @Nullable String helpURL) {
      showError(errorText, type, helpURL, false);
    }

    void showError(@NlsContexts.DetailedDescription String errorText, @MagicConstant(intValues = {ERROR_TYPE, INFO_TYPE}) int type, @NlsSafe @Nullable String helpURL, boolean initialError) {
      HtmlBuilder builder = new HtmlBuilder();
      builder.append(HtmlChunk.raw(errorText));
      if (helpURL != null) {
        builder.append(" ");
        builder.append(HtmlChunk.link(helpURL, LicenseBundle.message("link.text.help")).style("text-decoration:none;"));
      }

      setText(builder.wrapWith(HtmlChunk.body().style(getStyle(type))).wrapWith(HtmlChunk.html()).toString());
      myInitialError = initialError;
    }

    void clear(boolean rewriteInitial) {
      if (myInitialError && !rewriteInitial) return;
      setText(null);
    }

    void setClearOnLinkPress(boolean clear) {
      myClearOnNavigate = clear;
    }

    private static String getStyle(@MagicConstant(intValues = {ERROR_TYPE, INFO_TYPE}) int type) {
      Font font = UIUtil.getLabelFont();
      String fontFamily = font.getFamily();
      int fontSize = font.getSize();

      @NonNls StringBuilder builder = new StringBuilder();
      builder.append("font-family:").append(fontFamily).append(";");
      builder.append("font-size:").append(fontSize).append("pt;");
      builder.append("width:230px;");

      Color color = type == ERROR_TYPE ? ERROR_COLOR : INFO_COLOR;
      builder.append("color:").append(ColorUtil.toHtmlColor(color)).append(";");
      return builder.toString();
    }
  }

  interface LicenseUpdateListener {                // todo: do we need this?
    void licenseUpdated(ProductDescriptor pd);
  }

  private class NewLicenseAction extends AbstractAction {

    NewLicenseAction() {
      TextWithMnemonic textWithMnemonic = TextWithMnemonic.parse(LicenseBundle.message("button.new.license"));
      putValue(NAME, textWithMnemonic.getText());
      putValue(MNEMONIC_KEY, textWithMnemonic.getMnemonic());
    }

    @Override
    public void actionPerformed(ActionEvent e) {
      updateState(getCurrentLicenseState());
      setBackToLicenseInfoPossible(true);
    }

    private State getCurrentLicenseState() {
      LicenseSource<?> source = getSelectedLicenseSource();
      if (source instanceof JetprofileSource) return State.JETPROFILE_INPUT;
      if (source instanceof LicenseKeySource) return State.KEY_INPUT;
      if (source instanceof ServerSource) return State.SERVER_INPUT;

      return LicensePanel.DEFAULT;
    }
  }

  private class RemoveLicenseAction extends AbstractAction {

    RemoveLicenseAction() {
      TextWithMnemonic textWithMnemonic = TextWithMnemonic.parse(LicenseBundle.message("button.remove.license"));
      putValue(NAME, textWithMnemonic.getText());
      putValue(MNEMONIC_KEY, textWithMnemonic.getMnemonic());
    }

    @Override
    public void actionPerformed(ActionEvent e) {
      String productName = myModel.getDisplayName(myProductDescriptor);
      String message = LicenseBundle.message("dialog.text.html.are.you.sure.you.want.to.remove.the.license.for.0", productName, productName);
      int answer = Messages.showYesNoDialog(message, LicenseBundle.message("dialog.title.remove.license"),
                                            LicenseBundle.message("dialog.title.remove.license"), LicenseBundle.message("button.cancel"), Messages.getQuestionIcon());

      if (answer == Messages.YES) {
        myModel.forgetLicense(myProductDescriptor);
        updateState(LicensePanel.DEFAULT);
        notifyLicenseListenersIfNeeded();
      }
    }
  }

  private final class BackToLicenseInfoAction extends AbstractAction {

    private BackToLicenseInfoAction() {
      super(LicenseBundle.message("button.cancel"));
    }

    @Override
    public void actionPerformed(ActionEvent e) {
      updateState(State.LICENSED);
      setBackToLicenseInfoPossible(false);
    }
  }
}