package com.intellij.ide.license.ui;

import com.intellij.ide.license.LicenseBundle;
import com.intellij.idea.SplashManager;
import com.intellij.openapi.MnemonicHelper;
import com.intellij.openapi.application.ApplicationNamesInfo;
import com.intellij.openapi.ui.DialogWrapper;
import com.intellij.openapi.ui.MessageDialogBuilder;
import com.intellij.openapi.ui.Messages;
import com.intellij.openapi.util.Disposer;
import com.intellij.openapi.util.SystemInfoRt;
import com.intellij.ui.AppUIUtil;
import com.intellij.ui.JBColor;
import com.intellij.ui.components.panels.HorizontalBox;
import com.intellij.util.ui.JBDimension;
import com.intellij.util.ui.JBUI;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import javax.swing.*;
import javax.swing.border.Border;
import java.awt.*;
import java.awt.event.ActionListener;
import java.awt.event.KeyEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;

import static java.awt.event.WindowEvent.WINDOW_CLOSING;

public final class LicenseUtils {
  private static final int MIN_DIALOG_HEIGHT = 480;
  private static final int MIN_DIALOG_WIDTH = 800;

  private LicenseUtils() {}

  public static void showLicenseDialog(final LicenseEditingModel model) {
    LicenseEntryUI ui = null;
    for (Frame frame : Frame.getFrames()) {
      if (frame.isVisible()) { // frame already exists
        ui = createDialog(model);
        break;
      }
    }
    if (ui == null) {
      ui = createFrame(model);
    }

    ui.showUI();
  }

  private static LicenseEntryUI createFrame(LicenseEditingModel model) {
    LicenseFrame frame = new LicenseFrame(model);

    frame.pack();
    Dimension size = frame.getPreferredSize();
    size.width = Math.max(JBUI.scale(MIN_DIALOG_WIDTH), size.width);
    size.height = Math.max(JBUI.scale(MIN_DIALOG_HEIGHT), size.height);
    frame.setSize(size);
    frame.setMinimumSize(new JBDimension(MIN_DIALOG_WIDTH, MIN_DIALOG_HEIGHT));

    return frame;
  }

  @NotNull
  private static LicenseEntryUI createDialog(LicenseEditingModel model) {
    LicenseDialog dialog = new LicenseDialog(model);
    dialog.setModal(true);
    dialog.setTitle(LicenseBundle.message("dialog.title.licenses"));

    dialog.pack();
    Dimension size = dialog.getPreferredSize();
    size.width = Math.max(JBUI.scale(MIN_DIALOG_WIDTH), size.width);
    size.height = Math.max(JBUI.scale(MIN_DIALOG_HEIGHT), size.height);
    dialog.setSize(size.width, size.height);
    dialog.getWindow().setMinimumSize(new JBDimension(MIN_DIALOG_WIDTH, MIN_DIALOG_HEIGHT));

    return dialog;
  }

  private static final class LicenseDialog extends DialogWrapper implements LicenseEntryUI {

    private final FlatMultiLicensePanel myContentPanel;
    private final LicenseEditingModel myModel;
    private final JLabel myWarningLabel;
    private boolean myCanBeClosed = true;

    private LicenseDialog(final LicenseEditingModel model) {
      super(null, true, DialogWrapper.IdeModalityType.IDE);

      myWarningLabel = new JLabel();
      myWarningLabel.setForeground(JBColor.GRAY);

      myModel = model;
      myContentPanel = new FlatMultiLicensePanel(model, this);
      myContentPanel.setBorder(JBUI.Borders.customLine(JBColor.border(), 0, 0, 1, 0));
      setOKButtonText(LicenseBundle.message("button.close"));
      setCancelButtonText(LicenseBundle.message("button.exit"));
      init();

      Disposer.register(getDisposable(), myContentPanel);
      JButton closeButton = getButton(getCancelAction());
      if (closeButton != null && myCanBeClosed) {
        closeButton.setVisible(false);
      }
    }

    @Override
    public @Nullable JComponent getPreferredFocusedComponent() {
      JComponent focusedComponent = myContentPanel.getPreferredFocusedComponent();
      return focusedComponent != null ? focusedComponent : super.getPreferredFocusedComponent();
    }

    @Override
    protected JComponent createCenterPanel() {
      return myContentPanel;
    }

    @Nullable
    @Override
    protected Border createContentPaneBorder() {
      return null;
    }

    @Override
    protected JComponent createSouthPanel() {
      JComponent southPanel = super.createSouthPanel();
      southPanel.setBorder(createDefaultBorder());
      return southPanel;
    }

    @Override
    protected void createDefaultActions() {
      super.createDefaultActions();
      //init();
      setAutoAdjustable(false);
    }

    @Nullable
    @Override
    protected ActionListener createCancelAction() {
      return e -> hideDialogOrShutdown();
    }

    @Override
    public void doCancelAction(AWTEvent source) {
      if (source instanceof WindowEvent
          && source.getID() == WINDOW_CLOSING) {
        hideDialogOrShutdown();
        return;
      }

      super.doCancelAction(source);
    }

    private void hideDialogOrShutdown() {
      if (!myCanBeClosed) {
        if (confirmExit(getContentPane())) doCancelAction();
        else return;
      }

      doOKAction();
    }

    @Override
    public void doCancelAction() {
      try {
        super.doCancelAction();
      }
      finally {
        myModel.close(true);
      }
    }

    @Override
    protected void doOKAction() {
      try {
        super.doOKAction();
      }
      finally {
        myModel.close(false);
      }
    }

    @Override
    public void showUI() {
      SplashManager.hideBeforeShow(getWindow());
      show();
    }

    @Override
    public void setCanBeClosed(boolean canBeClosed) {
      myCanBeClosed = canBeClosed;
      getOKAction().setEnabled(canBeClosed);
      if (!canBeClosed) {
        showExit();
      }
    }

    private void showExit() {
      JButton closeButton = getButton(getCancelAction());
      if (closeButton != null && !closeButton.isVisible()) {
        closeButton.setVisible(true);
        setOKButtonText(LicenseBundle.message("button.continue"));
      }
    }

    @Override
    public void setWarning(@Nullable String warning) {
      myWarningLabel.setText(warning == null ? "" : warning);
    }

    @Override
    public JButton getOkButton() {
      return getButton(getOKAction());
    }

    @Nullable
    @Override
    protected JPanel createSouthAdditionalPanel() {
      HorizontalBox res = new HorizontalBox();
      res.add(myWarningLabel);
      return res;
    }
  }

  private static class LicenseFrame extends JFrame implements LicenseEntryUI {

    private final LicenseEditingModel myModel;
    private boolean myIsInitialized;

    @NotNull private final JButton myContinueButton;
    @NotNull private final JButton myExitButton;
    @NotNull private final JPanel myPnlButtonsHolder;
    @NotNull private final JLabel myWarningLabel;
    @NotNull private final FlatMultiLicensePanel myContentPanel;

    LicenseFrame(LicenseEditingModel model) {
      super(LicenseBundle.message("dialog.title.license.activation"));
      myModel = model;

      AppUIUtil.updateWindowIcon(this);

      myWarningLabel = new JLabel();
      myWarningLabel.setForeground(JBColor.GRAY);

      myPnlButtonsHolder = new JPanel();
      myContinueButton = new JButton(LicenseBundle.message("button.continue"));
      myExitButton = new JButton(LicenseBundle.message("button.exit"));
      myExitButton.setVisible(false);
      resetButtons(false);
      myContentPanel = new FlatMultiLicensePanel(myModel, this);

      setDefaultCloseOperation(WindowConstants.DO_NOTHING_ON_CLOSE);
      getRootPane().registerKeyboardAction(
        e -> hideDialogOrShutdown(), KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT
      );
      addWindowListener(new WindowAdapter() {
        @Override
        public void windowClosing(WindowEvent e) {
          hideDialogOrShutdown();
        }
      });

      Container rootContent = getContentPane();
      rootContent.setLayout(new BorderLayout());
      rootContent.add(myContentPanel, BorderLayout.CENTER);

      myContinueButton.addActionListener(e -> doCloseAction(false));
      myExitButton.addActionListener(e -> doCloseAction(true));

      JPanel pnlButtons = new JPanel(new GridBagLayout());

      GridBagConstraints gbc = new GridBagConstraints();
      gbc.gridx = 0;
      gbc.gridy = 0;
      gbc.weightx = 1;
      gbc.anchor = GridBagConstraints.EAST;
      gbc.fill = GridBagConstraints.NONE;
      gbc.insets = JBUI.insets(0);
      pnlButtons.add(myWarningLabel, gbc);

      gbc.gridx++;
      gbc.fill = GridBagConstraints.NONE;
      gbc.weightx = 0;
      gbc.insets = JBUI.insets(5, 27, 5, 8);
      pnlButtons.add(myPnlButtonsHolder, gbc);
      pnlButtons.setBorder(JBUI.Borders.customLine(JBColor.border(), 1, 0, 0, 0));

      rootContent.add(pnlButtons, BorderLayout.SOUTH);

      MnemonicHelper.init(rootContent);
      myIsInitialized = true;
    }

    private void hideDialogOrShutdown() {
      if (!isCanBeClosed()) {
        if (confirmExit(getContentPane())) doCloseAction(true);
        else return;
      }

      doCloseAction(false);
    }

    private void resetButtons(boolean showExit) {
      myPnlButtonsHolder.removeAll();
      int cols = showExit ? 2 : 1;
      myPnlButtonsHolder.setLayout(new GridLayout(1, cols, SystemInfoRt.isMac ? 0 : 5, 0));
      if (showExit) {
        if (SystemInfoRt.isMac) {
          myPnlButtonsHolder.add(myExitButton);
          myPnlButtonsHolder.add(myContinueButton);
        }
        else {
          myPnlButtonsHolder.add(myContinueButton);
          myPnlButtonsHolder.add(myExitButton);
        }
      }
      else {
        myPnlButtonsHolder.add(myContinueButton);
      }
    }

    private void doCloseAction(boolean isCancel) {
      try {
        Disposer.dispose(myContentPanel);
        dispose();
      }
      finally {
        myModel.close(isCancel);
        //todo stop all async processes #idea-217767
        //LicenseMessagesImpl.ourReopenLock.set(false);
      }
    }

    @Override
    public void setCanBeClosed(boolean canBeClosed) {
      myContinueButton.setEnabled(canBeClosed);
      if (myIsInitialized && canBeClosed && isOnlyProduct()) {
        doCloseAction(false);
        return;
      }

      if (!canBeClosed && !myExitButton.isVisible()) {
        myExitButton.setVisible(true);
        resetButtons(true);
      }
    }

    private boolean isOnlyProduct() {
      return myModel.getPlatformProduct() != null
             ? myModel.getPluginProducts().isEmpty()
             : myModel.getPluginProducts().size() == 1;
    }

    private boolean isCanBeClosed() {
      return myContinueButton.isEnabled();
    }

    @Override
    public void setWarning(@Nullable String warning) {
      if (warning == null) {
        myWarningLabel.setText("");
      }
      else {
        myWarningLabel.setText(warning);
      }
    }

    @Override
    public void showUI() {
      SplashManager.hideBeforeShow(this);
      pack();
      setLocationRelativeTo(null);
      setMinimumSize(getPreferredSize());
      setVisible(true);
    }

    @Override
    public JButton getOkButton() {
      return myContinueButton;
    }
  }

  private static boolean confirmExit(Component parent) {
    String productName = ApplicationNamesInfo.getInstance().getFullProductName();
    String message = LicenseBundle.message("dialog.text.you.need.to.obtain.the.license.are.you.sure.you.want.to.exit.0", productName);
    return MessageDialogBuilder.yesNo(LicenseBundle.message("dialog.title.confirmation"), message).show() == Messages.YES;
  }

}
