package com.intellij.ide.misc;

import com.intellij.openapi.util.io.FileUtilRt;
import com.intellij.util.Time;
import com.jetbrains.license.util.Base64Support;
import com.jetbrains.license.util.Pair;
import com.jetbrains.license.util.Signatures;
import com.jetbrains.license.util.VerifyingSignatureFactory;
import org.jetbrains.annotations.NotNull;

import java.io.*;
import java.nio.charset.StandardCharsets;
import java.nio.file.Path;

/**
 * @author Eugene Zhuravlev
 */
public abstract class UpdatableResource<T> {
  private static final long CACHE_DROP_PERIOD = Time.HOUR;
  private static final int DEFAULT_VERSION = 0;
  private long myLastAccessed;

  private final File myDataFile;
  private Pair<Integer, T> myCache;

  public UpdatableResource(@NotNull Path dataFile) {
    myDataFile = dataFile.toFile();
  }

  public synchronized final int getVersion(){
    checkCacheValidity();
    Pair<Integer, T> data = myCache;
    if (data == null) {
      myCache = data = readContent(myDataFile);
    }
    return data != null? data.fst : DEFAULT_VERSION;
  }

  public synchronized final T getData() {
    checkCacheValidity();
    Pair<Integer, T> data = myCache;
    if (data == null) {
      myCache = data = readContent(myDataFile);
    }
    return data != null? data.snd : null;
  }

  private interface Executable<E extends Throwable> {
    void execute() throws E;
  }

  public synchronized void updateData(String content) throws IOException {
    final Executable<IOException> update = ()-> {
      try (DataOutputStream out = new DataOutputStream(new BufferedOutputStream(new FileOutputStream(myDataFile)))) {
        out.writeBytes(content);
      }
    };
    try {
      update.execute();
    }
    catch (FileNotFoundException e) {
      // can throw FileNotFoundException if the file is marked readonly or is a directory
      if (FileUtilRt.delete(myDataFile)) { // fallback
        update.execute();
      }
      else {
        throw e;
      }
    }
    myCache = null;
  }

  protected abstract T parseData(String data) throws IOException;

  private Pair<Integer, T> readContent(File from){
    try {
      final int size = (int)from.length();
      if (size > 0) {
        byte[] bytes = new byte[size];
        try (DataInputStream in = new DataInputStream(new BufferedInputStream(new FileInputStream(from)))) {
          in.readFully(bytes);
        }
        bytes = Base64Support.decode(bytes);

        try (DataInputStream in = new DataInputStream(new ByteArrayInputStream(bytes))) {
          final int version = in.readInt();
          final String data = new String(bytes, 4, bytes.length - 4, StandardCharsets.UTF_8);
          // there must be a signature part followed by '\n' separator
          final Pair<String, String> signedContent = Signatures.verifyContent(data, VerifyingSignatureFactory.X509_CERTIFICATE_GENERIC_DATA_SIGNING);
          return new Pair<>(version, parseData(signedContent.snd));
        }
      }
    }
    catch (Throwable ignored) {
    }
    finally {
      myLastAccessed = System.currentTimeMillis();
    }
    return new Pair<>(DEFAULT_VERSION, null);
  }

  private void checkCacheValidity() {
    final long now = System.currentTimeMillis();
    if (Math.abs(now - myLastAccessed) > CACHE_DROP_PERIOD) {
      myCache = null;
    }
  }
}
