package com.intellij.indexing.composite;

import com.google.common.annotations.VisibleForTesting;
import com.intellij.indexing.shared.platform.impl.CompositeInvertedIndex;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.util.Computable;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.util.Processor;
import com.intellij.util.SmartList;
import com.intellij.util.SystemProperties;
import com.intellij.util.indexing.*;
import com.intellij.util.indexing.impl.AbstractUpdateData;
import com.intellij.util.indexing.impl.InputDataDiffBuilder;
import com.intellij.util.indexing.impl.MergedValueContainer;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.TestOnly;

import java.io.IOException;
import java.util.List;
import java.util.Map;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.function.Consumer;

public abstract class CompositeInvertedIndexBase<Key, Value>
  implements FileBasedIndexInfrastructureExtensionUpdatableIndex<Key, Value, FileContent> {

  @NotNull
  protected final UpdatableIndex<Key, Value, FileContent> myBaseIndex;
  protected final ID<Key, Value> myId;

  public CompositeInvertedIndexBase(@NotNull ID<Key, Value> id,
                                    @NotNull UpdatableIndex<Key, Value, FileContent> baseIndex) throws IOException {
    myId = id;
    myBaseIndex = baseIndex;
  }

  @Override
  public final @NotNull IndexInfrastructureExtensionUpdateComputation mapInputAndPrepareUpdate(int inputId, @Nullable FileContent content) {
    Computable<Boolean> secondaryIndexUpdate;
    Computable<Boolean> baseIndexUpdate;

    boolean hasSharedIndex;

    if (content != null && !IndexingDataKeys.REBUILD_REQUESTED.isIn(content.getFile())) {
      IndexInfrastructureUpdate update = tryToMapWithSecondaryIndex(inputId, content);
      secondaryIndexUpdate = update.update;
      hasSharedIndex = update.isIndexProvided;

      if (hasSharedIndex) {
        baseIndexUpdate = updateBaseIndex(inputId, null);
      }
      else {
        baseIndexUpdate = updateBaseIndex(inputId, content);
      }
    }
    else {
      secondaryIndexUpdate = removeFromSecondaryIndex(inputId);
      baseIndexUpdate = updateBaseIndex(inputId, content);
      hasSharedIndex = false;
    }

    return new IndexInfrastructureExtensionUpdateComputation(
      () -> Boolean.TRUE.equals(secondaryIndexUpdate.get()) && Boolean.TRUE.equals(baseIndexUpdate.get()),
      hasSharedIndex
    );
  }

  @Override
  public final void updateWithMap(@NotNull AbstractUpdateData<Key, Value> updateData) throws StorageException {
    // TODO ensure it's called only for stub index extension
    myBaseIndex.updateWithMap(updateData);
  }

  @Override
  public final void setBufferingEnabled(boolean enabled) {
    myBaseIndex.setBufferingEnabled(enabled);
  }

  @Override
  public final void cleanupMemoryStorage() {
    myBaseIndex.cleanupMemoryStorage();
  }

  @TestOnly
  @Override
  public final void cleanupForNextTest() {
    myBaseIndex.cleanupForNextTest();
  }

  @NotNull
  @Override
  public final ValueContainer<Value> getData(@NotNull Key key) throws StorageException {
    List<ValueContainer<Value>> data = new SmartList<>();
    data.add(myBaseIndex.getData(key));
    getSecondaryIndexData(key, data::add);
    return new MergedValueContainer<>(data);
  }

  @Override
  public final boolean processAllKeys(@NotNull Processor<? super Key> processor,
                                      @NotNull GlobalSearchScope scope,
                                      @Nullable IdFilter idFilter)
    throws StorageException {
    if (!myBaseIndex.processAllKeys(processor, scope, idFilter)) {
      return false;
    }
    return processAllKeysBySecondaryIndex(processor, scope, idFilter);
  }

  @NotNull
  @Override
  public final ReadWriteLock getLock() {
    return myBaseIndex.getLock();
  }

  @NotNull
  @Override
  public final Map<Key, Value> getIndexedFileData(int fileId) throws StorageException {
    // During indexing, base index diff builder is allowed to access previously indexed data.
    // And we should provide data only from base index
    VirtualFile indexingFile = FileBasedIndex.getInstance().getFileBeingCurrentlyIndexed();
    if (indexingFile != null) {
      int indexingFileId = FileBasedIndex.getFileId(indexingFile);
      Map<Key, Value> data = myBaseIndex.getIndexedFileData(fileId);
      if (indexingFileId != fileId) {
        String message = "Indexing process should not rely on non-indexed file data.\n" +
                         "Indexing file = " + indexingFile.getPath() + ", file id = " + indexingFileId + ", queried file id = " + fileId;
        if (ApplicationManager.getApplication().isUnitTestMode()) {
          FileBasedIndexImpl.LOG.warn(message);
        }
        else {
          FileBasedIndexImpl.LOG.error(message);
        }
      }
      return data;
    }

    Map<Key, Value> secondaryFileData = getSecondaryIndexFileData(fileId);
    if (secondaryFileData != null) return secondaryFileData;
    return myBaseIndex.getIndexedFileData(fileId);
  }

  @Override
  public final void setIndexedStateForFile(int fileId, @NotNull IndexedFile file, boolean isProvidedByInfrastructureExtension) {
    if (isProvidedByInfrastructureExtension) {
      //TODO write to stamps only once
      myBaseIndex.setUnindexedStateForFile(fileId);
      IndexingStamp.setFileIndexedStateCurrent(fileId, myId);
    }
    else {
      myBaseIndex.setIndexedStateForFile(fileId, file);
    }
  }

  @Override
  public final void setIndexedStateForFile(int fileId, @NotNull IndexedFile file) {
    myBaseIndex.setIndexedStateForFile(fileId, file);
  }

  @Override
  public final void invalidateIndexedStateForFile(int fileId) {
    myBaseIndex.invalidateIndexedStateForFile(fileId);
  }

  @Override
  public final void setUnindexedStateForFile(int fileId) {
    myBaseIndex.setUnindexedStateForFile(fileId);
  }

  @Override
  public final @NotNull FileIndexingState getIndexingStateForFile(int fileId, @NotNull IndexedFile file) {
    return myBaseIndex.getIndexingStateForFile(fileId, file);
  }

  @Override
  public final long getModificationStamp() {
    return myBaseIndex.getModificationStamp();
  }

  @Override
  public final void removeTransientDataForFile(int inputId) {
    myBaseIndex.removeTransientDataForFile(inputId);
  }

  @Override
  public final void removeTransientDataForKeys(int inputId, @NotNull InputDataDiffBuilder<Key, Value> diffBuilder) {
    myBaseIndex.removeTransientDataForKeys(inputId, diffBuilder);
  }

  @NotNull
  @Override
  public final IndexExtension<Key, Value, FileContent> getExtension() {
    return myBaseIndex.getExtension();
  }


  @Override
  public final void clear() throws StorageException {
    try {
      clearSecondaryIndex();
    }
    finally {
      myBaseIndex.clear();
    }
  }

  @Override
  public final void dispose() {
    try {
      disposeSecondaryIndex();
    }
    finally {
      myBaseIndex.dispose();
    }
  }

  @Override
  public final void flush() throws StorageException {
    try {
      myBaseIndex.flush();
    }
    finally {
      flushSecondaryIndex();
    }
  }


  @NotNull
  protected final Computable<Boolean> updateBaseIndex(int inputId, @Nullable FileContent content) {
    return mustUseOnlyIndexInfrastructureExtension() ? () -> Boolean.TRUE : myBaseIndex.mapInputAndPrepareUpdate(inputId, content);
  }


  /**
   * We must not cache the property value in static final field because {@link CompositeInvertedIndex}
   * class may be initialized too early (before the property is programmatically set in some startup activity).
   */
  private static boolean mustUseOnlyIndexInfrastructureExtension() {
    return SystemProperties.getBooleanProperty("idea.use.only.index.infrastructure.extension", false);
  }

  @VisibleForTesting
  public final @NotNull UpdatableIndex<Key, Value, FileContent> getBaseIndex() {
    return myBaseIndex;
  }


  protected abstract void getSecondaryIndexData(@NotNull Key key, Consumer<ValueContainer<Value>> dataConsumer) throws StorageException;

  protected abstract void disposeSecondaryIndex();

  protected abstract void flushSecondaryIndex() throws StorageException;

  protected abstract void clearSecondaryIndex() throws StorageException;


  protected abstract @NotNull IndexInfrastructureUpdate tryToMapWithSecondaryIndex(int inputId,
                                                                                   @NotNull FileContent content);


  protected abstract @NotNull Computable<Boolean> removeFromSecondaryIndex(int inputId);


  @Nullable
  protected abstract Map<Key, Value> getSecondaryIndexFileData(int fileId) throws StorageException;

  protected abstract boolean processAllKeysBySecondaryIndex(@NotNull Processor<? super Key> processor,
                                                            @NotNull GlobalSearchScope scope,
                                                            @Nullable IdFilter idFilter)
    throws StorageException;

  protected static class IndexInfrastructureUpdate {
    private final boolean isIndexProvided;
    private final Computable<Boolean> update;

    public IndexInfrastructureUpdate(boolean isIndexProvided, Computable<Boolean> update) {
      this.isIndexProvided = isIndexProvided;
      this.update = update;
    }
  }
}
