package com.intellij.indexing.shared.download

import com.intellij.openapi.Disposable
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.components.Service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.withPushPop
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import java.util.concurrent.atomic.AtomicReference

@Service
class SharedIndexAuthLifetimeHolder : Disposable {
  private val currentLifetime = AtomicReference<Disposable?>(null)

  override fun dispose() = Unit

  companion object {
    fun getNextDisposable(project: Project?): Disposable {

      val host = project ?: ApplicationManager.getApplication()
      val service = host.getService(SharedIndexAuthLifetimeHolder::class.java)
      val newDispose = Disposer.newDisposable("Shared Indexes login lifetime")

      listOfNotNull(service, project).forEach { parent ->
        Disposer.register(parent, newDispose)
      }

      val prev = service.currentLifetime.getAndSet(newDispose)
      if (prev != null) {
        Disposer.dispose(prev)
      }

      return newDispose
    }
  }
}

class SharedIndexAuthException(message: String,
                               cause: Throwable? = null
) : Exception(message, cause)

@Service
class SharedIndexAuthHolder {
  private val LOG = logger<SharedIndexAuthHolder>()

  fun resolveHeaders(fullRequestUrl: String,
                     request: SharedIndexLookupRequest,
                     indicator: ProgressIndicator
  ): Map<String, String> {
    val authId = request.authId
    if (authId == null) {
      //no headers if no auth was required
      return mapOf()
    }

    val it = resolveIndexDownloadAuthExtension(authId)

    if (it == null) {
      //request must fail to avoid it from being cached
      throw SharedIndexAuthException("Failed to find the shared index auth extension '$authId'")
    }

    val headers = indicator.withPushPop {
      try {
        it.tryGenerateToken(fullRequestUrl, indicator)
      }
      catch (t: Throwable) {
        LOG.warn("Failed to get shared indexes request headers " +
                 "from the auth extension '${authId}' for ${fullRequestUrl} " +
                 "and $request. ${t.message}", t)

        null
      }
    }

    if (headers == null) {
      throw SharedIndexAuthException("Failed to get shared indexes request headers " +
                                     "from the auth extension '${authId}' for ${fullRequestUrl} " +
                                     "and $request")
    }
    return headers
  }


  fun resolveHeaders(fullRequestUrl: String,
                     response: SharedIndexLookupResult,
                     indicator: ProgressIndicator): Map<String, String> = resolveHeaders(fullRequestUrl, response.request, indicator)
}
