package com.intellij.indexing.shared.download

import com.intellij.indexing.shared.util.compressSharedIndex
import com.intellij.openapi.diagnostic.thisLogger
import com.intellij.openapi.progress.ProgressIndicator
import org.tukaani.xz.LZMA2Options
import org.tukaani.xz.XZInputStream
import org.tukaani.xz.XZOutputStream
import java.io.InputStream
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.StandardCopyOption
import java.util.zip.GZIPInputStream
import java.util.zip.GZIPOutputStream

@Suppress("unused")
enum class SharedIndexCompression(
  val type: String,
  val extensionSuffix: String
) {

  XZ(type = "xz", extensionSuffix = ".xz") {
    override fun createUnpackingStream(packedStream: InputStream) = XZInputStream(packedStream)
    override fun compress(file: Path, output: Path, indicator: ProgressIndicator) {
      thisLogger().warn("Packing shared index with XZ")
      compressSharedIndex(file, output, indicator, compressorStream = { XZOutputStream(it, LZMA2Options()) })
    }
  },

  GZIP(type = "gzip", extensionSuffix = ".gz") {
    override fun createUnpackingStream(packedStream: InputStream) = GZIPInputStream(packedStream)

    override fun compress(file: Path, output: Path, indicator: ProgressIndicator) {
      thisLogger().warn("Packing shared index with GZip")
      compressSharedIndex(file, output, indicator, compressorStream = { GZIPOutputStream(it, 1024 * 1024) })
    }
  },

  PLAIN(type = "plain", extensionSuffix = "") {
    override fun createUnpackingStream(packedStream: InputStream) = packedStream
    override fun compress(file: Path, output: Path, indicator: ProgressIndicator) {
      Files.move(file, output, StandardCopyOption.REPLACE_EXISTING)
    }
  },

  ;

  abstract fun createUnpackingStream(packedStream: InputStream) : InputStream

  /**
   * Takes the [file] and create the packed version of it in the [output].
   * The implementation is allowed to (re)move the initial file
   */
  abstract fun compress(file: Path, output: Path, indicator: ProgressIndicator)

  companion object {
    @JvmStatic
    fun findByType(type: String?): SharedIndexCompression? = values().firstOrNull { it.type == type }
  }
}
