package com.intellij.indexing.shared.download

import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.openapi.components.service
import com.intellij.openapi.extensions.BaseExtensionPointName
import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.openapi.options.Configurable
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.ComboBox
import com.intellij.openapi.ui.DialogPanel
import com.intellij.ui.layout.*
import javax.swing.DefaultComboBoxModel
import javax.swing.JComponent

private val EP_NAME = ExtensionPointName<SharedIndexSuggesterConfigurable>("com.intellij.sharedIndexSuggesterConfigurable")

enum class SharedIndexConfigurableMode : Comparable<SharedIndexConfigurableMode> {
  Additional,
  PreBuilt,
  CustomBuilt,
}

interface SharedIndexSuggesterConfigurable {
  val mode : SharedIndexConfigurableMode

  /**
   * Predicate to test if a given extension is available for the project.
   *
   * Note. This extension can be executed for default project
   */
  fun isAvailableFor(project: Project) : Boolean

  /**
   * Builds project-specific settings panel.
   *
   * Note. This extension can be executed for default project
   */
  fun buildSettingsPanel(project: Project, scope: SharedIndexesConfigurableScope, builder: RowBuilder)
}

interface SharedIndexesConfigurableScope {
  fun Cell.consentComboBox(key: IndexDownloadConsentDecisionKey, includeUrl: Boolean): CellBuilder<ComboBox<*>>
}

private class SharedIndexesConfigurableScopeBase(val key: String) : SharedIndexesConfigurableScope {
  override fun Cell.consentComboBox(key : IndexDownloadConsentDecisionKey, includeUrl: Boolean): CellBuilder<ComboBox<*>> {
    val model = DefaultComboBoxModel(arrayOf(
      IndexDownloadConsentDecision.ALLOWED,
      IndexDownloadConsentDecision.UNKNOWN,
      IndexDownloadConsentDecision.DENIED))

    val comment = SharedIndexesBundle.message("configuration.combobox.comment.with.url", key.url)
    val cb = comboBox<IndexDownloadConsentDecision>(
      model,
      getter = { service<SharedIndexDownloadConsentStore>().decisionFor(key) },
      setter = { it?.let { service<SharedIndexDownloadConsentStore>().setDecisionFor(key, it) } },
      renderer = listCellRenderer { value, _, _ ->
        text = when (value) {
          IndexDownloadConsentDecision.ALLOWED -> SharedIndexesBundle.message("configuration.combobox.option.allow")
          IndexDownloadConsentDecision.UNKNOWN -> SharedIndexesBundle.message("configuration.combobox.option.ask")
          IndexDownloadConsentDecision.DENIED -> SharedIndexesBundle.message("configuration.combobox.option.deny")
        }
      }).sizeGroup("consent-checkbox-$key")

    if (includeUrl) {
      return cb.comment(comment)
    }

    return cb
  }
}

internal class SharedIndexConfigurable(val project: Project) : Configurable, Configurable.VariableProjectAppLevel, Configurable.WithEpDependencies {
  override fun getDisplayName() = SharedIndexesBundle.message("configuration.displayName")

  private var panel: DialogPanel? = null

  override fun isProjectLevel(): Boolean = false

  override fun getDependencies(): MutableCollection<BaseExtensionPointName<*>> = mutableListOf(EP_NAME)

  override fun createComponent(): JComponent {
    val newSuggestions = EP_NAME.extensionList
      .filter { it.isAvailableFor(project) }
      .groupBy { it.mode }

    val panel = panel {
      newSuggestions[SharedIndexConfigurableMode.Additional]?.let {
        for (suggestion in it) {
          suggestion.buildSettingsPanel(project, SharedIndexesConfigurableScopeBase("additional"), this)
        }
      }

      newSuggestions[SharedIndexConfigurableMode.PreBuilt]?.let {
        titledRow(SharedIndexesBundle.message("configuration.title.public.shared.indexes")) {
          row {
            label(SharedIndexesBundle.message("configuration.text.public.about.comment"))
          }

          for (suggestion in it) {
            suggestion.buildSettingsPanel(project, SharedIndexesConfigurableScopeBase("public"), this)
          }
        }
      }

      newSuggestions[SharedIndexConfigurableMode.CustomBuilt]?.let {
        for (suggestion in it) {
          suggestion.buildSettingsPanel(project, SharedIndexesConfigurableScopeBase("custom"), this)
        }
      }
    }

    this.panel = panel
    return panel
  }

  override fun reset() {
    panel?.reset()
  }

  override fun isModified(): Boolean = panel?.isModified() ?: false
  override fun apply() {
    val wasModified = isModified
    panel?.apply()

    if (wasModified) {
      service<SharedIndexDownloadService>().rescanNow(project)
    }
  }

  override fun disposeUIResources() {
    panel = null
  }

  // Make the "?" button on the Shared Indexes Settings page open the official help page:
  // example for 2020.2: https://www.jetbrains.com/help/idea/2020.2/shared-indexes.html
  override fun getHelpTopic(): String = "shared-indexes-settings"
}
