package com.intellij.indexing.shared.download

import com.intellij.openapi.components.BaseState
import com.intellij.openapi.components.SimplePersistentStateComponent
import com.intellij.openapi.components.State
import com.intellij.openapi.components.Storage
import com.intellij.openapi.util.registry.Registry
import com.intellij.util.xmlb.annotations.Attribute
import com.intellij.util.xmlb.annotations.Tag
import com.intellij.util.xmlb.annotations.XCollection
import java.util.concurrent.locks.ReentrantLock
import kotlin.concurrent.withLock

private fun shouldAllowAllDownloads() = Registry.`is`("shared.indexes.download.auto.consent")

/**
 * Identifier for download consent
 */
data class IndexDownloadConsentDecisionKey(
  val kind: String,
  val url: String
)

fun SharedIndexSuggestion.toKey() = IndexDownloadConsentDecisionKey(sharedIndexId.kind, sharedIndexId.url)

@Tag("decision")
enum class IndexDownloadConsentDecision {
  UNKNOWN,
  ALLOWED,
  DENIED
}

@Tag("item")
class IndexDownloadConsentItem : BaseState() {
  @get:Attribute
  var kind by string()

  @get:Attribute
  var url by string() // optional

  @get:Attribute
  var download by enum(IndexDownloadConsentDecision.UNKNOWN)

  fun toKey(): IndexDownloadConsentDecisionKey? {
    return IndexDownloadConsentDecisionKey(kind ?: return null, url ?: return null)
  }
}

class IndexDownloadConsentState : BaseState() {
  @get:XCollection
  var consentItems by list<IndexDownloadConsentItem>()

  @get:Attribute
  var pauseIndexing: String? by this.string()
}

@State(name = "download-consent", storages = [Storage("shared-indexes.xml")], allowLoadInTests = true)
class SharedIndexDownloadConsentStore : SimplePersistentStateComponent<IndexDownloadConsentState>(IndexDownloadConsentState()) {
  private val lock = ReentrantLock()

  override fun loadState(state: IndexDownloadConsentState) = lock.withLock {
    super.loadState(state)
  }

  fun reset() = loadState(IndexDownloadConsentState())

  var pauseIndexing: Boolean
    get() = lock.withLock { state.pauseIndexing?.toBoolean() ?: Registry.`is`("shared.indexes.wait.default")  }
    set(value) { lock.withLock { state.pauseIndexing = value.toString(); state.intIncrementModificationCount() } }

  fun filterAllowed(suggestions: List<SharedIndexSuggestion>): List<SharedIndexSuggestion> {
    if (shouldAllowAllDownloads()) return suggestions

    val allowedKeys = lock.withLock {
      state.consentItems.filter { it.download == IndexDownloadConsentDecision.ALLOWED }.mapNotNull { it.toKey() }.toSet()
    }

    return suggestions.filter { it.toKey() in allowedKeys }
  }

  fun loadAllDecisions(kind : String): List<IndexDownloadConsentDecisionKey> = lock.withLock {
    state.consentItems.mapNotNull { it.toKey() }.filter { it.kind == kind }
  }

  fun decisionFor(key: IndexDownloadConsentDecisionKey): IndexDownloadConsentDecision {
    if (shouldAllowAllDownloads()) return IndexDownloadConsentDecision.ALLOWED

    return lock.withLock {
      state.consentItems.find { it.toKey() == key }?.download ?: IndexDownloadConsentDecision.UNKNOWN
    }
  }

  fun isAllowedFor(key: IndexDownloadConsentDecisionKey) = decisionFor(key) == IndexDownloadConsentDecision.ALLOWED



  fun setDecisionFor(key: IndexDownloadConsentDecisionKey, allow: IndexDownloadConsentDecision) = setDecisionFor(listOf(key), allow)

  fun setDecisionFor(keys: Collection<IndexDownloadConsentDecisionKey>, decision: IndexDownloadConsentDecision) = lock.withLock {
    val keysSet = keys.toSet()
    state.consentItems.removeAll {
      val key = it.toKey()
      key == null || key in keysSet
    }

    for (key in keysSet) {
      val new = IndexDownloadConsentItem()
      new.kind = key.kind
      new.download = decision
      new.url = key.url.trim()
      state.consentItems.add(new)
    }
    state.intIncrementModificationCount()
  }
}
