package com.intellij.indexing.shared.download

import com.intellij.openapi.Disposable
import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.openapi.progress.ProgressIndicator
import org.jetbrains.annotations.Nls


val sharedIndexesDownloadExtensionEP =
  ExtensionPointName.create<SharedIndexDownloadAuthExtension>("com.intellij.sharedIndexDownloadExtension")

/**
 * This extension point allows to pass authentication headers to a
 * shared indexes download requests, for example:
 * ```
 *   sharedIndex:
 *     project:
 *       - url: <the URL with authentication>
 *         auth: the 'type' of your SharedIndexDownloadAuthExtension
 * ```
 */
interface SharedIndexDownloadAuthExtension {
  /** Download extension identifier, can be used in the settings **/
  val type: String

  /** Name of the authenticator to be shown in the UI **/
  @get:Nls
  val presentableName: String

  /**
   * This method is executed from a background thread. The implementation must first
   * synchronously check for the current authentication state. Asynchronous authentication
   * should be started if needed.
   *
   * The [indicator] can be only used in the method, not in the asynchronous part of the login,
   * an implementation may start own [com.intellij.openapi.progress.Task.Backgroundable]
   * to show any needed progress for the asynchronous authentication phase.
   *
   * The method should not use the [lifetime] or [onAuthCompleted] if it was able to
   * ensure the authentication is valid.
   *
   * User interaction should be performed after the method is completed. It should call the
   * [onAuthCompleted] callback once the auth is finished successfully. It should ignore the callback
   * and [lifetime] if login process stuck or failed.
   *
   * The [lifetime] is used to allow the caller to unsubscribe the [onAuthCompleted] notification when needed
   *
   * The [presentableName] is a human readable text with the reason, why the authentication is needed
   */
  fun checkAuthenticated(
    presentableReason: String,
    indicator: ProgressIndicator,
    lifetime: Disposable,
    onAuthCompleted: () -> Unit
  )

  /**
   * This method should return necessary HTTP headers (as Map) that will be sent for HTTP requests behind
   * shared indexes download for the link form the feed [feedUrl].
   *
   * For the cases, when it is impossible to get the authentication token without a user interaction,
   * the method should return `null` to indicate it's failure.
   *
   * The implementation may use [indicator] to check for cancellation and to report additional progress
   */
  fun tryGenerateToken(feedUrl: String, indicator: ProgressIndicator): Map<String, String>?
}


fun resolveIndexDownloadAuthExtension(type: String?): SharedIndexDownloadAuthExtension? = type?.let {
  sharedIndexesDownloadExtensionEP.findFirstSafe { it.type == type }
}
