package com.intellij.indexing.shared.download

import com.intellij.ide.nls.NlsMessages
import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.notification.*
import com.intellij.notification.NotificationAction.createSimpleExpiring
import com.intellij.openapi.Disposable
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.application.invokeLater
import com.intellij.openapi.components.service
import com.intellij.openapi.options.ShowSettingsUtil
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import java.util.concurrent.atomic.AtomicReference

class IndexDownloadConsentServiceDebug : DumbAwareAction() {
  override fun actionPerformed(e: AnActionEvent) {
    val project = e.project ?: return

    val suggestions = project.service<SharedIndexSuggestionService>().resolveSuggestions()
    project.service<IndexDownloadConsentService>().showNotifications(suggestions)
  }
}

class IndexDownloadConsentService(val project: Project) : Disposable {
  private val myPendingNotification = AtomicReference<Pair<Set<IndexDownloadConsentDecisionKey>, Notification>?>(null)

  init {
    project.messageBus
      .connect(this)
      .subscribe(ShadeIndexDumbModeTrackerTopic, object : ShadeIndexDumbModeTrackerTopicListener {
        override fun onStableDumbModeDetected() {
          myPendingNotification.get()?.second?.let {
            invokeLater {
              it.hideBalloon()
            }
          }
        }
      })
  }

  override fun dispose() {
    myPendingNotification.getAndSet(null)?.second?.expire()
  }

  fun askForMoreDownloads(suggestions: List<SharedIndexSuggestion>) {
    val consentStore = service<SharedIndexDownloadConsentStore>()
    val keysToAsk = suggestions
      .associateBy { it.toKey() }
      .filter { consentStore.decisionFor(it.key) == IndexDownloadConsentDecision.UNKNOWN }
      .toSortedMap(compareBy { it.kind })

    showNotifications(keysToAsk.values.toList())
  }

  fun showNotifications(suggestions: List<SharedIndexSuggestion>) {
    if (suggestions.isEmpty()) return

    val keysToAsk = suggestions.map { it.toKey() }.distinct()
    if (keysToAsk.isEmpty() || myPendingNotification.get()?.first?.containsAll(keysToAsk) == true) return

    val mergedKinds = NlsMessages.formatAndList(suggestions.map { it.sharedIndexConsentNotificationKindPresentableText }.toSortedSet(String.CASE_INSENSITIVE_ORDER).toList())
    val notificationContent = SharedIndexesBundle.message("notification.consent.content", mergedKinds)

    val notification = NotificationGroupManager.getInstance()
      .getNotificationGroup("Download Shared Index")
      .createNotification(type = NotificationType.INFORMATION)
      .apply { isImportant = true }
      .apply { collapseActionsDirection = Notification.CollapseActionsDirection.KEEP_LEFTMOST }
      .setTitle(SharedIndexesBundle.message("notification.consent.title"))
      .setContent(notificationContent)
      .setDropDownText(SharedIndexesBundle.message("notification.consent.action.moreActions"))
      .addAction(
        createSimpleExpiring(
          SharedIndexesBundle.message("notification.consent.action.alwaysAllow"),
          NotificationAlwaysDownload(project, keysToAsk)
        )
      )
      .addAction(
        createSimpleExpiring(
          SharedIndexesBundle.message("notification.consent.action.allowOnce"),
          NotificationAllowOnce(project, suggestions)
        )
      )
      .addAction(
        createSimpleExpiring(
          SharedIndexesBundle.message("notification.consent.action.doNotShow"),
          NotificationDeny(keysToAsk)
        )
      )
      .addAction(
        createSimpleExpiring(
          SharedIndexesBundle.message("notification.consent.action.configure"),
          NotificationConfigure(project)
        )
      )

    notification.whenExpired {
      myPendingNotification.updateAndGet { oldNotification ->
        when {
          oldNotification == null -> null
          oldNotification.second == notification -> null
          else -> oldNotification
        }
      }
    }

    //hide previous one if it exists there
    myPendingNotification.getAndSet(keysToAsk.toSet() to notification)?.second?.expire()
    notification.notify(project)
  }

  // FUS hack: use names classes for Runnables to make them visible in FUS reported for consent notifications.

  private class NotificationAlwaysDownload(
    private val project: Project,
    private val toAllowKeys: List<IndexDownloadConsentDecisionKey>
  ) : Runnable {
    override fun run() {
      service<SharedIndexDownloadConsentStore>().setDecisionFor(toAllowKeys, IndexDownloadConsentDecision.ALLOWED)
      //it's fine to re-start download if needed
      service<SharedIndexDownloadService>().rescanNow(project)
    }
  }

  private class NotificationAllowOnce(
    private val project: Project,
    private val suggestions: List<SharedIndexSuggestion>
  ) : Runnable {
    override fun run() {
      service<SharedIndexDownloadService>().applySuggestions(project, suggestions)
    }
  }

  private class NotificationDeny(private val toDenyKeys: List<IndexDownloadConsentDecisionKey>) : Runnable {
    override fun run() {
      service<SharedIndexDownloadConsentStore>().setDecisionFor(toDenyKeys, IndexDownloadConsentDecision.DENIED)
    }
  }

  private class NotificationConfigure(private val project: Project) : Runnable {
    override fun run() {
      ShowSettingsUtil.getInstance().showSettingsDialog(project, SharedIndexConfigurable::class.java)
    }
  }
}

