package com.intellij.indexing.shared.download

import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.openapi.Disposable
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.Task
import com.intellij.openapi.progress.runUnderNestedProgressAndRelayMessages
import com.intellij.openapi.progress.util.*
import com.intellij.openapi.project.DumbService
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Condition
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.registry.Registry
import com.intellij.openapi.wm.ex.StatusBarEx
import com.intellij.openapi.wm.ex.WindowManagerEx
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.concurrency.EdtScheduledExecutorService
import java.util.concurrent.CountDownLatch
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean
import java.util.concurrent.atomic.AtomicReference


interface DownloadTaskContext {
  val indicator: ProgressIndicator
  fun showBackgroundableProgress()

  /**
   * Any sub activities must be queued with
   * the helper function call to make sure
   * we do not lose the modalityState and any other
   * sophisticated execution context part
   */
  fun queueChildTask(action: () -> Unit)

  fun <Y> suspendIndexingIfNeeded(action: () -> Y) : Y
}

@Service
class SharedIndexDownloadExecutor : Disposable {
  private val LOG = logger<SharedIndexDownloadExecutor>()
  private val isDisposed = AtomicBoolean(false)
  private val oneThreadExecutor = AppExecutorUtil.createBoundedApplicationPoolExecutor("SharedIndexesDownloader", 1)

  override fun dispose() {
    isDisposed.set(true)
  }

  fun submitPreProgressSync(project: Project?,
                            action: DownloadTaskContext.() -> Unit
  ) {
    if (isDisposed.get()) return

    val rootIndicator = ProgressIndicatorBase()

    val isTaskCompleted = AtomicBoolean(false)
    val condition = object : Condition<Any?> {
      override fun value(t: Any?) = isExpired
      val isExpired get() = rootIndicator.isCanceled || isDisposed.get() || (project ?: ApplicationManager.getApplication()).isDisposed
    }

    // The callback can be executed from another thread which is likely running
    // under Progress Manager task background thread. The application may be showing
    // a dialog (e.g. Project Structure or New Project Wizard). We need to have our
    // progress started under the same modality. So we remember the actual modality
    // state to start our activity later on
    //
    // In fact, ProgressManager#run(Task.Backgroundable) tries to find the best suitable
    // modality state for the progress to show up and the start to start.
    // Such progress begins with an EDT invoke later, it it is vital to have the right
    // ModalityState to allow it.
    //
    // By default, in the case of a new thread, it picks the ModalityState.NON_MODAL
    // (you may find that in ProgressManagerCore). So our download only starts after a
    // modal dialog is closed (which is too late), while JDK download is running with
    // the dialog open! We need a better behaviour here, so we have this trick applied.
    //
    // It is an example of a leaky abstraction, unfortunately, we have to have that hack here.
    val modalityState = ApplicationManager.getApplication().defaultModalityState

    // we need to ensure we use the only thread to run the shared indexes download.
    // it seems there is no need to attempt downloading/resolving shared indexes in
    // parallel (e.g. say per project or per downloads). Thanks to caches, the
    // further iterations of the task would re-use computed results from the previous
    // task calls. Right now we pay with one locked background thread for the trick.
    oneThreadExecutor.submit {
      if (condition.isExpired) return@submit

      val progressCancellation = object : Disposable {
        // it is not possible to remove a Disposable from the Disposer
        // without invocation, so we have to have the flag here instead
        private val shouldCancelTheProgress = AtomicBoolean(true)

        fun doNotCancelProgress() = shouldCancelTheProgress.set(false)

        override fun dispose() {
          if (shouldCancelTheProgress.get()) {
            rootIndicator.cancel()
          }
        }
      }

      listOfNotNull(this, project).forEach { parent ->
        Disposer.register(parent, progressCancellation)
      }

      val context = object : DownloadTaskContext {
        private val hasProgress = AtomicBoolean(false)
        override val indicator: ProgressIndicator
          get() = rootIndicator

        override fun showBackgroundableProgress() {
          if (ApplicationManager.getApplication().isHeadlessEnvironment) return
          if (!hasProgress.compareAndSet(false, true)) return

          rootIndicator.checkCanceled()
          val taskCompletionSemaphore = CountDownLatch(1)
          Disposer.register(progressCancellation, Disposable { taskCompletionSemaphore.countDown() })

          //we create an empty progress indicator that broadcasts everything we have in the actual background thread
          val task = object : Task.Backgroundable(project,
                                                  SharedIndexesBundle.message("progress.title.resolving.shared.indexes")) {

            override fun run(indicator: ProgressIndicator) {
              //cancellation passing back from the progress indicator to the task
              runUnderNestedProgressAndRelayMessages(parentProgress = rootIndicator, childProgress = indicator) {
                ProgressIndicatorUtils.awaitWithCheckCanceled(taskCompletionSemaphore)
              }
            }
          }

          object: Runnable {
            override fun run() {
              if (condition.isExpired || isTaskCompleted.get()) return

              if (project == null) {
                //see the long comment above on why it's necessary
                queueChildTask { task.queue() }
                return
              }

              /// workaround for IDEA-252119
              val hasStatusBar = WindowManagerEx.getInstanceEx().findFrameHelper(project)?.statusBar is StatusBarEx
              if (hasStatusBar) {
                queueChildTask { task.queue() }
                return
              }

              val future = EdtScheduledExecutorService.getInstance().schedule(this, modalityState, 1, TimeUnit.SECONDS)
              val disposable = Disposable { future.cancel(false) }
              if (!Disposer.tryRegister(progressCancellation, disposable)) {
                Disposer.dispose(disposable)
              }
            }
          }.run()
        }

        override fun queueChildTask(action: () -> Unit) {
          if (condition.isExpired) return

          //see the long comment above on why it's necessary
          val task = Runnable {
            if (condition.isExpired) return@Runnable

            action()
          }

          ApplicationManager.getApplication()
            .invokeLater(task,
                         modalityState,
                         condition)
        }

        override fun <Y> suspendIndexingIfNeeded(action: () -> Y): Y {
          if (project == null || !service<SharedIndexDownloadConsentStore>().pauseIndexing) {
            return action()
          }

          val result = AtomicReference<Result<Y>>()
          DumbService
            .getInstance(project)
            .suspendIndexingAndRun(SharedIndexesBundle.message("suspend.indexing.message"), Runnable {
              result.set(runCatching(action))
            })

          return result.get().getOrThrow()
        }
      }

      try {
        val actualTask = Runnable {
          if (project != null && Registry.`is`("shared.indexes.download.autoCancel")) {
            project.messageBus
              .connect(progressCancellation)
              .subscribe(ShadeIndexDumbModeTrackerTopic, object : ShadeIndexDumbModeTrackerTopicListener {
                override fun onStableDumbModeDetected() {
                  LOG.warn("Shared indexes download was cancelled. Indexing has completed faster")
                  rootIndicator.cancel()
                }
              })
          }

          action.invoke(context)
        }

        ProgressManager.getInstance().runProcess(actualTask, context.indicator)

        // we check the progress cancellation in the queueChildTask,
        // it should not be cancelled on the normal task completion
        progressCancellation.doNotCancelProgress()
      } catch (t: Throwable) {
        if (t is ControlFlowException) throw t
        LOG.warn("Shared Indexes download task failed unexpectedly. ${t.message}", t)
      }
      finally {
        isTaskCompleted.set(true)
        Disposer.dispose(progressCancellation)
      }
    }
  }
}
