package com.intellij.indexing.shared.download

import com.google.common.annotations.VisibleForTesting
import com.intellij.indexing.shared.platform.impl.SharedIndexChunkConfiguration
import com.intellij.openapi.components.*
import com.intellij.util.xmlb.annotations.OptionTag
import com.intellij.util.xmlb.annotations.Tag
import com.intellij.util.xmlb.annotations.XCollection
import java.time.ZonedDateTime
import java.time.format.DateTimeFormatter
import java.util.*
import java.util.concurrent.locks.ReentrantLock
import kotlin.concurrent.withLock

@Tag("entry")
class IndexDownloadHistoryEntry : BaseState() {
  @get:OptionTag
  var indexKind by string()

  @get:OptionTag
  var indexId by string()

  @get:OptionTag
  var lastDeliveredChunkUniqueId by string()

  @get:OptionTag
  var lastDownloadDate by string()

  @get:OptionTag
  var lastDownloadSize by string()
}

class IndexDownloadHistoryState : BaseState() {
  private var myLookupModificationId : Long = -1
  private var myLookup = mapOf<String, IndexDownloadHistoryEntry>()

  @get:XCollection
  val entries by list<IndexDownloadHistoryEntry>()

  fun filterEntry(indexId: SharedIndexId) = toLookup()[indexId.key()]

  private fun toLookup(): Map<String, IndexDownloadHistoryEntry> {
    var lookup = myLookup
    if (myLookupModificationId != modificationCount) {
      lookup = entries.associateBy { it.key() }
      myLookup = lookup
      myLookupModificationId = modificationCount
    }
    return lookup
  }

  fun removeEntry(candidate: IndexDownloadHistoryEntry) {
    entries.remove(candidate)
    intIncrementModificationCount()
  }

  private fun IndexDownloadHistoryEntry.key() = "${indexId}$$${indexKind}"
  private fun SharedIndexId.key() = "${indexId}$$${kind}"
}

@State(name = "shared-index-download-history", storages = [Storage(StoragePathMacros.CACHE_FILE)], allowLoadInTests = true)
class SharedIndexDownloadHistoryStore : SimplePersistentStateComponent<IndexDownloadHistoryState>(IndexDownloadHistoryState()) {
  private val lock = ReentrantLock()

  override fun loadState(state: IndexDownloadHistoryState) = lock.withLock {
    super.loadState(state)
  }

  fun reset() = loadState(IndexDownloadHistoryState())

  fun updateLastDownload(indexIds: List<SharedIndexId>, chunkId: String, downloadSize: Long) {
    val indexesToProcess = indexIds.distinct()
    if (indexesToProcess.isEmpty()) return

    lock.withLock {
      val state = state
      //cleanup old state - collect first to avoid caches rebuild inside in the `state`
      val toRemove = indexesToProcess.mapNotNull { state.filterEntry(it) }
      val toAdd = indexesToProcess.map { indexId ->
        IndexDownloadHistoryEntry().also {
          it.indexKind = indexId.kind
          it.indexId = indexId.indexId
          it.lastDeliveredChunkUniqueId = chunkId
          it.lastDownloadSize = downloadSize.toString()
          it.lastDownloadDate = ZonedDateTime.now().format(DateTimeFormatter.ISO_ZONED_DATE_TIME.withLocale(Locale.US))
        }
      }

      state.entries.removeAll(toRemove)
      state.entries.addAll(toAdd)
      state.intIncrementModificationCount()
    }
  }

  fun isDownloadMakeSenseFor(indexId: SharedIndexId,
                             @VisibleForTesting checkDescriptor: Boolean = true): Boolean {
    lock.withLock {
      val state = state

      val candidate = state.filterEntry(indexId) ?: return true
      val lastChunkId = candidate.lastDeliveredChunkUniqueId

      // TODO: check if chunk is not too old
      if (lastChunkId == null || candidate.lastDownloadDate == null) {
        state.removeEntry(candidate)
        return true
      }

      if (!checkDescriptor) return false

      //TODO: check there are enough compatible shared indexes in the chunk
      if (!SharedIndexChunkConfiguration.getInstance().isAvailableChunk(lastChunkId)) {
        state.removeEntry(candidate)
        return true
      }

      return false
    }
  }
}
