package com.intellij.indexing.shared.download

import com.intellij.ide.nls.NlsMessages
import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.notification.NotificationGroupManager
import com.intellij.notification.NotificationType
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.text.StringUtil
import java.util.*

data class SharedIndexDownloadInfo(
  val suggestions: List<SharedIndexSuggestion>,
  val downloadSize: Long,
  val downloadTimeMillis: Long,
)

private fun <Y> Collection<Y>.sumByLong(extractValue: Y.() -> Long) : Long {
  var result = 0L
  for (y in this) {
    result += y.extractValue()
  }
  return result
}

@Service
class SharedIndexDownloadNotifications {
  fun notifySharedIndexInstalled(project: Project?,
                                 infos: List<SharedIndexDownloadInfo>
  ) {

    if (infos.isEmpty()) return

    val messageText =
      infos.singleOrNull()?.let { info ->
      info.suggestions.singleOrNull()?.let { suggestion ->
        SharedIndexesBundle.message(
            "eventLog.downloaded.textAndSize",
            suggestion.sharedIndexDownloadedPresentableText,
            StringUtil.formatFileSize(info.downloadSize),
            NlsMessages.formatDurationApproximate(info.downloadTimeMillis)
          )
      }
    } ?: run {
        val totalDownloadSize = infos.sumByLong { downloadSize }
        val totalDownloadTime = infos.sumByLong { downloadTimeMillis }
        val kinds = infos.asSequence().flatMap { it.suggestions }.mapTo(TreeSet()) { it.sharedIndexConsentNotificationKindPresentableText }

        SharedIndexesBundle.message(
            "eventLog.downloaded.textAndSize.merged",
            infos.size,
            StringUtil.formatFileSize(totalDownloadSize),
            NlsMessages.formatDurationApproximate(totalDownloadTime),
            NlsMessages.formatAndList(kinds)
          )
      }

    NotificationGroupManager.getInstance()
      .getNotificationGroup("Download Shared Index Log")
      .createNotification(NotificationType.INFORMATION)
      .setContent(messageText)
      .notify(project)
  }
}


class SharedIndexDownloadNotificationsDebugAction : DumbAwareAction() {
  override fun actionPerformed(e: AnActionEvent) {
    val project = e.project ?: return

    val suggestions = project.service<SharedIndexSuggestionService>().resolveSuggestions()

    service<SharedIndexDownloadNotifications>()
      .notifySharedIndexInstalled(project, suggestions.map {
        SharedIndexDownloadInfo(listOf(it), (342 * 1024 * 1024 * (1 + 0.4 * Math.random())).toLong(), 42_000) }
      )
  }
}
