package com.intellij.indexing.shared.download

import com.google.common.collect.MultimapBuilder
import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.indexing.shared.platform.api.ChunkDescriptor
import com.intellij.indexing.shared.platform.api.IdeIndexInfrastructureVersion
import com.intellij.indexing.shared.platform.impl.DownloadIndexResult
import com.intellij.indexing.shared.platform.impl.SharedIndexChunkConfiguration
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.forEachWithProgress
import com.intellij.openapi.progress.mapWithProgress
import com.intellij.openapi.progress.withPushPop
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.isNotificationSilentMode
import com.intellij.openapi.startup.StartupActivity
import com.intellij.openapi.util.registry.Registry
import java.nio.file.Path
import kotlin.system.measureTimeMillis

private fun isEnabled() = !ApplicationManager.getApplication().isUnitTestMode &&
                          Registry.`is`("shared.indexes.download")


class IndexDownloadServiceProjectStartupActivity : StartupActivity.DumbAware {
  override fun runActivity(project: Project) {
    service<SharedIndexDownloadService>().rescanNow(project)
  }
}

class ShareIndexDownloadAction : DumbAwareAction() {
  override fun actionPerformed(e: AnActionEvent) {
    service<SharedIndexLookup>().invalidateCaches()
    service<SharedIndexDownloadService>().rescanNow(e.project)
  }
}

class SharedIndexDownloadService {
  private val LOG = logger<SharedIndexDownloadService>()
  private val executor get() = service<SharedIndexDownloadExecutor>()

  /**
   * Resolves possible [SharedIndexSuggester], checks and asks for the [IndexDownloadConsentService].
   * Processes only the [SharedIndexSuggestion]s that have the ALLOW consent
   */
  fun rescanNow(project: Project?) {
    if (!isEnabled() || project == null) return

    val expiration = SharedIndexPerProjectTaskExpiration.registerNewTask(project)
    executor.submitPreProgressSync(project) {
      //we filter previously submitted tasks to avoid too many re-computations
      if (expiration.isExpired()) return@submitPreProgressSync

      val suggestions = try {
        collectSuggestionsSuitableForDownload(project)
      } catch (t: Throwable) {
        if (t is ControlFlowException) throw t
        LOG.warn("Failed to collect suggested shared indexes. ${t.message}", t)
        return@submitPreProgressSync
      }.mergeSameSuggestions()

      if (suggestions.isEmpty()) return@submitPreProgressSync

      if (!isNotificationSilentMode(project)) {
        //ask for consent to download more indexes
        project.service<IndexDownloadConsentService>().askForMoreDownloads(suggestions)
      }

      //only allowed suggestions goes for the download
      val actualSuggestions = service<SharedIndexDownloadConsentStore>().filterAllowed(suggestions)
      if (actualSuggestions.isEmpty()) return@submitPreProgressSync

      //we need to make sure the same (on less modal) modality state is re-use here
      queueChildTask {
        applySuggestions(project, actualSuggestions)
      }
    }
  }

  /**
   * Applies given suggestions once, without any checks with the [IndexDownloadConsentService]
   */
  fun applySuggestions(project: Project?, suggestions: List<SharedIndexSuggestion>) {
    if (!isEnabled() || suggestions.isEmpty()) return

    executor.submitPreProgressSync(project) {
      doDownloadSharedIndexes(project, suggestions.mergeSameSuggestions())
    }
  }

  private fun DownloadTaskContext.doDownloadSharedIndexes(project: Project?,
                                                          actualSuggestions: List<SharedIndexSuggestion>) {

    processSuggestions(project, actualSuggestions)
  }

  private fun Collection<SharedIndexSuggestion>.mergeSameSuggestions(): List<SharedIndexSuggestion>
    = associateBy { it.sharedIndexId }.values.toList()

  private fun collectSuggestionsSuitableForDownload(project: Project) : List<SharedIndexSuggestion> {
    val history = service<SharedIndexDownloadHistoryStore>()

    //take 1 - collect what's available for the project
    val allSuggestions = project
      .service<SharedIndexSuggestionService>()
      .resolveSuggestions()
      .groupBy { history.isDownloadMakeSenseFor(it.sharedIndexId) }

    allSuggestions[false]?.let {
      LOG.info("There is no need to download shared indexes for ${it.joinToString()}. Similar compatible shared indexes are available locally.")
    }

    return allSuggestions[true] ?: listOf()
  }

  private fun DownloadTaskContext.processSuggestions(project: Project?,
                                                     suggestions: List<SharedIndexSuggestion>) {
    val requests = collectRequests(indicator, suggestions)
    if (requests.isEmpty()) return

    initiateAuthIfNeeded(project, requests, indicator)

    val indexes = resolveIndexes(indicator, requests)
    if (indexes.isEmpty()) return

    showBackgroundableProgress()
    downloadAndInstallIndexes(project, indicator, indexes)
  }

  private fun collectRequests(
    indicator: ProgressIndicator,
    suggestions: List<SharedIndexSuggestion>
  ): List<Pair<SharedIndexSuggestion, SharedIndexLookupRequest>> {
    indicator.text = SharedIndexesBundle.message("progress.text.resolving.shared.indexes.requests")
    return suggestions.mapWithProgress(indicator) { it, subIndicator ->
      try {
        val request = it.resolveRequest(subIndicator)
        if (request == null) return@mapWithProgress null
        it to request
      }
      catch (e: Throwable) {
        if (e is ControlFlowException) throw e
        LOG.warn("Failed to create for shared index request for $it. ${e.message}", e)
        null
      }
    }.filterNotNull()
  }

  private fun initiateAuthIfNeeded(
    project: Project?,
    requests: List<Pair<SharedIndexSuggestion, SharedIndexLookupRequest>>,
    indicator: ProgressIndicator
  ) {
    indicator.checkCanceled()

    val auths = requests
      .mapNotNull {
        val auth = resolveIndexDownloadAuthExtension(it.second.authId) ?: return@mapNotNull null
        auth to it.first
      }.groupBy({ it.first }, { it.second })

    if (auths.isEmpty()) return

    indicator.text = SharedIndexesBundle.message("progress.text.auth")
    auths.entries.forEachWithProgress(indicator) { (auth, suggestions), subProgress ->
      try {
        subProgress.text2 = auth.presentableName
        auth.checkAuthenticated(
          presentableReason = SharedIndexesBundle.message("auth.presentable.text"),
          indicator = subProgress,
          lifetime = SharedIndexAuthLifetimeHolder.getNextDisposable(project),
          onAuthCompleted = { applySuggestions(project, suggestions) }
        )
      }
      catch (t: Throwable) {
        if (t is ControlFlowException) throw t
        LOG.warn("Failed to check for authentication with $auth. ${t.message}", t)
      }
    }
  }

  private fun resolveIndexes(
    indicator: ProgressIndicator,
    requests: List<Pair<SharedIndexSuggestion, SharedIndexLookupRequest>>
  ): Map<SharedIndexLookupResult, List<SharedIndexSuggestion>> = indicator.withPushPop {
    val ideVersion = IdeIndexInfrastructureVersion.getIdeVersion()
    indicator.text = SharedIndexesBundle.message("progress.text.resolving.shared.indexes.checking")
    indicator.isIndeterminate = false

    //group all requests by SharedIndexLookupRequest (all inheritors implement equals)
    val uniqueRequests = requests
      .groupBy({ it.second }, { it.first }).entries

    val indexLookup = service<SharedIndexLookup>()

    val responses = MultimapBuilder
      .linkedHashKeys()
      .linkedHashSetValues()
      .build<SharedIndexLookupResult, SharedIndexSuggestion>()

    val basesToProcess = ArrayDeque<Pair<SharedIndexLookupRequest, List<SharedIndexSuggestion>>>()
    var processed = 0.0
    for ((lookupRequest, suggestions) in uniqueRequests) {
      indicator.fraction = (processed++) / (basesToProcess.size + uniqueRequests.size)

      try {
        val lookupResult = indexLookup.lookupSharedIndex(lookupRequest, ideVersion, indicator) ?: continue
        responses.putAll(lookupResult, suggestions)
        basesToProcess += lookupResult.resolveBaseRequests().map { it to suggestions }
      }
      catch (e: Throwable) {
        if (e is ControlFlowException) throw e
        LOG.warn("Failed to lookup for shared index $lookupRequest ($suggestions). ${e.message}", e)
      }
    }

    val totalItems = basesToProcess.size + uniqueRequests.size
    while (basesToProcess.isNotEmpty()) {
      indicator.fraction = (processed++) / totalItems

      val (lookupRequest, suggestions) = basesToProcess.removeFirstOrNull() ?: break
      val lookupResult = indexLookup.lookupSharedIndex(lookupRequest, ideVersion, indicator) ?: continue
      if (responses.containsKey(lookupResult)) continue
      responses.putAll(lookupResult, suggestions)

      basesToProcess += lookupResult.resolveBaseRequests().map { it to suggestions }
    }

    //merge same responses together
    return responses.entries()
      .groupBy({ it.key }, { it.value })
      .mapValues { it.value.mergeSameSuggestions() }
  }

  private fun DownloadTaskContext.downloadAndInstallIndexes(
    project: Project?,
    indicator: ProgressIndicator,
    requests: Map<SharedIndexLookupResult, List<SharedIndexSuggestion>>
  ) {
    indicator.text = SharedIndexesBundle.message("progress.text.resolving.shared.indexes.installing")

    val infos = mutableListOf<SharedIndexDownloadInfo>()
    requests
      .entries
      //download smallest shared index first
      .sortedBy { (lookupResult) -> lookupResult.size }
      .forEachWithProgress(indicator) { (lookupResult, suggestions), subIndicator ->
        try {
          LOG.info("Fetching shared index: $lookupResult ($suggestions)")

          val chunkDescriptorToDownload = object : ChunkDescriptor {
            override val chunkUniqueId: String = lookupResult.chunkUniqueId
            override val kind: String = lookupResult.request.kind
            override fun downloadChunk(targetFile: Path, project: Project?, indicator: ProgressIndicator) = suspendIndexingIfNeeded {
              lookupResult.downloadChunk(targetFile, project, indicator)
            }

            override fun toString(): String {
              return chunkUniqueId
            }
          }

          val downloadTime = measureTimeMillis {
            val hadDownloadedSomething = SharedIndexChunkConfiguration.getInstance().downloadChunk(
              chunkDescriptorToDownload, project, subIndicator)
            if (hadDownloadedSomething != DownloadIndexResult.JUST_DOWNLOADED) return@forEachWithProgress
          }

          service<SharedIndexDownloadHistoryStore>().updateLastDownload(
            indexIds = suggestions.map { it.sharedIndexId },
            chunkId = lookupResult.chunkUniqueId,
            downloadSize = lookupResult.size
          )

          infos += SharedIndexDownloadInfo(suggestions, lookupResult.size, downloadTime)
        }
        catch (e: Throwable) {
          if (e is ControlFlowException) throw e
          LOG.warn("Failed to download shared index $lookupResult ($suggestions). ${e.message}", e)
        }
    }
    service<SharedIndexDownloadNotifications>().notifySharedIndexInstalled(project, infos)
  }
}
