package com.intellij.indexing.shared.download

import com.intellij.openapi.Disposable
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.project.DumbService
import com.intellij.openapi.project.Project
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.messages.Topic
import java.util.concurrent.Future
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicReference

val ShadeIndexDumbModeTrackerTopic = Topic.create("com.intellij.sharedIndexesDumbModeEnd",
                                                  ShadeIndexDumbModeTrackerTopicListener::class.java,
                                                  Topic.BroadcastDirection.NONE)

interface ShadeIndexDumbModeTrackerTopicListener {
  fun onStableDumbModeDetected()
}

@Service
class ShadeIndexDumbModeTracker(val project: Project) : Disposable {
  private var lastDumbModeEnter: Long = now()
  private val myPendingFuture = AtomicReference<Future<*>?>(null)

  private fun now() = System.nanoTime()

  override fun dispose() {
    myPendingFuture.getAndSet(null)?.cancel(false)
  }

  fun enteredDumbMode() {
    lastDumbModeEnter = now()
  }

  fun exitDumbMode() {
    val lastDumbModeExit = now()

    val hideNotification = {
      if (lastDumbModeEnter <= lastDumbModeExit) {
        onStableDumbModeDetected()
      }
    }

    val newFuture = AppExecutorUtil.getAppScheduledExecutorService()
      .schedule(hideNotification, 15L, TimeUnit.SECONDS)

    //cancel the future to avoid a memory leak
    myPendingFuture.getAndSet(newFuture)?.cancel(false)
  }

  private fun onStableDumbModeDetected() {
    project.messageBus.syncPublisher(ShadeIndexDumbModeTrackerTopic).onStableDumbModeDetected()
  }
}

class ShadeIndexDumbModeTrackerListener(private val project: Project) : DumbService.DumbModeListener {
  private val service
    get() = project.service<ShadeIndexDumbModeTracker>()

  override fun enteredDumbMode() { service.enteredDumbMode() }
  override fun exitDumbMode() { service.exitDumbMode() }
}
