package com.intellij.indexing.shared.download

import com.fasterxml.jackson.databind.node.ArrayNode
import com.fasterxml.jackson.databind.node.ObjectNode
import com.google.common.annotations.VisibleForTesting
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.util.registry.Registry
import java.time.Instant
import java.util.*

enum class SharedIndexUrlDecision {
  UNKNOWN,
  TRY,
  SKIP
}

interface SharedIndexListLookupResult {
  /**
   * Checks if a given base URL for a later download of `index.json.xz`
   * is worth a network check. There can be too many URLs that does
   * not exist. This service helps to avoid useless network requests
   * for URLs that may not exist for sure.
   *
   * The base URL means a URL from [SharedIndexLookupRequest.indexUrls()],
   * and it would try to check a `URL/index.json.xz` later in the [SharedIndexLookup]
   * implementation
   */
  fun testUrl(request: SharedIndexLookupRequest, url: String) : SharedIndexUrlDecision
}

private val rejectAllListLookupResult = object : SharedIndexListLookupResult {
  override fun testUrl(request: SharedIndexLookupRequest, url: String) = SharedIndexUrlDecision.SKIP
  override fun toString() = "RejectAllListLookupResult"
}

@Service
class SharedIndexListLookup {
  private val LOG = logger<SharedIndexListLookup>()

  fun groupUrsByListPaths(urls: List<String>) = urls
    .mapNotNull { extractListBaseUrl(it)?.to(it) }
    .groupBy({ it.first }, { it.second })

  /**
   * It can be too many URLs to process here or with respect to the outer loops
   * We attempt to download and use a list.json.xz metadata files to avoid
   * some non-existing URLs from check. The implementation uses a cached results
   * for list.json.xz with respect to the TTL. It is still possible the method
   * includes a URL into the result that does not exists (false positives in the filter).
   *
   * Any failure in checking list.json.xz would mean empty list to be returned.
   * In fact it could be either a broken CDN layout or network error. Both
   * results are fine to skip any further processing
   */
  fun collectUrlsToProcess(request: SharedIndexLookupRequest,
                           indicator: ProgressIndicator): List<String> {
    val urls = request.indexUrls()
    if (!request.useListJsonIndexToSkipUrls || !Registry.`is`("shared.indexes.use.cdn.list")) return urls
    val checkBases = groupUrsByListPaths(urls)

    val skipUrls = HashSet<String>()
    for ((listBase, listUrls) in checkBases) {
      val lookup = lookupListFor(request, indicator, listBase)

      for (url in listUrls) {
        if (lookup.testUrl(request, url) == SharedIndexUrlDecision.SKIP) {
          skipUrls += url
          LOG.debug("Shared index download URL $url is not listed, will be skipped")
        }
      }
    }
    //it is essential to keep the original order
    return urls.filter { it !in skipUrls }
  }

  //there are list.json.xz files in the following layout:
  // - <base>
  //    - list.json.xz
  //    - <some folder>/index.json.xz
  //    - <another folder>/index.json.xz
  // this methods resolves a list.json.xz with cache and returns a filter
  // for all folders that may are likely to exist (false-positives are possible)
  fun extractListBaseUrl(url: String) : String? {
    @Suppress("NAME_SHADOWING")

    val url = url.trim('/')
    val idx = url.lastIndexOf('/')
    if (idx <= 0 || idx + 1 == url.length) return null
    return url.substring(0, idx).trim('/')
  }

  private data class ListLookupRequest(
    val listUrl: String
  )

  private data class ListLookupCacheEntry(
    val computed : Instant,
    val checker: SharedIndexListLookupResult?
  )

  private val cache = object : SharedIndexNetworkCache<ListLookupRequest, ListLookupCacheEntry>() {
    override val ListLookupCacheEntry.created: Instant
      get() = this.computed
  }

  fun invalidateCaches() {
    cache.invalidateCaches()
  }

  @VisibleForTesting
  fun cacheSize() = cache.cacheSize()

  /**
   * @see extractListBaseUrl
   */
  fun lookupListFor(request: SharedIndexLookupRequest,
                    indicator: ProgressIndicator,
                    listUrl: String
  ): SharedIndexListLookupResult = cache.lookupOrCompute(request = ListLookupRequest(listUrl),
                                                          indicator = indicator) {
    val lookupUrl = listUrl.trim('/') + "/list.json.xz"
    val json = service<SharedIndexDownloadJson>().downloadJsonXz(request, lookupUrl, indicator)

    val checker = if (json == null) {
      LOG.info("Failed to download the list of allowed URLs from $lookupUrl. It looks like there is no internet connection. The $request is skipped")
      // failed to download the list, we do not know what mean, but we cannot afford testing it!
      rejectAllListLookupResult
    } else {
      val siblings = parseListJson(lookupUrl, request, json)
      if (siblings == null || siblings.isEmpty()) {
        LOG.debug("Invalid data is downloaded as the list of allowed URLs from $lookupUrl. The $request is skipped")
        //is a CDN broken or updated?
        rejectAllListLookupResult
      } else {
        UrlSiblingsChecker(listUrl, siblings)
      }
    }

    ListLookupCacheEntry(computed = now, checker = checker)
  }?.checker ?: rejectAllListLookupResult /* e.g. when auth failed */

  private fun parseListJson(lookupUrl: String, request: SharedIndexLookupRequest, json: ObjectNode) : Set<String>? {
    val listVersion = json.get("list_version")?.asText()
    if (listVersion != "1") {
      LOG.trace("List Index data version mismatch. The current version is $listVersion in $lookupUrl for $request")
      return null
    }

    val names = (json.get("names") as? ArrayNode) ?: run {
      LOG.trace("Index data format is incomplete. Missing 'entries' element in $lookupUrl for $request")
      return null
    }

    return names.mapNotNull { it.asText() }.toSet()
  }
}

private data class UrlSiblingsChecker(
  val listUrl: String,
  val siblings: Set<String>
) : SharedIndexListLookupResult {
  override fun testUrl(request: SharedIndexLookupRequest, url: String): SharedIndexUrlDecision {
    val prefix = listUrl.trim('/') + "/"
    if (!url.startsWith(prefix)) return SharedIndexUrlDecision.UNKNOWN

    val sibling = url.removePrefix(prefix).trim('/')
    if (sibling.contains("/")) return SharedIndexUrlDecision.UNKNOWN

    if (sibling in siblings) return SharedIndexUrlDecision.TRY
    return SharedIndexUrlDecision.SKIP
  }
}

