// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.download

import com.fasterxml.jackson.databind.node.ArrayNode
import com.fasterxml.jackson.databind.node.ObjectNode
import com.google.common.annotations.VisibleForTesting
import com.intellij.indexing.shared.metadata.SharedIndexMetadata
import com.intellij.indexing.shared.platform.api.IdeIndexInfrastructureVersion
import com.intellij.openapi.components.service
import com.intellij.openapi.components.serviceIfCreated
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.forEachWithProgress
import java.time.Instant

class SharedIndexLookup {
  private val LOG = logger<SharedIndexLookup>()

  private data class CacheEntry(
    val ideVersion: IdeIndexInfrastructureVersion, //dynamic plugins may change the version!
    val computed: Instant,
    val value: SharedIndexLookupResult?
  )

  private val cache = object: SharedIndexNetworkCache<SharedIndexLookupRequest, CacheEntry>() {
    override val CacheEntry.created: Instant
      get() = this.computed
  }

  fun lookupSharedIndex(
    request: SharedIndexLookupRequest,
    ideVersion: IdeIndexInfrastructureVersion,
    indicator: ProgressIndicator
  ): SharedIndexLookupResult? {
    return cache.lookupOrCompute(
      request = request,
      indicator = indicator,
      validKeyPredicate = { it.ideVersion == ideVersion },
      computeFunction = {
        val result = try {
          lookupSharedIndexWithoutCache(request, ideVersion, indicator)
        }
        catch (t: Throwable) {
          if (t is ControlFlowException) throw t
          if (t is SharedIndexAuthException) throw t
          indicator.checkCanceled()
          LOG.warn("Failed to lookup shared index for $request. ${t.message}", t)
          null
        }

        CacheEntry(
          ideVersion = ideVersion,
          computed = now,
          value = result
        )
      }
    )?.value
  }

  private fun lookupSharedIndexWithoutCache(
    request: SharedIndexLookupRequest,
    ourVersion: IdeIndexInfrastructureVersion,
    indicator: ProgressIndicator
  ): SharedIndexLookupResult? = lookupSharedIndexWithoutCache(request, indicator) { _, results ->
    val bestSuitableIndex = SharedIndexMetadata.selectBestSuitableIndex(ourVersion, results)
    if (bestSuitableIndex == null && LOG.isTraceEnabled) {
      LOG.trace(buildString {
        appendLine("Among ${results.size} candidate indexes no one is compatible with current IDE")
        appendLine("Current IDE version:")
        appendLine("    ${ourVersion.version}")
        appendLine("Candidates:")
        for ((index, result) in results.withIndex()) {
          appendLine("${(index + 1)}) Result : $result")
          appendLine("${(index + 1)}) Version: ${result.indexVersion}")
        }
      })
    }
    bestSuitableIndex
  }

  fun <T: Any> lookupSharedIndexWithoutCache(
    request: SharedIndexLookupRequest,
    indicator: ProgressIndicator,
    matchItem: (url: String, List<SharedIndexLookupResult>) -> T?
  ): T? {
    val urlsToProcess = service<SharedIndexListLookup>().collectUrlsToProcess(request, indicator)

    urlsToProcess
      .forEachWithProgress(indicator) { url, subIndicator ->
      subIndicator.text2 = url
      val results = downloadIndexList(request, url.trim('/') + "/index.json.xz", subIndicator)
      subIndicator.checkCanceled()
      matchItem(url, results)?.let { best ->
        LOG.info("Selected pre-built shared index $best from $url for $request")
        return best
      }
    }

    LOG.debug("No matching shared indexes are found for $request. The following URLs were tried: " + urlsToProcess.joinToString())
    return null
  }

  fun invalidateCaches() {
    serviceIfCreated<SharedIndexListLookup>()?.invalidateCaches()
    cache.invalidateCaches()
  }

  @VisibleForTesting
  fun cacheSize() = cache.cacheSize()

  @VisibleForTesting
  fun downloadIndexList(
    request: SharedIndexLookupRequest,
    indexUrl: String,
    indicator: ProgressIndicator
  ): List<SharedIndexLookupResult> {

    val json = service<SharedIndexDownloadJson>()
                 .downloadJsonXz(request, indexUrl, indicator) ?: return listOf()

    val listVersion = json.get("list_version")?.asText()
    if (listVersion != "1") {
      LOG.trace("Index data version mismatch. The current version is $listVersion in $indexUrl for $request")
      return listOf()
    }

    val entries = (json.get("entries") as? ArrayNode) ?: run {
      LOG.trace("Index data format is incomplete. Missing 'entries' element in $indexUrl for $request")
      return listOf()
    }

    return entries.elements().asSequence().mapNotNull { node ->
      if (node !is ObjectNode) return@mapNotNull null

      val url = node.get("url")?.asText() ?: return@mapNotNull null
      val sha = node.get("sha256")?.asText() ?: return@mapNotNull null
      val size = node.get("size")?.asLong() ?: return@mapNotNull null
      val indexVersionJson = node.get("metadata") as? ObjectNode ?: return@mapNotNull null
      val bases = (indexVersionJson.get("bases") as? ArrayNode)?.mapNotNull {
        (it as? ObjectNode)?.get("sha256")?.asText()?.let(::SharedIndexBaseId)
      }?.distinct() ?: listOf()

      val compression = SharedIndexCompression.findByType(node.get("format")?.asText()) ?: return@mapNotNull null
      val indexVersion = SharedIndexMetadata.tryReadIndexesVersion(indexVersionJson)
                         ?: return@mapNotNull null

      SharedIndexLookupResult(
        request,
        url,
        sha,
        size,
        compression,
        indexVersion,
        bases
      )
    }.toList()
  }
}
