package com.intellij.indexing.shared.download

import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.indexing.shared.metadata.SharedIndexMetadata
import com.intellij.indexing.shared.platform.api.IndexInfrastructureVersion
import com.intellij.indexing.shared.platform.impl.SharedIndexesFusCollector
import com.intellij.indexing.shared.util.downloadUrlWithValidation
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.project.Project
import com.intellij.util.progress.SubTaskProgressIndicator
import java.nio.file.Path
import java.util.*
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicLong

private val LOG = logger<SharedIndexLookupResult>()

class SharedIndexLookupResult(
  val request: SharedIndexLookupRequest,
  val url: String,
  val sha256: String,
  val size: Long,
  val compression: SharedIndexCompression,
  val sharedIndexMetadata: SharedIndexMetadata,
  private val basesIndexes: List<SharedIndexBaseId>
) {
  override fun equals(other: Any?): Boolean {
    if (this === other) return true
    if (other !is SharedIndexLookupResult) return false
    return this.url == other.url && this.sha256 == other.sha256
  }

  val indexVersion: IndexInfrastructureVersion by sharedIndexMetadata::indexInfrastructureVersion

  override fun hashCode() = Objects.hash(url, sha256)

  fun resolveBaseRequests(): List<SharedIndexLookupRequest> = request.resolveBasesRequests(basesIndexes.distinct())

  val chunkUniqueId = "${request.kind}-$sha256-${indexVersion.weakVersionHash}"

  override fun toString() = "SharedIndexResult(request='${request}', " +
                            "url='$url', " +
                            "weakHash=${indexVersion.weakVersionHash}, " +
                            "weakBaseVersionHash=${indexVersion.weakBaseVersionHash}, " +
                            "sha256='$sha256')"
}

fun SharedIndexLookupResult.downloadChunk(targetFile: Path, project: Project?, indicator: ProgressIndicator): Boolean {
  val message = SharedIndexesBundle.message("downloading.shared.indexes", request.presentableChunkName)
  if (indicator is SubTaskProgressIndicator) {
    indicator.text = message
  }
  else {
    indicator.text = message
    indicator.text2 = null
  }

  SharedIndexesFusCollector.reportIndexDownloadStarted(project, request.kind, chunkUniqueId)
  val headers = service<SharedIndexAuthHolder>().resolveHeaders(url, this, indicator)

  val downloadTime = AtomicLong(System.nanoTime())
  val downloadResult = runCatching {
    try {
      handleHttpDownloadExceptions(LOG, request, url) {
        downloadUrlWithValidation(
          url = url,
          targetFile = targetFile,
          expectedSize = size,
          expectedSha256 = sha256,
          headers = headers,
          compression = compression,
          indicator = indicator
        )
      }
    }
    finally {
      downloadTime.set(System.nanoTime() - downloadTime.get())
    }
  }

  val sharedIndexSize = downloadResult.getOrNull()
  val finishType = when {
    downloadResult.isFailure && downloadResult.exceptionOrNull() is ControlFlowException -> SharedIndexesFusCollector.FinishType.CANCELLED
    downloadResult.isFailure -> SharedIndexesFusCollector.FinishType.ERROR
    sharedIndexSize == null || sharedIndexSize <= 0 -> SharedIndexesFusCollector.FinishType.ERROR
    else -> SharedIndexesFusCollector.FinishType.SUCCESS
  }

  SharedIndexesFusCollector.reportIndexDownloadFinished(
    project,
    finishType,
    request.kind,
    chunkUniqueId,
    TimeUnit.NANOSECONDS.toMillis(downloadTime.get()),
    size,
    sharedIndexSize ?: 0,
    toSpeedKBytesPerSecond(size, downloadTime.get())
  )

  return finishType == SharedIndexesFusCollector.FinishType.SUCCESS && sharedIndexSize != null && sharedIndexSize > 0
}

private fun toSpeedKBytesPerSecond(size: Long, durationNano: Long): Long {
  val millis = TimeUnit.NANOSECONDS.toMillis(durationNano)
  if (millis == 0L) {
    return 0
  }
  val kbPerMs = (size / 1024) / millis.toDouble()
  return (kbPerMs * 1000).toLong()
}

