package com.intellij.indexing.shared.download

import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProgressIndicator
import java.net.SocketTimeoutException
import java.net.UnknownHostException
import java.time.Duration
import java.time.Instant
import com.intellij.openapi.util.registry.Registry
import java.util.concurrent.ConcurrentHashMap

interface SharedIndexNetworkCacheScope {
  val now: Instant
}

private object DefaultSharedIndexNetworkCacheScope : SharedIndexNetworkCacheScope {
  override val now: Instant
    get() = Instant.now()
}

abstract class SharedIndexNetworkCache<Key : Any, Value : Any>(
  private val scope: SharedIndexNetworkCacheScope = DefaultSharedIndexNetworkCacheScope,
  private val TTL: Duration = Duration.ofMinutes(30)
) {
  private val LOG = Logger.getInstance(javaClass)
  private val cache = ConcurrentHashMap<Key, Value?>()

  fun invalidateCaches() {
    cache.clear()
  }

  fun cacheSize() = cache.size

  protected abstract val Value.created: Instant

  fun lookupOrCompute(
    request: Key,
    indicator: ProgressIndicator,
    validKeyPredicate: (Value) -> Boolean = { true },
    computeFunction: SharedIndexNetworkCacheScope.() -> Value?
  ): Value? {
    val compute = java.util.function.Function<Key, Value?> {
      try {
        computeFunction(scope)
      }
      catch (t: SharedIndexAuthException) {
        null
      }
      catch (e: UnknownHostException) {
        null
      }
      catch (e: SocketTimeoutException) {
        null
      }
      catch (t: Throwable) {
        if (t is ControlFlowException) throw t

        indicator.checkCanceled()
        LOG.warn("Failed to compute cache value for $request. ${t.message}", t)
        null
      }
    }

    if (!Registry.`is`("shared.indexes.cache.cdn")) {
      return compute.apply(request)
    }

    while (true) {
      val cachedEntry = cache.computeIfAbsent(request, compute) ?: return null

      if (Duration.between(cachedEntry.created, scope.now) < TTL && validKeyPredicate(cachedEntry)) {
        return cachedEntry
      }

      cache.remove(request, cachedEntry)
    }
  }
}
