package com.intellij.indexing.shared.download

import com.intellij.openapi.Disposable
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.extensions.ExtensionPointListener
import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.openapi.extensions.PluginDescriptor
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.project.Project
import org.jetbrains.annotations.Nls

/**
 * An offline description of Shared Index that we are (in theory)
 * able to deliver from Internet, if allowed,
 *
 * this class is used as a key to merge a possible duplicate suggestions
 */
data class SharedIndexId(
  val kind: String,
  val url: String,
  val indexId: String
)

interface SharedIndexSuggestion {
  val sharedIndexId: SharedIndexId

  /**
   * The function to execute to get the information on the [SharedIndexLookupRequest]
   */
  fun resolveRequest(indicator: ProgressIndicator): SharedIndexLookupRequest?

  /**
   * The short name of that shared index kind for the download shared index consent notification,
   * it must not include 'shared indexes' or similar text
   */
  @get:Nls
  val sharedIndexConsentNotificationKindPresentableText: String

  /**
   * "Shared Indexes are downloaded" message localized and specified for that suggestion
   */
  @get:Nls
  val sharedIndexDownloadedPresentableText: String
}

private val sharedIndexSuggesterEP = ExtensionPointName.create<SharedIndexSuggester>("com.intellij.sharedIndexSuggester")

/**
 * The Shared Indexes lookup extension, it is executed to collect possible
 * downloadable shared indexes
 */
interface SharedIndexSuggester {
  fun suggestRequests(project: Project): List<SharedIndexSuggestion>
}

class SharedIndexSuggestionService(val project: Project) : Disposable {
  private val LOG = logger<SharedIndexSuggestionService>()

  override fun dispose() = Unit

  init {
    sharedIndexSuggesterEP.addExtensionPointListener(object : ExtensionPointListener<SharedIndexSuggester> {
      override fun extensionAdded(extension: SharedIndexSuggester, pluginDescriptor: PluginDescriptor) {
        service<SharedIndexDownloadService>().rescanNow(project)
      }
    }, this)
  }

  fun resolveSuggestions(): List<SharedIndexSuggestion> {
    val suggestions = mutableListOf<SharedIndexSuggestion>()
    sharedIndexSuggesterEP.forEachExtensionSafe {
      suggestions += it.suggestRequests(project)
    }

    val candidates = suggestions.groupBy { it.sharedIndexId }.mapValues {
      if (it.value.size > 1) {
        LOG.warn("Several extensions reported the same kind/id for shared index. Only one will be used from: " + it.value)
      }
      it.value.last()
    }.values.toList()

    return candidates.toList()
  }
}
