package com.intellij.indexing.shared.download

import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.io.HttpRequests
import java.io.IOException
import java.net.ConnectException
import java.net.SocketTimeoutException
import java.net.UnknownHostException

inline fun <T> handleHttpDownloadExceptions(LOG: Logger,
                                            request: Any,
                                            url: String,
                                            action: () -> T): T? {
  try {
    return action()
  }
  catch (e: ConnectException) {
    LOG.debug("Cannot connect to the host for $url for $request. ${e.message}")
  }
  catch (e: UnknownHostException) {
    LOG.debug("Cannot resolve host for $url for $request. ${e.message}")
  }
  catch (e: SocketTimeoutException) {
    LOG.debug("Connection expired by timeout at $url for $request. ${e.message}")
  }
  catch (e: HttpRequests.HttpStatusException) {
    //it is OK to CloudFront backed CDN to return 403 instead of 404 for paths that does not exist
    //the message is node a bit cryptic to lower questions from the customers
    if (e.statusCode < 500) {
      LOG.debug("No data available for $url for $request. ${e.statusCode}")
    }
    else {
      LOG.info("No data available for $url for $request. ${e.statusCode}")
    }
  }
  catch (e: IOException) {
    LOG.debug("Failed to download $url for $request. ${e.message}")
  }
  catch (t: Throwable) {
    if (t is ControlFlowException) throw t
    LOG.warn("Failed to download $url for $request. ${t.message}")
  }

  return null
}
