// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.generate;

import com.intellij.indexing.shared.platform.api.IdeIndexInfrastructureVersion;
import com.intellij.indexing.shared.platform.api.IndexInfrastructureVersion;
import com.intellij.indexing.shared.platform.impl.SharedIndexExtension;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.util.SystemProperties;
import com.intellij.util.indexing.*;
import com.intellij.util.indexing.impl.*;
import com.intellij.util.indexing.impl.forward.ForwardIndex;
import com.intellij.util.indexing.impl.forward.ForwardIndexAccessor;
import com.intellij.util.indexing.impl.forward.MapForwardIndexAccessor;
import com.intellij.util.indexing.impl.forward.PersistentMapBasedForwardIndex;
import com.intellij.util.io.IOUtil;
import com.intellij.util.io.PersistentHashMap;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Collections;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.function.IntConsumer;

public class HashBasedIndexGenerator<K, V> {
  private static final Logger LOG = Logger.getInstance(HashBasedIndexGenerator.class);

  @NotNull
  private final SharedIndexExtension<K, V> mySharedExtension;

  @NotNull
  private final FileBasedIndex.InputFilter myInputFilter;

  @NotNull
  private final Path myStorageFile;

  private final AtomicInteger myIndexedFilesNumber = new AtomicInteger();

  private final AtomicBoolean myIsEmpty = new AtomicBoolean(true);
  private final AtomicBoolean myCollisionReported = new AtomicBoolean();

  private InvertedIndex<K, V, FileContent> myIndex;
  @NotNull
  private final ID<K, V> myIndexId;

  public HashBasedIndexGenerator(@NotNull FileBasedIndexExtension<K, V> originalExtension, @NotNull Path chunkRoot) {
    myIndexId = originalExtension.getName();
    Path indexPath = SharedIndexExtension.getSharedIndexDir(chunkRoot, myIndexId);
    myStorageFile = indexPath.resolve(myIndexId.getName());
    IndexInfrastructureVersion ideVersion = IdeIndexInfrastructureVersion.getIdeVersion().getVersion();
    mySharedExtension = SharedIndexExtension.findExtensionAndOpen(originalExtension,
                                                                  indexPath,
                                                                  false,
                                                                  ideVersion);

    FileBasedIndex.InputFilter filter = originalExtension.getInputFilter();

    if (filter instanceof FileBasedIndex.FileTypeSpecificInputFilter) {
      Set<FileType> fileTypes = new HashSet<>();
      ((FileBasedIndex.FileTypeSpecificInputFilter)filter).registerFileTypesUsedForIndexing(fileTypes::add);
      myInputFilter = FileBasedIndexImpl.composeInputFilter(filter, file -> fileTypes.contains(file.getFileType()));
    }
    else {
      myInputFilter = filter;
    }
  }

  public void openIndex() throws IOException {
    FileBasedIndexExtension<K, V> extension = mySharedExtension.getFileBasedIndexExtension();
    IndexStorage<K, V> indexStorage = new MapIndexStorage<>(myStorageFile,
                                                            extension.getKeyDescriptor(),
                                                            extension.getValueExternalizer(),
                                                            extension.getCacheSize(),
                                                            extension.keyIsUniqueForIndexedFile()) {
      @Override
      protected void checkCanceled() {
        //ignore
      }

      @Override
      public void addValue(K k, int inputId, V v) throws StorageException {
        super.addValue(k, inputId, v);
        myIsEmpty.set(false);
      }

      @Override
      public void removeAllValues(@NotNull K k, int inputId) {
        reportHashCollision(inputId);
      }

      @Override
      protected boolean compactOnClose() {
        return true;
      }
    };

    boolean isSingleEntryIndex = extension instanceof SingleEntryFileBasedIndexExtension;
    ForwardIndex forwardIndex = createForwardIndex(myStorageFile, false);
    ForwardIndexAccessor<K, V> forwardIndexAccessor = createForwardIndexAccessor(mySharedExtension, this::reportHashCollision);
    myIndex = new MapReduceIndex<>(extension, indexStorage, forwardIndex, forwardIndexAccessor) {
      @NotNull
      @Override
      protected Map<K, V> mapByIndexer(int inputId, @NotNull FileContent content) {
        Map<K, V> data = super.mapByIndexer(inputId, content);
        if (isSingleEntryIndex && !data.isEmpty()) {
          data = Collections.singletonMap((K)(Integer)inputId, data.values().iterator().next());
        }
        return transformIndexedData(data, content);
      }

      @Override
      protected void updateForwardIndex(int inputId, @NotNull InputData<K, V> data) throws IOException {
        super.updateForwardIndex(inputId, data);
        try {
          visitInputData(inputId, data);
        }
        catch (StorageException e) {
          throw new IOException(e);
        }
      }

      @Override
      public void updateWithMap(@NotNull AbstractUpdateData<K, V> updateData) throws StorageException {
        super.updateWithMap(updateData);
        myIndexedFilesNumber.incrementAndGet();
      }

      @Override
      public void checkCanceled() {
        //ignore
      }

      @Override
      protected void requestRebuild(@NotNull Throwable e) {
        throw new RuntimeException("Error while processing " + getExtension().getName(), e);
      }
    };
  }

  private void reportHashCollision(int hashId) {
    if (myCollisionReported.compareAndSet(false, true)) {
      LOG.error("Index \"" + mySharedExtension.getFileBasedIndexExtension().getName() + "\" contains data collision for hashId = " + hashId,
                (Throwable)null);
    }
  }

  public static <K, V> @NotNull MapForwardIndexAccessor<K, V> createForwardIndexAccessor(@NotNull SharedIndexExtension<K, V> extension,
                                                                                         @NotNull IntConsumer collisionReporter) {
    return new MapForwardIndexAccessor<>(new InputMapExternalizer<>(extension.getFileBasedIndexExtension())) {
      @Override
      public @NotNull InputDataDiffBuilder<K, V> createDiffBuilderByMap(int inputId, @Nullable Map<K, V> map) {
        return new MapInputDataDiffBuilder<>(inputId, map) {
          @Override
          public boolean differentiate(@NotNull Map<K, V> newData,
                                       @NotNull KeyValueUpdateProcessor<? super K, ? super V> addProcessor,
                                       @NotNull KeyValueUpdateProcessor<? super K, ? super V> updateProcessor,
                                       @NotNull RemovedKeyProcessor<? super K> removeProcessor) throws StorageException {
            return super.differentiate(newData, addProcessor, (k, v, inputId) -> {
              collisionReporter.accept(myInputId);
              updateProcessor.process(k, v, inputId);
            }, (k, inputId1) -> {
              collisionReporter.accept(myInputId);
              removeProcessor.process(k, inputId1);
            });
          }
        };
      }
    };
  }

  public static @NotNull PersistentMapBasedForwardIndex createForwardIndex(@NotNull Path storageFile,
                                                                           boolean readOnly) throws IOException {
    return new PersistentMapBasedForwardIndex(getForwardIndexPath(storageFile), readOnly);
  }

  public static boolean shouldCreateForwardIndex(@NotNull SharedIndexExtension<?, ?> sharedExtension) {
    FileBasedIndexExtension<?, ?> extension = sharedExtension.getFileBasedIndexExtension();
    boolean needsForwardIndex =
      extension.needsForwardIndexWhenSharing() || SystemProperties.is("shared.indexes.always.use.forward.index");
    return needsForwardIndex && !(extension instanceof SingleEntryFileBasedIndexExtension);
  }

  @NotNull
  private static Path getForwardIndexPath(@NotNull Path storageFile) {
    return storageFile.resolveSibling(storageFile.getFileName() + ".forward");
  }

  protected void visitInputData(int hashId, @NotNull InputData<K, V> data) throws StorageException {

  }

  @NotNull
  protected Map<K, V> transformIndexedData(@NotNull Map<K, V> data, @NotNull FileContent content) {
    return data;
  }

  public void closeIndex() throws IOException {
    if (myIndex != null) {
      myIndex.dispose();
    }
    mySharedExtension.closeResources();
    if (!shouldCreateForwardIndex(mySharedExtension)) {
      IOUtil.deleteAllFilesStartingWith(getForwardIndexPath(myStorageFile));
    }
  }

  public void indexFile(int hashId, @NotNull FileContent fileContent) {
    if (!FileBasedIndexImpl.acceptsInput(myInputFilter, fileContent)) {
      return;
    }
    if (!myIndex.mapInputAndPrepareUpdate(hashId, fileContent).compute()) {
      throw new RuntimeException("Index computation returned false for hashId = " + hashId + ", " +
                                 "file = " + fileContent.getFile().getPath() + ", " +
                                 "index = " + getIndexName());
    }
  }

  @NotNull
  public Path getIndexRoot() {
    return myStorageFile.getParent();
  }

  public int getIndexedFilesNumber() {
    return myIndexedFilesNumber.get();
  }

  public boolean isEmpty() {
    return myIsEmpty.get();
  }

  @NotNull
  public String getIndexName() {
    return myIndexId.getName();
  }

  public InvertedIndex<K, V, FileContent> getIndex() {
    return myIndex;
  }
}
