// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.generate;

import com.google.common.collect.Maps;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.psi.PsiElement;
import com.intellij.psi.stubs.*;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.indexing.FileBasedIndexExtension;
import com.intellij.util.indexing.FileContent;
import com.intellij.util.indexing.StorageException;
import com.intellij.util.indexing.impl.EmptyInputDataDiffBuilder;
import com.intellij.util.indexing.impl.InputData;
import com.intellij.util.indexing.impl.MapReduceIndex;
import com.intellij.util.indexing.impl.UpdateData;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;

public final class StubHashBasedIndexGenerator extends HashBasedIndexGenerator<Integer, SerializedStubTree> {
  private static final Logger LOG = Logger.getInstance(StubHashBasedIndexGenerator.class);

  private final Map<StubIndexKey<?, ?>, HashBasedIndexGenerator<?, ?>> myStubIndexesGeneratorMap = new HashMap<>();
  @NotNull
  private final Set<FileType> myNoStubTreeFileTypes;

  private final Set<FileType> myIndexedFileTypes = ContainerUtil.newConcurrentSet();

  public StubHashBasedIndexGenerator(@NotNull Path chunkRoot,
                                     @NotNull List<StubIndexExtension<?, ?>> stubIndexExtensions,
                                     @NotNull Set<FileType> noStubTreeFileTypes) {
    super(new StubUpdatingIndex(), chunkRoot);
    for (StubIndexExtension<?, ?> stubIndexExtension : stubIndexExtensions) {
      myStubIndexesGeneratorMap.put(stubIndexExtension.getKey(), createGenerator(stubIndexExtension));
    }
    myNoStubTreeFileTypes = noStubTreeFileTypes;
    LOG.info("Stub trees will not be saved for " + myNoStubTreeFileTypes + " file types");
  }

  @NotNull
  public Set<FileType> getIndexedFileTypes() {
    return myIndexedFileTypes;
  }

  @Override
  public void indexFile(int hashId, @NotNull FileContent fileContent) {
    super.indexFile(hashId, fileContent);
    myIndexedFileTypes.add(fileContent.getFileType());
  }

  @NotNull
  private <K, V extends PsiElement> HashBasedIndexGenerator<K, Void> createGenerator(StubIndexExtension<K, V> stubIndexExtension) {
    FileBasedIndexExtension<K, Void> extension = StubIndexImpl.wrapStubIndexExtension(stubIndexExtension);
    return new HashBasedIndexGenerator<>(extension, getIndexRoot());
  }

  @NotNull
  public List<HashBasedIndexGenerator<?, ?>> getStubGenerators() {
    return new ArrayList<>(myStubIndexesGeneratorMap.values());
  }

  @SuppressWarnings({"rawtypes", "unchecked"})
  @Override
  protected void visitInputData(int hashId, @NotNull InputData<Integer, SerializedStubTree> data) throws StorageException {
    super.visitInputData(hashId, data);
    SerializedStubTree tree = ContainerUtil.getFirstItem(data.getKeyValues().values());
    if (tree == null) return;
    Map<StubIndexKey<?, ?>, Map<Object, StubIdList>> map = tree.getStubIndicesValueMap();
    for (Map.Entry<StubIndexKey<?, ?>, Map<Object, StubIdList>> entry : map.entrySet()) {
      StubIndexKey key = entry.getKey();
      Map<Object, StubIdList> value = entry.getValue();
      MapReduceIndex index = (MapReduceIndex)myStubIndexesGeneratorMap.get(key).getIndex();
      Map<Object, Object> reducedValue = Maps.asMap(value.keySet(), k -> null);
      UpdateData<?, ?> updateData = new UpdateData(
        hashId,
        reducedValue,
        () -> new EmptyInputDataDiffBuilder(hashId),
        index.getExtension().getName(),
        null
      );
      index.updateWithMap(updateData);
    }
  }

  @Override
  protected @NotNull Map<Integer, SerializedStubTree> transformIndexedData(@NotNull Map<Integer, SerializedStubTree> data,
                                                                           @NotNull FileContent content) {
    if (!myNoStubTreeFileTypes.contains(content.getFileType())) {
      return data;
    }
    if (data.isEmpty()) return data;
    SerializedStubTree tree = ContainerUtil.getFirstItem(data.values());
    if (tree == null) return data;
    tree = tree.withoutStub();
    Integer id = ContainerUtil.getFirstItem(data.keySet());
    return Collections.singletonMap(id, tree);
  }

  @Override
  public void openIndex() throws IOException {
    super.openIndex();
    for (HashBasedIndexGenerator<?, ?> generator : myStubIndexesGeneratorMap.values()) {
      generator.openIndex();
    }
  }

  @Override
  public void closeIndex() throws IOException {
    super.closeIndex();
    for (Map.Entry<StubIndexKey<?, ?>, HashBasedIndexGenerator<?, ?>> entry : myStubIndexesGeneratorMap.entrySet()) {
      entry.getValue().closeIndex();
    }
  }
}
