package com.intellij.indexing.shared.java.hash

import com.google.common.hash.HashCode
import com.google.common.hash.Hashing
import com.intellij.indexing.shared.platform.hash.SharedIndexContentHash
import com.intellij.indexing.shared.platform.hash.SharedIndexContentHashProvider
import com.intellij.openapi.Disposable
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.vfs.*
import com.intellij.openapi.vfs.newvfs.events.VFileEvent
import com.intellij.util.indexing.IndexedFile
import com.intellij.util.indexing.IndexedHashesSupport
import com.intellij.util.io.inputStream
import java.io.IOException
import java.util.concurrent.ConcurrentHashMap

object SharedIndexJarArchiveHash : SharedIndexContentHash {
  override val providerId: String = "jar-archive"
  override val version: String = "1"
  override val hashLength: Int = IndexedHashesSupport.HASH_SIZE_IN_BYTES
  override val performanceRank: Double = 500.0
}

class SharedIndexJarArchiveHashProvider : SharedIndexContentHashProvider {
  private val hasher = service<SharedIndexArchiveFileHashStore>()

  override val info: SharedIndexContentHash = SharedIndexJarArchiveHash

  override fun computeHash(content: IndexedFile): ByteArray? {
    val file = content.file
    val fileSystem = file.fileSystem

    val hash: HashCode
    val relPath: String

    if (fileSystem is JarFileSystem) {
      val archiveRoot = fileSystem.getRootByEntry(file) ?: return null
      relPath = VfsUtil.getRelativeLocation(file, archiveRoot) ?: return null
      val localEntry = fileSystem.getLocalByEntry(file) ?: return null
      if (localEntry.isDirectory) return null

      hash = hasher.getOrComputeHash(localEntry) ?: return null
    }
    else return null

    val input = hash.asBytes() + relPath.toByteArray()
    return IndexedHashesSupport.calculateIndexedHash(content, input)
  }
}

@Service //application
private class SharedIndexArchiveFileHashStore : Disposable {
  private val log = logger<SharedIndexArchiveFileHashStore>()
  private val myComputedHashes = ConcurrentHashMap<VirtualFile, HashCode?>()

  override fun dispose() = Unit

  private fun computeHash(file: VirtualFile): HashCode? {
    if (!file.isValid || file.isDirectory) return null
    if (file !is VirtualFileWithId) return null

    @Suppress("DEPRECATION")
    val hash = Hashing.sha1().newHasher()

    val buff = ByteArray(1024 * 256)

    try {
      // use nio to avoid content caching
      file.toNioPath().inputStream().use { stream ->
        while (true) {
          val sz = stream.read(buff)
          ProgressManager.checkCanceled()
          if (sz <= 0) break
          hash.putBytes(buff, 0, sz)
        }
      }
    }
    catch (e: IOException) {
      log.info(e)
      return null
    }

    return hash.hash()
  }

  fun getOrComputeHash(file: VirtualFile) = myComputedHashes.computeIfAbsent(file) {
    computeHash(file)
  }

  init {
    VirtualFileManager.getInstance().addAsyncFileListener(object : AsyncFileListener {
      override fun prepareChange(events: List<VFileEvent>): AsyncFileListener.ChangeApplier? {
        val allFiles = events.mapNotNull { it.file }
        if (allFiles.isEmpty()) return null
        return object : AsyncFileListener.ChangeApplier {
          override fun beforeVfsChange() {
            allFiles.forEach {
              myComputedHashes.remove(it)
            }
          }
        }
      }
    }, this)
  }
}
