package com.intellij.indexing.shared.java.jdk

import com.intellij.indexing.shared.download.IndexDownloadConsentDecisionKey
import com.intellij.indexing.shared.download.SharedIndexLookupRequest
import com.intellij.openapi.application.runReadAction
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.projectRoots.Sdk
import com.intellij.openapi.projectRoots.impl.jdkDownloader.JdkInstaller
import com.intellij.openapi.util.registry.Registry
import com.intellij.util.lang.JavaVersion
import java.util.*

data class JdkIndexLookupRequest(
  private val sdkName: String? = null,
  val hash: String? = null,
  val aliases: List<String> = listOf(),
  override val authId: String? = null
) : SharedIndexLookupRequest {

  companion object {
    const val KIND = "jdk"

    val baseJdkUrl
      get() = Registry.stringValue("shared.indexes.jdk.download.url").trim('/')

    val decisionKey get() = IndexDownloadConsentDecisionKey(KIND, baseJdkUrl)
  }

  /**
   * JDK indexes lookup may have a number if requests
   * because of aliases, worth to avoid unneeded HTTP requests
   */
  override val useListJsonIndexToSkipUrls: Boolean
    get() = true

  override fun hashCode() = Objects.hash(hash ?: "0", aliases)

  override fun equals(other: Any?): Boolean {
    if (this === other) return true
    if (other !is JdkIndexLookupRequest) return false
    return this.hash == other.hash && this.aliases == other.aliases
  }

  override val presentableChunkName get() = if (sdkName.isNullOrEmpty()) "JDK" else sdkName

  override val kind = KIND

  override fun indexUrls() = sequence {
    if (hash != null) {
      yield("$baseJdkUrl/$hash")
    }

    for (alias in aliases) {
      yield("$baseJdkUrl/-$alias")
    }
  }.toList().distinct()

  override fun toString() = "JdkIndexRequest(sdkName=$sdkName, hash=$hash, aliases=$aliases, kind=$kind)"
}

object JdkAliases {
  private val LOG = logger<JdkAliases>()

  fun computeAliasesForJdk(sdk: Sdk): List<String> {
    val versionString = try {
      runReadAction { sdk.versionString } ?: return listOf()
    }
    catch (e: Throwable) {
      LOG.warn("Failed to compute SDK version for: " + sdk + ". " + e.message, e)
      return listOf()
    }
    return (computeAliasesForInstalledJdk(sdk.homePath) + computeAliasesForJdkVersion(versionString)).distinct()
  }

  fun computeAliasesForJdkVersion(versionString: String): List<String> {
    try {
      val version = JavaVersion.tryParse(versionString) ?: return listOf()
      return listOf(
        version.toFeatureMinorUpdateString(), version.toFeatureString()
      )
    }
    catch (e: Throwable) {
      LOG.warn("Failed to compute aliases for: " + versionString + ". " + e.message, e)
      return listOf()
    }
  }

  private fun computeAliasesForInstalledJdk(jdkHome: String?): List<String> {
    if (jdkHome == null) return listOf()

    try {
      return JdkInstaller.getInstance().findJdkItemForInstalledJdk(jdkHome)?.sharedIndexAliases ?: return listOf()
    }
    catch (e: Throwable) {
      LOG.warn("Failed to get installed JDK aliases for: " + jdkHome + ". " + e.message, e)
      return listOf()
    }
  }
}
