// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.java.jdk;

import com.intellij.execution.wsl.WslDistributionManager;
import com.intellij.openapi.projectRoots.JavaSdk;
import com.intellij.openapi.projectRoots.Sdk;
import com.intellij.openapi.util.Pair;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.io.DigestUtil;
import com.intellij.util.io.PathKt;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.charset.StandardCharsets;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.MessageDigest;
import java.util.stream.Stream;

public final class JdkSupport {
  @NotNull
  public static JavaSdk getSupportedSdkType() {
    return JavaSdk.getInstance();
  }

  public static boolean isJdk(@NotNull Sdk sdk) {
    return JavaSdk.getInstance() == sdk.getSdkType();
  }

  @Nullable
  public static String computeJdkFingerprint(@NotNull Sdk sdk) {
    if (!isJdk(sdk)) {
      return null;
    }

    final byte[] fingerprintVersion = {1, 0, 0, 0};
    MessageDigest hasher = DigestUtil.sha1();
    hasher.update(fingerprintVersion);

    String path = sdk.getHomePath();
    if (path == null) return null;
    Path homePath = Paths.get(path);
    //see IDEA-263337, there is no need to compute that hash anyways
    if (WslDistributionManager.isWslPath(sdk.getHomePath())) return null;
    if (!PathKt.isDirectory(homePath)) return null;

    boolean hasRelease = hashIfExists(hasher, homePath, "release");
    boolean hasVersion = hashIfExists(hasher, homePath, "version.txt");
    boolean hasLib = hashIfExists(hasher, homePath, "lib");
    hashIfExists(hasher, homePath, "jmod");

    if (!hasRelease && !hasVersion && !hasLib) {
      return null;
    }
    return StringUtil.toHexString(hasher.digest());
  }

  private static boolean hashIfExists(@NotNull MessageDigest hasher,
                                      @NotNull Path homePath,
                                      @NotNull String rel) {
    Path relPath = homePath.resolve(rel);
    if (!PathKt.exists(relPath)) return false;

    if (PathKt.isFile(relPath)) {
      hasher.update(rel.getBytes(StandardCharsets.UTF_8));
      DigestUtil.updateContentHash(hasher, relPath);
      return true;
    }

    if (PathKt.isDirectory(relPath)) {
      try(Stream<Path> stream = Files.walk(relPath)) {
        stream
          .filter(file -> PathKt.isFile(file))
          .map(file -> Pair.create(homePath.relativize(file).toString(), file))
          .sorted(Pair.comparingByFirst())
          .forEach(pair -> {
            String relativeName = pair.first;
            Path file = pair.second;
            if (!PathKt.isFile(file)) return;

            if (relativeName.endsWith("modules") ||
                relativeName.endsWith(".jar") ||
                relativeName.endsWith(".zip") ||
                relativeName.endsWith(".jmod")) {
              hasher.update(relativeName.getBytes(StandardCharsets.UTF_8));
              DigestUtil.updateContentHash(hasher, file);
            }
          });
      }
      catch (IOException e) {
        throw new RuntimeException("Failed to walk " + relPath + ". " + e.getMessage(), e);
      }
      return true;
    }

    return false;
  }
}
