package com.intellij.indexing.shared.java.jdk

import com.google.common.hash.Hashing
import com.intellij.openapi.components.*
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.projectRoots.Sdk
import com.intellij.openapi.vfs.VfsUtil
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.VirtualFileVisitor
import com.intellij.util.containers.toMutableSmartList
import com.intellij.util.text.nullize
import com.intellij.util.xmlb.annotations.OptionTag
import com.intellij.util.xmlb.annotations.Tag
import com.intellij.util.xmlb.annotations.XCollection
import org.jetbrains.annotations.TestOnly
import java.util.*
import java.util.concurrent.atomic.AtomicInteger
import java.util.concurrent.locks.ReentrantLock
import kotlin.concurrent.withLock

@Tag("entry")
class SharedJdkChunkStoreEntry() : BaseState() {
  @get:OptionTag
  var name: String? by string()

  @get:OptionTag
  var hash: String? by string()

  @get:OptionTag
  var aliases by list<String>()

  @get:OptionTag
  var snapshotHash: String? by string()

  constructor(sdkName: String, snapshotHash: String, state: JdkIndexLookupRequest?) : this() {
    this.name = sdkName
    this.hash = state?.hash
    this.aliases = (state?.aliases ?: listOf()).toMutableSmartList()
    this.snapshotHash = snapshotHash
  }

  fun toJdkIndexRequest(): JdkIndexLookupRequest? {
    val hash = hash.nullize()
    if (hash == null && aliases.isEmpty()) return null
    return JdkIndexLookupRequest(if (name.isNullOrEmpty()) null else name, hash, aliases)
  }
}

class SharedJdkChunkStoreState : BaseState() {
  @get:XCollection
  val entries by list<SharedJdkChunkStoreEntry>()
}

@State(name = "shared-index-jdk-hash", storages = [Storage(value = StoragePathMacros.CACHE_FILE)], allowLoadInTests = true)
class SharedJdkChunkStore : SimplePersistentStateComponent<SharedJdkChunkStoreState>(SharedJdkChunkStoreState()) {
  private val lock = ReentrantLock()

  override fun loadState(state: SharedJdkChunkStoreState) = lock.withLock {
    super.loadState(state)
  }

  fun registerDownloadingJdkRequest(lookupName: String, home: VirtualFile, newState: JdkIndexLookupRequest) {
    val snapshotHash = snapshot(home) ?: return
    lock.withLock {
      state.entries.removeAll { it.snapshotHash == snapshotHash }
      state.entries.add(SharedJdkChunkStoreEntry(lookupName, snapshotHash, newState))
      state.intIncrementModificationCount()
    }
  }

  fun getOrComputeJdkRequest(sdk: Sdk): JdkIndexLookupRequest? {
    if (!JdkSupport.isJdk(sdk)) return null
    val snapshotHash = sdk.homeDirectory?.let { snapshot(it) } ?: return null

    lock.withLock {
      state.entries.find { it.snapshotHash == snapshotHash }?.let {
        return it.toJdkIndexRequest()
      }

      val newState = doComputeJdkRequest(sdk)
      state.entries.add(SharedJdkChunkStoreEntry(sdk.name, snapshotHash, newState))
      state.intIncrementModificationCount()
      return newState
    }
  }

  private fun doComputeJdkRequest(sdk: Sdk): JdkIndexLookupRequest? {
    if (!JdkSupport.isJdk(sdk)) return null

    val jdkFingerprint = JdkSupport.computeJdkFingerprint(sdk)
    val aliases = doComputeAliases(sdk)

    return JdkIndexLookupRequest(if (sdk.name.isEmpty()) null else sdk.name, jdkFingerprint, aliases)
  }

  @TestOnly
  val doComputeCounter = AtomicInteger()

  private fun doComputeAliases(sdk: Sdk): List<String> {
    doComputeCounter.incrementAndGet()
    return JdkAliases.computeAliasesForJdk(sdk)
  }

  private data class FileInfo(val size: Long, val timestamp: Long)

  private fun snapshot(home: VirtualFile): String? {
    //we need to ensure files are sorted
    val allFiles = TreeMap<String, FileInfo>()
    VfsUtil.visitChildrenRecursively(home, object : VirtualFileVisitor<Unit>() {
      override fun visitFile(file: VirtualFile): Boolean {
        if (!file.isDirectory) {
          allFiles[file.path] = FileInfo(file.length, file.timeStamp)
        }
        ProgressManager.checkCanceled()
        return super.visitFile(file)
      }
    })

    val hasher = Hashing.sha256().newHasher()
    hasher.putUnencodedChars(JdkSupport.getSupportedSdkType().name)

    for ((path, info) in allFiles) {
      hasher.putUnencodedChars(path)
      hasher.putLong(info.size)
      hasher.putLong(info.timestamp)
    }

    return hasher.hash().toString()
  }
}
