// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.java.jdk

import com.intellij.execution.*
import com.intellij.indexing.shared.download.*
import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.indexing.shared.util.sha256
import com.intellij.openapi.Disposable
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.application.runReadAction
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.module.ModuleManager
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.project.Project
import com.intellij.openapi.projectRoots.ProjectJdkTable
import com.intellij.openapi.projectRoots.Sdk
import com.intellij.openapi.projectRoots.impl.jdkDownloader.JdkInstallRequest
import com.intellij.openapi.projectRoots.impl.jdkDownloader.JdkInstallerListener
import com.intellij.openapi.roots.JdkOrderEntry
import com.intellij.openapi.roots.ModuleRootManager
import com.intellij.openapi.startup.StartupActivity
import com.intellij.openapi.util.io.FileUtil
import com.intellij.openapi.util.registry.Registry
import com.intellij.openapi.vfs.LocalFileSystem
import com.intellij.openapi.vfs.VfsUtil
import com.intellij.ui.layout.*
import com.intellij.util.concurrency.AppExecutorUtil

private fun isEnabled() = !ApplicationManager.getApplication().isUnitTestMode &&
                          Registry.`is`("shared.indexes.download") &&
                          Registry.`is`("shared.indexes.jdk.download")

interface SharedJdkIndexSuggestion : SharedIndexSuggestion {
  val sdkName : String

  override val sharedIndexDownloadedPresentableText: String
    get() = SharedIndexesBundle.message("eventLog.downloaded.jdk", sdkName)

  override val sharedIndexConsentNotificationKindPresentableText: String
    get() = SharedIndexesBundle.message("notification.consent.content.kind.jdk")

}

class SharedJdkIndexPreloader : JdkInstallerListener {
  private fun JdkInstallRequest.toSuggestion(): Pair<JdkIndexLookupRequest, SharedJdkIndexSuggestion>? {
    val request = this
    val id = request.toSharedIndexId() ?: return null
    val jdk = request.item

    val jdkRequest = JdkIndexLookupRequest(
      sdkName = jdk.suggestedSdkName,
      aliases = (jdk.sharedIndexAliases + JdkAliases.computeAliasesForJdkVersion(jdk.versionString)).distinct()
    )

    return jdkRequest to object : SharedJdkIndexSuggestion {
      override val sdkName: String = jdk.suggestedSdkName
      override val sharedIndexId: SharedIndexId = id
      override fun resolveRequest(indicator: ProgressIndicator) = jdkRequest
      override fun toString() = "JdkSharedIndexSuggestion($id, $jdkRequest)"
    }
  }

  override fun onJdkDownloadStarted(request: JdkInstallRequest, project: Project?) {
    if (!isEnabled()) return
    val (_, suggestion) = request.toSuggestion() ?: return
    service<SharedIndexDownloadService>().applySuggestions(project, listOf(suggestion))
  }

  override fun onJdkDownloadFinished(request: JdkInstallRequest, project: Project?) {
    if (!isEnabled()) return

    AppExecutorUtil.getAppExecutorService().submit {
      // we assume here that the call to `request.toSuggestion()` generates the same result
      // here and above.
      // Once SDK is installed, we'd try to seek for a shared index for it once again,
      // that is fine, but we need to ensure the same index would be used. So we
      // pre-register it to the cache with the same lookup request as it was here
      val (jdkLookupRequest, _) = request.toSuggestion() ?: return@submit
      val home = request.javaHome
      val virtualHome = LocalFileSystem.getInstance().refreshAndFindFileByNioFile(home) ?: return@submit
      //this may need modality state, it must not be executed in the listener directly
      VfsUtil.markDirtyAndRefresh(false, true, true, virtualHome)
      service<SharedJdkChunkStore>().registerDownloadingJdkRequest(request.item.suggestedSdkName, virtualHome, jdkLookupRequest)
    }
  }
}

private class RunConfigurationsUnknownJdkContributorRefresher : StartupActivity.DumbAware {
  override fun runActivity(project: Project) {
    if (!isEnabled()) return
    project.service<RunConfigurationsUnknownJdkContributorRefresherService>()
  }
}

@Service(Service.Level.PROJECT)
private class RunConfigurationsUnknownJdkContributorRefresherService(project: Project) : Disposable {
  override fun dispose() = Unit

  init {
    project.messageBus.connect(this).subscribe(RunManagerListener.TOPIC, object : RunManagerListener {
      override fun runConfigurationAdded(settings: RunnerAndConfigurationSettings) {
        scheduleReload(settings)
      }

      override fun runConfigurationChanged(settings: RunnerAndConfigurationSettings) {
        scheduleReload(settings)
      }

      private fun scheduleReload(settings: RunnerAndConfigurationSettings) {
        if (!AlternativeSdkRootsProvider.shouldIndexAlternativeJre()) return
        if (!AlternativeSdkRootsProvider.hasEnabledAlternativeJre(settings)) return

        service<SharedIndexDownloadService>().rescanNow(project)
      }
    })
  }
}

internal class JdkSharedIndexSuggesterConfigurable : SharedIndexSuggesterConfigurable {
  override val mode = SharedIndexConfigurableMode.PreBuilt
  override fun isAvailableFor(project: Project) = true
  override fun buildSettingsPanel(project: Project, scope: SharedIndexesConfigurableScope, builder: RowBuilder) {
    scope.run {
      builder.row(SharedIndexesBundle.message("configuration.section.jdk")) {
        consentComboBox(JdkIndexLookupRequest.decisionKey, includeUrl = false)
      }
    }
  }
}

internal class JdkSharedIndexSuggester : SharedIndexSuggester {
  private fun listJdks(project: Project) = runReadAction {
    ModuleManager
      .getInstance(project)
      .modules
      .flatMap { ModuleRootManager.getInstance(it).orderEntries.asIterable() }
      .asSequence()
      .filterIsInstance<JdkOrderEntry>()
      .mapNotNull { it.jdk }
      .distinct()
      .filter { JdkSupport.isJdk(it) }
      .toList() + AlternativeSdkRootsProvider.getAdditionalProjectJdksToIndex(project)
  }

  override fun suggestRequests(project: Project): List<SharedIndexSuggestion> {
    if (!isEnabled()) return listOf()

    val jdks = listJdks(project)

    val result = mutableListOf<SharedIndexSuggestion>()
    for (jdk in jdks) {
      val id = jdk.toSharedIndexId() ?: continue
      val jdkName = jdk.name
      val item = newSuggestion(jdkName, id)
      result += item
    }

    return result
  }
}

private fun newSuggestion(jdkName: String, id: SharedIndexId) = object : SharedJdkIndexSuggestion {
  override val sdkName: String = jdkName
  override val sharedIndexId: SharedIndexId = id

  override fun resolveRequest(indicator: ProgressIndicator): JdkIndexLookupRequest? {
    val jdk = ProjectJdkTable.getInstance().findJdk(jdkName) ?: return null
    return service<SharedJdkChunkStore>().getOrComputeJdkRequest(jdk)
  }

  override fun toString() = "JdkSharedIndexSuggestion($id, lazy)"
}

private fun JdkInstallRequest.toSharedIndexId(): SharedIndexId? =
  runCatching { toSharedIndexId(installDir.toString(), item.versionString) }.getOrNull()

private fun Sdk.toSharedIndexId(): SharedIndexId? =
  runCatching { toSharedIndexId(homePath, runCatching { versionString }.getOrNull()) }.getOrNull()

private fun toSharedIndexId(homePath: String?, versionString: String?): SharedIndexId {
  val indexId = (homePath?.let { FileUtil.toSystemIndependentName(it) } ?: "no home path").sha256() + "#" + (versionString ?: "no version")
  return SharedIndexId(JdkIndexLookupRequest.KIND, JdkIndexLookupRequest.baseJdkUrl, indexId)
}
