package com.intellij.indexing.shared.java.maven

object MavenJarNamingHelper {
  fun inferPackageCoordinates(path: String) : MavenPackageId? {
    val reverseElements = path.split("/").reversed()

    // we support the following scenarios:
    //  - a local maven repository ::  .m2/]repository/gr/out/pId/artifact-id/version/artifact-id-version.jar
    //  - a gradle cache           ::  .gradle/]caches/modules-2/files-2.1/gr.out.pId/artifact-id/version/hash/artifact-id-version.jar
    //

    //aka base/[groupId]/[artifactId]/[version]/file.jar
    if (reverseElements.size <= 5) return null

    //Gradle IJ SDK layout
    if (path.contains("/caches/modules-2/files-2.1/")) {
      fun testIJSdk() : MavenPackageId? {
        // - a IJ SDK layout  :: ".gradle/caches/modules-2/files-2.1/com.jetbrains.intellij.idea/ideaIU/2020.2.3/<hash>/ideaIU-2020.2.3/lib/FastInfoset-1.2.15.jar"
        val fileName = reverseElements.getOrNull(0) ?: return null
        val libName = reverseElements.getOrNull(1) ?: return null
        val folderName = reverseElements.getOrNull(2) ?: return null
        val version = reverseElements.getOrNull(4) ?: return null
        val artifactId = reverseElements.getOrNull(5) ?: return null
        val groupId = reverseElements.getOrNull(6) ?: return null

        if (!fileName.endsWith(".jar")) return null
        if (libName != "lib") return null
        if (folderName != "$artifactId-$version") return null

        return MavenPackageId(groupId, artifactId, version)
      }

      val result = testIJSdk()
      if (result != null) return result
    }

    //Gradle layout
    if (path.contains("/caches/modules-2/files-2.1/")) {
      //  - a gradle cache           ::  .gradle/]caches/modules-2/files-2.1/gr.out.pId/artifact-id/version/<hash>/artifact-id-version.jar
      val fileName = reverseElements.firstOrNull() ?: return null
      val version = reverseElements.getOrNull(2) ?: return null
      val artifactId = reverseElements.getOrNull(3) ?: return null

      val expectedFileName = "$artifactId-$version.jar"

      //a check to make sure we indeed have Gradle layout
      if (expectedFileName != fileName) return null
      val groupId = reverseElements.drop(4).takeWhile { it != "files-2.1" }.singleOrNull() ?: return null

      return MavenPackageId(groupId, artifactId, version)
    }


    //Maven layout
    if (path.contains("/repository/")) {
      //  - a local maven repository ::  .m2/]repository/gr/out/pId/artifact-id/version/artifact-id-version.jar
      val fileName = reverseElements.firstOrNull() ?: return null
      val version = reverseElements.getOrNull(1) ?: return null
      val artifactId = reverseElements.getOrNull(2) ?: return null

      val expectedFileName = "$artifactId-$version.jar"
      //a check to make sure we indeed have Gradle layout
      if (expectedFileName != fileName) return null

      val groupIdElements = reverseElements.drop(3).takeWhile { it != "repository" }.reversed()
      if (groupIdElements.isEmpty()) return null
      val groupId = groupIdElements.joinToString(".")

      return MavenPackageId(groupId, artifactId, version)
    }

    return null
  }
}
