// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.metadata

import com.fasterxml.jackson.databind.JsonNode
import com.fasterxml.jackson.databind.ObjectMapper
import com.fasterxml.jackson.databind.node.ArrayNode
import com.fasterxml.jackson.databind.node.ObjectNode
import com.google.common.collect.ImmutableSortedMap
import com.intellij.indexing.shared.download.SharedIndexBaseId
import com.intellij.indexing.shared.download.SharedIndexLookupResult
import com.intellij.indexing.shared.platform.api.*
import com.intellij.openapi.application.ApplicationInfo
import com.intellij.openapi.util.SystemInfo
import com.intellij.util.text.nullize
import java.io.IOException
import java.nio.file.Files
import java.nio.file.Path
import java.time.ZonedDateTime
import java.time.format.DateTimeFormatter
import java.util.*

data class SharedIndexMetadataInfo(
  val jdkAliases: List<String> = listOf(),
  val jdkHash: String? = null,
  val projectId: String? = null,
  val projectVcsCommitId: String? = null,
  val mavenPackage: String? = null,
  val mavenIncludedPackages: Set<String> = setOf(),
  val bases: Set<SharedIndexBaseId> = setOf(),
)

class SharedIndexMetadata @JvmOverloads constructor(
  val indexName: String,
  val indexKind: String,
  val sharedIndexInfo: SharedIndexMetadataInfo,
  val indexInfrastructureVersion: IndexInfrastructureVersion,
  val generationTime: ZonedDateTime = ZonedDateTime.now()
) {

  override fun equals(other: Any?): Boolean {
    if (this === other) return true
    if (other !is SharedIndexMetadata) return false
    if (indexName != other.indexName) return false
    if (indexKind != other.indexKind) return false
    if (sharedIndexInfo != other.sharedIndexInfo) return false
    if (indexInfrastructureVersion != other.indexInfrastructureVersion) return false
    return true
  }

  override fun hashCode(): Int {
    var result = indexName.hashCode()
    result = 31 * result + indexKind.hashCode()
    result = 31 * result + sharedIndexInfo.hashCode()
    result = 31 * result + indexInfrastructureVersion.hashCode()
    return result
  }

  override fun toString(): String = "SharedIndexMetadata($indexName, $indexKind', indexInfrastructureVersion, $sharedIndexInfo)"

  companion object {
  private const val METADATA_VERSION = "4"
  private val METADATA_DATE_FORMAT get() = DateTimeFormatter.ISO_DATE_TIME.withLocale(Locale.US)

  fun tryReadInfrastructureVersionOnly(node: ObjectNode) : IndexInfrastructureVersion? {
    val indexesNode = node["indexes"] as? ObjectNode ?: return null
    val caps = mutableMapOf<IndexInfrastructureCapability, Map<String, String>>()
    (indexesNode.get("capabilities") as? ObjectNode)?.let { capabilitiesNode ->
      for (capName in capabilitiesNode.fieldNames()) {
        val cap = IndexInfrastructureCapability.find(capName) ?: continue
        val data = (capabilitiesNode[capName] as? ObjectNode)?.toMap() ?: continue
        caps[cap] = data
      }
    }

    val baseVersions = indexesNode.mapFromPath("base_versions") ?: return null
    val fileIndexVersions = indexesNode.mapFromPath(FiledBasedIndexInfoValue,"file_index_versions") ?: return null
    val stubIndexVersions = indexesNode.mapFromPath("stub_index_versions") ?: return null
    val fileElementTypeVersions = indexesNode.mapFromPath("stub_file_element_type_versions") ?: return null
    val hashProviderVersions = indexesNode.mapFromPath("hash_provider_versions") ?: return null

    val osNames = (indexesNode["os"] as? ArrayNode)
                    ?.mapNotNull { it.asText() }
                    ?.toSortedSet()
                    ?.mapNotNull(IndexInfrastructureOs::find)?.toSet() ?: setOf()

    return IndexInfrastructureVersion(
      caps,
      osNames,
      baseVersions,
      fileIndexVersions,
      stubIndexVersions,
      fileElementTypeVersions,
      hashProviderVersions,
    )
  }

  fun tryReadIndexesVersion(node: ObjectNode): SharedIndexMetadata? {
    if (node["metadata_version"]?.asText() != METADATA_VERSION) return null

    val sources = node["sources"] as? ObjectNode ?: return null
    val indexKind = sources.get("kind")?.asText()?: return null
    val indexName = sources.get("name")?.asText()?: return null
    val generated = runCatching {
      ZonedDateTime.parse(sources.get("generated")?.asText() ?: return null, METADATA_DATE_FORMAT)
    }.getOrElse { return null }

    val bases = (node["bases"] as? ArrayNode)
                  ?.filterIsInstance<ObjectNode>()
                  ?.mapNotNull { it["sha256"]?.asText().nullize(true) }
                  ?.distinct()
                  ?.map(::SharedIndexBaseId)
                  ?.toSet()
                  ?: setOf()

    val metadataInfo = SharedIndexMetadataInfo(
      jdkHash = sources.get("jdk_hash")?.asText()?.nullize(),
      jdkAliases = (sources.get("jdk_aliases") as? ArrayNode)?.mapNotNull { it?.asText()?.nullize() }?.distinct() ?: listOf(),

      projectId = sources.get("project_id")?.asText()?.nullize(),
      projectVcsCommitId = sources.get("project_vcs_commit_id")?.asText()?.nullize(),

      mavenPackage = sources.get("maven_package")?.asText()?.nullize(),
      mavenIncludedPackages = (sources.get("maven_included_packages") as? ArrayNode)?.mapNotNull { it?.asText()?.nullize() }?.toSet() ?: setOf(),

      bases = bases
   )

    val indexVersions = tryReadInfrastructureVersionOnly(node) ?: return null
    return SharedIndexMetadata(indexName, indexKind, metadataInfo, indexVersions, generated)
  }

  fun selectBestSuitableIndex(
    ideVersion: IdeIndexInfrastructureVersion,
    candidates: List<SharedIndexLookupResult>
  ): SharedIndexLookupResult? {
    val nodes = candidates.groupBy { it.indexVersion }
    val best = ideVersion.pickBestSuitableVersion(nodes.keys) ?: return null
    return nodes[best]?.firstOrNull()
  }

  private inline fun renderJson(builder: (ObjectNode) -> Unit): ByteArray {
    try {
      val om = ObjectMapper()

      val root = om.createObjectNode()
      root.put("metadata_version", METADATA_VERSION)

      builder(root)

      return om.writerWithDefaultPrettyPrinter().writeValueAsBytes(root)
      //NOTE: should we include information about index sizes here too?
    }
    catch (t: Throwable) {
      throw RuntimeException("Failed to generate shared index metadata JSON. ${t.message}", t)
    }
  }

  @JvmStatic
  fun getMetadataPath(chunkRootPath: Path): Path {
    return chunkRootPath.resolve(getMetadataPath())
  }

  @JvmStatic
  fun getMetadataPath() = "metadata.json"

  @JvmStatic
  fun writeIndexMetadata(metadata: SharedIndexMetadata): ByteArray = renderJson { root ->
    val addon = metadata.sharedIndexInfo

    root.putObject("sources").also { sources ->
      sources.put("generated", metadata.generationTime.format(METADATA_DATE_FORMAT))
      sources.put("os", IndexInfrastructureOs.getOs().osName)
      sources.put("os_name", SystemInfo.getOsNameAndVersion())
      sources.put("kind", metadata.indexKind)
      sources.put("name", metadata.indexName)

      addon.jdkHash?.let {
        sources.put("jdk_hash", it)
      }

      addon.jdkAliases.ifEmpty { null }?.let { aliases ->
        sources.putArray("jdk_aliases").let { aliasesNode ->
          aliases.map { it.toLowerCase().trim() }.toSortedSet().forEach { aliasesNode.add(it) }
        }
      }

      addon.projectId?.let {
        sources.put("project_id", it)
      }

      addon.projectVcsCommitId?.let {
        sources.put("project_vcs_commit_id", it)
      }

      addon.mavenPackage?.let {
        sources.put("maven_package", it)
      }

      if (addon.mavenIncludedPackages.isNotEmpty()) {
        sources.putArray("maven_included_packages").let {
          addon.mavenIncludedPackages.toSortedSet().forEach { pkg -> it.add(pkg) }
        }
      }
    }

    addon.bases.ifEmpty { null }?.let { bases ->
      root.putArray("bases").let {
        for (baseId in bases.distinct()) {
          it.addObject().put("sha256", baseId.sha256)
        }
      }
    }

    root.putObject("environment").also { build ->
      build.put("os", IndexInfrastructureOs.getOs().osName)
      build.put("os_name", SystemInfo.getOsNameAndVersion())
      build.putObject("intellij").also { ij ->
        ij.put("product_code", ApplicationInfo.getInstance().build.productCode)
        ij.put("version", ApplicationInfo.getInstance().fullVersion)
        ij.put("build", ApplicationInfo.getInstance().build.toString())
      }
    }

    putIndexesVersion(root, metadata.indexInfrastructureVersion)
  }

  /**
   * @return `null` if the metadata format of the chunk is incompatible with current IDE.
   * @throws IOException if the metadata file does not exist or it cannot be read.
   */
  @JvmStatic
  @Throws(IOException::class)
  fun readIndexesVersion(chunkRoot: Path): SharedIndexMetadata? {
    val metadataPath = getMetadataPath(chunkRoot)
    val metadataBytes = Files.readAllBytes(metadataPath)
    return tryReadIndexesVersion(metadataBytes)
  }

  fun tryReadIndexesVersion(metadata: ByteArray?): SharedIndexMetadata? {
    if (metadata == null) return null
    val om = ObjectMapper()
    val root = om.readTree(metadata) as? ObjectNode ?: return null
    return tryReadIndexesVersion(root)
  }

  fun writeVersionInfo(infrastructureVersion: IndexInfrastructureVersion) = renderJson { root ->
    putIndexesVersion(root, infrastructureVersion)
  }

  private fun putIndexesVersion(
    root: ObjectNode,
    infrastructureVersion: IndexInfrastructureVersion
  ) {
    root.putObject("indexes").also { indexes ->
      indexes.put("weak_hash", infrastructureVersion.weakVersionHash)
      indexes.put("weak_base_hash", infrastructureVersion.weakBaseVersionHash)

      if (infrastructureVersion.capabilities.isNotEmpty()) {
        indexes.putObject("capabilities").also { caps ->
          for ((cap, versions) in infrastructureVersion.capabilities) {
            caps.putObjectFromMap(cap.capabilityName, versions)
          }
        }
      }

      indexes.putArray("os").let { os ->
        infrastructureVersion.supportedOs.map { it.osName }.toSortedSet().forEach { os.add(it) }
      }

      indexes.putObjectFromMap("base_versions", infrastructureVersion.baseIndexes)
      indexes.putObjectFromMap(FiledBasedIndexInfoValue, "file_index_versions", infrastructureVersion.fileBasedIndexVersionInfos)
      indexes.putObjectFromMap("stub_index_versions", infrastructureVersion.stubIndexVersions)
      indexes.putObjectFromMap("stub_file_element_type_versions", infrastructureVersion.stubFileElementTypeVersions)
      indexes.putObjectFromMap("hash_provider_versions", infrastructureVersion.hashProviderVersions)
    }
  }

  @JvmName("putObjectFromMap")
  private fun ObjectNode.putObjectFromMap(name: String, map: Map<String, String>) = putObjectFromMap(TextValue, name, map)

  private fun <Y: Any> ObjectNode.putObjectFromMap(saveValue: LoadSaveValue<Y>, name: String, map: Map<String, Y>) {
    putObject(name).also { obj ->
      map.entries.sortedBy { it.key.toLowerCase() }.forEach { (k, v) -> saveValue.saveValue(obj, k, v) }
    }
  }

  private fun ObjectNode.mapFromPath(vararg path: String): SortedMap<String, String>? = mapFromPath(TextValue, *path)
  private fun ObjectNode.toMap() = toMap(TextValue)


  private fun <Y : Any> ObjectNode.mapFromPath(loadValue: LoadSaveValue<Y>, vararg path: String): SortedMap<String, Y>? {
    if (path.isEmpty()) return toMap(loadValue)
    val first = path.first()
    val child = (get(first) as? ObjectNode) ?: return null
    return child.mapFromPath(loadValue, *path.drop(1).toTypedArray())
  }

  private fun <Y : Any> ObjectNode.toMap(loadValue: LoadSaveValue<Y>): SortedMap<String, Y>? {
    val result = HashMap<String, Y>()
    for (key in this.fieldNames()) {
      val value = loadValue.loadValue(get(key)) ?: return null
      result[key] = value
    }
    return ImmutableSortedMap.copyOf(result)
  }

  private interface LoadSaveValue<Y : Any> {
    fun loadValue(node: JsonNode?) : Y?
    fun saveValue(node: ObjectNode, key: String, value: Y)
  }

  private val TextValue = object : LoadSaveValue<String> {
    override fun loadValue(node: JsonNode?): String? = node?.asText()
    override fun saveValue(node: ObjectNode, key: String, value: String) {
      node.put(key, value)
    }
  }

  private val BooleanValue = object : LoadSaveValue<Boolean> {
    override fun loadValue(node: JsonNode?): Boolean? = node?.asBoolean()
    override fun saveValue(node: ObjectNode, key: String, value: Boolean) {
      node.put(key, value)
    }
  }

  private val FiledBasedIndexInfoValue = object : LoadSaveValue<IndexInfrastructureFileBasedIndexVersionInfo> {
    override fun loadValue(node: JsonNode?): IndexInfrastructureFileBasedIndexVersionInfo? {
      if (node !is ObjectNode) return null
      val version = TextValue.loadValue(node["version"]) ?: return null
      val needsForwardIndex = BooleanValue.loadValue(node["needs_forward_index"]) ?: return null
      return IndexInfrastructureFileBasedIndexVersionInfo(version, needsForwardIndex)
    }

    override fun saveValue(node: ObjectNode, key: String, value: IndexInfrastructureFileBasedIndexVersionInfo) {
      node.putObject(key).apply {
        TextValue.saveValue(this, "version", value.version)
        BooleanValue.saveValue(this, "needs_forward_index", value.needsForwardIndex)
      }
    }
  }

}
}
