package com.intellij.indexing.shared.performanceTesting

import com.fasterxml.jackson.annotation.JsonAutoDetect
import com.fasterxml.jackson.annotation.PropertyAccessor
import com.fasterxml.jackson.databind.ObjectMapper
import com.intellij.indexing.shared.platform.impl.FileContentHashIndex
import com.intellij.indexing.shared.platform.impl.SharedIndexChunkConfiguration
import com.intellij.openapi.application.runReadAction
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.Task
import com.intellij.openapi.project.DumbService
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.playback.PlaybackContext
import com.intellij.openapi.ui.playback.commands.AbstractCommand
import com.intellij.openapi.util.Comparing
import com.intellij.openapi.util.io.FileUtil
import com.intellij.openapi.util.text.StringUtil
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.newvfs.ManagingFS
import com.intellij.psi.search.GlobalSearchScope
import com.intellij.util.Processors
import com.intellij.util.indexing.*
import com.intellij.util.indexing.roots.IndexableFilesIterator
import com.jetbrains.performancePlugin.PerformanceTestingBundle
import com.jetbrains.performancePlugin.utils.ActionCallbackProfilerStopper
import com.jetbrains.performancePlugin.utils.errors.ErrorCollector
import com.jetbrains.performancePlugin.utils.errors.ToDirectoryWritingErrorCollector
import com.jetbrains.performancePlugin.utils.indexes.CurrentIndexedFileResolver.getAllToBeIndexedFilesInProject
import org.jetbrains.concurrency.Promise
import org.jetbrains.concurrency.toPromise
import java.io.File
import java.io.IOException
import java.nio.file.Path
import java.nio.file.Paths
import java.util.*
import java.util.stream.Collectors

class CheckSharedIndices(text: String?, line: Int) : AbstractCommand(text, line) {
  override fun _execute(context: PlaybackContext): Promise<Any?> {
    val actionCallback = ActionCallbackProfilerStopper()
    val project = context.project
    DumbService.getInstance(project).smartInvokeLater {
      object : Task.Modal(project, PerformanceTestingBundle.message("checking.shared.indexes"), false) {
        override fun run(indicator: ProgressIndicator) {
          val failureDiagnosticDirectory = failureDiagnosticDirectory
          try {
            val errorCollector: ErrorCollector = ToDirectoryWritingErrorCollector(
              "errors-of-check-shared-indices",
              failureDiagnosticDirectory,
              limitOfErrors = 100
            )
            checkFilesAvailableInSharedIndex(indicator, errorCollector, project)
            if (errorCollector.numberOfErrors > 0) {
              throw AssertionError("Some errors have been collected")
            }
            actionCallback.setDone()
          }
          catch (e: Throwable) {
            LOG.error("Check shared indices command has failed. See failure details in $failureDiagnosticDirectory", e)
            actionCallback.reject(e.message)
          }
        }
      }.queue()
    }
    return actionCallback.toPromise()
  }

  private fun checkFilesAvailableInSharedIndex(indicator: ProgressIndicator, errorCollector: ErrorCollector, project: Project) {
    check(SharedIndexChunkConfiguration.getInstance().hasAnyChunk()) {
      "No shared indexes chunks have been attached to the project"
    }
    (FileBasedIndex.getInstance() as FileBasedIndexImpl).flushIndexes()
    runReadAction {
      val allHashIds = collectAllSharedIndexHashIds(project, indicator)
      LOG.info(
        "Shared hash index contains the following number of hashes per indexId " +
        allHashIds.entries.joinToString("") { "\n    (${it.key.name}: ${it.value.size})" }
      )
      val fileIds = mapHashIdsToLocalFileIds(allHashIds, errorCollector)
      val sharedIndexFiles = collectSharedIndexFiles(fileIds, indicator, project)
      val sharedIndexFilesNumber = sharedIndexFiles.size
      LOG.info("Project files indexed by shared indexes: $sharedIndexFilesNumber")

      val allFiles = getAllToBeIndexedFilesInProject(project, indicator).mapValues { it.value.toMutableSet() }
      val numberOfProjectFiles = allFiles.values.map { set -> set.count { !it.isDirectory } }.sum()
      LOG.info("Total number of project files: $numberOfProjectFiles")
      filterOnlyIndexedFilesInPlace(allFiles)

      val numberOfIndexedFiles = allFiles.values.map { it.size }.sum()
      LOG.info("Total number of indexed project files: $numberOfIndexedFiles")

      val logInfo = SharedIndexLogInfo(sharedIndexFilesNumber, numberOfProjectFiles, numberOfIndexedFiles)
      saveLogInfo(logInfo)

      if (sharedIndexFiles.isEmpty()) {
        throw AssertionError("Shared indexes don't contain any files")
      }
      else {
        val message = getMessageForFilesNotAvailableInSharedIndex(sharedIndexFiles, allFiles)
        if (message.isEmpty()) {
          LOG.info("All project files are available in the shared index.")
        }
        else {
          LOG.info("Files without shared index: ")
          LOG.info(message)
        }
      }
    }
  }

  private fun filterOnlyIndexedFilesInPlace(allFiles: Map<IndexableFilesIterator, MutableSet<VirtualFile>>) {
    val contentDependentIndexes = contentDependentIndexes
    for ((_, value) in allFiles) {
      value.removeIf { fileOrDir: VirtualFile ->
        if (fileOrDir.isDirectory) {
          return@removeIf true
        }
        val fileId = FileBasedIndex.getFileId(fileOrDir)

        val nonTrivialIds = IndexingStamp.getNontrivialFileIndexedStates(fileId)
        val hasBeenIndexed = nonTrivialIds.any { contentDependentIndexes.contains(it) }
        !hasBeenIndexed
      }
    }
  }

  private data class SharedIndexLogInfo(
    val filesIndexedBySharedIndex: Int,
    val allProjectFiles: Int,
    val allIndexedFiles: Int
  )

  companion object {

    const val PREFIX = CMD_PREFIX + "checkSharedIndices"

    private val LOG = Logger.getInstance(CheckSharedIndices::class.java)

    private const val NUMBER_OF_ERRORS_LIMIT = 100

    private val failureDiagnosticDirectory: Path
      get() {
        val property = System.getProperty("check.shared.indices.command.failure.diagnostic.directory")
        return if (property != null) {
          Paths.get(property)
        }
        else try {
          FileUtil.createTempDirectory("compare-indices", "failure").toPath()
        }
        catch (e: IOException) {
          throw RuntimeException(e)
        }
      }

    private fun stopIfTooManyErrors(errorCollector: ErrorCollector) {
      if (errorCollector.numberOfErrors > NUMBER_OF_ERRORS_LIMIT) {
        throw AssertionError("Too many errors have been collected")
      }
    }

    private fun saveLogInfo(logInfo: SharedIndexLogInfo) {
      val sharedIndexInfoJson = System.getProperty("json.shared.index.info") ?: return
      val objectMapper = ObjectMapper()
      try {
        objectMapper.setVisibility(PropertyAccessor.FIELD, JsonAutoDetect.Visibility.ANY)
        objectMapper.writeValue(File(sharedIndexInfoJson), logInfo)
      }
      catch (ex: IOException) {
        LOG.error(ex.message)
      }
    }

    private fun getMessageForFilesNotAvailableInSharedIndex(
      sharedIndexFiles: Set<VirtualFile>,
      projectFiles: Map<IndexableFilesIterator, Set<VirtualFile>>
    ): String {
      val message = StringBuilder()
      projectFiles.forEach { (provider, providerFiles) ->
        val nonSharedFiles = providerFiles - sharedIndexFiles
        if (nonSharedFiles.isNotEmpty()) {
          val limit = 5
          message.append(provider.debugName)
            .append(" has ")
            .append(nonSharedFiles.size)
            .append(" ")
            .append(StringUtil.pluralize("file", nonSharedFiles.size))
            .append(" not available in shared index (only ").append(limit).append(" first are shown): ")
            .append("\n")
          nonSharedFiles.stream().limit(limit.toLong())
            .map { file: VirtualFile -> file.path }
            .forEach { path: String? -> message.append("    ").append(path).append("\n") }
        }
      }
      return message.toString()
    }

    private fun collectSharedIndexFiles(
      fileIds: Set<Int>,
      indicator: ProgressIndicator,
      project: Project
    ): Set<VirtualFile> {
      indicator.isIndeterminate = false
      indicator.fraction = 0.0
      var finished = 0
      val files: MutableSet<VirtualFile> = HashSet()
      val allScope = GlobalSearchScope.everythingScope(project)
      for (fileId in fileIds) {
        val file = ManagingFS.getInstance().findFileById(fileId) ?: continue
        if (!file.isDirectory && allScope.contains(file)) {
          files.add(file)
        }
        indicator.fraction = ++finished * 1.0 / fileIds.size
        indicator.text2 = file.path
      }
      return files
    }

    @Throws(StorageException::class)
    private fun mapHashIdsToLocalFileIds(indexIdToHashIds: Map<ID<*, *>, Set<Long>>, errorCollector: ErrorCollector): Set<Int> {
      val allLocalFileIds: MutableSet<Int> = HashSet()
      for ((indexId, hashIds) in indexIdToHashIds) {
        val hashIndex = FileContentHashIndex.getFileContentHashIndex(indexId)
        val fileIdToHashId: MutableMap<Int, Long> = HashMap()
        for (compositeHashId in hashIds) {
          hashIndex.getData(compositeHashId).forEach { fileId: Int, _: Void? ->
            val previousHashId = fileIdToHashId.put(fileId, compositeHashId)
            if (previousHashId != null && !Comparing.equal(previousHashId, compositeHashId)) {
              val file = ManagingFS.getInstance().findFileById(fileId)
              val fileMsg = if (file != null) file.path + " (fileId=" + fileId + ")" else "id=$fileId"
              errorCollector.addError(
                AssertionError("Duplicated file $fileMsg for hashes $previousHashId and $compositeHashId of index ${indexId.name}")
              )
            }
            stopIfTooManyErrors(errorCollector)
            true
          }
        }
        allLocalFileIds.addAll(fileIdToHashId.keys)
      }
      return allLocalFileIds
    }

    @Throws(StorageException::class)
    private fun collectAllSharedIndexHashIds(
      project: Project,
      indicator: ProgressIndicator
    ): Map<ID<*, *>, Set<Long>> {
      indicator.text = PerformanceTestingBundle.message("checking.shared.indexes.collecting.shared.files")
      indicator.isIndeterminate = true
      val allScope = GlobalSearchScope.allScope(project)
      val contentDependentIndexIds = contentDependentIndexes
      val indexIdToHashIds = HashMap<ID<*, *>, Set<Long>>()
      for (id in contentDependentIndexIds) {
        val hashIndex = FileContentHashIndex.getFileContentHashIndex(id)
        val allHashIds: Set<Long> = HashSet()
        hashIndex.processAllKeys(Processors.cancelableCollectProcessor(allHashIds), allScope, null)
        indexIdToHashIds[id] = allHashIds
      }
      return indexIdToHashIds
    }

    private val contentDependentIndexes: Set<ID<*, *>>
      get() = FileBasedIndexExtension.EXTENSION_POINT_NAME
        .extensions()
        .filter { it.dependsOnFileContent() }
        .map { it.name }
        .collect(Collectors.toSet())
  }

}