package com.intellij.indexing.shared.platform.api;

import com.google.common.collect.Maps;
import com.google.common.collect.Sets;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.diagnostic.Logger;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.annotations.TestOnly;

import java.util.*;

/**
 * Computes IDE own {@link IndexInfrastructureVersion}.
 * Note, the version may change when plugins are loaded or unloaded
 */
public final class IdeIndexInfrastructureVersion {
  private final IndexInfrastructureVersion myVersion;

  private IdeIndexInfrastructureVersion(@NotNull IndexInfrastructureVersion version) {
    myVersion = version;
  }

  /**
   * Computes current IDE version. Note, the version may change while IDE is running
   *
   * @return
   */
  @NotNull
  public static IdeIndexInfrastructureVersion getIdeVersion() {
    return new IdeIndexInfrastructureVersion(new IdeIndexesState().getIndexVersion());
  }

  @NotNull
  @TestOnly
  public static IdeIndexInfrastructureVersion forTests(@NotNull IndexInfrastructureVersion version) {
    Logger.getInstance(IdeIndexInfrastructureVersion.class).assertTrue(ApplicationManager.getApplication().isUnitTestMode());
    Logger.getInstance(IdeIndexInfrastructureVersion.class).assertTrue(version.getSupportedOs().size() == 1);
    return new IdeIndexInfrastructureVersion(version);
  }

  @Override
  public String toString() {
    return "IdeIndexInfrastructureVersion{" + myVersion + '}';
  }

  @NotNull
  public IndexInfrastructureVersion getVersion() {
    return myVersion;
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof IdeIndexInfrastructureVersion)) return false;
    IdeIndexInfrastructureVersion version = (IdeIndexInfrastructureVersion)o;
    return myVersion.equals(version.myVersion);
  }

  @Override
  public int hashCode() {
    return Objects.hash(myVersion);
  }

  /**
   * Checks if a given {@param version} is loadable with the IDE.
   * It may also check if it has a decent intersection with other indexes
   *
   * @see #pickBestSuitableVersion(Collection)
   */
  public boolean isSuitableVersion(@NotNull IndexInfrastructureVersion version) {
    return pickBestSuitableVersion(Collections.singletonList(version)) == version;
  }

  /**
   * Selects the versions from the given list {@param versions}'s that has the best similar versions
   * for shared indexes to apply. We try to pick the version with would allow us to make the minimal
   * extra work to get to the fully indexed state.
   *
   * @return It returns the best matching element from {@param versions} or {@code null} if none are good enough or suitable
   */
  @Nullable
  public IndexInfrastructureVersion pickBestSuitableVersion(@NotNull Collection<IndexInfrastructureVersion> versions) {
    IndexInfrastructureVersion ideVersion = this.getVersion();
    Integer bestScore = null;
    IndexInfrastructureVersion bestVersion = null;

    for (IndexInfrastructureVersion version : versions) {
      //we make sure supported OS contain the OS of the IDE
      if (!version.getSupportedOs().containsAll(ideVersion.getSupportedOs())) continue;

      //we make sure the persistent data structures, VFS version are exactly the same
      //it is required to be able to read indexes files on the IDE
      if (!version.getBaseIndexes().equals(ideVersion.getBaseIndexes())) continue;

      //how many common indexes are there?
      int commonFileBasedIndexesCount = Sets.intersection(
        version.getFileBasedIndexVersions().entrySet(),
        ideVersion.getFileBasedIndexVersions().entrySet()
      ).size();

      //a magic number - we skip loading if there are too few indexes matched
      //should we check for the StubUpdatingIndex here as well?
      if (commonFileBasedIndexesCount < 3) continue;

      // how many of included in the file based index versions are not present in IDE?
      int fileBasedIndexMiss = Sets.difference(
        ideVersion.getFileBasedIndexVersions().entrySet(),
        version.getFileBasedIndexVersions().entrySet()
      ).size();

      //how may of included stub file type elements in the shared index versions are not present in IDE?
      int stubFileElementsMiss = Sets.difference(
        ideVersion.getStubFileElementTypeVersions().entrySet(),
        version.getStubFileElementTypeVersions().entrySet()
      ).size();

      int negativeScope = fileBasedIndexMiss + stubFileElementsMiss;

      //computing min of the score
      if (bestScore == null || bestScore > negativeScope) {
        bestScore = negativeScope;
        bestVersion = version;
      }
    }

    return bestVersion;
  }

  public boolean isStubIndexVersionMatch(@NotNull IndexInfrastructureVersion version) {
    boolean hasMismatchingStubIndexVersions = !Maps.difference(
      version.getStubIndexVersions(),
      this.getVersion().getStubIndexVersions()
    ).entriesDiffering().isEmpty();

    if (hasMismatchingStubIndexVersions) return false;

    return true;
  }
}
