package com.intellij.indexing.shared.platform.api;

import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.ImmutableSortedSet;
import com.intellij.indexing.shared.platform.hash.SharedIndexContentHash;
import com.intellij.indexing.shared.platform.hash.SharedIndexHashProviders;
import com.intellij.indexing.shared.platform.impl.SharedIndexExtension;
import com.intellij.indexing.shared.platform.impl.StubSharedIndexExtension;
import com.intellij.lang.LanguageParserDefinitions;
import com.intellij.lang.ParserDefinition;
import com.intellij.openapi.application.ApplicationInfo;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.FileTypeManager;
import com.intellij.openapi.fileTypes.LanguageFileType;
import com.intellij.openapi.vfs.newvfs.persistent.FSRecords;
import com.intellij.psi.stubs.StubIndexExtension;
import com.intellij.psi.stubs.StubUpdatingIndex;
import com.intellij.psi.tree.IFileElementType;
import com.intellij.psi.tree.IStubFileElementType;
import com.intellij.util.hash.ContentHashEnumerator;
import com.intellij.util.indexing.FileBasedIndexExtension;
import com.intellij.util.indexing.IndexedHashesSupport;
import com.intellij.util.indexing.impl.storage.VfsAwareMapReduceIndex;
import com.intellij.util.io.PersistentEnumerator;
import org.jetbrains.annotations.NotNull;

import java.util.*;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.stream.Collectors;

public class IdeIndexesState {
  private static final Logger LOG = Logger.getInstance(IdeIndexesState.class);

  public final List<FileBasedIndexExtension<?, ?>> exportableFileBasedIndexExtensions;
  public final List<StubIndexExtension<?, ?>> exportableStubIndexExtensions;

  private final AtomicBoolean myIsFrozen = new AtomicBoolean(false);
  private IndexInfrastructureVersion myIndexVersion;

  public IdeIndexesState() {
    exportableFileBasedIndexExtensions = FileBasedIndexExtension
      .EXTENSION_POINT_NAME
      .extensions()
      .filter(FileBasedIndexExtension::dependsOnFileContent)
      .filter(FileBasedIndexExtension::canBeShared)
      .filter(ex -> !(ex instanceof StubUpdatingIndex))
      .collect(Collectors.toList());

    exportableStubIndexExtensions = StubIndexExtension.EP_NAME.getExtensionList();

    ImmutableSortedMap.Builder<String, String> global = ImmutableSortedMap.naturalOrder();
    //register your index here if that version is vital for shared indexes loading
    global.put("shared_indexes_format", IndexInfrastructureVersion.SHARED_INDEXES_VERSIONS_FORMAT_VERSION);
    global.put("vfs_version", String.valueOf(FSRecords.getVersion()));
    global.put("persistent_enumerator_version", String.valueOf(PersistentEnumerator.getVersion()));
    global.put("use_btree", String.valueOf(true));
    global.put("hashes_enumerator_version", String.valueOf(ContentHashEnumerator.getVersion()));
    global.put("hashes_algorithm_version", String.valueOf(IndexedHashesSupport.getVersion()));
    global.put("inverted_index_version", String.valueOf(VfsAwareMapReduceIndex.VERSION));
    global.put("product_version", getProductVersion());

    List<SharedIndexContentHash> allHashProviders = SharedIndexHashProviders.getSharedIndexHashingProvidersForMetadata();

    //NOTE[jo]: include all possible capabilities here too (we need a dedicated EP for that)
    myIndexVersion = new IndexInfrastructureVersion(Collections.emptyMap(),
                                                    ImmutableSortedSet.of(IndexInfrastructureOs.getOs()),
                                                    global.build(),
                                                    filedBasedIndexVersions(exportableFileBasedIndexExtensions),
                                                    stubIndexVersions(exportableStubIndexExtensions),
                                                    stubFileElementTypeVersions(getAllStubFileElementTypes()),
                                                    hashProviderVersions(allHashProviders));
  }


  public void addSupportedOs(@NotNull Collection<IndexInfrastructureOs> osToAdd) {
    assertIsNotFrozen();
    myIndexVersion = myIndexVersion.withSupportedOs(osToAdd);
  }

  public void registerCapability(@NotNull IndexInfrastructureCapability capability, @NotNull Map<String, String> versions) {
    assertIsNotFrozen();
    myIndexVersion = myIndexVersion.withCapability(capability, versions);
  }

  public void withStubFileTypeElements(@NotNull Set<IFileElementType> fileElementTypes) {
    assertIsNotFrozen();
    myIndexVersion = myIndexVersion.withStubFileElementTypes(stubFileElementTypeVersions(fileElementTypes));
  }

  public void withActualHashGeneratorVersions(@NotNull Collection<SharedIndexContentHash> providers) {
    assertIsNotFrozen();
    myIndexVersion = myIndexVersion.withHashProviderVersions(hashProviderVersions(providers));
  }

  @NotNull
  public IndexInfrastructureVersion getIndexVersion() {
    myIsFrozen.set(true);
    return myIndexVersion;
  }

  @NotNull
  private static String getProductVersion() {
    ApplicationInfo info = ApplicationInfo.getInstance();
    return info.getMajorVersion() + "." + info.getMinorVersionMainPart();
  }

  @NotNull
  private static Map<String, IndexInfrastructureFileBasedIndexVersionInfo> filedBasedIndexVersions(@NotNull List<FileBasedIndexExtension<?, ?>> fileBasedIndexExtensions) {
    var builder = new HashMap<String, IndexInfrastructureFileBasedIndexVersionInfo>();
    for (FileBasedIndexExtension<?, ?> fileBasedIndexExtension : fileBasedIndexExtensions) {
      SharedIndexExtension<?, ?> sharedIndexExtension = SharedIndexExtension.findExtension(fileBasedIndexExtension);

      String name = fileBasedIndexExtension.getName().getName();
      var newValue = new IndexInfrastructureFileBasedIndexVersionInfo(
        fileBasedIndexExtension.getVersion() + ":" + sharedIndexExtension.getVersion(),
        fileBasedIndexExtension.needsForwardIndexWhenSharing()
      );


      var oldValue = builder.put(name, newValue);

      if (oldValue != null && !oldValue.equals(newValue)) {
        LOG.warn("Multiple declarations of the same file based index: " + name + ", old value " + oldValue + ", new value: " + newValue);
      }
    }
    return builder;
  }

  @NotNull
  private static Map<String, String> stubIndexVersions(@NotNull List<StubIndexExtension<?, ?>> stubIndexExtensions) {
    var builder = new HashMap<String, String>();

    FileBasedIndexExtension<?, ?> stubUpdatingIndex =
      FileBasedIndexExtension.EXTENSION_POINT_NAME.findFirstSafe(ex -> ex.getName().equals(StubUpdatingIndex.INDEX_ID));

    if (stubUpdatingIndex == null) {
      throw new RuntimeException("Failed to find " + StubUpdatingIndex.INDEX_ID);
    }

    String commonPrefix = stubUpdatingIndex.getVersion() + ":";
    for (StubIndexExtension<?, ?> ex : stubIndexExtensions) {
      String name = ex.getKey().getName();
      String newValue = commonPrefix + ex.getVersion();
      var oldValue = builder.put(name, newValue);
      if (oldValue != null && !oldValue.equals(newValue)) {
        LOG.warn("Multiple declarations of the same stub based index: " + name + ", old value " + oldValue + ", new value: " + newValue);
      }
    }

    return builder;
  }

  @NotNull
  private static Map<String, String> hashProviderVersions(@NotNull Collection<SharedIndexContentHash> providers) {
    var builder = new HashMap<String, String>();

    for (var provider : providers) {
      String name = provider.getProviderId();
      String newValue = provider.getVersion();
      var oldValue = builder.put(name, newValue);
      if (oldValue != null && !oldValue.equals(newValue)) {
        LOG.warn("Multiple declarations of the same stub based index: " + name + ", old value " + oldValue + ", new value: " + newValue);
      }
    }

    return builder;
  }

  @NotNull
  private static Collection<IFileElementType> getAllStubFileElementTypes() {
    return Arrays.stream(FileTypeManager.getInstance().getRegisteredFileTypes())
      .filter(type -> type instanceof LanguageFileType)
      .map(type -> ((LanguageFileType)type).getLanguage())
      .map(LanguageParserDefinitions.INSTANCE::forLanguage)
      .filter(Objects::nonNull)
      .map(ParserDefinition::getFileNodeType)
      .collect(Collectors.toList());
  }

  @NotNull
  private static Map<String, String> stubFileElementTypeVersions(Collection<IFileElementType> fileElementTypes) {
    var builder = new HashMap<String, String>();

    for (IFileElementType fileNodeType : fileElementTypes) {
      if (fileNodeType instanceof IStubFileElementType) {
        int stubVersion = StubSharedIndexExtension.getStubFileElementBaseVersion((IStubFileElementType<?>)fileNodeType);
        String name = StubSharedIndexExtension.getStubFileElementTypeKey(fileNodeType);
        String newValue = Integer.toString(stubVersion);

        var oldValue = builder.put(name, newValue);
        if (oldValue != null && !oldValue.equals(newValue)) {
          LOG.warn("Multiple declarations of the same IFileElementType: " + name + ", old value " + oldValue + ", new value: " + newValue);
        }
      }
    }

    return builder;
  }

  private void assertIsNotFrozen() {
    if (myIsFrozen.get()) {
      throw new IllegalStateException("getIndexVersion() has already been executed, " +
                                      "it's too late to change IndexInfrastructureVersion");
    }
  }

  @NotNull
  public Set<FileType> getNoSizeLimitApplicableFileTypes() {
    return exportableFileBasedIndexExtensions
      .stream()
      .flatMap(g -> g.getFileTypesWithSizeLimitNotApplicable().stream())
      .collect(Collectors.toSet());
  }
}
