// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.api;

import com.google.common.collect.ImmutableMap;
import com.google.common.collect.ImmutableSet;
import com.google.common.collect.ImmutableSortedMap;
import com.google.common.collect.ImmutableSortedSet;
import com.google.common.hash.Hasher;
import com.google.common.hash.Hashing;
import com.intellij.indexing.shared.platform.impl.SharedIndexInfrastructure;
import com.intellij.openapi.util.text.StringUtil;
import org.jetbrains.annotations.NotNull;

import java.nio.charset.StandardCharsets;
import java.util.*;

/**
 * Represents Shared Index version, use to check (or select the best)
 * if a given shared index is compatible with an IDE
 *
 * @see IdeIndexInfrastructureVersion
 */
public final class IndexInfrastructureVersion {
  public static final String SHARED_INDEXES_VERSIONS_FORMAT_VERSION =
    String.valueOf(SharedIndexInfrastructure.SHARED_INDEX_INFRASTRUCTURE_VERSION);

  // base versions: it is required to have 100% match on that indexes in order to load it to an IDE.
  private final SortedMap<String, String> myBaseIndexes;

  // files-based indexes versions: indexes are loadable even if some indexes does not match
  private final SortedMap<String, IndexInfrastructureFileBasedIndexVersionInfo> myFileBasedIndexVersions;

  // stub indexes versions: it is on if some indexes does not match
  private final SortedMap<String, String> myStubIndexVersions;

  // stub file element type versions:
  private final SortedMap<String, String> myStubFileElementTypeVersions;

  // list of capabilities that are included
  private final SortedMap<IndexInfrastructureCapability, Map<String, String>> myCapabilities;

  // list of supported OS, for {@link IdeIndexInfrastructureVersion} it is the only current OS
  private final SortedSet<IndexInfrastructureOs> mySupportedOs;

  // list of supported HashProviders, see {@link SharedIndexContentHash} for details
  private final SortedMap<String, String> myHashProviderVersions;

  public IndexInfrastructureVersion(@NotNull Map<IndexInfrastructureCapability, Map<String, String>> capabilities,
                                    @NotNull Set<IndexInfrastructureOs> supportedOs,
                                    @NotNull Map<String, String> baseIndexes,
                                    @NotNull Map<String, IndexInfrastructureFileBasedIndexVersionInfo> fileBasedIndexVersions,
                                    @NotNull Map<String, String> stubIndexVersions,
                                    @NotNull Map<String, String> stubFileElementTypeVersions,
                                    @NotNull Map<String, String> hashProviderVersions) {
    myBaseIndexes = ImmutableSortedMap.copyOf(baseIndexes);
    myFileBasedIndexVersions = ImmutableSortedMap.copyOf(fileBasedIndexVersions);
    myStubIndexVersions = ImmutableSortedMap.copyOf(stubIndexVersions);
    myStubFileElementTypeVersions = ImmutableSortedMap.copyOf(stubFileElementTypeVersions);
    mySupportedOs = ImmutableSortedSet.copyOf(supportedOs);
    myHashProviderVersions = ImmutableSortedMap.copyOf(hashProviderVersions);

    ImmutableSortedMap.Builder<IndexInfrastructureCapability, Map<String, String>> builder = ImmutableSortedMap.naturalOrder();
    for (Map.Entry<IndexInfrastructureCapability, Map<String, String>> entry : capabilities.entrySet()) {
      builder.put(entry.getKey(), ImmutableSortedMap.copyOf(entry.getValue()));
    }

    myCapabilities = builder.build();
  }

  @Override
  public String toString() {
    return "IndexInfrastructureVersion{" +
           "base=" + myBaseIndexes +
           ", capabilities=" + myCapabilities +
           ", fileIndex=" + myFileBasedIndexVersions +
           ", stubIndex=" + myStubIndexVersions +
           ", stubFilesElementTypeVersions=" + myStubFileElementTypeVersions +
           ", supportedOs=" + mySupportedOs +
           '}';
  }

  /**
   * @return a weak global version, to be used in file names or other
   * places to refer to the global state
   */
  @NotNull
  public String getWeakVersionHash() {
    return computeWeakHash(Arrays.asList(getBaseIndexes(),
                                         getFileBasedIndexVersions(),
                                         getStubIndexVersions(),
                                         getStubFileElementTypeVersions()));
  }

  /**
   * @return a base version hash, o be used in file names or other
   * places to refer to the global state
   */
  @NotNull
  public String getWeakBaseVersionHash() {
    return computeWeakHash(Collections.singletonList(myBaseIndexes));
  }

  @NotNull
  private static String computeWeakHash(@NotNull List<Map<String, String>> maps) {
    Hasher hasher = Hashing.sha256().newHasher();
    for (Map<String, String> versions : maps) {
      for (Map.Entry<String, String> entry : ImmutableMap.copyOf(versions).entrySet()) {
        hasher.putString(entry.getKey(), StandardCharsets.UTF_8);
        hasher.putString(entry.getValue(), StandardCharsets.UTF_8);
      }
    }
    return StringUtil.first(hasher.hash().toString(), 12, false);
  }

  @NotNull
  public Map<String, String> getBaseIndexes() {
    return myBaseIndexes;
  }

  @NotNull
  public Map<String, String> getFileBasedIndexVersions() {
    ImmutableSortedMap.Builder<String, String> builder = ImmutableSortedMap.naturalOrder();
    for (var e : myFileBasedIndexVersions.entrySet()) {
      builder.put(e.getKey(), e.getValue().getVersion());
    }
    return builder.build();
  }

  @NotNull
  public Map<String, Boolean> getFileBasedIndexThatNeedsForwardIndex() {
    ImmutableSortedMap.Builder<String, Boolean> builder = ImmutableSortedMap.naturalOrder();
    for (var e : myFileBasedIndexVersions.entrySet()) {
      builder.put(e.getKey(), e.getValue().getNeedsForwardIndex());
    }
    return builder.build();
  }

  @NotNull
  public Map<String, IndexInfrastructureFileBasedIndexVersionInfo> getFileBasedIndexVersionInfos() {
    return myFileBasedIndexVersions;
  }

  @NotNull
  public Map<String, String> getStubIndexVersions() {
    return myStubIndexVersions;
  }

  @NotNull
  public Map<String, String> getStubFileElementTypeVersions() {
    return myStubFileElementTypeVersions;
  }

  @NotNull
  public Map<IndexInfrastructureCapability, Map<String, String>> getCapabilities() {
    return myCapabilities;
  }

  @NotNull
  public Set<IndexInfrastructureOs> getSupportedOs() {
    return mySupportedOs;
  }

  @NotNull
  public Map<String, String> getHashProviderVersions() {
    return myHashProviderVersions;
  }

  @NotNull
  public IndexInfrastructureVersion withCapability(@NotNull IndexInfrastructureCapability capability, @NotNull Map<String, String> versions) {
    ImmutableSortedMap<IndexInfrastructureCapability, Map<String, String>> caps =
      ImmutableSortedMap.<IndexInfrastructureCapability, Map<String, String>>naturalOrder().putAll(myCapabilities)
        .put(capability, ImmutableSortedMap.copyOf(versions)).build();

    return new IndexInfrastructureVersion(
      caps,
      mySupportedOs,
      myBaseIndexes,
      myFileBasedIndexVersions,
      myStubIndexVersions,
      myStubFileElementTypeVersions,
      myHashProviderVersions);
  }

  @NotNull
  public IndexInfrastructureVersion withStubFileElementTypes(@NotNull Map<String, String> versions) {
    return new IndexInfrastructureVersion(
      myCapabilities,
      mySupportedOs,
      myBaseIndexes,
      myFileBasedIndexVersions,
      myStubIndexVersions,
      versions,
      myHashProviderVersions);
  }

  @NotNull
  public IndexInfrastructureVersion withHashProviderVersions(@NotNull Map<String, String> hashProviderVersions) {
    return new IndexInfrastructureVersion(
      myCapabilities,
      mySupportedOs,
      myBaseIndexes,
      myFileBasedIndexVersions,
      myStubIndexVersions,
      myStubFileElementTypeVersions,
      hashProviderVersions);
  }

  @NotNull
  public IndexInfrastructureVersion withSupportedOs(@NotNull Collection<IndexInfrastructureOs> osToAdd) {
    Set<IndexInfrastructureOs> newSupportedOs = ImmutableSet.<IndexInfrastructureOs>builder().addAll(mySupportedOs).addAll(osToAdd).build();

    return new IndexInfrastructureVersion(
      myCapabilities,
      newSupportedOs,
      myBaseIndexes,
      myFileBasedIndexVersions,
      myStubIndexVersions,
      myStubFileElementTypeVersions,
      myHashProviderVersions
    );
  }

  @Override
  public boolean equals(Object o) {
    if (this == o) return true;
    if (o == null || getClass() != o.getClass()) return false;
    IndexInfrastructureVersion version = (IndexInfrastructureVersion)o;
    return myBaseIndexes.equals(version.myBaseIndexes) &&
           myFileBasedIndexVersions.equals(version.myFileBasedIndexVersions) &&
           myStubIndexVersions.equals(version.myStubIndexVersions) &&
           myStubFileElementTypeVersions.equals(version.myStubFileElementTypeVersions) &&
           myCapabilities.equals(version.myCapabilities) &&
           mySupportedOs.equals(version.mySupportedOs);
  }

  @Override
  public int hashCode() {
    return Objects.hash(myBaseIndexes, myFileBasedIndexVersions, myStubIndexVersions, myCapabilities, mySupportedOs, myStubFileElementTypeVersions);
  }
}
