package com.intellij.indexing.shared.platform.api

import com.intellij.openapi.util.text.StringUtil
import java.io.Closeable
import java.io.IOException
import java.nio.file.Path

/**
 * Represents shared index storage which may hold several shared index chunks (set of shared indexes + content hash enumerators).
 *
 * @see MutableSharedIndexStorage to have an instance of mutable shared index storage (where some new chunks might be appended).
 * @see SharedIndexStorageProvider to create a new instance of [SharedIndexStorage]
 */
interface SharedIndexStorage : Closeable {
  val storageId: String

  fun getSharedIndexStats(): SharedIndexStats

  fun isCompatibleChunk(chunkId: String): Boolean

  /**
   * @return root of the chunk with the given chunk ID or `null` if a chunk with such ID does not exist
   */
  fun getChunkRoot(chunkId: String): Path?

  fun containsChunk(chunkId: String): Boolean = getChunkRoot(chunkId) != null
}

/**
 * Represents a mutable [SharedIndexStorage].
 */
interface MutableSharedIndexStorage : SharedIndexStorage {
  @Throws(IOException::class)
  fun appendChunk(chunkSourcePath: Path, chunkId: String)
}

/**
 * A base interface to instantiate new shared index storage
 */
interface SharedIndexStorageProvider<Storage : SharedIndexStorage?> {
  @Throws(IOException::class)
  fun tryOpenOrResetIfFailed(corruptionHandler: Runnable): Storage
}

data class SharedIndexStats(
  val sizeInBytes: Long,
  val chunks: Int
) {
  fun toLogMessage(): String {
    if (chunks < 0) return "<failed to compute stats>"
    if (chunks == 0) return "<empty>"
    return "${chunks} chunk(s), ${StringUtil.formatFileSize(sizeInBytes)}"
  }

  operator fun plus(other: SharedIndexStats) = SharedIndexStats(sizeInBytes = this.sizeInBytes + other.sizeInBytes,
                                                                chunks = this.chunks + other.chunks)
}