package com.intellij.indexing.shared.platform.hash

import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.vfs.IntegrityCheckCapableFileSystem
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.newvfs.ArchiveFileSystem
import com.intellij.util.indexing.IndexedFile
import com.intellij.util.indexing.IndexedHashesSupport
import com.intellij.util.text.nullize
import org.jetbrains.annotations.VisibleForTesting

private object SharedIndexCrcArchiveHash : SharedIndexContentHash {
  override val providerId: String get() = "archive-crc"
  override val version: String get() = "1"
  override val performanceRank: Double get() = 700.0
  override val hashLength: Int get() = IndexedHashesSupport.HASH_SIZE_IN_BYTES // see ContentHashEnumerator.SIGNATURE_LENGTH
}

class SharedIndexCrcArchiveHashProvider : SharedIndexContentHashProvider {
  override val info: SharedIndexContentHash get() = SharedIndexCrcArchiveHash

  override fun computeHash(content: IndexedFile): ByteArray? {
    val file = content.file
    val fs = file.fileSystem
    if (fs !is ArchiveFileSystem ||
        fs !is IntegrityCheckCapableFileSystem) return null

    val archiveFile = getArchiveFile(file, fs) ?: return null

    val archiveId = getArchiveId(archiveFile.name) ?: return null
    val stamp = fs.getEntryCrc(file)
    if (stamp == 0L) return null
    val fileName = file.name

    val message = "$archiveId:$fileName:$stamp".toByteArray()
    return IndexedHashesSupport.calculateIndexedHash(content, message)
  }

  private fun getArchiveFile(archiveEntry: VirtualFile, fs: ArchiveFileSystem): VirtualFile? {
    val localFile = fs.getLocalByEntry(archiveEntry) ?: return null
    check(localFile.isInLocalFileSystem)
    return localFile
  }
}

@VisibleForTesting
fun getArchiveId(fileName: String): String? = fileName.substringBeforeLast("-").substringBeforeLast(".").nullize()

class SharedIndexCrcArchiveHashExporter : SharedIndexContentHashExporter {
  override val info: SharedIndexContentHash get() = SharedIndexCrcArchiveHash

  override fun createHashProvider(context: SharedIndexContentHashExporterContext,
                                  indicator: ProgressIndicator) = SharedIndexCrcArchiveHashProvider()
}
