package com.intellij.indexing.shared.platform.hash

import com.intellij.openapi.Disposable
import com.intellij.openapi.components.Service
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.text.StringUtil
import com.intellij.util.io.*
import java.io.DataInput
import java.io.DataOutput
import java.io.IOException
import java.nio.file.Path

@Service // application
class SharedIndexHashStorages {
  private val STORAGE_VERSION = "1"

  private fun SharedIndexContentHash.getHashPath(sharedIndexRoot: Path) = sharedIndexRoot.resolve("hashes-${providerId}-${version}-i$STORAGE_VERSION")

  fun openHashEnumerator(sharedIndexRoot: Path,
                         hash: SharedIndexContentHash,
                         lifetime: Disposable
  ): SharedIndexContentHashEnumerator? {
    // special case to handle here explicitly,
    // at least for now to avoid breaking the shared index format
    if (hash is SharedIndexMainContentHash) {
      return SharedIndexMainContentHashEnumerator(lifetime,
                                                  SharedIndexContentEnumerators.openEnumeratorForAttachedSharedIndex(sharedIndexRoot))
    }

    val indexPath = hash.getHashPath(sharedIndexRoot)
    if (!indexPath.isFile()) return null
    return SharedIndexHashStorageImpl(indexPath, hash, true, lifetime)
  }

  fun createGeneratorStorage(sharedIndexRoot: Path,
                             hash: SharedIndexContentHash,
                             lifetime: Disposable) : SharedIndexHashStorage? {

    // the main index, the special case, is handled separately
    // at least for now to avoid breaking the shared index format
    if (hash is SharedIndexMainContentHash) {
      return null
    }

    val indexPath = hash.getHashPath(sharedIndexRoot)
    if (indexPath.isFile()) throw IOException("Indexes hash file for ${hash.providerId} already exists at $indexPath")

    return SharedIndexHashStorageImpl(indexPath, hash, false, lifetime)
  }
}

interface SharedIndexHashStorage {
  fun enumerate(key: ByteArray, value: Int)
}

private class SharedIndexHashStorageImpl(
  path: Path,
  hash: SharedIndexContentHash,
  readonly: Boolean,
  lifetime: Disposable
) : SharedIndexContentHashEnumerator, SharedIndexHashStorage {
  override val info: SharedIndexContentHash = hash

  private val storage = PersistentMapImpl(
    PersistentMapBuilder.newBuilder(
      path,
      HashKeyDescriptor(hash.hashLength),
      InternalHashIdExternalizer).withReadonly(readonly).withCompactOnClose(!readonly)
  ).also {
    Disposer.register(lifetime, Disposable {
      it.close()
    })
  }

  override fun enumerate(key: ByteArray, value: Int) {
    require(value != 0) { "Value = 0, it cannot be persisted as we use 0 as 'null' replacement" }

    val oldValue = storage.get(key)
    if (oldValue != null && oldValue != 0 && oldValue != value) {
      storage.remove(key)
      throw RuntimeException(
        "Old value for ${StringUtil.toHexString(key)} is $oldValue != $value. " +
        "Could it be that this provider returns different values for the files which have " +
        "the same hash with the main hash provider? A weaker hash provider may cause " +
        "incorrect mapping between a file and a shared index, this key is ignored to avoid it"
      )
    }

    storage.put(key, value)
  }

  override fun tryEnumerate(hash: ByteArray): Int = storage.get(hash) ?: 0
}

private object InternalHashIdExternalizer : IntInlineKeyDescriptor() {
  override fun isCompactFormat(): Boolean = true
}

private class HashKeyDescriptor(val SIGNATURE_LENGTH: Int) : KeyDescriptor<ByteArray>, DifferentSerializableBytesImplyNonEqualityPolicy {
  override fun save(out: DataOutput, value: ByteArray) = out.write(value)

  override fun read(`in`: DataInput): ByteArray {
    val b = ByteArray(SIGNATURE_LENGTH)
    `in`.readFully(b)
    return b
  }

  override fun getHashCode(value: ByteArray) = value.contentHashCode()
  override fun isEqual(val1: ByteArray, val2: ByteArray): Boolean = val1.contentEquals(val2)
}
