package com.intellij.indexing.shared.platform.hash

import com.intellij.openapi.Disposable
import com.intellij.openapi.util.Disposer
import com.intellij.util.hash.ContentHashEnumerator
import com.intellij.util.indexing.FileContentImpl
import com.intellij.util.indexing.IndexedFile
import com.intellij.util.indexing.IndexedHashesSupport
import com.intellij.util.io.DataEnumeratorEx
import java.io.IOException
import java.nio.file.Path

object SharedIndexContentEnumerators {
  private const val hashFileName = "hashes"

  fun isRelatedToEnumerator(path: String): Boolean = path.startsWith(hashFileName)
  fun getHashEnumeratorPath(chunkRootPath: Path): Path = chunkRootPath.resolve(hashFileName)
  fun openEnumeratorForAttachedSharedIndex(chunkRootPath: Path) = object: ContentHashEnumerator(getHashEnumeratorPath(chunkRootPath)) {
    //
    // Sometimes our persistent enumerator may be corrupted due to unexpected input value.
    // Actually, it doesn't mean real data structure corruption. Just ignore it.
    //

    @Volatile
    private var corrupted : Boolean = false

    override fun markCorrupted() {
      corrupted = true
    }

    override fun isCorrupted(): Boolean = corrupted

    override fun enumerate(value: ByteArray?): Int = throw UnsupportedOperationException("Enumerator is read-only")

    override fun tryEnumerate(value: ByteArray?): Int {
      try {
        return super.tryEnumerate(value)
      }
      catch (e: IOException) {
        if (isCorrupted) {
          return DataEnumeratorEx.NULL_ID
        }
        throw e
      }
    }
  }
  fun newEnumeratorForGenerator(chunkRootPath: Path) = ContentHashEnumerator(getHashEnumeratorPath(chunkRootPath))
}

object SharedIndexMainContentHash : SharedIndexContentHash, DefaultSharedIndexContentHash {
  override val providerId: String get() = "main-hash"
  override val version: String get() = "1"
  override val performanceRank: Double get() = 1000.0
  override val hashLength: Int get() = 20 // see ContentHashEnumerator.SIGNATURE_LENGTH
}

class SharedIndexMainContentHashEnumerator(
  val lifetime: Disposable,
  val enumerator: ContentHashEnumerator
) : SharedIndexContentHashEnumerator {
  override val info: SharedIndexContentHash get() = SharedIndexMainContentHash

  init {
    Disposer.register(lifetime, Disposable { enumerator.close() })
  }

  override fun tryEnumerate(hash: ByteArray): Int = enumerator.tryEnumerate(hash)
}

class SharedIndexMainContentHashProvider : SharedIndexContentHashProvider {
  override val info: SharedIndexContentHash get() = SharedIndexMainContentHash

  override fun computeHash(content: IndexedFile): ByteArray? {
    if (content !is FileContentImpl) return null

    return computeHash(content)
  }

  fun computeHash(content: FileContentImpl): ByteArray? = IndexedHashesSupport.getOrInitIndexedHash(content)
}
