package com.intellij.indexing.shared.platform.hash

import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.project.Project
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.util.indexing.FileContent
import com.intellij.util.indexing.IndexedFile
import com.intellij.util.indexing.IndexedHashesSupport

/**
 * This service is responsible in returning different
 * pluggable hashes for files in the project.
 *
 * There are some examples for it:
 *  - archive hashes (based on archive hash and relative path)
 *  - archive hashes (based on file CRC)
 *  - bundled resources hashes (we may use shared index ID as base)
 *  - git hashes
 *
 *  All hashes are pluggable and optional. The ain is to map a
 *  given VirtualFile into the internal hashId in a given shared
 *  index.
 *
 *  The implementation is split into two sides:
 *   - the computation extension - that computes and return a (hash, providerId, version)
 *   - the usage extension - that tries providers one-by-one
 *
 *  Each provider can he fast or slow, depending on:
 *   - the method used
 *   - the need to fetch contents
 *   - CPU activity
 *
 *  Should a provider need a heavy preparations, it must do the preparations in the background
 *  thread and return `null`'s on all requests until the preparation work is finished.
 *
 */


/**
 * This service is used in IDE mode when indexing is running
 */
val SharedIndexContentHashProviderEP
  = ExtensionPointName.create<SharedIndexContentHashProvider>("com.intellij.sharedIndexHashProvider")

/**
 * This object is used as a key in the code, so it is
 * supposed there is a decent implementation of [Any.equals] and [Any.hashCode]
 */
interface SharedIndexContentHash {
  /**
   * The provider ID, it will be persisted in the shared index
   */
  val providerId: String

  /**
   * The version of that provider, so if the implementation of
   * [#computeHash] is changed, the version has to be updated
   */
  val version: String

  /**
   * Returns hash length in bytes for the returned data from
   * [#computeHash]. This is necessary for optimal hash persistence
   */
  val hashLength : Int

  /**
   * Priority is should be in [0, 1000]. The bigger the value,
   * the (relatively) slower is the implementation of that provider.
   * IDE is free to use this to sort providers to optimize indexing
   * performance
   */
  val performanceRank : Double
}

/**
 * A marker interface to make a default shared index hash provider, that is included by default
 * We check if [SharedIndexContentHash] implements this interface
 */
interface DefaultSharedIndexContentHash


interface SharedIndexContentHashProvider {
  /**
   * description of this hash provider
   */
  val info: SharedIndexContentHash

  /**
   * This method is supposed to work fast, if an extra work
   * is needed, the implementation may fork a background preparation
   * task and keep returning `null` immediately while the task
   * is running.
   *
   * The implementation is free to use [FileContent.getProject] to
   * get the project from the current scope
   *
   * The implementation may try to cast [content] to [FileContent]
   * in order to access the contents, but the operation may fail and
   * must he handled so that `null` is returned
   *
   * It is highly unrecommended to access the file contents through
   * [VirtualFile] APIs.
   *
   * The implementation may use [IndexedHashesSupport.calculateIndexedHash]
   * to mix-in standard add-ons to the hash. We are not calling it implicitly
   * to allow flexibility in the API
   *
   * It is highly recommended to use the [IndexedHashesSupport.calculateIndexedHash]
   * and to delegate ANY received hash to the function to make sure the final
   * hash comply with internal requirements
   */
  fun computeHash(content: IndexedFile): ByteArray?
}

/**
 * This interface defines access level to
 * the <hash bytes> -> internal chunk id mapping
 * table inside a given shared index
 */
interface SharedIndexContentHashEnumerator {
  /**
   * description of this hash provider
   */
  val info: SharedIndexContentHash

  /**
   * Returns 0 if no such hash is found
   */
  fun tryEnumerate(hash: ByteArray) : Int
}

/**
 * This service is used for shared indexes exporter to generate a shared index,
 * can be headless mode as well
 */
val SharedIndexContentHashExporterEP =
  ExtensionPointName.create<SharedIndexContentHashExporter>("com.intellij.sharedIndexHashExporter")

/**
 * Shared Indexes exporting extension that works (probably slow)
 * in the exporting process. The implementation should prepare the
 * best possible exporter (may take time for it) for a given project
 * and return it
 */
interface SharedIndexContentHashExporter {
  /**
   * description of this hash provider
   */
  val info: SharedIndexContentHash

  /**
   * Executed from Indexes Exported to prepare ready-to-go hashes
   * provider for a given project
   */
  fun createHashProvider(context: SharedIndexContentHashExporterContext, indicator: ProgressIndicator) : SharedIndexContentHashProvider?
}

/**
 * A context for [SharedIndexContentHashExporter].
 * This interface may be extended later with more data to use with hash provider
 */
interface SharedIndexContentHashExporterContext {
  val project : Project

  val allFiles: Set<VirtualFile>
}

object SharedIndexHashProviders {
  @JvmStatic
  fun getSharedIndexHashingProvidersForMetadata() = getSortedSharedIndexHashingProviders().map { it.info }

  @JvmStatic
  fun getNonDefaultSharedIndexHashingProviders() = getSharedIndexHashingProvidersForMetadata()
    .filter { it !is DefaultSharedIndexContentHash }
    .sortedBy { it.providerId }

  @JvmStatic
  fun getSortedSharedIndexHashingProviders() = SharedIndexContentHashProviderEP.extensionList.sortedBy { it.info.performanceRank }.toList()

  @JvmStatic
  fun getSharedIndexHashingGenerators(): List<SharedIndexContentHashExporter> = SharedIndexContentHashExporterEP.extensionList
}
