package com.intellij.indexing.shared.platform.impl

import com.intellij.indexing.shared.platform.api.AttachChunkResult
import com.intellij.indexing.shared.platform.hash.SharedIndexContentHashEnumerator
import com.intellij.indexing.shared.platform.hash.SharedIndexContentHashProvider
import com.intellij.indexing.shared.platform.hash.SharedIndexHashProviders
import com.intellij.indexing.shared.platform.hash.SharedIndexHashStorages
import com.intellij.openapi.Disposable
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProcessCanceledException
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.util.indexing.IndexedFile
import org.jetbrains.annotations.VisibleForTesting
import java.io.IOException
import java.nio.file.Path

private val LOG = Logger.getInstance(ChunkManager::class.java)

//Not thread safe. Must be synchronized on client side.
class ChunkManager {
  private val myChunkEnumerators = hashMapOf<Int, SharedIndexChunk>()

  @Throws(IOException::class)
  fun tryEnumerateContentHash(fileContent: IndexedFile): Long {
    // TODO in case of multiple chunks we can pass file's origin (module, library, etc) as an argument to reduce chunk search

    for ((chunkId, chunk) in myChunkEnumerators) {
      if (!chunk.attachedProjects.contains(fileContent.project)) continue

      val internalHashId = chunk.tryEnumerate(fileContent)
      if (internalHashId != 0) {
        return FileContentHashIndexExtension.getHashId(internalHashId, chunkId)
      }
    }

    return FileContentHashIndexExtension.NULL_HASH_ID
  }

  fun getOpenChunkIds(): Set<Int> {
    return myChunkEnumerators.keys
  }

  fun attachChunk(chunkId: Int, chunkRootPath: Path, indexAttachChunkResult: AttachChunkResult) {
    myChunkEnumerators[chunkId] = SharedIndexChunk(chunkRootPath, indexAttachChunkResult)
  }

  fun attachProjectToChunk(chunkId: Int, project: Project) {
    myChunkEnumerators[chunkId]?.attachProject(project)
  }

  fun getIndexAttachResult(chunkId: Int): AttachChunkResult? {
    return myChunkEnumerators[chunkId]?.indexAttachChunkResult
  }

  fun isProjectAttachedToChunk(chunkId: Int, project: Project): Boolean =
    myChunkEnumerators[chunkId]?.attachedProjects?.contains(project) ?: false

  fun disposeChunks(project: Project): List<Int> {
    val closedChunkIds = arrayListOf<Int>()
    for ((chunkId, enumerator) in myChunkEnumerators) {
      enumerator.detachProject(project)
      if (enumerator.attachedProjects.isEmpty()) {
        try {
          Disposer.dispose(enumerator)
        }
        catch (e: Exception) {
          LOG.info(e)
        }
        closedChunkIds += chunkId
      }
    }
    closedChunkIds.forEach { myChunkEnumerators.remove(it) }
    return closedChunkIds
  }

  fun disposeChunks(): Set<Int> {
    for (enumerator in myChunkEnumerators.values) {
      try {
        Disposer.dispose(enumerator)
      }
      catch (e: Exception) {
        LOG.info(e)
      }
    }
    val chunkIds = myChunkEnumerators.keys.toHashSet()
    myChunkEnumerators.clear()
    return chunkIds
  }
}

@VisibleForTesting
class SharedIndexChunk(chunkRootPath: Path, val indexAttachChunkResult: AttachChunkResult) : Disposable {

  fun tryEnumerate(fileContent: IndexedFile): Int {
    for ((hashProvider, hashEnumerator) in hashEnumerators) {
      val hash = try {
        hashProvider.computeHash(fileContent) ?: continue
      }
      catch (pce: ProcessCanceledException) {
        throw pce
      }
      catch (e: Exception) {
        LOG.error(e)
        continue
      }
      val hashId = hashEnumerator.tryEnumerate(hash)
      if (hashId == 0) continue
      return hashId
    }

    return 0
  }

  fun attachProject(project: Project) {
    attachedProjects.add(project)
  }

  fun detachProject(project: Project) {
    attachedProjects.remove(project)
  }

  private val hashEnumerators: Array<Pair<SharedIndexContentHashProvider, SharedIndexContentHashEnumerator>> =
    SharedIndexHashProviders.getSortedSharedIndexHashingProviders().mapNotNull {
      try {
        val enumerator = service<SharedIndexHashStorages>().openHashEnumerator(chunkRootPath, it.info, this)
        if (enumerator == null) return@mapNotNull null
        it to enumerator
      }
      catch (t: Throwable) {
        if (t is ControlFlowException) throw t
        LOG.warn("Failed to open enumerator for ${it.info.providerId} from $chunkRootPath. ${t.message}", t)
        null
      }
    }.toTypedArray()

  val attachedProjects = hashSetOf<Project>()

  override fun dispose() = Unit
}
