// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.intellij.indexing.composite.CompositeInvertedIndexBase;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.Computable;
import com.intellij.openapi.util.Ref;
import com.intellij.openapi.vfs.newvfs.NewVirtualFile;
import com.intellij.openapi.vfs.newvfs.persistent.PersistentFS;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.psi.stubs.StubIndexKey;
import com.intellij.psi.stubs.StubUpdatingIndex;
import com.intellij.util.Processor;
import com.intellij.util.indexing.*;
import com.intellij.util.indexing.impl.*;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.util.Collections;
import java.util.Map;
import java.util.function.Consumer;

public final class CompositeInvertedIndex<Key, Value> extends CompositeInvertedIndexBase<Key, Value> {
  private static final com.intellij.openapi.util.Key<Ref<@Nullable FileContent>> INDEXED_HASH_KEY =
    com.intellij.openapi.util.Key.create("indexed.file.shared.index.metadata.hash");
  private static final Logger LOG = Logger.getInstance(CompositeInvertedIndex.class);

  private final @NotNull SharedIndexChunkConfiguration mySharedIndexChunkConfiguration;

  @NotNull
  private final FileContentHashIndex myHashIndex;

  public CompositeInvertedIndex(@NotNull ID<Key, Value> id,
                                @NotNull UpdatableIndex<Key, Value, FileContent> baseIndex) throws IOException {
    super(id, baseIndex);
    mySharedIndexChunkConfiguration = SharedIndexChunkConfiguration.getInstance();

    myHashIndex = id instanceof StubIndexKey
                  ? FileContentHashIndex.getFileContentHashIndex(StubUpdatingIndex.INDEX_ID)
                  : FileContentHashIndex.createFileContentHashIndex(myId);
  }

  public boolean tryIndexWithoutContent(int inputId, @NotNull IndexedFile indexedFile) {
    var hashSuppliedIndexedFileRef = indexedFile.getUserData(INDEXED_HASH_KEY);
    if (hashSuppliedIndexedFileRef == null) {
      long hashId = mySharedIndexChunkConfiguration.tryEnumerateContentHash(indexedFile);
      hashSuppliedIndexedFileRef = new Ref<>();
      if (hashId != FileContentHashIndexExtension.NULL_HASH_ID) {
        hashSuppliedIndexedFileRef.set(new HashSuppliedIndexedFile(indexedFile, hashId));
      }
      indexedFile.putUserData(INDEXED_HASH_KEY, hashSuppliedIndexedFileRef);
    }

    var hashSuppliedIndexedFile = hashSuppliedIndexedFileRef.get();
    if (hashSuppliedIndexedFile == null) return false;
    Computable<Boolean> sharedIndexUpdate = myHashIndex.mapInputAndPrepareUpdate(inputId, hashSuppliedIndexedFile);
    long hashId = getHashId((MapReduceIndex<Long, Void, ?>.IndexUpdateComputable)sharedIndexUpdate);

    return hashId != FileContentHashIndexExtension.NULL_HASH_ID &&
           LOG.assertTrue(updateBaseIndex(inputId, null).compute()) &&
           LOG.assertTrue(sharedIndexUpdate.compute());
  }

  @Override
  @NotNull
  protected Computable<Boolean> removeFromSecondaryIndex(int inputId) {
    return myHashIndex.mapInputAndPrepareUpdate(inputId, null);
  }


  @Override
  protected void getSecondaryIndexData(@NotNull Key key, Consumer<ValueContainer<Value>> dataConsumer) throws StorageException {
    mySharedIndexChunkConfiguration.processSharedIndexes(myId, index -> {
      dataConsumer.accept(index.getData(key));
      return true;
    });
  }


  @Override
  protected @NotNull IndexInfrastructureUpdate tryToMapWithSecondaryIndex(int inputId,
                                                                          @NotNull FileContent content) {
    Computable<Boolean> sharedIndexUpdate = myHashIndex.mapInputAndPrepareUpdate(inputId, content);

    boolean hasSharedIndex;
    long hashId = getHashId((MapReduceIndex<Long, Void, ?>.IndexUpdateComputable)sharedIndexUpdate);

    int chunkId = FileContentHashIndexExtension.getChunkId(hashId);
    hasSharedIndex = mySharedIndexChunkConfiguration.hasSharedIndex(myId, chunkId);

    if (LOG.isTraceEnabled() && myId.equals(StubUpdatingIndex.INDEX_ID)) {
      if (hasSharedIndex) {
        LOG.trace("Shared stub index is found for " + content.getFileName() + " in chunk " + chunkId);
      }
      else {
        LOG.trace("Shared stub index is not found for " + content.getFileName());
      }
    }
    return new IndexInfrastructureUpdate(hasSharedIndex, sharedIndexUpdate);
  }

  private static long getHashId(MapReduceIndex<Long, Void, ?>.IndexUpdateComputable sharedIndexUpdate) {
    Map<Long, Void> map = sharedIndexUpdate.getInputData().getKeyValues();
    return FileContentHashIndexExtension.getHashId(map);
  }

  @Override
  @Nullable
  protected Map<Key, Value> getSecondaryIndexFileData(int fileId) throws StorageException {
    long hashId = hasAnyAttachedChunk()
                  ? myHashIndex.getHashId(fileId)
                  : FileContentHashIndexExtension.NULL_HASH_ID;
    int chunkId = FileContentHashIndexExtension.getChunkId(hashId);
    if (hashId != FileContentHashIndexExtension.NULL_HASH_ID && mySharedIndexChunkConfiguration.hasSharedIndex(myId, chunkId)) {
      int internalHashId = FileContentHashIndexExtension.getInternalHashId(hashId);

      Map<Key, Value> sharedFileData = mySharedIndexChunkConfiguration.querySharedIndex(
        myId, chunkId, index -> {
          if (!doesSharedIndexHaveForwardIndex(index)) {
            NewVirtualFile vFile = PersistentFS.getInstance().findFileByIdIfCached(fileId);
            if (vFile != null) {
              FileBasedIndex.getInstance().requestReindex(vFile);
            }
            LOG.error("Shared index chunk for " + getExtension().getName() + " doesn't have forward index. Index data might be incomplete");
            return Collections.emptyMap();
          }
          return index.getIndexedFileData(internalHashId);
        }
      );
      if (sharedFileData != null) {
        return sharedFileData;
      }
      throw new StorageException(String.format(
        "File data for %d hasn't been found although chunk %d is attached with index %s",
        fileId, chunkId, myId.getName()
      ));
    }
    return null;
  }

  @Override
  protected boolean processAllKeysBySecondaryIndex(@NotNull Processor<? super Key> processor,
                                                   @NotNull GlobalSearchScope scope,
                                                   @Nullable IdFilter idFilter)
    throws StorageException {
    return mySharedIndexChunkConfiguration.processSharedIndexes(myId, index -> index.processAllKeys(processor, scope, idFilter));
  }

  @Override
  protected void disposeSecondaryIndex() {
    if (!(myId instanceof StubIndexKey)) {
      myHashIndex.dispose();
    }
  }

  @Override
  protected void flushSecondaryIndex() throws StorageException {
    myHashIndex.flush();
  }

  @Override
  protected void clearSecondaryIndex() {
    myHashIndex.clear();
  }

  @NotNull
  public FileContentHashIndex getHashIndex() {
    return myHashIndex;
  }

  private boolean hasAnyAttachedChunk() {
    return !mySharedIndexChunkConfiguration.processSharedIndexes(myId, index -> false);
  }

  private static boolean doesSharedIndexHaveForwardIndex(@NotNull UpdatableIndex<?, ?, FileContent> index) {
    if (index instanceof EmptyIndex || index.getExtension() instanceof SingleEntryFileBasedIndexExtension) {
      return true;
    }
    return ((HashBasedMapReduceIndex<?, ?>)index).hasForwardIndex();
  }
}
