package com.intellij.indexing.shared.platform.impl

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.thisLogger
import com.intellij.openapi.util.ClearableLazyValue
import java.io.IOException
import java.nio.charset.StandardCharsets
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.StandardOpenOption
import java.util.*

private val log = logger<ExcludedChunkList>()

class ExcludedChunkList(private val excludedListDir: Path) {

  private val excludedChunks: ClearableLazyValue<Set<String>?> = ClearableLazyValue.create {
    readExcludedChunks()
  }

  fun markChunksAsExcluded(chunkId: String) {
    markChunksAsExcluded(Collections.singleton(chunkId))
  }

  @Synchronized
  fun markChunksAsExcluded(chunkIds: Collection<String>) {
    val chunkIdsToExclude = chunkIds.filterNot { excludedChunks.value.contains(it) }

    try {
      Files.createDirectories(excludedListDir)
      Files.write(getStoragePath(), chunkIdsToExclude, StandardCharsets.UTF_8, StandardOpenOption.CREATE, StandardOpenOption.APPEND)
    }
    catch (e: IOException) {
      log.error("can't add chunk ids to excluded chunk list", e)
      try {
        Files.createDirectories(excludedListDir)
        Files.write(getStoragePath(), excludedChunks.value union chunkIdsToExclude, StandardCharsets.UTF_8)
      }
      catch (e: IOException) {
        log.error("can't add chunk ids to excluded chunk list", e)
      }
    }
    finally {
      excludedChunks.drop()
    }
  }

  @Synchronized
  fun isChunkExcluded(chunkId: String): Boolean {
    return excludedChunks.value.contains(chunkId)
  }

  private fun readExcludedChunks(): Set<String> {
    try {
      val storagePath = getStoragePath()
      if (!Files.exists(storagePath)) return emptySet()
      return Files.readAllLines(storagePath, StandardCharsets.UTF_8).toHashSet()
    }
    catch (e: IOException) {
      log.error("excluded chunk list is corrupted", e)
      return emptySet()
    }
  }

  private fun getStoragePath(): Path {
    return excludedListDir.resolve("excluded-chunks.txt")
  }
}