// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.psi.stubs.StubIndexKey;
import com.intellij.psi.stubs.StubUpdatingIndex;
import com.intellij.util.ArrayUtil;
import com.intellij.util.indexing.*;
import com.intellij.util.indexing.impl.IndexStorage;
import com.intellij.util.indexing.impl.ValueContainerInputRemapping;
import com.intellij.util.indexing.impl.forward.ForwardIndex;
import com.intellij.util.indexing.impl.forward.ForwardIndexAccessor;
import com.intellij.util.indexing.impl.forward.MapForwardIndexAccessor;
import com.intellij.util.indexing.impl.forward.PersistentMapBasedForwardIndex;
import com.intellij.util.indexing.impl.storage.VfsAwareIndexStorageLayout;
import com.intellij.util.indexing.impl.storage.VfsAwareMapIndexStorage;
import com.intellij.util.indexing.impl.storage.VfsAwareMapReduceIndex;
import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.nio.file.Path;
import java.util.Map;

public final class FileContentHashIndex extends VfsAwareMapReduceIndex<Long, Void> {
  private static final Logger LOG = Logger.getInstance(FileContentHashIndex.class);

  public FileContentHashIndex(@NotNull FileContentHashIndexExtension extension,
                       @NotNull VfsAwareIndexStorageLayout<Long, Void> layout) throws IOException {
    super(extension, layout, null);
  }

  @Override
  protected void requestRebuild(@NotNull Throwable ex) {
    FileContentHashIndexExtension extension = getExtension();
    FileBasedIndex.getInstance().requestRebuild(extension.getName().getBaseIndexId());
  }

  @Override
  @NotNull
  public FileContentHashIndexExtension getExtension() {
    return (FileContentHashIndexExtension)super.getExtension();
  }

  public int getAssociatedChunkId(int fileId, VirtualFile file) {
    try {
      return FileContentHashIndexExtension.getChunkId(getHashId(fileId));
    }
    catch (StorageException e) {
      LOG.error(e);
      FileBasedIndex.getInstance().requestReindex(file);
      return FileContentHashIndexExtension.NULL_CHUNK_ID;
    }
  }

  public long getHashId(int fileId) throws StorageException {
    Map<Long, Void> data = getIndexedFileData(fileId);
    if (data.isEmpty()) return FileContentHashIndexExtension.NULL_HASH_ID;
    return data.keySet().iterator().next();
  }

  @NotNull
  public ValueContainerInputRemapping getHashIdToFileIdsFunction(int chunkId) {
    return hash -> {
      try {
        ValueContainer<Void> data = getData(FileContentHashIndexExtension.getHashId(hash, chunkId));
        if (data.size() == 0) return ArrayUtil.EMPTY_INT_ARRAY;
        return collect(data.getValueIterator().getInputIdsIterator());
      }
      catch (StorageException e) {
        throw new RuntimeException(e);
      }
    };
  }

  @Override
  public void dispose() {
    FileContentHashIndexExtension extension = getExtension();
    HashIndexKey key = extension.getName();
    try {
      super.dispose();
    }
    finally {
      ID.unloadId(key);
    }
  }

  private static int[] collect(@NotNull ValueContainer.IntIterator intIterator) {
    int[] result = new int[intIterator.size()];
    for (int i = 0; i < result.length; i++) {
      result[i] = intIterator.next();
    }
    return result;
  }

  @NotNull
  public static FileContentHashIndex getFileContentHashIndex(@NotNull ID<?, ?> indexId) {
    UpdatableIndex<?, ?, ?> index;
    if (indexId instanceof StubIndexKey) {
      indexId = StubUpdatingIndex.INDEX_ID;
    }
    FileBasedIndexImpl fileBasedIndex = (FileBasedIndexImpl)FileBasedIndex.getInstance();
    index = fileBasedIndex.getIndex(indexId);
    if (!(index instanceof CompositeInvertedIndex)) {
      throw new IllegalArgumentException(indexId.getName() + " is not supported by shared index. Actual index type: " + index.getClass());
    }
    return ((CompositeInvertedIndex<?, ?>)index).getHashIndex();
  }

  static @NotNull Path getFileContentHashIndexDir(@NotNull HashIndexKey indexId) {
    Path sharedIndexRoot = SharedIndexChunkConfigurationImpl.getSharedIndexConfigurationRoot();
    String name = indexId.getName();
    return sharedIndexRoot.resolve(name);
  }

  @NotNull
  static FileContentHashIndex createFileContentHashIndex(@NotNull ID<?, ?> index) throws IOException {
    FileContentHashIndexExtension extension = new FileContentHashIndexExtension(index);
    HashIndexKey sharedIndexKey = extension.getName();
    String name = sharedIndexKey.getName();
    Path indexDir = getFileContentHashIndexDir(sharedIndexKey);

    try {
      return new FileContentHashIndex(extension, new VfsAwareIndexStorageLayout<>() {
        @Override
        public void clearIndexData() {
          FileUtil.deleteWithRenaming(indexDir.toFile());
        }

        @Override
        public @NotNull IndexStorage<Long, Void> openIndexStorage() throws IOException {
          Path storagePath = indexDir.resolve(name + "_storage");
          return new VfsAwareMapIndexStorage<>(
            storagePath,
            extension.getKeyDescriptor(),
            extension.getValueExternalizer(),
            extension.getCacheSize(),
            extension.keyIsUniqueForIndexedFile(),
            extension.traceKeyHashToVirtualFileMapping()
          );
        }

        @Override
        public @NotNull ForwardIndex openForwardIndex() throws IOException {
          Path storagePath = indexDir.resolve(name + "_inputs");
          return new PersistentMapBasedForwardIndex(storagePath, false);
        }

        @Override
        public @NotNull ForwardIndexAccessor<Long, Void> getForwardIndexAccessor() {
          return new MapForwardIndexAccessor<>(new InputMapExternalizer<>(extension));
        }
      });
    }
    catch (IOException e) {
      try {
        FileUtil.delete(indexDir);
      } catch (IOException e1) {
        LOG.error(e1);
      }
      throw e;
    }
  }
}
