// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.intellij.openapi.util.Key;
import com.intellij.util.indexing.*;
import com.intellij.util.io.DataExternalizer;
import com.intellij.util.io.DataInputOutputUtil;
import com.intellij.util.io.KeyDescriptor;
import com.intellij.util.io.VoidDataExternalizer;
import org.jetbrains.annotations.NotNull;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.util.Collection;
import java.util.Collections;
import java.util.Map;

public class FileContentHashIndexExtension extends FileBasedIndexExtension<Long, Void>
  implements CustomInputsIndexFileBasedIndexExtension<Long> {
  private final @NotNull SharedIndexChunkConfiguration myIndexChunkConfiguration;
  private final @NotNull HashIndexKey myIndexId;

  public FileContentHashIndexExtension(@NotNull ID<?, ?> baseIndexId) {
    this(SharedIndexChunkConfiguration.getInstance(), baseIndexId);
  }

  FileContentHashIndexExtension(@NotNull SharedIndexChunkConfiguration configuration, @NotNull ID<?, ?> baseIndexId) {
    myIndexChunkConfiguration = configuration;
    myIndexId = HashIndexKey.createHashIndexKey(baseIndexId);
  }

  @NotNull
  @Override
  public HashIndexKey getName() {
    return myIndexId;
  }

  @NotNull
  @Override
  public FileBasedIndex.InputFilter getInputFilter() {
    return file -> !file.isDirectory();
  }

  @Override
  public boolean dependsOnFileContent() {
    return true;
  }

  @NotNull
  @Override
  public DataIndexer<Long, Void, FileContent> getIndexer() {
    return new DataIndexer<>() {
      @Override
      public @NotNull Map<Long, Void> map(@NotNull FileContent fc) {
        if (!isPhysical(fc)) return Collections.emptyMap();
        long hashId = getHashId(fc);
        if (hashId == NOT_YET_CALCULATED) {
          hashId = myIndexChunkConfiguration.tryEnumerateContentHash(fc);
        }

        int chunkId = getChunkId(hashId);

        if (hashId != NULL_HASH_ID) {
          //TODO: what if we have 2 shared indexes, one without data for the given myIndexId and the other with.
          //TODO: as of now, it may turn out we skip the updates for the same hashes if they map to the
          //TODO: first shared index ID
          if (myIndexChunkConfiguration.isSharedIndexAcceptable(myIndexId.getBaseIndexId(), chunkId, fc)) {
            setHashId(fc, hashId);
          }
          else {
            hashId = NULL_HASH_ID;
          }
        }

        return asMap(hashId);
      }

      @NotNull
      private Map<Long, Void> asMap(long hashId) {
        return hashId == NULL_HASH_ID ? Collections.emptyMap() : Collections.singletonMap(hashId, null);
      }

      private boolean isPhysical(@NotNull FileContent fc) {
        return !(fc instanceof FileContentImpl) || ((FileContentImpl)fc).isPhysicalContent();
      }
    };
  }

  @NotNull
  @Override
  public KeyDescriptor<Long> getKeyDescriptor() {
    return new KeyDescriptor<>() {
      @Override
      public int getHashCode(Long value) {
        return value.hashCode();
      }

      @Override
      public boolean isEqual(Long val1, Long val2) {
        return val1.longValue() == val2.longValue();
      }

      @Override
      public void save(@NotNull DataOutput out, Long value) throws IOException {
        DataInputOutputUtil.writeLONG(out, value);
      }

      @Override
      public Long read(@NotNull DataInput in) throws IOException {
        return DataInputOutputUtil.readLONG(in);
      }
    };
  }

  @NotNull
  @Override
  public DataExternalizer<Void> getValueExternalizer() {
    return VoidDataExternalizer.INSTANCE;
  }

  // TODO right now it's not used anywhere
  @Override
  public int getVersion() {
    return SharedIndexInfrastructure.SHARED_INDEX_INFRASTRUCTURE_VERSION;
  }

  @NotNull
  @Override
  public DataExternalizer<Collection<Long>> createExternalizer() {
    return new DataExternalizer<>() {
      @Override
      public void save(@NotNull DataOutput out, Collection<Long> value) throws IOException {
        assert value.isEmpty() || value.size() == 1;
        DataInputOutputUtil.writeLONG(out, value.isEmpty() ? 0 : value.iterator().next());
      }

      @Override
      public Collection<Long> read(@NotNull DataInput in) throws IOException {
        long id = DataInputOutputUtil.readLONG(in);
        return id == 0 ? Collections.emptyList() : Collections.singleton(id);
      }
    };
  }

  public static long getHashId(@NotNull Map<Long, Void> inputData) {
    return inputData.isEmpty() ? NULL_HASH_ID : inputData.keySet().iterator().next().longValue();
  }

  public static int getChunkId(long hashId) {
    return (int)(hashId >> 32);
  }

  public static int getInternalHashId(long hashId) {
    return (int)hashId;
  }

  public static long getHashId(int internalHashId, int chunkId) {
    return (((long)chunkId) << 32) | (internalHashId & 0xffffffffL);
  }

  public static final int NULL_CHUNK_ID = -1;
  public static final long NULL_HASH_ID = getHashId(0, NULL_CHUNK_ID);
  private static final long NOT_YET_CALCULATED = getHashId(0, -2);

  private static final Key<Long> HASH_ID_KEY = Key.create("file.content.hash.id");

  private static long getHashId(@NotNull FileContent content) {
    Long value = HASH_ID_KEY.get(content);
    return value == null ? NOT_YET_CALCULATED : value;
  }

  private static void setHashId(@NotNull FileContent content, long hashId) {
    HASH_ID_KEY.set(content, hashId);
  }
}
