// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.google.common.annotations.VisibleForTesting;
import com.intellij.indexing.shared.generate.HashBasedIndexGenerator;
import com.intellij.openapi.progress.ProcessCanceledException;
import com.intellij.openapi.progress.ProgressManager;
import com.intellij.openapi.util.Computable;
import com.intellij.psi.search.GlobalSearchScope;
import com.intellij.util.Processor;
import com.intellij.util.indexing.*;
import com.intellij.util.indexing.impl.MapReduceIndexMappingException;
import com.intellij.util.indexing.impl.IndexStorage;
import com.intellij.util.indexing.impl.MapIndexStorage;
import com.intellij.util.indexing.impl.ValueContainerInputRemapping;
import com.intellij.util.indexing.impl.forward.ForwardIndex;
import com.intellij.util.indexing.impl.forward.ForwardIndexAccessor;
import com.intellij.util.indexing.impl.storage.VfsAwareIndexStorageLayout;
import com.intellij.util.indexing.impl.storage.VfsAwareMapReduceIndex;
import com.intellij.util.io.DataExternalizer;
import com.intellij.util.io.KeyDescriptor;
import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Map;
import java.util.function.LongSupplier;

@ApiStatus.Internal
public class HashBasedMapReduceIndex<Key, Value> extends VfsAwareMapReduceIndex<Key, Value> {
  @NotNull
  private final SharedIndexExtension<Key, Value> mySharedExtension;
  @NotNull
  private final LongSupplier myHashIndexStampSupplier;
  private volatile long myLastFileContentIndexStamp = -1;

  @VisibleForTesting
  public HashBasedMapReduceIndex(@NotNull Path chunkPath,
                                 @NotNull SharedIndexExtension<Key, Value> sharedExtension,
                                 @NotNull ValueContainerInputRemapping hashIdRemapping,
                                 @NotNull LongSupplier hashIndexStampSupplier) throws IOException {
    super(assertIndexerNotUsed(sharedExtension.getFileBasedIndexExtension()),
          createIndexLayout(chunkPath, sharedExtension, hashIdRemapping),
          null);
    mySharedExtension = sharedExtension;
    myHashIndexStampSupplier = hashIndexStampSupplier;
  }

  @NotNull
  private static <Key, Value> VfsAwareIndexStorageLayout<Key, Value> createIndexLayout(@NotNull Path chunkPath,
                                                                                       @NotNull SharedIndexExtension<Key, Value> sharedExtension,
                                                                                       @NotNull ValueContainerInputRemapping hashIdRemapping) {
    if (HashBasedIndexGenerator.shouldCreateForwardIndex(sharedExtension)) {
      return new VfsAwareIndexStorageLayout<>() {
        @Override
        public void clearIndexData() {
          throw new UnsupportedOperationException();
        }

        @Override
        public @NotNull IndexStorage<Key, Value> openIndexStorage() throws IOException {
          return createStorage(chunkPath, sharedExtension, hashIdRemapping);
        }

        @Override
        public @NotNull ForwardIndex openForwardIndex() throws IOException {
          return HashBasedIndexGenerator.createForwardIndex(getStorageFileName(chunkPath, sharedExtension), true);
        }

        @Override
        public @NotNull ForwardIndexAccessor<Key, Value> getForwardIndexAccessor() {
          return HashBasedIndexGenerator.createForwardIndexAccessor(sharedExtension, hashId -> {
            throw new AssertionError();
          });
        }
      };
    }
    else {
      return new VfsAwareIndexStorageLayout<>() {
        @Override
        public void clearIndexData() {
          throw new UnsupportedOperationException();
        }

        @Override
        public @NotNull IndexStorage<Key, Value> openIndexStorage() throws IOException {
          return createStorage(chunkPath, sharedExtension, hashIdRemapping);
        }
      };
    }
  }

  public boolean acceptsFile(@NotNull IndexedFile file) {
    return mySharedExtension.acceptsInputFile(file);
  }

  @Override
  public @NotNull Computable<Boolean> mapInputAndPrepareUpdate(int inputId, @Nullable FileContent content)
    throws MapReduceIndexMappingException, ProcessCanceledException {
    return super.mapInputAndPrepareUpdate(inputId, content);
  }

  @Override
  public @NotNull Map<Key, Value> getIndexedFileData(int hashId) throws StorageException {
    ensureCachesUpToDate();
    if (getExtension() instanceof SingleEntryFileBasedIndexExtension || hasForwardIndex()) {
      return super.getIndexedFileData(hashId);
    }
    throw new AssertionError();
  }

  @Override
  public @NotNull ValueContainer<Value> getData(@NotNull Key key) throws StorageException {
    ensureCachesUpToDate();
    return super.getData(key);
  }

  @Override
  protected void doDispose() throws StorageException {
    try {
      super.doDispose();
    }
    finally {
      mySharedExtension.closeResources();
    }
  }

  boolean hasForwardIndex() {
    return getForwardIndex() != null;
  }

  private static <Key, Value> IndexStorage<Key, Value> createStorage(@NotNull Path baseFile,
                                                                     @NotNull SharedIndexExtension<Key, Value> sharedExtension,
                                                                     @NotNull ValueContainerInputRemapping hashToFileId)
    throws IOException {
    return new MyMapIndexStorage<>(getStorageFileName(baseFile, sharedExtension), sharedExtension, hashToFileId);
  }

  @NotNull
  private static <Key, Value> Path getStorageFileName(@NotNull Path baseFile, @NotNull SharedIndexExtension<Key, Value> extension) {
    return baseFile.resolve(extension.getFileBasedIndexExtension().getName().getName());
  }

  private static final class MyMapIndexStorage<Key, Value>
    extends MapIndexStorage<Key, Value>
    implements VfsAwareIndexStorage<Key, Value> {
    private MyMapIndexStorage(Path baseFile,
                              SharedIndexExtension<Key, Value> sharedExtension,
                              ValueContainerInputRemapping hashToFileIds) throws IOException {
      super(baseFile,
            sharedExtension.getFileBasedIndexExtension().getKeyDescriptor(),
            sharedExtension.getFileBasedIndexExtension().getValueExternalizer(),
            sharedExtension.getFileBasedIndexExtension().getCacheSize(),
            sharedExtension.getFileBasedIndexExtension().keyIsUniqueForIndexedFile(),
            true, true, hashToFileIds);
    }

    @Override
    protected void checkCanceled() {
      ProgressManager.checkCanceled();
    }

    @Override
    public boolean processKeys(@NotNull Processor<? super Key> processor, GlobalSearchScope scope, @Nullable IdFilter idFilter)
      throws StorageException {
      l.lock();
      try {
        myCache.clear(); // this will ensure that all new keys are made into the map
        return doProcessKeys(processor);
      }
      catch (IOException e) {
        throw new StorageException(e);
      }
      catch (RuntimeException e) {
        return unwrapCauseAndRethrow(e);
      }
      finally {
        l.unlock();
      }
    }
  }

  private void ensureCachesUpToDate() {
    long currentStamp = myHashIndexStampSupplier.getAsLong();
    if (myLastFileContentIndexStamp != currentStamp) {
      ((MapIndexStorage<?, ?>)myStorage).clearCachedMappings();
      myLastFileContentIndexStamp = currentStamp;
    }
  }

  // TODO should be properly fixed after 201
  // VfsAwareMapReduceIndex implementation should be properly split to read only and updateable part
  // please do not do it for now
  @NotNull
  private static <K, V> FileBasedIndexExtension<K, V> assertIndexerNotUsed(FileBasedIndexExtension<K, V> extension) {
    if (extension instanceof SingleEntryFileBasedIndexExtension) {
      //noinspection unchecked
      return (FileBasedIndexExtension<K, V>)new SingleEntryFileBasedIndexExtension<V>() {

        @Override
        public @NotNull ID<Integer, V> getName() {
          //noinspection unchecked
          return (ID<Integer, V>)extension.getName();
        }

        @Override
        public FileBasedIndex.@NotNull InputFilter getInputFilter() {
          throw error();
        }

        @Override
        public @NotNull SingleEntryIndexer<V> getIndexer() {
          return new SingleEntryIndexer<>(false) {
            @Nullable
            @Override
            protected V computeValue(@NotNull FileContent inputData) {
              throw error();
            }
          };
        }

        @Override
        public @NotNull DataExternalizer<V> getValueExternalizer() {
          return new MyKeyDescriptor<>();
        }

        @Override
        public int getVersion() {
          throw error();
        }
      };
    }
    else {
      return new FileBasedIndexExtension<>() {
        @Override
        public @NotNull ID<K, V> getName() {
          return extension.getName();
        }

        @Override
        public FileBasedIndex.@NotNull InputFilter getInputFilter() {
          throw error();
        }

        @Override
        public boolean dependsOnFileContent() {
          return extension.dependsOnFileContent();
        }

        @Override
        public @NotNull DataIndexer<K, V, FileContent> getIndexer() {
          // should not be a composite indexer
          return fc -> {
            throw error();
          };
        }

        @Override
        public @NotNull KeyDescriptor<K> getKeyDescriptor() {
          return new MyKeyDescriptor<>();
        }

        @Override
        public @NotNull DataExternalizer<V> getValueExternalizer() {
          return new MyKeyDescriptor<>();
        }

        @Override
        public int getVersion() {
          throw error();
        }
      };
    }
  }

  private static class MyKeyDescriptor<K> implements KeyDescriptor<K> {
    @Override
    public int getHashCode(K value) {
      throw error();
    }

    @Override
    public boolean isEqual(K val1, K val2) {
      throw error();
    }

    @Override
    public void save(@NotNull DataOutput out, K value) {
      throw error();
    }

    @Override
    public K read(@NotNull DataInput in) {
      throw error();
    }
  }

  @NotNull
  private static AssertionError error() {
    return new AssertionError("Should not be called for shared index");
  }
}
