// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl

import com.intellij.indexing.shared.platform.api.IndexInfrastructureVersion
import com.intellij.indexing.shared.platform.impl.SharedIndexExtension.getSharedIndexDir
import com.intellij.util.indexing.FileBasedIndexExtension
import com.intellij.util.indexing.FileContent
import com.intellij.util.indexing.ID
import com.intellij.util.indexing.UpdatableIndex
import com.intellij.util.io.PagedFileStorage
import com.intellij.util.io.StorageLock
import com.intellij.util.io.StorageLockContext
import org.jetbrains.annotations.ApiStatus
import java.io.Closeable
import java.nio.file.Path
import java.util.concurrent.ConcurrentHashMap
import java.util.function.LongSupplier

@ApiStatus.Internal
sealed class SharedIndex<K, V>(val indexName: ID<K, V>, val chunkId: Int) : Closeable {
  abstract val index: UpdatableIndex<K, V, FileContent>

  class Impl<K, V>(
    chunkRoot: Path,
    indexId: ID<K, V>,
    chunkId: Int,
    private val originalExtension: FileBasedIndexExtension<K, V>,
    private val hashIndex: FileContentHashIndex,
    private val chunkVersion: IndexInfrastructureVersion
  ) : SharedIndex<K, V>(indexId, chunkId) {

    private val chunkPath: Path = getSharedIndexDir(chunkRoot, indexId)

    override val index: HashBasedMapReduceIndex<K, V> = withReadWriteStorageLock(chunkId) {
      HashBasedMapReduceIndex(
        chunkPath,
        SharedIndexExtension.findExtensionAndOpen(originalExtension, chunkPath, true, chunkVersion),
        hashIndex.getHashIdToFileIdsFunction(chunkId),
        LongSupplier { hashIndex.modificationStamp }
      )
    }
  }

  class Empty<K, V>(indexName: ID<K, V>, chunkId: Int) : SharedIndex<K, V>(indexName, chunkId) {
    override val index: UpdatableIndex<K, V, FileContent>
      get() = EmptyIndex.getInstance()
  }

  override fun close() {
    index.dispose()
  }

  companion object {
    /**
     * We use per chunk lock to reduce file count limitation for [PagedFileStorage]
     *
     * Additionally we supply it with r-w lock since it provides more scalable concurrent read access.
     */
    private val sharedIndexLock = ConcurrentHashMap<Int, StorageLockContext>()

    @JvmStatic
    fun <T> withReadWriteStorageLock(chunkId: Int, operation: () -> T): T {
      if (PagedFileStorage.THREAD_LOCAL_STORAGE_LOCK_CONTEXT.get() != null) {
        throw AssertionError("reentrant storage lock option set-up")
      }
      try {
        PagedFileStorage.THREAD_LOCAL_STORAGE_LOCK_CONTEXT.set(sharedIndexLock.computeIfAbsent(chunkId) {
          StorageLockContext(StorageLock(), true, true)
        })
        return operation()
      }
      finally {
        PagedFileStorage.THREAD_LOCAL_STORAGE_LOCK_CONTEXT.remove()
      }
    }
  }
}
