// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.intellij.indexing.shared.platform.api.ChunkDescriptor;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.project.Project;
import com.intellij.util.indexing.*;
import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.Collection;

/**
 * Main service is used to manage (load / unload / download) shared index chunks.
 */
@ApiStatus.Internal
public interface SharedIndexChunkConfiguration {

  FileBasedIndexInfrastructureExtension.@NotNull InitializationResult initialize();

  void shutdown();

  boolean hasAnyChunk();

  boolean isSharedIndexAcceptable(@NotNull ID<?, ?> indexId, int chunkId, @NotNull IndexedFile indexedFile);

  @NotNull
  static SharedIndexChunkConfiguration getInstance() {
    return ApplicationManager.getApplication().getService(SharedIndexChunkConfiguration.class);
  }

  @Nullable
  static SharedIndexChunkConfiguration getInstanceIfCreated() {
    return ApplicationManager.getApplication().getServiceIfCreated(SharedIndexChunkConfiguration.class);
  }

  // Consider replacing with something more descriptive.
  @Nullable
  String getChunkDebugName(int chunkId);

  /**
   * @return true if an index for a given {@param indexId} which are stored in {@param chunkId} exist.
   */
  <Key, Value> boolean hasSharedIndex(@NotNull ID<Key, Value> indexId, int chunkId);

  /**
   * Processes all available shared indexes (among all of loaded chunks) for a given {@param indexId}.
   *
   * @param indexId
   * @param processor
   */
  <Key, Value, E extends Exception> boolean processSharedIndexes(@NotNull ID<Key, Value> indexId,
                                                                 @NotNull SharedIndexProcessor<Key, Value, E> processor)
    throws E;

  /**
   * Processes shared index by a given {@param indexId} and {@param chunkId}
   *
   * @return result calculated by {@link #querySharedIndex(ID, int, SharedIndexQuery)}
   */
  <Key, Value, Result, E extends Exception> Result querySharedIndex(@NotNull ID<Key, Value> indexId,
                                                                    int chunkId,
                                                                    @NotNull SharedIndexQuery<Key, Value, Result, E> query)
    throws E;

  /**
   * Tries to map a given {@link IndexedFile} to a shared index entry. The implementation
   * may access file content for that if it's available (See {@link FileContent}).
   *
   * @return hash_id = (internal_hash_id | chunk_id) if {@param hash} is present in some chunk,
   * otherwise returns {@link FileContentHashIndexExtension#NULL_HASH_ID}.
   */
  long tryEnumerateContentHash(@NotNull IndexedFile content);

  /**
   * Downloads shared index synchronously by requested {@param descriptor}.
   * Chunk is only registered for a project if the {@param project} is not null,
   * otherwise, it'll do all preparations only.
   *
   * @return true if downloaded chunk has been attached
   */
  @NotNull
  DownloadIndexResult downloadChunk(@NotNull ChunkDescriptor descriptor,
                                    @Nullable Project project,
                                    @NotNull ProgressIndicator indicator);

  /**
   * Attaches an existing chared index chunk to the project
   * @return true on success, false if no such chunk, it is not compatible or any other failure
   */
  boolean attachExistingChunk(@NotNull Project project, @NotNull String chunkUniqueId);

  /**
   * Attaches existing (already downloaded) chunk to a project
   *
   * @return collection of successfully attached chunks
   */
  @NotNull
  Collection<ID<?, ?>> attachExistingChunk(int chunkId, @NotNull Project project);

  /**
   * Tests is a given {@link ChunkDescriptor#getChunkUniqueId()}
   * is available and compatible with the current IDE version
   */
  boolean isAvailableChunk(@NotNull String chunkUniqueId);

  void markOpenChunksAsExcluded(boolean keepBundledIndexes);

  @FunctionalInterface
  interface SharedIndexProcessor<K, V, E extends Exception> {
    boolean process(@NotNull UpdatableIndex<K, V, FileContent> index) throws E;
  }

  @FunctionalInterface
  interface SharedIndexQuery<K, V, R, E extends Exception> {
    R calculate(@NotNull UpdatableIndex<K, V, FileContent> index) throws E;
  }
}
