// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.intellij.indexing.shared.message.SharedIndexesBundle;
import com.intellij.indexing.shared.platform.api.AttachChunkResult;
import com.intellij.indexing.shared.platform.api.ChunkDescriptor;
import com.intellij.indexing.shared.platform.api.ChunkDownloader;
import com.intellij.indexing.shared.platform.api.MutableSharedIndexStorage;
import com.intellij.openapi.Disposable;
import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.application.PathManager;
import com.intellij.openapi.components.ServiceManager;
import com.intellij.openapi.diagnostic.ControlFlowException;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.progress.ProcessCanceledException;
import com.intellij.openapi.progress.ProgressIndicator;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.project.ProjectManager;
import com.intellij.openapi.project.ProjectManagerListener;
import com.intellij.openapi.util.Ref;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.util.ExceptionUtil;
import com.intellij.util.indexing.*;
import com.intellij.util.io.IOUtil;
import com.intellij.util.io.MappingFailedException;
import com.intellij.util.io.SimpleStringPersistentEnumerator;
import org.jetbrains.annotations.ApiStatus;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.*;
import java.util.concurrent.locks.ReadWriteLock;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.stream.Collectors;

@ApiStatus.Internal
public final class SharedIndexChunkConfigurationImpl implements SharedIndexChunkConfiguration, Disposable {
  private static final Logger LOG = Logger.getInstance(SharedIndexChunkConfigurationImpl.class);

  private final ReadWriteLock myConfigurationLock = new ReentrantReadWriteLock();
  @Nullable
  private FileBasedIndexInfrastructureExtension.InitializationResult myInitialized;

  private ChunkManager myChunkManager;
  private SimpleStringPersistentEnumerator myChunkDescriptorEnumerator;
  private Map<ID<?, ?>, Map<Integer, SharedIndex<?, ?>>> myChunkMap;
  private MutableSharedIndexStorage mySharedIndexStorage;
  private final Map<Project, Long> myProjectStructureStamps = Collections.synchronizedMap(new WeakHashMap<>());
  private final ExcludedChunkList myExcludedChunkList = new ExcludedChunkList(getSharedIndexConfigurationRoot());

  public SharedIndexChunkConfigurationImpl() {
    ApplicationManager.getApplication().getMessageBus().connect(this).subscribe(ProjectManager.TOPIC, new ProjectManagerListener() {
      @Override
      public void projectClosed(@NotNull Project project) {
        myConfigurationLock.writeLock().lock();
        try {
          if (myInitialized != null) {
            List<Integer> closedChunkIds = myChunkManager.disposeChunks(project);
            for (Map<Integer, SharedIndex<?, ?>> sharedIndexes : myChunkMap.values()) {
              for (int id : closedChunkIds) {
                IOUtil.closeSafe(LOG, sharedIndexes.remove(id));
              }
            }
          }
        }
        finally {
          myConfigurationLock.writeLock().unlock();
        }
      }
    });
  }

  @Override
  public void dispose() { }

  @Override
  public FileBasedIndexInfrastructureExtension.@NotNull InitializationResult initialize() {
    myConfigurationLock.writeLock().lock();
    try {
      if (myInitialized == null) {
        Ref<FileBasedIndexInfrastructureExtension.InitializationResult> result =
          Ref.create(FileBasedIndexInfrastructureExtension.InitializationResult.SUCCESSFULLY);

        myChunkManager = new ChunkManager();
        myChunkMap = new HashMap<>();
        myChunkDescriptorEnumerator = new SimpleStringPersistentEnumerator(getSharedIndexConfigurationRoot().resolve("descriptors"));

        try {
          mySharedIndexStorage = SharedIndexZipStorageProvider.INSTANCE
            .tryOpenOrResetIfFailed(() -> {
              LOG.info("Shared index chunk storage corrupted and will be rebuilt");
              result.set(FileBasedIndexInfrastructureExtension.InitializationResult.INDEX_REBUILD_REQUIRED);
            });
        }
        catch (Exception e) {
          LOG.error("Failed to initialize shared indexes infrastructure", e);
          result.set(FileBasedIndexInfrastructureExtension.InitializationResult.INDEX_REBUILD_REQUIRED);
          shutdown();
        }

        myInitialized = result.get();
      }

      return myInitialized;
    }
    finally {
      myConfigurationLock.writeLock().unlock();
    }
  }

  @Override
  public void shutdown() {
    myConfigurationLock.writeLock().lock();
    try {
      if (myInitialized != null) {
        Set<Integer> disposedChunkIds = myChunkManager.disposeChunks();
        myChunkManager = null;
        myChunkDescriptorEnumerator = null;
        shutDownChunks(myChunkMap, disposedChunkIds);
        myChunkMap = null;
        myProjectStructureStamps.clear();
        myInitialized = null;

        IOUtil.closeSafe(LOG, mySharedIndexStorage);
        mySharedIndexStorage = null;
      }
    }
    finally {
      myConfigurationLock.writeLock().unlock();
    }

    LOG.trace("Shared index chunks shutdown was performed successful");
  }

  @Override
  public boolean hasAnyChunk() {
    myConfigurationLock.readLock().lock();
    try {
      if (myInitialized == null) return false;
      return mySharedIndexStorage.getSharedIndexStats().getChunks() > 0
             || !myChunkDescriptorEnumerator.isEmpty()
             || myChunkManager.getOpenChunkIds().size() > 0;
    } finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  @Override
  public boolean isSharedIndexAcceptable(@NotNull ID<?, ?> indexId,
                                         int chunkId,
                                         @NotNull IndexedFile indexedFile) {
    Boolean result = querySharedIndex(indexId, chunkId, index -> {
      if (index instanceof HashBasedMapReduceIndex) {
        return ((HashBasedMapReduceIndex<?, ?>)index).acceptsFile(indexedFile);
      }
      if (index instanceof EmptyIndex) {
        // TODO probably it should be with input filter too
        return true;
      }
      return false;
    });
    return Boolean.TRUE.equals(result);
  }

  @Override
  public <Key, Value> boolean hasSharedIndex(@NotNull ID<Key, Value> indexId, int chunkId) {
    if (chunkId == FileContentHashIndexExtension.NULL_CHUNK_ID) return false;
    myConfigurationLock.readLock().lock();
    try {
      if (myInitialized == null) {
        return false;
      }
      Map<Integer, SharedIndex<?, ?>> map = myChunkMap.get(indexId);
      return map != null && map.containsKey(chunkId);
    }
    finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  @Override
  public <Key, Value, E extends Exception> boolean processSharedIndexes(@NotNull ID<Key, Value> indexId,
                                                                        @NotNull SharedIndexProcessor<Key, Value, E> processor)
    throws E {
    myConfigurationLock.readLock().lock();
    try {
      if (myInitialized == null) {
        return true;
      }
      Map<Integer, SharedIndex<?, ?>> map = myChunkMap.get(indexId);
      if (map == null) return true;
      for (SharedIndex<?, ?> sharedIndex : map.values()) {
        //noinspection unchecked
        if (!processor.process(((SharedIndex<Key, Value>)sharedIndex).getIndex())) {
          return false;
        }
      }
      return true;
    }
    finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  @Override
  @Nullable
  public <Key, Value, Result, E extends Exception> Result querySharedIndex(@NotNull ID<Key, Value> indexId,
                                                                           int chunkId,
                                                                           @NotNull SharedIndexQuery<Key, Value, Result, E> query)
    throws E {
    myConfigurationLock.readLock().lock();
    try {
      if (myInitialized == null) {
        return null;
      }
      Map<Integer, SharedIndex<?, ?>> chunkToIndex = myChunkMap.get(indexId);
      if (chunkToIndex == null) return null;
      SharedIndex<?, ?> sharedIndex = chunkToIndex.get(chunkId);
      if (sharedIndex == null) return null;
      //noinspection unchecked
      return query.calculate((UpdatableIndex<Key, Value, FileContent>)sharedIndex.getIndex());
    }
    finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  @Override
  public long tryEnumerateContentHash(@NotNull IndexedFile indexedFile) {
    if (indexedFile instanceof HashSuppliedIndexedFile) {
      return ((HashSuppliedIndexedFile)indexedFile).getHashId();
    }

    myConfigurationLock.readLock().lock();
    try {
      if (myInitialized == null) return FileContentHashIndexExtension.NULL_HASH_ID;
      return myChunkManager.tryEnumerateContentHash(indexedFile);
    } catch (IOException t) {
      LOG.debug("Failed to compute hash for " + indexedFile + ". " + t.getMessage(), t);
      return FileContentHashIndexExtension.NULL_HASH_ID;
    }
    finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  @NotNull
  @Override
  public DownloadIndexResult downloadChunk(@NotNull ChunkDescriptor descriptor,
                                           @Nullable Project project,
                                           @NotNull ProgressIndicator indicator) {
    initialize();

    String chunkUniqueId = descriptor.getChunkUniqueId();

    Boolean isDownloadNeeded = isDownloadNeeded(chunkUniqueId);
    if (isDownloadNeeded == null) return DownloadIndexResult.FAILED;

    final DownloadIndexResult outcome;
    if (isDownloadNeeded) {
      indicator.pushState();
      try {
        indicator.setIndeterminate(true);
        if (!doDownloadChunk(chunkUniqueId, descriptor, indicator, project)) {
          return DownloadIndexResult.FAILED;
        }
        outcome = DownloadIndexResult.JUST_DOWNLOADED;
      }
      finally {
        indicator.popState();
      }
    }
    else {
      outcome = DownloadIndexResult.ALREADY_EXISTS;
    }

    if (project != null) {
      AttachChunkResult attachResult = openChunkForProject(chunkUniqueId, project, indicator);
      if (attachResult == null || !attachResult.isSuccess()) {
        return DownloadIndexResult.FAILED;
      }
    }

    return outcome;
  }

  @Nullable
  private Boolean isDownloadNeeded(@NotNull String chunkUniqueId) {
    try {
      myConfigurationLock.readLock().lock();
      if (myInitialized == null) return null;
      return !mySharedIndexStorage.containsChunk(chunkUniqueId);
    }
    finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  private boolean doDownloadChunk(@NotNull String chunkUniqueId, @NotNull ChunkDownloader descriptor, @NotNull ProgressIndicator indicator, @Nullable Project project) {
    Path chunkTempFile = downloadChunkToTempFile(chunkUniqueId, descriptor, indicator, project);
    if (chunkTempFile == null) return false;

    try {
      indicator.setText(SharedIndexesBundle.message("configuring.shared.indexes"));
      myConfigurationLock.writeLock().lock();
      try {
        if (myInitialized == null) return false;
        long startTime = System.currentTimeMillis();
        mySharedIndexStorage.appendChunk(chunkTempFile, chunkUniqueId);
        LOG.info("Chunk " + chunkUniqueId + " is appended into the store in " + (System.currentTimeMillis() - startTime) + " ms");
      }
      catch (Throwable e) {
        if (e instanceof ControlFlowException) ExceptionUtil.rethrow(e);
        if (e instanceof IOException || e instanceof MappingFailedException) myExcludedChunkList.markChunksAsExcluded(chunkUniqueId);
        LOG.error("Failed to attach shared index " + descriptor + ". " + e.getMessage(), e);
        return false;
      }
      finally {
        myConfigurationLock.writeLock().unlock();
      }

      return true;
    }
    finally {
      FileUtil.delete(chunkTempFile.toFile());
    }
  }

  @Nullable
  private static Path downloadChunkToTempFile(@NotNull String chunkUniqueId,
                                              @NotNull ChunkDownloader descriptor,
                                              @NotNull ProgressIndicator indicator,
                                              @Nullable Project project) {
    Path chunkTempDir = getSharedIndexConfigurationRoot().resolve("downloading");
    Path chunkTempFile = chunkTempDir.resolve(chunkUniqueId + "_" + System.currentTimeMillis() + "_temp.zip");
    try {
      Files.createDirectories(chunkTempDir);

      long startTime = System.currentTimeMillis();
      if (!descriptor.downloadChunk(chunkTempFile, project, indicator)) {
        LOG.debug("Failed to download shared index " + descriptor + ".");
        return null;
      }
      LOG.info("Chunk " + chunkUniqueId + " is downloaded in " + (System.currentTimeMillis() - startTime) + " ms");
    }
    catch (Throwable e) {
      if (e instanceof ControlFlowException) ExceptionUtil.rethrow(e);
      LOG.warn("Failed to download shared index " + descriptor + ". " + e.getMessage(), e);
      return null;
    }
    return chunkTempFile;
  }

  @Override
  public boolean attachExistingChunk(@NotNull Project project, @NotNull String chunkUniqueId) {
    AttachChunkResult result = openChunkForProject(chunkUniqueId, project, null);
    return result != null && result.isSuccess();
  }

  @Nullable
  private AttachChunkResult openChunkForProject(@NotNull String chunkUniqueId,
                                                @NotNull Project project,
                                                @Nullable ProgressIndicator indicator) {

    //we need to make sure FileBasedIndex and it's extensions are ready
    ((FileBasedIndexImpl)FileBasedIndex.getInstance()).loadIndexes();
    ((FileBasedIndexImpl)FileBasedIndex.getInstance()).waitUntilIndicesAreInitialized();

    myConfigurationLock.writeLock().lock();
    AttachChunkResult attachChunkResult;
    try {
      if (myInitialized == null) return null;
      int chunkEnumeratedId = myChunkDescriptorEnumerator.enumerate(chunkUniqueId);

      if (indicator != null) {
        indicator.setText(SharedIndexesBundle.message("configuring.shared.indexes"));
        indicator.setIndeterminate(true);
      }

      attachChunkResult = registerChunk(chunkEnumeratedId, chunkUniqueId, project);
    }
    catch (Throwable e) {
      if (e instanceof ControlFlowException) ExceptionUtil.rethrow(e);
      if (e instanceof IOException || e instanceof MappingFailedException) myExcludedChunkList.markChunksAsExcluded(chunkUniqueId);
      LOG.error("Failed to register shared indexes chunk: " + chunkUniqueId + ". " + e.getMessage(), e);
      return null;
    }
    finally {
      myConfigurationLock.writeLock().unlock();
    }
    String kind = guessKindFromChunkUniqueId(chunkUniqueId);
    SharedIndexesFusCollector.INSTANCE.reportIndexAttached(project, kind, chunkUniqueId, attachChunkResult);
    return attachChunkResult;
  }

  private static @NotNull String guessKindFromChunkUniqueId(@NotNull String chunkUniqueId) {
    // We do not store kinds of attached shared indexes in persistent structures. As can be seen from `ChunkDescriptor.getKind` the kind can
    // be extracted from the 'chunkUniqueId', which is constructed as "${request.kind}-$sha256-${indexVersion.weakVersionHash}"
    // (see SharedIndexLookupResult.getChunkUniqueId)
    String kind = StringUtil.substringBefore(chunkUniqueId, "-");
    return StringUtil.isEmpty(kind) ? "unknown" : kind;
  }

  @Override
  @Nullable
  public String getChunkDebugName(int chunkId) {
    return getChunkUniqueId(chunkId);
  }

  @Nullable
  private String getChunkUniqueId(int chunkId) {
    return myChunkDescriptorEnumerator.valueOf((short)chunkId);
  }

  @Override
  public @NotNull Collection<ID<?, ?>> attachExistingChunk(int chunkId, @NotNull Project project) {
    LOG.assertTrue(chunkId != FileContentHashIndexExtension.NULL_CHUNK_ID);
    String chunkUniqueId = getChunkUniqueId(chunkId);
    return chunkUniqueId == null ? Collections.emptyList() : attachExistingIndex(chunkUniqueId, chunkId, project);
  }

  @NotNull
  private List<ID<?, ?>> attachExistingIndex(@NotNull String chunkUniqueId, int chunkId, @NotNull Project project) {
    try {
      AttachChunkResult result = registerChunk(chunkId, chunkUniqueId, project);
      return result.getMatchingFbIndexes();
    }
    catch (Throwable e) {
      if (e instanceof ControlFlowException) ExceptionUtil.rethrow(e);
      SharedIndexStorageUtil.logAttachError("Failed to attach a known chunkId = " + chunkId + ". " + e.getMessage(), e);
      return Collections.emptyList();
    }
  }

  @Override
  public boolean isAvailableChunk(@NotNull String chunkUniqueId) {
    initialize();
    myConfigurationLock.readLock().lock();
    try {
      if (myInitialized == null) return false;
      return mySharedIndexStorage.isCompatibleChunk(chunkUniqueId);
    } finally {
      myConfigurationLock.readLock().unlock();
    }
  }

  @Override
  public void markOpenChunksAsExcluded(boolean keepBundledIndexes) {
    List<String> chunkNames;
    myConfigurationLock.readLock().lock();
    try {
      chunkNames = myChunkManager
        .getOpenChunkIds()
        .stream()
        .map(chunkId -> {
          try {
            return getChunkUniqueId(chunkId);
          }
          catch (Exception e) {
            LOG.error(e);
            return null;
          }
        })
        .filter(Objects::nonNull)
        .collect(Collectors.toList());
    }
    finally {
      myConfigurationLock.readLock().unlock();
    }

    if (keepBundledIndexes) {
      List<String> bundledChunkIds = ServiceManager.getService(BundledSharedIndexesResolver.class).listChunkIds();
      chunkNames.removeAll(bundledChunkIds);
    }

    myExcludedChunkList.markChunksAsExcluded(chunkNames);
  }

  @NotNull
  private AttachChunkResult registerChunk(int chunkId, @NotNull String chunkUniqueId, @NotNull Project project) throws IOException {
    if (myExcludedChunkList.isChunkExcluded(chunkUniqueId)) return AttachChunkResult.Incompatible.INSTANCE;

    myConfigurationLock.writeLock().lock();
    try {
      if (myInitialized == null) {
        throw new ProcessCanceledException();
      }

      AttachChunkResult indexAttachResult = myChunkManager.getIndexAttachResult(chunkId);
      if (indexAttachResult != null) {
        if (!myChunkManager.isProjectAttachedToChunk(chunkId, project)) {
          myChunkManager.attachProjectToChunk(chunkId, project);
        }
        return indexAttachResult;
      }

      AttachChunkResult indexAttachChunkResult = openFileBasedIndexChunk(chunkId, chunkUniqueId);
      if (!(indexAttachChunkResult instanceof AttachChunkResult.Success)) {
        myExcludedChunkList.markChunksAsExcluded(chunkUniqueId);
        LOG.warn("Chunk " + chunkUniqueId + " is incompatible and is not registered for project '" + project.getName() + "'");
        return indexAttachChunkResult;
      }

      AttachChunkResult.Success successResult = (AttachChunkResult.Success)indexAttachChunkResult;
      myChunkManager.attachChunk(chunkId, successResult.getChunkRoot(), indexAttachChunkResult);
      myChunkManager.attachProjectToChunk(chunkId, project);

      for (SharedIndex<?, ?> index : successResult.getAllMatchingIndexes()) {
        myChunkMap.computeIfAbsent(index.getIndexName(), __ -> new HashMap<>()).putIfAbsent(index.getChunkId(), index);
      }
      LOG.info("Chunk " + chunkUniqueId + " is registered for project '" + project.getName() + ": " + indexAttachChunkResult.toLogMessage());
      return indexAttachChunkResult;
    }
    finally {
      myConfigurationLock.writeLock().unlock();
    }
  }

  @NotNull
  private AttachChunkResult openFileBasedIndexChunk(int chunkId, @NotNull String chunkUniqueId) throws IOException {
    var chunkRoot = mySharedIndexStorage.getChunkRoot(chunkUniqueId);
    if (chunkRoot == null) {
      // We might use a separate result type for not-found chunks, but for now it is enough.
      return AttachChunkResult.Incompatible.INSTANCE;
    }

    var sharedIndexStats = mySharedIndexStorage.getSharedIndexStats();
    return SharedIndexStorageUtil.openFileBasedIndexChunks(chunkId, chunkRoot, sharedIndexStats);
  }

  static @NotNull Path getSharedIndexConfigurationRoot() {
    Path root = PathManager.getIndexRoot().resolve("shared_indexes");
    if (!Files.exists(root)) {
      try {
        Files.createDirectories(root);
      }
      catch (IOException e) {
        throw new RuntimeException(e);
      }
    }
    return root;
  }

  private static void shutDownChunks(@NotNull Map<ID<?, ?>, Map<Integer, SharedIndex<?, ?>>> myChunkMap,
                                     @NotNull Set<Integer> expectedChunkIds) {
    for (Map<Integer, SharedIndex<?, ?>> chunksForId : myChunkMap.values()) {
      for (Map.Entry<Integer, SharedIndex<?, ?>> entry : chunksForId.entrySet()) {
        Integer chunkId = entry.getKey();

        if (!expectedChunkIds.contains(chunkId)) {
          LOG.error("Inconsistent shared index chunk configuration: chunkId " + chunkId + " can't be found among " + expectedChunkIds);
        }

        SharedIndex<?, ?> index = entry.getValue();
        IOUtil.closeSafe(LOG, index);
      }
    }
  }
}
