package com.intellij.indexing.shared.platform.impl;

import com.intellij.openapi.project.Project;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.util.SystemProperties;
import com.intellij.util.indexing.*;
import it.unimi.dsi.fastutil.ints.IntOpenHashSet;
import it.unimi.dsi.fastutil.ints.IntSet;
import org.jetbrains.annotations.NotNull;

import java.util.Collection;
import java.util.HashSet;
import java.util.Objects;
import java.util.Set;

final class SharedIndexFileIndexingStatusProcessor implements FileBasedIndexInfrastructureExtension.FileIndexingStatusProcessor {
  private static final boolean SHARED_INDEX_USES_METADATA_HASHES = SystemProperties
    .getBooleanProperty("shared.index.uses.metadata.hashes", true);

  private final @NotNull IntSet myAttachedChunks = new IntOpenHashSet();
  private final @NotNull Set<IndexChunk> myInvalidatedChunkIndexes = new HashSet<>();
  private final @NotNull FileBasedIndexImpl myFileBasedIndex = ((FileBasedIndexImpl)FileBasedIndex.getInstance());
  private final @NotNull Project myProject;

  SharedIndexFileIndexingStatusProcessor(@NotNull Project project) {
    myProject = project;
  }

  @Override
  public boolean hasIndexForFile(@NotNull VirtualFile file, int inputId, @NotNull FileBasedIndexExtension<?, ?> extension) {
    if (!extension.dependsOnFileContent()) {
      return false;
    }
    int chunkId = FileContentHashIndex.getFileContentHashIndex(extension.getName()).getAssociatedChunkId(inputId, file);
    return chunkId != FileContentHashIndexExtension.NULL_CHUNK_ID &&
           SharedIndexChunkConfiguration.getInstance().hasSharedIndex(extension.getName(), chunkId);
  }

  @Override
  public boolean shouldProcessUpToDateFiles() {
    return SharedIndexChunkConfiguration.getInstance().hasAnyChunk();
  }

  @Override
  public boolean processUpToDateFile(@NotNull IndexedFile file, int inputId, @NotNull ID<?, ?> indexId) {
    int chunkId = FileContentHashIndex.getFileContentHashIndex(indexId).getAssociatedChunkId(inputId, file.getFile());
    if (chunkId == FileContentHashIndexExtension.NULL_CHUNK_ID) return true;

    boolean shouldAttach;
    synchronized (myAttachedChunks) {
      shouldAttach = myAttachedChunks.add(chunkId);
    }
    if (shouldAttach) {
      Collection<ID<?, ?>> attachedChunkIndexIds = SharedIndexChunkConfiguration.getInstance().attachExistingChunk(chunkId, myProject);
      Collection<ID<?, ?>> allContentDependentIndexes = ((FileBasedIndexImpl)FileBasedIndex.getInstance()).getContentDependentIndexes();

      synchronized (myInvalidatedChunkIndexes) {
        for (ID<?, ?> id : allContentDependentIndexes) {
          if (!attachedChunkIndexIds.contains(id)) {
            myInvalidatedChunkIndexes.add(new IndexChunk(chunkId, id));
          }
        }
      }
    }

    boolean isInvalidatedChunk;
    IndexChunk chunk = new IndexChunk(chunkId, indexId);
    synchronized (myInvalidatedChunkIndexes) {
      isInvalidatedChunk = myInvalidatedChunkIndexes.contains(chunk);
    }

    boolean isSharedIndexAcceptable = SharedIndexChunkConfiguration.getInstance().isSharedIndexAcceptable(indexId, chunkId, file);
    if (isInvalidatedChunk || !isSharedIndexAcceptable) {
      FileContentHashIndex.getFileContentHashIndex(indexId).mapInputAndPrepareUpdate(inputId, null).compute();
      myFileBasedIndex.dropNontrivialIndexedStates(inputId, indexId);
      return false;
    }

    return true;
  }

  @Override
  public boolean tryIndexFileWithoutContent(@NotNull IndexedFile file, int inputId, @NotNull ID<?, ?> indexId) {
    return SHARED_INDEX_USES_METADATA_HASHES &&
           ((CompositeInvertedIndex<?, ?>)myFileBasedIndex.getIndex(indexId)).tryIndexWithoutContent(inputId, file);
  }

  private static class IndexChunk {
    private final int myChunkId;
    private final ID<?, ?> myIndexId;

    private IndexChunk(int id, ID<?, ?> indexId) {
      myChunkId = id;
      myIndexId = indexId;
    }

    @Override
    public boolean equals(Object o) {
      if (this == o) return true;
      if (o == null || getClass() != o.getClass()) return false;

      IndexChunk chunk = (IndexChunk)o;

      if (myChunkId != chunk.myChunkId) return false;
      if (!Objects.equals(myIndexId, chunk.myIndexId)) return false;

      return true;
    }

    @Override
    public int hashCode() {
      int result = myChunkId;
      result = 31 * result + (myIndexId != null ? myIndexId.hashCode() : 0);
      return result;
    }
  }
}
