// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.io.FileUtil;
import com.intellij.openapi.util.io.FileUtilRt;
import com.intellij.psi.stubs.StubIndexKey;
import com.intellij.util.indexing.*;
import com.intellij.util.io.DataInputOutputUtil;
import org.jetbrains.annotations.NotNull;

import java.io.DataInputStream;
import java.io.DataOutputStream;
import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Path;

public final class SharedIndexInfrastructure implements FileBasedIndexInfrastructureExtension {
  public static final int SHARED_INDEX_INFRASTRUCTURE_VERSION = 6;

  private static final int REBUILD_COUNTER = 1;
  private static final int NULL_VERSION = -1;
  private static final String SHARED_INDEX_INFRASTRUCTURE_VERSION_FILE_NAME = "version";

  private static final Logger LOG = Logger.getInstance(SharedIndexInfrastructure.class);

  @Override
  public FileIndexingStatusProcessor createFileIndexingStatusProcessor(@NotNull Project project) {
    return new SharedIndexFileIndexingStatusProcessor(project);
  }

  @Override
  public <K, V> UpdatableIndex<K, V, FileContent> combineIndex(@NotNull FileBasedIndexExtension<K, V> indexExtension,
                                                               @NotNull UpdatableIndex<K, V, FileContent> baseIndex) throws IOException {
    return indexExtension.dependsOnFileContent() ? new CompositeInvertedIndex<>(indexExtension.getName(), baseIndex) : null;
  }

  @Override
  public void onFileBasedIndexVersionChanged(@NotNull ID<?, ?> indexId) {
    //TODO
  }

  @Override
  public void onStubIndexVersionChanged(@NotNull StubIndexKey<?, ?> indexId) {
    //TODO
  }

  @NotNull
  @Override
  public synchronized InitializationResult initialize() {
    //TODO replace with proper invalidation after version changed
    InitializationResult versionReadResult;
    int version = readPersistedVersion();

    if (version == NULL_VERSION && !Files.exists(SharedIndexChunkConfigurationImpl.getSharedIndexConfigurationRoot())) {
      getPersistVersion();
      versionReadResult = InitializationResult.SUCCESSFULLY;
    }
    else if (version != getVersion()) {
      LOG.info("Shared index version updated from " + version + " to " + getVersion());
      versionReadResult = InitializationResult.INDEX_REBUILD_REQUIRED;
      try {
        resetPersistentState();
      }
      catch (Exception e) {
        LOG.error(e);
      }
    }
    else {
      versionReadResult = InitializationResult.SUCCESSFULLY;
    }

    InitializationResult initializationResult = SharedIndexChunkConfiguration.getInstance().initialize();

    return (versionReadResult == InitializationResult.INDEX_REBUILD_REQUIRED ||
            initializationResult == InitializationResult.INDEX_REBUILD_REQUIRED)
           ? InitializationResult.INDEX_REBUILD_REQUIRED
           : InitializationResult.SUCCESSFULLY;
  }

  @Override
  public void resetPersistentState() {
    FileUtil.deleteWithRenaming(SharedIndexChunkConfigurationImpl.getSharedIndexConfigurationRoot().toFile());
    getPersistVersion();
    LOG.info("Shared index persistent data has been cleared");
  }

  @Override
  public void resetPersistentState(@NotNull ID<?, ?> indexId) {
    FileUtil.deleteWithRenaming(FileContentHashIndex.getFileContentHashIndexDir(HashIndexKey.createHashIndexKey(indexId)).toFile());
  }

  @Override
  public synchronized void shutdown() {
    runSafe(() -> {
      SharedIndexChunkConfiguration sharedIndexChunkConfiguration = SharedIndexChunkConfiguration.getInstanceIfCreated();
      if (sharedIndexChunkConfiguration != null) {
        sharedIndexChunkConfiguration.shutdown();
      }
    });
  }

  @Override
  public int getVersion() {
    return SHARED_INDEX_INFRASTRUCTURE_VERSION + REBUILD_COUNTER;
  }

  private void getPersistVersion() {
    persistVersion(getVersion());
  }

  private static void persistVersion(int version) {
    try {
      Files.createDirectories(getVersionFile().getParent());
      try (DataOutputStream os = FileUtilRt.doIOOperation(lastAttempt -> {
        try {
          return new DataOutputStream(Files.newOutputStream(getVersionFile()));
        }
        catch (IOException ex) {
          if (lastAttempt) throw ex;
          return null;
        }
      })) {
        assert os != null;
        DataInputOutputUtil.writeINT(os, version);
      }
    }
    catch (IOException e) {
      LOG.error(e);
    }
  }

  private static int readPersistedVersion() {
    Path versionFile = getVersionFile();
    try (DataInputStream in = new DataInputStream(Files.newInputStream(versionFile))) {
      return DataInputOutputUtil.readINT(in);
    }
    catch (IOException ignore) {
    }
    return NULL_VERSION;
  }

  @NotNull
  private static Path getVersionFile() {
    return SharedIndexChunkConfigurationImpl.getSharedIndexConfigurationRoot().resolve(SHARED_INDEX_INFRASTRUCTURE_VERSION_FILE_NAME);
  }

  private static void runSafe(Runnable r) {
    try {
      r.run();
    }
    catch (Exception e) {
      LOG.error(e);
    }
  }
}
