package com.intellij.indexing.shared.platform.impl

import com.google.common.collect.Maps
import com.intellij.indexing.shared.metadata.SharedIndexMetadata
import com.intellij.indexing.shared.platform.api.IdeIndexInfrastructureVersion
import com.intellij.indexing.shared.platform.api.MutableSharedIndexStorage
import com.intellij.indexing.shared.platform.api.SharedIndexStats
import com.intellij.indexing.shared.util.getFileOrDirectorySize
import com.intellij.indexing.shared.util.zipFs.UncompressedZipFileSystem
import com.intellij.indexing.shared.util.zipFs.Zip64Util
import com.intellij.openapi.diagnostic.Attachment
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProcessCanceledException
import com.intellij.openapi.util.io.FileUtil
import com.intellij.util.io.isDirectory
import org.jetbrains.annotations.VisibleForTesting
import java.io.IOException
import java.nio.file.*
import java.nio.file.attribute.BasicFileAttributes
import java.util.*
import java.util.concurrent.atomic.AtomicReference
import java.util.zip.ZipEntry
import kotlin.streams.toList

private val LOG = logger<SharedIndexMainZipStorage>()
/**
 * A Mutable shared index storage, usually located in `system/index/shared_index/shared-index.chunks`.
 */
class SharedIndexMainZipStorage @Throws(IOException::class) @VisibleForTesting constructor(oldStoragePath: Path,
                                                                                           private val storagePath: Path,
                                                                                           override val storageId: String)
  : MutableSharedIndexStorage {

  private val readSystem: UncompressedZipFileSystem
  private val myStats = AtomicReference<SharedIndexStats?>(null)
  private val myCompatibleChunksCache = mutableMapOf<String, Boolean>()

  init {
    if (Files.exists(oldStoragePath)) {
      if (Files.exists(storagePath)) throw IOException()
      FileUtil.rename(oldStoragePath.toFile(), storagePath.fileName.toString())
    }
    if (!Files.exists(storagePath)) {
      Zip64Util.openZip64File(storagePath, false).close()
    }
    readSystem = UncompressedZipFileSystem.create(storagePath)
    val stats = updateStats()
    LOG.info("Shared Indexes Storage is opened: " + stats.toLogMessage() + ", chunks: " + listChunks().sorted())
  }

  @Synchronized
  private fun updateStats(): SharedIndexStats {
    val result = try {
      val size = storagePath.getFileOrDirectorySize()
      val chunks = Files.list(readSystem.rootDirectory).use { stream -> stream.filter { it.isDirectory() }.count().toInt() }
      SharedIndexStats(size, chunks)
    }
    catch (t: Throwable) {
      //we'd better to avoid NPE later with stats
      LOG.info("Failed to compute shared indexes stats. ${t.message}", t)
      SharedIndexStats(0, 0)
    }

    myStats.set(result)
    return result
  }

  override fun getSharedIndexStats(): SharedIndexStats = myStats.get() ?: updateStats()

  @Synchronized
  override fun close() {
    readSystem.close()
  }

  @Synchronized
  override fun isCompatibleChunk(chunkId: String): Boolean = myCompatibleChunksCache.computeIfAbsent(chunkId) {
    try {
      val chunkRoot = getChunkRoot(chunkId) ?: return@computeIfAbsent false
      val chunkVersion = SharedIndexMetadata.readIndexesVersion(chunkRoot) ?: return@computeIfAbsent false
      // TODO: check there are enough compatible shared indexes in the chunk
      IdeIndexInfrastructureVersion.getIdeVersion().isSuitableVersion(chunkVersion.indexInfrastructureVersion)
    } catch (t: Throwable) {
      LOG.info("Failed to check compatibility of $chunkId in $storagePath")
      false
    }
  }

  private fun listChunks(): Collection<String> {
    return Files
      .list(readSystem.rootDirectory)
      .use { list ->
        list.filter { it.isDirectory() }
          .map { it.fileName.toString() }
          .toList()
          .toSortedSet()
      }
  }

  @Synchronized
  override fun getChunkRoot(chunkId: String): Path? {
    val chunkRoot = readSystem.getPath(chunkId)
    return if (Files.isDirectory(chunkRoot)) chunkRoot else null
  }

  @Throws(IOException::class)
  @Synchronized
  override fun appendChunk(chunkSourcePath: Path, chunkId: String) {
    if (!Files.exists(chunkSourcePath)) {
      throw NoSuchFileException(chunkSourcePath.toString())
    }

    if (containsChunk(chunkId)) {
      return
    }

    val sourceSnapshot = takeSourceChunkSnapshot(chunkSourcePath)

    try {
      val timestamp = System.currentTimeMillis()

      Zip64Util.openZip64File(storagePath, false).use { storageZipFile ->
        UncompressedZipFileSystem.create(chunkSourcePath).use { sourceFs ->
          Files.walkFileTree(sourceFs.rootDirectory, object : SimpleFileVisitor<Path>() {
            override fun visitFile(file: Path, attrs: BasicFileAttributes): FileVisitResult {
              if (!Files.isDirectory(file)) {
                val createdEntry = storageZipFile.getOrCreateEntry("$chunkId/$file")

                createdEntry.method = ZipEntry.STORED
                createdEntry.time = timestamp
                Files.newInputStream(file).use {
                  createdEntry.setDataFromStream(it)
                }

              }
              return FileVisitResult.CONTINUE
            }
          })
        }
      }
    }
    catch (e: ProcessCanceledException) {
      LOG.error("ProcessCanceledException should not be thrown there", Attachment("pce", e))
    }
    catch (e: Throwable) {
      LOG.error("Failed to append chunk", e)
      throw e
    }
    finally {
      try {
        readSystem.sync()
      }
      catch (e: Exception) {
        LOG.error("Failed to sync shared index storage", e)
        throw e
      }

      val chunkRoot = getChunkRoot(chunkId) ?: throw IOException(
        "Can't find root for added chunk $chunkId; available roots: ${listChunks()}")
      val storedSnapshot = takeStoredChunkSnapshot(chunkRoot)

      ensureSnapshotsEqual(sourceSnapshot, storedSnapshot)

      val stats = updateStats()
      LOG.info("Shared Indexes Storage is updated with $chunkId. Total: " + stats.toLogMessage())
    }
  }

  override fun toString(): String = storageId
}

@Throws(IOException::class)
private fun ensureSnapshotsEqual(expectedSnapshot: Map<String, Long>, actualSnapshot: Map<String, Long>) {
  val diff = Maps.difference(actualSnapshot, expectedSnapshot)
  val missedFiles = diff.entriesOnlyOnLeft()
  val differedFiles = diff.entriesDiffering()
  val superfluousFiles = diff.entriesOnlyOnRight()
  if (missedFiles.isEmpty() && differedFiles.isEmpty() && superfluousFiles.isEmpty()) return
  if (missedFiles.isNotEmpty()) {
    throw IOException("Shared index chunk files ${missedFiles.keys} were not copied to shared index central storage")
  }
  if (superfluousFiles.isNotEmpty()) {
    LOG.error("Superfluous files has been observed after chunk was copied ${superfluousFiles.keys}")
    return
  }
  if (differedFiles.isNotEmpty()) {
    throw IOException("Shared index chunk files ${differedFiles} corrupted")
  }
}

private fun takeSourceChunkSnapshot(chunkPath: Path): Map<String, Long> =
  UncompressedZipFileSystem.create(chunkPath).use { chunkFs ->
    doTakeSnapshotRecursively(chunkFs.rootDirectory)
  }


private fun takeStoredChunkSnapshot(chunkRoot: Path): Map<String, Long> {
  return doTakeSnapshotRecursively(chunkRoot)
}

private fun doTakeSnapshotRecursively(root: Path): Map<String, Long> {
  val snapshot = mutableMapOf<String, Long>()
  Files.walkFileTree(root, object : SimpleFileVisitor<Path>() {
    override fun visitFile(file: Path, attrs: BasicFileAttributes): FileVisitResult {
      if (!Files.isDirectory(file)) {
        val normalized = file.toAbsolutePath().toString().removePrefix(root.toString()).removePrefix("/")
        snapshot[normalized] = Files.size(file)
      }
      return FileVisitResult.CONTINUE
    }
  })
  return Collections.unmodifiableMap(snapshot)
}
