// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.platform.impl;

import com.google.common.annotations.VisibleForTesting;
import com.google.common.collect.Maps;
import com.google.common.collect.Streams;
import com.intellij.indexing.shared.metadata.SharedIndexMetadata;
import com.intellij.indexing.shared.platform.api.AttachChunkResult;
import com.intellij.indexing.shared.platform.api.IdeIndexInfrastructureVersion;
import com.intellij.indexing.shared.platform.api.IndexInfrastructureVersion;
import com.intellij.indexing.shared.platform.api.SharedIndexStats;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.psi.stubs.StubIndexExtension;
import com.intellij.psi.stubs.StubIndexImpl;
import com.intellij.psi.stubs.StubIndexKey;
import com.intellij.psi.stubs.StubUpdatingIndex;
import com.intellij.util.containers.ContainerUtil;
import com.intellij.util.indexing.FileBasedIndexExtension;
import com.intellij.util.indexing.ID;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.file.DirectoryStream;
import java.nio.file.Files;
import java.nio.file.Path;
import java.util.ArrayList;
import java.util.List;
import java.util.Objects;
import java.util.Set;
import java.util.concurrent.atomic.AtomicInteger;
import java.util.stream.Collectors;
import java.util.stream.Stream;

public final class SharedIndexStorageUtil {
  private static final Logger LOG = Logger.getInstance(SharedIndexStorageUtil.class);
  private static final AtomicInteger CORRUPTION_COUNTER = new AtomicInteger();
  @SuppressWarnings("StaticNonFinalField") @VisibleForTesting
  public static volatile boolean FAIL_ON_CORRUPTED_CHUNK_ATTACHMENT = true;

  /**
   * Opens all available indexes for all loaded {@link FileBasedIndexExtension}-s and {@link StubIndexExtension}-s
   */
  @NotNull
  static AttachChunkResult openFileBasedIndexChunks(int chunkId,
                                                    @NotNull Path chunkRoot,
                                                    @NotNull SharedIndexStats sharedIndexStats) throws IOException {
    var metadata = SharedIndexMetadata.readIndexesVersion(chunkRoot);
    if (metadata == null) {
      //TODO: cleanup outdated versions from the .zip
      return AttachChunkResult.Incompatible.INSTANCE;
    }

    var chunkVersion = metadata.getIndexInfrastructureVersion();
    IdeIndexInfrastructureVersion ideVersion = IdeIndexInfrastructureVersion.getIdeVersion();
    if (!ideVersion.isSuitableVersion(chunkVersion)) {
      //TODO: cleanup outdated versions from the .zip
      return AttachChunkResult.Incompatible.INSTANCE;
    }

    Set<String> suitableFbIndexes = Maps.filterEntries(chunkVersion.getFileBasedIndexVersions(), e -> {
      assert e != null;
      boolean areVersionsEqual = e.getValue().equals(ideVersion.getVersion().getFileBasedIndexVersions().get(e.getKey()));
      if (!StubUpdatingIndex.INDEX_ID.getName().equals(e.getKey())) {
        return areVersionsEqual;
      }
      return areVersionsEqual && ideVersion.isStubIndexVersionMatch(chunkVersion);
    }).keySet();

    Set<String> suitableStubIndexes = Maps.filterEntries(chunkVersion.getStubIndexVersions(), e -> {
      assert e != null;
      return e.getValue().equals(ideVersion.getVersion().getStubIndexVersions().get(e.getKey()));
    }).keySet();

    List<SharedIndex<?, ?>> matchingFbIndexes = new ArrayList<>();
    List<ID<?, ?>> mismatchingFbIndexes = new ArrayList<>();
    List<String> redundantFbIndexes = new ArrayList<>();

    List<SharedIndex<?, ?>> matchingStubIndexes = new ArrayList<>();
    List<StubIndexKey<?, ?>> mismatchingStubIndexes = new ArrayList<>();
    List<String> redundantStubIndexes = new ArrayList<>();

    //TODO filter stub index versions
    try (DirectoryStream<Path> directoryStream = Files.newDirectoryStream(chunkRoot)) {
      for (Path indexDir : directoryStream) {
        String indexId = indexDir.getFileName().toString();
        ID<?, ?> id = ID.findByName(indexId);
        //TODO filter stub index versions
        if (id == null) {
          redundantFbIndexes.add(indexId);
          continue;
        }
        if (suitableFbIndexes.contains(id.getName()) || id.equals(StubUpdatingIndex.INDEX_ID)) {
          FileBasedIndexExtension<?, ?> extension = findExtension(id);
          if (extension != null) {
            try {
              // noinspection unchecked,rawtypes
              matchingFbIndexes.add(new SharedIndex.Impl(chunkRoot,
                                                         id,
                                                         chunkId,
                                                         extension,
                                                         FileContentHashIndex.getFileContentHashIndex(id),
                                                         chunkVersion));
            } catch (Exception e) {
              logAttachError("failed to open shared index: extension = " + id + ", chunk = " + chunkRoot, e);
            }
          }
          if (id.equals(StubUpdatingIndex.INDEX_ID)) {
            StubChunks stubChunks = openStubIndexChunks(indexDir, chunkVersion, chunkId, suitableStubIndexes);
            matchingStubIndexes.addAll(stubChunks.matchingStubIndexes);
            mismatchingStubIndexes.addAll(stubChunks.mismatchingStubIndexes);
            redundantStubIndexes.addAll(stubChunks.redundantStubIndexes);
          }
        } else {
          mismatchingFbIndexes.add(id);
        }
      }
    }

    Set<String> emptyIndexNames = EmptyIndexEnumerator.readEmptyIndexes(chunkRoot);
    Set<String> emptyStubIndexNames = EmptyIndexEnumerator.readEmptyStubIndexes(chunkRoot);

    //noinspection IOResourceOpenedButNotSafelyClosed
    List<SharedIndex.Empty<?, Void>> matchingEmptyStubIndexes = emptyStubIndexNames
      .stream()
      .map(ID::findByName)
      .filter(id -> id instanceof StubIndexKey)
      .map(id -> findStubExtension(id))
      .filter(Objects::nonNull)
      .map(extension -> new SharedIndex.Empty<>(extension.getName(), chunkId))
      .collect(Collectors.toList());

    //noinspection IOResourceOpenedButNotSafelyClosed
    List<SharedIndex.Empty<Object, Object>> matchingEmptyFbIndexes = emptyIndexNames
      .stream()
      .map(ID::findByName)
      .filter(id -> Objects.nonNull(id) && suitableFbIndexes.contains(id.getName()))
      .map(id -> new SharedIndex.Empty<>(id, chunkId)).collect(Collectors.toList());

    return new AttachChunkResult.Success(
      chunkRoot,
      ContainerUtil.concat(matchingFbIndexes, matchingStubIndexes, matchingEmptyFbIndexes, matchingEmptyStubIndexes),
      Streams.concat(matchingFbIndexes.stream(), matchingEmptyFbIndexes.stream())
        .map(i -> i.getIndexName()).collect(Collectors.toList()),
      Streams.concat(matchingStubIndexes.stream(), matchingEmptyStubIndexes.stream())
        .map(i -> i.getIndexName()).collect(Collectors.toList()),
      mismatchingFbIndexes,
      mismatchingStubIndexes,
      redundantFbIndexes,
      redundantStubIndexes,
      sharedIndexStats
    );
  }

  private static class StubChunks {
    public final List<SharedIndex<?, ?>> matchingStubIndexes;
    public final List<StubIndexKey<?, ?>> mismatchingStubIndexes;
    public final List<String> redundantStubIndexes;

    private StubChunks(List<SharedIndex<?, ?>> matchingStubIndexes,
                       List<StubIndexKey<?, ?>> mismatchingStubIndexes,
                       List<String> redundantStubIndexes) {
      this.matchingStubIndexes = matchingStubIndexes;
      this.mismatchingStubIndexes = mismatchingStubIndexes;
      this.redundantStubIndexes = redundantStubIndexes;
    }
  }

  @NotNull
  private static StubChunks openStubIndexChunks(@NotNull Path stubIndexDir,
                                                @NotNull IndexInfrastructureVersion chunkVersion,
                                                int chunkId,
                                                @NotNull Set<String> suitableStubIndexes) throws IOException {
    List<SharedIndex<?, ?>> matchingStubIndexes = new ArrayList<>();
    List<StubIndexKey<?, ?>> mismatchingStubIndexes = new ArrayList<>();
    List<String> redundantStubIndexes = new ArrayList<>();

    try (Stream<Path> dirs = Files.list(stubIndexDir)) {
      dirs.forEach(stubIndexExDir -> {
        String stubIndexId = stubIndexExDir.getFileName().toString();
        ID<Object, Object> id = ID.findByName(stubIndexId);
        if (id instanceof StubIndexKey) {
          FileBasedIndexExtension<?, Void> extension = findStubExtension(id);
          if (extension != null) {
            try {
              @SuppressWarnings("unchecked")
              SharedIndex.Impl<?, ?> sharedIndex = new SharedIndex.Impl(stubIndexDir,
                                                                        extension.getName(),
                                                                        chunkId,
                                                                        extension,
                                                                        FileContentHashIndex.getFileContentHashIndex(extension.getName()),
                                                                        chunkVersion);
              matchingStubIndexes.add(sharedIndex);
            } catch (Exception e) {
              logAttachError("failed to open shared stub index: extension = " + id + ", chunk = " + stubIndexDir.getParent().getFileName().toString(), e);
            }
            // TODO: Actually, we don't check versions of stub indexes now... See the to-do above.
            if (!suitableStubIndexes.contains(stubIndexId)) {
              //noinspection rawtypes
              mismatchingStubIndexes.add(((StubIndexKey)id));
            }
          }
        }
        else {
          redundantStubIndexes.add(stubIndexId);
        }
      });
    }
    return new StubChunks(matchingStubIndexes, mismatchingStubIndexes, redundantStubIndexes);
  }

  @Nullable
  private static FileBasedIndexExtension<?, ?> findExtension(@Nullable ID<?, ?> id) {
    if (id == null) return null;
    return FileBasedIndexExtension.EXTENSION_POINT_NAME.findFirstSafe(ex -> ex.getName().equals(id));
  }

  @Nullable
  private static FileBasedIndexExtension<?, Void> findStubExtension(@Nullable ID<?, ?> id) {
    if (id == null) return null;
    StubIndexExtension<?, ?> stubEx = StubIndexExtension.EP_NAME.findFirstSafe(ex -> ex.getKey().equals(id));
    return stubEx == null ? null : StubIndexImpl.wrapStubIndexExtension(stubEx);
  }

  static void logAttachError(@NotNull String message, @NotNull Throwable error) {
    CORRUPTION_COUNTER.incrementAndGet();
    if (FAIL_ON_CORRUPTED_CHUNK_ATTACHMENT) {
      LOG.error(message, error);
    }
    else {
      LOG.info(message, error);
    }
  }

  @VisibleForTesting
  public static int getCorruptionCount() {
    return CORRUPTION_COUNTER.get();
  }
}
