package com.intellij.indexing.shared.platform.impl

import com.intellij.indexing.shared.platform.api.SharedIndexStorage
import com.intellij.openapi.application.ApplicationInfo
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.application.PathManager
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.diagnostic.thisLogger
import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.openapi.extensions.PluginAware
import com.intellij.openapi.extensions.PluginDescriptor
import com.intellij.openapi.project.Project
import com.intellij.openapi.startup.StartupActivity
import com.intellij.openapi.util.registry.Registry
import com.intellij.openapi.util.text.StringUtil
import com.intellij.util.io.isDirectory
import com.intellij.util.io.isFile
import com.intellij.util.io.size
import com.intellij.util.xmlb.annotations.Attribute
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import kotlin.streams.toList

private val EP_NAME = ExtensionPointName.create<BundledSharedIndexProvider>("com.intellij.sharedIndexBundled")

/**
 * Defines bundled shared index which is pre-loaded for every project.
 *
 * On the first start, IDE will attach the given shared index
 * to the shared indexes storage (locally). It will not re-attach
 * this shared index (e.g. after update), if the filename is the
 * same as it was. There is no need to update the index, when indexes
 * are already constructed.
 *
 * Change the filename in your plugin code to force the update,
 * but please do not change it too often.
 *
 * For example, it is quite OK to use SDK versions in the name,
 * and change it when version update.
 *
 * @see PreSharedIndexesGenerator in build scripts
 */
class BundledSharedIndexProvider : PluginAware {
  /**
   * Specify product home relative path
   * to the directory with a prebuilt shared index (the .ijx file)/
   * This directory is an --output parameter of the `dump-shared-index` command
   * @see com.intellij.indexing.shared.generate.MainGenerateArgs
   */
  @JvmField
  @Attribute("productPath")
  var productPath : String? = null

  /**
   * Specify plugin-home related relative path
   * to the directory with a prebuilt shared index (the .ijx file).
   * This directory is an --output parameter of the `dump-shared-index` command
   * @see com.intellij.indexing.shared.generate.MainGenerateArgs
   */
  @JvmField
  @Attribute("pluginPath")
  var pluginPath : String? = null

  private var pluginDescriptor: PluginDescriptor? = null

  override fun setPluginDescriptor(pluginDescriptor: PluginDescriptor) {
    this.pluginDescriptor = pluginDescriptor
  }

  val chunkId : String? get() {
    val (chunkId, _) = resolvedChunkIdAndPath ?: return null
    return chunkId
  }

  fun openChunkStorage(): SingleChunkReadonlyZipStorage? {
    val (chunkId, resolvedPath) = resolvedChunkIdAndPath ?: return null
    return try {
      SingleChunkReadonlyZipStorage(chunkId, resolvedPath, "bundled-${chunkId}")
    }
    catch (t: Throwable) {
      thisLogger().warn("Failed to load bundled shared index: $resolvedPath. ${t.message}", t)
      return null
    }
  }

  private val resolvedChunkIdAndPath : Pair<String, Path>? by lazy {
    try {
      val productPath = productPath
      val pluginPath = pluginPath

      if (!((productPath != null) xor (pluginPath != null))) {
        thisLogger().error("Only one parameter must be used in $this")
        return@lazy null
      }

      val resolvedPath: Path? = when {
        productPath != null -> Paths.get(PathManager.getHomePath()).resolve(productPath)
        pluginPath != null -> pluginDescriptor?.pluginPath?.resolve(pluginPath)
        else -> null
      }

      if (resolvedPath == null) {
        return@lazy null
      }

      if (!resolvedPath.isDirectory()) {
        thisLogger().warn("Bundled shared index is not found at: $resolvedPath")
        return@lazy null
      }

      val sharedIndexExtension = ".ijx"
      val indexes = Files.list(resolvedPath).use { stream ->
        stream.filter { it.isFile() && it.fileName.toString().endsWith(sharedIndexExtension) }.toList()
      }

      if (indexes.isEmpty()) {
        thisLogger().warn("Failed to find any bundled shared indexes at: $resolvedPath")
        return@lazy null
      }

      val theBundledIndex = indexes.singleOrNull() ?: run {
        thisLogger().warn("There are several shared index files found at $resolvedPath: $indexes. Please make sure IntelliJ was installed into an empty folder")
        return@lazy null
      }

      thisLogger().debug("Resolved bundled shared shared index: $theBundledIndex (${StringUtil.formatFileSize(theBundledIndex.size())}")
      val chunkId = run {
        val name = theBundledIndex.fileName.toString().removeSuffix(sharedIndexExtension)
        val source = pluginDescriptor?.pluginId?.idString ?: "global"
        // we include build into the chunk ID to make sure every IntelliJ build comes with new ID
        // and avoids from internal hashId clashes
        "bundled-$name-$source-${ApplicationInfo.getInstance().build}"
      }

      chunkId to theBundledIndex
    } catch (t: Throwable) {
      if (t is ControlFlowException) throw t
      thisLogger().warn("Failed to locate bundled shared index for ${pluginDescriptor} and ${pluginPath}. ${t.message}", t)
      null
    }
  }
}



@Service //application
class BundledSharedIndexesResolver {
  private fun isEnabled() = !ApplicationManager.getApplication().isUnitTestMode && Registry.`is`("shared.indexes.bundled")

  fun openBundledStorages(): List<SharedIndexStorage> {
    if (!isEnabled()) return listOf()

    return EP_NAME.iterable.mapNotNull {
      runCatching { it.openChunkStorage() }.getOrNull()
    }
  }

  fun listChunkIds() : List<String> {
    if (!isEnabled()) return listOf()
    return EP_NAME.iterable.mapNotNull {
      runCatching { it.chunkId }.getOrNull()
    }
  }
}

class BundledSharedIndexPostStartupActivity : StartupActivity.DumbAware {
  private val LOG = thisLogger()

  override fun runActivity(project: Project) {
    val preBuiltIndexes = service<BundledSharedIndexesResolver>().listChunkIds()
    if (preBuiltIndexes.isEmpty()) return

    val attached = mutableListOf<String>()
    val skipped = mutableListOf<String>()

    for (chunkId in preBuiltIndexes) {
      when (SharedIndexChunkConfiguration.getInstance().attachExistingChunk(project, chunkId)) {
        true -> attached.add(chunkId)
        false -> skipped.add(chunkId)
      }
    }

    if (attached.isNotEmpty()) {
      LOG.info("Attached ${attached.size} pre-built shared indexes: ${attached.joinToString(", ")}")
    }
    if (skipped.isNotEmpty()) {
      LOG.warn("Skipped ${skipped.size} pre-built shared indexes: ${skipped.joinToString(", ")}")
    }
  }
}
