package com.intellij.indexing.shared.platform.impl

import com.intellij.indexing.shared.platform.api.MutableSharedIndexStorage
import com.intellij.indexing.shared.platform.api.SharedIndexStats
import com.intellij.indexing.shared.platform.api.SharedIndexStorage
import com.intellij.openapi.diagnostic.logger
import java.nio.file.Path

class SharedIndexZipCompositeStorage(
  private val mutableStorage: MutableSharedIndexStorage,
  private val providedStorages: List<SharedIndexStorage>
) : MutableSharedIndexStorage {
  private val LOG = logger<SharedIndexZipCompositeStorage>()

  override fun toString() = buildString {
    append("CompositeStorage{")
    foreachStorage { append(it.toString()).append(", ") }
    append("}")
  }

  private inline fun foreachStorage(action: (SharedIndexStorage) -> Unit) {
    action(mutableStorage)
    providedStorages.forEach { action(it) }
  }

  override fun getChunkRoot(chunkId: String): Path? {
    foreachStorage {
      val chunkRoot = it.getChunkRoot(chunkId)
      if (chunkRoot != null) {
        return chunkRoot
      }
    }
    return null
  }

  override fun appendChunk(chunkSourcePath: Path, chunkId: String) {
    mutableStorage.appendChunk(chunkSourcePath, chunkId)
  }

  override val storageId: String
    get() = providedStorages.union(listOf(mutableStorage)).joinToString(", ") { it.storageId }

  override fun getSharedIndexStats(): SharedIndexStats {
    return providedStorages.fold(mutableStorage.getSharedIndexStats(), { acc, it -> acc + it.getSharedIndexStats() })
  }

  override fun isCompatibleChunk(chunkId: String): Boolean {
    foreachStorage {
      if (it.containsChunk(chunkId) && it.isCompatibleChunk(chunkId)) return true
    }
    return false
  }

  override fun close() {
    foreachStorage {
      try {
        it.close()
      } catch (t: Throwable) {
        LOG.warn("Failed to close $it. ${t.message}", t)
      }
    }
  }
}
