package com.intellij.indexing.shared.platform.impl

import com.intellij.indexing.shared.metadata.SharedIndexMetadata
import com.intellij.indexing.shared.platform.api.*
import com.intellij.indexing.shared.util.zipFs.UncompressedZipFileSystem
import com.intellij.openapi.components.service
import com.intellij.openapi.util.io.FileUtil
import com.intellij.util.io.IOUtil
import java.io.IOException
import java.nio.file.Files
import java.nio.file.Path

object SharedIndexZipStorageProvider : SharedIndexStorageProvider<MutableSharedIndexStorage> {
  private const val MAIN_STORAGE_OLD_NAME = "chunks.zip"
  const val MAIN_STORAGE_NAME = "shared-index.chunks"

  @Throws(IOException::class)
  override fun tryOpenOrResetIfFailed(corruptionHandler: Runnable): MutableSharedIndexStorage {
    /// NOTE. it is not possible to allow mounting any random shared index here without
    /// a risk of indexes inconsistency (when a file receives a wrong index data from a shared index)
    ///
    /// When a shared index is mounted, we build a mapping from every
    /// VirtualFile to a pair of <chunkId, hashId>, where `chunkId` is the internal
    /// enumerated int from the respective `uniqueChunkId` string of the index, and
    /// the `hashId` is the shared-index specific enumeration from a VirtualFile hash to int.
    ///
    /// Thus we need to make sure that every used `chunkId` will keep having the same
    /// contents.
    ///
    /// To fix that, we may assign a new chunkId, but it would cost us a user-visible
    /// re-index (file hashes has to be re-computed). We would like to avoid that
    /// side-effect. So far it means the shared index storage is a closed world, and
    /// it is not possible to allow a pluggable add-ons to it.

    val oldStoragePath = SharedIndexChunkConfigurationImpl.getSharedIndexConfigurationRoot().resolve(MAIN_STORAGE_OLD_NAME)
    val newStoragePath = SharedIndexChunkConfigurationImpl.getSharedIndexConfigurationRoot().resolve(MAIN_STORAGE_NAME)
    val mutableStorage = IOUtil.openCleanOrResetBroken({
                                                         SharedIndexMainZipStorage(oldStoragePath, newStoragePath, "shared index central storage")
                                                       },
                                                       {
                                                         corruptionHandler.run()
                                                         if (Files.exists(oldStoragePath)) {
                                                           FileUtil.delete(oldStoragePath)
                                                         }
                                                         if (Files.exists(newStoragePath)) {
                                                           FileUtil.delete(newStoragePath)
                                                         }
                                                       })

    val bundledIndexes = service<BundledSharedIndexesResolver>().openBundledStorages()
    if (bundledIndexes.isEmpty()) return mutableStorage
    return SharedIndexZipCompositeStorage(mutableStorage, bundledIndexes)
  }
}

class SingleChunkReadonlyZipStorage(
  val chunkId: String,
  private val storagePath: Path,
  override val storageId: String
) : SharedIndexStorage {
  private val readSystem = UncompressedZipFileSystem.create(storagePath)
  private val chunkRoot get() = readSystem.rootDirectory
  private val metadata = SharedIndexMetadata.readIndexesVersion(chunkRoot)

  override fun getSharedIndexStats() = SharedIndexStats(Files.size(storagePath), 1)
  override fun isCompatibleChunk(chunkId: String) = chunkId == this.chunkId
                                                    && metadata != null
                                                    && IdeIndexInfrastructureVersion.getIdeVersion().isSuitableVersion(metadata.indexInfrastructureVersion)

  override fun getChunkRoot(chunkId: String): Path? = if (chunkId == this.chunkId) chunkRoot else null

  override fun close() {
    readSystem.close()
  }

  override fun toString() = "SingleChunkReadonlyZipStorage(chunkId='$chunkId', storagePath=$storagePath)"
}