package com.intellij.indexing.shared.platform.impl

import com.intellij.indexing.shared.platform.api.AttachChunkResult
import com.intellij.internal.statistic.eventLog.EventLogGroup
import com.intellij.internal.statistic.eventLog.FeatureUsageData
import com.intellij.internal.statistic.eventLog.events.EventFields
import com.intellij.internal.statistic.eventLog.events.PrimitiveEventField
import com.intellij.internal.statistic.service.fus.collectors.CounterUsagesCollector
import com.intellij.openapi.project.Project
import com.intellij.util.indexing.ID

object SharedIndexesFusCollector : CounterUsagesCollector() {
  override fun getGroup() = GROUP

  private val GROUP = EventLogGroup("shared.indexes", 4)

  // Common fields
  private val kind = EventFields.Enum<Kind>("kind") { it.name.toLowerCase() }

  private val indexId = object : PrimitiveEventField<String>() {
    override val name = "indexId"

    override fun addData(fuData: FeatureUsageData, value: String) {
      fuData.addAnonymizedValue(name, value)
    }

    override val validationRule: List<String>
      get() = listOf("{regexp#hash}")
  }

  // Fields for "download.finished" event
  private val finishType = EventFields.Enum<FinishType>("finishType") { it.name.toLowerCase() }
  private val downloadTime = EventFields.Long("downloadTime") // Milliseconds
  private val packedSize = EventFields.Long("packedSize") // Bytes
  private val unpackedSize = EventFields.Long("unpackedSize") // Bytes
  private val averageDownloadSpeed = EventFields.Long("averageDownloadSpeed") // KBytes/s

  // Fields for "attached" event
  private val matchingFbIndexes = EventFields.Int("matchingFbIndexes")
  private val matchingStubIndexes = EventFields.Int("matchingStubIndexes")

  private val mismatchingFbIndexes = EventFields.StringListValidatedByCustomRule("mismatchingFbIndexes", "index_id")
  private val mismatchingStubIndexes = EventFields.StringListValidatedByCustomRule("mismatchingStubIndexes", "index_id")

  private val redundantFbIndexes = EventFields.Int("redundantFbIndexes")
  private val redundantStubIndexes = EventFields.Int("redundantStubIndexes")

  private val numberOfLocalSharedIndexes = EventFields.Int("numberOfLocalSharedIndexes")
  private val totalSizeOfLocalSharedIndexes = EventFields.Long("totalSizeOfLocalSharedIndexes") // Bytes

  private val localIndexLoaded = GROUP.registerVarargEvent(
    "local.index.loaded",
    kind,
    indexId,
    unpackedSize
  )

  private val downloadStarted = GROUP.registerVarargEvent(
    "download.started",
    kind,
    indexId
  )

  private val downloadFinished = GROUP.registerVarargEvent(
    "download.finished",
    kind,
    indexId,
    finishType,
    downloadTime,
    packedSize,
    unpackedSize,
    averageDownloadSpeed
  )

  private val attached = GROUP.registerVarargEvent(
    "attached",
    kind,
    indexId,
    matchingFbIndexes,
    matchingStubIndexes,
    mismatchingFbIndexes,
    mismatchingStubIndexes,
    redundantFbIndexes,
    redundantStubIndexes,
    numberOfLocalSharedIndexes,
    totalSizeOfLocalSharedIndexes
  ) // Warning: FUS allows max 10 fields. Consider splitting it if you want to add a new field.

  private val attachFailedIncompatible = GROUP.registerVarargEvent(
    "attach.failed.incompatible",
    kind,
    indexId
  )

  fun reportLocalIndexLoaded(
    project: Project?,
    kind: String,
    indexId: String,
    unpackedSizeBytes: Long
  ) {
    localIndexLoaded.log(
      project,
      this.kind.with(Kind.fromString(kind)),
      this.indexId.with(indexId),
      this.unpackedSize.with(unpackedSizeBytes)
    )
  }

  fun reportIndexDownloadStarted(
    project: Project?,
    kind: String,
    indexId: String
  ) {
    downloadStarted.log(
      project,
      this.kind.with(Kind.fromString(kind)),
      this.indexId.with(indexId)
    )
  }

  fun reportIndexDownloadFinished(
    project: Project?,
    finishType: FinishType,
    kind: String,
    indexId: String,
    downloadTimeMs: Long,
    packedSizeBytes: Long,
    unpackedSizeBytes: Long,
    averageDownloadSpeedKBytesPerSec: Long
  ) {
    downloadFinished.log(
      project,
      this.kind.with(Kind.fromString(kind)),
      this.indexId.with(indexId),
      this.finishType.with(finishType),
      this.downloadTime.with(downloadTimeMs),
      this.packedSize.with(packedSizeBytes),
      this.unpackedSize.with(unpackedSizeBytes),
      this.averageDownloadSpeed.with(averageDownloadSpeedKBytesPerSec)
    )
  }

  fun reportIndexAttached(
    project: Project?,
    kind: String,
    indexId: String,
    attachChunkResult: AttachChunkResult
  ) {
    return when (attachChunkResult) {
      is AttachChunkResult.Success -> {
        attached.log(
          project,
          this.kind.with(Kind.fromString(kind)),
          this.indexId.with(indexId),
          this.matchingFbIndexes.with(attachChunkResult.matchingFbIndexes.size),
          this.matchingStubIndexes.with(attachChunkResult.matchingStubIndexes.size),
          this.mismatchingFbIndexes.with(limitMismatchingIndexesList(attachChunkResult.mismatchingFbIndexes).map { it.name }),
          this.mismatchingStubIndexes.with(limitMismatchingIndexesList(attachChunkResult.mismatchingStubIndexes).map { it.name }),
          this.redundantFbIndexes.with(attachChunkResult.redundantFbIndexes.size),
          this.redundantStubIndexes.with(attachChunkResult.redundantStubIndexes.size),
          this.numberOfLocalSharedIndexes.with(attachChunkResult.sharedIndexStats.chunks),
          this.totalSizeOfLocalSharedIndexes.with(attachChunkResult.sharedIndexStats.sizeInBytes)
        )
      }
      is AttachChunkResult.Incompatible -> {
        attachFailedIncompatible.log(
          project,
          this.kind.with(Kind.fromString(kind)),
          this.indexId.with(indexId)
        )
      }
    }
  }

  /**
   * To not overflow the FUS logs with very long string lists.
   * Although in general we do not expect more than this number of mismatching indexes.
   */
  private const val INDEX_LIST_LIMIT = 50

  private fun limitMismatchingIndexesList(list: List<ID<*, *>>): List<ID<*, *>> {
    if (list.size <= INDEX_LIST_LIMIT) {
      return list
    }
    val randomIds = list.toMutableList()
    randomIds.shuffle()
    return randomIds.take(INDEX_LIST_LIMIT)
  }

  enum class FinishType {
    SUCCESS,
    CANCELLED,
    ERROR
  }

  private enum class Kind {
    PROJECT,
    JDK,
    MVN,
    OTHER;

    companion object {
      @JvmStatic
      fun fromString(kind: String) = when (kind) {
        "project" -> PROJECT
        "jdk" -> JDK
        "mvn" -> MVN
        else -> OTHER
      }
    }
  }
}