package com.intellij.indexing.shared.platform.impl;

import com.intellij.indexing.shared.platform.api.IndexInfrastructureVersion;
import com.intellij.lang.Language;
import com.intellij.lang.LanguageParserDefinitions;
import com.intellij.lang.ParserDefinition;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.fileTypes.FileType;
import com.intellij.openapi.fileTypes.LanguageFileType;
import com.intellij.openapi.util.Disposer;
import com.intellij.psi.stubs.*;
import com.intellij.psi.templateLanguages.TemplateLanguage;
import com.intellij.psi.tree.IFileElementType;
import com.intellij.psi.tree.IStubFileElementType;
import com.intellij.util.containers.ConcurrentFactoryMap;
import com.intellij.util.indexing.FileBasedIndexExtension;
import com.intellij.util.indexing.ID;
import com.intellij.util.indexing.IndexedFile;
import com.intellij.util.io.DataInputOutputUtil;
import com.intellij.util.io.SimpleStringPersistentEnumerator;
import org.jetbrains.annotations.NotNull;

import java.io.DataInput;
import java.io.DataOutput;
import java.io.IOException;
import java.nio.file.Path;
import java.util.Map;
import java.util.Set;

public class StubSharedIndexExtension implements SharedIndexExtension<Integer, SerializedStubTree> {
  private static final Logger LOG = Logger.getInstance(StubSharedIndexExtension.class);

  private volatile StubUpdatingIndex myStubUpdatingIndex;
  private volatile SerializationManagerImpl mySerializationManager;
  private volatile SimpleStringPersistentEnumerator myIndexIdEnumerator;
  private volatile StubBuilderVersionFilter myStubBuilderVersionFilter;

  @Override
  public void openResources(@NotNull Path indexRoot, boolean readOnly, @NotNull IndexInfrastructureVersion chunkVersion) {
    mySerializationManager = new SerializationManagerImpl(getStubSerializerNamesStorageFile(indexRoot), readOnly);
    myIndexIdEnumerator = new SimpleStringPersistentEnumerator(indexRoot.resolve("stubIndexIds.txt"));
    myStubUpdatingIndex = new StubUpdatingIndex(new StubForwardIndexExternalizer<Void>() {
      @Override
      protected Void createStubIndexKeySerializationState(@NotNull DataOutput out, @NotNull Set<StubIndexKey<?, ?>> set) {
        return null;
      }

      @Override
      protected void writeStubIndexKey(@NotNull DataOutput out, @NotNull StubIndexKey key, Void unused) throws IOException {
        DataInputOutputUtil.writeINT(out, myIndexIdEnumerator.enumerate(key.getName()));
      }

      @Override
      protected Void createStubIndexKeySerializationState(@NotNull DataInput input, int stubIndexKeyCount) {
        return null;
      }

      @Override
      protected ID<?, ?> readStubIndexKey(@NotNull DataInput input, Void unused) throws IOException {
        int indexId = DataInputOutputUtil.readINT(input);
        if (indexId > Short.MAX_VALUE) {
          throw new IOException("invalid index id");
        }
        String name = myIndexIdEnumerator.valueOf((short)indexId);
        if (name == null) {
          throw new IOException("name can't be found for id");
        }
        ID<Object, Object> id = ID.findByName(name);
        if (!(id instanceof StubIndexKey)) {
          // TODO get rid of StubSharedIndexExtension.StubBuilderVersionFilter.isJsOrDialect
          LOG.warn("id with name `" + name + "`, enumerated id " + indexId +
                   " is not " + (id == null ? "found" : "a StubIndexKey"));
          return null;
        }
        return id;
      }
    }, mySerializationManager);
    myStubBuilderVersionFilter = readOnly ? new StubBuilderVersionFilter(chunkVersion) : null;
  }

  @Override
  public boolean acceptsInputFile(@NotNull IndexedFile file) {
    return myStubBuilderVersionFilter.acceptsFileType(file.getFileType());
  }

  @Override
  public void closeResources() {
    Disposer.dispose(mySerializationManager);
  }

  @Override
  public @NotNull FileBasedIndexExtension<Integer, SerializedStubTree> getFileBasedIndexExtension() {
    return myStubUpdatingIndex;
  }

  @Override
  public int getVersion() {
    return 1;
  }

  @NotNull
  private static Path getStubSerializerNamesStorageFile(@NotNull Path stubIndexesRoot) {
    return stubIndexesRoot.resolve("serializerNames").resolve("names");
  }

  private static class StubBuilderVersionFilter {
    private final @NotNull Map<String, String> myVersions;
    private final @NotNull Map<FileType, Boolean> myResolutionMap;

    private StubBuilderVersionFilter(@NotNull IndexInfrastructureVersion chunkVersion) {
      myVersions = chunkVersion.getStubFileElementTypeVersions();
      myResolutionMap = ConcurrentFactoryMap.createMap(fileType -> {
        if (!(fileType instanceof LanguageFileType)) return Boolean.TRUE;
        Language language = ((LanguageFileType)fileType).getLanguage();
        if (isJsOrDialect(language)) return true;
        ParserDefinition parserDefinition = LanguageParserDefinitions.INSTANCE.forLanguage(language);
        if (parserDefinition == null) return Boolean.TRUE;
        IFileElementType fileNodeType = parserDefinition.getFileNodeType();
        if (!(fileNodeType instanceof IStubFileElementType)) return Boolean.TRUE;
        String key = getStubFileElementTypeKey(fileNodeType);
        int version = getStubFileElementBaseVersion((IStubFileElementType<?>)fileNodeType);
        return Boolean.valueOf(String.valueOf(version).equals(myVersions.get(key)));
      });
    }

    boolean acceptsFileType(@NotNull FileType fileType) {
      return myResolutionMap.get(fileType).booleanValue();
    }

    // workaround JS prebuilt indexes.
    private static boolean isJsOrDialect(@NotNull Language language) {
      Language baseLanguage = language.getBaseLanguage();
      if (baseLanguage != null) {
        language = baseLanguage;
      }
      return language.getID().equals("JavaScript");
    }
  }

  @NotNull
  public static String getStubFileElementTypeKey(@NotNull IFileElementType fileNodeType) {
    String externalId = ((IStubFileElementType<?>)fileNodeType).getExternalId();
    return externalId + ":" + fileNodeType.getLanguage().getID();
  }

  public static int getStubFileElementBaseVersion(@NotNull IStubFileElementType<?> fileNodeType) {
    int stubVersion = fileNodeType.getStubVersion();
    return fileNodeType.getLanguage() instanceof TemplateLanguage
           ? stubVersion - IStubFileElementType.getTemplateStubBaseVersion()
           : stubVersion;
  }
}
