package com.intellij.indexing.shared.util

import com.google.common.hash.Hashing
import com.intellij.indexing.shared.download.SharedIndexCompression
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.util.io.FileUtil
import com.intellij.util.io.HttpRequests
import com.intellij.util.io.createDirectories
import com.intellij.util.io.outputStream
import java.io.InputStream
import java.nio.file.Files
import java.nio.file.Path
import kotlin.math.min

/**
 * Downloads file by [url] to [targetFile].
 * Checks expected size ([expectedSize]) and hash ([expectedSha256]).
 * @return size of the target file
 */
fun downloadUrlWithValidation(
  url: String,
  targetFile: Path,
  expectedSize: Long,
  expectedSha256: String,
  headers: Map<String, String>,
  compression: SharedIndexCompression,
  indicator: ProgressIndicator
): Long {
  require(expectedSize > 0L)

  indicator.isIndeterminate = false
  indicator.fraction = 0.0

  try {
    HttpRequests.request(url)
      .productNameAsUserAgent()
      .tuner { tuner ->
        headers.forEach { (k, v) -> tuner.addRequestProperty(k, v) }
      }
      .accept("application/octet-stream, application/xz")
      .throwStatusCodeException(true)
      .connect {
        targetFile.parent.createDirectories()

        val countingStream = object : InputStream() {
          val networkStream = it.inputStream.buffered(1024 * 1024)
          val hasher = Hashing.sha256().newHasher()
          var bytesRead = 0L

          override fun read(): Int {
            val tmp = ByteArray(1)
            if (read(tmp, 0, 1) != 1) return -1
            return tmp[0].toInt().and(0xff)
          }

          override fun read(b: ByteArray, off: Int, len: Int): Int {
            val sz = networkStream.read(b, off, len)
            if (sz > 0) {
              bytesRead += sz
              hasher.putBytes(b, off, sz)
            }
            return sz
          }

          override fun close() {
            networkStream.close()
          }
        }

        compression.createUnpackingStream(countingStream).use { input ->
          targetFile.outputStream().buffered(1024 * 1024).use { output ->
            val buff = ByteArray(1024 * 1024)
            while (true) {
              indicator.checkCanceled()
              val off = input.read(buff)
              if (off < 0) break
              indicator.checkCanceled()
              if (off <= 0) continue

              output.write(buff, 0, off)

              if (countingStream.bytesRead > expectedSize) {
                throw RuntimeException("Connection is not closed at byte ${countingStream.bytesRead}. Expected $expectedSize bytes")
              }
              indicator.fraction = min(1.0, countingStream.bytesRead.toDouble() / expectedSize)
            }
          }
        }

        if (countingStream.bytesRead < expectedSize) {
          throw RuntimeException("Connection closed at byte ${countingStream.bytesRead}. Expected $expectedSize bytes")
        }

        val actualHashCode = countingStream.hasher.hash().toString()
        if (!actualHashCode.equals(expectedSha256, ignoreCase = true)) {
          throw RuntimeException("SHA-256 checksums does not match. Actual value is $actualHashCode, expected $expectedSha256")
        }
      }
    return Files.size(targetFile)
  }
  catch (e: Throwable) {
    FileUtil.delete(targetFile)
    throw e
  }
}