package com.intellij.indexing.shared.util

import com.intellij.indexing.shared.message.SharedIndexesBundle
import com.intellij.indexing.shared.util.zipFs.Zip64Util
import com.intellij.openapi.diagnostic.ControlFlowException
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.util.io.FileUtil
import com.intellij.openapi.util.text.StringUtil
import com.intellij.util.io.*
import java.io.File
import java.io.OutputStream
import java.nio.file.Files
import java.nio.file.Path
import java.util.zip.ZipEntry
import kotlin.streams.toList

fun Path.sha256(): String {
  val digest = DigestUtil.sha256()
  DigestUtil.updateContentHash(digest, this)
  return StringUtil.toHexString(digest.digest())
}

fun String.sha256(): String {
  val digest = DigestUtil.sha256()
  digest.update(this.toByteArray())
  return StringUtil.toHexString(digest.digest())
}

fun recreateDir(file: Path): Path {
  FileUtil.delete(file)
  Files.createDirectories(file)
  return file
}

fun File.totalSize(): Long {
  if (isFile) return length()
  if (isDirectory) {
    return Files.walk(this.toPath()).use { stream ->
      stream.mapToLong {
        when {
          it.isFile() -> java.lang.Long.max(it.sizeOrNull(), 0L)
          else -> 0L
        }
      }.sum()
    }
  }
  return 0
}

fun <Y : Any> runAndCatchNotNull(errorMessage: String, action: () -> Y?): Y {
  try {
    return action() ?: error("<null> was returned!")
  }
  catch (t: Throwable) {
    throw Error("Failed to $errorMessage. ${t.message}", t)
  }
}


fun uncompressedZip(
  dir: Path,
  targetFile: Path,
  indicator: ProgressIndicator
) {
  indicator.pushState()
  indicator.text = SharedIndexesBundle.message("progress.text.zipping.index.pack")
  try {
    Zip64Util.openZip64File(targetFile, false).use { file ->
      indicator.text2 = SharedIndexesBundle.message("progress.text2.scanning.files")
      val allFiles = Files.walk(dir).use { stream -> stream.filter { !Files.isDirectory(it) }.toList() }
      indicator.isIndeterminate = false
      var count = 0.0
      for (p in allFiles) {
        indicator.checkCanceled()
        val relativePath = FileUtil.toSystemIndependentName(dir.relativize(p).toString())
        try {
          indicator.fraction = count++ / allFiles.size
          val entry = file.getOrCreateEntry(relativePath)
          entry.method = ZipEntry.STORED
          entry.setDataFromPath(p)
        }
        catch (e: Exception) {
          throw RuntimeException("Failed to add " + relativePath + " entry to the target archive. " + e.message, e)
        }
      }
    }
  }
  catch (e: Throwable) {
    FileUtil.delete(targetFile)
    if (e is ControlFlowException) throw e
    throw RuntimeException("Failed to generate indexes archive at " + targetFile + ". " + e.message, e)
  }
  finally {
    indicator.popState()
  }
}

fun compressSharedIndex(file: Path, output: Path, indicator: ProgressIndicator, compressorStream: (OutputStream) -> OutputStream) {
  indicator.pushState()
  val bufferSize = 8 * 1024 * 1024
  indicator.text = SharedIndexesBundle.message("packing.to.xz")
  indicator.isIndeterminate = false
  val total = Files.size(file)
  var completed = 0L
  try {
    output.outputStream().buffered(bufferSize).use { outputStream ->
      compressorStream(outputStream).use { output ->
        file.inputStream().use { inputStream ->
          val buffer = ByteArray(bufferSize)
          while (true) {
            val sz = inputStream.read(buffer)
            if (sz <= 0) break
            completed += sz
            output.write(buffer, 0, sz)
            indicator.fraction = completed.toDouble() / total
            indicator.checkCanceled()
          }
        }
      }
    }
  }
  catch (e: Throwable) {
    FileUtil.delete(output)
    if (e is ControlFlowException) throw e
    throw RuntimeException("Failed to generate compressed shared index from $file to $output. ${e.message}", e)
  }
  finally {
    indicator.popState()
  }
}

fun Path.getFileOrDirectorySize(): Long {
  if (!Files.exists(this)) {
    return 0
  }
  return if (!Files.isDirectory(this)) {
    Files.size(this)
  }
  else Files.walk(this).use { it
    .filter { p: Path -> p.toFile().isFile }
    .mapToLong { p: Path -> p.toFile().length() }
    .sum()
  }
}
