// Copyright 2000-2019 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.intellij.indexing.shared.util.zipFs;

import org.jetbrains.annotations.NotNull;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.MappedByteBuffer;
import java.nio.channels.FileChannel;
import java.nio.channels.FileLock;
import java.nio.channels.ReadableByteChannel;
import java.nio.channels.WritableByteChannel;

public abstract class FileBlockReadOnlyFileChannel extends FileChannel {
  private final String myEntryName;
  @NotNull
  private final TraceableFileChannel myUnderlying;
  private final long myStartOffset;
  private final long myEndOffset;
  private final long mySize;

  // TODO duplicates?
  private volatile long myGlobalPosition;
  private volatile long myLocalPosition;

  public FileBlockReadOnlyFileChannel(@NotNull String entryName,
                                      @NotNull TraceableFileChannel underlying,
                                      long startOffset,
                                      long size) {
    myEntryName = entryName;
    myUnderlying = underlying;

    myStartOffset = startOffset;
    myEndOffset = startOffset + size;
    mySize = size;

    myGlobalPosition = myStartOffset;
    myLocalPosition = 0;
  }

  @Override
  public String toString() {
    return "FileBlockReadOnlyFileChannel{" +
           "myEntryName='" + myEntryName + '\'' +
           ", myStartOffset=" + myStartOffset +
           ", mySize=" + mySize +
           '}';
  }

  @Override
  public int read(ByteBuffer dst) throws IOException {
    int read = read(dst, myLocalPosition);
    if (read > 0) {
      position(myLocalPosition + read);
    }
    return read;
  }

  @Override
  public long read(ByteBuffer[] dsts, int offset, int length) throws IOException {
    throw new IOException("Not implemented for " + this);
  }

  @Override
  public int read(ByteBuffer dst, long position) throws IOException {
    long globalPosition = myStartOffset + position;
    if (position >= mySize) {
      return -1;
    }
    int read = myUnderlying.read(dst, globalPosition);
    if (read <= 0) return read;
    if (read > mySize - position) {
      return (int)(mySize - position);
    }
    return read;
  }

  @Override
  public int write(ByteBuffer src) throws IOException {
    throw createWriteNotSupportedException();
  }

  @Override
  public long write(ByteBuffer[] srcs, int offset, int length) throws IOException {
    throw createWriteNotSupportedException();
  }

  @Override
  public int write(ByteBuffer src, long position) throws IOException {
    throw createWriteNotSupportedException();
  }

  @Override
  public long position() {
    return myGlobalPosition - myStartOffset;
  }

  @Override
  public FileChannel position(long newPosition) {
    myGlobalPosition = myStartOffset + newPosition;
    myLocalPosition = newPosition;
    assert myGlobalPosition <= myEndOffset;
    assert myLocalPosition <= mySize;
    return this;
  }

  @Override
  public long size() {
    return mySize;
  }

  @Override
  public FileChannel truncate(long size) {
    throw new UnsupportedOperationException();
  }

  @Override
  public void force(boolean metaData) {
    // do nothing
  }

  @Override
  public long transferTo(long position, long count, WritableByteChannel target) {
    throw new UnsupportedOperationException("Implement it for " + this);
  }

  @Override
  public long transferFrom(ReadableByteChannel src, long position, long count) {
    throw new UnsupportedOperationException("transferFrom is not implemented for " + this);
  }

  @Override
  public MappedByteBuffer map(MapMode mode, long position, long size) {
    throw new UnsupportedOperationException("map is not implemented for " + this);
  }

  @Override
  public FileLock lock(long position, long size, boolean shared) {
    return null;
  }

  @Override
  public FileLock tryLock(long position, long size, boolean shared) {
    return null;
  }

  @Override
  protected void implCloseChannel() throws IOException {
    // do nothing
  }

  @NotNull
  private IOException createWriteNotSupportedException() {
    return new IOException("Write operation is not supported for " + this);
  }
}
