package com.intellij.indexing.shared.util.zipFs;

import com.intellij.openapi.diagnostic.Attachment;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.util.io.FileChannelUtil;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.io.IOException;
import java.nio.ByteBuffer;
import java.nio.channels.ClosedChannelException;
import java.nio.channels.FileChannel;
import java.nio.file.Path;
import java.nio.file.StandardOpenOption;

class TraceableFileChannel {
  private static final Logger LOG = Logger.getInstance(TraceableFileChannel.class);

  @NotNull
  private final FileChannel myFileChannel;
  @NotNull
  private final Throwable myOpenTrace;
  @Nullable
  private volatile Throwable myCloseTrace;

  TraceableFileChannel(@NotNull Path path, StandardOpenOption... options) throws IOException {
    myFileChannel = FileChannelUtil.unInterruptible(FileChannel.open(path, options));
    myOpenTrace = new Throwable(path + " channel creation trace");
  }

  int read(@NotNull ByteBuffer buffer, long position) throws IOException {
    try {
      return myFileChannel.read(buffer, position);
    }
    catch (ClosedChannelException e) {
      reportOpenAndCloseTraces(e);
      throw e;
    }
  }

  long size() throws IOException {
    try {
      return myFileChannel.size();
    }
    catch (ClosedChannelException e) {
      reportOpenAndCloseTraces(e);
      throw e;
    }
  }

  void close() throws IOException {
    try {
      myFileChannel.close();
    }
    finally {
      myCloseTrace = new Throwable();
    }
  }

  boolean isOpen() {
    return myFileChannel.isOpen();
  }

  private void reportOpenAndCloseTraces(ClosedChannelException e) {
    Throwable closeTrace = myCloseTrace;
    Attachment closeTraceAttachment = closeTrace == null
                                      ? new Attachment("closed externally!", new Throwable())
                                      : new Attachment("close trace", closeTrace);
    LOG.error("Channel is already closed ", e,
              new Attachment("open trace", myOpenTrace),
              closeTraceAttachment);
  }
}
