package com.intellij.licensecommon.client.impl;

import com.intellij.licensecommon.client.LicenseClientApplication;
import com.intellij.licensecommon.client.StartUpFailedException;
import com.intellij.licensecommon.license.LicenseBroadcastData;
import com.intellij.licensecommon.message.Message;
import com.intellij.licensecommon.message.MessageClient;
import com.intellij.licensecommon.message.MessageFactory;
import com.intellij.licensecommon.net.Address;
import com.intellij.licensecommon.net.Network;
import com.intellij.licensecommon.net.NetworkTarget;
import com.intellij.openapi.util.Pair;
import com.intellij.util.concurrency.Semaphore;
import org.jetbrains.annotations.NotNull;
import org.jetbrains.annotations.Nullable;

import java.util.concurrent.Future;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.atomic.AtomicReference;

public class ClientImpl implements NetworkTarget, MessageClient {
  private final Network myNetwork;
  private final MessageFactory myMessageFactory;
  private final ScheduledExecutorService myScheduler;
  private final LicenseClientApplication myApplication;
  private final AtomicReference<Pair<LicenseBroadcastData, Future>>
    myCurrentBroadcastData = new AtomicReference<>(null);

  private static final long SECOND = 1000L;
  private static final long MINUTE = 60 * SECOND;
  private static final long BROADCAST_PERIOD = 10 * MINUTE; // 10 minutes
  private static final long INITIAL_BROADCAST_PERIOD = 100;

  public ClientImpl(LicenseClientApplication application, Network network, MessageFactory messageFactory, ScheduledExecutorService scheduler) {
    myApplication = application;
    myNetwork = network;
    myMessageFactory = messageFactory;
    myScheduler = scheduler;
  }

  private void ensureNetworkStarted() throws StartUpFailedException {
    try {
      if (!myNetwork.isStarted()) {
        myNetwork.startWith(this);
      }
    }
    catch (Throwable t) {
      myNetwork.shutDown();
      throw new StartUpFailedException(t);
    }
  }

  public void broadcastLicenseData(@Nullable LicenseBroadcastData license) {
    if (license != null) {
      scheduleBroadcasts(license);
    }
    else {
      cancelBroadcastTask();
      myNetwork.shutDown();
    }
  }

  private void scheduleBroadcasts(@NotNull final LicenseBroadcastData broadcastData) {
    final BroadcastTask broadcastTask = new BroadcastTask(broadcastData);
    final ScheduledFuture<?> scheduledFuture = myScheduler.scheduleWithFixedDelay(
      broadcastTask, INITIAL_BROADCAST_PERIOD, BROADCAST_PERIOD, TimeUnit.MILLISECONDS
    );

    final Pair<LicenseBroadcastData, Future> previous = myCurrentBroadcastData.getAndSet(
      Pair.create(broadcastData, scheduledFuture)
    );

    broadcastTask.enable();

    if (previous != null) {
      previous.second.cancel(false);
    }
  }

  public void cancelBroadcastTask() {
    final Pair<LicenseBroadcastData, Future> task = myCurrentBroadcastData.getAndSet(null);
    if (task != null) {
      task.second.cancel(false);
    }
  }

  @Override
  @Nullable
  public LicenseBroadcastData getBroadcastData() {
    final Pair<LicenseBroadcastData, Future> pair = myCurrentBroadcastData.get();
    return Pair.getFirst(pair);
  }

  @Override
  public void duplicateLicenseDetected(Address from, LicenseBroadcastData conflictingData) {
    myApplication.duplicateLicenseDetected(from, conflictingData);
  }

  @Override
  public void receivedData(Message message, Address from) throws Exception {
    message.receivedAt(this, from);
  }

  private class BroadcastTask implements Runnable {
    private final LicenseBroadcastData myBroadcastData;
    private final Semaphore myEnabledSemaphore = new Semaphore();
    BroadcastTask(LicenseBroadcastData broadcastData) {
      myBroadcastData = broadcastData;
      myEnabledSemaphore.down();
    }

    public void enable() {
      myEnabledSemaphore.up();
    }

    @Override
    public void run() {
      try {
        ensureNetworkStarted();
        final Message msg = myMessageFactory.createBroadcastMessage(myBroadcastData);
        if (msg != null) {
          myEnabledSemaphore.waitFor();
          final Pair<LicenseBroadcastData, Future> currentData = myCurrentBroadcastData.get();
          if (currentData != null && myBroadcastData.equals(currentData.first)) {
            myNetwork.broadcast(msg);
          }
        }
      }
      catch(Throwable ignored) {
      }
    }
  }
}
