package com.intellij.licensecommon.license;

import com.intellij.licensecommon.util.InvalidLicenseKeyException;
import com.intellij.licensecommon.util.LicenseInfo;
import com.intellij.licensecommon.util.LicenseType;

import java.util.Date;

/**
 * @author qx
 *
 * Base class for legacy key-based licenses.
 * Currently such licenses are obsolete and replaced with X509 Certificate-based keys.
 */
public abstract class LegacyKeyBasedLicenseData extends AbstractLicenseData {
  private final String myKey;
  private final String myUserName;
  protected LicenseInfo myInfo = null;

  public LegacyKeyBasedLicenseData(String key, String userName) {
    myKey = key;
    myUserName = userName;
  }

  @Override
  public String getKey() {
    return myKey;
  }

  @Override
  public String getUserName() {
    return myUserName;
  }

  public boolean equals(Object o) {
    if (this == o) return true;
    if (!(o instanceof LegacyKeyBasedLicenseData)) return false;

    final LegacyKeyBasedLicenseData abstractLicenseData = (LegacyKeyBasedLicenseData)o;

    if (myKey != null ? !myKey.equals(abstractLicenseData.myKey) : abstractLicenseData.myKey != null) return false;
    if (myUserName != null
        ? !myUserName.equals(abstractLicenseData.myUserName)
        : abstractLicenseData.myUserName != null) {
      return false;
    }

    return true;
  }

  public int hashCode() {
    int result;
    result = (myKey != null ? myKey.hashCode() : 0);
    result = 29 * result + (myUserName != null ? myUserName.hashCode() : 0);
    return result;
  }

  public String toString() {
    return getUserName() + ":" + getKey();
  }

  @Override
  public boolean isExpired() {
    final Date expirationDate = getExpirationDate();
    if (expirationDate == null) {
      return false;
    }
    return new Date().after(expirationDate);
  }

  @Override
  public boolean isValid() {
    decode();
    return isDecoded();
  }

  protected void decode() {
    if (isDecoded()) return;
    try {
      myInfo = decodeInternal();
    }
    catch (InvalidLicenseKeyException e) {
      myInfo = getNullObjectLicense();
    }
  }

  protected abstract LicenseInfo decodeInternal() throws InvalidLicenseKeyException;

  protected LicenseInfo getNullObjectLicense() {
    return LicenseInfo.NULL_OBJECT;
  }

  private boolean isDecoded() {
    return myInfo != null && myInfo != getNullObjectLicense();
  }

  @Override
  public Date getExpirationDate() {
    decode();
    return myInfo.expirationDate;
  }

  @Override
  public boolean isNonCommercial() {
    decode();
    return myInfo.licenseType == LicenseType.NON_COMMERCIAL;
  }

  public boolean isSite() {
    decode();
    return myInfo.licenseType == LicenseType.SITE;
  }

  @Override
  public boolean isOpenSource() {
    decode();
    return myInfo.licenseType == LicenseType.OPENSOURCE;
  }

  @Override
  public boolean isYearAcademic() {
    decode();
    return myInfo.licenseType == LicenseType.YEARACADEMIC;
  }

  @Override
  public boolean shouldDetectDuplicates() {
    return !(isSite() || isOpenSource() || isYearAcademic());
  }

  @Override
  public boolean isPersonal() {
    decode();
    return myInfo.licenseType == LicenseType.PERSONAL;
  }

  @Override
  public Date getGenerationDate() {
    decode();
    return myInfo.generationDate;
  }

  @Override
  public int getMajorVersion() {
    decode();
    return myInfo.majorVersion;
  }

  public int getProductId() {
    decode();
    return myInfo.productId;
  }
}
