package com.intellij.licensecommon.net.impl;

import com.intellij.licensecommon.message.Message;
import com.intellij.licensecommon.net.*;
import com.intellij.licensecommon.protocol.Protocol;
import com.intellij.licensecommon.protocol.ProtocolDecodingException;
import com.intellij.licensecommon.protocol.ProtocolEncodingException;
import com.intellij.util.concurrency.SequentialTaskExecutor;
import org.jetbrains.annotations.NotNull;

import java.util.List;
import java.util.concurrent.Executor;
import java.util.concurrent.ExecutorService;

public class NetworkImpl implements Network {
  private volatile boolean myIsStarted;
  private final NetworkSocketFactory mySocketFactory;
  private final Protocol myProtocol;
  private volatile NetworkSocket myLocalSocket;
  private volatile List<Address> myMulticastAddresses;
  private NetworkTarget myTarget;
  private NetworkSocket myMulticastSocket;
  private Address myLocalAddress;
  @NotNull
  private final ExecutorService myMessageDispatcher;
  @NotNull
  private final Executor myTaskExecutor;

  public NetworkImpl(@NotNull NetworkSocketFactory socketFactory, @NotNull Protocol protocol, @NotNull Executor taskExecutor) {
    mySocketFactory = socketFactory;
    myProtocol = protocol;
    myTaskExecutor = taskExecutor;
    myMessageDispatcher = SequentialTaskExecutor.createSequentialApplicationPoolExecutor("NetworkImpl Pool", taskExecutor);
  }

  @Override
  public synchronized void startWith(NetworkTarget target) throws CouldNotCreateSocketException, FatalNetworkException {
    if (myIsStarted) {
      return;
    }
    myTarget = target;

    final NetworkSocket localSocket = mySocketFactory.createLocalSocket();
    final NetworkSocket multicastSocket = mySocketFactory.createMulticastSocket();

    myTaskExecutor.execute(new SocketListener(this, localSocket));
    myTaskExecutor.execute(new SocketListener(this, multicastSocket));

    try {
      localSocket.waitUntilReceiving();
      multicastSocket.waitUntilReceiving();
    }
    catch (InterruptedException e) {
      throw new FatalNetworkException(e);
    }
    myLocalSocket = localSocket;
    myLocalAddress = localSocket.getAddress();
    myMulticastSocket = multicastSocket;
    myMulticastAddresses = mySocketFactory.getMulticastAddresses();
    myIsStarted = true;
  }

  @Override
  public synchronized void shutDown() {
    final NetworkSocket localSocket = myLocalSocket;
    if (localSocket != null) {
      myLocalSocket = null;
      localSocket.close();
    }
    final NetworkSocket multicastSocket = myMulticastSocket;
    if (multicastSocket != null) {
      myMulticastSocket = null;
      multicastSocket.close();
    }
    myMulticastAddresses = null;
    myIsStarted = false;
  }

  @Override
  public void send(Message message, Address to) throws NetworkSendException {
    try {
      final byte[] bytes = myProtocol.encode(message);
      synchronized (this) {
        send(bytes, to);
      }
    }
    catch (ProtocolEncodingException e) {
      throw new NetworkSendException(e);
    }
  }

  private void send(byte[] bytes, Address to) {
    myLocalSocket.send(new Packet(bytes, to));
  }

  @Override
  public boolean isStarted() {
    return myIsStarted;
  }

  @Override
  public void receive(Packet packet) {
    final Address address = packet.getAddress();
    if (isAddressIgnored(address)) {
      return;
    }
    try {
      final Message data = myProtocol.decode(packet.getBytes());
      myMessageDispatcher.execute(() -> {
        try {
          myTarget.receivedData(data, address);
        }
        catch (Throwable ignored) {
        }
      });
    }
    catch (ProtocolDecodingException ignored) {
    }
  }

  private boolean isAddressIgnored(Address address) {
    return address.equals(myLocalAddress) || address.isAnyLocalAddress();
  }

  @Override
  public void broadcast(Message message) throws NetworkSendException {
    try {
      final byte[] bytes = myProtocol.encode(message);
      synchronized (this) {
        for (Address address : myMulticastAddresses) {
          send(bytes, address);
        }
      }
    }
    catch (ProtocolEncodingException e) {
      throw new NetworkSendException(e);
    }

  }
}
