package com.intellij.licensecommon.net.impl;

import com.intellij.licensecommon.net.Address;
import com.intellij.licensecommon.net.CouldNotCreateSocketException;
import com.intellij.licensecommon.net.NetworkSocket;
import com.intellij.licensecommon.net.NetworkSocketFactory;
import org.jetbrains.annotations.NotNull;

import java.net.*;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.List;

public class NetworkSocketFactoryImpl implements NetworkSocketFactory {
  private static final int[] PORTS = {8976,8977,8978,8979};
  private static final String MULTICAST_HOST = "230.230.230.230"; // 0xE6E6E6E6

  private static final Address[] MULTICAST_ADDRESSES;
  static {
    MULTICAST_ADDRESSES = new Address[PORTS.length];
    for (int i = 0; i < PORTS.length; i++) {
      MULTICAST_ADDRESSES[i] = new AddressImpl(MULTICAST_HOST, PORTS[i]);
    }
  }

  @Override
  public NetworkSocket createLocalSocket() throws CouldNotCreateSocketException {
    try {
      // send using wildcard address for reliable message delivery
      return createLocalSocket(new AddressImpl(new InetSocketAddress(0)));
    }
    catch (Exception e) {
      throw new CouldNotCreateSocketException(e);
    }
  }

  @Override
  public NetworkSocket createMulticastSocket() throws CouldNotCreateSocketException {
    for (Address address : MULTICAST_ADDRESSES) {
      try {
        return createMulticastSocket(address);
      }
      catch (CouldNotCreateSocketException ignored) {
      }
    }
    throw new CouldNotCreateSocketException();
  }

  @Override
  public List<Address> getMulticastAddresses() {
    return Arrays.asList(MULTICAST_ADDRESSES);
  }

  private static NetworkSocket createLocalSocket(@NotNull Address address) throws CouldNotCreateSocketException {
    try {
      return new SocketImpl(new DatagramSocket(address.getSocketAddress()), SocketImpl.DEFAULT_TIMEOUT);
    }
    catch (SocketException e) {
      throw new CouldNotCreateSocketException(address.getDisplay(), e);
    }
  }

  private static NetworkSocket createMulticastSocket(Address address) throws CouldNotCreateSocketException {
    MulticastSocket socket = null;
    try {
      socket = new MulticastSocket(address.getSocketAddress().getPort());
      final Enumeration<NetworkInterface> networkInterfaces = NetworkInterface.getNetworkInterfaces();
      while (networkInterfaces.hasMoreElements()) {
        final NetworkInterface anInterface = networkInterfaces.nextElement();
        if (isValid(anInterface)) {
          socket.joinGroup(address.getSocketAddress(), anInterface);
        }
      }
      return new SocketImpl(socket, SocketImpl.DEFAULT_TIMEOUT);
    }
    catch (Throwable e) {
      if (socket != null) {
        socket.close();
      }
      throw new CouldNotCreateSocketException(e);
    }
  }

  private static boolean isValid(NetworkInterface anInterface) throws SocketException {
    final Enumeration<InetAddress> addresses = anInterface.getInetAddresses();
    if (!addresses.hasMoreElements()) {
      return false;
    }
    if (anInterface.isLoopback() || anInterface.isPointToPoint() || anInterface.isVirtual()) {
      return false;
    }
    do {
      final InetAddress ifaceAddress = addresses.nextElement();
      if (ifaceAddress.getAddress().length == 4) {
        return true;
      }
    }
    while (addresses.hasMoreElements());

    return false;
  }

}
