package com.intellij.licensecommon.protocol.impl;

import com.intellij.licensecommon.license.LicenseBroadcastData;
import com.intellij.licensecommon.message.Message;
import com.intellij.licensecommon.message.impl.ResharperMessage;
import com.intellij.licensecommon.protocol.Protocol;
import com.intellij.licensecommon.protocol.ProtocolDecodingException;
import com.intellij.licensecommon.protocol.ProtocolEncodingException;
import com.intellij.util.io.DataOutputStream;

import java.io.*;

/**
 * format:
 * <int-code>;<length><encrypted-license-id-string>;<length><licensee-name>;<length><system-login-name>
 */
public class ResharperProtocol implements Protocol {
  //private static final Key KEY = new SecretKeySpec(new byte[] {
  //  0x35, (byte)0x9a, 0x0, (byte)0x97, 0x32, (byte)0xf1, 0x32, (byte)0xdf, 0x5f, 0x54, 0x59, (byte)0x93, (byte)0xc1, (byte)0xc1, 0x10, 0xe
  //}, "AES");

  private static final String DATA_ENCODING = "UTF-16le";

  public ResharperProtocol() {
  }

  @Override
  public byte[] encode(Message message) throws ProtocolEncodingException {
    try {
      final LicenseBroadcastData data = message.getBroadcastData();
      ByteArrayOutputStream out = new ByteArrayOutputStream();
      try (DataOutputStream dataOut = new DataOutputStream(out)) {
        writeInt(dataOut, 0); // code
        writeString(dataOut, data.getLicenseKeyId(), true);
        writeString(dataOut, data.getLicenseeName(), false);
        writeString(dataOut, data.getUserLogin(), false);
      }
      return out.toByteArray();
    }
    catch (Exception e) {
      throw new ProtocolEncodingException(e);
    }
  }

  @Override
  public ResharperMessage decode(byte[] buf) throws ProtocolDecodingException {
    final DataInputStream in = new DataInputStream(new ByteArrayInputStream(buf));
    try {
      final int code = readInt(in);
      final String keyId = readString(in, true);
      final String licenseeName = readString(in, false);
      final String userLogin = readString(in, false);
      return new ResharperMessage(new LicenseBroadcastData(code, keyId, licenseeName, userLogin));
    }
    catch (Exception e) {
      throw new ProtocolDecodingException(e);
    }
  }

  // according to the protocol, int should be stored in low-to-upper sequence
  private void writeInt(DataOutput out, int value) throws IOException {
    out.write(value & 0xFF);
    out.write((value >> 8) & 0xFF);
    out.write((value >> 16) & 0xFF);
    out.write((value >> 24) & 0xFF);
  }

  private int readInt(DataInput in) throws IOException {
    int value = in.readByte();
    value |= in.readByte() << 8;
    value |= in.readByte() << 16;
    value |= in.readByte() << 24;
    return value;
  }

  private void writeString(DataOutput out, String value, boolean encrypt) throws Exception {
    final byte[] bytes = (encrypt? EncryptionUtil.encrypt(value) : value).getBytes(DATA_ENCODING);
    writeInt(out, bytes.length);
    out.write(bytes);
  }

  private String readString(DataInput in, boolean decrypt) throws Exception {
    final int length = readInt(in);
    final byte[] bytes = new byte[length];
    in.readFully(bytes);
    final String str = new String(bytes, DATA_ENCODING);
    return decrypt? EncryptionUtil.decrypt(str) : str;
  }

}
