package com.intellij.marketplace;

import com.intellij.compiler.instrumentation.FailSafeClassReader;
import com.intellij.compiler.instrumentation.FailSafeMethodVisitor;
import com.intellij.ide.BytecodeTransformer;
import org.jetbrains.annotations.Nullable;
import org.jetbrains.org.objectweb.asm.*;

import java.lang.reflect.Method;
import java.lang.reflect.Modifier;
import java.security.ProtectionDomain;
import java.util.List;
import java.util.concurrent.CompletionStage;

/**
 * @author Eugene Zhuravlev
 */
public final class Bootstrap implements BytecodeTransformer {
  static final String MAIN_CLASS_NAME = "com.intellij.idea.MainImpl";
  static final String IDEA_METHOD_OWNER_CLASS_NAME = "com.intellij.idea.ApplicationLoader";
  static final String INIT_APP_METHOD_NAME = "initApplication";

  static final String CE_PLUGINS_LICENSE_MANAGER_CLASS = "com.intellij.ide.license.impl.CEPluginsLicenseManager";
  private boolean myInstallComplete = false;

  @Override
  public boolean isApplicable(String className, ClassLoader loader, @Nullable ProtectionDomain protectionDomain) {
    return !myInstallComplete && MAIN_CLASS_NAME.equals(className);
  }

  @Override
  public byte[] transform(ClassLoader loader, String className, ProtectionDomain protectionDomain, byte[] classBytes) {
    if (!isApplicable(className, loader, protectionDomain)) {
      return null;
    }
    try {
      if (!isApplicable(className, loader, protectionDomain)) {
        return null;
      }

      final Method subst = findWrappingMethod(loader);
      if (subst == null) {
        return null;
      }
      final ClassReader reader = new FailSafeClassReader(classBytes);
      final ClassWriter writer = new ClassWriter(reader, ClassWriter.COMPUTE_FRAMES);
      final BooleanValue instrumented = new BooleanValue(false);
      reader.accept(new ClassVisitor(Opcodes.API_VERSION, writer) {
        @Override
        public MethodVisitor visitMethod(int access, String name, String descriptor, String signature, String[] exceptions) {
          final MethodVisitor mv = super.visitMethod(access, name, descriptor, signature, exceptions);
          return instrumented.get()? mv : new MyInitAppCallInstrumenter(mv, subst, instrumented);
        }
      }, 0);

      return instrumented.get()? writer.toByteArray() : classBytes;
    }
    finally {
      myInstallComplete = true;
    }
  }

  // wraps a call to com/intellij/idea/ApplicationLoader.initApplication
  private static final class MyInitAppCallInstrumenter extends FailSafeMethodVisitor {
    private final Method myWrapper;
    private final BooleanValue myIsIntrumented;

    MyInitAppCallInstrumenter(MethodVisitor mv, Method wrapper, BooleanValue isIntrumented) {
      super(Opcodes.API_VERSION, mv);
      myWrapper = wrapper;
      myIsIntrumented = isIntrumented;
    }

    @Override
    public void visitMethodInsn(int opcode, String owner, String name, String descriptor, boolean isInterface) {

      // debugging
      //final String qualifier = owner != null? owner.replace('/', '.')  + ".": "";
      //System.out.println("Visit method call " + qualifier + name + descriptor);

      if (owner != null && INIT_APP_METHOD_NAME.equals(name) && IDEA_METHOD_OWNER_CLASS_NAME.equals(Type.getObjectType(owner).getClassName())) {
        final String wrapperDescriptor = Type.getMethodDescriptor(myWrapper);
        if (wrapperDescriptor.equals(descriptor)) { // signatures should match
          mv.visitMethodInsn(
            Opcodes.INVOKESTATIC,
            Type.getInternalName(myWrapper.getDeclaringClass()),
            myWrapper.getName(),
            wrapperDescriptor,
            false
          );
          myIsIntrumented.set(Boolean.TRUE);
          return;
        }
      }

      super.visitMethodInsn(opcode, owner, name, descriptor, isInterface);
    }
  }

  static Method findWrappingMethod(ClassLoader loader) {
    try {
      // static void method containing among its parameter types String[]
      for (Method method : Class.forName(CE_PLUGINS_LICENSE_MANAGER_CLASS, false, loader).getDeclaredMethods()) {
        if (!method.isSynthetic() && !method.isBridge() && Modifier.isStatic(method.getModifiers()) && void.class.equals(method.getReturnType())) {
          final Class<?>[] paramTypes = method.getParameterTypes();
          if (paramTypes.length == 2 && List.class.equals(paramTypes[0]) && CompletionStage.class.equals(paramTypes[1])) {
            return method;
          }
        }
      }
    }
    catch (Throwable ignored) {
    }
    return null;
  }
}
