// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ide

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import java.util.*

class Log<TVersion : Comparable<TVersion>, TValue>(private val id: String, initialLastVersion: TVersion) : Iterable<Map.Entry<TVersion, TValue>> {
  private val history = TreeMap<TVersion, TValue>()
  var lastKnownVersion = initialLastVersion
    private set

  fun updateLastKnownVersion(newVersion: TVersion) {
    if (lastKnownVersion < newVersion) {
      lastKnownVersion = newVersion
      shrinkHistory(newVersion)
    }
  }

  fun contains(version: TVersion) = history.contains(version)

  fun getValue(version: TVersion): TValue? {
    return history[version]
  }

  fun put(version: TVersion, change: TValue) {
    require(!history.containsKey(version)) { "Log $id already contains change on $version" }

    history[version] = change
  }

  private fun shrinkHistory(lastKnownVersion: TVersion) {
    val it = history.keys.iterator()
    var removedEntries = 0
    while (it.hasNext()) {
      val version = it.next()
      if (version >= lastKnownVersion) {
        break
      }
      it.remove()
      removedEntries++
    }
    logger.trace { "Log $id was shrunk to $lastKnownVersion version; removed entries: ${removedEntries}, new size: ${history.size}" }
  }

  fun rollbackHistoryToVersion(v: TVersion) {
    val it = history.descendingKeySet().iterator()
    var removedEntries = 0
    while (it.hasNext()) {
      val version = it.next()
      if (version <= v) {
        break
      }
      it.remove()
      removedEntries++
    }
    logger.trace { "Log $id rollback history to $v; removed entries: ${removedEntries}, new size: ${history.size}" }
  }

  override fun iterator(): Iterator<Map.Entry<TVersion, TValue>> {
    return history.iterator()
  }

  companion object {
    private val logger = logger<Log<*, *>>()
  }
}