// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ide.document

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import com.jetbrains.rd.ide.model.RdDocumentChange
import com.jetbrains.rd.ide.model.RdDocumentVersion
import java.util.*

private val logger = logger<DocumentLog>()

open class DocumentLog : Iterable<Map.Entry<RdDocumentVersion, RdDocumentChange>> {
  private val history = TreeMap<RdDocumentVersion, RdDocumentChange> { x, y -> x.value.compareTo(y.value)}
  var lastKnownVersion = InitialDocVersion
    private set

  open fun updateLastKnownVersion(newVersion: RdDocumentVersion) {
    if (lastKnownVersion.value < newVersion.value) {
      lastKnownVersion = newVersion
      shrinkHistory(newVersion)
    }
  }

  fun contains(version: RdDocumentVersion) = history.contains(version)

  fun getChange(version: RdDocumentVersion): RdDocumentChange? {
    return history[version]
  }

  fun put(version: RdDocumentVersion, change: RdDocumentChange) {
    require(!history.containsKey(version)) { "Document log already contains change on $version" }

    history[version] = change
  }

  private fun shrinkHistory(lastKnownVersion: RdDocumentVersion) {
    val it = history.keys.iterator()
    var removedEntries = 0
    while (it.hasNext()) {
      val version = it.next()
      if (version.value >= lastKnownVersion.value) {
        break
      }
      it.remove()
      removedEntries++
    }
    logger.trace { "DocumentLog was shrunk to $lastKnownVersion version; removed entries: ${removedEntries}, new size: ${history.size}" }
  }

  open fun rollbackHistoryToVersion(v: RdDocumentVersion) {
    val it = history.descendingKeySet().iterator()
    var removedEntries = 0
    while (it.hasNext()) {
      val version = it.next()
      if (version.value <= v.value) {
        break
      }
      it.remove()
      removedEntries++
    }
    logger.trace { "DocumentLog rollback history to $v; removed entries: ${removedEntries}, new size: ${history.size}" }
  }

  override fun iterator(): Iterator<Map.Entry<RdDocumentVersion, RdDocumentChange>> {
    return history.iterator()
  }
}

inline fun DocumentLog.getOrPut(version: RdDocumentVersion, crossinline factory: () -> RdDocumentChange): RdDocumentChange {
  return getChange(version) ?: run {
    val change = factory()
    put(version, change)
    change
  }
}