// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ide.document

import com.jetbrains.rd.ide.model.AbstractDocumentVersion
import com.jetbrains.rd.ide.model.RdDocumentVersion
import com.jetbrains.rd.ide.model.RdTextModificationKind
import com.jetbrains.rd.platform.Range
import com.jetbrains.rd.platform.util.IDocumentRangeShifter

class DocumentLogRangeShifter(
  override val documentSynchronizer: ProtocolDocumentSynchronizer
) : IDocumentRangeShifter {
  private val documentLog get() = documentSynchronizer.documentLog

  override fun shrinkHistory() {
    //NOP
  }

  override fun shiftRange(v: AbstractDocumentVersion,
                          range: Range,
                          isVerifiable: Boolean,
                          isGreedyToLeft: Boolean,
                          isGreedyToRight: Boolean): Range {
    assert(range.isNormalized) { "Range: $range is not normalized" }
    val initialVersion = v as RdDocumentVersion
    if (!checkVersions(initialVersion)) return Range.invalidRange

    var resultStart = range.start
    var resultEnd = range.end
    for ((version, change) in documentLog) {
      if (version.value <= initialVersion.value) continue

      val oldLength = change.oldText.length
      val newLength = change.newText.length
      val start = change.offset
      val end = start + oldLength
      val delta = newLength - oldLength

      if (oldLength == 0 && change.moveOffset != start) { // text was moved
        // offset before insertion
        val moveOffset = change.moveOffset.let { if (it > start) it - newLength else it }

        if (moveOffset <= resultStart && resultEnd <= moveOffset + newLength) { // range is inside moved text
          val moveDelta = start - moveOffset
          resultStart += moveDelta
          resultEnd += moveDelta
          continue
        }
      }

      // optimized replace at the endOffset or greedy range marker
      if (resultEnd == start && (change.modificationKind == RdTextModificationKind.RightSide || isGreedyToRight)) {
        resultEnd += delta
        continue
      }

      if (resultEnd <= start) // change happened after specified range, no need for transformation
        continue

      // optimized replace at the startOffset or greedy range marker
      if (resultStart == end && (change.modificationKind == RdTextModificationKind.LeftSide || isGreedyToLeft)) {
        resultEnd += delta
        continue
      }

      if (resultStart >= end) { // change happened before specified range, shifting
        resultStart += delta
        resultEnd += delta
        continue
      }

      if (resultStart <= start && end <= resultEnd && !isVerifiable) { // change is inside result range and we allow text mismatch (e.g. for folding)
        resultEnd += delta
        continue
      }

      // result range intersects with document change (can't be shifted and will be removed by the backend shortly)
      return Range.invalidRange
    }
    return Range(resultStart, resultEnd)
  }

  override fun shiftOnePoint(v: AbstractDocumentVersion, offset: Int, isStickingToRight: Boolean): Int {
    val initialVersion = v as RdDocumentVersion
    if (!checkVersions(initialVersion)) return -1

    var resultOffset = offset
    for ((version, change) in documentLog) {
      if (version.value <= initialVersion.value) continue
      val oldLength = change.oldText.length
      val start = change.offset
      val end = start + oldLength
      val delta = change.newText.length - change.oldText.length


      if (resultOffset < start) continue

      if (resultOffset > end || isStickingToRight && oldLength == 0) {
        resultOffset += delta
        continue
      }

      return -1
    }
    return resultOffset
  }

  private fun checkVersions(initialRangeVersion: RdDocumentVersion): Boolean {
    return documentLog.contains(initialRangeVersion)
  }
}