package com.jetbrains.rd.ide.document

import com.intellij.openapi.Disposable
import com.intellij.openapi.command.CommandProcessor
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.ex.PrioritizedDocumentListener
import com.intellij.openapi.editor.impl.DocumentImpl
import com.intellij.openapi.editor.impl.EditorDocumentPriorities
import com.intellij.openapi.editor.impl.event.DocumentEventImpl
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.Key
import com.intellij.openapi.util.TextRange
import com.intellij.util.LocalTimeCounter
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ide.requests.IdeGateway
import com.jetbrains.rd.platform.util.application
import kotlin.math.max
import kotlin.math.min

interface DocumentSynchronizer : Disposable {
  val modelDocument: RdDocumentModel // todo drop it
  val documentId: RdDocumentId
  val isSandbox: Boolean
  val isReceivingChange: Boolean
  val version: AbstractDocumentVersion
  fun resetVersion(v: AbstractDocumentVersion) { throw UnsupportedOperationException() }
}

class ProtocolDocumentSynchronizer(
  project: Project,
  val document: Document,
  override val modelDocument: RdDocumentModel, // todo drop it
  override val documentId: RdDocumentId,
  private val versionHolder: DocumentVersionHolder,
  val documentLog: DocumentLog,
  private val ideGateway: IdeGateway,
  private val sendDocumentText: Boolean,
  private val forceWrite: Boolean
) : PrioritizedDocumentListener, Disposable, DocumentSynchronizer {

  companion object {
    private var debugId = 0
  }

  private var backendUpdateInProgressMarker = false
  private val logger: Logger

  override val isReceivingChange: Boolean
    get() {
      application.assertIsDispatchThread()
      return backendUpdateInProgressMarker
    }

  init {
    val fileName = documentId.createLogCategory()
    logger = Logger.getInstance("#com.jetbrains.rd.ide.document.Document#${debugId++}($fileName)")
    document.addDocumentListener(this, this)
    Disposer.register(project, this)
  }

  override fun documentChanged(event: DocumentEvent) {
    versionHolder.incrementVersion(document.modificationStamp)

    if (backendUpdateInProgressMarker) return

    logger.trace {
      "Document changed from IDEA side. Offset: ${event.offset}, New fragment: '${escape(event.newFragment)}', OldFragment: '${escape(event.oldFragment)}'"
    }

    val change = documentLog.getOrPut(versionHolder.version) { toRdDocumentChange(documentId, event, sendDocumentText) }

    ideGateway.sendDocumentChange(change, versionHolder.previousVersion, versionHolder.version)
  }

  fun sendInitialTextChange() {
    val change = RdDocumentChange(
      documentId,
      0,
      0,
      document.text,
      "",
      RdTextModificationKind.Reset,
      -1,
      null
    )

    // put initial text only if new document is created
    if (versionHolder.incrementVersion(document.modificationStamp)) {
      documentLog.put(versionHolder.version, change)
    }

    ideGateway.sendDocumentChange(change, versionHolder.previousVersion, versionHolder.version)
  }

  fun receiveDocumentChange(change: RdDocumentChange) {
    try {
      backendUpdateInProgressMarker = true

      if (change.modificationKind != RdTextModificationKind.Reset && document.textLength != change.oldDocumentLength) {
        error("expected document length: ${change.oldDocumentLength}, but actual length: ${document.textLength}")
      }

      val offset = change.offset
      val newText = change.newText
      val oldText = change.oldText

      fun applyChange() {
        // optimization for rdserver, cache changes for all clients before `documentChanged` listeners execution
        documentLog.put(versionHolder.version.increment(), change)

        if (change.modificationKind == RdTextModificationKind.Reset) {
          document.setText(change.newText)
          return
        }

        if (oldText.isEmpty() && change.modificationKind == RdTextModificationKind.LeftSide) {
          val initialEnd = min(document.textLength, offset + 1)
          val initialNewText = newText + document.getText(TextRange(offset, initialEnd))
          // needed to update range markers properly (internally tracked in IJ by DocumentEventImpl#initialStartOffset/EndOffset)
          document.replaceString(offset, initialEnd, initialNewText)
          return
        }

        if (oldText.isEmpty() && change.modificationKind == RdTextModificationKind.RightSide) {
          val initialStart = max(0, offset - 1)
          val initialNewText = document.getText(TextRange(initialStart, offset)) + newText
          // needed to update range markers properly (internally tracked in IJ by DocumentEventImpl#initialStartOffset/EndOffset)
          document.replaceString(initialStart, offset, initialNewText)
          return
        }

        if (document is DocumentImpl) {
          document.replaceString(offset, offset + oldText.length, change.moveOffset, newText, LocalTimeCounter.currentTime(), false)
          return
        }

        document.replaceString(offset, offset + oldText.length, newText)
      }
      
      fun applyChangeCheckForce() {
        if (forceWrite && !document.isWritable) {
          document.setReadOnly(false)
          try {
            applyChange()
          } finally {
            document.setReadOnly(true)
          }
        } else {
          applyChange()
        }
      }

      application.runWriteAction {
        val instance = CommandProcessor.getInstance()
        if (instance.currentCommand != null) {
          applyChangeCheckForce()
        }
        else {
          CommandProcessor.getInstance().runUndoTransparentAction {
            applyChangeCheckForce()
          }
        }

        if (change.text != null && document.text != change.text) {
          error("Not matched document text on version: ${versionHolder.version.value}")
        }
      }
    }
    finally {
      backendUpdateInProgressMarker = false
    }
  }


  override fun beforeDocumentChange(event: DocumentEvent) {}
  override fun getPriority() = EditorDocumentPriorities.RANGE_MARKER - 1

  override val isSandbox: Boolean get() = modelDocument.sandboxInfo != null

  override val version: RdDocumentVersion get() = versionHolder.version

  override fun resetVersion(v: AbstractDocumentVersion) {
    v as RdDocumentVersion
    versionHolder.resetVersion(v, document.modificationStamp)
    documentLog.rollbackHistoryToVersion(v)
  }

  override fun dispose() {
    logger.trace { "Document synchronizer is released" }
  }

  private fun escape(s: CharSequence) = s.toString().replace("\n", "\\n").replace("\r", "\\r")
}

fun <TExtension : DocumentExtension> DocumentSynchronizer.getExtension(key: Key<TExtension>): TExtension? {
  @Suppress("UNCHECKED_CAST")
  return modelDocument.extensions.firstOrNull { it.key == key.toString() } as? TExtension
}

private fun toRdDocumentChange(documentId: RdDocumentId, event: DocumentEvent, sendDocumentText: Boolean): RdDocumentChange {
  val modificationSide = when {
    event.oldFragment.isEmpty() && event is DocumentEventImpl -> when {
      event.initialStartOffset < event.offset -> RdTextModificationKind.RightSide
      event.initialStartOffset + event.initialOldLength > event.offset -> RdTextModificationKind.LeftSide
      else -> RdTextModificationKind.NotSpecified
    }
    else -> RdTextModificationKind.NotSpecified
  }

  val change = RdDocumentChange(
    documentId,
    event.offset,
    event.moveOffset,
    event.newFragment.toString(),
    event.oldFragment.toString(),
    modificationSide,
    event.document.textLength + event.oldLength - event.newLength,
    if (sendDocumentText) event.document.text else null
  )
  return change
}