package com.jetbrains.rd.ide.editor

import com.intellij.codeInsight.daemon.impl.VisibleHighlightingPassFactory
import com.intellij.openapi.Disposable
import com.intellij.openapi.editor.*
import com.intellij.openapi.editor.ex.util.EditorUtil
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.jetbrains.rd.ide.document.toDebugString
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.userData

var Caret.id: RdCaretId? by userData()

fun ScrollKind.toScrollType(): ScrollType = when (this) {
  ScrollKind.Smart -> ScrollType.MAKE_VISIBLE
  ScrollKind.Relative -> ScrollType.RELATIVE
  ScrollKind.Center -> ScrollType.CENTER
  ScrollKind.LowerHalf -> ScrollType.CENTER_DOWN
  ScrollKind.UpperHalf -> ScrollType.CENTER_UP
}

private fun Editor.docOffsetAndVirtualToVisualPosition(pos: DocOffsetAndVirtualModel): VisualPosition {
  require(pos.virtualDelta >= 0)
  val vPos = offsetToVisualPosition(pos.offset)
  return if (pos.virtualDelta == 0) vPos
  else VisualPosition(vPos.line, vPos.column + pos.virtualDelta, vPos.leansRight)
}

fun calculateVisibleRange(editor: Editor): VisibleRange {
  val visibleRange = VisibleHighlightingPassFactory.calculateVisibleRange(editor)
  return VisibleRange(visibleRange.startOffset, visibleRange.endOffset)
}

fun applyStateTo(newState: RdCaretState, editor: Editor) {
  val states = newState.carets.map { caret ->
    val caretPosition = editor.offsetToLogicalPosition(caret.offset.offset)
    val selectionStart = editor.offsetToLogicalPosition(caret.selection.start.offset)
    val selectionEnd = editor.offsetToLogicalPosition(caret.selection.end.offset)
    CaretState(caretPosition, caret.offset.virtualDelta, selectionStart, selectionEnd)
  }
  editor.caretModel.caretsAndSelections = states

  // Restore IDs for all carets
  for (newCaret in newState.carets) {
    val visualPosition = editor.docOffsetAndVirtualToVisualPosition(newCaret.offset)
    val caret = editor.caretModel.getCaretAt(visualPosition) ?: error("Cannot find caret at $visualPosition")
    caret.id = newCaret.id
    if (newCaret.id == newState.primaryCaretId && editor.caretModel.primaryCaret !== caret) {
      editor.caretModel.removeCaret(caret)
      addCaretTo(newCaret, editor, isPrimary = true)
    }
  }
}

fun applySelectionTo(selectionModel: RdSelection, caretToSync: Caret, editor: Editor) {
  val (startOffset, startVDelta) = selectionModel.start
  val (endOffset, endVDelta) = selectionModel.end

  if (startVDelta > 0 || endVDelta > 0) {
    val startVPos = editor.docOffsetAndVirtualToVisualPosition(selectionModel.start)
    val endVPos = editor.docOffsetAndVirtualToVisualPosition(selectionModel.end)
    caretToSync.setSelection(startVPos, startOffset, endVPos, endOffset)
  }
  else {
    caretToSync.setSelection(startOffset, endOffset)
  }
}

fun addCaretTo(caret: RdCaret, editor: Editor, isPrimary: Boolean = true): Caret? {
  val caretToSync = addCaretTo(caret.offset, editor, isPrimary) ?: return null
  applySelectionTo(caret.selection, caretToSync, editor)
  caretToSync.id = caret.id
  return caretToSync
}

fun addCaretTo(positionModel: DocOffsetAndVirtualModel, editor: Editor, isPrimary: Boolean = true): Caret? {
  val pos = editor.docOffsetAndVirtualToVisualPosition(positionModel)
  return editor.caretModel.addCaret(pos, isPrimary)
}

fun CaretModel.createModelCarets(): List<RdCaret> {
  return allCarets.map { caret ->
    val id = caret.id ?: error("Caret has no ID assigned")
    val offset = caret.createModelPosition(caret.offset) { caret.visualPosition }
    val startPos = caret.createModelPosition(caret.selectionStart) { caret.selectionStartPosition }
    val endPos = caret.createModelPosition(caret.selectionEnd) { caret.selectionEndPosition }
    val selection = RdSelection(startPos, endPos)
    RdCaret(id, offset, selection)
  }
}

fun CaretModel.createModelCaretState(): RdCaretState {
  val primaryId = primaryCaret.id ?: error("Primary caret has no ID assigned")
  return RdCaretState(primaryId, createModelCarets())
}

private inline fun Caret.createModelPosition(offset: Int, crossinline calcVisualPosition: () -> VisualPosition): DocOffsetAndVirtualModel {
  val isInBulkUpdate = editor.document.isInBulkUpdate
  if (isInBulkUpdate)
    return DocOffsetAndVirtualModel(offset, 0)

  val position = calcVisualPosition()
  val virtualDelta = calcVirtualDelta(editor, offset, position)
  return DocOffsetAndVirtualModel(offset, virtualDelta)
}

// todo drop copy-pasted method and use EditorModificationUtil.calcAfterLineEnd
private fun calcVirtualDelta(editor: Editor, offset: Int, visualPosition: VisualPosition): Int {
  val document = editor.document
  val logicalPosition = editor.visualToLogicalPosition(visualPosition)
  val lineNumber = logicalPosition.line
  val columnNumber = logicalPosition.column
  if (lineNumber >= document.lineCount) {
    return columnNumber
  }

  var anchorLineEndOffset = document.getLineEndOffset(lineNumber)
  val softWraps = editor.softWrapModel.getSoftWrapsForLine(logicalPosition.line)
  for (softWrap in softWraps) {
    if (!editor.softWrapModel.isVisible(softWrap)) {
      continue
    }

    val softWrapOffset = softWrap.start
    if (softWrapOffset == offset) {
      // There are two possible situations:
      //     *) caret is located on a visual line before soft wrap-introduced line feed;
      //     *) caret is located on a visual line after soft wrap-introduced line feed;
      val position = editor.offsetToVisualPosition(offset - 1)
      if (position.line == visualPosition.line) {
        val delta = visualPosition.column - position.column - 1
        return if (delta < 0) 0 else delta
      }
    }
    if (softWrapOffset > offset) {
      anchorLineEndOffset = softWrapOffset
      break
    }

    // Same offset corresponds to all soft wrap-introduced symbols, however, current method should behave differently in
    // situations when the caret is located just before the soft wrap and at the next visual line.
    if (softWrapOffset == offset) {
      val visuallyBeforeSoftWrap = visualPosition.line < editor.offsetToVisualPosition(offset).line
      if (visuallyBeforeSoftWrap) {
        anchorLineEndOffset = softWrapOffset
        break
      }
    }
  }

  val lineEndColumnNumber = editor.offsetToLogicalPosition(anchorLineEndOffset).column
  val delta = columnNumber - lineEndColumnNumber
  return if (delta < 0) 0 else delta
}

fun RdCaretStateChange.reverse(): RdCaretStateChange {
  return RdCaretStateChange(newState, oldState, editorId)
}

fun RdDocumentChange.reverse(): RdDocumentChange {
  if (modificationKind == RdTextModificationKind.Reset)
    error("Can't revert initial change")

  val oldDocumentText = text?.replaceRange(offset, offset + newText.length, oldText)
  return RdDocumentChange(this.documentId, this.offset, this.moveOffset,
                          this.oldText, this.newText, this.modificationKind, oldDocumentLength - oldText.length + newText.length, oldDocumentText)
}

/**
 * Returns Disposable which will be disposed on exposing the editor
 * @param project if null -> [Editor.getProject] will be used
 */
fun Editor.getDisposable(project: Project?): Disposable {
  val selectedProject = project ?: this.project
  val debugName = "Disposable for editor $this"
  val disposable =
    if (selectedProject == null) {
      Disposer.newDisposable(debugName)
    }
    else {
      Disposer.newDisposable(selectedProject, debugName)
    }
  EditorUtil.disposeWithEditor(this, disposable)
  return disposable
}

fun RdCaretState.toDebugString(): String = "(primary: ${primaryCaretId.toDebugString()}, carets: ${carets.map { it.toDebugString() }})"
fun RdCaret.toDebugString(): String = "(id: ${id.toDebugString()}, offset: ${offset.toDebugString()}, selection: ${selection.toDebugString()})"
fun RdCaretId.toDebugString(): String = "${if (isClient) "c" else "s"}$number"
fun DocOffsetAndVirtualModel.toDebugString(): String = "($offset, $virtualDelta)"
fun RdSelection.toDebugString(): String = "(start: ${start.toDebugString()} - end: ${end.toDebugString()})"
fun TextControlId.toDebugString(): String = "(${documentId.toDebugString()}, $tabIndex, $kind)"

operator fun VisualPosition?.component1() = this?.line ?: -1
operator fun VisualPosition?.component2() = this?.column ?: -1

operator fun LogicalPosition.component1() = this.line
operator fun LogicalPosition.component2() = this.column
