// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ide.requests

import com.intellij.openapi.Disposable
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.use
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.isNotAlive
import java.util.*

fun IdeGateway.queueUntilBulkOperationIsFinished(lifetime: Lifetime, name: String, block: () -> Unit) {
  application.invokeLater {
    if (lifetime.isNotAlive) return@invokeLater
    logger<IdeGateway>().debug("Queuing '$name'")
    if (bulkUpdateInProgress) {
      queueUntilBulkOperationIsFinished(lifetime, name, block)
    } else {
      block()
    }
  }
}

abstract class IdeGateway(lifetime: Lifetime) {
  // allow nested bulk updates to support LookupTypedHandler case that can insert text or complete item
  private val buildersStack = ArrayDeque<PatchBuilder>()
  protected val currentPatchBuilder: PatchBuilder? get() = buildersStack.peek()
  private val unknownOperationLifetimes = SequentialLifetimes(lifetime)

  val bulkUpdateInProgress: Boolean
    get() = currentPatchBuilder != null

  protected abstract fun sendPatch(patch: RdPatch)

  fun sendEditorChange(change: RdEditorChange, oldVersion: RdEditorVersion, newVersion: RdEditorVersion, documentVersion: RdDocumentVersion) {
    val builder = (currentPatchBuilder ?: PatchBuilder()).pushEditorStep(change, oldVersion, newVersion, documentVersion)

    if (currentPatchBuilder != null) {
      return
    }

    val patch = builder.build()!!
    sendPatch(patch)
  }

  fun sendDocumentChange(change: RdDocumentChange, oldVersion: RdDocumentVersion, newVersion: RdDocumentVersion) {
    val builder = (currentPatchBuilder ?: PatchBuilder()).pushDocumentStep(change, oldVersion, newVersion)

    if (currentPatchBuilder != null) {
      return
    }

    val patch = builder.build()!!
    sendPatch(patch)
  }

  fun sendClipboardChange(change: RdClipboardChange, oldVersion: Int, newVersion: Int) {
    val builder = (currentPatchBuilder ?: PatchBuilder()).pushClipboardChange(change, oldVersion, newVersion)

    if (currentPatchBuilder != null) {
      return
    }

    val patch = builder.build()!!
    sendPatch(patch)
  }

  /**
   * Use this method if editor action/typed action is executed.
   *
   * Helps to revert caret model properly, if patch contains only document changes (for example backspace action's patch).
   * Also saves initial editor and document versions.
   */
  abstract suspend fun executeBulkUpdate(editor: Editor, block: suspend () -> Unit): RdPatch?

  suspend fun executeBulkUpdate(block: suspend () -> Unit): RdPatch? {
    createBulkUpdateCookie().use {
      block()
      return it.builder.build()
    }
  }

  protected fun startBulkUpdateOnUnknownOperation(reason: String): Lifetime {
    LOG.trace { ">> grouping changes for foreign activity: $reason" }
    val cookie = createBulkUpdateCookie()
    val nextLifetime = unknownOperationLifetimes.next()
    beforeBulkUpdateOnUnknownOperation(nextLifetime)
    nextLifetime.onTermination {
      val patch = cookie.builder.build()
      if (patch != null) {
        sendPatch(patch)
      }

      Disposer.dispose(cookie)
      LOG.trace { "<< grouping changes" }
    }
    return nextLifetime
  }

  protected open fun beforeBulkUpdateOnUnknownOperation(lifetime: Lifetime) { }

  protected val unknownOperationInProgress: Boolean get() = !unknownOperationLifetimes.isTerminated

  protected fun endBulkUpdateOnUnknownOperation() {
    unknownOperationLifetimes.terminateCurrent()
  }

  private fun createBulkUpdateCookie(): BulkUpdateCookie {
    endBulkUpdateOnUnknownOperation()
    return BulkUpdateCookie(buildersStack)
  }

  private class BulkUpdateCookie(private val buildersStack:  ArrayDeque<PatchBuilder>): Disposable {
    val builder = PatchBuilder()

    init {
      require(buildersStack.peek()?.build() == null) { "Outer builder must be empty" }
      buildersStack.addFirst(builder)
    }

    override fun dispose() {
      require(buildersStack.peek() == builder)
      buildersStack.removeFirst()
    }
  }

  companion object {
    private val LOG = logger<IdeGateway>()
  }
}