// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ide.requests

import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.impl.EditorImpl
import com.jetbrains.rd.ide.editor.createModelCaretState
import com.jetbrains.rd.ide.model.*

class PatchBuilder {
  private val changes = mutableListOf<RdPatchStep>()
  private val features = mutableListOf<RdEditorFeatureStep>()
  private val expectedDocumentVersions = mutableMapOf<RdDocumentId, RdDocumentVersion>()
  private val expectedEditorVersions = mutableMapOf<TextControlId, RdEditorVersion>()
  private val updatedEditorVersions = mutableMapOf<TextControlId, RdEditorVersion>()
  private val updatedDocumentVersions = mutableMapOf<RdDocumentId, RdDocumentVersion>()
  private var initialCaretStateHint: RdCaretStateSnapshot? = null

  private var expectedClipboardVersion: Int? = null
  private var updatedClipboardVersion: Int? = null
  private var primaryEditor: Editor? = null
  private var disableCaretShiftOnWhitespace = false


  fun build(): RdPatch? {
    if (changes.isEmpty() && features.isEmpty()) {
      return null
    }

    val nullableFeatures = if (features.isEmpty()) null else features

    return RdPatch(changes, nullableFeatures,
                   initialCaretStateHint,
                   disableCaretShiftOnWhitespace,
                   expectedDocumentVersions.keys.toList(),
                   updatedDocumentVersions.keys.toList(),
                   expectedDocumentVersions.values.toList(),
                   updatedDocumentVersions.values.toList(),
                   expectedEditorVersions.keys.toList(),
                   updatedEditorVersions.keys.toList(),
                   expectedEditorVersions.values.toList(),
                   updatedEditorVersions.values.toList(),
                   expectedClipboardVersion, updatedClipboardVersion)
  }

  fun pushInitialSelectionHintAndEditorVersion(editorId: TextControlId,
                                               editor: Editor,
                                               editorVersion: RdEditorVersion,
                                               documentVersion: RdDocumentVersion) {
    primaryEditor = editor
    val caretState = editor.caretModel.createModelCaretState()
    expectedEditorVersions.putIfAbsent(editorId, editorVersion)
    expectedDocumentVersions.putIfAbsent(editorId.documentId, documentVersion)
    initialCaretStateHint = RdCaretStateSnapshot(caretState, editorId)
  }

  fun pushEditorStep(step: RdEditorChange, oldVersion: RdEditorVersion, newVersion: RdEditorVersion, documentVersion: RdDocumentVersion): PatchBuilder {
    val editorId = step.editorId
    expectedEditorVersions.putIfAbsent(editorId, oldVersion)
    expectedDocumentVersions.putIfAbsent(editorId.documentId, documentVersion)

    if (oldVersion != newVersion) {
      require(step !is RdTransparentEditorChange)
      updatedEditorVersions[editorId] = newVersion
    }

    if (step is RdEditorFeatureStep) {
      pushEditorFeatureStep(step)
    }
    else {
      pushChangeStep(step)
    }

    return this
  }

  fun pushDocumentStep(step: RdDocumentChange, oldVersion: RdDocumentVersion, newVersion: RdDocumentVersion): PatchBuilder {
    require(oldVersion != newVersion)
    disableCaretShiftOnWhitespace = primaryEditor?.getUserData(EditorImpl.DISABLE_CARET_SHIFT_ON_WHITESPACE_INSERTION) ?: false
    val documentId = step.documentId
    expectedDocumentVersions.putIfAbsent(documentId, oldVersion)
    updatedDocumentVersions[documentId] = newVersion
    pushChangeStep(step)
    return this
  }

  fun pushClipboardChange(change: RdClipboardChange, oldVersion: Int, newVersion: Int): PatchBuilder {
    if (expectedClipboardVersion == null) {
      expectedClipboardVersion = oldVersion
    }

    if (oldVersion != newVersion) {
      updatedClipboardVersion = newVersion
    }

    pushChangeStep(change)
    return this
  }

  private fun pushEditorFeatureStep(change: RdEditorFeatureStep) {
    features.add(change)
  }

  private fun pushChangeStep(step: RdPatchStep) {
    changes.add(step)
  }
}