// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.lang

import com.intellij.lang.Language
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.extensions.ExtensionPointName
import com.jetbrains.rd.ide.model.RdLanguage
import com.intellij.util.containers.forEachLoggingErrors

class ProtocolLanguageManagerImpl : ProtocolLanguageManager {
  private val EP_NAME = ExtensionPointName.create<ProtocolLanguageAssociationBean>("com.intellij.rd.languageAssociation")
  private val backendToFrontendLanguageCache: HashMap<RdLanguage, String> = hashMapOf()
  private val frontendToBackendLanguageCache: HashMap<String, RdLanguage> = hashMapOf()

  init {
    EP_NAME.extensionList.forEachLoggingErrors(ProtocolLanguageManager.logger) {
      addAssociation(it.backendLanguage!!, it.frontendLanguage!!, it.isDefault)
    }
  }

  override fun toRdLanguage(language: Language): RdLanguage? {
    return frontendToBackendLanguageCache[language.id]
  }

  override fun toIdeaLanguage(rdLanguage: RdLanguage): Language? {
    var frontendLanguageId = backendToFrontendLanguageCache[rdLanguage]
    if (frontendLanguageId != null) {
      val frontendLanguage = Language.findLanguageByID(frontendLanguageId)
      if (frontendLanguage == null) {
        logger<ProtocolLanguageManager>().error("Unable to find '$frontendLanguageId'. Possibly it's incorrect or com.intellij.lang.Language is not initialized")
      }
      return frontendLanguage
    }

    val foundLanguage = Language.getRegisteredLanguages().find { it.id.equals(rdLanguage.name, true) } ?: return null
    frontendLanguageId = foundLanguage.id

    backendToFrontendLanguageCache[rdLanguage] = frontendLanguageId
    if (!frontendToBackendLanguageCache.containsKey(frontendLanguageId)) {
      frontendToBackendLanguageCache[frontendLanguageId] = rdLanguage
    }
    return foundLanguage
  }

  override fun addAssociation(backendLanguageId: String, frontendLanguageId: String, default: Boolean) {
    if (default) {
      val rdLanguage = RdLanguage(backendLanguageId)
      frontendToBackendLanguageCache[frontendLanguageId] = rdLanguage
      backendToFrontendLanguageCache[rdLanguage] = frontendLanguageId
      return
    }

    val backendLanguage = RdLanguage(backendLanguageId)

    if (!frontendToBackendLanguageCache.containsKey(frontendLanguageId)) {
      frontendToBackendLanguageCache[frontendLanguageId] = backendLanguage
    }

    if (!backendToFrontendLanguageCache.containsKey(backendLanguage)) {
      backendToFrontendLanguageCache[backendLanguage] = frontendLanguageId
    }
  }
}