package com.jetbrains.rd.platform.codeWithMe

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.startup.StartupActivity
import com.intellij.openapi.util.Version
import com.intellij.openapi.util.io.FileUtil
import com.jetbrains.rd.platform.util.application
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.StandardCopyOption
import java.nio.file.attribute.BasicFileAttributes
import java.time.Instant
import java.time.temporal.ChronoUnit
import kotlin.io.path.*
import kotlin.streams.asSequence

class CodeWithMeCleanupManager : StartupActivity.Background {
  override fun runActivity(project: Project) {
    require(!application.isDispatchThread) { "executing cleanup activity in the main thread, should be in the background instead" }

    val systemPath = getJetBrainsSystemCachesDir()
    val configPath = getJetBrainsConfigCachesDir()
    val logsPath = getJetBrainsSpecialLogsDir()
    CodeWithMeCleanup.cleanupCodeWithMeCaches(systemPath, configPath, logsPath)
  }
}
  
object CodeWithMeCleanup {
  private const val JbrNamePrefix = "jbr"
  private const val MaxJbrFindDepth = 6
  private val productNames = listOf("CodeWithMeGuest", "IntelliJClient")
  private val installationFolders = listOf("CodeWithMeClientDist", "IntelliJClient")
  private val logger = logger<CodeWithMeCleanup>()
  
  fun cleanupCodeWithMeCaches(systemPath: Path, configPath: Path, logsPath: Path?) {
    logger.info("running activity to cleanup old thin clients... Root path is '$systemPath'")

    try {
      val keptVersions = removeOldSystemFolders(systemPath, logsPath)
      removeOldConfigFolders(configPath, keptVersions)
      val keptRuntimes = removeOldInstallations(systemPath, keptVersions)
      removeOldJavaRuntimes(systemPath, keptRuntimes)
      removeEmptyFolders(systemPath)
    } catch (e: Exception) {
      logger.error(e)
    }
  }
  
  private fun removeEmptyFolders(rootPath: Path) {
    for (folderName in installationFolders) {
      val path = rootPath / folderName
      try {
        if (path.isDirectory() && !path.useDirectoryEntries { it.iterator().hasNext() }) {
          path.deleteIfExists()
        }
      }
      catch (e: Exception) {
        logger.warn("Failed to delete an installation folder: $path", e)
      }
    }
  }

  private fun removeOldJavaRuntimes(rootPath: Path, keptRuntimes: Map<Path, String>) {
    val homeJbrPath = System.getProperty("java.home")?.let { Path(it) }

    val runtimes = installationFolders.flatMap {
      val path = rootPath / it
      path.listDirectoryEntriesSafe("$JbrNamePrefix*") ?: emptyList()
    }
    
    val monthAgo = Instant.now().minus(30L, ChronoUnit.DAYS)
    runtimes
      .asSequence()
      .filter {
        if (homeJbrPath == null || !homeJbrPath.startsWith(it))
          return@filter true

        logger.info("keep Java runtime ${it.name} because the current process is running under this runtime, full path is '$it'")
        return@filter false
      }
      .mapNotNull { folder ->
        try {
          folder to folder.listDirectoryEntriesAndSelf().maxOf { (_, attrs) -> attrs.lastModifiedTime() }
        } catch (e: Exception) {
          logger.warn("Failed to get last modified time", e)
          null
        }
      }
      .sortedByDescending { it.second }
      .mapIndexed { index, pair ->
        if (index < 2) {
          val order = if (index == 0) "most" else "second most"
          logger.info("keep Java runtime ${pair.first.name} because it is the $order recent runtime downloaded, path is '${pair.first}'")
        }
        return@mapIndexed pair
      }
      .drop(2)
      .forEach { (folder, lastModified) ->
        val installedOrUsedTime = lastModified.toInstant()
        when {
          installedOrUsedTime >= monthAgo -> {
            logger.info("keep Java runtime ${folder.name} because it is not old enough, path is '$folder'")
          }
          folder in keptRuntimes -> {
            logger.info("keep Java runtime ${folder.name} because ${keptRuntimes[folder]} still uses it, path is '$folder'")
          }
          else -> {
            logger.info("remove Java runtime ${folder.name} because it is older than a month, path is '$folder'")
            removeFolder(folder)
          }
        }
      }
  }

  private fun removeOldInstallations(rootPath: Path, keptVersions: Set<Version>): Map<Path, String> {
    val installations = installationFolders.flatMap { folderName -> 
      val path = rootPath / folderName
      productNames.flatMap { productName ->
        path.listDirectoryEntriesSafe("$productName*") ?: emptyList()
      }
    }
    
    val keptRuntimes = mutableMapOf<Path, String>()
    installations.forEach { folder ->
      val versionString = folder.name.split('-').getOrNull(1) ?: return@forEach
      val version = Version.parseVersion(versionString) ?: return@forEach
      if (version in keptVersions) {
        logger.info("keep Code With Me client installation with version $version because its system folder still exists, path is '$folder'")
        
        val jbrLink = Files.find(folder, MaxJbrFindDepth, { path, _ -> 
          path.name == JbrNamePrefix && path.seemsLikeLink
        }).use { it.asSequence().firstOrNull() }
        
        if (jbrLink != null) {
          keptRuntimes[jbrLink.toRealPath().parent] = folder.name
        }
        return@forEach
      }
      
      logger.info("remove Code With Me client installation with version $version because its system folder is removed, path is '$folder'")
      removeFolder(folder)
    }
    return keptRuntimes
  }

  private fun removeOldSystemFolders(path: Path, logsPath: Path?): Set<Version> {
    val systemFolders = productNames.flatMap { productName ->
      path.listDirectoryEntriesSafe("$productName?*") ?: emptyList()
    }

    logger.info("found ${systemFolders.size} Code With Me client system folders to check")
    val keptVersions = mutableSetOf<Version>()
    for (folder in systemFolders) {
      val currentProductName = productNames.find { folder.name.startsWith(it) }
      if (currentProductName == null) {
        logger.error("Name of a config folder '${folder.name}' doesn't start with any of known product names: ${productNames.joinToString()}")
        continue
      }
      
      val version = Version.parseVersion(folder.name.removePrefix(currentProductName))
      if (version == null) {
        logger.info("failed to parse version for the folder '${folder.name}', path is $folder")
        continue
      }
      
      if (folder.systemFolderIsTooOld(logsPath)) {
        logger.info("remove system folder with version $version because it is older than two weeks, path is '$folder'")
        removeFolder(folder)
        if (logsPath != null) {
          val logFolder = logsPath.resolve(folder.name)
          logger.info("remove log folder with version $version because its system folder was removed,  path is '$logFolder'")
          removeFolder(logFolder)
        }
      } else {
        logger.info("keep system folder with version $version, it is not old enough, path is '$folder'")
        keptVersions.add(version)
      }
    }
    return keptVersions
  }
  
  private fun removeOldConfigFolders(path: Path, keptVersions: Set<Version>) {
    val configFolders = productNames.flatMap { productName ->
      path.listDirectoryEntriesSafe("$productName?*") ?: emptyList()
    }

    logger.info("found ${configFolders.size} Code With Me client config folders to check")
    val configsByMajor = configFolders.mapNotNull {
      val currentProductName = productNames.find { name ->  it.name.startsWith(name) } ?: run {
        logger.error("Name of a system folder '${it.name}' doesn't start with any of known product names: ${productNames.joinToString()}")
        return@mapNotNull null
      }
      
      val version = Version.parseVersion(it.name.removePrefix(currentProductName)) 
                    ?: return@mapNotNull null

      it to version
    }.groupBy { it.second.major }
    
    val twoLastMajors = configsByMajor.keys
      .asSequence()
      .filter { it < 1000 } // filter out old version notation (like 2020.2)
      .sortedDescending()
      .take(2)
      .toSet()
    
    logger.info("keep only $twoLastMajors major versions")
    
    configsByMajor.forEach { (major, folders) ->
      val filteredFolders = if (major in twoLastMajors)
        filterOutLatestConfig(folders) else folders
      
      for ((folder, version) in filteredFolders) {
        if (folder.configFolderIsTooOld()) {
          if (version in keptVersions) {
            logger.info("keep config folder with version $version because its system folder still exists, path is '$folder'")
          } else {
            logger.info("remove config folder with version $version because it is older than two weeks, path is '$folder'")
            removeFolder(folder)
          }
        } else {
          logger.info("keep config folder with version $version, it is not old enough, path is '$folder'")
        }
      }
    }
  }
  
  private fun removeFolder(folder: Path) {
    try {
      val tempFolder = FileUtil.createTempDirectory(folder.toFile().parentFile, folder.name, "", false).toPath()
      Files.move(folder, tempFolder, StandardCopyOption.REPLACE_EXISTING)
      FileUtil.delete(tempFolder)
    } catch (e: Exception) {
      logger.error("failed to remove folder $folder", e)
    }
  }
  
  private fun filterOutLatestConfig(folders: List<Pair<Path, Version>>): List<Pair<Path, Version>> {
    val sorted = folders.sortedByDescending { it.second }
    sorted.getOrNull(0)?.let { 
      logger.info("keep config folder with version ${it.second}, it is the most up-to-date config folder with ${it.second.major} major version, path is '${it.first}'")
    }
    return sorted.drop(1)
  }

  private fun Path.configFolderIsTooOld(): Boolean {
    val twoWeeksAgo = Instant.now().minus(14, ChronoUnit.DAYS)
    return listDirectoryEntriesAndSelf().all { (_, attrs) -> 
      attrs.isTooOld(twoWeeksAgo) 
    }
  }

  private fun Path.systemFolderIsTooOld(logsPath: Path?): Boolean {
    val twoWeeksAgo = Instant.now().minus(14, ChronoUnit.DAYS)
    val files = listDirectoryEntriesAndSelf()
    val logFolder = logsPath?.resolve(name) ?: resolve("log")
    val logFiles = logFolder.listDirectoryEntriesAndSelf()
    return files.all { (_, attrs) -> attrs.isTooOld(twoWeeksAgo) } && 
           logFiles.all { (_, attrs) -> attrs.isTooOld(twoWeeksAgo) }
  }
  
  private fun BasicFileAttributes.isTooOld(expiration: Instant): Boolean {
    return lastModifiedTime().toInstant() < expiration
  }

  private fun Path.listDirectoryEntriesAndSelf(): List<Pair<Path, BasicFileAttributes>> {
    val result = mutableListOf<Pair<Path, BasicFileAttributes>>()
    try {
      Files.find(this, 1, { path, attrs ->
        result.add(path to attrs)
        true
      }).use { it.count() }
    } catch (e: Exception) {
      logger.warn("failed to list directory entries with attributes for path '$this'", e)
    }
    return result
  }
  
  private fun Path.listDirectoryEntriesSafe(glob: String = "*"): List<Path>? {
    if (!exists()) return null
    return try {
      listDirectoryEntries(glob)
    } catch (e: Exception) {
      logger.warn("failed to list directory entries for path '$this'", e)
      null
    }
  }
  
  private val Path.seemsLikeLink
    get() = try {
      this != toRealPath()
    } catch (e: Exception) {
      false
    }
}