package com.jetbrains.rd.platform.codeWithMe

import com.intellij.execution.configurations.GeneralCommandLine
import com.intellij.execution.process.OSProcessHandler
import com.intellij.execution.process.ProcessAdapter
import com.intellij.execution.process.ProcessEvent
import com.intellij.ide.impl.ProjectUtil
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.application.impl.LaterInvocator
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.util.BackgroundTaskUtil
import com.intellij.openapi.project.DumbService
import com.intellij.openapi.project.ProjectManager
import com.intellij.openapi.util.BuildNumber
import com.intellij.openapi.util.Key
import com.intellij.openapi.util.SystemInfo
import com.intellij.openapi.util.io.FileSystemUtil
import com.intellij.openapi.util.io.FileUtil
import com.intellij.openapi.wm.WindowManager
import com.intellij.util.BitUtil
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.concurrency.EdtScheduledExecutorService
import com.intellij.util.download.DownloadableFileDescription
import com.intellij.util.download.DownloadableFileService
import com.intellij.util.io.BaseOutputReader
import com.intellij.util.io.Decompressor
import com.intellij.util.io.exists
import com.jetbrains.codeWithMe.lobby.SessionInfoResponse
import com.jetbrains.rd.platform.codeWithMe.util.FileManifestUtil
import com.jetbrains.rd.platform.codeWithMe.util.WindowsFileUtil
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.applicationEx
import java.awt.Frame
import java.io.File
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths
import java.nio.file.attribute.FileTime
import java.util.concurrent.CompletableFuture
import java.util.concurrent.TimeUnit
import javax.swing.SwingUtilities
import kotlin.io.path.div
import kotlin.io.path.name

@Service
class CodeWithMeClientDownloadsManager {
  companion object {
    fun getInstance() = service<CodeWithMeClientDownloadsManager>()
    private val LOG = logger<CodeWithMeClientDownloadsManager>()

    private val trustedUrls = listOf(
      "https://download.jetbrains.com/",
      "https://bintray.com/jetbrains/intellij-jbr/",
      "https://www.jetbrains.com/",
      "https://cache-redirector.jetbrains.com/jetbrains.bintray.com/intellij-jbr/",
      "https://cache-redirector.jetbrains.com/download.jetbrains.com/",
      "https://cache-redirector.jetbrains.com/www.jetbrains.com/",
      "https://buildserver.labs.intellij.net/"
    )

    private val urlAllowedChars = Regex("^[._\\-a-zA-Z0-9:/]+$")
    fun isValidDownloadUrl(url: String): Boolean {
      return urlAllowedChars.matches(url) && !url.contains("..")
    }

    fun isTrustedDownloadUrl(url: String) = isValidDownloadUrl(url) && trustedUrls.any { url.startsWith(it) }

    private const val extractDirSuffix = "-ide"

    private fun skipJbrSymlinkFileFilter(file: Path): Boolean {
      return !(file.name == "jbr" && FileSystemUtil.getAttributes(file.toFile())?.isSymLink == true)
    }

    fun versionsMatch(hostBuildNumberString: String, localBuildNumberString: String): Boolean {
      try {
        val hostBuildNumber = BuildNumber.fromString(hostBuildNumberString)!!
        val localBuildNumber = BuildNumber.fromString(localBuildNumberString)!!

        // Any guest in that branch compatible with SNAPSHOT version (it's used by IDEA developers mostly)
        if ((localBuildNumber.isSnapshot || hostBuildNumber.isSnapshot) && hostBuildNumber.baselineVersion == localBuildNumber.baselineVersion) {
          return true
        }

        return hostBuildNumber.asStringWithoutProductCode() == localBuildNumber.asStringWithoutProductCode()
      }
      catch (t: Throwable) {
        LOG.error("Error comparing versions $hostBuildNumberString and $localBuildNumberString: ${t.message}", t)
        return false
      }
    }
  }

  private data class DownloadableFileData(
    val fileName: String,
    val url: String,
    val targetPath: Path,
    val filesFilter: (Path) -> Boolean = { true }
  )

  fun downloadClientAndJdk(sessionInfoResponse: SessionInfoResponse): CompletableFuture<Pair<Path, Path>?> {
    val guestName = sessionInfoResponse.compatibleClientName
    val guestData = DownloadableFileData(
      fileName = guestName,
      url = sessionInfoResponse.compatibleClientUrl,
      targetPath = getCwmGuestCachesDir() / (guestName + extractDirSuffix),
      filesFilter = ::skipJbrSymlinkFileFilter
    )

    val jdkFullName = sessionInfoResponse.compatibleJreName
    val jdkData = DownloadableFileData(
      fileName = jdkFullName,
      url = sessionInfoResponse.compatibleJreUrl,
      targetPath = getCwmGuestCachesDir() / (jdkFullName + extractDirSuffix)
    )

    val downloadableCollection = hashMapOf<DownloadableFileDescription, DownloadableFileData>()

    val collectResult = collectDownloadableDescriptions(downloadableCollection, guestData, jdkData)
    if (!collectResult) {
      LOG.warn("Unable to collect downloadable file description. Not downloading")
      return CompletableFuture.completedFuture(null)
    }

    if (downloadableCollection.isEmpty()) {
      LOG.info("All directories are up-to-date. Skip downloading.")
      return CompletableFuture.completedFuture(guestData.targetPath to jdkData.targetPath)
    }

    val tempDir = File.createTempFile("jb-cwm-dl", "").also { it.delete() }.parent

    activateWindow()

    val downloadableFileService = DownloadableFileService.getInstance()
    val fileDescriptionList = downloadableCollection.keys.toList()

    val downloader = downloadableFileService.createDownloader(fileDescriptionList, "Code With Me Guest")
    val filesFuture = downloader.downloadWithBackgroundProgress(tempDir, null)

    return filesFuture.thenApplyAsync({ files ->
      application.assertIsNonDispatchThread()

      files ?: let {
        LOG.warn("Download failed. Files argument: $files")
        return@thenApplyAsync null
      }

      if (files.size != fileDescriptionList.size) {
        LOG.warn("Download failed. Expected to process files (${fileDescriptionList.size}): " +
                 "${fileDescriptionList.joinToString("', '", "'", "'")}, " +
                 "but processed (${files.size}): ${files.joinToString("', '", "'", "'")}")
        return@thenApplyAsync null
      }

      files.forEach { filePair ->
        val fileDescription = filePair.second
        val fileData = downloadableCollection[fileDescription] ?: let {
          LOG.warn("Unable to find data for file '${fileDescription.defaultFileName}' in collection")
          return@forEach
        }
        val archivePath = filePair.first.toNioPath()

        extractArchive(fileData.url, archivePath, fileData.targetPath)
        Files.delete(archivePath)
        FileManifestUtil.writeDirectoryManifest(fileData.targetPath, fileData.filesFilter)
      }

      return@thenApplyAsync guestData.targetPath to jdkData.targetPath
    }, AppExecutorUtil.getAppExecutorService())
  }

  private fun collectDownloadableDescriptions(
    map: HashMap<DownloadableFileDescription, DownloadableFileData>,
    vararg dataFiles: DownloadableFileData
  ): Boolean {
    dataFiles.forEach { data ->
      val extractDirectory = FileManifestUtil.getExtractDirectory(data.targetPath, data.filesFilter)
      if (extractDirectory.isUpToDate) return@forEach

      val descriptor = getDownloadableFileDescription(data.url, data.fileName) ?: let {
        LOG.debug("Invalid download url ${data.url}, not downloading")
        return false
      }
      map[descriptor] = data
    }
    return true
  }

  private fun getDownloadableFileDescription(downloadUrl: String, fileName: String): DownloadableFileDescription? {
    val downloadableFileService = DownloadableFileService.getInstance()

    if (!isValidDownloadUrl(downloadUrl)) {
      return null
    }

    LOG.debug("Got downloadable file description for file '$fileName'")
    return downloadableFileService.createFileDescription(downloadUrl, fileName)
  }

  private fun extractArchive(originalUrl: String, archiveFile: Path, targetDir: Path) {
    FileUtil.delete(targetDir)

    when {
      originalUrl.endsWith(".zip", ignoreCase = true) || originalUrl.endsWith(".sit", ignoreCase = true) ->
        Decompressor.Zip(archiveFile).withZipExtensions().extract(targetDir)
      originalUrl.endsWith(".tar.gz", ignoreCase = true) ->
        Decompressor.Tar(archiveFile).extract(targetDir)
      else -> error("Unknown archive extension in file name: ${originalUrl.substringAfterLast('/')}")
    }
  }

  private fun findCwmGuestHome(guestRoot: Path): Path {
    // maxDepth 2 for Mac OS's .app/Contents
    Files.walk(guestRoot, 2).use {
      for (dir in it) {
        if (dir.resolve("bin").exists() && dir.resolve("lib").exists()) {
          return dir
        }
      }
    }

    error("Code With Me Guest home is not found under $guestRoot")
  }

  private fun findLauncher(guestRoot: Path, launcherName: Path): Path? {
    // maxDepth 2 for Mac OS's .app/Contents
    Files.walk(guestRoot, 2).use {
      for (dir in it) {
        val candidate = dir.resolve(launcherName)
        if (candidate.exists()) {
          return candidate
        }
      }
    }

    return null
  }

  private fun findLauncherUnderCwmGuestRoot(guestRoot: Path): List<String> {
    when {
      SystemInfo.isWindows -> {
        val exeLauncherName = Path.of("bin", "cwm_guest64.exe")
        val exeLauncher = findLauncher(guestRoot, exeLauncherName)
        if (exeLauncher != null) {
          return listOf(exeLauncher.toString())
        }

        val oldExeLauncherName = Path.of("bin", "intellij_client64.exe")
        val oldExeLauncher = findLauncher(guestRoot, oldExeLauncherName)
        if (oldExeLauncher != null) {
          return listOf(oldExeLauncher.toString())
        }

        val batLauncherName = Path.of("bin", "intellij_client.bat")
        val batLauncher = findLauncher(guestRoot, batLauncherName)
        if (batLauncher == null) {
          return listOf(batLauncher.toString())
        }

        error("Both '$exeLauncherName' and '$batLauncherName' are missing under $guestRoot")
      }

      SystemInfo.isUnix -> {
        if (SystemInfo.isMac) {
          val app = guestRoot.toFile().listFiles { file -> file.name.endsWith(".app") && file.isDirectory }!!.singleOrNull()
          if (app != null) {
            return listOf("open", app.toString(), "--args")
          }
        }

        val shLauncherName = Path.of("bin", "cwm_guest.sh")
        val shLauncher = findLauncher(guestRoot, shLauncherName)
        if (shLauncher != null) {
          return listOf(shLauncher.toString())
        }

        val oldShLauncherName = Path.of("bin", "intellij_client.sh")
        val oldShLauncher = findLauncher(guestRoot, oldShLauncherName)
        if (oldShLauncher != null) {
          return listOf(oldShLauncher.toString())
        }

        error("Could not find launcher '$shLauncherName' under $guestRoot")
      }

      else -> error("Unsupported OS: ${SystemInfo.OS_NAME}")
    }
  }

  fun runCwmGuestProcessFromDownload(url: String, guestRoot: Path, jdkRoot: Path) {
    val launcher = findLauncherUnderCwmGuestRoot(guestRoot)
    val guestHome = findCwmGuestHome(guestRoot)

    val linkTarget = if (SystemInfo.isMac) jdkRoot / "jbr" else detectTrueJdkRoot(jdkRoot)
    createSymlink(guestHome / "jbr", linkTarget)

    // Update mtime on JRE & CWM Guest roots. The cleanup process will use it later.
    listOf(guestRoot, jdkRoot).forEach { path ->
      Files.setLastModifiedTime(path, FileTime.fromMillis(System.currentTimeMillis()))
    }

    val parameters = listOf("thinClient", url)

    if (SystemInfo.isWindows) {
      WindowsFileUtil.windowsShellExecute(
        executable = Path.of(launcher.single()),
        workingDirectory = guestRoot,
        parameters = parameters
      )
    }
    else {
      // Mac gets multiple start attempts because starting it fails occasionally (CWM-2244, CWM-1733)
      var attemptCount = if(SystemInfo.isMac) 5 else 1
      var lastProcessStartTime: Long

      fun doRunProcess() {
        val commandLine = GeneralCommandLine(launcher + parameters)

        LOG.info("Starting Code With Me Guest process (attempts left: $attemptCount): ${commandLine}")

        attemptCount--
        lastProcessStartTime = System.currentTimeMillis()

        val processHandler = object : OSProcessHandler(commandLine) {
          override fun readerOptions(): BaseOutputReader.Options = BaseOutputReader.Options.forMostlySilentProcess()
        }

        processHandler.addProcessListener(object : ProcessAdapter() {
          override fun onTextAvailable(event: ProcessEvent, outputType: Key<*>) {
            super.onTextAvailable(event, outputType)
            LOG.info("OUTPUT: ${event.text}")
          }

          override fun processTerminated(event: ProcessEvent) {
            super.processTerminated(event)
            LOG.info("Guest process terminated, exit code " + event.exitCode)

            // if process exited abnormally but took longer than 10 seconds, it's likely to be an issue with connection instead of Mac-specific bug
            if (event.exitCode != 0 && (System.currentTimeMillis() - lastProcessStartTime) < 10_000) {
              if (attemptCount > 0) {
                LOG.info("Previous attempt to start guest process failed, will try again in one second")
                EdtScheduledExecutorService.getInstance().schedule({ doRunProcess() }, ModalityState.any(), 1, TimeUnit.SECONDS)
              } else {
                LOG.warn("Running client process failed after specified number of attempts")
              }
            }
          }
        })

        processHandler.startNotify()
      }

      doRunProcess()
    }
  }

  private fun createSymlink(link: Path, target: Path) {
    if (link.exists() && link.toRealPath() == target.toRealPath()) {
      LOG.info("Symlink/junction '$link' is UP-TO-DATE and points to '$target'")
    }
    else {
      Files.deleteIfExists(link)

      LOG.info("Creating symlink/junction '$link' -> '$target'")

      if (SystemInfo.isWindows) {
        WindowsFileUtil.createJunction(junctionFile = link, targetFile = target)
      }
      else {
        Files.createSymbolicLink(link, target)
      }

      if (link.toRealPath() != target.toRealPath()) {
        LOG.error("Symlink/junction '$link' should point to '${target.toRealPath()}', but points to '${link.toRealPath()}' instead")
      }
    }
  }

  private fun detectTrueJdkRoot(jdkDownload: Path): Path {
    jdkDownload.toFile().walk(FileWalkDirection.TOP_DOWN).forEach {
      if (File(it, "bin").isDirectory && File(it, "lib").isDirectory) {
        return it.toPath()
      }
    }

    return jdkDownload
  }
}

fun getJetBrainsSystemCachesDir(): Path {
  if (SystemInfo.isWindows)
    return Paths.get(System.getenv("LOCALAPPDATA")) / "JetBrains"

  if (SystemInfo.isMac)
    return Paths.get(System.getenv("HOME")) / "Library" / "Caches" / "JetBrains"

  val homeDir = System.getenv("XDG_CACHE_HOME")
  val homeDirFile = if (homeDir.isNullOrEmpty())
    Paths.get(System.getenv("HOME")) / ".cache"
  else
    Paths.get(homeDir)

  return homeDirFile / "JetBrains"
}

fun getJetBrainsConfigCachesDir(): Path {
  if (SystemInfo.isWindows)
    return Paths.get(System.getenv("APPDATA")) / "JetBrains"

  if (SystemInfo.isMac)
    return Paths.get(System.getenv("HOME")) / "Library" / "Application Support" / "JetBrains"

  val homeDir = System.getenv("XDG_CONFIG_HOME")
  val homeDirFile = if (homeDir.isNullOrEmpty())
    Paths.get(System.getenv("HOME")) / ".config"
  else
    Paths.get(homeDir)

  return homeDirFile / "JetBrains"
}

fun getJetBrainsSpecialLogsDir(): Path? {
  if (SystemInfo.isMac)
    return Paths.get(System.getenv("HOME")) / "Library" / "Logs" / "JetBrains"

  return null
}

fun getCwmGuestCachesDir(): Path {
  return getJetBrainsSystemCachesDir() / "CodeWithMeClientDist"
}

fun activateWindow() {
  application.invokeLater {
    val guessedProject = ProjectManager.getInstance().openProjects.filterNot { it.isDefault }.firstOrNull()
    var frame = WindowManager.getInstance().findVisibleFrame()

    if (guessedProject != null) {
      ProjectUtil.focusProjectWindow(guessedProject, true)
      frame = WindowManager.getInstance().getFrame(guessedProject)

      DumbService.getInstance(guessedProject).showDumbModeNotification("An appropriate version of the Code With Me Client is being downloaded.\nOnce finished, it will be automatically launched.")
    }

    if (frame != null) {
      if (BitUtil.isSet(frame.extendedState, Frame.ICONIFIED))
        frame.extendedState = BitUtil.set(frame.extendedState, Frame.ICONIFIED, false)
    }

  }
}
