package com.jetbrains.rd.platform.codeWithMe.carets

import com.intellij.openapi.editor.CaretVisualAttributes
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.ex.MarkupModelEx
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.impl.DocumentMarkupModel
import com.intellij.openapi.editor.impl.event.MarkupModelListener
import com.intellij.openapi.editor.markup.RangeHighlighter
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.ui.scale.JBUIScale
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.codeWithMe.model.RemoteCaretData
import java.awt.geom.Rectangle2D
import kotlin.math.max

data class RemoteCaret(
  val editor: Editor,
  private val highlighter: RangeHighlighter,
  val data: RemoteCaretData
) {
  fun addRemoveListener(lifetime: Lifetime, action: () -> Unit) {
    val markup = DocumentMarkupModel.forDocument(editor.document, editor.project, true) as MarkupModelEx
    markup.addMarkupModelListener(lifetime.createNestedDisposable(), object : MarkupModelListener {
      override fun beforeRemoved(h: RangeHighlighterEx) {
        if (h == highlighter) {
          action()
        }
      }
    })
  }

  val isValid get() = highlighter.isValid
  val startOffset get() = highlighter.startOffset
  val endOffset get() = highlighter.endOffset
  val offset get() = if (data.showCaretAtStart) startOffset else endOffset
  val userName get() = data.user.name
}

object RemoteCaretRenderingUtil {
  fun calculatePinRectangle(caret: RemoteCaret, scale: Double): Rectangle2D.Double {
    val caretRectangle = calculateCaretRectangle(caret, scale)
    return calculatePinRectangle(caretRectangle)
  }

  fun calculateCaretRectangle(caret: RemoteCaret, scale: Double): Rectangle2D.Double {
    assert(scale > 0.0)
    val editor = caret.editor
    val isRtl = false // todo editor.myView.isRtlLocation(getVisualPosition())
    val visualPosition = editor.offsetToVisualPosition(caret.offset).leanRight(isRtl)
    val caretPoint = editor.visualPositionToPoint2D(visualPosition)

    val width = CaretVisualAttributes.DEFAULT.getWidth(editor.settings.lineCursorWidth)
    val caretWidth = JBUIScale.scale(width).toDouble()
    val caretShift = if (caretWidth <= 1) 0.0 else 1.0 / scale
    val caretHeight = editor.lineHeight.toDouble() // todo overhangs
    val x = max(caretPoint.x - caretShift, getPinSize(caretWidth) / 2)
    return Rectangle2D.Double(x, caretPoint.y, caretWidth, caretHeight)
  }

  fun calculatePinRectangle(caretRectangle: Rectangle2D.Double): Rectangle2D.Double {
    val pinSize = getPinSize(caretRectangle.width)
    return Rectangle2D.Double(caretRectangle.centerX - pinSize / 2, caretRectangle.y, pinSize, pinSize)
  }

  private fun getPinSize(caretWidth: Double) = caretWidth * 2
}
