package com.jetbrains.rd.platform.codeWithMe.carets

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.Disposable
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.editor.Caret
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.event.CaretEvent
import com.intellij.openapi.editor.event.CaretListener
import com.intellij.openapi.editor.event.SelectionEvent
import com.intellij.openapi.editor.event.SelectionListener
import com.intellij.openapi.editor.ex.MarkupModelEx
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.impl.CaretModelImpl
import com.intellij.openapi.editor.impl.DocumentMarkupModel
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.editor.impl.RangeMarkerImpl
import com.intellij.openapi.editor.impl.event.MarkupModelListener
import com.intellij.openapi.editor.markup.CustomHighlighterRenderer
import com.intellij.openapi.editor.markup.HighlighterLayer
import com.intellij.openapi.editor.markup.HighlighterTargetArea
import com.intellij.openapi.editor.markup.RangeHighlighter
import com.intellij.openapi.fileEditor.FileEditorManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.ui.scale.JBUIScale
import com.intellij.util.Alarm
import com.jetbrains.rd.platform.codeWithMe.control.CodeWithMeColors
import com.jetbrains.rd.platform.codeWithMe.editors.CodeWithMeEditorService
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.IPropertyView
import com.jetbrains.rd.util.reactive.Property
import com.jetbrains.codeWithMe.model.RemoteCaretData
import java.awt.Graphics
import java.awt.Graphics2D

@Service
class RemoteCaretHighlighterService(val project: Project) {
  companion object {
    fun getInstance(project: Project): RemoteCaretHighlighterService = project.service()
    const val REMOTE_CARET_INITIAL_TOOLTIP_DELAY = 1000
  }

  private val tooltipAlarm = Alarm()

  //called only from backend
  fun addRemoteCaret(lifetime: Lifetime,
                     editor: Editor,
                     startOffset: Int,
                     endOffset: Int,
                     remoteCaretData: RemoteCaretData): RangeHighlighter {

    val markup = DocumentMarkupModel.forDocument(editor.document, editor.project, true) as MarkupModelEx
    val highlighter = addRemoteCaret(markup, lifetime, startOffset, endOffset, remoteCaretData)

    CodeWithMeEditorService.getInstance(project).serverNotify(editor.document, ClientId.current)

    return highlighter
  }

  private fun addRemoteCaret(markup: MarkupModelEx,
                             lifetime: Lifetime,
                             startOffset: Int,
                             endOffset: Int,
                             remoteCaretData: RemoteCaretData): RangeHighlighterEx {
    assert(remoteCaretData.lead != null || remoteCaretData.isSetManually)
    val highlighter = markup.addRangeHighlighterAndChangeAttributes(
      null, startOffset, endOffset,
      HighlighterLayer.SELECTION - 1, HighlighterTargetArea.EXACT_RANGE, false) {
      initializeHighlighter(it, remoteCaretData)
    }
    lifetime.onTermination {
      highlighter.dispose()
    }
    return highlighter
  }

  fun findRemoteCaret(offset: Int, editor: Editor): RemoteCaret? {
    val markupModel = DocumentMarkupModel.forDocument(editor.document, editor.project, false) as MarkupModelEx? ?: return null

    var h: RangeHighlighterEx? = null
    markupModel.processRangeHighlightersOverlappingWith(offset, offset) p@{
      val renderer = it.customRenderer
      if (renderer is RemoteCaretHighlighterRenderer && renderer.shouldShow()) {
        h = it
        return@p false
      }
      return@p true
    }
    val highlighter = h
    if (highlighter == null) return null
    val renderer = highlighter.customRenderer as RemoteCaretHighlighterRenderer

    return RemoteCaret(editor, highlighter, renderer.remoteCaretData)
  }

  fun initializeHighlighter(h: RangeHighlighterEx, remoteCaretData: RemoteCaretData) {
    assert(remoteCaretData.lead != null || remoteCaretData.isSetManually)

    if (remoteCaretData.lead?.clientId == ClientId.localId && !remoteCaretData.isSetManually) {
      showFromLocalCaret(h, remoteCaretData)
      return
    }

    val renderer = RemoteCaretHighlighterRenderer(remoteCaretData)
    if (h.startOffset == h.endOffset) {
      (h as? RangeMarkerImpl)?.isStickingToRight = true
    }
    h.customRenderer = renderer

    if (renderer.shouldShow()) {
      val key = CodeWithMeColors.getSelectionAttributesKey(remoteCaretData.user)
      h.setTextAttributesKey(key)

      if (!remoteCaretData.isSetManually) {
        tooltipAlarm.cancelAllRequests()
        tooltipAlarm.addRequest(
          {
            val editor = FileEditorManager.getInstance(project).selectedTextEditor ?: return@addRequest
            if (h.document != editor.document) return@addRequest

            val remoteCaret = RemoteCaret(editor, h, remoteCaretData)
            RemoteCaretTooltipManager.getInstance().showUserTooltip(
              editor, remoteCaret, RemoteCaretTooltipManager.REMOTE_CARET_TOOLTIP_TIMEOUT)
          }, REMOTE_CARET_INITIAL_TOOLTIP_DELAY)
      }
    }
  }

  /**
   * for better UX for a lead we show remote carets of the followers instantly without waiting for a round trip:
   * the lead changes a caret ->
   * a follower receives a change ->
   * the follower moves a caret to match the lead ->
   * the remote caret is sent to the lead
   * The heuristic is based on [RemoteCaretData.isSetManually]
   * which states whether a caret of a follower had the same position as the one of the lead at the moment it was set
   */
  private fun showFromLocalCaret(h: RangeHighlighterEx, remoteCaretData: RemoteCaretData) {
    val editor = FileEditorManager.getInstance(project).selectedTextEditor ?: return
    if (h.document != editor.document) return

    val listenerDisposable = editor.caretModel as CaretModelImpl

    val highlighterLifetime = listenerDisposable.defineNestedLifetime()
    val markup = DocumentMarkupModel.forDocument(editor.document, editor.project, true) as MarkupModelEx
    markup.addMarkupModelListener(listenerDisposable, object : MarkupModelListener {
      override fun beforeRemoved(highlighter: RangeHighlighterEx) {
        if (highlighter == h) {
          highlighterLifetime.terminate()
        }
      }
    })

    val caretData = editor.createViewableCarets(listenerDisposable) {
      it.run {
        Triple(selectionStart, selectionEnd, selectionEnd != offset)
      }
    }

    caretData.view(highlighterLifetime) { lt, triples ->
      for ((start, end, showAtStart) in triples) {
        val data = remoteCaretData.copy(showCaretAtStart = showAtStart, isSetManually = true)
        addRemoteCaret(editor.markupModel as MarkupModelEx, lt, start, end, data)
      }
    }

    return
  }
}

fun <T> Editor.createViewableCarets(listenerDisposable: Disposable, func: (Caret) -> T): IPropertyView<List<T>> {
  fun mapAllCarets(): List<T> {
    return caretModel.allCarets.map(func)
  }

  val caretData = Property(mapAllCarets())

  caretModel.addCaretListener(object : CaretListener {
    override fun caretPositionChanged(event: CaretEvent) {
      caretData.set(mapAllCarets())
    }

    override fun caretAdded(event: CaretEvent) {
      caretData.set(mapAllCarets())
    }

    override fun caretRemoved(event: CaretEvent) {
      caretData.set(mapAllCarets())
    }
  }, listenerDisposable)

  selectionModel.addSelectionListener(object : SelectionListener {
    override fun selectionChanged(e: SelectionEvent) {
      caretData.set(mapAllCarets())
    }
  }, listenerDisposable)
  return caretData
}

class RemoteCaretHighlighterRenderer(val remoteCaretData: RemoteCaretData) : CustomHighlighterRenderer {
  override fun paint(editor: Editor, highlighter: RangeHighlighter, g: Graphics) {
    if (!shouldShow() || g !is Graphics2D) return

    val caret = RemoteCaret(editor as EditorImpl, highlighter, remoteCaretData)
    g.color = CodeWithMeColors.getMarkerAttributes(remoteCaretData.user).backgroundColor
    val caretRectangle = RemoteCaretRenderingUtil.calculateCaretRectangle(caret, JBUIScale.sysScale(g).toDouble())
    val pinRectangle = RemoteCaretRenderingUtil.calculatePinRectangle(caretRectangle)
    g.fill(caretRectangle)
    g.fill(pinRectangle)
  }

  // ignore own highlighters
  fun shouldShow(): Boolean {
    return remoteCaretData.user.clientId != ClientId.localId
  }

  override fun isForeground() = true
}