package com.jetbrains.rd.platform.codeWithMe.carets

import com.intellij.codeInsight.hint.*
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.event.EditorMouseEvent
import com.intellij.openapi.editor.event.EditorMouseEventArea
import com.intellij.openapi.editor.event.EditorMouseMotionListener
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.editor.impl.CaretModelImpl
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.ui.HintHint
import com.intellij.ui.LightweightHint
import com.intellij.ui.scale.JBUIScale
import com.intellij.util.ui.StartupUiUtil
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.rd.platform.codeWithMe.control.CodeWithMeColors
import java.awt.Graphics
import java.awt.Point
import javax.swing.JLabel
import javax.swing.SwingConstants

@Service
class RemoteCaretTooltipManager {
  companion object {
    fun getInstance() = service<RemoteCaretTooltipManager>()
    const val REMOTE_CARET_TOOLTIP_TIMEOUT = 2000
  }

  private val mergingUpdateQueue = MergingUpdateQueue("RemoteCaretTooltipProvider", 100, true, null)
  private val REMOTE_CARET_GROUP = TooltipGroup("Remote carets tooltips", 10)

  class MyMouseEventListener : EditorMouseMotionListener {
    override fun mouseMoved(e: EditorMouseEvent) {
      val editor = e.editor as? EditorEx ?: return
      val project = editor.project ?: return
      val offset = getTargetOffset(e)
      if (offset == -1) return

      val remoteCaret = RemoteCaretHighlighterService.getInstance(project).findRemoteCaret(offset, editor) ?: return
      val tooltipManager = getInstance()
      tooltipManager.showUserTooltip(editor, remoteCaret, REMOTE_CARET_TOOLTIP_TIMEOUT)
    }

    private fun getTargetOffset(event: EditorMouseEvent): Int {
      val editor = event.editor
      val point = event.mouseEvent.point
      if (editor is EditorEx &&
          editor.getProject() != null &&
          event.area === EditorMouseEventArea.EDITING_AREA &&
          event.mouseEvent.modifiers == 0 &&
          editor.foldingModel.getFoldingPlaceholderAt(point) == null
      ) {
        val logicalPosition = editor.xyToLogicalPosition(point)
        val offset = editor.logicalPositionToOffset(logicalPosition)
        if (editor.offsetToPoint2D(offset).distance(point) < JBUIScale.scale(30)) {
          return offset
        }
      }
      return -1
    }
  }

  fun showUserTooltip(editor: Editor, caret: RemoteCaret, timeout: Int = 0) {
    mergingUpdateQueue.queue(Update.create(caret) {
      doShowUserTooltip(editor, caret, timeout)
    })
  }

  private fun doShowUserTooltip(editor: Editor, caret: RemoteCaret, timeout: Int = 0) {
    if (!caret.isValid) return

    // todo JBUIScale.sysScale(g) to match EditorPainter.Session.paintCaret
    val scale = JBUIScale.scale(1.0f).toDouble()
    val pinRectangle = RemoteCaretRenderingUtil.calculatePinRectangle(caret, scale)
    val editorPoint = Point(pinRectangle.x.toInt(), pinRectangle.maxY.toInt())
    val hintHint = HintHint(editor, editorPoint)
    val tooltipRenderer = RemoteCaretTooltipRenderer(caret, timeout)

    TooltipController.getInstance().showTooltip(
      editor,
      hintHint.targetPoint.point,
      tooltipRenderer,
      false,
      REMOTE_CARET_GROUP,
      hintHint
    )
  }
}

class RemoteCaretTooltipRenderer(val caret: RemoteCaret, private val timeout: Int = 0) : TooltipRenderer {
  init {
    assert(timeout >= 0)
  }

  override fun show(editor: Editor, p: Point, alignToRight: Boolean, group: TooltipGroup, hintHint: HintHint): LightweightHint? {
    if (!caret.isValid) return null
    val tooltipComponent = UsernameComponent(caret)

    val lifetimeDef = (editor.caretModel as CaretModelImpl).defineNestedLifetime()
    val hint = object : LightweightHint(tooltipComponent) {
      override fun hide() {
        lifetimeDef.terminate(true)
        super.hide()
      }
    }

    caret.addRemoveListener(lifetimeDef) {
      hint.hide()
    }

    hintHint.apply {
      initStyleFrom(tooltipComponent)
    }

    val height = tooltipComponent.preferredSize.height
    val targetPoint = hintHint.targetPoint
    val enoughHeight = targetPoint.originalPoint.y - editor.scrollingModel.visibleArea.y > height
    // show tooltip on top of remote caret
    val point = targetPoint.point
    val resultPoint = Point(point.x, if (enoughHeight) point.y - height else point.y)

    var flags = HintManager.HIDE_BY_ESCAPE or HintManager.HIDE_BY_ANY_KEY
    // todo more fine-grained control over hint's lifetime
    flags = flags or (if (timeout > 0) HintManager.HIDE_BY_SCROLLING else HintManager.UPDATE_BY_SCROLLING)
    HintManagerImpl.getInstanceImpl().showEditorHint(
      hint,
      editor,
      resultPoint,
      flags,
      timeout,
      false,
      hintHint)

    return hint
  }

  override fun equals(other: Any?): Boolean {
    return caret == caret
  }

  override fun hashCode(): Int {
    return caret.hashCode()
  }
}

private class UsernameComponent(val caret: RemoteCaret) : JLabel(caret.userName, SwingConstants.CENTER) {
  init {
    text = " " + caret.userName + " "
    val attributes = CodeWithMeColors.getMarkerAttributes(caret.data.user)
    background = attributes.backgroundColor
    if (attributes.foregroundColor != null) {
      foreground = attributes.foregroundColor
    }
    font = StartupUiUtil.getLabelFont().deriveFont(attributes.fontType)
  }

  override fun paintComponent(g: Graphics) {
    g.color = background
    g.fillRect(0, 0, width, height)
    super.paintComponent(g)
  }
}
