// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.connection

import com.intellij.DynamicBundle
import com.intellij.internal.statistic.eventLog.EventLogConfiguration
import com.intellij.openapi.application.ApplicationInfo
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.io.HttpRequests
import com.intellij.util.net.HTTPMethod
import com.jetbrains.codeWithMe.license.CMSDataDecoder
import com.jetbrains.codeWithMe.license.CertificateDecoder
import com.jetbrains.codeWithMe.license.LicenseException
import com.jetbrains.codeWithMe.lobby.*
import com.jetbrains.rd.platform.util.fragmentParameters
import com.jetbrains.rd.platform.util.withFragment
import kotlinx.serialization.decodeFromString
import kotlinx.serialization.encodeToString
import kotlinx.serialization.json.*
import java.io.IOException
import java.net.HttpURLConnection
import java.net.URI
import java.nio.charset.StandardCharsets
import java.util.zip.GZIPInputStream

@Suppress("NOTHING_TO_INLINE")
object LobbyServerApi {
  private val logger = Logger.getInstance(javaClass)

  private val cmsDataDecoder = CMSDataDecoder(
    if (isUnitTestOrLocalRun()) CertificateDecoder.getSignaturesDecoderWithTestingCA() else CertificateDecoder.getSignaturesDecoder(),
    false
  )

  private fun isUnitTestOrLocalRun(): Boolean {
    val application = ApplicationManager.getApplication() ?: return false
    return application.isUnitTestMode || ApplicationInfo.getInstance().build.isSnapshot
  }

  // Some random string. Used to seed EventLogConfiguration.anonymize to prevent cross-referencing this install ids to other install IDs
  @Suppress("SpellCheckingInspection")
  private const val INSTALL_ID_CONSTANT_SEED = "8JEwEr7tyEjqtH07iGw3DOmihNsgGmsuUfHG7vb2ScuIrpTHS6yeFGuWJuIS"

  val anonymizedInstallId by lazy {
    EventLogConfiguration.anonymize(INSTALL_ID_CONSTANT_SEED)
  }

  // TODO Handle correct json error message and convert it to user error on the screen

  fun serverInfo(
    serverUrl: URI
  ): Pair<ServerInfoResponse, CertificateDecoder.CertificateData> = jsonPostRequest<Unit, ServerInfoResponse>("$serverUrl/info", HTTPMethod.GET, requestObject = null)

  fun createSession(
    serverUrl: URI,
    request: CreateSessionRequest
  ): Pair<CreateSessionResponse, CertificateDecoder.CertificateData> = jsonPostRequest("$serverUrl/createSession", HTTPMethod.POST, request)

  private fun getSessionUrl(joinLinkUrl: URI): String = joinLinkUrl.withFragment(null).toString().trimEnd('/')

  fun heartbeat(
    serverUrl: URI,
    request: HeartbeatRequest
  ): Pair<HeartbeatResponse, CertificateDecoder.CertificateData> = jsonPostRequest("$serverUrl/heartbeat", HTTPMethod.POST, request)

  fun sessionInfo(
    joinLinkUrl: URI,
    request: SessionInfoRequest
  ): Pair<SessionInfoResponse, CertificateDecoder.CertificateData> = jsonPostRequest(getSessionUrl(joinLinkUrl) + "/info", HTTPMethod.POST, request)

  fun join(
    joinLinkUrl: URI,
    request: JoinRequest
  ): Pair<JoinResponse, CertificateDecoder.CertificateData> = jsonPostRequest(getSessionUrl(joinLinkUrl), HTTPMethod.POST, request)

  fun clientConnected(
    sessionUrl: URI,
    request: ClientConnectedRequest
  ) {
    jsonPostRequest<ClientConnectedRequest, EmptyResponse>("$sessionUrl/clientConnected", HTTPMethod.POST, request)
  }

  fun clientDisconnected(
    sessionUrl: URI,
    request: ClientDisconnectedRequest
  ) {
    jsonPostRequest<ClientDisconnectedRequest, EmptyResponse>("$sessionUrl/clientDisconnected", HTTPMethod.POST, request)
  }

  fun removeSession(
    joinLinkUrl: URI,
    sessionUpdateToken: String?,
  ) {
    val fingerprint = joinLinkUrl.fragmentParameters["fp"] ?: error("Join link lacks host fingerprint: $joinLinkUrl")
    val joinApiUrl = joinLinkUrl.withFragment(null)
    jsonPostRequest<RemoveSessionRequest, EmptyResponse>(
      joinApiUrl.toString(),
      HTTPMethod.DELETE,
      RemoveSessionRequest(hostFingerprint = fingerprint, sessionUpdateToken = sessionUpdateToken),
      allowedHttpErrorCodes = listOf(404)
    )
  }

  private fun getErrorPayload(connection: HttpURLConnection, nonce: LobbyReplyNonce): ByteArray {
    val errorStream = connection.errorStream ?: throw IOException("Connection errorStream is null")
    val stream = if (connection.contentEncoding == "gzip") GZIPInputStream(errorStream) else errorStream
    if (LOBBY_REPLY_CMS_CONTENT_TYPE == connection.contentType) {
      try {
        return stream.use { cmsDataDecoder.decode(it.readBytes(), nonce.bytes).payload }
      }
      catch (ex: LicenseException) {
        throw IOException("Error validating Lobby server error reply signature: ${ex.message}", ex)
      }
    }
    return stream.use { it.readAllBytes() }
  }

  private fun getErrorMessage(responseBodyString: String?): String? {
    if (responseBodyString == null) {
      return null
    }

    try {
      val tree = Json.Default.parseToJsonElement(responseBodyString) as? JsonObject
      val text = (tree?.get("message") as? JsonPrimitive)?.content
      return if (text.isNullOrBlank()) null else text
    }
    catch (_: Throwable) {
      return null
    }
  }

  private fun headers2String(headers: Iterable<Map.Entry<String, Iterable<String>>>) = headers
    .map { entry -> entry.value.map { headerValue -> entry.key to headerValue } }
    .flatten()
    .joinToString("\n") { header -> "${header.first}: ${header.second}" }

  private inline fun <reified TRequest, reified TResponse> jsonPostRequest(
    url: String, method: HTTPMethod, requestObject: TRequest?, allowedHttpErrorCodes: List<Int>? = null): Pair<TResponse, CertificateDecoder.CertificateData> {
    val requestString = requestObject?.let { lobbyModelClientJsonFormat.encodeToString(it) }
    var requestHeaders: String? = null
    var responseString: String? = null
    var responseCode: Int? = null
    var responseHeaders: String? = null

    val startTime = System.currentTimeMillis()

    return try {
      if (method == HTTPMethod.GET) {
        require(requestString == null) {
          "request body must be null for GET method"
        }
      }
      else {
        require(requestString != null) {
          "request body must NOT be null for $method method"
        }
      }

      val requestBuilder = when (method) {
        HTTPMethod.GET -> HttpRequests.request(url)
        HTTPMethod.POST -> HttpRequests.post(url, HttpRequests.JSON_CONTENT_TYPE)
        HTTPMethod.DELETE -> HttpRequests.delete(url, HttpRequests.JSON_CONTENT_TYPE)
        HTTPMethod.PUT -> HttpRequests.put(url, HttpRequests.JSON_CONTENT_TYPE)
        else -> error("Unsupported method: $method (url $url)")
      }

      val nonce = LobbyReplyNonce.random()

      requestBuilder
        .throwStatusCodeException(false)
        .productNameAsUserAgent()
        .tuner {
          it.setRequestProperty("Accept", LOBBY_REPLY_CMS_CONTENT_TYPE)
          it.setRequestProperty(LOBBY_REPLY_NONCE_HEADER, nonce.string)

          val languageBundle = DynamicBundle.findLanguageBundle()
          if (languageBundle?.locale != null) {
            it.setRequestProperty("Accept-Language", languageBundle.locale)
          }

          it.setRequestProperty("X-Anonymized-Install-Id", anonymizedInstallId)

          requestHeaders = headers2String(it.requestProperties.entries).prependIndent("  ")
        }
        .connect { request ->
          if (requestString != null) {
            request.write(requestString)
          }

          val connection = request.connection as HttpURLConnection
          val certData: CertificateDecoder.CertificateData
          try {
            if (connection.contentType != LOBBY_REPLY_CMS_CONTENT_TYPE) {
              throw IOException("Lobby server at '$url' returned unsigned result (Content-Type ${connection.contentType}) which is forbidden")
            }

            val decodeResult = cmsDataDecoder.decode(request.readBytes(null), nonce.bytes)
            certData = decodeResult.certificateData
            responseString = String(decodeResult.payload, StandardCharsets.UTF_8)
          }
          catch (ex: LicenseException) {
            throw IOException("Error validating Lobby server reply signature: ${ex.message}", ex)
          }
          catch (ioException: IOException) {
            val errorPayload = getErrorPayload(connection, nonce)
            responseString = String(errorPayload, Charsets.UTF_8)

            if (connection.responseCode == 402) {
              val element = Json.Default.parseToJsonElement(responseString!!) as? JsonObject
              if (element?.get("messageId")?.jsonPrimitive?.content == LOBBY_MAXIMUM_NUMBER_OF_CONCURRENT_SESSIONS_REACHED) {
                val sessionsLimit = element["sessionsLimit"]?.jsonPrimitive?.content?.toIntOrNull()
                val message = element["message"]?.jsonPrimitive?.content ?: "<no description provided>"
                throw LobbySessionsLimitReachedException(message, sessionsLimit)
              }
            }

            val decodedResponseString = getErrorMessage(responseString) ?: responseString
            throw IOException("HTTP response ${connection.responseCode}: $url\n$decodedResponseString", ioException)
          }
          finally {
            responseCode = connection.responseCode
            responseHeaders = headers2String(connection.headerFields.entries).prependIndent("  ")
          }

          if (responseCode!! >= 400 && allowedHttpErrorCodes?.contains(responseCode!!) != true) {
            throw ErrorHttpResponseException("Request to $url ($method) failed with status code $responseCode", responseCode!!)
          }

          if (TResponse::class == EmptyResponse::class) {
            return@connect Pair(EmptyResponse() as TResponse, certData)
          }

          val tResponse = lobbyModelClientJsonFormat.decodeFromString<TResponse>(responseString!!)
          return@connect Pair(tResponse, certData)
        }
    }
    finally {
      logger.info(
        "HTTP ${method.name} $url in ${System.currentTimeMillis() - startTime} ms. Response code: $responseCode\n" +
        "Request headers:\n$requestHeaders\n" +
        "Request:\n~~~$requestString~~\n\n" +
        "Response headers:\n$responseHeaders\nResponse:\n${responseString?.let { "~~~$it~~~" } ?: "n/a"}")
    }
  }

  fun pollForConnectedClients(listenForClientConnectionEndpoint: URI, hostFingerprint: String, sessionUpdateToken: String?): List<ClientJoinedEvent> =
    jsonPostRequest<JoinPollRequest, JoinPollResponse>(
      listenForClientConnectionEndpoint.toString(),
      HTTPMethod.POST,
      JoinPollRequest(hostFingerprint = hostFingerprint, sessionUpdateToken = sessionUpdateToken)
    ).first.events

  class EmptyResponse

  class LobbySessionsLimitReachedException(message: String, val maximumSessionsLimit: Int?): Exception(message)

  inline fun CertificateDecoder.CertificateData.isOnPremise() : Boolean = isOuFlagSet(CertificateDecoder.ON_PREMISES_LOBBY_OU_FLAG)
}
