package com.jetbrains.rd.platform.codeWithMe.control

import com.intellij.ide.DataManager
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.ActionPlaces
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.project.guessCurrentProject
import com.intellij.openapi.ui.popup.JBPopupFactory
import com.intellij.ui.LayeredIcon
import com.intellij.util.containers.ComparatorUtil
import com.intellij.util.ui.JBDimension
import com.intellij.util.ui.JBUI
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.control.icons.CircleFactories
import com.jetbrains.rd.platform.codeWithMe.control.icons.TextIcon
import com.jetbrains.rd.platform.codeWithMe.statistics.CodeWithMeLifecycleLogger
import com.jetbrains.rd.platform.codeWithMe.toolbar.CodeWithMeUsersGroup
import com.jetbrains.rd.platform.codeWithMe.toolbar.createActionGroup
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.isHost
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import net.miginfocom.swing.MigLayout
import java.awt.Font
import java.awt.Graphics
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import javax.swing.*

class CWMCirclesPane {
  private val separatorGap = JBUI.scale(CWMDrawingData.STATE_HOVER_THICKNESS)
  private val separator = object : JPanel() {
    init {
      preferredSize = JBDimension(1, 10)
      minimumSize = JBDimension(1, 10)
    }

    override fun paint(g: Graphics?) {
      //in case if the theme has changed
      background = UIManager.getColor("Separator.separatorColor")
      super.paint(g)
    }
  }

  private val additionCounter = object : JLabel() {
    private val hoverIcon = CircleFactories.createHoverIcon()

    private val symbolIcon = TextIcon({ hoverIcon.iconWidth },
                                      { hoverIcon.iconWidth },
                                      "K",
                                      null) { CWMDrawingData.iconFont }

    private var hovered = false

    private val hoveredIcon = LayeredIcon(2).apply {
      setIcon(hoverIcon, 0, SwingConstants.CENTER)
      setIcon(symbolIcon, 1, SwingConstants.CENTER)
    }

    init {
      icon = symbolIcon
      isOpaque = false
      addMouseListener(object : MouseAdapter() {
        override fun mouseClicked(e: MouseEvent) {
          if (SwingUtilities.isLeftMouseButton(e)) {
            val dataContext = DataManager.getInstance().getDataContext(view)
            val actionEvent =
              AnActionEvent.createFromInputEvent(e, ActionPlaces.TOOLBAR, null, dataContext)
            userManager?.getCWMActionName()?.let { actionName ->
              ActionManager.getInstance().getAction(actionName).actionPerformed(actionEvent)
            }
          }
        }

        override fun mouseEntered(e: MouseEvent) {
          hovered = true
          updateIcon()
        }

        override fun mouseExited(e: MouseEvent) {
          hovered = false
          updateIcon()
        }
      })
    }

    private fun updateIcon() {
      icon = if (hovered) hoveredIcon else symbolIcon
    }

    fun updateCounter() {
      isVisible = userManager?.let {
        val size = it.users.size
        if (size > CWMDrawingData.MAX_CIRCLES_COUNT) {
          val afterCircles = size - CWMDrawingData.MAX_CIRCLES_COUNT
          symbolIcon.text = if (afterCircles <= CWMDrawingData.MAX_BEFORE_COUNTER) "+$afterCircles" else "${CWMDrawingData.MAX_BEFORE_COUNTER}+"
          updateIcon()
          true
        }
        else {
          false
        }
      } ?: false
    }

    override fun getFont(): Font {
      return CWMDrawingData.iconFont
    }
  }

  private val userCircles = Array(CWMDrawingData.MAX_CIRCLES_COUNT) {
    RdUserCircle(circleLeftClick(), circleRightClick(),
                 {
                   if(userManager?.lead?.value == it)
                     CodeWithMePlatformBundle.message(
                       "circle.control.second.tooltip") else
                     CodeWithMePlatformBundle.message(
                       "circle.control.first.tooltip")
                 })
  }

  private val usualPane = JPanel(MigLayout("ins 0, gap 0, novisualpadding, hidemode 3"))
    .apply {

      userCircles.forEachIndexed { i, rdUserCircle ->
        add(rdUserCircle.view)
        if (i == 0) {
          add(separator, "gapbefore $separatorGap, gapafter $separatorGap")
        }
      }

      add(additionCounter)
      isOpaque = false
    }

  private val noUserPane = CWMDrawingData.createLabel(JBUI.Fonts.toolbarFont(),
                                                      CodeWithMePlatformBundle.message("circle.control.state.no.one.joined"))
  private val container = JPanel(MigLayout("ins 0, gap 0, novisualpadding, hidemode 3")).apply {
    add(noUserPane)
    add(usualPane)
    checkState()
    this.isOpaque = false
  }

  val view: JComponent = container
  private val lifetimes = SequentialLifetimes(Lifetime.Eternal)
  private var userManager: UserManager? = null

  fun show(userManager: UserManager) {
    this.userManager = userManager
    val lifetime = lifetimes.next().lifetime

    userManager.users.advise(lifetime) {
      update(userManager)
    }
    userManager.lead.advise(lifetime) {
      update(userManager)
    }
    userManager.followers.advise(lifetime) {
      update(userManager)
    }
  }

  private fun update(userManager: UserManager) {
    val users = userManager.users.toMutableList()
    val lead = userManager.lead.value
    val host = users.firstOrNull { it.isHost }

    checkState()

    if (users.isEmpty()) {
      return
    }

    val count = ComparatorUtil.min(CWMDrawingData.MAX_CIRCLES_COUNT, users.size)
    separator.isVisible = host?.let {
      if (users.indexOf(host) != 0) {
        users.remove(host)
        users.add(0, host)
      }
      users.size > 1
    } ?: false

    val list = users.subList(0, count).toMutableList()
    lead?.let {
      if (!list.contains(it)) {
        list.removeAt(list.size - 1)
        list.add(it)
      }
    }

    for (i in userCircles.indices) {
      val rdUserCircle = userCircles[i]
      rdUserCircle.view.isVisible = if (i < count) {
        val userModel = list[i]
        rdUserCircle.update(userModel)
        rdUserCircle.state = if (userModel == lead)
          if (userManager.followers.contains(userModel)) Circle.CircleState.FULL_SYNC
          else Circle.CircleState.FOLLOWED
        else Circle.CircleState.NORMAL

        true
      }
      else false
    }

    additionCounter.updateCounter()
  }

  fun hide() {
    this.userManager = null
    checkState()
  }

  private fun circleLeftClick(): (RdUserModel) -> Unit = {
    userManager?.let { um ->
      if (um.lead.value == it) {
        CodeWithMeLifecycleLogger.onCircleLeftClickStop(guessCurrentProject(view), it)
        um.stopFollowingOrSync()
      }
      else {
        CodeWithMeLifecycleLogger.onCircleLeftClickStart(guessCurrentProject(view), it)
        um.jumpTo(it)
        um.startFollowing(it)
      }
    }
  }

  private fun circleRightClick(): (RdUserModel) -> Unit {
    return { user ->
      CodeWithMeLifecycleLogger.onCircleRightClick(guessCurrentProject(view), user)

      userManager?.let {
        val group = it.createActionGroup(user)
        val name = CodeWithMeUsersGroup.getName(user)
        JBPopupFactory.getInstance().createActionGroupPopup(name, group, DataManager.getInstance().getDataContext(view), null,
                                                            false).showUnderneathOf(view)
      }
    }
  }

  private fun checkState() {
    val active = userManager?.users?.isNotEmpty() ?: false
    usualPane.isVisible = active
    noUserPane.isVisible = !active
  }
}