// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.control

import com.intellij.openapi.ui.popup.JBPopupFactory
import com.intellij.ui.ColorUtil
import com.intellij.ui.LayeredIcon
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.control.icons.*
import com.jetbrains.rd.platform.codeWithMe.toolbar.CodeWithMeUsersGroup
import java.awt.Color
import java.awt.Font
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import javax.swing.Icon
import javax.swing.JLabel
import javax.swing.SwingConstants
import javax.swing.SwingUtilities

class RdUserCircle(val onLeftClick: (RdUserModel) -> Unit,
                   val onRightClick: (RdUserModel) -> Unit,
                   val getTooltip: (RdUserModel) -> String?,
                   private var data: RdUserModel? = null) : Circle() {
  fun update(userModel: RdUserModel) {
    data = userModel
    updateCircleColor(userModel.backgroundColor)
    updateName(userModel.name, userModel.foregroundColor)
  }

  init {
    data?.let {
      update(it)
    }

    view.addMouseListener(object : MouseAdapter() {
      override fun mouseClicked(e: MouseEvent) {
        data?.let {
          if (SwingUtilities.isLeftMouseButton(e)) {
            onLeftClick(it)
          }
          else if (SwingUtilities.isRightMouseButton(e)) {
            onRightClick(it)
          }
        }
      }

      override fun mouseEntered(e: MouseEvent) {
        data?.let {
          hovered = true
          showToolTip()
        }
      }

      override fun mouseExited(e: MouseEvent) {
        data?.let {
          hovered = false
        }
      }
    })
  }


  private fun showToolTip() {
    if (!JBPopupFactory.getInstance().isPopupActive())
      data?.let {
        val info = getTooltip(it)
        view.toolTipText = if (info.isNullOrEmpty())
           CodeWithMeUsersGroup.getName(it, escapeHtml = true) else
           CodeWithMePlatformBundle.message("circle.control.tooltip",
                                           CodeWithMeUsersGroup.getName(it, escapeHtml = true),
                                           ColorUtil.toHex(CWMDrawingData.toolTipForeground), info)
      }
  }
}

open class Circle {
  private val circleIcon = CircleIcon(null, CWMDrawingData.CIRCLE_DIAMETER, CWMDrawingData.STATE_HOVER_THICKNESS)
  private val fullSyncIcon = FullSyncIcon(null, CWMDrawingData.CIRCLE_DIAMETER, CWMDrawingData.STATE_HOVER_HEIGHT,
                                          CWMDrawingData.STATE_HOVER_THICKNESS)
  private val followedIcon = FollowedIcon(null, CWMDrawingData.CIRCLE_DIAMETER, CWMDrawingData.STATE_HOVER_HEIGHT,
                                          CWMDrawingData.STATE_HOVER_THICKNESS)
  private val hoverIcon = CircleFactories.createHoverIcon()
  private val symbolIcon = TextIcon({hoverIcon.iconWidth},
                                    {hoverIcon.iconHeight},
                                    "K",
                                    CWMDrawingData.avatarForeground) { CWMDrawingData.iconFont }

  private val compositeIcon = LayeredIcon(2).apply {
    setIcon(circleIcon, 0, SwingConstants.CENTER)
    setIcon(symbolIcon, 1, SwingConstants.CENTER)
  }

  enum class CircleState {
    NORMAL,
    FOLLOWED,
    FULL_SYNC
  }

  var hovered: Boolean = false
    set(value) {
      if (field == value) return
      field = value
      updateIcon()
    }

  var state: CircleState = CircleState.NORMAL
    set(value) {
      if(field == value) return
      field = value
      updateIcon()
    }

  private fun updateIcon() {
    val stateIcon = when(state) {
      CircleState.FOLLOWED -> followedIcon
      CircleState.FULL_SYNC -> fullSyncIcon
      else -> null
    }

    val iconArray = ArrayList<Icon>()
    if(hovered) iconArray.add(hoverIcon)
    iconArray.add(compositeIcon)
    stateIcon?.let {
      iconArray.add(it)
    }

    val lIcon = LayeredIcon(iconArray.size)
    iconArray.forEachIndexed { i, icon ->
      lIcon.setIcon(icon, i, SwingConstants.CENTER)
    }

    view.icon = lIcon

    redraw()
  }

  val view = object : JLabel() {
    override fun getFont(): Font {
      return CWMDrawingData.iconFont
    }
  }.apply { icon = compositeIcon }

  fun updateCircleColor(color: Color) {
    circleIcon.color = color
    fullSyncIcon.color = color
    followedIcon.color = color

    redraw()
  }

  fun updateName(name: String, color: Color) {
    symbolIcon.text = if (name.isNotEmpty()) name[0].toUpperCase().toString() else ""
    symbolIcon.color = color
    redraw()
  }

  private fun redraw() {
    view.revalidate()
    view.repaint()
  }
}





