package com.jetbrains.rd.platform.codeWithMe.control.following

import com.intellij.openapi.components.service
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.fileEditor.FileEditorManager
import com.intellij.openapi.fileEditor.TextEditor
import com.intellij.openapi.fileEditor.impl.FileEditorManagerImpl
import com.intellij.openapi.project.Project
import com.intellij.openapi.wm.ToolWindowManager
import com.intellij.openapi.wm.impl.InternalDecoratorImpl
import com.intellij.ui.AncestorListenerAdapter
import com.intellij.ui.ComponentUtil
import com.intellij.ui.tabs.impl.JBEditorTabs
import com.intellij.ui.tabs.impl.JBTabsImpl
import com.intellij.util.ui.JBUI
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.editors.CodeWithMeEditorService
import com.jetbrains.rd.platform.codeWithMe.users.UserFocusManager
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.AddRemove
import com.jetbrains.rd.util.reactive.Property
import com.jetbrains.rd.util.reactive.adviseWithPrev
import javax.swing.JComponent
import javax.swing.Timer
import javax.swing.event.AncestorEvent

class CodeWithMeEditorMarkService(project: Project) {

  companion object {
    fun getInstance(project: Project): CodeWithMeEditorMarkService = project.service()
  }

  private val userFocusManager = UserFocusManager.getInstance(project)
  private val toolWindowManager = ToolWindowManager.getInstance(project)
  private val userManager = UserManager.getInstance(project)
  private val editorService = CodeWithMeEditorService.getInstance(project)
  private val fileEditorManager = FileEditorManager.getInstance(project) as FileEditorManagerImpl

  private val state = Property(State.NONE)
  private val active = Property(false)
  private var activeComponentUpdater: Timer? = null

  private val markedComponent: Property<JComponent?> = Property(null)
  private val statePaneController: StatePaneController

  init {
    val start = Property(false)

    val projectLifetime = project.lifetime
    userManager.users.advise(projectLifetime) {
      start.set(userManager.users.isNotEmpty())
    }

    val userForResume = Property<RdUserModel?>(null)
    statePaneController = StatePaneController(userManager, state, {
      val tabs = fileEditorManager.splitters.currentWindow?.tabbedPane?.tabs
      if(tabs == null || tabs !is JBTabsImpl) {
        JBUI.scale(1)
      } else {
        tabs.borderThickness
      }
    }, userForResume)

    start.view(projectLifetime) { lifetime, started ->
      if (!started) return@view

      userManager.followers.advise(lifetime) {
        checkState()
      }

      val leadSLT = SequentialLifetimes(lifetime)
      userManager.lead.adviseWithPrev(lifetime) { pr, curr ->
        val lt = leadSLT.next()
        val prev = pr.asNullable

        curr ?: run {
          prev?.let {
            if (userManager.users.contains(it)) {
              markedComponent.value?.let { component ->
                state.set(State.FOLLOWING_STOPPED)

                userForResume.set(it)

                val timer = Timer(5000) {
                  leadSLT.terminateCurrent()
                }

                val ancestorListener = object : AncestorListenerAdapter() {
                  override fun ancestorRemoved(event: AncestorEvent?) {
                    leadSLT.terminateCurrent()
                  }
                }

                component.addAncestorListener(ancestorListener)

                lt.onTermination {
                  component.removeAncestorListener(ancestorListener)
                  timer.stop()
                  checkState(true)
                  userForResume.set(null)
                }

                statePaneController.start(component, lt)
                timer.start()

                userManager.users.adviseAddRemove(lt) { addRemove, i, rdUserModel ->
                  if (addRemove == AddRemove.Remove && rdUserModel == prev) {
                    leadSLT.terminateCurrent()
                  }
                }
                return@adviseWithPrev
              }

            }
          }
        }

        checkState(true)
      }

      active.view(lifetime) { activeLT, value ->
        if (value) {
          activeComponentUpdater = Timer(200) {
            checkMarkedComponent()
            activeComponentUpdater?.start()
          }
          activeComponentUpdater?.start()

          activeLT.onTerminationIfAlive {
            activeComponentUpdater?.stop()
            activeComponentUpdater = null
            markedComponent.set(null)
          }
        }

        markedComponent.view(activeLT) { compLT, comp ->
          comp?.let { c ->
            state.view(compLT) { slt, st ->
              when (st) {
                State.FULL_SYNC, State.FOLLOWING, State.FOLLOWED_BY -> {
                  statePaneController.start(comp, slt)
                }
                else -> {
                }
              }
            }
          }
        }
      }
    }
  }

  private fun checkMarkedComponent() {
    markedComponent.set(getToolWindowComponent() ?: getEditorComponent())
  }

  private fun getEditorComponent(): JComponent? {
    val selectedEditor = fileEditorManager.selectedEditor?.let {
      if (it is TextEditor) it.editor else null
    }

    val editor = if (isSelectedEditorActive()) {
      selectedEditor
    }
    else {
      selectedEditor?.let {se->
        if (editorService.isLeadEditor(se)) se
        else {
          fileEditorManager.allEditors.filterIsInstance<TextEditor>().map { it.editor }.filter {
            editorService.isLeadEditor(it)
          }.firstOrNull()
        }
      }
    }

    return editor?.let {
      if (it !is EditorEx) return null

      val ancestorOfClass = ComponentUtil.getParentOfType(JBEditorTabs::class.java, it.scrollPane)?.parent

      if (ancestorOfClass is JComponent) {
        ancestorOfClass
      }
      else null
    }
  }

  private fun isSelectedEditorActive(): Boolean {
    return state.value == State.FOLLOWED_BY || state.value == State.FOLLOWING_STOPPED
  }

  private fun getToolWindowComponent(): JComponent? {
    val toolWindowsId: String? = when (state.value) {
      State.FOLLOWED_BY,
      State.FOLLOWING_STOPPED -> {
        userFocusManager.currentActiveToolWindowId()
      }
      State.FULL_SYNC,
      State.FOLLOWING -> {
        userFocusManager.leadActiveToolWindowId()
      }
      else -> {
        null
      }
    }

    return toolWindowsId?.let {
      val toolWindow = toolWindowManager.getToolWindow(it)
      toolWindow?.let {
        if (toolWindow.isVisible.not()) {
          if (state.value == State.FOLLOWING) {
            CodeWithMeToolWindowCookie.autoOpen {
              toolWindow.show()
            }
          }
          if (state.value == State.FULL_SYNC) {
            CodeWithMeToolWindowCookie.autoOpen {
              toolWindow.show {
                userFocusManager.fireCurrentActiveToolWindowChanged(toolWindowsId)
              }
            }
          }
        }

        val parent = it.component.parent
        if (parent is InternalDecoratorImpl) {
          parent
        }
        else null
      }
    }
  }

  private fun checkState(clear: Boolean = false) {
    if (state.value == State.FOLLOWING_STOPPED && !clear) return

    state.set(when {
                userManager.users.size > 0 -> {
                  userManager.lead.value?.let {
                    if (userManager.isFullSyncWith(it)) {
                      State.FULL_SYNC
                    }
                    else {
                      State.FOLLOWING
                    }
                  } ?: run {
                    val followers = userManager.followers
                    if (followers.isEmpty()) {
                      State.NONE
                    }
                    else {
                      State.FOLLOWED_BY
                    }
                  }
                }
                else -> {
                  State.NONE
                }
              })
    active.set(state.value != State.NONE)
  }
}