package com.jetbrains.rd.platform.codeWithMe.control.following

import com.intellij.openapi.rd.fill2DRect
import com.intellij.ui.awt.RelativePoint
import com.intellij.ui.awt.RelativeRectangle
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.control.backgroundColor
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.IProperty
import com.jetbrains.rd.util.reactive.Property
import java.awt.*
import java.awt.event.ComponentAdapter
import java.awt.event.ComponentEvent
import javax.swing.JComponent
import javax.swing.JLayeredPane

class StatePaneController(val userManager: UserManager,
                          val state: IProperty<State>,
                          val getTabThickness: () -> Int,
                          private val userForResume: IProperty<RdUserModel?>) {

  private val statePane: Property<StatePane?> = Property(null)
  private val layeredPane: Property<JLayeredPane?> = Property(null)

  private val followingStoppedPane = FollowingStoppedPane {
    if (userManager.users.contains(it)) {
      userManager.startFollowing(it)
    }
  }

  private val followingPane = FollowingPane(
    {
      userManager.stopFollowingOrSync()
    })

  private val top = createBorderLine()
  private val bottom = createBorderLine()
  private val left = createBorderLine()
  private val right = createBorderLine()

  private fun createBorderLine(): JComponent {
    return object : JComponent() {
      override fun paint(g: Graphics) {
        super.paint(g)
        g as Graphics2D

        val rectangle = Rectangle(0, 0, width, height)
        g.fill2DRect(rectangle, background)
      }
    }
  }

  private val fullSyncPane = FullSyncPane {
    userManager.stopFollowingOrSync()
  }
  private val followedByPane = FollowedByPane()

  fun start(component: JComponent, lifetime: Lifetime) {
    if (layeredPane.value != UIUtil.getRootPane(component)?.layeredPane)
      layeredPane.set(UIUtil.getRootPane(component)?.layeredPane)

    state.view(lifetime) { slt, st ->
      statePane.set(
        when (st) {
          State.FOLLOWING_STOPPED -> {
            followingStoppedPane
          }
          State.FULL_SYNC -> {
            fullSyncPane
          }
          State.FOLLOWING -> {
            followingPane
          }
          State.FOLLOWED_BY -> {
            followedByPane
          }
          else -> {
            null
          }
        }
      )
    }


    layeredPane.view(lifetime) { layLT, layer ->
      layer?.let { l ->
        state.view(lifetime) { stLT, st ->
            when (st) {
              State.FOLLOWING_STOPPED -> {
                followingStoppedPane
              }
              State.FULL_SYNC -> {
                fullSyncPane
              }
              State.FOLLOWING -> {
                followingPane
              }
              State.FOLLOWED_BY -> {
                followedByPane
              }
              else -> {
                null
              }
            }?.let { p ->
              val view = p.view
              val listener = object : ComponentAdapter() {
                override fun componentMoved(e: ComponentEvent?) {
                  updatePaneBounds(component, layer, view)
                  updateBorderBounds(component, layer)
                }

                override fun componentResized(e: ComponentEvent?) {
                  updatePaneBounds(component, layer, view)
                  updateBorderBounds(component, layer)
                }
              }

              stLT.bracket({
                             var le = 12
                             when (st) {
                               State.FULL_SYNC, State.FOLLOWING -> {
                                 userManager.lead.view(stLT) { leadLT, it ->
                                   it?.let {

                                     top.background = it.backgroundColor
                                     bottom.background = it.backgroundColor
                                     left.background = it.backgroundColor
                                     right.background = it.backgroundColor


                                     l.add(top, Integer(le))
                                     l.add(bottom, Integer(le++))
                                     l.add(left, Integer(le++))
                                     l.add(right, Integer(le++))

                                     updateBorderBounds(component, l)

                                     leadLT.onTermination {
                                       l.remove(top)
                                       l.remove(bottom)
                                       l.remove(left)
                                       l.remove(right)
                                     }
                                   }
                                 }
                               }
                               else -> {
                               }
                             }

                             l.add(view, Integer(le++))

                             when (p) {
                               is FollowingPane -> {
                                 userManager.lead.advise(stLT) {
                                   p.user = userManager.lead.value
                                   updatePaneBounds(component, layer, view)
                                 }
                               }
                               is FollowedByPane -> {
                                 userManager.followers.advise(stLT) {
                                   p.followers = userManager.followers
                                   updatePaneBounds(component, layer, view)
                                 }
                               }
                               is FollowingStoppedPane -> {
                                 userForResume.advise(stLT) {
                                   p.user = it
                                   updatePaneBounds(component, layer, view)
                                 }
                               }

                             }
                             component.addComponentListener(listener)
                           },
                           {
                             component.removeComponentListener(listener)
                             l.remove(view)
                             l.repaint()
                           })
            }
        }
      }
    }
  }

  private val t: Int = JBUI.scale(3)

  private fun getBounds(component: JComponent, layer: JLayeredPane): Rectangle {
    val bounds = RelativeRectangle(component).getRectangleOn(layer)

    bounds.width -= t
    bounds.height -= t
    return bounds
  }

  private fun updatePaneBounds(component: JComponent,
                               layer: JLayeredPane,
                               view: JComponent) {
    val bounds = getBounds(component, layer)

    val d = view.preferredSize

    view.bounds = Rectangle((bounds.maxX - d.width + t).toInt(), (bounds.maxY - d.height + t).toInt(), d.width, d.height)
  }

  private fun updateBorderBounds(component: JComponent,
                               layer: JLayeredPane) {
    val bounds = getBounds(component, layer)

    top.setBounds(bounds.x, bounds.y, bounds.width, t)

    bottom.setBounds(bounds.x, bounds.y + bounds.height, bounds.width, t)
    left.setBounds(bounds.x, bounds.y, t, bounds.height)
    right.setBounds(bounds.x + bounds.width, bounds.y, t, bounds.height)
  }
}