package com.jetbrains.rd.platform.codeWithMe.debugger

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.VirtualFileManager
import com.intellij.util.containers.BidirectionalMap
import com.intellij.xdebugger.XDebuggerManager
import com.intellij.xdebugger.breakpoints.*
import com.jetbrains.codeWithMe.model.CWMBreakpointModelBase
import com.jetbrains.codeWithMe.model.CWMLineBreakpointModel
import com.jetbrains.codeWithMe.model.CodeWithMeBreakpointsModel
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.ide.model.RdFileId
import com.jetbrains.rd.platform.codeWithMe.util.fromModel
import com.jetbrains.rd.platform.codeWithMe.util.toModel
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeStatus
import com.jetbrains.rd.util.reactive.adviseEternal

typealias CommonBreakpointId = String

abstract class CommonBreakpointHost(
  val lifetime: Lifetime,
  val protocol: IProtocol,
  val project: Project
) {
  protected val model = CodeWithMeBreakpointsModel.create(lifetime, protocol)
  protected val breakpointMap = BidirectionalMap<XBreakpoint<*>, CommonBreakpointId>()

  init {
    syncBreakpointsFromModel()
    syncBreakpointsToModel()
  }

  abstract fun findFile(fileId: RdFileId): VirtualFile?
  abstract fun findFileId(file: VirtualFile): RdFileId?

  abstract fun beforeBreakpointAdd(breakpoint: XBreakpoint<*>)

  abstract fun doIgnoreBreakpoint(breakpoint: XBreakpoint<*>): Boolean
  
  fun getBreakpointById(breakpointId: String): XBreakpoint<*>? {
    val keys = breakpointMap.getKeysByValue(breakpointId)
    if (keys.isNullOrEmpty())
      return null
    
    if (keys.size == 1)
      return keys.single()
    
    logger.error("Multiple breakpoints have the same id = $breakpointId, breakpoints are ${keys.joinToString()}")
    throw IllegalStateException("Multiple breakpoints with the same id")
  }

  private fun syncBreakpointsFromModel() {
    val breakpointManager = XDebuggerManager.getInstance(project).breakpointManager

    model.breakpoints.view(lifetime) { breakpointLifetime, breakpointId, breakpointModel ->
      when (breakpointModel) {
        is CWMLineBreakpointModel -> {
          val breakpointType = XBreakpointType.EXTENSION_POINT_NAME.extensionList.find {
            it.id == breakpointModel.breakpointTypeId
          }
          if (breakpointType == null) {
            logger.error("Can't find breakpointType with id = ${breakpointModel.breakpointTypeId}")
            return@view
          }
          if (breakpointType !is XLineBreakpointType<*>) {
            logger.error("Found breakpointType with id = ${breakpointModel.breakpointTypeId}, but it doesn't implement XLineBreakpointType")
            return@view
          }

          fun registerLineBreakpoint(file: VirtualFile) {
            //workaround for KT-43630. Var declaration could be removed once the ticket is fixed
            //discussion in slack https://jetbrains.slack.com/archives/C0288G57R/p1606468620399600
            val breakpointType = breakpointType as XLineBreakpointType<*>

            breakpointLifetime.bracket({
              if (breakpointMap.containsValue(breakpointId))
                return@bracket

              application.runWriteAction {
                val breakpoint = breakpointType.addLineBreakpoint(breakpointManager, file, breakpointModel)
                if (breakpoint != null) {
                  breakpointMap[breakpoint] = breakpointId
                  adviseOnBreakpointProperties(breakpoint, breakpointModel)
                }
              }
            }) {
              if (lifetime.status == LifetimeStatus.Terminating)
                return@bracket

              application.runWriteAction {
                val breakpoint = breakpointManager.findBreakpointAtLine(breakpointType, file, breakpointModel.line) ?: run {
                  logger.warn("Not found a breakpoint to delete: type = $breakpointType, file = $file, line = ${breakpointModel.line}")
                  return@runWriteAction
                }

                breakpointManager.removeBreakpoint(breakpoint)
              }
            }
          }

          val file = findFile(breakpointModel.fileId)
          if (file == null) {
            logger.warn("File ${breakpointModel.fileId} was not found")
            return@view
          }

          registerLineBreakpoint(file)
        }
        else -> {
          // TODO
        }
      }
    }
  }

  private fun setBreakpointProperties(breakpoint: XBreakpoint<*>, breakpointModel: CWMBreakpointModelBase) {
    application.assertIsDispatchThread()
    breakpointModel.isEnabled.set(breakpoint.isEnabled)
    breakpointModel.suspendPolicy.set(breakpoint.suspendPolicy.toModel())
  }

  private fun adviseOnBreakpointProperties(breakpoint: XBreakpoint<*>, breakpointModel: CWMBreakpointModelBase) {
    application.assertIsDispatchThread()
    breakpointModel.isEnabled.adviseEternal { breakpoint.isEnabled = it }
    breakpointModel.suspendPolicy.adviseEternal { breakpoint.suspendPolicy = it.fromModel() }
  }

  private fun createLineBreakpointModel(
    breakpoint: XLineBreakpoint<*>,
    breakpointId: String
  ): CWMLineBreakpointModel? {
    application.assertIsDispatchThread()
    val file = VirtualFileManager.getInstance().findFileByUrl(breakpoint.fileUrl) ?: return null
    val fileId = findFileId(file) ?: return null
    return CWMLineBreakpointModel(
      fileId,
      breakpoint.line,
      breakpointId,
      breakpoint.type.id
    ).apply {
      setBreakpointProperties(breakpoint, this)
      adviseOnBreakpointProperties(breakpoint, this)
    }
  }

  protected fun addBreakpointToModel(breakpoint: XBreakpoint<*>) {
    if (breakpoint in breakpointMap)
      return

    val breakpointId = ClientId.localId.value + id.nextId()
    breakpointMap[breakpoint] = breakpointId

    beforeBreakpointAdd(breakpoint)
    when (breakpoint) {
      is XLineBreakpoint<*> -> {
        val breakpointModel = createLineBreakpointModel(breakpoint, breakpointId)!!
        model.breakpoints[breakpointId] = breakpointModel
      }
      else -> {

      }
    }
    // TODO
  }

  private fun syncBreakpointsToModel() {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())

    val clientId = ClientId.currentOrNull
    connection.subscribe(XBreakpointListener.TOPIC, object : XBreakpointListener<XBreakpoint<*>> {
      override fun breakpointAdded(breakpoint: XBreakpoint<*>) {
        if (doIgnoreBreakpoint(breakpoint))
          return

        application.invokeLater {
          ClientId.withClientId(clientId) {
            addBreakpointToModel(breakpoint)
          }
        }
      }

      override fun breakpointRemoved(breakpoint: XBreakpoint<*>) {
        if (doIgnoreBreakpoint(breakpoint))
          return

        application.invokeLater {
          val breakpointId = breakpointMap.remove(breakpoint) ?: run {
            logger.error("Breakpoint $breakpoint was not found, can't remove it from model")
            return@invokeLater
          }

          model.breakpoints.remove(breakpointId)
        }
      }

      override fun breakpointChanged(breakpoint: XBreakpoint<*>) {
        if (doIgnoreBreakpoint(breakpoint))
          return

        application.invokeLater {
          val breakpointId = breakpointMap[breakpoint] ?: run {
            logger.error("Breakpoint $breakpoint was not found, can't remove it from model")
            return@invokeLater
          }

          val breakpointModel = model.breakpoints[breakpointId] ?: run {
            logger.error("Can't find breakpoint model for id='$breakpointId'")
            return@invokeLater
          }

          setBreakpointProperties(breakpoint, breakpointModel)
        }
      }
    })
  }

  private fun <P: XBreakpointProperties<*>> XLineBreakpointType<P>.addLineBreakpoint(breakpointManager: XBreakpointManager, file: VirtualFile, model: CWMLineBreakpointModel): XLineBreakpoint<P>? {
    val existingBreakpoint = breakpointManager.findBreakpointAtLine(this, file, model.line)
    if (existingBreakpoint != null)
      return null

    return breakpointManager.addLineBreakpoint(this, file.url, model.line, createBreakpointProperties(file, model.line))
  }

  companion object {
    private val logger = getLogger<CommonBreakpointHost>()

    private val id = object {
      private var nextId = 0
      fun nextId() = nextId++
    }
  }
}