package com.jetbrains.rd.platform.codeWithMe.diagnostics

import com.intellij.CommonBundle
import com.intellij.codeWithMe.ClientId
import com.intellij.ide.IdeBundle
import com.intellij.ide.actions.RevealFileAction
import com.intellij.notification.*
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.Task
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.completeWith
import java.io.File
import java.util.concurrent.CompletableFuture

abstract class CodeWithMeCollectZippedLogsActionBase : AnAction(), DumbAware {

  private val logger = Logger.getInstance(javaClass)

  override fun actionPerformed(e: AnActionEvent) {
    runAction(e.project, askAboutClients = true, askAboutSensitiveData = true).handleAsync { zippedLogsFile, throwable ->
      when {
        throwable != null -> {
          logger.warn(throwable)
          val errorNotification = Notification(
            NOTIFICATION_GROUP.displayId,
            CodeWithMePlatformBundle.message("product.name"),
            IdeBundle.message("notification.content.can.t.create.zip.file.with.logs.0", throwable.localizedMessage),
            NotificationType.ERROR)
          Notifications.Bus.notify(errorNotification)
        }

        zippedLogsFile != null -> {
          if (RevealFileAction.isSupported()) {
            application.invokeLater {
              RevealFileAction.openFile(zippedLogsFile)
            }
          }
          else {
            val logNotification = Notification(
              NOTIFICATION_GROUP.displayId,
              CodeWithMePlatformBundle.message("product.name"),
              IdeBundle.message("notification.content.log.file.is.created.0", zippedLogsFile.absolutePath),
              NotificationType.INFORMATION)
            Notifications.Bus.notify(logNotification)
          }
        }
      }
    }
  }

  fun runAction(project: Project?, askAboutSensitiveData: Boolean, askAboutClients: Boolean): CompletableFuture<File?> {
    application.assertIsDispatchThread()

    ClientId.withClientId(ClientId.localId) {
      if (askAboutSensitiveData) {
        val result = Messages.showOkCancelDialog(
          project, IdeBundle.message("message.included.logs.and.settings.may.contain.sensitive.data"),
          IdeBundle.message("dialog.title.sensitive.data"),
          CommonBundle.getOkButtonText(), CommonBundle.getCancelButtonText(), Messages.getWarningIcon()
        )
        if (result != Messages.OK) return CompletableFuture.completedFuture(null)
      }

      val users = project?.let { UserManager.getInstance(it).users.toList() }
      val collectClientLogs = when {
        !askAboutClients -> true
        users == null -> false
        users.isEmpty() -> false
        else -> {
          // Always collect for CWM EAP. TODO Enable again later
          true
          //val connectsClientsLogDialogResult = Messages.showYesNoDialog(
          //  project,
          //  CodeWithMePlatformBundle.message("collect.log.action.confirm.collecting.from.clients"),
          //  CodeWithMePlatformBundle.message("product.name"),
          //  Messages.getQuestionIcon()
          //)
          //if (connectsClientsLogDialogResult == Messages.YES) {
          //  true
          //}
          //else {
          //  logger.info("User does not want to include log files from connected clients")
          //  false
          //}
        }
      }

      val fileResult = CompletableFuture<File?>()

      ProgressManager.getInstance().run(
        object : Task.Backgroundable(null, IdeBundle.message("progress.title.collecting.logs"), true) {
          override fun run(indicator: ProgressIndicator) {
            fileResult.completeWith {
              collectLogs(project, indicator, collectClientLogs)
            }
          }
        })

      return fileResult
    }
  }

  abstract fun collectLogs(project: Project?, indicator: ProgressIndicator, collectClientLogs: Boolean): File?

  companion object {
    val NOTIFICATION_GROUP = NotificationGroup("Code With Me Zipped Logs", NotificationDisplayType.BALLOON, true)
  }
}