// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.diagnostics

import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.util.io.FileUtil
import com.jetbrains.rd.platform.protocol.asRdDispatcher
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.ide.model.FileTransmissionChunk
import com.jetbrains.rd.ide.model.FileTransmissionModel
import com.jetbrains.rd.ide.model.FileTransmissionRequest
import com.jetbrains.rd.ide.model.FileTransmissionSession
import java.io.File
import java.util.concurrent.CompletableFuture
import java.util.concurrent.ConcurrentHashMap

class FileTransmissionHandler(lifetime: Lifetime, private val transmissionModel: FileTransmissionModel): FileTransmission {
  private val LOG = Logger.getInstance(javaClass)

  private val filesToReceive = ConcurrentHashMap<FileTransmissionRequest, CompletableFuture<File>>()
  private val filesToSend = ConcurrentHashMap<FileTransmissionRequest, Pair<File, CompletableFuture<Unit>>>()

  init {
    transmissionModel.sessions.view(lifetime) { sessionLt, request, session ->
      val onFileReceived = filesToReceive.remove(request)
      if (onFileReceived != null) {
        handleReceive(sessionLt, session, request, onFileReceived)
        return@view
      }

      val sendData = filesToSend.remove(request)
      if (sendData != null) {
        val (fileToSend, onFileSent) = sendData
        handleSend(sessionLt, session, request, fileToSend, onFileSent)
        return@view
      }

      LOG.error("Received unexpected FileTransmissionRequest: fileName=${request.name} uuid=${request.uuid}")
    }
  }

  private fun handleSend(sessionLt: Lifetime, session: FileTransmissionSession, request: FileTransmissionRequest, fileToSend: File, onFileSent: CompletableFuture<Unit>) {
    fun <R> catchAndTerminateSession(block: () -> R): R? {
      try {
        return block()
      } catch (t: Throwable) {
        session.chunk.fire(FileTransmissionChunk(byteArrayOf(), true, "${t.javaClass.name}: ${t.message}"))
        onFileSent.completeExceptionally(t)
        return null
      }
    }

    val stream = catchAndTerminateSession {
      fileToSend.inputStream().also { sessionLt.onTermination(it) }
    } ?: return

    val buffer = ByteArray(512 * 1024)

    fun sendNext(): Boolean = catchAndTerminateSession {
        val bytesRead = stream.read(buffer)
        if (bytesRead < 0) {
          session.chunk.fire(FileTransmissionChunk(byteArrayOf(), true, null))
          onFileSent.complete(Unit)
          false
        } else {
          session.chunk.fire(FileTransmissionChunk(buffer.copyOf(bytesRead), false, null))
          true
        }
      } ?: false

    val sendNextRc = sendNext()
    if (sendNextRc) {
      session.ack.advise(sessionLt) {
        sendNext()
      }
    }

    sessionLt.onTermination {
      LOG.debug("Terminating sending file $request from $fileToSend")
    }
  }

  private fun handleReceive(sessionLt: Lifetime, session: FileTransmissionSession, request: FileTransmissionRequest, onFileReceived: CompletableFuture<File>) {
    val tempFile = FileUtil.createTempFile(request.name, ".tmp", true)
    val stream = tempFile.outputStream().also { sessionLt.onTermination(it) }

    session.chunk.adviseOn(sessionLt, session.protocol.scheduler.asRdDispatcher.backgroundScheduler) { chunk ->
      stream.write(chunk.bytes)

      if (chunk.eof) {
        val errorString = chunk.error
        LOG.debug("Received EOF signal, errorString=$errorString")

        catch { stream.close() }

        transmissionModel.protocol.scheduler.invokeOrQueue {
          transmissionModel.sessions.remove(request)
        }

        if (errorString != null) {
          CompletableFuture.runAsync {
            catch {
              tempFile.delete()
            }
          }
          onFileReceived.completeExceptionally(IllegalStateException("File sending error on remote side: $errorString"))
        }
        else {
          onFileReceived.complete(tempFile)
        }
      } else {
        session.ack.fire(Unit)
      }
    }

    sessionLt.onTermination {
      LOG.debug("Terminating receiving file $request")
    }
  }

  override fun sendFileAsync(file: File, request: FileTransmissionRequest): CompletableFuture<Unit> {
    LOG.info("Sending file '${request.name} ${request.uuid}' from $file via Rd protocol")

    val future = CompletableFuture<Unit>()

    future.handleAsync { _, exception ->
      if (exception != null) {
        LOG.warn("File sending of '${request.name} ${request.uuid}' completed with an error: ${exception.message}", exception)
      } else {
        LOG.info("Successfully sent file '${request.name} ${request.uuid}' to the other side")
      }
    }

    filesToSend[request] = file to future

    return future
  }

  override fun receiveFileAsync(request: FileTransmissionRequest): CompletableFuture<File> {
    LOG.info("Receiving file '${request.name}' via Rd protocol")

    val future = CompletableFuture<File>()

    future.handleAsync { _, exception ->
      if (exception != null) {
        LOG.warn("File receiving of '${request.name} ${request.uuid}' completed with an error: ${exception.message}", exception)
      } else {
        LOG.info("Successfully received file '${request.name} ${request.uuid}'")
      }
    }

    filesToReceive[request] = future

    transmissionModel.protocol.scheduler.invokeOrQueue {
      transmissionModel.sessions[request] = FileTransmissionSession()
    }

    return future
  }
}