// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.util.io.ByteArraySequence
import com.intellij.util.containers.MultiMap
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.catchAndDrop
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.Signal
import java.io.IOException
import java.net.DatagramPacket
import java.net.DatagramSocket
import java.net.InetSocketAddress
import java.nio.ByteBuffer
import kotlin.concurrent.thread

class DatagramSocketWrapper(val lifetime: Lifetime) {
  private val socket = DatagramSocket().also { it.reuseAddress = true }

  val localAddress: InetSocketAddress?
    get() = socket.localSocketAddress as? InetSocketAddress

  private val remainingPacketsEvent = Signal<Pair<InetSocketAddress, ByteArray>>()
  fun adviseOnRemainingPackets(adviseLifetime: Lifetime, handler: (InetSocketAddress, ByteArray) -> Unit) {
    remainingPacketsEvent.advise(adviseLifetime) {
      handler(it.first, it.second)
    }
  }

  private val mappedEvents = MultiMap.createConcurrent<InetSocketAddress, (ByteArray) -> Unit>()
  fun adviseOnSpecificAddressPackets(adviseLifetime: Lifetime, address: InetSocketAddress, handler: (ByteArray) -> Unit) {
    mappedEvents.putValue(address, handler)
    adviseLifetime.onTermination {
      mappedEvents.remove(address, handler)
    }
  }

  // ByteArraySequence for equals
  private val prefixedEvents = MultiMap.createConcurrent<ByteArraySequence, (InetSocketAddress, ByteArray) -> Unit>()
  fun adviseOnSpecificContentPrefix(adviseLifetime: Lifetime, prefix: ByteArray, handler: (InetSocketAddress, ByteArray) -> Unit) {
    val seq = ByteArraySequence(prefix)
    prefixedEvents.putValue(seq, handler)
    adviseLifetime.onTermination {
      prefixedEvents.remove(seq, handler)
    }
  }

  fun send(buf: ByteArray, address: InetSocketAddress) {
    val p = DatagramPacket(buf, 0, buf.size, address)
    socket.send(p)
  }

  init {
    val threadName = "UDP receive on ${socket.localSocketAddress}"
    val thread = thread(isDaemon = true, name = threadName) {
      LOG.debug("Receive thread '$threadName' is operational")
      val buffer = ByteBuffer.allocate(64 * 1024)
      try {
        receivePacketLoop@ while (true) {
          val packet = DatagramPacket(buffer.array(), 0, buffer.capacity())

          try {
            // TODO Error handling? Should we ignore unreachable destinations?
            socket.receive(packet)
          }
          catch (t: IOException) {
            if (socket.isClosed) {
              LOG.debug("Receive thread '$threadName': socket is closed")
              break
            }

            if (lifetime.isNotAlive) {
              LOG.debug("Receive thread '$threadName': lifetime is terminated")
              break
            }

            if (Thread.currentThread().isInterrupted) {
              LOG.debug("Receive thread '$threadName': thread was interrupted")
              break
            }

            // Should no happen, but
            LOG.warn("I/O exception receiving UDP packet: ${t.message}", t)

            continue
          }

          LOG.trace("received an UDP packet from ${packet.socketAddress} size ${packet.length}")

          // TODO In the future we may require in-place packet processing to reduce allocations
          // Probably not a priority in IDE scenarios
          val packetData = packet.data.copyOf(packet.length)
          val remoteAddress = packet.socketAddress as InetSocketAddress

          for ((prefixSeq, handlers) in prefixedEvents.entrySet()) {
            if (packetData.size >= prefixSeq.length && ByteArraySequence(packetData, 0, prefixSeq.length) == prefixSeq) {
              for (handler in handlers) {
                catch("UDP packet handler for specific prefix, remoteAddress: $remoteAddress, ${handler.javaClass.name}") {
                  handler(remoteAddress, packetData)
                }
              }

              continue@receivePacketLoop
            }
          }

          val specificHandlers = mappedEvents[remoteAddress]
          if (specificHandlers.isNotEmpty()) {
            for (specificHandler in specificHandlers) {
              catch("UDP packet handler for $remoteAddress: ${specificHandler.javaClass.name}") {
                specificHandler(packetData)
              }
            }
          }
          else {
            remainingPacketsEvent.fire(remoteAddress to packetData)
          }
        }
      }
      catch (_: InterruptedException) {
      }
      catch (t: Throwable) {
        LOG.warn("Receive thread '$threadName' exception: ${t.message}", t)
      }
      finally {
        LOG.debug("Receive thread '$threadName' is finished")
      }
    }

    lifetime.onTermination {
      if (!socket.isClosed) {
        catchAndDrop {
          socket.close()
        }
      }

      QuicheCommon.gracefullyJoinThread(thread)
    }
  }

  companion object {
    private val LOG = Logger.getInstance(DatagramSocketWrapper::class.java)
  }
}