package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.intellij.openapi.diagnostic.Logger
import com.jetbrains.quiche.stubs.QuicheLibrary
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionState
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionStateProperty
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.assertQuicheError
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toBoolean
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toByte
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toNativeSize
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.IPropertyView
import com.ochafik.lang.jnaerator.runtime.NativeSize
import com.sun.jna.Pointer
import com.sun.jna.ptr.LongByReference
import com.sun.jna.ptr.PointerByReference
import java.io.IOException
import java.net.InetSocketAddress
import java.nio.ByteBuffer
import java.util.*
import java.util.concurrent.BlockingDeque
import java.util.concurrent.LinkedBlockingDeque
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.TimeUnit
import kotlin.concurrent.thread
import kotlin.random.Random

class QuicClientStreamTransport(private val lifetime: Lifetime, private val localSocket: DatagramSocketWrapper, private val peer: InetSocketAddress, hostName: String): QuicStreamTransport() {
    private val conn: PointerByReference

    private val socketReceivedDatagrams = LinkedBlockingQueue<Pair<InetSocketAddress, ByteArray>>()
    override val receivedQueue: BlockingDeque<ByteArray> = LinkedBlockingDeque()
    override val toSendQueue: BlockingDeque<ByteArray> = LinkedBlockingDeque()
    override val wakeupQueue: BlockingDeque<Boolean> = LinkedBlockingDeque()

    private val stateProperty = ConnectionStateProperty()
    override val connected: IPropertyView<ConnectionState> = stateProperty

    init {
        val quicConnectionId = Random.nextBytes(16)

        val config = QuicheConfig().also { it.setCommonSensibleOptions() }
        config.setSupportedProtocols("hq-27", "http/0.9")

        conn = Quiche.library.quiche_connect(hostName, quicConnectionId, NativeSize(quicConnectionId.size.toLong()), config.pointer)

        val quicThread = thread(isDaemon = true, name = "QUIC connect to $peer") { quicThread() }
        lifetime.onTermination {
            QuicheCommon.gracefullyJoinThread(quicThread)
        }
    }

    private fun quicThread() {
        try {
            val buffer = ByteBuffer.allocate(64 * 1024)
            val finBuf = ByteBuffer.allocate(1)

            LOG.debug("QUIC Client '$peer' thread operational")

            while (stateProperty.value !is ConnectionState.FAULTED && stateProperty.value != ConnectionState.CLOSED) {
                LOG.trace("TILL NEXT QUICHE TIMEOUT: ${Quiche.library.quiche_conn_timeout_as_millis(conn)} ms")

                val quicheTimeoutNanos = Quiche.library.quiche_conn_timeout_as_nanos(conn)
                val timeoutNanos = if (quicheTimeoutNanos < 0) 1000 * 1000000L else quicheTimeoutNanos

                wakeupQueue.poll(timeoutNanos, TimeUnit.NANOSECONDS)

                if (lifetime.isNotAlive || stateProperty.value is ConnectionState.FAULTED && stateProperty.value == ConnectionState.CLOSED) {
                    LOG.debug("QUIC Client '$peer' thread: terminating normally")
                    break
                }

                // Read what needs to be read
                while (true) {
                    val arr = socketReceivedDatagrams.poll() ?: break

                    val readWrapped = Quiche.library.quiche_conn_recv(conn, ByteBuffer.wrap(arr.second), arr.second.size.toNativeSize())
                    val read = readWrapped.toLong()
                    assertQuicheError("quiche_conn_recv", read)
                }

                Quiche.library.quiche_conn_on_timeout(conn)

                if (stateProperty.value == ConnectionState.NOT_CONNECTED && Quiche.library.quiche_conn_is_established(conn).toBoolean()) {
                    stateProperty.set(ConnectionState.CONNECTED)
                }

                if (stateProperty.value == ConnectionState.CONNECTED && Quiche.library.quiche_conn_is_closed(conn).toBoolean()) {
                    stateProperty.set(ConnectionState.CLOSED)
                    break
                }

                if (stateProperty.value == ConnectionState.CONNECTED) {
                    val readIterator = Quiche.library.quiche_conn_readable(conn)
                    try {
                        while (true) {
                            val streamId = LongByReference()
                            if (!Quiche.library.quiche_stream_iter_next(readIterator, streamId).toBoolean()) break

                            val streamRecvRc = Quiche.library.quiche_conn_stream_recv(conn, streamId.value, buffer,
                                                                                      buffer.capacity().toNativeSize(),
                                                                                      finBuf).toLong()
                            assertQuicheError("quiche_conn_stream_recv", streamRecvRc)

                            LOG.trace("$peer: received from app stream ${streamId.value} $streamRecvRc bytes")
                            receivedQueue.add(Arrays.copyOf(buffer.array(), streamRecvRc.toInt()))

                            // TODO Currently close entire conn if one of the streams closed
                            if (finBuf.get(0).toBoolean()) {
                                LOG.debug("$peer: app stream closed, closing entire connection")
                                val closeRc = Quiche.library.quiche_conn_close(conn, true.toByte(), 0L, Pointer.NULL, 0.toNativeSize())
                                if (closeRc < 0) {
                                    LOG.warn("quiche_conn_close error: ${QuicheCommon.getQuicheErrorName(closeRc)}")
                                }

                                stateProperty.set(ConnectionState.CLOSED)
                                break
                            }
                        }
                    }
                    finally {
                        Quiche.library.quiche_stream_iter_free(readIterator)
                    }
                }

                if (Quiche.library.quiche_conn_is_established(conn).toBoolean() || Quiche.library.quiche_conn_is_in_early_data(conn).toBoolean()) {
                    // Send data to quic
                    loop@ while (true) {
                        val arr = toSendQueue.poll() ?: break

                        // TODO: when to send fin=true?
                        val sendRcRaw = Quiche.library.quiche_conn_stream_send(
                          conn,
                          QuicheCommon.DEFAULT_BIDI_STREAM_ID,
                          arr,
                          arr.size.toNativeSize(),
                          false.toByte())
                        val sendRc = sendRcRaw.toLong()
                        assertQuicheError("quiche_conn_stream_send", sendRc)

                        if (sendRc > arr.size.toLong()) {
                            throw IOException(
                              "Trying to send more than requested. Reported sent size from quiche_conn_stream_send ($sendRc) != requested bytes to send (${arr.size})")
                        }
                        else if (sendRc < arr.size.toLong()) {
                            LOG.trace("Partial send?! reported sent size from quiche_conn_stream_send ($sendRc) != requested bytes to send (${arr.size})")
                            toSendQueue.addFirst(arr.copyOfRange(sendRc.toInt(), arr.size))
                            wakeupQueue.add(true)

                            if (sendRc == 0L) {
                                break@loop
                            }
                        }
                    }
                }

                // Send what needs to be sent to socket
                while (true) {
                    buffer.rewind()
                    val writtenPacked = Quiche.library.quiche_conn_send(conn, buffer, buffer.capacity().toNativeSize())
                    val written = writtenPacked.toLong()
                    if (written == QuicheLibrary.quiche_error.QUICHE_ERR_DONE.toLong()) {
                        break
                    }
                    assertQuicheError("quiche_conn_send", written)

                    LOG.trace("$peer: sending $written bytes to UDP socket")

                    try {
                        localSocket.send(buffer.array().copyOf(written.toInt()), peer)
                    }
                    catch (t: IOException) {
                        // Quiche will retry sending or will fail by timeout
                        LOG.warn("socket.send I/O exception: ${t.message}", t)
                    }
                }
            }
        } catch (_: InterruptedException) {
            LOG.debug("QUIC Client '$peer' thread received an interrupt")
        } catch (t: Throwable) {
            LOG.warn("QUIC Client '$peer' thread exception: ${t.message}", t)
            stateProperty.set(ConnectionState.FAULTED(t))
        } finally {
            Quiche.library.quiche_conn_free(conn)
            LOG.debug("QUIC Client '$peer' thread finished")
        }
    }

    init {
        localSocket.adviseOnSpecificAddressPackets(lifetime, peer) { datagramBytes ->
            socketReceivedDatagrams.add(peer to datagramBytes)
            wakeupQueue.add(true)
        }

        wakeupQueue.add(true)
    }

    companion object {
        private val LOG = Logger.getInstance(QuicClientStreamTransport::class.java)
    }
}