package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.intellij.openapi.diagnostic.Logger
import com.jetbrains.quiche.stubs.QuicheLibrary
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionState
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionStateProperty
import com.jetbrains.rd.platform.codeWithMe.protocol.StreamTransport
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.MAX_SIGNAL_DATAGRAM_SIZE
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.isNull
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toBoolean
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toByte
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toHexString
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toNativeSize
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.lifetime.*
import com.jetbrains.rd.util.reactive.IPropertyView
import com.jetbrains.rd.util.reactive.ISignal
import com.jetbrains.rd.util.reactive.Signal
import com.ochafik.lang.jnaerator.runtime.NativeSizeByReference
import com.sun.jna.ptr.LongByReference
import java.io.ByteArrayOutputStream
import java.io.DataOutputStream
import java.io.IOException
import java.net.InetSocketAddress
import java.nio.ByteBuffer
import java.nio.IntBuffer
import java.util.*
import java.util.concurrent.*
import kotlin.concurrent.thread

class QuicServer(private val lifetime: Lifetime, private val datagramSocketWrapper: DatagramSocketWrapper) {
    private val config: QuicheConfig = QuicheConfig()
    init {
        config.setCommonSensibleOptions()

        config.loadCertChainFromPemBytes(javaClass.getResourceAsStream("/quic/cert.crt")!!.use { it.readBytes() })
        config.loadPrivateKeyFromPemBytes(javaClass.getResourceAsStream("/quic/cert.key")!!.use { it.readBytes() })

        //        Quiche.library.quiche_config_verify_peer(config, false.toQuicheBool())

        // TODO?
        config.setSupportedProtocols("hq-27", "http/0.9")
    }

    private val socketReceivedDatagrams = LinkedBlockingQueue<Pair<InetSocketAddress, ByteArray>>()

    private val quicWakeupQueue = LinkedBlockingDeque<Boolean>()

    class AcceptData(
      val lifetimeOnQuicThread: Lifetime,
      val streamTransport: StreamTransport,
      val peer: InetSocketAddress
    ) {
        override fun toString(): String = "QUIC:$peer"
    }

    val acceptSignal: ISignal<AcceptData> = Signal()

    private class DestinationConnectionId(val bytes: ByteArray) {
        init {
            require(bytes.size <= QuicheLibrary.QUICHE_MAX_CONN_ID_LEN)
        }

        override fun equals(other: Any?): Boolean = (other as DestinationConnectionId?)?.bytes?.contentEquals(bytes) == true
        override fun hashCode(): Int = bytes.contentHashCode()
        override fun toString(): String = "DCID:${bytes.toHexString()}"
    }

    private class SourceConnectionId(val bytes: ByteArray) {
        init {
            require(bytes.size <= QuicheLibrary.QUICHE_MAX_CONN_ID_LEN)
        }

        override fun equals(other: Any?): Boolean = (other as SourceConnectionId?)?.bytes?.contentEquals(bytes) == true
        override fun hashCode(): Int = bytes.contentHashCode()
        override fun toString(): String = "SCID:${bytes.toHexString()}"
    }

    private class ConnectionData(
      val lifetimeDefinition: LifetimeDefinition,
      val quic: QuicheConnection,
      val streamTransport: QuicServerStreamTransport,
      val peer: InetSocketAddress,
      val id: DestinationConnectionId
    )

    private val connections = ConcurrentHashMap<DestinationConnectionId, ConnectionData>()
    private val LOCAL_CONN_ID_LEN = 16.toNativeSize()
    private val TOKEN_LEN = 31 // TODO

    private val PACKET_TYPE_INITIAL = 1.toByte()

    private fun quicThread(localAddress: InetSocketAddress) {
        try {
            val buffer = ByteBuffer.allocate(64 * 1024)

            LOG.debug("quic server thread operational")

            while (lifetime.isAlive) {
                val nextTimeoutNs = connections.values
                                      .map { connection ->
                                          Quiche.library.quiche_conn_timeout_as_nanos(connection.quic.pointer)
                                      }
                                      .filter { it > 0 }
                                      .minOrNull() ?: (5000 * 1000000L)

                LOG.trace("TILL NEXT QUICHE TIMEOUT: ${nextTimeoutNs / 1000000L} ms")
                quicWakeupQueue.poll(nextTimeoutNs, TimeUnit.NANOSECONDS)
                if (lifetime.isNotAlive) {
                    break
                }

                // Read what needs to be read
                while (true) {
                    val (packetRemoteAddress, packetData) = socketReceivedDatagrams.poll() ?: break

                    val scidRaw = ByteBuffer.allocate(QuicheLibrary.QUICHE_MAX_CONN_ID_LEN)
                    val scidLen = NativeSizeByReference(scidRaw.limit().toNativeSize())
                    val dcidRaw = ByteBuffer.allocate(QuicheLibrary.QUICHE_MAX_CONN_ID_LEN)
                    val dcidLen = NativeSizeByReference(dcidRaw.limit().toNativeSize())
                    val typeBuf = ByteBuffer.allocate(1)
                    val versionBuf = IntBuffer.allocate(1)
                    val token = ByteBuffer.allocate(TOKEN_LEN)
                    val tokenLen = NativeSizeByReference(token.limit().toNativeSize())
                    val rc = Quiche.library.quiche_header_info(
                            packetData, packetData.size.toNativeSize(),
                            LOCAL_CONN_ID_LEN,
                            versionBuf, typeBuf,
                            scidRaw, scidLen,
                            dcidRaw, dcidLen,
                            token, tokenLen
                    )
                    // TODO Do not fail, ignore. Probably log WARN
                    QuicheCommon.assertQuicheError("quiche_header_info", rc)

                    val version = versionBuf.get(0)
                    val type = typeBuf.get(0)
                    val scid = SourceConnectionId(scidRaw.array().copyOf(scidLen.value.toInt()))
                    val dcid = DestinationConnectionId(dcidRaw.array().copyOf(dcidLen.value.toInt()))

                    val existingConnection: ConnectionData? = connections[dcid]

                    val connection = if (existingConnection == null) {
                        if (type != PACKET_TYPE_INITIAL) {
                            LOG.debug("No connection established and got packet type $type (not INITIAL), skipping")
                            continue
                        }

                        if (!Quiche.library.quiche_version_is_supported(version).toBoolean()) {
                            LOG.debug("version negotiation")

                            val out = ByteBuffer.allocate(MAX_SIGNAL_DATAGRAM_SIZE)
                            val written = Quiche.library.quiche_negotiate_version(
                                    scid.bytes, scid.bytes.size.toNativeSize(),
                                    dcid.bytes, dcid.bytes.size.toNativeSize(),
                                    out, out.limit().toNativeSize()
                            ).toLong()
                            // TODO Just log WARN and continue?
                            QuicheCommon.assertQuicheError("quiche_negotiate_version", written)

                            if (written < 0) {
                                LOG.debug("failed to create vneg packet: $written")
                                continue
                            }

                            require(written > 0)
                            require(written <= MAX_SIGNAL_DATAGRAM_SIZE)

                            // TODO Handle errors
                            datagramSocketWrapper.send(out.array().copyOf(written.toInt()), packetRemoteAddress)

                            LOG.debug("sent $written bytes as vneg packet to $packetRemoteAddress")
                            continue
                        }

                        // Do stateless retry if the client didn't send a token.
                        if (tokenLen.value.toLong() == 0L) {
                            LOG.debug("No token from $packetRemoteAddress, doing stateless retry")

                            val newToken = mintToken(scid, packetRemoteAddress)

                            val out = ByteBuffer.allocate(MAX_SIGNAL_DATAGRAM_SIZE)
                            val written = Quiche.library.quiche_retry(
                                    scid.bytes, scid.bytes.size.toNativeSize(),
                                    dcid.bytes, dcid.bytes.size.toNativeSize(),
                                    dcid.bytes, dcid.bytes.size.toNativeSize(),
                                    newToken, newToken.size.toNativeSize(),
                                    version,
                                    out, out.limit().toNativeSize()
                            ).toLong()

                            // TODO Extract common part of sending packets

                            if (written < 0) {
                                LOG.debug("failed to create retry packet: $written")
                                continue
                            }

                            require(written > 0)
                            require(written <= MAX_SIGNAL_DATAGRAM_SIZE)

                            // TODO Handle errors
                            datagramSocketWrapper.send(out.array().copyOf(written.toInt()), packetRemoteAddress)

                            LOG.debug("send $written bytes as retry packet to $packetRemoteAddress")
                            continue
                        }

                        // TODO Validate token here
                        // TODO Strange logic about extracting connection id from token, let's skip it for now
                        // server.rs and server.c differs there

                        // TODO Reuse? or now? server.c generates new session id on accept, server.rs is not
                        @Suppress("UnnecessaryVariable")
                        val connScid = dcid

                        val connectionRaw = Quiche.library.quiche_accept(
                                connScid.bytes, connScid.bytes.size.toNativeSize(),
                                dcid.bytes, dcid.bytes.size.toNativeSize(),
                                config.pointer
                        )

                        if (connectionRaw.value.isNull()) {
                            LOG.debug("accept failed")
                            continue
                        }

                        val newConnection = QuicheConnection(connectionRaw)

                        val streamTransport = QuicServerStreamTransport(quicWakeupQueue)
                        streamTransport.stateProperty.set(ConnectionState.NOT_CONNECTED)
                        val data = ConnectionData(
                                lifetimeDefinition = lifetime.createNested(),
                                quic = newConnection,
                                streamTransport = streamTransport,
                                peer = packetRemoteAddress,
                                id = connScid
                        )

                        data.lifetimeDefinition.onTermination {
                            // Just in cause it was not set before
                            catch {
                                streamTransport.stateProperty.set(ConnectionState.CLOSED)
                            }

                            Quiche.library.quiche_conn_free(newConnection.pointer)
                        }

                        connections[connScid] = data

                        acceptSignal.fire(
                          AcceptData(
                            lifetimeOnQuicThread = data.lifetimeDefinition.lifetime,
                            streamTransport = data.streamTransport,
                            peer = data.peer,
                          )
                        )

                        LOG.info("Accepted new connection from $packetRemoteAddress")
                        data
                    } else existingConnection

                    val read = Quiche.library.quiche_conn_recv(connection.quic.pointer,
                            ByteBuffer.wrap(packetData), packetData.size.toNativeSize()).toLong()
                    // TODO Fail only one connection
                    QuicheCommon.assertQuicheError("quiche_conn_recv", read)

                    if (connection.streamTransport.connected.value == ConnectionState.NOT_CONNECTED && Quiche.library.quiche_conn_is_established(connection.quic.pointer).toBoolean()) {
                        connection.streamTransport.stateProperty.set(ConnectionState.CONNECTED)
                    }

                    if (connection.streamTransport.stateProperty.value == ConnectionState.CONNECTED && Quiche.library.quiche_conn_is_closed(connection.quic.pointer).toBoolean()) {
                        LOG.debug("CLOSING CONNECTION")
                        connection.streamTransport.stateProperty.set(ConnectionState.CLOSED)
                        connection.lifetimeDefinition.terminate()
                        connections.remove(connection.id)
                    }

                    if (connection.streamTransport.stateProperty.value == ConnectionState.CONNECTED) {
                        val iter = Quiche.library.quiche_conn_readable(connection.quic.pointer)
                        try {
                            while (true) {
                                val streamId = LongByReference()
                                if (!Quiche.library.quiche_stream_iter_next(iter, streamId).toBoolean()) break

                                LOG.trace("DATA FOR STREAM ID ${streamId.value}")


                                val x = ByteBuffer.allocate(1)

                                val streamRecvRaw = Quiche.library.quiche_conn_stream_recv(connection.quic.pointer, streamId.value, buffer, buffer.capacity().toNativeSize(), x)
                                val streamRecvRc = streamRecvRaw.toLong()
                                if (streamRecvRc == 0L || streamRecvRc == QuicheLibrary.quiche_error.QUICHE_ERR_DONE.toLong()) {
                                    break
                                }
                                QuicheCommon.assertQuicheError("quiche_conn_stream_recv", streamRecvRc)

                                LOG.trace("received from app stream ${streamId.value} $streamRecvRc bytes")
                                connection.streamTransport.receivedQueue.add(Arrays.copyOf(buffer.array(), streamRecvRc.toInt()))

                                // TODO Currently close entire conn if one of the streams closed
/*
                                if (x.get(0).toBoolean()) {
                                    println("app stream closed, closing entire connection")
                                    val closeRc = Quiche.library.quiche_conn_close(connection.quic.pointer, true.toByte(), 0L, Pointer.NULL, 0.toNativeSize())
                                    assertQuicheError("quiche_conn_close upon fin", closeRc)
                                }
*/
                            }
                        } finally {
                            Quiche.library.quiche_stream_iter_free(iter)
                        }
                    }
                }

                for (connection in connections.values) {
                    Quiche.library.quiche_conn_on_timeout(connection.quic.pointer)

                    // Send data to quic
                    loop@ while (true) {
                        val arr = connection.streamTransport.toSendQueue.poll() ?: break

                        val streamReadReturn = Quiche.library.quiche_conn_stream_send(
                          connection.quic.pointer,
                          QuicheCommon.DEFAULT_BIDI_STREAM_ID,
                          arr,
                          arr.size.toNativeSize(),
                          false.toByte())
                        val sendRc = streamReadReturn.toLong()
                        QuicheCommon.assertQuicheError("quiche_conn_stream_send", sendRc)

                        if (sendRc > arr.size.toLong()) {
                            throw IOException(
                              "Trying to send more than requested. Reported sent size from quiche_conn_stream_send ($sendRc) != requested bytes to send (${arr.size})")
                        }
                        else if (sendRc < arr.size.toLong()) {
                            LOG.trace("Partial send?! reported sent size from quiche_conn_stream_send ($sendRc) != requested bytes to send (${arr.size})")
                            connection.streamTransport.toSendQueue.addFirst(arr.copyOfRange(sendRc.toInt(), arr.size))
                            quicWakeupQueue.add(true)

                            if (sendRc == 0L) {
                                break@loop
                            }
                        }
                    }

                    // Send what needs to be sent to socket
                    while (true) {
                        buffer.rewind()

                        val written = Quiche.library.quiche_conn_send(
                                connection.quic.pointer, buffer, buffer.capacity().toNativeSize()).toLong()
                        if (written == QuicheLibrary.quiche_error.QUICHE_ERR_DONE.toLong()) {
//                        println("done writing")
                            break
                        }

                        QuicheCommon.assertQuicheError("quiche_conn_send", written)

                        LOG.trace("${connection.peer}: sending $written bytes to UDP socket")

                        // TODO Error handling? Should we ignore unreachable destinations?
                        try {
                          datagramSocketWrapper.send(buffer.array().copyOf(written.toInt()), connection.peer)
                        } catch (t: Throwable) {
                          LOG.info("${connection.peer}: sending $written bytes to UDP socket: ${t.message}", t)
                        }
                    }
                }
            }
        } catch (t: Throwable) {
            LOG.error("QUIC server thread '$localAddress' failed with exception: ${t.message}", t)
        } finally {
            LOG.debug("QUIC server thread '$localAddress' is finished")
        }
    }

    init {
        val localAddress = datagramSocketWrapper.localAddress!!
        thread(isDaemon = true, name = "Quic Server on $localAddress") { quicThread(localAddress) }
        quicWakeupQueue.add(true)
    }

    private class QuicServerStreamTransport(override val wakeupQueue: BlockingDeque<Boolean>): QuicStreamTransport() {
        public override val receivedQueue: BlockingDeque<ByteArray> = LinkedBlockingDeque()
        public override val toSendQueue: BlockingDeque<ByteArray> = LinkedBlockingDeque()

        val stateProperty = ConnectionStateProperty()
        override val connected: IPropertyView<ConnectionState> = stateProperty
    }

    // TODO Note that this function is only an example and doesn't do any cryptographic
    // TODO authenticate of the token. *It should not be used in production system*.
    // TODO Probably HashMap is ok for us (stateful), since number of clients is not that big
    private fun mintToken(scid: SourceConnectionId, remoteAddress: InetSocketAddress): ByteArray {
        val out = ByteArrayOutputStream()
        out.write(0x30)

        out.write(scid.bytes.size)
        out.write(scid.bytes, 0, scid.bytes.size)

        out.write(remoteAddress.address.address.size)
        out.write(remoteAddress.address.address, 0, remoteAddress.address.address.size)

        DataOutputStream(out).writeInt(remoteAddress.port)

        val arr = out.toByteArray()
        require(arr.size <= getMaxTokenLen())

        return arr
    }

    // See above
    private fun getMaxTokenLen() = 1 + (1 + QuicheLibrary.QUICHE_MAX_CONN_ID_LEN) + (1 + 16) + 4

    init {
        datagramSocketWrapper.adviseOnRemainingPackets(lifetime) { peer, bytes ->
            socketReceivedDatagrams.add(peer to bytes)
            quicWakeupQueue.add(true)
        }
    }

    companion object {
        val LOG = Logger.getInstance(QuicServer::class.java)
    }
}