// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.intellij.ide.plugins.PluginManagerCore
import com.intellij.openapi.application.PathManager
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.util.SystemInfo
import com.intellij.util.io.isDirectory
import com.intellij.util.system.CpuArch
import com.jetbrains.quiche.stubs.QuicheLibrary
import com.sun.jna.NativeLibrary
import com.sun.jna.Platform
import org.jetbrains.intellij.build.dependencies.QuicheDependencyDownloader
import java.nio.file.Files
import java.nio.file.Path
import java.nio.file.Paths

object Quiche {
  const val disableQuicheSystemProperty = "codewithme.disable.quiche"

  val isAvailable: Boolean
  val logger = Logger.getInstance(javaClass)

  init {
    isAvailable = if (System.getProperty(disableQuicheSystemProperty).toBoolean()) {
      logger.info("$disableQuicheSystemProperty property is set to false, will not attempt to check Quiche availability.")
      false
    }
    else if (SystemInfo.isMac && CpuArch.isArm64()) {
      logger.info("Quiche library binary is not yet in a working state for darwin-aarch64. QUIC support will be disabled")
      false
    }
    else try {
      checkFilesArePresent()
      @Suppress("DEPRECATION")
      QuicheLibrary.INSTANCE.quiche_version()
      true
    }
    catch (t: Throwable) {
      logger.info("Quiche library is not available. QUIC support will be disabled: ${t.message}", t)
      false
    }
  }

  // careful: will try to download if running from sources
  private fun checkFilesArePresent() {
    if (PluginManagerCore.isRunningFromSources()) {
      val quicheBinPath = try {
        QuicheDependencyDownloader.downloadQuicheBin(Paths.get(PathManager.getCommunityHomePath()))
      }
      catch (e: Throwable) {
        logger.error("Failed to download Quiche", e)
        return
      }

      val quicheDebugBinaries = quicheBinPath.resolve("debug")

      val dll1 = quicheDebugBinaries.resolve("quiche.dll")
      val dll2 = quicheDebugBinaries.resolve("win32-x86-64").resolve("quiche.dll")
      require(Files.exists(dll1) || Files.exists(dll2)) { "Internal error: $dll2 or $dll1 are both missing after extracting quiche bin" }

      addLibrarySearchPath(quicheDebugBinaries)
    }
    else {
      val plugin = PluginManagerCore.getPlugin(PluginManagerCore.getPluginByClassName(javaClass.name))
      if (plugin != null) {
        // plugin context

        val pluginPath: Path? = plugin.pluginPath
        if (pluginPath != null) {
          val quichePath = pluginPath.resolve("quiche-native")
          if (quichePath.isDirectory()) {
            addLibrarySearchPath(quichePath)
          }
          else {
            logger.error("Unable to find Quiche native libraries at $pluginPath")
          }
        }
        else {
          logger.error("Unable to find plugin path by plugin ${plugin.pluginId}")
        }
      }
      else {
        // IDE context (such as Code With Me Guest)
        val nativePath = Paths.get(PathManager.getLibPath(), "quiche-native")
        if (nativePath.isDirectory()) {
          addLibrarySearchPath(nativePath)
        }
        else {
          logger.error("Unable to find Quiche native libraries at $nativePath")
        }
      }
    }
  }

  private fun addLibrarySearchPath(path: Path) {
    // JNA resource prefix on Mac is always "darwin" since they're using universal binaries
    // We don't use them to pack only required binary code on every arch
    // since Quiche library is quite big
    val resourcePrefix = when {
      SystemInfo.isMac && CpuArch.isArm64() -> "darwin-aarch64"
      SystemInfo.isMac && CpuArch.isIntel64() -> "darwin-x86-64"
      else -> Platform.RESOURCE_PREFIX
    }

    val archSpecific = path.resolve(resourcePrefix)
    logger.info("Adding $archSpecific to JNA native library search path for '${QuicheLibrary.JNA_LIBRARY_NAME}'")
    NativeLibrary.addSearchPath(QuicheLibrary.JNA_LIBRARY_NAME, archSpecific.toString())
  }

  @Suppress("DEPRECATION")
  val library: QuicheLibrary
    get() = if (isAvailable) QuicheLibrary.INSTANCE else error("Quiche library is not available")
}