// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.google.common.net.InetAddresses
import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.quiche.stubs.QuicheLibrary
import com.jetbrains.rd.platform.util.TestUtils
import com.jetbrains.rd.util.catch
import com.ochafik.lang.jnaerator.runtime.NativeSize
import com.sun.jna.Pointer
import java.lang.reflect.Modifier
import java.net.InetSocketAddress
import java.net.URI

object QuicheCommon {
    const val MAX_SIGNAL_DATAGRAM_SIZE = 1600
    const val DEFAULT_BIDI_STREAM_ID = 4L

    fun endpointUrlToInetSocketAddress(uri: URI): InetSocketAddress {
        require(uri.scheme == "quic") {
            "Required 'quic' URI scheme: $uri"
        }

        require(uri.port > 0) {
            "Port must be specified for 'quic' URI scheme: $uri"
        }

        val inetAddress = InetAddresses.forString(uri.host)
        return InetSocketAddress(inetAddress, uri.port)
    }

    fun gracefullyJoinThread(thread: Thread) {
        val terminationFunc = {
            if (thread.isAlive) {
                thread.interrupt()
                thread.join(500)
                if (thread.isAlive) {
                    Logger.getInstance(javaClass).error("${thread.name}: thread is still operation after termination")
                }
            }
        }

        if (TestUtils.isRunningUnderUnitTest) {
            terminationFunc()
        }
        else {
            AppExecutorUtil.getAppExecutorService().execute {
                catch {
                    terminationFunc()
                }
            }
        }
    }


    fun ByteArray.toHexString() = joinToString("") { "%02x".format(it) }

    fun Pointer.isNull() = Pointer.nativeValue(this) == 0L

    fun Int.toNativeSize() = this.toLong().toNativeSize()
    fun Long.toNativeSize() = NativeSize(this)
    fun Byte.toBoolean() = this != 0.toByte()
    fun Boolean.toByte(): Byte = if (this) 1 else 0

    fun enableDebugLogging() {
        Quiche.library.quiche_enable_debug_logging(
          { line, _ ->
              LOG.debug(line.getString(0))
          }, Pointer.NULL)
    }

    private val quicheErrors = QuicheLibrary.quiche_error::class.java.declaredFields
            .filter { Modifier.isStatic(it.modifiers) }
            .map { it.getInt(null) to it.name }.toMap().also {
                // Assert it generated correctly
                require(it[-1] == "QUICHE_ERR_DONE")
                require(it[-5] == "QUICHE_ERR_INVALID_PACKET")
                require(it[-14] == "QUICHE_ERR_CONGESTION_CONTROL")
            }
    fun getQuicheErrorName(value: Int) = quicheErrors[value] ?: "QUICHE_ERR_UNKNOWN ($value)"

    fun assertQuicheError(errorText: String, rc: Int) {
        if (rc < 0) {
            throw IllegalStateException("$errorText: ${getQuicheErrorName(rc)}")
        }
    }

    fun assertQuicheError(errorText: String, rc: Long) {
        if (rc > Int.MAX_VALUE || rc < Int.MIN_VALUE) return
        assertQuicheError(errorText, rc.toInt())
    }

    private val LOG = Logger.getInstance(javaClass)
}