package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.intellij.openapi.util.io.FileUtil
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toByte
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon.toNativeSize
import com.jetbrains.quiche.stubs.QuicheLibrary
import com.sun.jna.Memory
import com.sun.jna.Pointer
import java.io.ByteArrayOutputStream
import java.io.File

class QuicheConfig {
    val pointer = Quiche.library.quiche_config_new(QuicheLibrary.QUICHE_PROTOCOL_VERSION)!!

    override fun toString(): String = "CONFIG:${pointer}"
    override fun equals(other: Any?): Boolean = (other as QuicheConnection?)?.pointer == pointer
    override fun hashCode(): Int = pointer.hashCode()

    fun setCommonSensibleOptions() {
        Quiche.library.quiche_config_set_max_idle_timeout(pointer, 15000)

        // In absence of MTU discovery, 1200 is a pretty safe value
        Quiche.library.quiche_config_set_max_send_udp_payload_size(pointer, 1200.toNativeSize())

        Quiche.library.quiche_config_set_initial_max_data(pointer, 10000000)
        Quiche.library.quiche_config_set_initial_max_stream_data_bidi_local(pointer, 1000000)
        Quiche.library.quiche_config_set_initial_max_stream_data_bidi_remote(pointer, 1000000)
        Quiche.library.quiche_config_set_initial_max_stream_data_uni(pointer, 1000000)
        Quiche.library.quiche_config_set_initial_max_streams_bidi(pointer, 100)
        Quiche.library.quiche_config_set_initial_max_streams_uni(pointer, 100)
        Quiche.library.quiche_config_set_disable_active_migration(pointer, true.toByte())
//        Quiche.library.quiche_config_set_cc_algorithm(config, QuicheLibrary.quiche_cc_algorithm.QUICHE_CC_RENO)
    }

    // in production (installers) JNA settings are different (doesn't use UTF8 for String->char*)
    private fun File.pathAsUTF8Ptr() : Pointer {
      val utf8path = this.path.toByteArray(Charsets.UTF_8)
      val pointer = Memory(utf8path.size + 1L)
      pointer.write(0, utf8path, 0, utf8path.size)
      pointer.setByte(utf8path.size.toLong(), 0)
      return pointer
    }

    fun loadCertChainFromPemFile(file: File) = QuicheCommon.assertQuicheError(
      "quiche_config_load_cert_chain_from_pem_file",
      Quiche.library.quiche_config_load_cert_chain_from_pem_file(pointer, file.pathAsUTF8Ptr()))

    fun loadPrivateKeyFromPemFile(file: File) = QuicheCommon.assertQuicheError(
      "quiche_config_load_priv_key_from_pem_file",
      Quiche.library.quiche_config_load_priv_key_from_pem_file(pointer, file.pathAsUTF8Ptr()))

    fun loadPrivateKeyFromPemBytes(bytes: ByteArray) {
        val file = FileUtil.createTempFile("quiche-private-key", ".pem")
        try {
            file.writeBytes(bytes)
            loadPrivateKeyFromPemFile(file)
        } finally {
          file.delete()
        }
    }

    fun loadCertChainFromPemBytes(bytes: ByteArray) {
        val file = FileUtil.createTempFile("quiche-cert-chain", ".crt")
        try {
            file.writeBytes(bytes)
            loadCertChainFromPemFile(file)
        } finally {
          file.delete()
        }
    }

    fun setSupportedProtocols(vararg protocol: String) {
        val memory = ByteArrayOutputStream()
        for (proto in protocol) {
            require(proto.length in 1..127)
            memory.write(proto.length)

            val stringByteArray = proto.toByteArray()
            memory.write(stringByteArray, 0, stringByteArray.size)
        }

        val arr = memory.toByteArray()
        Quiche.library.quiche_config_set_application_protos(pointer, arr, arr.size.toNativeSize())
    }

}