// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.network.quic

import com.intellij.openapi.diagnostic.Logger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import de.javawi.jstun.attribute.ChangeRequest
import de.javawi.jstun.attribute.MappedAddress
import de.javawi.jstun.attribute.MessageAttributeInterface
import de.javawi.jstun.attribute.MessageAttributeParsingException
import de.javawi.jstun.header.MessageHeader
import de.javawi.jstun.header.MessageHeaderInterface
import java.net.InetAddress
import java.net.InetSocketAddress
import java.util.concurrent.CompletableFuture
import kotlin.concurrent.thread

object StunConnector {
  private val logger = Logger.getInstance(javaClass)

  fun maintainStunServerConnection(lifetime: Lifetime, datagramSocketWrapper: DatagramSocketWrapper, stunServer: StunTurnUrl): CompletableFuture<InetSocketAddress> {
    val result = CompletableFuture<InetSocketAddress>()

    // TODO Separate thread?
    val threadName = "Ping STUN server from ${datagramSocketWrapper.localAddress}"
    val thread = thread(name = threadName, isDaemon = true) {
      try {
        logger.info("Pinging STUN server: $stunServer")

        if (stunServer.protocol != StunTurnUrl.Protocol.STUN || stunServer.transport == StunTurnUrl.Transport.TCP) {
          error("Only STUN over UDP is supported in StunConnector")
        }

        val stunServerAddress = InetSocketAddress(InetAddress.getByName(stunServer.host), stunServer.portNotNull)

        var loggedMappedAddress = false

        if (lifetime.isNotAlive) {
          logger.debug("$threadName: lifetime was already terminated. exiting STUN ping thread")
          return@thread
        }

        datagramSocketWrapper.adviseOnSpecificAddressPackets(lifetime, stunServerAddress) { packetBytes ->
          val receiveMH = MessageHeader(MessageHeaderInterface.MessageHeaderType.BindingResponse)
          try {
            receiveMH.parseAttributes(packetBytes)
          }
          catch (e: Throwable) {
            logger.warn("Error parsing response from stun server: ${stunServer.host}: ${e.message}", e)
            return@adviseOnSpecificAddressPackets
          }
          val ma: MappedAddress = receiveMH
            .getMessageAttribute(MessageAttributeInterface.MessageAttributeType.MappedAddress) as MappedAddress

          val inetSocketAddress = InetSocketAddress(ma.address.inetAddress, ma.port)
          if(loggedMappedAddress)
            logger.trace("Received mapped address from stun server: $inetSocketAddress")
          else {
            logger.info("Received mapped address from stun server: $inetSocketAddress")
            loggedMappedAddress = true
          }

          result.complete(inetSocketAddress)
        }

        val sendMH = MessageHeader(MessageHeaderInterface.MessageHeaderType.BindingRequest)
        // add an empty ChangeRequest attribute. Not required by the
        // standard, but some servers require it
        sendMH.addMessageAttribute(ChangeRequest())

        val bindingRequestBytes = sendMH.bytes

        while (lifetime.isAlive && !Thread.interrupted()) {
          logger.trace("Sending STUN packet from ${datagramSocketWrapper.localAddress} to $stunServerAddress")

          try {
            datagramSocketWrapper.send(bindingRequestBytes, stunServerAddress)
          } catch (sendException: Throwable) {
            logger.debug("Unable to send STUN packet from ${datagramSocketWrapper.localAddress} to $stunServerAddress: ${sendException.message}", sendException)
          }

          Thread.sleep(5000)
        }
      }
      catch(_: InterruptedException) {
      }
      catch(t: Throwable) {
        logger.error("$threadName: ${t.message}", t)
      }
    }

    lifetime.onTermination {
      QuicheCommon.gracefullyJoinThread(thread)
    }

    return result
  }
}