// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.network.quic

data class StunTurnUrl(val protocol: Protocol, val host: String, val port: Int, val transport: Transport?) {

  override fun toString(): String {
    val sb = StringBuilder()
    sb.append(when (protocol) {
      Protocol.STUN -> "stun"
      Protocol.STUNS -> "stuns"
      Protocol.TURN -> "turn"
      Protocol.TURNS -> "turns"
    })
    sb.append(":")
    sb.append(host)
    if (port > 0) {
      sb.append(':')
      sb.append(port)
    }
    if (transport != null) {
      sb.append("?transport=")
      sb.append(when (transport) {
        Transport.TCP -> "tcp"
        Transport.UDP -> "udp"
      })
    }

    return sb.toString()
  }

  enum class Transport {
    TCP,
    UDP
  }

  enum class Protocol {
    STUN,
    STUNS,
    TURN,
    TURNS,
  }

  val portNotNull: Int
    get() {
      if (port > 0) {
        return port
      }

      return when (protocol) {
        Protocol.STUN -> DEFAULT_PORT
        Protocol.TURN -> DEFAULT_PORT
        Protocol.STUNS -> DEFAULT_TLS_PORT
        Protocol.TURNS -> DEFAULT_TLS_PORT
      }
    }

  companion object {
    const val DEFAULT_PORT = 3478
    const val DEFAULT_TLS_PORT = 5349

    fun tryParse(string: String): StunTurnUrl? {
      val matchResult = STUN_TURN_REGEX.matchEntire(string) ?: return null

      val protocol = when (matchResult.groups["scheme"]?.value) {
        "stun" -> Protocol.STUN
        "stuns" -> Protocol.STUNS
        "turn" -> Protocol.TURN
        "turns" -> Protocol.TURNS
        else -> return null
      }

      val host = matchResult.groups["host"]?.value ?: return null
      val port = matchResult.groups["port"]?.value?.toIntOrNull() ?: 0
      val transport = when (matchResult.groups["transport"]?.value) {
        null -> null
        "udp" -> Transport.UDP
        "tcp" -> Transport.TCP
        else -> return null
      }

      return StunTurnUrl(protocol = protocol, host = host, port = port, transport = transport)
    }

    // https://tools.ietf.org/html/rfc7065 Traversal Using Relays around NAT (TURN) Uniform Resource Identifiers
    // https://tools.ietf.org/html/rfc7064 URI Scheme for the Session Traversal Utilities for NAT (STUN) Protocol
    private val STUN_TURN_REGEX = Regex(
      "(?<scheme>stun|stuns|turn|turns)\\:(?<host>[^?:]+)(\\:(?<port>[0-9]+?))?" +
      "(\\?transport=(?<transport>udp|tcp))?"
    )
  }
}