// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.protocol

import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.readAll
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import tlschannel.NeedsWriteException
import java.nio.ByteBuffer
import java.nio.channels.ByteChannel
import java.nio.channels.ClosedByInterruptException
import java.nio.channels.ClosedChannelException
import java.nio.channels.GatheringByteChannel
import java.util.concurrent.BlockingQueue
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicReference
import kotlin.concurrent.thread

class BufferedByteChannel(private val lifetime: Lifetime,
                          private val underlying: ByteChannel,
                          private val mergingTimeSpanMs: Int,
                          private val id: String) : ByteChannel {
  // *Unbound* blocking queue
  private val buffersQueue: BlockingQueue<ByteBuffer> = LinkedBlockingQueue()

  private val lastWriteException = AtomicReference<Throwable>()

  init {
    val flushThread = thread(isDaemon = true, name = "Flush:$id") {
      try {
        while (lastWriteException.get() == null) {
          logger.trace("BufferedByteChannel '$id' FLUSH: poll")
          val firstElement = buffersQueue.poll(1000, TimeUnit.MILLISECONDS)
          logger.trace("BufferedByteChannel '$id' FLUSH: awake, buffersQueue.size = ${buffersQueue.size}")

          Thread.sleep(mergingTimeSpanMs.toLong())

          completeBuffersQueue(firstElement)
          while (!buffersQueue.isEmpty() && lastWriteException.get() == null) {
            completeBuffersQueue()
          }
        }
      }
      catch (_: InterruptedException) {
        // Do nothing
      }
      catch (t: Throwable) {
        logger.error("Uncaught exception in BufferedByteChannel '$id' Flush Thread: ${t.message}", t)
      }

      logger.trace("BufferedByteChannel '$id' FLUSH: thread exit")
    }

    lifetime.onTermination {
      // don't block UI thread with waiting for termination
      AppExecutorUtil.getAppExecutorService().execute {
        catch {
          flushThread.join(500) // give some time for socket to finish
          flushThread.interrupt()
          flushThread.join(500)
          if (flushThread.isAlive) {
            logger.warn("BufferedByteChannel '$id' Flush Thread is still alive after lifetime termination")
          }
        }
      }
    }
  }

  private fun completeBuffersQueue(firstElement: ByteBuffer? = null) {
    val buffers = mutableListOf<ByteBuffer>()
    firstElement?.let { buffers.add(it) }
    buffersQueue.drainTo(buffers)
    if (buffers.isEmpty()) return

    val handleCloseException = {
      if (buffers.any { it.hasRemaining() }) {
        logger.warn("$id: some data were not written to underlying channel since channel was closed")
      }
      if (isOpen) close()
    }

    try {
      val buffersArray = buffers.toTypedArray()

      logger.trace("BufferedByteChannel '$id': flushing ${buffers.sumBy { it.remaining() }} bytes")
      if (underlying is GatheringByteChannel) {
        while (buffers.any { it.hasRemaining() }) {
          underlying.write(buffersArray)
        }
      }
      else {
        throw UnsupportedOperationException("Only GatheringByteChannels are supported")
      }
    }
    catch (_: ClosedByInterruptException) {
      handleCloseException()
    }
    catch (_: InterruptedException) {
      handleCloseException()
    }
    catch (_: ClosedChannelException) {
      handleCloseException()
    }
    catch (_: NeedsWriteException) {
      handleCloseException()
    }
    catch (t: Throwable) {
      logger.warn("$id: error writing to underlying channel. Subsequent writes will fail", t)
      lastWriteException.set(t)
      handleCloseException()
    }
  }

  override fun isOpen(): Boolean = underlying.isOpen
  override fun close() = underlying.close()
  override fun read(dst: ByteBuffer): Int = underlying.read(dst)

  override fun write(src: ByteBuffer): Int {
    if (!isOpen || lifetime.isNotAlive) throw ClosedChannelException()

    val throwable = lastWriteException.get()
    if (throwable != null) throw throwable

    val remaining = src.remaining()
    if (remaining <= 0) return 0

    buffersQueue.put(ByteBuffer.wrap(src.readAll()))
    logger.trace("BufferedByteChannel '$id': queued $remaining bytes")
    return remaining
  }

  companion object {
    private val logger = getLogger<BufferedByteChannel>()
  }
}