// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.protocol

import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.reactive.IProperty
import com.jetbrains.rd.util.reactive.ISource
import com.jetbrains.rd.util.reactive.Signal
import java.util.concurrent.locks.ReentrantLock
import kotlin.concurrent.withLock

class ConnectionStateProperty: IProperty<ConnectionState> {
  private val lock = ReentrantLock()

  override fun set(newValue: ConnectionState) {
    value = newValue
  }

  override var value: ConnectionState = ConnectionState.NOT_CONNECTED
    get() = lock.withLock { field }
    set(newValue) {
      lock.withLock {
        if (field == newValue) return

        if (newValue == ConnectionState.NOT_CONNECTED) {
          error("Could not return ConnectionState back from $field to NOT_CONNECTED")
        }

        if (field is ConnectionState.FAULTED) {
          // FAULTED state is a final state. Some clients may change it to CLOSED, skip them
          when (newValue) {
            is ConnectionState.NOT_CONNECTED -> error("NOT_CONNECTED unexpected")
            is ConnectionState.FAULTED -> Unit
            is ConnectionState.CLOSED -> Unit
            is ConnectionState.DECLINED -> error("Trying to set ConnectionState from $field to $newValue")
            is ConnectionState.CONNECTED -> error("Trying to set ConnectionState from $field to $newValue")
          }.let {  }
          return@withLock
        }

        if (field is ConnectionState.DECLINED) {
          // DECLINED state is a final state. Some clients may change it to CLOSED, skip them
          when (newValue) {
            is ConnectionState.NOT_CONNECTED -> error("NOT_CONNECTED unexpected")
            is ConnectionState.FAULTED -> error("Trying to set ConnectionState from $field to $newValue")
            is ConnectionState.CLOSED -> Unit
            is ConnectionState.DECLINED -> Unit
            is ConnectionState.CONNECTED -> error("Trying to set ConnectionState from $field to $newValue")
          }.let {  }
          return@withLock
        }

        if (field is ConnectionState.CLOSED) {
          // CLOSED state is a final state. No changes are possible
          when (newValue) {
            is ConnectionState.NOT_CONNECTED -> error("NOT_CONNECTED unexpected")
            is ConnectionState.FAULTED, is ConnectionState.CONNECTED, is ConnectionState.DECLINED -> error("Trying to set ConnectionState from $field to $newValue")
            is ConnectionState.CLOSED -> Unit
          }.let {  }

          return@withLock
        }

        field = newValue

        _change.fire(newValue)
      }
    }

  override fun advise(lifetime: Lifetime, handler: (ConnectionState) -> Unit) {
    if (!lifetime.isAlive)
      return

    lock.withLock {
      handler(value)
      change.advise(lifetime, handler)
    }
  }

  private val _change = Signal<ConnectionState>()
  override val change: ISource<ConnectionState> get() = _change
}