package com.jetbrains.rd.platform.codeWithMe.protocol

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicheCommon
import com.jetbrains.rd.platform.util.readAll
import com.jetbrains.rd.platform.util.writeAll
import com.jetbrains.rd.util.catchAndDrop
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.Signal
import tlschannel.NeedsReadException
import java.io.IOException
import java.nio.ByteBuffer
import java.nio.channels.ByteChannel
import java.nio.channels.ClosedByInterruptException
import java.nio.channels.ClosedChannelException
import java.util.concurrent.atomic.AtomicReference

interface FramedByteChannel {
  fun send(bytes: ByteBuffer)

  /**
   * handler must be instantly read from position 0 (with rewind).
   * handler may be reused in subsequent reads, so no queuing on another thread
   */
  fun adviseOnFrameReceived(lifetime: Lifetime, handler: (ByteBuffer) -> Unit)

  fun adviseOnErrorReceived(lifetime: Lifetime, handler: (Throwable) -> Unit)

  /**
   * Start reading from underlying channel on a separate thread.
   * Upon receiving a frame frameReceived will be fired.
   */
  fun startReceiving()
}

class FramedByteChannelImpl(lifetime: Lifetime, underlyingByteChannel: ByteChannel, private val id: String) : FramedByteChannel {
  private val chan = underlyingByteChannel
  private val frameReceived = Signal<ByteBuffer>()
  private val errorReceived = Signal<Throwable>()

  init {
    lifetime.onTermination {
      stopReceiving()
    }
  }

  override fun send(bytes: ByteBuffer) {
    val len = bytes.remaining()
    logger.trace { "FramedByteChannel '$id': sending $len bytes" }

    // TODO Is it any good compared to usual allocations?
    val buf = if (len + 4 <= REUSED_BUFFER_SIZE) {
      threadLocalWriteBuffer.get().slice().also { it.limit(len + 4) }
    }
    else {
      ByteBuffer.allocate(len + 4)
    }

    buf.putInt(len)
    buf.put(bytes)
    buf.rewind()
    if (!chan.writeAll(buf)) error("$id: channel EOF")
  }

  override fun adviseOnFrameReceived(lifetime: Lifetime, handler: (ByteBuffer) -> Unit) = frameReceived.advise(lifetime, handler)
  override fun adviseOnErrorReceived(lifetime: Lifetime, handler: (Throwable) -> Unit) = errorReceived.advise(lifetime, handler)

  private val readThread = AtomicReference<Thread>()

  private fun stopReceiving() {
    val thread = readThread.getAndSet(null)
    if (thread != null) {
      QuicheCommon.gracefullyJoinThread(thread)
    }
  }

  override fun startReceiving() {
    val oldReadThread = readThread.getAndSet(Thread {
      try {
        logger.info("FramedByteChannel '$id' read thread: waiting for channel to open")
        while (!chan.isOpen) {
          Thread.sleep(10)
        }
        logger.info("FramedByteChannel '$id' read thread: chan connected")

        while (chan.isOpen) {
          val reusedBuffer = threadLocalReadBuffer.get()

          reusedBuffer.rewind()
          if (!chan.readAll(reusedBuffer.slice().also { it.limit(4) })) break

          reusedBuffer.rewind()
          val len = reusedBuffer.int
          if (len < 0) error("len < 0")
          if (len >= 300_000_000) error("Possibly broken frames, len is too big, len = $len")

          val buf = if (len <= REUSED_BUFFER_SIZE) {
            reusedBuffer.rewind()
            reusedBuffer
          }
          else ByteBuffer.allocate(len)

          if (!chan.readAll(buf.slice().also { it.limit(len) })) error("Got EOF while reading a frame of len=$len")
          buf.rewind()

          val slice = buf.slice()
          slice.limit(len)

          logger.trace { "FramedByteChannel '$id': received a frame len=$len" }
          frameReceived.fire(slice)
        }

        Thread.sleep(500) // wait a bit to let written data get flushed
      }
      catch (_: ClosedByInterruptException) {
        // do nothing
      }
      catch (_: InterruptedException) {
        // do nothing
      }
      catch (_: ClosedChannelException) {
        // do nothing
      }
      catch (_: NeedsReadException) {
        // do nothing
      }
      catch (t: Throwable) {
        // do nothing if the channel was just closed

        if (t.message?.contains("Output closed") == false) {
          if (t !is IOException) {
            logger.error("Error while reading protocol frames ($id): ${t.message}", t)
          }

          errorReceived.fire(t)
        }
      }

      if (chan.isOpen) {
        catchAndDrop {
          chan.close() // close the channel when read thread ends
        }
      }
    }.also {
      it.name = "ProtocolFromChan:$id"
      it.isDaemon = true
      it.start()
    })

    if (oldReadThread != null) {
      error("Old receiving thread is already present")
    }
  }

  companion object {
    private const val REUSED_BUFFER_SIZE = 65536
    private val threadLocalReadBuffer = ThreadLocal.withInitial { ByteBuffer.allocate(REUSED_BUFFER_SIZE) }
    private val threadLocalWriteBuffer = ThreadLocal.withInitial { ByteBuffer.allocate(REUSED_BUFFER_SIZE) }
    private val logger = logger<FramedByteChannelImpl>()
  }
}