// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.protocol

import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import java.nio.ByteBuffer
import java.nio.channels.ByteChannel
import java.nio.channels.GatheringByteChannel
import java.util.concurrent.atomic.AtomicReference

class RetryingConnectStreamTransport(lifetime: Lifetime, private val moniker: String, val streamTransportFactory: (Lifetime) -> StreamTransport): StreamTransport {

  private val transportSequentialLifetimes = SequentialLifetimes(lifetime)

  private val underlyingTransport = AtomicReference<StreamTransport>()

  private fun getUnderlying(): StreamTransport {
    if (connected.value != ConnectionState.CONNECTED) {
      error("Unable to get underlying transport when it's not ready yet (current state: ${connected.value})")
    }

    val current: StreamTransport? = underlyingTransport.get()
    require(current != null) { "No underlying transport" }

    return current
  }

  fun tryGetUnderlying(): StreamTransport? = underlyingTransport.get()

  override val byteChannel: ByteChannel = object : ByteChannel, GatheringByteChannel {

    override fun close() = getUnderlying().byteChannel.close()

    override fun isOpen(): Boolean = connected.value == ConnectionState.CONNECTED

    override fun write(srcs: Array<out ByteBuffer>, offset: Int, length: Int): Long =
      (getUnderlying().byteChannel as GatheringByteChannel).write(srcs, offset, length)

    override fun write(srcs: Array<out ByteBuffer>): Long =
      (getUnderlying().byteChannel as GatheringByteChannel).write(srcs)

    override fun write(src: ByteBuffer): Int  = getUnderlying().byteChannel.write(src)

    override fun read(dst: ByteBuffer): Int = getUnderlying().byteChannel.read(dst)
  }

  override val connected: ConnectionStateProperty = ConnectionStateProperty()

  private fun retry() {
    require(connected.value == ConnectionState.NOT_CONNECTED)

    val transportLifetime = transportSequentialLifetimes.next()

    @Suppress("ThrowableNotThrown")
    val transport = try {
      streamTransportFactory(transportLifetime)
    } catch (t: Throwable) {
      catch {
        transportSequentialLifetimes.terminateCurrent()
      }

      connected.set(ConnectionState.FAULTED(IllegalStateException("Unable to create stream transport", t)))
      return
    }

    underlyingTransport.set(transport)

    transport.connected.advise(transportLifetime) {
      when (it) {
        ConnectionState.NOT_CONNECTED -> Unit

        // Final states
        ConnectionState.CONNECTED, ConnectionState.CLOSED, is ConnectionState.DECLINED ->
          connected.set(it)

        is ConnectionState.FAULTED -> {
          if (connected.value == ConnectionState.NOT_CONNECTED) {
            LOG.info("Connect to $moniker failed, retrying in $RETRY_INTERVAL_MS ms: ${it.throwable.message}", it.throwable)
            AppExecutorUtil.getAppExecutorService().execute {
              catch("Retrying connection to $moniker") {
                // Not too frequent
                Thread.sleep(RETRY_INTERVAL_MS)
                retry()
              }
            }
          } else {
            // Retry only establishing connection
            connected.set(it)
          }
        }
      }.let {  }
    }
  }

  init {
    retry()
  }

  companion object {
    private val LOG = Logger.getInstance(RetryingConnectStreamTransport::class.java)
    private const val RETRY_INTERVAL_MS = 500L
  }
}