@file:Suppress("MoveVariableDeclarationIntoWhen")

package com.jetbrains.rd.platform.codeWithMe.protocol

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.trace
import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.rd.framework.*
import com.jetbrains.rd.framework.base.IRdWireable
import com.jetbrains.rd.framework.impl.ProtocolContexts
import com.jetbrains.rd.framework.util.nextValue
import com.jetbrains.rd.platform.util.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.*
import kotlinx.coroutines.*
import kotlinx.coroutines.future.await
import tlschannel.ClientTlsChannel
import tlschannel.ServerTlsChannel
import tlschannel.TlsChannel
import java.nio.ByteBuffer
import java.nio.channels.ClosedChannelException
import java.security.cert.X509Certificate
import java.util.concurrent.CompletableFuture
import java.util.concurrent.LinkedBlockingQueue
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicInteger
import javax.net.ssl.KeyManagerFactory
import javax.net.ssl.SSLContext
import javax.net.ssl.X509TrustManager

class SecureWireOverStreamTransport(
  private val lifetime: Lifetime,
  val streamTransport: StreamTransport,
  val scheduler: IScheduler,
  val id: String,
  private val server: Boolean,
  private val certificateData: () -> SelfSignedCertificate.SelfSignedCertificateData,
  private val checkHandshakeTrusted: (chain: List<X509Certificate>) -> Unit = { throw UnsupportedOperationException() },
  private val checkBeforeProtocolTrusted: SecureWireOverStreamTransport.(chain: List<X509Certificate>) -> WireConnectStatus = { throw UnsupportedOperationException() },
  private val pinHandler: (SecureWireOverStreamTransport.(chain : List<X509Certificate>, pin: ByteArray) -> Unit) = { _, _ -> },
  private val passwordHandler: IPasswordAuthHandler? = null
) : IWire {

  var clientId: ClientId? = null
  private lateinit var contextsInternal: ProtocolContexts
  private val messageBroker = MessageBroker(scheduler)
  val wireStats = WireStatsCounter(lifetime)
  private val connectionStateProperty = ConnectionStateProperty()
  val connectionState: IPropertyView<ConnectionState> = connectionStateProperty
  override val heartbeatAlive: Property<Boolean> = Property(true)
  override var heartbeatIntervalMs: Long = Long.MAX_VALUE

  init {
    connectionState.advise(lifetime) { state ->
      logger.debug("SecureWireOverStreamTransport '$id': state = $state")
    }
  }

  override val connected = connectionState.map { it is ConnectionState.CONNECTED }

  private fun setFaultedState(t: Throwable) {
    setStateIfNotTerminal(ConnectionState.FAULTED(t))
  }

  override val contexts: ProtocolContexts
    get() = contextsInternal

  override fun advise(lifetime: Lifetime, entity: IRdWireable) {
    val wrapper = object: IRdWireable by entity {
      override fun onWireReceived(buffer: AbstractBuffer) {
        if (assetNullClientIdOnReceive && clientId == null)
          logger.error("No ClientId set in onWireReceived")
        ClientId.withClientId(clientId) {
          entity.onWireReceived(buffer)
        }
      }
    }
    messageBroker.adviseOn(lifetime, wrapper)
    wireStats.onAdvise(lifetime, wrapper)
  }

  override fun setupContexts(newContexts: ProtocolContexts) {
    require(!this::contextsInternal.isInitialized) { "Can't replace ProtocolContexts in IWire" }
    contextsInternal = newContexts
  }

  private val messagePushed = Signal<Unit>()
  private val accumulatedMessages = LinkedBlockingQueue<ByteArray>()
  private var doDeliverMessages: Boolean = false
  private val deliverMessagesFuture = CompletableFuture<Unit>()

  private val trustedChainFuture = CompletableFuture<List<X509Certificate>>()

  private val tlsChannel: TlsChannel = run {
    val sslContext = SSLContext.getInstance("TLSv1.3")!!.also { ctx ->
      val trustManager = object : X509TrustManager {
        private fun checkTrusted(chain: List<X509Certificate>) {
          try {
            checkHandshakeTrusted(chain)
            trustedChainFuture.complete(chain)
          } catch (t: Throwable) {
            trustedChainFuture.completeExceptionally(t)
            throw t
          }
        }

        override fun checkClientTrusted(chain: Array<out X509Certificate>, authType: String) {
          if (server) {
            checkTrusted(chain.toList())
          }
          else {
            throw UnsupportedOperationException()
          }
        }

        override fun checkServerTrusted(chain: Array<out X509Certificate>, authType: String) {
          if (server) {
            throw UnsupportedOperationException()
          }
          else {
            checkTrusted(chain.toList())
          }
        }

        override fun getAcceptedIssuers(): Array<X509Certificate> = emptyArray()
      }

      val certData = certificateData()

      val keyManagerFactory = KeyManagerFactory.getInstance(KeyManagerFactory.getDefaultAlgorithm())
        .also { it.init(certData.keyStore, certData.keyPassword) }

      ctx.init(keyManagerFactory.keyManagers, arrayOf(trustManager), null)
    }

    val tlsChan = when (server) {
      true -> ServerTlsChannel
        .newBuilder(streamTransport.byteChannel, sslContext)
        .withEngineFactory {
          val sslEngine = it.createSSLEngine()
          sslEngine.useClientMode = false
          sslEngine.needClientAuth = true
          sslEngine
        }.build()
      false -> ClientTlsChannel.newBuilder(streamTransport.byteChannel, sslContext).build()
    }

    tlsChan
  }

  private val bufChan = BufferedByteChannel(lifetime, tlsChannel, 3, id)

  private val framedByteChannel = FramedByteChannelImpl(lifetime, bufChan, id)

  private val pingsWithoutReceivedData = AtomicInteger()

  private val checkForTimeouts: Boolean

  var assetNullClientIdOnReceive = false

  private fun runConnectionFlow(tlsChannel: TlsChannel): Job {
    val job = lifetime.launchLongBackground {
      try {
        logger.info("$id: running connection flow; wait for transport")
        val connectedState = streamTransport.connected.nextValue(lifetime) { it != ConnectionState.NOT_CONNECTED }

        when (connectedState) {
          ConnectionState.CLOSED -> {
            logger.info("$id: transport closed, aborting connection flow")
            return@launchLongBackground
          }
          is ConnectionState.DECLINED -> throw PeerVisibleErrorCode(WireConnectStatus.DECLINED, "Non-user declined by other side")
          is ConnectionState.FAULTED -> throw connectedState.throwable
          ConnectionState.CONNECTED -> { } // fall through
          ConnectionState.NOT_CONNECTED -> error("Wire in not connected state despite explicit wait for connected state")
        }.let {}

        logger.info("$id: running TLS handshake")
        tlsChannel.handshake()

        if (!trustedChainFuture.isDone) {
          error("$id: trustedChainFuture must be completed after handshake")
        }

        val certList = trustedChainFuture.await()
        logger.info("$id: got cert chain")

        if (server) {
          val pinBytes = ByteArray(4)
          pinHandler(certList, pinBytes)
          sendServiceMessage(WireConnectStatus.JOIN_PIN, pinBytes)

          val judgement = checkBeforeProtocolTrusted(certList)
          logger.info("$id: protocol trusted judgement: $judgement")
          when (judgement) {
            WireConnectStatus.ACCEPTED -> {
            } // fall through
            WireConnectStatus.USER_DECLINED, WireConnectStatus.DECLINED, WireConnectStatus.INTERNAL_ERROR, WireConnectStatus.GUEST_LIMIT_REACHED -> {
              throw PeerVisibleErrorCode(judgement)
            }
            WireConnectStatus.JOIN_PIN -> error("Client sent illegal pin message")
            WireConnectStatus.PASSWORD_REQUIRED -> {
              require(server) { "Only server end can request passwords" }
              require(passwordHandler != null) { "Password handler must be set in order to request passwords" }

              logger.trace("$id: requesting password from client")
              val passwordRequest = passwordHandler.getPasswordRequestBytes()
              sendServiceMessage(WireConnectStatus.PASSWORD_REQUIRED, passwordRequest)

              val (remoteReply, argument) = decodeServiceMessage(readServiceMessage())
              logger.trace("$id: client replied with $remoteReply and ${argument.size} bytes")

              require(remoteReply == WireConnectStatus.PASSWORD_REQUIRED) { "Remote did not reply with password response when one was expected" }

              if (!passwordHandler.checkPasswordFromClient(argument)) {
                logger.trace("$id: client password was wrong")
                throw PeerVisibleErrorCode(WireConnectStatus.DECLINED, "Wrong password")
              }

              logger.trace("$id: client password was correct")

              Unit
            }
          }.let { }

          // server sends accepted immediately; client waits for ACCEPTED or PASSWORD_REQUIRED instead
          logger.info("$id: connection flow achieved acceptance")
          sendServiceMessage(WireConnectStatus.ACCEPTED)
        } else {
          // receive pin and pass it outwards
          val (remoteReply, argument) = decodeServiceMessage(readServiceMessage())
          require(remoteReply == WireConnectStatus.JOIN_PIN) { "Remote send unexpected message when PIN was expected" }
          require(argument.size == 4) { "Remote's PIN was of wrong length" }
          require(argument.all { it in 0..9 }) { "Remote's PIN was out of range" }

          pinHandler(certList, argument)
        }

        var (remoteReply, argument) = decodeServiceMessage(readServiceMessage())
        logger.info("$id: remote responded with $remoteReply and ${argument.size} bytes")
        if (remoteReply == WireConnectStatus.PASSWORD_REQUIRED) {
          require(!server) { "Only client end can receive password requests" }
          require(passwordHandler != null) { "Password handler was not set, but remote requested a password" }

          logger.trace("$id: asking user for password input")

          val password = passwordHandler.askForPassword(argument).await()

          logger.trace("$id: sending password to server")

          sendServiceMessage(WireConnectStatus.PASSWORD_REQUIRED, password)

          val (remoteReply2, argument2) = decodeServiceMessage(readServiceMessage())
          remoteReply = remoteReply2
          argument = argument2

          logger.trace("$id: remote responded again with $remoteReply and ${argument.size} bytes")
        }

        require(argument.isEmpty()) { "Remote must not respond with arguments" }

        if (remoteReply != WireConnectStatus.ACCEPTED)
          throw PeerVisibleErrorCode(remoteReply)

        if (!server) { // server has already sent accepted
          logger.trace("$id: connection flow achieved acceptance")
          sendServiceMessage(WireConnectStatus.ACCEPTED)
        }

        setStateIfNotTerminal(ConnectionState.CONNECTED)

        logger.info("$id: awaiting for message delivery permission")
        deliverMessagesFuture.await()
        logger.info("$id: delivering accumulated messages")
        synchronized(accumulatedMessages) {
          assert(!doDeliverMessages) { "Already delivering messages" }
          accumulatedMessages.forEach {
            logger.runCatching {
              dispatchByteBuffer(it)
            }
          }
          accumulatedMessages.clear()
          doDeliverMessages = true
        }
        logger.debug("$id: connection flow done")
      } catch(ex: PeerVisibleErrorCode) {
        require(ex.code != WireConnectStatus.ACCEPTED) { "Accepted status can't be indicated with an exception" }

        logger.info("$id: connection flow failed with code ${ex.code} and message ${ex.message}")

        when(ex.code) {
          WireConnectStatus.USER_DECLINED -> setStateIfNotTerminal(ConnectionState.DECLINED(ConnectionState.DeclineReason.UserDeclined))
          WireConnectStatus.GUEST_LIMIT_REACHED -> setStateIfNotTerminal(ConnectionState.DECLINED(ConnectionState.DeclineReason.GuestLimitReached))
          else -> setFaultedState(ex)
        }

        try {
          sendServiceMessage(if (ex.code == WireConnectStatus.PASSWORD_REQUIRED) WireConnectStatus.DECLINED else ex.code)
        } catch(_: ClosedChannelException) {
          // do nothing
        }
      } catch(ex : CancellationException) {
        logger.debug("$id: connection flow was cancelled")
        try {
          sendServiceMessage(WireConnectStatus.DECLINED)
        } catch(_: ClosedChannelException) {
          // do nothing
        }
        throw ex
      } catch (t: ClosedChannelException) {
        logger.debug("$id: ClosedChannelException", t)
        setStateIfNotTerminal(ConnectionState.CLOSED)
      } catch (t: Throwable) { // todo: more exception cases?
        logger.info("$id: exception in connection flow", t)
        setFaultedState(t)
      }
    }
    return job
  }

  private fun setStateIfNotTerminal(state: ConnectionState) {
    if (connectionStateProperty.value == ConnectionState.NOT_CONNECTED || connectionStateProperty.value == ConnectionState.CONNECTED)
      connectionStateProperty.set(state)
  }

  private fun sendServiceMessage(code: WireConnectStatus, argument: ByteArray? = null) {
    val byteArray = code.name.toByteArray() + if(argument != null) byteArrayOf(0) + argument else byteArrayOf()
    framedByteChannel.send(ByteBuffer.wrap(byteArray))
  }

  private fun decodeServiceMessage(byteArray: ByteArray): Pair<WireConnectStatus, ByteArray> {
    val firstZeroIndex = byteArray.indexOf(0)
    val connectStatus = WireConnectStatus.valueOf(byteArray.decodeToString(0, if(firstZeroIndex < 0) byteArray.size else firstZeroIndex, true))
    val argument = if(firstZeroIndex < 0) byteArrayOf() else byteArray.copyOfRange(firstZeroIndex + 1, byteArray.size)
    return connectStatus to argument
  }

  private suspend fun readServiceMessage(): ByteArray {
    while(lifetime.isAlive) {
      synchronized(accumulatedMessages) {
        assert(!doDeliverMessages) { "Can't read service messages when already delivering messages to protocol" }
        if (accumulatedMessages.isNotEmpty())
          return accumulatedMessages.poll()
      }

      messagePushed.nextValue(lifetime) // wait suspending
    }
    throw CancellationException()
  }

  fun startDeliveringMessages() {
    deliverMessagesFuture.complete(Unit)
  }

  private fun dispatchByteBuffer(byteArray: ByteArray) {
    logger.trace { "'$id': received ${byteArray.size} bytes" }
    val reader = UnsafeBuffer(byteArray)
    val id = reader.readRdId()
    // TODO It may save allocations if buffer will be deserialized in-place
    wireStats.onReceive(id, byteArray.size.toLong())
    messageBroker.dispatch(id, reader)
  }

  init {
    runConnectionFlow(tlsChannel)

    framedByteChannel.adviseOnFrameReceived(lifetime) { buffer ->
      pingsWithoutReceivedData.set(0)

      val byteArray = buffer.readAll()

      logger.trace("$id: frame received, doDeliverMessages=${doDeliverMessages}, len=${byteArray.size}")

      // pings
      if (byteArray.isEmpty()) {
        return@adviseOnFrameReceived
      }

      if (connectionState.value is ConnectionState.DECLINED || connectionState.value == ConnectionState.CLOSED) {
        logger.debug("$id: frame received when wire is not connected: ${connectionState.value}")
        return@adviseOnFrameReceived
      }

      if(!doDeliverMessages) {
        synchronized(accumulatedMessages) {
          if (doDeliverMessages)
            dispatchByteBuffer(byteArray)
          else {
            accumulatedMessages.add(byteArray)
            messagePushed.fire(Unit)
          }
        }
      }
      else
        dispatchByteBuffer(byteArray)
    }

    framedByteChannel.adviseOnErrorReceived(lifetime) { throwable ->
      logger.warn("${javaClass.name} '${id}' switching to faulted state due to I/O error from framedByteChannel: ${throwable.message}")
      setFaultedState(throwable)
    }

    streamTransport.connected.advise(lifetime) { transportState ->
      if (transportState is ConnectionState.FAULTED || transportState is ConnectionState.CLOSED) {
        logger.info("$id: received $transportState from stream transport. Setting wire state to $transportState")
        connectionStateProperty.set(transportState)
      }
    }

    streamTransport.connected.advise(lifetime) {
      if (it == ConnectionState.CONNECTED)
        framedByteChannel.startReceiving()
    }

    lifetime.onTermination {
      logger.info("$id lifetime is being terminated. setting state to CLOSED")
      setStateIfNotTerminal(ConnectionState.CLOSED)
    }

    val noTimeoutsEnvVar = System.getenv("CWM_NO_TIMEOUTS")

    checkForTimeouts = noTimeoutsEnvVar.isNullOrEmpty() || noTimeoutsEnvVar == "0"

    connectionState.view(lifetime) { connectedLt, state ->
      if (state is ConnectionState.CONNECTED || state is ConnectionState.NOT_CONNECTED) {
        val timerTask = AppExecutorUtil.getAppScheduledExecutorService().scheduleWithFixedDelay({ pingAndCheckTimeout() }, 1, 1, TimeUnit.SECONDS)
        connectedLt.onTermination {
          timerTask.cancel(false)
        }
      }
    }
  }

  private fun pingAndCheckTimeout() {
    framedByteChannel.send(ByteBuffer.wrap(byteArrayOf())) // send empty frame as ping

    if (!checkForTimeouts) return

    val pings = pingsWithoutReceivedData.incrementAndGet()
    if (pings > 15) {
      logger.info("15 unsuccessful pings. set wire.connected to false...")
      setFaultedState(Exception("connection timeout"))
    }
  }

  override fun toString(): String = "${javaClass.simpleName}($id)"

  companion object {
    private val threadLocalWriteBuffers = ThreadLocal.withInitial { UnsafeBuffer(ByteArray(16384)) }
    private val logger = getLogger<SecureWireOverStreamTransport>()
  }

  override fun send(id: RdId, writer: (AbstractBuffer) -> Unit) {
    require(!id.isNull) { "id mustn't be null" }

    if (!bufChan.isOpen) {
      logger.debug("Channel is closed, send() will drop data silently for rd id $id")
      return
    }

    if (!lifetime.isAlive || connectionState.value is ConnectionState.CLOSED) {
      logger.debug("${this.id} wire's closed, send() will drop data silently for rd id $id")
      return
    }

    if (connectionState.value is ConnectionState.FAULTED) {
      logger.debug("${this.id} wire is in faulted state, send() will drop data silently for rd id $id")
      return
    }

    if (connectionState.value is ConnectionState.NOT_CONNECTED) {
      logger.error("${this.id}: wire is not connected in send() but also not faulted for rd id $id")
      return
    }

    val unsafeBuffer = threadLocalWriteBuffers.get()
    val initialPosition = unsafeBuffer.position
    try {
      id.write(unsafeBuffer) //write id
      contexts.writeCurrentMessageContext(unsafeBuffer)
      writer(unsafeBuffer) //write rest

      val len = unsafeBuffer.position - initialPosition

      wireStats.onSend(id, len.toLong())
      val bytes = unsafeBuffer.getArray()
      logger.trace { "'$id': sending $len bytes" }
      framedByteChannel.send(ByteBuffer.wrap(bytes.copyOfRange(initialPosition, initialPosition + len)))
    }
    catch (ex: ClosedChannelException) {
      logger.debug("Channel got closed during send(), data dropped for rd id $id")
      setFaultedState(ex)
    }
    finally {
      if (initialPosition == 0)
        unsafeBuffer.reset() // apply shrinking logic
      else
        unsafeBuffer.position = initialPosition
    }
  }

  class PeerVisibleErrorCode(val code: WireConnectStatus, message: String? = null, cause: Throwable? = null) : Exception(message, cause)
}