// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.protocol

import com.intellij.openapi.diagnostic.Logger
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.adviseUntil
import java.io.IOException
import java.nio.channels.ByteChannel
import java.nio.channels.SocketChannel
import kotlin.concurrent.thread

class SocketStreamTransport(private val socketChannel: SocketChannel, id: String): StreamTransport {
  override val byteChannel: ByteChannel
    get() = socketChannel

  // Socket is ready as soon as we get socket channel
  override val connected: ConnectionStateProperty = ConnectionStateProperty()

  init {
    require(socketChannel.isOpen) {
      "$id: socket channel must be open at the time of ${javaClass.simpleName} ctor"
    }

    connected.adviseUntil(Lifetime.Eternal) { state ->
      LOG.debug("socketStreamTransport '$id': connected = $state")
      state is ConnectionState.CLOSED || state is ConnectionState.FAULTED
    }

    // TODO Rewrite with selectors?
    thread(name = "SocketState:$id", isDaemon = true) {
      catch {
        LOG.debug("$id: waiting for a socket state change in a separate thread")

        try {
          socketChannel.finishConnect()
        } catch(ex: Throwable) {
          connected.set(ConnectionState.FAULTED(ex))
        }

        while (!socketChannel.isConnected && socketChannel.isOpen) {
          Thread.sleep(10)
        }

        if (!socketChannel.isOpen) {
          connected.set(ConnectionState.FAULTED(IOException("Connection was closed")))
          return@thread
        }

        connected.set(ConnectionState.CONNECTED)

        while (socketChannel.isConnected) {
          Thread.sleep(50)
        }

        connected.set(ConnectionState.CLOSED)

        LOG.debug("$id: quit waiting for socket close")
      }
    }
  }

  companion object {
    private val LOG = Logger.getInstance(SocketStreamTransport::class.java)
  }
}