// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.protocol

import com.jetbrains.rd.platform.util.NetworkUtils
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.closeNormal
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import tlschannel.impl.ByteBufferSet
import java.io.IOException
import java.lang.IllegalStateException
import java.net.URI
import java.net.http.WebSocket
import java.nio.ByteBuffer
import java.nio.channels.ByteChannel
import java.nio.channels.GatheringByteChannel
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CompletionStage
import java.util.concurrent.LinkedBlockingDeque
import java.util.concurrent.TimeUnit

class WebSocketStreamTransport(lifetime: Lifetime, uri: URI, httpHeaders: Map<String, String>, private val id: String)
  : StreamTransport {

  override val connected: ConnectionStateProperty = ConnectionStateProperty()

  private val inputQueue: LinkedBlockingDeque<ByteArray> = LinkedBlockingDeque()

  val ws: CompletableFuture<WebSocket>

  init {
    logger.debug("WebSocket client '$id' connecting to $uri")

    val httpClient = NetworkUtils.createHttpClientWithGoodDefaults()

    val wsBuilder = httpClient.newWebSocketBuilder()

    for ((headerName, headerValue) in httpHeaders) {
      wsBuilder.header(headerName, headerValue)
    }

    ws = wsBuilder
      .buildAsync(uri, object : WebSocket.Listener {
        override fun onOpen(webSocket: WebSocket) {
          logger.info("WebSocket client '$id' connected at $uri")
          connected.set(ConnectionState.CONNECTED)

          super.onOpen(webSocket)
        }

        override fun onText(webSocket: WebSocket, data: CharSequence, last: Boolean): CompletionStage<*>? {
          val error = "Received string message in ${javaClass.simpleName} id '$id'. String messages are not supported and will be ignored"
          if (application.isEAP || application.isInternal || application.isUnitTestMode) {
            logger.error(error)
          } else {
            logger.warn(error)
          }

          return super.onText(webSocket, data, last)
        }

        override fun onBinary(webSocket: WebSocket, data: ByteBuffer, last: Boolean): CompletionStage<*>? {
          val remaining = data.remaining()
          logger.trace("WebSocket client '$id' received message, size = $remaining")

          val arr = ByteArray(remaining)
          data.get(arr)
          inputQueue.put(arr)

          return super.onBinary(webSocket, data, last)
        }

        override fun onClose(webSocket: WebSocket, statusCode: Int, reason: String): CompletionStage<*>? {
          logger.info("WebSocket client '$id' disconnected, code: $statusCode, reason: $reason")
          connected.set(ConnectionState.CLOSED)
          return super.onClose(webSocket, statusCode, reason)
        }

        override fun onError(webSocket: WebSocket, ex: Throwable?) {
          logger.error("Error processing in WebSocket $id", ex)
          connected.set(ConnectionState.FAULTED(IOException("Error processing in WebSocket $id: ${ex?.message}", ex)))
          super.onError(webSocket, ex)
        }
      })

    ws.handleAsync { _, exception ->
      if (exception != null) {
        connected.set(ConnectionState.FAULTED(IllegalStateException("Unable to connect to $uri: ${exception.message}", exception)))
      } else {
        logger.debug("WebSocket client '$id' created for $uri")
      }
    }

    lifetime.onTermination {
      logger.debug("${javaClass.simpleName} $id disconnected and closing")
      ws.closeNormal()
    }
  }

  override val byteChannel: ByteChannel = object : ByteChannel, GatheringByteChannel {
    override fun isOpen(): Boolean = connected.value == ConnectionState.CONNECTED

    private fun send(buf: ByteBuffer) {
      val future = ws.get().sendBinary(buf, true)
      future.get()
    }

    override fun write(srcs: Array<out ByteBuffer>, offset: Int, length: Int): Long {
      val bufferSet = ByteBufferSet(srcs, offset, length)
      if (bufferSet.hasRemaining()) {
        val total = bufferSet.remaining().toInt()
        val buf = ByteBuffer.allocate(total)
        val copied = bufferSet.getRemaining(buf)
        assert(copied == total)

        buf.rewind()
        send(buf)
        return total.toLong()
      } else {
        return 0
      }
    }

    override fun write(srcs: Array<out ByteBuffer>): Long = write(srcs, 0, srcs.size)

    override fun write(src: ByteBuffer): Int {
      logger.trace("WebSocket client '$id' sending message, size = ${src.remaining()}")

      val length = src.remaining()
      send(src)
      return length
    }

    override fun close() = ws.closeNormal()

    override fun read(dst: ByteBuffer): Int {
      // Assumes one reader
      while (connected.value == ConnectionState.CONNECTED) {
        val buffer = inputQueue.poll(100, TimeUnit.MILLISECONDS)
        if (buffer == null) {
          logger.trace("WebSocket client '$id': waiting on read")
          continue
        }

        val available = dst.remaining()

        if (available >= buffer.size) {
          dst.put(buffer)
          return buffer.size
        }
        else {
          dst.put(buffer.sliceArray(0 until available))
          inputQueue.addFirst(buffer.sliceArray(available until buffer.size))
          return available
        }
      }

      return -1
    }
  }

  companion object {
    private val logger = getLogger<WebSocketStreamTransport>()
  }
}