// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.protocol

import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.rd.framework.RdId
import com.jetbrains.rd.framework.base.IRdWireable
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.onTermination
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicLong

class WireStatsCounter(lifetime: Lifetime) {
  val totalStats = EntityStats()

  var statsMode = StatsMode.Simple
    set(value) {
      field = value
      perEntityStats.clear()
      storedNames.clear()
      if (value == StatsMode.PerEntity) {
        rdidToEntity.forEach { (rdid: RdId, wireable: IRdWireable) ->
          storedNames[rdid] = wireable.location.toString()
        }
      }
    }

  val perEntityStats = ConcurrentHashMap<RdId, EntityStats>()
  private val rdidToEntity = ConcurrentHashMap<RdId, IRdWireable>()
  val storedNames = ConcurrentHashMap<RdId, String>()

  init {
    val task = AppExecutorUtil.getAppScheduledExecutorService().scheduleWithFixedDelay(this::updateMomentarySpeeds, 0, 1, TimeUnit.SECONDS)
    lifetime.onTermination {
      task.cancel(false)
    }
  }

  private fun updateMomentarySpeeds() {
    totalStats.updateSpeeds()

    perEntityStats.forEach { (_, stats) -> stats.updateSpeeds() }
  }

  fun onSend(id: RdId, numBytes: Long) {
    totalStats.logSend(numBytes)

    if (statsMode == StatsMode.Simple) return

    perEntityStats.getOrPut(id, ::EntityStats).logSend(numBytes)
  }

  fun onReceive(id: RdId, numBytes: Long) {
    totalStats.logReceive(numBytes)

    if (statsMode == StatsMode.Simple) return

    perEntityStats.getOrPut(id, ::EntityStats).logReceive(numBytes)
  }

  fun onAdvise(lifetime: Lifetime, entity: IRdWireable) {
    if (!lifetime.isAlive) return

    val rdid = entity.rdid
    rdidToEntity[rdid] = entity
    lifetime.onTermination {
      rdidToEntity.remove(rdid)
    }

    if(statsMode == StatsMode.Simple) return

    storedNames[rdid] = entity.location.toString()
  }

  enum class StatsMode {
    Simple,
    PerEntity
  }

  class EntityStats {
    val totalSend = AtomicLong(0)
    val totalReceive = AtomicLong(0)

    var sendSpeed = 0L
      private set
    var receiveSpeed = 0L
      private set

    private var lastSend = 0L
    private var lastReceive = 0L

    fun updateSpeeds() {
      val currentSend = totalSend.get()
      sendSpeed = currentSend - lastSend
      lastSend = currentSend

      val currentReceive = totalReceive.get()
      receiveSpeed = currentReceive - lastReceive
      lastReceive = currentReceive
    }

    fun logSend(bytes: Long) {
      totalSend.addAndGet(bytes)
    }

    fun logReceive(bytes: Long) {
      totalReceive.addAndGet(bytes)
    }
  }
}
