package com.jetbrains.rd.platform.codeWithMe.statistics

import com.intellij.internal.statistic.eventLog.EventLogGroup
import com.intellij.internal.statistic.eventLog.events.AnonymizedEventField
import com.intellij.internal.statistic.eventLog.events.EventFields
import com.intellij.internal.statistic.eventLog.events.VarargEventId
import com.intellij.internal.statistic.service.fus.collectors.CounterUsagesCollector
import com.intellij.openapi.project.Project
import com.jetbrains.codeWithMe.lobby.CodeWithMeTransport
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.SessionInfoProvider

object CodeWithMeLifecycleLogger: CounterUsagesCollector() {
  private const val GROUP_ID = "cwm.lifecycle"

  internal enum class SessionParticipantsSize { OneOnOne, Group }
  internal enum class SessionDuration { Below15mins, Below25mins, Below1hr, Above1hr }

  fun onClientConnected(project: Project, user: RdUserModel, transport: CodeWithMeTransport) {
    val sessionId = SessionInfoProvider.getInstance(project).getSessionId()

    CLIENT_CONNECTED.log(
      project,
      SESSION_ID_FIELD.with(sessionId),
      EventFields.CodeWithMeClientId.with(user.clientIdValue),
      ID_FIELD.with(user.id),
      CONNECTION_TYPE.with(transport)
    )
  }

  //explicit session Id cause session might be off already
  fun onClientFinished(project: Project, user: RdUserModel, sessionId: String) {
    CLIENT_FINISHED.log(project, SESSION_ID_FIELD.with(sessionId), EventFields.CodeWithMeClientId.with(user.clientIdValue), ID_FIELD.with(user.id))
  }

  fun onSessionCreated(project: Project, sessionId: String, telephonyEnabled: Boolean) {
    SESSION_CREATED.log(project, SESSION_ID_FIELD.with(sessionId), TELEPHONY_ENABLED_FIELD.with(telephonyEnabled))
  }

  fun onSessionTerminated(project: Project, sessionId: String, telephonyEnabled: Boolean) {
    SESSION_TERMINATED.log(project, SESSION_ID_FIELD.with(sessionId), TELEPHONY_ENABLED_FIELD.with(telephonyEnabled))
  }

  fun onSessionStarted(project: Project, sessionId: String) {
    SESSION_STARTED.log(project, SESSION_ID_FIELD.with(sessionId))
  }

  fun onSessionFinished(project: Project,
                        sessionId: String,
                        maxParticipants: Int,
                        sessionDurationMinutes: Int,
                        disconnectsCount: Int
  ) {
    val participantsType = calculateSessionParticipantsSize(maxParticipants)
    val durationType = calculateSessionDurationType(sessionDurationMinutes)

    SESSION_FINISHED.log(
      project,
      SESSION_ID_FIELD.with(sessionId),
      SESSION_PARTICIPANTS_SIZE_TYPE.with(participantsType),
      SESSION_PARTICIPANTS_MAX.with(maxParticipants),
      SESSION_DURATION_TYPE.with(durationType),
      SESSION_DURATION_MINUTES.with(sessionDurationMinutes),
      SESSION_DISCONNECTS_COUNT.with(disconnectsCount)
    )
  }

  private fun calculateSessionDurationType(durationMinutes: Int): SessionDuration =
    when {
      durationMinutes < 15 -> SessionDuration.Below15mins
      durationMinutes < 25 -> SessionDuration.Below25mins
      durationMinutes < 60 -> SessionDuration.Below1hr
      else -> SessionDuration.Above1hr
    }

  private fun calculateSessionParticipantsSize(maxParticipants: Int): SessionParticipantsSize =
    if (maxParticipants <= 2) SessionParticipantsSize.OneOnOne
    else SessionParticipantsSize.Group

  fun onSessionExpired(project: Project, sessionId: String) {
    SESSION_EXPIRED.log(project, SESSION_ID_FIELD.with(sessionId))
  }

  fun onEditorFollowingLabelResume(project: Project?, user: RdUserModel) {
    logEvent(project, ON_EDITOR_FOLLOWING_LABEL_RESUME, user)
  }

  fun onEditorFollowingLabelStop(project: Project?, user: RdUserModel) {
    logEvent(project, ON_EDITOR_FOLLOWING_LABEL_STOP, user)
  }

  fun onEditorFullSyncLabelStop(project: Project?, user: RdUserModel) {
    logEvent(project, ON_EDITOR_FULL_SYNC_LABEL_STOP, user)
  }

  fun onCircleLeftClickStart(project: Project?, user: RdUserModel) {
    logEvent(project, ON_CIRCLE_LEFT_CLICK_START, user)
  }

  fun onCircleLeftClickStop(project: Project?, user: RdUserModel) {
    logEvent(project, ON_CIRCLE_LEFT_CLICK_STOP, user)
  }

  fun onCircleRightClick(project: Project?, user: RdUserModel) {
    logEvent(project, ON_CIRCLE_RIGHT_CLICK, user)
  }

  fun onConnectionFailed(project: Project, sessionId: String) {
    ON_CONNECTION_FAILED.log(project, SESSION_ID_FIELD.with(sessionId))
  }

  fun onReconnection(project: Project, sessionId: String) {
    ON_RECONNECTION.log(project, SESSION_ID_FIELD.with(sessionId))
  }

  fun onReconnectionFailed(project: Project, sessionId: String) {
    ON_RECONNECTION_FAILED.log(project, SESSION_ID_FIELD.with(sessionId))
  }

  private fun logEvent(project: Project?, event: VarargEventId, user: RdUserModel) {
    val sessionId = project?.let {
      SessionInfoProvider.getInstance(it).getSessionId()
    }
    event.log(project, SESSION_ID_FIELD.with(sessionId), EventFields.CodeWithMeClientId.with(user.clientIdValue), ID_FIELD.with(user.id))
  }

  private val EVENT_GROUP = EventLogGroup(GROUP_ID, 7)

  private val ID_FIELD = EventFields.Int("id")
  private val TELEPHONY_ENABLED_FIELD = EventFields.Boolean("telephonyEnabled")
  private val SESSION_ID_FIELD = AnonymizedEventField("sessionId")
  private val CONNECTION_TYPE = EventFields.Enum("connectionType", CodeWithMeTransport::class.java)

  private val CLIENT_CONNECTED =
    EVENT_GROUP.registerVarargEvent("connected", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD, CONNECTION_TYPE)

  private val CLIENT_FINISHED = EVENT_GROUP.registerVarargEvent("finished", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val SESSION_CREATED = EVENT_GROUP.registerVarargEvent("sessionCreated", SESSION_ID_FIELD, TELEPHONY_ENABLED_FIELD)
  private val SESSION_TERMINATED = EVENT_GROUP.registerVarargEvent("sessionTerminated", SESSION_ID_FIELD, TELEPHONY_ENABLED_FIELD)
  private val SESSION_STARTED = EVENT_GROUP.registerVarargEvent("sessionStarted", SESSION_ID_FIELD)

  private val SESSION_PARTICIPANTS_SIZE_TYPE = EventFields.Enum("participantsSize", SessionParticipantsSize::class.java)
  private val SESSION_PARTICIPANTS_MAX = EventFields.Int("participantsMax")
  private val SESSION_DURATION_TYPE = EventFields.Enum("sessionDurationType", SessionDuration::class.java)
  private val SESSION_DURATION_MINUTES = EventFields.Int("sessionDurationMinutes")
  private val SESSION_DISCONNECTS_COUNT = EventFields.Int("disconnectsCount")
  private val SESSION_FINISHED = EVENT_GROUP.registerVarargEvent(
    "sessionFinished",
    SESSION_ID_FIELD,
    SESSION_PARTICIPANTS_SIZE_TYPE,
    SESSION_PARTICIPANTS_MAX,
    SESSION_DURATION_TYPE,
    SESSION_DURATION_MINUTES,
    SESSION_DISCONNECTS_COUNT
  )

  private val SESSION_EXPIRED = EVENT_GROUP.registerVarargEvent("sessionExpired", SESSION_ID_FIELD)
  private val ON_EDITOR_FOLLOWING_LABEL_RESUME = EVENT_GROUP.registerVarargEvent("onEditorFollowingLabelResume", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val ON_EDITOR_FOLLOWING_LABEL_STOP = EVENT_GROUP.registerVarargEvent("onEditorFollowingLabelStop", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val ON_EDITOR_FULL_SYNC_LABEL_STOP = EVENT_GROUP.registerVarargEvent("onEditorFullSyncLabelStop", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val ON_CIRCLE_LEFT_CLICK_START = EVENT_GROUP.registerVarargEvent("onCircleLeftClickStart", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val ON_CIRCLE_LEFT_CLICK_STOP = EVENT_GROUP.registerVarargEvent("onCircleLeftClickStop", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val ON_CIRCLE_RIGHT_CLICK = EVENT_GROUP.registerVarargEvent("onCircleRightClick", SESSION_ID_FIELD, EventFields.CodeWithMeClientId, ID_FIELD)
  private val ON_CONNECTION_FAILED = EVENT_GROUP.registerVarargEvent("onConnectionFailed", SESSION_ID_FIELD)
  private val ON_RECONNECTION = EVENT_GROUP.registerVarargEvent("onReconnection", SESSION_ID_FIELD)
  private val ON_RECONNECTION_FAILED = EVENT_GROUP.registerVarargEvent("onReconnectionFailed", SESSION_ID_FIELD)

  override fun getGroup(): EventLogGroup = EVENT_GROUP
}