package com.jetbrains.rd.platform.codeWithMe.telephony

import com.intellij.ide.BrowserUtil
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.wm.ToolWindowManager
import com.intellij.ui.jcef.JBCefBrowser
import com.intellij.ui.jcef.JBCefJSQuery
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.toolWindow.CodeWithMeToolWindowFactory
import com.jetbrains.rd.platform.codeWithMe.users.UserNotifications
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.fragmentParameters
import com.jetbrains.rd.platform.util.newURIWithFragmentParameters
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.reactive.Property
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.JsonObject
import kotlinx.serialization.json.JsonPrimitive
import org.cef.CefSettings
import org.cef.browser.CefBrowser
import org.cef.browser.CefFrame
import org.cef.handler.*
import org.cef.network.CefRequest
import org.intellij.lang.annotations.Language
import java.net.URL
import java.net.URLDecoder
import javax.swing.JComponent


class TelephonySession(val lifetime: Lifetime, val project: Project, jitsiRoomUrlString: String) {
  companion object {
    val LOG = logger<TelephonySession>()
  }

  private val browser: JBCefBrowser

  private val notificationsQuery: JBCefJSQuery
  private val conferenceLeaveQuery: JBCefJSQuery
  private val participantEnabledCameraQuery: JBCefJSQuery
  private var participantEnabledCameraNotificationShown = false
  private val cameraEnabledQuery: JBCefJSQuery
  private val microphoneEnabledQuery: JBCefJSQuery
  private val conferenceJoinedQuery: JBCefJSQuery
  private var myCameraEnabledHandled = false
  private val failedToInstallConferenceHooksQuery: JBCefJSQuery
  private val notifyPrejoinScreenQuery: JBCefJSQuery
  private var failedHooks = mutableListOf<String>()

  private val jitsiRoomUrl = run {
    URL(jitsiRoomUrlString).toURI()
  }
  private val localUserName = jitsiRoomUrl.fragmentParameters["userInfo.displayName"]?.let {
    val trimmed = it.trim('"')
    URLDecoder.decode(trimmed, Charsets.UTF_8)
  }
  val cameraEnabled : Property<Boolean> = Property(false)
  val microphoneEnabled : Property<Boolean> = Property(false)
  val conferenceJoined : Property<Boolean> = Property(false)

  fun toggleMicrophone() {
    browser.cefBrowser.executeJavaScript(
      "APP.conference.toggleAudioMuted()", "", 4
    )
  }

  fun toggleCamera() {
    browser.cefBrowser.executeJavaScript(
      "APP.conference.toggleVideoMuted()", "", 3
    )
  }


  fun openSettings() {
    @Language("JavaScript")
    val jsCode = """
      function openSettings(tries) {
        if (tries < 228) {
          if (document.querySelector('div.settings-dialog')) {
            return;
          }
          // mega crutch because I'm unable to find a legal solution with API usage
          document.querySelector('[aria-label=\"Toggle more actions menu\"]').click(); 
          document.querySelector('li[aria-label=\"Toggle settings\"]').click();
          setTimeout(() => openSettings(tries + 1), 100);
        }
      }
      openSettings(0);
    """.trimIndent()
    browser.cefBrowser.executeJavaScript(jsCode, "", 3)
  }

  // used to bypass prejoin screen or go back to prejoin screen
  fun joinConference(join: Boolean) {
    if (conferenceJoined.value == join) return
    if (join) {
      browser.cefBrowser.executeJavaScript(
        // mega crutch because I'm unable to find a legal solution with API usage
        "document.querySelectorAll('[data-testid=\"prejoin.joinMeeting\"]')[0].click()", "", 1
      )
    } else {
      browser.cefBrowser.executeJavaScript("APP.conference.hangup()", "", 2)
    }
  }

  init {
    val resultUrl = buildJitsiUrl(null)
    browser = JBCefBrowser(resultUrl)
    val client = browser.jbCefClient

    lifetime.onTerminationIfAlive {
      Disposer.dispose(browser)
    }

    notificationsQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    conferenceLeaveQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    cameraEnabledQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    microphoneEnabledQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    conferenceJoinedQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    participantEnabledCameraQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    failedToInstallConferenceHooksQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")
    notifyPrejoinScreenQuery = JBCefJSQuery.create(browser) ?: error("TelephonySession(): unable to create JBCefJSQuery")

    notificationsQuery.addHandler {
      val messageText: String
      val messageNick: String
      try {
        val tree: JsonObject = (Json.Default.parseToJsonElement(it) as? JsonObject) ?: error("Json Object is expected")
        @Suppress("HardCodedStringLiteral", "DialogTitleCapitalization")
        messageText = (tree["text"] as? JsonPrimitive)?.content ?: error("empty message text")
        @Suppress("HardCodedStringLiteral")
        messageNick = (tree["nick"] as? JsonPrimitive)?.content ?: error("empty message nickname")
      }
      catch (e: IllegalStateException) {
        LOG.warn(e)
        return@addHandler null
      }
      if (messageNick != localUserName) {
        val title = CodeWithMePlatformBundle.message("notification.telephonyChatMessage.title", messageNick)
        UserNotifications.showInfoNotification(project, title, messageText, true)
      }
      return@addHandler null
    }

    conferenceLeaveQuery.addHandler {
      conferenceJoined.set(false)
      LOG.info("conference left, going to reload page")
      val resultUrl_ = buildJitsiUrl(true)
      browser.loadURL(resultUrl_)
      return@addHandler null
    }

    cameraEnabledQuery.addHandler {
      cameraEnabled.set(it == "true")
      if (cameraEnabled.value && !myCameraEnabledHandled) {
        myCameraEnabledHandled = true
        openToolWindow()
      }
      return@addHandler null
    }

    microphoneEnabledQuery.addHandler {
      microphoneEnabled.set(it == "true")
      return@addHandler null
    }

    conferenceJoinedQuery.addHandler {
      conferenceJoined.set(true)
      if (failedHooks.size > 0) {
        LOG.error("Failed to install some voice chat hooks ($failedHooks), jitsi version: $it")

      }

      return@addHandler null
    }

    participantEnabledCameraQuery.addHandler {
      if (!participantEnabledCameraNotificationShown) {
        participantEnabledCameraNotificationShown = true
        UserNotifications.showClickableLinkNotification(project,
          CodeWithMePlatformBundle.message("notification.telephonyParticipantVideoEnabled.title"),
          CodeWithMePlatformBundle.message("notification.telephonyParticipantVideoEnabled.message")
        ) {
          openToolWindow()
        }
      }
      return@addHandler null
    }

    failedToInstallConferenceHooksQuery.addHandler {
      failedHooks.add(it)
      return@addHandler null
    }

    notifyPrejoinScreenQuery.addHandler {
      UserNotifications.showClickableLinkNotification(
        project,
        CodeWithMePlatformBundle.message("notification.telephonyPrejoinScreenShowing.title"),
        CodeWithMePlatformBundle.message("notification.telephonyPrejoinScreenShowing.message"),
        -1
      ) {
        openToolWindow()
      }
      return@addHandler null
    }

    client.addDisplayHandler(createDisplayHandler(), browser.cefBrowser)
    client.addLoadHandler(createLoadHandler(), browser.cefBrowser)
    //client.addRequestHandler(createRequestHandler(), browser.cefBrowser)
    client.addLifeSpanHandler(createLifeSpanHandler(), browser.cefBrowser)

    browser.cefBrowser.createImmediately()
  }

  // I don't like it to have any knowledge of container, but whatever. If there will be more such cases - should use
  // some listener instead... ?
  private fun openToolWindow() {
    application.invokeLater {
      ToolWindowManager.getInstance(project).getToolWindow(CodeWithMeToolWindowFactory.codeWithMeToolWindowID)?.show()
    }
  }

  fun getComponent(): JComponent {
    return browser.component
  }

  private fun buildJitsiUrl(overridePrejoin: Boolean?): String {
    val currentParameters = jitsiRoomUrl.fragmentParameters.toMutableMap()

    if (overridePrejoin != null) {
      currentParameters["config.prejoinPageEnabled"] = if (overridePrejoin) "true" else "false"
    }

    //if (currentParameters.containsKey("config.prejoinPageEnabled")) {
    //  if (overridePrejoin != null) {
    //    currentParameters["config.prejoinPageEnabled"] = if (overridePrejoin) "true" else "false"
    //  }
    //} else {
    //  currentParameters["config.prejoinPageEnabled"] = "true"
    //}

    return jitsiRoomUrl.newURIWithFragmentParameters(currentParameters).toString()
  }

  private fun createDisplayHandler(): CefDisplayHandler {
    return object : CefDisplayHandlerAdapter() {
      override fun onConsoleMessage(browser: CefBrowser,
                                    level: CefSettings.LogSeverity,
                                    message: String,
                                    source: String,
                                    line: Int): Boolean
      {
        LOG.debug(message)
        return true
      }
    }
  }

  private fun createLoadHandler(): CefLoadHandlerAdapter {
    return object : CefLoadHandlerAdapter() {

      override fun onLoadEnd(browser_: CefBrowser?, frame: CefFrame?, httpStatusCode: Int) {
        LOG.debug("onLoadEnd: page loading ended with $httpStatusCode status code")
        browser_ ?: error("onLoadEnd: Jitsi loaded, but onLoadEnd got no browser")
        @Suppress("JSIgnoredPromiseFromCall")
        @Language("JavaScript")
        val jsCode = """
          function sleep(ms) {
            return new Promise(resolve => setTimeout(resolve, ms));
          }
          
          function installHook(event, handler) {
            try {
              var addListenerFunc = APP.conference.addConferenceListener || function(e, h) {APP.conference._room.on(e, h);};
              addListenerFunc(event, handler);
              console.log("hook installed: " + event);
            } catch (e) {
              console.error("failed to install hook: " + event + ", error: " + e);
              ${failedToInstallConferenceHooksQuery.inject("event")};
            }
          }
          
          if (!JitsiMeetJS) {          
            JitsiMeetJS = {};
          }
          
          var message_received = JitsiMeetJS.events?.conference?.MESSAGE_RECEIVED || "conference.messageReceived";
          var track_muted = JitsiMeetJS.events?.conference?.TRACK_MUTE_CHANGED || "conference.trackMuteChanged";
          var track_added = JitsiMeetJS.events?.conference?.TRACK_ADDED || "conference.trackAdded";
          var track_removed = JitsiMeetJS.events?.conference?.TRACK_REMOVED || "conference.trackRemoved";
          var conference_left = JitsiMeetJS.events?.conference?.CONFERENCE_LEFT || "conference.left";
          
          async function hookInstaller() {
            if (document.querySelector('div.premeeting-screen')) {
              ${notifyPrejoinScreenQuery.inject("")};
            }
            while (!APP.conference.isJoined()) {
              console.log("Conference is not joined. Waiting...");
              await sleep(1000);
            }
            installHook(message_received, function(id, text, ts, nick)
              {
                var messageInfo = JSON.stringify({
                  "nick": nick,
                  "text": text
                });
                ${notificationsQuery.inject("messageInfo")};
              }
            );
            installHook(conference_left,
              function(){
                ${conferenceLeaveQuery.inject("")};
              }
            );
            track_mon_func = function(track) {
              console.log(track);
              if (track.isLocal()) {
                if (track.isAudioTrack()) {
                  ${microphoneEnabledQuery.inject("!track.isMuted()")};
                } else if (track.isVideoTrack()) {
                  ${cameraEnabledQuery.inject("!track.isMuted()")};
                }
              } else {
                if (track.isAudioTrack()) {
                  // nothing
                } else if (track.isVideoTrack()) {
                  if (!track.isMuted()) {
                    ${participantEnabledCameraQuery.inject("")};
                  }
                }
              }
            };
            
            installHook(track_muted, track_mon_func);
            installHook(track_added, track_mon_func);
            installHook(track_removed, function(track) 
              {
                if (track.isLocal()) {
                  if (track.isAudioTrack()) {
                    ${microphoneEnabledQuery.inject("false")};
                  } else if (track.isVideoTrack()) {
                    ${cameraEnabledQuery.inject("false")};
                  } 
                }
              }  
            );

            ${microphoneEnabledQuery.inject("!APP.conference.isLocalAudioMuted()")};
            ${cameraEnabledQuery.inject("!APP.conference.isLocalVideoMuted()")};
            ${conferenceJoinedQuery.inject("JitsiMeetJS.version")};
          }

          hookInstaller();
        """.trimIndent()


        //APP.store.getState()['features/base/media']
        // api.executeCommand('toggleVideo');
        //APP.conference.addListener(JitsiMeetJS.events.conference.CONFERENCE_JOINED, window.myCustomAddMessageListener);
        browser_.executeJavaScript(jsCode, "", 0)
      }
    }
  }

  private fun createRequestHandler(): CefRequestHandlerAdapter {
    return object : CefRequestHandlerAdapter() {
      override fun onBeforeBrowse(browser: CefBrowser?,
                                  frame: CefFrame?,
                                  request: CefRequest?,
                                  user_gesture: Boolean,
                                  is_redirect: Boolean): Boolean {
        if (request != null && request.transitionType == CefRequest.TransitionType.TT_LINK) {
          BrowserUtil.browse(request.url)
          return true
        }
        return false
      }

      override fun onOpenURLFromTab(browser: CefBrowser?, frame: CefFrame?, target_url: String?, user_gesture: Boolean): Boolean {
        if (target_url != null) {
          BrowserUtil.browse(target_url)
          return true
        }
        return false
      }
    }
  }

  private fun createLifeSpanHandler(): CefLifeSpanHandlerAdapter {
    return object : CefLifeSpanHandlerAdapter() {
      override fun onBeforePopup(browser: CefBrowser?, frame: CefFrame?, target_url: String?, target_frame_name: String?): Boolean {
        if (target_url != null) {
          BrowserUtil.browse(target_url)
          return true
        }
        return false
      }
    }
  }
}