package com.jetbrains.rd.platform.codeWithMe.tests.impl

import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isLocal
import com.intellij.diagnostic.ThreadDumper
import com.intellij.ide.IdeEventQueue
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.notification.Notifications
import com.intellij.openapi.application.PathManager
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ex.ProjectManagerEx
import com.intellij.openapi.util.SystemInfo
import com.intellij.openapi.util.text.StringUtil
import com.jetbrains.codeWithMe.model.tests.DistributedTestModel
import com.jetbrains.codeWithMe.model.tests.RdTestSessionException
import com.jetbrains.rd.framework.*
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.platform.codeWithMe.tests.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.reactive.viewNotNull
import org.apache.log4j.Level
import org.apache.log4j.LogManager
import org.jetbrains.annotations.ApiStatus
import java.io.File
import java.net.InetAddress
import java.time.LocalTime
import java.time.format.DateTimeFormatter
import kotlin.reflect.full.createInstance


@ApiStatus.Internal
class DistributedTestHost(private val project: Project) {

  companion object {
    private val logger = getLogger<DistributedTestHost>()
  }

  init {
    val hostAddress = when (SystemInfo.isLinux) {
      true -> System.getenv(AgentConstants.dockerHostIpEnvVar)?.let {
        logger.info("${AgentConstants.dockerHostIpEnvVar} env var is set=$it, will try to get address from it.")
        // this won't work when we do custom network setups as the default gateway will be overridden
        // val hostEntries = File("/etc/hosts").readText().lines()
        // val dockerInterfaceEntry = hostEntries.last { it.isNotBlank() }
        // val ipAddress = dockerInterfaceEntry.split("\\s".toRegex()).first()
        InetAddress.getByName(it)
      } ?: InetAddress.getLoopbackAddress()
      false -> InetAddress.getLoopbackAddress()
    }

    logger.info("Host address=${hostAddress}")

    val port = System.getenv(AgentConstants.protocolPortEnvVar)?.toIntOrNull()

    if (port != null) {
      logger.info("Queue creating protocol on port $port")
      application.invokeLater { createProtocol(hostAddress, port) }
    }
  }

  private fun createProtocol(hostAddress: InetAddress, port: Int) {
    logger.info("Creating protocol...")
    val lifetime = project.lifetime

    val wire = SocketWire.Client(lifetime, DistributedTestIdeScheduler, port, AgentConstants.protocolName, hostAddress)
    val protocol = Protocol(AgentConstants.protocolName, Serializers(), Identities(IdKind.Client), DistributedTestIdeScheduler, wire, lifetime)
    val model = DistributedTestModel.create(lifetime, protocol)

    logger.info("Advise for session...")
    model.session.viewNotNull(lifetime) { lt, session ->
      try {
        val context = AgentContext(session.agentId, application, project, protocol)
        logger.info("New test session: ${session.testClassName}.${session.testMethodName}")
        logger.info("Setting up loggers")
        AgentTestLoggerFactory.bindSession(lt, session)

        // Create test class
        val testClass = Class.forName(session.testClassName)
        val testClassObject = testClass.kotlin.createInstance() as DistributedTestPlayer

        // Tell test we are running it inside an agent
        val agentInfo = AgentInfo(session.agentId, session.testMethodName)
        val queue = testClassObject.initAgent(agentInfo)

        // Play test method
        val testMethod = testClass.getMethod(session.testMethodName)
        testMethod.invoke(testClassObject)

        // Advice for processing events
        session.runNextAction.set { _, _ ->
          try {
            assert(ClientId.current.isLocal) { "ClientId '${ClientId.current}' should be local when test method starts" }

            val action = queue.remove()
            showNotification(action.title)

            // Flush all events to process pending protocol events and other things
            //   before actual test method execution
            IdeEventQueue.getInstance().flushQueue()

            // Execute test method
            logger.info("Execute next action ${action.title}...")
            val result = action.action.invoke(context)
            logger.info("Executing completed")

            // Sync state across all IDE agents to maintain proper order in protocol events
            logger.info("Sync protocol events after execution...")
            DistributedTestBridge.getInstance(project).syncProtocolEvents()
            IdeEventQueue.getInstance().flushQueue()
            logger.info("Protocol state sync completed")

            // Assert state
            assertStateAfterTestMethod()

            return@set result
          }
          catch (ex: Throwable) {
            logger.error("Exception during test action execution", ex)
            session.sendException.fire(ex.toModel())
            return@set RdTask.fromResult(false)
          }
        }

        session.shutdown.advise(lifetime) {
          logger.info("Close project...")
          try {
            ProjectManagerEx.getInstanceEx().forceCloseProject(project)
          }
          catch (e: Exception) {
            logger.error(e)
          }

          logger.info("Shutdown application...")
          application.exit(true, true, false)
        }

        session.dumpThreads.adviseOn(lifetime, DistributedTestInplaceScheduler) {
          logger.info("Dump threads...")
          val threadDump = ThreadDumper.dumpThreadsToString()
          val threadDumpStamp = LocalTime.now().format(DateTimeFormatter.ofPattern("HH_mm_ss_SSS"))
          val threadDumpFileName = "${AgentConstants.threadDumpFilePrefix}_$threadDumpStamp.log"
          val threadDumpFile = File(PathManager.getLogPath()).resolve(threadDumpFileName)
          threadDumpFile.writeText(threadDump)
        }

        // Initialize loggers
        for (category in session.traceCategories) {
          logger.info("Enabling TRACE for $category")
          val logger = LogManager.getLogger(category)
          logger.level = Level.TRACE
        }

        logger.info("Test session ready!")
        session.ready.value = true
      }
      catch (ex: Throwable) {
        logger.error("Exception during test session initializing", ex)
        session.sendException.fire(ex.toModel())
        session.ready.value = false
      }
    }
  }

  private fun assertStateAfterTestMethod() {
    assert(Logger.getFactory() is AgentTestLoggerFactory) {
      "Logger Factory was overridden during test method execution. " +
      "Inspect logs to find stack trace of the overrider. " +
      "Overriding logger factory leads to breaking distributes test log processing."
    }
  }

  @Suppress("HardCodedStringLiteral", "DialogTitleCapitalization")
  private fun showNotification(text: String?): Notification? {
    if (application.isHeadlessEnvironment || text.isNullOrBlank())
      return null

    val notification = Notification("TestFramework",
                                    "Test Framework",
                                    text,
                                    NotificationType.INFORMATION)
    Notifications.Bus.notify(notification)
    return notification
  }

  private fun Throwable.toModel(): RdTestSessionException {
    return RdTestSessionException(
      this.javaClass.typeName,
      this.message ?: "",
      StringUtil.getThrowableText(this))
  }
}