package com.jetbrains.rd.platform.codeWithMe.toolbar

import com.intellij.icons.AllIcons
import com.intellij.ide.BrowserUtil
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.actionSystem.ex.CustomComponentAction
import com.intellij.openapi.actionSystem.ex.TooltipDescriptionProvider
import com.intellij.openapi.actionSystem.ex.TooltipLinkProvider
import com.intellij.openapi.actionSystem.impl.ActionButton
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.isNotificationSilentMode
import com.intellij.openapi.ui.popup.Balloon
import com.intellij.openapi.util.EmptyRunnable
import com.intellij.openapi.util.Key
import com.intellij.openapi.wm.IdeFrame
import com.intellij.openapi.wm.WindowManager
import com.intellij.ui.GotItTooltip
import com.intellij.ui.LayeredIcon
import com.intellij.ui.awt.RelativePoint
import com.intellij.ui.scale.JBUIScale
import com.intellij.util.IconUtil
import com.intellij.util.ui.JBUI
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.toolbar.combobox.CWMPopup
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rd.util.reactive.Property
import org.jetbrains.annotations.Nls
import java.awt.Dimension
import java.awt.Point
import java.awt.event.HierarchyEvent
import java.awt.event.HierarchyListener
import java.net.URL
import javax.swing.*

abstract class CWMShowPopupAction: AnAction(), CustomComponentAction, DumbAware, TooltipLinkProvider, TooltipDescriptionProvider {
  companion object {
    val key = Key<JComponent>("CodeWithMeControl")
    const val MAX_ROWS_COUNT = 30
    var DEFAULT_MINIMUM_BUTTON_SIZE: Dimension = JBUI.size(32, 22)
  }

  private var sessionWasNotInitialized: Property<Boolean?>? = null
  private val lifetimes = SequentialLifetimes(Lifetime.Eternal)

  override fun actionPerformed(e: AnActionEvent) {
    val project = e.project ?: return

    val frame = WindowManager.getInstance().getFrame(project)
    if (frame !is IdeFrame) return

    val popup = createPopup(e.dataContext, EmptyRunnable.INSTANCE)

    val component = project.getUserData(key) ?: e.inputEvent?.component
    if (component != null) {
      popup.show(RelativePoint(component, Point(0, component.height)))
    }
    else {
      popup.showCenteredInCurrentWindow(project)
    }
  }

  private fun createPopup(dataContext: DataContext, onDispose: Runnable): CWMPopup {
    val group = ActionManager.getInstance().getAction("CodeWithMeGroup") as DefaultActionGroup
    return CWMPopup(group, dataContext, false, onDispose, MAX_ROWS_COUNT, true, null)
  }

  override fun update(e: AnActionEvent) {
    e.presentation.icon = AllIcons.CodeWithMe.CwmAccess
    e.presentation.hoveredIcon = e.presentation.icon

    if (e.place == ActionPlaces.MAIN_MENU) {
      e.presentation.text = "${CodeWithMePlatformBundle.message("product.name")}..."
      e.presentation.description = "${CodeWithMePlatformBundle.message("product.name")}..."
      return
    }

    val project = e.project ?: return

    val userManager = UserManager.getInstance(project)
    if (sessionWasNotInitialized == null) {
      sessionWasNotInitialized = Property(true)
      userManager.sessionEnabled.advise(lifetimes.next(), handler = {
        if (sessionWasNotInitialized!!.value == true && it) {
          sessionWasNotInitialized!!.set(false)
        }
      })
    }
    else if (!(sessionWasNotInitialized!!.value as Boolean)) {
      lifetimes.terminateCurrent()
    }

    e.presentation.icon = if (sessionWasNotInitialized!!.value == true) {
      layeredIconWithArrowDown(AllIcons.CodeWithMe.CwmAccess)
    }

    else if ((!userManager.users.isEmpty() && java.lang.Boolean.getBoolean("cwm.ui.test.mode")) || userManager.sessionEnabled.value) {
      layeredIconWithArrowDown(AllIcons.CodeWithMe.CwmAccessOn)
    }
    else {
      layeredIconWithArrowDown(AllIcons.CodeWithMe.CwmAccessOff)
    }
    updateHoveredIcon(e.presentation)
  }

  private fun layeredIconWithArrowDown(icon: Icon): LayeredIcon {
    val icon1 = LayeredIcon(2)
    icon1.setIcon(icon, 0, 0, 0)
    icon1.setIcon(AllIcons.General.ButtonDropTriangle, 1, JBUI.scale(18), JBUIScale.scale(7.5f).toInt())
    return icon1
  }

  private fun updateHoveredIcon(presentation: Presentation) {
    presentation.hoveredIcon = LayeredIcon(2).apply {
      setIcon(IconUtil.colorize(IconUtil.getEmptyIcon(true),
                                JBUI.CurrentTheme.ActionButton.hoverBackground()), 0, SwingConstants.CENTER)
      setIcon(presentation.icon, 1, SwingConstants.CENTER)
    }
  }

  private val learnMore: @Nls String = CodeWithMePlatformBundle.message("code.with.me.learn.more")
  private val learnMoreLink: @Nls String = CodeWithMePlatformBundle.message("code.with.me.learn.more.link")

  abstract val text: @Nls String
  abstract val description: @Nls String
  abstract val gotItHeader: @Nls String
  abstract val gotItMessage: @Nls String

  override fun createCustomComponent(presentation: Presentation, place: String, dataContext: DataContext): JComponent {
    val project = CommonDataKeys.PROJECT.getData(dataContext)
    return if (place == ActionPlaces.MAIN_MENU) {
      ActionButton(this, presentation, place, DEFAULT_MINIMUM_BUTTON_SIZE)
    }
    else {
      presentation.text = text
      presentation.description = description

      if (project != null && !isNotificationSilentMode(project)) {
        GotItTooltip("cwm.invitation.notification", gotItMessage,
                     CWMPluginDisposableService.getInstance(project)).withHeader(gotItHeader)
          .withPosition(Balloon.Position.below)
          .withBrowserLink(learnMore, URL(learnMoreLink))
          .assignTo(presentation, GotItTooltip.BOTTOM_MIDDLE)
      }


      object : ActionButton(this, presentation, place, DEFAULT_MINIMUM_BUTTON_SIZE) {
        private val listener = HierarchyListener { e ->
          if (e.changeFlags and HierarchyEvent.SHOWING_CHANGED.toLong() != 0L) {
            check()
          }
        }

        override fun addNotify() {
          super.addNotify()
          addHierarchyListener(listener)
          check()
        }

        override fun removeNotify() {
          removeHierarchyListener(listener)
          super.removeNotify()
        }

        private fun check() {
          SwingUtilities.getWindowAncestor(this)?.let {
            if (it is IdeFrame) {
              it.project?.let { project ->
                if (isShowing) {
                  project.putUserData(key, this)
                }
              }
            }
          }
        }
      }.apply {
        isOpaque = false
        border = BorderFactory.createEmptyBorder(1, 2, 1, 2)
      }
    }
  }

  override fun getTooltipLink(owner: JComponent?): TooltipLinkProvider.TooltipLink? {
    return TooltipLinkProvider.TooltipLink(learnMore, Runnable {
      BrowserUtil.browse(learnMoreLink)
    })
  }
}

class CWMThinClientShowPopupAction: CWMShowPopupAction() {
  override val text: String
    get() = CodeWithMePlatformBundle.message("action.CWMThinClientShowPopupAction.text")
  override val description: String
    get() = CodeWithMePlatformBundle.message("action.CWMThinClientShowPopupAction.description")
  override val gotItHeader: String
    get() = CodeWithMePlatformBundle.message("action.CWMThinClientShowPopupAction.gotIt.header")
  override val gotItMessage: String
    get() = description
}