package com.jetbrains.rd.platform.codeWithMe.toolbar

import com.intellij.icons.AllIcons
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.util.text.StringUtil
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.control.icons.CircleFactories
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.isHost
import com.jetbrains.rd.platform.util.project
import com.jetbrains.codeWithMe.model.RdUserModel
import java.util.function.Supplier
import javax.swing.Icon
import kotlin.math.min

class CodeWithMeUsersGroup : ActionGroup(), DumbAware {
  companion object {
    private const val nameTrunc = 22
    private const val nameSeparator = " \u00b7 "
    val userComparator = Comparator<RdUserModel> { o0, o1 ->
      when {
        o1.isHost -> 1
        o0.isHost -> -1
        else -> {
          val compare = StringUtil.compare(o0.name, o1.name, true)
          if (compare == 0) -1 else compare
        }
      }
    }

    fun getUsersActionGroup(userManager: UserManager, users: List<RdUserModel>): MutableList<AnAction> {
      val actions = mutableListOf<AnAction>()

      if (users.isNotEmpty()) {
        actions.add(Separator(CodeWithMePlatformBundle.message("participants.separator.text", users.size)))
        ActionManager.getInstance().getAction("ForceOthersToFollowAction")?.let {
          actions.add(it)
        }

        userManager.me.value?.let {
          val name = getName(it, CodeWithMePlatformBundle.message("circle.control.user.state.you"))
          actions.add(object : AnAction({ name }, CircleFactories.createUsersIcon(it)) {
            override fun actionPerformed(e: AnActionEvent) {
            }
          })
        }

        actions.addAll(users.map { userManager.createActionGroup(it) })
      }

      return actions
    }

    fun getName(user: RdUserModel, additionalText: String? = null, escapeHtml: Boolean = false): String {
      val length = min(nameTrunc, user.name.length)
      // var name = "${user.id} ${user.name.substring(0, length)}"
      var name = user.name.substring(0, length)

      if (user.name.length - 3 > nameTrunc) {
        name += "..."
      }

      additionalText?.let {
        name += " $it"
      }

      if (escapeHtml)
        name = StringUtil.escapeXmlEntities(name)

      return if (user.isHost) {
        "$name$nameSeparator${CodeWithMePlatformBundle.message("circle.control.user.state.host")}"
      } else {
        name
      }
    }
  }


  override fun getChildren(e: AnActionEvent?): Array<AnAction> {
    val project = e?.dataContext?.project ?: return arrayOf()
    val userManager = UserManager.getInstance(project)

    val users = userManager.users

    val actions = getUsersActionGroup(userManager, users)
    return actions.toTypedArray()
  }
}

fun UserManager.createActionGroup(user: RdUserModel,
                                  icon: Icon = CircleFactories.createUsersIcon(user)): ActionGroup {
  return RdUserActionGroup(this@createActionGroup, user, icon)
}

fun UserManager.createJumpToAction(user: RdUserModel): AnAction {
  return object : AnAction(Supplier { CodeWithMePlatformBundle.message("action.jump.to") }, AllIcons.CodeWithMe.CwmJump), DumbAware {
    override fun actionPerformed(e: AnActionEvent) {
      this@createJumpToAction.jumpTo(user)
    }

    override fun update(e: AnActionEvent) {
      val youFollow = this@createJumpToAction.lead.value == user
      e.presentation.isEnabled = !youFollow
    }
  }
}

fun UserManager.createFollowAction(user: RdUserModel): AnAction {
  val value = object : AnAction(Supplier { CodeWithMePlatformBundle.message("action.follow") }, AllIcons.CodeWithMe.CwmFollow), DumbAware {
    override fun actionPerformed(e: AnActionEvent) {
      this@createFollowAction.jumpTo(user)          // Firstly jump
      this@createFollowAction.startFollowing(user)  // Then follow
    }

    override fun update(e: AnActionEvent) {
      val youFollow = this@createFollowAction.lead.value == user
      e.presentation.isEnabled = !youFollow
    }

    public override fun setShortcutSet(shortcutSet: ShortcutSet) {
      super.setShortcutSet(shortcutSet)
    }
  }
  val userManager: UserManager = UserManager.getInstance(project)
  if (SingleUserFollowAction().isDefaultFollowee(user, userManager)) {
    val customShortcutSet = CustomShortcutSet.fromString("control shift alt Y")
    value.shortcutSet = customShortcutSet
  }
  return value
}

fun UserManager.createFullSyncAction(user: RdUserModel): AnAction {
  return object : AnAction(Supplier { CodeWithMePlatformBundle.message("action.full.sync") }, AllIcons.CodeWithMe.CwmFullSync), DumbAware {
    override fun actionPerformed(e: AnActionEvent) {
      this@createFullSyncAction.jumpTo(user)                      // Firstly jump
      this@createFullSyncAction.startFollowing(user)              // Then follow
      this@createFullSyncAction.requestFollowing(user, true) // Then ask for following us
    }

    override fun update(e: AnActionEvent) {
      val fullSyncWith = this@createFullSyncAction.isFullSyncWith(user)
      e.presentation.isEnabled = !fullSyncWith
    }
  }
}

fun UserManager.createStopFollowingAction(user: RdUserModel): AnAction {
  val userManager: UserManager = this@createStopFollowingAction

  val value = object : AnAction(), DumbAware {
    override fun actionPerformed(e: AnActionEvent) {
      userManager.stopFollowingOrSync()
    }

    override fun update(e: AnActionEvent) {
      val available = userManager.lead.value == user
      e.presentation.isEnabledAndVisible = available
      if (!available) return

      val isFullSync = userManager.isFullSyncWith(user)
      e.presentation.text = if (isFullSync)
        CodeWithMePlatformBundle.message("action.stop.fullSync")
      else
        CodeWithMePlatformBundle.message("action.stop.following")
    }

    public override fun setShortcutSet(shortcutSet: ShortcutSet) {
      super.setShortcutSet(shortcutSet)
    }
  }
  if (SingleUserFollowAction().isDefaultFollowee(user, userManager)) {
    val customShortcutSet = CustomShortcutSet.fromString("control shift alt Y")
    value.shortcutSet = customShortcutSet
  }
  return value
}