package com.jetbrains.rd.platform.codeWithMe.ui

import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.DialogWrapper
import com.intellij.openapi.util.NlsContexts
import com.intellij.openapi.util.NlsSafe
import com.intellij.ui.BrowserHyperlinkListener
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import icons.CodeWithMeCommonIcons
import net.miginfocom.swing.MigLayout
import org.jetbrains.annotations.Nls
import javax.swing.*
import javax.swing.text.html.HTMLEditorKit

class CWMDefaultDialog private constructor(project: Project?,
                                           builder: CWMDefaultDialogBuilder) : DialogWrapper(project) {
  companion object {
    val LOG = Logger.getInstance(CWMDefaultDialog::class.java)

    fun createBuilder(project: Project?, dialogTitle: @NlsContexts.DialogTitle String, message: @NlsSafe String, onAfterClose: (() -> Unit)? = null): CWMDefaultDialogBuilder {
      return CWMDefaultDialogBuilder(project, dialogTitle, message, onAfterClose)
    }

    fun multiLineLabelPane(@Nls txt: String) = JTextPane().apply {
      contentType = "text/html"
      isEditable = false
      background = null
      font = JBUI.Fonts.label()
      minimumSize = null
      border = JBUI.Borders.empty()

      addHyperlinkListener(BrowserHyperlinkListener.INSTANCE)
      putClientProperty(JEditorPane.HONOR_DISPLAY_PROPERTIES, true)

      val editorKit = HTMLEditorKit()
      val styleSheet = editorKit.styleSheet
      styleSheet.addRule("body { font-family: ${font.family}; }")
      styleSheet.addRule("p { margin: 4 0 4 0; }")
      setEditorKit(editorKit)

      text = txt
    }
  }

  private var okAction = builder.okAction
  private var cancelAction = builder.cancelAction
  private var header: String? = builder.header
  private var icon: Icon = builder.icon ?: CodeWithMeCommonIcons.CodeWithMeGuest_64

  private val actions = mutableListOf<Action>()

  private val logoPane = JPanel(MigLayout("ins 0, novisualpadding, gap 0")).apply {
    add(JLabel(icon))
    border = JBUI.Borders.empty(3, 8, 0, 18)
  }

  private val onAfterClose = builder.onAfterClose
  private val message = builder.message

  init {
    setResizable(false)
    setUndecorated(false)
    title = builder.dialogTitle
    isModal = builder.builderIsModal

    okAction?.let {
      it.name?.let {
        myOKAction.putValue(Action.NAME, it)
      }

      actions.add(myOKAction)
    }

    cancelAction?.let {
      it.name?.let {
        myCancelAction.putValue(Action.NAME, it)
      }

      actions.add(myCancelAction)
    }

    if(LOG.isTraceEnabled) {
      LOG.trace("$title ok: ${builder.okAction?.name}|${myOKAction.getValue(Action.NAME)} cancel: ${builder.cancelAction?.name}|${
        myCancelAction.getValue(Action.NAME)
      }")
    }

    init()
  }

  override fun doOKAction() {
    okAction?.action?.let {
      it()
    }
    super.doOKAction()
  }

  override fun doCancelAction() {
    cancelAction?.action?.let {
      it()
    }
    super.doCancelAction()
  }

  override fun dispose() {
    super.dispose()
    onAfterClose?.let { it() }
  }

  override fun createCenterPanel(): JComponent {
    val rightPanel = JPanel(MigLayout("fillx, ins 0 0 0 0, gapy 8, novisualpadding, flowy")).apply {
      header?.let {
        add(multiLineLabelPane(CodeWithMePlatformBundle.message("bold.text", it)), "growx, wmax 475, wmin 250")
      }
      add(multiLineLabelPane(message), "pushx, wmax 475, wmin 250")

      foreground = UIUtil.getContextHelpForeground()
      border = JBUI.Borders.empty(0, 0, 15, 10)
    }

    return JPanel(MigLayout("fillx, ins 10 0 0 0, novisualpadding, gap 0, w pref", "[pref!][grow]")).apply {
      add(logoPane, "ay top")
      add(rightPanel, "growx")
    }
  }

  override fun createActions(): Array<Action> {
    return actions.toTypedArray()
  }

  internal data class DialogAction(val action: (() -> Unit)?, val name: @NlsSafe String? = null)

  class CWMDefaultDialogBuilder(val project: Project?,
                                val dialogTitle: @NlsContexts.DialogTitle String,
                                val message: @NlsSafe String,
                                val onAfterClose: (() -> Unit)?) {
    internal var header: @NlsSafe String? = null

    internal var okAction: DialogAction? = null
    internal var cancelAction: DialogAction? = null

    internal var icon: Icon? = null

    internal var builderIsModal = false

    fun withModalFlag(value: Boolean): CWMDefaultDialogBuilder {
      this.builderIsModal = value
      return this
    }

    fun withHeader(value: @NlsSafe String): CWMDefaultDialogBuilder {
      this.header = value
      return this
    }

    fun withIcon(icon: Icon): CWMDefaultDialogBuilder {
      this.icon = icon
      return this
    }

    fun withOk(action: () -> Unit, name: @NlsSafe String? = null): CWMDefaultDialogBuilder {
      okAction = DialogAction(action, name)
      return this
    }

    fun withCancel(action: (() -> Unit)? = null, name: @NlsSafe String? = null): CWMDefaultDialogBuilder {
      cancelAction = DialogAction(action, name)
      return this
    }

    fun build(): DialogWrapper {
      return CWMDefaultDialog(project, this)
    }
  }
}

