// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.ui

import com.intellij.ide.ui.fullRow
import com.intellij.openapi.components.Service
import com.intellij.openapi.project.Project
import com.intellij.ui.components.JBScrollPane
import com.intellij.ui.layout.*
import com.intellij.ui.table.JBTable
import com.intellij.util.concurrency.EdtExecutorService
import com.intellij.util.containers.ContainerUtil
import com.intellij.util.ui.JBUI
import com.jetbrains.rd.framework.RdId
import com.jetbrains.rd.platform.codeWithMe.protocol.SecureWireOverStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.WireStatsCounter
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.swing.visibleProperty
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.Signal
import java.awt.BorderLayout
import java.awt.Dialog
import java.util.concurrent.TimeUnit
import javax.swing.JFrame
import javax.swing.JTable
import javax.swing.table.AbstractTableModel

@Service
class WireStatsRegistryService(val project: Project) {
  val registeredWires = ContainerUtil.createConcurrentList<SecureWireOverStreamTransport>()

  /**
   * Fired on UI thread when wires change
   */
  val wiresChanged = Signal<Unit>()

  fun registerWire(lifetime: Lifetime, wire: SecureWireOverStreamTransport) {
    registeredWires.addUnique(lifetime, wire)
    application.invokeLater { wiresChanged.fire(Unit) }
  }

  private var hasFrameShown = false
  fun showUi() {
    if (hasFrameShown) return
    val frame = JFrame("Protocol transmission stats")

    frame.defaultCloseOperation = JFrame.HIDE_ON_CLOSE
    frame.modalExclusionType = Dialog.ModalExclusionType.NO_EXCLUDE

    val lifetimeDef = project.lifetime.createNested()
    lifetimeDef.lifetime.onTermination {
      frame.isVisible = false
      hasFrameShown = false
      frame.dispose()
    }
    frame.visibleProperty().change.advise(lifetimeDef.lifetime) { if(!it) lifetimeDef.terminate() }

    val tableModel = WireStatsTableModel()
    val table = JBTable(tableModel)
    table.autoCreateColumnsFromModel = true
    table.autoCreateRowSorter = true
    table.setShowColumns(true)
    val panel = panel(LCFlags.fill) {
      fullRow { checkBox("Stay on top") { _, cb -> frame.isAlwaysOnTop = cb.isSelected }.constraints(CCFlags.growX) }
      fullRow { component(JBScrollPane(table).apply { minimumSize = JBUI.size(300, 300) }).constraints(CCFlags.grow) }
      fullRow {
        button("Show protocol tree stats for selected wire") {
          val selected = table.selectedRow
          if (selected >= 0) {
            val wire = registeredWires.drop(selected).first()
            showWireFrame(wire)
          }
        }.constraints(CCFlags.growX)
      }
    }

    frame.contentPane.add(panel, BorderLayout.CENTER)

    wiresChanged.advise(lifetimeDef.lifetime) {
      table.withPreserveSelection {
        tableModel.fireTableDataChanged()
      }
    }

    val future = EdtExecutorService.getScheduledExecutorInstance().scheduleWithFixedDelay({
      table.withPreserveSelection {
        tableModel.fireTableDataChanged()
      }
    }, 1, 1, TimeUnit.SECONDS)
    lifetimeDef.lifetime.onTermination {
      future.cancel(false)
    }

    frame.pack()
    frame.isVisible = true
    hasFrameShown = true
  }

  fun showWireFrame(wire: SecureWireOverStreamTransport) {
    val frame = JFrame("Protocol stats for ${wire.id}")

    frame.defaultCloseOperation = JFrame.HIDE_ON_CLOSE
    frame.modalExclusionType = Dialog.ModalExclusionType.NO_EXCLUDE

    val lifetimeDef = project.lifetime.createNested()
    lifetimeDef.lifetime.onTermination {
      frame.isVisible = false
      hasFrameShown = false
      frame.dispose()
    }
    frame.visibleProperty().change.advise(lifetimeDef.lifetime) { if(!it) lifetimeDef.terminate() }

    val table = JBTable()
    val tableModel = ProtocolEntitiesStatsModel(lifetimeDef.lifetime, wire.wireStats, table)
    table.model = tableModel
    table.autoCreateColumnsFromModel = true
    table.autoCreateRowSorter = true
    table.setShowColumns(true)
    val panel = panel(LCFlags.fill) {
      fullRow { checkBox("Stay on top") { _, cb -> frame.isAlwaysOnTop = cb.isSelected }.constraints(CCFlags.growX) }
      fullRow { component(JBScrollPane(table).apply { minimumSize = JBUI.size(300, 300) }).constraints(CCFlags.grow) }
    }

    frame.contentPane.add(panel, BorderLayout.CENTER)

    frame.pack()
    frame.isVisible = true
  }

  companion object {
    private inline fun JTable.withPreserveSelection(block: () -> Unit) {
      val oldSelection = this.selectedRow
      block()
      if (oldSelection >= 0 && oldSelection < this.rowCount)
        this.changeSelection(oldSelection, 0, false, false)
    }
  }

  private class ProtocolEntitiesStatsModel(val lifetime: Lifetime, val wireStats: WireStatsCounter, val table: JTable): AbstractTableModel() {
    val rdIdList = ArrayList<RdId>()

    init {
      wireStats.statsMode = WireStatsCounter.StatsMode.PerEntity
      val future = EdtExecutorService.getScheduledExecutorInstance().scheduleWithFixedDelay({
        rdIdList.clear()
        rdIdList.addAll(wireStats.perEntityStats.keys.filter { wireStats.storedNames[it]?.contains("DiagnosticsModel") != true })
        table.withPreserveSelection {
          fireTableDataChanged()
        }
      }, 1, 1, TimeUnit.SECONDS)
      lifetime.onTermination {
        future.cancel(false)
        wireStats.statsMode = WireStatsCounter.StatsMode.Simple
      }
    }

    override fun getRowCount(): Int {
      return rdIdList.size
    }

    override fun getColumnName(column: Int): String {
      return when(column) {
        0 -> "Entity"
        1 -> "Speed down"
        2 -> "Speed up"
        3 -> "Total down"
        4 -> "Total up"
        else -> ""
      }
    }

    override fun getColumnClass(columnIndex: Int): Class<*> {
      return when(columnIndex) {
        0 -> String::class.java
        1, 2, 3, 4-> java.lang.Long::class.java
        else -> Any::class.java
      }
    }

    override fun getColumnCount(): Int {
      return 5
    }

    val filenameRegex = Regex("""filePath\s*=\s*"[^"]*[/\\]([^"\\/]+)"""")
    override fun getValueAt(rowIndex: Int, columnIndex: Int): Any? {
      if (rowIndex >= rdIdList.size) return null
      val rdid = rdIdList[rowIndex]
      val wire = wireStats.perEntityStats[rdid] ?: return "-"
      return when(columnIndex) {
        0 -> wireStats.storedNames[rdid]?.replace(filenameRegex) { "fileName=\"${it.groupValues[1]}\""} ?: rdid.toString()
        1 -> wire.receiveSpeed
        2 -> wire.sendSpeed
        3 -> wire.totalReceive.get()
        4 -> wire.totalSend.get()
        else -> null
      }
    }
  }

  private inner class WireStatsTableModel : AbstractTableModel() {
    override fun getRowCount(): Int {
      return registeredWires.size
    }

    override fun getColumnCount(): Int {
      return 5
    }

    override fun getColumnName(column: Int): String {
      return when(column) {
        0 -> "Wire"
        1 -> "Speed down"
        2 -> "Speed up"
        3 -> "Total down"
        4 -> "Total up"
        else -> ""
      }
    }

    override fun getColumnClass(columnIndex: Int): Class<*> {
      return when(columnIndex) {
        0 -> String::class.java
        1, 2, 3, 4-> java.lang.Long::class.java
        else -> Any::class.java
      }
    }

    override fun getValueAt(rowIndex: Int, columnIndex: Int): Any? {
      val wire = registeredWires[rowIndex]
      val totalStats = wire.wireStats.totalStats
      return when(columnIndex) {
        0 -> wire.id
        1 -> totalStats.receiveSpeed
        2 -> totalStats.sendSpeed
        3 -> totalStats.totalReceive.get()
        4 -> totalStats.totalSend.get()
        else -> null
      }
    }
  }
}