// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.ui

import com.intellij.openapi.components.service
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.text.Formats.formatFileSize
import com.intellij.openapi.wm.CustomStatusBarWidget
import com.intellij.openapi.wm.StatusBar
import com.intellij.openapi.wm.StatusBarWidget
import com.intellij.openapi.wm.StatusBarWidgetFactory
import com.intellij.openapi.wm.impl.status.TextPanel
import com.intellij.util.concurrency.EdtExecutorService
import com.jetbrains.rd.platform.codeWithMe.protocol.RetryingConnectStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.SocketStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.StreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.WebSocketStreamTransport
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicStreamTransport
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.swing.mouseClicked
import java.util.concurrent.Future
import java.util.concurrent.TimeUnit
import javax.swing.JComponent

class WireStatsStatusBarWidget(val project: Project) : TextPanel(), CustomStatusBarWidget {
  private val myLifetimeDef = project.lifetime.createNested()
  private var myFuture: Future<*>? = null

  override fun ID() = WireStatsWidgetFactory.ID
  override fun getComponent(): JComponent = this

  init {
    setTextAlignment(1f)
  }

  override fun install(statusBar: StatusBar) {
    myFuture = EdtExecutorService.getScheduledExecutorInstance().scheduleWithFixedDelay(this::update, 1, 1, TimeUnit.SECONDS)
    mouseClicked().advise(myLifetimeDef) {
      project.service<WireStatsRegistryService>().showUi()
    }
  }

  private fun update() {
    val service = project.service<WireStatsRegistryService>()
    val wireCount = service.registeredWires.size
    if (wireCount == 0) {
      text = ""
      isVisible = false
      return
    }
    isVisible = true
    val totalSend = service.registeredWires.sumByDouble { it.wireStats.totalStats.sendSpeed.toDouble() }.toLong()
    val totalReceive = service.registeredWires.sumByDouble { it.wireStats.totalStats.receiveSpeed.toDouble() }.toLong()
    if (wireCount == 1) {
      text = "(${getTransportTypeLetter(service.registeredWires[0].streamTransport)}) ↑${formatFileSize(totalSend)}/s / ↓${formatFileSize(totalReceive)}/s"
    } else {
      text = "($wireCount) ↑${formatFileSize(totalSend)}/s / ↓${formatFileSize(totalReceive)}/s"
    }
  }

  private fun getTransportTypeLetter(transport: StreamTransport): String {
    return when(transport) {
      is SocketStreamTransport -> "D"
      is QuicStreamTransport -> "P" // P2P
      is RetryingConnectStreamTransport -> transport.tryGetUnderlying()?.let { getTransportTypeLetter(it) } ?: "?"
      is WebSocketStreamTransport -> "R"
      else -> "?"
    }
  }

  override fun getTextForPreferredSize(): String {
    val service = project.service<WireStatsRegistryService>()
    val wireCount = service.registeredWires.size
    return when(wireCount) {
      0 -> ""
      1 -> "(D) ↑999.99 MB/s / ↓999.99 MB/s"
      else -> "(9) ↑999.99 MB/s / ↓999.99 MB/s"
    }
  }

  override fun dispose() {
    myLifetimeDef.terminate()
    myFuture?.cancel(true)
  }
}

class WireStatsWidgetFactory : StatusBarWidgetFactory {
  companion object {
    const val ID = "CwmWireStatsWidgetFactory"
  }

  override fun getId(): String {
    return ID
  }

  override fun getDisplayName() = "Code With Me Wire stats"

  override fun isAvailable(project: Project): Boolean {
    return true
  }

  override fun createWidget(project: Project): StatusBarWidget {
    return WireStatsStatusBarWidget(project)
  }

  override fun disposeWidget(widget: StatusBarWidget) {
    Disposer.dispose(widget)
  }

  override fun canBeEnabledOn(statusBar: StatusBar): Boolean {
    return true
  }
}
