package com.jetbrains.rd.platform.codeWithMe.users

import com.intellij.ide.util.PropertiesComponent
import com.intellij.openapi.util.SystemInfo
import com.intellij.util.EnvironmentUtil
import com.intellij.util.system.CpuArch
import com.jetbrains.rd.platform.util.getLogger
import com.sun.jna.platform.win32.*
import kotlinx.serialization.json.Json
import kotlinx.serialization.json.JsonObject
import kotlinx.serialization.json.JsonPrimitive
import java.io.File
import java.net.URI

object LobbyUrlSettings {
  val logger = getLogger<LobbyUrlSettings>()

  class LobbyUrlHolder(val uri: URI, val osOriginLocation: String?) {
    val isSetFromOs = osOriginLocation != null
  }

  private const val lobbyUrlKey = "code.with.me.lobby.url"
  val lobbyUrl: LobbyUrlHolder
    get() {
      val externalValue = getFromOs()
      val settingValue =  PropertiesComponent.getInstance().getValue(lobbyUrlKey)?.let { URI(it.trimEnd('/')) }
      return LobbyUrlHolder(externalValue?.uri ?: settingValue ?: DEFAULT_LOBBY_URL, externalValue?.osOriginLocation)
    }


  // Common values
  //   https://code-with-me.jetbrains.com
  //   https://codewithme-lobby-staging.api.jetbrains.com
  //   http://127.0.0.1:9980
  val DEFAULT_LOBBY_URL get() = URI("https://code-with-me.jetbrains.com")

  fun setLobbyUrl(uri: URI) {
    val current = lobbyUrl
    if (uri == current.uri) return
    if (current.isSetFromOs)
      error("Lobby URL is not editable.")
    else
      PropertiesComponent.getInstance().setValue(lobbyUrlKey, uri.toString())
  }

  private fun getFromOs(): LobbyUrlHolder? = when {
    SystemInfo.isWindows -> getFromRegistry()
    SystemInfo.isLinux -> getFromXdgConfig()
    SystemInfo.isMac -> getFromLibraryApplicationSupport()
    else -> error("Unknown OS")
  }

  // key: SOFTWARE\JetBrains\CodeWithMe, value: lobbyServerUrl, value type: REG_SZ
  // prio: HKLM64 -> 32 -> HKCU64 -> 32
  private fun getFromRegistry(): LobbyUrlHolder? {
    val key = "SOFTWARE\\JetBrains\\CodeWithMe"
    val value = "lobbyServerUrl"

    logger.info("Looking for lobby url in registry, is32Bit=${CpuArch.isIntel32()}")

    // these WOW64 keys are ignored by 32-bit Windows, see https://docs.microsoft.com/en-us/windows/win32/sysinfo/registry-key-security-and-access-rights
    val uris = listOf(
      "HKLM_64" to { Advapi32Util.registryGetStringValue(WinReg.HKEY_LOCAL_MACHINE, key, value, WinNT.KEY_WOW64_64KEY) },
      "HKLM_32" to { Advapi32Util.registryGetStringValue(WinReg.HKEY_LOCAL_MACHINE, key, value, WinNT.KEY_WOW64_32KEY) },

      "HKCU_64" to { Advapi32Util.registryGetStringValue(WinReg.HKEY_CURRENT_USER, key, value, WinNT.KEY_WOW64_64KEY) },
      "HKCU_32" to { Advapi32Util.registryGetStringValue(WinReg.HKEY_CURRENT_USER, key, value, WinNT.KEY_WOW64_32KEY) }
    )
    fun Pair<String, () -> String>.readable() = "${this.first}\\$key\\$value"

    return uris.mapNotNull {
      val uri = try {
        logger.info("Searching for lobby url at ${it.readable()}")
        it.second()
      }
      catch (t: Throwable) {
        when (t) {
          is Win32Exception -> {
            if (t.errorCode == WinError.ERROR_FILE_NOT_FOUND) logger.info("Lobby url registry entry not found at ${it.readable()}")
            else logger.warn("Failed to get lobby url registry entry at ${it.readable()}, HRESULT=${t.hr.toInt()}", t)
          }
          else -> {
            logger.warn("Failed to get lobby url registry entry at ${it.readable()}", t)
          }
        }
        null
      }

      if (uri != null) {
        logger.info("Found lobby url registry entry at ${it.readable()}, value=$uri")
        LobbyUrlHolder(URI(uri), it.readable())
      }
      else {
        null
      }
    }.firstOrNull()
  }

  // https://specifications.freedesktop.org/basedir-spec/basedir-spec-0.6.html
  private fun getFromXdgConfig(): LobbyUrlHolder? {
    val configPath = "JetBrains/CodeWithMe/config.json"
    val env = EnvironmentUtil.getEnvironmentMap()
    val home = System.getProperty("user.home")

    // non-user writable location first, even if that's wrong according to spec as that's what we want
    val configLookupDirs = listOf("/etc/xdg/", env["XDG_CONFIG_HOME"] ?: "$home/.config/").toMutableList()

    // we already set /etc/xdg as the first one, so no need to default to it
    val xdgConfigDirs = (env["XDG_CONFIG_DIRS"] ?: "").split(":").filter { it.isNotEmpty() }
    configLookupDirs.addAll(xdgConfigDirs)

    return getLobbyUrlFromDirectories(configLookupDirs.map { File(it) }, configPath)
  }

  private fun getLobbyUrlFromDirectories(dirs: List<File>, configPath: String) : LobbyUrlHolder? =
    dirs.mapNotNull {
      val file = File(it, configPath)
      getLobbyUrlFromJsonFile(file)
    }.firstOrNull()

  private fun getLobbyUrlFromJsonFile(file: File): LobbyUrlHolder? {
    logger.info("Trying to get lobby url from file=${file.canonicalPath}")

    if (!file.exists()) {
      logger.info("File=${file.canonicalPath} does not exist.")
      return null
    }
    return try {
      val root = Json.Default.parseToJsonElement(file.readText()) as? JsonObject
      val uri = (root?.get("lobbyServerUrl") as? JsonPrimitive)?.content
      if (uri != null) {
        logger.info("Found lobby url in file=${file.canonicalPath}, value=$uri")
        LobbyUrlHolder(URI(uri), file.canonicalPath)
      } else null
    }
    catch (t: Throwable) {
      logger.warn("Failed to read json for lobby server url at location $file", t)
      null
    }
  }

  private fun getFromLibraryApplicationSupport(): LobbyUrlHolder? {
    val home = System.getProperty("user.home")
    val dirs = listOf("/", "$home/")
    val configPath = "Library/Application Support/JetBrains/CodeWithMe/config.json"
    return getLobbyUrlFromDirectories(dirs.map { File(it) }, configPath)
  }
}
