package com.jetbrains.rd.platform.codeWithMe.users

import com.intellij.openapi.Disposable
import com.intellij.openapi.editor.impl.EditorComponentImpl
import com.intellij.openapi.project.Project
import com.intellij.openapi.wm.ToolWindowManager
import com.intellij.openapi.wm.ex.ToolWindowManagerListener
import com.jetbrains.rd.util.reactive.IPropertyView
import com.jetbrains.rd.util.reactive.Property
import java.awt.KeyboardFocusManager
import java.beans.PropertyChangeListener


abstract class UserFocusManager(protected val project: Project) : Disposable {

  companion object {
    fun getInstance(project: Project): UserFocusManager = project.getService(UserFocusManager::class.java)
  }

  protected val leadActiveToolWindowIdProperty = Property<String?>(null)
  protected val currentActiveToolWindowIdProperty = Property<String?>(null)
  private val focusPropertyChangeListener: PropertyChangeListener

  /**
   * Provides lead's (the person we are following now) active tool window ID
   * If there is no following, returns null
   */
  val leadActiveToolWindowId: IPropertyView<String?>
    get() = leadActiveToolWindowIdProperty

  /**
   * Provides current active tool window ID
   */
  val currentActiveToolWindowId: IPropertyView<String?>
    get() = currentActiveToolWindowIdProperty

  init {
    project.messageBus.connect(project).subscribe(
      ToolWindowManagerListener.TOPIC,
      object : ToolWindowManagerListener {
        override fun stateChanged(toolWindowManager: ToolWindowManager) {
          fireCurrentActiveToolWindowChanged(toolWindowManager.activeToolWindowId)
        }
      })

    focusPropertyChangeListener = PropertyChangeListener { evt ->
      val focused = evt.newValue
      if (focused is EditorComponentImpl) {
        fireCurrentActiveToolWindowChanged(null)
      }
    }

    KeyboardFocusManager.getCurrentKeyboardFocusManager().addPropertyChangeListener("focusOwner", focusPropertyChangeListener)
    KeyboardFocusManager.getCurrentKeyboardFocusManager().addPropertyChangeListener("permanentFocusOwner", focusPropertyChangeListener)
  }

  override fun dispose() {
    KeyboardFocusManager.getCurrentKeyboardFocusManager().removePropertyChangeListener("focusOwner", focusPropertyChangeListener)
    KeyboardFocusManager.getCurrentKeyboardFocusManager().removePropertyChangeListener("permanentFocusOwner", focusPropertyChangeListener)
  }

  protected abstract fun currentActiveToolWindowChanged(toolWindowId: String?)

  fun fireCurrentActiveToolWindowChanged(toolWindowId: String?) {
    stopFollowingIfNeeded(toolWindowId)

    currentActiveToolWindowIdProperty.set(toolWindowId)
    currentActiveToolWindowChanged(toolWindowId)
  }

  private fun stopFollowingIfNeeded(toolWindowId: String?) {
    val userManager = UserManager.getInstance(project)
    if (userManager.isInFullSync()) return

    val leadToolWindowId = leadActiveToolWindowId.value ?: return
    val activeToolWindowId = toolWindowId ?: return
    if (leadToolWindowId == activeToolWindowId) return

    val toolWindowManager = ToolWindowManager.getInstance(project)
    val leadToolWindow = toolWindowManager.getToolWindow(leadToolWindowId) ?: return
    val activeToolWindow = toolWindowManager.getToolWindow(activeToolWindowId) ?: return

    val shouldBreakFollowing = leadToolWindow.anchor == activeToolWindow.anchor
    if (shouldBreakFollowing) {
      userManager.stopFollowing()
    }
  }
}