// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.codeWithMe.users

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.project.Project
import com.jetbrains.rd.platform.codeWithMe.toolbar.createFollowAction
import com.jetbrains.rd.platform.codeWithMe.toolbar.createFullSyncAction
import com.jetbrains.rd.platform.codeWithMe.toolbar.createJumpToAction
import com.jetbrains.rd.platform.codeWithMe.toolbar.createStopFollowingAction
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.*
import com.jetbrains.codeWithMe.model.RdUserModel

/**
 * Contains information about connected users and provides user-related operations for the current (local) user
 * To control remove users use FollowMeManager on backend
 */
abstract class UserManager(val project: Project) {

  companion object {
    private val logger = getLogger<UserManager>()

    fun getInstance(project: Project): UserManager {
      return project.getService(UserManager::class.java)
    }
  }

  protected val lifetime = project.lifetime
  protected val followersList = ViewableList<RdUserModel>()
  protected val leadProperty = Property<RdUserModel?>(null)
  protected val sessionEnabledProperty = Property<Boolean>(false)

  private val usersList = ViewableList<RdUserModel>()
  private val meProperty = Property<RdUserModel?>(null)

  protected fun registerUser(lifetime: Lifetime, user: RdUserModel) {
    if (user.isMe) {
      logger.debug("Me: ${user.id} : ${user.name}")
      meProperty.setValue(lifetime, user)
    } else {
      logger.debug("New user: ${user.id} : ${user.name}")
      usersList.addUnique(lifetime, user)
    }
  }

  /**
   * All users connected to the same backend at the moment, current [this/me] user is not presented in this list
   */
  val users: IViewableList<RdUserModel> = usersList

  /**
   * User model for me (current user)
   */
  val me: IPropertyView<RdUserModel?> = meProperty

  /**
   * Users which follows current client
   */
  val followers: IViewableList<RdUserModel> = followersList

  /**
   * User which current client follows, can be null which means the client does not follow anyone
   */
  val lead: IPropertyView<RdUserModel?> = leadProperty

  var recentLead: IPropertyView<RdUserModel?> = Property<RdUserModel?>(null)

  /**
   * True if session link is available
   */
  val sessionEnabled: IPropertyView<Boolean> = sessionEnabledProperty

  /**
   * Navigates current user to the specified user
   */
  abstract fun jumpTo(user: RdUserModel)

  /**
   * Start to follow specified user
   */
  abstract fun startFollowing(user: RdUserModel)

  /**
   * Stop following current user -> other user if following exists
   * Terminates full sync relation if exists
   */
  fun stopFollowingOrSync() {
    // Stop following us if we are in FullSync mode
    val userWeAreFollowingNow = lead.value
    if (userWeAreFollowingNow != null && isFullSyncWith(userWeAreFollowingNow)) {
      requestStopFollowing(userWeAreFollowingNow)
    }
    stopFollowing()
  }

  /**
   * Stop following some user if following exists
   */
  abstract fun stopFollowing()

  /**
   * Send request for specified user to follow us
   */
  abstract fun requestFollowing(user: RdUserModel, force: Boolean)

  /**
   * Send request for specified user to stop following us
   */
  abstract fun requestStopFollowing(user: RdUserModel)

  /**
   * Returns true if we are in full sync with specified user
   */
  fun isFullSyncWith(user: RdUserModel): Boolean {
    return lead.value == user && followers.contains(user)
  }

  open fun createUserActions(user: RdUserModel) : List<AnAction> {
    return listOf(
      this.createJumpToAction(user),
      this.createFollowAction(user),
      this.createFullSyncAction(user),
      this.createStopFollowingAction(user))
  }

  fun hasRecentLead(users: IViewableList<RdUserModel>): Boolean {
    if (recentLead.value == null) return false
    val recentClientId: String? = recentLead.value?.clientIdValue
    return users.stream().anyMatch { user -> user.clientIdValue == recentClientId }
  }

  fun getCWMActionName(): String {
    return if (me.value?.isHost == true) {
      "CWMHostShowPopupAction"
    }
    else {
      "CWMThinClientShowPopupAction"
    }
  }
}

fun UserManager.getHost(): RdUserModel? {
  return users.find { it.isHost }
}

fun UserManager.isInFullSync(): Boolean {
  return users.any { isFullSyncWith(it) }
}

val RdUserModel.isMe: Boolean
  get() = clientId == ClientId.localId

val RdUserModel.isHost: Boolean
  get() = clientId == ClientId.defaultLocalId

val RdUserModel.clientId: ClientId
  get() = ClientId(clientIdValue)

fun broadcastToAllGuests(project: Project, actionBlock: () -> Unit) {
  UserManager.getInstance(project).users.forEach {
    ClientId.withClientId(it.clientId) {
      actionBlock()
    }
  }
}