package com.jetbrains.rd.platform.codeWithMe.users

import com.intellij.codeWithMe.ClientId
import com.intellij.icons.AllIcons
import com.intellij.notification.Notification
import com.intellij.notification.NotificationListener
import com.intellij.notification.NotificationType
import com.intellij.notification.impl.NotificationsManagerImpl
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.MessageType
import com.intellij.openapi.ui.popup.Balloon
import com.intellij.openapi.ui.popup.JBPopupFactory
import com.intellij.openapi.util.NlsContexts
import com.intellij.openapi.util.Ref
import com.intellij.openapi.util.text.StringUtil
import com.intellij.openapi.wm.IdeFrame
import com.intellij.openapi.wm.WindowManager
import com.intellij.ui.BalloonLayoutData
import com.intellij.ui.GotItTooltip
import com.intellij.ui.awt.RelativePoint
import com.intellij.util.concurrency.EdtScheduledExecutorService
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.CwmNotifications.CWM_NOTIFICATION_GROUP_ID
import com.jetbrains.rd.platform.codeWithMe.toolbar.CWMShowPopupAction
import com.jetbrains.rd.platform.util.applicationEx
import java.awt.Point
import java.util.concurrent.TimeUnit
import javax.swing.JComponent
import javax.swing.event.HyperlinkEvent

object UserNotifications {
  fun askForFollowOtherUser(project: Project, user: RdUserModel, force: Boolean, onAccepted: () -> Unit) {

    if (UserManager.getInstance(project).lead.value == user) {
      // Already follow
      return
    }

    if (force) {
      showInfoNotification(project,
                           CodeWithMePlatformBundle.message("notification.following.started.title"),
                           CodeWithMePlatformBundle.message("notification.following.started", StringUtil.escapeXmlEntities(user.name)),
                           true)
      onAccepted()
      return
    }

    val notification = Notification(
      CWM_NOTIFICATION_GROUP_ID,
      CodeWithMePlatformBundle.message("notification.following.request.title"),
      CodeWithMePlatformBundle.message("notification.following.request", StringUtil.escapeXmlEntities(user.name),
                                       "<br><a href=\"/accept\">Accept</a> <a href=\"/decline\">Decline</a>"),
      NotificationType.INFORMATION,
      object : NotificationListener.Adapter() {
        override fun hyperlinkActivated(notificationFromListener: Notification, e: HyperlinkEvent) {
          notificationFromListener.hideBalloon()
          notificationFromListener.expire()
          if (e.description == "/accept") {
            onAccepted()
          }
        }
      })
    notification.icon = AllIcons.CodeWithMe.CwmAccess
    showNotification(project, notification, -1)
  }

  @Suppress("HardCodedStringLiteral")
  fun showClickableLinkNotification(project: Project, @NlsContexts.NotificationTitle title: String, @NlsContexts.NotificationTitle clickableMessage: String, autoHidePeriod: Int = 10, onLinkClick: () -> Unit) {
    val notification = Notification(
      CWM_NOTIFICATION_GROUP_ID,
      title,
      "<a href=\"/accept\">${StringUtil.escapeXmlEntities(clickableMessage)}</a>",
      NotificationType.INFORMATION,
      object : NotificationListener.Adapter() {
        override fun hyperlinkActivated(notificationFromListener: Notification, e: HyperlinkEvent) {
          notificationFromListener.hideBalloon()
          notificationFromListener.expire()
          if (e.description == "/accept") {
            onLinkClick()
          }
        }
      })
    showNotification(project, notification, autoHidePeriod)
  }

  fun showFollowingRetargetNotification(project: Project, fromUser: String, toUser: String) {
    showInfoNotification(project,
                         CodeWithMePlatformBundle.message("notification.following.user.changed.title"),
                         CodeWithMePlatformBundle.message("notification.following.user.changed", fromUser, toUser),
                         true)
  }

  fun showInfoNotification(
    project: Project,
    @NlsContexts.NotificationTitle notificationTitle: String,
    @NlsContexts.NotificationContent notificationText: String,
    autoHide: Boolean
  ) {
    val notification = Notification(
      CWM_NOTIFICATION_GROUP_ID,
      notificationTitle,
      notificationText,
      NotificationType.INFORMATION)
    notification.icon = AllIcons.CodeWithMe.CwmAccess
    showNotification(project, notification)
  }

  fun showGreenNotification(
    project: Project,
    @NlsContexts.PopupContent message: String,
    @NlsContexts.NotificationTitle title: String? = null
  ) {
    if (applicationEx.isHeadlessEnvironment) return

    val notification = Notification(
      CWM_NOTIFICATION_GROUP_ID,
      AllIcons.CodeWithMe.CwmAccess,
      title,
      null,
      message,
      NotificationType.INFORMATION, null)

    showNotification(project, notification, true) { component, _ ->
      val balloon = JBPopupFactory.getInstance()
        .createHtmlTextBalloonBuilder(
          message, notification.icon,
          MessageType.INFO.popupBackground, null)
        .setShadow(false)
        .setFadeoutTime(5000)
        .createBalloon()
      val relativePoint = RelativePoint(
        component,
        Point(component.width / 2, component.height))
      balloon.show(relativePoint, Balloon.Position.below)
    }
  }

  fun showGotItPopup(project: Project, id: String, @NlsContexts.NotificationTitle title: String? = null, @NlsContexts.NotificationContent text: String) {
    val notification = Notification(
      CWM_NOTIFICATION_GROUP_ID,
      AllIcons.CodeWithMe.CwmAccess,
      title,
      null,
      text,
      NotificationType.INFORMATION, null)

    showNotification(project, notification, false) { comp, _ ->
      val message = GotItTooltip(id, text, project).withPosition(Balloon.Position.below)
      title?.let {
        message.withHeader(it)
      }

      notification.icon?.let {
        message.withIcon(it)
      }

      message.show(comp, GotItTooltip.BOTTOM_MIDDLE)
    }
  }

  private fun showNotification(project: Project,
                               notification: Notification,
                               autoHide: Int = 5,
                               layoutData: Ref<BalloonLayoutData>? = null) {

    showNotification(project, notification) { component, frame ->
      val layout = layoutData ?: BalloonLayoutData.fullContent()
      val balloon = NotificationsManagerImpl.createBalloon(
        frame, notification, true, false,
        layout, project)

      val relativePoint = RelativePoint(
        component,
        Point(component.width / 2, component.height))
      balloon.show(relativePoint, Balloon.Position.below)

      if (autoHide >= 0) {
        EdtScheduledExecutorService.getInstance().schedule({ balloon.hide() }, autoHide.toLong(), TimeUnit.SECONDS)
      }
    }
  }

  private fun showNotification(project: Project,
                               notification: Notification,
                               notifyIfControlHidden: Boolean = true,
                               showBalloon: (JComponent, IdeFrame) -> Unit) {
    ClientId.withClientId(ClientId.localId) {
      project.getUserData(CWMShowPopupAction.key)?.let { component ->
        if (component.isShowing) {
          WindowManager.getInstance().getIdeFrame(project)?.let { frame ->
            if (frame.component.isAncestorOf(component)) {
              showBalloon(component, frame)
              return
            }
          }
        }

        if (notifyIfControlHidden) {
          notification.notify(project)
        }
      }
    }
  }
}