package com.jetbrains.rd.platform.codeWithMe.util

import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.io.DigestUtil
import java.io.File
import java.nio.charset.StandardCharsets
import java.nio.file.*
import java.nio.file.attribute.BasicFileAttributes
import kotlin.io.path.*

object FileManifestUtil {
  private val logger = Logger.getInstance(javaClass)
  const val ManifestFileName = ".manifest.txt"
  const val HashSeed = "e69b0a64-b91a-4da6-bc80-35828c9a97d1"

  fun generateDirectoryManifest(root: Path, filterPaths: (Path) -> Boolean): String {
    val list = mutableListOf<String>()

    Files.walkFileTree(root, object : SimpleFileVisitor<Path>() {
      override fun visitFile(file: Path, attrs: BasicFileAttributes): FileVisitResult {
        if (file.name != ManifestFileName && filterPaths(file)) {
          val relativeToRoot = file.relativeToOrNull(root) ?: error("walkFileTree: file '$file' must be relative to '$root'")
          list.add("${relativeToRoot.toString().replace('\\', '/')} ${attrs.size()} ${attrs.lastModifiedTime().toMillis() / 1000}")
        }

        return FileVisitResult.CONTINUE
      }

      override fun preVisitDirectory(dir: Path, attrs: BasicFileAttributes?): FileVisitResult {
        if (dir == root) return FileVisitResult.CONTINUE

        if (filterPaths(dir)) {
          val relativeToRoot = dir.relativeToOrNull(root) ?: error("walkFileTree: directory '$dir' must be relative to '$root'")
          list.add(relativeToRoot.toString().replace('\\', '/'))
          return FileVisitResult.CONTINUE
        } else {
          return FileVisitResult.SKIP_SUBTREE
        }
      }
    })

    val contentPart = list.sorted().joinToString("\n")
    // Prevent easy tampering
    return "${DigestUtil.sha256Hex((HashSeed + contentPart).toByteArray(StandardCharsets.UTF_8))}\n$contentPart"
  }

  fun writeDirectoryManifest(root: Path, filterPaths: (Path) -> Boolean = { true }) {
    root.resolve(ManifestFileName).writeText(generateDirectoryManifest(root, filterPaths), StandardCharsets.UTF_8)
  }

  fun isUpToDate(root: Path, filterPaths: (Path) -> Boolean = { true }): Boolean {
    val manifestFile = root.resolve(ManifestFileName)
    if (manifestFile.notExists()) {
      logger.info("isUpToDate false for '$root': manifest file $manifestFile does not exist")
      return false
    }

    val manifestFileContent = manifestFile.readText(StandardCharsets.UTF_8)
    val actualOnDiskContent = generateDirectoryManifest(root, filterPaths)

    if (manifestFileContent == actualOnDiskContent) {
      logger.info("isUpToDate true for '$root': manifest file $manifestFile contains actual information")
      return true
    }

    logger.info("isUpToDate false for '$root': manifest file $manifestFile differs from on disk content\n" +
                "on disk content:\n$actualOnDiskContent\n" +
                "saved manifest in $manifestFile content:\n$manifestFileContent")
    return false
  }

  data class ExtractDirectory(val path: Path, val isUpToDate: Boolean)

  /**
   * Get extract directory state (path to directory and if directory is up-to-date).
   *
   * @param path - base path to extract directory
   * @param filterPaths - filter files to check for up-to-date state
   */
  fun getExtractDirectory(path: Path, filterPaths: (Path) -> Boolean = { true }): ExtractDirectory {
    val suffix = ".${ProcessHandle.current().pid()}.${System.currentTimeMillis()}"

    val retriesCount = 100

    (1..retriesCount).forEach moveFolder@{ attempt ->
      val destinationPath = if (attempt <= 1) path else Path.of("$path-$attempt")

      val isUpToDate = isUpToDate(destinationPath, filterPaths)
      if (isUpToDate) {
        logger.info("All files inside extract directory '$destinationPath' are up-to-date")
        return ExtractDirectory(destinationPath, true)
      }

      if (!destinationPath.exists()) {
        logger.info("Destination does not exist, returning $destinationPath")
        return ExtractDirectory(destinationPath, false)
      }

      // Detect any locked files inside. If a directory was safe to rename, it should be safe to write to (or delete).
      val renamedPath = File("${destinationPath}${suffix}").toPath()
      try {
        // Try to move the whole base path into a temp directory.
        Files.move(destinationPath, renamedPath, StandardCopyOption.ATOMIC_MOVE)
      } catch (t: Throwable) {
        logger.debug(
          "Rename '$destinationPath' to '$renamedPath' has failed. Probably file is locked by another process. Trying the next one.")
        return@moveFolder
      } finally {
        try {
          renamedPath.deleteIfExists()
        }
        catch (t: Throwable) {
          logger.warn("Unable to delete renamed destination (generally should not happen): $renamedPath", t)
        }
      }

      logger.info("Destination was deleted, returning path '$destinationPath'")
      return ExtractDirectory(destinationPath, false)
    }

    throw IllegalStateException("Exceeded $retriesCount retries to get safe destination based on '$path'.")
  }
}