package com.jetbrains.rd.platform.diagnostics

import com.intellij.ide.util.PropertiesComponent
import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.util.SingleAlarm
import com.intellij.util.containers.ContainerUtil
import com.jetbrains.rd.platform.util.application
import org.apache.log4j.Level
import org.apache.log4j.LogManager
import java.util.*

class LogTraceScenariosRegistry {
  companion object {
    val HOLDER_EP_NAME = ExtensionPointName.create<Any>("com.intellij.rd.platform.traceScenarioHolder")
    val TRACE_CATEGORIES_LISTENER_EP_NAME = ExtensionPointName.create<LogTraceCategoriesListener>("com.intellij.rd.platform.logTraceCategoriesListener")

    fun getInstance(): LogTraceScenariosRegistry = application.getService(LogTraceScenariosRegistry::class.java)

    // slow init (300-400 ms)
    fun loadTraceScenarioHolders() {
      // load classes and trigger auto-registration
      val scenariosObjects = mutableListOf<Any>(*HOLDER_EP_NAME.extensions)

      for (holder in scenariosObjects) {
        holder::class.objectInstance
        holder::class.nestedClasses.map { it.objectInstance }
      }
    }

    // do not move to instance because it may be accessed when Application is not ready
    val allValues = ContainerUtil.createConcurrentList<LogTraceScenario>() // called from ui and background

    private const val TRACE_SCENARIOS_KEY = "rider.trace.scenarios"
    private const val delimiter = ","
  }


  private val syncAlarm = SingleAlarm(Runnable { synchronizeToLogger() }, 100)

  fun enabledScenarios() = PropertiesComponent.getInstance().getValue(TRACE_SCENARIOS_KEY, "")
    .split(delimiter).mapNotNull { name -> allValues.firstOrNull { it.name == name } }.sorted()

  fun save(enabledScenarios: List<String>) {
    val newEnabledScenarios = enabledScenarios.joinToString(delimiter)
    PropertiesComponent.getInstance().setValue(TRACE_SCENARIOS_KEY, newEnabledScenarios)

    synchronizeToLogger(true)
  }

  private val categoriesOldLevels = HashMap<String, Level>()

  fun synchronizeToLogger(immediately: Boolean = false) {
    if (immediately) {
      syncAlarm.cancelAllRequests()
      synchronizeToLoggerImmediately()
    } else syncAlarm.request()
  }

  private fun synchronizeToLoggerImmediately() {
    val enabledScenarios = enabledScenarios()
    val traceCategories = enabledScenarios.flatMap { it.categories }.distinct().sorted()

    //backend
    updateTraceCategories(traceCategories)

    //frontend, restore old
    for ((category, level) in categoriesOldLevels.entries) {
      LogManager.getLogger(category).level = level
    }
    categoriesOldLevels.clear()

    //frontend, set new & store old
    for (category in traceCategories) {
      val logger = LogManager.getLogger(category)
      categoriesOldLevels[category] = logger.effectiveLevel
      logger.level = Level.TRACE
    }
  }

  open fun updateTraceCategories(traceCategories: List<String>) {
    for (listener in TRACE_CATEGORIES_LISTENER_EP_NAME.extensions) {
      listener.traceCategoriesUpdated(traceCategories)
    }
    application.getService(LogTraceScenariosRegistrySpecial::class.java)?.traceCategoriesUpdated(traceCategories)
  }
}


object RdLogTraceScenarios {
  object Protocol : LogTraceScenario("protocol")

  object AltEnter : LogTraceScenario(
    "#com.jetbrains.rdclient.services.popups.FrontendIntentionMenuContributor",
    "#com.jetbrains.rdclient.services.popups.FrontendPopupModelHost"
  )

  object Completion : LogTraceScenario(
    "#com.jetbrains.rdclient.completion",
    "#com.jetbrains.rdclient.document"
  )

  object Documents : LogTraceScenario(
    "#com.jetbrains.rdclient.document"
  )

  object Editors : LogTraceScenario(
    "#com.jetbrains.rdclient.document",
    "#com.jetbrains.rdclient.editors",
    "#com.jetbrains.rdclient.editorActions",
    "#com.jetbrains.rdclient.template"
  )

  object Commands : LogTraceScenario(
    "#com.intellij.openapi.command.impl"
  )

  object ActionsAndPopups : LogTraceScenario(
    "#com.jetbrains.rdclient.actions",
    "#com.jetbrains.rdclient.services.UiContextManager"
  )

  object Daemon : LogTraceScenario(
    "#com.jetbrains.rdclient.daemon"
  )

  object BeControls : LogTraceScenario(
    "#com.jetbrains.rdclient.ui"
  )

  object ParameterInfo : LogTraceScenario(
    "#com.jetbrains.rdclient.parameterInfo"
  )

  object TrafficLightIcon : LogTraceScenario(
    "#com.jetbrains.rdclient.daemon.FrontendTrafficLightRenderer",
    "#com.jetbrains.rdclient.daemon.components.FrontendErrorStripeUpdateManager"
  )

  object TypeHierarchy : LogTraceScenario(
    "#com.jetbrains.rider.ideaInterop.vfs",
    "JetBrains.ProjectModel.Impl",
    "JetBrains.ReSharper.Feature.Services.Actions",
    "JetBrains.ReSharper.Features.Inspections.Hierarchies.TypeHierarchy",
    "JetBrains.ReSharper.Host.Features.Documents",
    "JetBrains.ReSharper.Host.Features.FileSystem",
    "JetBrains.ReSharper.Host.Features.TextControls",
    "JetBrains.ReSharper.Host.Features.TypeHierarchy",
    "JetBrains.ReSharper.Psi.Caches",
    "JetBrains.ReSharper.Psi.Files",
    "JetBrains.Rider.Test.Framework.Core.Documents"
  )

}