package com.jetbrains.rd.platform.diagnostics

import com.intellij.ide.DataManager
import com.intellij.ide.util.PropertiesComponent
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.ActionPlaces
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.Presentation
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.DialogWrapper
import com.intellij.ui.ColoredTableCellRenderer
import com.intellij.ui.TableSpeedSearch
import com.intellij.ui.components.JBScrollPane
import com.intellij.ui.scale.JBUIScale
import com.intellij.ui.speedSearch.SpeedSearchUtil
import com.intellij.ui.table.JBTable
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.rd.RdPlatformBundle
import java.awt.*
import java.awt.event.*
import javax.swing.*
import javax.swing.event.ListSelectionEvent
import javax.swing.table.AbstractTableModel
import javax.swing.table.JTableHeader
import javax.swing.table.TableCellEditor
import javax.swing.table.TableCellRenderer


class LogTraceScenariosDialog(val project: Project?) : DialogWrapper(project) {
  companion object {
    const val TOUCHED = "log.trace.scenarios.preset.1"
    internal const val dialogTitle = "Choose Trace Scenarios"
    private val columnNames = arrayOf("Scenario", "")
    private val defaultSize = Dimension(JBUI.scale(200), JBUI.scale(500))
  }

  private val enabledScenarios = LogTraceScenariosRegistry.getInstance().enabledScenarios()
  val scenarios = LogTraceScenariosRegistry.allValues.sorted().associateBy(
    { it.name },
    {
      JCheckBox("", it in enabledScenarios).apply {
        horizontalAlignment = SwingConstants.CENTER
        background = UIUtil.getTableBackground()
      }
    }
  )

  private val tableModel = object : AbstractTableModel() {
    override fun getRowCount() = scenarios.size
    override fun getColumnCount() = 2
    override fun getColumnName(column: Int) = columnNames[column]
    override fun isCellEditable(rowIndex: Int, columnIndex: Int) = columnIndex == 1

    override fun getValueAt(row: Int, column: Int): Any? = scenarios.asSequence().elementAt(
      row
    ).run { if (column == 0) key else value.isSelected }

    override fun getColumnClass(c: Int): Class<*> {
      return if (c == 0) String::class.java else java.lang.Boolean::class.java
    }
  }

  var renderer = MyRendererEditor()

  private val table = object : JBTable(tableModel) {
    override fun changeSelection(rowIndex: Int, columnIndex: Int, toggle: Boolean, extend: Boolean) {
      super.changeSelection(rowIndex, columnIndex, toggle, extend)
      if(columnIndex == 1) {
        getCellEditor(rowIndex, columnIndex).getTableCellEditorComponent(this, null, true, rowIndex, columnIndex)
      }
    }
  }.apply table@{
    selectionBackground = UIUtil.getTableSelectionBackground(true)

    //checkbox column
    columnModel.getColumn(1).apply {
      resizable = false
      val width = JBUIScale.scale(30)
      minWidth = JBUI.scale(width)
      maxWidth = JBUI.scale(width)
      preferredWidth = JBUI.scale(width)
    }
    tableHeader = JTableHeader(columnModel)
    selectionModel.selectionMode = ListSelectionModel.SINGLE_SELECTION

    columnModel.apply {
      getColumn(1).apply {

        cellRenderer = renderer
        cellEditor = renderer
      }

      getColumn(0).apply {
        // setting up speed search
        TableSpeedSearch(this@table) { _, cell -> scenarios.asSequence().elementAt(cell.row).run { key } }

        cellRenderer = object : ColoredTableCellRenderer() {
          init {
            selectionBackground = UIUtil.getTableSelectionBackground(true)
          }

          override fun paintBorder(g: Graphics?) {
            //
          }

          override fun customizeCellRenderer(
            table: JTable,
            value: Any?,
            selected: Boolean,
            hasFocus: Boolean,
            row: Int,
            column: Int
          ) {
            if (value != null) {
              append(value.toString())
            }
            SpeedSearchUtil.applySpeedSearchHighlighting(this@table, this, true, selected)
          }
        }
      }
    }
    val listener = object : KeyAdapter() {
      override fun keyPressed(e: KeyEvent) {
        fun isToggleEvent(e: KeyEvent): Boolean {
          return e.keyCode == KeyEvent.VK_ENTER && e.isShiftDown
        }
        if (isToggleEvent(e)) {
          val selectedRow = this@table.selectedRow
          scenarios.asSequence().elementAt(selectedRow).run {
            value.isSelected = !value.isSelected
          }

          val cellRect = this@table.getCellRect(selectedRow, 1, false)
          this@table.repaint(cellRect)
          e.consume()
        }
      }
    }
    addKeyListener(listener)
  }


  inner class MyRendererEditor : AbstractCellEditor(), TableCellEditor, TableCellRenderer {

    override fun getCellEditorValue(): Any {
      return ""
    }

    override fun getTableCellEditorComponent(table: JTable?, value: Any?, isSelected: Boolean, row: Int, column: Int): Component? {
      return getComponent(row, table)
    }

    override fun getTableCellRendererComponent(
      table: JTable?,
      value: Any?,
      isSelected: Boolean,
      hasFocus: Boolean,
      row: Int,
      column: Int
    ): Component? {
      return getComponent(row, table)
    }

    private fun getComponent(row: Int, table: JTable?): Component? {
      val cb = scenarios.asSequence().elementAt(row).value

      return if (table?.selectedRow == row) {
        cb.apply {
          background = UIUtil.getTableSelectionBackground(true)
        }
      } else {
        cb.apply {
          background = UIUtil.getTableBackground()
        }
      }
    }
  }

  override fun createActions(): Array<Action> {
    val baseActions = super.createActions()
    val configure = object : AbstractAction(RdPlatformBundle.message("logs.advanced.settings")) {
      override fun actionPerformed(e: ActionEvent?) {
        val logConfigureAction = ActionManager.getInstance().getAction("LogDebugConfigure")
        logConfigureAction.actionPerformed(
          AnActionEvent(
            null, DataManager.getInstance().getDataContext(preferredFocusedComponent),
            ActionPlaces.UNKNOWN, Presentation(),
            ActionManager.getInstance(), 0
          )
        )
      }

    }
    return arrayOf(configure, *baseActions)
  }

  override fun doOKAction() {
    LogTraceScenariosRegistry.getInstance().save(scenarios.filter { it.value.isSelected }.map { it.key })
    PropertiesComponent.getInstance().setValue(TOUCHED, true)
    super.doOKAction()
  }

  override fun createCenterPanel() = JPanel(BorderLayout()).apply {
    add(JBScrollPane(table), BorderLayout.CENTER)
  }

  override fun getPreferredFocusedComponent() = table

  init {
    title = dialogTitle
    init()
    pack()
    setSize(defaultSize.width, defaultSize.height)
  }
}