package com.jetbrains.rd.platform.diagnostics

import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProcessCanceledException
import org.apache.log4j.spi.Filter
import org.apache.log4j.spi.LoggingEvent
import java.io.Closeable

const val structuredSuffix = "\u200b\u200c\u200b" // ZWSP ZWNJ ZWSP - should be invisible in clear text
const val structuredLogMeasuredEvent = "Measured"
const val structuredLogMeasuredDataEvent = "MeasuredData"
const val structuredLogMarkerEvent = "Marker"
const val structuredLogStartEvent = "Start"
const val structuredLogCancelledEvent = "Cancelled"
const val structuredLogExceptionEvent = "Exception"
const val structuredLogAssertionEvent = "Assertion"
const val structuredLogCommandRejectEvent = "CommandReject"
const val structuredLogFinishEvent = "Finish"

class StructuredLogFilter : Filter() {
  override fun decide(event: LoggingEvent): Int {
    return if (event.message.let { it is String && it.endsWith(structuredSuffix) }) NEUTRAL else DENY
  }
}

fun Logger.logMeasuredTime(key: String, timeMs: Long, message: String? = null) {
  debug("<$structuredLogMeasuredEvent> \"$key\" ${message ?: ""} time=${timeMs}ms$structuredSuffix")
}

inline fun <T> Logger.doActivity(key: String, message: String? = null, action: () -> T): T {
  var hadException = false
  debug("<$structuredLogStartEvent> \"$key\" ${message ?: ""}$structuredSuffix")
  val startTime = System.nanoTime()
  try {
    return action()
  } catch(ex: ProcessCanceledException) {
    hadException = true

    debug("<$structuredLogCancelledEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
    throw ex
  } catch (t: Throwable) {
    hadException = true
    debug("<$structuredLogExceptionEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
    throw t
  }
  finally {
    if (!hadException)
      debug("<$structuredLogFinishEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
  }
}

inline fun Logger.doActivity(key: String, message: String? = null, action: () -> Unit) {
  doCalculation(key, message, action)
}

inline fun <T> Logger.doCalculation(key: String, message: String? = null, action: () -> T): T {
  var hadException = false
  debug("<$structuredLogStartEvent> \"$key\" ${message ?: ""}$structuredSuffix")
  val startTime = System.nanoTime()
  try {
    return action()
  }
  catch (ex: ProcessCanceledException) {
    hadException = true

    debug("<$structuredLogCancelledEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
    throw ex
  }
  catch (t: Throwable) {
    hadException = true
    debug("<$structuredLogExceptionEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
    throw t
  }
  finally {
    if (!hadException)
      debug("<$structuredLogFinishEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
  }
}

fun Logger.stopwatchCookie(key: String, message: String? = null): Closeable {
  debug("<$structuredLogStartEvent> \"$key\" ${message ?: ""}$structuredSuffix")
  val startTime = System.nanoTime()

  return Closeable {
    debug("<$structuredLogFinishEvent> \"$key\" ${message ?: ""} time=${((System.nanoTime() - startTime) / 1e6).toLong()}ms$structuredSuffix")
  }
}

fun Logger.structuredMarker(key: String, message: String? = null) {
  debug("<$structuredLogMarkerEvent> \"$key\" ${message ?: ""}$structuredSuffix")
}

fun Logger.logAssertion(message: String) {
  error("<$structuredLogAssertionEvent> \"Assertion\" $message$structuredSuffix")
}

fun Logger.logCommandReject(message: String) {
  error("<$structuredLogCommandRejectEvent> \"CommandReject\" $message$structuredSuffix")
}
