package com.jetbrains.rd.platform.internal

import com.intellij.ide.actions.CloseTabToolbarAction
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.DefaultActionGroup
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.SimpleToolWindowPanel
import com.intellij.openapi.wm.ToolWindow
import com.intellij.openapi.wm.ToolWindowFactory
import com.intellij.ui.content.ContentFactory
import com.intellij.ui.content.ContentManagerAdapter
import com.intellij.ui.content.ContentManagerEvent
import com.intellij.util.ContentsUtil
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.platform.internal.widgets.InternalWidgetBuilder
import com.jetbrains.rd.platform.util.lifetime
import java.awt.FlowLayout
import java.util.*
import javax.swing.BorderFactory
import javax.swing.BoxLayout
import javax.swing.JButton
import javax.swing.JPanel

class InternalToolWindowFactory : ToolWindowFactory, DumbAware {
  override fun isApplicable(project: Project) = ApplicationManager.getApplication().isInternal

  override fun createToolWindowContent(project: Project, toolWindow: ToolWindow) {
    val lifetime = project.lifetime
    val tools = InternalTools.getInstance(project)

    val dashboardContent = JPanel().apply { layout = BoxLayout(this, BoxLayout.Y_AXIS) }
    val dashboardsBoxes = HashMap<String, JPanel>()
    val dashboardTab = ContentFactory.SERVICE.getInstance().createContent(null, "Dashboard", false).apply {
      component = SimpleToolWindowPanel(false).setProvideQuickActions(true).apply {
        setContent(dashboardContent)
      }
    }
    toolWindow.contentManager.addContent(dashboardTab)

    fun addTabBuilder(widgetBuilder: InternalWidgetBuilder) {
      if (!dashboardsBoxes.containsKey(widgetBuilder.category)) {
        val box = JPanel().apply {
          border = BorderFactory.createTitledBorder(widgetBuilder.category)
          layout = FlowLayout(FlowLayout.LEFT)
        }
        dashboardsBoxes[widgetBuilder.category] = box
        dashboardContent.add(box)
      }
      val box = dashboardsBoxes[widgetBuilder.category]!!
      val button = JButton(widgetBuilder.title, widgetBuilder.icon)
      button.addActionListener { toolWindow.addTab(widgetBuilder, button, lifetime) }
      box.add(button)
    }

    for (tab in tools.tabBuilders)
      addTabBuilder(tab)
    tools.tabBuilderAdded.advise(lifetime) { addTabBuilder(it) }
  }

  private fun ToolWindow.addTab(widgetBuilder: InternalWidgetBuilder, button: JButton, globalLifetime: Lifetime) {
    val lifetimeDef = globalLifetime.createNested()
    val lifetime = lifetimeDef.lifetime
    val tab = widgetBuilder.build(lifetime)

    val content = ContentFactory.SERVICE.getInstance().createContent(null, widgetBuilder.category + ": " + widgetBuilder.title, false)
    content.isCloseable = true
    content.icon = widgetBuilder.icon
    content.component = SimpleToolWindowPanel(false).setProvideQuickActions(true).apply {
      setContent(tab.content)

      val actions = ArrayList<AnAction>()
      actions.addAll(tab.toolbarActions())
      actions.add(object : CloseTabToolbarAction() {
        override fun actionPerformed(e: AnActionEvent) {
          ContentsUtil.closeContentTab(contentManager, content)
        }
      })
      val actionGroup = DefaultActionGroup(actions)
      val toolbar = ActionManager.getInstance().createActionToolbar("", actionGroup, false).component
      setToolbar(toolbar)
    }

    contentManager.addContentManagerListener(object : ContentManagerAdapter() {
      override fun contentRemoved(event: ContentManagerEvent) {
        if (event.content == content) {
          lifetimeDef.terminate()
          button.isEnabled = true
        }
      }
    })
    button.isEnabled = false
    contentManager.addContent(content)
    contentManager.setSelectedContent(content, true)

    globalLifetime.onTermination {
      if (!button.isEnabled)
        ContentsUtil.closeContentTab(contentManager, content)
    }
  }
}