// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.platform.util

import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.net.ssl.CertificateManager
import com.intellij.util.proxy.CommonProxy
import java.net.InetAddress
import java.net.NetworkInterface
import java.net.URI
import java.net.http.HttpClient
import java.net.http.WebSocket
import java.util.concurrent.CompletableFuture
import javax.net.ssl.SSLContext

fun URI.withFragment(newFragment: String?): URI = URI(
  scheme,
  userInfo,
  host,
  port,
  path,
  query,
  newFragment
)

fun URI.withScheme(newScheme: String): URI = URI(
  newScheme,
  userInfo,
  host,
  port,
  path,
  query,
  fragment
)

fun URI.withPath(newPath: String?): URI = URI(
  scheme,
  userInfo,
  host,
  port,
  newPath,
  query,
  fragment
)

val URI.fragmentParameters: Map<String, String>
  get() {
    // TODO Url-decode values?
    if (fragment.isNullOrBlank()) return emptyMap()

    return fragment
      .split('&')
      .mapNotNull {
        val split = it.split('=', limit = 2)
        if (split.size != 2) return@mapNotNull null
        split[0] to split[1]
      }
      .toMap()
  }

fun URI.newURIWithFragmentParameters(fragmentParameters: Map<String, String>): URI {
  val newFragment = fragmentParameters.toList().joinToString(separator = "&") { "${it.first}=${it.second}" }
  return URI(scheme, userInfo, host, port, path, query, newFragment)
}

fun CompletableFuture<WebSocket>.closeNormal() {
  if (isDone && !isCompletedExceptionally) {
    val webSocket = get()

    webSocket.sendClose(WebSocket.NORMAL_CLOSURE, "ok").thenRun {
      webSocket.abort()
    }
  }
}

object NetworkUtils {
  private val LOG = Logger.getInstance(javaClass)

  fun getSelfAddresses(): List<InetAddress> {
    try {
      return NetworkInterface.getNetworkInterfaces()
        .toList()
        .flatMap { it.inetAddresses.toList() }
        .filter { !it.isLoopbackAddress && !it.isLinkLocalAddress }
        .distinctBy { it.hostAddress }
        .sortedBy { it.hostAddress }
    }
    catch (t: Throwable) {
      LOG.error("Cannot get list of local interfaces", t)
      return emptyList()
    }
  }

  fun getMacAddress(): ByteArray? {
    fun isGoodAddress(address: ByteArray?) = address != null && address.size == 6 && address.any { it.toInt() != 0 }

    val en = NetworkInterface.getNetworkInterfaces()
    if (en != null) {
      while (en.hasMoreElements()) {
        val networkInterface = en.nextElement()
        if (networkInterface.isLoopback) continue

        val address = networkInterface.hardwareAddress
        if (isGoodAddress(address)) {
          return address
        }
      }
    }

    return null
  }

  fun createHttpClientWithGoodDefaults(customizer: (HttpClient.Builder) -> Unit = { }): HttpClient {
    val httpClientBuilder = HttpClient.newBuilder()

    if (ApplicationManager.getApplication() != null) {
      httpClientBuilder.sslContext(CertificateManager.getInstance().sslContext)
    } else {
      // Reasonable defaults
      val sslContext = SSLContext.getInstance("TLSv1.2")
      sslContext.init(null, null, null)
      httpClientBuilder.sslContext(sslContext)
    }

    httpClientBuilder.proxy(CommonProxy.getInstance())

    customizer(httpClientBuilder)

    return httpClientBuilder.build()
  }
}