package com.jetbrains.rd.platform.util

import com.jetbrains.rd.framework.IRdTask
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.reactive.adviseOnce
import org.jetbrains.concurrency.AsyncPromise
import org.jetbrains.concurrency.Promise
import java.util.concurrent.CancellationException


fun <T> IRdTask<T>.toPromise(lifetime: Lifetime? = null): Promise<T> = toAsyncPromise(lifetime ?: Lifetime.Eternal)

/**
 * Creates an [AsyncPromise] that may be resolved in one of the following ways:
 * - if the task isn't finished yet, then the promise will be resolved on the current dispatcher of the task after it is finished
 * - if the task is already finished, then the promise will be resolved **synchronously on the current thread**
 */
fun <T> IRdTask<T>.toAsyncPromise(lifetime: Lifetime? = null) = toAsyncPromise(lifetime ?: Lifetime.Eternal, { it }, { it })

/**
 * Creates an [AsyncPromise] that may be resolved in one of the following ways:
 * - if the task isn't finished yet, then the promise will be resolved on the current dispatcher of the task after it is finished
 * - if the task is already finished, then the promise will be resolved **synchronously on the current thread**
 */
fun <T, R> IRdTask<T>.toAsyncPromise(lifetime: Lifetime, map: (T) -> R)= toAsyncPromise(lifetime, map, { it })

/**
 * Creates an [AsyncPromise] that may be resolved in one of the following ways:
 * - if the task isn't finished yet, then the promise will be resolved on the current dispatcher of the task after it is finished
 * - if the task is already finished, then the promise will be resolved **synchronously on the current thread**
 */
fun <T, R> IRdTask<T>.toAsyncPromise(outerLifetime: Lifetime, mapResult: (T) -> R, mapError: (Throwable) -> Throwable): AsyncPromise<R> {
  val promise = AsyncPromise<R>()
  val def = outerLifetime.createNested().apply {
    synchronizeWith(promise, true)
  }
  val lt = def.lifetime
  result.adviseOnce(lt) {
    when (it) {
      is RdTaskResult.Success -> promise.setResult(mapResult(it.value))
      is RdTaskResult.Fault -> promise.setError(mapError(it.error))
      is RdTaskResult.Cancelled -> promise.cancel()
    }
  }
  return promise
}

/**
 * Creates new [AsyncPromise] that is translated [State] into [State.REJECTED] if this lifetime becomes terminated (and promise wasn't completed before).
 * @return New [AsyncPromise]. Possibly already canceled if this lifetime already terminated.
 **/
fun <T> Lifetime.createNestedAsyncPromise(): AsyncPromise<T> {
  return AsyncPromise<T>().apply {
    createNested().synchronizeWith(this)
  }
}

/**
 * Terminates this lifetime once the passed [asyncPromise] becomes terminated.
 * @param asyncPromise promise to listen for
 * @param supportsTerminationUnderExecution pass `true` for cases when you expect for the [asyncPromise] to become
 * processed while an action is executing under this lifetime. For example, this:
 */
fun <T> LifetimeDefinition.synchronizeWith(asyncPromise: AsyncPromise<T>, supportsTerminationUnderExecution: Boolean = false) {
  if (asyncPromise.isDone) {
    terminate(supportsTerminationUnderExecution)
    return
  }

  if (onTerminationIfAlive { asyncPromise.cancel() }) {
    asyncPromise.onProcessed { terminate(supportsTerminationUnderExecution) }
  } else {
    asyncPromise.cancel()
  }
}

fun <T> Promise<T>.toRdTask() = RdTask<T>().also { rdTask ->
  onSuccess { rdTask.set(it) }
  onError {
    when (it) {
      is CancellationException -> rdTask.result.set(RdTaskResult.Cancelled())
      else -> rdTask.result.set(RdTaskResult.Fault(it))
    }
  }
}