package com.jetbrains.rd.platform.util

import com.intellij.credentialStore.CredentialAttributes
import com.intellij.credentialStore.generateServiceName
import com.intellij.ide.passwordSafe.PasswordSafe
import com.intellij.openapi.application.PathManager
import com.intellij.openapi.diagnostic.Logger
import com.jetbrains.rd.util.UUID
import org.bouncycastle.asn1.oiw.OIWObjectIdentifiers
import org.bouncycastle.asn1.x500.X500Name
import org.bouncycastle.asn1.x500.X500NameBuilder
import org.bouncycastle.asn1.x500.style.RFC4519Style
import org.bouncycastle.asn1.x509.*
import org.bouncycastle.cert.X509ExtensionUtils
import org.bouncycastle.cert.X509v3CertificateBuilder
import org.bouncycastle.cert.jcajce.JcaX509CertificateConverter
import org.bouncycastle.jce.provider.BouncyCastleProvider
import org.bouncycastle.operator.DigestCalculator
import org.bouncycastle.operator.bc.BcDigestCalculatorProvider
import org.bouncycastle.operator.jcajce.JcaContentSignerBuilder
import java.io.File
import java.math.BigInteger
import java.security.KeyPairGenerator
import java.security.KeyStore
import java.security.PublicKey
import java.security.cert.X509Certificate
import java.time.Duration
import java.time.Instant
import java.util.*
import javax.security.auth.x500.X500Principal

object SelfSignedCertificate {
  // Use in-memory certificates by default
  // Persistent certificates are bringing UX problem of asking keychain password which looks strange for our service
  // See https://youtrack.jetbrains.com/issue/CWM-1283 https://youtrack.jetbrains.com/issue/CWM-126
  // With correctly implemented fingerprinting, persistent certificates are no longer required,
  // see https://jetbrains.team/p/ij/review/5813 for the implementation
  private val memoryOnlyCertificate = System.getProperty("codeWithMe.memory.only.certificate", "true")!!.toBoolean()

  private fun getOrCreateKeyStorePassword(): CharArray {
    if (memoryOnlyCertificate) {
      LOG.info("Using memory only password for persistent certificate. Certificate won't be reused on the next run")
      return UUID.randomUUID().toString().toCharArray()
    }

    val attributes = CredentialAttributes(generateServiceName("Code With Me", "Key Store Password"), "KeyStorePassword")

    val existingPassword = PasswordSafe.instance.get(attributes)?.password
    if (existingPassword != null) {
      return existingPassword.toCharArray()
    }

    val newPassword = UUID.randomUUID().toString()

    PasswordSafe.instance.setPassword(attributes, newPassword)
    val newCredentials = PasswordSafe.instance.get(attributes)
    if (newCredentials == null) {
      LOG.error("Unable to save Code With Me keystore password, attributes: $attributes")
    }
    else
      if (newCredentials.getPasswordAsString() != newPassword) {
        LOG.error("Code With Me keystore password is different after get/set PasswordSafe round-trip")
      }

    return newPassword.toCharArray()
  }

  private fun getKeyStoreLocation(): File {
    return PathManager.getConfigDir().resolve("codeWithMe").resolve("app.jks").toFile()
  }

  val permanentCertificateData: SelfSignedCertificateData by lazy {
    val keyPassword = getOrCreateKeyStorePassword()
    val keystoreFile = getKeyStoreLocation()

    // Best effort to save key across restarts
    val existingKeyStore: KeyStore? = try {
      if (keystoreFile.isFile && !memoryOnlyCertificate) {
        val keystore = KeyStore.getInstance("jks")

        keystoreFile.inputStream().use {
          keystore.load(it, keyPassword)
        }

        val certificate = keystore.getCertificate(KEY_ALIAS) as? X509Certificate
        if (certificate == null) {
          error("X509 certificate with alias '$KEY_ALIAS' was not found")
        }

        val key = keystore.getKey(KEY_ALIAS, keyPassword)
        if (key == null) {
          error("Key with alias '$KEY_ALIAS' was not found")
        }

        certificate.checkValidity()

        val otherAliases = keystore.aliases().toList().filter { it != KEY_ALIAS }
        if (otherAliases.isNotEmpty()) {
          error("Other aliases discovered in keystore: ${otherAliases.joinToString()}")
        }

        keystore
      }
      else null
    }
    catch (t: Throwable) {
      LOG.warn("Unable to get certificate or password from existing keystore at $keystoreFile, will generate a new one: ${t.message}", t)
      null
    }

    return@lazy if (existingKeyStore != null) {
      val certificate = existingKeyStore.getCertificate(KEY_ALIAS) as X509Certificate
      SelfSignedCertificateData(
        certificate = certificate,
        fingerprint = certificate.sha3Fingerprint,
        keyStore = existingKeyStore,
        keyEntryAlias = KEY_ALIAS,
        keyPassword = keyPassword
      )
    }
    else {
      // Issue a long time certificate
      val data = generateNewCertificate(
        keyPassword = keyPassword,
        subjectName = X500NameBuilder().addRDN(RFC4519Style.cn, UUID.randomUUID().toString()).build(),
        issuer = null,
        days = 3650
      )

      if (!memoryOnlyCertificate) {
        keystoreFile.parentFile.mkdirs()
        keystoreFile.delete()

        keystoreFile.outputStream().use {
          data.keyStore.store(it, keyPassword)
        }
      }

      data
    }
  }

  class SelfSignedCertificateData(
    val certificate: X509Certificate,
    val fingerprint: String,
    val keyStore: KeyStore,
    val keyEntryAlias: String,
    val keyPassword: CharArray
  )

  fun generateNewCertificate(keyPassword: CharArray,
                             subjectName: X500Name,
                             issuer: X509Certificate?,
                             days: Int): SelfSignedCertificateData {
    val keyPairGenerator = KeyPairGenerator.getInstance("RSA").also {
      it.initialize(2048)
    }
    val subjectKeyPair = keyPairGenerator.genKeyPair()

    val now: Instant = Instant.now()
    val notBefore = Date.from(now)
    val notAfter = Date.from(now.plus(Duration.ofDays(days.toLong())))

    val subjectPrincipal = X500Principal(subjectName.encoded)
    val subjectSerial = BigInteger.valueOf(now.toEpochMilli())

    val isCA = issuer == null
    val issuerPrincipal = if (issuer != null) issuer.subjectX500Principal else subjectPrincipal
    val issuerPublicKey = if (issuer != null) issuer.publicKey else subjectKeyPair.public

    val certificateBuilder: X509v3CertificateBuilder = X509v3CertificateBuilder(
      X500Name.getInstance(issuerPrincipal.encoded),
      subjectSerial,
      notBefore,
      notAfter,
      Locale.ENGLISH,
      X500Name.getInstance(subjectPrincipal.encoded),
      SubjectPublicKeyInfo.getInstance(subjectKeyPair.public.encoded))
      .addExtension(Extension.subjectKeyIdentifier, false, createSubjectKeyId(subjectKeyPair.public))
      .addExtension(Extension.authorityKeyIdentifier, false, createAuthorityKeyId(issuerPublicKey))
      .addExtension(Extension.basicConstraints, true, BasicConstraints(isCA))

    val contentSigner = JcaContentSignerBuilder("SHA512WithRSA").build(subjectKeyPair.private)
    val certificate = JcaX509CertificateConverter().setProvider(BouncyCastleProvider())
      .getCertificate(certificateBuilder.build(contentSigner))!!

    val keystore = KeyStore.getInstance("jks")
    keystore.load(null, null)
    val chain = if (issuer != null) arrayOf(certificate, issuer) else arrayOf(certificate)
    keystore.setKeyEntry(KEY_ALIAS, subjectKeyPair.private, keyPassword, chain)

    return SelfSignedCertificateData(
      certificate = certificate,
      fingerprint = certificate.sha3Fingerprint,
      keyStore = keystore,
      keyEntryAlias = KEY_ALIAS,
      keyPassword = keyPassword
    )
  }

  private fun createSubjectKeyId(publicKey: PublicKey): SubjectKeyIdentifier? {
    val publicKeyInfo: SubjectPublicKeyInfo = SubjectPublicKeyInfo.getInstance(publicKey.encoded)
    val digCalc: DigestCalculator = BcDigestCalculatorProvider().get(AlgorithmIdentifier(OIWObjectIdentifiers.idSHA1))
    return X509ExtensionUtils(digCalc).createSubjectKeyIdentifier(publicKeyInfo)
  }

  private fun createAuthorityKeyId(publicKey: PublicKey): AuthorityKeyIdentifier? {
    val publicKeyInfo: SubjectPublicKeyInfo = SubjectPublicKeyInfo.getInstance(publicKey.encoded)
    val digCalc: DigestCalculator = BcDigestCalculatorProvider().get(AlgorithmIdentifier(OIWObjectIdentifiers.idSHA1))
    return X509ExtensionUtils(digCalc).createAuthorityKeyIdentifier(publicKeyInfo)
  }

  private const val KEY_ALIAS = "codewithme"
  private val LOG = Logger.getInstance(SelfSignedCertificate::class.java)
}
