// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui

import java.awt.Dimension
import java.awt.GridLayout
import javax.swing.JComponent
import javax.swing.JPanel
import kotlin.math.min

/**
 * @author Nikolay.Kuznetsov
 */
class MaximumAndPreferredSizeLimitedComponentWrapper(val target: JComponent, val maxSize:Dimension? = null) : JPanel(GridLayout(1, 1, 0, 0)) {
  init {
    add(target)
  }

  override fun getPreferredSize(): Dimension {
    val prefSize = target.preferredSize
    val maxSize = maxSize ?: target.maximumSize
    val widthPadding = 10

    return Dimension(min(prefSize.width, maxSize.width) + widthPadding, min(prefSize.height, maxSize.height))
  }

  override fun getMaximumSize() = preferredSize
}

/**
 * This wraps the component in a JPanel.
 * The maximum and preferred sizes of that panel are equal, and are the minimum of original component's maximum and preferred sizes
 */
fun JComponent.withSensibleMaxAndPreferredSizes(maxDimension: Dimension? = null): JComponent =
  MaximumAndPreferredSizeLimitedComponentWrapper(this, maxDimension)

fun JComponent.withOverriddenPreferredWidth(width: Int): JComponent {
  val panel = object : JPanel(GridLayout(1, 1)) {
    override fun getPreferredSize(): Dimension {
      val calculated = super.getPreferredSize()
      return Dimension(width, calculated.height)
    }
  }
  panel.add(this)
  return panel
}

fun JComponent.withOverriddenPreferredSize(width: Int, height:Int, getPreferred:  (Int, Int) -> Int): JComponent {
  val panel = object : JPanel(GridLayout(1, 1)) {
    override fun getPreferredSize(): Dimension {
      val calculated = super.getPreferredSize()
      return Dimension(getPreferred(calculated.width, width), getPreferred(calculated.height, height))
    }
  }
  panel.add(this)
  return panel
}
/**
 * This wraps a label into two panels that ensure the following:
 *  * The label occupies full width of containing panel
 *  * The label's preferred size does not cause containing panel to become wider
 *  * If the label is multi-line (e.g. with HTML in it), it's free to reflow the text to occupy multiple lines
 *    while still not exceeding the width of containing panel
 */
fun JComponent.fillWidthButDoNotOverflow(opaque: Boolean = true): JComponent {
  val containingPanel = object : JPanel(GridLayout(1, 1, 0, 0)) {
    override fun resize(width: Int, height: Int) {
      super.resize(width, height)
      this@fillWidthButDoNotOverflow.maximumSize = Dimension(width, 9999999)
    }

    override fun getPreferredSize(): Dimension {
      val result = super.getPreferredSize()
      return Dimension(minimumSize.width, result.height)
    }
  }.apply {
    isOpaque = opaque
  }
  containingPanel.add(this.withSensibleMaxAndPreferredSizes().apply { isOpaque = opaque })
  return containingPanel
}