// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui

import com.intellij.openapi.util.text.StringUtil
import com.intellij.ui.ColorUtil
import com.jetbrains.rdclient.util.idea.ui.toColor
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ui.util.getContrastedForegroundIfNotThemed
import java.awt.Color
import java.lang.IllegalArgumentException

fun RgbColor.isGrayed(): Boolean {
  return red == green && green == blue && red < 255 && red > 0
}

fun RichTextModel.toSimpleText(): String {
  return this.parts.fold("") { s, modelString -> s + modelString.text }
}

fun RichTextModel.splitByGrayColor(): Pair<String, String> {
  val appender = StringBuilder()
  val grayBuilder = StringBuilder()
  for (part in this.parts) {

    val foregroundColor = part.foregroundColor as? RgbColor
    if (foregroundColor != null && foregroundColor.isGrayed()) {
      grayBuilder.append(part.text)
    }
    else {
      if (grayBuilder.isNotEmpty()) //abort if the rich text is inconsistent
        return Pair(this.toSimpleText(), "")
      appender.append(part.text)
    }
  }

  return Pair(appender.toString(), grayBuilder.toString())
}

fun RichTextModel.toHtml(stringBuilder: StringBuilder, bgColor: Color) {
  for (part in parts) {
    stringBuilder.append("<span style=\"")
    stringBuilder.append("background-color:")
      .append(ColorUtil.toHtmlColor(bgColor))
      .append(";")

    val fgColor = part.foregroundColor?.toColor()
    if (fgColor != null) {
      val contrastedFgColor = getContrastedForegroundIfNotThemed(bgColor, fgColor, true)
      stringBuilder.append("color:")
        .append(ColorUtil.toHtmlColor(contrastedFgColor))
        .append(';')
    }

    val effectStyle = part.effectStyle
    if (effectStyle != EffectStyle.None) {
      when (effectStyle) {
        EffectStyle.WeavyUnderline -> stringBuilder.append("text-decoration: underline;")
        EffectStyle.StrikeOut -> stringBuilder.append("text-decoration: line-through;")
        else -> throw IllegalArgumentException("$effectStyle is unsupported effect.")
      }
    }

    val fs = part.fontStyle
    if (fs != FontStyle.Regular) {
      if (fs == FontStyle.Bold) {
        stringBuilder.append("font-weight: bold;")
      }
      if (fs == FontStyle.Italic) {
        stringBuilder.append("font-style: italic;")
      }
      if (fs == FontStyle.Strikeout) {
        stringBuilder.append("text-decoration: line-through;")
      }
      if (fs == FontStyle.Underline) {
        stringBuilder.append("text-decoration: underline;")
      }
    }
    stringBuilder.append("\">").append(StringUtil.escapeXmlEntities(part.text)).append("</span>")
  }
}

fun RichTextBlockModel.toHtml(stringBuilder: StringBuilder, bgColor: Color) {
  for ((i, line) in lines.withIndex()) {
    if (i in 1 until lines.size)
      stringBuilder.append("<br/>")
    line.toHtml(stringBuilder, bgColor)
  }
}

fun RichTextBlockModel.toHtml(bgColor: Color): String {
  val builder = StringBuilder()
  toHtml(builder, bgColor)
  return builder.toString()
}