@file:Suppress("unused")

package com.jetbrains.rd.ui.bedsl.dsl

import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.reactive.*
import com.jetbrains.rd.ui.bedsl.dsl.util.BeLocalCookie
import com.jetbrains.rd.ui.bedsl.extensions.bind

private fun createCombobox(
  isEditable: Boolean = false,
  id: String? = "",
  updateSourceTrigger: UpdateSourceTrigger = UpdateSourceTrigger.TextChanged
): BeComboBox {
  val comboBox = BeComboBox(BeTextBoxSettings(updateSourceTrigger))
  comboBox.beInit(id)

  comboBox.isEditable.value = isEditable
  comboBox.text.value = ""
  comboBox.selectedIndex.value = 0

  comboBox.validationResult.set(ValidationResult().apply { icon.value = null })
  comboBox.validationStyle.set(ValidationStyle.Border)

  comboBox.settings.textSettings.fontStyle.set(BeFontStyle.DEFAULT)
  comboBox.settings.textSettings.fontSize.set(BeFontSize.DEFAULT)
  comboBox.settings.readonly.value = false
  comboBox.settings.useCamelHumps.value = false

  return comboBox
}

private fun <T> getControlFromValue(t: T, present: ((T) -> BeControl)? = null): BeControl? {
  val invalidValue = label("Invalid")

  val text: BeControl
  if (present != null) {
    text = present(t)
  }
  else {
    var description = ""
    /* if (t is Enum @enum)
     {
         description = @enum.Description()
     }*/
    if (description.isNullOrEmpty())
      description = t.toString()
    if (description.isNullOrEmpty()) {
      return invalidValue
    }

    text = label(description)
    text.controlId.value = t.toString()
  }
  return text
}

private fun <T> buildItem(value: T, presentation: ((T) -> BeControl?)?): BeControl? {
  val control: BeControl?
  if (presentation != null) {
    control = presentation(value)
    if (control == null) return null
  }
  else
    control = getControlFromValue(value)

  return control
}

private fun <T : Any> addControls(
  comboBox: BeComboBox,
  lifetime: Lifetime,
  comboValues: ViewableList<T>,
  presentation: ((T) -> BeControl?)?) {
  comboValues.adviseAddRemove(lifetime) { h, index, e ->
    val comboBoxValues = comboBox.values
    val selectedIndex = comboBox.selectedIndex.value
    if (h == AddRemove.Add) {
      val buildItem = buildItem(e, presentation)
      if (buildItem != null) {
        comboBoxValues.add(index, buildItem)
        if (selectedIndex <= 0 || comboBoxValues.count() == 1) {
          if (!comboBox.isEditable.value) {
            // || selectedIndex >= 0
            comboBox.selectedIndex.value = -1
            comboBox.selectedIndex.value = 0
          }
        }
        else if (selectedIndex >= index)
          comboBox.selectedIndex.value++
      }
    }
    else {
      if (comboBoxValues.count() > index) {
        comboBoxValues.removeAt(index)
        if (!lifetime.isAlive) return@adviseAddRemove

        val count = comboBoxValues.count()
        if (selectedIndex >= index && count >= 0) {
          if (selectedIndex == index) {
            comboBox.selectedIndex.value = -1
            comboBox.selectedIndex.value = index + (if (index == count - 1) -1 else 0)
          }
          else
            comboBox.selectedIndex.value += if (selectedIndex == 0) 1 else -1
        }
      }
    }
  }
}

private fun <T : Any> bindToProperty(property: IProperty<T>, lifetime: Lifetime, comboBox: BeComboBox, valueList: IMutableViewableList<T>) {
  val value = property.value
  comboBox.selectedIndex.value = if (valueList.contains(value)) valueList.indexOf(value) else 0
  var localCookie = false
  comboBox.selectedIndex.advise(lifetime)
  {
    val outsideValues = valueList.count() <= it || it < 0
    //if (outsideValues && !property.isNullValueAllowed) return

    if (localCookie) return@advise
    localCookie = true
    property.value = if (outsideValues) value else valueList[it]
    localCookie = false
  }
  property.change.adviseWithPrev(lifetime
  ) { m, it ->
    if (localCookie || !m.hasValue) return@adviseWithPrev
    localCookie = true
    comboBox.selectedIndex.value = if (valueList.contains(it)) valueList.indexOf(it) else if (comboBox.isEditable.value) -1 else 0
    localCookie = false
  }

}

fun combobox(lifetime: Lifetime,
             values: List<String>,
             selectedValue: String = "",
             id: String? = "",
             handleSelected: (String) -> Unit): BeComboBox {
  val property = Property(if (selectedValue.isNotEmpty()) selectedValue else values.first())
  property.advise(lifetime, handleSelected)
  return combobox(property, lifetime, values, id) { element -> label(element) }
}

fun <T : Any> combobox(lifetime: Lifetime, values: List<T>, selectedValue: T? = null,
                       id: String? = "",
                       handleSelected: (T) -> Unit,
                       presentation: ((element: T) -> BeControl?)? = null): BeComboBox {
  val property = Property<T>(selectedValue ?: values.first())
  property.advise(lifetime, handleSelected)
  return combobox(property, lifetime, values, id, presentation)
}

fun <T : Any> combobox(
  property: IProperty<T>,
  lifetime: Lifetime,
  values: ViewableList<T>,
  id: String? = "",
  presentation: ((T) -> BeControl?)? = null
): BeComboBox {
  val comboBox = createCombobox(id = id)
  addControls(comboBox, lifetime, values, presentation)
  bindToProperty(property, lifetime, comboBox, values)
  return comboBox
}


fun <T : Any> combobox(
  property: IProperty<T>,
  lifetime: Lifetime,
  values: List<T>,
  id: String? = "",
  presentation: ((T) -> BeControl?)? = null
): BeComboBox {
  val valueList = ViewableList<T>()
  for (element in values) {
    valueList.add(element)
  }
  return combobox(property, lifetime, valueList, id, presentation)
}


inline fun <reified T : Enum<T>> comboboxFromEnum(
  lifetime: Lifetime,
  id: String = "",
  selectedValue: T? = null,
  noinline handleSelected: (T) -> Unit,
  crossinline getText: (T) -> String = { it.name }
): BeControl {
  return combobox(lifetime, enumValues<T>().toList(), selectedValue, id, handleSelected) { label(getText(it)) }
}

inline fun <reified T : Enum<T>> comboboxFromEnum(
  property: IProperty<T>,
  lifetime: Lifetime,
  id: String = "",
  crossinline getText: (T) -> String = { it.name }
): BeControl {
  return combobox(property, lifetime, enumValues<T>().toList(), id) { it -> label(getText(it)) }
}

fun editableCombobox(
  lifetime: Lifetime,
  values: List<String>,
  initialValue: String = "",
  id: String? = "",
  updateSourceTrigger: UpdateSourceTrigger = UpdateSourceTrigger.TextChanged,
  handleSelected: (String) -> Unit
): BeComboBox {
  val property = Property(if (initialValue.isNotEmpty()) initialValue else values.firstOrNull() ?: "")
  property.advise(lifetime, handleSelected)
  return editableCombobox(property, lifetime, values, id, updateSourceTrigger, { label(it) }, { it })
}

fun editableCombobox(
  property: IProperty<String>,
  lifetime: Lifetime,
  values: List<String>,
  id: String? = "",
  updateSourceTrigger: UpdateSourceTrigger = UpdateSourceTrigger.TextChanged): BeComboBox {
  return editableCombobox(property, lifetime, values, id, updateSourceTrigger, { label(it) }, { it })
}

fun editableCombobox(
  property: IProperty<String>,
  lifetime: Lifetime,
  values: ViewableList<String>,
  id: String? = "",
  updateSourceTrigger: UpdateSourceTrigger = UpdateSourceTrigger.TextChanged): BeComboBox {
  return editableCombobox(property, lifetime, values, id, updateSourceTrigger, { label(it) }, { it })
}

fun <T : Any> editableCombobox(
  property: IProperty<String>,
  lifetime: Lifetime,
  values: List<T>,
  id: String? = "",
  updateSourceTrigger: UpdateSourceTrigger = UpdateSourceTrigger.TextChanged,
  presentation: ((T) -> BeControl?),
  getElementFromString: ((String) -> T?)
): BeComboBox {
  return editableCombobox(property, lifetime, ViewableList(values.toMutableList()), id, updateSourceTrigger, presentation,
                          getElementFromString)
}

fun <T : Any> editableCombobox(
  property: IProperty<String>,
  lifetime: Lifetime,
  values: ViewableList<T>,
  id: String? = "",
  updateSourceTrigger: UpdateSourceTrigger = UpdateSourceTrigger.TextChanged,
  presentation: ((T) -> BeControl?),
  getElementFromString: ((String) -> T?)
): BeComboBox {
  val comboBox = createCombobox(true, id, updateSourceTrigger)
  addControls(comboBox, lifetime, values, presentation)
  bindToPropertyForEdit(property, lifetime, comboBox, values, getElementFromString)
  return comboBox
}

fun <T : Any> bindToPropertyForEdit(
  property: IProperty<String>,
  lifetime: Lifetime,
  comboBox: BeComboBox,
  valueList: ViewableList<T>,
  elementFromString: (String) -> T?
) {
  val value = property.value
  val valueFromString = elementFromString(value)

  val selectedIndex = if (valueList.contains(valueFromString))
    valueList.indexOf(valueFromString)
  else -1
  comboBox.selectedIndex.value = selectedIndex
  comboBox.text.value = value
  val localCookie = BeLocalCookie()
  val realTextProperty = property
  /*
      if (propagateValueWhenInvalid != BeInvalidValuePropagation.SAVE_ANY)
      {
          realTextProperty = BeUtil.GetValidatedProperty<T>(property, lifetime, comboBox, propagateValueWhenInvalid)
      }
  */

  property.bind(lifetime, comboBox.text)

  comboBox.text.change.advise(lifetime)
  {
    localCookie.Execute {
      val argsFromString = elementFromString(it)
      comboBox.selectedIndex.value = when {
        valueList.contains(argsFromString) -> valueList.indexOf(argsFromString)
        comboBox.isEditable.value -> -1
        else -> 0
      }
    }
  }

  comboBox.selectedIndex.change.advise(lifetime)
  {
    localCookie.Execute {
      val currentValue = realTextProperty.value
      comboBox.text.value = when {
        it >= 0 && it < comboBox.values.count() ->
          comboBox.values[it].getText() ?: currentValue
        else -> currentValue
      }
    }
  }
}
