@file:Suppress("unused")

package com.jetbrains.rd.ui.bedsl.dsl

import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.*
import com.jetbrains.rd.ui.bedsl.dsl.util.BeLocalCookie
import com.jetbrains.rd.ui.bedsl.dsl.util.ColumnDefinition
import com.jetbrains.rd.ui.bedsl.dsl.util.TreeConfiguration
import com.jetbrains.rd.ui.bedsl.dsl.util.TreeSettings
import com.jetbrains.rd.ui.bedsl.extensions.bind
import com.jetbrains.rd.ui.bedsl.extensions.view

fun BeTreeGrid.withIndefiniteProgress(lifetime: Lifetime, inProgress: Property<Boolean>): BeTreeGrid {
  val progress = indefiniteProgress()
  this.progress.value = progress
  inProgress.bind(lifetime, progress.inProgress)
  return this
}

typealias presentElement<T> = (Lifetime, T, ListNodeProperties) -> BeControl
fun <T : Any> list(
  items: List<T>,
  lifetime: Lifetime,
  id: String? = "",
  presentLine: presentElement<T>
): BeTreeGrid {
  return table(items, lifetime, TreeConfiguration.oneColumn(), id){lt, e, p -> listOf(presentLine(lt, e, p))}
}

fun <T : Any> singleSelectionTable(resultProperty: IProperty<T?>,
                                   initialList: List<T>,
                                   lifetime: Lifetime,
                                   columnsAndSizes: List<String>? = null,
                                   id: String? = "",
                                   hasHeader: Boolean = true,
                                   backendContextMenu: BeBackendContextMenu? = null,
                                   presentLine: presentTableLine<T>
): BeTreeGrid {
  return singleSelectionTable(resultProperty,
                              initialList,
                              lifetime,
                              id,
                              TreeConfiguration.columnNames(columnsAndSizes,
                                                            TreeSettings(
                                                              hasHeader = hasHeader,
                                                              backendContextMenu = backendContextMenu
                                                            )
                              ),
                              presentLine)
}

fun <T : Any> singleSelectionTable(resultProperty: IProperty<T?>,
                                   initialList: List<T>,
                                   lifetime: Lifetime,
                                   id: String? = "",
                                   treeConfiguration: TreeConfiguration,
                                   presentLine: presentTableLine<T>
): BeTreeGrid {
  val listPresentation = table(initialList, lifetime,
                               treeConfiguration, id) { lt, e, c ->

    val localCookie = BeLocalCookie()
    lt.onTermination {
      if (c.selected.value) // && resultProperty.IsNullValueAllowed)
        resultProperty.value = null
    }
    c.selected.change.advise(lt
    ) {
      localCookie.Execute {
        if (it)
          resultProperty.value = e
      }
    }
    resultProperty.view(lt
    ) { _, it ->
      localCookie.Execute {
        //if (e == it)
        c.selected.value = e == it
      }
    }
    return@table presentLine(lt, e, c)
  }

  if (resultProperty.value == null) {
    val beTreeGridLine = listPresentation.fakeRoot.children.firstOrNull()
    if (beTreeGridLine is BeTreeGridLine)
      beTreeGridLine.isSelected.value = true
  }

  //listPresentation.controlId.Value = resultProperty.Id.Id
  return listPresentation
}



typealias presentCheckedListLine<T> = (Lifetime, T, CheckBoxListNodeProperties) -> List<BeControl>

//todo: should be boolean but clashes signature elsewhere
class CheckBoxListNodeProperties(val listNodeProperties: ListNodeProperties, val included: Property<Boolean?> = Property(true))

typealias presentCheckedTreeLine<T> = (Lifetime, T?, CheckBoxTreeNodeProperties, TreeNodeProperties) -> List<BeControl>

class CheckBoxTreeNodeProperties(val listNodeProperties: ListNodeProperties, val included: Property<Boolean?> = Property(true))

fun <T : Any> treeWithCheckBoxes(initialList: List<T>,
                                 lifetime: Lifetime,
                                 configuration: TreeConfiguration,
                                 getChildren: (T?) -> List<T>,
                                 id: String? = "", getLinePresentation: presentCheckedTreeLine<T>
): BeTreeGrid {
  val elements = initialList.map { ElementWithCheckBox(it, false) }
  //configuration.NodeConverter = o => (o as ElementWithCheckBox<T>).Value

  return tree(elements, lifetime, configuration, { t: ElementWithCheckBox<T>? -> t?.children ?: ViewableList() }, id)
  { nodeLt, e, lp, tp ->
    if (e == null) return@tree mutableListOf()

    val checkBoxTreeNodeProperties = CheckBoxTreeNodeProperties(lp)
    e.initChildren(getChildren, nodeLt)
    val checkBox = checkbox("")
    checkBox.isThreeState.value = true

    val backendCookie = BeLocalCookie()
    checkBoxTreeNodeProperties.included.change.advise(nodeLt
    ) {
      val updateFrom = "backend"
      //ourLogger.Trace($"{updateFrom} valueChange: {element}, newValue: {v.New}")
      if (it == null)
        return@advise
      //Assertion.Assert(v.New != null, "v.New != null")
      updateIsIncluded(backendCookie, updateFrom, e, it)
    }

    val frontendCookie = BeLocalCookie()
    checkBox.property.change.advise(nodeLt
    ) {
      val updateFrom = "frontend"
      //ourLogger.Trace($"{updateFrom} valueChange: {element}, newValue: {v}")
      if (it == null)
        return@advise
      updateIsIncluded(frontendCookie, updateFrom, e, it)
    }

    val localCookie = BeLocalCookie()
    e.isIncluded.change.advise(nodeLt
    ) {
      //var updateFrom = "model"
      //ourLogger.Trace($"{updateFrom} valueChange: {element}, newValue: {v.New}")
      localCookie.Execute {
        checkBoxTreeNodeProperties.included.value = it
        // This check is necessary to prevent cyclic update of frontend property
        if (/*!checkBox.property.maybe.hasValue ||*/ checkBox.property.value == null || checkBox.property.value != it)
          checkBox.property.value = it
      }
    }
    val controlAutomations = getLinePresentation(nodeLt, e.value, checkBoxTreeNodeProperties, tp)
    e.isIncluded.value = checkBoxTreeNodeProperties.included.value
    checkBox.property.value = checkBoxTreeNodeProperties.included.value
    val grid = horizontalGrid {
      column { checkBox }
      column { controlAutomations.first() } //todo: fill
    }
    return@tree controlAutomations.drop(1).toMutableList().apply { add(0, grid) }
  }
}

private fun <T : Any> updateIsIncluded(localCookie: BeLocalCookie, updateFrom: String, element: ElementWithCheckBox<T>, v: Boolean) {
  localCookie.Execute {
    //ourLogger.Trace($"{updateFrom} valueChange, firing to model: {v}")
    element.isIncluded.value = v
  }
}

class ElementWithCheckBox<T : Any>(val value: T, isIncluded: Boolean?) {
  val isIncluded: IProperty<Boolean?> = Property(isIncluded)
  val children: ViewableList<ElementWithCheckBox<T>> = ViewableList()

  /*  public override string ToString()
  {
      return $"Value: {Value}, IsIncluded: {IsIncluded.Value}"
  }*/
  private fun calculateIsIncluded(): Boolean? {
    val childrenCount = children.count()
    val includedChildrenCount = children.count { it.isIncluded.value ?: false }
    //ourLogger.Trace($"CalculateIsIncluded {Value.ToString()} children count: {childrenCount}, " +
    //$"includedChildrenCount: {includedChildrenCount}, IsIncluded: {IsIncluded}")
    if (childrenCount == 0)
      return isIncluded.value
    if (includedChildrenCount == 0)
      return false
    if (includedChildrenCount == childrenCount)
      return true
    return null
  }

  internal fun initChildren(getChildren: (T) -> List<T>, lifetime: Lifetime) {
    when (val modelChildren = getChildren(value)) {
      is ViewableList<T> -> modelChildren
      else -> ViewableList<T>().apply { addAll(modelChildren) }
    }.view(lifetime) { lt, i, e, k ->
      when (k) {
        AddRemove.Add -> {
          val localCookie = BeLocalCookie()
          val element = ElementWithCheckBox(e, isIncluded.value)
          //ourLogger.Trace($"into: ({this}) added: ({element})")
          isIncluded.change.advise(lt!!)
          {
            localCookie.Execute {
              if (it != null)
                element.isIncluded.value = it
            }
          }
          element.isIncluded.change.advise(lt)
          {
            localCookie.Execute()
            {
              isIncluded.value = calculateIsIncluded()
            }
          }
          children.add(i, element)
        }
        AddRemove.Remove -> {

          val element = children.firstOrNull { e == it.value }
          //ourLogger.Trace($"from: ({this}) removed: ({element})")
          children.remove(element)
          isIncluded.value = calculateIsIncluded()
        }
      }
    }
  }

}

fun <T : Any> tableWithCheckBoxes(initialList: List<T>,
                                  lifetime: Lifetime,
                                  configuration: TreeConfiguration,
                                  id: String? = "",
                                  getLinePresentation: presentCheckedListLine<T>
): BeTreeGrid {
  return table(initialList, lifetime, getConfigurationWithCheckBoxes(configuration), id) { l, e, c ->
    present(l, e, c, getLinePresentation)
  }
}

private fun <T> present(
  lifetime: Lifetime,
  node: T,
  listNodeProperties: ListNodeProperties,
  linePresentation: presentCheckedListLine<T>
): List<BeControl> {
  val checkBoxNodeConfig = CheckBoxListNodeProperties(listNodeProperties)
  val checkBox = checkbox("", checkBoxNodeConfig.included, lifetime)
  listNodeProperties.enabled.flowInto(lifetime, checkBox.enabled)
  return linePresentation(lifetime, node, checkBoxNodeConfig).toMutableList().apply { add(0, checkBox) }
}

private fun getConfigurationWithCheckBoxes(configuration: TreeConfiguration): TreeConfiguration {
  return TreeConfiguration.columnDefinitions(configuration.columnElements?.toMutableList()?.apply
  {
    add(0,
        ColumnDefinition(
          "",
          getUnitSize(BeSizingType.Fit),
          BeAlignment.Center
        )
    )
  }, configuration.settings)
}
