package com.jetbrains.rd.ui.bedsl.dsl

import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.flowInto
import com.jetbrains.rd.util.reactive.hasValue
import com.jetbrains.rd.util.reactive.valueOrDefault
import com.jetbrains.rd.ui.bedsl.dsl.util.BeControlSizes
import com.jetbrains.rd.ui.bedsl.dsl.util.BeMarginsBuilder
import com.jetbrains.rd.ui.bedsl.extensions.descendants


fun BeControl.beInit(id: String? = "",
                     tooltip: String = "",
                     enabled: Boolean = true,
                     visibility: ControlVisibility = ControlVisibility.Visible) {
  this.enabled.value = enabled
  this.visible.set(visibility)
  this.controlId.value = id ?: ""
  this.tooltip.value = if (tooltip.isBlank()) null else tooltip
}

fun getUnitSize(beSizingType: BeSizingType, multiplier: Int = 1): BeUnitSize {
  return BeUnitSize(beSizingType, multiplier)
}

@ContainerMarker
open class GridBuilder {
  protected fun addElement(
    grid: BeGrid,
    content: () -> BeControl,
    size: BeSizingType
  ) {
    addElement(grid, content(), size)
  }

  protected fun addElement(
    grid: BeGrid,
    content: BeControl,
    size: BeSizingType
  ) {
    val oldItems = grid.items.valueOrNull
    val currentElements = oldItems ?: listOf()
    val newElements = currentElements.toMutableList().apply { add(content.getBeGridElement(size)) }
    grid.items.set(newElements)
  }
}

@ContainerMarker
class HorizontalGridBuilder(private val grid: BeGrid) : GridBuilder() {
  fun column(size: BeSizingType = BeSizingType.Fit, content: () -> BeControl) {
    addElement(grid, content, size)
  }

}

fun horizontalGrid(intercellSpacing: BeMarginType = BeMarginType.DefaultIntercellSpacing, init: HorizontalGridBuilder.() -> Unit): BeGrid {
  val grid = grid(GridOrientation.Horizontal, intercellSpacing)
  val builder = HorizontalGridBuilder(grid)
  builder.init()
  return grid
}

@DslMarker
annotation class ContainerMarker


class VerticalGridBuilder(private val grid: BeGrid) : GridBuilder() {

  fun row(size: BeSizingType = BeSizingType.Fit, content: () -> BeControl) {
    addElement(grid, content, size)
  }

  fun rows(rowContents: () -> List<BeControl>) {
    rowContents().forEach { addElement(grid, it, BeSizingType.Fit) }
  }

  operator fun plus(beControl: BeControl): VerticalGridBuilder {
    addElement(grid, beControl, BeSizingType.Fit)
    return this
  }
}

@ContainerMarker
class SpanGridBuilder(private val spanGrid: BeSpanGrid) {
  fun newRowChild(
    rowSize: BeSizingType = BeSizingType.Fit,
    contentSize: BeSizingType = BeSizingType.Fill,
    columnSpan: Int = -1,
    expandGrid: Boolean = true,
    content: () -> BeControl
  ) {
    val rowElement = BeSpanRowElement(BeUnitSize(rowSize, 1))
    addElementToRow(spanGrid, rowElement, columnSpan, content(), contentSize, expandGrid)
    if (rowElement.content.valueOrNull?.any() == true) {
      val oldItems = spanGrid.items.valueOrNull
      val currentElements = oldItems ?: listOf()
      val newElements = currentElements.toMutableList().apply { add(rowElement) }
      spanGrid.items.set(newElements)
    }
  }

  private fun addElementToRow(
    spanGrid: BeSpanGrid,
    rowElement: BeSpanRowElement,
    columnSpan: Int,
    content: BeControl,
    contentSize: BeSizingType,
    expandGrid: Boolean = true
  ) {
    var rdGridElements = mutableListOf<BeSpanColumnElement>()
    if (rowElement.content.valueOrNull?.any() == true)
      rdGridElements = rowElement.content.valueOrNull!!.toMutableList()
    var resultColumnSpan = columnSpan
    if (resultColumnSpan < 0)
      resultColumnSpan = spanGrid.columns.count()

    val beSpanColumnElement = BeSpanColumnElement(content.getBeGridElement(contentSize), resultColumnSpan)
    if (!expandGrid)
      rdGridElements.add(beSpanColumnElement)
    else {
      when (content) {
        is BeGrid -> {
          val gridElements = content.items.valueOrNull ?: return
          if (content.orientation == GridOrientation.Horizontal) {
            addGridElementsToSpan(rdGridElements, gridElements, content.intercellSpacing, beSpanColumnElement, resultColumnSpan)
          }
          else {
            for (gridElement in gridElements) {
              newRowChild(gridElement.size.unit, contentSize, resultColumnSpan, expandGrid) { gridElement.content }
            }
          }
        }
        is BeStyleControl -> {
          val contents = content.content
          if (content.margin.value != null && contents is BeGrid && contents.orientation == GridOrientation.Horizontal) {
            if (contents.items.valueOrNull == null) return
            val gridElements = content.margin.value!!.unwrapWrappedGrid(contents)
            addGridElementsToSpan(rdGridElements, gridElements, contents.intercellSpacing, beSpanColumnElement, resultColumnSpan)
          }
        }
        else -> rdGridElements.add(beSpanColumnElement)
      }
    }
    rowElement.content.set(rdGridElements)
  }

  private fun BeMarginModifier.unwrapWrappedGrid(grid: BeGrid): List<BeGridElement> {
    val result = mutableListOf<BeGridElement>()
    val wrapMargin = this.margin
    val beGridElements = grid.items.valueOrNull ?: return result
    val count = beGridElements.count()
    for (index in 0 until count) {
      var (left, top, right, bottom) = wrapMargin
      //(wrapMargin.left, wrapMargin.Top, wrapMargin.Right, wrapMargin.Bottom);
      if (index != 0) {
        left = BeMarginItem(BeMarginType.None, 1)
      }
      if (index != count - 1) {
        right = BeMarginItem(BeMarginType.None, 1)
      }

      val gridElement = beGridElements[index]
      result.add(gridElement.content.withMargin { BeMargin(left, top, right, bottom) }.getBeGridElement(gridElement.size))
    }
    return result
  }

  private fun getAvailableSpan(beGridElements: List<BeGridElement>, columnSpan: Int): Int {
    val gridElements = beGridElements
    val availableSpan = columnSpan - gridElements.count { it.size.unit == BeSizingType.Fit }
    val starCount = gridElements.count { it.size.unit == BeSizingType.Fill }
    if (starCount == 0 && gridElements.count() < columnSpan || (starCount != 0 && availableSpan == 0))
      return -1

    var starSpan = 1
    if (starCount != 0) {
      //Assertion.Assert(availableSpan % starCount == 0, "BeGrid items count cannot be applied to this span");
      starSpan = availableSpan / starCount
    }

    return starSpan
  }

  private fun addGridElementsToSpan(
    rdGridElements: MutableList<BeSpanColumnElement>,
    gridElements: List<BeGridElement>,
    gridIntercellSpacing: BeMarginType,
    beSpanColumnElement: BeSpanColumnElement,
    columnSpan: Int
  ) {
    //todo: alisa: do not forget about custom intercell spacing!
    val starSpan = getAvailableSpan(gridElements, columnSpan)
    if (starSpan < 0)
      rdGridElements.add(beSpanColumnElement)
    else {
      for (index in 0 until gridElements.count()) {
        val gridElement = gridElements[index]
        val span = if (gridElement.size.unit == BeSizingType.Fill) starSpan else 1
        rdGridElements.add(BeSpanColumnElement(if (spanGrid.intercellSpacing == gridIntercellSpacing || index == gridElements.count() - 1)
                                                 gridElement
                                               else {
          val margin = BeMarginsBuilder().margin(right = gridIntercellSpacing)
          gridElement.content.withMargin { margin }.getBeGridElement(gridElement.size)
        }, span))
      }
    }
  }

  fun newRowChildren(content: () -> BeControl) {

  }

  fun existingRowChild(
    contentSize: BeSizingType = BeSizingType.Fill,
    columnSpan: Int = -1,
    expandGrid: Boolean = false,
    content: () -> BeControl
  ) {
    val rowElement = (spanGrid.items.valueOrNull ?: mutableListOf()).lastOrNull()
    val columnElement = content()
    val customExpand = expandGrid && columnElement is BeGrid && columnElement.orientation != GridOrientation.Vertical
    val elementWithMargin = columnElement
    if (rowElement == null)
      return newRowChild(BeSizingType.Fit, contentSize, columnSpan, customExpand) { elementWithMargin }

    addElementToRow(spanGrid, rowElement, columnSpan, elementWithMargin, contentSize, customExpand)
  }
}

private fun spanGrid(columnSizes: String, intercellSpacing: BeMarginType = BeMarginType.DefaultIntercellSpacing): BeSpanGrid {
  val columns = columnSizes.split(',').map {
    if (it.endsWith("*")) getUnitSize(BeSizingType.Fill, Integer.getInteger(it.substringBefore("*"), 1)) else getUnitSize(BeSizingType.Fit)
  }
  val beSpanGrid = BeSpanGrid(columns, intercellSpacing)
  beSpanGrid.beInit()
  return beSpanGrid
}

fun verticalGrid(columnsSizes: String,
                 intercellSpacing: BeMarginType = BeMarginType.DefaultIntercellSpacing,
                 init: SpanGridBuilder.() -> Unit): BeSpanGrid {
  val grid = spanGrid(columnsSizes, intercellSpacing)
  val builder = SpanGridBuilder(grid)
  builder.init()
  return grid
}

fun verticalGrid(intercellSpacing: BeMarginType = BeMarginType.DefaultIntercellSpacing, init: VerticalGridBuilder.() -> Unit): BeGrid {
  val grid = grid(GridOrientation.Vertical, intercellSpacing)
  val builder = VerticalGridBuilder(grid)
  builder.init()
  return grid
}

private fun grid(orientation: GridOrientation = GridOrientation.Vertical,
                 intercellSpacing: BeMarginType = BeMarginType.DefaultIntercellSpacing): BeGrid {
  val emptyGrid = BeGrid(orientation, intercellSpacing)
  emptyGrid.beInit()
  emptyGrid.items.set(mutableListOf())
  return emptyGrid
}


fun BeGrid.setElement(control: BeControl, size: BeSizingType = BeSizingType.Fill): BeGrid {
  if (this.items.hasValue)
    this.items.valueOrDefault(emptyList())
  val newElements = mutableListOf<BeGridElement>()
  newElements.add(control.getBeGridElement(size))
  this.items.set(newElements)
  return this
}

private fun BeControl.getBeGridElement(size: BeSizingType = BeSizingType.Fit, multiplier: Int = 1): BeGridElement {
  return this.getBeGridElement(BeControlSizes.getUnitSize(size, multiplier))
}

private fun BeControl.getBeGridElement(size: BeUnitSize): BeGridElement {
  return BeGridElement(this, size)
}


/*@ContainerMarker
class DescriptionBuilder(private val grid: BeGrid){

}*/

fun description(description: String,
                lifetime: Lifetime,
                orientation: GridOrientation = GridOrientation.Horizontal,
                customTargetId: String? = null,
                target: () -> BeControl): BeGrid {

  val control = target()
  val label = descriptionLabel(description, control, customTargetId)
  val isBigControl = (control is BeGrid && control.orientation == GridOrientation.Vertical)
                     || control is BeSpanGrid || control is BeTreeGrid
  val labelControl = if (isBigControl) label/*InAutoGrid()*/ else label
  val grid = when (orientation) {
    GridOrientation.Horizontal -> horizontalGrid {
      column { labelControl }
      column(BeSizingType.Fill) { control }
    }
    GridOrientation.Vertical -> verticalGrid {
      row { labelControl }
      row(BeSizingType.Fill) { control }
    }
  }
  for (descendant in grid.descendants()) {
    if (descendant == grid) continue

    grid.enabled.flowInto(lifetime, descendant.enabled)
  }
  return grid
}
