package com.jetbrains.rd.ui.bedsl.dsl

import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.*
import com.jetbrains.rd.ui.bedsl.dsl.util.TreeConfiguration
import com.jetbrains.rd.ui.bedsl.extensions.bind
import com.jetbrains.rd.ui.bedsl.extensions.descendantsOfType
import com.jetbrains.rd.ui.bedsl.extensions.view

typealias presentTableLine<T> = (Lifetime, T, ListNodeProperties) -> List<BeControl>
typealias presentTreeLineWithContext<T, TRootContext> = (Lifetime, TRootContext?, T?, ListNodeProperties, TreeNodePropertiesWithChildren<T>) -> List<BeControl>
typealias presentTreeLine<T> = (Lifetime, T?, ListNodeProperties, TreeNodeProperties) -> List<BeControl>


fun <T : Any> table(
  items: List<T>,
  lifetime: Lifetime,
  configuration: TreeConfiguration,
  id: String? = "",
  presentLine: presentTableLine<T>
): BeTreeGrid {
  val tree = tree<T>(null, lifetime, configuration.copy(
    configuration.settings.copy().apply { showAsList = true },
    configuration.columnElements
  ).getBeConfiguration(), id) { lt, rc, e, lp, tp ->
    if (e == null) {
      when (items) {
        is ViewableList<T> -> items
        else -> ViewableList<T>().apply { addAll(items) }
      }.bindOneWay(lt, tp.children)
      return@tree mutableListOf<BeControl>()
    }
    presentLine(lt, e, lp)
  }
  return tree.apply {
    if (this.fakeRoot.children.all { l -> (l as? BeTreeGridLine)?.isSelected?.value == false }) {
      val beTreeGridLine = this.fakeRoot.children.first { l -> l as? BeTreeGridLine != null && l.enabled.value }
      (beTreeGridLine as BeTreeGridLine).isSelected.value = true
    }
  }
}


fun <T : Any> tree(
  items: List<T>,
  lifetime: Lifetime,
  configuration: TreeConfiguration,
  getChildren: (T?) -> List<T>,
  id: String? = "",
  presentLine: presentTreeLine<T>
): BeTreeGrid {
  val viewableList: ViewableList<T> = when (items) {
    is ViewableList<T> -> items
    else -> ViewableList<T>().apply { addAll(items) }
  }
  return tree(viewableList, lifetime, configuration, id, {
    when (val children = getChildren(it)) {
      is ViewableList<T> -> children
      else -> ViewableList<T>().apply { addAll(children) }
    }
  }, presentLine)
}

fun <T : Any> tree(
  items: ViewableList<T>,
  lifetime: Lifetime,
  configuration: TreeConfiguration,
  id: String?,
  getChildren: (T?) -> ViewableList<T>,
  presentLineWithContext: presentTreeLine<T>
): BeTreeGrid {
  return tree<T>(null, lifetime, configuration, id)
  { nodeLt, r, e, lp, tp ->

    if (r == e) {
      items.bindOneWay(nodeLt, tp.children)
      return@tree listOf<BeControl>()
    }
    val childrenEvents = getChildren(e)
    bindOneWay(nodeLt, childrenEvents, tp.children)
    presentLineWithContext(nodeLt, e, lp, TreeNodeProperties(tp.expanded.value))
  }
}

fun <T : Any> ViewableList<T>.bindOneWay(
  lifetime: Lifetime,
  secondList: ViewableList<T>
) {
  bindOneWay(lifetime, this, secondList)
}

fun <T : Any> bindOneWay(
  lifetime: Lifetime,
  firstList: ViewableList<T>,
  secondList: ViewableList<T>
) {
  firstList.view(lifetime) { lt, index, c, k ->
    when (k) {
      AddRemove.Add -> secondList.add(index, c)
      AddRemove.Remove -> secondList.remove(c)
    }
  }
}


private fun <T : Any> tree(
  root: T?,
  lifetime: Lifetime,
  configuration: TreeConfiguration,
  id: String?,
  presentLineWithContext: presentTreeLineWithContext<T, T>
): BeTreeGrid {
  return tree(root, lifetime, configuration.getBeConfiguration(), id, presentLineWithContext)
}


private fun <T : Any> tree(
  root: T?,
  lifetime: Lifetime,
  configuration: BeTreeConfiguration,
  id: String? = "",
  presentLineWithContext: presentTreeLineWithContext<T, T>
): BeTreeGrid {
  return tree(Property(root), lifetime, { _, t -> t }, configuration, id, presentLineWithContext)
}

fun <TRootContext, T : Any> tree(
  rootContext: IProperty<TRootContext?>,
  lifetime: Lifetime,
  getRootNode: (Lifetime, TRootContext?) -> T?,
  configuration: BeTreeConfiguration,
  id: String? = "",
  presentLineWithContext: presentTreeLineWithContext<T, TRootContext>
): BeTreeGrid {
  val fakeRoot = BeTreeGridLine(mutableListOf())
  //val idMap = di
  rootContext.view(lifetime) { rootLifetime, currentRootContext ->
    val present = presentLineWithContext
    //if (configuration.BackendContextMenu != null) todo
    BuildTreeNode(rootLifetime, currentRootContext, getRootNode(rootLifetime, currentRootContext), present, { fakeRoot })
  }
  return BeTreeGrid(fakeRoot, configuration).apply {
    this.beInit(id)
    validationResult.set(ValidationResult())
    if (configuration.backendContextMenu != null) {
      if (id.isNullOrBlank()) throw AssertionError("Control id should be specified")
      // this.extend(idMap)
    }
  }
}


fun <TRootContext, T : Any> BuildTreeNode(
  lifetime: Lifetime,
  rootContext: TRootContext?,
  model: T?,
  presentLineWithContext: presentTreeLineWithContext<T, TRootContext>,
  nodeBuilder: (List<BeControl>) -> BeTreeGridLine
): BeTreeGridLine {
  val listNodeProperties = ListNodeProperties()
  val treeNodeProperties = TreeNodePropertiesWithChildren<T>()
  val nodePresentation = presentLineWithContext(lifetime, rootContext, model, listNodeProperties, treeNodeProperties)
  nodePresentation.flatMap { it.descendantsOfType(BeControlWithValidation::class.java).asIterable() }.forEach {
    it.handledByParent.value = true
  }
  val node = nodeBuilder(nodePresentation)
  node.beInit(listNodeProperties.id.value)
  bindListProperties(lifetime, listNodeProperties, node)
  bindTreeProperties(lifetime, rootContext, presentLineWithContext, treeNodeProperties, node)
  return node
}

private fun <TRootContext, T : Any> BuildTreeNode(
  lifetime: Lifetime,
  rootContext: TRootContext?,
  model: T?,
  presentLineWithContext: presentTreeLineWithContext<T, TRootContext>
): BeTreeGridLine {
  return BuildTreeNode(lifetime, rootContext, model, presentLineWithContext, { BeTreeGridLine(it) })
}

private fun <TRootContext, T : Any> bindTreeProperties(
  lifetime: Lifetime,
  rootContext: TRootContext?,
  presentLineWithContext: presentTreeLineWithContext<T, TRootContext>,
  treeNodeProperties: TreeNodePropertiesWithChildren<T>,
  node: BeTreeGridLine
) {
  treeNodeProperties.expanded.bind(lifetime, node.isExpanded)
  subscribeChildren(lifetime, rootContext, treeNodeProperties.children, presentLineWithContext, node.children)
}

private fun <TRootContext, T : Any> subscribeChildren(
  lifetime: Lifetime,
  rootContext: TRootContext,
  children: ViewableList<T>,
  presentLineWithContext: presentTreeLineWithContext<T, TRootContext>,
  nodeChildren: IMutableViewableList<BeControl>
) {
  children.view(lifetime) { nl, i, v, k ->
    when (k) {
      AddRemove.Add -> {
        nodeChildren.add(i, BuildTreeNode(nl!!, rootContext, v, presentLineWithContext))
      }
      AddRemove.Remove -> {
        nodeChildren.removeAt(i)
      }
    }
  }
}


private fun bindListProperties(lifetime: Lifetime, listNodeProperties: ListNodeProperties, node: BeTreeGridLine) {
  listNodeProperties.selected.bind(lifetime, node.isSelected)
  node.signal.advise(lifetime, { listNodeProperties.signal.fire(true) })
}

class ListNodeProperties {
  val id: Property<String> = Property("")
  val selected: Property<Boolean> = Property(false)
  val enabled: Property<Boolean> = Property(true)
  val signal: Signal<Boolean> = Signal()
}

open class TreeNodeProperties(expanded: Boolean) {
  val expanded: Property<Boolean> = Property(expanded)
}

class TreeNodePropertiesWithChildren<T : Any> : TreeNodeProperties(false) {
  val children: ViewableList<T> = ViewableList()
}
//todo: validation binding
//todo: data rule provider