package com.jetbrains.rd.ui.bedsl.dsl

import com.intellij.openapi.actionSystem.AnAction
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.dsl.util.CommonBehavior
import com.jetbrains.rd.ui.bedsl.dsl.util.Description
import com.jetbrains.rd.ui.bedsl.dsl.util.getAnnotation
import com.jetbrains.rd.ui.bedsl.extensions.getExtended
import com.jetbrains.rd.ui.bedsl.predefinedBehaviorButton
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.Property
import com.jetbrains.rd.util.reactive.flowInto
import com.jetbrains.rd.ui.bedsl.actionButton as getActionButton

class BeToolbarBuilder(val toolbar: BeToolbar) {
  fun <T : Any> buttonWithAction(
    action: BeTableAddAction,
    getNewElement: (Int) -> T,
    canExecute: (Int) -> Boolean = { true },
    id: String = "",
    style: BeButtonStyle = BeButtonStyle.ICON,
    customTooltip: String = ""): BeToolbar {
    val data = toolbar.getExtended<TreeGridActionData<T>>("ToolbarTableActionData") ?: return toolbar
    //Assertion.AssertNotNull(data, "This method should be called after GetSingleSelectionListWithToolbar or TreeGridActionData should be specified in extension")
    val disableWhenEmpty = action == BeTableAddAction.CLONE

    val enabledProperty = Property(true)
    val onClick = initOnClick(data, action, getNewElement)
    initActions(data, canExecute, disableWhenEmpty, enabledProperty)
    return addItem(getButton(action, id, style, customTooltip, data, onClick, enabledProperty))
  }

  private inline fun <reified E : Enum<E>> shortcut(action: Enum<E>): BeCommonBehavior {
    return action.getAnnotation<CommonBehavior, E>()?.behavior ?: BeCommonBehavior.NONE
  }

  fun addItem(control: BeControl): BeToolbar {
    return toolbar.apply { leftItems.add(control) }
  }

  fun <T : Any> buttonWithAction(
    action: BeTableAction,
    onExecute: (Int) -> Unit = {},
    canExecute: (Int) -> Boolean = { true },
    id: String = "",
    style: BeButtonStyle = BeButtonStyle.ICON,
    customTooltip: String = ""): BeToolbar {
    val data = toolbar.getExtended<TreeGridActionData<T>>("ToolbarTableActionData") ?: return toolbar
    //Assertion.AssertNotNull(data, "This method should be called after GetSingleSelectionListWithToolbar or TreeGridActionData should be specified in extension")

    val enabledProperty = Property(true)
    val onClick = initOnClick(data, action, onExecute)
    val button = getButton(action, id, style, customTooltip, data, onClick, enabledProperty)
    initActions(data, canExecute, true, enabledProperty)
    return addItem(button)
  }

  /*fun AddButton(
      icon : IconModel,
      tooltip : String,
      onClick:()->Unit,
      lifetime: Lifetime,
      enabled:Property<Boolean> = Property(true),
       id: String = "",
       withShortcut:Boolean = true,
       style:BeButtonStyle = BeButtonStyle.DEFAULT): BeToolbar {
      val button = toolbarButton(icon, tooltip, lifetime, onClick, id)
      enabled.bind(lifetime, button.enabled)
      return AddItem(button)
  }*/


  private inline fun <T : Any, reified E : Enum<E>> getButton(action: Enum<E>,
                                                              id: String,
                                                              style: BeButtonStyle,
                                                              customTooltip: String,
                                                              data: TreeGridActionData<T>,
                                                              noinline onClick: () -> Unit = {},
                                                              enabledProperty: Property<Boolean>): BeAbstractButton {
    val annotationDescription = action.getAnnotation<Description, E>()?.description ?: "Any"
    val description = if (customTooltip.isEmpty()) annotationDescription else customTooltip
    //var icon = action.IconId(data.IconHost)
    val label = labelWithShortcut(description) //, iconId = icon)
    val button = predefinedBehaviorButton(label, data.lifetime, shortcut(action),
                                          style, id, onClick)
    if (style == BeButtonStyle.ICON)
      button.tooltip.value = label.getText()
    enabledProperty.flowInto(data.lifetime, button.enabled)
    return button
  }


  private fun <T : Any> initActions(
    data: TreeGridActionData<T>,
    canExecute: (Int) -> Boolean = { true },
    disableWhenEmpty: Boolean,
    enabledProperty: Property<Boolean>
  ) {
    data.selectedElement.view(data.lifetime)
    { _, it ->
      if (it == null) return@view
      val index = data.collection.indexOf(it)
      enabledProperty.value = canExecute(index)
    }
    if (disableWhenEmpty) {
      data.collection.change.advise(data.lifetime
      ) {
        val count = data.collection.count()
        val index = data.collection.indexOf(data.selectedElement.value)
        enabledProperty.value = count > 0 && canExecute(index)
      }
    }
  }

  private fun <T : Any> onCreate(data: TreeGridActionData<T>, getNewElement: (Int) -> T?) {
    val value = data.selectedElement.value
    data.selectedElement.value = null
    val index = data.collection.indexOf(value)
    val newElement = getNewElement(index) ?: return
    data.collection.add(newElement)
    data.selectedElement.value = newElement
  }

  private fun <T : Any> onCopy(data: TreeGridActionData<T>, getNewElement: (Int) -> T?) {
    val value = data.selectedElement.value
    data.selectedElement.value = null
    val index = data.collection.indexOf(value)
    if (index < 0) return
    val newElement = getNewElement(index) ?: return
    data.collection.add(index + 1, newElement)
    data.selectedElement.value = newElement
  }

  private fun <T : Any> onMove(
    data: TreeGridActionData<T>,
    movePosition: Int,
    onExecute: (Int) -> Unit = {}
  ) {
    val value = data.selectedElement.value ?: return
    data.selectedElement.value = null
    val index = data.collection.indexOf(value)
    if (index + movePosition < 0 || index + movePosition >= data.collection.count()) return
    onExecute(index)

    data.collection.remove(value)
    data.collection.add(index + movePosition, value)
    data.selectedElement.value = value
  }

  private fun <T : Any> onRemove(
    data: TreeGridActionData<T>,
    onExecute: (Int) -> Unit = {}
  ) {
    val value = data.selectedElement.value
    data.selectedElement.value = null
    val index = data.collection.indexOf(value)
    if (index < 0) return
    onExecute(index)
    data.collection.remove(value)
    data.selectedElement.value = if (data.collection.count() > index)
      data.collection[index]
    else data.collection.lastOrNull()
  }

  private fun <T : Any> initOnClick(
    data: TreeGridActionData<T>,
    action: BeTableAddAction,
    getNewElement: (Int) -> T?
  ): () -> Unit {
    var onClick: () -> Unit = {}
    if (action == BeTableAddAction.ADD || action == BeTableAddAction.ADD_AFTER_SELECTED) {
      onClick = { onCreate(data, getNewElement) }
    }

    if (action == BeTableAddAction.CLONE) {
      onClick = { onCopy(data, getNewElement) }
    }

    return {
      onClick()
      data.list.focus.fire(Unit)
    }
  }

  private fun <T : Any> initOnClick(
    data: TreeGridActionData<T>,
    action: BeTableAction,
    onExecute: (Int) -> Unit
  ): () -> Unit {
    var onClick: () -> Unit = {}

    if (action == BeTableAction.REMOVE) {
      onClick = { onRemove(data, onExecute) }
    }

    if (action == BeTableAction.MOVE_UP || action == BeTableAction.MOVE_DOWN) {
      val movePosition = if (action == BeTableAction.MOVE_UP) -1 else 1
      onClick = { onMove(data, movePosition, onExecute) }
    }
    return {
      onClick()
      data.list.focus.fire(Unit)
    }
  }

  fun actionButton(action: AnAction, lifetime: Lifetime): BeToolbar {
    val button = getActionButton(action, lifetime)
    return addItem(button)
  }

  fun actionButton(actionId: String, lifetime: Lifetime): BeToolbar {
    val button = getActionButton(actionId, lifetime)
    return addItem(button)
  }
}

enum class BeTableAction {
  @Description("_Delete")
  @CommonBehavior(BeCommonBehavior.TABLE_DELETE)
  REMOVE,

  @Description("Move _up")
  @CommonBehavior(BeCommonBehavior.TABLE_MOVE_UP)
  MOVE_UP,

  @Description("Move _down")
  @CommonBehavior(BeCommonBehavior.TABLE_MOVE_DOWN)
  MOVE_DOWN
}

enum class BeTableAddAction {

  @Description("_Add")
  @CommonBehavior(BeCommonBehavior.TABLE_ADD)
  ADD,

  @Description("_Clone")
  CLONE,

  @Description("_Add")
  @CommonBehavior(BeCommonBehavior.TABLE_ADD)
  ADD_AFTER_SELECTED
}


