package com.jetbrains.rd.ui.bedsl.extensions

import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.reactive.*
import kotlin.collections.map

fun <T : BeControl> BeControl.getBeControlById(name: String): T {
  return tryGetBeControlById(name) ?: throw AssertionError("Control with name $name doesn't exist in ${this.javaClass.name}")
}

fun <T : BeControl> BeControl.tryGetBeControlById(name: String): T? {

  @Suppress("UNCHECKED_CAST")
  return descendants().firstOrNull { it.controlId.value == name } as? T
}

fun BeControl.contentToString(separator: CharSequence = ", ", richTextPartsSeparator: CharSequence = ", "): String {
  return descendants().mapNotNull {
    when (it) {
        is BeRichText -> it.text.valueOrNull?.parts?.joinToString(richTextPartsSeparator) { it.text }
      is BeLabel -> it.text.value
      else -> null
    }
  }.joinToString(separator)
}

@Suppress("UNCHECKED_CAST")
fun <T : BeControl> BeControl.descendantsOfType(clazz: Class<T>): Sequence<T> {
  return descendants().filter { clazz.isInstance(it) }.map { it as T }
}

fun <T : Any> IOptProperty<List<T>>.valueOrEmpty(): List<T> = valueOrNull ?: emptyList()

/**
 * @param additionalGetChildren This func is helpful when you are using custom BeControl-derived views in plugins which can't be checked in common code.
 */
fun BeControl.descendants(additionalGetChildren: ((BeControl) -> Sequence<BeControl>)? = null): Sequence<BeControl> = sequence {
  @Suppress("UnnecessaryVariable") // Alias to shorten the name inside of the method body
  val getChildren = additionalGetChildren
  yield(this@descendants)
  when (this@descendants) {
    is BeDialog -> yieldAll(content.valueOrNull?.descendants(getChildren) ?: emptySequence())
    is BeScrollablePanel -> yieldAll(content.descendants() ?: emptySequence())
    is BeBehaviorControl -> yieldAll(content.descendants())
    is BeTabbedControl -> tabs.forEach { t -> yieldAll(t.content.descendants(getChildren)) }
    is BeCollapsiblePanel -> yieldAll(content.descendants(getChildren))
    is BeDynamicControlContainer -> content.valueOrNull?.let{yieldAll(content.valueOrNull!!.descendants(getChildren))}

    is BeGrid -> items.valueOrEmpty().forEach { yieldAll(it.content.descendants(getChildren)) }
    is BeStyleControl -> yieldAll(content.descendants(getChildren))
    is BeSpanGrid -> items.valueOrEmpty()
      .forEach {
        it.content.valueOrEmpty()
          .forEach {
            yieldAll(it.content.content.descendants(getChildren))
          }
      }
    is BeTreeGridLine -> columnElements.forEach { yieldAll(it.descendants(getChildren)) }
    is BeComboBox -> values.forEach { yieldAll(it.descendants(getChildren)) }
    is BeTreeGrid -> fakeRoot.children.forEach { yieldAll(it.descendants(getChildren)) }
    is BeSplitControl -> yieldAll(firstPanel.descendants().plus(secondPanel.descendants()))
    is BeButton -> yieldAll(content.descendants(getChildren))
    is BeRadioButton -> yield(description)
    is BeCheckbox -> yield(description)
    is BeToolbar -> {
      leftItems.plus(content).forEach { yieldAll(it.descendants(getChildren)) }
    }
    else -> {
      getChildren?.let {
        getChildren(this@descendants).forEach { yieldAll(it.descendants(getChildren)) }
      }
    }
  }
}

fun BeTreeGridLine.subtree(): Sequence<BeTreeGridLine> = sequence {
  yield(this@subtree)
  this@subtree.children
    .map { it as BeTreeGridLine }
    .forEach { yieldAll(it.subtree()) }
}

fun BeTreeGrid.tryGetElementByIndex(index: Int): BeTreeGridLine? {
  return this.fakeRoot.children.asSequence()
    .flatMap { (it as BeTreeGridLine).subtree() }.elementAtOrNull(index)
}
private fun IPropertyBase<BeControl>.applyToLiveDescendants(lifetime: Lifetime,
                                                            action: (BeControl) -> Unit) {
    var nested : LifetimeDefinition? = null
    this.advise(lifetime){
      if(nested != null)
        nested?.terminate()
      nested = lifetime.createNested()
      it.applyToLiveDescendants(nested!!.lifetime, action)
    }
}

private fun <T> IOptProperty<List<T>>.applyToLiveDescendants(lifetime: Lifetime, action: ((T, Lifetime) -> Unit)) {
  var nested: LifetimeDefinition? = null
  this.advise(lifetime) {
    if (nested != null)
      nested?.terminate()
    nested = lifetime.createNested()
    it.forEach { child ->  action(child, nested!!.lifetime)}
  }
}
private fun IViewableList<BeControl>.applyToLiveDescendants(lifetime: Lifetime,
                                                            action: (BeControl) -> Unit) {
  val nestedLifetimes = mutableListOf<LifetimeDefinition>()
  this.adviseAddRemove(lifetime) { e, i, it ->
    if (e == AddRemove.Add) {
      val nested = lifetime.createNested()
      nestedLifetimes.add(i, nested)
      it.applyToLiveDescendants(nested.lifetime, action)
    }
    if (e == AddRemove.Remove) {
      val nested = nestedLifetimes[i]
      nested.terminate()
      nestedLifetimes.removeAt(i)
    }
  }
}

fun BeControl.applyToLiveDescendants(lifetime: Lifetime, action: ((BeControl) -> Unit)) {
  action(this)
  when (this@applyToLiveDescendants) {
    is BeDialog -> content.applyToLiveDescendants(lifetime, action)
    is BeScrollablePanel -> content.applyToLiveDescendants(lifetime, action)
    is BeBehaviorControl -> content.applyToLiveDescendants(lifetime, action)
    is BeSplitControl -> {
      firstPanel.applyToLiveDescendants(lifetime, action)
      secondPanel.applyToLiveDescendants(lifetime, action)
    }
    is BeCollapsiblePanel -> content.applyToLiveDescendants(lifetime, action)
    is BeButton -> content.applyToLiveDescendants(lifetime, action)
    is BeCheckbox -> description.applyToLiveDescendants(lifetime, action)
    is BeRadioButton -> description.applyToLiveDescendants(lifetime, action)
    is BeTabbedControl -> tabs.applyToLiveDescendants(lifetime, action)
    is BeStyleControl -> content.applyToLiveDescendants(lifetime, action)
    is BeTreeGridLine -> {
      columnElements.forEach { it.applyToLiveDescendants(lifetime, action) }
      children.applyToLiveDescendants(lifetime, action)
    }
    is BeComboBox -> values.applyToLiveDescendants(lifetime, action)
    is BeDynamicControlContainer -> content.applyToLiveDescendants(lifetime, action)
    is BeTreeGrid -> fakeRoot.children.applyToLiveDescendants(lifetime, action)
    is BeToolbar -> {
      leftItems.applyToLiveDescendants(lifetime, action)
      content.applyToLiveDescendants(lifetime, action)
    }
    is BeGrid -> items.applyToLiveDescendants(lifetime) { it, lt -> it.content.applyToLiveDescendants(lt, action) }
    is BeSpanGrid -> items.applyToLiveDescendants(lifetime) { it, rowLt ->
      it.content.applyToLiveDescendants(rowLt) { e, colLt ->
        e.content.content.applyToLiveDescendants(colLt, action)
      }
    }
    /*    else -> {
          getChildren?.let {
            getChildren(this@descendants).forEach { yieldAll(it.descendants(getChildren)) }
          }*/
  }
}

  @Suppress("UNCHECKED_CAST")
  fun <T : BeControl> BeControl.applyToLiveDescendantsOfType(clazz: Class<T>, lifetime: Lifetime, action:((T) -> Unit)) {
    applyToLiveDescendants(lifetime) {
      if (clazz.isInstance(it))
        action(it as T)
    }
  }

fun IVoidSignal.grouppedFire(mergingQueue: MergingUpdateQueue){
  mergingQueue.cancelAllUpdates()
  mergingQueue.queue(object : Update("") {
  override fun run() {this@grouppedFire.fire()} })
}
