// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable

import com.intellij.ide.DataManager
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.CommonShortcuts
import com.intellij.openapi.actionSystem.CustomShortcutSet
import com.intellij.openapi.actionSystem.IdeActions
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.keymap.KeymapUtil
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.popup.ActiveIcon
import com.intellij.openapi.ui.popup.JBPopup
import com.intellij.openapi.util.Computable
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.Pair
import com.intellij.ui.ActiveComponent
import com.intellij.ui.ScreenUtil
import com.intellij.ui.popup.AbstractPopup
import com.intellij.util.ui.EmptyIcon
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.dsl.getText
import com.jetbrains.rd.ui.util.getShortcut
import com.jetbrains.rd.ui.util.getStringPresentation
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.fire
import java.awt.AWTEvent
import java.awt.event.ActionEvent
import java.awt.event.ActionListener
import java.awt.event.KeyEvent
import java.awt.event.MouseEvent
import javax.swing.*
import javax.swing.plaf.LayerUI

fun BeAbstractPopup.showPopup(project: Project, lifetime: Lifetime) {
  showPopup(project, lifetime, false)
}

fun BeAbstractPopup.showPopup(project: Project, lifetime: Lifetime, isReflection: Boolean) {
  val viewControl =
    if (!isReflection) PopupViewControlImpl(lifetime, this, project)
    else ReflectedPopupViewControlImpl(lifetime, this, project)

  /*if (modality == DialogModality.MODAL) {
    viewControl.withPumpProtocol { viewControl.loadAndShowPopup() }
  }
  if (modality != DialogModality.MODAL) {
  */
  viewControl.loadAndShowPopup()
  //}
}

interface PopupViewControl {
  fun loadAndShowPopup()
}

class PopupViewControlImpl(val lifetime: Lifetime,
                           private val model: BeAbstractPopup,
                           private val project: Project) : PopupViewControl {
  private lateinit var myPopup: AbstractPopup
  private val myKeyboardActions: MutableList<Pair<ActionListener, KeyStroke>> = mutableListOf()

  private fun loadPopup(): JBPopup {
    val content = model.content./*(scrollablePanel { label("Control is not loaded") }).*/getView(lifetime, project)
    addChooserActions(content)

    myPopup = object : AbstractPopup() {
      init {
        val cancelCallback = Computable {
          when (model.result.valueOrNull) {
            DialogResult.CLOSE -> true
            else -> {
              model.result.set(DialogResult.CANCEL)
              false // signal for closing should come from backend, cause there are some cancel callbacks which should be called, and may cance cancelliing of popup
            }
          }
          //we already checked this
        }
        super.init(
          project,
          content,
          findFirstFocusableControl(content), /*beAbstractPopup.requestFocus*/
          model.isFocused,
          model.focusable,
          model.isMovable,
          model.controlId.value.takeIf { it.isNotEmpty() },
          model.isResizable,
          model.title?.getText(),
          cancelCallback,
          model.cancelOnClickOutside,
          LinkedHashSet(), // myListeners,
          model.layoutPersistenceMode == DialogLayoutPersistenceMode.SIZE_AND_POSITION,
          object : ActiveComponent.Adapter() {
            override fun getComponent(): JComponent {
              return model.rightTitleToolbar.getView(lifetime, this@PopupViewControlImpl.project)
            }
          },
          null, //rarely used
          null, //myCancelOnMouseOutCallback, todo?
          model.cancelOnOtherWindowOpen,
          ActiveIcon(EmptyIcon.ICON_0), //rarely used
          model.cancelKeyEnabled,
          true, //myLocateByContent,
          true,
          null, //myMinSize,
          0f,
          null, //myMaskProvider,
          true, //myInStack,
          model.modality == DialogModality.MODAL,
          arrayOfNulls(0), //focus owners
          model.adText,
          SwingConstants.LEFT,
          false,
          myKeyboardActions,
          model.leftTitleToolbar.getView(lifetime, this@PopupViewControlImpl.project),
          null, //myPinCallback,
          false, //myMayBeParent,
          true, //myShowShadow,
          true, //myShowBorder,
          null, //myBorderColor,
          model.cancelOnWindowDeactivation, //myCancelOnWindowDeactivation,
          null //myKeyEventHandler
        )
      }
    }

    // popup.setNormalWindowLevel(myNormalWindowLevel)
    myPopup.setOkHandler {
      model.accept.fire()
      model.result.set(DialogResult.OK)
      model.result.set(DialogResult.CANCEL)

    }
    model.result.advise(lifetime) {
      if (it == DialogResult.CLOSE && !myPopup.isDisposed)
        myPopup.cancel(null)
    }
    model.title?.text?.change?.advise(lifetime){
      myPopup.setCaption(it)
    }
    Disposer.register(ApplicationManager.getApplication(), myPopup)
    return myPopup
  }

  private fun addChooserActions(content: JComponent) {

    registerClosePopupKeyboardAction(content, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), false)
    if (model.closeOnEnter) {
      registerClosePopupKeyboardAction(content, KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0), true)
    }
    else {
      registerKeyboardAction(KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0),
                             ActionListener { model.accept.fire() })
    }

    /*if (myUserData != null) {
      popup.setUserData(myUserData)
    }*/
  }

  private fun registerKeyboardAction(keyStroke: KeyStroke?, actionListener: ActionListener) {
    val pair = com.intellij.openapi.util.Pair.create(actionListener, keyStroke!!)
    myKeyboardActions.add(pair)
  }

  private fun registerClosePopupKeyboardAction(component: JComponent, keyStroke: KeyStroke, shouldPerformAction: Boolean) {
    registerPopupKeyboardAction(component, keyStroke, object : AbstractAction() {
      override fun actionPerformed(e: ActionEvent) {
        // if (!shouldPerformAction /*&& myChooserComponent.checkResetFilter()*/) return
        if (shouldPerformAction) {
          myPopup.closeOk(null)
        }
        else {
          myPopup.cancel(null)
        }
      }
    })
  }

  private fun registerPopupKeyboardAction(component: JComponent, keyStroke: KeyStroke, action: AbstractAction) {
    component.registerKeyboardAction(action, keyStroke, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT)
  }

  override fun loadAndShowPopup() {
    val popup = loadPopup()
    if(popup.isDisposed) return
    this.model.resize.advise(lifetime){
      popup.pack(true, true)
      val content = popup.content
      val window = SwingUtilities.windowForComponent(content)
      val bounds = window?.bounds
      if (bounds != null) {
        ScreenUtil.fitToScreen(bounds)
        window.bounds = bounds
      }

    }
    this.model.loaded.advise(lifetime) {
      DataManager.getInstance().dataContextFromFocusAsync.onSuccess { dataContext ->
        SwingUtilities.invokeLater {
          if(popup.isDisposed) return@invokeLater
          when(model.context.valueOrNull){
            is BeFocusCenterPopupContext ->{
              popup.showInFocusCenter()
            }
            is BeWindowCenterPopupContext ->{
              popup.showCenteredInCurrentWindow(project)
            }
            is BeScreenPointPopupContext ->{
              //popup.showInScreenCoordinates()
              popup.showInBestPositionFor(dataContext)
            }
            else -> {
              popup.showInBestPositionFor(dataContext)
            }

          }

        }
      }
    }
  }
}

class ReflectedPopupViewControlImpl(val lifetime: Lifetime,
                                    private val model: BeAbstractPopup,
                                    private val project: Project) : PopupViewControl {
  private lateinit var myPopup: AbstractPopup
  private val myKeyboardActions: MutableList<Pair<ActionListener, KeyStroke>> = mutableListOf()

  private fun loadPopup(): JBPopup {
    val originalContent = model.content.getView(lifetime, project)
    val layerUI = MyLayerUISubclass()
    val content = JLayer(originalContent, layerUI)

    suppressKeyEvents(content)
    addCloseAction(content)

    myPopup = object : AbstractPopup() {
      init {
        val cancelCallback = Computable { true }
        super.init(
          project,
          content,
          findFirstFocusableControl(content),
          false,
          false,
          model.isMovable,
          model.controlId.value.takeIf { it.isNotEmpty() },
          model.isResizable,
          model.title?.getText(),
          cancelCallback,
          true,
          LinkedHashSet(), // myListeners,
          model.layoutPersistenceMode == DialogLayoutPersistenceMode.SIZE_AND_POSITION,
          null,
          null, //rarely used
          null, //myCancelOnMouseOutCallback
          false,
          ActiveIcon(EmptyIcon.ICON_0), //rarely used
          false,
          true, //myLocateByContent,
          true,
          null, //myMinSize,
          0f,
          null, //myMaskProvider,
          true, //myInStack,
          model.modality == DialogModality.MODAL,
          arrayOfNulls(0), //focus owners
          model.adText,
          SwingConstants.LEFT,
          false,
          myKeyboardActions,
          null,
          null, //myPinCallback,
          false, //myMayBeParent,
          true, //myShowShadow,
          true, //myShowBorder,
          null, //myBorderColor,
          false, //myCancelOnWindowDeactivation,
          null //myKeyEventHandler
        )
      }
    }

    model.result.advise(lifetime) {
      if (it == DialogResult.CLOSE && !myPopup.isDisposed)
        myPopup.cancel(null)
    }

    Disposer.register(ApplicationManager.getApplication(), myPopup)
    return myPopup
  }

  private fun suppressKeyEvents(content: JLayer<JComponent>) {
    val shortcutsTexts = arrayOf(
      KeymapUtil.getActiveKeymapShortcuts(IdeActions.ACTION_EDITOR_MOVE_CARET_LEFT),
      KeymapUtil.getActiveKeymapShortcuts(IdeActions.ACTION_EDITOR_MOVE_CARET_RIGHT),
      CommonShortcuts.getMoveUp(),
      CommonShortcuts.getMoveDown(),
      CommonShortcuts.getMovePageUp(),
      CommonShortcuts.getMovePageDown(),
      KeymapUtil.getActiveKeymapShortcuts(IdeActions.ACTION_EDITOR_TEXT_START),
      KeymapUtil.getActiveKeymapShortcuts(IdeActions.ACTION_EDITOR_TEXT_END),
    ).flatMap { it.getStringPresentation() }

    val shortcuts = shortcutsTexts.mapNotNull { getShortcut(it, content, true) }
    val shortcutsSet = CustomShortcutSet(*shortcuts.toTypedArray())
    shortcutsSet.let {
      object : DumbAwareAction() {
        override fun actionPerformed(e: AnActionEvent) {
          // suppress
        }

        override fun displayTextInToolbar(): Boolean = false

        override fun update(e: AnActionEvent) {}
      }.registerCustomShortcutSet(it, content)
    }
  }

  private fun addCloseAction(content: JComponent) {
    registerPopupKeyboardAction(content, KeyStroke.getKeyStroke(KeyEvent.VK_ESCAPE, 0), object : AbstractAction() {
      override fun actionPerformed(e: ActionEvent) {
        model.result.set(DialogResult.CANCEL)
        myPopup.cancel(null)
      }
    })
  }

  private fun registerPopupKeyboardAction(component: JComponent, keyStroke: KeyStroke, action: AbstractAction) {
    component.registerKeyboardAction(action, keyStroke, JComponent.WHEN_ANCESTOR_OF_FOCUSED_COMPONENT)
  }

  override fun loadAndShowPopup() {
    val popup = loadPopup()
    if (popup.isDisposed) return
    this.model.resize.advise(lifetime) {
      popup.pack(true, true)
      val content = popup.content
      val window = SwingUtilities.windowForComponent(content)
      val bounds = window?.bounds
      if (bounds != null) {
        ScreenUtil.fitToScreen(bounds)
        window.bounds = bounds
      }

    }
    this.model.loaded.advise(lifetime) {
      DataManager.getInstance().dataContextFromFocusAsync.onSuccess { dataContext ->
        SwingUtilities.invokeLater {
          if (popup.isDisposed) return@invokeLater
          when (model.context.valueOrNull) {
            is BeFocusCenterPopupContext -> {
              popup.showInFocusCenter()
            }
            is BeWindowCenterPopupContext -> {
              popup.showCenteredInCurrentWindow(project)
            }
            is BeScreenPointPopupContext -> {
              //popup.showInScreenCoordinates()
              popup.showInBestPositionFor(dataContext)
            }
            else -> {
              popup.showInBestPositionFor(dataContext)
            }

          }

        }
      }
    }
  }
}

internal class MyLayerUISubclass : LayerUI<JComponent>() {

  override fun installUI(c: JComponent) {
    super.installUI(c)
    val jlayer = c as JLayer<*>
    jlayer.layerEventMask = AWTEvent.MOUSE_EVENT_MASK or
      AWTEvent.MOUSE_MOTION_EVENT_MASK or
      AWTEvent.KEY_EVENT_MASK
  }

  override fun uninstallUI(c: JComponent) {
    val jlayer = c as JLayer<*>
    jlayer.layerEventMask = 0
    super.uninstallUI(c)
  }



  override fun processMouseMotionEvent(e: MouseEvent?, l: JLayer<out JComponent>?) {
    e?.consume()
  }

  override fun processMouseEvent(e: MouseEvent?, l: JLayer<out JComponent>?) {
    e?.consume()
  }

  override fun processKeyEvent(e: KeyEvent,
                               l: JLayer<out JComponent?>?) {
    e.consume()
  }
}