// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable

import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.components.service
import com.intellij.openapi.extensions.KeyedFactoryEPBean
import com.intellij.openapi.extensions.ProjectExtensionPointName
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Comparing
import com.jetbrains.ide.model.popups.ProtoViewModel
import com.jetbrains.ide.model.uiautomation.BeControl
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.ui.bindable.views.listControl.IActionOwner
import com.jetbrains.rd.ui.bindable.views.listControl.IRendererOwner
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import java.util.concurrent.ConcurrentHashMap
import javax.swing.JComponent

interface ViewModel

interface ViewBinder<in TViewModel> {
  fun bind(viewModel: TViewModel, lifetime: Lifetime = Lifetime.Eternal): JComponent
}

private class ProtocolViewModelAdapter<out T : ProtoViewModel>(val vm: T) : ViewModel

fun <T : ProtoViewModel> T.wrap(): ViewModel = ProtocolViewModelAdapter(this)

class ViewRegistry(private val project: Project) {
  companion object {
    //todo replace "rdclient" "rd"
    private val EP_NAME = ProjectExtensionPointName<KeyedFactoryEPBean>("com.intellij.rdclient.view")

    fun getInstance(project: Project): ViewRegistry = project.service()
  }

  private val logger = getLogger<ViewRegistry>()

  private val myCache = ConcurrentHashMap<String, ViewBinder<Any>>()

  private fun <T> getKey(key: Class<T>): String = key.name

  @Suppress("MemberVisibilityCanBePrivate")
  fun <T : Any> getByKey(clazz: Class<T>): ViewBinder<T>? {
    val key = getKey(clazz)
    var cached = myCache[key]
    if (cached != null) return cached

    cached = getByKey(key)
    if (cached != null) {
      myCache[key] = cached
    }

    return cached
  }


  private fun <T> getByKey(key: String): ViewBinder<T>? {
    val epBeans = getExtensions()
    for (epBean in epBeans) {
      if (Comparing.strEqual(key, epBean.key)) {
        try {
          if (epBean.implementationClass != null) {
            return project.instantiateClass<ViewBinder<T>>(epBean.implementationClass, epBean.pluginDescriptor)
          }
        }
        catch (e: Exception) {
          throw RuntimeException(e)
        }

      }
    }
    return null
  }

  private fun getExtensions() = EP_NAME.getExtensions(project)

  fun <T : ViewModel> getView(viewModel: T, lifetime: Lifetime = Lifetime.Eternal): JComponent {
    if (viewModel is ProtocolViewModelAdapter<*>) return getView(viewModel.vm, lifetime)
    val view = getByKey(viewModel)
               ?: throw IllegalArgumentException(
                 "There's no registered view for '${getKey(viewModel.javaClass)}' viewModel (fqn:'${viewModel.javaClass}').")
    return view.bind(viewModel)
  }

  fun <T : ProtoViewModel> getView(viewModel: T, lifetime: Lifetime): JComponent {
    val view = getByKey(viewModel)
               ?: throw IllegalArgumentException(
                 "There's no registered view for '${getKey(viewModel.javaClass)}' viewModel (fqn:'${viewModel.javaClass}').")
    return view.bind(viewModel, lifetime)
  }


  private fun <T : Any> getByKey(viewModel: T): ViewBinder<T>? {
    val clazz = viewModel.javaClass
    val viewBinder = getByKey(clazz)
    if (viewBinder != null)
      return viewBinder
    if (clazz.superclass == viewModel) return null

    @Suppress("UNCHECKED_CAST")
    val superclass = clazz.superclass as? Class<T>
    if (superclass != null) return getByKey(superclass)
    return null
  }

  fun <T : BeControl> getRenderer(viewModel: T): ITreeGridRenderer<T>? {
    @Suppress("UNCHECKED_CAST")
    val rendererOwner = getByKey(viewModel) as? IRendererOwner<T>
    if (rendererOwner == null)
      logger.warn(viewModel.javaClass.simpleName + " is not an IRendererOwner")
    val view = rendererOwner ?: return null
    return view.getRenderer(viewModel)
  }

  fun <T : BeControl> getAction(viewModel: T, lifetime: Lifetime): AnAction? {
    @Suppress("UNCHECKED_CAST")
    val actionOwner = getByKey(viewModel) as? IActionOwner<T>
    if (actionOwner == null)
      logger.warn(viewModel.javaClass.simpleName + " is not an IActionOwner")
    val view = actionOwner ?: return null
    return view.getAction(lifetime, viewModel)
  }
}

fun <T : BeControl> T.getView(lifetime: Lifetime, project: Project): JComponent {
  return ViewRegistry.getInstance(project).getView(this, lifetime)
}

fun <T : BeControl> T.getRenderer(project: Project): ITreeGridRenderer<T>? {
  return ViewRegistry.getInstance(project).getRenderer(this)
}

fun <T : BeControl> T.getAction(lifetime: Lifetime, project: Project): AnAction? {
  return ViewRegistry.getInstance(project).getAction(this, lifetime)
}