// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.actionSystem.ActionPlaces
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.impl.ActionButton
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.util.ui.JBUI
import com.jetbrains.ide.model.uiautomation.BeButton
import com.jetbrains.ide.model.uiautomation.BeButtonStyle
import com.jetbrains.ide.model.uiautomation.BeRichText
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.ui.icons.RdIconHost
import com.jetbrains.rd.ui.icons.toIdeaIcon
import com.jetbrains.rd.ui.ComponentActionWrapper
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.views.listControl.IActionOwner
import com.jetbrains.rd.ui.bindable.views.listControl.IRendererOwner
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.BeButtonRenderer
import com.jetbrains.rd.ui.bindable.views.utils.*
import javax.swing.*

class ButtonViewControl(private val project: Project) : ViewBinder<BeButton>,
                                                        IRendererOwner<BeButton>,
                                                        IActionOwner<BeButton> {
  override fun getAction(lifetime: Lifetime, viewModel: BeButton): AnAction =
    tryGetRealAction(lifetime, viewModel) ?: ComponentActionWrapper { bind(viewModel, lifetime) }

  override fun getRenderer(viewModel: BeButton): ITreeGridRenderer<BeButton> = BeButtonRenderer(project)

  override fun bind(viewModel: BeButton, lifetime: Lifetime): JComponent {
    val jButton: JComponent = when (viewModel.style) {
      BeButtonStyle.DEFAULT,
      BeButtonStyle.ICON_TEXT -> {
        return createDefaultButton(viewModel, lifetime, false)
      }
      BeButtonStyle.TEXT -> {
        return createDefaultButton(viewModel, lifetime, true)
      }
      BeButtonStyle.ICON,
      BeButtonStyle.HYBRID_TEXT_ONLY,
      BeButtonStyle.HYBRID -> {
        return createIconButton(viewModel, lifetime)
      }
      BeButtonStyle.NO_BORDER -> {
        createNoBorderButton(viewModel, lifetime)
      }

    }
    jButton.isRequestFocusEnabled = true
    return jButton
  }

  private fun tryGetRealAction(lifetime: Lifetime, viewModel: BeButton): AnAction? {
    if (viewModel.style == BeButtonStyle.NO_BORDER) return null
    if (viewModel.content.shouldBeWrappedIntoHtml) return null
    if (viewModel.content is BeRichText) return null

    val action = object : DumbAwareAction() {
      override fun actionPerformed(e: AnActionEvent) = viewModel.action.click.fire()

      override fun displayTextInToolbar(): Boolean = viewModel.style != BeButtonStyle.ICON && viewModel.style != BeButtonStyle.HYBRID && viewModel.style != BeButtonStyle.HYBRID_TEXT_ONLY

      override fun update(e: AnActionEvent) = viewModel.updateActionEvent(e, project)
    }
    viewModel.prepareActionWithUpdate(lifetime) { update(action) }
    return action
  }

  private fun createIconButton(
    viewModel: BeButton,
    lifetime: Lifetime
  ): ActionButton {
    val actionObject = getAction(lifetime, viewModel) //todo: resursion?
    //val actionObject = tryGetRealAction(lifetime, viewModel) ?: ComponentActionWrapper { bind(viewModel, lifetime) }
    return ActionButton(
      actionObject,
      actionObject.templatePresentation,
      ActionPlaces.TOOLBAR,
      JBUI.size(24)
    )
  }

  private fun createDefaultButton(
    viewModel: BeButton,
    lifetime: Lifetime,
    widthFromContent: Boolean
  ): JComponent {
    return JButton().apply {
      addActionListener { viewModel.action.click.fire() }
      initFromText(lifetime, viewModel.content, RdIconHost.getInstance(project), widthFromContent)
      viewModel.bindDefaultProperties(lifetime, this)
    }
  }

  private fun createNoBorderButton(
    viewModel: BeButton,
    lifetime: Lifetime
  ): JLabel {
    val component = JLabel()
    val beAbstractText = viewModel.content

    beAbstractText.onTextChanged(lifetime) { _, s ->
      component.text = s
      component.revalidate()
    }

    beAbstractText.icon.view(lifetime) { _, s ->
      component.icon = s?.toIdeaIcon(RdIconHost.getInstance(project))
      component.revalidate()
    }

    viewModel.addMouseListener(component)

    return component.apply {
      viewModel.enabled.advise(lifetime) { isEnabled = it }
      viewModel.tooltip.advise(lifetime) { toolTipText = it }
    }
  }
}


