// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.project.Project
import com.intellij.ui.components.JBCheckBox
import com.intellij.util.ui.JBEmptyBorder
import com.jetbrains.ide.model.uiautomation.BeCheckbox
import com.jetbrains.ide.model.uiautomation.ControlVisibility
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.bind
import com.jetbrains.rd.ui.icons.RdIconHost
import com.jetbrains.rd.ui.ComponentActionWrapper
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.listControl.IActionOwner
import com.jetbrains.rd.ui.bindable.views.listControl.IRendererOwner
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.BeCheckBoxRenderer
import com.jetbrains.rd.ui.bindable.views.utils.BeComponents
import com.jetbrains.rd.ui.bindable.views.utils.initFromText
import com.jetbrains.rd.ui.bindable.views.utils.shouldBeWrappedIntoHtml
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import javax.swing.JComponent

class CheckBoxViewControl(private val project: Project) : ViewBinder<BeCheckbox>,
                                                          IRendererOwner<BeCheckbox>,
                                                          IActionOwner<BeCheckbox> {
  override fun getAction(lifetime: Lifetime, viewModel: BeCheckbox): AnAction {
    val jComponent = bind(viewModel, lifetime).apply {
      val insets = this.insets
      this.border = JBEmptyBorder(insets.top, 5, insets.bottom, 5)
    }
    return ComponentActionWrapper { jComponent }
  }

  override fun getRenderer(viewModel: BeCheckbox): ITreeGridRenderer<BeCheckbox> = BeCheckBoxRenderer()

  override fun bind(viewModel: BeCheckbox, lifetime: Lifetime): JComponent {
    val result = JBCheckBox()

    viewModel.enabled.advise(lifetime) {
      result.isEnabled = it
      viewModel.description.enabled.value = it
    }
    viewModel.tooltip.advise(lifetime) { result.toolTipText = it }

    viewModel.visible.advise(lifetime) {
      result.isVisible = it == ControlVisibility.Visible
      result.repaint()
    }

    viewModel.property.value?.let { result.isSelected = it }
    viewModel.property.bind(lifetime, { result.isSelected = it!! },
                            { change ->
                              result.addItemListener { change(result.isSelected) }
                            })


    return if (viewModel.description.shouldBeWrappedIntoHtml) initWrapped(result, viewModel, lifetime)
    else result.initFromText(lifetime, viewModel.description, RdIconHost.getInstance(project), false)

  }

  private fun initWrapped(checkbox: JBCheckBox, viewModel: BeCheckbox, lifetime: Lifetime): JComponent {
    val jPanel = BeComponents.getNewGridInstance({ it.columnConstraints = "[min!]5[fill,grow]" })
    jPanel.add(checkbox, "aligny top")

    val label = ViewRegistry.getInstance(project).getView(viewModel.description, lifetime)
    label.addMouseListener(object : MouseAdapter() {
      override fun mousePressed(e: MouseEvent?) {
        if (checkbox.isEnabled)
          checkbox.isSelected = !checkbox.isSelected
      }
    })
    jPanel.add(label)
    return jPanel
  }
}