// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.project.Project
import com.intellij.util.DocumentUtil
import com.jetbrains.ide.model.uiautomation.BeEditor
import com.jetbrains.ide.model.uiautomation.BeIdeaEditorSettings
import com.jetbrains.rd.ui.bedsl.extensions.UpdateFlag
import com.jetbrains.rd.ui.bedsl.extensions.ignoreOnReentrancy
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import javax.swing.JComponent

class EditorViewControl(private val project: Project) : ViewBinder<BeEditor> {
  private val reentrancyPreventer = UpdateFlag()

  override fun bind(viewModel: BeEditor, lifetime: Lifetime): JComponent {
    val editorFactory = EditorFactory.getInstance()
    val document = editorFactory.createDocument("")
    document.setReadOnly(viewModel.isReadonly.value)

    val editor = editorFactory.createEditor(document, project) as EditorEx
    lifetime.onTermination {
      EditorFactory.getInstance().releaseEditor(editor)
    }

    val settings = viewModel.settings.value
    if (settings is BeIdeaEditorSettings) {
      editor.settings.additionalColumnsCount = settings.additionalColumnsCount
      editor.settings.additionalLinesCount = settings.additionalLinesCount
      editor.settings.isCaretRowShown = settings.isCaretRowShown
      editor.settings.isFoldingOutlineShown = settings.isFoldingOutlineShown
      editor.settings.isIndentGuidesShown = settings.isIndentGuidesShown
      editor.settings.isLineMarkerAreaShown = settings.isLineMarkerAreaShown
      editor.settings.isLineNumbersShown = settings.isLineNumbersShown
      editor.settings.isUseSoftWraps = settings.isUseSoftWraps
      editor.settings.isWhitespacesShown = settings.isWhitespacesShown
    }

    editor.setHorizontalScrollbarVisible(true)
    editor.setVerticalScrollbarVisible(true)

    viewModel.text.text.advise(lifetime) {
      reentrancyPreventer.ignoreOnReentrancy {
        ApplicationManager.getApplication().invokeLaterOnWriteThread {
          DocumentUtil.writeInRunUndoTransparentAction {
            doIgnoringReadOnly(document) {
              document.setText(it.parts.joinToString("") { it.text })
            }
          }
        }
      }
    }

    return editor.component
  }

  private fun doIgnoringReadOnly(document: Document, action: () -> Unit) {
    if (document.isWritable) {
      action()
    }
    else {
      document.setReadOnly(false)
      try {
        action()
      }
      finally {
        document.setReadOnly(true)
      }
    }
  }
}