// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.DefaultActionGroup
import com.intellij.openapi.project.Project
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.extensions.valueOrEmpty
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.viewNotNull
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.listControl.IActionOwner
import com.jetbrains.rd.ui.bindable.views.listControl.IRendererOwner
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.BeGridRenderer
import com.jetbrains.rd.ui.bindable.views.utils.bindDefaultProperties
import com.jetbrains.rd.ui.bindable.views.utils.generateName
import com.jetbrains.rd.ui.bindable.views.utils.getLabelShortcut
import com.jetbrains.rd.ui.bindable.views.utils.marginTypeToString
import com.jetbrains.rdclient.ui.bindableUi.views.utils.contentHasFocusBorder
import net.miginfocom.swing.MigLayout
import javax.swing.JComponent
import javax.swing.JLabel
import javax.swing.JPanel


class GridViewControl(private val project: Project) : ViewBinder<BeGrid>, IRendererOwner<BeGrid>, IActionOwner<BeGrid> {
  override fun getRenderer(viewModel: BeGrid): ITreeGridRenderer<BeGrid> {
    return BeGridRenderer(project)
  }

  override fun bind(viewModel: BeGrid, lifetime: Lifetime): JComponent {
    this.labelCache.clear()
    val intercellMargin = marginTypeToString(viewModel.intercellSpacing)
    val isVertical = viewModel.orientation == GridOrientation.Vertical

    val layoutConstraints = "fill, left, insets 0, gap 0, novisualpadding,hidemode 2, ${if (isVertical) "wrap" else ""}"
    val panel = JPanel(MigLayout(layoutConstraints, "", "")).apply {
      name = viewModel.generateName()
    }

    // this is a way to keep size of * columns unchanged and similar. May be should be rewritten, since sg can give sideeffects
    val commonsize = "sg fit${viewModel.rdid}"

    viewModel.items.viewNotNull(lifetime) { innerLifetime, it ->
      panel.removeAll()
      myFocusBorderGap = "gap 2 2" //if(viewModel.shouldHaveCustomFocusMargin()) "gap 2 2" else ""
      var rows = it.joinToString(" $intercellMargin ")
      { gridElement ->
        if (gridElement.size.unit == BeSizingType.Fit)
          "[pref!]"
        else
          "[fill, grow ${gridElement.size.multiplier}, ${commonsize}]"
      }
      if (it.all { it.size.unit == BeSizingType.Fit })
        rows += "[fill, grow]"
      var columns = "[fill, grow, ${commonsize}]"
      if (!isVertical) {
        columns = rows.also { rows = columns }
      }
      val migLayout = MigLayout(layoutConstraints, columns, rows)
      panel.layout = migLayout

      it.forEach {
        loadElements(it, innerLifetime, viewModel, panel)
      }
      labelCache.clear()
      panel.revalidate()
      //panel.rootPane?.revalidate()
    }
    panel.isOpaque = false
    viewModel.bindDefaultProperties(lifetime, panel)

    return panel
  }

  private var myFocusBorderGap: String = ""
  private val BE_PREDEFINED_KEY = "GridViewControl.BePredefined"
  private val HAS_BE_PREDEFINED_KEY = "GridViewControl.HasBePredefined"
  private val labelCache = mutableMapOf<String, JLabel>()

  private fun loadElements(it: BeGridElement, lifetime: Lifetime, grid: BeGrid, panel: JPanel) {
    val content = it.content
    content.getOrCreateExtension("gridUnitSize") { it.size.unit }

    val component = ViewRegistry.getInstance(project).getView(content, lifetime)
    when (content) {
      is BeAbstractText -> {
        val shortcut = content.shortcutIndex.value
        if (shortcut >= 0) {
          val shortcutChar = content.getLabelShortcut()
          val label = when (component) {
            is JPanel -> component.components.firstOrNull() as? JLabel
            is JLabel -> component
            else -> null
          }
          if (label != null) {
            label.setDisplayedMnemonic(shortcutChar)
            label.displayedMnemonicIndex = shortcut
            val key = content.target.value
            if (key.isNotEmpty() && !labelCache.containsKey(key))
              labelCache.addUnique(lifetime, key, label)
          }
        }
      }

      /*
      //todo:[alisa] shortcut button initialization should be discussed
      is BePredefinedBehaviorButton -> {
          component.putClientProperty(BE_PREDEFINED_KEY, content)
          panel.putClientProperty(HAS_BE_PREDEFINED_KEY, true)
      }
      is BeGrid -> {
          val hasPredefinedShortcutsInside = component.getClientProperty(HAS_BE_PREDEFINED_KEY) as? Boolean
          if(hasPredefinedShortcutsInside == true){
              component.components.forEach {
                  val button = (it as? JComponent)?.getClientProperty(BE_PREDEFINED_KEY) as? BePredefinedBehaviorButton
                  button?.tryAddCustomShortCut(it, panel, lifetime)
              }
          }
      }*/
    }

    val id = content.controlId.value
    if (labelCache.containsKey(id))
      labelCache[id]!!.labelFor = component

    val isVertical = grid.orientation == GridOrientation.Vertical
    var constraints = when (it.size.unit) {
      BeSizingType.Fill -> ", grow"
      BeSizingType.Fit -> if (isVertical) ", aligny top" else ", align left "
    }
    constraints += when (it.alignment) {
      BeAlignment.Left -> if (isVertical || it.size.unit == BeSizingType.Fill) ", align left" else ""
      BeAlignment.Right -> ", align right"
      BeAlignment.Top -> if (!isVertical || it.size.unit == BeSizingType.Fill) ", align top" else ""
      BeAlignment.Bottom -> ", align bottom"
      BeAlignment.Center -> ", align center"
      BeAlignment.Stretch -> ""
      BeAlignment.Default -> ""
      BeAlignment.JustifiedDefault ->
        if (!content.contentHasFocusBorder()) ", gap 2 0" else ""
      /*
      ${if(isVertical) "2 0" else "0 2"}
      else
        if(it.content is BeAbstractText)
        "alignx label"
        else ""*/
      else -> ""
    }
    panel.add(component, constraints.trim(',', ' '))
  }



  override fun getAction(lifetime: Lifetime, viewModel: BeGrid): AnAction {
    val registry = ViewRegistry.getInstance(project)
    val items = viewModel.items.valueOrEmpty().map { registry.getAction(it.content, lifetime) }
    return object : DefaultActionGroup(items) {
      override fun displayTextInToolbar(): Boolean = true
    }
  }

}

