// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.project.Project
import com.jetbrains.ide.model.uiautomation.BeAbstractText
import com.jetbrains.ide.model.uiautomation.BeMutableSpanGrid
import com.jetbrains.ide.model.uiautomation.BeSizingType
import com.jetbrains.ide.model.uiautomation.BeSpanRowElement
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.utils.getLabelShortcut
import com.jetbrains.rd.ui.bindable.views.utils.marginTypeToString
import net.miginfocom.swing.MigLayout
import javax.swing.JComponent
import javax.swing.JLabel
import javax.swing.JPanel


class MutableSpanGridViewControl(private val project: Project) : ViewBinder<BeMutableSpanGrid> {
  private val labelCache = mutableMapOf<String, JLabel>()

  override fun bind(viewModel: BeMutableSpanGrid, lifetime: Lifetime): JComponent {
    this.labelCache.clear()

    val intercellMargin = marginTypeToString(viewModel.intercellSpacing)
    val fillRow = "[fill, grow]"
    val fillColumn = "[fill, grow]"
    var columns = viewModel.columns.joinToString(" $intercellMargin ", "", "", -1, "...") {
      if (it.unit == BeSizingType.Fill) fillColumn else "[min!]"
    }

    if (!columns.contains(fillColumn))
      columns += fillColumn
    val migLayout = MigLayout("fill, left, insets 0, gap 0, hidemode 2, novisualpadding", columns)

    val panel = JPanel(migLayout)
    viewModel.items.adviseAddRemove(lifetime) { e, _, _ ->
      this.labelCache.clear()

      var rows = viewModel.items.joinToString(" $intercellMargin", "", "", -1, "...") {
        if (it.size.unit == BeSizingType.Fit) "[min!]" else fillRow
      }
      if (!rows.contains(fillRow))
        rows += fillRow
      migLayout.rowConstraints = rows
      panel.removeAll()
      viewModel.items.forEachIndexed { index, gridElement ->
        loadElements(gridElement, lifetime, panel)
      }
      migLayout.invalidateLayout(panel)
      panel.revalidate()
    }
    panel.isOpaque = false
    return panel
  }


  private fun loadElements(
    rowElement: BeSpanRowElement,
    lifetime: Lifetime,
    panel: JPanel) {
    val content = rowElement.content
    content.valueOrNull?.forEachIndexed { index, columnElement ->
      val isLastColumn = index == content.valueOrNull?.lastIndex ?: 0
      val wrap = if (isLastColumn) "wrap" else ""

      val viewModel = columnElement.content.content
      viewModel.getOrCreateExtension("gridUnitSize") { columnElement.content.size.unit }

      val component = ViewRegistry.getInstance(project).getView(viewModel, lifetime)
      if (viewModel is BeAbstractText) {
        val shortcut = viewModel.shortcutIndex.value
        if (shortcut >= 0) {

          if (component is JPanel) {
            val label = component.components.firstOrNull() as? JLabel
            if (label != null) {
              val shortcutChar = viewModel.getLabelShortcut()
              label.setDisplayedMnemonic(shortcutChar)
              label.displayedMnemonicIndex = shortcut
              val key = viewModel.target.value
              if (key.isNotEmpty() && !labelCache.containsKey(key))
                labelCache.addUnique(lifetime, key, label)
            }
          }
        }
      }
      val id = viewModel.controlId.value
      if (labelCache.containsKey(id)) {
        labelCache[id]!!.labelFor = component
      }

      var constraints = wrap

      if (columnElement.columnSpan != 1)
        constraints += ", spanx ${columnElement.columnSpan}"

      var grow: String? = when (columnElement.content.size.unit) {
        BeSizingType.Fill -> "growx"
        BeSizingType.Fit -> "align left, w min, gapafter push"
      }

      grow +=
        when (rowElement.size.unit) {
          BeSizingType.Fill -> ", growy"
          BeSizingType.Fit -> ", growy"
        }
      grow.let { constraints += ", $grow" }

      panel.add(component, constraints.trim(',', ' '))
    }
  }
}