// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.project.Project
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.utils.getLabelShortcut
import com.jetbrains.rd.ui.bindable.views.utils.marginTypeToString
import com.jetbrains.rdclient.ui.bindableUi.views.utils.contentHasFocusBorder
import net.miginfocom.swing.MigLayout
import javax.swing.JComponent
import javax.swing.JLabel
import javax.swing.JPanel


class SpanGridViewControl(private val project: Project) : ViewBinder<BeSpanGrid> {
  private val labelCache = mutableMapOf<String, JLabel>()

  override fun bind(viewModel: BeSpanGrid, lifetime: Lifetime): JComponent {
    this.labelCache.clear()
    val s = "sg fit${viewModel.rdid}"
    val intercellMargin = marginTypeToString(viewModel.intercellSpacing)
    var columns = viewModel.columns.joinToString(" $intercellMargin ") {
      if (it.unit == BeSizingType.Fill)
        "[fill, grow ${it.multiplier}, ${s}]"
      else "[pref!]"
    }
    if (viewModel.columns.all { it.unit == BeSizingType.Fit })
      columns += "[fill, grow]"
    var rows = viewModel.items.valueOrNull?.joinToString(" $intercellMargin ") { it ->
      if (it.size.unit == BeSizingType.Fill)
        "[fill, grow ${it.size.multiplier}, ${s}r]"
      else "[pref!]"
    }
    if (viewModel.items.valueOrNull?.all { it.size.unit == BeSizingType.Fit } == true)
      rows += "[fill, grow]"
    val migLayout = MigLayout("fill, left, insets 0, gap 0, hidemode 2, novisualpadding", columns, rows)
    val panel = JPanel(migLayout)
    viewModel.items.advise(lifetime) {
      it.forEachIndexed { index, gridElement ->
        loadElements(gridElement, lifetime, panel, it.count())
        panel.repaint()
      }
      this.labelCache.clear()
    }
    panel.isOpaque = false
    return panel
  }


  private fun loadElements(
    rowElement: BeSpanRowElement,
    lifetime: Lifetime,
    panel: JPanel,
    rowCount: Int
  ) {
    val content = rowElement.content
    content.valueOrNull?.forEachIndexed { index, columnElement ->
      val isLastColumn = index == content.valueOrNull?.lastIndex ?: 0
      val wrap = if (isLastColumn && rowCount > 1) "wrap" else ""

      val viewModel = columnElement.content.content
      viewModel.getOrCreateExtension("gridUnitSize") { columnElement.content.size.unit }

      val component = ViewRegistry.getInstance(project).getView(viewModel, lifetime)
      if (viewModel is BeAbstractText) {
        val shortcut = viewModel.shortcutIndex.value
        if (shortcut >= 0) {

          if (component is JPanel) {
            val label = component.components.firstOrNull() as? JLabel
            if (label != null) {
              val shortcutChar = viewModel.getLabelShortcut()
              label.setDisplayedMnemonic(shortcutChar)
              label.displayedMnemonicIndex = shortcut
              val key = viewModel.target.value
              if (key.isNotEmpty() && !labelCache.containsKey(key))
                labelCache.addUnique(lifetime, key, label)
            }
          }
        }
      }
      val id = viewModel.controlId.value
      if (labelCache.containsKey(id))
        labelCache[id]!!.labelFor = component

      var constraints = wrap

      if (columnElement.columnSpan != 1)
        constraints += ", spanx ${columnElement.columnSpan}"

      var grow: String? = when (columnElement.content.size.unit) {
        BeSizingType.Fill -> "growx"
        BeSizingType.Fit -> "align left, w min"//, gapafter push"
      }

      grow =  when (rowElement.size.unit) {
        BeSizingType.Fill -> if(grow == "growx") "grow" else grow + ", growy"
        BeSizingType.Fit -> grow
      }

      grow.let { constraints += ", $grow" }
      constraints += when(columnElement.content.alignment){
        BeAlignment.Left -> if (columnElement.content.size.unit == BeSizingType.Fill)  ", align left" else ""
        BeAlignment.Right -> ", align right"
        BeAlignment.Top -> if ( columnElement.content.size.unit == BeSizingType.Fill) ", align top" else ""
        BeAlignment.Bottom -> ", align bottom"
        BeAlignment.Center -> ", align center"
        BeAlignment.Stretch -> ""
        BeAlignment.Default -> ""
        BeAlignment.JustifiedDefault -> if(!columnElement.content.content.contentHasFocusBorder())
          ", gap 2 2"
        /*else
          if(it.content is BeAbstractText)
          "alignx label"
          else ""*/
        else  ""
      }
      panel.add(component, constraints.trim(',', ' '))
    }

  }
}