// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.project.Project
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.ui.ComponentActionWrapper
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.listControl.IActionOwner
import com.jetbrains.rd.ui.bindable.views.listControl.IRendererOwner
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.BeStyleRenderer
import com.jetbrains.rd.ui.bindable.views.utils.generateName
import com.jetbrains.rd.ui.bindable.views.utils.getSize
import com.jetbrains.rd.ui.bindable.views.utils.marginToMig
import com.jetbrains.rdclient.util.idea.ui.toColor
import net.miginfocom.swing.MigLayout
import java.awt.Cursor
import javax.swing.JComponent
import javax.swing.JPanel

class StyleControlViewControl(private val project: Project) : ViewBinder<BeStyleControl>, IRendererOwner<BeStyleControl>, IActionOwner<BeStyleControl> {
  override fun getRenderer(viewModel: BeStyleControl): ITreeGridRenderer<BeStyleControl> {
    return BeStyleRenderer(project)
  }

  override fun bind(viewModel: BeStyleControl, lifetime: Lifetime): JComponent {
    val layoutConstraints = "fill, insets 0, gap 0, novisualpadding, alignx left"
    val panel = JPanel(MigLayout(layoutConstraints)).apply {
      name = viewModel.generateName()
    }
    panel.isOpaque = false
    val content = ViewRegistry.getInstance(project).getView(viewModel.content, lifetime)
    var constraints = "grow, push"
    var xConstraints: String = ""
    var yConstraints: String = ""
    val bounds = viewModel.sizes.forEach { bounds ->
      if (bounds != null) {
        //TODO(Korovin): Bug? Why only width* fields are used for x and y constraints?
        val widthType = bounds.type
        val widthSize = bounds.size
        val size = widthSize.getSize(content.preferredSize)
        content.preferredSize = size
        xConstraints = if (widthSize is BeControlSizePredefined && widthSize.width.type == BeControlSizeType.FIT_TO_CONTENT) "pushx, growx" else ""
        yConstraints =
          if (widthSize is BeControlSizePredefined && widthSize.height.type == BeControlSizeType.FIT_TO_CONTENT) "pushy, growy" else ""

        if (xConstraints.isEmpty())
          xConstraints = when (widthType) {
            BeBoundsType.FIXED_SIZE -> "east, w ${size.width}px"
            BeBoundsType.MAX_SIZE -> "east, wmax ${size.width}px, pushx, growx"
            BeBoundsType.MIN_SIZE -> "east, wmin ${size.width}px, pushx, growx"
          }
        if (yConstraints.isEmpty())
          yConstraints = when (widthType) {
            BeBoundsType.FIXED_SIZE -> "north, h ${size.height}px"
            BeBoundsType.MAX_SIZE -> "north, hmax ${size.height}px, pushy, growy"
            BeBoundsType.MIN_SIZE -> "north, hmin ${size.height}px, pushy, growy"
          }
      }
    }
    if(!xConstraints.isEmpty() && !yConstraints.isEmpty())
    constraints = "$xConstraints, $yConstraints"
    if (viewModel.backgroundColor.value != null) {
      val it = viewModel.backgroundColor.value!!
      panel.background = it.color.toColor()
      panel.isOpaque = true
    }
    if (viewModel.margin.value != null) {
      val it = viewModel.margin.value!!
      constraints += ", " + marginToMig(it.margin)
    }

    panel.add(content, constraints)

    viewModel.cursor.advise(lifetime) {
      if (it != null) {
        content.cursor = it.toCursor()
      }
    }
    viewModel.enabled.advise(lifetime) {
      content.isEnabled = it
      content.repaint()
    }
    viewModel.tooltip.advise(lifetime) { content.toolTipText = it }

    viewModel.visible.advise(lifetime) {
      content.isVisible = it == ControlVisibility.Visible
      content.repaint()
    }

    return panel
  }

  override fun getAction(lifetime: Lifetime, viewModel: BeStyleControl): AnAction {
    return ComponentActionWrapper { bind(viewModel, lifetime) }
  }
}

private fun BeCursorModifier.toCursor(): Cursor {
  return when(this.cursor){
    BeCursor.HAND -> Cursor.getPredefinedCursor(Cursor.HAND_CURSOR)
    BeCursor.ARROW -> Cursor.getDefaultCursor()
    else -> Cursor.getDefaultCursor()
  }
}
