// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.ui.TabbedPaneWrapper
import com.jetbrains.ide.model.uiautomation.BeTab
import com.jetbrains.ide.model.uiautomation.BeTabbedControl
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.getView
import com.jetbrains.rd.ui.bindable.views.utils.BeComponents
import com.jetbrains.rd.ui.bindable.views.utils.getLabelText
import com.jetbrains.rd.ui.icons.toIdeaIcon
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.IViewableList
import com.jetbrains.rd.util.reactive.valueOrDefault
import java.awt.Dimension
import javax.swing.JComponent

class TabbedViewControl(private val project: Project) :
  ViewBinder<BeTabbedControl> {

  override fun bind(viewModel: BeTabbedControl, lifetime: Lifetime): JComponent {
    val tabbedControl = TabbedPaneWrapper(lifetime.createNestedDisposable())

    val processingContents = Any()

    viewModel.tabs.advise(lifetime) {
      noReenter(processingContents) {
        when (it) {
          is IViewableList.Event.Remove -> tabbedControl.removeTabAt(it.index)
          is IViewableList.Event.Add -> addTab(tabbedControl, it.newValue, it.index, lifetime)
          is IViewableList.Event.Update -> {
            tabbedControl.removeTabAt(it.index)
            addTab(tabbedControl, it.newValue, it.index, lifetime)
          }
        }

        val selectedIndex = viewModel.selectedIndex.valueOrDefault(0)
        if (selectedIndex < tabbedControl.tabCount) {
          tabbedControl.selectedIndex = selectedIndex
        }
      }
    }

    tabbedControl.addChangeListener {
      noReenter(processingContents) {
        val selectedIndex = tabbedControl.selectedIndex
        if (selectedIndex >= 0 && selectedIndex < viewModel.tabs.size) {
          viewModel.selectedIndex.set(selectedIndex)
        }
      }
    }
    viewModel.selectedIndex.advise(lifetime) { index ->
      noReenter(processingContents) {
        if (index < tabbedControl.tabCount) {
          tabbedControl.selectedIndex = index
        }
      }
    }

    return BeComponents.getNewGridInstance().apply {
      add(tabbedControl.component, "grow")
      minimumSize = Dimension(300, 60)
    }
  }

  private fun addTab(tabbedControl: TabbedPaneWrapper,
                     tab: BeTab,
                     index: Int,
                     lifetime: Lifetime) {
    tabbedControl.insertTab(tab.tabName.getLabelText(),
                            tab.tabName.icon.value?.toIdeaIcon(project),
                            tab.content.getView(lifetime, project),
                            null,
                            index)
  }
}

