// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.editor.event.*
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.editor.impl.DocumentImpl
import com.intellij.ui.EditorTextField
import com.intellij.util.DocumentUtil
import com.jetbrains.ide.model.uiautomation.BeTextControl
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.ui.bindable.ViewBinder
import java.lang.Integer.max
import java.lang.Integer.min
import javax.swing.JComponent

class TextControlViewControl : ViewBinder<BeTextControl> {
  override fun bind(viewModel: BeTextControl, lifetime: Lifetime): JComponent {
    var localCookie = false

    val caretListener: CaretListener = object : CaretListener {
      override fun caretPositionChanged(event: CaretEvent) {
        if (localCookie) return
        localCookie = true
        if (event.caret != null)
          viewModel.settings.caret.set(event.caret!!.offset)
        localCookie = false
      }
    }

    val selectionListener : SelectionListener = object : SelectionListener {
      override fun selectionChanged(e: SelectionEvent) {
        if (localCookie) return
        localCookie = true
        val left = e.newRange.startOffset
        val right = e.newRange.endOffset
        val caretOffset = e.editor.caretModel.currentCaret.offset
        if (caretOffset == left)
          viewModel.settings.selectionLength.set(right - left)
        else
          viewModel.settings.selectionLength.set(left - right)
        localCookie = false
      }
    }

    val textField = object : EditorTextField(viewModel.text.value) {
      override fun createEditor(): EditorEx {
        return super.createEditor().apply {
          settings.isUseSoftWraps = true
          settings.isWhitespacesShown = false
          setHorizontalScrollbarVisible(true)
          setVerticalScrollbarVisible(true)

          setCaretPosition(viewModel.settings.caret.value)
          caretModel.addCaretListener(caretListener)
//          todo: set starting selection to selectionModel if anyone needs it
          selectionModel.addSelectionListener(selectionListener)
          lifetime.onTermination {
            caretModel.removeCaretListener(caretListener)
            selectionModel.removeSelectionListener(selectionListener)
          }
        }
      }
    }

    (textField.document as? DocumentImpl)?.setAcceptSlashR(true)

    // text
    val value: DocumentListener = object : DocumentListener {
      override fun documentChanged(event: DocumentEvent) {
        if (localCookie) return
        localCookie = true
        viewModel.text.set(textField.text)
        localCookie = false
      }
    }
    textField.addDocumentListener(value)

    viewModel.text.advise(lifetime) {
      if (localCookie) return@advise
      localCookie = true
      DocumentUtil.writeInRunUndoTransparentAction { textField.document.setText(it) }
      localCookie = false
    }
    lifetime.onTermination { textField.removeDocumentListener(value) }
    textField.setOneLineMode(false)

    // caret
    viewModel.settings.caret.advise(lifetime) {
      if (localCookie) return@advise
      localCookie = true
      textField.setCaretPosition(it)
      localCookie = false
    }

    // selection
    viewModel.settings.selectionLength.advise(lifetime) {
      if (localCookie) return@advise
      localCookie = true
      if (it == 0)
        textField.removeSelection()
      else {
        val left = min(viewModel.settings.caret.value, it)
        val right = max(viewModel.settings.caret.value, it)
        textField.editor?.selectionModel?.setSelection(left, right)
      }
      localCookie = false
    }

    // enabled
    viewModel.enabled.advise(lifetime) {
      textField.isEnabled = it
    }
    return textField
  }
}