// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.IdeActions
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.colors.EditorFontType
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.event.DocumentListener
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.editor.highlighter.EditorHighlighterFactory
import com.intellij.openapi.fileTypes.FileType
import com.intellij.openapi.fileTypes.FileTypes
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.psi.PsiDocumentManager
import com.intellij.psi.PsiFileFactory
import com.intellij.refactoring.ui.MethodSignatureComponent
import com.intellij.ui.EditorTextField
import com.intellij.util.DocumentUtil
import com.intellij.util.LocalTimeCounter
import com.jetbrains.ide.model.uiautomation.BeTextControlWithLanguage
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.lang.toIdeaFileType
import com.jetbrains.rd.ui.bindable.ViewBinder
import java.awt.Dimension
import java.awt.event.FocusEvent
import java.awt.event.FocusListener
import javax.swing.JComponent

class TextControlWithLanguageViewControl(private val project: Project) :
  ViewBinder<BeTextControlWithLanguage> {
  override fun bind(viewModel: BeTextControlWithLanguage, lifetime: Lifetime): JComponent {
    val fileType: FileType = viewModel.language?.toIdeaFileType() ?: FileTypes.PLAIN_TEXT
    val result: JComponent
    if (viewModel.textControl.settings.readonly.value) {
      result = object : MethodSignatureComponent("", project, fileType) {
        override fun createEditor(): EditorEx {
          return super.createEditor().apply {
            settings.isUseSoftWraps = true
            minimumSize = Dimension(300, 60)
          }
        }
      }

      viewModel.textControl.text.advise(lifetime) {
        result.setSignature(it)
      }


    }
    else {
      val fileName = viewModel.customFileName
      val document = createDocument(fileName, fileType, viewModel.createPsiFile)

      val textField = object : EditorTextField(document, project, fileType) {
        override fun createEditor(): EditorEx {
          return super.createEditor().apply {
            settings.isUseSoftWraps = true


            highlighter = EditorHighlighterFactory.getInstance()
              .createEditorHighlighter(project, fileType)

            settings.isWhitespacesShown = false
            settings.isCaretRowShown = true
            setHorizontalScrollbarVisible(true)
            setVerticalScrollbarVisible(true)
            minimumSize = Dimension(300, 60)
          }
        }

        // fixed find action under welcome screen
        override fun getData(dataId: String): Any? {
          if (CommonDataKeys.PROJECT.`is`(dataId)) {
            return project
          }

          return super.getData(dataId)
        }
      }

      textField.font = EditorFontType.PLAIN.globalFont
      textField.setCaretPosition(0)

      var localCookie = false
      val value: DocumentListener = object : DocumentListener {
        override fun documentChanged(event: DocumentEvent) {
          if (localCookie) return
          localCookie = true
          viewModel.textControl.text.set(textField.text)
          localCookie = false
        }
      }
      textField.addDocumentListener(value)

      viewModel.textControl.text.advise(lifetime) {
        if (localCookie) return@advise
        localCookie = true
        DocumentUtil.writeInRunUndoTransparentAction { textField.document.setText(it) }
        localCookie = false
      }
      lifetime.onTermination { textField.removeDocumentListener(value) }
      textField.setOneLineMode(false)

      val findAction = ActionManager.getInstance().getAction(IdeActions.ACTION_FIND)
      if (findAction != null) {
        textField.addFocusListener(object : FocusListener {
          private val sequentialLifetimes = SequentialLifetimes(lifetime)
          override fun focusGained(e: FocusEvent?) {
            val component = textField.component
            findAction.registerCustomShortcutSet(findAction.shortcutSet, component, sequentialLifetimes.next().createNestedDisposable())
          }

          override fun focusLost(e: FocusEvent?) {
            sequentialLifetimes.terminateCurrent()
          }
        })
      }

      result = textField
    }
    viewModel.enabled.advise(lifetime) {
      result.isEnabled = it
    }
    return result
  }

  private fun createDocument(fileName: String?, fileType: FileType, createPsi: Boolean): Document? {
    if (!createPsi)
      return EditorFactory.getInstance().createDocument("")

    val factory = PsiFileFactory.getInstance(project)
    val stamp = LocalTimeCounter.currentTime()
    val psiFile = factory.createFileFromText(fileName ?: "Dummy." + fileType.defaultExtension, fileType, "", stamp, true, false)
    return PsiDocumentManager.getInstance(project).getDocument(psiFile)
  }
}