// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views

import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.project.DumbAwareToggleAction
import com.intellij.openapi.project.Project
import com.jetbrains.ide.model.uiautomation.BeButtonStyle
import com.jetbrains.ide.model.uiautomation.BeToggleButton
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.bind
import com.jetbrains.rd.util.reactive.map
import com.jetbrains.rd.ui.icons.RdIconHost
import com.jetbrains.rd.ui.icons.toIdeaIcon
import com.jetbrains.rd.ui.bindable.ViewBinder
import com.jetbrains.rd.ui.bindable.views.customControls.IconToggleButton
import com.jetbrains.rd.ui.bindable.views.listControl.IActionOwner
import com.jetbrains.rd.ui.bindable.views.listControl.IRendererOwner
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.BeToggleButtonLabelRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.BeToggleButtonSCCRenderer
import com.jetbrains.rd.ui.bindable.views.utils.onTextChanged
import com.jetbrains.rd.ui.bindable.views.utils.updateActionEvent
import javax.swing.JComponent

class ToggleButtonViewControl(private val project: Project)
  : ViewBinder<BeToggleButton>
    , IRendererOwner<BeToggleButton>
    , IActionOwner<BeToggleButton> {
  override fun getAction(lifetime: Lifetime, viewModel: BeToggleButton): AnAction {
    return object : DumbAwareToggleAction() {
      override fun update(e: AnActionEvent) {
        super.update(e)
        viewModel.updateActionEvent(e, project)
      }

      override fun isSelected(e: AnActionEvent): Boolean = viewModel.isChecked.value

      override fun setSelected(e: AnActionEvent, state: Boolean) = viewModel.isChecked.set(state)

      override fun displayTextInToolbar(): Boolean =
        viewModel.style != BeButtonStyle.ICON && viewModel.style != BeButtonStyle.HYBRID && viewModel.style != BeButtonStyle.HYBRID_TEXT_ONLY
    }
  }

  override fun getRenderer(viewModel: BeToggleButton): ITreeGridRenderer<BeToggleButton> {
    return if (viewModel.useLabelRenderer.value) BeToggleButtonLabelRenderer(project) else BeToggleButtonSCCRenderer(project)
  }

  override fun bind(viewModel: BeToggleButton, lifetime: Lifetime): JComponent {
    return IconToggleButton().apply {
      viewModel.tooltip.view(lifetime) { _, text ->
        toolTipText = text
      }
      val ideaIcon = viewModel.content.icon.map { it?.toIdeaIcon(RdIconHost.getInstance(project)) }
      ideaIcon.advise(lifetime) { icon ->
        this.icon = icon
      }
      viewModel.content.onTextChanged(lifetime) { _, s ->
        this.text = s
      }
      viewModel.isChecked.bind(lifetime, { isChecked.value = it },
                               { change ->
                                 isChecked.advise(lifetime) { change(it) }
                               })
    }
  }
}