// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.customControls

import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.ComboBox
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.swing.textProperty
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.AddRemove
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.RdCellRenderer
import com.jetbrains.rd.ui.bindable.views.utils.getSimpleText
import java.awt.Color
import java.awt.Component
import javax.swing.AbstractListModel
import javax.swing.JList
import javax.swing.JTextField
import javax.swing.ListCellRenderer
import javax.swing.plaf.basic.BasicComboBoxEditor

class CustomComboBoxControl(private val lifetime: Lifetime,
                            private val isEditor: Boolean = false,
                            private val project: Project,
                            private val useSmallPreferredWidth: Boolean = false) : ComboBox<Any?>() {
  var lock = false
  var initPreferred = false

  fun init(beComboBox: BeComboBox) {
    val comboBoxModel: javax.swing.ComboBoxModel<Any?> =
      if (beComboBox.isEditable.value)
        EditableComboBoxModel(lifetime, beComboBox, this /*[alisa] may be it's hack, to rewrite*/)
      else
        ComboBoxModel(lifetime, beComboBox, this /*[alisa] may be it's hack, to rewrite*/)

    model = comboBoxModel
    isEnabled = beComboBox.enabled.value
    beComboBox.enabled.advise(lifetime) {
      isEnabled = it
      if (beComboBox.isEditable.value)
        beComboBox.revalidate.fire()
    }
    //NOTE(Korovin): Ideally we should just set
    // setRenderer(RdCellRenderer(iconHost, lifetime, project, false))
    // here, but it cause our renderer to be wrapped and there are default margins in the wrapper
    // Default margins are bad when we want to render whole line control, like BeHeader
    // From the other side without wrapper we have no default margins for text values
    val unitSize = beComboBox.getOrCreateExtension("gridUnitSize", { BeSizingType.Fit })
    if (unitSize == BeSizingType.Fill)
      setMinimumAndPreferredWidth(0)
    isEditable = beComboBox.isEditable.value
    if (isEditable) {
      setMinimumAndPreferredWidth(72)

      val editor = BasicComboBoxEditor()
      selectedItem = beComboBox.text.value

      this.setEditor(editor)
      //val editorComponent = editor.editorComponent as EditorTextField

      val editorComponent = editor.editorComponent as JTextField
      editorComponent.isEditable = !beComboBox.settings.readonly.value

      editorComponent.text = beComboBox.text.value

      val textProperty = editorComponent.textProperty()
      textProperty.change.advise(lifetime) {
        if (lock/* comboBoxModel.lock*/) {
          editorComponent.repaint()
          return@advise
        }
        lock = true
        if (selectedIndex < 0 || beComboBox.selectedIndex.value < 0)
          beComboBox.text.set(editorComponent.text)
        lock = false
      }

      val default = editorComponent.foreground
      val error = Color.RED
      val warning = Color.ORANGE
      beComboBox.validationResult.view(lifetime) { _, it ->

        if (beComboBox.validationStyle.valueOrNull == ValidationStyle.Text) {
          when {
            it.state == ValidationStates.validationError -> editorComponent.foreground = error
            it.state == ValidationStates.validationWarning -> editorComponent.foreground = warning
            it.state == ValidationStates.validationPassed -> editorComponent.foreground = default
          }
        }
        else {
          when {
            it.state == ValidationStates.validationError -> this.putClientProperty("JComponent.outline", "error")
            it.state == ValidationStates.validationWarning -> this.putClientProperty("JComponent.outline", "warning")
            it.state == ValidationStates.validationPassed -> this.putClientProperty("JComponent.outline", null)
          }
        }
        beComboBox.tooltip.set(it.resultMessage)
        editorComponent.toolTipText = it.resultMessage
        repaint()
      }

      beComboBox.text.advise(lifetime) {
        if (lock /*|| comboBoxModel.lock*/) return@advise
        lock = true
        editorComponent.text = it
        selectedItem = null
        lock = false
      }

      beComboBox.revalidate.fire()

    }
    else {
      //NOTE(Korovin): See note above for explanation of this hack
      renderer = RdCellRenderer(
        project,
        true,
        arrayListOf(BeAlignment.Left)
      )
      beComboBox.selectedIndex.advise(lifetime) {
        if (lock || it < 0) return@advise
        lock = true
        selectedItemChanged()
        repaint()
        lock = false
      }
    }

    beComboBox.enabled.advise(lifetime) {
      isEnabled = it
      if (beComboBox.isEditable.value)
        beComboBox.revalidate.fire()
    }
    //NOTE(Korovin): See note above for explanation of this hack
    if (beComboBox.isEditable.value)
      popup?.list?.cellRenderer = RdWrapperCellRenderer(project)
    else
      popup?.list?.cellRenderer = RdCellRenderer(
        project,
        false,
        arrayListOf(BeAlignment.Left)
      )
  }

  class RdWrapperCellRenderer(project: Project) :
    ListCellRenderer<Any> {
    val renderer = RdCellRenderer(
      project,
      false,
      arrayListOf(BeAlignment.Left)
    )

    override fun getListCellRendererComponent(
      list: JList<out Any>?,
      value: Any?,
      index: Int,
      isSelected: Boolean,
      cellHasFocus: Boolean
    ): Component {
      return when (value) {
        is BeControlWrapper -> renderer.getListCellRendererComponent(list, value.content, index, isSelected, cellHasFocus)
        else -> renderer.getListCellRendererComponent(list, value, index, isSelected, cellHasFocus)
      }
    }
  }

  override fun addNotify() {
    super.addNotify()
    /*if(useSmallPreferredWidth)
        border = JBUI.Borders.empty()*/
    //border = JBUI.Borders.empty()
    if (isEditor)
      showPopup()
  }

  /* override fun getPreferredSize(): Dimension {
       val result = super.getPreferredSize()
       if (useSmallPreferredWidth)
           return Dimension(10, result.height)
       val preferredSize = super.getPreferredSize()
       return preferredSize
   }*/
  private class BeControlWrapper(val content: BeControl?) {

    override fun toString(): String {
      return when (content) {
               is BeAbstractText -> content.getSimpleText()
               else -> content?.toString()
             } ?: ""
    }
  }

  private class ComboBoxModel(
    lifetime: Lifetime,
    val model: BeComboBox,
    val customComboBoxControl: CustomComboBoxControl
  ) : AbstractListModel<Any?>(),
      javax.swing.ComboBoxModel<Any?> {

    init {
      model.values.adviseAddRemove(lifetime) { e, i, _ ->
        if (e == AddRemove.Remove)
          fireIntervalRemoved(this, i, i)
        if (e == AddRemove.Add)
          fireIntervalAdded(this, i, i)
      }
    }

    override fun getElementAt(index: Int): Any? = if (index >= 0) model.values[index] else null

    override fun getSize(): Int = model.values.size

    override fun getSelectedItem(): Any? {
      val index = model.selectedIndex.value
      return model.values.getOrNull(index)
    }

    override fun setSelectedItem(anItem: Any?) {
      val beControl = anItem as? BeControl
      if (beControl is BeHeader) return

      val value = model.values
      val first = value.firstOrNull { it == beControl }
      if (first != null)
        model.selectedIndex.set(value.indexOf(first))
    }
  }

  private class EditableComboBoxModel(
    lifetime: Lifetime,
    val model: BeComboBox,
    val customComboBoxControl: CustomComboBoxControl
  ) : AbstractListModel<Any?>(),
      javax.swing.ComboBoxModel<Any?> {

    private var viewableList: MutableList<BeControlWrapper>

    init {
      viewableList = mutableListOf()/*model.values.map {i -> BeControlWrapper(i as? BeAbstractText) }.toMutableList()*/
      model.values.adviseAddRemove(lifetime) { e, i, it ->

        lock = true
        if (e == AddRemove.Remove) {
          viewableList.removeAt(i)
          fireIntervalRemoved(this, i, i)
        }
        if (e == AddRemove.Add) {
          viewableList.add(i, BeControlWrapper(it))
          fireIntervalAdded(this, i, i)
        }
        lock = false
      }
    }


    override fun getElementAt(index: Int): Any? = if (index >= 0) viewableList[index] else null

    override fun getSize(): Int = viewableList.size

    override fun getSelectedItem(): Any? {
      val index = model.selectedIndex.value
      val currentText = (customComboBoxControl.editor?.editorComponent as? JTextField)?.text

      val existingListElement = viewableList.getOrNull(index)
      val beControl = if (existingListElement != null)
        existingListElement
      else {
        viewableList.firstOrNull { c -> c.toString() == currentText }
      }
      return if (beControl.toString() == currentText) {
        beControl
      }
      else {
        model.selectedIndex.set(-1)
        currentText
      }
    }

    var lock: Boolean = false

    override fun setSelectedItem(anItem: Any?) {
      val beControl = anItem as? BeControlWrapper
      if (beControl?.content is BeHeader) return

      lock = beControl != null

      if (!lock) {
        //model.selectedIndex.set(-1)
        return
      }
      val first = viewableList.firstOrNull { it == beControl }
      if (first != null)
        model.selectedIndex.set(viewableList.indexOf(first))
      else
        model.selectedIndex.set(-1)
      lock = false
    }
  }
}