// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.customControls

import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.event.*
import com.intellij.openapi.fileTypes.FileTypes
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.ui.ComponentValidator
import com.intellij.openapi.ui.ValidationInfo
import com.intellij.openapi.ui.popup.*
import com.intellij.ui.EditorTextField
import com.intellij.ui.components.JBList
import com.intellij.ui.speedSearch.NameFilteringListModel
import com.intellij.util.DocumentUtil
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.dsl.tryGetSettings
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.util.reactive.valueOrDefault
import com.jetbrains.rd.ui.bedsl.dsl.util.BeDslMargin
import com.jetbrains.rd.ui.bedsl.dsl.withMargin
import com.jetbrains.rd.ui.bedsl.extensions.addElement
import com.jetbrains.rd.ui.bedsl.extensions.bindFocus
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.RdCellRenderer
import com.jetbrains.rd.ui.bindable.views.utils.BeMagicMargin
import com.jetbrains.rd.ui.bindable.views.utils.getFont
import java.awt.Component
import java.awt.Dimension
import java.awt.Event
import java.awt.event.FocusAdapter
import java.awt.event.FocusEvent
import java.awt.event.KeyAdapter
import java.awt.event.KeyEvent
import java.util.*
import javax.swing.DefaultListCellRenderer
import javax.swing.JList
import javax.swing.SwingUtilities

open class TextBoxControl : EditorTextField {
  private val myListenerList = HashSet<(Boolean) -> Unit>()
  private var myDocumentListener: MyDocumentListener? = null
  private var mySuggestions: List<BeCompletionItem> = listOf()
  private var myCurrentPopup: JBPopup? = null
  private var myJList: JBList<BeCompletionItem>? = null
  private var showNextResults = false
  private var myPopupIsShown = false
  private var myBuilder: PopupChooserBuilder<BeCompletionItem>? = null
  private var isStaticSuggestionsList = false
  private var localCookie: Boolean = false
  private lateinit var myModel: BeTextBox

  constructor(project: Project) : super(
    EditorFactory.getInstance().createDocument(""),
    project,
    FileTypes.PLAIN_TEXT
  ) {
    this.ensureWillComputePreferredSize()
  }

  constructor(lifetime: Lifetime, viewModel: BeTextBox, project: Project) : super(
    EditorFactory.getInstance().createDocument(""), project,
    FileTypes.PLAIN_TEXT, viewModel.settings.readonly.value, true) {
    init(viewModel, lifetime)
    isFocusable = !viewModel.settings.readonly.value
  }

  fun init(model: BeTextBox, lifetime: Lifetime) {
    // val textField = EditorTextField(model.text.value, project, FileTypes.PLAIN_TEXT)
    myModel = model

    if(model.settings.updateSourceTrigger.equals(UpdateSourceTrigger.TextChanged))
      addDataChangedListener {
        setText()
      }

    text = model.text.value
    model.completionItems.change.advise(lifetime) {
      if (isDisplayable && editor?.selectionModel?.hasSelection() != true)
        setSuggestions(it, lifetime) //todo: use real addremove instead of full replacement
    }
    this.ensureWillComputePreferredSize()
    model.text.advise(lifetime) {
      if (localCookie || document.text == it) return@advise
      localCookie = true
      DocumentUtil.writeInRunUndoTransparentAction {
        document.setText(it)
      }
      localCookie = false
    }
    model.enabled.advise(lifetime) {
      isEnabled = it
      model.revalidate.fire()
    }
    /* val focusManager = IdeFocusManager.getGlobalInstance()
     focusManager.doWhenFocusSettlesDown {
         SwingUtilities.invokeLater {
             if (component is TreeTable) {
                 focusManager.doWhenFocusSettlesDown {
                     //component.cellEditor?.stopCellEditing() // store user input
                     *//*val row = component.selectedRow
                        val col = component.selectedColumn
                        component.scrollRectToVisible(component.getCellRect(row, 0, true))
                        if (component.editCellAt(row, 0)) {
                            val editorComponent = component.editorComponent as? JComponent
*//*
                        if (editor != null && focusManager.focusOwner != editor) {

                                focusManager.getFocusTargetFor(this)?.requestFocusInWindow()
                            }
                        }
                    }
                }
            }
        }*/


    model.bindFocus(lifetime, this) { beTextBox, control ->
      control.selectAll()
      control.setCaretPosition(beTextBox.text.value.length)
    }

    model.settings.caret.advise(lifetime) {
      if (localCookie) return@advise
      localCookie = true
      if (it > document.textLength || it < 0) return@advise
      this.setCaretPosition(it)
      localCookie = false
    }

    ComponentValidator(lifetime.createNestedDisposable())
      .withValidator { ->
        val it = model.validationResult.valueOrNull
        val validationInfo = if (it == null) null
        else
          when (it.state) {
            ValidationStates.validationError -> ValidationInfo(
              it.resultMessage
            )
            ValidationStates.validationWarning -> ValidationInfo(
              it.resultMessage
            ).asWarning()
            ValidationStates.validationPassed -> null
            else -> null
          }?.forComponent(this)
        validationInfo

      }.installOn(this)

    model.validationResult.view(lifetime) { _, _ ->
      ComponentValidator.getInstance(this)
        .ifPresent(ComponentValidator::revalidate)
    }
    this.setCaretPosition(model.settings.caret.value)

    lifetime.onTermination { detachListeners() }

    isOneLineMode = true
    font = model.settings.textSettings.getFont()
    //font.size = com.jetbrains.rd.ui.bindable.views.utils.getFontSize(font.size, model.settings.textSettings.fontSize.valueOrDefault(BeFontSize.DEFAULT))
    minimumSize = Dimension(BeMagicMargin.MinimumTextBoxWidth, BeMagicMargin.MinimumTextBoxHeight)
  }

  private fun setText() {
    if (localCookie) return
    localCookie = true
    myModel.text.set(text)
    if (editor != null)
      myModel.settings.caret.value = editor!!.caretModel.offset
    localCookie = false
  }

  override fun addNotify() {
    super.addNotify()

    registerEditorListeners()
    //editor?.setBorder(null)
    val startOffset = myModel!!.settings.caret.value
    val selectionLength = myModel!!.settings.caret.value + myModel!!.settings.selectionLength.value
    if (editor?.document?.text?.length ?: 0 >= selectionLength) {
      this.editor?.selectionModel?.setSelection(
        startOffset,
        selectionLength
      )
    }
    this.editor?.settings?.isCamelWords = myModel!!.settings.useCamelHumps.value
  }

  private fun registerEditorListeners() {
    editor?.caretModel?.addCaretListener(object : CaretListener {
      override fun caretPositionChanged(e: CaretEvent) {
        if (e.caret == null) return
        if (localCookie) return
        localCookie = true
        myModel?.settings?.caret?.set(e.caret!!.offset)
        //   myModel?.selectionLength?.set( editor?.selectionModel?.selectedText?.length ?:0)
        localCookie = false
      }
    })
    editor?.selectionModel?.addSelectionListener(object : SelectionListener {
      override fun selectionChanged(e: SelectionEvent) {
        if (localCookie) return
        localCookie = true
        var newValue = e.newRange.length
        if (e.newRange.endOffset == e.editor.caretModel.primaryCaret.offset)
          newValue = -newValue
        myModel?.settings?.selectionLength?.set(newValue)
        localCookie = false


      }
    })
    this.focusTarget.addKeyListener(object : KeyAdapter() {
      private fun keyIsForList(keyCode: Int): Boolean {
        return keyCode == KeyEvent.VK_DOWN || keyCode == KeyEvent.VK_UP || keyCode == KeyEvent.VK_PAGE_UP || keyCode == KeyEvent.VK_PAGE_DOWN
      }

      private fun keyIsForCompletion(keyCode: Int, modifiers: Int): Boolean {
        return keyCode == KeyEvent.VK_SPACE && modifiers == Event.CTRL_MASK
      }

      override fun keyPressed(e: KeyEvent) {
        if (e.keyCode == KeyEvent.VK_ENTER) {
          if (myPopupIsShown) {
            showNextResults = false
            if (myJList != null && myJList!!.selectedIndex >= 0) {
              myJList!!.dispatchEvent(e)
              e.consume()
            }
            hidePopup()
            myCurrentPopup = null
          }
          else {
            if(myModel.settings.updateSourceTrigger == UpdateSourceTrigger.EnterOrTab)
              setText()
            showNextResults = true
          }
        }
        else showNextResults = true

        if (e.keyCode == KeyEvent.VK_TAB) {
          if(myModel.settings.updateSourceTrigger == UpdateSourceTrigger.EnterOrTab)
            setText()
        }

        if (e.keyCode == KeyEvent.VK_ESCAPE || e.keyCode == KeyEvent.VK_ENTER) {

          if (myPopupIsShown) {
            hidePopup()
            myCurrentPopup = null
            e.consume()
          }
          return
        }
        if (keyIsForCompletion(e.keyCode, e.modifiers)) {
          if (myCurrentPopup == null)
            showPopup()
          e.consume()
        }

        if (keyIsForList(e.keyCode)) {
          if (myCurrentPopup == null) {
            showPopup()
            e.consume()
          }
          else {
            myJList?.dispatchEvent(e)
          }

        }
        if (e.keyCode == KeyEvent.VK_DELETE || e.keyCode == KeyEvent.VK_BACK_SPACE) {
          if (editor != null)
            editor!!.caretModel.currentCaret.moveToOffset(editor!!.caretModel.offset - 1)
        }
        showNextResults = true
      }

      override fun keyReleased(e: KeyEvent) {
        /*   if (keyIsForList(e.keyCode) || e.keyCode == KeyEvent.VK_ENTER && myCurrentPopup != null) {
               myJList?.dispatchEvent(e)
               e.consume()
           }*/
      }

      override fun keyTyped(e: KeyEvent) {
        /*if (myCurrentPopup == null && !e.keyChar.isWhitespace() && !isStaticSuggestionsList) {
            showPopup()
            showNextResults = true
        }*/
      }
    })

    this.focusTarget.addFocusListener(object : FocusAdapter() {
      override fun focusGained(p0: FocusEvent?) {
        super.focusGained(p0)
        if (!myPopupIsShown) {
          showNextResults = true
          //                 showPopup()
        }
      }

      override fun focusLost(e: FocusEvent) {
        /*if (e is CausedFocusEvent && e.cause == CausedFocusEvent.Cause.ACTIVATION)
            return*/
        hidePopup()
        if(myModel.settings.updateSourceTrigger == UpdateSourceTrigger.LostFocus)
          setText()
      }
    })
  }

  fun hidePopup() {
    myCurrentPopup?.cancel()
    myPopupIsShown = false
  }

  fun showPopup() {
    if (!showNextResults || myPopupIsShown || myCurrentPopup != null) return
    val filteredByLengthSuggestions = mySuggestions.filter { !isStaticSuggestionsList || it.translatedText.startsWith(this.text) }

    if (filteredByLengthSuggestions.isEmpty())
      return

    //val optionsJList = JBList(*filteredByLengthSuggestions.toTypedArray())


    myBuilder = JBPopupFactory.getInstance().createPopupChooserBuilder(
      filteredByLengthSuggestions.toList()) as PopupChooserBuilder<BeCompletionItem>
    val defaultControlRenderer = RdCellRenderer(
      project,
      false,
      arrayListOf(BeAlignment.Left)
    )
    myBuilder!!
      .setMovable(true)
      .setRequestFocus(false)
      .setModalContext(false)
      .setRequestFocus(false)
      .setVisibleRowCount(5)
      .setRenderer(object : DefaultListCellRenderer() {
        override fun getListCellRendererComponent(
          list: JList<*>?,
          value: Any?,
          index: Int,
          isSelected: Boolean,
          cellHasFocus: Boolean
        ): Component {

          val completionItem = value as? BeCompletionItem ?: return this
          val localGrid = BeGrid(
            GridOrientation.Horizontal,
            BeMarginType.DefaultIntercellSpacing
          )
          val result = BeGrid(
            GridOrientation.Horizontal,
            BeMarginType.DefaultIntercellSpacing
          )
          val control = completionItem.leftIcon
          var content: BeControl = completionItem.leftText
          if (control != null) {
            val beIcon = BeIcon()
            beIcon.icon.set(control)
            localGrid.addElement(beIcon)
          }
          else {
            content = content.withMargin {
              margin(
                BeDslMargin(
                  BeMarginType.OnePx,
                  3
                )
              )
            }
          }
          val leftText = completionItem.leftText
          leftText.tryGetSettings()?.apply {
            this.fontStyle.set(myModel!!.settings.textSettings.fontStyle.valueOrDefault(BeFontStyle.DEFAULT))
            this.fontSize.set(myModel!!.settings.textSettings.fontSize.valueOrDefault(BeFontSize.DEFAULT))
          }
          localGrid.addElement(content, BeSizingType.Fill)
          result.addElement(localGrid)
          return defaultControlRenderer.getListCellRendererComponent(null, result, index, isSelected, cellHasFocus)
        }
      })
      .setAutoselectOnMouseMove(false)
      .setAutoSelectIfEmpty(false)
      .setResizable(true)
      .setMinSize(Dimension(size.width - 2, 0))
      .setItemChosenCallback {
        this.text = myJList?.selectedValue?.translatedText ?: text
        this.editor?.caretModel?.moveToOffset(text.length)
        showNextResults = false
      }

    val popup = myBuilder!!.createPopup()
    popup.addListener(object : JBPopupListener {
      override fun onClosed(event: LightweightWindowEvent) {
        myCurrentPopup = null
        myPopupIsShown = false
      }
    })
    @Suppress("UNCHECKED_CAST")
    myJList = myBuilder!!.chooserComponent as JBList<BeCompletionItem>?

    val pane = myBuilder!!.scrollPane
    pane?.border = null
    pane?.viewportBorder = null
    SwingUtilities.invokeLater { myCurrentPopup?.pack(true, true) }

    popup.showUnderneathOf(this)
    myCurrentPopup = popup
    myPopupIsShown = true
  }

  private fun setSuggestions(suggestions: List<BeCompletionItem>, lifetime: Lifetime) {
    mySuggestions = suggestions //.groupBy { it.translatedText }.map { it.value.first() }
    if (mySuggestions.isEmpty()) {
      hidePopup()
      return
    }

    if (myCurrentPopup != null) {
      val model = myJList?.model
      if (model is NameFilteringListModel<*>) {
        (model as NameFilteringListModel<BeCompletionItem>).replaceAll(suggestions)
        SwingUtilities.invokeLater({ myCurrentPopup?.pack(true, true) })
      }
    }
    else
      if (showNextResults)
        showPopup()

  }

  private fun addDataChangedListener(listener: (Boolean) -> Unit) {
    myListenerList.add(listener)
    attachListeners()
  }

  fun removeDataChangedListener(listener: (Boolean) -> Unit) {
    myListenerList.remove(listener)
    if (myListenerList.size == 0) {
      detachListeners()
    }
  }

  private fun attachListeners() {
    if (myDocumentListener == null) {
      val listener = MyDocumentListener()
      myDocumentListener = listener
      this.addDocumentListener(listener)
    }
  }

  private fun detachListeners() {
    val listener = myDocumentListener
    this.removeDocumentListener(myDocumentListener!!)
    myDocumentListener = null
  }

  private fun fireDataChanged(isFromCombo: Boolean) {
    for (aList in myListenerList) {
      aList(isFromCombo)
    }
  }
  /*
      override fun requestFocusInWindow(): Boolean {
          //return this.requestFocusInWindow()
      }*/

  override fun setEnabled(enabled: Boolean) {
    val wasEnabled = isEnabled
    super.setEnabled(enabled)
    if (enabled && enabled != wasEnabled) {
      registerEditorListeners()
    }
  }

  private inner class MyDocumentListener : DocumentListener {
    override fun beforeDocumentChange(event: DocumentEvent) {
    }

    override fun documentChanged(event: DocumentEvent) {
      fireDataChanged(false)
    }
  }
}