// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl

import com.jetbrains.ide.model.uiautomation.OrderBy
import com.jetbrains.ide.model.uiautomation.SortKey
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.IOptPropertyView

import java.util.*
import javax.swing.RowSorter
import javax.swing.SortOrder
import javax.swing.table.TableModel
import kotlin.collections.ArrayList

class BackendRowSorter(
  lifetime: Lifetime,
  private val tableModel: TableModel,
  private val columnToIsSortable: Array<Boolean>,
  viewToModelMapUpdates: IOptPropertyView<IntArray>,
  private val fireSortKeysUpdate: ((Array<com.jetbrains.ide.model.uiautomation.SortKey>) -> Unit)
) : RowSorter<TableModel>() {
  private var modelToViewMap: IntArray? = null
  private var viewToModelMap: IntArray? = null
  private var sortKeys: MutableList<SortKey> = mutableListOf()

  init {
    viewToModelMapUpdates.advise(lifetime) {
      updateViewToModelMap(it)
    }
  }

  private fun updateViewToModelMap(newViewToModelMap: IntArray) {
    val oldModelToViewMap = modelToViewMap
    viewToModelMap = newViewToModelMap
    val newModelToViewMap = IntArray(newViewToModelMap.size)
    newViewToModelMap.forEachIndexed { viewIdx, modelIdx -> newModelToViewMap[modelIdx] = viewIdx }
    modelToViewMap = newViewToModelMap
    fireRowSorterChanged(oldModelToViewMap)
  }


  override fun getModel(): TableModel {
    return tableModel
  }

  override fun setSortKeys(newKeys: MutableList<out SortKey>?) {
    val old = this.sortKeys
    if (newKeys != null && newKeys.isNotEmpty()) {
      val max = getColumnCount()
      for (key in newKeys) {
        require(!(key.column < 0 || key.column >= max)) { "Invalid SortKey" }
      }
      this.sortKeys = Collections.unmodifiableList(ArrayList(newKeys))
    }
    else {
      this.sortKeys = mutableListOf()
    }
    if (this.sortKeys != old) {
      fireSortOrderChanged()
      val keysToSend = this.sortKeys
      fireSortKeysUpdate(keysToSend.map { toSortKey(it) }.toTypedArray())
    }
  }

  private fun toSortKey(it: SortKey) = SortKey(it.column, toOrder(it.sortOrder))

  private fun toOrder(sortOrder: SortOrder): OrderBy {
    return when (sortOrder) {
      SortOrder.ASCENDING -> OrderBy.ASC
      SortOrder.DESCENDING -> OrderBy.DESC
      SortOrder.UNSORTED -> OrderBy.UNSORTED
    }
  }

  private fun getColumnCount(): Int {
    return columnToIsSortable.size
  }

  private fun isSortable(column: Int): Boolean {
    return columnToIsSortable[column]
  }

  override fun convertRowIndexToModel(index: Int): Int {
    return getMappingSafe(index, viewToModelMap)
  }


  override fun convertRowIndexToView(index: Int): Int {
    return getMappingSafe(index, modelToViewMap)
  }

  private fun getMappingSafe(index: Int, map: IntArray?): Int {
    if (map == null || isUpdating(map)) {
      if (index < 0 || index >= model.rowCount) {
        throw IndexOutOfBoundsException("Invalid index")
      }
      return index
    }
    return map[index]
  }

  private fun isUpdating(map: IntArray): Boolean {
    return model.rowCount != map.count()
  }

  override fun getViewRowCount(): Int {
    //NOTE(Korovin): No filtering supported on the frontend
    return modelRowCount
  }


  override fun getSortKeys(): MutableList<out SortKey> {
    return sortKeys
  }


  private fun toggle(key: SortKey): SortKey {
    return if (key.sortOrder == SortOrder.ASCENDING) {
      SortKey(key.column, SortOrder.DESCENDING)
    }
    else SortKey(key.column, SortOrder.ASCENDING)
  }

  override fun toggleSortOrder(column: Int) {
    if (isSortable(column)) {
      var keys: MutableList<SortKey> = ArrayList(sortKeys)
      val sortKey: SortKey
      var sortIndex: Int
      sortIndex = keys.size - 1
      while (sortIndex >= 0) {
        if (keys[sortIndex].column == column) {
          break
        }
        sortIndex--
      }
      when (sortIndex) {
        -1 -> {
          // Key doesn't exist
          sortKey = SortKey(column, SortOrder.ASCENDING)
          keys.add(0, sortKey)
        }
        0 -> // It's the primary sorting key, toggle it
          keys[0] = toggle(keys[0])
        else -> {
          // It's not the first, but was sorted on, remove old
          // entry, insert as first with ascending.
          keys.removeAt(sortIndex)
          keys.add(0, SortKey(column, SortOrder.ASCENDING))
        }
      }
      if (keys.size > getMaxSortKeys()) {
        keys = keys.subList(0, getMaxSortKeys())
      }
      setSortKeys(keys)
    }
  }

  private fun getMaxSortKeys(): Int {
    return 3
  }

  override fun getModelRowCount(): Int {
    return model.rowCount
  }

  override fun allRowsChanged() = Unit
  override fun modelStructureChanged() = Unit
  override fun rowsUpdated(firstRow: Int, endRow: Int) = Unit
  override fun rowsUpdated(firstRow: Int, endRow: Int, column: Int) = Unit
  override fun rowsInserted(firstRow: Int, endRow: Int) = Unit
  override fun rowsDeleted(firstRow: Int, endRow: Int) = Unit
}
