// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl

import com.intellij.openapi.actionSystem.DataProvider
import com.intellij.openapi.actionSystem.LangDataKeys
import com.intellij.ui.treeStructure.Tree
import com.intellij.util.ui.JBEmptyBorder
import com.jetbrains.ide.model.uiautomation.BeTreeGrid
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.ui.bindable.views.listControl.utils.getDeepestRendererComponentByEvent
import com.jetbrains.rd.ui.bindable.views.listControl.utils.pathToRoot
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdclient.util.idea.ui.bindToBeEmptyText
import java.awt.event.KeyEvent
import java.awt.event.MouseEvent
import javax.swing.JComponent
import javax.swing.KeyStroke
import javax.swing.SwingUtilities
import javax.swing.tree.DefaultTreeModel

class BeTree(
  lifetime: Lifetime,
  treeModel: DefaultTreeModel,
  val showAsList: Boolean,
  private val viewModel: BeTreeGrid
) : Tree(treeModel), DataProvider {
  private val logger = getLogger<BeTree>()

  init {
    // NOTE(Korovin): Without setting border on JTree top level expanders are being clipped by the left border
    if (!showAsList)
      border = JBEmptyBorder(0, 4, 0, 0)
    else {
      border = JBEmptyBorder(0)
      // otherwise tree handles enter and we can't register another enter action for popup with list
      val pressedEnter = KeyStroke.getKeyStroke(KeyEvent.VK_ENTER, 0)
      SwingUtilities.getUIInputMap(this, JComponent.WHEN_IN_FOCUSED_WINDOW)?.remove(pressedEnter)
      SwingUtilities.getUIInputMap(this, JComponent.WHEN_FOCUSED)?.remove(pressedEnter)
    }

    bindProgress(lifetime, viewModel.progress)
    {
      setPaintBusy(it)
    }

    this.emptyText.bindToBeEmptyText(lifetime, viewModel.emptyText)
  }

  override fun getData(dataId: String): Any? {
    return when {
      LangDataKeys.NO_NEW_ACTION.`is`(dataId) -> true
      else -> treeGetData(this, viewModel.controlId.value, dataId, logger)
    }
  }


  override fun shouldShowBusyIconIfNeeded(): Boolean {
    return true
  }

  private fun getRendererComponentToolTip(event: MouseEvent): String? {
    val point = event.point
    val row = getRowForLocation(point.x, point.y)
    val renderer = getCellRenderer()
    if (row == -1 || renderer == null)
      return null
    val path = getPathForRow(row)
    val lastPath = path.lastPathComponent
    val rendererComponent = renderer.getTreeCellRendererComponent(
      this, lastPath, isRowSelected(row),
      super.isExpanded(row), model.isLeaf(lastPath), row,
      true
    )
    if (rendererComponent !is JComponent)
      return null
    val pathBounds = getPathBounds(path) ?: return null
    val activeComponent = getDeepestRendererComponentByEvent(this, rendererComponent, pathBounds, event) ?: return null
    return activeComponent.pathToRoot(this).mapNotNull { (it as? JComponent)?.toolTipText }.firstOrNull()
  }

  override fun getToolTipText(event: MouseEvent?): String? {
    if (event == null) return toolTipText
    return getRendererComponentToolTip(event) ?: toolTipText
  }
}