// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl

import com.intellij.openapi.actionSystem.ActionGroup
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.CustomShortcutSet
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.ui.PopupHandler
import com.intellij.ui.treeStructure.Tree
import com.intellij.ui.treeStructure.treetable.TreeTable
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.IProperty
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rd.util.reactive.viewNotNull
import com.jetbrains.rd.util.string.printToString
import com.jetbrains.rd.ui.bindable.BeControlsConstantProvider
import com.jetbrains.rd.ui.actions.FrontendActionPlaces
import com.jetbrains.rd.ui.bindable.BeActionSupport
import com.jetbrains.rd.platform.util.TreeTraverse
import java.awt.Point
import javax.swing.JComponent
import javax.swing.event.TreeSelectionEvent
import javax.swing.plaf.basic.BasicTreeUI
import javax.swing.tree.DefaultMutableTreeNode
import javax.swing.tree.TreePath

fun Tree.setChildIndent(size: ChildOffsetSize) {
  val ui = ui as BasicTreeUI
  when (size) {
    ChildOffsetSize.NONE -> {
      ui.leftChildIndent = 0
      ui.rightChildIndent = 0
    }
    ChildOffsetSize.SMALL -> {
      ui.leftChildIndent = 3
      ui.rightChildIndent = 9
    }
    ChildOffsetSize.DEFAULT -> {
      // Leave as is
    }
  }
}

fun spaces(level: Int): String {
  return with(StringBuilder()) {
    repeat(level) { append(' ') }
    toString()
  }
}

fun BeTreeGridLine.propsToLog(): String {
  with(this) {
    return "isExpanded: ${isExpanded.value}, isSelected: ${isSelected.value}, children: ${children.count()}"
  }
}

fun DefaultMutableTreeNode.toLog(): String {
  if (this.isRoot)
    return "@root"
  return treeGridUserObject().toLog()
}

fun TreeSelectionEvent.toLog(): String {
  with(this) {
    return "TreeSelectionEvent \n" +
           "olsp: ${this.oldLeadSelectionPath}\n" +
           "nlsp: ${this.newLeadSelectionPath}\n" +
           "path: ${this.path}\n" +
           "isAdded: ${this.isAddedPath}\n" +
           "paths.size: ${paths.size}"
  }
}


fun TreeTable.getPathForLocation(p: Point): TreePath? {
  val row = this.rowAtPoint(p)
  val column = this.columnAtPoint(p)
  return if (row < 0 || column < 0) null else tree.getPathForRow(row)
}

fun TreeTable.nodeAtPoint(p: Point): Any? = getPathForLocation(p)?.lastPathComponent

fun Tree.nodeAtPoint(p: Point) =
  getClosestPathForLocation(p.x, p.y)?.lastPathComponent

fun getComponentAndTree(component: Any?): Pair<JComponent, Tree> {
  return when (component) {
    is TreeTable -> Pair(component, component.tree)
    is Tree -> Pair(component, component)
    else -> throw IllegalArgumentException("Component is not a Tree nor TreeTable")
  }
}

fun treeGetData(
  tree: Tree,
  enclosingControlId: String?,
  dataId: String,
  logger: Logger? = null
): List<BeControlConstant>? {
  if (dataId != BeControlsConstantProvider.dataId || tree.selectionCount == 0)
    return null
  val nodeIds = mutableListOf<String>()
  for (path in tree.selectionPaths) {
    nodeIds.add(treeNodeId(path))
  }
  logger?.trace { "treeGetData dataId: $dataId, controlId: $enclosingControlId, nodeIds ${nodeIds.printToString()}" }
  return listOf(BeControlConstant(enclosingControlId, nodeIds))
}

private fun treeNodeId(path: TreePath): String {
  val selectedNode = path.lastPathComponent as DefaultMutableTreeNode
  val selectedCell = selectedNode.treeGridUserObject().line
  val selectedNodeId = selectedCell.controlId.value
  return selectedNodeId
}

fun tableGetData(
  table: TreeTable,
  enclosingControlId: String?,
  dataId: String,
  logger: Logger? = null
): List<BeControlConstant>? {
  if (dataId != BeControlsConstantProvider.dataId || table.selectedRowCount == 0)
    return null
  val nodeIds = mutableListOf<String>()
  for (selectedRowViewIndex in table.selectedRows) {
    val modelIndex = table.convertRowIndexToModel(selectedRowViewIndex)
    val path = table.tree.getPathForRow(modelIndex)
    val selectedNodeId = treeNodeId(path)
    nodeIds.add(selectedNodeId)
  }
  logger?.trace({ "tableGetData dataId: $dataId, controlId: $enclosingControlId, nodeIds ${nodeIds.printToString()}" })
  return listOf(BeControlConstant(enclosingControlId, nodeIds))
}


fun JComponent.addBackendContextMenu(lt: Lifetime, project: Project, resharperGroupId: String) {
  val riderActionGroup = BeActionSupport.getInstance(project).createBackendActionGroup(resharperGroupId) ?: return
  val popupMenuListener =
    PopupHandler.installPopupHandler(this, riderActionGroup, FrontendActionPlaces.BE_CONTROLS_CONTEXT_MENU, ActionManager.getInstance())
  lt.onTermination { this.removeMouseListener(popupMenuListener) }
  registerShortcuts(lt, project, resharperGroupId)
}

fun JComponent.registerShortcuts(
  lt: Lifetime,
  project: Project,
  resharperGroupId: String
) {
  val beActionSupport = BeActionSupport.getInstance(project)
  val group = beActionSupport.getBackendActionGroup(resharperGroupId) ?: return
  val disposable = lt.createNestedDisposable()
  TreeTraverse.thisAndDescendantsDfs(group) {
    when (it) {
      is ActionGroup -> it.getChildren(null).asSequence()
      else -> emptySequence()
    }
  }.mapNotNull { action -> beActionSupport.getShortcut(action)?.let { action to it } }
    .forEach { (action, text) ->
      action.registerCustomShortcutSet(
        CustomShortcutSet(text),
        this,
        disposable
      )
    }
}

fun bindProgress(
  lifetime: Lifetime,
  progress: IProperty<AbstractProgress?>,
  setBusy: (Boolean) -> Unit
) {
  progress.viewNotNull(lifetime) { lt, progress ->
    when (progress) {
      is IndefiniteProgress -> {
        progress.inProgress.adviseNotNull(lt) {
          setBusy(it)
        }
      }
    }
  }
}