// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl

import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.wm.IdeFocusManager
import com.intellij.ui.*
import com.intellij.ui.speedSearch.SpeedSearchSupply
import com.intellij.ui.treeStructure.Tree
import com.intellij.ui.treeStructure.treetable.TreeTable
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.BeCheckbox
import com.jetbrains.ide.model.uiautomation.BeTreeGrid
import com.jetbrains.ide.model.uiautomation.ControlVisibility
import com.jetbrains.rd.ui.bedsl.extensions.UpdateFlag
import com.jetbrains.rd.ui.bedsl.extensions.descendantsOfType
import com.jetbrains.rd.ui.bedsl.extensions.ignoreOnReentrancy
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.util.reactive.valueOrDefault
import java.awt.Point
import java.awt.event.*
import javax.swing.JComponent
import javax.swing.JTree
import javax.swing.SwingUtilities
import javax.swing.event.*
import javax.swing.tree.DefaultMutableTreeNode
import javax.swing.tree.TreePath

fun toggleCheckboxOnSpaceBar(): KeyListener {
  return object : KeyAdapter() {
    fun isToggleEvent(component: JComponent, e: KeyEvent): Boolean {
      return e.keyCode == KeyEvent.VK_SPACE && SpeedSearchSupply.getSupply(component) == null
    }

    override fun keyPressed(e: KeyEvent) {
      val (component, tree) = getComponentAndTree(e.source)

      if (isToggleEvent(component, e)) {
        val selectionPaths = tree.selectionPaths
        if (selectionPaths != null) {
          for (path in selectionPaths) {
            val node = path.lastPathComponent as DefaultMutableTreeNode
            toggle(node)
            component.repaint()
          }
        }
        e.consume()
        component.repaint()
      }
    }
  }
}

private fun toggle(o: DefaultMutableTreeNode): Boolean {
  val userObject = o.treeGridUserObject()
  val checkBox = userObject.line.descendantsOfType(BeCheckbox::class.java).firstOrNull() ?: return false
  if(checkBox.visible.valueOrDefault(ControlVisibility.Visible) != ControlVisibility.Visible) return false

  checkBox.property.value = !(checkBox.property.value ?: false)
  return true
}

fun navigateOnDoubleClick(navigateOverride: ((MouseEvent) -> Unit)? = null): DoubleClickListener {
  return object : DoubleClickListener() {
    override fun onDoubleClick(event: MouseEvent): Boolean {
      event
      if (navigateOverride != null) {
        navigateOverride(event)
        return true
      }
      else {
        val node = findTreeNodeByEvent(event)
        if (node == null)
          return false
        if (node.childCount > 0)
          return false
        val userObject = node.treeGridUserObject()
        userObject.line.signal.fire()
        return true
      }
    }

    private fun findTreeNodeByEvent(e: MouseEvent): DefaultMutableTreeNode? {
      val source = e.source
      return when (source) {
        is TreeTable -> source.nodeAtPoint(e.point) as? DefaultMutableTreeNode
        is Tree -> source.nodeAtPoint(e.point) as? DefaultMutableTreeNode
        else -> null
      }
    }
  }
}

fun sendExpandToModel(
  updateFromUi: UpdateFlag,
  logger: Logger
): TreeWillExpandListener {
  return object : TreeWillExpandListener {
    override fun treeWillCollapse(event: TreeExpansionEvent) {
      setExpanded(event.path.lastPathComponent, false)
    }

    override fun treeWillExpand(event: TreeExpansionEvent) {
      setExpanded(event.path.lastPathComponent, true)
    }

    private fun setExpanded(inputNode: Any?, isExpanded: Boolean) {
      updateFromUi.ignoreOnReentrancy {
        val node = inputNode as DefaultMutableTreeNode
        if (node.isRoot)
          return@ignoreOnReentrancy
        logger.trace { "setExpanded (${node.toLog()}) : $isExpanded" }
        val userObject = node.treeGridUserObject()
        userObject.line.isExpanded.set(isExpanded)
      }
    }
  }
}

fun sendSelectionToModel(
  updateFromUi: UpdateFlag,
  logger: Logger
): TreeSelectionListener {
  return object : TreeSelectionListener {
    override fun valueChanged(e: TreeSelectionEvent?) {
      e ?: return
      logger.trace { "changed ${e.toLog()}" }
      for (i in e.paths.indices) {
        updateModelSelection(e.paths[i], e.isAddedPath(i))
      }
    }

    private fun updateModelSelection(path: TreePath, isSelected: Boolean) {
      updateFromUi.ignoreOnReentrancy {
        val node = path.lastPathComponent as DefaultMutableTreeNode
        logger.trace({ "updateModelSelection set (${node.toLog()}), isSelected: $isSelected" })
        node.treeGridUserObject().line.isSelected.set(isSelected)
      }
    }
  }
}

@Suppress("ObjectLiteralToLambda")
fun autoScrollToSelection(tree: Tree) : TreeSelectionListener {
  return object : TreeSelectionListener {
    override fun valueChanged(e: TreeSelectionEvent?) {
      tree.scrollPathToVisible(e?.newLeadSelectionPath)
    }
  }
}

fun autoScrollToEnd(tree: Tree): TreeModelListener {
  return object : TreeModelListener {
    private fun scrollToEnd() {
      val row = tree.rowCount - 1
      val lastRowPath = tree.getPathForRow(row)
      tree.scrollPathToVisible(lastRowPath)
    }

    override fun treeNodesInserted(e: TreeModelEvent?) {
      scrollToEnd()
    }

    override fun treeStructureChanged(e: TreeModelEvent?) {
      scrollToEnd()
    }

    override fun treeNodesChanged(e: TreeModelEvent?) {
    }

    override fun treeNodesRemoved(e: TreeModelEvent?) {
      scrollToEnd()
    }
  }
}

fun JComponent.enableStartEditingOnFocus(
  viewModel: BeTreeGrid,
  lifetime: Lifetime
) {
  putClientProperty("JTable.autoStartsEdit", true)
  val focusManager = IdeFocusManager.getGlobalInstance()
  viewModel.focus.advise(lifetime) {
    focusManager.doWhenFocusSettlesDown {
      SwingUtilities.invokeLater {
        if (this is TreeTable) {
          startEditing()
        }
        else if (this is JTree) {
          startEditingAtPath(selectionPath)
        }
      }
    }
  }
}

private fun TreeTable.startEditing() {
  cellEditor?.stopCellEditing() // store user input
  var row = selectedRow
  if (row < 0) {
    row = tree.selectionRows.min() ?: -1
  }
  scrollRectToVisible(getCellRect(row, 0, true))
  if (editCellAt(row, 0)) {
    val editorComponent = this.editorComponent as? JComponent
    val ideFocusManager = IdeFocusManager.getGlobalInstance()
    if (editorComponent != null && ideFocusManager.focusOwner != editorComponent) {
      ideFocusManager.getFocusTargetFor(editorComponent)?.requestFocusInWindow()
    }
  }
}

fun sendMouseClickToModel(
  component: JComponent,
  viewModel: BeTreeGrid
):MouseAdapter{
  return object: MouseAdapter() {
    override fun mouseReleased(e: MouseEvent?) {
      if(e == null || UIUtil.isActionClick(e, MouseEvent.MOUSE_RELEASED) && !UIUtil.isSelectionButtonDown(e) && !e.isConsumed)
        viewModel.mouseClicked.fire()
    }
  }
}
fun sendKeyboardEnterToModel(
  component: JComponent,
  viewModel: BeTreeGrid
): KeyAdapter {
  return object : KeyAdapter() {
    override fun keyPressed(e: KeyEvent) {
      if (e.keyCode == KeyEvent.VK_ENTER) {
        viewModel.enter.fire()
      }
    }
  }
}
fun sendFocusGainedToModel(
  component: JComponent,
  viewModel: BeTreeGrid
): FocusAdapter {
  return object : FocusAdapter() {
    override fun focusGained(e: FocusEvent?) {
      if (!component.isFocusOwner)
        viewModel.focus.fire()
    }
  }
}
internal fun JComponent.getMouseMotionListener(): MouseMotionListener? {
 return when(this){
   is BeTree -> TreeMouseMotionListener(this)
   is BeTreeTable -> TreeTableMouseMotionListener(this)
   else -> null
 }
}
internal fun JComponent.addSpeedSearch() {
   when(this){
    is BeTree ->{
      val speedSearch = TreeSpeedSearch(this)
      speedSearch.comparator = SpeedSearchComparator(false)
    }

    is BeTreeTable -> {
      val speedSearch = TreeTableSpeedSearch(this)
      speedSearch.comparator = SpeedSearchComparator(false)
    }
  }

}
  private class TreeMouseMotionListener(private val tree:BeTree) : MouseMotionAdapter() {
    private var myLastSelectedIndex = -2
    private var myIsEngaged: Boolean = false
    private var myLastMouseLocation: Point? = null
    private fun isMouseMoved(location: Point): Boolean {
      if (myLastMouseLocation == null) {
        myLastMouseLocation = location
        return false
      }
      return myLastMouseLocation != location
    }

    override fun mouseMoved(e: MouseEvent) {
      if (!isMouseMoved(e.locationOnScreen)) return
      val point = e.point
      if (myIsEngaged && !UIUtil.isSelectionButtonDown(e) && tree.selectionCount <= 1){
          /*&& tree.selectionModel.selectionMode == ListSelectionModel.SINGLE_SELECTION) && focusOwner !is JRootPane*/
        val index: Int = tree.getRowForLocation(point.x, point.y)//.nodeAtPoint(point)
        //if (tree.isSelectableAt(index)) {
        if (index != myLastSelectedIndex) {
          tree.putClientProperty(ListUtil.SELECTED_BY_MOUSE_EVENT, java.lang.Boolean.TRUE)
          tree.setSelectionRow(index)
          //restartTimer()
          tree.putClientProperty(ListUtil.SELECTED_BY_MOUSE_EVENT, java.lang.Boolean.FALSE)
          myLastSelectedIndex = index
        }
        //}
        else {
          //treeTable.clearSelection()
          //myLastSelectedIndex = -1
        }
        //treeTable.notifyParentOnChildSelection()
      }
      else {
        myIsEngaged = true
      }
    }
  }


class TreeTableMouseMotionListener(private val treeTable: BeTreeTable) : MouseMotionAdapter() {
  private var myIsEngaged: Boolean = false
  private var myLastSelectedIndex = -2
    private var myLastMouseLocation: Point? = null
    private fun isMouseMoved(location: Point): Boolean {
      if (myLastMouseLocation == null) {
        myLastMouseLocation = location
        return false
      }
      return myLastMouseLocation != location
    }

    override fun mouseMoved(e: MouseEvent) {
      if (!isMouseMoved(e.locationOnScreen)) return
      val point = e.point
      if (myIsEngaged && !UIUtil.isSelectionButtonDown(e) && treeTable.selectedRowCount <= 1){
          /*&& treeTable.selectionModel.selectionMode == ListSelectionModel.SINGLE_SELECTION) && focusOwner !is JRootPane*/
        val index: Int = treeTable.rowAtPoint(point)
        //if (treeTable.isSelectableAt(index)) {
          if (index != myLastSelectedIndex && index >= 0) {
            treeTable.putClientProperty(ListUtil.SELECTED_BY_MOUSE_EVENT, java.lang.Boolean.TRUE)
              treeTable.setRowSelectionInterval(index, index)
                //restartTimer()
            treeTable.putClientProperty(ListUtil.SELECTED_BY_MOUSE_EVENT, java.lang.Boolean.FALSE)
            myLastSelectedIndex = index
          }
        //}
        else {
          //treeTable.clearSelection()
          //myLastSelectedIndex = -1
        }
        //treeTable.notifyParentOnChildSelection()
      }
      else {
        myIsEngaged = true
      }
    }
  }
