// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl

import com.intellij.openapi.actionSystem.ActionPlaces
import com.intellij.openapi.actionSystem.ex.ActionUtil
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.ui.FilteringTree
import com.intellij.ui.JBColor
import com.intellij.ui.ScrollingUtil
import com.intellij.ui.TableUtil
import com.intellij.ui.components.JBScrollPane
import com.intellij.ui.speedSearch.SpeedSearch
import com.intellij.ui.treeStructure.Tree
import com.intellij.ui.treeStructure.treetable.ListTreeTableModelOnColumns
import com.intellij.ui.treeStructure.treetable.TreeTable
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.intellij.util.ui.tree.TreeUtil
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.dsl.getText
import com.jetbrains.rd.ui.bedsl.extensions.UpdateFlag
import com.jetbrains.rd.ui.bedsl.extensions.descendantsOfType
import com.jetbrains.rd.ui.bedsl.extensions.ignoreOnReentrancy
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.ui.addMouseMotionListener
import com.jetbrains.rd.ui.addTreeModelListener
import com.jetbrains.rd.ui.addTreeSelectionListener
import com.jetbrains.rd.ui.bindable.BeActionSupport
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.RdCellEditor
import com.jetbrains.rd.ui.bindable.views.listControl.renderers.RdCellRenderer
import com.jetbrains.rd.ui.bindable.views.utils.PassFocusThroughTo
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.AddRemove
import com.jetbrains.rd.util.reactive.IMutableViewableList
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.util.reactive.valueOrDefault
import java.awt.BorderLayout
import java.awt.Color
import java.awt.Component
import java.awt.Dimension
import java.awt.event.ComponentAdapter
import java.awt.event.ComponentEvent
import java.awt.event.MouseEvent
import javax.swing.*
import javax.swing.border.Border
import javax.swing.tree.DefaultMutableTreeNode
import javax.swing.tree.DefaultTreeModel
import javax.swing.tree.TreePath
import javax.swing.tree.TreeSelectionModel
import kotlin.math.min

fun BeShowBorders.toSwingBorder(): Border {
  return when (this) {
    BeShowBorders.All -> JBUI.Borders.customLine(JBColor.border())
    BeShowBorders.OnlyTop -> JBUI.Borders.customLine(JBColor.border(), 1, 0, 0, 0)
    BeShowBorders.OnlyBottom -> JBUI.Borders.customLine(JBColor.border(), 0, 0, 1, 0)
    BeShowBorders.TopAndBottom -> JBUI.Borders.customLine(JBColor.border(), 1, 0, 1, 0)
    BeShowBorders.None -> JBUI.Borders.customLine(JBColor.border(), 0, 0, 0, 0)
  }
}

class BeTreeOrTable(lifetime: Lifetime,
                    private val viewModel: BeTreeGrid,
                    private val project: Project
){
  private val panel = JPanel(BorderLayout())
  private val logger = getLogger<BeTreeOrTable>()
  private val selectionUpdateFromUi = UpdateFlag()
  private val expandedUpdateFromUi = UpdateFlag()
  private val scrollPane = JBScrollPane()
  private val mergingQueue: MergingUpdateQueue = MergingUpdateQueue("BeTreeOrTableResize", 200, true, scrollPane, lifetime.createNestedDisposable())


  init {
    val columnAlignments = viewModel.configuration.columnDefinitions.map { it.alignment }
    val treeTableCellRenderer = RdCellRenderer(project, false, columnAlignments,
                                               viewModel.configuration.speedSearch.valueOrNull == BeSpeedSearch.ENABLED_HIGHLIGHT)


    val columns = viewModel.configuration.columnDefinitions
      .mapIndexed { index, rdColumnDefinition ->
        TreeGridListColumn(
          project,
          treeTableCellRenderer,
          rdColumnDefinition,
          index,
          columnAlignments
        )
      }.toTypedArray()

    val root = DefaultMutableTreeNode(null)
    val showAsList = viewModel.configuration.showAsList

    val (treeModel: DefaultTreeModel, tree: Tree, component: JComponent) = buildTreeOrTableCore(lifetime, columns, root, showAsList, viewModel)

    scrollPane.setViewportView(component)
    panel.add(scrollPane, BorderLayout.CENTER)

    scrollPane.border = viewModel.configuration.showBorders.toSwingBorder()
    // todo: border = JBUI.Borders.customLine(RiderTheme.BeFramework.separatorColor)
    val visibleRowCount = viewModel.configuration.visibleRowsCount

    if (component is TreeTable) {
      val selectionMode = if (viewModel.configuration.selection == BeTreeSelection.Multi)
        ListSelectionModel.MULTIPLE_INTERVAL_SELECTION
      else ListSelectionModel.SINGLE_SELECTION
      component.setSelectionMode(selectionMode)
      component.tree.isRequestFocusEnabled = false
      component.border = JBUI.Borders.empty()

      //todo: change defaults here, reuse visible row count
      component.preferredScrollableViewportSize = preferredSize(component.rowHeight)
      enableManualCellSizer(component)
      if(showAsList && viewModel.configuration.infiniteCycleScroll.valueOrDefault(false)) {
        ScrollingUtil.installActions(component)
      }
    }
    if (component is Tree) {
      val selectionMode = if (viewModel.configuration.selection == BeTreeSelection.Multi)
        TreeSelectionModel.DISCONTIGUOUS_TREE_SELECTION
      else TreeSelectionModel.SINGLE_TREE_SELECTION
      component.selectionModel.selectionMode = selectionMode
      //todo: change defaults here
      if(visibleRowCount > 0)
        component.visibleRowCount = visibleRowCount
      /*if (viewModel.configuration.showAsList) {
        //SwingUtilities.invokeLater {
        this.viewport.preferredSize = preferredSize(tree.preferredSize.width, min(viewModel.configuration.visibleRowsCount, tree.rowCount))
        // }
      }*/
    }

    scrollPane.addFocusListener(PassFocusThroughTo(component))

    tree.cellRenderer = treeTableCellRenderer
    tree.cellEditor = RdCellEditor(project, tree, columnAlignments)
    tree.rowHeight = 0
    component.addKeyListener(toggleCheckboxOnSpaceBar())
    navigateOnDoubleClick(overrideNavigateIfDoubleClickActionIdSet(component, viewModel)).installOn(component)
    component.addFocusListener(sendFocusGainedToModel(component, viewModel))
    component.addMouseListener(sendMouseClickToModel(component, viewModel))
    component.addKeyListener(sendKeyboardEnterToModel(component, viewModel))
    if (viewModel.configuration.autoStartEditing)
      component.enableStartEditingOnFocus(viewModel, lifetime)

    viewModel.configuration.backendContextMenu?.let {
      component.addBackendContextMenu(lifetime, project, it.backendMenuGroupId)
    }

    tree.isEditable = true
    tree.isRootVisible = false
    tree.setChildIndent(viewModel.configuration.childOffsetSize)
    tree.addTreeSelectionListener(sendSelectionToModel(selectionUpdateFromUi, logger))
    tree.addTreeWillExpandListener(sendExpandToModel(expandedUpdateFromUi, logger))
    tree.showsRootHandles = !showAsList
    tree.setExpandableItemsEnabled(!showAsList)

    viewModel.configuration.autoScroll.view(lifetime){ lt, scrollSettings ->
      when(scrollSettings) {
        BeScrollSettings.AUTOSCROLL_TO_END -> treeModel.addTreeModelListener(lt, autoScrollToEnd(tree))
        BeScrollSettings.AUTOSCROLL_TO_SELECTION -> tree.addTreeSelectionListener(lt, autoScrollToSelection(tree))
        else -> {}
      }
    }

    viewModel.configuration.selectOnMouseOver.view(lifetime){ lt, it ->
      if(it)
        component.addMouseMotionListener(lt, component.getMouseMotionListener())
    }

    addFiltersAndSpeedSearchIfNecessary(component, lifetime, treeTableCellRenderer)
    subscribeChildrenToUpdates(lifetime, tree, component, treeModel, root, viewModel.fakeRoot.children)
    viewModel.enabled.advise(lifetime) { component.isEnabled = it }
    val currentBorder = this.scrollPane.border
    viewModel.validationResult.advise(lifetime) {
      this.scrollPane.border = when (it.state) {
        ValidationStates.validationError -> JBUI.Borders.customLine(Color.red.darker())
        //this.putClientProperty("JComponent.outline", "error")
        ValidationStates.validationWarning -> JBUI.Borders.customLine(Color.orange)
        //this.putClientProperty("JComponent.outline", "warning")
        ValidationStates.validationPassed -> currentBorder
        //this.putClientProperty("JComponent.outline", null)
        else -> currentBorder
      }
      component.toolTipText = it.resultMessage
    }
    viewModel.revalidate.fire()


  }

  private fun addFiltersAndSpeedSearchIfNecessary(component:JComponent, lifetime:Lifetime, cellRenderer:RdCellRenderer) {

    when (viewModel.configuration.speedSearch.valueOrNull) {
      BeSpeedSearch.ENABLED_FILTER_AUTO, BeSpeedSearch.ENABLED_FILTER_ALWAYS_VISIBLE -> {

        when (component) {
          is Tree -> {
            val filteringTree = object : FilteringTree<DefaultMutableTreeNode, TreeGridUserObject>(project, component,
                                                                                                   DefaultMutableTreeNode(
                                                                                                     TreeGridUserObject(
                                                                                                       lifetime.createNested(),
                                                                                                       this.viewModel.fakeRoot))) {
              override fun getNodeClass(): Class<out DefaultMutableTreeNode> {
                return DefaultMutableTreeNode::class.java
              }

              override fun createNode(obj: TreeGridUserObject): DefaultMutableTreeNode {
                return DefaultMutableTreeNode(obj)
              }

              override fun getChildren(obj: TreeGridUserObject): MutableIterable<TreeGridUserObject> {
                val lifetimeDefinition = lifetime.createNested()
                return obj.line.children.toList().map { TreeGridUserObject(lifetimeDefinition, it as BeTreeGridLine) }.toMutableList()
              }

              override fun getText(obj: TreeGridUserObject?): String? {
                return obj?.line?.columnElements?.firstOrNull()?.descendantsOfType(
                  BeAbstractText::class.java)?.toList()?.firstOrNull()?.getText()
              }

              override fun useIdentityHashing(): Boolean {
                return false
              }
            }
            val searchField = filteringTree.installSearchField()
            panel.add(searchField, BorderLayout.NORTH)
            val speedSearch = filteringTree.searchModel.speedSearch as SpeedSearch

            if(viewModel.configuration.speedSearch.valueOrNull == BeSpeedSearch.ENABLED_FILTER_AUTO) {
              component.addKeyListener(speedSearch)
              var inUpdate = true
              var searchFieldShown = false
              searchField.isVisible = false
              searchField.textEditor.isFocusable = false
              speedSearch.addChangeListener {

                searchField.textEditor.background = UIUtil.getTextFieldBackground()
                //onSpeedSearchPatternChanged()
                //searchField.setText(getFilter())
                if (!searchFieldShown) {
                  searchField.isVisible = true
                  searchFieldShown = true
                }
                else {
                  if ((it.newValue as String).isEmpty() && searchFieldShown) {
                    searchField.isVisible = false
                    searchFieldShown = false
                  }
                }

                inUpdate = false
                panel.revalidate()
              }
            }
          }
        }
      }
      BeSpeedSearch.ENABLED_HIGHLIGHT -> {
          component.addSpeedSearch()
      }
    }

  }


  private fun overrideNavigateIfDoubleClickActionIdSet(component: Component, viewModel: BeTreeGrid): ((MouseEvent) -> Unit)? {
    var navigateOverride: ((MouseEvent) -> Unit)? = null
    val backendContextMenu = viewModel.configuration.backendContextMenu
    if (backendContextMenu != null) {
      val doubleClickActionId = backendContextMenu.doubleClickActionId
      if (doubleClickActionId != null) {
        val beActionSupport = BeActionSupport.getInstance(project)
        val action = beActionSupport.tryFindBackendAction(backendContextMenu.backendMenuGroupId, doubleClickActionId)
        if (action != null) {
          navigateOverride = {
            ActionUtil.invokeAction(action, component, ActionPlaces.UNKNOWN, it, null)
          }
        }
      }
    }
    return navigateOverride
  }

  private fun preferredSize(preferredWidth:Int = 400, visibleRowCount:Int = 7,  rowHeight: Int = 22): Dimension {
    return Dimension(preferredWidth, rowHeight * visibleRowCount)
  }

  /**
   * Enabling this for TreeTable with tree in first column breaks tree layouting pretty hard.
   * If we could debug it one day it will be good
   */
  private fun enableManualCellSizer(component: TreeTable) {
    component.autoResizeMode = JTable.AUTO_RESIZE_OFF
    val adjuster = TableCellSizeAdjuster(component,{scrollPane.viewport.size.width} , 10)
    adjuster.adjustCells()
    scrollPane.addComponentListener(object : ComponentAdapter() {
      override fun componentResized(e: ComponentEvent?) {
        mergingQueue.cancelAllUpdates()
        mergingQueue.queue(object : Update("") {
          override fun run() {
            adjuster.adjustCells()
          }
        })
      }
    })
  }

  private fun buildTreeOrTableCore(
    lifetime: Lifetime,
    columns: Array<TreeGridListColumn>,
    root: DefaultMutableTreeNode,
    showAsList: Boolean,
    viewModel: BeTreeGrid
  ): Triple<DefaultTreeModel, Tree, JComponent> {
    val treeModel: DefaultTreeModel
    val tree: Tree
    val component: JComponent
    if (columns.count() > 1) {
      val tableModel = ListTreeTableModelOnColumns(root, columns)
      val treeTable = BeTreeTable(
        lifetime,
        tableModel,
        showAsList,
        columns,
        viewModel
      )
      treeModel = tableModel
      tree = treeTable.tree
      if (columns.all { it.columnDefinition.columnName.isBlank() } || !viewModel.configuration.hasHeader)
        treeTable.tableHeader = null
      val sortableColumns = columns.map { it.columnDefinition.isSortable }.toTypedArray()
      if (sortableColumns.any { it }) {
        treeTable.rowSorter = BackendRowSorter(lifetime, treeTable.model, sortableColumns, viewModel.viewToModelMap) { newKeys ->
          viewModel.sortKeys.set(newKeys)
        }
      }
      component = treeTable
    }
    else {
      treeModel = DefaultTreeModel(root)
      tree = BeTree(lifetime, treeModel, showAsList, viewModel)
      component = tree
    }

    return Triple(treeModel, tree, component)
  }

  private fun invalidateTreeOnRendererUpdate(
    it: BeTreeGridLine,
    lifetime: Lifetime,
    treeModel: DefaultTreeModel,
    newChild: DefaultMutableTreeNode
  ) {
    it.columnElements.forEach {
      ViewRegistry.getInstance(project).getRenderer(it)?.addUpdate(it, lifetime) {
        treeModel.nodeChanged(newChild)
      }
    }
  }

  private fun subscribeChildrenToUpdates(
    parentLifetime: Lifetime,
    tree: Tree,
    component: JComponent,
    treeModel: DefaultTreeModel,
    parent: DefaultMutableTreeNode,
    children: IMutableViewableList<BeControl>
  ) {
    val name = "${spaces(parent.level)}subChildren"
    logger.trace { "$name (${parent.toLog()})" }
    children.adviseAddRemove(
      parentLifetime
    ) { addRemove: AddRemove, i: Int, line: BeControl ->
      if (addRemove == AddRemove.Add) {
        val nodeLifetime = parentLifetime.createNested()
        val newUserObject = TreeGridUserObject(
          nodeLifetime,
          line as BeTreeGridLine
        )
        logger.trace { "$name newUserObject ${newUserObject.toLog()}" }
        var newChild = DefaultMutableTreeNode(newUserObject)
        treeModel.insertNodeInto(newChild, parent, i)
        val searchTreeModel = tree.model as? FilteringTree.SearchTreeModel<*, TreeGridUserObject>
        newChild = searchTreeModel?.getNode(newUserObject) ?: newChild
        searchTreeModel?.refilter()
        logger.trace { "$name newChild (${parent.toLog()}):(${newChild.toLog()})" }
        subscribeChildrenToUpdates(nodeLifetime.lifetime, tree, component, treeModel, newChild, line.children)
        if (parent.isRoot || parent.treeGridUserObject().line.isExpanded.value) {
          logger.trace { "$name expand parent path (${parent.toLog()})" }
          tree.expandPath(TreePath(parent.path))
        }
        invalidateTreeOnRendererUpdate(line, nodeLifetime.lifetime, treeModel, newChild)
        bindExpanded(nodeLifetime.lifetime, line, name, newChild, tree)
        bindSelection(nodeLifetime.lifetime, component, line, newChild)
      }
      if (addRemove == AddRemove.Remove) {
        val node = parent.getChildAt(i) as DefaultMutableTreeNode
        logger.trace { "$name remove (${parent.toLog()}):(${node.toLog()})" }
        treeModel.removeNodeFromParent(node)

        val userObject = node.treeGridUserObject()
        userObject.lifetimeDef.terminate()
      }
      (tree.model as?FilteringTree.SearchTreeModel<*, *>)?.refilter()

      if (viewModel.configuration.showAsList) {
        mergingQueue.cancelAllUpdates()
        mergingQueue.queue(object : Update("") {
          override fun run() {
            scrollPane.viewport.preferredSize = preferredSize(tree.preferredSize.width,
                                                              min(viewModel.configuration.visibleRowsCount, tree.rowCount))
          }
        })
      }
    }
  }

  private fun bindExpanded(
    lifetime: Lifetime,
    line: BeTreeGridLine,
    name: String,
    node: DefaultMutableTreeNode,
    tree: Tree
  ) {
    line.isExpanded.advise(lifetime) { isExpand ->
      expandedUpdateFromUi.ignoreOnReentrancy {
        val pathToNewChild = TreePath(node.path)
        //NOTE(Korovin): Swing expandPath and collapsePath are expanding whole path from root to the node
        //and after that make last node expand/collapse respectively. We don't want to change parent states
        //so we skip all updates if some of our parents is collapsed.
        if (!tree.isExpanded(pathToNewChild.parentPath)) {
          logger.trace { "$name path to (${node.toLog()}) has collapsed nodes. Ignoring." }
          return@ignoreOnReentrancy
        }
        if (isExpand) {
          logger.trace { "$name expand path (${node.toLog()})" }
          tree.expandPath(pathToNewChild)
        }
        else {
          logger.trace { "$name collapse path (${node.toLog()})" }
          tree.collapsePath(pathToNewChild)
        }
      }
    }
  }

  private fun bindSelection(
    lifetime: Lifetime,
    component: JComponent,
    line: BeTreeGridLine,
    node: DefaultMutableTreeNode
  ) {
    logger.trace { "bindSelection (${node.toLog()})" }
    if (line.isSelected.value) {
      selectNode(component, node, "static")
    }
    line.isSelected.advise(lifetime) { isSelected ->
      selectionUpdateFromUi.ignoreOnReentrancy {
        if (isSelected)
          selectNode(component, node, "dynamic")
      }
    }
  }

  private fun selectNode(
    component: JComponent,
    node: DefaultMutableTreeNode,
    logAction: String
  ) {
    logger.trace { "selectNode $logAction (${node.toLog()})" }
    if(component is TreeTable) {
      val row = TreeUtil.getRowForNode(component.tree, node)
      TableUtil.selectRows(component, intArrayOf(row))
    }
    if(component is JTree)
      TreeUtil.selectPath(component, TreeUtil.getPathFromRoot(node), false)
  }

  fun tryGetComponent(): JComponent {

    return panel
  }
}





