// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl

import com.intellij.openapi.diagnostic.trace
import com.intellij.ui.treeStructure.treetable.ListTreeTableModelOnColumns
import com.intellij.ui.treeStructure.treetable.TreeTable
import com.intellij.ui.treeStructure.treetable.TreeTableCellRenderer
import com.jetbrains.ide.model.uiautomation.BeSizingType
import com.jetbrains.rd.platform.util.getLogger
import java.awt.Dimension
import javax.swing.SwingUtilities
import javax.swing.UIManager
import javax.swing.event.TableModelEvent
import javax.swing.event.TableModelListener
import javax.swing.table.TableCellRenderer
import javax.swing.table.TableModel
import kotlin.math.roundToInt

class TableCellSizeAdjuster(
  private val table: TreeTable,
  private val getAvailableWidth: () -> Int,
  private val headerExtraWidth: Int
) :
  TableModelListener {
  private val logger = getLogger<TableCellSizeAdjuster>()


  init {
    table.addPropertyChangeListener {
      //  When the TableModel changes we need to update the listeners
      //  and column widths
      if ("model" == it.propertyName) {
        var model = it.oldValue as TableModel
        model.removeTableModelListener(this)
        model = it.newValue as TableModel
        model.addTableModelListener(this)
        adjustCells()
      }
    }
    table.model.addTableModelListener(this)
  }


  companion object {
    private data class ColumnData(
      val type: BeSizingType,
      val multiplier: Int?,
      val minWidth: Int
    )

    /*
   *  Adjust the widths and heigths of all cells in the table
   */
    fun adjustCells(treeTable: TreeTable, maxWidth: Int, headerExtraWidth: Int) {
      if (treeTable.tableModel !is ListTreeTableModelOnColumns) return
      val columnInfos = (treeTable.tableModel as ListTreeTableModelOnColumns).columnInfos
      val tableColumns = treeTable.columnModel.columns.toList()
      var fillMultiplierSum = 0
      val columnsData = mutableListOf<ColumnData>()
      var lastFillColumnIndex = 0
      for (columnIndex in tableColumns.indices) {
        val columnInfo = columnInfos[columnIndex] as TreeGridListColumn
        val type = columnInfo.columnDefinition.size.unit
        val multiplier = columnInfo.columnDefinition.size.multiplier
        if (type == BeSizingType.Fill) {
          fillMultiplierSum += multiplier
          lastFillColumnIndex = columnIndex
        }
        val headerWidth = getColumnHeaderWidth(treeTable, columnIndex) + optionalSortIconWidth(columnInfo) + headerExtraWidth
        columnsData.add(ColumnData(type, multiplier, headerWidth))
      }

      var preferredRowHeight = 0
      for (rowIndex in 0 until treeTable.rowCount) {
        for (columnIndex in 0 until tableColumns.count()) {
          val columnData = columnsData[columnIndex]
          val cellDataSize = getCellDataSize(
            treeTable,
            rowIndex,
            columnIndex
          )
          preferredRowHeight = maxOf(cellDataSize.height, preferredRowHeight)
          if (columnData.type == BeSizingType.Fit) {
            val newMinWidth = maxOf(cellDataSize.width, columnData.minWidth)
            if (newMinWidth > columnData.minWidth)
              columnsData[columnIndex] = ColumnData(columnData.type, columnData.multiplier, newMinWidth)
          }
        }
      }

      // TreeTable can't handle reducing rowHeight, so we never reduce
      if (preferredRowHeight > treeTable.rowHeight)
        treeTable.rowHeight = preferredRowHeight

      val widthToFill = maxWidth - columnsData.filter { it.type == BeSizingType.Fit }.sumBy { it.minWidth }
      var remainingWidthToFill = widthToFill

      for (columnIndex in columnsData.indices) {
        val tableColumn = treeTable.columnModel.getColumn(columnIndex)
        val columnData = columnsData[columnIndex]
        if (columnData.type == BeSizingType.Fit) {
          tableColumn.minWidth = columnData.minWidth
          tableColumn.width = columnData.minWidth
          tableColumn.preferredWidth = columnData.minWidth
        }
        else if (columnData.type == BeSizingType.Fill) {
          require(fillMultiplierSum > 0 && columnData.multiplier != null)
          val fillWidth = if (lastFillColumnIndex == columnIndex)
            remainingWidthToFill
          else
            (((widthToFill).toDouble() / fillMultiplierSum) * columnData.multiplier).roundToInt()
          remainingWidthToFill -= maxOf(fillWidth, columnData.minWidth)
          tableColumn.minWidth = columnData.minWidth
          tableColumn.width = fillWidth
          tableColumn.preferredWidth = fillWidth
        }
      }
      treeTable.tableHeader?.resizingColumn = null
      treeTable.repaint()
    }

    /*
   *  Calculated the width based on the column name
   */
    private fun getColumnHeaderWidth(treeTable: TreeTable, column: Int): Int {
      val tableColumn = treeTable.columnModel.getColumn(column)
      val value = tableColumn.headerValue
      var renderer: TableCellRenderer? = tableColumn.headerRenderer

      if (renderer == null) {
        renderer = treeTable.tableHeader?.defaultRenderer
      }

      if (renderer == null)
        return 0

      val c = renderer.getTableCellRendererComponent(treeTable, value, false, false, -1, column)
      return c.preferredSize.width
    }

    private fun optionalSortIconWidth(columnInfo: TreeGridListColumn): Int {
      if (columnInfo.canBeSorted) {
        return UIManager.getIcon("Table.ascendingSortIcon").iconWidth
      }
      return 0
    }

    /*
   *  Get the preferred width for the specified cell
   */
    private fun getCellDataSize(treeTable: TreeTable, row: Int, column: Int): Dimension {
      //  Inovke the renderer for the cell to calculate the preferred width

      val cellRenderer = treeTable.getCellRenderer(row, column)
      val c = treeTable.prepareRenderer(cellRenderer, row, column)

      /**
       * This is hack for TreeTable, renderer for tree column does not calculate preferred height properly
       */
      var preferredHeight = c.preferredSize.height + treeTable.intercellSpacing.height
      if (cellRenderer is TreeTableCellRenderer)
        preferredHeight = 0

      return Dimension(c.preferredSize.width + treeTable.intercellSpacing.width, preferredHeight)
    }

  }

  override fun tableChanged(e: TableModelEvent) {
    logger.trace({ "tableChanged ${e.toLog()}" })
    adjustCellsLater()
  }

  fun adjustCells() {
    adjustCells(
      table,
      getAvailableWidth(),
      headerExtraWidth
    )
  }

  private fun adjustCellsLater() {
    SwingUtilities.invokeLater {
      adjustCells()
    }
  }
}/*
   *  Specify the table and use default spacing
   */
