// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl.renderers

import com.intellij.openapi.project.Project
import com.intellij.ui.SimpleColoredComponent
import com.intellij.util.ui.JBUI
import com.jetbrains.ide.model.uiautomation.BeButton
import com.jetbrains.ide.model.uiautomation.BeButtonStyle
import com.jetbrains.ide.model.uiautomation.BeLabel
import com.jetbrains.ide.model.uiautomation.BeRichText
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.ui.icons.RdIconHost
import com.jetbrains.rd.ui.bindable.views.listControl.RdCellSettings
import com.jetbrains.rd.ui.bindable.views.listControl.utils.BeGridUI
import com.jetbrains.rd.ui.bindable.views.utils.BeComponents
import com.jetbrains.rd.ui.bindable.views.utils.addMouseListener
import com.jetbrains.rd.ui.bindable.views.utils.getLabelShortcut
import com.jetbrains.rd.ui.bindable.views.utils.getLabelText
import javax.swing.JButton
import javax.swing.JComponent

class BeButtonRenderer(private val project: Project) : RdMultiRenderer<JComponent, BeButton>() {
  override fun isEditable() = true
  override fun containsEditor(model: BeButton) = true

  override fun getNewInstance(isEditor: Boolean, model: BeButton): JComponent {
    return when (model.style) {
      BeButtonStyle.DEFAULT,
      BeButtonStyle.ICON_TEXT,
      BeButtonStyle.TEXT,
      BeButtonStyle.HYBRID,
      BeButtonStyle.HYBRID_TEXT_ONLY,
      BeButtonStyle.ICON -> JButton()
      BeButtonStyle.NO_BORDER -> when (model.content) {
        is BeLabel -> BeComponents.getNewSCCInstance(true)
        is BeRichText -> BeComponents.getNewSCCInstance(true)
        else -> JButton()
      }
    }
  }

  override fun reset(
    component: JComponent?,
    value: BeButton,
    settings: RdCellSettings
  ): JComponent {
    val panel = getElement(settings.isEditor, value)
    panel.background = BeGridUI.getBackground(settings.isSelected, settings.hasFocus)
    val abstractText = value.content
    //val button = panel.components.first()
    val button = panel
    val iconHost = RdIconHost.getInstance(project)

    when (button) {
      is JButton -> {
        button.text = abstractText.getLabelText()
        val shortcut = abstractText.shortcutIndex.value
        if (shortcut >= 0) {
          button.setMnemonic(abstractText.getLabelShortcut())
        }
        if (settings.isEditor)
          button.addActionListener { value.action.click.fire() }
      }
      is SimpleColoredComponent -> {
        when (val label = value.content) {
          is BeLabel -> {
            button.apply {
              BeLabelRenderer.resetLabel(
                this,
                label.getLabelText(),
                label.icon.value,
                settings,
                iconHost,
                label.settings,
                label.tooltip.value
              )
            }
          }
          is BeRichText ->
            BeRichTextRenderer.resetRichText(
              button,
              label.text.valueOrNull,
              label.icon.value,
              settings,
              iconHost,
              label.settings,
              label.tooltip.value
            )
        }
        button.foreground = if (!value.enabled.value)
          JBUI.CurrentTheme.Label.disabledForeground(settings.isSelected)
        else BeGridUI.getForeground(
          settings.isSelected,
          settings.hasFocus)

        button.isOpaque = true

        if (settings.isEditor)
          value.addMouseListener(button)
      }
    }
    //panel.toolTipText = tooltip

    //todo: custom variations such as icon actions are currently not supported
    return panel
  }
}


