// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
@file:Suppress("JAVA_MODULE_DOES_NOT_EXPORT_PACKAGE")
package com.jetbrains.rd.ui.bindable.views.listControl.renderers

import com.intellij.CommonBundle
import com.intellij.openapi.project.Project
import com.intellij.ui.treeStructure.treetable.TreeTableTree
import com.intellij.util.ui.AbstractTableCellEditor
import com.jetbrains.ide.model.uiautomation.BeAlignment
import com.jetbrains.ide.model.uiautomation.BeControl
import com.jetbrains.ide.model.uiautomation.BeSpacer
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.ui.bindable.views.listControl.RdCellSettings
import com.jetbrains.rd.ui.bindable.views.listControl.treeGridUserObject
import com.jetbrains.rd.ui.bindable.views.listControl.utils.getDeepestRendererComponentByEvent
import com.jetbrains.rd.ui.bindable.views.listControl.utils.pathToRoot
import java.awt.Component
import java.awt.event.MouseEvent
import java.util.*
import javax.swing.*
import javax.swing.tree.DefaultMutableTreeNode
import javax.swing.tree.TreeCellEditor

/**
 * tree is null when editor created for TreeTable control and not null for Tree control
 */
class RdCellEditor(
  private val project: Project,
  private val tree: JTree?,
  private val columnAlignments: List<BeAlignment>,
  private val backgroundIsVisible: Boolean = false,
  private val applyHighlighting: Boolean = false
) : AbstractTableCellEditor(), TreeCellEditor {
  override fun getTableCellEditorComponent(table: JTable?, value: Any?, isSelected: Boolean, row: Int, column: Int): Component {
    return BeGridRenderer(project, backgroundIsVisible)
             .reset(value as? BeControl ?: BeSpacer(), RdCellSettings(true, true, true, true, applyHighlighting = applyHighlighting), table, columnAlignments[column])
           ?: JLabel(CommonBundle.getErrorTitle())
  }

  override fun getTreeCellEditorComponent(
    tree: JTree?,
    value: Any?,
    isSelected: Boolean,
    isExpanded: Boolean,
    isLeaf: Boolean,
    row: Int
  ): Component {
    hackTreeForEditors(tree)
    val node = value as DefaultMutableTreeNode
    val userObject = node.treeGridUserObject()
    return BeGridRenderer(project, tree = tree, isBackgroundVisible = backgroundIsVisible)
             .reset(userObject.line.columnElements[0], RdCellSettings(true, true, true, true, applyHighlighting = applyHighlighting), tree, columnAlignments[0])
           ?: JLabel(CommonBundle.getErrorTitle())
  }

  val logger = getLogger<RdCellEditor>()

  /**
   * The problem is - JTree treats any click on node as start editing
   * We want to start editing only if user clicked on "editable" control inside row
   */
  override fun isCellEditable(e: EventObject?): Boolean {
    tree ?: return true
    // e is notMouse event only if user want to start editing programatically via startEditingAtPath
    val e = e as? MouseEvent ?: return true
    val path = tree.getClosestPathForLocation(e.x, e.y)
    val nodeBounds = tree.getPathBounds(path) ?: return false
    val row = tree.getRowForPath(path)
    val editingComponent = getTreeCellEditorComponent(tree, path.lastPathComponent, true, false, false, row)
    val activeComponent = getDeepestRendererComponentByEvent(tree, editingComponent, nodeBounds, e) ?: return false
    return checkIsEditable(activeComponent, tree)
  }

  /**
   * Check if activeComponent or any parent above before root has client property isEditable == true
   */
  private fun checkIsEditable(activeComponent: JComponent, root: JComponent): Boolean {
    return activeComponent.pathToRoot(root)
      .mapNotNull { it as? JComponent }
      .fold(false) { acc, jComponent ->
        acc || (jComponent.getClientProperty(BeControls.isEditableKey) as? Boolean == true)
      }
  }

  /**
   * This is hack for allowing editors in first column. The problem is TreeTableTree is not present
   * in visual tree therefore cannot hit-test components inside its rows. When editor inserted into such
   * tree validate() call can't measure inner components because peer == null, so we are adding tree into table
   *
   * This needs to be done before getting editor, as cellRenderer removes all renderers and our tree after full pass
   * (can be investigated further)
   */
  private fun hackTreeForEditors(tree: JTree?) {
    (tree as? TreeTableTree)?.treeTable?.add(tree)
  }

  override fun getCellEditorValue(): Any {
    return "Value"
  }

  override fun stopCellEditing(): Boolean {
    return super.stopCellEditing()
  }

  override fun shouldSelectCell(anEvent: EventObject?): Boolean {
    return anEvent?.source !is TreeTableTree
  }

}

