// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl.renderers

import com.intellij.openapi.project.Project
import com.intellij.util.ui.JBUI
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.extensions.valueOrEmpty

import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.valueOrDefault
import com.jetbrains.rd.ui.bindable.ViewRegistry
import com.jetbrains.rd.ui.bindable.views.listControl.BeTree
import com.jetbrains.rd.ui.bindable.views.listControl.ITreeGridRenderer
import com.jetbrains.rd.ui.bindable.views.listControl.RdCellSettings
import com.jetbrains.rd.ui.bindable.views.listControl.utils.BeGridUI
import com.jetbrains.rd.ui.bindable.views.listControl.utils.RdRendererCache
import com.jetbrains.rd.ui.bindable.views.utils.marginTypeToPx
import com.jetbrains.rdclient.ui.bindableUi.views.utils.toDock
import net.miginfocom.swing.MigLayout
import java.awt.Component
import java.awt.Dimension
import javax.swing.JComponent
import javax.swing.JPanel
import javax.swing.JTree

class BeGridRenderer(
  private val project: Project,
  private val isBackgroundVisible: Boolean = false,
  private val tree: JTree? = null
) : JPanel(MigLayout("insets 0, gap 0, novisualpadding, left, fill")),
    ITreeGridRenderer<BeGrid> {
  private val myCache = RdRendererCache(project)

  init {
    border = JBUI.Borders.empty(0, 0, 0, 0)
    isOpaque = !isBackgroundVisible
  }

  override fun resetCounter() {
    myCache.resetCounter()
  }

  override fun addUpdate(element: BeGrid, lifetime: Lifetime, update: () -> Unit) {
    element.items.advise(lifetime) {
      it.forEach { item ->
        item.content.visible.change.advise(lifetime) { update() }
        item.content.enabled.change.advise(lifetime) { update() }
        ViewRegistry.getInstance(project).getRenderer(item.content)?.addUpdate(item.content, lifetime, update)
      }
      update()
    }

    // value.items.change.advise(lifetime) { this.repaint() }
  }


  override fun reset(
    component: JComponent?,
    value: BeGrid,
    settings: RdCellSettings
  ): JComponent {
    this.removeAll()
    this.layout = MigLayout("insets 0, gap 0, novisualpadding, left")
      .apply {
        val marginType = if (value.intercellSpacing == BeMarginType.DefaultIntercellSpacing) BeMarginType.None else value.intercellSpacing
        val interCellMargin = marginTypeToPx(marginType)
        val rows = value.items.valueOrEmpty().joinToString(" $interCellMargin ")
        { gridElement ->
          if (gridElement.size.unit == BeSizingType.Fit)
            "[pref!]"
          else
            "[fill, grow ${gridElement.size.multiplier}]" //, ${commonsize}
        }
        this.columnConstraints = rows
      }
    value.items.valueOrNull?.forEachIndexed { _, it ->
      myCache.getComponent(it.content, component, settings).let { jComponent ->
        if (value.visible.valueOrDefault(ControlVisibility.Visible) != ControlVisibility.Visible)
          jComponent.isVisible = false
        addComponent(
          jComponent,
          value.orientation == GridOrientation.Horizontal,
          it.size
        )
      }
    }

    if (!isBackgroundVisible)
      background = BeGridUI.getBackground(settings.isSelected, settings.hasFocus)

    return this
  }

  fun reset(
    control: BeControl,
    settings: RdCellSettings,
    table: JComponent?,
    alignment: BeAlignment
  ): JComponent {
    this.removeAll()
    addComponent(myCache.getComponent(control, table, settings), alignment = alignment)
    if (!isBackgroundVisible)
      background = BeGridUI.getBackground(settings.isSelected, settings.hasFocus)
    return this
  }

  fun isCellEditable(model: BeControl?): Boolean {
    return myCache.isEditable(model)
  }

  override fun containsEditor(model: BeGrid): Boolean {
    return model.items.valueOrNull?.any {
      it.content.visible.valueOrDefault(ControlVisibility.Visible) == ControlVisibility.Visible &&
      it.content.enabled.value &&
      myCache.isEditable(it.content)
    } ?: false
  }

  override fun isEditable() = false

  private fun addComponent(
    jComponent: JComponent,
    isHorizontal: Boolean = true,
    size: BeUnitSize? = null,
    alignment: BeAlignment = BeAlignment.Left
  ) {
    jComponent.alignmentY = Component.CENTER_ALIGNMENT
    var constraints =
      "gap 0" +
      when {
        alignment != BeAlignment.Stretch -> when {
          size == null -> ", ${alignment.toDock()}" //", grow, push"
          size.unit == BeSizingType.Fill -> ", grow, push"
          else -> ", grow 0"
        }
        else -> ", ${alignment.toDock()}"
      }

    if (!isHorizontal)
      constraints += ", wrap"

    add(jComponent, constraints)
  }

  override fun getToolTipText(): String? =
    (components.firstOrNull() as? JComponent)?.toolTipText ?: super.getToolTipText()

  override fun getPreferredSize(): Dimension {
    val prefSize = super.getPreferredSize()
    if (tree != null && tree is BeTree && tree.showAsList) {
      return Dimension(tree.width, prefSize.height)
    }
    return prefSize
  }
}
