// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.listControl.renderers

import com.intellij.ide.ui.laf.darcula.DarculaUIUtil
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.ui.JBColor
import com.intellij.ui.SideBorder
import com.intellij.ui.SimpleColoredComponent
import com.intellij.ui.treeStructure.treetable.TreeTableTree
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.BeTextBox
import com.jetbrains.ide.model.uiautomation.ValidationResult
import com.jetbrains.ide.model.uiautomation.ValidationStates
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.ui.icons.RdIconHost
import com.jetbrains.rd.ui.bindable.views.customControls.TextBoxControl
import com.jetbrains.rd.ui.bindable.views.listControl.RdCellSettings
import com.jetbrains.rd.ui.bindable.views.listControl.utils.BeGridUI
import com.jetbrains.rd.ui.bindable.views.utils.BeComponents
import com.jetbrains.rd.ui.bindable.views.utils.BeMagicMargin
import java.awt.*
import javax.swing.JComponent
import javax.swing.JPanel

class BeTextBoxRenderer(
  private val project: Project
) : RdMultiRenderer<JPanel, BeTextBox>() {
  override fun isEditable() = true
  override fun containsEditor(model: BeTextBox) = true
  var definition = project.defineNestedLifetime()


  override fun getNewInstance(isEditor: Boolean, model: BeTextBox): JPanel {
    val jPanel = JPanel(GridBagLayout()).apply {
      isOpaque = true
      border = JBUI.Borders.empty(0, 0, 0, 0)
    }
    val c = GridBagConstraints().apply {
      fill = GridBagConstraints.BOTH
      gridx = 0
      weightx = 1.0
      gridy = 0
      weighty = 1.0
    }

    if (isEditor) {
      val textBoxControl = TextBoxControl(project)
      textBoxControl.ensureWillComputePreferredSize()
      textBoxControl.putClientProperty("JComboBox.isTableCellEditor", java.lang.Boolean.TRUE)
      textBoxControl.isEnabled = true
      textBoxControl.addSettingsProvider { editor ->
        editor.setBorder(TextEditorBorder())
      }
      return jPanel.apply { add(textBoxControl.focusTarget, c) }
    }
    return jPanel.apply {
      minimumSize = Dimension(BeMagicMargin.MinimumTextBoxWidth, BeMagicMargin.MinimumTextBoxHeight)
      //this.border = TextEditorBorder()

      add(BeComponents.getNewSCCInstance(true), c)
    }

  }

  private fun getLabelForeground(
    settings: RdCellSettings,
    validationState: ValidationStates
  ) = when (validationState) {
    ValidationStates.validationError -> JBUI.CurrentTheme.Focus.errorColor(true)
    ValidationStates.validationWarning -> JBUI.CurrentTheme.Focus.warningColor(true)
    else -> if (!settings.isEnabled) JBUI.CurrentTheme.Label.disabledForeground(settings.isSelected)
    else BeGridUI.getForeground(
      settings.isSelected,
      settings.hasFocus
    )
  }

  private fun getToolTipText(validationResult: ValidationResult?, defaultToolTip: String?): String? =
    validationResult?.resultMessage ?: defaultToolTip

  override fun reset(
    component: JComponent?,
    value: BeTextBox,
    settings: RdCellSettings
  ): JComponent {
    val panel = getElement(settings.isEditor, value)
    panel.background = BeGridUI.getBackground(settings.isSelected, settings.hasFocus) //UIUtil.getTextFieldBackground()
    if (definition.isAlive)
      definition.terminate()
    definition = project.defineNestedLifetime()

    if (settings.isEditor) {
      val textBoxControl = panel.components.first() as TextBoxControl
      textBoxControl.apply {
        text = value.text.value
        focusTarget.border = JBUI.Borders.empty()
        //if (component !is TreeTableTree)
          init(value, definition.lifetime)
      }
    }
    else {
      val label = panel.components.first() as SimpleColoredComponent
      label.apply {
        BeLabelRenderer.resetLabel(
          this,
          value.text.value,
          null,
          settings,
          RdIconHost.getInstance(project),
          value.settings.textSettings,
          value.tooltip.value
        )
        /*
        May be it doesn't work (settings)
           value.enabled.value,

        * */
        val validationState = value.validationResult.valueOrNull?.state ?: ValidationStates.validationPassed
        foreground = getLabelForeground(settings, validationState)
        isOpaque = true
      }

      /* val iconLabel = panel.components.last() as JLabel
       iconLabel.apply {
           iconLabel.icon =value.validationResult.valueOrNull?.icon?.value?.toIdeaIcon(iconHost)
       }*/
    }
    panel.revalidate()
    return panel
  }
}

/**
 * The default border implementation based on [EditorImpl.TablessBorder] that is used by [TextBoxControl].
 */
class TextEditorBorder : SideBorder(JBColor.border(), SideBorder.ALL, 1) {
  override fun paintBorder(c: Component, g: Graphics, x: Int, y: Int, width: Int, height: Int) {
    if (c is JComponent) {
      val textBoxControl = UIUtil.getParentOfType(TextBoxControl::class.java, c)
      val outline = textBoxControl?.getClientProperty("JComponent.outline")?.let { DarculaUIUtil.Outline.valueOf(it.toString()) }
      val hasFocus = textBoxControl?.focusTarget?.hasFocus() ?: false
      outline?.setGraphicsColor(g as Graphics2D, hasFocus)

      if (outline != null) {
        g.fillRect(x, y, thickness, height)
        g.fillRect(x, y, width, thickness)
        g.fillRect(x + width - thickness, y, thickness, height)
        g.fillRect(x, y + height - thickness, width, thickness)
      }
      else super.paintBorder(c, g, x, y, width, height)
    }
  }
}