// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.utils

import com.intellij.ide.ui.UISettings
import com.intellij.openapi.actionSystem.ActionToolbar
import com.intellij.openapi.editor.colors.EditorColorsManager
import com.intellij.openapi.editor.ex.util.EditorUtil
import com.intellij.openapi.project.Project
import com.intellij.ui.BalloonImpl
import com.intellij.ui.SimpleColoredComponent
import com.intellij.ui.SimpleTextAttributes
import com.intellij.ui.components.JBLabel
import com.intellij.uiDesigner.core.Util
import com.intellij.util.ui.JBFont
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.dsl.tryGetSettings
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rd.util.reactive.valueOrDefault
import com.jetbrains.rd.ui.icons.RdIconHost
import com.jetbrains.rd.ui.icons.toIdeaIcon
import com.jetbrains.rd.ui.fillWidthButDoNotOverflow
import com.jetbrains.rd.ui.toSimpleText
import com.jetbrains.rdclient.util.idea.ui.toSimpleTextAttributes
import com.jetbrains.rd.ide.model.RichTextModel
import javax.swing.*
import com.intellij.uiDesigner.core.Util.getPreferredSize
import java.awt.*


internal fun getFontSize(size: BeFontSize, editable: Boolean): Float {
  val scheme = EditorColorsManager.getInstance().globalScheme
  val fontSize = if (editable) {
    (if (UISettings.instance.presentationMode)
      UISettings.instance.presentationModeFontSize - 4
    else
      scheme.editorFontSize)
  }
  else {
    val settings = UISettings.instance
    if (settings.overrideLafFonts) settings.fontSize else JBFont.label().size
  }
  val result: Int = when (size) {
    BeFontSize.DEFAULT -> fontSize
    BeFontSize.SMALLER -> UIUtil.getFontSize(UIUtil.FontSize.SMALL).toInt()
    BeFontSize.BIGGER -> fontSize + 2
    else -> fontSize
  }
  return result.toFloat()
}

internal fun BeTextSettings.getFont(): Font? {
  val style = this.fontStyle.valueOrDefault(BeFontStyle.DEFAULT)
  val newSize = this.getFontSize(style != BeFontStyle.DEFAULT)
  val styledFont = when (style) {
    BeFontStyle.DEFAULT -> {
      JBFont.label()
    }
    BeFontStyle.MONOSPACE -> {
      EditorUtil.getEditorFont()
    }
  }
  return styledFont.deriveFont(newSize)
}


private fun BeTextSettings.getFontSize(editable: Boolean): Float {
  return getFontSize(this.fontSize.valueOrDefault(BeFontSize.DEFAULT), editable)
}

fun SimpleColoredComponent.applyModel(modelRichText: RichTextModel?, bg: Color, overrideFg: Color? = null): SimpleColoredComponent {
  return this.apply {
    clear()
    if (modelRichText == null)
      return@apply
    for (part in modelRichText.parts) {
      val attributes = part.toSimpleTextAttributes(bg, overrideFg, overrideFg == null)
      append(part.text, attributes)
    }

  }

}

fun SimpleColoredComponent.applyModel(beRichText: BeRichText, bg: Color, overrideFg: Color? = null): SimpleColoredComponent {
  return this.apply {
    val modelRichText = beRichText.text.valueOrNull
    clear()
    if (modelRichText == null)
      return@apply
    for (part in modelRichText.parts) {
      var attributes = part.toSimpleTextAttributes(bg, overrideFg, overrideFg == null)
      if (beRichText.settings.fontSize.valueOrDefault(BeFontSize.DEFAULT) == BeFontSize.SMALLER)
        attributes = attributes.derive(SimpleTextAttributes.STYLE_SMALLER, null, null, null)
      append(part.text, attributes)
    }
  }
}

internal fun BeRichText.getHtmlLabelFromBeRichText(): String {
  val valueOrNull = this.text.valueOrNull ?: return "invalid"
  if (shortcutIndex.value >= 0) return valueOrNull.toSimpleText()
  object : SimpleColoredComponent() {
    fun toLabel(): JLabel = formatToLabel(JLabel())
  }.apply {
    applyModel(this@getHtmlLabelFromBeRichText, UIUtil.getLabelBackground())
  }.toLabel().text.let {
    return if (this.wrapText)
      it.replace("nowrap", "normal") //.replace("<span>", "").replace("</span>", "")
    else
      it
  }
}

fun BeAbstractText.getLabelText(): String {
  when (this) {
      is BeLabel -> {
        if (shouldBeWrappedIntoHtml)
          text.value.wrapToHtmlWrap()
        else
          return text.value
      }
    is BeRichText -> getHtmlLabelFromBeRichText()
    else -> ""
  }.let {
    val result = it.replace("\n", "<br/>")
    return if (!this@getLabelText.enabled.value)
      convertToDisabled(result)
    else result
  }
}

fun BeAbstractText.getSimpleText(): String =
  when (this) {
    is BeLabel -> text.value
    is BeRichText -> text.valueOrNull?.toSimpleText() ?: ""
    is BeRichAttributedText -> text.valueOrNull?.parts?.joinToString { it.text } ?: ""
    else -> ""
  }

internal fun convertToDisabled(result: String): String {
  var result1 = result
  val disabledColor = UIUtil.getLabelDisabledForeground()
  result1 = result1.replace(
    "<html>",
    "<html><font color='rgb(${disabledColor.red}, ${disabledColor.green}, ${disabledColor.blue})'>"
  )
  return result1
}

fun String.wrapToHtmlWrap(wrap: Boolean = true): String {

  val wrapping = if (wrap) "normal" else "nowrap"
  return "<html><body style=\"white-space:$wrapping\"> $this <html><body style=\"white-space:$wrapping\">"
}

internal fun BeAbstractText.getLabelShortcut(): Char {
  return this.getLabelText()[this.shortcutIndex.value]
}

internal fun BeAbstractText.onTextChanged(lifetime: Lifetime, onChange: (Lifetime, String) -> Unit) {
  when (this) {
    is BeLabel -> {
      this.text.view(lifetime) { lt, _ ->
        onChange(lt, this.getLabelText())
      }
    }
    is BeRichText -> this.text.view(lifetime) { lt, _ ->
      onChange(lt, this.getLabelText())
    }
  }
}

internal fun BeAbstractText.getIdeaIcon(project: Project): Icon? {
  return icon.value?.toIdeaIcon(project)
}

internal val BeAbstractText.shouldBeWrappedIntoHtml: Boolean
  get() {
    return this.wrapText || return when (this) {
      is BeLabel -> this.text.value.contains("\n")
      is BeRichText -> this.text.valueOrNull?.toSimpleText()?.contains("\n") ?: false || this.shortcutIndex.value >= 0
      else -> false
    }

  }

internal fun AbstractButton.initFromText(lifetime: Lifetime, beAbstractText: BeAbstractText, iconHost: RdIconHost, widthFromText: Boolean): JComponent {
  val label = beAbstractText.getLabelText()
  this.text = label
  beAbstractText.onTextChanged(lifetime) { _, s ->
    this.text = s
    this.revalidate()
  }

  val shortcut = beAbstractText.shortcutIndex.value
  if (shortcut >= 0) {
    this.setMnemonic(beAbstractText.getLabelShortcut())
    this.displayedMnemonicIndex = shortcut
  }

  this.isRequestFocusEnabled = true

  beAbstractText.icon.advise(lifetime) {
    this.icon = it?.toIdeaIcon(iconHost)
    this.disabledIcon = it?.toIdeaIcon(iconHost)
  }

  beAbstractText.enabled.advise(lifetime) {
    if (beAbstractText.shouldBeWrappedIntoHtml)
      this.text = beAbstractText.getLabelText()
    else
      this.isEnabled = it
  }

  beAbstractText.tooltip.advise(lifetime) { toolTipText = it }

  val settings = beAbstractText.tryGetSettings()
  if (settings != null)
    font = settings.getFont()

  if(widthFromText) {
    val metrics: FontMetrics = getFontMetrics(font)
    val width: Int = metrics.stringWidth(text)
    val newDimension = Dimension(width + 20, ActionToolbar.DEFAULT_MINIMUM_BUTTON_SIZE.height)
    preferredSize = newDimension
    minimumSize = newDimension
  }

  return beAbstractText.allowWrapped(this)
}


internal fun BeAbstractText.allowWrapped(
  component: JComponent
): JComponent {
  return if (this.wrapText) {
    component.fillWidthButDoNotOverflow(false)
  }
  else {
    JBUI.Panels.simplePanel().apply { add(component, BorderLayout.WEST) }
  }
}


internal fun BeAbstractText.getLabelForAbstractText(
  lifetime: Lifetime, iconHost: RdIconHost
): JLabel {
  val component = JBLabel().apply {
    border = JBUI.Borders.empty()
    horizontalAlignment = SwingConstants.LEFT
    verticalAlignment = SwingConstants.CENTER
    isOpaque = false
  }

  this.onTextChanged(lifetime) { _, s ->
    component.text = s
  }
  val shortcut = shortcutIndex.value
  if (shortcut >= 0) {
    val shortcutChar = getLabelShortcut()
    component.setDisplayedMnemonic(shortcutChar)
    component.displayedMnemonicIndex = shortcut
  }
  this.icon.advise(lifetime) {
    component.icon = it?.toIdeaIcon(iconHost)
    component.disabledIcon = it?.toIdeaIcon(iconHost)
  }

  this.enabled.advise(lifetime) {
    if (this.shouldBeWrappedIntoHtml)
      component.text = getLabelText()
    else
      component.isEnabled = it
    component.repaint()
  }

  this.visible.adviseNotNull(lifetime) {
    component.isVisible = it == ControlVisibility.Visible
  }

  this.tooltip.advise(lifetime) { component.toolTipText = it }

  val settings = this.tryGetSettings()
  if (settings != null)
    component.font = settings.getFont()
  return component
}
