// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.utils

import com.intellij.icons.AllIcons
import com.intellij.ide.DataManager
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.actionSystem.impl.ActionButton
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.text.TextWithMnemonic
import com.intellij.ui.AnActionButton
import com.intellij.util.IconUtil
import com.jetbrains.ide.model.uiautomation.BeAbstractButton
import com.jetbrains.ide.model.uiautomation.BeButton
import com.jetbrains.ide.model.uiautomation.BeCommonBehavior
import com.jetbrains.ide.model.uiautomation.BePredefinedBehaviorButton
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.fire
import java.awt.event.MouseAdapter
import java.awt.event.MouseEvent
import javax.swing.Action
import javax.swing.Icon
import javax.swing.JComponent
import javax.swing.JLabel

internal fun BePredefinedBehaviorButton.tryAddCustomShortCut(component: JComponent, container: JComponent, lifetime: Lifetime) {
  val disposable = lifetime.createNestedDisposable()
  val shortcutSet = predefinedBehavior.getCustomShortCut() ?: return
  when (component) {
    is ActionButton -> component.action.registerCustomShortcutSet(shortcutSet, container, disposable)
    is AnActionButton -> component.registerCustomShortcutSet(shortcutSet, container, disposable)
  }
}

internal fun BePredefinedBehaviorButton.tryAddCustomShortCut(action: AnAction?, container: JComponent, lifetime: Lifetime) {
  val shortcutSet = predefinedBehavior.getCustomShortCut() ?: return
  action?.registerCustomShortcutSet(shortcutSet, container, lifetime.createNestedDisposable())
}

internal fun BeCommonBehavior.getCustomIcon(): Icon? {
  return when (this) {
    BeCommonBehavior.TABLE_ADD -> IconUtil.getAddIcon()
    BeCommonBehavior.TABLE_DELETE -> IconUtil.getRemoveIcon()
    BeCommonBehavior.TABLE_MOVE_UP -> IconUtil.getMoveUpIcon()
    BeCommonBehavior.TABLE_MOVE_DOWN -> IconUtil.getMoveDownIcon()
    BeCommonBehavior.TABLE_EDIT -> IconUtil.getEditIcon()
    BeCommonBehavior.DIALOG_ACCEPT_BUTTON -> null
    BeCommonBehavior.DIALOG_CANCEL_BUTTON -> null
    BeCommonBehavior.DIALOG_HELP -> null
    BeCommonBehavior.OPEN_FILE_DIALOG -> AllIcons.General.OpenDisk
    BeCommonBehavior.RESET_VALUE -> AllIcons.General.Reset
    BeCommonBehavior.GEAR -> AllIcons.General.GearPlain
    BeCommonBehavior.NONE -> null
    BeCommonBehavior.SEARCH -> AllIcons.Actions.Search
    BeCommonBehavior.NONE -> null
    BeCommonBehavior.INFO -> AllIcons.General.Information
    else -> null
  }
}

internal fun BeCommonBehavior.getCustomShortCut(): ShortcutSet? {
  return when (this) {
    BeCommonBehavior.TABLE_ADD -> CommonShortcuts.getNew()
    BeCommonBehavior.TABLE_DELETE -> CommonShortcuts.getDelete()
    BeCommonBehavior.TABLE_MOVE_UP -> CommonShortcuts.getMoveUp()
    BeCommonBehavior.TABLE_MOVE_DOWN -> CommonShortcuts.getMoveDown()
    BeCommonBehavior.TABLE_EDIT -> CommonShortcuts.getEditSource()
    BeCommonBehavior.DIALOG_ACCEPT_BUTTON -> CommonShortcuts.ENTER
    BeCommonBehavior.DIALOG_CANCEL_BUTTON -> CommonShortcuts.ESCAPE
    BeCommonBehavior.DOWN -> PredefinedBehaviorCommonShortcuts.DOWN
    BeCommonBehavior.UP -> PredefinedBehaviorCommonShortcuts.UP
    BeCommonBehavior.SEARCH -> CommonShortcuts.getFind()
    else -> null
  }
}

internal fun BeAbstractButton.prepareActionWithUpdate(
  lifetime: Lifetime,
  update: () -> Unit
) {
  val beAbstractText = this.content
  beAbstractText.icon.advise(lifetime) {
    update()
  }
  this.tooltip.advise(lifetime) {
    update()
  }
  beAbstractText.onTextChanged(lifetime) { _, s ->
    update()
    //beAbstractText.revalidate()
  }
  beAbstractText.tooltip.advise(lifetime) {
    update()
  }
  this.enabled.advise(lifetime) {
    update()
  }
}

fun BeAbstractButton.prepareActionWithUpdate(
  lifetime: Lifetime,
  action: Action
) {
  val beAbstractText = this.content
  beAbstractText.icon.advise(lifetime) {
    update(action)
  }

  beAbstractText.onTextChanged(lifetime) { _, s ->
    update(action)
    //beAbstractText.revalidate()
  }
  this.enabled.advise(lifetime) {
    update(action)
  }
}

internal fun BeAbstractButton.tryGetPredefinedIcon(project: Project): Icon? {
  var icon = this.content.getIdeaIcon(project)
  if (this is BePredefinedBehaviorButton) {
    val customIcon = this.predefinedBehavior.getCustomIcon()
    if (customIcon != null) icon = customIcon
  }
  return icon
}

internal fun BeAbstractButton.updateActionEvent(e: AnActionEvent, project: Project) {
  e.presentation.icon = this.tryGetPredefinedIcon(project)
  e.presentation.description = this.getButtonActionText()
  e.presentation.setTextWithMnemonic {
    TextWithMnemonic.fromPlainText(e.presentation.description).apply {
      val index = content.shortcutIndex.value
      if (index >= 0)
        setMnemonicAt(index)
    }
  }
  e.presentation.isEnabled = this.enabled.value
  //e.putValue(Action.NAME, value);
}

internal fun update(action: AnAction) {
  val actionManager = ActionManager.getInstance()
  val dataManager = DataManager.getInstance()
  val presentation = action.templatePresentation //.clone()
  //todo: may be it should be written more carefully
  val event = AnActionEvent(null, dataManager.getDataContext(JLabel()), "Button", presentation, actionManager, 0)
  action.update(event)
}

internal fun BeAbstractButton.update(action: Action) {
  action.isEnabled = this.enabled.value
  action.putValue(Action.NAME, this.content.getLabelText())
  //    myOKAction.putValue(Action.MNEMONIC_KEY, c);
}

internal fun BeAbstractButton.getButtonActionText(): String {
  val simpleText = this.content.getSimpleText()
  return when {
    simpleText.isEmpty() -> {
      val valueOrDefault = this.tooltip.value ?: ""
      when {
        valueOrDefault.isEmpty() -> this.content.tooltip.value ?: ""
        else -> valueOrDefault
      }
    }
    else -> simpleText
  }
}

internal fun BeButton.addMouseListener(button: JComponent) {
  button.addMouseListener(object : MouseAdapter() {
    override fun mouseEntered(e: MouseEvent?) {
      this@addMouseListener.isMouseOver.set(true)
    }

    override fun mouseExited(e: MouseEvent?) {
      this@addMouseListener.isMouseOver.set(false)
    }

    override fun mousePressed(e: MouseEvent?) {
      this@addMouseListener.action.click.fire()
    }
  })

}