// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.bindable.views.utils

import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.*
import java.awt.Dimension
import javax.swing.JCheckBox
import javax.swing.JComponent


internal val checkboxMarginSize: Int by lazy { UIUtil.getCheckBoxTextHorizontalOffset(JCheckBox("a")) }
private var intercellSpacing = 5
internal val DEFAULT_VERTICAL = MarginType(0, 0, 0, intercellSpacing)
internal val DEFAULT_HORIZONTAL = MarginType(0, 0, intercellSpacing, 0)
internal val DEFAULT_SPAN = MarginType(0, 0, intercellSpacing, intercellSpacing)
internal val DEFAULT_NONE = MarginType(0, 0, 0, 0)

data class MarginType(val left: Int, val right: Int, val top: Int, val bottom: Int)

internal fun marginTypeToPx(type: BeMarginType): Int {
  return when (type) {
    BeMarginType.None -> 0
    BeMarginType.OnePx -> 1
    BeMarginType.Checkbox -> checkboxMarginSize
    BeMarginType.RadioButton -> checkboxMarginSize
    BeMarginType.ExtractClassTempLineMargin -> 0
    BeMarginType.DefaultIntercellSpacing -> 5
    BeMarginType.IncreasedIntercellSpacing -> 20
    BeMarginType.ReducedIntercellSpacing -> 2
  }
}

internal fun marginItemToString(item: BeMarginItem): String {
  return "${marginTypeToPx(item.unit) * item.multiplier}px"
}

internal fun marginTypeToString(type: BeMarginType): String {
  return "${marginTypeToPx(type)}px"
}

internal fun marginToMig(margin: BeMargin): String? {
  val left = margin.left
  val right = margin.right
  val top = margin.top
  val bottom = margin.bottom
  return "gap " +
         "${marginItemToString(left)} " +
         "${marginItemToString(right)} " +
         "${marginItemToString(top)} " +
         "${marginItemToString(bottom)} "
}

fun BeControlSize.applyToContent(content: JComponent) {
  content.preferredSize = getSize(content.preferredSize)
}


fun BeControlSizeType.toSwing(): Int {
  return JBUI.scale(when (this) {
                      BeControlSizeType.TINY -> 100
                      BeControlSizeType.SMALL -> 200
                      BeControlSizeType.MEDIUM -> 300
                      BeControlSizeType.LARGE -> 400
                      BeControlSizeType.HUGE -> 500
                      BeControlSizeType.EXTRA_HUGE -> 800
                      BeControlSizeType.FIT_TO_CONTENT -> -1
                      BeControlSizeType.REFACTORING_WIDTH -> 600
                    })
}

fun BeControlSize.getSize(): Pair<Int, Int> {
  var resultWidth = -1
  var resultHeight = -1
  return when (this) {
    is BeControlSizePredefined -> {
      val localWidth = width.type.toSwing()
      if (localWidth >= 0)
        resultWidth = localWidth * width.multiplier
      val localHeight = height.type.toSwing()
      if (localHeight >= 0)
        resultHeight = localHeight * height.multiplier
      Pair(resultWidth, resultHeight)
    }
    is BeControlSizeCustom -> Pair(width, height)
    is BeControlSizeProportional -> {
      val baseSize = this.size.toSwing()
      if (proportion < 1) {
        Pair(baseSize, (baseSize / proportion).toInt())
      }
      else {
        Pair((baseSize * proportion).toInt(), baseSize)
      }
    }
    else -> Pair(resultWidth, resultHeight)
  }
}
fun BeControlSize.getSize(dimension: Dimension): Dimension {
  return when (this) {
    is BeControlSizePredefined -> {
      val localSize = dimension
      val localWidth = width.type.toSwing()
      if (localWidth >= 0)
        localSize.width = localWidth * width.multiplier
      val localHeight = height.type.toSwing()
      if (localHeight >= 0)
        localSize.height = localHeight * height.multiplier
      localSize
    }
    is BeControlSizeCustom -> JBUI.size(width, height)
    is BeControlSizeProportional -> {
      val baseSize = this.size.toSwing()
      val dimension = if (proportion < 1) {
        Dimension(baseSize, (baseSize / proportion).toInt())
      }
      else {
        Dimension((baseSize * proportion).toInt(), baseSize)
      }
      dimension
    }
    else -> dimension
  }
}

