// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.icons

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProcessCanceledException
import com.intellij.openapi.util.SystemInfo
import com.intellij.reference.SoftReference
import com.intellij.ui.icons.LazyImageIcon
import com.intellij.ui.scale.DerivedScaleType
import com.intellij.ui.scale.ScaleContext
import com.intellij.ui.scale.ScaleType
import com.intellij.util.BitUtil
import com.intellij.util.ImageLoader
import com.intellij.util.containers.FixedHashMap
import com.intellij.util.ui.ImageUtil
import com.intellij.util.ui.JBImageIcon
import com.intellij.util.ui.MultiResolutionImageProvider
import com.jetbrains.rd.ide.model.IconSource
import java.awt.image.BufferedImage
import java.util.*
import java.util.concurrent.CancellationException
import javax.swing.Icon
import javax.swing.ImageIcon

// todo somehow unify with CachedImageIcon
class ProtocolCachedImageIcon(
  private val originalPath: String,
  private val isDark: Boolean,
  private val sources: List<IconSource>
) : LazyImageIcon() {
  private val scaledIconCache = MyScaledIconsCache()

  override fun getRealIcon(context: ScaleContext?): ImageIcon {
    val realIcon = myRealIcon
    synchronized(myLock) {
      if (!updateScaleContext(context) && realIcon != null) {
        val icon = unwrapIcon(realIcon)
        if (icon != null) return icon
      }
      val icon = scaledIconCache.getOrScaleIcon(1f)
      val smallEnough = icon.iconWidth < 50 && icon.iconHeight < 50
      myRealIcon = if (smallEnough) icon else SoftReference(icon)
      return icon
    }
  }

  override fun scale(scale: Float): Icon {
    if (scale == 1f) return this
    return scaledIconCache.getOrScaleIcon(scale)
  }

  override fun getDarkIcon(isDark: Boolean): Icon {
    return ProtocolCachedImageIcon(originalPath, isDark, sources)
  }

  fun createImage(ctx: ScaleContext, dark: Boolean) = runCatching l@{
    var flags = ImageLoader.USE_SVG or ImageLoader.ALLOW_FLOAT_SCALING or ImageLoader.USE_CACHE
    flags = BitUtil.set(flags, ImageLoader.USE_DARK, dark)

    val descriptors = ImageLoader.getImageDescriptors(originalPath, flags, ctx)

    for (descriptor in descriptors) {
      val source = sources.find { it.path == descriptor.path } ?: continue
      val body = source.body ?: continue
      val origUsrSize = ImageLoader.Dimension2DDouble(0.0, 0.0)

      flags = BitUtil.set(flags, ImageLoader.USE_SVG, descriptor.isSvg)
      flags = BitUtil.set(flags, ImageLoader.USE_DARK, descriptor.isDark)
      val image = ImageLoader.loadFromStream(body.inputStream(), null, descriptor.scale, origUsrSize, flags)
      val usrScale = ctx.getScale(DerivedScaleType.EFF_USR_SCALE)
      return@l ImageUtil.ensureHiDPI(image, ctx, origUsrSize.width * usrScale, origUsrSize.height * usrScale)
    }
    return@l null
  }.onFailure {
    e ->
    when (e) {
      is ProcessCanceledException,
      is CancellationException -> throw e
      else -> LOG.error("Path $originalPath. Exception $e")
    }
  }.getOrNull()

  override fun getMenuBarIcon(isDark: Boolean): Icon {
    val useMRI = SystemInfo.isMac
    val ctx = if (useMRI) ScaleContext.create() else ScaleContext.createIdentity()
    ctx.setScale(ScaleType.USR_SCALE.of(1.0))
    var img = createImage(ctx, isDark)
    if (img != null && useMRI) {
      img = MultiResolutionImageProvider.convertFromJBImage(img)
    }
    return img?.let { ImageIcon(it) } ?: EMPTY_ICON
  }

  override fun copy(): ProtocolCachedImageIcon {
    return ProtocolCachedImageIcon(originalPath, isDark, sources)
  }

  private inner class MyScaledIconsCache {
    private val scaledIconsCache = Collections.synchronizedMap(FixedHashMap<ScaleContext, SoftReference<JBImageIcon>>(
      SCALED_ICONS_CACHE_LIMIT))

    fun getOrScaleIcon(scale: Float): ImageIcon {
      var ctx = scaleContext
      if (scale != 1f) {
        ctx = ctx.copy()
        ctx.setScale(ScaleType.OBJ_SCALE.of(scale.toDouble()))
      }
      val cachedIcon = SoftReference.dereference(scaledIconsCache[ctx])
      if (cachedIcon != null) {
        return cachedIcon
      }

      val icon = createImage(ctx, isDark)?.let { JBImageIcon(it) } ?: EMPTY_ICON
      if (ImageLoader.ImageCache.isIconTooLargeForCache(icon)) {
        ctx = ctx.copy() // ctx is mutable
        scaledIconsCache[ctx] = SoftReference(icon)
      }
      return icon
    }
  }

  companion object {
    private val LOG = logger<ProtocolCachedImageIcon>()

    private const val SCALED_ICONS_CACHE_LIMIT = 5

    private val EMPTY_ICON = object : JBImageIcon(
      ImageUtil.createImage(1, 1, BufferedImage.TYPE_3BYTE_BGR)) {
      override fun toString(): String {
        return "Empty icon " + super.toString()
      }
    }
  }
}