// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.icons

import com.intellij.openapi.diagnostic.logger
import com.intellij.util.ui.EmptyIcon
import com.intellij.util.ui.JBCachingScalableIcon
import com.jetbrains.rd.framework.IRdCall
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.adviseOnce
import com.jetbrains.rd.ide.model.IconModel
import java.awt.Component
import java.awt.Graphics
import java.util.concurrent.atomic.AtomicBoolean
import javax.swing.Icon

class ProtocolLazyIcon<T>(
  private val iconHost: RdIconHost,
  baseIcon: Icon?,
  private val param: T,
  private val evaluator: IRdCall<T, IconModel?>,
  private val invalidator: (ProtocolLazyIcon<T>) -> Long
) : JBCachingScalableIcon<ProtocolLazyIcon<T>>() {
  companion object {
    private val LOG = logger<ProtocolLazyIcon<*>>()
    val EMPTY_ICON = EmptyIcon.create(16).withIconPreScaled(false)
  }

  private var currentIcon = baseIcon ?: EMPTY_ICON
  private var requested = AtomicBoolean()
  private var currentTimestamp = -1L

  private fun requestIcon() {
    val lifetime = iconHost.project.lifetime
    if (lifetime.isNotAlive) return

    val requestTimestamp = invalidator(this)
    if (requestTimestamp == currentTimestamp) return
    if (!requested.compareAndSet(false, true)) return
    currentTimestamp = requestTimestamp

    return evaluator.start(lifetime, param).result.adviseOnce(lifetime) {
      when (it) {
        is RdTaskResult.Success -> {
          val model = it.value
          if (requestTimestamp == invalidator(this) && model != null) {
            currentIcon = iconHost.toIdeaIcon(model)
          }
          requested.set(false)
        }
        is RdTaskResult.Cancelled -> {
          requested.set(false)
        }
        is RdTaskResult.Fault -> {
          LOG.error("evaluation faulted. evaluator: $evaluator, ${it.error}")
        }
      }
    }
  }

  override fun getIconHeight(): Int {
    return currentIcon.iconHeight
  }

  override fun getIconWidth(): Int {
    return currentIcon.iconWidth
  }

  override fun paintIcon(c: Component?, g: Graphics?, x: Int, y: Int) {
    requestIcon()
    currentIcon.paintIcon(c, g, x, y)
  }

  override fun copy(): ProtocolLazyIcon<T> {
    return ProtocolLazyIcon(
      iconHost,
      currentIcon,
      param,
      evaluator,
      invalidator).also { copy ->
      copy.requested = requested
      copy.currentTimestamp = currentTimestamp
    }
  }
}