// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.icons

import com.intellij.openapi.application.ApplicationManager
import com.intellij.ui.PaintingParent
import com.intellij.ui.tabs.impl.TabLabel
import com.intellij.util.Alarm
import com.intellij.util.IconUtil
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.containers.TransferToEDTQueue
import com.intellij.util.ui.UIUtil
import com.jetbrains.rd.ide.model.RotatedIconModel
import java.awt.*
import java.util.*
import javax.swing.*

class RotatedIconWrapper(iconNode: RotatedIconModel, iconHost: RdIconHost) : IconWrapper {
  private val myStaticIcon = iconHost.createWrapper(iconNode.staticNode).createIcon()

  override fun createIcon(): Icon {
    return RotatedIcon(myStaticIcon)
  }

  class RotatedIcon(val staticIcon: Icon, val speedFactor: Int = 5, val direction: Int = 1) : Icon {

    override fun paintIcon(c: Component?, g: Graphics, x: Int, y: Int) {
      val time = System.nanoTime() / 1000 / 1000 / speedFactor
      val angle = Math.toRadians((time % 360).toDouble()) * direction

      staticIcon.paintRotatedIcon(c, g, x, y, angle)

      requestRepaint(c, x, y)
    }

    private fun requestRepaint(c: Component?, x: Int, y: Int) {
      if (c == null) return
      val target = getTarget(c)
      val paintingParent = SwingUtilities.getAncestorOfClass(PaintingParent::class.java, c)
      val paintingParentRec = if (paintingParent == null) null else (paintingParent as PaintingParent).getChildRec(c)
      ourIconsCalculatingExecutor.execute {
        ourLaterInvocator.offer(Runnable {
          var actualTarget: Component? = target
          if (actualTarget != null && SwingUtilities.getWindowAncestor(actualTarget) == null) {
            actualTarget = paintingParent
            if (actualTarget == null || SwingUtilities.getWindowAncestor(actualTarget) == null) {
              actualTarget = null
            }
          }

          if (actualTarget == null) return@Runnable

          ourRepaintScheduler.pushDirtyComponent(actualTarget, paintingParentRec)
        })
      }
    }

    override fun getIconWidth() = staticIcon.iconWidth
    override fun getIconHeight() = staticIcon.iconHeight

    private class RepaintScheduler {
      private val myAlarm = Alarm()
      private val myQueue = LinkedHashSet<RepaintRequest>()

      fun pushDirtyComponent(c: Component, rec: Rectangle?) {
        ApplicationManager.getApplication().assertIsDispatchThread() // assert myQueue accessed from EDT only
        myAlarm.cancelAllRequests()
        myAlarm.addRequest({
                             for (request in myQueue) {
                               val r = request.rectangle
                               if (r == null) {
                                 request.component.repaint()
                               }
                               else {
                                 request.component.repaint(r.x, r.y, r.width, r.height)
                               }
                             }
                             myQueue.clear()
                           }, 50)

        myQueue.add(RepaintRequest(c, rec))
      }
    }

    private class RepaintRequest(val component: Component, val rectangle: Rectangle?)

    companion object {
      private val ourRepaintScheduler = RepaintScheduler()
      private val ourIconsCalculatingExecutor = AppExecutorUtil.createBoundedApplicationPoolExecutor("IconsCalculatingExecutor", 1)
      private val ourLaterInvocator = TransferToEDTQueue.createRunnableMerger("Rotated icon later invocator", 200)

      private fun getTarget(c: Component): Component {
        val list = SwingUtilities.getAncestorOfClass(JList::class.java, c)
        if (list != null)
          return list

        val tree = SwingUtilities.getAncestorOfClass(JTree::class.java, c)
        if (tree != null)
          return tree

        val table = SwingUtilities.getAncestorOfClass(JTable::class.java, c)
        if (table != null)
          return table

        val box = SwingUtilities.getAncestorOfClass(JComboBox::class.java, c)
        if (box != null)
          return box

        val tabLabel = SwingUtilities.getAncestorOfClass(TabLabel::class.java, c)
        if (tabLabel != null)
          return tabLabel

        return c
      }

      private fun Icon.paintRotatedIcon(c: Component?, g: Graphics, x: Int, y: Int, angle: Double) {
        val cx = x + this.iconWidth / 2.0
        val cy = y + this.iconHeight / 2.0

        val gg = g.create() as Graphics2D
        gg.setRenderingHint(RenderingHints.KEY_INTERPOLATION, RenderingHints.VALUE_INTERPOLATION_BILINEAR)
        gg.rotate(angle, cx, cy)
        UIUtil.drawImage(gg, IconUtil.toImage(this), x, y, c)
        gg.dispose()
      }
    }
  }

}