// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rd.ui.util

import com.intellij.ui.JBColor
import java.awt.Color


fun getContrastedForegroundIfNotThemed(background: Color,
                                       foreground: Color,
                                       addContrastToForegroundIfNeeded: Boolean = true): Color {
  // JBColor already has two different colors for dark and light themes
  // So we don't need to change it
  if (foreground is JBColor || !addContrastToForegroundIfNeeded)
    return foreground
  return getContrastedForeground(background, foreground)
}

/**
 * Corrects foreground color by background based on lightness difference (copy-pasted from R#)
 */
fun getContrastedForeground(background: Color, foreground: Color): Color {
  val contrast = 0.45f
  val margin = 0.02f
  val whiteBalance = 0.66f

  var (fL, fA, fB) = CIELabUtils.fromRGB(foreground.getColorComponents(null))
  val (bL, bA, bB) = CIELabUtils.fromRGB(background.getColorComponents(null))

  // Check whether we have to tune lightness
  if (Math.abs(bL - fL) / 100f < contrast) {
    // Make possible contrast, considering the given preferences of
    // balance of white/black and margin (to prevent color missing)
    if (bL / 100f > whiteBalance) {
      val lightness = Math.min(fL / 100f, margin);
      if ((bL / 100f) - lightness < contrast) fL = lightness * 100f;
      else fL = ((bL / 100f) - contrast) * 100f;
    }
    else {
      val lightness = Math.max(fL / 100f, 1f - margin);
      if (lightness - (bL / 100f) < contrast) fL = lightness * 100f;
      else fL = ((bL / 100f) + contrast) * 100f;
    }

    val (r, g, b) = CIELabUtils.toRGB(floatArrayOf(fL, fA, fB));
    val rgb = Color(r, g, b)

    // Hue preserving!
    // Result color can be outside the color gamut of monitor,
    // so we can see light purple (instead of blue) etc.
    // Here we try to preserve color hue

    var min = Math.min(Math.min(foreground.red, foreground.green), foreground.blue);
    var max = Math.max(Math.max(foreground.red, foreground.green), foreground.blue);
    if (max != min) {
      // Original portions
      val redPortion = (foreground.red - min) / (max - min).toDouble()
      val greenPortion = (foreground.green - min) / (max - min).toDouble()
      val bluePortion = (foreground.blue - min) / (max - min).toDouble()

      val sorted = arrayOf(rgb.red, rgb.green, rgb.blue);
      sorted.sort()


      // Do some strange thing
      min = (sorted[0] + sorted[1]) / 2;
      max = sorted[2];

      return Color(
        (redPortion * (max - min) + min).toInt(),
        (greenPortion * (max - min) + min).toInt(),
        (bluePortion * (max - min) + min).toInt(),
        foreground.alpha);
    }
    return Color(rgb.red, rgb.green, rgb.blue, foreground.alpha);
  }
  else return foreground;
}

object CIELabUtils {
  private val xyz_D65 = floatArrayOf(0.9505f, 1.0f, 1.0890f)

  fun fromRGB(rgbvalue: FloatArray): FloatArray {
    return XYZtoLab(RGBtoXYZ(rgbvalue))
  }

  fun toRGB(colorvalue: FloatArray): FloatArray {
    return XYZtoRGB(LabtoXYZ(colorvalue))
  }

  private fun RGBtoXYZ(rgb: FloatArray): FloatArray {
    // normalized red, green, blue values
    val (rLinear, gLinear, bLinear) = rgb

    // convert to a sRGB form
    val r = if (rLinear > 0.04045f) Math.pow((rLinear + 0.055) / (1 + 0.055), 2.2).toFloat() else rLinear / 12.92f
    val g = if (gLinear > 0.04045f) Math.pow((gLinear + 0.055) / (1 + 0.055), 2.2).toFloat() else gLinear / 12.92f
    val b = if (bLinear > 0.04045f) Math.pow((bLinear + 0.055) / (1 + 0.055), 2.2).toFloat() else bLinear / 12.92f

    // converts
    // matrix for white point D65
    var x = r * 0.4124f + g * 0.3576f + b * 0.1805f
    var y = r * 0.2126f + g * 0.7152f + b * 0.0722f
    var z = r * 0.0193f + g * 0.1192f + b * 0.9505f

    x = if (x > 0.9505f) 0.9505f else if (x < 0f) 0f else x
    y = if (y > 1.0f) 1.0f else if (y < 0f) 0f else y
    z = if (z > 1.089f) 1.089f else if (z < 0f) 0f else z
    return floatArrayOf(x, y, z)
  }

  private fun XYZtoLab(xyz: FloatArray): FloatArray {
    val (x, y, z) = xyz
    val (X, Y, Z) = xyz_D65
    return floatArrayOf(
      116f * f(y / Y) - 16,
      500f * (f(x / X) - f(y / Y)),
      200f * (f(y / Y) - f(z / Z))
    )
  }

  private fun LabtoXYZ(lab: FloatArray): FloatArray {
    val theta = 6f / 29f
    val (l, a, b) = lab
    val fy = (l + 16f) / 116f
    val fx = fy + a / 500f
    val fz = fy - b / 200f
    val (X, Y, Z) = xyz_D65

    var x = if (fx > theta) X * (fx * fx * fx) else (fx - 16f / 116f) * 3 * (theta * theta) * X
    var y = if (fy > theta) Y * (fy * fy * fy) else (fy - 16f / 116f) * 3 * (theta * theta) * Y
    var z = if (fz > theta) Z * (fz * fz * fz) else (fz - 16f / 116f) * 3 * (theta * theta) * Z

    x = if (x > 0.9505f) 0.9505f else if (x < 0f) 0f else x
    y = if (y > 1.0f) 1.0f else if (y < 0f) 0f else y
    z = if (z > 1.089f) 1.089f else if (z < 0f) 0f else z
    return floatArrayOf(x, y, z)
  }

  private fun XYZtoRGB(xyz: FloatArray): FloatArray {
    val (x, y, z) = xyz
    val cLinear = floatArrayOf(
      x * 3.2410f - y * 1.5374f - z * 0.4986f,
      -x * 0.9692f + y * 1.8760f + z * 0.0416f,
      x * 0.0556f - y * 0.2040f + z * 1.0570f)
    for (i in 0..2) {
      cLinear[i] = if (cLinear[i] <= 0.0031308f) 12.92f * cLinear[i]
      else (1 + 0.055f) * Math.pow(cLinear[i].toDouble(), 1.0 / 2.2).toFloat() - 0.055f
      if (cLinear[i] > 1f) cLinear[i] = 1f
      if (cLinear[i] < 0) cLinear[i] = 0f
    }
    return cLinear
  }

  private fun f(t: Float): Float {
    return if (t > 0.008856f) Math.pow(t.toDouble(), 1.0 / 3.0).toFloat() else 7.787f * t + 16f / 116f
  }
}