// Copyright 2000-2021 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdclient.util.idea.ui

import com.intellij.openapi.editor.colors.EditorFontType
import com.intellij.ui.ColoredTextContainer
import com.intellij.ui.JBColor
import com.intellij.ui.SimpleTextAttributes
import com.intellij.usages.TextChunk
import com.intellij.util.ui.StatusText
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.BeEmptyText
import com.jetbrains.ide.model.uiautomation.BeUserAction
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.IProperty
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ui.RdUiBundle
import com.jetbrains.rd.ui.themes.ColorHost
import com.jetbrains.rd.ui.themes.rgbToColor
import com.jetbrains.rd.ui.themes.rgbaToColor
import com.jetbrains.rd.ui.util.getContrastedForegroundIfNotThemed
import java.awt.Color
import java.awt.Font
import java.awt.event.ActionListener
import java.util.*

fun RichTextModel.toTextChunks(): Array<TextChunk> {
  val chunks = ArrayList<TextChunk>()
  for (part in parts) {
    // todo patch SimpleColoredComponent to calculate foreground color in contrast normalization mode
    val attributes = part.toSimpleTextAttributes(UIUtil.getTreeTextBackground())
    val chunk = TextChunk(attributes.toTextAttributes(), part.text)
    chunks.add(chunk)
  }
  return chunks.toTypedArray()
}

fun AbstractColor.toColor(): Color {
  return when (this) {
    is RgbaColor -> this.rgbaToColor()
    is RgbColor -> this.rgbToColor()
    is DualColor -> JBColor(this.lightColor.rgbaToColor(), this.darkColor.rgbaToColor())
    is ThemeColorId -> ColorHost.getInstance().getColor(this)
    else -> throw IllegalArgumentException("Unexpected AbstractColor value: ${this}")
  }
}


fun RichStringModel.toSimpleTextAttributes(globalBgColor: Color? = null, overrideFgColor: Color? = null, addContrastToForegroundIfNeeded:Boolean = true): SimpleTextAttributes {
  var style = this.fontStyle.toFont()
  when (this.effectStyle) {
    EffectStyle.StraightUnderline -> style = style or SimpleTextAttributes.STYLE_UNDERLINE
    EffectStyle.WeavyUnderline -> style = style or SimpleTextAttributes.STYLE_WAVED
    EffectStyle.StrikeOut -> style = style or SimpleTextAttributes.STYLE_STRIKEOUT
    EffectStyle.SearchMatch_IDEA -> style = style or SimpleTextAttributes.STYLE_SEARCH_MATCH
    EffectStyle.None -> {
    }
    else -> throw IllegalArgumentException("Unexpected EffectStyle value:  ${this.effectStyle.toString()}")
  }

  var fgColor = overrideFgColor ?: this.foregroundColor?.toColor()
  val localBgColor = this.backgroundColor?.toColor()
  val bgColor = localBgColor ?: globalBgColor
  if (fgColor != null && bgColor != null) {
    fgColor = getContrastedForegroundIfNotThemed(bgColor, fgColor, addContrastToForegroundIfNeeded)
  }

  return SimpleTextAttributes(
    localBgColor,
    fgColor,
    this.effectColor?.toColor(),
    style
  )
}

fun ColoredTextContainer.append(richText: RichTextModel?, bgColor: Color?) {
  if (richText == null) {
    append(RdUiBundle.message("becontrols.unnamed.item"), SimpleTextAttributes.REGULAR_ATTRIBUTES)
    return
  }

  for (part in richText.parts) {
    val attributes = part.toSimpleTextAttributes(bgColor)
    append(part.text, attributes)
  }
}

fun FontStyle.toFont() = when (this) {
  FontStyle.Regular -> SimpleTextAttributes.STYLE_PLAIN
  FontStyle.Italic -> SimpleTextAttributes.STYLE_ITALIC
  FontStyle.Bold -> SimpleTextAttributes.STYLE_BOLD
  FontStyle.Strikeout -> SimpleTextAttributes.STYLE_STRIKEOUT
  FontStyle.Underline -> SimpleTextAttributes.STYLE_UNDERLINE
  else -> throw IllegalArgumentException()
}

fun FontStyle.toEditorFontType() = when (this) {
  FontStyle.Regular -> EditorFontType.PLAIN
  FontStyle.Italic -> EditorFontType.ITALIC
  FontStyle.Bold -> EditorFontType.BOLD
  else -> throw IllegalArgumentException("Style $this can't be converted to editor font style")
}

fun FontStyle.toSwingFontStyle() = when (this) {
  FontStyle.Italic -> Font.ITALIC
  FontStyle.Bold -> Font.BOLD
  else -> Font.PLAIN
}

fun StatusText.bindToBeEmptyText(
  lifetime: Lifetime,
  emptyText: IProperty<BeEmptyText?>
) {
  emptyText.advise(lifetime) {
    clear()
    if (it == null)
      text = StatusText.getDefaultEmptyText()
    else {
      appendRichText(it.text, it.actions)
      val secondaryText = it.secondaryText
      if (secondaryText != null)
        appendSecondaryRichText(secondaryText, it.secondaryActions)
    }
  }
}

fun StatusText.appendRichText(richText: RichTextModel, actions: List<BeUserAction>?) {
  appendRichText(richText, actions, this::appendText)
}

fun StatusText.appendSecondaryRichText(richText: RichTextModel, actions: List<BeUserAction>?) {
  appendRichText(richText, actions, this::appendSecondaryText)
}

private fun appendRichText(
  richText: RichTextModel,
  actions: List<BeUserAction>?,
  appender: (String, SimpleTextAttributes, ActionListener?) -> StatusText
) {
  val onlyOnePartAndOneActionExist = richText.parts.size == 1 && actions?.size == 1
  var actionIdx = 0
  for (part in richText.parts) {
    val partIsUnderlined = part.fontStyle == FontStyle.Underline
    if (actions != null && (partIsUnderlined || onlyOnePartAndOneActionExist)) {
      // Appending link
      val userAction = actions[actionIdx++]
      val actionListener = ActionListener { userAction.click.fire(Unit); }
      appender(part.text, SimpleTextAttributes.LINK_PLAIN_ATTRIBUTES, actionListener)
    }
    else {
      // Appending text
      appender(part.text, StatusText.DEFAULT_ATTRIBUTES, null)
    }
  }
}