package com.jetbrains.rdserver.unattendedHost.deploy

import com.intellij.ide.plugins.PluginManager
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.application.ApplicationInfo
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.Task
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.popup.PopupStep
import com.intellij.openapi.ui.popup.util.BaseListPopupStep
import com.intellij.remote.RemoteCredentials
import com.intellij.ui.popup.PopupFactoryImpl
import com.intellij.util.io.HttpRequests
import com.intellij.util.io.ZipUtil
import com.jetbrains.cwm.unattended.common.UnattendedHostConstants
import com.jetbrains.cwm.unattended.common.UnattendedHostControlConnection
import com.jetbrains.cwm.unattended.common.UnattendedHostProjectListDialog
import com.jetbrains.cwm.unattended.ssh.SftpChannelFacade
import com.jetbrains.cwm.unattended.ssh.SshConfigFacade
import com.jetbrains.cwm.unattended.ssh.SshUtil
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.adviseNotNull
import org.jetbrains.io.JsonReaderEx
import org.jetbrains.io.JsonUtil
import java.io.File
import java.nio.file.Path
import java.util.concurrent.TimeUnit

class RunHostOnRemoteAction : AnAction() {
  override fun actionPerformed(e: AnActionEvent) {
    val configs = SshUtil.getConfigs(e.project)
    if (configs.isEmpty()) {
      val builder = PopupFactoryImpl.getInstance().createListPopup(object : BaseListPopupStep<String>("No remote hosts available", "Open SSH Configurations") {
        override fun onChosen(selectedValue: String?, finalChoice: Boolean): PopupStep<*>? {
          if (finalChoice && selectedValue != null)
            SshUtil.showConfigConfigurable(e.project)
          return null
        }

        override fun hasSubstep(selectedValue: String?) = false
        override fun canceled() = Unit
        override fun getTextFor(value: String) = value
      })

      builder.showInBestPositionFor(e.dataContext)

      return
    }

    val builder = PopupFactoryImpl.getInstance().createListPopup(object : BaseListPopupStep<SshConfigFacade>("Select remote host to run Code With Me Host on", configs) {
      override fun onChosen(selectedValue: SshConfigFacade?, finalChoice: Boolean): PopupStep<*>? {
        if (selectedValue != null && finalChoice) {
          // todo: ask for password if it's not saved
          RunHostOnRemoteTask(e.project, selectedValue.copyToCredentials()).queue()
        }
        return null
      }

      override fun hasSubstep(selectedValue: SshConfigFacade?) = false
      override fun canceled() = Unit
      override fun getTextFor(value: SshConfigFacade) = value.presentableFullName
    })

    builder.showInBestPositionFor(e.dataContext)
  }

  companion object {
    private var logger = Logger.getInstance(RunHostOnRemoteAction::class.java)

    private const val Timeout: Long = 10_000

    private val nightlyDownloadLinksPerCode = mapOf(
      "IIC" to "https://buildserver.labs.intellij.net/guestAuth/repository/download/ijplatform_master_Idea_Installers/%s/ideaIC-%s",
      "IC" to "https://buildserver.labs.intellij.net/guestAuth/repository/download/ijplatform_master_Idea_Installers/%s/ideaIC-%s",
      "IU" to "https://buildserver.labs.intellij.net/guestAuth/repository/download/ijplatform_master_Idea_Installers/%s/ideaIU-%s",
    )

    private const val nightlyPluginDownloadTemplate = "https://buildserver.labs.intellij.net/guestAuth/repository/download/ijplatform_master_Idea_Experiments_CodeWithMeClient/%s/CWMG-plugins/codeWithMe-plugin-%s.zip"
    private const val nightlyLastBuildUrlTemplate = "https://buildserver.labs.intellij.net/guestAuth/app/rest/builds?locator=buildType:%s,count:1&fields=build(number)"
    private val nightlyVersionExtractRegex = Regex("build\\s+number=\"([\\d.]+)\"")
    private val nightlyLastBuildConfigurationPerCode = mapOf(
      "IIC" to "ijplatform_master_Idea_Installers",
      "IC" to "ijplatform_master_Idea_Installers",
      "IU" to "ijplatform_master_Idea_Installers",
    )

  }

  private class RunHostOnRemoteTask(project: Project?, val credentials: RemoteCredentials) : Task.Backgroundable(project, "Waiting to run host on remote", true) {
    override fun shouldStartInBackground(): Boolean {
      return false
    }

    override fun onCancel() {
      // todo: this
      super.onCancel()
    }

    override fun run(indicator: ProgressIndicator) {
      val remoteOsType = guessRemoteOS(credentials)
      logger.debug("Remote OS guessed as $remoteOsType")

      if (remoteOsType == OSKind.MacOS)
        error("Remote host on mac is not supported yet")

      val buildNumber = ApplicationInfo.getInstance().build
      val productCode = buildNumber.productCode
      val requestedVersion = if(buildNumber.isSnapshot) getLatestNightlyBuildNumber(buildNumber.productCode) else buildNumber.asStringWithoutProductCode()

      val downloadLink = getDownloadLinkForOs(productCode, requestedVersion, remoteOsType) ?: error("No host download link")

      val sftpChannel = SshUtil.openSftpChannel(credentials)
      val hostDir = getHostStoreDir(productCode, requestedVersion, remoteOsType)
      if (!sftpChannel.file("$hostDir/.done").exists()) {
        downloadAndExtractHost(sftpChannel, hostDir, downloadLink, remoteOsType)

        val pluginUploadList = getPluginUploadList(requestedVersion)

        val pluginDir = "$hostDir/plugins/"
        pluginUploadList.forEach { (localPath, remotePathSuffix) ->
          logger.debug("Uploading $localPath to remote dir $pluginDir relative path $remotePathSuffix")
          sftpChannel.uploadFileOrDir(localPath.toFile(), pluginDir, remotePathSuffix.toString())
        }
        logger.debug("Removing download zip on remote")
        sftpChannel.file("$hostDir/download${if (remoteOsType == OSKind.Windows) ".zip" else ".tar.gz"}").let { if (it.exists()) it.rm() }
        sftpChannel.file("$hostDir/.done").outputStream(false).close()
      }

      // todo: correct launcher path
      val statusResponse = credentials.execCmd("\"$hostDir/bin/idea${if(remoteOsType == OSKind.Windows) ".bat" else ".sh"}\" cwmHostStatus")
      logger.debug("Remote status response: $statusResponse")
      if (statusResponse.isEmpty() || statusResponse.contains("already running", true))
        error("Remote has an instance without Code With Me plugin running")

      if (statusResponse.contains(UnattendedHostConstants.notRunningInUnattendedMode))
        error("Remote has an instance running in user-interactive mode")

      val remoteVersion = UnattendedHostConstants.currentlyRunningVersionRegex.find(statusResponse)?.groupValues?.get(1) ?: error("Remote did not provide version info")
      if (remoteVersion != "$productCode-$requestedVersion")
        error("Remote has a different version already running (have $remoteVersion, want $productCode-$requestedVersion")

      val controlPort = UnattendedHostConstants.controlPortRegex.find(statusResponse)?.groupValues?.get(1)
      if (controlPort != null && controlPort != "0") {
        logger.debug("Remote is already running an instance, connecting to control socket on remote port $controlPort instead")
        application.invokeLater {
          connectToControlSocket(controlPort)
        }
        return
      }

      // todo: proper launcher path (it's not always idea)
      if (remoteOsType == OSKind.Windows)
        credentials.execCmdSucceed("powershell Invoke-WmiMethod -Class Win32_Process -Name Create -ArgumentList '\"$hostDir/bin/idea.bat\" cwmHostUi'")
      else
        credentials.execCmdSucceed("nohup \"$hostDir/bin/idea.sh\" cwmHostUi > /dev/null")

      // Wait for the host to start
      Thread.sleep(5_000) // todo: replace with something proper

      val statusResponse2 = credentials.execCmd("\"$hostDir/bin/idea${if(remoteOsType == OSKind.Windows) ".bat" else ".sh"}\" cwmHostStatus")
      val controlPort2 = UnattendedHostConstants.controlPortRegex.find(statusResponse2)?.groupValues?.get(1)
      if (controlPort2 == null || controlPort2 == "0") {
        error("Remote did not specify control port")
      }

      connectToControlSocket(controlPort2)
    }

    private fun connectToControlSocket(controlPort2: String) {
      val controlConnection = UnattendedHostControlConnection(credentials, controlPort2.toInt())
      controlConnection.controlModel.view(Lifetime.Eternal) { sessionLt, session ->
        session.session.adviseNotNull(sessionLt) {
          val dialog = UnattendedHostProjectListDialog(sessionLt, it)
          dialog.show()

          //Disposer.dispose(controlConnection)
        }
      }
    }

    private fun getPluginUploadList(requestedVersion: String): List<Pair<Path, Path>> {
      if (ApplicationInfo.getInstance().build.isSnapshot) {
        return getTempExtractedNightlyPlugin(requestedVersion)
      }
      val plugin = PluginManager.getLoadedPlugins().single { it.name == "Code With Me" }

      val pluginRoot = plugin.pluginPath
      val pluginRootFile = plugin.pluginPath.toFile()

      return pluginRoot.toFile().walk().mapNotNull { if(it.isFile) it.toPath() to it.relativeTo(pluginRootFile.parentFile).toPath() else null }.toList()
    }

    private fun getTempExtractedNightlyPlugin(requestedVersion: String): List<Pair<Path, Path>> {
      logger.debug("Downloading and extracting nightly plugin to temp locally")
      val pluginZipUrl = nightlyPluginDownloadTemplate.format(requestedVersion, requestedVersion)
      val tempFile = File.createTempFile("cwm-plugin-", ".zip")
      HttpRequests.request(pluginZipUrl).connect { request ->
        request.saveToFile(tempFile, null) // todo: progress indicators everywhere?
      }
      logger.debug("Downloaded plugin to $tempFile")
      val tempDir = File.createTempFile("cwm-plugin-ext", "", null)
      tempDir.delete()
      tempDir.mkdir()
      tempDir.deleteOnExit()
      ZipUtil.extract(tempFile.toPath(), tempDir.toPath(), null)
      logger.debug("Extracted plugin zip to $tempDir")
      tempFile.delete()

      tempDir.walkBottomUp().forEach { it.deleteOnExit() }

      return tempDir.walk().mapNotNull { if(it.isFile) it.toPath() to it.relativeTo(tempDir).toPath() else null }.toList()
    }

    private fun downloadAndExtractHost(sftpChannel: SftpChannelFacade, hostDir: String, downloadLink: String, remoteOsType: OSKind) {
      logger.debug("Downloading/extracting host on the remote")
      if (sftpChannel.file(hostDir).exists()) {
        logger.debug("Remote dir $hostDir already exists, deleting")
        sftpChannel.file(hostDir).rm()
      }

      when(remoteOsType) {
        OSKind.Windows -> downloadAndExtractHostWindows(downloadLink, hostDir)
        else -> downloadAndExtractHostUnix(downloadLink, hostDir)
      }
    }

    private fun downloadAndExtractHostUnix(downloadLink: String, hostDir: String) {
      val remoteHasCurl = credentials.execCmdSucceed("which curl") == 0
      val remoteHasWget = credentials.execCmdSucceed("which wget") == 0

      logger.debug("Remote has: curl $remoteHasCurl wget $remoteHasWget")

      credentials.execCmdAssertSucceed("mkdir \"$hostDir\"")

      if (!remoteHasCurl && !remoteHasWget)
        TODO("Handle local download/upload pair")

      if (remoteHasWget) {
        credentials.execCmdAssertSucceed("wget -O \"$hostDir/download.tar.gz\" \"$downloadLink\"", 600_000)
      } else {
        credentials.execCmdAssertSucceed("wget -f -o \"$hostDir/download.tar.gz\" \"$downloadLink\"", 600_000)
      }

      credentials.execCmdAssertSucceed("tar -x -f \"$hostDir/download.tar.gz\"", 120_000)
    }

    private fun downloadAndExtractHostWindows(downloadLink: String, hostDir: String) {
      credentials.execCmdAssertSucceed("mkdir \"$hostDir\"") // todo: check success for all commands
      credentials.execCmdAssertSucceed("powershell \"(New-Object System.Net.WebClient).DownloadFile(\\\"$downloadLink\\\", \\\"$hostDir/download.zip\\\")\"", 600_000)
      credentials.execCmdAssertSucceed("powershell Expand-Archive -LiteralPath '$hostDir/download.zip' -DestinationPath '$hostDir'", 120_000)
    }

    private fun getRemoteAppDataFolder(os: OSKind): String {
      return when (os) {
        OSKind.Windows -> credentials.execCmd("cmd /C \"echo %LOCALAPPDATA%\"").trim()
        OSKind.MacOS -> credentials.execCmd("sh -c \"echo \$HOME/.local/share\"").trim()
        OSKind.Linux -> credentials.execCmd("sh -c \"echo \$HOME/.local/share\"").trim()
      }
    }

    private fun getHostStoreDir(productCode: String, buildNumber: String, os: OSKind): String {
      val appdata = getRemoteAppDataFolder(os)
      return "$appdata/JetBrains/CwmHost/$productCode-$buildNumber"
    }

    private fun getDownloadLinkForOs(productCode: String, buildNumber: String, remoteOsType: OSKind): String? {
      val targetUrl = "https://data.services.jetbrains.com/products?code=$productCode"
      logger.debug("Getting product list from $targetUrl")
      val releaseOrEapLink = HttpRequests.request(targetUrl).connect { request ->
        val jsonString = request.readString()
        val json = JsonUtil.nextList<Map<String, Any>>(JsonReaderEx(jsonString))
        val productObject = json[0]
        val releaseList = productObject["releases"] as List<Map<String, Any>>
        val selectedBuild = releaseList.firstOrNull { it["build"]?.toString() == buildNumber }
        if(selectedBuild == null) return@connect null
        val osString = when (remoteOsType) {
          OSKind.Windows -> "windowsZip"
          OSKind.Linux -> "linux"
          OSKind.MacOS -> "mac"
        }
        val downloadLink = (selectedBuild["downloads"] as Map<String, Map<String, String>>)[osString]!!["link"]
        logger.debug("Download link extracted: $downloadLink")
        downloadLink
      }

      if (releaseOrEapLink != null)
        return releaseOrEapLink

      return getNightlyDownloadLink(productCode, buildNumber, remoteOsType)
    }

    private fun getLatestNightlyBuildNumber(productCode: String): String {
      val buildName = nightlyLastBuildConfigurationPerCode[productCode] ?: error("Nightly build names for current product are not set up")
      val lastNightly = HttpRequests.request(nightlyLastBuildUrlTemplate.format(buildName)).connect { request ->
        val xml = request.readString() // todo: use progress indicator versions for working with network?
        return@connect nightlyVersionExtractRegex.find(xml)?.groups?.get(1)?.value ?: error("Unable to parse latest nightly version")
      }
      logger.debug("Fetched nightly version is $lastNightly")
      return lastNightly
    }

    private fun getNightlyDownloadLink(productCode: String,
                                       buildNumber: String,
                                       remoteOsType: OSKind): String? {
      val nightlyLinkFormat = nightlyDownloadLinksPerCode[productCode] ?: return null
      val nightlyLinkNoExt = String.format(nightlyLinkFormat, buildNumber, buildNumber)

      // todo: check availability locally via head request?
      return nightlyLinkNoExt + when (remoteOsType) {
        OSKind.Windows -> ".win.zip"
        OSKind.Linux -> ".tar.gz"
        OSKind.MacOS -> ".portable.zip"
      }
    }

    private fun guessRemoteOS(credentials: RemoteCredentials): OSKind {
      val output = credentials.execCmd("uname")
      logger.info("Output of 'uname' command: \"${output.trim()}\"")
      if (output.contains("windows", ignoreCase = true) ||
          output.contains("msys", ignoreCase = true) ||
          output.contains("uname", ignoreCase = true) ||
          output.isBlank())
        return OSKind.Windows
      if (output.contains("darwin", ignoreCase = true))
        return OSKind.MacOS
      return OSKind.Linux
    }

    private fun RemoteCredentials.execCmdAssertSucceed(cmd: String, timeoutMs: Long = Timeout) {
      val errorCode = execCmdSucceed(cmd, timeoutMs)
      if (errorCode != 0)
        error("Command \"$cmd\" failed with error code $errorCode")
    }

    private fun RemoteCredentials.execCmdSucceed(cmd: String, timeoutMs: Long = Timeout): Int {
      logger.debug("Executing on remote host: $cmd")
      val process = SshUtil.execute(this, cmd)
      val exited = process.waitFor(timeoutMs, TimeUnit.MILLISECONDS)
      if (exited) {
        logger.info("execCmd for '$cmd' exited normally with code ${process.exitValue()}")
      } else {
        logger.warn("execCmd for '$cmd' was killed due to timeout of $Timeout ms")
      }
      process.destroyForcibly()
      process.inputStream.bufferedReader().readText().also { logger.debug("Command output: $it") }
      return process.exitValue()
    }

    private fun RemoteCredentials.execCmd(cmd: String): String {
      logger.debug("Executing on remote host: $cmd")
      val process = SshUtil.execute(this, cmd)
      val exited = process.waitFor(Timeout, TimeUnit.MILLISECONDS)
      if (exited) {
        logger.info("execCmd for '$cmd' exited normally")
      } else {
        logger.warn("execCmd for '$cmd' was killed due to timeout of $Timeout ms")
      }
      process.destroyForcibly()
      return process.inputStream.bufferedReader().readText().also { logger.debug("Command output: $it") }
    }

    private enum class OSKind {
      Windows,
      Linux,
      MacOS
    }
  }
}