// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.completion

import com.intellij.codeInsight.completion.*
import com.intellij.codeInsight.completion.CompletionPhase.CommittingDocuments
import com.intellij.codeInsight.completion.impl.CompletionServiceImpl
import com.intellij.codeInsight.lookup.LookupElement
import com.intellij.codeInsight.lookup.impl.LookupImpl
import com.intellij.codeInsight.template.postfix.completion.PostfixTemplateLookupElement
import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isLocal
import com.intellij.codeWithMe.ClientId.Companion.withClientId
import com.intellij.openapi.Disposable
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.application.ReadAction
import com.intellij.openapi.command.CommandProcessor
import com.intellij.openapi.command.WriteCommandAction
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.progress.ProcessCanceledException
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.util.ProgressIndicatorUtils
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.roots.ProjectRootManager
import com.intellij.psi.PsiDocumentManager
import com.intellij.psi.PsiManager
import com.intellij.psi.impl.source.tree.injected.InjectedLanguageUtil
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.containers.ContainerUtil
import com.intellij.util.indexing.DumbModeAccessType
import com.intellij.util.indexing.FileBasedIndex
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.rd.ide.completion.markCompletionSessionExecuting
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.concurrentMapOf
import com.jetbrains.rd.util.lifetime.*
import com.jetbrains.rd.util.reactive.viewNotNull
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.completion.BackendCompletionHost.Companion.LIGHT_ITEMS_THRESHOLD_MS
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.BackendEditorHostListener
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.permissions.PermissionManager
import org.jetbrains.annotations.NotNull
import java.util.concurrent.Callable
import java.util.concurrent.atomic.AtomicBoolean
import java.util.function.Supplier
import kotlin.math.max
import kotlin.math.min

@Suppress("NonDefaultConstructor")
class BackendCompletionHost(
  private val project: Project,
  private val clientId: ClientId,
  componentLifetime: Lifetime
) {
  var session: CompletionSession? = null
  var sequentialLifetimes = SequentialLifetimes(componentLifetime)

  class MyEditorHostListener : BackendEditorHostListener {
    override fun editorBound(lifetime: Lifetime, editorModel: TextControlModel, editor: Editor) {
      val project = editor.project ?: return
      RdServerPerClientComponents.getService<BackendCompletionHost>(project).adviseTextControl(lifetime, editorModel, editor)
    }
  }

  private fun adviseTextControl(lifetime: Lifetime, textControlModel: TextControlModel, editor: Editor) {
    val textControlId = editor.getTextControlId() ?: return
    textControlModel.completionModel.completionSession.viewNotNull(lifetime) { sessionLifetime, completionTuple ->
      val currentSessionLifetime = sequentialLifetimes.next().intersect(sessionLifetime)
      viewCompletionSession(textControlId, completionTuple, currentSessionLifetime)
    }
  }

  private fun viewCompletionSession(textControlId: TextControlId,
                                    completionTuple: CompletionTuple,
                                    sessionLifetime: Lifetime) {
    val permissionManager = PermissionManager.getInstance(project)
    if (!permissionManager.hasFileWriteAccess(ClientId.current)) return

    val editor = BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(textControlId) ?: return
    if (!permissionManager.isWritable(ClientId.current, editor.document)) return

    LOG.info("viewing new completion session for offset ${completionTuple.parameters.offset}")
    require(session == null)
    session = CompletionSession(project, clientId, editor, completionTuple, sessionLifetime).apply { this.runOrReschedule() }
    editor.markCompletionSessionExecuting(sessionLifetime)
    sessionLifetime.onTermination {
      LOG.info("completion session for offset '${completionTuple.parameters.offset}' is about to be terminated")
      session = null
    }
  }

  fun completeItem(item: CompletionItem, char: Char, editor: Editor): Boolean {
    val completionSession = session ?: return false
    if (completionSession.editor != editor) return false

    CommandProcessor.getInstance().executeCommand(project, Runnable {
      completionSession.insertSelectedLookupItem(item, char)
    }, null, null)

    return true
  }

  fun getCorrespondingLookupElement(completionItem: CompletionItem): LookupElement? {
    return session?.lookupElementMap?.get(completionItem)
  }

  companion object {
    val LOG = Logger.getInstance(BackendCompletionHost::class.java)
    val LIGHT_ITEMS_THRESHOLD_MS = 200

    fun getInstance(project: Project) = RdServerPerClientComponents.getService(project, BackendCompletionHost::class.java)
  }
}

class CompletionSession(
  private val project: Project,
  private val clientId: ClientId,
  val editor: Editor,
  val completionTuple: CompletionTuple,
  val sessionLifetime: Lifetime
) {

  init {
    LOG.info("creating new backend calculation session for $clientId on offset ${completionTuple.parameters.offset}")
  }
  private val completionService = CompletionService.getCompletionService() as CompletionServiceImpl
  private val sessionOffset = completionTuple.parameters.offset
  private var isCalculating :AtomicBoolean = AtomicBoolean(false)

  private var calculationStartTime: Long = 0
  private val rerunQueue = MergingUpdateQueue("rerunQueue", 50, true, MergingUpdateQueue.ANY_COMPONENT,
                                              sessionLifetime.createNestedDisposable())

  private val completionType = when (completionTuple.parameters.mode) {
    CompletionMode.Basic -> CompletionType.BASIC
    CompletionMode.Smart -> CompletionType.SMART
    CompletionMode.Import -> CompletionType.CLASS_NAME
  }

  private val completionHandler = CodeCompletionHandlerBase.createHandler(completionType, !completionTuple.parameters.isAutoCompletion,
                                                                          completionTuple.parameters.isAutoCompletion, true)
  internal val lookupElementMap = concurrentMapOf<CompletionItem, LookupElement>()

  private val sentItems = ContainerUtil.newConcurrentSet<LookupElementIdentity>()

  /* set up params */

  val file = PsiDocumentManager.getInstance(project).getPsiFile(editor.document) ?: error("cannot get psi file")
  private val hostOffsets = OffsetsInFile(file)
  private val offsetMap = hostOffsets.offsets
  private var identifierStart = sessionOffset
  private val completionProcess: CompletionProcessEx = BackendCompletionProcess(this, editor, hostOffsets, offsetMap,
                                                                                completionTuple.parameters.isAutoCompletion)
  private val arranger = BaseCompletionLookupArranger(completionProcess)

  //background thread
  private fun prepareCompletion(context: CompletionInitializationContextImpl): Supplier<OffsetsInFile>? {
    val startOffset = CompletionInitializationContext.calcStartOffset(editor.caretModel.currentCaret)

    LOG.info(
      "refreshing completion params for offset ${editor.caretModel.currentCaret.offset}. Start offset = $startOffset. Identifier start = $identifierStart")

    offsetMap.addOffset(CompletionInitializationContext.START_OFFSET, startOffset)
    offsetMap.addOffset(CompletionInitializationContext.SELECTION_END_OFFSET,
                        CompletionInitializationContext.calcSelectionEnd(editor.caretModel.currentCaret))
    offsetMap.addOffset(
      CompletionInitializationContext.IDENTIFIER_END_OFFSET,
      CompletionInitializationContext.calcDefaultIdentifierEnd(editor, CompletionInitializationContext.calcSelectionEnd(
        editor.caretModel.currentCaret))
    )

    var createNestedDisposable: Disposable? = null
    sessionLifetime.executeIfAlive {
      createNestedDisposable = sessionLifetime.createNestedDisposable()
    }

    // retrieve the injected file from scratch since our typing might have destroyed the old one completely
    val topLevelFile = PsiDocumentManager.getInstance(project).getPsiFile(editor.document)
    val completionEditor = InjectedLanguageUtil.getEditorForInjectedLanguageNoCommit(editor,
                                                                                     topLevelFile,
                                                                                     editor.caretModel.offset)
    val file = PsiDocumentManager.getInstance(project).getPsiFile(completionEditor.document)
    if (file == null || completionTuple.parameters.isAutoCompletion
        && CommittingDocuments.shouldSkipAutoPopup(completionEditor, file)) {
      return null
    }

    if (!file.isWritable)
      return null

    CommittingDocuments.loadContributorsOutsideEdt(completionEditor, file)

    return CompletionInitializationUtil.insertDummyIdentifier(context, hostOffsets, createNestedDisposable, true)
  }

  fun runOrReschedule(refresh : Boolean = false) {
    LOG.info("scheduling to rerun completion for ${ClientId.current}")

    val runDelegate = object : Update("rerun") {
      override fun run() {
        if (editor.isDisposed) return
        if (refresh) completionTuple.session.refreshFrontend.fire("prefixUpdated")

        withClientId(clientId) {
          this@CompletionSession.run()
        }
      }
    }

    withClientId(clientId) {
      if (notTheBestTimeForCompletion()) {
        LOG.info("Not the best time for completion. Reschedule.")
        if (sessionLifetime.isAlive) {
          rerunQueue.queue(runDelegate)
        }
      }
      else {
        LOG.info("good time to start completion calculation. Begin.")
        runDelegate.run()
      }
    }
  }

  private fun notTheBestTimeForCompletion(): Boolean {
    val lead = FollowMeManager.getInstance(project).followers[clientId] ?: return false

    if (lead.isLocal) {
      return CompletionServiceImpl.isPhase(
        CommittingDocuments::class.java,
        CompletionPhase.BgCalculation::class.java,
        CompletionPhase.Synchronous::class.java
      )
    }
    else {
      withClientId(lead) {
        val leadsSession = BackendCompletionHost.getInstance(project).session ?: return true
        return leadsSession.isCalculating.get()
      }
    }
  }

  fun run() {
    require(!ClientId.isCurrentlyUnderLocalId) {
      "this method should not be run under local client id"
    }

    LOG.info("running backend completion for ${ClientId.current}")

    completionTuple.session.readyToReceiveItems.advise(sessionLifetime) { args ->
      application.assertIsDispatchThread()
      application.assertIsWriteThread()

      LOG.info("ready to receive items. Begin calculation ${ClientId.current}")

      if (!args || sessionLifetime.isNotAlive || sessionOffset > editor.document.textLength)
        return@advise

      PsiDocumentManager.getInstance(project).performWhenAllCommitted {
        if (sessionLifetime.isNotAlive || sessionOffset > editor.document.textLength)
          return@performWhenAllCommitted

        runOnCommittedPsi()
      }
    }
  }

  private fun runOnCommittedPsi() {
    val nestedDisposable = sessionLifetime.createNestedDisposable()

    identifierStart = calcDefaultIdentifierStart(editor, sessionOffset)
    val completionContext = CompletionInitializationUtil.runContributorsBeforeCompletion(editor, file, 1, editor.caretModel.currentCaret,
                                                                                         completionType)

    ReadAction
      .nonBlocking(
        Callable { prepareCompletion(completionContext) })
      .expireWith(nestedDisposable)
      .withDocumentsCommitted(project)
      .finishOnUiThread(
        ModalityState.defaultModalityState()) { supplier ->

        if (supplier == null || sessionLifetime.isNotAlive)
          return@finishOnUiThread

        withClientId(clientId) {
          val copyOffsets = supplier.get()
          val finalOffsets = CompletionInitializationUtil.toInjectedIfAny(file, copyOffsets)
          val parameters = CompletionInitializationUtil.createCompletionParameters(completionContext, completionProcess, finalOffsets)
          completionProcess.parameters = parameters

          arranger.clearClassifierCache()

          LOG.info(
            "refreshed completion params. params offset - ${parameters.offset}, real editor offset - ${parameters.editor.caretModel.offset}")
          //RUN!
          calculateCompletionVariants(parameters)
        }

      }
      .submit(AppExecutorUtil.getAppExecutorService())
  }

  private fun calculateCompletionVariants(parameters: com.intellij.codeInsight.completion.CompletionParameters) {
    LOG.info("calculating completion variants...")

    require(!ClientId.isCurrentlyUnderLocalId) {
      "should not be called for the host"
    }

    require(PsiDocumentManager.getInstance(project).isCommitted(editor.document)
    ) { "starting completion calculation for an uncommitted document ${editor.document}" }

    if (!PsiManager.getInstance(project).isInProject(parameters.originalFile)) {
      val projectRoots = ProjectRootManager.getInstance(project).contentRoots
      LOG.info("Invoking completion on non-project file: path ${parameters.originalFile.virtualFile.path}, roots $projectRoots")
    }
    calculationStartTime = System.currentTimeMillis()
    ApplicationManager.getApplication().executeOnPooledThread {
      var batchHasItems = false

      val result = ProgressIndicatorUtils.runInReadActionWithWriteActionPriority(
        {
          FileBasedIndex.getInstance().ignoreDumbMode(DumbModeAccessType.RELIABLE_DATA_ONLY) {
            try {
              isCalculating.set(true)

              completionService.performCompletion(parameters, object : BatchConsumer<CompletionResult> {
                override fun consume(result: CompletionResult) {
                  ProgressManager.checkCanceled()
                  if (!sessionLifetime.isAlive) {
                    throw ProcessCanceledException()
                  }
                  batchHasItems = true
                  arranger.addElement(result)
                }

                override fun endBatch() {
                  if (batchHasItems) {
                    batchHasItems = false
                    sendItemBatch(false)
                  }
                }
              })

              if (sessionLifetime.isAlive) {
                LOG.info(
                  "[${completionTuple.parameters.offset}] All ${arranger.matchingItems.size} results received in ${System.currentTimeMillis() - calculationStartTime} ms")
                sendItemBatch(true)
              }
            } finally {
              isCalculating.set(false)
            }
          }
        }, null)

      if (!result) {
        LOG.info("Completion process was interrupted. Reschedule...")
        application.invokeLater {
          if (sessionLifetime.isNotAlive) return@invokeLater
          runOrReschedule()
        }
      }
    }
  }

  private fun passLightItemsThreshold(): Boolean {
    require(calculationStartTime > 0) { "Start time was not set" }

    val calculationTime = System.currentTimeMillis() - calculationStartTime
    return calculationTime > LIGHT_ITEMS_THRESHOLD_MS
  }

  private fun sendItemBatch(final: Boolean) {
    val backendItems = arranger.arrangeItems().first

    //suppress for a bit light items to prevent blinking. Replicates IJ's frozen items logic
    if (!final && !passLightItemsThreshold()) {
      LOG.trace("don't send ${backendItems.size} non-final items. Threshold = $LIGHT_ITEMS_THRESHOLD_MS ")
      return
    }

    val newProtocolItems = backendItems.registerNewElements()
    ApplicationManager.getApplication().invokeLater(
      {
        sendNewItems(newProtocolItems)
        if (final) {
          LOG.info("sending FullItemsPrepared signal")
          completionTuple.session.sessionState.fire(SessionState.FullItemsPrepared)
        }
      }, { !sessionLifetime.isAlive })
  }

  private fun sendNewItems(newProtocolItems: List<CompletionItem>) {
    if (newProtocolItems.isEmpty())
      return

    val completionList = CompletionList(newProtocolItems)
    completionTuple.session.completionItems.fire(completionList)
  }

  private fun extractIdentity(lookupElement: LookupElement) =
    LookupElementIdentity(lookupElement.javaClass.name, lookupElement.lookupString)

  fun insertSelectedLookupItem(item: CompletionItem, char: Char) {
    val hostOffsetsNotNull = hostOffsets ?: error("hostOffsets")
    val lookupElement = lookupElementMap[item] ?: error("Can't find lookup element for completion item $item")

    val itemPattern = arranger.itemPattern(lookupElement)
    val initialPrefixLength = itemPattern.length

    val additionalOffset = guessAdditionalOffset(lookupElement, itemPattern)

    WriteCommandAction.runWriteCommandAction(project) {
      editor.document.startGuardedBlockChecking()
      try {
        LookupImpl.insertLookupString(
          project,
          editor,
          lookupElement,
          arranger.itemMatcher(lookupElement),
          itemPattern,
          initialPrefixLength + additionalOffset
        )
      }
      finally {
        editor.document.stopGuardedBlockChecking()
      }
    }

    if (lookupElement.requiresCommittedDocuments()) {
      PsiDocumentManager.getInstance(project).commitAllDocuments()
    }

    completionHandler.handleCompletionElementSelected(
      lookupElement,
      char,
      hostOffsetsNotNull.offsets,
      hostOffsetsNotNull,
      editor,
      editor.caretModel.currentCaret.offset
    )

  }

  private fun guessAdditionalOffset(lookupElement: LookupElement,
                                    itemPattern: @NotNull String): Int {
    val relativeOffset = calculateRelativeOffset()
    val truncatedRelativeOffset = relativeOffset - itemPattern.length
    var additionalOffset = 0.coerceAtLeast(truncatedRelativeOffset)

    if (lookupElement is PostfixTemplateLookupElement) {
      if (additionalOffset != 0)
        additionalOffset++
    }
    //such hacks can be wiped out when CWM-2111 will be ready
    else if (additionalOffset == 0) {
      if (itemPattern.contains(".")) {
        val qualificationLength = itemPattern.lastIndexOf('.') + 1  // Qualification of "os.Environ" is "os."
        additionalOffset = truncatedRelativeOffset + qualificationLength
      } else if (itemPattern.contains("-")) {
        val qualificationLength = itemPattern.lastIndexOf('-') + 1  // Qualification of "font-family" is "font-"
        additionalOffset = truncatedRelativeOffset + qualificationLength
      }
    }

    return additionalOffset
  }

  private fun calculateRelativeOffset(): Int {
    // Note: find a caret which is the closest to the identifier
    val relativeOffsets = editor.caretModel.allCarets.asSequence().mapNotNull { caret ->
      //if selection take the selection start, without selection - caret offset
      val completionOffset = CompletionInitializationContext.calcStartOffset(caret)
      (completionOffset - identifierStart).takeIf { it >= 0 }
    }
    return relativeOffsets.minOrNull() ?: 0
  }

  private fun List<LookupElement>.registerNewElements(): List<CompletionItem> {
    //take only new elements, dont send the same items twice
    val newElements = this.filter {
      //plugins and all millions of LookupElement creators can provide LEs without hashCode/Equals
      val identity = extractIdentity(it)
      val isNew = !sentItems.contains(identity)
      if (isNew)
        sentItems.add(identity)

      return@filter isNew
    }

    val pairs = newElements.attachProtocolItems(arranger)

    pairs.forEach {
      lookupElementMap[it.first] = it.second
    }

    return pairs.map { it.first }
  }

  companion object {
    val LOG = Logger.getInstance(CompletionSession::class.java)
  }

  private fun calcDefaultIdentifierStart(editor: Editor, offset: Int): Int {
    val text = editor.document.charsSequence
    var identifierStartOffset = max(offset - 1, 0)

    while (identifierStartOffset > 0 && Character.isJavaIdentifierPart(text[identifierStartOffset])) {
      identifierStartOffset--
    }

    return min(offset, identifierStartOffset + 1)
  }
}

data class LookupElementIdentity(val elementClassName : String, val lookupString : String)

