package com.jetbrains.rdserver.completion

import com.intellij.codeInsight.completion.*
import com.intellij.codeInsight.lookup.Lookup
import com.intellij.codeInsight.lookup.LookupElement
import com.intellij.openapi.Disposable
import com.intellij.openapi.editor.Caret
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.event.DocumentListener
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Pair
import com.intellij.patterns.ElementPattern
import com.jetbrains.rd.platform.util.getLogger
import java.util.*
import java.util.concurrent.ConcurrentHashMap
import java.util.function.Supplier
import javax.swing.Icon

class BackendCompletionProcess(
    private val completionSession: CompletionSession,
    private val myEditor: Editor,
    private val myHostOffsets: OffsetsInFile,
    private val myOffsetMap: OffsetMap,
    private val isAutopopup: Boolean) : CompletionProcessEx {

  init {
    completionSession.editor.document.addDocumentListener(object : DocumentListener {
      override fun documentChanged(event: DocumentEvent) {
        //naive approach
        if (event.newFragment.all { it.isLetterOrDigit() }) {
          prefixUpdated()
        }
      }
    }, completionSession.sessionLifetime.createNestedDisposable())
  }

  private val logger = getLogger<BackendCompletionProcess>()
  private val restartingPrefixConditions = Collections.newSetFromMap<Pair<Int, ElementPattern<String>>>(ConcurrentHashMap())

  private var myParameters: CompletionParameters? = null

  override fun isAutopopupCompletion(): Boolean = isAutopopup

  override fun getProject(): Project = myEditor.project ?: error("cannot get project from an editor $myEditor")

  override fun getEditor(): Editor = myEditor

  override fun getCaret(): Caret = myEditor.caretModel.currentCaret

  override fun getOffsetMap(): OffsetMap = myOffsetMap

  override fun getHostOffsets(): OffsetsInFile = myHostOffsets

  override fun getLookup(): Lookup = throw UnsupportedOperationException()

  override fun registerChildDisposable(child: Supplier<out Disposable>) {
  }

  override fun itemSelected(item: LookupElement, completionChar: Char) = throw UnsupportedOperationException()

  override fun addWatchedPrefix(startOffset: Int, restartCondition: ElementPattern<String>?) {
    restartingPrefixConditions.add(Pair.create(startOffset, restartCondition))
  }

  override fun addAdvertisement(message: String, icon: Icon?) {
  }

  override fun getParameters(): CompletionParameters? = myParameters

  override fun setParameters(parameters: CompletionParameters) {
    myParameters = parameters
  }

  override fun scheduleRestart() {
    logger.info("schedule Restart api. Rescheduling completion...")

    completionSession.runOrReschedule(true)
  }

  override fun prefixUpdated() {
    logger.info("prefix updated. Check if we need to restart the completion")
    if (CompletionProgressIndicator.shouldRestartCompletion(completionSession.editor, restartingPrefixConditions, "")) {
      restartingPrefixConditions.clear()
      completionSession.runOrReschedule(true)
    }
  }
}