package com.jetbrains.rdserver.completion

import com.intellij.codeInsight.completion.impl.CamelHumpMatcher
import com.intellij.codeInsight.lookup.*
import com.intellij.codeInsight.lookup.impl.LookupImpl
import com.intellij.codeInsight.lookup.impl.LookupManagerImpl
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.Disposable
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.openapi.util.Disposer
import com.intellij.ui.HintHint
import com.jetbrains.rd.ide.document.ProtocolDocumentSynchronizer
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.editors.BackendEditorHost.Companion.synchronizer
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.requests.BackendIdeGateway
import javax.swing.JComponent

internal class BackendLookupManagerImpl(private val project: Project) : LookupManagerImpl(project) {
  companion object {
    private fun FocusDegreeModel.toFocusDegree(): LookupFocusDegree = when (this) {
      FocusDegreeModel.FOCUSED -> LookupFocusDegree.FOCUSED
      FocusDegreeModel.SEMI_FOCUSED -> LookupFocusDegree.SEMI_FOCUSED
      FocusDegreeModel.UNFOCUSED -> LookupFocusDegree.UNFOCUSED
    }

    private fun LookupFocusDegree.toProtocol(): FocusDegreeModel = when (this) {
      LookupFocusDegree.FOCUSED -> FocusDegreeModel.FOCUSED
      LookupFocusDegree.SEMI_FOCUSED -> FocusDegreeModel.SEMI_FOCUSED
      LookupFocusDegree.UNFOCUSED -> FocusDegreeModel.UNFOCUSED
    }
  }

  private val clientId2Lookups = mutableMapOf<ClientId, LookupImpl>()

  // TemplateState.runLookup invokes showLookup(editor: items: prefix: arranger:)

  override fun getActiveLookup(): LookupEx? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getActiveLookup()
    }
    return clientId2Lookups[ClientId.current]
  }

  override fun hideActiveLookup() {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.hideActiveLookup()
    }
    clientId2Lookups[ClientId.current]?.hide()
  }

  override fun createLookup(editor: Editor, items: Array<out LookupElement>?, prefix: String, arranger: LookupArranger): LookupImpl {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createLookup(editor, items, prefix, arranger)

    // Note: using the same arranger as on a client side
    val lookup = object : LookupImpl(editor.project, editor, LookupArranger.DefaultArranger()) {
      override fun doShowLookup(): Boolean {
        return true
      }

      override fun show(parentComponent: JComponent, x: Int, y: Int, focusBackComponent: JComponent?, hintHint: HintHint) {
      }

      override fun isAvailableToUser(): Boolean {
        return myShown
      }

      override fun suppressHidingOnChange(): Boolean {
        return true
      }
    }
    if (items != null && items.isNotEmpty()) {
      val matcher = CamelHumpMatcher(prefix)
      for (item in items) {
        lookup.addItem(item, matcher)
      }
    }
    hideActiveLookup()
    val clientId = ClientId.current
    clientId2Lookups[clientId] = lookup
    addLookupModel(lookup, editor, items)
    Disposer.register(lookup, Disposable {
      clientId2Lookups.remove(clientId)
      ClientId.withClientId(clientId) {
        removeLookupModel(editor)
      }
    })
    return lookup
  }

  private fun addLookupModel(lookup: LookupImpl, editor: Editor, items: Array<out LookupElement>?) {
    val editorId = editor.getTextControlId() ?: return
    val textControlModel = tryGetTextControlModel(editorId) ?: return
    val lookupModel = createLookupModel(lookup, items)
    textControlModel.lookupModel.customLookup.apply {
      set(lookupModel)
      advise(lookup.defineNestedLifetime()) { model ->
        // Note: a server-side lookup is almost eternal due to suppression
        // so it has to be hidden manually after its
        // client-side counterpart disappears
        if (model == null) {
          lookup.hide()
        }
      }
    }
    sendLookupShowStep(editor, editorId)
  }

  private fun createLookupModel(lookup: LookupImpl, items: Array<out LookupElement>?): CustomLookupModel {
    val lookupLifetime = lookup.defineNestedLifetime()
    return CustomLookupModel().also { lookupModel ->
      lookup.addLookupListener(object : LookupListener {
        override fun focusDegreeChanged() {
          lookupModel.focusDegree.set(lookup.lookupFocusDegree.toProtocol())
        }
      })
      lookupModel.selectedIndex.adviseNotNull(lookupLifetime) { index ->
        lookup.selectedIndex = index
      }
      lookupModel.focusDegree.adviseNotNull(lookupLifetime) { degree ->
        val newFocusDegree = degree.toFocusDegree()
        if (lookup.lookupFocusDegree != newFocusDegree) {  // Note: check in order to prevent recursive model updates
          lookup.lookupFocusDegree = newFocusDegree
        }
      }
      if (items != null) {
        val completionItems = items.asList().attachProtocolItems().map { it.first }
        lookupModel.items.value = CustomLookupItemsModel(completionItems, completionItems.size, selectedIndex = 0)
      }
    }
  }

  private fun sendLookupShowStep(editor: Editor, editorId: TextControlId) {
    val step = RdLookupShowStep(editorId)
    val editorVersion = editor.synchronizer?.version ?: return
    val documentVersion = tryGetDocumentSynchronizer(editorId)?.version ?: return
    BackendIdeGateway.getInstance(project).sendEditorChange(step, editorVersion, editorVersion, documentVersion)
  }

  private fun removeLookupModel(editor: Editor) {
    val editorId = editor.getTextControlId() ?: return
    val model = tryGetTextControlModel(editorId) ?: return
    model.lookupModel.customLookup.set(null)
  }

  private fun tryGetTextControlModel(editorId: TextControlId): TextControlModel? {
    return tryGetDocumentSynchronizer(editorId)?.modelDocument?.textControls?.get(editorId)
  }

  private fun tryGetDocumentSynchronizer(editorId: TextControlId): ProtocolDocumentSynchronizer? {
    return BackendDocumentHost.getInstance(project).tryGetSynchronizer(editorId.documentId)
  }
}