package com.jetbrains.rdserver.completion

import com.intellij.codeInsight.AutoPopupController
import com.intellij.codeInsight.AutoPopupControllerImpl
import com.intellij.codeInsight.completion.CompletionType
import com.intellij.codeWithMe.ClientId
import com.intellij.ide.PowerSaveMode
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Condition
import com.intellij.psi.PsiDocumentManager
import com.intellij.psi.PsiFile
import com.intellij.psi.impl.source.tree.injected.InjectedLanguageEditorUtil
import com.intellij.psi.impl.source.tree.injected.InjectedLanguageUtil
import com.intellij.util.codeInsight.CommentUtilCore
import com.jetbrains.rd.ide.Suppressor
import com.jetbrains.rd.ide.model.AutoPopupResult
import com.jetbrains.rd.ide.model.CompletionMode
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.editors.*
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.getTextControlId

class RdServerAutoPopupController(private val project: Project) : AutoPopupControllerImpl(project) {
  private val suppressLifetimes = SequentialLifetimes(project.lifetime)

  val suppressor = Suppressor()

  override fun scheduleAutoPopup(editor: Editor, completionType: CompletionType, condition: Condition<in PsiFile>?) {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.scheduleAutoPopup(editor, completionType, condition)

    val topLevelEditor = InjectedLanguageEditorUtil.getTopLevelEditor(editor)

    val textControlId = topLevelEditor.getTextControlId() ?: return
    val completionModel = topLevelEditor.completionModel ?: return

    if (PowerSaveMode.isEnabled()) {
      return
    }

    val dataContext = (editor as? EditorImpl)?.dataContext
    if (dataContext != null) {
      val psiFile = CommonDataKeys.PSI_FILE.getData(dataContext)
      val element = psiFile?.findElementAt(editor.caretModel.primaryCaret.offset)?.prevSibling
      if (element != null && CommentUtilCore.isComment(element)) {
        return //do not show in comments
      }
    }

    if (suppressor.shouldSuppress()) {
      LOG.trace { "Autopopup request was suppressed." }
      suppressor.onRelease(suppressLifetimes.next()) {
        try {
          requestAutoPopup(textControlId, completionModel, editor, condition)
        }
        finally {
          suppressLifetimes.terminateCurrent()
        }
      }
      return
    }

    requestAutoPopup(textControlId, completionModel, editor, condition)
  }

  private fun runIfConditionSatisfied(editor: Editor, condition: Condition<in PsiFile>?, task: () -> Unit) {
    if (condition != null) {
      val topLevelEditor = InjectedLanguageEditorUtil.getTopLevelEditor(editor)
      val offset = topLevelEditor.caretModel.offset
      val documentManager = PsiDocumentManager.getInstance(project)
      documentManager.performForCommittedDocument(topLevelEditor.document) {
        val topLevelFile = documentManager.getPsiFile(topLevelEditor.document)
        val completionEditor = InjectedLanguageUtil.getEditorForInjectedLanguageNoCommit(topLevelEditor, topLevelFile, offset)
        val file = documentManager.getPsiFile(completionEditor.document)
        if (file != null && condition.value(file)) {
          task()
        }
      }
    }
    else {
      task()
    }
  }

  private fun requestAutoPopup(editorId: TextControlId,
                               completionModel: CompletionModel,
                               editor: Editor,
                               condition: Condition<in PsiFile>?) {
    runIfConditionSatisfied(editor, condition) {
      requestAutoPopup(editorId, completionModel)
    }
  }

  private fun requestAutoPopup(editorId: TextControlId, completionModel: CompletionModel) {
    LOG.trace { "Sending autopopup request to frontend" }
    val documentHost = BackendDocumentHost.getInstance(project)
    val documentSynchronizer = documentHost.tryGetSynchronizer(editorId.documentId) ?: return

    val editorHost = BackendEditorHost.getInstance(project)
    val editorSynchronizer = editorHost.tryGetSynchronizer(editorId) ?: return
    val documentVersion = documentSynchronizer.version
    val editorVersion = editorSynchronizer.version
    val result = AutoPopupResult(shouldBeShown = true, typingTimeStamp = null, forceSoft = true, mode = CompletionMode.Basic,
                                 documentVersion = documentVersion, editorVersion = editorVersion)
    completionModel.autoPopupResult.fire(result)
  }

  companion object {
    val LOG = Logger.getInstance(RdServerAutoPopupController::class.java)
    fun getInstance(project: Project) = project.service<AutoPopupController>() as RdServerAutoPopupController
  }
}
