package com.jetbrains.rdserver.connection

import com.intellij.codeWithMe.ui.AccessNotification
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionState
import com.jetbrains.rd.platform.codeWithMe.protocol.SecureWireOverStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.SocketStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.WireConnectStatus
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.delayFuture
import com.jetbrains.rd.platform.util.launchLongBackground
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.startup.RdServerConnectionManager
import com.jetbrains.rdserver.startup.SecureClientData
import com.jetbrains.rdserver.unattendedHost.UnattendedHostManager
import kotlinx.coroutines.future.await
import java.time.Duration
import java.util.concurrent.CompletableFuture
import java.util.concurrent.atomic.AtomicBoolean

class CwmClientConnectionGroup(val project: Project, val data: SecureClientData, private val canRequestPassword: Boolean) {
  val id: String = data.userName
  val groupKey: String = data.clientFingerprint

  companion object {
    private val logger = logger<CwmClientConnectionGroup>()
  }

  enum class AcceptState {
    Accepted,
    Declined,
    Password,
    GuestLimit
  }

  private val lifetimeDefinition = LifetimeDefinition()
  private val userAcceptFuture = CompletableFuture<AcceptState>()
  private val selectedWireFuture = CompletableFuture<Pair<SecureWireOverStreamTransport?, AcceptState>>()
  private val wires = ArrayList<SecureWireOverStreamTransport>()
  private val askedForConfirmation = AtomicBoolean()
  private var wireToUse: SecureWireOverStreamTransport? = null

  fun terminateLifetime() {
    lifetimeDefinition.terminate()
  }

  init {
    userAcceptFuture.thenAccept {
      when(it) {
        AcceptState.Declined, AcceptState.GuestLimit -> selectedWireFuture.complete(null to it)
        else -> if (wireToUse == null)
          delayFuture(Duration.ofSeconds(2)).thenAccept { _ -> selectSingleWire(it) }
        else
          selectSingleWire(it)
      }
    }
    lifetimeDefinition.onTermination {
      logger.debug("Wire group $id lifetime termination")
      application.invokeLater { // delayed to make ConcurrentHashMap behave better
        RdServerConnectionManager.getInstance(project).removeGroup(this)
      }
      try {
        wireToUse?.streamTransport?.byteChannel?.let {
          if (it.isOpen) it.close()
        }
      } catch(t: Throwable) {
        logger.error("Exception while closing wire byte channel from wire group $id", t)
      }
    }
    selectedWireFuture.handle { (wire, _), _ ->
      wire?.connectionState?.advise(lifetimeDefinition.lifetime) {
        if (it != ConnectionState.CONNECTED && it != ConnectionState.NOT_CONNECTED)
          lifetimeDefinition.terminate()
      } ?: lifetimeDefinition.terminate()
    }
  }

  private fun selectSingleWire(state: AcceptState) {
    if (selectedWireFuture.isDone) return

    val selectedWire = wireToUse ?: wires.first()

    logger.debug("Selected wire ${selectedWire.id} for group for $id")
    selectedWireFuture.complete(selectedWire to state)
  }

  private fun getConfirmationFuture(data: SecureClientData) : CompletableFuture<AcceptState> {
    if (UnattendedHostManager.getInstance().isUnattendedMode) {
      val accepted = UnattendedHostManager.getInstance().shouldAutomaticallyAcceptGuest(project, data)
      logger.info("Host in unattended mode, auto-accept check for connection group $id: $accepted")
      return accepted
    }

    val fingerprintString = RdServerConnectionManager.getInstance(project).getPinString(data.clientFingerprint)

    return AccessNotification.showNewAccessNotification(
      project, Lifetime.Eternal, // todo: use proper lifetime (union of all wire lifetimes?)
      data.userName, data.clientId, fingerprintString).thenApply { if (it) AcceptState.Accepted else AcceptState.Declined }
  }

  private fun askForConfirmation(data: SecureClientData) {
    if (!askedForConfirmation.compareAndSet(false, true)) return

    project.lifetime.launchLongBackground {
      logger.debug("Asking for confirmation for client group $id")
      if (!project.service<BackendGuestLimitNotificator>().isAtConcurrentGuestLimit && RdServerConnectionManager.getInstance(project).shouldAutoAccept(data)) {
        logger.info("Auto accepting client group $id for ${data.userName}")
        userAcceptFuture.complete(AcceptState.Accepted)
        return@launchLongBackground
      }

      val acceptFuture = getConfirmationFuture(data).await()
      if (project.service<BackendGuestLimitNotificator>().isAtConcurrentGuestLimit && (acceptFuture == AcceptState.Declined)) {
        logger.debug("Group $id was decided with ${AcceptState.GuestLimit}")
        userAcceptFuture.complete(AcceptState.GuestLimit)
      }
      else {
        logger.debug("Group $id was decided with $acceptFuture")
        userAcceptFuture.complete(acceptFuture)
      }
    }
  }

  private fun shouldSelectImmediately(wire: SecureWireOverStreamTransport): Boolean {
    return wire.streamTransport.let { it is SocketStreamTransport || it is QuicStreamTransport }
  }

  fun registerWire(wire: SecureWireOverStreamTransport, data: SecureClientData): CompletableFuture<WireConnectStatus> {
    if (shouldSelectImmediately(wire) && wireToUse == null) {
      logger.debug("Fast-track selecting wire ${wire.id} because of its priority")
      wireToUse = wire // still need to wait for user accept, can't return immediately
      if (userAcceptFuture.isDone)
        selectSingleWire(userAcceptFuture.get())
    }

    synchronized(wires) {
      wires.add(wire)
    }
    logger.debug("Registered wire ${wire.id} for client group $id")
    askForConfirmation(data)
    return selectedWireFuture.handle { (it, state), _ ->
        val isSelectedWire = it === wire || it == null
        if (!isSelectedWire) return@handle WireConnectStatus.DECLINED

        return@handle when (state) {
          AcceptState.Accepted -> WireConnectStatus.ACCEPTED
          AcceptState.Password -> WireConnectStatus.PASSWORD_REQUIRED
          AcceptState.GuestLimit -> WireConnectStatus.GUEST_LIMIT_REACHED
          AcceptState.Declined -> WireConnectStatus.USER_DECLINED
        }
    }
  }
}