// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
@file:Suppress("NonDefaultConstructor")

package com.jetbrains.rdserver.daemon

import com.intellij.codeInsight.daemon.DaemonCodeAnalyzer
import com.intellij.codeInsight.daemon.impl.DaemonCodeAnalyzerImpl
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Document
import com.intellij.openapi.fileEditor.FileEditor
import com.intellij.openapi.fileEditor.FileEditorManager
import com.intellij.openapi.fileEditor.TextEditor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.psi.PsiDocumentManager
import com.jetbrains.rd.ide.document.toDebugString
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.document.BackendDocumentHostListener

class BackendDaemonStatusHost(
  lifetime: Lifetime,
  clientId: ClientId,
  val project: Project
) {

  companion object {
    private val LOG = logger<BackendDaemonStatusHost>()
  }

  init {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(DaemonCodeAnalyzer.DAEMON_EVENT_TOPIC, object : DaemonCodeAnalyzer.DaemonListener {
      override fun daemonStarting(files: Collection<FileEditor>) {
        ClientId.withClientId(clientId) {
          foreachDaemonModel(files) {
            it.state.set(BackendDaemonStateEnum.INVALIDATED)
            it.highlightingProgressState.set(BackendHighlightingProgressStateEnum.LOCAL_STAGES)
          }
        }
      }

      override fun daemonFinished(files: Collection<FileEditor>) {
        ClientId.withClientId(clientId) {
          foreachDaemonModel(files) {
            it.state.set(BackendDaemonStateEnum.UP_TO_DATE)
            it.highlightingProgressState.set(BackendHighlightingProgressStateEnum.OFF)
          }
        }
      }
    })

    connection.subscribe(BackendDocumentHostListener.TOPIC, object : BackendDocumentHostListener {
      override fun documentBound(
        lifetime: Lifetime,
        documentId: RdDocumentId,
        model: RdDocumentModel,
        document: Document
      ) {
        if (ClientId.current != clientId) return

        model.markup.advise(lifetime) { protocolMarkupModel ->
          if (protocolMarkupModel !is RdMarkupModel) {
            LOG.info("No frontend markup model for ${documentId.toDebugString()}")
            return@advise
          }

          protocolMarkupModel.daemon.restart.advise(lifetime) {
            restartDaemon(document)
          }
        }
      }
    })
  }

  private fun foreachDaemonModel(
    fileEditors: Collection<FileEditor>,
    action: (RdDaemonModel) -> Unit
  ) {
    val documentHost = BackendDocumentHost.getInstance(project)
    for (fileEditor in fileEditors) {
      val editor = (fileEditor as? TextEditor)?.editor ?: continue
      val documentModel = documentHost.tryGetSynchronizer(editor.document)?.modelDocument ?: continue
      val documentMarkup = documentModel.markup.valueOrNull as? RdMarkupModel ?: continue
      action(documentMarkup.daemon)
      for (textControl in documentModel.textControls.values) {
        val textControlMarkup = textControl.markup.valueOrNull as? RdMarkupModel ?: continue
        action(textControlMarkup.daemon)
      }
    }
  }

  private fun restartDaemon(document: Document) {
    val psiFile = PsiDocumentManager.getInstance(project).getPsiFile(document) ?: return
    val daemonCodeAnalyzer = DaemonCodeAnalyzer.getInstance(project) as DaemonCodeAnalyzerImpl
    if (application.isUnitTestMode) {
      PsiDocumentManager.getInstance(project).commitDocument(document)
      val textEditor = FileEditorManager.getInstance(project).getSelectedEditor(psiFile.virtualFile) as TextEditor
      daemonCodeAnalyzer.runPasses(psiFile, document, listOf(textEditor), intArrayOf(), false, null)
    }
    else {
      daemonCodeAnalyzer.restart(psiFile)
    }
  }
}
