package com.jetbrains.rdserver.daemon

import com.intellij.application.options.colors.ColorSettingsUtil
import com.intellij.application.options.colors.highlighting.HighlightData
import com.intellij.application.options.colors.highlighting.HighlightsExtractor
import com.intellij.codeHighlighting.RainbowHighlighter
import com.intellij.codeInsight.daemon.impl.HighlightInfoType
import com.intellij.codeInsight.daemon.impl.SeverityRegistrar
import com.intellij.ide.highlighter.HighlighterFactory
import com.intellij.lang.annotation.HighlightSeverity
import com.intellij.openapi.editor.colors.EditorColorsManager
import com.intellij.openapi.editor.colors.TextAttributesKey
import com.intellij.openapi.editor.colors.impl.AbstractColorsScheme
import com.intellij.openapi.editor.colors.impl.EditorColorsManagerImpl
import com.intellij.openapi.editor.ex.util.LexerEditorHighlighter
import com.intellij.openapi.editor.markup.EffectType
import com.intellij.openapi.editor.markup.RangeHighlighter
import com.intellij.openapi.editor.markup.TextAttributes
import com.intellij.openapi.options.Scheme
import com.intellij.openapi.options.colors.ColorSettingsPage
import com.intellij.openapi.util.Segment
import com.intellij.psi.codeStyle.DisplayPriority
import com.intellij.psi.codeStyle.DisplayPrioritySortable
import com.jetbrains.ide.model.highlighterRegistration.*
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.ide.editor.markup.HighlighterTextAnnotator
import com.jetbrains.rd.ide.editor.markup.explicitAttributeId
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.daemon.syntax.LexerBasedSyntaxIterator
import com.jetbrains.rdserver.toRgbColor

@Suppress("NonDefaultConstructor")
class BackendHighlighterRegistrationsHost(
  lifetime: Lifetime,
  protocol: IProtocol
) {
  private val model = HighlighterRegistrationHostModel.create(lifetime, protocol)

  init {
    LOG.info("Preparing highlighter registrations")
    model.registry.view(lifetime) { _, registry ->
      if (registry !is DefaultRegistryModel) return@view

      // to force attributesKey classes to load
      ColorSettingsPage.EP_NAME.extensions.forEach { it.attributeDescriptors }

      val packs = getHighlighterRegistrationPacks()
      registry.registrationPacks.fire(packs.values.toList())
      registry.severities.fire(getSeverities())
      // todo can be limited to AdditionalTextAttributesEP
      //  ?and sending of bundled schemes? (not yet clear if we want to pollute guests schemes list)
      val manager = (EditorColorsManager.getInstance() as EditorColorsManagerImpl)
      for (scheme in manager.schemeManager.allSchemes) {
        if (scheme !is AbstractColorsScheme) continue
        if (scheme.name.startsWith(Scheme.EDITABLE_COPY_PREFIX)) continue
        registry.colorsScheme.fire(createSchemeModel(scheme))
      }

      ColorSettingsPage.EP_NAME.forEachExtensionSafe { page ->
        registry.colorSettingsPage.fire(createPageModel(page))
      }
    }
    model.backendReady.set(true)
    LOG.info("Submitted highlighter registrations")
  }

  private fun createSchemeModel(scheme: AbstractColorsScheme): ColorsSchemeModel {
    val entries = scheme.directlyDefinedAttributes.mapNotNull {
      val attributeId = getAttributeId(TextAttributesKey.find(it.key)) ?: return@mapNotNull null
      val attributes = it.value ?: return@mapNotNull null
      val isInherited = attributes === AbstractColorsScheme.INHERITED_ATTRS_MARKER
      ColorsSchemeEntry(attributeId, isInherited, convertAttributes(attributes))
    }
    return ColorsSchemeModel(scheme.name, entries)
  }

  private fun createPageModel(page: ColorSettingsPage): ColorSettingsPageModel {
    fun Map<String, TextAttributesKey>?.toModel(): List<HighlighterData> {
      return this?.mapNotNull { (key, value) -> getAttributeId(value)?.let { HighlighterData(key, it) } } ?: emptyList()
    }

    val priority = page.getPriority()
    val demoText = getDemoText(page)
    val highlighterTags = page.additionalHighlightingTagToDescriptorMap.toModel()
    val inlineElementsTags = page.additionalInlineElementToDescriptorMap.toModel()
    val presentableNames = page.attributeDescriptors.map { it.displayName to it.key }.toMap().toModel()
    // todo color keys
    return ColorSettingsPageModel(
      page.displayName,
      priority,
      demoText,
      highlighterTags,
      inlineElementsTags,
      presentableNames
    )
  }

  private fun getDemoText(page: ColorSettingsPage): String {
    val scheme = EditorColorsManager.getInstance().globalScheme // just to pass it
    val highlighter = HighlighterFactory.createHighlighter(page.highlighter, scheme) as? LexerEditorHighlighter
    if (highlighter == null) return page.demoText

    val highlightExtractor = HighlightsExtractor(page.additionalHighlightingTagToDescriptorMap,
                                                 page.additionalInlineElementToDescriptorMap,
                                                 page.additionalHighlightingTagToColorKeyMap)
    val data = mutableListOf<HighlightData>()
    val extractedText = highlightExtractor.extractHighlights(page.demoText, data)
    val highlightData = data.map { MyHighlightData(it.startOffset, it.endOffset, it.highlightKey) }.toMutableList()
    highlighter.setText(extractedText)

    val iterator = LexerBasedSyntaxIterator(highlighter, 0, page.demoText.length)
    try {
      while (!iterator.atEnd()) {
        // todo pass all keys, can't be done currently due to compatibility with Rider
        val attributesKey = iterator.textAttributesKeys?.lastOrNull()
        if (attributesKey != null) {
          highlightData.add(MyHighlightData(iterator.start, iterator.end, attributesKey))
        }
        iterator.advance()
      }
      highlightData.sortBy { it.startOffset }
    }
    finally {
      iterator.dispose()
    }

    return HighlighterTextAnnotator.annotateWithTags(
      extractedText, highlightData, compareBy(MyHighlightData::getStartOffset), { getAttributeId(it.highlightKey) }, { null }
    )
  }

  private class MyHighlightData(start: Int, end: Int, key: TextAttributesKey) : HighlightData(start, end, key), Segment

  private fun getSeverities(): List<HighlightSeverityModel> {
    val allSeverities = SeverityRegistrar.getSeverityRegistrar(null).allSeverities
    return allSeverities
      .map { HighlightSeverityModel(it.name, it.myVal) }
  }

  private fun getHighlighterRegistrationPacks(): MutableMap<String, HighlighterRegistrationPack> {
    val result = mutableMapOf<String, HighlighterRegistrationPack>()

    val severityRegistrar = SeverityRegistrar.getSeverityRegistrar(null)
    for (severity in severityRegistrar.allSeverities) {
      result.addPack(severityRegistrar.getHighlightInfoTypeBySeverity(severity))
    }
    result.addPack(HighlightInfoType.ELEMENT_UNDER_CARET_READ)
    result.addPack(HighlightInfoType.ELEMENT_UNDER_CARET_WRITE)

    RainbowHighlighter.createRainbowHLM().forEach {
      result.addPack(it.value, HighlightSeverity.INFORMATION)
    }

    for (colorSettingsPage in ColorSettingsPage.EP_NAME.extensions) {
      for (descriptor in ColorSettingsUtil.getAllAttributeDescriptors(colorSettingsPage)) {
        result.addPack(descriptor.key, HighlightSeverity.INFORMATION)
      }
    }

    for (key in TextAttributesKey.getAllKeys()) {
      result.addPack(key, HighlightSeverity.INFORMATION)
    }

    result.addDummyPack(GUTTER_MARK_ID)
    result.addDummyPack(LINE_MARKER_ID)
    result.addDummyPack(FOLD_REGION_ID, false, false)
    result.addDummyPack(INLAY_ID, false, true)
    result.addDummyPack(VCS_MARK_ID, false, false)
    result.addDummyPack(CUSTOM_HIGHLIGHTING_ID)
    result.addDummyPack(DIFF_MARK_ID, false, false)

    HighlightInfoType.VISIBLE_IF_FOLDED.forEach {
      val pack = result[getAttributeId(it.attributesKey)]!!
      val newPack = pack.copy(properties = pack.properties.copy(isVisibleIfFolded = true))
      result[getAttributeId(it.attributesKey)!!] = newPack
    }

    return result
  }

  private fun MutableMap<String, HighlighterRegistrationPack>.addDummyPack(attributeId: String,
                                                                           isBulkUpdatable: Boolean = false,
                                                                           isVerifiable: Boolean = true) {
    _allRegisteredAttributeIds.add(attributeId)
    val properties = HighlighterProperties(HighlighterKind.USUAL, false, GreedySide.NONE, isVerifiable, isBulkUpdatable, false)
    this[attributeId] = HighlighterRegistrationPack(attributeId, attributeId, null, properties)
  }

  private fun MutableMap<String, HighlighterRegistrationPack>.addPack(infoType: HighlightInfoType) {
    infoType.let { addPack(it.attributesKey, it.getSeverity(null)) }
  }

  private fun MutableMap<String, HighlighterRegistrationPack>.addPack(textAttributesKey: TextAttributesKey,
                                                                      severity: HighlightSeverity) {
    var key = textAttributesKey
    while (!containsKey(getAttributeId(key))) {
      val attributeId = createAttributeId(key)
      _allRegisteredAttributeIds.add(attributeId)
      val properties = HighlighterProperties(HighlighterKind.USUAL, true, GreedySide.NONE, true, false, false)
      val infoType = IdeaHighlightInfoType(severity.name, IdeaTextAttributesKey(key.externalName, key.fallbackAttributeKey?.externalName))
      this[attributeId] = HighlighterRegistrationPack(attributeId, attributeId, infoType, properties)
      key = key.fallbackAttributeKey ?: break
    }
  }

  companion object {
    val LOG = getLogger<BackendHighlighterRegistrationsHost>()

    fun getAttributeId(highlighter: RangeHighlighter): String {
      return highlighter.textAttributesKey?.let { getAttributeId(it) } ?: highlighter.explicitAttributeId
             ?: error("No attributeId for $highlighter, key ${highlighter.textAttributesKey}")
    }

    private fun createAttributeId(key: TextAttributesKey): String {
      return "IJ." + key.externalName
    }

    fun getAttributeId(key: TextAttributesKey): String? {
      val element = createAttributeId(key)
      return if (allRegisteredAttributeIds.contains(element)) element else null
    }

    private val _allRegisteredAttributeIds = mutableSetOf<String>()
    val allRegisteredAttributeIds: Set<String> get() = _allRegisteredAttributeIds

    const val GUTTER_MARK_ID = "IJ.GUTTER_MARK"
    const val LINE_MARKER_ID = "IJ.LINE_MARKER"
    const val FOLD_REGION_ID = "IJ.FOLD_REGION"
    const val INLAY_ID = "IJ.INLAY"
    const val CUSTOM_HIGHLIGHTING_ID = "IJ.CUSTOM_HIGHLIGHTING"
    const val VCS_MARK_ID = "IJ.VCS_MARK"
    const val DIFF_MARK_ID = "IJ.DIFF_MARK"

    fun convertAttributes(textAttributes: TextAttributes): IdeaTextAttributes {
      return IdeaTextAttributes(
        textAttributes.foregroundColor.toRgbColor(),
        textAttributes.backgroundColor.toRgbColor(),
        if (textAttributes.effectType != null) textAttributes.effectColor.toRgbColor() else null, // IdeaTextAttributes.effectType is not nullable
        textAttributes.errorStripeColor.toRgbColor(),
        textAttributes.effectType.toModel(),
        textAttributes.fontType
      )
    }
  }
}

private fun EffectType?.toModel(): IdeaEffectType {
  return when (this) {
    EffectType.LINE_UNDERSCORE -> IdeaEffectType.LINE_UNDERSCORE
    EffectType.WAVE_UNDERSCORE -> IdeaEffectType.WAVE_UNDERSCORE
    EffectType.BOXED -> IdeaEffectType.BOXED
    EffectType.STRIKEOUT -> IdeaEffectType.STRIKEOUT
    EffectType.BOLD_LINE_UNDERSCORE -> IdeaEffectType.BOLD_LINE_UNDERSCORE
    EffectType.BOLD_DOTTED_LINE -> IdeaEffectType.BOLD_DOTTED_LINE
    else -> IdeaEffectType.BOXED
  }
}

private fun ColorSettingsPage.getPriority(): SettingsPageDisplayPriority {
  if (this !is DisplayPrioritySortable) {
    return SettingsPageDisplayPriority.LANGUAGE_SETTINGS
  }
  return when (this.priority) {
    DisplayPriority.GENERAL_SETTINGS -> SettingsPageDisplayPriority.GENERAL_SETTINGS
    DisplayPriority.FONT_SETTINGS -> SettingsPageDisplayPriority.FONT_SETTINGS
    DisplayPriority.COMMON_SETTINGS -> SettingsPageDisplayPriority.COMMON_SETTINGS
    DisplayPriority.CODE_SETTINGS -> SettingsPageDisplayPriority.CODE_SETTINGS
    DisplayPriority.KEY_LANGUAGE_SETTINGS -> SettingsPageDisplayPriority.KEY_LANGUAGE_SETTINGS
    DisplayPriority.LANGUAGE_SETTINGS -> SettingsPageDisplayPriority.LANGUAGE_SETTINGS
    DisplayPriority.OTHER_SETTINGS -> SettingsPageDisplayPriority.OTHER_SETTINGS
    else -> SettingsPageDisplayPriority.LANGUAGE_SETTINGS
  }
}
