package com.jetbrains.rdserver.daemon

import com.intellij.codeInsight.daemon.DaemonCodeAnalyzer
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.ex.MarkupModelEx
import com.intellij.openapi.editor.impl.DocumentMarkupModel
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Key
import com.intellij.psi.PsiManager
import com.jetbrains.rd.ide.document.toDebugString
import com.jetbrains.rd.ide.model.RdDocumentId
import com.jetbrains.rd.ide.model.RdDocumentModel
import com.jetbrains.rd.ide.model.RdMarkupModel
import com.jetbrains.rd.ide.model.TextControlModel
import com.jetbrains.rd.platform.util.getOrCreateUserData
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.adviseOnce
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.document.BackendDocumentHostListener
import com.jetbrains.rdserver.editors.BackendEditorHostListener
import com.jetbrains.rdserver.editors.getTextControlId
import java.util.function.Predicate

class BackendMarkupHost(val project: Project) {
  companion object {
    fun getInstance(project: Project): BackendMarkupHost = project.service()

    private val LOG = logger<BackendMarkupHost>()
  }

  private val markupSuppressorFactory get() = RdServerPerClientComponents.getService<BackendMarkupSuppressors>(project)

  fun createMarkupContributor(
    lifetime: Lifetime,
    documentId: RdDocumentId,
    documentModel: RdDocumentModel,
    document: Document
  ) {
    documentModel.markup.advise(lifetime) { protocolMarkupModel ->
      if (protocolMarkupModel !is RdMarkupModel) {
        LOG.info("No frontend markup model for ${documentId.toDebugString()}")
        return@advise
      }

      val documentMarkupModel = DocumentMarkupModel.forDocument(document, project, true) as MarkupModelEx?

      if (documentMarkupModel == null) {
        LOG.info("No document markup model for ${documentId.toDebugString()}")
        return@advise
      }


      val documentSynchronizer = BackendDocumentHost.getInstance(project).tryGetSynchronizer(documentId)
      if (documentSynchronizer == null) {
        LOG.info("No document synchronizer for ${documentId.toDebugString()}")
        return@advise
      }
      val suppressor = markupSuppressorFactory.createSuppressor(lifetime)
      val contributor = BackendMarkupModelContributor(project, document, documentSynchronizer, suppressor, protocolMarkupModel)
      initMarkupExtensions(lifetime, protocolMarkupModel, contributor)

      contributor.bind(lifetime, documentMarkupModel)
      document.getOrCreateUserData(DOCUMENT_CONTRIBUTOR_KEY) { HashMap() }.addUnique(lifetime, ClientId.current, contributor)
    }
  }

  fun createMarkupContributor(
    lifetime: Lifetime,
    textControlModel: TextControlModel,
    editor: Editor
  ) {
    textControlModel.markup.adviseOnce(lifetime) {
      if (it !is RdMarkupModel) return@adviseOnce

      val documentId = editor.getTextControlId()?.documentId ?: return@adviseOnce
      val documentSynchronizer = BackendDocumentHost.getInstance(project).tryGetSynchronizer(documentId)
      if (documentSynchronizer == null) {
        LOG.info("No document synchronizer for ${documentId.toDebugString()}")
        return@adviseOnce
      }
      val document = documentSynchronizer.document
      val parentSuppressor = document.markupModelContributor!!.suppressor
      val suppressor = markupSuppressorFactory.createSuppressor(lifetime, "document suppressor", parentSuppressor)
      val contributor = BackendEditorMarkupModelContributor(lifetime, project, editor, documentSynchronizer, suppressor, it)
      initMarkupExtensions(lifetime, it, contributor)
      //addUnique doesn't work for multiple synchronizers case
      val userData = editor.getOrCreateUserData(EDITOR_CONTRIBUTOR_KEY) { HashMap() }
      val clientId = ClientId.current
      userData[clientId] = contributor
      lifetime.onTermination {
        userData.remove(clientId)
      }

      contributor.bind(lifetime, editor.markupModel as MarkupModelEx)
    }
  }

  fun initMarkupExtensions(lifetime: Lifetime, protocolMarkupModel: RdMarkupModel,
                                   modelContributor: BackendMarkupModelContributor) {
    BackendProtocolMarkupModelExtensionHandler.EP_NAME.findFirstSafe(project, Predicate {
      it.consume(lifetime, protocolMarkupModel.extensions, modelContributor)
      false
    })
  }

  class MyDocumentHostListener(private val project: Project) : BackendDocumentHostListener {
    override fun documentBound(
      lifetime: Lifetime,
      documentId: RdDocumentId,
      model: RdDocumentModel,
      document: Document
    ) {
      getInstance(project).createMarkupContributor(lifetime, documentId, model, document)
    }
  }

  class MyEditorHostListener(private val project: Project) : BackendEditorHostListener {
    override fun editorBound(lifetime: Lifetime, editorModel: TextControlModel, editor: Editor) {
      getInstance(project).createMarkupContributor(lifetime, editorModel, editor)

      val virtualFile = FileDocumentManager.getInstance().getFile(editor.document) ?: return
      val psiFile = PsiManager.getInstance(project).findFile(virtualFile)
      if (psiFile != null) {
        DaemonCodeAnalyzer.getInstance(project).restart()
      }
    }
  }
}

private val DOCUMENT_CONTRIBUTOR_KEY = Key.create<HashMap<ClientId, BackendMarkupModelContributor>>("BackendMarkupContributor")
private val EDITOR_CONTRIBUTOR_KEY = Key.create<HashMap<ClientId, BackendEditorMarkupModelContributor>>("BackendEditorMarkupContributor")

val Document.markupModelContributor
  get() = getUserData(DOCUMENT_CONTRIBUTOR_KEY)?.get(ClientId.current)

val Editor.markupModelContributor
  get() = getUserData(EDITOR_CONTRIBUTOR_KEY)?.get(ClientId.current)

suspend inline fun Editor.suppressMarkupSynchronization(reason: String, crossinline block: suspend () -> Unit) {
  val markupModelContributor = this.document.markupModelContributor ?: error("markupModelContributor is null")
  Lifetime.using { lifetime ->
    markupModelContributor.suppressor.suppress(lifetime, reason)
    block()
  }
}

fun Editor.suppressMarkupSynchronization(lifetime: Lifetime, reason: String) {
  val markupModelContributor = this.document.markupModelContributor ?: error("markupModelContributor is null")
  markupModelContributor.suppressor.suppress(lifetime, reason)
}