package com.jetbrains.rdserver.daemon

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.RangeMarker
import com.intellij.openapi.editor.ex.MarkupModelEx
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.ex.RangeMarkerEx
import com.intellij.openapi.editor.impl.event.MarkupModelListener
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Key
import com.intellij.util.SmartList
import com.jetbrains.rd.ide.document.ProtocolDocumentSynchronizer
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rd.ide.model.*
import java.util.function.Predicate

open class BackendMarkupModelContributor(
  val project: Project,
  val document: Document,
  private val documentSynchronizer: ProtocolDocumentSynchronizer,
  val suppressor: BackendMarkupSuppressor,
  val protocolMarkupModel: RdMarkupModel
) {
  val clientId = ClientId.current
  val storage = BackendRangeMarkerStorage()

  init {
    suppressor.onRelease { flushRangeMarkers() }
  }

  fun addOrUpdateRangeMarkers(rangeMarkers: Iterable<RangeMarker>) = ClientId.withClientId(clientId) {
    if (suppressor.shouldSuppress()) {
      rangeMarkers.forEach {
        if (!storage.contains(it as RangeMarkerEx)) {
          storage.queueRegistration(it)
        }
        else {
          storage.queueUpdate(it)
        }
      }
      return@withClientId
    }

    val toAdd = mutableListOf<HighlighterModel>()
    val toUpdate = mutableListOf<HighlighterModel>()
    val documentVersion = currentDocumentVersion()
    for (rangeMarker in rangeMarkers) {
      if (!isSupported(rangeMarker as RangeMarkerEx)) continue
      val model = createModel(rangeMarker, documentVersion) ?: continue

      if (!storage.contains(rangeMarker)) {
        storage.add(rangeMarker)
        toAdd.add(model)
      }
      else {
        toUpdate.add(model)
      }
    }

    if (toAdd.isNotEmpty()) {
      protocolMarkupModel.highlightersAdded.fire(HighlighterList(toAdd))
    }
    if (toUpdate.isNotEmpty()) {
      protocolMarkupModel.highlightersUpdated.fire(HighlighterList(toUpdate))
    }
  }

  fun removeRangeMarkers(rangeMarkers: Iterable<RangeMarker>) = ClientId.withClientId(clientId) {
    if (suppressor.shouldSuppress()) {
      rangeMarkers.forEach {
        storage.queueDeletion(it as RangeMarkerEx)
      }
      return@withClientId
    }

    val toRemove = rangeMarkers.mapNotNull { if (storage.remove(it as RangeMarkerEx)) it.id else null }
    if (toRemove.isEmpty()) return@withClientId
    protocolMarkupModel.highlightersRemoved.fire(HighlighterIdList(toRemove))
  }

  private fun flushRangeMarkers() {
    addOrUpdateRangeMarkers(storage.flushPendingToAdd())
    addOrUpdateRangeMarkers(storage.flushPendingToUpdate())
    removeRangeMarkers(storage.flushPendingToRemove())
  }

  fun recoverMarkupState() {
    val allRangeMarkers = storage.getAllRegistered()
    performBulkOperation(allRangeMarkers) {
      removeRangeMarkers(allRangeMarkers)
    }
    performBulkOperation(allRangeMarkers) {
      addOrUpdateRangeMarkers(allRangeMarkers)
    }
  }

  fun currentDocumentVersion(): RdDocumentVersion {
    return documentSynchronizer.version
  }

  protected open fun isSupported(rangeMarker: RangeMarkerEx): Boolean {
    return BackendRangeMarkerModelCreator.EP_NAME.findFirstSafe(project, Predicate { it.accept(rangeMarker) }) != null
  }

  open fun createModel(rangeMarker: RangeMarkerEx, documentVersion: RdDocumentVersion): HighlighterModel? {
    val creator = BackendRangeMarkerModelCreator.EP_NAME.findFirstSafe(project, Predicate { it.accept(rangeMarker) })
    return creator?.createModel(rangeMarker, documentVersion)
  }

  @Suppress("UNCHECKED_CAST")
  fun <T : MarkupModelExtension> getExtension(key: Key<T>): T? {
    return protocolMarkupModel.extensions.find { it.key == key.toString() } as? T
  }

  fun getRangeMarker(id: Long): RangeMarkerEx? {
    return storage.get(id)
  }

  companion object {
    val LOG = getLogger<BackendMarkupModelContributor>()
  }
}


internal const val CHUNK_SIZE = 200

fun <T> BackendMarkupModelContributor?.performInBulkIfPossible(
  elements: Iterable<T>,
  chuckSize: Int = CHUNK_SIZE,
  action: (T) -> Unit
) {
  if (this != null) {
    performBulkOperation(elements, chuckSize, action)
  }
  else {
    elements.forEach(action)
  }
}

fun <T> BackendMarkupModelContributor.performBulkOperation(
  elements: Iterable<T>,
  chunkSize: Int = CHUNK_SIZE,
  action: (T) -> Unit
) {
  assert(chunkSize > 0 || chunkSize == -1)
  Lifetime.using { lt ->
    val lifetimes = SequentialLifetimes(lt)

    if (chunkSize == -1) {
      suppressor.suppress(lifetimes.next(), "performBulkAddOperation")
      elements.forEach(action)
      return@using
    }

    for ((counter, element) in elements.withIndex()) {
      if (counter % chunkSize == 0) {
        suppressor.suppress(lifetimes.next(), "performBulkAddOperation")
      }
      action(element)
    }
  }
}


fun BackendMarkupModelContributor.bind(lifetime: Lifetime, markupModel: MarkupModelEx) {
  val listener = object : MarkupModelListener {
    override fun attributesChanged(
      highlighter: RangeHighlighterEx,
      renderersChanged: Boolean,
      fontStyleOrColorChanged: Boolean
    ) {
      addOrUpdateRangeMarkers(SmartList(highlighter))
    }

    override fun beforeRemoved(highlighter: RangeHighlighterEx) {
      removeRangeMarkers(SmartList(highlighter))
    }

    override fun afterAdded(highlighter: RangeHighlighterEx) {
      addOrUpdateRangeMarkers(SmartList(highlighter))
    }
  }

  markupModel.allHighlighters
    .asIterable()
    .chunked(CHUNK_SIZE)
    .forEach {
      addOrUpdateRangeMarkers(it)
    }

  markupModel.addMarkupModelListener(lifetime.createNestedDisposable(), listener)
}