package com.jetbrains.rdserver.daemon

import com.intellij.codeInsight.daemon.impl.EditorTracker
import com.intellij.codeInsight.daemon.impl.EditorTrackerListener
import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isLocal
import com.intellij.openapi.diagnostic.debug
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.fileEditor.FileEditorManagerEvent
import com.intellij.openapi.fileEditor.FileEditorManagerListener
import com.intellij.openapi.project.Project
import com.intellij.util.containers.MultiMap
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rdserver.editors.RemoteEditors
import com.jetbrains.rdserver.editors.clientId

class RdServerEditorTracker(private val project: Project) : EditorTracker(project) {

  companion object {
    private val LOG = logger<RdServerEditorTracker>()
  }

  private var currentEditors = emptyList<Editor>()
  private val remoteTracedEditors = MultiMap.createLinkedSet<ClientId, Editor>()

  init {
    val connection = project.messageBus.connect(this)
    connection.subscribe(
      FileEditorManagerListener.FILE_EDITOR_MANAGER,
      object : FileEditorManagerListener {
        override fun selectionChanged(event: FileEditorManagerEvent) {
          update()
        }
      }
    )
  }

  override fun createEditorImpl(editor: Editor) {
    if (ClientId.isCurrentlyUnderLocalId) {
      LOG.debug { "Add local editor. $editor" }
      super.createEditorImpl(editor)
      return
    }

    LOG.debug { "Add remote editor. $editor" }

    remoteTracedEditors.putValue(ClientId.current, editor)
    update()
  }

  private fun update() {
    // it's important that local ones go first
    val editors = currentEditors.filter { it.clientId().isLocal }.toMutableList()
    @Suppress("UsePropertyAccessSyntax")
    setActiveEditors(editors)
  }

  override fun editorReleasedImpl(editor: Editor) {
    if (ClientId.isCurrentlyUnderLocalId) {
      LOG.debug { "Release remote editor. $editor" }
      super.editorReleasedImpl(editor)
      return
    }

    LOG.debug { "Release remote editor. $editor" }
    remoteTracedEditors.remove(ClientId.current, editor)
  }

  override fun getActiveEditors(): MutableList<Editor> {
    application.assertIsDispatchThread()
    return currentEditors.toMutableList()
  }

  override fun setActiveEditors(editors: MutableList<Editor>) {
    application.assertIsDispatchThread()
    val allEditors = editors + collectActiveRemoteEditors()

    LOG.debug { "All active Editors. [${allEditors.joinToString(separator = " ,") { it.toString() }}]" }

    if (allEditors != currentEditors) {
      currentEditors = allEditors
      project.messageBus.syncPublisher(EditorTrackerListener.TOPIC).activeEditorsChanged(editors)
    }
  }

  private fun collectActiveRemoteEditors(): List<Editor> {
    val remoteEditors = mutableListOf<Editor>()
    for ((clientId, editors) in remoteTracedEditors.entrySet()) {
      if (isActive(clientId)) {
        val activeRemoteEditors = editors.filter { isActive(it) }
        remoteEditors.addAll(activeRemoteEditors)
      }
    }
    return remoteEditors
  }

  private fun isActive(editor: Editor): Boolean {
    return editor.getUserData(RemoteEditors.REMOTELY_VISIBLE) == true
  }

  private fun isActive(clientId: ClientId): Boolean {
    return true
  }
}