package com.jetbrains.rdserver.daemon.inlays

import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.Inlay
import com.intellij.openapi.editor.InlayModel
import com.intellij.openapi.editor.RangeMarker
import com.intellij.openapi.editor.ex.RangeMarkerEx
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.util.SmartList
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.daemon.BackendEditorMarkupModelContributor
import com.jetbrains.rdserver.daemon.BackendMarkupModelContributor
import com.jetbrains.rdserver.daemon.BackendProtocolMarkupModelExtensionHandler
import com.jetbrains.rdserver.daemon.CHUNK_SIZE
import com.jetbrains.rd.ide.model.MarkupModelExtension

class BackendInlayModelBinder : BackendProtocolMarkupModelExtensionHandler {
  override fun consume(lifetime: Lifetime, extensions: List<MarkupModelExtension>, contributor: BackendMarkupModelContributor) {
    if (contributor !is BackendEditorMarkupModelContributor) return
    val editor = contributor.editor
    contributor.bind(lifetime, editor.inlayModel)
  }

  private fun BackendEditorMarkupModelContributor.bind(lifetime: Lifetime, inlayModel: InlayModel) {
    val listener = object : InlayModel.Listener {
      val toAddOrUpdate = mutableSetOf<RangeMarkerEx>()
      val toRemove = mutableSetOf<RangeMarkerEx>()
      override fun onAdded(inlay: Inlay<*>) {
        val rangeMarker = inlay as RangeMarkerEx
        if (inlayModel.isInBatchMode) {
          toAddOrUpdate.add(rangeMarker)
        }
        else {
          addOrUpdateRangeMarkers(SmartList(rangeMarker))
        }
      }

      override fun onUpdated(inlay: Inlay<*>) {
        onAdded(inlay)
      }

      override fun onRemoved(inlay: Inlay<*>) {
        val rangeMarker = inlay as RangeMarkerEx
        if (inlayModel.isInBatchMode) {
          toAddOrUpdate.remove(rangeMarker)
          toRemove.add(rangeMarker)
        }
        else {
          removeRangeMarkers(SmartList(rangeMarker))
        }
      }

      override fun onBatchModeFinish(editor: Editor) {
        try {
          removeRangeMarkers(toRemove)
          addOrUpdateRangeMarkers(toAddOrUpdate)
        }
        finally {
          toRemove.clear()
          toAddOrUpdate.clear()
        }
      }
    }

    inlayModel.getBlockElementsInRange(0, Int.MAX_VALUE)
      .map { it as RangeMarker }
      .chunked(CHUNK_SIZE)
      .forEach { addOrUpdateRangeMarkers(it) }

    inlayModel.addListener(listener, lifetime.createNestedDisposable())
  }
}