package com.jetbrains.rdserver.daemon.syntax

import com.intellij.openapi.editor.colors.TextAttributesKey
import com.intellij.openapi.editor.ex.MarkupIterator
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.ex.util.LexerEditorHighlighter
import com.intellij.openapi.editor.highlighter.HighlighterIterator
import com.intellij.openapi.editor.markup.RangeHighlighter
import com.jetbrains.rd.platform.util.userData


fun <TOld, TNew> calculateDiff(old: THighlighterIterator<TOld>,
                               new: THighlighterIterator<TNew>): Pair<List<TOld>, List<TNew>> {
  val toRemove = mutableListOf<TOld>()
  val toAdd = mutableListOf<TNew>()

  while (!old.atEnd() || !new.atEnd()) {
    // [->
    //    [
    while (!old.atEnd() && (new.atEnd() || old.start < new.start)) {
      toRemove.add(old.current)
      old.advance()
    }

    //    ]
    // ]->
    while (!new.atEnd() && (old.atEnd() || old.end > new.end)) {
      toAdd.add(new.current)
      new.advance()
    }

    if (!old.atEnd() && !new.atEnd()) {
      if (old.start != new.start || old.end != new.end ||
          !old.textAttributesKeys.contentEquals(new.textAttributesKeys)) {
        toRemove.add(old.current)
        toAdd.add(new.current)
      }
      new.advance()
      old.advance()
    }
  }
  assert(old.atEnd() && new.atEnd())
  return Pair(toRemove, toAdd)
}

interface THighlighterIterator<T> : HighlighterIterator {
  val current: T

  override fun retreat() = error("not supported")
  override fun getTokenType() = error("not supported")
  override fun getTextAttributes() = error("not supported")
  override fun getDocument() = error("not supported")

  /**
   * must be called after working with iterator
   */
  fun dispose()
}

var RangeHighlighter.textAttributesKeys: Array<TextAttributesKey>? by userData()

class MarkupBasedSyntaxIterator(
  private val delegate: MarkupIterator<RangeHighlighterEx>
) : THighlighterIterator<RangeHighlighterEx> {

  override val current get() = delegate.peek()!!
  override fun getTextAttributesKeys() = current.textAttributesKeys
  override fun getStart() = current.startOffset
  override fun getEnd() = current.endOffset
  override fun atEnd() = !delegate.hasNext()

  override fun advance() {
    delegate.next()
  }

  override fun dispose() {
    delegate.dispose()
  }
}

class SyntaxHighlighterData(val startOffset: Int, val endOffset: Int, val textAttributesKeys: Array<TextAttributesKey>)

class LexerBasedSyntaxIterator(
  lexerHighlighter: LexerEditorHighlighter,
  fragmentStart: Int,
  private val fragmentEnd: Int
) : THighlighterIterator<SyntaxHighlighterData> {
  private val delegate = lexerHighlighter.createIterator(fragmentStart)

  init {
    if (!atEnd() && !calculateCurrentKey()) {
      advance()
    }
  }

  override val current get() = SyntaxHighlighterData(start, end, attributesKeys!!)
  private var attributesKeys: Array<TextAttributesKey>? = null
  override fun getTextAttributesKeys() = attributesKeys
  override fun getStart() = delegate.start
  override fun getEnd() = delegate.end
  override fun atEnd() = delegate.atEnd() || delegate.start >= fragmentEnd

  override fun advance() {
    do {
      delegate.advance()
    }
    while (!atEnd() && !calculateCurrentKey())
  }

  private fun calculateCurrentKey(): Boolean {
      attributesKeys = delegate.textAttributesKeys
    return !attributesKeys.isNullOrEmpty()
  }

  override fun dispose() {
  }
}

fun <T> THighlighterIterator<T>.toList(): List<T> {
  val result = mutableListOf<T>()
  try {
    while (!atEnd()) {
      result.add(current)
      advance()
    }
  }
  finally {
    this.dispose()
  }
  return result
}
