package com.jetbrains.rdserver.daemon.syntax

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.ex.MarkupModelEx
import com.intellij.openapi.editor.ex.util.LexerEditorHighlighter
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.editor.impl.FilteringMarkupIterator
import com.intellij.openapi.editor.impl.HighlighterListener
import com.intellij.openapi.editor.markup.HighlighterLayer
import com.intellij.openapi.editor.markup.HighlighterTargetArea
import com.intellij.openapi.fileEditor.impl.text.AsyncEditorLoader
import com.intellij.openapi.rd.createNestedDisposable
import com.jetbrains.rd.ide.model.TextControlModel
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rdserver.daemon.markupModelContributor
import com.jetbrains.rdserver.daemon.performInBulkIfPossible
import com.jetbrains.rdserver.editors.BackendEditorHostListener
import org.jetbrains.annotations.NotNull
import kotlin.math.max
import kotlin.math.min


object MarkupBasedLexerHighlighting {

  class MyEditorHostListener : BackendEditorHostListener {
    override fun editorBound(lifetime: Lifetime, editorModel: TextControlModel, editor: Editor) {
      if (editor !is EditorImpl) return
      if (ClientId.isLocalId(ClientId.current)) return

      performWhenLoaded(lifetime, editor) {
        updateSyntaxHighlighters(editor, 0, editor.document.textLength)

        editor.addHighlighterListener(
          HighlighterListener { fragmentStart, fragmentEnd ->
            updateSyntaxHighlighters(editor, fragmentStart, fragmentEnd)
          }, lifetime.createNestedDisposable())
      }
    }
  }

  private fun updateSyntaxHighlighters(editor: EditorImpl, fragmentStart: Int, fragmentEnd: Int) {
    val lexerHighlighter = editor.highlighter as? LexerEditorHighlighter ?: return

    // implementations of segment arrays and interval trees treat boundaries differently
    val hStart = max(fragmentStart - 1, 0)
    val hEnd = min(fragmentEnd + 1, editor.document.textLength)
    val editorMarkup = editor.markupModel
    val old = createMarkupIterator(editorMarkup, hStart, hEnd)

    // implementations of segment arrays and interval trees treat boundaries differently
    val lStart = max(fragmentStart - 2, 0)
    val lEnd = min(fragmentEnd + 2, editor.document.textLength)
    val new = createLexerIterator(lexerHighlighter, lStart, lEnd)

    val (toRemove, toAdd) = try {
      calculateDiff(old, new)
    }
    finally {
      old.dispose()
      new.dispose()
    }

    editor.markupModelContributor.performInBulkIfPossible(toRemove) {
      editorMarkup.removeHighlighter(it)
    }

    editor.markupModelContributor.performInBulkIfPossible(toAdd) {
      editorMarkup
        .addRangeHighlighterAndChangeAttributes(it.textAttributesKeys.last(), it.startOffset, it.endOffset,
                                                HighlighterLayer.SYNTAX,
                                                HighlighterTargetArea.EXACT_RANGE, false) { ex ->
          ex.textAttributesKeys = it.textAttributesKeys
        }
    }
  }

  private fun createMarkupIterator(markup: @NotNull MarkupModelEx,
                                   start: Int,
                                   end: Int): MarkupBasedSyntaxIterator {
    val overlappingIterator = markup.overlappingIterator(start, end)
    val markupIterator = FilteringMarkupIterator(overlappingIterator) { it.textAttributesKeys != null }
    return MarkupBasedSyntaxIterator(markupIterator)
  }

  private fun createLexerIterator(lexerHighlighter: LexerEditorHighlighter,
                                  start: Int,
                                  end: Int): LexerBasedSyntaxIterator {
    return LexerBasedSyntaxIterator(lexerHighlighter, start, end)
  }

  private fun performWhenLoaded(lifetime: Lifetime, editor: Editor, action: () -> Unit) {
    // EditorFactoryListener#editorCreated (the source of textControlHost.openedEditors)
    // is fired too early to be able to use AsyncEditorLoader#performWhenLoaded
    // No commonly used event for subscribing at the right moment
    ApplicationManager.getApplication().invokeLater {
      AsyncEditorLoader.performWhenLoaded(editor) loaded@{
        if (editor.isDisposed || lifetime.isNotAlive) return@loaded

        action()
      }
    }
  }
}
