package com.jetbrains.rdserver.debugger

import com.intellij.codeWithMe.ClientId
import com.intellij.execution.ExecutionManager
import com.intellij.execution.RunManager
import com.intellij.execution.configurations.RunConfiguration
import com.intellij.execution.executors.DefaultDebugExecutor
import com.intellij.execution.runners.ExecutionEnvironmentBuilder
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Key
import com.intellij.util.containers.ContainerUtil
import com.intellij.util.pooledThreadSingleAlarm
import com.intellij.xdebugger.*
import com.intellij.xdebugger.frame.XExecutionStack
import com.intellij.xdebugger.frame.XStackFrame
import com.intellij.xdebugger.impl.XDebugSessionImpl
import com.jetbrains.codeWithMe.model.*
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.platform.codeWithMe.debugger.CommonDebuggerHost
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.ViewableMap
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.adviseOnTargetOutput
import com.jetbrains.rdserver.execution.configurations.BackendRunContentHost
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.permissions.RunAccessLevel
import com.jetbrains.rdserver.toProtocolModel

@Suppress("NonDefaultConstructor")
class BackendDebuggerHost(
  val clientId: ClientId,
  lifetime: Lifetime,
  project: Project,
  protocol: Protocol
) : CommonDebuggerHost(lifetime, protocol, project) {
  val processMap = ViewableMap<XDebugProcess, DebugProcessInfo>()
  private val connection = project.messageBus.connect(lifetime.createNestedDisposable())
  private val frameStackMap = ContainerUtil.createWeakKeyWeakValueMap<XStackFrame, XExecutionStack>()
  private val frameIdMap = ContainerUtil.createWeakMap<XStackFrame, Int>()
  private val idFrameMap = ContainerUtil.createWeakValueMap<Int, XStackFrame>()
  private var nextId = 0

  init {
    XDebuggerManager.getInstance(project).debugSessions.forEach { debugSession ->
      val model = syncDebugProcessToProtocol(debugSession.debugProcess)
      debugSession.suspendContext?.let { xSuspendContext ->
        model?.suspendContext?.set(xSuspendContext.createSuspendContextModel(project, clientId))
      }
    }

    connection.subscribe(XDebuggerManager.TOPIC, object : XDebuggerManagerListener {
      override fun processStarted(debugProcess: XDebugProcess) {
        ClientId.withClientId(clientId) {
          // TODO remove application.invokeLater
          application.invokeLater {
            if (lifetime.isNotAlive) return@invokeLater
            initSessionFromRemote(debugProcess.session)
            syncDebugProcessToProtocol(debugProcess)
          }
        }
      }

      override fun processStopped(debugProcess: XDebugProcess) {
        if (lifetime.isNotAlive) return
        ClientId.withClientId(clientId) {
          val info = processMap[debugProcess]
                     ?: error("debugProcess was not found in the processMap")

          processMap.remove(debugProcess)
          pooledThreadSingleAlarm(1000, lifetime.createNestedDisposable()) {
            info.processLifetime.terminate()
            application.invokeLater {
              debuggerModel.sessions.remove(info.model)
            }
          }.request()
        }
      }

      override fun currentSessionChanged(previousSession: XDebugSession?, currentSession: XDebugSession?) {
        // TODO what?
      }
    })

    debuggerModel.startCommands.advise(lifetime) { (settingsId, commandId, sessionData) ->
      if (PermissionManager.getInstance(project).getRunAccess(clientId) != RunAccessLevel.FullAccess)
        return@advise

      val runManager = RunManager.getInstance(project)
      val settings = runManager.allSettings.find { it.uniqueID == settingsId }
                     ?: return@advise

      val executor = DefaultDebugExecutor()
      val builder = ExecutionEnvironmentBuilder.createOrNull(executor, settings)
                    ?: return@advise

      val executionManager = ExecutionManager.getInstance(project)
      val environment = builder.activeTarget().build().apply {
        putUserData(START_COMMAND_ID, commandId)
        putUserData(STARTER_CLIENT_ID, ClientId.currentOrNull)
        putUserData(SESSION_DATA, sessionData)
      }
      ClientId.withClientId(ClientId.localId) {
        executionManager.restartRunProfile(environment);
      }
    }
  }

  private fun initSessionFromRemote(session: XDebugSession) {
    val sessionData = (session as? XDebugSessionImpl)?.executionEnvironment?.getUserData(SESSION_DATA) ?: return
    session.setBreakpointMuted(sessionData.areBreakpointsMuted)
  }

  internal fun getOrCreateFrameId(stack: XExecutionStack?, frame: XStackFrame): Int {
    var id = frameIdMap[frame]
    if (id == null) {
      id = nextId++
      frameIdMap[frame] = id
      idFrameMap[id] = frame
      if (stack != null) {
        frameStackMap[frame] = stack
      }
    }
    return id
  }

  internal fun tryGetFrameById(id: Int): XStackFrame? {
    return idFrameMap[id]
  }

  internal fun tryGetStackByFrame(frame: XStackFrame): XExecutionStack? {
    return frameStackMap[frame]
  }

  private fun syncDebugProcessToProtocol(debugProcess: XDebugProcess): CodeWithMeDebuggerSessionModel? {
    check(debugProcess !in processMap) { "$debugProcess is already in processMap" }

    fun createSessionModel(): CodeWithMeDebuggerSessionModel {
      val processLifetime = lifetime.createNested()
      val sessionModel = createDebuggerSessionModel(processLifetime.lifetime, debugProcess.session)
      processMap[debugProcess] = DebugProcessInfo(sessionModel, processLifetime)
      debuggerModel.sessions.add(sessionModel)
      return sessionModel
    }
    
    var sessionModel: CodeWithMeDebuggerSessionModel? = null
    if (debugProcess.session.isInitialized()) {
      sessionModel = createSessionModel()
    }

    debugProcess.session.addSessionListener(object : XDebugSessionListener {
      override fun sessionPaused() {
        application.invokeLater {
          if (lifetime.isNotAlive) return@invokeLater
          // If debug session had 'showTabOnSuspend' then create a session model on the first pause
          if (sessionModel == null) {
            sessionModel = createSessionModel() 
          }
          
          val xSuspendContext = debugProcess.session.suspendContext
          val model = xSuspendContext.createSuspendContextModel(project, clientId)
          sessionModel!!.suspendContext.set(model)
        }
      }

      override fun sessionResumed() {
        application.invokeLater {
          if (lifetime.isNotAlive) return@invokeLater
          sessionModel!!.suspendContext.set(null)
        }
      }
    }, lifetime.createNestedDisposable())

    return sessionModel
  }
  
  private fun XDebugSession.isInitialized() = (this as XDebugSessionImpl).sessionTab != null

  private fun createDebuggerSessionModel(sessionLifetime: Lifetime, session: XDebugSession): CodeWithMeDebuggerSessionModel {
    val environment = (session as? XDebugSessionImpl)?.executionEnvironment
    val originalClientId = environment?.getUserData(STARTER_CLIENT_ID)
    val startCommandId = if (originalClientId == clientId) environment.getUserData(START_COMMAND_ID) else null
    return CodeWithMeDebuggerSessionModel(
      session.sessionName,
      session.runContentDescriptor.icon?.toProtocolModel(),
      session.runContentDescriptor.executionId,
      guessTestFramework(session),
      isIdBased(session),
      getUniqueID(session),
      BackendFollowMeDebuggerHost.getInstance(project).getOrCreateGlobalSessionId(session),
      startCommandId,
      BackendRunContentHost.getExecutionConsoleKind(session.consoleView)
    ).apply {
      adviseOnDebuggerActions(sessionLifetime, session, this)
      adviseOnTargetOutput(sessionLifetime, session.project, clientId, session.consoleView, targetOutput)
      syncSessionSettingsWithModel(session, this, startCommandId == null)
    }
  }

  private fun getUniqueID(session: XDebugSession): String {
    if (session.runProfile is RunConfiguration) {
      val settings = RunManager.getInstance(session.project).findSettings(session.runProfile as RunConfiguration)
      if (settings != null) return settings.uniqueID
    }
    return ""
  }

  private fun guessTestFramework(session: XDebugSession): String {
    val executor = (session as? XDebugSessionImpl)?.executionEnvironment?.executor
    if (executor == null) return ""
    return BackendRunContentHost.tryGetTestFramework(project, session.runProfile, executor)
  }
  
  private fun isIdBased(session: XDebugSession): Boolean {
    val executor = (session as? XDebugSessionImpl)?.executionEnvironment?.executor
    if (executor == null) return false
    return BackendRunContentHost.getConsoleProperties(session.runProfile, executor)?.isIdBasedTestTree?:false
  }

  private fun adviseOnDebuggerActions(lifetime: Lifetime, session: XDebugSession, model: CodeWithMeDebuggerSessionModel) {
    val permissionManager = PermissionManager.getInstance(project)
    model.debuggerAction.advise(lifetime) { action ->
      logger.trace("Got debugger action from remote client: ${action.javaClass}")
      when (action) {
        is StopActionModel -> {
          if (permissionManager.getRunAccess(clientId) != RunAccessLevel.FullAccess) return@advise
          session.stop()
        }
        is ResumeActionModel -> {
          if (permissionManager.getRunAccess(clientId) != RunAccessLevel.FullAccess) return@advise
          session.resume()
        }
        is StepIntoActionModel -> {
          if (permissionManager.getRunAccess(clientId) != RunAccessLevel.FullAccess) return@advise
          session.stepInto()
        }
        is StepOverActionModel -> {
          if (permissionManager.getRunAccess(clientId) != RunAccessLevel.FullAccess) return@advise
          session.stepOver(false)
        }
        is StepOutActionModel -> {
          if (permissionManager.getRunAccess(clientId) != RunAccessLevel.FullAccess) return@advise
          session.stepOut()
        }
        is RunToPositionActionModel -> {
          if (permissionManager.getRunAccess(clientId) != RunAccessLevel.FullAccess) return@advise
          val position = action.sourcePosition.fromModel(project) ?: run {
            logger.warn("RunToPositionActionModel: source position == null")
            return@advise
          }

          session.runToPosition(position, false)
        }
      }
    }
  }

  companion object {
    private val logger = getLogger<BackendDebuggerHost>()
    private val START_COMMAND_ID = Key.create<Int>("BackendDebuggerHost.StartCommandId")
    private val STARTER_CLIENT_ID = Key.create<ClientId?>("BackendDebuggerHost.StarterClientId")
    private val SESSION_DATA = Key.create<InitialSessionData>("BackendDebuggerHost.SessionData") 
    
    fun getInstance(project: Project): BackendDebuggerHost = RdServerPerClientComponents.getService(project)
    fun getInstance(project: Project, clientId: ClientId) = ClientId.withClientId(clientId) { getInstance(project) }

    data class DebugProcessInfo(
      val model: CodeWithMeDebuggerSessionModel,
      val processLifetime: LifetimeDefinition
    )
  }
}