package com.jetbrains.rdserver.debugger

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.util.containers.ContainerUtil
import com.intellij.xdebugger.XDebugSession
import com.intellij.xdebugger.XDebugSessionListener
import com.jetbrains.codeWithMe.model.CodeWithMeDebuggerSessionModel
import com.jetbrains.rd.platform.util.idea.LifetimedProjectService
import com.jetbrains.rd.util.lifetime.intersect
import com.jetbrains.rdserver.followMe.FollowMeManager

class BackendFollowMeDebuggerHost(project: Project) : LifetimedProjectService(project) {
  private val sessionIdMap = ContainerUtil.createWeakMap<XDebugSession, Long>()
  private var nextId = 0L
  
  init {
    val followMeManager = FollowMeManager.getInstance(project)
    followMeManager.followers.view(projectServiceLifetime) { followingLifetime, who, whom ->
      when {
        whom == ClientId.localId -> {
          syncGuestWithHost(who)
        }
        who == ClientId.localId -> {
          syncHostWithGuest(whom)
        }
        else -> {
          // TODO live subscription on new sessions?
          val whoHost = ClientId.withClientId(who) { BackendDebuggerHost.getInstance(project) }
          val whomHost = ClientId.withClientId(whom) { BackendDebuggerHost.getInstance(project) }
          for (whomSessionModel in whomHost.debuggerModel.sessions) {
            val globalId = whomSessionModel.globalSessionId
            val whoSessionModel = whoHost.debuggerModel.sessions.find { it.globalSessionId == globalId }
                                  ?: continue
            
            syncGuestModels(whoHost, whoSessionModel, whomHost, whomSessionModel)
          }
        }
      } 
    }
  }
  
  private fun syncHostWithGuest(whom: ClientId) {
    val host = ClientId.withClientId(whom) { BackendDebuggerHost.getInstance(project) }
    host.debuggerModel.sessions.view(projectServiceLifetime) { sessionLifetime, _, sessionModel ->
      val lifetime = projectServiceLifetime.intersect(sessionLifetime)
      sessionModel.currentFrame.advise(lifetime) { frameId ->
        val frame = host.tryGetFrameById(frameId) ?: return@advise
        val stack = host.tryGetStackByFrame(frame) ?: return@advise
        val (process, _) = host.processMap.toList().find { it.second.model === sessionModel } ?: return@advise
        process.session.setCurrentStackFrame(stack, frame)
      }
    }
  }
  
  private fun syncGuestWithHost(who: ClientId) {
    val host = ClientId.withClientId(who) { BackendDebuggerHost.getInstance(project) }
    host.processMap.advise(projectServiceLifetime) { event ->
      val process = event.key
      val info = event.newValueOpt ?: return@advise
      val lifetime = projectServiceLifetime.intersect(info.processLifetime)
      process.session.addSessionListener(object : XDebugSessionListener {
        override fun stackFrameChanged() {
          if (!ClientId.isCurrentlyUnderLocalId)
            return
          
          val currentFrame = process.session.currentStackFrame ?: return
          val currentStack = process.session.suspendContext.activeExecutionStack ?: return
          val frameId = host.getOrCreateFrameId(currentStack, currentFrame)
          info.model.changeFrame.fire(frameId)
        }
      }, lifetime.createNestedDisposable())
    }
  }
  
  private fun syncGuestModels(whoHost: BackendDebuggerHost, whoSession: CodeWithMeDebuggerSessionModel, 
                              whomHost: BackendDebuggerHost, whomSession: CodeWithMeDebuggerSessionModel) {
    whomSession.currentFrame.advise(projectServiceLifetime) { whomFrameId ->
      val frame = whomHost.tryGetFrameById(whomFrameId) ?: return@advise
      val whoFrameId = whoHost.getOrCreateFrameId(null, frame)
      whoSession.changeFrame.fire(whoFrameId)
    }
  }
  
  fun getOrCreateGlobalSessionId(session: XDebugSession): Long {
    return sessionIdMap.getOrPut(session) { nextId++ }
  }
  
  companion object {
    fun getInstance(project: Project): BackendFollowMeDebuggerHost = 
      project.getService(BackendFollowMeDebuggerHost::class.java)
  }
}