package com.jetbrains.rdserver.debugger

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.extensions.ExtensionPointListener
import com.intellij.openapi.extensions.PluginDescriptor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.VirtualFileManager
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.intellij.xdebugger.XDebugSession
import com.intellij.xdebugger.XDebuggerManager
import com.intellij.xdebugger.breakpoints.*
import com.intellij.xdebugger.impl.breakpoints.XBreakpointBase
import com.jetbrains.codeWithMe.model.BreakpointPresentationUpdatedArgs
import com.jetbrains.codeWithMe.model.CustomBreakpointTypeModel
import com.jetbrains.codeWithMe.model.LineBreakpointTypeModel
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.ide.model.RdFileId
import com.jetbrains.rd.platform.codeWithMe.debugger.CommonBreakpointHost
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.vfs.BackendVirtualFileSystemHost
import com.jetbrains.rdserver.toProtocolModel

@Suppress("NonDefaultConstructor")
class CodeWithMeBreakpointHost(
  val clientId: ClientId,
  lifetime: Lifetime,
  project: Project,
  protocol: Protocol
) : CommonBreakpointHost(lifetime, protocol, project) {
  private val breakpointsUpdateQueue = 
    MergingUpdateQueue("Code With Me breakpoints", 1000, true, null, project)
  
  init {
    application.assertIsDispatchThread()

    syncBreakpointTypes()
    syncExistingBreakpointsToModel()
    listenBreakpointPresentationUpdates()

    model.getCanPutAtTypes.set { args ->
      val file = findFile(args.fileId) ?: return@set emptyList()
      val point = XBreakpointType.EXTENSION_POINT_NAME.point
      point.extensionList
        .filterIsInstance<XLineBreakpointType<*>>()
        .filter { it.canPutAt(file, args.line, project) }
        .map { it.id }
    }
  }

  private fun listenBreakpointPresentationUpdates() {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(XBreakpointListener.TOPIC, object : XBreakpointListener<XBreakpoint<*>> {
      override fun breakpointPresentationUpdated(breakpoint: XBreakpoint<*>, session: XDebugSession?) {
        if (session == null)
          return // TODO[moklev] support global presentation
        
        // TODO[moklev] support different debug sessions, not only active one
        val breakpointId = breakpointMap[breakpoint] ?: return
        val debuggerHost = BackendDebuggerHost.getInstance(project, clientId)
        val (sessionModel, _) = debuggerHost.processMap[session.debugProcess] ?: return
        
        // TODO[moklev] this update must be called after XLineBreakpointManager's queue update
        breakpointsUpdateQueue.queue(object : Update(breakpointId) {
          override fun run() {
            val icon = (breakpoint as? XBreakpointBase<*, *, *>)?.icon
            val errorMessage = (breakpoint as? XBreakpointBase<*, *, *>)?.errorMessage
            sessionModel.breakpointPresentationUpdated.fire(
              BreakpointPresentationUpdatedArgs(breakpointId, icon?.toProtocolModel(), errorMessage)
            )
          }
        })
      }
    })
  }
  
  private fun syncBreakpointTypes() {
    fun addBreakpointTypeToModel(type: XBreakpointType<*, *>) {
      try {
        val enabledIcon = type.enabledIcon.toProtocolModel()
        val disabledIcon = type.disabledIcon.toProtocolModel()
        val suspendNoneIcon = type.suspendNoneIcon.toProtocolModel()
        val mutedEnabledIcon = type.mutedEnabledIcon.toProtocolModel()
        val mutedDisabledIcon = type.mutedDisabledIcon.toProtocolModel()
        val pendingIcon = type.pendingIcon?.toProtocolModel()
        val inactiveDependentIcon = type.inactiveDependentIcon.toProtocolModel()

        val typeModel = if (type is XLineBreakpointType<*>)
          LineBreakpointTypeModel(type.temporaryIcon.toProtocolModel(), type.id, type.title, enabledIcon, disabledIcon, 
                                  suspendNoneIcon, mutedEnabledIcon, mutedDisabledIcon, pendingIcon, inactiveDependentIcon)
        else
          CustomBreakpointTypeModel(type.id, type.title, enabledIcon, disabledIcon, suspendNoneIcon, 
                                    mutedEnabledIcon, mutedDisabledIcon, pendingIcon, inactiveDependentIcon)

        model.breakpointTypes[typeModel.id] = typeModel
      } catch (t: Throwable) {
        logger.error("Failed to add the breakpoint type: $type", t)
      }
    }

    fun removeBreakpointTypeFromModel(type: XBreakpointType<*, *>) {
      model.breakpointTypes.remove(type.id)
    }

    XBreakpointType.EXTENSION_POINT_NAME.extensionList.forEach { addBreakpointTypeToModel(it) }
    XBreakpointType.EXTENSION_POINT_NAME.addExtensionPointListener(object : ExtensionPointListener<XBreakpointType<*, *>> {
      override fun extensionAdded(extension: XBreakpointType<*, *>, pluginDescriptor: PluginDescriptor) {
        addBreakpointTypeToModel(extension)
      }

      override fun extensionRemoved(extension: XBreakpointType<*, *>, pluginDescriptor: PluginDescriptor) {
        removeBreakpointTypeFromModel(extension)
      }
    }, lifetime.createNestedDisposable())
  }

  override fun findFile(fileId: RdFileId): VirtualFile? {
    return BackendVirtualFileSystemHost.getInstance(project).findVirtualFile(fileId)
  }

  override fun findFileId(file: VirtualFile): RdFileId? {
    return BackendVirtualFileSystemHost.getInstance(project).bindFile(file)
  }

  override fun beforeBreakpointAdd(breakpoint: XBreakpoint<*>) {
    if (breakpoint !is XLineBreakpoint<*>) return
    val virtualFile = VirtualFileManager.getInstance().findFileByUrl(breakpoint.fileUrl) ?: return
    BackendVirtualFileSystemHost.getInstance(project).bindFile(virtualFile)
  }

  override fun doIgnoreBreakpoint(breakpoint: XBreakpoint<*>): Boolean {
    return false
  }

  private fun syncExistingBreakpointsToModel() {
    val breakpointManager = XDebuggerManager.getInstance(project).breakpointManager
    breakpointManager.allBreakpoints.forEach(::addBreakpointToModel)
  }
  
  companion object {
    private val logger = getLogger<CodeWithMeBreakpointHost>()
  }
}