package com.jetbrains.rdserver.debugger

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.editor.colors.TextAttributesKey
import com.intellij.openapi.project.Project
import com.intellij.ui.GuiUtils
import com.intellij.ui.SimpleTextAttributes
import com.intellij.util.DeprecatedMethodException
import com.intellij.xdebugger.XSourcePosition
import com.intellij.xdebugger.frame.*
import com.intellij.xdebugger.frame.presentation.XValuePresentation
import com.jetbrains.codeWithMe.model.*
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.IOptProperty
import com.jetbrains.rd.util.reactive.adviseEternal
import com.jetbrains.rd.util.reactive.adviseOnce
import com.jetbrains.rdserver.toProtocolModel
import javax.swing.Icon

abstract class CodeWithMeCompositeNodeBase(protected val clientId: ClientId, protected val project: Project) : XCompositeNode {
  private val values = mutableListOf<ValueEntityModelBase>()

  override fun addChildren(children: XValueChildrenList, last: Boolean) {
    for (group in children.topGroups) {
      values.add(group.createValueGroupModel(clientId, project))
    }
    
    for (childIndex in 0 until children.size()) {
      val name = children.getName(childIndex)
      val value = children.getValue(childIndex)
      values.add(value.createValueModel(clientId, name, project))
    }

    for (group in children.bottomGroups) {
      values.add(group.createValueGroupModel(clientId, project))
    }

    if (last) {
      taskSetChildrenList(values.toList())
      values.clear()
    }
  }

  override fun setErrorMessage(errorMessage: String) {
    taskSetErrorMessage(errorMessage)
  }

  override fun setErrorMessage(errorMessage: String, link: XDebuggerTreeNodeHyperlink?) {
    /*TODO*/
    setErrorMessage(errorMessage)
  }

  override fun tooManyChildren(remaining: Int) {
    DeprecatedMethodException.report("Use tooManyChildren(Int, Runnable) instead")
  }

  override fun tooManyChildren(remaining: Int, addNextChildren: Runnable) {
    // TODO Migrate all composite nodes to the live node
    setErrorMessage("Too many children, remaining: $remaining")
  }

  override fun setAlreadySorted(alreadySorted: Boolean) { /*TODO*/ }
  override fun setMessage(message: String, icon: Icon?, attributes: SimpleTextAttributes, link: XDebuggerTreeNodeHyperlink?) { /*TODO*/ }
  
  abstract fun taskSetChildrenList(children: List<ValueEntityModelBase>)
  abstract fun taskSetErrorMessage(errorMessage: String)
}

class CodeWithMeCompositeNode(clientId: ClientId, val task: RdTask<List<ValueEntityModelBase>>, project: Project) 
  : CodeWithMeCompositeNodeBase(clientId, project) {
  
  override fun taskSetChildrenList(children: List<ValueEntityModelBase>) {
    application.invokeLater {
      task.set(children)
    }
  }

  override fun taskSetErrorMessage(errorMessage: String) {
    task.result.set(RdTaskResult.Fault(Exception(errorMessage)))
  }
}

class CodeWithMeCompositeNodeLive(clientId: ClientId, task: RdTask<CodeWithMeChildrenListLive>, 
                                  project: Project, private val lifetime: Lifetime)
  : CodeWithMeCompositeNodeBase(clientId, project) {
  private val liveList = CodeWithMeChildrenListLive()   

  init {
    application.invokeLater {
      task.set(liveList)
    }
  }
  
  override fun taskSetChildrenList(children: List<ValueEntityModelBase>) {
    if (lifetime.isNotAlive)
      return

    application.invokeLater {
      liveList.children.add(ImmutableListChildrenWrapper(children))
    }
  }

  override fun taskSetErrorMessage(errorMessage: String) {
    // TODO
  }

  override fun tooManyChildren(remaining: Int, addNextChildren: Runnable) {
    if (lifetime.isNotAlive)
      return

    application.invokeLater {
      liveList.getNextChildren.adviseOnce(lifetime) {
        addNextChildren.run()
      }
      liveList.tooManyChildren.set(remaining)
    }
  }
}

fun XValueGroup.createValueGroupModel(clientId: ClientId, project: Project): CodeWithMeValueGroupModel {
  val xValueGroup = this
  return CodeWithMeValueGroupModel(icon?.toProtocolModel(), isAutoExpand, isRestoreExpansion, separator, comment, name).apply {
    computeChildren.set { _, _ ->
      xValueGroup.computeChildrenAsync(clientId, project)
    }
  }
}

fun XValue.createValueModel(clientId: ClientId, name: String, project: Project): CodeWithMeValueModel {
  val xValue = this
  return CodeWithMeValueModel(xValue.canNavigateToTypeSource(), null, name).apply {
    application.invokeLater {
      computeObjectPropertiesLive.set { _, _ ->
        xValue.createValuePropertiesModel(clientId, project)
      }
      computeChildren.set { _, _ ->
        throw RuntimeException("Use computeChildrenLive instead")
      }
      computeChildrenLive.set { callLifetime, _ ->
        xValue.computeChildrenLiveAsync(clientId, project, callLifetime)
      }
      computeSourcePosition.set { _, _ ->
        xValue.computeSourcePositionAsync(clientId, project)
      }
      computeTypeSourcePosition.set { _, _ ->
        xValue.computeTypeSourcePositionAsync(clientId, project)
      }
      inlineDebuggerDataCallback.adviseEternal { callbackModel ->
        xValue.computeInlineDebuggerData(object : XInlineDebuggerDataCallback() {
          override fun computed(position: XSourcePosition?) {
            application.invokeLater computedLater@{
              val positionModel = position?.toModel(clientId, project) ?: return@computedLater
              callbackModel.computed.start(positionModel)
            }
          }
        })
      }
    }
  }
}

fun XValue.computeTypeSourcePositionAsync(
  clientId: ClientId,
  project: Project
): RdTask<CodeWithMeSourcePositionModel?> {
  return RdTask<CodeWithMeSourcePositionModel?>().apply {
    computeTypeSourcePosition { sourcePosition ->
      GuiUtils.invokeLaterIfNeeded(Runnable {
        set(sourcePosition?.toModel(clientId, project))
      }, ModalityState.defaultModalityState())
    }
  }
}

fun XValue.computeSourcePositionAsync(
  clientId: ClientId,
  project: Project
): RdTask<CodeWithMeSourcePositionModel?> {
  return RdTask<CodeWithMeSourcePositionModel?>().apply {
    computeSourcePosition { sourcePosition ->
      GuiUtils.invokeLaterIfNeeded(Runnable {
        set(sourcePosition?.toModel(clientId, project))
      }, ModalityState.defaultModalityState())
    }
  }
}

fun XValueContainer.computeChildrenAsync(
  clientId: ClientId,
  project: Project
): RdTask<List<ValueEntityModelBase>> {
  return RdTask<List<ValueEntityModelBase>>().apply {
    computeChildren(CodeWithMeCompositeNode(clientId, this, project))
  }
}

fun XValueContainer.computeChildrenLiveAsync(
  clientId: ClientId,
  project: Project,
  valueLifetime: Lifetime
): RdTask<CodeWithMeChildrenListLive> {
  return RdTask<CodeWithMeChildrenListLive>().apply {
    computeChildren(CodeWithMeCompositeNodeLive(clientId, this, project, valueLifetime))
  }
}

private fun XValue.createValuePropertiesModel(clientId: ClientId, project: Project): RdTask<LiveProperties> {
  val properties = LiveProperties()
  computePresentation(CodeWithMeValueNode(clientId, properties.value, project), XValuePlace.TREE)
  return RdTask<LiveProperties>().apply { set(properties) }
}

private class CodeWithMeValueNode(val clientId: ClientId, val property: IOptProperty<ValuePropertiesModelBase>, private val project: Project) : XValueNode {
  private fun finishTask(icon: Icon?, type: String?, value: List<ValuePresentationPart>, hasChildren: Boolean) {
    val propertiesModel = CodeWithMeValuePropertiesModel(icon?.toProtocolModel(), type ?: "", value, hasChildren)
    ClientId.withClientId(clientId) {
      property.set(propertiesModel)
    }
  }

  override fun setPresentation(icon: Icon?, type: String?, value: String, hasChildren: Boolean) {
    finishTask(icon, type, listOf(ValuePresentationPart(value, ValuePresentationPartKind.Default)), hasChildren)
  }

  override fun setPresentation(icon: Icon?, presentation: XValuePresentation, hasChildren: Boolean) {
    val valueList = mutableListOf<ValuePresentationPart>()
    presentation.renderValue(object : XValuePresentation.XValueTextRenderer {
      override fun renderValue(value: String) {
        valueList.add(ValuePresentationPart(value, ValuePresentationPartKind.Default))
      }

      override fun renderStringValue(value: String) {
        valueList.add(ValuePresentationPart(value, ValuePresentationPartKind.String))
      }

      override fun renderNumericValue(value: String) {
        valueList.add(ValuePresentationPart(value, ValuePresentationPartKind.Numerical))
      }

      override fun renderKeywordValue(value: String) {
        valueList.add(ValuePresentationPart(value, ValuePresentationPartKind.Keyword))
      }

      override fun renderComment(comment: String) {
        valueList.add(ValuePresentationPart(comment, ValuePresentationPartKind.Comment))
      }

      override fun renderSpecialSymbol(symbol: String) {
        valueList.add(ValuePresentationPart(symbol, ValuePresentationPartKind.SpecialSymbol))
      }

      override fun renderError(error: String) {
        valueList.add(ValuePresentationPart(error, ValuePresentationPartKind.Error))
      }

      override fun renderValue(value: String, key: TextAttributesKey) =
        renderValue(value) // TODO

      override fun renderStringValue(value: String, additionalSpecialCharsToHighlight: String?, maxLength: Int) =
        renderStringValue(value) // TODO
    })
    finishTask(icon, presentation.type, valueList, hasChildren)
  }

  @Deprecated("Deprecated in XValueNode", ReplaceWith("setPresentation(Icon, XValuePresentation, boolean)"))
  override fun setPresentation(icon: Icon?, type: String?, separator: String, value: String?, hasChildren: Boolean) {
    setPresentation(icon, type, value ?: separator, hasChildren)
  }

  override fun setFullValueEvaluator(fullValueEvaluator: XFullValueEvaluator) {
    // TODO
  }
}