package com.jetbrains.rdserver.debugger

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.project.Project
import com.intellij.util.concurrency.annotations.RequiresEdt
import com.intellij.xdebugger.evaluation.XDebuggerEvaluator
import com.intellij.xdebugger.frame.XExecutionStack
import com.intellij.xdebugger.frame.XStackFrame
import com.intellij.xdebugger.frame.XValue
import com.jetbrains.codeWithMe.model.CodeWithMeStackFrameModel
import com.jetbrains.codeWithMe.model.ValueModelBase
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.toRdTask
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.toRdTextRange

@RequiresEdt
fun XStackFrame.createStackFrameModel(project: Project, clientId: ClientId, stack: XExecutionStack): CodeWithMeStackFrameModel {
  val xStackFrame = this
  val frameId = ClientId.withClientId(clientId) {
    BackendDebuggerHost.getInstance(project).getOrCreateFrameId(stack, xStackFrame)
  }
  val sourcePosition = xStackFrame.sourcePosition?.toModel(clientId, project)
  val presentation = CodeWithMeColoredTextContainer().apply {
    xStackFrame.customizePresentation(this)
  }.createModel(project)

  return CodeWithMeStackFrameModel(frameId, presentation, sourcePosition).apply {
    computeLocals.set { lifetime, _ ->
      xStackFrame.computeChildrenAsync(clientId, project)
    }

    evaluateExpression.set { _, arguments ->
      val evaluator = xStackFrame.getEvaluatorIfAvailable(project, clientId) {
        return@set RdTask.faulted(Exception(CodeWithMePlatformBundle.message("debugger.evaluation.no.access")))
      }
      if (evaluator == null) {
        return@set RdTask.faulted(Exception("Evaluator is null"))
      }
      
      RdTask<ValueModelBase>().apply rdTask@{
        evaluator.evaluate(arguments.expression, object : XDebuggerEvaluator.XEvaluationCallback {
          override fun errorOccurred(errorMessage: String) {
            application.invokeLater {
              result.set(RdTaskResult.Fault(Exception(errorMessage)))
            }
          }

          override fun evaluated(result: XValue) {
            application.invokeLater {
              set(result.createValueModel(clientId, arguments.expression, project))
            }
          }
        }, null)
      }
    }
    
    getExpressionInfoAtOffset.set { _, (documentId, offset, sideEffectsAllowed) ->
      val evaluator = xStackFrame.getEvaluatorIfAvailable(project, clientId) {
        return@set RdTask.faulted(Exception(CodeWithMePlatformBundle.message("debugger.evaluation.no.access")))
      }
      if (evaluator == null) {
        return@set RdTask.fromResult(null)
      }

      val document = BackendDocumentHost.getInstance(project).getDocument(documentId)
      evaluator.getExpressionInfoAtOffsetAsync(project, document, offset, sideEffectsAllowed)
        .then { info -> info.textRange.toRdTextRange() }
        .toRdTask()
    }
  }
}

private inline fun XStackFrame.getEvaluatorIfAvailable(project: Project, clientId: ClientId, blockIfNot: () -> Unit = {}): XDebuggerEvaluator? {
  val manager = PermissionManager.getInstance(project)
  if (!manager.hasFileWriteAccess(clientId)) {
    blockIfNot()
    return null
  }
  
  return evaluator
} 