package com.jetbrains.rdserver.diagnostics

import com.intellij.ide.actions.CollectZippedLogsAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.intellij.openapi.util.io.FileUtil
import com.intellij.openapi.util.text.StringUtil
import com.intellij.util.io.Compressor
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.diagnostics.CodeWithMeCollectZippedLogsActionBase
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.hasValue
import com.jetbrains.rd.util.threading.SpinWait
import com.jetbrains.rd.ide.model.FileTransmissionRequest
import com.jetbrains.codeWithMe.model.RdUserModel
import java.io.File
import java.util.*
import java.util.concurrent.CompletableFuture
import java.util.zip.ZipInputStream

class BackendCollectZippedLogsAction : CodeWithMeCollectZippedLogsActionBase(), DumbAware {
  companion object {
    private val LOG = logger<BackendCollectZippedLogsAction>()
  }

  override fun update(e: AnActionEvent) {
    val project = e.project ?: return
    val userManager = UserManager.getInstance(project)
    e.presentation.isEnabledAndVisible = userManager.users.isNotEmpty()
  }

  override fun collectLogs(project: Project?, indicator: ProgressIndicator, collectClientLogs: Boolean): File {
    val userLogs = if (collectClientLogs) {
      collectClientLogs(project!!, indicator)
    }
    else emptyList()

    return CollectZippedLogsAction.createZip(project) { compressor: Compressor ->
      for ((userModel, zipFile) in userLogs) {
        val prefix = "Guest-${FileUtil.sanitizeFileName(userModel.name)}/"

        ZipInputStream(zipFile.inputStream()).use { zip ->
          while (true) {
            val entry = zip.nextEntry ?: break

            val path = entry.name!!
            if (path.contains("..")) {
              LOG.warn(".. in path '$path', skipping zip item")
              continue
            }

            compressor.addFile(
              (prefix + path).replace('\\', '/'),
              zip,
              entry.lastModifiedTime.toMillis()
            )
          }
        }
      }
    }
  }

  private fun collectClientLogs(project: Project, indicator: ProgressIndicator): List<Pair<RdUserModel, File>> {
    val userLogs = Collections.synchronizedList(mutableListOf<Pair<RdUserModel, File>>())

    val toReceive = mutableListOf<Pair<RdUserModel, FileTransmissionRequest>>()

    val statusToShow = Collections.synchronizedList(mutableListOf<String>())
    fun addStatus(userModel: RdUserModel, status: String) {
      val message = "${userModel.name}: $status"
      LOG.info(message)
      statusToShow.add(message)
    }

    for (userModel in UserManager.getInstance(project).users.toList()) {
      indicator.text = "Asking ${userModel.name}"

      val diagnosticsHost = BackendDiagnosticsHost.getInstance(project, userModel.clientId)
      val logResult = diagnosticsHost.getLog()

      if (!SpinWait.spinUntil(Lifetime.Eternal, 60000) {
          logResult.hasValue
        }) {
        addStatus(userModel, "Timeout")
        continue
      }

      val logRequest = logResult.valueOrNull
      if (logRequest == null) {
        addStatus(userModel, "Not ready?")
      }
      else {
        when (logRequest) {
          is RdTaskResult.Success -> {
            val request = logRequest.value
            if (request == null) {
              addStatus(userModel, "Rejected by user")
            }
            else {
              toReceive.add(userModel to request)
              Unit
            }
          }
          is RdTaskResult.Cancelled -> {
            addStatus(userModel, "Cancelled")
          }
          is RdTaskResult.Fault -> {
            val errorMessage = "Error asking user ${userModel.name}: ${logRequest.error.message}"
            LOG.error(errorMessage, logRequest.error.cause)
            addStatus(userModel, "ERROR: ${logRequest.error.message}")
          }
        }.let { } // exhaustive when
      }
    }

    indicator.text = "Downloading files"

    val futures = toReceive.map { (userModel, transmissionRequest) ->
      val fileTransmissionHost = BackendFileTransmissionHost.getInstance(project, userModel.clientId)
      val future = fileTransmissionHost.receiveFileAsync(transmissionRequest)
      future.whenCompleteAsync { file, throwable ->
        if (file != null && throwable == null) {
          userLogs.add(userModel to file)
          LOG.info("${userModel.name}: logs were downloaded, size ${StringUtil.formatFileSize(file.length())}")
        }
        else {
          LOG.warn("Error downloading file $transmissionRequest from ${userModel.name} clientId ${userModel.clientIdValue}", throwable)
          addStatus(userModel, "Download ERROR: ${throwable?.message}")
        }
      }
      future
    }

    CompletableFuture.allOf(*futures.toTypedArray()).get()

    if (statusToShow.isNotEmpty()) {
      application.invokeAndWait {
        Messages.showInfoMessage(project, statusToShow.joinToString("\n"), CodeWithMePlatformBundle.message("product.name"))
      }
    }

    return userLogs
  }
}