@file:Suppress("RedundantLambdaArrow")

package com.jetbrains.rdserver.diagnostics

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.IdeBundle
import com.intellij.openapi.application.PathManager
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.IOptPropertyView
import com.jetbrains.rd.util.reactive.map
import com.jetbrains.rd.util.threading.SynchronousScheduler
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.codeWithMe.model.CodeWithMeDiagnosticsModel
import com.jetbrains.rd.ide.model.FileTransmissionRequest
import com.jetbrains.codeWithMe.model.LogFileData
import com.jetbrains.rdserver.unattendedHost.UnattendedHostManager
import java.io.File
import java.util.*

@Suppress("NonDefaultConstructor")
class BackendDiagnosticsHost(
  lifetime: Lifetime,
  project: Project,
  protocol: Protocol,
  private val clientId: ClientId
) {
  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendDiagnosticsHost =
      RdServerPerClientComponents.getService(project, clientId, BackendDiagnosticsHost::class.java)

    private val LOG = logger<BackendDiagnosticsHost>()
  }

  private val model = CodeWithMeDiagnosticsModel.create(lifetime, protocol)

  init {
    model.pingViaUiThread.set { _ -> Unit }
    model.pingDirect.set(SynchronousScheduler) { _ -> Unit }
    model.getLog.set { _, _ ->
      val rdTask = RdTask<LogFileData?>()

      LOG.info("A CWM client requested logs")

      val userModel = UserManager.getInstance(project).users.singleOrNull { it.clientId == clientId }
                      ?: error("Unknown client id in user model: $clientId")

      val unattendedHostManager = UnattendedHostManager.getInstance()
      if (!unattendedHostManager.isUnattendedMode) {
        ClientId.withClientId(ClientId.localId) {
          val result = Messages.showOkCancelDialog(
            project,
            IdeBundle.message("dialog.title.request.to.submit.logs", userModel.name, PathManager.getLogPath()),
            IdeBundle.message("dialog.title.submit.logs"),
            "Allow access to your logs", "Cancel", Messages.getWarningIcon()
          )
          if (result != Messages.OK) return@set RdTask.fromResult(null)
        }
      }
      else if (!unattendedHostManager.isControllerClientId(ClientId.current)) {
        // don't ask if non-controller is asking for logs on unattended host - there's nobody to see the UI anyway
        return@set RdTask.fromResult(null)
      }

      // initialize model on dispatch thread
      BackendFileTransmissionHost.getInstance(project, clientId)

      val fileFuture = BackendCollectZippedLogsAction().runAction(project, askAboutClients = false, askAboutSensitiveData = false)
      fileFuture.handleAsync { file: File?, throwable: Throwable? ->
        when {
          throwable != null -> rdTask.result.set(RdTaskResult.Fault(throwable))
          file != null -> {
            val transmissionHost = BackendFileTransmissionHost.getInstance(project, clientId)
            val request = FileTransmissionRequest("logs.zip", UUID.randomUUID())
            transmissionHost.sendFileAsync(file, request)
            rdTask.set(LogFileData(request.name, request.uuid))
          }
          // Cancelled
          else -> rdTask.set(null)
        }
      }

      return@set rdTask
    }
  }

  fun getLog(): IOptPropertyView<RdTaskResult<FileTransmissionRequest?>> =
    ClientId.withClientId(clientId) {
      model.getLog.start(Unit).result.map { result ->
        result.map { logFileData -> logFileData?.let { FileTransmissionRequest(it.name, it.uuid) } }
      }
    }
}