package com.jetbrains.rdserver.diagnostics

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.BrowserUtil
import com.intellij.ide.IdeBundle
import com.intellij.ide.actions.SendFeedbackAction
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.Task
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.Messages
import com.intellij.util.io.URLUtil
import com.jetbrains.rd.platform.diagnostics.SupportInformationUploader
import com.jetbrains.rd.platform.util.application

class BackendReportIssueAction : AnAction(), DumbAware {

  override fun update(e: AnActionEvent) {
    e.presentation.isEnabledAndVisible = e.project != null
  }

  override fun displayTextInToolbar() = true

  override fun actionPerformed(e: AnActionEvent) {
    val project = e.project ?: return
    reportIssue(project)
  }

  companion object {
    private const val cwmYoutrackUrlTemplate = "https://youtrack.jetbrains.com/newissue?project=CWM&clearDraft=true&description=\$DESCR"

    private val logger = Logger.getInstance(BackendReportIssueAction::class.java)

    fun reportIssue(project: Project) {
      application.assertIsDispatchThread()

      ClientId.withClientId(ClientId.localId) {
        val result = Messages.showYesNoDialog(
          project, IdeBundle.message("message.included.logs.and.settings.may.contain.sensitive.data"),
          IdeBundle.message("dialog.title.sensitive.data"), Messages.getWarningIcon())
        if (result != Messages.YES) return
      }

      ProgressManager.getInstance().run(
        object : Task.Backgroundable(null, IdeBundle.message("progress.title.collecting.logs"), true) {
          override fun run(indicator: ProgressIndicator) {
            val file = BackendCollectZippedLogsAction().collectLogs(project, indicator, true)

            indicator.text = "Uploading logs..."
            val uploadedLogsUrl = SupportInformationUploader.uploadFile(file, "cwm-" + file.name, indicator)

            logger.info("Uploaded logs to $uploadedLogsUrl")

            val navigateToUrl = getYouTrackUrl(
              "\n\nLogs (visible to JetBrains only): $uploadedLogsUrl\n\n" +
              SendFeedbackAction.getDescription(project).trim()
            )

            logger.info("Prepared a youtrack url for reporting issue: $navigateToUrl")

            BrowserUtil.browse(navigateToUrl, project)
          }
        })
    }

    fun getYouTrackUrl(description: String): String {
      val replaced = cwmYoutrackUrlTemplate.replace("\$DESCR", URLUtil.encodeURIComponent(description))
      if (replaced.contains("\$")) {
        logger.error("Some values were not substituted: $replaced")
      }
      return replaced
    }
  }
}