package com.jetbrains.rdserver.document

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Document
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.util.messages.MessageBus
import com.jetbrains.rd.ide.document.*
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.tests.isDistributedTestMode
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.usingNullableFlag
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.vfs.BackendVirtualFileSystemHost
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.requests.BackendIdeGateway
import java.util.*

@Suppress("NonDefaultConstructor")
class BackendDocumentHost(
  lifetime: Lifetime,
  private val clientId: ClientId,
  solution: Solution,
  private val project: Project,
  private val bus: MessageBus) : ProtocolDocumentHost {
  companion object {
    fun getInstance(project: Project): BackendDocumentHost {
      return RdServerPerClientComponents.getService(project)
    }

    private val LOG = getLogger<BackendDocumentHost>()
  }

  private val documentVersions = BackendDocumentVersions()
  private val documentLogManager = BackendDocumentLogManager()

  private val idToSynchronizersMap = hashMapOf<RdDocumentId, ProtocolDocumentSynchronizer>()
  private val documentToSynchronizersMap = hashMapOf<Document, ProtocolDocumentSynchronizer>()

  private var currentlyBindingDocument: Document? = null
  private val operationModel = solution.documentsOperationModel

  init {
    operationModel.documents.view(lifetime, this::synchronizeDocument)
  }

  private fun synchronizeDocument(lifetime: Lifetime, documentId: RdDocumentId, documentModel: RdDocumentModel) {
    val document = currentlyBindingDocument ?: findOrCreateDocument(documentId, documentModel)
    if (document == null) {
      application.invokeLater {
        if (lifetime.isNotAlive) return@invokeLater
        operationModel.documents.remove(documentId)
      }
      return
    }

    val documentLog = documentLogManager.register(lifetime, clientId, documentId, InitialDocVersion)
    val versionHolder = documentVersions.createDocumentVersionHolder(lifetime, documentId)
    val ideGateway = BackendIdeGateway.getInstance(project)

    val documentSynchronizer = ProtocolDocumentSynchronizer(
      project, document, documentModel, documentId,
      versionHolder, documentLog, ideGateway, isDistributedTestMode, false
    )
    documentSynchronizer.sendInitialTextChange()

    lifetime.onTerminationIfAlive {
      Disposer.dispose(documentSynchronizer)
    }
    documentToSynchronizersMap.addUnique(lifetime, document, documentSynchronizer)
    idToSynchronizersMap.addUnique(lifetime, documentId, documentSynchronizer)
    bus.syncPublisher(BackendDocumentHostListener.TOPIC).documentBound(lifetime, documentId, documentModel, document)
  }

  fun applyChange(change: RdDocumentChange) {
    val documentId = change.documentId
    val documentSynchronizer = tryGetSynchronizer(documentId) ?: error("document synchronizer is null")
    documentSynchronizer.receiveDocumentChange(change)
  }

  fun tryGetSynchronizer(document: Document): ProtocolDocumentSynchronizer? {
    return documentToSynchronizersMap[document]
  }

  fun tryGetSynchronizer(documentId: RdDocumentId): ProtocolDocumentSynchronizer? {
    return idToSynchronizersMap[documentId]
  }

  fun getDocument(documentId: RdDocumentId): Document {
    return tryGetSynchronizer(documentId)!!.document
  }

  private fun findOrCreateDocument(documentId: RdDocumentId, documentModel: RdDocumentModel): Document? {
    val documentValue = BackendDocumentProvider.EP_NAME.computeSafeIfAny(project) l@{
      if (!it.accept(documentId, documentModel)) return@l null
      return@l (it.document(documentId, documentModel) to it)
    }

    if (documentValue == null) {
      LOG.warn("No document providers found for ${documentId.toDebugString()}")
      return null
    }

    val (document, provider) = documentValue

    if (document == null) {
      LOG.debug("Suitable provider refused from a document. Provider: $provider, document: ${documentId.toDebugString()}")
      return null
    }

    return document
  }

  override fun createId(document: Document): RdDocumentId {
    val file = FileDocumentManager.getInstance().getFile(document)
    if (file != null) {
      val fileId = BackendVirtualFileSystemHost.getInstance(project).bindFile(file)
      require(fileId != null) { "not enough permissions to open ${file.name}" }
      return RdDocumentIdWithFileId(fileId)
    }
    return super.createId(document)
  }

  fun bindDocument(document: Document): RdDocumentId {
    tryGetSynchronizer(document)?.let { return it.documentId }

    val documentId = createId(document)
    require(PermissionManager.getInstance(project).isVisible(clientId, document)) {
      "Permission denied clientId=$clientId documentId=${documentId.toDebugString()}"
    }

    usingNullableFlag(this::currentlyBindingDocument, document) {
      val extensions = BackendDocumentExtensionsProvider.EP_NAME.getExtensions(project).flatMap { it.getExtensions(document) }
      val documentModel = RdDocumentModel(null, extensions)

      val file = FileDocumentManager.getInstance().getFile(document)
      if (file != null) {
        // send file content so it will be available right at the moment of document text loading on the frontend
        BackendVirtualFileSystemHost.getInstance(project).bindFile(file, sendProperties = true, sendContent = true)
      }
      operationModel.documents[documentId] = documentModel
      return documentId
    }
  }
}