package com.jetbrains.rdserver.document

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.clipboard.RemoteClipboard
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.requests.BackendAsyncRequestDelegate

private val logger = logger<BackendPatchExecutorDelegate>()
private const val retryMessage = "not matched expected versions"

class BackendPatchExecutorDelegate : BackendAsyncRequestDelegate<RdApplyPathRequest> {
  override suspend fun execute(request: RdApplyPathRequest, project: Project, consumer: (RdResponse) -> Unit) {
    val patch = request.patch
    val documentHost = BackendDocumentHost.getInstance(project)
    val editorHost = BackendEditorHost.getInstance(project)
    val remoteClipboard = RdServerPerClientComponents.getService<RemoteClipboard>(project)
    val response = applyClientPatch(patch, request.id, documentHost, editorHost, remoteClipboard)
    consumer(response)
  }

  private fun applyClientPatch(patch: RdPatch, requestId: Int, documentHost: BackendDocumentHost, editorHost: BackendEditorHost, remoteClipboard: RemoteClipboard): RdResponse {
    require(patch.features == null)

    if (!checkExpectedVersions(patch, documentHost, editorHost, remoteClipboard)) {
      return RdRetryResponse(retryMessage, requestId)
    }

    editorHost.setEditorSettings(patch) {
      for (change in patch.changes) {
        when (change) {
          is RdDocumentChange -> documentHost.applyChange(change)
          is RdSelectionChangeBase -> editorHost.applyChange(change)
          is RdClipboardChange -> remoteClipboard.applyChange(change)
          else -> logger.error("Unexpected change type from client: ${change::class.simpleName}")
        }
      }
    }

    if (!checkUpdatedVersions(patch, documentHost, editorHost, remoteClipboard)) {
      logger.error("Not matched version after applying guest patch")
    }

    return RdSuccessResponse(requestId)
  }

  private fun checkExpectedVersions(patch: RdPatch, documentHost: BackendDocumentHost, editorHost: BackendEditorHost, remoteClipboard: RemoteClipboard) : Boolean {
    return checkVersions(patch.expectedDocIds, patch.expectedDocVersions,
                         patch.expectedEditorIds, patch.expectedEditorVersions, patch.expectedClipboardVersion,
                         documentHost, editorHost, remoteClipboard)
  }

  private fun checkUpdatedVersions(patch: RdPatch, documentHost: BackendDocumentHost, editorHost: BackendEditorHost, remoteClipboard: RemoteClipboard) : Boolean {
    return checkVersions(patch.updatedDocIds, patch.updatedDocVersions,
                         patch.updatedEditorIds, patch.updatedEditorVersions, patch.updatedClipboardVersion,
                         documentHost, editorHost, remoteClipboard)
  }

  private fun checkVersions(docIds: List<RdDocumentId>, docVersions: List<RdDocumentVersion>,
                            editorIds: List<TextControlId>, editorVersions: List<RdEditorVersion>, clipboardVersion: Int?,
                            documentHost: BackendDocumentHost, editorHost: BackendEditorHost, remoteClipboard: RemoteClipboard): Boolean {
    for ((i, id) in docIds.withIndex()) {
      val actualVersion = documentHost.tryGetSynchronizer(id)?.version ?: return false
      val expectedVersion = docVersions[i]
      if (actualVersion != expectedVersion)
        return false
    }

    for ((i, id) in editorIds.withIndex()) {
      val actualVersion = editorHost.tryGetSynchronizer(id)?.version ?: return false
      val expectedVersion = editorVersions[i]
      if (actualVersion != expectedVersion)
        return false
    }
    val expectedClipboardVersion = clipboardVersion ?: return true
    if (expectedClipboardVersion != remoteClipboard.version) {
      return false
    }

    return true
  }
}

private inline fun BackendEditorHost.setEditorSettings(patch: RdPatch, crossinline block: () -> Unit) {
  try {
    if (patch.disableCaretShiftOnWhitespace) {
      for (editorId in patch.expectedEditorIds) {
        val editor = tryGetOwnRemoteEditor(editorId) ?: continue
        editor.putUserData(com.intellij.openapi.editor.impl.EditorImpl.DISABLE_CARET_SHIFT_ON_WHITESPACE_INSERTION, true)
      }
    }

    block()
  } finally {
    if (patch.disableCaretShiftOnWhitespace) {
      for (editorId in patch.expectedEditorIds) {
        val editor = tryGetOwnRemoteEditor(editorId) ?: continue
        editor.putUserData(com.intellij.openapi.editor.impl.EditorImpl.DISABLE_CARET_SHIFT_ON_WHITESPACE_INSERTION, null)
      }
    }
  }
}