// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.impl.ScrollingModelImpl
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.fileEditor.impl.text.TextEditorProvider
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Pair
import com.intellij.openapi.util.ProperTextRange
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.util.messages.MessageBus
import com.jetbrains.rd.framework.util.setSuspend
import com.jetbrains.rd.ide.document.toDebugString
import com.jetbrains.rd.ide.editor.toDebugString
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.debug
import com.jetbrains.rd.platform.util.idea.toEditorKind
import com.jetbrains.rd.platform.util.idea.toLocation
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.platform.util.userData
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.getOrCreate
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rd.util.reflection.usingTrueFlag
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.daemon.markupModelContributor
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.document.BackendDocumentHostListener
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.requests.BackendIdeGateway
import org.jetbrains.annotations.TestOnly

@Suppress("NonDefaultConstructor")
class BackendEditorHost(
  private val clientId: ClientId,
  private val lifetime: Lifetime,
  private val project: Project,
  solution: Solution,
  private val bus: MessageBus
) {
  companion object {
    val LOG = logger<BackendEditorHost>()

    fun getInstance(project: Project): BackendEditorHost = RdServerPerClientComponents.getService(project)

    // important to keep in the editor, so other BackendEditorHosts won't try to synchronize it as their own
    var Editor.synchronizer: BackendEditorSynchronizer? by userData()
    var Editor.textControlModel: TextControlModel? by userData()
  }

  val documentsOperationModel = solution.documentsOperationModel

  private var openingFileEditor = false
  private var focusingEditor = false
  private var currentlyBindingEditor: Editor? = null

  private val idToSynchronizersMap = hashMapOf<TextControlId, BackendEditorSynchronizer>()

  private val fileEditorManager = FileEditorManager.getInstance(project) as BackendServerFileEditorManager
  private val documentHost = BackendDocumentHost.getInstance(project)
  private val ideGateway = RdServerPerClientComponents.getService<BackendIdeGateway>(project)

  init {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(BackendDocumentHostListener.TOPIC, MyDocumentHostListener())
    connection.subscribe(FileEditorManagerListener.FILE_EDITOR_MANAGER, MyFileEditorManagerListener())

    val model = solution.documentsOperationModel
    model.codeWithMeFocusesModel.lastFocusedTextControl.change.adviseNotNull(lifetime) { reportFocusFromModel(it) }
    model.synchronizationModel.getEditorStates.setSuspend { _, ids -> getEditorStates(ids) }
    model.isReady.set(true)
  }

  private inner class MyDocumentHostListener : BackendDocumentHostListener {
    override fun documentBound(
      lifetime: Lifetime,
      documentId: RdDocumentId,
      model: RdDocumentModel,
      document: Document
    ) {
      if (ClientId.current != clientId) return

      model.textControls.view(lifetime) editors@{ textControlModelLifetime, textControlId, textControlModel ->
        val editor = ensureBackendEditorCreated(textControlModelLifetime, textControlId)
        //this editor is already bound with this model. return
        if (editor.getSynchronizer() != null) return@editors

        val watchLifetimeDef = textControlModelLifetime.createNested()
        attachEditorListeners(watchLifetimeDef, editor, textControlId, textControlModel)
      }
    }
  }

  private inner class MyFileEditorManagerListener : FileEditorManagerListener {
    override fun fileOpenedSync(
      source: FileEditorManager,
      file: VirtualFile,
      editors: Pair<Array<FileEditor>, Array<FileEditorProvider>>
    ) {
      for (fileEditor in editors.first) {
        if (fileEditor is TextEditor) {
          reportToModel(file, fileEditor)
        }
      }
    }

    override fun fileOpened(source: FileEditorManager, file: VirtualFile) {
      val fileEditor = source.getSelectedEditor(file)
      if (fileEditor is TextEditor) {
        reportToModel(file, fileEditor)
      }
    }

    override fun selectionChanged(event: FileEditorManagerEvent) {
      val file = event.newFile ?: return
      val textEditor = event.newEditor as? TextEditor ?: return
      reportToModel(file, textEditor)
    }
  }

  fun viewLastFocusedTextControl(lifetime: Lifetime, handler: (Lifetime, TextControlId) -> Unit) {
    documentsOperationModel.codeWithMeFocusesModel.lastFocusedTextControl.view(lifetime) { focusedLt, id ->
      if (id != null)
        handler(focusedLt, id)
    }
  }

  private fun reportToModel(file: VirtualFile, textEditor: TextEditor) {
    if (ClientId.current != clientId) return
    if (textEditor.clientId() != clientId) return
    if (openingFileEditor) return

    val denied = !PermissionManager.getInstance(project).isVisible(clientId, file)
    LOG.info("Reporting of editor${if (denied) " denied (no permissions)" else ""}. " +
             "File $file, editor: $textEditor, clientId: $clientId")
    if (denied) return

    val editor = textEditor.editor
    if (editor.getSynchronizer() == null) {
      bindEditor(editor)
    }
    else {
      reportFocusToModel(editor)
    }
  }

  private fun reportFocusFromModel(textControlId: TextControlId) {
    val documentId = textControlId.documentId
    val document = documentHost.getDocument(documentId)
    val virtualFile = FileDocumentManager.getInstance().getFile(document)

    if (virtualFile == null) {
      LOG.info("Trying to focus editor for non-existing virtual file: ${documentId.toDebugString()}")
      return
    }
    noReenter(this::focusingEditor) {
      val editor = tryGetOwnRemoteEditor(textControlId) ?: return
      fileEditorManager.selectRemoteTextEditor(virtualFile, editor)
    }
  }

  private fun reportFocusToModel(selectedEditor: Editor) {
    noReenter(this::focusingEditor) {
      //do not report if there is no model yet
      val textControlId = selectedEditor.getTextControlId() ?: return@noReenter

      LOG.info("[$clientId] navigating to $textControlId")
      documentsOperationModel.navigateFromBackend.fire(textControlId)
    }
  }

  fun tryGetOwnRemoteEditor(textControlId: TextControlId): Editor? {
    return tryGetSynchronizer(textControlId)?.editor
  }

  fun tryGetSynchronizer(textControlId: TextControlId): BackendEditorSynchronizer? {
    return idToSynchronizersMap[textControlId]
  }

  private fun ensureBackendEditorCreated(lifetime: Lifetime, textControlId: TextControlId): Editor {
    require(textControlId.clientId == clientId.value && ClientId.current == clientId) {
      "ClientId mismatch. Current: ${ClientId.current}, Host: $clientId, Id: ${textControlId.clientId}"
    }

    val bindingEditor = currentlyBindingEditor
    if (bindingEditor != null) {
      require(bindingEditor.document == documentHost.getDocument(textControlId.documentId)) {
        "Simultaneous binding of editors for different documents"
      }
    }

    return if (textControlId.kind == TextControlLocation.EditorTab) {
      findOrOpenFileEditor(lifetime, textControlId)
    }
    else {
      // don't manage editor closing for general purpose editors created not by us (i.e. bindingEditor != null)
      bindingEditor ?: openEditor(lifetime, textControlId)
    }
  }

  /**
   * Open a file editor in 'tab'
   */
  private fun findOrOpenFileEditor(lifetime: Lifetime, textControlId: TextControlId): Editor {
    val bindingEditor = currentlyBindingEditor
    val document = documentHost.getDocument(textControlId.documentId)
    val file = FileDocumentManager.getInstance().getFile(document) ?: error("cannot find file for document $textControlId")

    if (bindingEditor == null) {
      val existingBackendEditor = tryFindExistingFileEditor(file, textControlId)
      if (existingBackendEditor != null) return existingBackendEditor
    }

    val editor = bindingEditor ?: usingTrueFlag(BackendEditorHost::openingFileEditor) {
      //not found (might be splitters case) -> force create new editorImpl
      LOG.info("Opening editor for ${file.canonicalFile}")
      fileEditorManager.forceCreateRemoteTextEditor(file)
    }

    if (editor.getSynchronizer() == null) {
      // pass closing to fileEditorManager even in case the editor was created not by us (i.e. bindingEditor != null)
      lifetime.onTermination {
        noReenter(file) {
          fileEditorManager.enableClosingRemoteEditor {
            LOG.info("Closing file editor for $textControlId")
            fileEditorManager.closeRemoteTextEditor(file, editor)
          }
        }
      }
    }
    return editor
  }

  /**
   * Open a general purpose editor (e.g. preview for find in path, or inside a tool window, debugger watches, etc.)
   */
  private fun openEditor(lifetime: Lifetime, textControlId: TextControlId): Editor {
    val factory = EditorFactory.getInstance()
    val document = documentHost.getDocument(textControlId.documentId)
    val kind = textControlId.kind.toEditorKind()
    val file = FileDocumentManager.getInstance().getFile(document)
    val editor = if (file != null) {
      factory.createEditor(document, project, file, false, kind)
    }
    else {
      factory.createEditor(document, project, kind)
    }
    lifetime.onTermination { EditorFactory.getInstance().releaseEditor(editor) }
    return editor
  }

  private fun tryFindExistingFileEditor(
    targetFile: VirtualFile,
    textControlId: TextControlId
  ): Editor? {
    for (fileEditor in fileEditorManager.getEditors(targetFile)) {
      val editor = (fileEditor as? TextEditor)?.editor ?: continue
      if (editor.getTextControlId() == textControlId) return editor

      if (textControlId.tabIndex == 1 && textControlId.kind == TextControlLocation.EditorTab) {
        //any bug here will cause another EditorImpl to be created which is very hard to diagnose. So important assert here.
        LOG.error("we've found an editor for virtualFile: ${targetFile} " +
                  "but its textControlId ${editor.getTextControlId()} doesn't match the backend's one: $textControlId")
      }
    }

    return null
  }

  fun bindEditor(editor: Editor): TextControlId {
    require(ClientId.current == clientId && editor.clientId() == clientId) {
      "ClientId mismatch. Current: ${ClientId.current}, Host: $clientId, Editor: ${editor.clientId()}"
    }

    val existingSynchronizer = editor.synchronizer
    if (existingSynchronizer != null) {
      val existingTextControlId = existingSynchronizer.textControlId
      LOG.warn("Trying to bind an already bound editor. ${existingTextControlId.toDebugString()}")
      return existingTextControlId
    }

    val documentId = documentHost.bindDocument(editor.document)
    val documentModel = documentHost.tryGetSynchronizer(documentId)!!.modelDocument
    val location = editor.editorKind.toLocation()
    val tabIndex = if (location == TextControlLocation.EditorTab) {
      documentModel.textControls.keys
        .filter { it.kind == location }
        .maxOfOrNull { it.tabIndex } ?: 1
    }
    else {
      documentModel.textControls.keys
        .filter { it.documentId == documentId && it.kind == location && it.clientId == clientId.value }
        .maxOfOrNull { it.tabIndex + 1 } ?: 1
    }
    val textControlId = TextControlId(documentId, tabIndex, location, clientId.value)

    try {
      require(currentlyBindingEditor == null) {
        "trying to bind $editor during binding of $currentlyBindingEditor"
      }

      currentlyBindingEditor = editor
      documentModel.textControls.getOrCreate(textControlId) {
        LOG.info("Adding ${textControlId.toDebugString()} to the model. clientId: ${ClientId.current}")

        TextControlModel(
          documentId,
          OriginKind.Backend,
          true
        ).apply {
          isVisible.set(true)
          isReadonly.set(editor.isViewer)
        }
      }
      return textControlId
    }
    finally {
      currentlyBindingEditor = null
    }
  }

  private fun attachEditorListeners(
    editorLifetime: Lifetime,
    editor: Editor,
    textControlId: TextControlId,
    textControlModel: TextControlModel
  ) {
    require(textControlId.clientId == clientId.value && ClientId.current == clientId) {
      "ClientId mismatch. Current: ${ClientId.current}, Host: $clientId, Id: ${textControlId.clientId}"
    }

    val textControlDisposable = editorLifetime.createNestedDisposable("attachEditorListeners: $textControlModel")
    val editorSynchronizer = createSynchronizer(editor, textControlId)
    idToSynchronizersMap.addUnique(editorLifetime, textControlId, editorSynchronizer)
    editor.textControlModel = textControlModel
    editor.synchronizer = editorSynchronizer
    editor.remoteEditorData = RemoteEditorData()

    editorLifetime.onTermination {
      editor.textControlModel = null
      editor.synchronizer = null
      editor.remoteEditorData = null
    }

    textControlModel.visibleRange.advise(editorLifetime) {
      val textRange = ProperTextRange(it.startOffset, it.endOffset)
      editor.remoteEditorData!!.visibleRange.set(textRange)
    }
    editor.selectionModel.addSelectionListener(editorSynchronizer, textControlDisposable)
    editor.caretModel.addCaretActionListener(editorSynchronizer, textControlDisposable)
    editor.caretModel.addCaretListener(editorSynchronizer, textControlDisposable)
    editor.document.addDocumentListener(editorSynchronizer, textControlDisposable)

    (editor.scrollingModel as ScrollingModelImpl).addScrollRequestListener(editorSynchronizer, textControlDisposable)
    textControlModel.isVisible.advise(editorLifetime) { visible ->
      editor.putUserData(RemoteEditors.REMOTELY_VISIBLE, if (visible) true else null)
    }

    bus.syncPublisher(BackendEditorHostListener.TOPIC).editorBound(editorLifetime, textControlModel, editor)
  }

  private fun createSynchronizer(editor: Editor, textControlId: TextControlId): BackendEditorSynchronizer {
    require(textControlId.clientId == clientId.value && ClientId.current == clientId) {
      "ClientId mismatch. Current: ${ClientId.current}, Host: $clientId, Id: ${textControlId.clientId}"
    }

    val documentSynchronizer = documentHost.tryGetSynchronizer(editor.document) ?: error("Document synchronizer is null")
    val editorSynchronizer = BackendEditorSynchronizer(editor, textControlId, documentSynchronizer, ideGateway)
    editorSynchronizer.sendInitialCaretState()
    return editorSynchronizer
  }

  fun applyChange(change: RdSelectionChangeBase) {
    val editorId = change.editorId

    val synchronizer = tryGetSynchronizer(editorId)
    requireNotNull(synchronizer) { "cannot find synchronizers for $editorId" }
    synchronizer.receiveSelection(change)
  }

  private suspend fun getEditorStates(ids: List<TextControlId>): RdPatch {
    LOG.debug { ">> Star recovering editor states after exception; client: $clientId" }
    try {
      val patch = ideGateway.executeBulkUpdate {
        for (id in ids) {
          val documentId = id.documentId
          val documentSynchronizer = documentHost.tryGetSynchronizer(documentId) ?: continue
          documentSynchronizer.sendInitialTextChange()
          val editorSynchronizer = tryGetSynchronizer(id) ?: continue
          editorSynchronizer.sendInitialCaretState()
        }
      }

      application.invokeLater {
        if (lifetime.isNotAlive) return@invokeLater
        val documents = hashSetOf<Document>()
        val editors = hashSetOf<Editor>()
        for (id in ids) {
          val documentId = id.documentId
          val documentSynchronizer = documentHost.tryGetSynchronizer(documentId) ?: continue
          documents.add(documentSynchronizer.document)
          val editorSynchronizer = tryGetSynchronizer(id) ?: continue
          editors.add(editorSynchronizer.editor)
        }
        for (document in documents) {
          document.markupModelContributor?.recoverMarkupState()
        }
        for (editor in editors) {
          editor.markupModelContributor?.recoverMarkupState()
        }
      }
      return patch!!
    }
    finally {
      LOG.debug { "<< End recovering editor states" }
    }
  }

  @TestOnly
  fun reportFocusedTextControl() {
    val selectedEditor = (fileEditorManager.selectedEditor as? TextEditor)?.editor ?: error("no selected editor found")
    reportFocusToModel(selectedEditor)
  }
}

class TextBackendFileEditorHandler : BackendFileEditorHandler {
  override fun isAvailable(project: Project, file: VirtualFile, provider: FileEditorProvider): Boolean = provider is TextEditorProvider

  override fun createFollowNavigatable(project: Project, editor: FileEditor): FileEditorNavigatable? {
    if (editor is TextEditor) {
      val fileToNavigate = editor.file ?: return null
      val offset = editor.editor.caretModel.offset
      return OpenFileDescriptor(project, fileToNavigate, offset)
    }
    return null
  }
}
