package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Pair
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.rd.platform.codeWithMe.editors.CodeWithMeEditorService
import com.jetbrains.rd.platform.codeWithMe.users.broadcastToAllGuests
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.startup.serverSolution

class BackendEditorService(private val project: Project) : CodeWithMeEditorService(project) {

  private val queue = MergingUpdateQueue("queue", 50, true, MergingUpdateQueue.ANY_COMPONENT, project)

  private val clientToActiveFileMap = hashMapOf<ClientId, VirtualFile>()

  init {
    val connection = project.messageBus.connect(project.lifetime.createNestedDisposable())
    connection.subscribe(FileEditorManagerListener.FILE_EDITOR_MANAGER, object : FileEditorManagerListener {
      override fun fileOpenedSync(
        source: FileEditorManager,
        file: VirtualFile,
        editors: Pair<Array<FileEditor>, Array<FileEditorProvider>>
      ) {
        val document = FileDocumentManager.getInstance().getDocument(file) ?: return
        val clientIds = FileEditorManager.getInstance(project)
        .allEditors
        .mapNotNull { (it as? TextEditor)?.editor }
        .filter { it.document == document }
        .map { it.clientId() }.toHashSet()

        for (clientId in clientIds + ClientId.current) {
          serverNotify(document, clientId)
        }
      }

      override fun selectionChanged(event: FileEditorManagerEvent) {
        val file = event.newEditor?.file ?: return
        val document = FileDocumentManager.getInstance().getDocument(file) ?: return
        serverNotify(document, ClientId.current)
      }

      override fun fileClosed(source: FileEditorManager, file: VirtualFile) {
        val currentClientId = ClientId.current
        val virtualFile = clientToActiveFileMap[currentClientId]
        if (virtualFile != file) return

        clientToActiveFileMap.remove(currentClientId)
        queue.queue(object : Update(currentClientId) {
          override fun run() {
            broadcastToAllGuests(project) {
              val operationModel = project.serverSolution.documentsOperationModel
              if (operationModel.activeClients.containsKey(currentClientId.value))
                operationModel.activeClients.remove(currentClientId.value)
            }
          }
        })
      }
    })
  }

  override fun serverNotify(document: Document, clientId: ClientId) {
    application.assertIsDispatchThread()

    val virtualFile = FileDocumentManager.getInstance().getFile(document) ?: return
    clientToActiveFileMap[clientId] = virtualFile

    queue.queue(object : Update(clientId) {
      override fun run() {
        broadcastToAllGuests(project) {
          val activeClients = project.serverSolution.documentsOperationModel.activeClients
          val synchronizer = BackendDocumentHost.getInstance(project).tryGetSynchronizer(document)
          if (synchronizer != null)
            activeClients[clientId.value] = synchronizer.documentId
          else
            activeClients.remove(clientId.value)
        }
      }
    })
  }

  override fun getActiveClients(editor: Editor): List<ClientId> {
    val virtualFile = FileDocumentManager.getInstance().getFile(editor.document) ?: return emptyList()
    return clientToActiveFileMap.filter { it.value == virtualFile }.map { it.key }
  }
}