package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.*
import com.intellij.openapi.editor.event.*
import com.intellij.openapi.editor.impl.ScrollRequestListener
import com.jetbrains.rd.ide.document.ProtocolDocumentSynchronizer
import com.jetbrains.rd.ide.editor.*
import com.jetbrains.rd.util.reflection.usingTrueFlag
import com.jetbrains.rdserver.requests.BackendIdeGateway
import com.jetbrains.rd.ide.model.*
import java.util.concurrent.atomic.AtomicLong

private val logger = logger<BackendEditorSynchronizer>()

class BackendEditorSynchronizer(
  val editor: Editor,
  val textControlId: TextControlId,
  private val documentSynchronizer: ProtocolDocumentSynchronizer,
  private val ideGateway: BackendIdeGateway
) : CaretListener, CaretActionListener, SelectionListener, ScrollRequestListener, DocumentListener {
  private val versionHolder = EditorVersionHolder(InitialEditorVersion)
  private val idNumberCounter = AtomicLong()
  private var oldState: RdCaretState
  private var dontSendChange: Boolean = false
  private var isInsideAllCaretsAction: Boolean = false
  private val documentVersion get() = documentSynchronizer.version

  init {
    for (caret in editor.caretModel.allCarets) {
      caret.id = createNextId(isClient = true)
    }
    oldState = editor.caretModel.createModelCaretState()
  }

  private fun createNextId(isClient: Boolean): RdCaretId {
    val number = idNumberCounter.incrementAndGet()
    return RdCaretId(isClient, number)
  }

  override fun caretAdded(e: CaretEvent) {
    diagnostics("caretAdded") {
      if (!dontSendChange) {
        // Note: if the control flow is here then a caret was added by a server
        e.caret!!.id = createNextId(isClient = false)
      }
      sendCaretStateChange()
    }
  }

  override fun caretRemoved(e: CaretEvent) {
    diagnostics("caretRemoved") {
      sendCaretStateChange()
    }
  }

  override fun caretPositionChanged(e: CaretEvent) {
    diagnostics("caretPositionChanged") {
      sendCaretStateChange()
    }
  }

  override fun selectionChanged(e: SelectionEvent) {
    diagnostics("selectionChanged") {
      sendCaretStateChange()
    }
  }

  override fun beforeAllCaretsAction() {
    sendCaretStateSnapshot(force = true)
    isInsideAllCaretsAction = true
  }

  override fun afterAllCaretsAction() {
    isInsideAllCaretsAction = false
    sendCaretStateSnapshot(force = true)
  }

  override fun beforeDocumentChange(event: DocumentEvent) {
    diagnostics("beforeDocumentChanged") {
      sendCaretStateSnapshot()
    }
  }

  override fun documentChanged(event: DocumentEvent) {
    diagnostics("documentChanged") {
      sendCaretStateSnapshot()
    }
  }

  fun sendInitialCaretState() {
    val snapshot = RdCaretStateSnapshot(oldState, textControlId)
    ideGateway.sendEditorChange(snapshot, versionHolder.version, versionHolder.incrementVersion(), documentVersion)
  }

  private fun sendCaretStateChange() {
    sendCaretState(SendingMode.CHANGE, force = false)
  }

  private fun sendCaretStateSnapshot(force: Boolean = false) {
    sendCaretState(SendingMode.SNAPSHOT, force)
  }

  private enum class SendingMode {
    SNAPSHOT,
    CHANGE,
  }

  private fun sendCaretState(mode: SendingMode, force: Boolean) {
    if (dontSendChange || isInsideAllCaretsAction) {
      return
    }
    val newState = editor.caretModel.createModelCaretState()
    if (newState == oldState && !force) {
      return
    }
    if (documentSynchronizer.isReceivingChange) {
      oldState = newState
      return
    }
    val change = when (mode) {
      SendingMode.CHANGE -> RdCaretStateChange(oldState, newState, textControlId)
      SendingMode.SNAPSHOT -> RdCaretStateSnapshot(newState, textControlId)
    }
    oldState = newState
    val oldVersion = versionHolder.version
    val newVersion = versionHolder.incrementVersion()
    ideGateway.sendEditorChange(change, oldVersion, newVersion, documentVersion)
  }

  fun receiveSelection(change: RdSelectionChangeBase) {
    when (change) {
      is RdCaretStateChange -> receiveCaretStateChange(change)
      is RdCaretStateSnapshot -> receiveCaretStateSnapshot(change)
      else -> error("Unexpected change type: ${change::class.simpleName}")
    }
  }

  private fun receiveCaretStateSnapshot(snapshot: RdCaretStateSnapshot) {
    usingTrueFlag(BackendEditorSynchronizer::dontSendChange) {
      receiveCaretState(snapshot.state)
    }
  }

  private fun receiveCaretStateChange(change: RdCaretStateChange) {
    usingTrueFlag(BackendEditorSynchronizer::dontSendChange) {
      if (change.oldState != oldState) {
        logger.error("expected oldState: ${change.oldState.toDebugString()}, actual: ${oldState.toDebugString()}")
      }
      receiveCaretState(change.newState)
    }
  }

  private fun receiveCaretState(newState: RdCaretState) {
    applyStateTo(newState, editor)
    versionHolder.incrementVersion()
    oldState = editor.caretModel.createModelCaretState()
  }

  override fun scrollRequested(targetPosition: LogicalPosition, scrollType: ScrollType) {
    val scrollingInfo = ScrollingInfo(scrollType.toScrollKind(), editor.logicalPositionToOffset(targetPosition))
    val editorVersion = versionHolder.version
    ideGateway.sendEditorChange(RdScrollingRequest(scrollingInfo, textControlId), editorVersion, editorVersion, documentVersion)
  }

  val version get() = versionHolder.version

  /**
   * Should help to diagnose specific editor when "Editor is disposed" is thrown
   */
  private fun<T> diagnostics(operation: String, action: () -> T): T{
    try {
        return action()
    }
    catch (t: Throwable) {
      if (logger.isTraceEnabled) {
        logger.error("Exception during $operation on text control $textControlId under client ${ClientId.current}", t)
      }
      throw t
    }
  }
}

private fun ScrollType.toScrollKind() = when (this) {
  ScrollType.RELATIVE -> ScrollKind.Relative
  ScrollType.CENTER -> ScrollKind.Center
  ScrollType.MAKE_VISIBLE -> ScrollKind.Smart
  ScrollType.CENTER_UP -> ScrollKind.UpperHalf
  ScrollType.CENTER_DOWN -> ScrollKind.LowerHalf
}