package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.event.DocumentListener
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Pair
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.testFramework.LightVirtualFile
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.idea.LifetimedProjectService
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rdserver.followMe.FollowMeManager

//normal project service - not lifted
@Suppress("NonDefaultConstructor")
class BackendFollowMeEditorsHost(
  project: Project
) : LifetimedProjectService(project) {

  companion object {
    fun getInstance(project: Project): BackendFollowMeEditorsHost = project.getService(BackendFollowMeEditorsHost::class.java)
  }

  private val logger = getLogger<BackendFollowMeEditorsHost>()

  //for diagnostic issues
  private val synchronizers = hashSetOf<FollowMeSynchronizer>()

  init {
    val followMeManager = FollowMeManager.getInstance(project)

    followMeManager.followers.view(projectServiceLifetime) { followingLifetime, who, whom ->
      logger.info("$who is not following $whom. Binding...")

      val connection = project.messageBus.connect(followingLifetime.createNestedDisposable())
      val fileSelectionSynchronizer = FollowMeSynchronizer(followingLifetime, who, whom, project)
      connection.subscribe(FileEditorManagerListener.FILE_EDITOR_MANAGER, fileSelectionSynchronizer)
      synchronizers.addUnique(followingLifetime, fileSelectionSynchronizer)
      EditorFactory.getInstance().eventMulticaster.addDocumentListener(object : DocumentListener {
        override fun beforeDocumentChange(event: DocumentEvent) {
          val manager = FollowMeManager.getInstance(project)
          if (ClientId.current == who && !manager.isFullSyncBetween(who, whom)) {
            val virtualFile = FileDocumentManager.getInstance().getFile(event.document)
            val isRealVirtualFile = virtualFile != null && virtualFile !is LightVirtualFile
            if (isRealVirtualFile) {
              logger.info("Stopping following $whom due to the document change from $who")
              manager.unregisterFollower(who)
            }
          }
        }
      }, followingLifetime.createNestedDisposable())
    }
    //
    //followMeManager.users.view(projectServiceLifetime) { userLifetime, (userClientId, _) ->
    //  project.messageBus.subscribe(
    //    userLifetime,
    //    FileEditorManagerListener.FILE_EDITOR_MANAGER,
    //    FollowMeFileEditorListener(userClientId, project)
    //  )
    //}
  }

  fun dump(): String {
    val result = StringBuilder()
    for (synchronizer in synchronizers) {
      result.appendln("+++")
      result.appendln(synchronizer.dump())
    }

    return result.toString()
  }
}

class FollowMeFileEditorListener(private val clientId: ClientId, private val project: Project) : FileEditorManagerListener {
  val followMeManager = FollowMeManager.getInstance(project)

  override fun fileOpenedSync(source: FileEditorManager, file: VirtualFile, editors: Pair<Array<FileEditor>, Array<FileEditorProvider>>) {
    reportToFollowers(file, true)
  }

  override fun fileOpened(source: FileEditorManager, file: VirtualFile) {
    reportToFollowers(file, true)
  }

  override fun selectionChanged(event: FileEditorManagerEvent) {
    reportToFollowers(event.newFile, false)
  }

  // todo ensure the same FileEditor providers are used
  private fun reportToFollowers(virtualFile: VirtualFile, forceOpen: Boolean) {
    if (ClientId.current != clientId) return

    for (follower in followMeManager.getFollowers(clientId)) {
      ClientId.withClientId(follower) {
        FileEditorManager.getInstance(project).openFile(virtualFile, true, !forceOpen)
      }
    }
  }
}