package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isLocal
import com.intellij.codeWithMe.ClientId.Companion.isValid
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.fileEditor.ex.FileEditorProviderManager
import com.intellij.openapi.fileEditor.ex.FileEditorWithProvider
import com.intellij.openapi.fileEditor.ex.IdeDocumentHistory
import com.intellij.openapi.fileEditor.impl.*
import com.intellij.openapi.fileEditor.impl.text.TextEditorProvider
import com.intellij.openapi.progress.ProcessCanceledException
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ProjectManager
import com.intellij.openapi.project.ProjectManagerListener
import com.intellij.openapi.util.Pair
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.ui.EditorNotificationPanel
import com.jetbrains.rd.ide.model.NotificationPanelCallbackHyperlinkModel
import com.jetbrains.rd.ide.model.NotificationPanelModel
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.userData
import com.jetbrains.rd.util.concurrentMapOf
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.editors.BackendEditorHost.Companion.textControlModel
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.startup.RdServerConnectionManager
import com.jetbrains.rdserver.ui.BackendNotificationsHost
import javax.swing.JComponent

//create rdServer editors, not visible for the host
class BackendServerFileEditorManager(project: Project) : PsiAwareFileEditorManagerImpl(project) {

  companion object {
    fun getInstance(project: Project) = FileEditorManager.getInstance(project) as BackendServerFileEditorManager
    private val logger = logger<BackendServerFileEditorManager>()

    private const val NOTIFICATION_GROUP_ID = "CWM BackendServerFileEditorManager"

    internal var Editor.editorClientId: ClientId? by userData()
    internal var FileEditor.fileEditorClientId: ClientId? by userData()
  }

  init {
    registerExtraEditorDataProvider(RemoteEditorDataProvider(this), null)

    project.messageBus.connect().subscribe(ProjectManager.TOPIC, object : ProjectManagerListener {
      override fun projectClosing(closedProject: Project) {
        if (project === closedProject) {
          closeRemoteFilesOnDispose()
        }
      }
    })
  }

  private val editorsPerClientId = concurrentMapOf<ClientId, RemoteEditors>()

  fun forceCreateRemoteTextEditor(file: VirtualFile) : Editor {
    val result = openRemoteEditor(file, true)
    val createdEditor = result.first.firstOrNull() ?: error("couldn't get a single created editor for $file")
    return (createdEditor as TextEditor).editor
  }

  override fun openFileImpl4(
    window: EditorWindow,
    file: VirtualFile,
    entry: HistoryEntry?,
    options: FileEditorOpenOptions
  ): Pair<Array<FileEditor>, Array<FileEditorProvider>> {

    if (ClientId.isCurrentlyUnderLocalId) {
      logger.debug("Opening host editor for ${file.canonicalPath}")
      return super.openFileImpl4(window, file, entry, options)
    }

    return openRemoteEditor(file, false)
  }

  private fun openRemoteEditor(file: VirtualFile, forceCreate: Boolean): Pair.NonNull<Array<FileEditor>, Array<FileEditorProvider>> {
    logger.debug("Opening remote editor for ${file.canonicalPath}")

    if (!PermissionManager.getInstance(project).isVisible(ClientId.current, file)) {
      logger.warn("abort an attempt to open not visible for client ${ClientId.current} file $file")
      return Pair.createNonNull(emptyArray(), emptyArray())
    }

    val newProviders = FileEditorProviderManager.getInstance().getProviders(project, file)
    if (newProviders.isEmpty()) {
      logger.info("Can not open the file $file for client ${ClientId.current}: no available providers")
      showUnableToOpenFileNotification(file)
      return Pair.createNonNull(emptyArray(), emptyArray())
    }

    var result = Pair.createNonNull(emptyArray<FileEditor>(), emptyArray<FileEditorProvider>())

    application.invokeAndWait {
      val remoteEditors = getOrCreateRemoteEditors()

      result = remoteEditors.findEditorsWithProviders(file)
      val createNewEditors = result.first.isEmpty() || forceCreate

      if (createNewEditors) {
        logger.info("creating remote file. ForceCreate = $forceCreate; noExisting = ${result.first.isEmpty()}")
        project.messageBus.syncPublisher(FileEditorManagerListener.Before.FILE_EDITOR_MANAGER).beforeFileOpened(this, file)

        val newEditors = mutableListOf<FileEditor>()
        for (provider in newProviders) {
          try {
            val fileEditor = remoteEditors.createRemoteEditor(file, provider)
            if (fileEditor != null) {
              newEditors.add(fileEditor)
            }
          }
          catch (t: Throwable) {
            logger.error(t)
          }
        }

        if (newEditors.isEmpty()) {
          val provider = TextEditorProvider.getInstance()
          if (provider.accept(project, file)) {
            val fileEditor = remoteEditors.createRemoteEditor(file, provider)
            if (fileEditor != null) {
              newEditors.add(fileEditor)
            }
          }
        }

        if (newEditors.isEmpty()) {
          logger.info("Can not open the file $file for client ${ClientId.current}: no available editors")
          showUnableToOpenFileNotification(file)
        }

        result = Pair.createNonNull(newEditors.toTypedArray(), newProviders)
      }
      else {
        //ignore options.isFocusEditor. I'm not sure we support case when we select a tab but not focus it...
        remoteEditors.setSelectedEditor(file)
      }

      (IdeDocumentHistory.getInstance(project) as IdeDocumentHistoryImpl).onSelectionChanged()

      if (createNewEditors) {
        project.messageBus.syncPublisher(FileEditorManagerListener.FILE_EDITOR_MANAGER)
          .fileOpenedSync(this, file, result)

        notifyPublisher {
          if (isFileOpen(file)) {
            project.messageBus.syncPublisher(FileEditorManagerListener.FILE_EDITOR_MANAGER)
              .fileOpened(this, file)
          }
        }
      }
    }

    return result
  }

  private fun showUnableToOpenFileNotification(file: VirtualFile) {
    BackendNotificationsHost.getInstance(project).notifyClient(Notification(
      NOTIFICATION_GROUP_ID, CodeWithMePlatformBundle.message("product.name"),
      CodeWithMePluginBundle.message("notification.content.unable.to.open.file", file.name),
      NotificationType.INFORMATION))
  }

  override fun getEditorsWithProviders(file: VirtualFile): Pair<Array<FileEditor>, Array<FileEditorProvider>> {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getEditorsWithProviders(file)
    }

    val remoteEditors = getOrCreateRemoteEditors()
    return remoteEditors.findEditorsWithProviders(file)
  }

  private fun closeRemoteFilesOnDispose() {
    for (remoteEditors in editorsPerClientId.values) {
      for (editorToRemove in remoteEditors.getAllEditors()) {
        remoteEditors.removeEditor(editorToRemove)
      }
    }
  }

  override fun closeFile(file: VirtualFile, window: EditorWindow, transferFocus: Boolean) {
    if (ClientId.isCurrentlyUnderLocalId) {
      super.closeFile(file, window, transferFocus)
      return
    }

    throw UnsupportedOperationException()
  }

  override fun closeFile(file: VirtualFile, moveFocus: Boolean, closeAllCopies: Boolean) {
    if (ClientId.isCurrentlyUnderLocalId) {
      super.closeFile(file, moveFocus, closeAllCopies)
      return
    }
    closeRemoteTextEditor(file, null)
  }

  /**
   * @param editor null means all editors
   */
  fun closeRemoteTextEditor(file: VirtualFile, editor: Editor?) {
    require(!ClientId.isCurrentlyUnderLocalId) { "Trying to close remote FileEditor $editor" }

    val remoteEditors = getOrCreateRemoteEditors()
    remoteEditors.assertIsClosingFromGuest()

    val editors = remoteEditors.findEditors(file).let {
      if (editor == null) return@let it
      it.filter { e -> (e as? TextEditor)?.editor == editor }
    }

    if (editors.isEmpty()) {
      logger.warn("cannot get editors for the closing file $file")
      return
    }

    try {
      val beforePublisher = project.messageBus.syncPublisher(FileEditorManagerListener.Before.FILE_EDITOR_MANAGER)
      beforePublisher.beforeFileClosed(this, file)

      for (editorToRemove in editors) {
        remoteEditors.removeEditor(editorToRemove)
      }
    } finally {
      notifyPublisher(Runnable {
        if (!project.isDisposed) {
          val afterPublisher = project.messageBus.syncPublisher(
            FileEditorManagerListener.FILE_EDITOR_MANAGER)
          afterPublisher.fileClosed(this, file)
        }
      })
    }
  }

  override fun getCurrentWindow(): EditorWindow? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getCurrentWindow()
    }
    return null
  }

  override fun setCurrentWindow(window: EditorWindow?) {
    if (ClientId.isCurrentlyUnderLocalId) {
      super.setCurrentWindow(window)
    }
  }

  override fun getCurrentFile(): VirtualFile? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getCurrentFile()
    }
    return getOrCreateRemoteEditors().getSelectedFile()
  }

  override fun getSelectedEditor(): FileEditor? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getSelectedEditor()
    }

    val remoteEditors = getOrCreateRemoteEditors()
    return remoteEditors.getSelectedEditor()
  }

  private fun getOrCreateRemoteEditors(): RemoteEditors {
    application.assertIsDispatchThread()
    assert(!ClientId.current.isLocal) { "Trying to get remote editors for local ClientId" }

    val currentClientId = ClientId.current
    return editorsPerClientId.getOrPut(currentClientId) {
      // todo replace with editorsPerClientId per-client components
      val remoteEditors = RemoteEditors(project, currentClientId)
      RdServerConnectionManager.getInstance(project).getClientLifetime(currentClientId).let {
        // todo remove
        if (it.isAlive) {
          it.onTermination {
            editorsPerClientId.remove(currentClientId)
          }
        }
      }

      return@getOrPut remoteEditors
    }
  }

  private fun tryGetRemoteEditors(): RemoteEditors? {
    val currentClientId = ClientId.current
    return editorsPerClientId[currentClientId]
  }

  override fun isFileOpen(file: VirtualFile): Boolean {
    val remoteEditors = tryGetRemoteEditors() ?: return super.isFileOpen(file)
    return remoteEditors.findEditors(file).isNotEmpty()
  }

  override fun getOpenFiles(): Array<VirtualFile> {
    val remoteEditors = tryGetRemoteEditors() ?: return super.getOpenFiles()
    return remoteEditors.getAllFiles()
  }

  override fun getOpenFilesWithRemotes(): Array<VirtualFile> {
    val remoteOpen = editorsPerClientId.values.flatMap { it.getAllFiles().toList() }.toTypedArray()
    return super.getOpenFiles() + remoteOpen
  }

  override fun getSelectedEditorWithRemotes(): Array<FileEditor> {
    val remoteSelected = editorsPerClientId.values.mapNotNull { it.getSelectedEditor() }.toTypedArray()
    return super.getSelectedEditors() + remoteSelected
  }

  override fun getSelectedTextEditorWithRemotes(): Array<Editor> {
    return selectedEditorWithRemotes.mapNotNull { (it as? TextEditor)?.editor }.toTypedArray()
  }

  override fun getFile(editor: FileEditor): VirtualFile? {
    val virtualFile = super.getFile(editor)
    if (virtualFile != null) return virtualFile

    return editor.file
  }

  override fun getSelectedFiles(): Array<VirtualFile> {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getSelectedFiles()
    }

    if (!application.isDispatchThread) {
      val remoteEditors = tryGetRemoteEditors() ?: return emptyArray()
      val selectedFiles = remoteEditors.getSelectedFile() ?: return emptyArray()
      return arrayOf(selectedFiles)
    }

    val remoteEditors = getOrCreateRemoteEditors()

    val selectedFile = remoteEditors.getSelectedFile() ?: return emptyArray()
    return arrayOf(selectedFile)
  }

  override fun getAllEditors(file: VirtualFile): Array<FileEditor> {
    val result = arrayListOf<FileEditor>()

    val physicalEditors = super.getAllEditors(file)
    result.addAll(physicalEditors)
    for (entry in editorsPerClientId) {
      result.addAll(entry.value.findEditors(file))
    }
    return result.toTypedArray()
  }

  override fun getAllEditors(): Array<FileEditor> {
    val result = arrayListOf<FileEditor>()
    val physicalEditors = super.getAllEditors()
    result.addAll(physicalEditors)
    for (entry in editorsPerClientId) {
      result.addAll(entry.value.getAllEditors())
    }

    return result.toTypedArray()
  }

  override fun setSelectedEditor(file: VirtualFile, fileEditorProviderId: String) {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.setSelectedEditor(file, fileEditorProviderId)
    }
    getOrCreateRemoteEditors().setSelectedEditor(file)
  }

  fun selectRemoteTextEditor(file: VirtualFile, editor: Editor) {
    require(!ClientId.isCurrentlyUnderLocalId) { "Trying to select remote FileEditor $editor with local id" }

    val remoteEditors = getOrCreateRemoteEditors()
    val fileEditor = remoteEditors.findEditors(file).find { (it as? TextEditor)?.editor == editor }
    if (fileEditor != null) {
      remoteEditors.setSelectedEditor(fileEditor)
    }
  }

  override fun getEditors(file: VirtualFile): Array<FileEditor> {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getEditors(file)
    }
    return getOrCreateRemoteEditors().findEditors(file).toTypedArray()
  }

  override fun getSelectedEditor(file: VirtualFile): FileEditor? {
    return getSelectedEditorWithProvider(file)?.fileEditor
  }

  override fun getSelectedEditorWithProvider(file: VirtualFile): FileEditorWithProvider? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getSelectedEditorWithProvider(file)
    }
    val remoteEditors = getOrCreateRemoteEditors()
    val fileEditor = remoteEditors.getSelectedEditor()
    if (fileEditor?.file == file) {
      val provider = remoteEditors.getProvider(fileEditor) ?: return null //no assert here is intended
      return FileEditorWithProvider(fileEditor, provider)
    }

    return null
  }

  override fun getSelectedTextEditor(): Editor? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getSelectedTextEditor()
    }
    return (selectedEditor as? TextEditor)?.editor
  }

  override fun getSelectedTextEditor(lockfree: Boolean): Editor? {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getSelectedTextEditor(lockfree)
    }
    return selectedTextEditor
  }

  override fun getSelectedEditors(): Array<FileEditor> {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.getSelectedEditors()
    }
    val selectedEditor = selectedEditor ?: return arrayOf()
    return arrayOf(selectedEditor)
  }

  fun enableClosingRemoteEditor(block : () -> Unit) {
    val remoteEditors = getOrCreateRemoteEditors()
    remoteEditors.enableClosingRemoteEditor(block)
  }

  // todo replace with BeControls
  private var notificationCounter = 0
  private val NOTIFICATION_ID = "CWM_NOTIFICATION_ID"

  override fun addTopComponent(editor: FileEditor, component: JComponent) {
    super.addTopComponent(editor, component)
    if (editor.clientId().isLocal) return

    val textControlModel = (editor as? TextEditor)?.editor?.textControlModel ?: return
    val panel = component as? EditorNotificationPanel ?: return

    val id = notificationCounter++
    panel.putClientProperty(NOTIFICATION_ID, id)
    // no link support for now. unclear security outcomes
    // todo permissions
    val closeLink = NotificationPanelCallbackHyperlinkModel("Close", true)
    val panelModel = NotificationPanelModel(id.toString(), panel.text, arrayOf(closeLink))
    textControlModel.notificationPanels[id] = panelModel
  }

  override fun removeTopComponent(editor: FileEditor, component: JComponent) {
    super.removeTopComponent(editor, component)
    if (editor.clientId().isLocal) return

    val textControlModel = (editor as? TextEditor)?.editor?.textControlModel ?: return
    val panel = component as? EditorNotificationPanel ?: return

    val id = panel.getClientProperty(NOTIFICATION_ID) ?: return
    textControlModel.notificationPanels.remove(id)
  }
}