package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.LogicalPosition
import com.intellij.openapi.editor.ScrollType
import com.intellij.openapi.editor.event.*
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.rd.ide.document.adjustScrollPosition
import com.jetbrains.rd.util.lifetime.Lifetime
import java.awt.Point

class EditorToEditorSynchronizer(private val from: Editor, private val to: Editor, private val lifetime: Lifetime) : CaretListener, SelectionListener, VisibleAreaListener {
  private val disposable = lifetime.createNestedDisposable("editorToEditorLifetime")
  private val visibleAreaChangesQueue = MergingUpdateQueue("editorToEditorAreaChangesQueue", 100, true, MergingUpdateQueue.ANY_COMPONENT, disposable, from.component, true)

  init {
    syncCaret(to, from.caretModel.primaryCaret.logicalPosition)
    visibleAreaChangesQueue.setRestartTimerOnAdd(true)
    adviseOnRemoteScroll(from)
  }

  fun install() {
    from.selectionModel.addSelectionListener(this, disposable)
    from.caretModel.addCaretListener(this, disposable)
    from.scrollingModel.addVisibleAreaListener(this, disposable)
  }

  private fun adviseOnRemoteScroll(editor: Editor) {
    editor.remoteEditorData?.visibleRange?.advise(lifetime) {
      headlessVisibleAreaChanged(editor, it.startOffset, it.endOffset)
    }
  }

  private fun getOther(editorImpl: Editor): Editor {
    if (editorImpl == from) return to
    if (editorImpl == to) return from

    throw UnsupportedOperationException("should be called either for 'from' or for 'to' editor")
  }

  override fun selectionChanged(e: SelectionEvent) {
    val editor = e.editor
    val otherEditor = getOther(editor)

    otherEditor.selectionModel.setSelection(e.newRange.startOffset, e.newRange.endOffset)

    adjustScrollPosition(otherEditor)
  }

  override fun caretPositionChanged(event: CaretEvent) {
    val editor = event.editor
    val otherEditor = getOther(editor)

    syncCaret(otherEditor, event.newPosition)
  }

  override fun visibleAreaChanged(e: VisibleAreaEvent) {
    if (!ClientId.isCurrentlyUnderLocalId)
      return
    
    val editor = e.editor
    val otherEditor = getOther(editor)
    
    visibleAreaChangesQueue.queue(object : Update("SendViewport") {
      override fun run() {
        val visibleArea = editor.scrollingModel.visibleArea

        val targetPosition = editor.xyToLogicalPosition(Point(0, visibleArea.y + visibleArea.height / 3))
        otherEditor.scrollingModel.disableAnimation()
        otherEditor.scrollingModel.scrollTo(targetPosition, ScrollType.CENTER)
        otherEditor.scrollingModel.enableAnimation()
      }
    })
  }

  private fun headlessVisibleAreaChanged(editor: Editor, startOffset: Int, endOffset: Int) {
    val otherEditor = getOther(editor)
    
    val startLine = editor.offsetToLogicalPosition(startOffset).line
    val endLine = editor.offsetToLogicalPosition(endOffset).line
    val centerLine = (2 * startLine + endLine) / 3
    val centerLogicalPosition = LogicalPosition(centerLine, 0)
    
    otherEditor.scrollingModel.disableAnimation()
    otherEditor.scrollingModel.scrollTo(centerLogicalPosition, ScrollType.CENTER)
    otherEditor.scrollingModel.enableAnimation()
  }

  private fun syncCaret(otherEditor: Editor, position : LogicalPosition) {
    otherEditor.caretModel.moveToLogicalPosition(position)

    adjustScrollPosition(otherEditor)
  }

  override fun caretAdded(event: CaretEvent) {
    val editor = event.editor
    val otherEditor = getOther(editor)

    val isPrimary = otherEditor.caretModel.primaryCaret == event.caret
    otherEditor.caretModel.addCaret(event.newPosition, isPrimary)
  }

  override fun caretRemoved(event: CaretEvent) {
    val editor = event.editor
    val otherEditor = getOther(editor)

    val caretToRemove = otherEditor.caretModel.allCarets.find { it.logicalPosition == event.oldPosition }
    caretToRemove?.let {
      otherEditor.caretModel.removeCaret(caretToRemove)
    }
  }

  fun dump(): String {
    val result = StringBuilder()
    result.appendln("from Editor: ${from.presentableName()}")
    result.appendln("to Editor: ${to.presentableName()}")
    return result.toString()
  }
}

fun Editor.presentableName(): String {
  val file = FileDocumentManager.getInstance().getFile(document) ?: return this.toString()
  return file.name
}