package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createLifetime
import com.intellij.openapi.util.Pair
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.testFramework.LightVirtualFile
import com.intellij.util.castSafelyTo
import com.jetbrains.rd.ide.document.isLocalFile
import com.jetbrains.rd.util.getLogger
import com.jetbrains.rd.util.info
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.intersect
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.followMe.FollowMeManager

class FollowMeSynchronizer(
  private val lifetime: Lifetime,
  private val who: ClientId,
  private val whom: ClientId,
  private val project: Project
) : FileEditorManagerListener {

  private val synchronizedEditors = hashMapOf<Editor, EditorToEditorSynchronizer>()
  private val logger = getLogger("FollowMeSynchronizer(who=$who;whom=$whom)")

  init {
    val leadEditor = ClientId.withClientId(whom) {
      FileEditorManager.getInstance(project).selectedTextEditor as? EditorImpl
    }

    leadEditor?.let {
      syncFollower(it)
    }
  }

  private fun syncFollower(leadEditor: EditorImpl) {
    val followerEditor = ClientId.withClientId(who) {
      val leadFile = FileDocumentManager.getInstance().getFile(leadEditor.document) ?: return
      if (shouldIgnore(leadFile, leadEditor)) return

      val openFile = FileEditorManager.getInstance(project).openFile(leadFile, true)
      openFile.firstOrNull()?.castSafelyTo<TextEditor>()?.editor as? EditorImpl
    }

    if (synchronizedEditors.containsKey(leadEditor)) return
    val followerEditorToBind = followerEditor ?: return

    val intersectWithLead = lifetime.intersect(leadEditor.disposable.createLifetime())
    val intersectAllLifetime = intersectWithLead.intersect(followerEditorToBind.disposable.createLifetime())

    val synchronizer = EditorToEditorSynchronizer(leadEditor, followerEditorToBind, intersectAllLifetime)
    synchronizer.install()

    synchronizedEditors[leadEditor] = synchronizer
    intersectAllLifetime.onTerminationIfAlive {
      synchronizedEditors.remove(leadEditor)
    }
  }

  // todo extensibility
  private fun shouldIgnore(leadFile: VirtualFile, leadEditor: EditorImpl): Boolean {
    if (leadFile is LightVirtualFile) return true
    if (ClientId.isLocalId(ClientId.current)) return false
    val existingSynchronizer = BackendDocumentHost.getInstance(project).tryGetSynchronizer(leadEditor.document)
    if (existingSynchronizer != null && !existingSynchronizer.documentId.isLocalFile()) return true
    return false
  }

  override fun fileOpenedSync(source: FileEditorManager, file: VirtualFile, editors: Pair<Array<FileEditor>, Array<FileEditorProvider>>) {

    val fileEditor = editors.first.find { fileEditor -> fileEditor.file == file }

    if (followerBreaksFollowing(fileEditor))
      return

    if (ClientId.current != whom)
      return

    val editorImpl = (fileEditor as? TextEditor)?.editor as? EditorImpl ?: return
    syncFollower(editorImpl)
  }

  override fun selectionChanged(event: FileEditorManagerEvent) {

    val editorImpl = (event.newEditor as? TextEditor)?.editor as? EditorImpl ?: return

    if (followerBreaksFollowing(event.newEditor))
      return

    if (ClientId.current != whom)
      return

    syncFollower(editorImpl)
  }

  private fun followerBreaksFollowing(fileEditor: FileEditor?): Boolean {
    if (ClientId.current == who && !FollowMeManager.getInstance(project).isFullSyncBetween(who, whom)) {
      val selectedLeadEditor = ClientId.withClientId(whom) {
        FileEditorManager.getInstance(project).selectedEditor
      }

      if (selectedLeadEditor?.file != fileEditor?.file) {
        logger.info { "Navigating outside of a lead's scope. Follower navigation: ${fileEditor}. Selected lead editor: $selectedLeadEditor" }
        logger.info { "stopping following $who to $whom" }
        FollowMeManager.getInstance(project).unregisterFollower(who)
        return true
      }
    }
    return false
  }

  fun dump(): String {
    return StringBuilder().apply {
      appendln("FollowMeSynchronizer")
      appendln("who: $who")
      appendln("whom: $whom")
      appendln("isAlive: ${lifetime.isAlive}")

      appendln("EditorToEditorSynchronizers:")

      for (synchronizedEditor in synchronizedEditors) {
        appendln(synchronizedEditor.value.dump())
      }
    }.toString()
  }
}