package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.editor.Caret
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.impl.CaretModelImpl
import com.intellij.openapi.fileEditor.FileEditorManager
import com.intellij.openapi.fileEditor.FileEditorManagerEvent
import com.intellij.openapi.fileEditor.FileEditorManagerListener
import com.intellij.openapi.fileEditor.FileEditorManagerListener.FILE_EDITOR_MANAGER
import com.intellij.openapi.fileEditor.TextEditor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createLifetime
import com.intellij.openapi.rd.createNestedDisposable
import com.jetbrains.codeWithMe.model.RemoteCaretData
import com.jetbrains.codeWithMe.model.codeWithMeSolutionExtModel
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.platform.codeWithMe.carets.RemoteCaretHighlighterService
import com.jetbrains.rd.platform.codeWithMe.carets.createViewableCarets
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.platform.util.subscribe
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.intersect
import com.jetbrains.rd.util.reactive.Property
import com.jetbrains.rdserver.followMe.FollowMeManager

@Suppress("NonDefaultConstructor")
class MultiUserCaretSynchronizer(
  val lifetime: Lifetime,
  val project: Project,
  solution: Solution
) {
  init {
    solution.codeWithMeSolutionExtModel // don't remove, force to initialize serializers from unused Ext
    MultiUserCaretSynchronizerProjectService.getInstance(project).handleGuest(lifetime)
  }
}

@Service
class MultiUserCaretSynchronizerProjectService(val project: Project) {
  init {
    ClientId.withClientId(ClientId.localId) {
      handleHost(project.lifetime)
    }
  }

  fun handleHost(hostLifetime: Lifetime) {
    val lastFocusedEditor = Property(FileEditorManager.getInstance(project).selectedTextEditor)

    project.messageBus.subscribe(hostLifetime, FILE_EDITOR_MANAGER, object : FileEditorManagerListener {
      override fun selectionChanged(event: FileEditorManagerEvent) {
        val editor = (event.newEditor as? TextEditor)?.editor ?: return
        if (!editor.isLocal()) return
        val lifetime = (editor.caretModel as? CaretModelImpl)?.createLifetime() ?: return
        lifetime.bracket(
          { lastFocusedEditor.set(editor) },
          { if (lastFocusedEditor.value == editor) lastFocusedEditor.set(null) }
        )
      }
    })

    lastFocusedEditor.view(hostLifetime) { editorLt, editor ->
      if (editor == null) return@view
      register(editorLt, ClientId.localId, editor)
    }
  }

  fun handleGuest(guestLifetime: Lifetime) {
    val editorsHost = BackendEditorHost.getInstance(project)
    val clientId = ClientId.current
    editorsHost.viewLastFocusedTextControl(guestLifetime) view@{ modelLt, textControlId ->
      val editor = editorsHost.tryGetOwnRemoteEditor(textControlId) ?: return@view
      val editorLt = (editor.caretModel as? CaretModelImpl)?.createLifetime() ?: return@view
      val lifetime = modelLt.intersect(editorLt) // todo viewLastFocused should handle it by default
      register(lifetime, clientId, editor)
    }
  }

  fun register(lifetime: Lifetime, clientId: ClientId, editor: Editor) {
    val listenerDisposable = lifetime.createNestedDisposable()
    val remoteCaretModels = editor.createViewableCarets(listenerDisposable) { it.createModel(clientId) }

    remoteCaretModels.view(lifetime) { caretLt, models ->
      for ((startOffset, endOffset, data) in models) {
        RemoteCaretHighlighterService.getInstance(project).addRemoteCaret(caretLt, editor, startOffset, endOffset, data)
      }
    }
  }

  private data class RemoteCaretModel(
    val start: Int,
    val end: Int,
    val data: RemoteCaretData
  )

  private fun Caret.createModel(clientId: ClientId): RemoteCaretModel {
    val manager = FollowMeManager.getInstance(project)
    val user = manager.getUser(clientId)
    val lead = manager.followers[clientId]?.let { manager.getUser(it) }
    val isSetManually = lead?.let { isSetManually(it.clientId) } ?: true
    val data = RemoteCaretData(user, selectionEnd != offset, lead, isSetManually)
    return RemoteCaretModel(selectionStart, selectionEnd, data)
  }


  private fun Caret.isSetManually(lead: ClientId): Boolean {
    ClientId.withClientId(lead) {
      val selectedEditor = BackendServerFileEditorManager.getInstance(project).selectedEditor
      val leadEditor = (selectedEditor as? TextEditor)?.editor ?: return true
      if (leadEditor.document != this.editor.document) return true
      val leadCaret = leadEditor.caretModel.primaryCaret

      return (selectionStart != leadCaret.selectionStart ||
              selectionEnd != leadCaret.selectionEnd ||
              offset != leadCaret.offset)
    }
  }

  companion object {
    fun getInstance(project: Project) = project.service<MultiUserCaretSynchronizerProjectService>()
  }
}