package com.jetbrains.rdserver.editors

import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isValid
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.fileEditor.ex.FileEditorManagerEx
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.Key
import com.intellij.openapi.util.Pair
import com.intellij.openapi.vfs.VirtualFile
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.util.CopyOnWriteArrayList
import com.jetbrains.rd.util.concurrentMapOf
import com.jetbrains.rdserver.editors.BackendServerFileEditorManager.Companion.editorClientId
import com.jetbrains.rdserver.editors.BackendServerFileEditorManager.Companion.fileEditorClientId

class RemoteEditors(private val project: Project, val clientId: ClientId) {
  companion object {
    private val logger = getLogger<RemoteEditors>()
    val REMOTELY_VISIBLE = Key<Boolean>("REMOTELY_VISIBLE")
  }

  private val editors = CopyOnWriteArrayList<FileEditor>()
  private val editorsToProviders = concurrentMapOf<FileEditor, FileEditorProvider>()
  private var controllableClosing = false

  @Volatile
  private var selectedEditor: FileEditor? = null

  init {
    require(!ClientId.isLocalId(clientId)) { "creating RemoteEditors with local clientId" }
  }

  fun createRemoteEditor(file: VirtualFile, provider: FileEditorProvider): FileEditor? {
    val handler = BackendFileEditorHandler.EP_NAME.extensionList.find { it.isAvailable(project, file, provider) }
    if (handler == null) {
      logger.info("Can't open remote editor for file=${file.presentableUrl} with provider=$provider for clientId=$clientId")
      return null
    }

    val editor = handler.createEditor(project, file, provider)
    if (editor == null) return null

    logger.info("opening and selecting remote editor ${file.presentableUrl} for clientId = $clientId")
    editor.fileEditorClientId = ClientId.current
    if (editor is TextEditor) editor.editor.editorClientId = ClientId.current

    editors.add(editor)
    editorsToProviders[editor] = provider
    setSelectedEditor(editor)

    return editor
  }

  fun getProvider(editor: FileEditor): FileEditorProvider? {
    return editorsToProviders[editor]
  }

  fun getSelectedEditor(): FileEditor? {
    selectedEditor?.let {
      require(!Disposer.isDisposed(it)) { "getting disposed selected editor $it" }
    }
    return selectedEditor
  }

  fun findEditors(file: VirtualFile): List<FileEditor> {
    return editors.filter {
      it.file == file
    }
  }

  fun findEditorsWithProviders(file: VirtualFile): Pair.NonNull<Array<FileEditor>, Array<FileEditorProvider>> {
    val editors = mutableListOf<FileEditor>()
    val providers = mutableListOf<FileEditorProvider>()
    for (fileEditor in findEditors(file)) {
      editors.add(fileEditor)
      val fileEditorProvider = editorsToProviders[fileEditor] ?: error("cannot find editors provider")
      providers.add(fileEditorProvider)
    }
    return Pair.createNonNull(editors.toTypedArray(), providers.toTypedArray())
  }

  fun removeEditor(editor: FileEditor) {
    logger.info("removing editor ${editor.file?.presentableUrl ?: editor}")

    val editorProvider = editorsToProviders[editor] ?: error("cannot find editors provider")
    editorProvider.disposeEditor(editor)

    editors.remove(editor)
    editorsToProviders.remove(editor)

    if (editors.isNotEmpty()) {
      setSelectedEditor(editors.last())
      logger.info("selected editor is now ${editor.file?.presentableUrl ?: editor}")
    }
    else {
      setSelectedEditor(null)
    }
  }

  fun getSelectedFile(): VirtualFile? {
    return selectedEditor?.file
  }

  fun setSelectedEditor(file: VirtualFile) {
    val editor = findEditors(file).firstOrNull() ?: return

    setSelectedEditor(editor)
  }

  fun setSelectedEditor(newEditor: FileEditor?) {
    if (!clientId.isValid) return

    val newFile = newEditor?.file
    val oldEditor = selectedEditor
    val oldVirtualFile = oldEditor?.file

    newFile?.let {
      require(findEditors(newFile).isNotEmpty()) { "selecting unregistered editor" }
    }

    val fileEditorManager = FileEditorManagerEx.getInstanceEx(project)
    fileEditorManager.notifyPublisher {
       //needed here so IJ listeners of 'selectionChanged' events will get proper visibility of the editor
      // (not after selectionChanged -> frontend -> backend loop in TextControlHost)
      (newEditor as? TextEditor)?.editor?.putUserData(REMOTELY_VISIBLE, true)

      selectedEditor = newEditor
      val event = FileEditorManagerEvent(fileEditorManager, oldVirtualFile, oldEditor, newFile, newEditor)
      project.messageBus.syncPublisher(FileEditorManagerListener.FILE_EDITOR_MANAGER).selectionChanged(event)
    }
  }

  fun getAllFiles(): Array<VirtualFile> {
    return editors.mapNotNull { it.file }.toTypedArray()
  }

  fun getAllEditors(): List<FileEditor> {
    return editors
  }

  fun enableClosingRemoteEditor(block: () -> Unit) {
    application.assertIsDispatchThread()
    noReenter(::controllableClosing) {
      block()
    }
  }

  fun assertIsClosingFromGuest() {
    application.assertIsDispatchThread()
    require(controllableClosing) {
      "Closing guests editor not from unknown places"
    }
  }
}