package com.jetbrains.rdserver.execution

import com.intellij.build.BuildProgressListener
import com.intellij.build.DefaultBuildDescriptor
import com.intellij.build.events.StartBuildEvent
import com.intellij.openapi.components.service
import com.intellij.openapi.externalSystem.model.task.ExternalSystemTaskId
import com.intellij.openapi.externalSystem.model.task.ExternalSystemTaskNotificationEvent
import com.intellij.openapi.externalSystem.model.task.ExternalSystemTaskNotificationListenerAdapter
import com.intellij.openapi.externalSystem.model.task.event.ExternalSystemTaskExecutionEvent
import com.intellij.openapi.externalSystem.service.execution.ExternalSystemProcessHandler
import com.intellij.openapi.externalSystem.service.execution.ExternalSystemRunConfigurationViewManager
import com.intellij.openapi.externalSystem.service.notification.ExternalSystemProgressNotificationManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.util.containers.ContainerUtil
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.build.toProtocolModel
import com.jetbrains.rdserver.execution.configurations.BackendRunContentHost

@Suppress("NonDefaultConstructor")
class BackendExternalSystemHost(
  lifetime: Lifetime,
  project: Project,
  protocol: Protocol
) {
  private val model = CWMExternalSystemModel.create(lifetime, protocol)
  private val isTest = ContainerUtil.createWeakMap<ExternalSystemTaskId, Boolean>()

  init {
    project.service<ExternalSystemRunConfigurationViewManager>().addListener(BuildProgressListener { buildId, event ->
      if (buildId !is ExternalSystemTaskId) return@BuildProgressListener
      if (isTest[buildId] == true) return@BuildProgressListener
      if (event is StartBuildEvent) {
        val task = ((event.buildDescriptor as? DefaultBuildDescriptor)?.processHandler as? ExternalSystemProcessHandler)?.task
        if (task != null && BackendRunContentHost.tryGuessIsTest(task)) {
          isTest[buildId] = true
          return@BuildProgressListener
        }
      }
      
      val eventModel = event.toProtocolModel(project, buildId)
      model.onStatusChangeBuild.fire(OnStatusChangeArgs2(eventModel, buildId.toModel(project)))
    }, lifetime.createNestedDisposable())
    
    service<ExternalSystemProgressNotificationManager>().addNotificationListener(object : ExternalSystemTaskNotificationListenerAdapter() {
      // TODO[moklev]: should not fire an event if a run configuration was started, only for sync and other tasks
      //override fun onStart(id: ExternalSystemTaskId, workingDir: String) {
      //  val task = service<ExternalSystemProcessingManager>().findTask(id)
      //  if (task is ExternalSystemExecuteTaskTask) {
      //    model.onStart.fire(OnStartArgs(
      //      id.toModel(project), 
      //      "${task.ideProject.name} [${task.tasksToExecute.joinToString()}]", 
      //      task.tasksToExecute
      //    ))
      //  } else {
      //    model.onStart.fire(OnStartArgs(id.toModel(project), "Unknown", emptyList()))
      //  }
      //}

      override fun onTaskOutput(id: ExternalSystemTaskId, text: String, stdOut: Boolean) {
        model.onTaskOutput.fire(OnTaskOutputArgs(text, stdOut, id.toModel(project)))
      }

      override fun onEnd(id: ExternalSystemTaskId) {
        model.onEnd.fire(OnEndArgs(id.toModel(project)))
      }

      override fun onSuccess(id: ExternalSystemTaskId) {
        model.onSuccess.fire(OnSuccessArgs(id.toModel(project)))
      }

      override fun onFailure(id: ExternalSystemTaskId, e: Exception) {
        model.onFailure.fire(OnFailureArgs(e.message ?: "", id.toModel(project)))
      }

      override fun onCancel(id: ExternalSystemTaskId) {
        model.onCancel.fire(OnCancelArgs(id.toModel(project)))
      }

      override fun onStatusChange(event: ExternalSystemTaskNotificationEvent) {
        when (event) {
          //is ExternalSystemBuildEvent -> {
          //  val eventModel = event.buildEvent.toProtocolModel(project, event.id) { 
          //    project.service<ThreadSafeIdHolder>().getId(it) 
          //  }
          //  model.onStatusChangeBuild.fire(OnStatusChangeArgs2(event.id.toModel(project), eventModel))
          //}
          is ExternalSystemTaskExecutionEvent -> {
            val eventModel = event.progressEvent.toModel()
            if (eventModel != null) {
              model.onStatusChange.fire(OnStatusChangeArgs(eventModel, event.id.toModel(project)))
            }
          }
        }
      }

      override fun beforeCancel(id: ExternalSystemTaskId) {}
    }, lifetime.createNestedDisposable())
  }
}