package com.jetbrains.rdserver.execution.configurations

import com.intellij.codeWithMe.ClientId
import com.intellij.execution.ExecutionManager
import com.intellij.execution.RunManager
import com.intellij.execution.RunManagerListener
import com.intellij.execution.RunnerAndConfigurationSettings
import com.intellij.execution.configurations.ConfigurationFactory
import com.intellij.execution.configurations.ConfigurationType
import com.intellij.execution.configurations.UnknownConfigurationType
import com.intellij.execution.executors.DefaultRunExecutor
import com.intellij.execution.runners.ExecutionEnvironmentBuilder
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.openapi.extensions.ExtensionPointListener
import com.intellij.openapi.extensions.PluginDescriptor
import com.intellij.openapi.externalSystem.service.execution.ExternalSystemRunConfiguration
import com.intellij.openapi.options.SettingsEditor
import com.intellij.openapi.options.SettingsEditorGroup
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.jetbrains.codeWithMe.model.*
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.permissions.RunAccessLevel
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rdserver.ui.BackendNotificationsHost
import com.jetbrains.rdserver.ui.converters.toBeLabel

@Suppress("NonDefaultConstructor")
class BackendRunConfigurationsHost(
  val clientId: ClientId,
  lifetime: Lifetime,
  val project: Project,
  protocol: Protocol
) {
  init {
    val runConfigurationsModel = CodeWithMeRunConfigurationsModel.create(lifetime, protocol)

    ConfigurationType.CONFIGURATION_TYPE_EP.getPoint().addExtensionPointListener(object : ExtensionPointListener<ConfigurationType> {
      override fun extensionAdded(extension: ConfigurationType, pluginDescriptor: PluginDescriptor) {
        ClientId.withClientId(clientId) {
          runConfigurationsModel.configurationTypes[extension.id] = extension.toModel()
          runConfigurationsModel.configurationFactories[extension.id] = extension.configurationFactories.map { it.toModel() }
        }
      }

      override fun extensionRemoved(extension: ConfigurationType, pluginDescriptor: PluginDescriptor) {
        ClientId.withClientId(clientId) {
          runConfigurationsModel.configurationTypes.remove(extension.id)
          runConfigurationsModel.configurationFactories.remove(extension.id)
        }
      }
    }, true, lifetime.createNestedDisposable())

    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(RunManagerListener.TOPIC, object : RunManagerListener {
      override fun runConfigurationAdded(settings: RunnerAndConfigurationSettings) {
        runConfigurationsModel.runConfigurations[settings.uniqueID] = settings.toModel() ?: return
      }

      override fun runConfigurationRemoved(settings: RunnerAndConfigurationSettings) {
        runConfigurationsModel.runConfigurations.remove(settings.uniqueID)
      }

      override fun runConfigurationChanged(settings: RunnerAndConfigurationSettings, existingId: String?) {
        // TODO (note that this method is called not from dispatch thread)
      }

      override fun runConfigurationSelected(settings: RunnerAndConfigurationSettings?) {
        val uniqueID = settings?.uniqueID ?: return
        runConfigurationsModel.selectedConfiguration.fire(uniqueID)
      }
    })

    val runManager = RunManager.getInstance(project)
    runManager.allSettings.forEach { settings ->
      runConfigurationsModel.runConfigurations[settings.uniqueID] = settings.toModel() ?: return@forEach
    }

    runManager.selectedConfiguration?.let { cfg ->
      runConfigurationsModel.selectedConfiguration.fire(cfg.uniqueID)
    }
  }

  private fun RunnerAndConfigurationSettings.toModel(): CodeWithMeRunConfigurationModel? {
    if (factory is UnknownConfigurationType)
      return null

    return CodeWithMeRunConfigurationModel(
      configuration.name,
      folderName,
      configuration.icon?.toProtocolModel(),
      type.id,
      factory.id,
      configuration is ExternalSystemRunConfiguration,
      configuration.configurationEditor.toModel()
    ).apply {
      execute.set { args ->
        if (PermissionManager.getInstance(project).getRunAccess(clientId) != RunAccessLevel.FullAccess) {
          logger.info("A client tries to run Run action without permissions: clientId=${ClientId}")
          BackendNotificationsHost.getInstance(project).notifyClient(Notification(
            PermissionManager.NOTIFICATION_GROUP_ID, "Can not execute Run action",
            "You do not have necessary permissions.",
            NotificationType.INFORMATION))
          return@set
        }
        
        if (args.executorId != DefaultRunExecutor.EXECUTOR_ID)
          throw IllegalStateException("Must be DefaultRunExecutor.EXECUTOR_ID")

        val executor = DefaultRunExecutor()
        logger.info("Trying to find runner and configuration settings for configuration \"$configuration\"")
        val settings = RunManager.getInstance(project).findSettings(configuration) 
                       ?: throw IllegalStateException("Can't find runner and configuration settings")

        logger.info("Trying to create an execution environment for executor \"${executor.id}\" and configuration \"$configuration\"")
        val builder = ExecutionEnvironmentBuilder.createOrNull(executor, settings) 
                      ?: throw IllegalStateException("Can't create an environment builder")

        val environment = builder.activeTarget().build()
        ExecutionManager.getInstance(project).restartRunProfile(environment)
      }
    }
  }

  private fun <T> SettingsEditor<T>.toModel(): SettingsEditorModelBase {
    (this as? SettingsEditorGroup<T>)?.let { group ->
      return SettingsEditorGroupModel(group.editors.map { pair ->
        val name = pair.first
        val editor = pair.second.toModel()
        NamedEditor(name, editor as SimpleSettingsEditorModel)
      })
    }

    // TODO leads to CWM-667 components conversion goes through the hierarchy, meets the project tree, and starts to render all NPM packages
    // TODO return beComponent when implementation will become reliable
    //val beComponent = try {
    //  component.toBeModel()
    //} catch (e: Throwable) {
    //  logger.warn("Failed to convert $component", e)
    //  "".toBeLabel()
    //}
    return SimpleSettingsEditorModel("".toBeLabel())
  }

  private fun ConfigurationFactory.toModel() = CodeWithMeConfigurationFactoryModel(id).also {
    it.isApplicable.set { _ -> isApplicable(project) }
    it.createTemplateConfiguration.set { _ -> TODO() }
  }

  private fun ConfigurationType.toModel() = CodeWithMeConfigurationTypeModel(
    displayName,
    id,
    icon.toProtocolModel(),
    tag,
    configurationTypeDescription,
    configurationFactories.map { it.id },
    helpTopic,
    isManaged
  )

  companion object {
    private val logger = getLogger<BackendRunConfigurationsHost>()
  }
}