package com.jetbrains.rdserver.execution.configurations

import com.intellij.codeWithMe.ClientId
import com.intellij.execution.ExecutionListener
import com.intellij.execution.ExecutionManager
import com.intellij.execution.Executor
import com.intellij.execution.configurations.RunConfiguration
import com.intellij.execution.configurations.RunProfile
import com.intellij.execution.process.ProcessHandler
import com.intellij.execution.runners.ExecutionEnvironment
import com.intellij.execution.testframework.actions.ConsolePropertiesProvider
import com.intellij.execution.testframework.sm.runner.SMTRunnerConsoleProperties
import com.intellij.execution.ui.ExecutionConsole
import com.intellij.openapi.externalSystem.model.task.ExternalSystemTask
import com.intellij.openapi.externalSystem.service.execution.DefaultExternalSystemExecutionConsoleManager
import com.intellij.openapi.externalSystem.service.execution.ExternalSystemProcessHandler
import com.intellij.openapi.externalSystem.service.execution.ExternalSystemRunConfiguration
import com.intellij.openapi.externalSystem.service.internal.ExternalSystemExecuteTaskTask
import com.intellij.openapi.externalSystem.util.ExternalSystemUtil
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.terminal.TerminalExecutionConsole
import com.intellij.util.pooledThreadSingleAlarm
import com.jetbrains.codeWithMe.model.BackendRunContentModel
import com.jetbrains.codeWithMe.model.BackendRunProcessModel
import com.jetbrains.codeWithMe.model.ExecutionConsoleKind
import com.jetbrains.codeWithMe.model.ExternalSystemInfo
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rdserver.adviseOnTargetOutput
import com.jetbrains.rdserver.execution.toModel
import com.jetbrains.rdserver.toProtocolModel

@Suppress("NonDefaultConstructor")
class BackendRunContentHost(
  clientId: ClientId,
  lifetime: Lifetime,
  project: Project,
  protocol: Protocol
) {
  init {
    val runContentModel = BackendRunContentModel.create(lifetime, protocol)

    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(ExecutionManager.EXECUTION_TOPIC, object : ExecutionListener {
      override fun processStarted(executorId: String, env: ExecutionEnvironment, handler: ProcessHandler) {
        val executionId = env.executionId
        val configuration = env.runProfile as? RunConfiguration ?: env.runnerAndConfigurationSettings?.configuration
        val configurationTypeId = configuration?.type?.id ?: ""
        val configurationName = configuration?.name ?: ""
        val testFramework = tryGetTestFramework(project, configuration, env.executor)
        val descriptor = env.contentToReuse ?: run {
          logger.warn("env.contentToReuse is null - can't synchronize a process with executionId = $executionId and " +
                      "configuration name = \"$configurationName\"")
          return
        }

        val isIdBased = getConsoleProperties(configuration, env.executor)?.isIdBasedTestTree ?: false
        val model = BackendRunProcessModel(
          executionId, descriptor.displayName, configurationTypeId, configurationName, testFramework?: "", isIdBased,
          env.executor.id, env.runner.runnerId, getExternalSystemInfo(env, handler), getExecutionConsoleKind(descriptor.executionConsole)
        )

        adviseOnTargetOutput(lifetime, project, clientId, descriptor.executionConsole, model.consoleOutput)
        model.terminateFromFrontend.advise(lifetime) {
          handler.destroyProcess()
        }

        runContentModel.runProcesses[executionId] = model
      }
      
      private fun getExternalSystemInfo(env: ExecutionEnvironment, handler: ProcessHandler): ExternalSystemInfo? {
        if (handler !is ExternalSystemProcessHandler)
          return null
        
        val task = handler.task ?: return null
        val taskId = task.id
        val configurationIcon = env.runnerAndConfigurationSettings?.configuration?.icon?.toProtocolModel()
        val taskNames: List<String>
        val executionName: String
        if (task is ExternalSystemExecuteTaskTask) {
          taskNames = task.tasksToExecute
          executionName = "${task.ideProject.name} [${task.tasksToExecute.joinToString()}]"
        } else {
          taskNames = emptyList()
          executionName = "Unknown" // TODO[moklev]
        }
        return ExternalSystemInfo(taskId.toModel(project), executionName, taskNames, configurationIcon)
      }

      override fun processTerminated(executorId: String, env: ExecutionEnvironment, handler: ProcessHandler, exitCode: Int) {
        application.invokeLater {
          if (lifetime.isNotAlive) return@invokeLater
          val model = runContentModel.runProcesses[env.executionId]
          if (model == null) {
            logger.warn("No process with executionId='${env.executionId}' was found")
            return@invokeLater
          }
          
          model.exitCode.set(exitCode)
          pooledThreadSingleAlarm(1000, lifetime.createNestedDisposable()) {
            runContentModel.runProcesses.remove(env.executionId)
          }.request()
        }
      }

      override fun processNotStarted(executorId: String, env: ExecutionEnvironment) {
        application.invokeLater {
          if (lifetime.isNotAlive) return@invokeLater
          runContentModel.runProcesses.remove(env.executionId)
        }
      }
    })
  }
  
  companion object {
    private val logger = getLogger<BackendRunContentHost>()

    fun tryGetTestFramework(project: Project, configuration: RunProfile?, executor: Executor): String {
      val testFramework = getConsoleProperties(configuration, executor)?.testFrameworkName

      // This is the hack to distinguish GradleTestsExecutionConsoleManager from GradleExecutionConsoleManager
      if (configuration is ExternalSystemRunConfiguration) {
        val task = ExternalSystemExecuteTaskTask(project, configuration.settings, null, configuration)
        if (!tryGuessIsTest(task))
          return ""
      }

      return testFramework?:""
    }

    fun getConsoleProperties(configuration: RunProfile?,
                                     executor: Executor) =
      ((configuration as? ConsolePropertiesProvider)?.createTestConsoleProperties(executor) as? SMTRunnerConsoleProperties)

    fun tryGuessIsTest(task: ExternalSystemTask): Boolean {
      val consoleManager = ExternalSystemUtil.getConsoleManagerFor(task)
      if (consoleManager is DefaultExternalSystemExecutionConsoleManager)
        return false
      
      return true
    }
    
    fun getExecutionConsoleKind(console: ExecutionConsole): ExecutionConsoleKind = when (console) {
      is TerminalExecutionConsole -> ExecutionConsoleKind.TerminalExecutionConsole
      else -> ExecutionConsoleKind.Default
    }
  }
}