package com.jetbrains.rdserver.filters

import com.intellij.execution.filters.CompositeFilter
import com.intellij.execution.filters.ConsoleFilterProvider
import com.intellij.execution.filters.Filter
import com.intellij.execution.impl.EditorHyperlinkSupport
import com.intellij.execution.ui.ConsoleViewContentType
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.application.ReadAction
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.runAndLogException
import com.intellij.openapi.editor.colors.CodeInsightColors.FOLLOWED_HYPERLINK_ATTRIBUTES
import com.intellij.openapi.editor.colors.CodeInsightColors.HYPERLINK_ATTRIBUTES
import com.intellij.openapi.editor.colors.EditorColorsManager
import com.intellij.openapi.editor.colors.EditorColorsScheme
import com.intellij.openapi.editor.colors.TextAttributesKey
import com.intellij.openapi.editor.impl.DocumentImpl
import com.intellij.openapi.editor.markup.TextAttributes
import com.intellij.openapi.project.Project
import com.intellij.unscramble.AnalyzeStacktraceUtil
import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rdserver.daemon.BackendHighlighterRegistrationsHost

@Suppress("NonDefaultConstructor")
class BackendConsoleFilterHost(
  lifetime: Lifetime,
  project: Project,
  solution: Solution
) {
  private var idCounter = 0L

  init {
    val filterProviderModel = solution.filterProviderModel

    filterProviderModel.navigate.advise(lifetime) { context ->
      ReadAction
        .nonBlocking<Filter.ResultItem?> {
          val models = calculateItems(project, context.text, context.filterKind)
          val item = models.find { createModel(it, context.hyperlink.id) == context.hyperlink }
          item
        }
        .expireWhen { lifetime.isNotAlive }
        .withDocumentsCommitted(project)
        .finishOnUiThread(ModalityState.any()) {
          it?.getHyperlinkInfo()?.navigate(project)
        }
        .submit(AppExecutorUtil.getAppExecutorService())
    }

    filterProviderModel.applyFilter.set { _, context ->
      val task = RdTask<List<FilterHighlighterModelBase>>()
      ReadAction
        .nonBlocking {
          val models = LOG.runAndLogException {
            val resultItems = calculateItems(project, context.text, context.filterKind)
            resultItems.mapNotNull { createModel(it, idCounter++) }
          }
          task.set(models ?: emptyList())
        }
        .expireWhen { lifetime.isNotAlive }
        .withDocumentsCommitted(project)
        .submit(AppExecutorUtil.getAppExecutorService())
      return@set task
    }
  }

  private fun calculateItems(project: Project, text: String, filterKind: FilterKind): List<Filter.ResultItem> {
    val document = DocumentImpl(text, true).apply { setReadOnly(true) }

    val compositeFilter = createFilter(project, filterKind)

    val resultItems = mutableListOf<Filter.ResultItem>()
    for (line in 0 until document.lineCount) {
      val lineStart = document.getLineStartOffset(line)
      val lineText = EditorHyperlinkSupport.getLineText(document, line, true)
      val result = compositeFilter.applyFilter(lineText, lineStart + lineText.length) ?: continue
      resultItems.addAll(result.resultItems)
    }

    if (compositeFilter.shouldRunHeavy()) {
      compositeFilter.applyHeavyFilter(document, 0, 0) {
        resultItems.addAll(it.resultItems)
      }
    }
    return resultItems
  }

  private fun createFilter(project: Project, filterKind: FilterKind): CompositeFilter {
    val includeStacktrace = filterKind == FilterKind.StackTrace || filterKind == FilterKind.All
    val includeConsole = filterKind == FilterKind.Console || filterKind == FilterKind.All
    val stacktraceFilters = if (includeStacktrace) AnalyzeStacktraceUtil.EP_NAME.getExtensions(project) else emptyList()
    val consoleFilters = if (includeConsole) {
      ConsoleFilterProvider.FILTER_PROVIDERS.extensionList.flatMap { it.getDefaultFilters(project).toList() }
    }
    else emptyList()
    val filters = stacktraceFilters + consoleFilters
    return CompositeFilter(project, filters)
  }

  private fun createModel(item: Filter.ResultItem, id: Long): FilterHighlighterModelBase? {
    // todo pull text attributes key through ResultItem
    val colorsScheme = EditorColorsManager.getInstance().globalScheme
    val isFollowed = item.followedHyperlinkAttributes != null
    val attributesKey = if (isFollowed) FOLLOWED_HYPERLINK_ATTRIBUTES else colorsScheme.tryRecoverKey(item.getHighlightAttributes())

    val hyperlinkInfo = item.getHyperlinkInfo()
    if (hyperlinkInfo != null) {
      val isNavigatable = hyperlinkInfo.includeInOccurenceNavigation()
      val resultAttributeKey = attributesKey ?: HYPERLINK_ATTRIBUTES
      val attributeId = BackendHighlighterRegistrationsHost.getAttributeId(resultAttributeKey)!!
      return FilterHyperlinkModel(isNavigatable, id, attributeId, item.highlightStartOffset, item.highlightEndOffset)
    }

    val attributeId = attributesKey?.let { BackendHighlighterRegistrationsHost.getAttributeId(attributesKey) }
    if (attributeId == null) {
      LOG.warn("Couldn't recover the attributes key for ${item.getHighlightAttributes().toString()}")
      return null
    }
    return FilterHighlighterModel(id, attributeId, item.highlightStartOffset, item.highlightEndOffset)
  }

  companion object {
    private val LOG = logger<BackendConsoleFilterHost>()
  }
}

// todo pull text attributes key through ResultItem
private fun EditorColorsScheme.tryRecoverKey(attributes: TextAttributes?): TextAttributesKey? {
  return when (attributes) {
    null -> null
    getAttributes(HYPERLINK_ATTRIBUTES) -> HYPERLINK_ATTRIBUTES
    getAttributes(FOLLOWED_HYPERLINK_ATTRIBUTES) -> FOLLOWED_HYPERLINK_ATTRIBUTES
    getAttributes(ConsoleViewContentType.LOG_DEBUG_OUTPUT_KEY) -> ConsoleViewContentType.LOG_DEBUG_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.LOG_DEBUG_OUTPUT_KEY) -> ConsoleViewContentType.LOG_DEBUG_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.LOG_VERBOSE_OUTPUT_KEY) -> ConsoleViewContentType.LOG_VERBOSE_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.LOG_INFO_OUTPUT_KEY) -> ConsoleViewContentType.LOG_INFO_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.LOG_WARNING_OUTPUT_KEY) -> ConsoleViewContentType.LOG_WARNING_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.LOG_ERROR_OUTPUT_KEY) -> ConsoleViewContentType.LOG_ERROR_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.LOG_EXPIRED_ENTRY) -> ConsoleViewContentType.LOG_EXPIRED_ENTRY
    getAttributes(ConsoleViewContentType.NORMAL_OUTPUT_KEY) -> ConsoleViewContentType.NORMAL_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.ERROR_OUTPUT_KEY) -> ConsoleViewContentType.ERROR_OUTPUT_KEY
    getAttributes(ConsoleViewContentType.USER_INPUT_KEY) -> ConsoleViewContentType.USER_INPUT_KEY
    getAttributes(ConsoleViewContentType.SYSTEM_OUTPUT_KEY) -> ConsoleViewContentType.SYSTEM_OUTPUT_KEY
    else -> null
  }
}
