package com.jetbrains.rdserver.findInProject

import com.intellij.codeWithMe.ClientId
import com.intellij.find.FindModel
import com.intellij.find.FindSettings
import com.intellij.find.impl.FindInProjectUtil
import com.intellij.openapi.fileEditor.OpenFileDescriptor
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.util.ProgressIndicatorUtils
import com.intellij.openapi.progress.util.ReadTask
import com.intellij.openapi.project.Project
import com.intellij.ui.SimpleTextAttributes
import com.intellij.usages.FindUsagesProcessPresentation
import com.intellij.usages.TextChunk
import com.intellij.usages.UsageInfo2UsageAdapter
import com.intellij.usages.UsageViewPresentation
import com.jetbrains.codeWithMe.model.FindInProjectModel
import com.jetbrains.codeWithMe.model.RdFindInProjectUsage
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.ide.model.RdSimpleTextAttributes
import com.jetbrains.rd.ide.model.RdTextChunk
import com.jetbrains.rd.platform.codeWithMe.util.toModel
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rd.util.reactive.viewNotNull
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.vfs.BackendVirtualFileSystemHost
import org.jetbrains.annotations.NotNull
import java.awt.Font

@Suppress("NonDefaultConstructor")
class BackendFindInProjectHost(
  lifetime: Lifetime,
  project: Project,
  protocol: IProtocol,
  clientId: ClientId
) {
  private val model = FindInProjectModel.create(lifetime, protocol)

  init {
    model.session.viewNotNull(lifetime) { sessionLifetime, session ->
      val findModel = FindModel().apply {
        stringToFind = session.model.stringToFind
        isWholeWordsOnly = session.model.isWholeWordsOnly
        isRegularExpressions = session.model.isRegularExpressions
        isCaseSensitive = session.model.isCaseSensitive
        isProjectScope = session.model.isProjectScope
        fileFilter = session.model.fileFilter
        moduleName = session.model.moduleName
        searchContext = FindModel.SearchContext.valueOf(session.model.searchContext)
      }

      ProgressIndicatorUtils.scheduleWithWriteActionPriority(object : ReadTask() {
        override fun performInReadAction(indicator: ProgressIndicator): Continuation? {
          val presentation = FindUsagesProcessPresentation(UsageViewPresentation())
          sessionLifetime.onTermination { presentation.isCanceled = true }

          val permissionManager = PermissionManager.getInstance(project)
          FindInProjectUtil.findUsages(findModel, project, { usageInfo ->
            if (sessionLifetime.isNotAlive)
              return@findUsages false

            val virtualFile = usageInfo.virtualFile
            if (virtualFile == null || !permissionManager.isVisible(clientId, virtualFile))
              return@findUsages true

            val fileId = BackendVirtualFileSystemHost.getInstance(project).bindFile(virtualFile, false)
            if (fileId == null)
              return@findUsages true

            val adapter = UsageInfo2UsageAdapter(usageInfo)
            val textChunks = adapter.text.drop(1).map {
              val attributes = createSimpleTextAttributes(it)
              RdTextChunk(it.text, attributes)
            }

            val rdFindInProjectUsage = RdFindInProjectUsage(textChunks,
                                                            adapter.line + 1,
                                                            usageInfo.navigationOffset,
                                                            usageInfo.rangeInElement?.length ?: 0,
                                                            fileId,
                                                            virtualFile.path)
            session.consumer.fire(rdFindInProjectUsage)
            true
          }, presentation)
          return Continuation {
            session.ready.fire()
          }
        }

        override fun onCanceled(indicator: ProgressIndicator) {}
      })
    }

    model.navigate.advise(lifetime) {
      val file = BackendVirtualFileSystemHost.getInstance(project).findVirtualFile(it.fileId) ?: return@advise
      val descriptor = OpenFileDescriptor(project, file, it.offset)
      descriptor.navigate(it.requestFocus)
    }

    // Sync current recent file masks to client
    for (fileMask in FindSettings.getInstance().recentFileMasks) {
      model.fileMasks.add(fileMask)
    }
  }

  // TODO: unity with UsageTableCellRenderer
  private fun createSimpleTextAttributes(textChunk: @NotNull TextChunk): RdSimpleTextAttributes {
    var at = textChunk.simpleAttributesIgnoreBackground
    val highlighted = textChunk.type != null || at.fontStyle == Font.BOLD
    if (highlighted)
      at = SimpleTextAttributes(
        null, at.fgColor, at.waveColor,
        at.style and SimpleTextAttributes.STYLE_BOLD.inv() or
          SimpleTextAttributes.STYLE_SEARCH_MATCH
      )

    return at.toModel()
  }
}