package com.jetbrains.rdserver.findUsages

import com.intellij.openapi.project.Project
import com.intellij.usages.Usage
import com.intellij.usages.UsageGroup
import com.intellij.usages.UsageTarget
import com.intellij.usages.rules.UsageGroupingRule
import com.intellij.usages.rules.UsageGroupingRuleEx
import com.intellij.usages.rules.UsageGroupingRuleProvider
import com.intellij.usages.rules.UsageGroupingRuleProviderEx
import com.jetbrains.rd.ide.model.RdUsageGroupingRule
import com.jetbrains.rd.ide.model.RdUsageGroupingRuleList
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.ide.model.findUsagesHost
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.toProtocolModel
import javax.swing.Icon

@Suppress("NonDefaultConstructor")
class BackendUsageGroupingRulesSupport(
  project: Project,
  solution: Solution
) {
  companion object {
    private val logger = getLogger<BackendUsageGroupingRulesSupport>()

    fun getInstance(project: Project) = RdServerPerClientComponents.getService(project, BackendUsageGroupingRulesSupport::class.java)
  }

  val supportedRules: List<UsageGroupingRuleEx>

  init {
    val rulesList = mutableListOf<RdUsageGroupingRule>()
    val originalRules = mutableListOf<UsageGroupingRuleEx>()
    UsageGroupingRuleProvider.EP_NAME.extensions.forEach { provider ->
      if (provider !is UsageGroupingRuleProviderEx) {
        logger.warn(
          "Provider $provider does not implement UsageGroupingRuleProviderEx, grouping rules will not be available on remote clients")
        return@forEach
      }
      val rules = provider.getAllRules(project, null)
      rules.map { it as? UsageGroupingRuleEx ?: GenericGroupingRuleWrapper(it) }.forEach { // TODO: better values for those parameters
        val ruleModel = RdUsageGroupingRule(
          it.id,
          it.title,
          it.icon?.toProtocolModel(),
          it.rank,
          true,
          true,
          it.isGroupingToggleable,
          true,
          it.groupingActionId,
          it.isGroupingActionInverted
        ).apply {
          enabled.set(true)
        }
        rulesList.add(ruleModel)
        originalRules.add(it)
      }
    }
    solution.findUsagesHost.groupingRules.set(RdUsageGroupingRuleList(rulesList))
    supportedRules = originalRules
  }

  class GenericGroupingRuleWrapper(private val builtinRule: UsageGroupingRule): UsageGroupingRule by builtinRule, UsageGroupingRuleEx {
    override fun getId() = builtinRule.javaClass.name
    override fun getIcon(): Icon? = null
    override fun getTitle() = builtinRule.javaClass.simpleName ?: "" // presentable title will be shown on the other end

    override fun getParentGroupsFor(usage: Usage, targets: Array<out UsageTarget>?): MutableList<UsageGroup> =
      builtinRule.getParentGroupsFor(usage, targets)

    override fun getRank() = builtinRule.rank

    override fun groupUsage(usage: Usage) = builtinRule.groupUsage(usage)
  }
}