// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.findUsages

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.Task
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.openapi.util.Factory
import com.intellij.psi.PsiDocumentManager
import com.intellij.usages.*
import com.intellij.usages.impl.UsageViewEx
import com.intellij.usages.impl.UsageViewManagerImpl
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.startup.serverSolution
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ide.requests.queueUntilBulkOperationIsFinished
import com.jetbrains.rdserver.requests.BackendIdeGateway
import java.util.concurrent.atomic.AtomicInteger

class BackendUsageViewManager(private val project: Project) : UsageViewManagerImpl(project) {
  private val sessionId = AtomicInteger()
  private val findUsagesModel: FindUsagesModel
    get() = project.serverSolution.findUsagesHost

  override fun createUsageView(
    targets: Array<out UsageTarget>,
    usages: Array<out Usage>,
    presentation: UsageViewPresentation,
    usageSearcherFactory: Factory<out UsageSearcher>?
  ): UsageViewEx {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createUsageView(targets, usages, presentation, usageSearcherFactory)

    val targetList = RdUsageTargetList(presentation.targetsNodeText,
                                       targets.mapIndexed { index, target -> convertUsageTarget(index, target) })
    val usageList = RdUsageList("", Unit)

    val session = RdFindUsagesSession(
      presentation.tabName ?: presentation.tabText,
      usageList, targetList, presentation.isShowCancelButton, true, null
    )
    val usageView = RdUsageView(project, targets, session, presentation)
    if (usages.isNotEmpty()) {
      usageView.appendUsagesInBulk(usages.toList())
    }
    val lifetime = project.defineNestedLifetime()
    BackendIdeGateway.getInstance(project).queueUntilBulkOperationIsFinished(lifetime, "Register new find usages session") {
      findUsagesModel.sessions[sessionId.incrementAndGet()] = session
    }
    return usageView
  }

  private fun convertUsageTarget(index: Int, usageTarget: UsageTarget): RdUsageTarget {
    val name = usageTarget.name ?: "<unnamed>"
    val rdUsageTarget = RdUsageTarget(index, name, usageTarget.presentation?.getIcon(true)?.toProtocolModel())
    rdUsageTarget.isValid.set(usageTarget.isValid)
    return rdUsageTarget
  }

  override fun showUsages(
    searchedFor: Array<out UsageTarget>,
    foundUsages: Array<out Usage>,
    presentation: UsageViewPresentation,
    factory: Factory<out UsageSearcher>?
  ): UsageView {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.showUsages(searchedFor, foundUsages, presentation, factory)

    return createUsageView(searchedFor, foundUsages, presentation, factory)
  }

  override fun showUsages(
    searchedFor: Array<out UsageTarget>,
    foundUsages: Array<out Usage>,
    presentation: UsageViewPresentation
  ): UsageView {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.showUsages(searchedFor, foundUsages, presentation)
    return createUsageView(searchedFor, foundUsages, presentation, null)
  }

  override fun searchAndShowUsages(
    searchFor: Array<out UsageTarget>,
    searcherFactory: Factory<out UsageSearcher>,
    showPanelIfOnlyOneUsage: Boolean,
    showNotFoundMessage: Boolean,
    presentation: UsageViewPresentation,
    listener: UsageViewStateListener?
  ): UsageView? {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.searchAndShowUsages(searchFor, searcherFactory, showPanelIfOnlyOneUsage, showNotFoundMessage, presentation, listener)

    val usageView = createUsageView(searchFor, arrayOf(), presentation, searcherFactory) as RdUsageView
    usageView.session.reRun.advise(Lifetime.Eternal) {
      PsiDocumentManager.getInstance(project).commitAllDocuments()
      usageView.clear()
      usageView.session.isSearchComplete.set(false)
      searchAndUpdateView(searcherFactory, searchFor, usageView)
    }

    searchAndUpdateView(searcherFactory, searchFor, usageView)
    return usageView
  }

  private fun searchAndUpdateView(
    searcherFactory: Factory<out UsageSearcher>,
    searchFor: Array<out UsageTarget>,
    usageView: RdUsageView
  ) {
    val usageSearcher = searcherFactory.create()
    val task = object : Task.Backgroundable(project, "Find Usages", true) {
      override fun run(indicator: ProgressIndicator) {
        usageSearcher.generate { usage ->
          if (!isSelfUsage(usage, searchFor)) {
            usageView.appendUsage(usage)
          }
          true
        }
      }

      override fun onFinished() {
        usageView.session.isSearchComplete.set(true)
        usageView.session.canReRun.set(true)
      }
    }
    ProgressManager.getInstance().run(task)
  }

  override fun searchAndShowUsages(
    searchFor: Array<out UsageTarget>,
    searcherFactory: Factory<out UsageSearcher>,
    processPresentation: FindUsagesProcessPresentation,
    presentation: UsageViewPresentation,
    listener: UsageViewStateListener?
  ) {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.searchAndShowUsages(searchFor, searcherFactory, processPresentation, presentation, listener)

    TODO("not implemented")
  }

  override fun getSelectedUsageView(): UsageView? {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.getSelectedUsageView()
    // TODO: implement
    return null
  }
}
