// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.findUsages

import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.usages.*
import com.intellij.usages.impl.UsageViewEx
import com.intellij.usages.rules.PsiElementUsage
import com.jetbrains.rd.framework.RdTaskResult
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rdserver.toRichStringModel
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ide.requests.queueUntilBulkOperationIsFinished
import com.jetbrains.rdserver.requests.BackendIdeGateway
import com.jetbrains.rdserver.vfs.createDocumentIdFromFile
import java.util.concurrent.CompletableFuture
import javax.swing.Action
import javax.swing.JComponent

class RdUsageView(val project: Project,
                  val targets: Array<out UsageTarget>,
                  val session: RdFindUsagesSession,
                  private val presentation: UsageViewPresentation) : UsageViewEx {
  private val usages = mutableListOf<Usage>()
  private val lifetimeDefinition = project.defineNestedLifetime()

  init {
    session.usages.updateItem.set { _ -> true }
    session.usages.navigateTo.advise(lifetimeDefinition.lifetime) { usageNavigationData ->
      usages[usageNavigationData.index].navigate(usageNavigationData.requestFocus)
    }
    session.targets.navigateTo.advise(lifetimeDefinition.lifetime) { usageTargetNavigationData ->
      targets[usageTargetNavigationData.index].navigate(usageTargetNavigationData.requestFocus)
    }
  }

  override fun appendUsage(usage: Usage) {
    addUsages(listOf(usage))
  }

  override fun appendUsagesInBulk(usages: Collection<Usage>): CompletableFuture<*> {
    return if (application.isDispatchThread) {
      addUsages(usages)
      CompletableFuture.completedFuture(Unit)
    }
    else {
      val future = CompletableFuture<Unit>()
      application.invokeLater {
        addUsages(usages)
        future.complete(Unit)
      }
      future
    }
  }

  private fun addUsages(usages: Collection<Usage>) {
    if (application.isDispatchThread)
      session.usages.items.add(usages.map { convertUsage(it) })
    else
      application.invokeLater {
        addUsages(usages)
      }
  }

  private fun convertUsage(usage: Usage): RdUsageBase {
    usages.add(usage)
    val position = convertPosition(usage)
    val isAccessedForReading = (usage as? ReadWriteAccessUsage)?.isAccessedForReading ?: true
    val isAccessedForWriting = (usage as? ReadWriteAccessUsage)?.isAccessedForWriting ?: true
    val psiElement = (usage as? PsiElementUsage)?.element
    val virtualFile = psiElement?.containingFile?.virtualFile
    // todo UX for usages in files hidden by permissions
    // todo use fileId in models
    val documentId = createDocumentIdFromFile(project, virtualFile)
    return if (position != null) {
      RdMergeableUsage(position, usages.size - 1,
                       usage.presentation.plainText,
                       isAccessedForReading,
                       isAccessedForWriting,
                       psiElement?.textRange?.startOffset ?: 0,
                       psiElement?.textRange?.endOffset ?: 0,
                       documentId,
                       usage.presentation.icon?.toProtocolModel(),
                       convertGroups(usage))
    }
    else {
      RdUsage(
        usage.presentation.text.toRichText(), usages.size - 1,
        usage.presentation.plainText,
        isAccessedForReading,
        isAccessedForWriting,
        psiElement?.textRange?.startOffset ?: 0,
        psiElement?.textRange?.endOffset ?: 0,
        documentId,
        usage.presentation.icon?.toProtocolModel(),
        convertGroups(usage)
      )
    }
  }

  private fun convertGroups(usage: Usage): List<RdUsageGroup> {
    return BackendUsageGroupingRulesSupport.getInstance(project).supportedRules.flatMap { rule ->
      rule.getParentGroupsFor(usage, targets).map {
        RdUsageGroup(rule.id, it.getText(this), it.getIcon(false)?.toProtocolModel())
      }
    }
  }

  private fun convertPosition(usage: Usage): RdUsagePosition? {
    val psiElement = (usage as? PsiElementUsage)?.element ?: return null
    val viewProvider = psiElement.containingFile?.viewProvider ?: return null
    val document = viewProvider.document ?: return null
    val textRange = (usage as? UsageInfo2UsageAdapter)?.navigationRange ?: psiElement.textRange
    val line = document.getLineNumber(textRange.startOffset)
    return RdUsagePosition(
      line,
      textRange.startOffset - document.getLineStartOffset(line),
      textRange.startOffset - document.getLineStartOffset(line),
      textRange.endOffset - document.getLineStartOffset(line),
      textRange.startOffset,
      textRange.endOffset,
      viewProvider.virtualFile.path
    )
  }

  fun clear() {
    usages.clear()
    session.usages.items.clear()
  }

  override fun includeUsages(usages: Array<out Usage>) {
    TODO("not implemented")
  }

  override fun excludeUsages(usages: Array<out Usage>) {
    TODO("not implemented")
  }

  override fun addButtonToLowerPane(runnable: Runnable, text: String, mnemonic: Char) {
    TODO("not implemented")
  }

  override fun addButtonToLowerPane(runnable: Runnable, text: String) {
    TODO("not implemented")
  }

  override fun addButtonToLowerPane(action: Action) {
    TODO("not implemented")
  }

  override fun setAdditionalComponent(component: JComponent?) {
    TODO("not implemented")
  }

  override fun addPerformOperationAction(processRunnable: Runnable,
                                         commandName: String?,
                                         cannotMakeString: String,
                                         shortDescription: String,
                                         checkReadOnlyStatus: Boolean) {
    BackendIdeGateway.getInstance(project).queueUntilBulkOperationIsFinished(lifetimeDefinition, "Add action to usage view") {
      isSearchInProgress = false
      val request = RdAddPerformOperationActionRequest(commandName, cannotMakeString, shortDescription, checkReadOnlyStatus)
      val task = session.addPerformOperationAction.start(lifetimeDefinition, request)
      task.result.advise(lifetimeDefinition) { result ->
        if (result is RdTaskResult.Success) {
          processRunnable.run()
        }
      }
    }
  }

  override fun getPresentation(): UsageViewPresentation {
    return presentation
  }

  override fun getExcludedUsages(): Set<Usage> {
    return emptySet()
  }

  override fun getSelectedUsages(): Set<Usage> {
    TODO("not implemented")
  }

  override fun getUsages(): Set<Usage> {
    return usages.toSet()
  }

  override fun getComponent(): JComponent {
    error("RdUsageView does not have UI components")
  }

  override fun searchHasBeenCancelled(): Boolean {
    TODO("not implemented")
  }

  override fun cancelCurrentSearch() {
    TODO("not implemented")
  }

  override fun associateProgress(indicator: ProgressIndicator) {
    TODO("not implemented")
  }

  override fun getSortedUsages(): MutableList<Usage> {
    return usages
  }

  override fun waitForUpdateRequestsCompletion() {
    TODO("not implemented")
  }

  override fun setSearchInProgress(searchInProgress: Boolean) {
    session.isSearchComplete.set(!searchInProgress)
  }

  override fun searchFinished() {
    session.isSearchComplete.set(true)
  }

  override fun removeUsage(usage: Usage) {
    TODO("not implemented")
  }

  override fun selectUsages(usages: Array<out Usage>) {
    TODO("not implemented")
  }

  override fun close() {
    lifetimeDefinition.terminate()
    TODO("Remove session")
  }

  override fun isSearchInProgress(): Boolean {
    return session.isSearchComplete.valueOrNull != true
  }

  override fun addPerformOperationAction(processRunnable: Runnable,
                                         commandName: String?,
                                         cannotMakeString: String,
                                         shortDescription: String) {
    TODO("not implemented")
  }

  override fun dispose() {
    TODO("not implemented")
  }

  override fun getUsagesCount(): Int {
    TODO("not implemented")
  }

  override fun removeUsagesBulk(usages: Collection<Usage>) {
    TODO("not implemented")
  }

  override fun setReRunActivity(runnable: Runnable) {
    TODO("not implemented")
  }
}

private fun Array<TextChunk>.toRichText(): RichTextModel {
  return RichTextModel(map { it.text.toRichStringModel(it.simpleAttributesIgnoreBackground) })
}
