package com.jetbrains.rdserver.followMe

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.project.Project
import com.jetbrains.rd.platform.codeWithMe.users.UserFocusManager
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.reactive.Signal

data class FocusChangeEvent(val leadId: ClientId, val source: FocusChangeSource)

enum class FocusChangeSource {
  BackendFocusChanged,
  FrontendFocusChanges,
  FollowingStarted
}

class BackendUserFocusManager(project: Project) : UserFocusManager(project) {

  companion object {
    fun getBackendInstance(project: Project): BackendUserFocusManager {
      return getInstance(project) as BackendUserFocusManager
    }
  }

  private val followMeManager = FollowMeManager.getInstance(project)

  val changeSignal = Signal<FocusChangeEvent>()

  init {
    followMeManager.followers.advise(project.lifetime) { event ->
      val leadId = event.newValueOpt
      if (leadId == null) {
        val followerId = event.key
        if (followerId == ClientId.localId)
          leadActiveToolWindowIdProperty.set(null)
        else
          BackendUserFocusHost.tryGetInstance(project, followerId)?.setLeadActiveToolWindow(null)
      }
      else {
        refreshActiveToolWindow(leadId, FocusChangeSource.FollowingStarted)
      }
    }
  }

  override fun currentActiveToolWindowChanged(toolWindowId: String?) {
    refreshActiveToolWindow(ClientId.localId, FocusChangeSource.BackendFocusChanged)
  }

  fun refreshActiveToolWindow(leadId: ClientId, source: FocusChangeSource) {
    // Obtain active toolwindow ID for the user 'where'
    val activeToolWindow =
      if (leadId == ClientId.localId) currentActiveToolWindowId.value
      else BackendUserFocusHost.getInstance(project, leadId).getCurrentActiveToolWindow()

    // Go through all users who follows 'where' and update lead ID
    var inFullSyncMode = false
    for (followerId in followMeManager.getFollowers(leadId)) {
      inFullSyncMode = inFullSyncMode || followMeManager.isFollowing(leadId, followerId)
      setLeadActiveToolWindowIdFor(followerId, activeToolWindow)
    }

    // In case of 'full sync mode' sync toolwindow id back to the lead user
    if (inFullSyncMode) {
      setLeadActiveToolWindowIdFor(leadId, activeToolWindow)
    }

    // Fire notifications
    changeSignal.fire(FocusChangeEvent(leadId, source))
  }

  private fun setLeadActiveToolWindowIdFor(clientId: ClientId, activeToolWindow: String?) {
    if (clientId == ClientId.localId) {
      // Update host state
      leadActiveToolWindowIdProperty.value = activeToolWindow
    }
    else {
      // Update other clients
      BackendUserFocusHost.getInstance(project, clientId).setLeadActiveToolWindow(activeToolWindow)
    }
  }
}

