package com.jetbrains.rdserver.followMe

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.project.Project
import com.jetbrains.codeWithMe.model.RdUserFollowingRequest
import com.jetbrains.codeWithMe.model.RdUserFollowingRetarget
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.codeWithMe.model.UserModel
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.framework.impl.RdProperty
import com.jetbrains.rd.platform.codeWithMe.users.UserRequestFollowingCookie
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rd.platform.codeWithMe.util.onTermination
import com.jetbrains.rd.platform.codeWithMe.util.view
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.RdServerPerClientComponents

@Suppress("NonDefaultConstructor")
class BackendUsersHost(
  lifetime: Lifetime,
  val project: Project,
  protocol: IProtocol
) {

  private val model = UserModel.create(lifetime, protocol)
  private val clientId = ClientId.current

  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendUsersHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendUsersHost::class.java)
    }
  }

  init {
    val followMeManager = FollowMeManager.getInstance(project)

    followMeManager.users.view(lifetime, clientId) { lt, (_, userModel) ->
      model.users[userModel.id] = userModel
      lt.onTermination(clientId) {
        model.users.remove(userModel.id)
      }
    }

    followMeManager.followers.view(lifetime, clientId) { followingLifetime, (who, whom) ->
      if (whom == clientId) {
        val user = followMeManager.users[who] ?: return@view
        model.followers.add(user.id)
        followingLifetime.onTermination(clientId) {
          model.followers.remove(user.id)
        }
      }

      if (who == clientId) {
        require(ClientId.current == clientId) { "invalid client id in users host ${ClientId.current}"}

        // Update current follow user for this client, it can be changed inside FollowMeManager
        val targetUser = followMeManager.users[whom]
        val targetUserId = targetUser?.id
        if (targetUserId != null) {
          model.follow.value = targetUserId
          followingLifetime.onTermination(clientId) {
            if (model.follow.value == targetUserId) {
              model.follow.value = null
            }
          }
        }
      }
    }

    val followProperty = model.follow as RdProperty<Int?>
    followProperty.change.advise(lifetime) { targetUserId ->
      if (followProperty.isLocalChange)
        return@advise
      if (targetUserId == null) {
        followMeManager.unregisterFollower(clientId)
        return@advise
      }

      val targetUser = model.users[targetUserId]!!
      followMeManager.registerFollower(clientId, targetUser.clientId)
    }

    model.jumpTo.advise(lifetime) { targetUserId ->
      val targetUser = model.users[targetUserId]!!
      BackendUserOperations.getInstance(project).jump(clientId, targetUser)
    }

    model.followingRequest.advise(lifetime) { request ->
      if (UserRequestFollowingCookie.isLocalChange()) return@advise
      val targetUser = model.users[request.userId]!!
      BackendUserOperations.getInstance(project).requestFollowing(clientId, targetUser, request.force)
    }

    model.stopFollowingRequest.advise(lifetime) { userId ->
      if (UserRequestFollowingCookie.isLocalChange()) return@advise
      val targetUser = model.users[userId]!!
      BackendUserOperations.getInstance(project).requestStopFollowing(clientId, targetUser)
    }
  }

  fun requestFollowing(who: RdUserModel, force: Boolean) {
    ClientId.withClientId(clientId) {
      UserRequestFollowingCookie.using {
        model.followingRequest.fire(RdUserFollowingRequest(who.id, force))
      }
    }
  }

  fun requestStopFollowing() {
    ClientId.withClientId(clientId) {
      UserRequestFollowingCookie.using {
        val currentClient = FollowMeManager.getInstance(project).getUser(clientId)
        model.stopFollowingRequest.fire(currentClient.id)
      }
    }
  }

  fun retargetFollowing(from: RdUserModel, to: RdUserModel) {
    ClientId.withClientId(clientId) {
      model.followingRetarget.fire(RdUserFollowingRetarget(from.name, to.name))
    }
  }
}