package com.jetbrains.rdserver.followMe

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.components.service
import com.intellij.openapi.project.Project
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.users.UserNotifications
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.IViewableMap
import com.jetbrains.rd.util.reactive.ViewableMap
import com.jetbrains.rdserver.BackendUserSettings
import kotlin.collections.set

class FollowMeManagerService(private val project: Project) : FollowMeManager() {

  init {
    application.invokeLater {
      if (project.isDisposed) return@invokeLater
      initializeHostUser(project.lifetime)
    }
  }

  override fun onFollowingRetargeted(transitiveFollowId: ClientId, who: ClientId, whomUser: RdUserModel) {
    // Show notification about automatic switching follower
    val transitiveFollowUser = getUser(transitiveFollowId)
    if (who == ClientId.localId) {
      UserNotifications.showFollowingRetargetNotification(project, whomUser.name, transitiveFollowUser.name)
    }
    else {
      BackendUsersHost.getInstance(project, who).retargetFollowing(whomUser, transitiveFollowUser)
    }
  }

  override fun assertIsDispatchThread() {
    application.assertIsDispatchThread()
  }
}

/**
 * Project service to manage all connected users and following relations between them
 * Contains all appropriate API to manage them on server-side
 */
abstract class FollowMeManager {

  companion object {
    private val LOG = getLogger<FollowMeManager>()
    fun getInstance(project: Project): FollowMeManager = project.service()

    private val ALLOWED_ACTIONS_FOR_POPUP_REFLECTION = listOf(
      "GotoImplementation",
      "GotoDeclaration",
      "ShowUsages"
      //actionId == "NewClass", // TODO: NewClass reflected popup doesn't work properly for now
      //IdeActions.ACTION_QUICK_IMPLEMENTATIONS,
    )
  }

  private var nextId = 0
  private val myUsers = ViewableMap<ClientId, RdUserModel>()
  private val myFollowers = ViewableMap<ClientId, ClientId>()

  /**
   * All users (including Host) connected to this project
   */
  val users: IViewableMap<ClientId, RdUserModel>
    get() = myUsers

  /**
   * All following relations between users
   */
  val followers: IViewableMap<ClientId, ClientId>
    get() = myFollowers

  protected abstract fun assertIsDispatchThread()
  protected abstract fun onFollowingRetargeted(transitiveFollowId: ClientId, who: ClientId, whomUser: RdUserModel)

  protected fun initializeHostUser(lifetime: Lifetime) {
    registerUser(lifetime, ClientId.localId, BackendUserSettings.getHostUserName())
  }

  /**
   * Registers new user with specified client ID
   */
  fun registerUser(lifetime: Lifetime, clientId: ClientId, userName: String) {
    assertIsDispatchThread()
    assert(!users.containsKey(clientId)) { "User with ClientId=$clientId already registered in FollowMeManager" }

    val userId = nextId++
    val userModel = RdUserModel(userId, clientId.value, userName)
    myUsers[clientId] = userModel
    LOG.info("Register new user $userId:'$userName' with ClientId=$clientId")

    lifetime.onTermination {
      assertIsDispatchThread()
      assert(users.containsKey(clientId)) { "Can not find user with ClientId=$clientId in FollowMeManager" }
      LOG.info("Unregister user $userId:'$userName' with ClientId=$clientId")

      // Remove user
      myUsers.remove(clientId)

      // Cleanup following
      myFollowers.remove(clientId)
      for ((who, whom) in followers.toList()) {
        if (whom == clientId) {
          myFollowers.remove(who)
        }
      }
    }
  }

  /**
   * Returns a user with specified client ID. Throws exception if a user was not found
   */
  fun getUser(clientId: ClientId): RdUserModel {
    assertIsDispatchThread()
    return users[clientId] ?: error("Can not find user for ClientId=$clientId")
  }

  /**
   * Registers new following relationship: user `who` now follows user `whom`
   */
  fun registerFollower(who: ClientId, whom: ClientId) {
    assertIsDispatchThread()
    assert(who != whom) { "Can not follower yourself, ClientId=$who" }

    val whoUser = getUser(who)
    val whomUser = getUser(whom)
    LOG.info("User '${whoUser.name}' with ClientId=$who follows user '${whomUser.name}' ith ClientId=$whom")

    val transitiveFollowId = followers[whom]
    if (transitiveFollowId != null && transitiveFollowId != who) {
      myFollowers[who] = transitiveFollowId
      onFollowingRetargeted(transitiveFollowId, who, whomUser)
    }
    else {
      myFollowers[who] = whom
    }
  }

  /**
   * Removes a following relationship: user `who` now follows no one
   */
  fun unregisterFollower(who: ClientId) {
    assertIsDispatchThread()

    val whoUser = getUser(who)
    LOG.info("User '${whoUser.name}' with ClientId=$who stops following")

    myFollowers.remove(who)
  }

  fun isHostFollowingClient(whom: ClientId): Boolean {
    return isFollowing(ClientId.localId, whom)
  }

  fun isClientFollowingHost(who: ClientId): Boolean {
    return isFollowing(who, ClientId.localId)
  }

  fun isFollowing(who: ClientId, whom: ClientId): Boolean {
    return followers[who] == whom
  }

  fun isFullSyncBetween(user1: ClientId, user2: ClientId): Boolean {
    return isFollowing(user1, user2) && isFollowing(user2, user1)
  }

  fun isFullSyncBetweenHostAndClient(client: ClientId): Boolean {
    return isFollowing(ClientId.localId, client) && isFollowing(client, ClientId.localId)
  }

  fun getFollowers(whom: ClientId): List<ClientId> {
    return followers.entries.filter { it.value == whom }.map { it.key }.toList()
  }

  fun getFollowersWithoutHost(whom: ClientId): List<ClientId> {
    return getFollowers(whom).filter { it != ClientId.localId }.toList()
  }

  fun canReflectPopupToFollowers(currentActionId: String?): Boolean {
    return currentActionId != null && ALLOWED_ACTIONS_FOR_POPUP_REFLECTION.contains(currentActionId)
  }
}