package com.jetbrains.rdserver.internal

import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.startup.StartupActivity
import com.intellij.refactoring.rename.inplace.InplaceRefactoring
import com.intellij.util.ui.EdtInvocationManager
import com.jetbrains.rd.platform.codeWithMe.users.LobbyUrlSettings
import com.jetbrains.rd.platform.util.RdCoroutineHost
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.CommonsLoggingLoggerFactory
import com.jetbrains.rd.util.ILoggerFactory
import com.jetbrains.rd.util.Statics
import com.jetbrains.rdserver.joinLinks.RdServerJoinLinkManager
import com.jetbrains.rdserver.startup.RdServerDirectConnectionsListener
import com.jetbrains.rdserver.ui.RdServerEdtInvocationManager
import java.io.File
import java.net.URI
import java.nio.file.Files
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicBoolean

class RdServerOnProjectOpen: StartupActivity.DumbAware {
  companion object {
    private val logger = Logger.getInstance(RdServerOnProjectOpen::class.java)
    private val initiatedListening = AtomicBoolean(false)
    const val listenOnProjectOpenPortProperty = "codeWithMe.internal.listenPortOnProjectOpen"
    private const val createSessionJoinLinkFileOnProjectOpenProperty = "codeWithMe.internal.createSessionJoinLinkFileOnProjectOpen"
  }

  override fun runActivity(project: Project) {
    try {
      // Listen only on first project
      if (initiatedListening.getAndSet(true)) return

      // initialize logging for protocol events and logging from rd library code
      Statics<ILoggerFactory>().push(CommonsLoggingLoggerFactory)
      RdCoroutineHost.initAsync()

      EdtInvocationManager.setEdtInvocationManager(RdServerEdtInvocationManager())
      // todo is it needed?
      InplaceRefactoring.setShowBalloonInHeadlessMode(true)

      val listenPort = System.getProperty(listenOnProjectOpenPortProperty)?.toIntOrNull()
      val joinLinkFile = System.getProperty(createSessionJoinLinkFileOnProjectOpenProperty)?.let { File(it) }
      if (listenPort != null && joinLinkFile != null)
        error("Auto-listening via TCP and a auto-creating a session are mutually exclusive.")

      if (listenPort != null) {
        logger.info("$listenOnProjectOpenPortProperty=$listenPort")
        val directConnectionListener = RdServerDirectConnectionsListener.getInstance(project)
        directConnectionListener.startListener(project.lifetime, listenPort, true)
      }

      if (joinLinkFile != null) {
        logger.info("$createSessionJoinLinkFileOnProjectOpenProperty=${joinLinkFile.canonicalPath}")
        if (joinLinkFile.exists()) Files.delete(joinLinkFile.toPath())

        if (!LobbyUrlSettings.lobbyUrl.isSetFromOs && LobbyUrlSettings.lobbyUrl.uri == LobbyUrlSettings.DEFAULT_LOBBY_URL) {
          logger.warn("Production server URI is used with the test-specific flag for generating join links on start. Setting lobby server URI to staging.")
          LobbyUrlSettings.setLobbyUrl(URI("https://codewithme-lobby-staging.api.jetbrains.com"))
        }

        logger.info("Lobby server URI=${LobbyUrlSettings.lobbyUrl.uri}")

        val joinLink = RdServerJoinLinkManager.getInstance(project).getJoinLink().toCompletableFuture().get(1, TimeUnit.MINUTES)
        joinLinkFile.writeText(joinLink.toString())
      }

    } catch (t: Throwable) {
      Logger.getInstance(javaClass).error("Unable to listen on project open: ${t.message}", t)
    }
  }
}