package com.jetbrains.rdserver.joinLinks

import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.ide.CopyPasteManager
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ProjectManager
import com.intellij.openapi.ui.Messages
import com.intellij.util.concurrency.annotations.RequiresEdt
import com.intellij.util.ui.TextTransferable
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.connection.LobbyServerApi
import com.jetbrains.rd.platform.codeWithMe.users.UserNotifications
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.internal.CwmEULA
import com.jetbrains.rdserver.permissions.ui.PermissionsDialog
import com.jetbrains.rdserver.ui.HostNotifications
import java.net.URI

class GetJoinLinkAction : AnAction(), DumbAware {

  override fun update(e: AnActionEvent) {
    val project = e.project
    if (project == null) {
      e.presentation.isEnabledAndVisible = false
      return
    }
    val hasNoOwnLink = RdServerJoinLinkManager.getInstance(project).getExistingJoinLink() == null
    val anyProjectHasLink = ProjectManager.getInstance().openProjects.any { RdServerJoinLinkManager.getInstance(it).getExistingJoinLink() != null }
    e.presentation.isVisible = hasNoOwnLink
    e.presentation.isEnabled = !anyProjectHasLink
  }

  override fun actionPerformed(e: AnActionEvent) {
    val project = e.project ?: return
    invokeAction(project)
  }

  @RequiresEdt
  private fun invokeAction(project: Project) {
    if (!CwmEULA.isAccepted() && !CwmEULA.showEulaAcceptanceDialog()) {
      logger<GetJoinLinkAction>().warn("CWM EULA declined, will not create join link")
      return
    }

    if (!PermissionsDialog(project, null, true).showAndGet())
      return

    val future = RdServerJoinLinkManager.getInstance(project).getJoinLink()
    future.handleAsync { link: URI?, exception: Throwable? ->
      when {
        exception is LobbyServerApi.LobbySessionsLimitReachedException -> {
          Logger.getInstance(javaClass).info("Sessions limit reached on lobby server: ${exception.message}", exception)

          application.invokeLater {
            if (HostNotifications.showSessionsLimitDialog(project) == HostNotifications.SessionsLimitDialogResult.RETRY) {
              application.invokeLater {
                invokeAction(project)
              }
            }
          }
        }

        exception != null -> {
          Logger.getInstance(javaClass).warn("Error creating join link: ${exception.message}", exception)

          application.invokeLater {
            Messages.showErrorDialog(
              project,
              CodeWithMePluginBundle.message("notification.join.link.error",
                                             exception.message ?: "<error message is not available>"),
              CodeWithMePlatformBundle.message("product.name"))
          }
        }

        else -> {
          val linkString = link!!.toString()
          CopyPasteManager.getInstance().setContents(TextTransferable(linkString as String?))

          UserNotifications.showGreenNotification(
            project,
            CodeWithMePluginBundle.message("notification.join.link.copied"))
        }
      }
    }
  }
}
