package com.jetbrains.rdserver.joinLinks

import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.ProgressManager
import com.intellij.openapi.progress.Task.Backgroundable
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.registry.Registry
import com.intellij.ui.jcef.JBCefApp
import com.jetbrains.codeWithMe.lobby.CodeWithMeFeature
import com.jetbrains.codeWithMe.lobby.CodeWithMeTransport
import com.jetbrains.rd.platform.codeWithMe.connection.LobbyServerApi.isOnPremise
import com.jetbrains.rd.platform.codeWithMe.network.CwmStunUtil
import com.jetbrains.rd.platform.codeWithMe.network.quic.DatagramSocketWrapper
import com.jetbrains.rd.platform.codeWithMe.network.quic.QuicServer
import com.jetbrains.rd.platform.codeWithMe.network.quic.Quiche
import com.jetbrains.rd.platform.codeWithMe.network.quic.StunConnector
import com.jetbrains.rd.platform.codeWithMe.statistics.CodeWithMeLifecycleLogger
import com.jetbrains.rd.platform.codeWithMe.statistics.TelephonyStatisticsCollector
import com.jetbrains.rd.platform.codeWithMe.users.LobbyUrlSettings
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.firstOrNull
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.Signal
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rdserver.startup.EdtScheduler
import com.jetbrains.rdserver.startup.RdServerConnectionManager
import com.jetbrains.rdserver.startup.RdServerDirectConnectionsListener
import com.jetbrains.rdserver.startup.RdServerRelayedConnectionsListener
import com.jetbrains.rdserver.telephony.BackendTelephonyManager
import java.net.URI
import java.time.Instant
import java.time.temporal.ChronoUnit
import java.util.concurrent.*
import kotlin.collections.set
import kotlin.math.abs

@Service
class RdServerJoinLinkManager(val project: Project) {

  companion object {
    private val LOG = Logger.getInstance(RdServerJoinLinkManager::class.java)

    fun getInstance(project: Project) = project.service<RdServerJoinLinkManager>()

    val p2pEnabled
      get() = Registry.`is`("codewithme.peer2peer.enabled")

    val relayEnabled
      get() = Registry.`is`("codewithme.relay.enabled")

    val directConnectionsEnabled
      get() = Registry.`is`("codewithme.direct.enabled")

    val telephonyEnabled
      get() = Registry.`is`("codeWithMe.voiceChat.enabled")

    val forceRegionRegistryValue: String?
      get() {
        val stringValue = Registry.stringValue("codewithme.forceRegion")
        return if (stringValue.isBlank()) null else stringValue
      }
  }

  private data class JoinLinkPrivateData(
    val sessionId: String,
    val sessionUpdateToken: String?,
    val lifetimeDefinition: LifetimeDefinition,
    val publicData: JoinLinkData
  )

  data class JoinLinkData(
    val lifetime: Lifetime,
    val telephonyUrl: String?
  )

  private val joinLinks = ConcurrentHashMap<URI, JoinLinkPrivateData>()

  val joinLinksUpdated = Signal<Unit>()

  fun getExistingJoinLink(): URI? = joinLinks.firstOrNull()?.key
  fun getJoinLinkData(link: URI): JoinLinkData? = joinLinks[link]?.publicData

  fun getCurrentSessionId() : String? {
    return joinLinks[getExistingJoinLink() ?: return null]?.sessionId
  }

  fun terminateJoinLinks() {
    joinLinks.values.toList().forEach {
      val telephonyEnabled = it.publicData.telephonyUrl != null && BackendTelephonyManager.getInstance(project).isTelephonyEnabled.value
      CodeWithMeLifecycleLogger.onSessionTerminated(project, it.sessionId, telephonyEnabled)
      TelephonyStatisticsCollector.cwmSessionEnded(project, telephonyEnabled)
      it.lifetimeDefinition.terminate()
    }
    joinLinksUpdated.fire()
  }

  fun getJoinLink(): CompletionStage<URI> {
    val lastJoinLink = getExistingJoinLink()
    if (lastJoinLink != null) {
      return CompletableFuture.completedFuture(lastJoinLink)
    }

    val result = CompletableFuture<URI>()
    ProgressManager.getInstance().run(object : Backgroundable(project, "Creating Join Link", false) {
      override fun run(indicator: ProgressIndicator) {
        val directListenerLifetimeDef = project.lifetime.createNested()
        val p2pLifetimeDef = project.lifetime.createNested()

        try {
          val serverInfo = RdServerLobbySessionManager.getInstance(project).serverInfo()
          LOG.info("Received server info from lobby: $serverInfo")

          val simpleStunServers = CwmStunUtil.filterSimpleStunServers(serverInfo.stunServers)

          val relayEnabledWithPolicies = relayEnabled && serverInfo.enabledFeatures.contains(CodeWithMeFeature.WsRelay)
          val directConnectionsEnabledWithPolicies = directConnectionsEnabled && serverInfo.enabledFeatures.contains(CodeWithMeFeature.DirectTcp)
          val telephonyEnabledWithPolicies = telephonyEnabled && JBCefApp.isSupported() && serverInfo.enabledFeatures.contains(CodeWithMeFeature.JitsiTelephony)
          val p2pEnabledWithPolicies = p2pEnabled
                                       && Quiche.isAvailable
                                       && serverInfo.enabledFeatures.contains(CodeWithMeFeature.Peer2PeerQuic)
                                       && simpleStunServers.isNotEmpty()

          LOG.info("P2P availability = $p2pEnabledWithPolicies: " +
                   "registry key = $p2pEnabled, " +
                   "simpleStunServers.isNotEmpty = ${simpleStunServers.isNotEmpty()}, " +
                   "Quiche.isAvailable = ${Quiche.isAvailable}, " +
                   "serverInfo.enabledFeatures.contains(CodeWithMeFeature.Peer2PeerQuic) = ${serverInfo.enabledFeatures.contains(CodeWithMeFeature.Peer2PeerQuic)}")

          // TODO Handle lobby errors and report it to the user
          val directLinkPort = if (directConnectionsEnabledWithPolicies) {
            RdServerDirectConnectionsListener.getInstance(project).startListener(directListenerLifetimeDef)
          }
          else null

          var hostPeer2PeerQuicEndpoint: String? = null
          var hostPeer2PeerDatagramSocketWrapper: DatagramSocketWrapper? = null

          val connectionManager = RdServerConnectionManager.getInstance(project)

          if (p2pEnabledWithPolicies) {
            hostPeer2PeerDatagramSocketWrapper = DatagramSocketWrapper(p2pLifetimeDef)
            val quicServer = QuicServer(p2pLifetimeDef, hostPeer2PeerDatagramSocketWrapper)

            quicServer.acceptSignal.advise(p2pLifetimeDef) { acceptData ->
              LOG.info("QUIC: accepted connection")

              val id = "QUIC:${acceptData.peer}"

              // createServerOnTransport expects lifetime to be terminated on EDT
              val connectionEdtLifetimeDefinition = project.lifetime.createNested()
              acceptData.lifetimeOnQuicThread.onTermination {
                application.invokeLater {
                  connectionEdtLifetimeDefinition.terminate()
                }
              }

              connectionManager.createServerOnTransport(
                connectionEdtLifetimeDefinition.lifetime, acceptData.streamTransport, EdtScheduler, id,
                CodeWithMeTransport.Peer2PeerQuic)
            }

            val ourExternalAddressFuture = StunConnector.maintainStunServerConnection(p2pLifetimeDef, hostPeer2PeerDatagramSocketWrapper, simpleStunServers.first())

            val stunStartTime = System.currentTimeMillis()

            val ourExternalAddress = try {
              ourExternalAddressFuture.get(500, TimeUnit.MILLISECONDS)
            } catch (t: TimeoutException) {
              LOG.info("Timeout getting external address via STUN")
              null
            }

            LOG.info("External address via STUN in ${System.currentTimeMillis() - stunStartTime} ms: $ourExternalAddress")

            if (ourExternalAddress != null) {
              hostPeer2PeerQuicEndpoint = "quic://${ourExternalAddress.hostString}:${ourExternalAddress.port}"
            }
          }

          val (sessionResponse, cert) = RdServerLobbySessionManager.getInstance(project).createSession(
            serverInfo = serverInfo,
            directTcpLocalPort = directLinkPort,
            hostPeer2PeerQuicEndpoint = hostPeer2PeerQuicEndpoint,
            relayEnabled = relayEnabledWithPolicies,
            jitsiTelephonyEnabled = telephonyEnabledWithPolicies,
            forceRegion = forceRegionRegistryValue,
          )
          LOG.info("Received response from lobby server: $sessionResponse")

          val link = sessionResponse.shareableJoinLink

          val sessionLifetimeDef = RdServerLobbySessionManager.getInstance(project).registerSession(
            LobbyUrlSettings.lobbyUrl.uri, sessionResponse.sessionId, sessionResponse.sessionUpdateToken!!, sessionResponse.nextHeartbeatIntervalMs, cert.isOnPremise()
          )

          if (relayEnabledWithPolicies) {
            RdServerRelayedConnectionsListener.getInstance(project).attachToSession(sessionResponse, sessionLifetimeDef)
          }

          if (p2pEnabledWithPolicies && hostPeer2PeerDatagramSocketWrapper != null) {
            RdServerLobbySessionManager.getInstance(project).listenForConnectedClients(
              sessionLifetimeDef.lifetime, sessionResponse.listenForClientConnectionEndpoint,
              hostPeer2PeerDatagramSocketWrapper, sessionResponse.sessionUpdateToken)
          }

          val telephonyEnabledOnSessionStart = sessionResponse.jitsiRoomUrl != null && BackendTelephonyManager.getInstance(project).isTelephonyEnabled.value
          CodeWithMeLifecycleLogger.onSessionCreated(project, sessionResponse.sessionId, telephonyEnabledOnSessionStart)
          TelephonyStatisticsCollector.cwmSessionStarted(project, telephonyEnabledOnSessionStart)
          val joinLinkData = JoinLinkPrivateData(
            sessionId = sessionResponse.sessionId,
            lifetimeDefinition = sessionLifetimeDef,
            sessionUpdateToken = sessionResponse.sessionUpdateToken,
            publicData = JoinLinkData(
              lifetime = sessionLifetimeDef.lifetime,
              telephonyUrl = sessionResponse.jitsiRoomUrl
            )
          )

          joinLinks[link] = joinLinkData
          sessionLifetimeDef.lifetime.onTermination {
            val sessionId = joinLinkData.sessionId

            val sessionStartTime = connectionManager.getSessionStartTime(sessionId)
            connectionManager.clearSessionStartTimeMap(sessionId)

            val sessionParticipantsCount = connectionManager.getSessionParticipantsMax(sessionId)
            connectionManager.clearSessionParticipantsMaxCountMap(sessionId)

            val sessionDisconnectsCount = connectionManager.getSessionDisconnectsCount(sessionId)
            connectionManager.clearSessionDisconnectsCountMap(sessionId)

            joinLinks.remove(link)
            // todo: use refactor these signals
            joinLinksUpdated.fire()

            catch { directListenerLifetimeDef.terminate() }
            catch { p2pLifetimeDef.terminate() }

            val sessionManager = RdServerLobbySessionManager.getInstance(project)

            application.executeOnPooledThread {
              // TODO Retry it?
              try {
                sessionManager.removeSession(link, joinLinkData.sessionUpdateToken)
              } catch (t: Throwable) {
                LOG.warn("Unable to remove session from lobby: ${t.message}", t)
              }
            }

            if (sessionStartTime != null) {
              CodeWithMeLifecycleLogger.onSessionFinished(
                project = project,
                sessionId = sessionId,
                maxParticipants = sessionParticipantsCount ?: 1,
                sessionDurationMinutes = abs(ChronoUnit.MINUTES.between(sessionStartTime, Instant.now())).toInt(),
                disconnectsCount = sessionDisconnectsCount ?: 0
              )
            }
          }
          joinLinksUpdated.fire()
          project.messageBus.syncPublisher(RdServerJoinLinkLifecycleListener.TOPIC).joinLinkCreated(sessionLifetimeDef.lifetime, link)
          result.complete(link)
        }
        catch (t: Throwable) {
          directListenerLifetimeDef.terminate()
          p2pLifetimeDef.terminate()
          result.completeExceptionally(t)
        }
      }
    })

    return result
  }
}