package com.jetbrains.rdserver.joinLinks

import com.google.common.net.InetAddresses
import com.intellij.openapi.application.ApplicationInfo
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.NlsSafe
import com.intellij.ui.LicensingFacade
import com.intellij.util.concurrency.AppExecutorUtil
import com.jetbrains.codeWithMe.license.CertificateDecoder
import com.jetbrains.codeWithMe.lobby.*
import com.jetbrains.codeWithMe.model.CWMSessionStatus
import com.jetbrains.codeWithMe.model.GuestTerminationReason
import com.jetbrains.rd.platform.codeWithMe.connection.*
import com.jetbrains.rd.platform.codeWithMe.connection.LobbyServerApi.isOnPremise
import com.jetbrains.rd.platform.codeWithMe.network.quic.DatagramSocketWrapper
import com.jetbrains.rd.platform.codeWithMe.statistics.CodeWithMeLifecycleLogger
import com.jetbrains.rd.platform.codeWithMe.users.LobbyUrlSettings
import com.jetbrains.rd.platform.codeWithMe.users.LocalUserSettings
import com.jetbrains.rd.platform.codeWithMe.util.CodeWithMePlatformUtil
import com.jetbrains.rd.platform.util.*
import com.jetbrains.rd.util.catch
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.startup.RdServerConnectionManager
import com.jetbrains.rdserver.ui.HostNotifications
import org.jetbrains.annotations.Nls
import java.net.*
import java.time.Duration
import java.time.Instant
import java.time.temporal.ChronoUnit
import java.util.concurrent.CompletableFuture
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.ScheduledFuture
import java.util.concurrent.TimeUnit
import java.util.function.Supplier
import kotlin.math.abs
import kotlin.math.max
import kotlin.math.min

@Service
class RdServerLobbySessionManager(val project: Project) {
  companion object {
    private val logger = Logger.getInstance(RdServerLobbySessionManager::class.java)
    private const val freeSessionDurationMinutes = 30
    val acceptedProducts = listOf(LicenseStampCheck.CWM_PRODUCT_CODE, "PRDEV")
    fun getInstance(project: Project) = project.service<RdServerLobbySessionManager>()

    @Suppress("NOTHING_TO_INLINE")
    inline fun hasLicenseStamp() = acceptedProducts.find { LicenseStampCheck.hasLicenseStamp(it) } != null

    @Suppress("NOTHING_TO_INLINE")
    inline fun isLicensed() = acceptedProducts.find { LicenseStampCheck.isLicensed(it) } != null
  }

  private val lifetime = project.lifetime
  private val knownSessions = ConcurrentHashMap<String, LifetimeDefinition>()
  private val heartbeatTasks = ConcurrentHashMap<String, ScheduledFuture<*>>()

  fun isKnownSession(sessionId: String) = knownSessions.containsKey(sessionId)

  private val connectionStatusService = CwmConnectionStatusService.getInstance(project)

  fun registerSession(serverUri: URI,
                      sessionId: String,
                      sessionUpdateToken: String,
                      heartbeatIntervalMs: Int?,
                      onPremiseLobby: Boolean): LifetimeDefinition {
    if (knownSessions.containsKey(sessionId)) {
      error("Session '$sessionId' already exists")
    }

    val lifetimeDefinition = lifetime.createNested()
    val hbInterval = heartbeatIntervalMs ?: 60 * 1000
    lifetimeDefinition.lifetime.bracket(
      {
        knownSessions[sessionId] = lifetimeDefinition
        logger.info("Registered new Code With Me session '$sessionId'")

        val sessionStart = Instant.now()
        if (!onPremiseLobby) {
          updateSessionStatus(sessionStart)
        }
        scheduleHeartbeat(serverUri, HeartbeatRequest(sessionId, sessionUpdateToken), hbInterval, sessionStart)
      },
      {
        val sessionLifetime = knownSessions.remove(sessionId)
        if (sessionLifetime != null) {
          logger.info("Removed Code With Me session '$sessionId'")
        }
        else {
          logger.info("Code With Me session '$sessionId' does not exists upon session termination")
        }
        heartbeatTasks.remove(sessionId)?.cancel(false)

        // reset connectionStatusService data
        connectionStatusService.sessionStatus.set(null)
        connectionStatusService.freeSessionRemainingTimeMs.set(null)
      })

    return lifetimeDefinition
  }

  private fun scheduleHeartbeat(serverUri: URI, request: HeartbeatRequest, timeout: Int, sessionStart: Instant, failedRequestsCount : Int = 0) {
    heartbeatTasks.put(request.sessionId, AppExecutorUtil.getAppScheduledExecutorService().schedule({
      var nextTimeout = timeout
      var failuresCount = failedRequestsCount
      var stopSession = false
      @NlsSafe
      var licensingError: String? = null
      var sessionStatus: CWMSessionStatus? = null
      try {
        val (response, cert) = LobbyServerApi.heartbeat(serverUri, request)
        failuresCount = 0 // reset counter after successful heartbeat
        nextTimeout = response.nextHeartbeatIntervalMs
        // Do not check license if talking to on-premise lobby server. In this case licence-related restrictions are managed on the server
        if (!cert.isOnPremise()) {
          licensingError = checkLicense(sessionStart)
          sessionStatus = updateSessionStatus(sessionStart)
        }
        if (licensingError != null) {
          stopSession = true
          logger.info(licensingError)
        }
      }
      catch (e: ErrorHttpResponseException) {
        if (e.responseCode >= 500) {
          failuresCount += 1
        }
        else {
          stopSession = true
          if (e.responseCode == 402) {
            // payment required code
            licensingError = e.message
          }
        }
        logger.info("Heartbeat request failed for session '${request.sessionId}': " + e.message)
      }
      catch (e: Throwable) {
        failuresCount += 1
        logger.info("Heartbeat request failed for session '${request.sessionId}': " + e.message)
      }
      finally {
        if (stopSession || failuresCount > 10) {
          AppExecutorUtil.getAppExecutorService().execute {
            applicationOrNull?.invokeLater({
              if (!project.isDisposed) {
                // This will terminate the session and gracefully disconnect all currently connected clients See ShutdownCodeWithMeAction
                val terminationReason = if (sessionStatus == CWMSessionStatus.FREE_SESSION_STOPPED) GuestTerminationReason.FREE_SESSION_ENDED else GuestTerminationReason.UNKNOWN
                val message = licensingError ?: CodeWithMePluginBundle.message("message.session.ended.title")
                RdServerConnectionManager.getInstance(project).terminateAllTemporaryListeners(GuestTerminationReasonsData(terminationReason, message))

                if (licensingError != null) {
                  if (sessionStatus == CWMSessionStatus.FREE_SESSION_STOPPED || sessionStatus == CWMSessionStatus.EXPIRED) {
                    CodeWithMeLifecycleLogger.onSessionExpired(project = project, sessionId = request.sessionId)
                  }

                  when(sessionStatus) {
                    CWMSessionStatus.FREE_SESSION_STOPPED -> {
                      HostNotifications.showTimeIsUpDialog(project)
                    }
                    CWMSessionStatus.EXPIRED -> {
                      HostNotifications.showLicenseExpiredDialog(project)
                    }
                    else -> {
                      HostNotifications.showErrorDialog(project)
                    }
                  }
                }
              }
            }, ModalityState.NON_MODAL)
          }
        }
        else {
          scheduleHeartbeat(serverUri, request, nextTimeout, sessionStart, failuresCount)
        }
      }
    }, timeout.toLong(), TimeUnit.MILLISECONDS))?.cancel(false)
  }

  fun removeSession(joinLink: URI, sessionUpdateToken: String?) = LobbyServerApi.removeSession(joinLink, sessionUpdateToken)

  fun serverInfo() = LobbyServerApi.serverInfo(LobbyUrlSettings.lobbyUrl.uri).first

  fun createSession(
    directTcpLocalPort: Int?,
    hostPeer2PeerQuicEndpoint: String?,
    relayEnabled: Boolean,
    jitsiTelephonyEnabled: Boolean,
    serverInfo: ServerInfoResponse,
    forceRegion: String? = null,
    existingSessionId: String? = null,
    existingSessionUpdateToken: String? = null,
  ): Pair<CreateSessionResponse, CertificateDecoder.CertificateData> {
    val serverUrl = LobbyUrlSettings.lobbyUrl.uri

    logger.info("Query lobby server $serverUrl for a session (existing session id = '$existingSessionId')")

    val enabledHostFeatures = serverInfo.enabledFeatures.toMutableSet()
    if (!relayEnabled) {
      enabledHostFeatures.remove(CodeWithMeFeature.WsRelay)
    }
    if (hostPeer2PeerQuicEndpoint == null) {
      enabledHostFeatures.remove(CodeWithMeFeature.Peer2PeerQuic)
    }
    if (directTcpLocalPort == null) {
      enabledHostFeatures.remove(CodeWithMeFeature.DirectTcp)
    }
    if (!jitsiTelephonyEnabled) {
      enabledHostFeatures.remove(CodeWithMeFeature.JitsiTelephony)
    }

    return LobbyServerApi.createSession(
      serverUrl = serverUrl,
      request = CreateSessionRequest(
        sessionId = existingSessionId,
        sessionUpdateToken = existingSessionUpdateToken,
        hostBuildNumber = ApplicationInfo.getInstance().build.asString(),
        hostProjectName = if (enabledHostFeatures.contains(CodeWithMeFeature.ProjectNames)) project.name else "",
        hostUserMoniker = if (enabledHostFeatures.contains(CodeWithMeFeature.UserNames)) LocalUserSettings.userName else "",
        hostPlatform = CodeWithMePlatformUtil.currentPlatform(),
        hostDirectTcpEndpoints = if(directTcpLocalPort == null) emptyList() else NetworkUtils.getSelfAddresses().mapNotNull { address ->
          when (address) {
            is Inet6Address -> "tcp://[${address.hostAddress}]:$directTcpLocalPort"
            is Inet4Address -> "tcp://${address.hostAddress}:$directTcpLocalPort"
            else -> {
              logger.error("Unknown address type ${address.javaClass.name}, toString:$address")
              null
            }
          }
        },
        hostFingerprint = hostFingerprint,
        hostUseRelay = relayEnabled,
        hostPeer2PeerQuicEndpoint = hostPeer2PeerQuicEndpoint,
        enabledHostFeatures = enabledHostFeatures,
        eventsLogging = true,
        forceRegion = forceRegion,
        metadata = getMetadata()
      )
    )
  }

  private fun getMetadata(): HashMap<String, String>? {
    val metadata = hashMapOf<String, String>()

    if (applicationOrNull?.isUnitTestMode == true)
      metadata[SessionMetadataKeys.JB_FROM_TESTS] = "true"

    if (applicationOrNull?.isInternal == true)
      metadata[SessionMetadataKeys.JB_IS_INTERNAL] = "true"

    if (isJetBrainsTeamLicense())
      metadata[SessionMetadataKeys.JB_TEAM_LICENSE] = "true"

    if (isInternalHostResolved())
      metadata[SessionMetadataKeys.JB_NET] = "true"

    return if (metadata.isEmpty()) null else metadata
  }

  private fun isJetBrainsTeamLicense(): Boolean {
    val licensedToMessage = LicensingFacade.getInstance()?.licensedToMessage
    if (licensedToMessage.isNullOrEmpty())
      return false

    val prefix = "Licensed to "
    if (!licensedToMessage.startsWith(prefix))
      return false

    val licenseTarget = licensedToMessage.substring(prefix.length)
    return licenseTarget.startsWith("JetBrains Team", ignoreCase = false)
  }

  /**
   * Check if internal JetBrains host can be resolved to identify internal network.
   */
  private fun isInternalHostResolved(): Boolean =
    try {
      CompletableFuture.supplyAsync(
        Supplier {
          val testHost = "repo.labs.intellij.net"
          try {
            InetAddress.getByName(testHost)
            true
          } catch (t: Throwable) {
            false
          }
        },
        AppExecutorUtil.getAppExecutorService()
      ).get(500, TimeUnit.MILLISECONDS)
    } catch (t: Throwable) {
      false
    }

  @Suppress("NOTHING_TO_INLINE")
  @Nls
  private inline fun checkLicense(sessionStart : Instant) : String? {
    val freeSessionDuration = freeSessionDuration()
    if (abs(ChronoUnit.MINUTES.between(sessionStart, Instant.now())) > freeSessionDuration && !isLicensed()) {
      return if (hasLicenseStamp())
        CodeWithMePluginBundle.message("message.subscription.expired")
      else
        CodeWithMePluginBundle.message("message.session.time.limit.reached", freeSessionDuration)
    }
    return null
  }

  private fun updateSessionStatus(sessionStart : Instant): CWMSessionStatus {
    val status : CWMSessionStatus?
    val time : Long?
    if (isLicensed()) {
      status = CWMSessionStatus.LICENSED_SESSION
      time = null
    }
    else {
      val remainingTimeMs = max(0, freeSessionDuration() - abs(ChronoUnit.MINUTES.between(sessionStart, Instant.now())))
      if (hasLicenseStamp()) {
        // The product is somehow licensed, but the license is not valid anymore. Apply default plan with session limit
        status = CWMSessionStatus.EXPIRED
        time = Duration.ofMinutes(remainingTimeMs).toMillis()
      }
      else {
        status = if (remainingTimeMs > 0) CWMSessionStatus.FREE_SESSION else CWMSessionStatus.FREE_SESSION_STOPPED
        time = Duration.ofMinutes(remainingTimeMs).toMillis()
      }

    }
    connectionStatusService.sessionStatus.set(status)
    connectionStatusService.freeSessionRemainingTimeMs.set(time)
    return status;
  }

  @Suppress("NOTHING_TO_INLINE")
  private inline fun freeSessionDuration() = min(
    freeSessionDurationMinutes,
    max(1, runCatching { Integer.parseInt(System.getProperty("cwm.free.session.duration.min")) }.getOrDefault(freeSessionDurationMinutes))
  )

  private val hostFingerprint get() = SelfSignedCertificate.permanentCertificateData.fingerprint

  fun listenForConnectedClients(
    lifetime: Lifetime,
    listenForClientConnectionEndpoint: URI,
    hostPeer2PeerDatagramSocketWrapper: DatagramSocketWrapper,
    sessionUpdateToken: String?,
  ) {
    AppExecutorUtil.getAppExecutorService().execute {
      catch {
        logger.info("Listening for connected clients via lobby polling at $listenForClientConnectionEndpoint")

        while (lifetime.isAlive) {
          try {
            val events = LobbyServerApi.pollForConnectedClients(listenForClientConnectionEndpoint, hostFingerprint, sessionUpdateToken)
            for (event in events) {
              logger.warn("Client connected: $event")

              val clientPeer2PeerQuicEndpoint = event.clientPeer2PeerQuicEndpoint
              if (clientPeer2PeerQuicEndpoint != null) {
                catch {
                  val uri = URI.create(clientPeer2PeerQuicEndpoint)
                  if (uri.scheme != "quic") {
                    error("Expected QUIC scheme from clientPeer2PeerQuicEndpoint, but got $uri")
                  }

                  val address = InetAddresses.forString(uri.host)
                  val port = uri.port
                  if (port <= 0) {
                    error("Expected port in URI from clientPeer2PeerQuicEndpoint, but got $uri")
                  }
                  val inetSocketAddress = InetSocketAddress(address, port)

                  application.executeOnPooledThread {
                    for (retry in 1..60) {
                      if (lifetime.isNotAlive) break

                      logger.trace("Sending dummy UDP packet to $inetSocketAddress")

                      try {
                        hostPeer2PeerDatagramSocketWrapper.send(ByteArray(5), inetSocketAddress)
                      } catch (t: Throwable) {
                        logger.debug("Unable to send dummy UDP packet to $inetSocketAddress: ${t.message}", t)
                      }

                      Thread.sleep(1000)
                    }
                  }
                }
              }
            }
          }
          catch (t: Throwable) {
            logger.info("listenForConnectedClients: pollForConnectedClients terminated with an error: ${t.message}", t)
          }

          // Don't poll too much
          Thread.sleep(1000)
        }

        if (!lifetime.isAlive) {
          logger.info("Lobby polling at $listenForClientConnectionEndpoint terminated since lifetime is not alive")
        }
      }

      logger.info("Lobby polling at $listenForClientConnectionEndpoint thread terminated")
    }
  }
}