package com.jetbrains.rdserver.joinLinks.permanent

import com.intellij.codeWithMe.ClientId
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.util.Alarm
import com.intellij.util.AlarmFactory
import com.jetbrains.rd.platform.codeWithMe.connection.LobbyServerApi.isOnPremise
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.CwmNotifications.CWM_NOTIFICATION_GROUP_ID
import com.jetbrains.rd.platform.codeWithMe.users.LobbyUrlSettings
import com.jetbrains.rd.platform.util.SelfSignedCertificate
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.joinLinks.RdServerJoinLinkManager
import com.jetbrains.rdserver.joinLinks.RdServerLobbySessionManager
import com.jetbrains.rdserver.joinLinks.shareableJoinLink
import com.jetbrains.rdserver.startup.RdServerRelayedConnectionsListener
import com.jetbrains.rdserver.unattendedHost.UnattendedHostManager
import java.net.URI
import java.time.Duration
import java.util.concurrent.CompletableFuture

@Service
class CwmPermanentLinkManager(private val project: Project) {
  private val alarm = AlarmFactory.getInstance().create(Alarm.ThreadToUse.POOLED_THREAD, project)

  var lastPermanentLink : URI? = null
   private set

  init {
    val permaLinkSettings = CwmPermanentLinkSettings.getInstance(project)
    permaLinkSettings.sessionId.view(project.lifetime) { sessionIdValueLifetime, sessionIdValue ->
      if (sessionIdValue == null) return@view

      if (!application.isInternal && !UnattendedHostManager.getInstance().isUnattendedMode) {
        LOG.info("Not in internal or unattended mode, not enabling permanent link")
        return@view
      }

      LOG.info("Loaded permanent link settings, will attach listener to session ID $sessionIdValue")

      val relayedConnectionListener = RdServerRelayedConnectionsListener.getInstance(project)
      application.executeOnPooledThread {
        val serverInfo = RdServerLobbySessionManager.getInstance(project).serverInfo()
        val (sessionResponse, cert) = RdServerLobbySessionManager.getInstance(project).createSession(
          serverInfo = serverInfo,
          existingSessionId = sessionIdValue,
          existingSessionUpdateToken = permaLinkSettings.state.lastSessionUpdateToken,
          directTcpLocalPort = null,
          relayEnabled = true,
          hostPeer2PeerQuicEndpoint = null,
          jitsiTelephonyEnabled = false,
          forceRegion = RdServerJoinLinkManager.forceRegionRegistryValue,
        )

        lastPermanentLink = sessionResponse.shareableJoinLink

        val sessionLifetimeDef = RdServerLobbySessionManager.getInstance(project).registerSession(
          LobbyUrlSettings.lobbyUrl.uri, sessionResponse.sessionId, sessionResponse.sessionUpdateToken!!, sessionResponse.nextHeartbeatIntervalMs, cert.isOnPremise()
        )

        relayedConnectionListener.attachToSession(sessionResponse, sessionLifetimeDef, true) {}

        sessionIdValueLifetime.onTermination {
          sessionLifetimeDef.terminate()
          lastPermanentLink = null
        }
      }
    }
  }

  fun disablePermanentLink() {
    alarm.cancelAllRequests()

    val settings = CwmPermanentLinkSettings.getInstance(project)

    val lastJoinLink = settings.state.lastJoinLink
    val lastSessionUpdateToken = settings.state.lastSessionUpdateToken
    if (lastJoinLink != null) {
      application.executeOnPooledThread {
        RdServerLobbySessionManager.getInstance(project).removeSession(URI(lastJoinLink), lastSessionUpdateToken)
      }
    }

    settings.updateState {
      it.enabled = false
      it.lastHostFingerprint = null
      it.lastSessionUpdateToken = null
      it.lastJoinLink = null
      it.lastSessionId = null
    }
  }

  fun enableAndGetPermanentLink(): CompletableFuture<URI> {
    val settings = CwmPermanentLinkSettings.getInstance(project)
    val currentJoinLink = settings.state.lastJoinLink
    if (settings.state.enabled && currentJoinLink != null) return CompletableFuture.completedFuture(URI(currentJoinLink))

    val future = registerOrUpdateSession()

    future.thenRun {
      application.invokeLater {
        if (!settings.state.enabled) {
          ClientId.withClientId(ClientId.localId) {
            // TODO Link to options page?
            Notification(CWM_NOTIFICATION_GROUP_ID,
                         CodeWithMePlatformBundle.message("product.name"),
                         CodeWithMePluginBundle.message("notification.permanent.link.created"),
                         NotificationType.INFORMATION).notify(project)
          }
        }

        settings.updateState {
          it.enabled = true
        }

        requestUpdateAfterInterval()
      }
    }

    return future
  }

  private fun requestUpdateAfterInterval() {
    alarm.addRequest({ updateSession() }, UPDATE_INTERVAL.toMillis())
  }

  private fun updateSession() {
    val state = CwmPermanentLinkSettings.getInstance(project).state
    if (!state.enabled) return

    registerOrUpdateSession()

    requestUpdateAfterInterval()
  }

  private fun registerOrUpdateSession(): CompletableFuture<URI> {
    val future = CompletableFuture<URI>()

    val state = CwmPermanentLinkSettings.getInstance(project).state
    val currentFingerprint = SelfSignedCertificate.permanentCertificateData.fingerprint

    val existingSessionId = if (state.lastHostFingerprint != null && currentFingerprint != state.lastHostFingerprint) {
      LOG.info("Permanent certificate changed, regenerating permanent join link")
      null
    } else state.lastSessionId

    application.executeOnPooledThread {
      try {
        val lobbySessionManager = RdServerLobbySessionManager.getInstance(project)

        LOG.info("Registering/updating permanent link on lobby server. existingSessionId=$existingSessionId")

        val serverInfo = lobbySessionManager.serverInfo()
        val (response, _) = lobbySessionManager.createSession(
          serverInfo = serverInfo,
          existingSessionId = existingSessionId,
          existingSessionUpdateToken = state.lastSessionUpdateToken,
          hostPeer2PeerQuicEndpoint = null,
          relayEnabled = true,
          directTcpLocalPort = null,
          jitsiTelephonyEnabled = false,
          forceRegion = RdServerJoinLinkManager.forceRegionRegistryValue,
        )

        application.invokeLater {
          val settings = CwmPermanentLinkSettings.getInstance(project)

          settings.updateState {
            it.lastSessionId = response.sessionId
            it.lastJoinLink = response.shareableJoinLink.toString()
            it.lastSessionUpdateToken = response.sessionUpdateToken
            it.lastHostFingerprint = currentFingerprint
          }

          future.complete(response.shareableJoinLink)
        }
      } catch (t: Throwable) {
        LOG.warn("Unable to create or refresh permanent session: ${t.message}", t)

        application.invokeLater {
          CwmPermanentLinkSettings.getInstance(project).updateState {
            it.lastSessionId = null
            it.lastJoinLink = null
            it.lastHostFingerprint = null
            it.lastSessionUpdateToken = null
          }
        }

        future.completeExceptionally(t)
      }
    }

    return future
  }

  companion object {
    private val LOG = Logger.getInstance(CwmPermanentLinkManager::class.java)
    private val UPDATE_INTERVAL = Duration.ofHours(6)

    fun getInstance(project: Project) = project.service<CwmPermanentLinkManager>()
  }
}