package com.jetbrains.rdserver.joinLinks.permanent

import com.intellij.openapi.components.*
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.jetbrains.rd.platform.util.ThreadSafeProperty
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.reactive.IPropertyView
import kotlin.reflect.KProperty

@Service
@State(name = "CodeWithMePermanentLinkSettings",
       storages = [(Storage(StoragePathMacros.WORKSPACE_FILE))])
class CwmPermanentLinkSettings
  : PersistentStateComponentWithModificationTracker<CwmPermanentLinkSettings.LinkSettingsState> {

  companion object {
    fun getInstance(project: Project): CwmPermanentLinkSettings = project.service()
    private val LOG = logger<CwmPermanentLinkSettings>()
  }

  class LinkSettingsState : BaseState() {
    private val enabledProperty = property(false)
    var enabled by enabledProperty

    private val lastSessionIdProperty = string()
    var lastSessionId by lastSessionIdProperty

    private val lastHostFingerprintProperty = string()
    var lastHostFingerprint by lastHostFingerprintProperty

    private val lastSessionUpdateTokenProperty = string()
    var lastSessionUpdateToken by lastSessionUpdateTokenProperty

    private val lastJoinLinkProperty = string()
    var lastJoinLink by lastJoinLinkProperty
  }

  @Volatile
  private var state: LinkSettingsState = LinkSettingsState()

  override fun getState() = LinkSettingsState().also { it.copyFrom(state) }

  override fun getStateModificationCount() = state.modificationCount

  override fun loadState(state: LinkSettingsState) {
    compareValues(this.state.enabled, state.enabled, LinkSettingsState::enabled)
    compareValues(this.state.lastHostFingerprint, state.lastHostFingerprint, LinkSettingsState::lastHostFingerprint)
    compareValues(this.state.lastJoinLink, state.lastJoinLink, LinkSettingsState::lastJoinLink)
    compareValues(this.state.lastSessionId, state.lastSessionId, LinkSettingsState::lastSessionId)
    compareValues(this.state.lastSessionUpdateToken, state.lastSessionUpdateToken, LinkSettingsState::lastSessionUpdateToken)

    this.state = state

    sessionIdProperty.value = state.lastSessionId
  }

  private fun <T> compareValues(before: T, after: T, property: KProperty<T>) {
    if (before != after) {
      LOG.info("Updated ${property.name}: '$before' -> '$after'")
    }
  }

  fun updateState(updater: (LinkSettingsState) -> Unit) {
    application.assertIsDispatchThread()

    val newState = getState()
    updater(newState)
    loadState(newState)
  }

  private val sessionIdProperty = ThreadSafeProperty<String?>(null)
  val sessionId: IPropertyView<String?> = sessionIdProperty
}
