package com.jetbrains.rdserver.navigation

import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.fileEditor.FileEditorManager
import com.intellij.openapi.fileEditor.TextEditor
import com.intellij.openapi.project.Project
import com.intellij.openapi.vfs.VfsUtil
import com.intellij.openapi.vfs.VirtualFile
import com.jetbrains.rd.ide.model.Navigation
import com.jetbrains.rd.ide.model.NavigationByFileId
import com.jetbrains.rd.ide.model.NavigationByPath
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.vfs.BackendVirtualFileSystemHost
import java.io.File

@Suppress("NonDefaultConstructor")
class BackendNavigationHost(
  private val project: Project,
  solution: Solution
) {

  companion object {
    private val LOG = logger<BackendNavigationHost>()
  }

  init {
    solution.navigate.advise(project.lifetime) { navigation ->
      LOG.info("receiving command to navigate to $navigation")

      val (file, offset) = getNavigationData(navigation) ?: return@advise

      val editor = FileEditorManager.getInstance(project).openFile(file, true)
      LOG.info("Navigation to ${file.name}. Moving to offset $offset")
      (editor as? TextEditor)?.editor?.caretModel?.moveToOffset(offset)
      BackendEditorHost.getInstance(project).reportFocusedTextControl()
    }
  }

  private fun getNavigationData(navigation: Navigation): Pair<VirtualFile, Int>? {
    when (navigation) {
      is NavigationByPath -> {
        val basePath = project.basePath ?: return null
        val file = VfsUtil.findFileByIoFile(File(basePath).resolve(navigation.pathToFile), true)
        if (file == null) {
          LOG.info("cannot find file by path ${navigation.pathToFile}")
          return null
        }
        return file to navigation.offset
      }
      is NavigationByFileId -> {
        val file = BackendVirtualFileSystemHost.getInstance(project).findVirtualFile(navigation.fileId)
        if (file == null) return null
        return file to navigation.offset
      }
    }
    return null
  }
}
