package org.jetbrains.idea.perforce.operations;

import com.intellij.openapi.application.ApplicationManager;
import com.intellij.openapi.diagnostic.Logger;
import com.intellij.openapi.project.Project;
import com.intellij.openapi.util.Ref;
import com.intellij.openapi.util.text.StringUtil;
import com.intellij.openapi.vcs.FilePath;
import com.intellij.openapi.vcs.FileStatus;
import com.intellij.openapi.vcs.VcsException;
import com.intellij.openapi.vcs.changes.*;
import com.intellij.openapi.vfs.VirtualFile;
import com.intellij.util.ProcessingContext;
import org.jetbrains.annotations.NonNls;
import org.jetbrains.idea.perforce.PerforceBundle;
import org.jetbrains.idea.perforce.application.PerforceVcs;
import org.jetbrains.idea.perforce.perforce.FStat;
import org.jetbrains.idea.perforce.perforce.P4File;
import org.jetbrains.idea.perforce.perforce.PerforceCachingContentRevision;
import org.jetbrains.idea.perforce.perforce.PerforceRunner;

/**
 * @author yole
 */
public class P4EditOperation extends VcsOperationOnPath {
  private static final Logger LOG = Logger.getInstance(P4EditOperation.class);
  @NonNls private static final String CANNOT_FIND_ERROR = "the system cannot find";
  @NonNls private static final String CANNOT_FIND_ERROR_2 = "no such file or directory";
  private boolean mySuppressErrors;

  @SuppressWarnings("unused") // used by deserialization reflection
  public P4EditOperation() {
  }

  public P4EditOperation(String changeList, final VirtualFile file) {
    super(changeList, file.getPath());
  }

  public P4EditOperation(String changeList, final String path) {
    super(changeList, path);
  }

  @SuppressWarnings("unused") // used by deserialization reflection
  public boolean isSuppressErrors() {
    return mySuppressErrors;
  }

  public void setSuppressErrors(boolean suppressErrors) {
    mySuppressErrors = suppressErrors;
  }

  @Override
  public void execute(final Project project, ProcessingContext context) throws VcsException {
    final Ref<PerforceVcs> vcs = new Ref<>();
    final Ref<PerforceRunner> runner = new Ref<>();
    ApplicationManager.getApplication().runReadAction(() -> {
      if (project.isDisposed()) return;
      vcs.set(PerforceVcs.getInstance(project));
      runner.set(PerforceRunner.getInstance(project));
    });

    if (vcs.isNull()) return;
    try {
      final P4File p4File = P4File.createInefficientFromLocalPath(myPath);
      FStat p4FStat = vcs.get().getFstatSafe(p4File);
      if (p4FStat == null) return;
      if ((p4FStat.status == FStat.STATUS_NOT_ADDED || p4FStat.status == FStat.STATUS_ONLY_LOCAL) &&
          p4FStat.local != FStat.LOCAL_BRANCHING) {
        throw new VcsException(
          PerforceBundle.message("confirmation.text.auto.edit.file.not.registered.on.server", p4File.getLocalPath()));
      }
      else if (p4FStat.status == FStat.STATUS_DELETED) {
        throw new VcsException(PerforceBundle.message("exception.text.file.deleted.from.server.cannot.edit", p4File.getLocalPath()));
      }
      else if (p4FStat.local == FStat.LOCAL_MOVE_DELETING) {
        return; // edit already done by rename/move handlers
      }
      else if (p4FStat.local != FStat.LOCAL_CHECKED_IN && p4FStat.local != FStat.LOCAL_INTEGRATING &&
               p4FStat.local != FStat.LOCAL_BRANCHING) {
        throw new VcsException(
          PerforceBundle.message("exception.text.file..should.not.be.readonly.cannot.edit", p4File.getLocalPath()));
      }

      long changeListNumber = getPerforceChangeList(project, p4File, context);
      runner.get().edit(p4File, changeListNumber);
    }
    catch (VcsException e) {
      if (mySuppressErrors) {
        // in allwrite workspace, 'p4 edit' is executed when modifying a file that's unchanged according to IDE status.
        // "not changed" file status might've been outdated, so the file might be not available for 'p4 edit'. Not a big deal.
        LOG.debug(e);
      } else {
        // check if file was deleted while we were waiting to perform background edit
        String message = StringUtil.toLowerCase(e.getMessage());
        if (!message.contains(CANNOT_FIND_ERROR) && !message.contains(CANNOT_FIND_ERROR_2)) {
          throw e;
        }
      }
    } finally {
      VirtualFile vFile = getFilePath().getVirtualFile();
      if (vFile != null) {
        vFile.refresh(true, false);
        vcs.get().asyncEditCompleted(vFile);
      }
    }
    final FilePath filePath = getFilePath();
    ApplicationManager.getApplication().runReadAction(() -> {
      if (!project.isDisposed()) {
        VcsDirtyScopeManager.getInstance(project).fileDirty(filePath);
      }
    });
  }

  @Override
  public Change getChange(final Project project, ChangeListManagerGate addGate) {
    FilePath path = getFilePath();
    final FileStatus status = addGate.getStatus(path);
    ContentRevision beforeRevision = FileStatus.ADDED.equals(status) ? null : PerforceCachingContentRevision.createOffline(project, path, path);
    ContentRevision afterRevision = CurrentContentRevision.create(path);
    return new Change(beforeRevision, afterRevision);
  }

}
