package org.jetbrains.intellij.build.dependencies

import groovy.transform.CompileStatic

import java.nio.file.Path

import static org.jetbrains.intellij.build.dependencies.BuildDependenciesUtil.createDocumentBuilder
import static org.jetbrains.intellij.build.dependencies.BuildDependenciesUtil.getSingleChildElement

@CompileStatic
class QuicheDependencyDownloader {
  static Path downloadQuicheBin(Path communityRoot) {
    def quicheVersion = getQuicheVersion(communityRoot)

    def quicheBinUri = new URI(
      "https://cache-redirector.jetbrains.com/jetbrains.bintray.com/intellij-third-party-dependencies/com/jetbrains/quiche/quiche-bin/$quicheVersion/quiche-bin-${quicheVersion}.zip")
    def quicheZip = BuildDependenciesDownloader.downloadFileToCacheLocation(communityRoot, quicheBinUri)
    def quicheBin = BuildDependenciesDownloader.extractFileToCacheLocation(communityRoot, quicheZip)
    return quicheBin
  }

  static private String getQuicheVersion(Path communityRoot) {
    BuildDependenciesDownloader.checkCommunityRoot(communityRoot)

    def quicheJnaStubsXml = communityRoot.parent.resolve(".idea").resolve("libraries").resolve("quiche_jna_stubs.xml")

    try {
      def documentBuilder = createDocumentBuilder()
      def document = documentBuilder.parse(quicheJnaStubsXml.toFile())

      def libraryElement = getSingleChildElement(document.documentElement, "library")
      def propertiesElement = getSingleChildElement(libraryElement, "properties")
      def mavenId = propertiesElement.getAttribute("maven-id")
      if (mavenId == null || mavenId.isBlank()) {
        throw new IllegalStateException("Invalid maven-id")
      }

      BuildDependenciesDownloader.debug("Quiche JNA stubs artifact coordinates: $mavenId")

      def quicheVersion = mavenId.substring(mavenId.lastIndexOf(':') + 1)
      BuildDependenciesDownloader.debug("Quiche JNA stubs version (build): $quicheVersion")

      return quicheVersion
    }
    catch (Throwable t) {
      throw new IllegalStateException("Unable to load quiche version from $quicheJnaStubsXml: ${t.message}", t)
    }
  }

  static void main(String[] args) {
    def path = downloadQuicheBin(BuildDependenciesDownloader.getCommunityRootFromWorkingDirectory())
    println("Quiche is at $path")
  }
}
