package com.jetbrains.rdserver.permissions.impl

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.impl.DocumentImpl
import com.intellij.openapi.editor.impl.DocumentWriteAccessGuard
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.vfs.VirtualFile
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.startup.RdServerApplicationService
import org.jetbrains.annotations.Nullable

class BackendDocumentWriteAccessGuard : DocumentWriteAccessGuard() {

  companion object {
    private val LOG = logger<BackendDocumentWriteAccessGuard>()
  }

  override fun isWritable(document: Document): Result {
    // usually !writeThreadOnly means that a document is created for internal purposes of a feature,
    // and modifications of such document is a part this feature work (e.g. a fragment mapping in LayeredLexerEditorHighlighter)
    if (document is DocumentImpl && !document.isWriteThreadOnly) return success()

    if (ClientId.isCurrentlyUnderLocalId) return success()

    val currentClientId = ClientId.current
    val project = RdServerApplicationService.getInstance().getProjectForClient(currentClientId) ?: run {
      LOG.info("There is no project for client($currentClientId). Deny write access attempt.")
      return fail("There is no project for client ($currentClientId).")
    }

    val file = FileDocumentManager.getInstance().getFile(document)
    if (file == null) return success()  // CWM-2063

    if (!PermissionManager.getInstance(project).hasFileWriteAccess(currentClientId)) {
      return fail("The client (${getUserName(currentClientId, project)}) has no write access at all." +
                  " The file bound to document is ${getFileInfo(file)}")
    }

    if (PermissionManager.getInstance(project).isWritable(currentClientId, file)) {
      return success()
    }
    else {
      return fail("The client (${getUserName(currentClientId, project)}) has no write access to the file ${getFileInfo(file)}.")
    }
  }

  private fun getFileInfo(file: @Nullable VirtualFile?) =
    if (file == null) "(null)" else "(class=${file.javaClass.simpleName} name=${file.name})"

  private fun getUserName(clientId: ClientId, project: Project): String {
    return UserManager.getInstance(project).users.singleOrNull { it.clientId == clientId }?.name
           ?: run {
             LOG.warn("Unknown clientId in user model: $clientId")
             return clientId.toString()
           }
  }
}