package com.jetbrains.rdserver.permissions.impl

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Document
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.guessProjectDir
import com.intellij.openapi.roots.FileIndexFacade
import com.intellij.openapi.vfs.VfsUtil
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.newvfs.ArchiveFileSystem
import com.intellij.testFramework.LightVirtualFileBase
import com.intellij.util.io.isAncestor
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.reactive.Signal
import com.jetbrains.rdserver.permissions.*

class PermissionManagerImpl(private val project: Project) : PermissionManager {
  val permissionModelChanged = Signal<Unit>()

  companion object {
    private val logger = getLogger<PermissionManager>()

    fun getInstanceImpl(project: Project): PermissionManagerImpl {
      return PermissionManager.getInstance(project) as PermissionManagerImpl
    }
  }

  private val locker = Object()

  private val projectPermissions: PermissionModel
    get() = project.getService(PermissionStore::class.java).model
  private var userPermissions = mutableMapOf<ClientId, PermissionModel>()

  init {
    BackendPermissionHost.update(project, null, projectPermissions)
  }

  fun getModel(client: ClientId?): PermissionModel {
    synchronized(locker) {
      return getPermissionsForClient(client)
    }
  }

  fun applyModel(client: ClientId?, model: PermissionModel) {
    synchronized(locker) {
      if (client == null) {
        project.getService(PermissionStore::class.java).model = model
      }
      else {
        userPermissions[client] = model
      }

      permissionModelChanged.fire(Unit)
    }

    BackendPermissionHost.update(project, client, model)
  }

  override fun hasFileWriteAccess(client: ClientId): Boolean {
    synchronized(locker) {
      return getPermissionsForClient(client).fileAccess == FileAccessLevel.FullAccess
    }
  }

  override fun isVisible(client: ClientId, virtualFile: VirtualFile): Boolean {
    var file = virtualFile
    if (virtualFile is LightVirtualFileBase) {
      // allow synchronization of in-memory files
      file = virtualFile.originalFile ?: return true
    }

    synchronized(locker) {
      val permissions = getPermissionsForClient(client)

      // Check if file is hidden in permission settings
      if (isHiddenInPermissionSettings(file, permissions)) return false

      // In FullAccess mode allow to sync any non-excluded file from project directory
      if (permissions.fileAccess == FileAccessLevel.FullAccess) {
        val projectDirectory = project.guessProjectDir()
        if (projectDirectory != null && VfsUtil.isAncestor(projectDirectory, file, false))
          return true
      }
    }

    return FileIndexFacade.getInstance(project).isInContent(file) ||
           FileIndexFacade.getInstance(project).isInLibrarySource(file) ||
           FileIndexFacade.getInstance(project).isInLibraryClasses(file)
  }

  private fun isHiddenInPermissionSettings(virtualFile: VirtualFile, permissions: PermissionModel): Boolean {
    val fileSystem = virtualFile.fileSystem
    val file =
      if (fileSystem is ArchiveFileSystem) fileSystem.getLocalByEntry(virtualFile)
      else virtualFile
    if (file == null || !file.isInLocalFileSystem) return false

    val path = file.toNioPath()
    for (invisibleFile in permissions.excludedFiles) {
      if (invisibleFile.isAncestor(path)) {
        logger.debug("Request to invisible file $virtualFile")
        return true
      }
    }

    return false
  }

  override fun isVisible(client: ClientId, document: Document): Boolean {
    val file = FileDocumentManager.getInstance().getFile(document)
    return file == null || isVisible(client, file)
  }

  override fun isWritable(client: ClientId, virtualFile: VirtualFile): Boolean {
    synchronized(locker) {
      val permissions = getPermissionsForClient(client)
      return permissions.fileAccess == FileAccessLevel.FullAccess &&
             isVisible(client, virtualFile)
    }
  }

  override fun isWritable(client: ClientId, document: Document): Boolean {
    val file = FileDocumentManager.getInstance().getFile(document)
    if (file == null) {
      return hasFileWriteAccess(client)
    }
    return isWritable(client, file)
  }

  override fun getTerminalAccess(client: ClientId): TerminalAccessLevel {
    synchronized(locker) {
      return getPermissionsForClient(client).terminalAccess
    }
  }

  override fun getRunAccess(client: ClientId): RunAccessLevel {
    synchronized(locker) {
      return getPermissionsForClient(client).runAccess
    }
  }

  override fun getToolWindowAccess(client: ClientId): ToolWindowAccessLevel {
    synchronized(locker) {
      return getPermissionsForClient(client).toolWindowAccess
    }
  }

  private fun getPermissionsForClient(client: ClientId?): PermissionModel {
    if (client == null) return projectPermissions
    return userPermissions[client] ?: projectPermissions
  }
}