package com.jetbrains.rdserver.permissions.impl

import com.intellij.openapi.components.PersistentStateComponent
import com.intellij.openapi.components.State
import com.intellij.openapi.components.Storage
import com.intellij.openapi.components.StoragePathMacros
import com.intellij.openapi.project.Project
import com.intellij.util.PathUtil
import com.intellij.util.io.systemIndependentPath
import com.jetbrains.rdserver.permissions.FileAccessLevel
import com.jetbrains.rdserver.permissions.RunAccessLevel
import com.jetbrains.rdserver.permissions.TerminalAccessLevel
import com.jetbrains.rdserver.permissions.ToolWindowAccessLevel
import org.jdom.Element
import java.nio.file.Path

@State(
  name = "PermissionStore",
  storages = [(Storage(StoragePathMacros.CACHE_FILE))])
class PermissionStore(private val project: Project) : PersistentStateComponent<Element> {

  private var modelState : PermissionModel? = null
  var model : PermissionModel
    get() {
      return modelState ?: PermissionModel.createDefault(project).apply {
        modelState = this
      }
    }
    set(value)  {
      modelState = value
    }

  override fun getState(): Element {
    val root = Element("permissions")

    val excludedFilesElement = Element("excludedFiles")
    for (file in model.excludedFiles) {
      excludedFilesElement.addContent(Element("Path").apply {
        text = PathUtil.toSystemIndependentName(file.systemIndependentPath)
      })
    }
    root.addContent(excludedFilesElement)

    root.addContent(Element("fileAccess").apply {
      text = model.fileAccess.name
    })
    root.addContent(Element("terminalAccess").apply {
      text = model.terminalAccess.name
    })
    root.addContent(Element("runAccess").apply {
      text = model.runAccess.name
    })
    root.addContent(Element("toolWindowAccess").apply {
      text = model.toolWindowAccess.name
    })
    return root
  }

  override fun loadState(state: Element) {
    val result = MutablePermissionModel(mutableListOf(),
                                        PermissionModel.defaultFileAccessLevel,
                                        PermissionModel.defaultTerminalAccessLevel,
                                        PermissionModel.defaultRunAccessLevel,
                                        PermissionModel.defaultToolWindowAccessLevel)

    val excludedFilesElement = state.getChild("excludedFiles")
    excludedFilesElement?.getChildren("Path")?.forEach {
      val file = Path.of(it.text)
      if (file != null) {
        result.excludedFiles.add(file)
      }
    }

    val fileAccessElement = state.getChild("fileAccess")
    if (fileAccessElement != null) {
      try {
        result.fileAccess = FileAccessLevel.valueOf(fileAccessElement.text)
      }
      catch (ex: IllegalArgumentException) {
        // ignore - unknown enum value
      }
    }

    val terminalAccessElement = state.getChild("terminalAccess")
    if (terminalAccessElement != null) {
      try {
        result.terminalAccess = TerminalAccessLevel.valueOf(terminalAccessElement.text)
      }
      catch (ex: IllegalArgumentException) {
        // ignore - unknown enum value
      }
    }

    val runAccessElement = state.getChild("runAccess")
    if (runAccessElement != null) {
      try {
        result.runAccess = RunAccessLevel.valueOf(runAccessElement.text)
      }
      catch (ex: IllegalArgumentException) {
        // ignore - unknown enum value
      }
    }

    val toolWindowAccessElement = state.getChild("toolWindowAccess")
    if (toolWindowAccessElement != null) {
      try {
        result.toolWindowAccess = ToolWindowAccessLevel.valueOf(toolWindowAccessElement.text)
      }
      catch (ex: IllegalArgumentException) {
        // ignore - unknown enum value
      }
    }

    model = result
  }
}