package com.jetbrains.rdserver.permissions.ui

import com.intellij.codeWithMe.ClientId
import com.intellij.icons.AllIcons
import com.intellij.ide.BrowserUtil
import com.intellij.openapi.project.Project
import com.intellij.openapi.projectRoots.impl.jdkDownloader.RuntimeChooserUtil
import com.intellij.openapi.ui.ComboBox
import com.intellij.openapi.ui.DialogWrapper
import com.intellij.ui.ContextHelpLabel
import com.intellij.ui.EnumComboBoxModel
import com.intellij.ui.HideableDecorator
import com.intellij.ui.RoundedLineBorder
import com.intellij.ui.components.ActionLink
import com.intellij.ui.components.JBRadioButton
import com.intellij.ui.jcef.JBCefApp
import com.intellij.util.ui.JBUI
import com.intellij.util.ui.UIUtil
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.enumSetOf
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.joinLinks.RdServerJoinLinkManager
import com.jetbrains.rdserver.permissions.FileAccessLevel
import com.jetbrains.rdserver.permissions.RunAccessLevel
import com.jetbrains.rdserver.permissions.TerminalAccessLevel
import com.jetbrains.rdserver.permissions.ToolWindowAccessLevel
import com.jetbrains.rdserver.permissions.impl.PermissionManagerImpl
import com.jetbrains.rdserver.permissions.impl.PermissionModel
import com.jetbrains.rdserver.telephony.BackendTelephonyManager
import com.jetbrains.rdserver.toolWindow.BackendToolWindowModeProvider
import icons.CodeWithMeCommonIcons
import net.miginfocom.swing.MigLayout
import org.jetbrains.annotations.Nls
import java.awt.event.ComponentAdapter
import java.awt.event.ComponentEvent
import java.nio.file.Path
import java.util.*
import javax.swing.*
import javax.swing.event.HyperlinkEvent

class PermissionsDialog(private val project: Project, private val client: ClientId?, private val isInitialAccess: Boolean) : DialogWrapper(
  project, true) {

  private val warningIcon = AllIcons.General.Warning
  private val logoPane = JPanel(MigLayout("ins 0, novisualpadding, gap 0")).apply {
    add(JLabel(CodeWithMeCommonIcons.CodeWithMeGuest_64))
    border = JBUI.Borders.empty(14, 8, 0, 18)
  }

  private val manager = PermissionManagerImpl.getInstanceImpl(project)

  private val model = manager.getModel(client)

  private val excludedFiles = model.excludedFiles.toMutableList()
  private val empty = JBUI.Borders.empty(7, 20, 9, 20)

  private val safeStateLabel = JPanel(MigLayout("ins 0, fillx")).apply {
    val text =
      if (client == null) CodeWithMePluginBundle.message("permission.dialog.state.safe.guests")
      else CodeWithMePluginBundle.message("permission.dialog.state.safe.guest")
    val label = multiLineLabelPane(text)
    val pane = JPanel(MigLayout("ins 0, fillx")).apply {
      add(label, "growx")
    }
    add(pane, "growx, wmin 100")
    label.border = empty
    pane.background = JBUI.CurrentTheme.NotificationInfo.backgroundColor()
    pane.foreground = JBUI.CurrentTheme.NotificationInfo.foregroundColor()
    pane.border = RoundedLineBorder(JBUI.CurrentTheme.NotificationInfo.borderColor(), 5)
  }

  private val unsafeStateLabel = JPanel(MigLayout("ins 0, fillx")).apply {
    val text =
      if (client == null) CodeWithMePluginBundle.message("permission.dialog.state.unsafe.guests")
      else CodeWithMePluginBundle.message("permission.dialog.state.unsafe.guest")
    val label = multiLineLabelPane(text)
    val pane = JPanel(MigLayout("ins 0, fillx")).apply {
      add(label, "growx")
    }
    add(pane, "growx, wmin 100")
    label.border = empty
    pane.background = JBUI.CurrentTheme.NotificationError.backgroundColor()
    pane.foreground = JBUI.CurrentTheme.NotificationError.foregroundColor()
    pane.border = RoundedLineBorder(JBUI.CurrentTheme.NotificationError.borderColor(), JBUI.scale(5))
  }


  private val help = ContextHelpLabel.createWithLink(null,
                                                     CodeWithMePluginBundle.message("permission.dialog.help.text"),
                                                     CodeWithMePlatformBundle.message("code.with.me.learn.more")
  ) {
    BrowserUtil.browse(CodeWithMePlatformBundle.message("code.with.me.learn.more.link"))
  }

  private val modeRadioButtons = linkedMapOf<Mode, JBRadioButton>()

  private val fileAccessWarning = JLabel(warningIcon)
  private val fileAccessBox = ComboBox(EnumComboBoxModel(FileAccessLevel::class.java)).apply {
    addActionListener { autoUpdateMode() }
    renderer = ListCellRenderer { _, value, _, _, _ ->
      JLabel(value.present())
    }
  }

  private val terminalAccessWarning = JLabel(warningIcon)
  private val terminalAccessBox = ComboBox(object : EnumComboBoxModel<TerminalAccessLevel>(TerminalAccessLevel::class.java) {
    override fun createEnumSet(en: Class<TerminalAccessLevel>): EnumSet<TerminalAccessLevel> {
      return enumSetOf(super.createEnumSet(en).filter { it != TerminalAccessLevel.Request }.toSet())
    }
  }).apply {
    addActionListener { autoUpdateMode() }
    renderer = ListCellRenderer { _, value, _, _, _ ->
      JLabel(value.present())
    }
  }

  private val runAccessWarning = JLabel(warningIcon)
  private val runAccessBox = ComboBox(EnumComboBoxModel(RunAccessLevel::class.java)).apply {
    addActionListener { autoUpdateMode() }
    renderer = ListCellRenderer { _, value, _, _, _ ->
      JLabel(value.present())
    }
  }

  private val toolWindowAccessWarning = JLabel(warningIcon)
  private val toolWindowAccessBox = ComboBox(EnumComboBoxModel(ToolWindowAccessLevel::class.java)).apply {
    addActionListener { autoUpdateMode() }
    renderer = ListCellRenderer { _, value, _, _, _ ->
      JLabel(value.present())
    }
  }

  init {
    // Setup initial state from model
    fileAccessBox.item = model.fileAccess
    terminalAccessBox.item = model.terminalAccess
    runAccessBox.item = model.runAccess
    toolWindowAccessBox.item = model.toolWindowAccess

    // Setup modes
    for (mode in Mode.values().sortedBy { it.ordinal }) {
      modeRadioButtons[mode] = JBRadioButton(mode.title).apply {
        addActionListener { setMode(mode) }
      }
    }
    autoUpdateMode()

    // Prepare UI
    title =
      if (client == null) CodeWithMePluginBundle.message("permission.dialog.title")
      else CodeWithMePluginBundle.message("permission.dialog.title.per.user", FollowMeManager.getInstance(project).getUser(client).name)
    init()
  }

  private fun multiLineLabelPane(@Nls text: String) = JTextPane().apply {
    isEditable = false
    background = null
    font = JBUI.Fonts.label()
    this.text = text
    minimumSize = null
  }

  override fun createCenterPanel(): JComponent {

    val radioPanel = JPanel(MigLayout("fillx, ins 0, novisualpadding")).apply {
      for (radioButton in modeRadioButtons.values) {
        add(radioButton)
      }

      add(help, "gapbefore 10")
    }

    val detailsPanel = JPanel(
      MigLayout("gap 0, fillx, ins 0, wrap 3, novisualpadding, gapx 0, gapy 4",
                "[min!]15[min!][grow]",
                "[al baseline]"))

    detailsPanel.add(JLabel(CodeWithMePluginBundle.message("permission.dialog.file.access.label")))
    detailsPanel.add(fileAccessWarning)
    detailsPanel.add(fileAccessBox, "grow")

    detailsPanel.add(ActionLink(CodeWithMePluginBundle.message("permission.dialog.manage.hidden")) { updateExcludedFiles() },
                     "skip 2, alignx left, gapleft 2, gapbottom 5")

    detailsPanel.add(JLabel(CodeWithMePluginBundle.message("permission.dialog.terminal.access.label")))
    detailsPanel.add(terminalAccessWarning)
    detailsPanel.add(terminalAccessBox, "grow")

    detailsPanel.add(JLabel(CodeWithMePluginBundle.message("permission.dialog.run.access.label")))
    detailsPanel.add(runAccessWarning)
    detailsPanel.add(runAccessBox, "grow")

    detailsPanel.add(JLabel(CodeWithMePluginBundle.message("permission.dialog.toolWindow.access.label")), "split 2")
    @Suppress("DialogTitleCapitalization")
    detailsPanel.add(ContextHelpLabel.create(BackendToolWindowModeProvider.getSupportedToolWindowsList()))
    detailsPanel.add(toolWindowAccessWarning)
    detailsPanel.add(toolWindowAccessBox, "grow")

    val hideable = JPanel(MigLayout("novisualpadding, ins 0, flowy, fillx, hidemode 3"))

    fun pack() {
      SwingUtilities.getWindowAncestor(hideable)?.let {
        SwingUtilities.invokeLater { it.pack() }
      }
    }

    val decorator = object : HideableDecorator(hideable, CodeWithMePluginBundle.message("permission.dialog.details"), false) {
      override fun on() {
        super.on()
        pack()
      }

      override fun off() {
        super.off()
        pack()
      }
    }
    decorator.setOn(false)
    decorator.setContentComponent(detailsPanel)
    hideable.add(detailsPanel, "growx, gaptop 9, gapleft 15")

    val telephony = JPanel(MigLayout("ins 10 0 0 0, novisualpadding"))

    val isTelephonyEnabled = BackendTelephonyManager.getInstance(project).isTelephonyEnabled
    val telephonyCheckbox = JCheckBox(CodeWithMePluginBundle.message("permission.dialog.auto.start.call"),
                                      isTelephonyEnabled.value)
    val isJcefSupported = JBCefApp.isSupported()
    telephony.add(telephonyCheckbox, "cell 0 0")
    val micOffLabel = JLabel(AllIcons.CodeWithMe.CwmMicOff)
    micOffLabel.toolTipText = CodeWithMePluginBundle.message("permission.dialog.auto.start.call.mic.cam")
    val camOffLabel = JLabel(AllIcons.CodeWithMe.CwmCamOff)
    camOffLabel.toolTipText = micOffLabel.toolTipText
    telephony.add(micOffLabel, "cell 1 0")
    telephony.add(camOffLabel, "cell 2 0, wrap")
    if (isJcefSupported) {
      telephonyCheckbox.addActionListener { isTelephonyEnabled.set(!isTelephonyEnabled.value) }
    } else {
      telephonyCheckbox.isSelected = false
      telephonyCheckbox.isEnabled = false
      val editorPane = JEditorPane()
      editorPane.isEditable = false
      editorPane.editorKit = UIUtil.getHTMLEditorKit()
      editorPane.text = CodeWithMePluginBundle.message("permission.dialog.call.unavailable")
      editorPane.background = UIUtil.getPanelBackground()
      editorPane.addHyperlinkListener {
        if (it.eventType == HyperlinkEvent.EventType.ACTIVATED) {
          if (it.description == "/runtime") {
            application.invokeLater {
              doCancelAction()
              RuntimeChooserUtil.showRuntimeChooserPopup()
            }
          }
        }
      }
      telephony.add(editorPane, "span 3, wmax 90%")
    }

    val headerText =
      if (client == null) CodeWithMePluginBundle.message("permission.dialog.header")
      else CodeWithMePluginBundle.message("permission.dialog.header.user", FollowMeManager.getInstance(project).getUser(client).name)
    val leftPanel = JPanel(MigLayout("fillx, ins 0, flowy, novisualpadding, hidemode 3, gapx 0, gapy 9", "")).apply {
      add(JLabel(headerText), "grow")
      add(radioPanel, "grow")

      add(JPanel(MigLayout("fillx, flowy, ins 0, gap 0, novisualpadding, hidemode 3")).apply {
        add(safeStateLabel,"grow, w 100")
        add(unsafeStateLabel,"grow, w 100")
        add(hideable, "growx, gaptop 4")
      }, "growx")

      if (RdServerJoinLinkManager.telephonyEnabled && client == null) {
        add(telephony, "growx")
      }
      border = JBUI.Borders.empty(11, 0, 10, 0)
    }

    return object : JPanel(MigLayout("ins 0 0 0 5, novisualpadding, gap 0, fill", "[pref!][grow]")) {
      override fun addNotify() {
        super.addNotify()

        leftPanel.addComponentListener(object : ComponentAdapter(){
          override fun componentResized(e: ComponentEvent?) {
            pack()
          }
        })

        modeRadioButtons.values.forEach{
          it.addActionListener { pack() }
        }
        pack()
      }
    }.apply {
      add(logoPane, "ay top")
      add(leftPanel, "grow")
    }
  }

  override fun getOKAction(): Action {
    return super.getOKAction().apply {
      val message =
        if (isInitialAccess) CodeWithMePluginBundle.message("permission.dialog.enable.access")
        else CodeWithMePluginBundle.message("permission.dialog.apply")
      putValue(Action.NAME, message)
    }
  }

  override fun doOKAction() {

    val model = object : PermissionModel {
      override val excludedFiles: List<Path>
        get() = this@PermissionsDialog.excludedFiles
      override val fileAccess: FileAccessLevel
        get() = fileAccessBox.item
      override val terminalAccess: TerminalAccessLevel
        get() = terminalAccessBox.item
      override val runAccess: RunAccessLevel
        get() = runAccessBox.item
      override val toolWindowAccess: ToolWindowAccessLevel
        get() = toolWindowAccessBox.item
    }

    manager.applyModel(client, model)
    super.doOKAction()
  }

  private fun updateExcludedFiles() {
    PermissionsFilesDialog(project, excludedFiles).showAndGet()
  }

  private fun setMode(mode: Mode) {
    // Update buttons state
    for ((radioButtonMode, radioButton) in modeRadioButtons) {
      radioButton.isSelected = radioButtonMode == mode
    }

    // Setup permissions for selected mode
    PreventAutoUpdateModeCookie.using {
      if (mode.fileAccess != null) {
        fileAccessBox.item = mode.fileAccess
      }
      if (mode.terminalAccess != null) {
        terminalAccessBox.item = mode.terminalAccess
      }
      if (mode.runAccess != null) {
        runAccessBox.item = mode.runAccess
      }
      if (mode.toolWindowAccess != null) {
        toolWindowAccessBox.item = mode.toolWindowAccess
      }
    }

    // Update state notifications
    fileAccessWarning.isVisible = fileAccessBox.item > FileAccessLevel.Readonly
    terminalAccessWarning.isVisible = terminalAccessBox.item > TerminalAccessLevel.Readonly
    runAccessWarning.isVisible = runAccessBox.item > RunAccessLevel.Readonly
    toolWindowAccessWarning.isVisible = toolWindowAccessBox.item > ToolWindowAccessLevel.Disabled
    safeStateLabel.isVisible = !fileAccessWarning.isVisible && !terminalAccessWarning.isVisible && !runAccessWarning.isVisible
    unsafeStateLabel.isVisible = !safeStateLabel.isVisible
  }

  private fun autoUpdateMode() {
    if (PreventAutoUpdateModeCookie.get()) return
    val mode = Mode.values().singleOrNull {
      it.fileAccess == fileAccessBox.item &&
      it.terminalAccess == terminalAccessBox.item &&
      it.runAccess == runAccessBox.item &&
      it.toolWindowAccess == toolWindowAccessBox.item
    } ?: Mode.Custom
    setMode(mode)
  }

  public enum class Mode(@Nls val title: String,
                         val icon: Icon,
                         val fileAccess: FileAccessLevel?,
                         val terminalAccess: TerminalAccessLevel?,
                         val runAccess: RunAccessLevel?,
                         val toolWindowAccess: ToolWindowAccessLevel?) {
    Readonly(CodeWithMePluginBundle.message("permission.mode.readonly"), AllIcons.CodeWithMe.CwmPermissionView,
             FileAccessLevel.Readonly,
             TerminalAccessLevel.Readonly,
             RunAccessLevel.Readonly,
             ToolWindowAccessLevel.Disabled),
    EditFiles(CodeWithMePluginBundle.message("permission.mode.editfiles"), AllIcons.CodeWithMe.CwmPermissionEdit,
              FileAccessLevel.FullAccess,
              TerminalAccessLevel.Readonly,
              RunAccessLevel.Readonly,
              ToolWindowAccessLevel.Disabled),
    FullAccess(CodeWithMePluginBundle.message("permission.mode.fullaccess"), AllIcons.CodeWithMe.CwmPermissionEdit,
               FileAccessLevel.FullAccess,
               TerminalAccessLevel.FullAccess,
               RunAccessLevel.FullAccess,
               ToolWindowAccessLevel.FullAccess),
    Custom(CodeWithMePluginBundle.message("permission.mode.custom"), AllIcons.CodeWithMe.CwmPermissionEdit, null, null, null, null)
  }

  private object PreventAutoUpdateModeCookie {
    private var underCookie = false

    fun get() = underCookie

    fun using(action: () -> Unit) {
      underCookie = true
      try {
        action()
      }
      finally {
        underCookie = false
      }
    }
  }
}