package com.jetbrains.rdserver.problems

import com.intellij.codeInsight.problems.WolfTheProblemSolverImpl
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ModalityState
import com.intellij.openapi.project.Project
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.problems.ProblemListener
import com.intellij.problems.WolfTheProblemSolver
import com.intellij.ui.GuiUtils
import com.jetbrains.codeWithMe.model.ProblemFileModel
import com.jetbrains.codeWithMe.model.RdProblemsEvent
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.vfs.BackendVirtualFileSystemHost
import java.io.IOException

@Suppress("NonDefaultConstructor")
class BackendProblemFileHost(
  private val lifetime: Lifetime,
  project: Project,
  protocol: IProtocol
) {

  private val model = ProblemFileModel.create(lifetime, protocol)

  private val wolf = WolfTheProblemSolver.getInstance(project)

  private val serverVfsHost = BackendVirtualFileSystemHost.getInstance(project)

  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendProblemFileHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendProblemFileHost::class.java)
    }
  }

  private val listener = object : ProblemListener {
    override fun problemsAppeared(file: VirtualFile) {
      fireProblem(file, true)
    }

    override fun problemsDisappeared(file: VirtualFile) {
      fireProblem(file, false)
    }
  }

  private fun syncFileProblems(virtualFile: VirtualFile) {
    fireProblem(virtualFile, wolf.isProblemFile(virtualFile))
  }

  private fun fireProblem(file: VirtualFile, hasProblems: Boolean) {
    GuiUtils.invokeLaterIfNeeded(l@{
      if (lifetime.isNotAlive) return@l
      if (file.isDirectory) return@l

      val fileId = try {
        ClientId.withClientId(ClientId.current) {
          serverVfsHost.bindFile(file, true)
        }
      }
      catch (e: IOException) {
        return@l
      }
      if (fileId != null) {
        model.problemsEvent.fire(RdProblemsEvent(hasProblems, fileId))
      }
    }, ModalityState.defaultModalityState())
  }

  init {
    (wolf as? WolfTheProblemSolverImpl)?.problemFiles?.forEach {
      syncFileProblems(it)
    }

    project.messageBus.connect().subscribe(ProblemListener.TOPIC, listener)
  }
}