package com.jetbrains.rdserver.projectView

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.util.treeView.AbstractTreeNode
import com.intellij.openapi.module.ModuleManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.roots.ProjectRootManager
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.psi.PsiElement
import com.intellij.ui.tree.TreeVisitor
import com.jetbrains.codeWithMe.model.ProjectViewModel
import com.jetbrains.codeWithMe.model.RdProjectViewPane
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.toProtocolModel

@Suppress("NonDefaultConstructor")
class BackendProjectViewHost(
  lifetime: Lifetime,
  project: Project,
  protocol: IProtocol
) {

  private val model = ProjectViewModel.create(lifetime, protocol)

  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendProjectViewHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendProjectViewHost::class.java)
    }
  }

  private val paneBindings = hashMapOf<String, BackendProjectViewPaneBinding>()

  init {
    for (provider in BackendProjectPaneProvider.EP_NAME.extensions(project)) {
      if (provider.isAvailable()) {
        val pane = provider.getProjectViewPane()
        val paneModel = RdProjectViewPane(pane.id, pane.title, pane.weight, pane.icon.toProtocolModel())
        val paneBinding = BackendProjectViewPaneBinding(lifetime, project, provider, paneModel)

        paneBindings[pane.id] = paneBinding
        model.panes.add(paneModel)
      }
    }

    // TODO: subscribe for changes
    model.contentModel.projectName.set(project.name)
    model.contentModel.moduleNames.addAll(ModuleManager.getInstance(project).modules.mapNotNull { it.name })
    model.contentModel.contentRoots.addAll(ProjectRootManager.getInstance(project).contentRoots.mapNotNull { it.path })
    model.contentModel.moduleRoots.addAll(ModuleManager.getInstance(project).modules.mapNotNull { it.moduleFile?.parent?.path })
  }

  fun fireModelSelectionChanged(requestFocus: Boolean) {
    paneBindings.values.forEach { it.fireModelSelectionChanged(requestFocus) }
  }

  fun firePaneSelectionChanged(requestFocus: Boolean) {
    paneBindings.values.forEach { it.firePaneSelectionChanged(requestFocus) }
  }

  fun getNodeById(paneId: String, id: Int?): AbstractTreeNode<*>? {
    return paneBindings[paneId]?.getNodeById(id)
  }

  fun isTreeModelBusy(): Boolean {
    return paneBindings.values.any { it.isTreeModelBusy() }
  }

  fun select(paneId: String, element: PsiElement?, virtualFile: VirtualFile?, requestFocus: Boolean) {
    paneBindings[paneId]?.select(element, virtualFile, requestFocus)
  }

  fun selectNode(paneId: String, visitor: TreeVisitor, requestFocus: Boolean) {
    paneBindings[paneId]?.selectNode(visitor, requestFocus)
  }
}