package com.jetbrains.rdserver.projectView

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.projectView.ProjectView
import com.intellij.ide.projectView.ProjectViewNode
import com.intellij.ide.projectView.impl.AbstractProjectViewPane
import com.intellij.ide.projectView.impl.ProjectViewImpl
import com.intellij.ide.projectView.impl.ProjectViewListener
import com.intellij.ide.projectView.impl.nodes.PsiDirectoryNode
import com.intellij.ide.util.treeView.AbstractTreeNode
import com.intellij.ide.util.treeView.PresentableNodeDescriptor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.vfs.VfsUtil
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.wm.ToolWindowId
import com.intellij.openapi.wm.ToolWindowManager
import com.intellij.psi.PsiElement
import com.intellij.ui.tree.TreePathUtil
import com.intellij.ui.tree.TreeVisitor
import com.intellij.util.ui.tree.TreeUtil
import com.jetbrains.codeWithMe.model.*
import com.jetbrains.rd.framework.impl.RdProperty
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.ide.model.RdColoredFragment
import com.jetbrains.rd.ide.model.RdSimpleTextAttributes
import com.jetbrains.rd.platform.codeWithMe.util.toModel
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rd.util.reactive.fire
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.toProtocolModel
import java.io.File
import javax.swing.event.TreeModelEvent
import javax.swing.event.TreeModelListener
import javax.swing.tree.TreePath

class BackendProjectViewPaneBinding(lifetime: Lifetime,
                                    val project: Project,
                                    val provider: BackendProjectPaneProvider,
                                    val model: RdProjectViewPane): TreeModelListener {

  companion object {
    private val logger = getLogger<BackendProjectViewHost>()
  }

  private val clientId = ClientId.current
  private val pane = provider.getProjectViewPane()
  private val treeModel = provider.createAsyncTreeModel(lifetime)

  private val idToNodeMap = hashMapOf<Int?, Any>()
  private val nodeToIdMap = hashMapOf<Any, Int?>()
  private var nextId = 1
  private var pendingOperations = 0

  private val followMeManager = FollowMeManager.getInstance(project)
  private val permissionManager = PermissionManager.getInstance(project)
  private val projectView = ProjectView.getInstance(project) as ProjectViewImpl

  init {
    treeModel.addTreeModelListener(this)
    lifetime.onTermination { treeModel.removeTreeModelListener(this) }

    model.expand.set { id ->
      idToNodeMap[id]?.let {
        expandNode(id, it)
      }
    }

    model.expandByPath.set { _, path ->
      val file = VfsUtil.findFileByIoFile(File(path), true)
                 ?: return@set RdTask<Int>().apply { this.set(-1) }
      logger.trace("Expanding project view to the file '${path}'...")

      val result = RdTask<Int>()
      val visitor = provider.createSelectVisitor(file)
      if (visitor == null) {
        logger.debug("Can not create tree visitor for '${file}'")
        result.set(-1)
        return@set result
      }

      treeModel.accept(visitor, true).onProcessed { treePath ->
        val node = treePath?.lastPathComponent?.let { nodeToIdMap[it] }
        if (node == null) {
          logger.debug("Can not expand tree to '$path'")
        }
        ClientId.withClientId(clientId) {
          result.set(node ?: -1)
        }
      }
      return@set result
    }

    model.navigate.set { navigation ->
      idToNodeMap[navigation.id]?.let {
        val node = getModelNode(it)
        if (node != null && node.canNavigateToSource() && node !is PsiDirectoryNode) {
          node.navigate(navigation.requestFocus)
        }
      }
    }

    val selectionProperty = model.selection as RdProperty<Int?>
    selectionProperty.advise(lifetime) { id ->
      if (selectionProperty.isLocalChange) return@advise
      onModelSelectionChanged(id, false)
    }

    project.messageBus
      .connect(lifetime.createNestedDisposable())
      .subscribe(ProjectViewListener.TOPIC, object : ProjectViewListener {
        override fun paneShown(current: AbstractProjectViewPane, previous: AbstractProjectViewPane?) {
          subscribeForSelection(current)
        }
      })

    subscribeForSelection(pane)

    // start syncing
    idToNodeMap[null] = treeModel.root
    nodeToIdMap[treeModel.root] = null
    expandNode(null, treeModel.root)
  }

  fun fireModelSelectionChanged(requestFocus: Boolean) = onModelSelectionChanged(model.selection.value, requestFocus)
  fun firePaneSelectionChanged(requestFocus: Boolean) = onPaneSelectionChanged(pane.selectedPath, requestFocus)

  private fun onModelSelectionChanged(id: Int?, requestFocus: Boolean) {
    val node = idToNodeMap[id] ?: return
    val modelNode = getModelNode(node) ?: return
    val visitor = createVisitorByTreeNodeValue(modelNode)

    // Select virtual file in current project view
    if (followMeManager.isHostFollowingClient(clientId)) {
      val projectViewToolWindow = ToolWindowManager.getInstance(project).getToolWindow(ToolWindowId.PROJECT_VIEW)
      projectViewToolWindow?.show {
        projectView.changeViewCB(pane.id, pane.subId).doWhenProcessed {
          TreeUtil.promiseSelect(pane.tree, visitor)
        }
      }
      if (requestFocus) {
        projectViewToolWindow?.activate(null)
      }
    }

    // Select virtual file in all other client views
    for (followerId in followMeManager.getFollowersWithoutHost(clientId)) {
      val host = RdServerPerClientComponents.getService(project, followerId, BackendProjectViewHost::class.java)
      host.selectNode(pane.id, visitor, requestFocus)
    }
  }

  private fun onPaneSelectionChanged(path: TreePath?, requestFocus: Boolean) {
    if (FollowMeManager.getInstance(project).isClientFollowingHost(clientId)) {
      val node = path?.lastPathComponent?.let { getModelNode(it) } ?: return
      val visitor = createVisitorByTreeNodeValue(node)
      selectNode(visitor, requestFocus)
    }
  }

  fun getNodeById(id: Int?): AbstractTreeNode<*>? {
    val treeNode = idToNodeMap[id] ?: return null
    return getModelNode(treeNode)
  }

  fun isTreeModelBusy(): Boolean {
    return pendingOperations > 0 || treeModel.isProcessing
  }

  private fun expandNode(id: Int?, node: Any) {
    logger.trace("Expand: $id")
    val count = treeModel.getChildCount(node)
    if (count == 0) {
      // this is an empty node
      model.update.fire(id)
      logger.trace("Refresh node: $id (children collection is empty)")
    }
  }

  fun select(element: PsiElement?, virtualFile: VirtualFile?, requestFocus: Boolean) {
    val visitor = AbstractProjectViewPane.createVisitor(element, virtualFile)
    if (visitor != null) {
      selectNode(visitor, requestFocus)
    }
  }

  fun selectNode(visitor: TreeVisitor, requestFocus: Boolean) {
    logger.trace("Start selecting...")
    pendingOperations++

    treeModel.accept(visitor).onProcessed {
      logger.trace("Finish selecting")
      pendingOperations--

      val lastPathComponent = it?.lastPathComponent ?: return@onProcessed
      val node = nodeToIdMap[lastPathComponent]
      if (node != null) {
        ClientId.withClientId(clientId) {
          logger.trace("Set selection into protocol model: $node")
          model.selection.set(node)
          if (requestFocus) {
            model.activate.fire()
          }
        }
      }
    }
  }

  private fun subscribeForSelection(pane: AbstractProjectViewPane) {
    val tree = pane.tree ?: return
    tree.selectionModel.addTreeSelectionListener { event ->
      onPaneSelectionChanged(event.newLeadSelectionPath, false)
    }
  }

  override fun treeNodesInserted(e: TreeModelEvent) {
    updateNode(e) { parentId ->
      refreshNodeOnChildrenChange(parentId)

      for ((index, child) in e.children.withIndex()) {
        val childDescriptor = presentNode(child, e.childIndices[index] - 1) ?: continue

        val childId = nextId++
        idToNodeMap[childId] = child
        nodeToIdMap[child] = childId

        model.items[childId] = RdProjectViewItem(parentId, childDescriptor)
        logger.trace("Insert child: $childId $child -> $parentId")
      }
    }
  }

  override fun treeStructureChanged(e: TreeModelEvent) {
    logger.warn("Not implemented event: treeStructureChanged for ${e.treePath}")
  }

  override fun treeNodesChanged(e: TreeModelEvent) {
    updateNode(e) { parentId ->
      for ((index, child) in e.children.withIndex()) {
        val childId = nodeToIdMap[child] ?: continue
        val childDescriptor = presentNode(child, index) ?: continue
        model.items[childId] = RdProjectViewItem(parentId, childDescriptor)
        logger.trace("Update child: $childId $child -> $parentId")
      }
    }
  }

  override fun treeNodesRemoved(e: TreeModelEvent) {
    updateNode(e) { parentId ->
      refreshNodeOnChildrenChange(parentId)
      for (child in e.children) {
        val childId = nodeToIdMap[child] ?: continue

        nodeToIdMap.remove(child)
        idToNodeMap.remove(childId)
        model.items.remove(childId)
        logger.trace("Remove child: $childId $child")
      }
    }
  }

  private fun updateNode(e: TreeModelEvent, action: (Int?) -> Unit) {
    val node = e.treePath.lastPathComponent
    if (!nodeToIdMap.containsKey(node))
      return
    if (e.children == null) {
      logger.trace("Skip updating: ${node}, children is null")
      return
    }

    ClientId.withClientId(clientId) {
      val id = nodeToIdMap[node]
      action(id)
      model.update.fire(id)
      logger.trace("Refresh: $id")
    }
  }

  private fun presentNode(node: Any, sortKey: Int): RdProjectViewItemDescriptor? {

    val modelNode = getModelNode(node) ?: return null

    var virtualFile = (modelNode as? ProjectViewNode<*>)?.virtualFile ?: provider.getVirtualFile(modelNode)
    if (virtualFile != null && !permissionManager.isVisible(clientId, virtualFile)) {
      if (provider.shouldSyncExternalNode(modelNode)) {
        logger.debug("Sync invisible node but without virtual file")
        virtualFile = null
      } else {
        logger.debug("Skip invisible node: $modelNode")
        return null
      }
    }

    val presentation = modelNode.presentation
    val path = virtualFile?.path

    val fragments = presentation.coloredText.map {
      RdColoredFragment(it.text, it.toolTip, it.attributes.toModel())
    }.toMutableList()

    if (fragments.isEmpty()) {
      fragments.add(
        RdColoredFragment(
          presentation.presentableText ?: "",
          presentation.tooltip,
          RdSimpleTextAttributes(null, null, null, 0)
        )
      )
    }

    val scopeModel = ProjectScopeUtils.getScope(project, virtualFile)
    val hasChildren = !treeModel.isLeaf(node)

    // Do not navigate to Project/Module nodes, we do not have such UI anyway
    val canNavigate = modelNode.canNavigateToSource() && modelNode !is PsiDirectoryNode

    // Get node icon under local ID, truly magic can happen inside
    // More info:
    //  calculating icon can trigger downloading schemes on a background thread so it is better
    //  to trigger such operations as 'local' ones to avoid handling download callback as a client event
    val icon = ClientId.withClientId(ClientId.localId) {
      try {
        presentation.getIcon(false)?.toProtocolModel()
      }
      catch (ex: Exception) {
        logger.warn(ex)
        null
      }
    }

    return RdProjectViewItemDescriptor(
      scopeModel,
      path,
      sortKey,
      fragments,
      icon,
      hasChildren,
      canNavigate
    )
  }

  /**
   * Refresh a node when node's children were changed to update `hasChildren` in protocol model
   */
  private fun refreshNodeOnChildrenChange(nodeId: Int?) {
    nodeId ?: return
    val node = idToNodeMap[nodeId] ?: return
    val viewItem = model.items[nodeId] ?: return

    val hasChildren = !treeModel.isLeaf(node)
    val descriptor = RdProjectViewItemDescriptor(
      viewItem.descriptor.scope,
      viewItem.descriptor.path,
      viewItem.descriptor.sortKey,
      viewItem.descriptor.presentation,
      viewItem.descriptor.icon,
      hasChildren,
      viewItem.descriptor.canNavigate
    )

    logger.trace("Update node because children were changed: $nodeId")
    model.items[nodeId] = RdProjectViewItem(viewItem.parentId, descriptor)
    model.update.fire(viewItem.parentId)
  }

  private fun createVisitorByTreeNodeValue(node: AbstractTreeNode<*>): TreeVisitor {
    logger.trace("Creating visitor for node '${node.name}'...")
    val path = TreePathUtil.pathToCustomNode<AbstractTreeNode<*>>(node) { it.parent }
    return TreeVisitor.ByTreePath(path) { getModelNode(it) }
  }

  private fun getModelNode(treeNode: Any): AbstractTreeNode<*>? {
    return TreeUtil.getUserObject(PresentableNodeDescriptor::class.java, treeNode) as? AbstractTreeNode<*>
  }
}