package com.jetbrains.rdserver.projectView.providers

import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isLocal
import com.intellij.ide.IdeView
import com.intellij.ide.projectView.ProjectViewNode
import com.intellij.ide.projectView.impl.ProjectViewPane
import com.intellij.ide.util.DirectoryChooserUtil
import com.intellij.ide.util.EditorHelper
import com.intellij.ide.util.treeView.AbstractTreeNode
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.DataProvider
import com.intellij.openapi.actionSystem.LangDataKeys
import com.intellij.openapi.actionSystem.PlatformDataKeys
import com.intellij.openapi.module.Module
import com.intellij.openapi.module.ModuleUtilCore
import com.intellij.openapi.project.Project
import com.intellij.psi.PsiDirectory
import com.intellij.psi.PsiElement
import com.intellij.psi.PsiFile
import com.intellij.psi.PsiFileSystemItem
import com.intellij.psi.util.PsiUtilCore
import com.jetbrains.rdserver.actions.BackendActionDataProvider
import com.jetbrains.rdserver.actions.BackendDataKeys
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.projectView.BackendProjectViewHost
import java.util.HashSet

class BackendProjectViewDateProvider : BackendActionDataProvider {
  override fun getData(dataId: String, dataProvider: DataProvider): Any? {
    val project = CommonDataKeys.PROJECT.getData(dataProvider) ?: return null
    val projectViewSelection = BackendDataKeys.PROJECT_VIEW_SELECTION.getData(dataProvider) ?: return null

    val nodes = projectViewSelection
      .getNodes(project)
      .mapNotNull { it as? ProjectViewNode<*> }
    if (nodes.isEmpty()) return null

    val psiElements = nodes
      .mapNotNull { findPsiElement(it) }
      .toTypedArray()
    if (psiElements.isEmpty()) return null

    val delegator = BackendCopyPasteDelegator(project, psiElements.toList().toTypedArray())

    return when {
      LangDataKeys.IDE_VIEW.`is`(dataId) -> object : IdeView {
        override fun selectElement(element: PsiElement?) = selectElementOnClient(element)
        override fun getOrChooseDirectory() = DirectoryChooserUtil.getOrChooseDirectory(this)
        override fun getDirectories() = psiElements.filterIsInstance<PsiDirectory>().toTypedArray()
      }
      LangDataKeys.PSI_ELEMENT_ARRAY.`is`(dataId) -> psiElements
      LangDataKeys.PASTE_TARGET_PSI_ELEMENT.`is`(dataId) -> psiElements.filterIsInstance<PsiDirectory>().firstOrNull()
                                                            ?: psiElements.filterIsInstance<PsiFile>().mapNotNull { it.containingDirectory }.firstOrNull()
      PlatformDataKeys.COPY_PROVIDER.`is`(dataId) -> withCheckingPermissions(project) { delegator.copyProvider }
      PlatformDataKeys.CUT_PROVIDER.`is`(dataId) -> withCheckingPermissions(project) { delegator.cutProvider }
      PlatformDataKeys.PASTE_PROVIDER.`is`(dataId) -> withCheckingPermissions(project) { delegator.pasteProvider }
      PlatformDataKeys.DELETE_ELEMENT_PROVIDER.`is`(dataId) -> withCheckingPermissions(project) { BackendProjectViewDeleteElementProvider }
      LangDataKeys.MODULE.`is`(dataId) -> {
        val modules: MutableSet<Module?> = HashSet()
        for (psiElement in psiElements) {
          modules.add(ModuleUtilCore.findModuleForPsiElement(psiElement))
        }
        return if (modules.size == 1) modules.iterator().next() else null
      }
      else -> null
    }
  }

  private fun withCheckingPermissions(project: Project, getter: () -> Any): Any? {
    if (PermissionManager.getInstance(project).hasFileWriteAccess(ClientId.current)) {
      return getter()
    }
    return null
  }

  private fun findPsiElement(node: AbstractTreeNode<*>): PsiFileSystemItem? {
    val elementFromNode = node.value as? PsiFileSystemItem
    if (elementFromNode != null) return elementFromNode

    val element = node.value as? PsiElement
    if (element != null) return element.containingFile

    val projectViewNode = node as? ProjectViewNode<*>
    val virtualFile = projectViewNode?.virtualFile
    if (virtualFile != null) {
      return PsiUtilCore.findFileSystemItem(projectViewNode.project, virtualFile)
    }

    return null
  }

  /**
   * Implements the same logic as IdeViewForProjectViewPane.selectElement, but for remote client
   */
  private fun selectElementOnClient(element: PsiElement?) {
    element ?: return

    val clientId = ClientId.current
    if (clientId.isLocal) return

    val file = PsiUtilCore.getVirtualFile(element)
    val requestFocus = element is PsiDirectory

    val host = BackendProjectViewHost.getInstance(element.project, ClientId.current)
    host.select(ProjectViewPane.ID, element, file, requestFocus)
    if (!requestFocus && null == EditorHelper.openInEditor(element, false, true)) {
      host.select(ProjectViewPane.ID, element, file, true)
    }
  }
}