package com.jetbrains.rdserver.quickDoc

import com.intellij.codeInsight.documentation.DocumentationComponent
import com.intellij.codeInsight.documentation.DocumentationManager
import com.intellij.codeInsight.lookup.LookupManager
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.ActionCallback
import com.intellij.psi.PsiElement
import com.intellij.psi.PsiFile
import com.intellij.ui.popup.PopupUpdateProcessor
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.editors.BackendServerFileEditorManager
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.followMe.FollowMeManager
import org.jetbrains.annotations.NotNull
import java.util.concurrent.CompletableFuture
import java.util.concurrent.atomic.AtomicInteger

class BackendDocumentationManager(project: @NotNull Project) : DocumentationManager(project) {
  private val idCounter = AtomicInteger(0)

  private var currentSessionId = 0  // Note: it's accessed from EDT only so there is no need to make it volatile

  override fun cancelAndFetchDocInfoByLink(component: DocumentationComponent, provider: DocumentationCollector): ActionCallback {
    provider.getElement(false)?.let { resolvedLink ->
      forEachFollower { followerId ->
        showJavaDocInfoFor(followerId, resolvedLink, currentSessionId)
      }
    }
    return super.cancelAndFetchDocInfoByLink(component, provider)
  }

  override fun doShowJavaDocInfo(elementFuture: CompletableFuture<out PsiElement>,
                                 requestFocus: Boolean,
                                 updateProcessor: PopupUpdateProcessor,
                                 originalElement: PsiElement?,
                                 closeCallback: Runnable?,
                                 documentation: String?,
                                 useStoredPopupSize: Boolean,
                                 onAutoUpdate: Boolean) {
    fun show(clientEditor: Editor?, psiFile: PsiFile, sessionId: Int, closeAllFollowers: Runnable) {
      if (ClientId.isCurrentlyUnderLocalId) {
        currentSessionId = sessionId
        super.doShowJavaDocInfo(elementFuture, requestFocus, updateProcessor, originalElement, closeAllFollowers,
                                documentation, useStoredPopupSize, onAutoUpdate)
      } else if (clientEditor != null) {
        BackendQuickDocHost.getInstance(myProject).fetchDocumentation(clientEditor, psiFile, sessionId)
      }
    }

    val currentEditor = editor ?: return
    val psiFile = originalElement?.containingFile ?: return
    val sessionId = idCounter.incrementAndGet()
    val closeAllFollowers = Runnable {
      closeCallback?.run()
      forEachFollower { followerId ->
        closeDocInfo(followerId, sessionId)
      }
    }
    show(currentEditor, psiFile, sessionId, closeAllFollowers)
    forEachFollower { followerId ->
      ClientId.withClientId(followerId) {
        val clientEditor = if (!ClientId.isCurrentlyUnderLocalId) findMatchingClientEditor(currentEditor.document) else null
        show(clientEditor, psiFile, sessionId, closeAllFollowers)
      }
    }
  }

  fun closeDocInfo(clientId: ClientId, sessionId: Int) {
    ClientId.withClientId(clientId) {
      if (clientId == ClientId.localId) {
        docInfoHint?.takeIf { it.isVisible }?.cancel()
      } else {
        BackendQuickDocHost.getInstance(myProject).closeDocSession(sessionId)
      }
    }
  }

  fun showJavaDocInfoFor(clientId: ClientId, targetElement: PsiElement, sessionId: Int) {
    ClientId.withClientId(clientId) {
      if (clientId == ClientId.localId) {
        currentSessionId = sessionId
        showJavaDocInfo(targetElement, null, LookupManager.getActiveLookup(editor) == null, null)
      } else {
        val document = editor?.document ?: return
        val psiFile = targetElement.containingFile ?: return
        findMatchingClientEditor(document)?.let { followerEditor ->
          BackendQuickDocHost.getInstance(myProject).fetchDocumentation(followerEditor, psiFile, targetElement, null, sessionId)
        }
      }
    }
  }

  private inline fun forEachFollower(task: (ClientId) -> Unit) {
    for (followerId in FollowMeManager.getInstance(project).getFollowers(ClientId.current)) {
      task(followerId)
    }
  }

  private fun findMatchingClientEditor(document: Document): Editor? {
    val backendDocumentHost = BackendDocumentHost.getInstance(project)
    val documentId = backendDocumentHost.tryGetSynchronizer(document)?.documentId ?: return null
    val followerEditor = BackendServerFileEditorManager.getInstance(project).selectedTextEditor ?: return null
    return followerEditor.takeIf { documentId == followerEditor.getTextControlId()?.documentId }
  }
}