package com.jetbrains.rdserver.quickDoc

import com.intellij.codeInsight.daemon.impl.DaemonCodeAnalyzerImpl
import com.intellij.codeInsight.daemon.impl.ShowIntentionsPass
import com.intellij.codeInsight.daemon.impl.tooltips.TooltipActionProvider
import com.intellij.codeInsight.documentation.DocumentationManager
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.EditorMouseHoverPopupManager
import com.intellij.openapi.project.Project
import com.intellij.psi.PsiDocumentManager
import com.intellij.psi.PsiElement
import com.intellij.psi.PsiFile
import com.jetbrains.rd.actions.TooltipActionId
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.ui.intentions.BackendIntentionActionFilter

@Suppress("NonDefaultConstructor")
class BackendEditorMouseHoverPopupHost(
  val project: Project,
  val solution: Solution
) {
  val model = solution.quickDocHostModel

  companion object {
    fun getInstance(project: Project): BackendEditorMouseHoverPopupHost {
      return RdServerPerClientComponents.getService(project, BackendEditorMouseHoverPopupHost::class.java)
    }
  }

  init {
    model.startHoverDocSession.set(null) { _, offset ->
      val focusedTextControlId = solution.documentsOperationModel.codeWithMeFocusesModel.lastFocusedTextControl.value
      val editor = focusedTextControlId?.let { BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(focusedTextControlId) }
                   ?: return@set RdTask.fromResult(Unit)
      val psiFile = PsiDocumentManager.getInstance(project).getPsiFile(editor.document) ?: return@set RdTask.fromResult(Unit)

      val hoverTooltip = getHoverTooltip(editor, psiFile, offset)
      model.hoverTooltip.set(hoverTooltip)

      val hoverDocSession = getHoverDocSession(editor, psiFile, offset)
      model.hoverDocSession.set(hoverDocSession)

      RdTask.fromResult(Unit)
    }
  }

  private fun getHoverTooltip(editor: Editor, psiFile: PsiFile, offset: Int): HoverTooltip? {
    val highlightInfo = (DaemonCodeAnalyzerImpl.getInstance(project) as DaemonCodeAnalyzerImpl)
                          .findHighlightByOffset(editor.document, offset, true) ?: return null
    val tooltipAction = TooltipActionProvider.calcTooltipAction(highlightInfo, editor)
    val myAction = ShowIntentionsPass.getAvailableFixes(editor, psiFile, -1, offset)
      .firstOrNull { it.action.text == tooltipAction?.text }?.action
    val highlighterId = highlightInfo.highlighter?.id ?: return null

    return if (myAction == null || !BackendIntentionActionFilter.acceptErrorFix(myAction)) {
      null
    }
    else {
      val text = myAction.text
      val actionId = TooltipActionId(text, highlighterId).toActionId()
      HoverTooltip(actionId, text)
    }
  }

  private fun getHoverDocSession(editor: Editor, psiFile: PsiFile, offset: Int): QuickDocSession? {
    val documentationManager = DocumentationManager.getInstance(project)
    val element = psiFile.findElementAt(offset) ?: return null
    val targetElement = documentationManager.findTargetElement(editor, offset, psiFile, element) ?: return null

    val textControl = editor.getTextControlId() ?: error("cannot find textControlId for editor")
    val info = generateDocInfo(targetElement, element)  ?: return null
    val session = QuickDocSession(info, textControl)

    session.fetchInfoByLink.set { _, url ->
      val providerFromElement = DocumentationManager.getProviderFromElement(targetElement)
      val resolvedLink = providerFromElement.getDocumentationElementForLink(psiFile.manager, url, targetElement)

      return@set if (resolvedLink != null) {
        RdTask.fromResult(generateDocInfo(resolvedLink, null)
                          ?: QuickDocElementInfo(BackendQuickDocHost.NO_DOCUMENTATION_TEXT.richText, null, null))
      }
      else {
        RdTask.fromResult(QuickDocElementInfo(BackendQuickDocHost.NO_DOCUMENTATION_TEXT.richText, null, null))
      }
    }
    return session
  }

  private fun generateDocInfo(targetElement: PsiElement, element: PsiElement?): QuickDocElementInfo? {
    val documentationManager = DocumentationManager.getInstance(project)

    var quickDocMessage = documentationManager.generateDocumentation(targetElement, element, true) ?: return null
    if (quickDocMessage.length > EditorMouseHoverPopupManager.MAX_QUICK_DOC_CHARACTERS) {
      quickDocMessage = quickDocMessage.take(EditorMouseHoverPopupManager.MAX_QUICK_DOC_CHARACTERS)
    }

    return QuickDocElementInfo(quickDocMessage.richText, null, null)
  }

  private val String.richText
    get() = RichAttributedTextModel(listOf(RichAttributedStringModel(this, null)))
}