// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.quickDoc

import com.intellij.codeInsight.CodeInsightBundle
import com.intellij.codeInsight.completion.CompletionUtil
import com.intellij.codeInsight.documentation.DocumentationManager
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.fileEditor.OpenFileDescriptor
import com.intellij.openapi.project.Project
import com.intellij.psi.PsiElement
import com.intellij.psi.PsiFile
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.IViewableMap
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.completion.BackendCompletionHost
import com.jetbrains.rdserver.editors.getTextControlId
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.followMe.FollowMeManager

@Suppress("NonDefaultConstructor")
class BackendQuickDocHost(
  private val lifetime: Lifetime,
  private val project: Project,
  private val solution: Solution
) {

  private val clientId = ClientId.current

  init {
    val sessions = solution.quickDocHostModel.quickDocSessions
    sessions.change.advise(lifetime) { event ->
      if (event is IViewableMap.Event.Remove) {
        val documentationManager = DocumentationManager.getInstance(project) as BackendDocumentationManager
        for (followerId in FollowMeManager.getInstance(project).getFollowers(clientId)) {
          documentationManager.closeDocInfo(followerId, event.key)
        }
      }
    }
  }

  fun fetchDocumentation(editor: Editor, psiFile: PsiFile, sessionId: Int) {
    if (lifetime.isNotAlive) return
    val documentationManager = DocumentationManager.getInstance(project)

    val element: PsiElement?
    val targetElement: PsiElement?

    val currentSelectedLookupItem = tryGetCurrentSelectedLookupItem(editor)
    if (currentSelectedLookupItem != null) {
      targetElement = getTargetElementForLookup(psiFile, currentSelectedLookupItem)
      element = null
    }
    else {
      element = psiFile.findElementAt(editor.caretModel.offset) ?: return
      targetElement = documentationManager.findTargetElement(editor, psiFile, element) ?: return
    }

    LOG.info("Fetching doc: targetElement=$targetElement, element=$element")

    if (targetElement == null) return
    fetchDocumentation(editor, psiFile, targetElement, element, sessionId)
  }

  fun fetchDocumentation(editor: Editor, psiFile: PsiFile, targetElement: PsiElement, element: PsiElement?, sessionId: Int) {
    val psiManager = psiFile.manager
    val documentationManager = DocumentationManager.getInstance(project)
    generateDocAndInvoke(documentationManager, targetElement, element) l1@{ info ->
      if (lifetime.isNotAlive) return@l1

      val textControlId = editor.getTextControlId() ?: error("cannot find textControlId for editor")
      val session = QuickDocSession(info, textControlId)

      session.fetchQuickDocAgain.set { _, _ ->
        generateQuickDocForCompletion(editor, psiFile)
      }

      session.navigate.advise(Lifetime.Eternal) {
        if (it.url == null) {
          LOG.info("Open file in editor for element $targetElement in offset ${targetElement.textOffset}")
          OpenFileDescriptor(project, psiFile.virtualFile, targetElement.textOffset).navigate(true)
        }
        else {
          LOG.info("Navigate to url ${it.url}")
        }
      }

      session.fetchInfoByLink.set { _, url ->
        val providerFromElement = DocumentationManager.getProviderFromElement(targetElement)
        val resolvedLink = providerFromElement.getDocumentationElementForLink(psiManager, url, targetElement)
        LOG.info("Fetching info by link: url=$url, resolvedLink=$resolvedLink")
        val result = RdTask<QuickDocElementInfo>()
        if (resolvedLink != null) {
          generateDocAndInvoke(documentationManager, resolvedLink, null) l2@{ info ->
            if (lifetime.isNotAlive) return@l2

            result.set(info)
          }
        }
        else {
          result.set(QuickDocElementInfo(NO_DOCUMENTATION_TEXT.richText, null, null))
        }
        if (resolvedLink != null) {
          val backendDocumentationManager = documentationManager as BackendDocumentationManager
          for (followerId in FollowMeManager.getInstance(project).getFollowers(clientId)) {
            backendDocumentationManager.showJavaDocInfoFor(followerId, resolvedLink, sessionId)
          }
        }
        result
      }

      solution.quickDocHostModel.quickDocSessions[sessionId] = session
    }
  }

  fun closeDocSession(sessionId: Int) {
    application.assertIsDispatchThread()
    val sessions = solution.quickDocHostModel.quickDocSessions
    if (sessions.containsKey(sessionId)) {
      sessions.remove(sessionId)
    }
  }

  private fun generateQuickDocForCompletion(editor: Editor, psiFile: PsiFile): RdTask<QuickDocElementInfo> {
    val result = RdTask<QuickDocElementInfo>()
    val docStr: String?


    val currentSelectedLookupItem = tryGetCurrentSelectedLookupItem(editor)
    if (currentSelectedLookupItem == null) {
      docStr = null
    }
    else {
      val targetElement = getTargetElementForLookup(psiFile, currentSelectedLookupItem)
      if (targetElement == null) {
        docStr = null
      }
      else {
        val documentationManager = DocumentationManager.getInstance(psiFile.project)
        // TODO [V.Petrenko] better to call doc generation in background thread
        docStr = documentationManager.generateDocumentation(targetElement, null, false)
      }
    }

    LOG.info("Generated doc for completion: currentSelectedLookupItem=${currentSelectedLookupItem?.shortText}, " +
             "docStr.isNullOrEmpty()=${docStr.isNullOrEmpty()}")
    result.set(QuickDocElementInfo((docStr ?: NO_DOCUMENTATION_TEXT).richText, null, null))
    return result
  }

  private fun tryGetCurrentSelectedLookupItem(editor: Editor): CompletionItem? {
    val project = editor.project ?: return null
    val backendDocumentHost = BackendDocumentHost.getInstance(project)
    val documentSynchronizer = backendDocumentHost.tryGetSynchronizer(editor.document)
      ?: return null
    val textControlModel = documentSynchronizer.modelDocument.textControls[editor.getTextControlId()] ?: return null

    return textControlModel.completionModel.completionSession.value?.session?.currentSelectedItem?.valueOrNull
  }


  private fun getTargetElementForLookup(psiFile: PsiFile,
                                        currentSelectedLookupItem: CompletionItem): PsiElement? {
    val correspondingLookupElement =
      BackendCompletionHost.getInstance(psiFile.project).getCorrespondingLookupElement(currentSelectedLookupItem)
    val documentationProvider = DocumentationManager.getProviderFromElement(psiFile)
    return documentationProvider.getDocumentationElementForLookupItem(
      psiFile.manager, correspondingLookupElement?.getObject(), null
    ) ?: CompletionUtil.getTargetElement(correspondingLookupElement)
  }

  private fun generateDocAndInvoke(documentationManager: DocumentationManager,
                                   targetElement: PsiElement,
                                   element: PsiElement?,
                                   withDoc: (QuickDocElementInfo) -> Unit) {
    ApplicationManager.getApplication().executeOnPooledThread {
      val description = documentationManager.generateDocumentation(targetElement, element, false) ?: NO_DOCUMENTATION_TEXT
      val info = QuickDocElementInfo(description.richText, null, null)
      ApplicationManager.getApplication().invokeLater {
        withDoc.invoke(info)
      }
    }
  }

  companion object {
    private val String.richText
      get() = RichAttributedTextModel(listOf(RichAttributedStringModel(this, null)))

    private val LOG = logger<BackendQuickDocHost>()
    internal val NO_DOCUMENTATION_TEXT = CodeInsightBundle.message("no.documentation.found")

    fun getInstance(project: Project) = RdServerPerClientComponents.getService(project, BackendQuickDocHost::class.java)
  }
}
