package com.jetbrains.rdserver.requests

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.project.Project
import com.intellij.util.SmartList
import com.jetbrains.rd.framework.util.adviseSuspend
import com.jetbrains.rd.ide.model.RdFailureResponse
import com.jetbrains.rd.ide.model.RdRequest
import com.jetbrains.rd.ide.model.RdResponse
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.daemon.BackendMarkupSuppressors

private val logger = logger<BackendAsyncRequestExecutor>()

@Suppress("NonDefaultConstructor")
class BackendAsyncRequestExecutor(
  lifetime: Lifetime,
  private val project: Project,
  solution: Solution,
) {
  private val protocolModel = solution.documentsOperationModel.synchronizationModel
  private val publisher get() = application.messageBus.syncPublisher(BackendClientOperationListener.TOPIC)

  init {
    protocolModel.requests.adviseSuspend(lifetime, protocolModel.protocol.scheduler, ::executeRequests)
  }

  private suspend inline fun executeRequest(request: RdRequest, crossinline consumer: (RdResponse) -> Unit) {
    val requestName = request.debugString()
    logger.trace { ">> $requestName" }
    val startTime = System.nanoTime()
    try {
      val executorDelegate = BackendAsyncRequestDelegates.getDelegate(request) ?: error(
        "Unknown request: ${request::class.simpleName}::${request.id}")
      executorDelegate.execute(request, project) { consumer(it) }
    }
    catch (t: Throwable) {
      logger.error(t)
      consumer(RdFailureResponse(t.message ?: "exception", request.id))
    }
    finally {
      logger.trace { "<< $requestName; execution time: ${(System.nanoTime() - startTime) / 1e6}, client: ${ClientId.current}" }
    }
  }

  private suspend fun executeRequests(requests: List<RdRequest>) {
    application.assertIsDispatchThread()
    val isBatchMode = requests.size > 1
    val operationName = if (isBatchMode) "Batch request" else requests.single().debugString()
    try {
      publisher.beforeGuestOperation(operationName)
      if (!isBatchMode) {
        val request = requests.single()
        executeRequest(request) { protocolModel.responses.fire(SmartList(it)) }
        return
      }

      val startBatchTime = System.nanoTime()
      val responses = SmartList<RdResponse>()
      BackendMarkupSuppressors.fullSuppressMarkupSynchronization(project, "Executing requests batch") {
        for (request in requests) {
          executeRequest(request) { responses.add(it) }
        }
        logger.trace { "Requests batch size: ${requests.size}, summarized execution time: ${(System.nanoTime() - startBatchTime) / 1e6}" }
        protocolModel.responses.fire(responses)
      }
    }
    finally {
      publisher.afterGuestOperation(operationName)
    }
  }
}

fun RdRequest.debugString(): String = "${this::class.simpleName}#${this.id}"