package com.jetbrains.rdserver.requests

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.actionSystem.ex.AnActionListener
import com.intellij.openapi.command.CommandEvent
import com.intellij.openapi.command.CommandListener
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.jetbrains.rd.ide.model.RdPatch
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.ide.requests.IdeGateway
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.daemon.BackendMarkupSuppressors
import com.jetbrains.rdserver.document.BackendDocumentHost
import com.jetbrains.rdserver.editors.getSynchronizer
import com.jetbrains.rd.util.getLogger
import com.jetbrains.rd.util.trace

@Suppress("NonDefaultConstructor")
class BackendIdeGateway(
  private val clientId: ClientId,
  lifetime: Lifetime,
  private val project: Project,
  solution: Solution
) : IdeGateway(lifetime), BackendClientOperationListener, AnActionListener, CommandListener {
  companion object {
    fun getInstance(project: Project): BackendIdeGateway = RdServerPerClientComponents.getService(project)
    val logger = getLogger<BackendIdeGateway>()
  }

  private val protocolModel = solution.documentsOperationModel.synchronizationModel
  private val backendMarkupSuppressors = RdServerPerClientComponents.getService<BackendMarkupSuppressors>(project)

  init {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(BackendClientOperationListener.TOPIC, this)
    connection.subscribe(AnActionListener.TOPIC, this)
    connection.subscribe(CommandListener.TOPIC, this)
  }

  override fun sendPatch(patch: RdPatch) {
    protocolModel.backendPatches.fire(patch)
  }

  override suspend fun executeBulkUpdate(editor: Editor, block: suspend () -> Unit): RdPatch? {
    return executeBulkUpdate {
      val editorSynchronizer = editor.getSynchronizer() ?: error("editorSynchronizer")
      val editorId = editorSynchronizer.textControlId
      val documentHost = BackendDocumentHost.getInstance(project)
      val documentSynchronizer = documentHost.tryGetSynchronizer(editorId.documentId) ?: error("documentSynchronizer")
      currentPatchBuilder!!.pushInitialSelectionHintAndEditorVersion(editorId, editor, editorSynchronizer.version, documentSynchronizer.version)
      block()

      logger.trace { "Start awaiting async modifications" }
      BackendIdeGatewayHelper.getInstance(project).awaitAsyncModifications()
      logger.trace { "End awaiting async modifications" }
    }
  }

  override fun beforeBulkUpdateOnUnknownOperation(lifetime: Lifetime) {
    backendMarkupSuppressors.getGlobalSuppressor().suppress(lifetime, "Bulk Update On Unknown Operation")
  }

  // region grouping other guests operations
  override fun beforeGuestOperation(name: String) {
    if (ClientId.current == clientId)  return
    startBulkUpdateOnUnknownOperation(name)
  }

  override fun afterGuestOperation(name: String) {
    if (ClientId.current == clientId)  return
    endBulkUpdateOnUnknownOperation()
  }
  // endregion

  // region grouping host operations
  override fun beforeActionPerformed(action: AnAction, dataContext: DataContext, event: AnActionEvent) {
    if (!ClientId.isCurrentlyUnderLocalId)  return
    startBulkUpdateOnUnknownOperation("host action")
  }

  override fun afterActionPerformed(action: AnAction, dataContext: DataContext, event: AnActionEvent) {
    if (!ClientId.isCurrentlyUnderLocalId)  return
    endBulkUpdateOnUnknownOperation()
  }

  override fun beforeEditorTyping(c: Char, dataContext: DataContext) {
    if (!ClientId.isCurrentlyUnderLocalId)  return
    startBulkUpdateOnUnknownOperation("host typing")
  }

  override fun afterEditorTyping(c: Char, dataContext: DataContext) {
    if (!ClientId.isCurrentlyUnderLocalId)  return
    endBulkUpdateOnUnknownOperation()
  }
  // endregion

  private var unknownUndoCommandInProgress = false

  override fun commandStarted(event: CommandEvent) {
    if (unknownOperationInProgress || bulkUpdateInProgress) return
    unknownUndoCommandInProgress = true
    startBulkUpdateOnUnknownOperation("Executing unknown command: ${event.commandName}, clientId: ${ClientId.current}").onTermination {
      require(unknownUndoCommandInProgress)
      unknownUndoCommandInProgress = false
    }
  }

  override fun commandFinished(event: CommandEvent) {
    if (!unknownUndoCommandInProgress) return
    endBulkUpdateOnUnknownOperation()
  }
}