package com.jetbrains.rdserver.startup

import com.intellij.openapi.application.ApplicationInfo
import com.intellij.openapi.util.registry.Registry
import com.intellij.util.io.host
import com.intellij.util.io.origin
import com.intellij.util.io.referrer
import com.jetbrains.codeWithMe.lobby.SessionInfoRequest
import com.jetbrains.rd.platform.codeWithMe.CodeWithMeClientDownloadsManager
import com.jetbrains.rd.platform.codeWithMe.connection.LobbyServerApi
import com.jetbrains.rd.platform.codeWithMe.users.LobbyUrlSettings
import com.jetbrains.rd.platform.codeWithMe.util.CodeWithMePlatformUtil
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rdserver.joinLinks.ConnectToRemoteFromHostInstallAction
import io.netty.buffer.ByteBufAllocator
import io.netty.buffer.ByteBufUtil
import io.netty.channel.ChannelHandlerContext
import io.netty.handler.codec.http.*
import org.jetbrains.ide.HttpRequestHandler
import org.jetbrains.io.send
import java.net.URI
import java.nio.CharBuffer

class RdHttpStarter : HttpRequestHandler() {

  private val PREFIX = "/codeWithMe"
  private val QUERY_PREFIX = "$PREFIX/queryThinClient"
  private val CONNECT_PREFIX = "$PREFIX/connectThinClient"
  private val VERSION = "4"
  private val SESSION_ID_REGEX = Regex("[a-zA-Z0-9_\\-]{22}")

  override fun isSupported(request: FullHttpRequest): Boolean {
    return Registry.`is`("codewithme.browser.activation.enabled") &&
           request.method() in arrayOf(HttpMethod.GET, HttpMethod.POST) &&
           request.uri().startsWith(PREFIX)
  }

  override fun isOriginAllowed(request: HttpRequest): OriginCheckResult {
    val logger = getLogger<RdHttpStarter>()
    val requestReferer = request.referrer
    logger.debug("isOriginAllowed: HOST:${request.host} ORIGIN:${request.origin} REFERER:$requestReferer")

    val lobbyUrl = LobbyUrlSettings.lobbyUrl.uri
    val allowedOrigin = URI(lobbyUrl.scheme, null, lobbyUrl.host, lobbyUrl.port, null, null, null)
    logger.debug("isOriginAllowed: expected Referer prefix: $lobbyUrl, expected Origin: $allowedOrigin")

    if (allowedOrigin.toASCIIString() != request.origin) {
      logger.warn("isOriginAllowed: accepting only Origin '$allowedOrigin', but got: ${request.origin}")
      return OriginCheckResult.FORBID
    }

    val expectedRefererPrefix = lobbyUrl.toASCIIString() + "/"
    if (requestReferer != null && !requestReferer.startsWith(expectedRefererPrefix)) {
      logger.warn("isOriginAllowed: accepting only Referer with prefix '$expectedRefererPrefix', but got: $requestReferer")
      return OriginCheckResult.FORBID
    }

    return OriginCheckResult.ALLOW
  }

  private fun checkUrl(url: String): Boolean {
    val prefix = LobbyUrlSettings.lobbyUrl.uri.toString() + "/"
    if (!url.startsWith(prefix)) {
      getLogger<RdHttpStarter>().warn(
        "CodeWithMe join link must start with currently configured lobby server to be activated within IDE. " +
        "Configured Lobby Server Url: '${LobbyUrlSettings.lobbyUrl.uri}', join link: '$url'")
      return false
    }

    val uri = try {
      URI(url)
    } catch (t: Throwable) {
      getLogger<RdHttpStarter>().warn("Unable to parse CodeWithMe join link from the browser: $url", t)
      return false
    }

    val sessionId = uri.path.trimStart('/')
    if (!SESSION_ID_REGEX.matches(sessionId)) {
      getLogger<RdHttpStarter>().warn("Session ID part of join link url '$url' does not match regex '${SESSION_ID_REGEX.pattern}': $sessionId")
      return false
    }

    try {
      val (response, _) = LobbyServerApi.sessionInfo(uri, SessionInfoRequest(
        clientBuildNumber = ApplicationInfo.getInstance().build.toString(),
        clientPlatform = CodeWithMePlatformUtil.currentPlatform()
      ))

      if (!CodeWithMeClientDownloadsManager.isTrustedDownloadUrl(response.compatibleClientUrl)) {
        getLogger<RdHttpStarter>().warn("CodeWithMe client url is invalid or does not start with a trusted prefix: ${response.compatibleClientUrl}")
        return false
      }

      if (!CodeWithMeClientDownloadsManager.isTrustedDownloadUrl(response.compatibleJreUrl)) {
        getLogger<RdHttpStarter>().warn("CodeWithMe JRE url is invalid or does not start with a trusted prefix: ${response.compatibleJreUrl}")
        return false
      }
    } catch (t: Throwable) {
      getLogger<RdHttpStarter>().warn("Unable to get session info from join link '$url': ${t.message}", t)
      return false
    }

    return true
  }

  override fun process(urlDecoder: QueryStringDecoder, request: FullHttpRequest, context: ChannelHandlerContext): Boolean {
    if (request.uri().startsWith(QUERY_PREFIX)) {
      val url = request.content().toString(Charsets.UTF_8)
      getLogger<RdHttpStarter>().info("CodeWithMe connection query for url '$url'")

      val checkResult = checkUrl(url)
      if (!checkResult) {
        return false
      }

      val response = DefaultFullHttpResponse(
        HttpVersion.HTTP_1_1, HttpResponseStatus.OK,
        ByteBufUtil.encodeString(ByteBufAllocator.DEFAULT, CharBuffer.wrap(VERSION), Charsets.UTF_8))
      response.headers().set(HttpHeaderNames.CONTENT_TYPE, "text/plain")
      response.send(context.channel(), request)
      return true
    }

    if (request.uri().startsWith(CONNECT_PREFIX)) {
      val url = request.content().toString(Charsets.UTF_8)
      getLogger<RdHttpStarter>().info("CodeWithMe connection request to '$url'")

      val checkResult = checkUrl(url)
      if (!checkResult) {
        return false
      }

      ConnectToRemoteFromHostInstallAction().connectToUserProvidedUrl(url)
      HttpResponseStatus.OK.send(context.channel(), request)
      return true
    }

    return false
  }
}