// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.startup

import com.intellij.openapi.diagnostic.Logger
import com.intellij.util.ExceptionUtil
import com.intellij.util.ui.UIUtil
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.ide.model.loggers.LogEntry
import com.jetbrains.ide.model.loggers.LogEntryType
import com.jetbrains.ide.model.loggers.LoggerModel
import com.jetbrains.rd.util.lifetime.Lifetime
import org.apache.log4j.Level
import org.apache.log4j.LogManager
import java.util.*

class RdLogger(private val loggerModel: LoggerModel, private val ideaLogger: Logger) : Logger() {
  override fun isDebugEnabled(): Boolean {
    return ideaLogger.isDebugEnabled
  }

  override fun debug(message: String) {
    ideaLogger.debug(message)
    sendToFrontend(message, null, LogEntryType.VERBOSE)
  }

  override fun debug(t: Throwable?) {
    ideaLogger.debug(t)
    sendToFrontend(null, t, LogEntryType.VERBOSE)
  }

  override fun debug(message: String?, t: Throwable?) {
    ideaLogger.debug(message, t)
    sendToFrontend(message, t, LogEntryType.VERBOSE)
  }

  override fun info(message: String?) {
    ideaLogger.info(message)
    sendToFrontend(message, null, LogEntryType.INFO)
  }

  override fun info(message: String?, t: Throwable?) {
    ideaLogger.info(message, t)
    sendToFrontend(message, t, LogEntryType.INFO)
  }

  override fun warn(message: String?, t: Throwable?) {
    ideaLogger.warn(message, t)
    sendToFrontend(message, t, LogEntryType.WARN)
  }

  override fun error(message: String?, t: Throwable?, vararg details: String) {
    ideaLogger.error(message, t, *details)
    sendToFrontend(message + details.joinToString("\n"), t, LogEntryType.ERROR)
  }

  override fun setLevel(level: Level) {
    ideaLogger.setLevel(level)
  }

  private fun sendToFrontend(message: String?, t: Throwable?, severity: LogEntryType) {
    UIUtil.invokeLaterIfNeeded {
      loggerModel.message.fire(LogEntry(
        message ?: "",
        t?.let { ExceptionUtil.getThrowableText(it) } ?: "",
        severity)
      )
    }
  }
}

class RdLoggerFactory(protocol: IProtocol, private val ideaLoggerFactory: Logger.Factory) : Logger.Factory {
  private val loggerModel = LoggerModel.create(Lifetime.Eternal, protocol)
  private val categoriesOldLevels = HashMap<String, Level>()

  init {
    loggerModel.traceCategories.view(Lifetime.Eternal) { _, categories ->
      for ((category, level) in categoriesOldLevels.entries) {
        LogManager.getLogger(category).level = level
      }
      categoriesOldLevels.clear()

      for (category in categories) {
        val logger = LogManager.getLogger(category)
        categoriesOldLevels[category] = logger.effectiveLevel
        logger.level = Level.TRACE
      }
    }
  }

  override fun getLoggerInstance(category: String): Logger {
    return RdLogger(loggerModel, ideaLoggerFactory.getLoggerInstance(category))
  }

  companion object {
    fun install(protocol: IProtocol) {
      val ideaLoggerFactory = Logger.getFactory()
      Logger.setFactory(RdLoggerFactory(protocol, ideaLoggerFactory))
    }
  }
}
