package com.jetbrains.rdserver.startup

import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isValid
import com.intellij.codeWithMe.ClientId.Companion.toDisposable
import com.intellij.icons.AllIcons
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.openapi.command.CommandListener
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.extensions.ExtensionPointName
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.SystemInfo
import com.intellij.ui.mac.foundation.MacUtil
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.pico.DefaultPicoContainer
import com.jetbrains.codeWithMe.lobby.ClientConnectedRequest
import com.jetbrains.codeWithMe.lobby.ClientDisconnectedRequest
import com.jetbrains.codeWithMe.lobby.CodeWithMeTransport
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.ide.model.uiautomation.UIAutomationInteractionModel
import com.jetbrains.rd.framework.IdKind
import com.jetbrains.rd.framework.Identities
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.framework.Serializers
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.CwmNotifications.CWM_NOTIFICATION_GROUP_ID
import com.jetbrains.rd.platform.codeWithMe.SessionInfoProvider
import com.jetbrains.rd.platform.codeWithMe.connection.CwmConnectionStatusService
import com.jetbrains.rd.platform.codeWithMe.connection.LobbyServerApi
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionState
import com.jetbrains.rd.platform.codeWithMe.protocol.SecureWireOverStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.StreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.WireConnectStatus
import com.jetbrains.rd.platform.codeWithMe.connection.*
import com.jetbrains.rd.platform.codeWithMe.protocol.*
import com.jetbrains.rd.platform.codeWithMe.statistics.CodeWithMeLifecycleLogger
import com.jetbrains.rd.platform.codeWithMe.ui.WireStatsRegistryService
import com.jetbrains.rd.platform.codeWithMe.users.LobbyUrlSettings
import com.jetbrains.rd.platform.codeWithMe.users.broadcastToAllGuests
import com.jetbrains.rd.platform.codeWithMe.users.isHost
import com.jetbrains.rd.platform.util.SelfSignedCertificate
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.ConcurrentHashMap
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.first
import com.jetbrains.rd.util.lifetime.*
import com.jetbrains.rd.util.reactive.IScheduler
import com.jetbrains.rd.util.reactive.OptProperty
import com.jetbrains.rd.util.reactive.adviseOnce
import com.jetbrains.rdserver.BackendUserSettings
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.connection.CwmClientConnectionGroup
import com.jetbrains.rdserver.editors.RdCommandListener
import com.jetbrains.rdserver.extension.RdServerComponentExtensionPoint
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.joinLinks.RdServerJoinLinkManager
import com.jetbrains.rdserver.joinLinks.RdServerLobbySessionManager
import com.jetbrains.rdserver.joinLinks.permanent.CwmPermanentLinkSettings
import com.jetbrains.rdserver.permissions.FileAccessLevel
import com.jetbrains.rdserver.permissions.RunAccessLevel
import com.jetbrains.rdserver.permissions.TerminalAccessLevel
import com.jetbrains.rdserver.permissions.ToolWindowAccessLevel
import com.jetbrains.rdserver.permissions.impl.MutablePermissionModel
import com.jetbrains.rdserver.permissions.impl.PermissionManagerImpl
import com.jetbrains.rdserver.unattendedHost.UnattendedHostControlSocketManager
import com.jetbrains.rdserver.unattendedHost.UnattendedHostManager
import com.jetbrains.rdserver.unattendedHost.UnattendedHostProjectService
import it.unimi.dsi.fastutil.ints.IntOpenHashSet
import java.net.URI
import java.security.SecureRandom
import java.time.Instant
import java.time.LocalTime
import java.time.temporal.ChronoUnit
import java.util.concurrent.TimeUnit
import java.util.function.BiConsumer

@Service
class RdServerConnectionManager(val project: Project) {
  companion object {
    fun getInstance(project: Project) = project.service<RdServerConnectionManager>()

    val LOG = logger<RdServerConnectionManager>()

    private val EP_NAME = ExtensionPointName<RdServerComponentExtensionPoint>("com.intellij.rdserver.component")
  }

  private val clientConnectionLifetimeDefs = HashMap<ClientId, LifetimeDefinition>()
  private val autoAcceptedClients = HashMap<ClientId, AutoAcceptedClient>()
  private val sessionStartTimeMap = HashMap<String, Instant>()
  private val sessionParticipantsMaxCountMap = HashMap<String, Int>()
  private val sessionDisconnectsCountMap = HashMap<String, Int>()

  private val pinsLock = Any()
  private val clientCachedPins = HashMap<String, Int>()
  private val usedPins = IntOpenHashSet()
  private val pinsRandom = SecureRandom()

  private fun getPinFor(clientRootFingerprint: String): Int {
    synchronized(pinsLock) {
      val pin = clientCachedPins[clientRootFingerprint]
      if (pin != null)
        return pin

      var newPin = pinsRandom.nextInt(10_000)
      if (usedPins.size > 5_000) error("Too many connections, please restart CWM session")
      while(!usedPins.add(newPin)) {
        newPin = pinsRandom.nextInt(10_000)
      }

      clientCachedPins[clientRootFingerprint] = newPin

      return newPin
    }
  }

  // stored pin should be removed once the client was accepted or declined
  private fun purgeStoredPin(clientRootFingerprint: String) {
    synchronized(pinsLock) {
      clientCachedPins.remove(clientRootFingerprint)
    }
  }

  fun getPinString(clientRootFingerprint: String): String {
    val clientPin = synchronized(pinsLock) {
      clientCachedPins[clientRootFingerprint]
    } ?: error("Client is not currently in connection process")
    val bytes = ByteArray(4)
    pinToBytes(clientPin, bytes)
    return bytes.joinToString(" ") { it.toString() }
  }

  private fun pinToBytes(pinNumber: Int, pinBytes: ByteArray) {
    assert(pinNumber in 0..9999) { "Pins should be between 0 and 9999, got $pinNumber" }

    pinBytes[0] = (pinNumber % 10).toByte()
    pinBytes[1] = (pinNumber / 10 % 10).toByte()
    pinBytes[2] = (pinNumber / 100 % 10).toByte()
    pinBytes[3] = (pinNumber / 1000 % 10).toByte()
  }

  // todo replace its usages with per-client services/ClientId.isValid checks
  fun getClientLifetime(clientId: ClientId): Lifetime {
    val lifetime = clientConnectionLifetimeDefs[clientId] ?: return Lifetime.Terminated
    return lifetime.createNestedWith(clientId)
  }

  fun disconnectClient(clientId: ClientId, disconnectReason: GuestTerminationReasonsData?) {
    if (disconnectReason != null) {
      ClientId.withClientId(clientId) {
        CwmConnectionStatusService.getInstance(project).terminated.fire(disconnectReason)
      }
    }

    clientConnectionLifetimeDefs.remove(clientId)?.terminate()
    autoAcceptedClients.remove(clientId)
  }

  fun terminateAllTemporaryListeners(disconnectReason: GuestTerminationReasonsData) {
    broadcastToAllGuests(project) {
      CwmConnectionStatusService.getInstance(project).terminated.fire(disconnectReason)
    }

    autoAcceptedClients.clear()

    RdServerJoinLinkManager.getInstance(project).terminateJoinLinks()

    synchronized(pinsLock) {
      usedPins.clear()
      clientCachedPins.clear()
    }
  }

  private fun <T>clearStatisticsEntriesMap(sessionId: String, map: HashMap<String, T>) {
    if (!map.containsKey(sessionId)) return
    map.remove(sessionId)
  }

  fun getSessionStartTime(sessionId: String): Instant? = sessionStartTimeMap[sessionId]
  fun clearSessionStartTimeMap(sessionId: String) = clearStatisticsEntriesMap(sessionId, sessionStartTimeMap)

  fun getSessionParticipantsMax(sessionId: String): Int? = sessionParticipantsMaxCountMap[sessionId]
  fun clearSessionParticipantsMaxCountMap(sessionId: String) = clearStatisticsEntriesMap(sessionId, sessionParticipantsMaxCountMap)

  fun getSessionDisconnectsCount(sessionId: String): Int? = sessionDisconnectsCountMap[sessionId]
  fun clearSessionDisconnectsCountMap(sessionId: String) = clearStatisticsEntriesMap(sessionId, sessionDisconnectsCountMap)

  private fun attachServerToWire(wireLifetimeDef: LifetimeDefinition, wire: SecureWireOverStreamTransport,
                                 clientData: SecureClientData, sessionUrl: URI?,
                                 transport: CodeWithMeTransport,
                                 scheduler: IScheduler) {
    ClientId.propagateAcrossThreads = true
    val clientId = clientData.clientId

    // it's important to terminate all the activities using the same ClientId they were initiated with.
    val clientLifetime = wireLifetimeDef.createNestedWith(clientId)

    val protocol = Protocol("CodeWithMe-$wire", Serializers(), Identities(IdKind.Server), scheduler, wire, clientLifetime)

    wire.assetNullClientIdOnReceive = true
    project.service<WireStatsRegistryService>().registerWire(clientLifetime, wire)

    clientConnectionLifetimeDefs.remove(clientId)?.let {
      if (it.isAlive) {
        LOG.info("Client $clientId joined before the previous client with the same ID was cleaned up, forcing cleanup now")
        it.terminate()
      }
    }
    if (clientLifetime.isNotAlive) {
      LOG.error("Trying to attach a server to a terminated wire. ClientId $clientId")
      return
    }
    clientConnectionLifetimeDefs[clientId] = wireLifetimeDef

    LOG.assertTrue(!clientId.isValid) { "ClientId $clientId is valid after cleanup" }
    LOG.info("New remote client, wire id ${wire}, ClientId $clientId")
    Disposer.register(clientLifetime.createNestedDisposable(), clientId.toDisposable())
    if (!clientId.isValid) {
      LOG.error("Registration of a disposable for ClientId $clientId failed")
      Disposer.dispose(clientId.toDisposable())
      Disposer.register(clientLifetime.createNestedDisposable(), clientId.toDisposable())
    }

    RdServerApplicationService.getInstance().registerClient(clientLifetime, clientId, project)
    FollowMeManager.getInstance(project).registerUser(clientLifetime, clientId, clientData.userName)

    if (!clientId.isValid) {
      LOG.error("Trying to handleNewClient for an invalid clientId. $clientId")
    }
    handleNewClient(clientLifetime, protocol, clientData, sessionUrl, transport)

    wire.startDeliveringMessages()
  }

  private fun collectParticipantsStatistics(sessionId: String?) {
    sessionId ?: return

    val usersCount = FollowMeManager.getInstance(project).users.size
    if (sessionParticipantsMaxCountMap.containsKey(sessionId)) {
      val currentCount = sessionParticipantsMaxCountMap[sessionId]
      if (currentCount != null && currentCount >= usersCount)
        return
    }

    sessionParticipantsMaxCountMap[sessionId] = usersCount
  }

  private fun handleNewClient(
    lifetime: Lifetime,
    protocol: Protocol,
    clientData: SecureClientData,
    sessionUrl: URI?,
    transport: CodeWithMeTransport
  ) {
    //            RdLoggerFactory.install(protocol) // TODO: protocol logger factory that can live in multi-client environment
    //            BackendTestHost(protocol).attach() // TODO: test host, adapt for Rider, rewrite for per-client?
    val currentClientId = clientData.clientId
    val uiInteractionModel = UIAutomationInteractionModel.create(lifetime, protocol)
    application.messageBus.connect(lifetime.createNestedDisposable()).subscribe(CommandListener.TOPIC, RdCommandListener())

    val solutionModel = SolutionModel.create(lifetime, protocol)
    LOG.info("Attached solution model to protocol")
    collectParticipantsStatistics(clientData.connectionSessionId)

    ClientId.withClientId(currentClientId) {
      val projectProp = openSolutionCallback(lifetime, solutionModel, clientData, sessionUrl, transport)

      projectProp.advise(lifetime) { project ->
        protocol.outOfSyncModels.adviseOnce(lifetime) {
          ClientId.withClientId(ClientId.localId) {
            Notification(CWM_NOTIFICATION_GROUP_ID,
                         CodeWithMePlatformBundle.message("product.name"),
                         CodeWithMePluginBundle.message("notification.version.mismatch", FollowMeManager.getInstance(project).getUser(
                           currentClientId).name),
                         NotificationType.ERROR).notify(project)
          }
        }
        ClientId.withClientId(currentClientId) {
          val projectService = UnattendedHostProjectService.getInstance(project)
          val joinedViaToken = projectService.wasClientJoinedViaToken(clientData.clientId)
                               || projectService.hasClientUsedValidTokenPreviously(clientData)
                               || UnattendedHostManager.getInstance().wasClientJoinedViaPassword(clientData.clientId)
          if (joinedViaToken)
            UnattendedHostManager.getInstance().setControllerClientId(lifetime, clientData.clientId)

          val solution = solutionModel.solutions.first().value

          val parentContainer = project.picoContainer as DefaultPicoContainer
          val container = DefaultPicoContainer(parentContainer).apply {
            registerComponentInstance(solution)
            registerComponentInstance(protocol)
            registerComponentInstance(lifetime)
            registerComponentInstance(currentClientId)
            registerComponentInstance(uiInteractionModel)
          }

          RdServerPerClientComponents.getInstance(project).registerServices(lifetime, currentClientId, container)

          EP_NAME.processWithPluginDescriptor(BiConsumer { it, pluginDescriptor ->
            val baseClass = it.interfaceClass
            if (baseClass != null) {
              container.registerComponentImplementation(baseClass, it.getExtensionClass(pluginDescriptor))
            }
            else {
              container.registerComponentImplementation(it.getExtensionClass(pluginDescriptor))
            }
          })
          container.componentAdapters.forEach {
            runCatching { it.getComponentInstance(container) }
              .onFailure { exception -> LOG.error(exception) }
          }

          solution.isLoaded.set(true)
          solution.solutionLifecycle.isReadyForTyping.set(true)
          solution.solutionLifecycle.fullStartupFinished.set(LoadResult(0, false))

          if (SystemInfo.isMac) {
            lifetime.executeIfAlive {
              val wakeUp = MacUtil.wakeUpNeo("Guest ${currentClientId.value} is connected")
              lifetime.onTermination { wakeUp.run() }
            }
          }

          if (joinedViaToken) {
            PermissionManagerImpl.getInstanceImpl(project).applyModel(clientData.clientId, MutablePermissionModel(mutableListOf(), FileAccessLevel.FullAccess, TerminalAccessLevel.FullAccess, RunAccessLevel.FullAccess, ToolWindowAccessLevel.FullAccess))
            UnattendedHostControlSocketManager.getInstance().createControlModelOnProtocol(lifetime, protocol)
          } else if (clientData.connectionSessionId != null && CwmPermanentLinkSettings.getInstance(project).sessionId.value == clientData.connectionSessionId) {
            // Grant full access to users joining off permalinks
            PermissionManagerImpl.getInstanceImpl(project).applyModel(clientData.clientId, MutablePermissionModel(mutableListOf(), FileAccessLevel.FullAccess, TerminalAccessLevel.FullAccess, RunAccessLevel.FullAccess, ToolWindowAccessLevel.FullAccess))
          }
        }
      }
    }
  }

  private fun openSolutionCallback(solutionModelLifetime: Lifetime,
                                   solutionModel: SolutionModel,
                                   clientData: SecureClientData,
                                   sessionUrl: URI?,
                                   transport: CodeWithMeTransport): OptProperty<Project> {

    val rdCodeWithMeProject = RdCodeWithMeProject(project.name, BackendUserSettings.getHostUserName())
    val solution = Solution(RdOpenSolution(rdCodeWithMeProject, false))
    solutionModel.solutions.addUnique(solutionModelLifetime, 0, solution)
    RdServerPerClientSolutions.putPerClientIdSolution(project, solutionModelLifetime, clientData.clientId, solution)

    val startTimeMs = System.currentTimeMillis()
    val user = FollowMeManager.getInstance(project).getUser(clientData.clientId)

    sendConnectStatistics(project = project, user = user, transport = transport)
    sendSessionStartedStatistics(project = project, sessionId = clientData.connectionSessionId)

    if (sessionUrl != null) {
      application.executeOnPooledThread {
        // TODO Retry it?
        try {
          LobbyServerApi.clientConnected(sessionUrl, ClientConnectedRequest(
            transport = transport,
            connectionClientId = clientData.clientAnonymizedInstallId,
          ))
        }
        catch (t: Throwable) {
          LOG.warn("Unable to send clientConnected event to lobby", t)
        }
      }

      solutionModelLifetime.onTermination {
        application.executeOnPooledThread {
          // TODO Retry it?
          try {
            LobbyServerApi.clientDisconnected(sessionUrl, ClientDisconnectedRequest(
              transport = transport,
              connectionClientId = clientData.clientAnonymizedInstallId,
              connectionTimeMs = System.currentTimeMillis() - startTimeMs,
              disconnectReason = "Unknown"
            ))
          }
          catch (t: Throwable) {
            LOG.warn("Unable to send clientConnected event to lobby", t)
          }
        }
      }
    }

    val userName = user.name
    val sessionId = SessionInfoProvider.getInstance(project).getSessionId() ?: ""

    ClientId.withClientId(ClientId.localId) {
      Notification(
        CWM_NOTIFICATION_GROUP_ID,
        AllIcons.CodeWithMe.CwmAccess,
        CodeWithMePlatformBundle.message("product.name"), null,
        CodeWithMePluginBundle.message("notification.user.joined", userName),
        NotificationType.INFORMATION, null
      ).notify(project)
    }

    solutionModelLifetime.onTermination {
      ClientId.withClientId(ClientId.localId) {
        CodeWithMeLifecycleLogger.onClientFinished(project, user, sessionId)
        Notification(
          CWM_NOTIFICATION_GROUP_ID,
          AllIcons.CodeWithMe.CwmAccess,
          CodeWithMePlatformBundle.message("product.name"), null,
          CodeWithMePluginBundle.message("notification.user.left", userName),
          NotificationType.INFORMATION, null
        ).notify(project)
      }
    }
    return OptProperty(project)
  }

  private fun sendConnectStatistics(project: Project, user: RdUserModel, transport: CodeWithMeTransport) {
    CodeWithMeLifecycleLogger.onClientConnected(project, user, transport)
  }

  private fun sendSessionStartedStatistics(project: Project, sessionId: String?) {
    sessionId ?: return
    val users = FollowMeManager.getInstance(project).users
    if (sessionStartTimeMap.containsKey(sessionId) || users.size > 2 || users.none { !it.value.isHost })
      return

    sessionStartTimeMap[sessionId] = Instant.now()
    sessionParticipantsMaxCountMap[sessionId] = users.size
    sessionDisconnectsCountMap[sessionId] = 0
    CodeWithMeLifecycleLogger.onSessionStarted(project, sessionId)
  }

  private val clientWireGroups = ConcurrentHashMap<String, CwmClientConnectionGroup>()
  private fun getWireGroup(data: SecureClientData, passwordAuthHandler: RdServerPasswordAuthHandler?): CwmClientConnectionGroup {
    return clientWireGroups.compute(data.clientFingerprint) { _, existingGroup ->
      if (existingGroup == null || existingGroup.data.issuedOn < data.issuedOn) {
        LOG.debug("New connection will override old group ${data.userName} (old ${existingGroup?.data?.issuedOn} new ${data.issuedOn}")
        existingGroup?.terminateLifetime()
        return@compute CwmClientConnectionGroup(project, data, passwordAuthHandler != null)
      }
      return@compute existingGroup
    } ?: error("Failed to register wire group for client ${data.userName} - ${data.clientFingerprint}")
  }

  internal fun removeGroup(group: CwmClientConnectionGroup) {
    LOG.debug("Removing group ${group.id} (${group.data.issuedOn})")
    clientWireGroups.computeIfPresent(group.groupKey) { _, oldGroup ->
      if(oldGroup === group) null else oldGroup
    }
  }

  internal fun shouldAutoAccept(data: SecureClientData) = autoAcceptedClients[data.clientId]?.autoConnect() == true

  private fun incrementSessionDisconnects() {
    val sessionId = SessionInfoProvider.getInstance(project).getSessionId() ?: return

    if (!sessionDisconnectsCountMap.containsKey(sessionId))
      sessionDisconnectsCountMap[sessionId] = 0

    val currentCount = sessionDisconnectsCountMap[sessionId] ?: 0
    sessionDisconnectsCountMap[sessionId] = currentCount + 1
  }

  fun createServerOnTransport(
    lifetime: Lifetime,
    transport: StreamTransport,
    scheduler: IScheduler,
    id: String,
    transportType: CodeWithMeTransport,
    autoAcceptConnection: Boolean = false,
    acceptAnySession: Boolean = false
  ) {

    lateinit var data1 : SecureClientData

    val passwordHandler = UnattendedHostManager.getInstance().createPasswordRequest()?.let { RdServerPasswordAuthHandler(it) }

    val wireLifetime = lifetime.createNested()
    val delayedWireLifetime = LifetimeDefinition()
    wireLifetime.onTermination {
      delayedWireLifetime.delayedTerminate()
    }
    val wire = SecureWireOverStreamTransport(
      lifetime = delayedWireLifetime,
      streamTransport = transport,
      scheduler = scheduler,
      id = id,
      server = true,
      passwordHandler = passwordHandler,
      certificateData = { SelfSignedCertificate.permanentCertificateData },
      checkHandshakeTrusted = { list ->
        val data = SecureClientData(list)
        data1 = data

        if (!acceptAnySession) {
          // TODO Assert we have created a session with sessionId
          // TODO Write a test
          val connectionSessionId = data.connectionSessionId
          if (connectionSessionId == null) {
            error("Could not get session id from provided certificate")
          }

          if (!RdServerLobbySessionManager.getInstance(project).isKnownSession(connectionSessionId)) {
            error("SessionId $connectionSessionId is unknown on this host")
          }
        }
      },
      checkBeforeProtocolTrusted = { list ->
        val data = SecureClientData(list)

        val clientFingerprint = data.clientFingerprint
        val userName = data.userName

        require(data.clientId.value.startsWith("${clientFingerprint}-")) {
          error("Connection with client '$userName' ($clientFingerprint) protocol error: clientId does not start with client fingerprint")
        }

        if (autoAcceptConnection) {
          LOG.info("Auto accepted connection with client '$userName'")
          purgeStoredPin(data.clientFingerprint)
          WireConnectStatus.ACCEPTED
        }
        else {
          // TODO Handle timeouts
          val group = getWireGroup(data, passwordHandler)
          val acceptedValue = group.registerWire(this@SecureWireOverStreamTransport, data).get()!!
          purgeStoredPin(data.clientFingerprint)
          when (acceptedValue) {
            WireConnectStatus.ACCEPTED -> {
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id was accepted by user")
            }
            WireConnectStatus.USER_DECLINED -> {
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id was declined by user")
              throw SecureWireOverStreamTransport.PeerVisibleErrorCode(WireConnectStatus.USER_DECLINED, "Declined by user")
            }
            WireConnectStatus.DECLINED -> {
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id was closed in favor of another connection")
              throw SecureWireOverStreamTransport.PeerVisibleErrorCode(WireConnectStatus.DECLINED, "Closed")
            }
            WireConnectStatus.GUEST_LIMIT_REACHED -> {
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id was declined due to guest limit")
              throw SecureWireOverStreamTransport.PeerVisibleErrorCode(WireConnectStatus.GUEST_LIMIT_REACHED, "Guest limit reached")
            }
            WireConnectStatus.JOIN_PIN -> { // todo: separate wire-level service messages from these status codes
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id received a bad accept value from group: $acceptedValue, interpreting as internal error")
              throw SecureWireOverStreamTransport.PeerVisibleErrorCode(WireConnectStatus.INTERNAL_ERROR, "Internal error")
            }
            WireConnectStatus.PASSWORD_REQUIRED -> {
              passwordHandler?.setClientId(data.clientId)
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id requires password")
            }
            WireConnectStatus.INTERNAL_ERROR -> {
              LOG.info("Connection with client '$userName' ($clientFingerprint) on wire $id experienced an internal error")
              throw SecureWireOverStreamTransport.PeerVisibleErrorCode(WireConnectStatus.INTERNAL_ERROR, "Internal error")
            }
          }.let { }

          acceptedValue
        }
      },
      pinHandler = { chain, pinBytes ->
        val clientFingerprint = SecureClientData(chain).clientFingerprint
        val pinNumber = getPinFor(clientFingerprint)
        pinToBytes(pinNumber, pinBytes)
      }
    )

    wire.connectionState.advise(wireLifetime.lifetime) { state ->
      application.invokeLater {
        when (state) {
          is ConnectionState.NOT_CONNECTED -> Unit
          is ConnectionState.CLOSED -> {
            if (wireLifetime.lifetime.isAlive) {
              LOG.info("$id: wire connection was closed, terminating its lifetime")
              wireLifetime.terminate()
            }
            incrementSessionDisconnects()
          }
          is ConnectionState.FAULTED -> {
            if (wireLifetime.lifetime.isAlive) {
              LOG.info("$id: wire connection faulted, terminating its lifetime")
              wireLifetime.terminate()
            }
            return@invokeLater
          }
          is ConnectionState.DECLINED -> {
            if (wireLifetime.lifetime.isAlive) {
              LOG.info("$id: wire connection declined (how?), terminating its lifetime")
              wireLifetime.terminate()
            }
            return@invokeLater
          }
          is ConnectionState.CONNECTED -> {
            if (wireLifetime.lifetime.isNotAlive) {
              LOG.info("$id: wire connection terminated")
              return@invokeLater
            }

            val clientId = data1.clientId
            val sessionId = data1.connectionSessionId

            wire.clientId = clientId

            // Allow this client to reconnect without explicit conformation
            autoAcceptedClients[clientId] = AutoAcceptedClient(wireLifetime)

            // TODO: get here real server url this session where created with
            val sessionUrl = if(sessionId.isNullOrEmpty()) null else URI(LobbyUrlSettings.lobbyUrl.uri.toString().trimEnd('/') + "/$sessionId")

            attachServerToWire(wireLifetime, wire, data1, sessionUrl, transportType, scheduler)
          }
        }.let {  }
      }
    }
  }

  private fun LifetimeDefinition.delayedTerminate() {
    AppExecutorUtil.getAppScheduledExecutorService().schedule({ this@delayedTerminate.terminate() }, 5, TimeUnit.SECONDS)
  }

  private class AutoAcceptedClient(lifetime: Lifetime) {

    private var disconnectedTimeStamp : LocalTime? = null

    init {
      lifetime.onTermination {
        disconnectedTimeStamp = LocalTime.now()
      }
    }

    fun autoConnect(): Boolean {
      return disconnectedTimeStamp == null ||
             ChronoUnit.MINUTES.between(disconnectedTimeStamp, LocalTime.now()) < 5.0
    }
  }
}

// todo questionable extension
/**
 * Creates a nested lifetime that will be terminated under specified ClientId
 */
private fun Lifetime.createNestedWith(clientId: ClientId): Lifetime {
  if (this.isNotAlive) return this
  return LifetimeDefinition().let { nested ->
    this.onTermination {
      ClientId.withClientId(clientId) {
        nested.terminate()
      }
    }
    nested.lifetime
  }
}