package com.jetbrains.rdserver.startup

import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.jetbrains.codeWithMe.lobby.CodeWithMeTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.SocketStreamTransport
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.internal.RdServerOnProjectOpen
import java.net.InetSocketAddress
import java.net.ServerSocket
import java.nio.channels.AsynchronousCloseException
import java.nio.channels.ServerSocketChannel
import kotlin.concurrent.thread

@Service
class RdServerDirectConnectionsListener(val project: Project) {
  companion object {
    fun getInstance(project: Project) = project.service<RdServerDirectConnectionsListener>()

    private val LOG = Logger.getInstance(RdServerDirectConnectionsListener::class.java)
  }

  // the same start port as vscode live share, easier for users
  private val START_PORT = 5990

  private fun findFistAvailablePort(): Int {
    val portRange = START_PORT..65535

    var lastThrowable: Throwable? = null
    for (port in portRange) {
      try {
        ServerSocket(port).close()
        return port
      } catch (t: Throwable) {
        LOG.debug("Unable to bind to port $port, trying the next one", t)
        lastThrowable = t
      }
    }

    throw IllegalStateException("Unable to find free port in entire port range $portRange", lastThrowable)
  }

  /**
   * Starts a new direct connection listener.
   * @return port of newly started listener
   */
  fun startListener(lifetime: Lifetime, port: Int? = null, autoAcceptConnection: Boolean = false): Int {
    val portToBind = port ?: findFistAvailablePort()
    return startRdServer(lifetime, portToBind, project, autoAcceptConnection)
  }

  private fun startRdServer(
    lifetime: Lifetime,
    port: Int,
    project: Project,
    autoAcceptConnection: Boolean
  ): Int {
    LOG.info("Preparing to listen for connections" + if (port > 0) " on port $port" else "")

    val serverSocketChannel = ServerSocketChannel.open().bind(InetSocketAddress(port))
    val localPort = (serverSocketChannel.localAddress as InetSocketAddress).port
    lifetime.onTermination { serverSocketChannel.close() }

    val listeningThread = thread(name = "RdServer", isDaemon = true) {
      LOG.info("Waiting for connections on port $localPort")

      while (lifetime.isAlive && serverSocketChannel.isOpen) {
        try {
          val socketChannel = serverSocketChannel.accept()
          socketChannel.socket().tcpNoDelay = true

          LOG.info("Accepted connection from ${socketChannel.remoteAddress}")

          application.invokeLater {
            val id = "Tcp:${socketChannel.remoteAddress}"

            val transport = SocketStreamTransport(socketChannel, id)
            if (autoAcceptConnection) {
              RdServerConnectionManager.LOG.info("Auto-accepting connection from socket transport because ${RdServerOnProjectOpen.listenOnProjectOpenPortProperty} system property is specified")
            }
            RdServerConnectionManager.getInstance(project).createServerOnTransport(
              lifetime, transport, EdtScheduler, id, CodeWithMeTransport.DirectTcp,
              autoAcceptConnection, autoAcceptConnection)
          }
        }
        catch (_: InterruptedException) {
          break
        }
        catch (_: AsynchronousCloseException) {
          break
        }
        catch (t: Throwable) {
          LOG.warn("Error accepting connection at port $localPort: ${t.message}", t)
        }
      }

      LOG.info("Listening on rd connections on port $localPort is closed")
    }

    lifetime.onTermination {
      listeningThread.interrupt()
      listeningThread.join(500)
      if (listeningThread.isAlive) {
        LOG.warn("Listening thread is still alive")
      }
    }

    return localPort
  }
}