package com.jetbrains.rdserver.startup

import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.concurrency.EdtExecutorService
import com.jetbrains.codeWithMe.lobby.CodeWithMeTransport
import com.jetbrains.codeWithMe.lobby.CreateSessionResponse
import com.jetbrains.rd.platform.codeWithMe.protocol.WebSocketStreamTransport
import com.jetbrains.rd.platform.util.NetworkUtils
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.closeNormal
import com.jetbrains.rd.platform.util.delayFuture
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.onTermination
import java.net.URI
import java.net.http.WebSocket
import java.nio.ByteBuffer
import java.time.Duration
import java.util.concurrent.CompletionStage
import java.util.concurrent.TimeUnit
import java.util.concurrent.atomic.AtomicInteger

@Service
class RdServerRelayedConnectionsListener(val project: Project) {
  companion object {
    private const val RelayServerRegisteredMessage = "________registered"

    fun getInstance(project: Project) = project.service<RdServerRelayedConnectionsListener>()
    private val logger = Logger.getInstance(RdServerRelayedConnectionsListener::class.java)
  }

  fun attachToSession(sessionResponse: CreateSessionResponse, sessionLifetime: Lifetime, autoAcceptConnections: Boolean = false, onRegistered: () -> Unit = {}) {
    assert(!application.isDispatchThread) { "Must not be executed on ui thread" }

    // TODO Handle lobby errors and report it to the user
    val wsRelay = sessionResponse.wsRelay ?: error("Could not find relay in session data")

    ControlWebSocketClient(
      lifetime = sessionLifetime,
      httpHeaders = wsRelay.httpHeaders,
      relayControlChannelUrl = wsRelay.controlChannelEndpoint,
      relayAcceptChannelUrlPrefix = wsRelay.acceptChannelEndpointPrefix,
      project = project,
      onRegistered = onRegistered,
      autoAcceptConnections = autoAcceptConnections
    )
  }

  private class ControlWebSocketClient(
    private val lifetime: Lifetime,
    private val httpHeaders: Map<String, String>,
    private val onRegistered: () -> Unit,
    private val relayControlChannelUrl: URI,
    private val relayAcceptChannelUrlPrefix: URI,
    private val project: Project,
    private val autoAcceptConnections: Boolean = false
  ) {

    private var registered: Boolean = false
    private val currentWebsocketCounter = AtomicInteger()

    init {
      createWebsocket(0)
    }

    private fun createWebsocket(targetCount: Int) {
      logger.info("Registering on relay server at $relayControlChannelUrl")

      if (!currentWebsocketCounter.compareAndSet(targetCount, targetCount + 1))
        return

      registered = false

      val newTargetCount = targetCount + 1

      val httpClient = NetworkUtils.createHttpClientWithGoodDefaults()

      val wsBuilder = httpClient.newWebSocketBuilder()

      for ((headerName, headerValue) in httpHeaders) {
        wsBuilder.header(headerName, headerValue)
      }

      val websocketLifetimeDef = lifetime.createNested()

      var pingsWithoutPong = 0

      val ws = wsBuilder
        .buildAsync(relayControlChannelUrl, object : WebSocket.Listener {
          override fun onOpen(webSocket: WebSocket) {
            application.invokeLater {
              logger.info("Connected to management cwm websocket at $relayControlChannelUrl")

              val pingMessage = "ping".toByteArray()
              val future = EdtExecutorService.getScheduledExecutorInstance().scheduleWithFixedDelay(
                { webSocket.sendPing(ByteBuffer.wrap(pingMessage)) }, 1, 5, TimeUnit.SECONDS)
              lifetime.onTermination {
                future.cancel(true)
              }
            }

            super.onOpen(webSocket)
          }

          override fun onPong(webSocket: WebSocket?, message: ByteBuffer?): CompletionStage<*>? {
            pingsWithoutPong = 0

            return super.onPong(webSocket, message)
          }

          override fun onText(webSocket: WebSocket, data: CharSequence, last: Boolean): CompletionStage<*>? {
            require(last)

            // Copy to string since `data` is available only inside onText and probably the buffer will be reused later
            val message = data.toString()

            application.invokeLater {
              logger.info("Received message on control connection: $message")

              if (message == RelayServerRegisteredMessage) {
                logger.info("Received we've successfully registered at relay at $relayControlChannelUrl")

                if (registered) {
                  logger.error("Got registered message when we're already have a join link at $relayControlChannelUrl")
                }
                else {
                  registered = true
                  onRegistered()
                }

                return@invokeLater
              }

              if (!registered) {
                logger.error("Got knock-knock message while we're not registered yet at $relayControlChannelUrl")
                return@invokeLater
              }

              // TODO Validate it's not a garbage
              @Suppress("UnnecessaryVariable") val clientId = message

              val acceptUrl = "${relayAcceptChannelUrlPrefix}/$clientId"
              logger.info("Creating new connection with client '$clientId', relay accept url = $acceptUrl")

              val scheduler = EdtScheduler
              val connectionId = "WebSocket-$clientId"
              val transport = WebSocketStreamTransport(lifetime, URI(acceptUrl), httpHeaders, connectionId)
              RdServerConnectionManager.getInstance(project).createServerOnTransport(
                lifetime, transport, scheduler, connectionId, CodeWithMeTransport.WsRelay,
                autoAcceptConnections)
            }

            return super.onText(webSocket, data, last)
          }

          override fun onBinary(webSocket: WebSocket, data: ByteBuffer, last: Boolean): CompletionStage<*>? {
            val error = "Received binary frame on control websocket '$relayControlChannelUrl', this is unsupported"

            if (application.isEAP || application.isInternal || application.isUnitTestMode) {
              logger.error(error)
            } else {
              logger.warn(error)
            }

            return super.onBinary(webSocket, data, last)
          }

          override fun onClose(webSocket: WebSocket, statusCode: Int, reason: String): CompletionStage<*>? {
            websocketLifetimeDef.terminate()

            application.invokeLater {
              logger.info("Management cwm websocket closed ($relayControlChannelUrl) code: $statusCode reason: $reason")

              handleClosedWebsocket(newTargetCount)
            }

            return super.onClose(webSocket, statusCode, reason)
          }

          override fun onError(webSocket: WebSocket, ex: Throwable?) {
            websocketLifetimeDef.terminate()

            logger.warn("Error in control websocket '$relayControlChannelUrl'", ex)

            handleClosedWebsocket(newTargetCount)

            super.onError(webSocket, ex)
          }
        })

      ws.handleAsync { webSocket, exception ->
        if (exception != null) {
          logger.warn("Error creating websocket '$relayControlChannelUrl'", exception)
          handleClosedWebsocket(newTargetCount)
        }

        val noTimeoutsEnvVar = System.getenv("CWM_NO_TIMEOUTS")
        if (!(noTimeoutsEnvVar.isNullOrEmpty() || noTimeoutsEnvVar == "0"))
          return@handleAsync

        if (webSocket != null) {
          val future = AppExecutorUtil.getAppScheduledExecutorService().scheduleWithFixedDelay(
            {
              if (!websocketLifetimeDef.isAlive) return@scheduleWithFixedDelay
              webSocket.sendPing(ByteBuffer.wrap(byteArrayOf()))

              if (pingsWithoutPong++ > 10) {
                logger.info("Control websocket spent 10 pings without a pong, reconnecting")
                handleClosedWebsocket(newTargetCount)
                webSocket.abort()
                websocketLifetimeDef.terminate()
              }
            }, 1, 1, TimeUnit.SECONDS)

          if (!websocketLifetimeDef.onTerminationIfAlive { future.cancel(false) })
            future.cancel(false)
        }
      }

      lifetime.onTermination {
        logger.info("CWM session lifetime terminated, closing management websocket $relayControlChannelUrl")
        ws.closeNormal()
      }
    }

    private fun handleClosedWebsocket(targetCount: Int) {
      if (!lifetime.isAlive || currentWebsocketCounter.get() != targetCount) return

      logger.debug("Management websocket connection closed, will retry connection in 2 seconds")

      delayFuture(Duration.ofSeconds(2), lifetime).thenApply {
        createWebsocket(targetCount)
      }
    }
  }
}