// Copyright 2000-2020 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.startup

import com.intellij.codeWithMe.ClientId
import com.jetbrains.rd.platform.RdConstants
import com.jetbrains.rd.platform.util.sha3Fingerprint
import org.bouncycastle.asn1.ASN1ObjectIdentifier
import org.bouncycastle.asn1.x500.X500Name
import org.bouncycastle.asn1.x500.style.BCStyle
import org.bouncycastle.asn1.x500.style.IETFUtils
import java.security.cert.X509Certificate
import java.util.*

class SecureClientData(certificates: List<X509Certificate>) {
  val connectionSessionId: String?
  val userName: String
  val clientId: ClientId
  val clientAnonymizedInstallId: String
  val clientFingerprint: String
  val singleUseToken: String?
  val issuedOn: Date

  companion object {
    private val notAllowedCharactersRegex = Regex("[^a-zA-Z0-9\\-_:\\s.,/+]")
  }

  init {
    require(certificates.size == 2) {
      "Exactly two certificates must be in guest certification chain, but got ${certificates.size}: " +
      certificates.joinToString { it.subjectX500Principal.toString() }
    }

    val (sessionCertificate, permanentCertificate) = when {
      certificates[0].subjectX500Principal == certificates[1].issuerX500Principal -> certificates[1] to certificates[0]
      certificates[0].issuerX500Principal == certificates[1].subjectX500Principal -> certificates[0] to certificates[1]
      else -> error("Certificates in guest certification chain look unrelated")
    }

    val connectionData = X500Name.getInstance(sessionCertificate.subjectX500Principal.encoded)

    issuedOn = sessionCertificate.notBefore

    connectionSessionId = connectionData.getComponent(BCStyle.SERIALNUMBER)?.let { sanitizeString(it) }
    userName = connectionData.getComponent(BCStyle.CN)?.let { sanitizeUserVisibleString(it) }
               ?: error("There is no UserName in certificate")

    clientAnonymizedInstallId = connectionData.getComponent(BCStyle.T)?.let { sanitizeString(it) }
               ?: error("There is no clientAnonymizedInstallId in certificate")

    val clientIdValue = connectionData.getComponent(BCStyle.UNIQUE_IDENTIFIER)?.let { sanitizeString(it) }
                        ?: error("There is no ClientId in certificate")
    clientId = ClientId(clientIdValue)

    clientFingerprint = permanentCertificate.sha3Fingerprint
    singleUseToken = connectionData.getComponent(BCStyle.PSEUDONYM)
  }

  private fun sanitizeString(s: String): String {
    return s.replace(notAllowedCharactersRegex, "_")
  }

  private fun sanitizeUserVisibleString(s: String): String {
    return s.replace(RdConstants.notAllowedNameCharactersRegex, "_")
  }

  private fun X500Name.getComponent(identifier: ASN1ObjectIdentifier): String? =
    getRDNs(identifier)
      .flatMap { it.typesAndValues.toList() }
      .map { IETFUtils.valueToString(it.value) }
      .singleOrNull()
}