package com.jetbrains.rdserver.tests

import com.intellij.codeWithMe.ClientId
import com.intellij.execution.filters.HyperlinkInfo
import com.intellij.execution.process.ProcessOutputTypes
import com.intellij.execution.testframework.Printable
import com.intellij.execution.testframework.Printer
import com.intellij.execution.testframework.sm.runner.SMTRunnerEventsAdapter
import com.intellij.execution.testframework.sm.runner.SMTRunnerEventsListener
import com.intellij.execution.testframework.sm.runner.SMTestProxy
import com.intellij.execution.testframework.sm.runner.events.TestOutputEvent
import com.intellij.execution.testframework.sm.runner.states.TestStateInfo
import com.intellij.execution.ui.ConsoleViewContentType
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.Key
import com.jetbrains.codeWithMe.model.*
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.RdServerPerClientComponents
import org.jetbrains.annotations.NotNull
import org.jetbrains.annotations.Nullable

@Suppress("NonDefaultConstructor")
class BackendTestsContentHost(
  val clientId: ClientId,
  lifetime: Lifetime,
  project: Project,
  protocol: Protocol) {
  private var ourProxies = mutableMapOf<Long, HashMap<String, SMTestProxy>>()

  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendTestsContentHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendTestsContentHost::class.java)
    }
  }

  fun getProxy(executionId : Long, locationUrl : String): SMTestProxy? {
    return ourProxies[executionId]?.get(locationUrl)
  }

  private fun registerProxy(proxy: SMTestProxy, executionId: Long) {
    val locationUrl = proxy.locationUrl
    if (locationUrl != null) {
      var hashMap = ourProxies[executionId]
      if (hashMap == null) {
        hashMap = HashMap()
        ourProxies[executionId] = hashMap
        Disposer.register(proxy) { ourProxies.remove(executionId) }
      }
      hashMap[locationUrl] = proxy
    }
  }

  init {
    val testsRunContentModel = TestsRunContentModel.create(lifetime, protocol)
    project.messageBus.connect(lifetime.createNestedDisposable()).subscribe(SMTRunnerEventsListener.TEST_STATUS, object : SMTRunnerEventsAdapter() {
      private fun doLaterWithClientId(action : () -> Unit) {
        ClientId.withClientId(clientId) {
          ApplicationManager.getApplication().invokeLater(action)
        }
      } 
      
      override fun onSuiteTreeStarted(suite: SMTestProxy?, nodeId: String?, parentNodeId: String?) {
        if (suite == null) return
        doLaterWithClientId {
          val executionId = suite.root.executionId
          val testsModel = SuiteTreeStarted(suite.locationUrl, parentNodeId, suite.name, nodeId, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onSuiteTreeNodeAdded(testProxy: SMTestProxy?, isSuite: Boolean, nodeId: String?, parentNodeId: String?) {
        if (testProxy == null) return
        doLaterWithClientId {
          val executionId = testProxy.root.executionId
          val testsModel = SuiteTreeNodeAdded(isSuite, testProxy.locationUrl, parentNodeId, testProxy.name, nodeId, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
          registerProxy(testProxy, executionId)
        }
      }

      override fun onSuiteTreeEnded(rootProxy: SMTestProxy.SMRootTestProxy,
                                    suiteName: String) {
        doLaterWithClientId {
          val executionId = rootProxy.executionId
          val testsModel = SuiteTreeEnded(suiteName, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onBuildTreeEnded(rootProxy: SMTestProxy.SMRootTestProxy) {
        doLaterWithClientId {
          val executionId = rootProxy.executionId
          val testsModel = BuildTreeEnded(executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onRootPresentationAdded(testsRoot: SMTestProxy.SMRootTestProxy,
                                           rootName: String?,
                                           comment: String?,
                                           rootLocation: String?) {
        doLaterWithClientId {
          val executionId = testsRoot.executionId
          registerProxy(testsRoot, executionId)
          testsRunContentModel.testEvent.fire(RootPresentation(rootName, comment, rootLocation, executionId))
        }
      }

      override fun onTestingStarted(testsRoot: SMTestProxy.SMRootTestProxy) {
        doLaterWithClientId {
          val executionId = testsRoot.executionId
          testsRunContentModel.testEvent.fire(ReporterAttached(executionId))
          val testsModel = Started(testsRoot.name, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onSuiteStarted(suite: SMTestProxy, nodeId : String?, parentNodeId: String?) {
        doLaterWithClientId {
          val executionId = suite.root.executionId
          val testsModel = SuiteStarted(suite.locationUrl, parentNodeId, suite.name, nodeId, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
          registerProxy(suite, executionId)
        }
      }

      override fun onSuiteStarted(suite: SMTestProxy) {
        onSuiteStarted(suite, null, null)
      }

      override fun onTestStarted(test: SMTestProxy, nodeId: String?, parentNodeId: String?) {
        doLaterWithClientId {
          val executionId = test.root.executionId
          val testsModel = TestStarted(test.locationUrl, parentNodeId, test.name, nodeId, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
          registerProxy(test, executionId)
        }
      }

      override fun onTestStarted(test: SMTestProxy) {
        onTestStarted(test, null, null)
      }


      override fun onTestFinished(test: @NotNull SMTestProxy,
                                  nodeId: @Nullable String?) {
        doLaterWithClientId {
          val executionId = test.root.executionId
          val testsModel = TestFinished(
            test.duration ?: 0,
            test.name,
            nodeId,
            executionId
          )
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onTestFinished(test: SMTestProxy) {
        onTestFinished(test, null)
      }

      override fun onTestFailed(test: @NotNull SMTestProxy,
                                nodeId: @Nullable String?) {
        doLaterWithClientId {
          val executionId = test.root.executionId
          val testsModel = TestFailure(
            test.errorMessage ?: "",
            test.stacktrace ?: "",
            test.magnitudeInfo == TestStateInfo.Magnitude.ERROR_INDEX,
            test.diffViewerProvider?.left,
            test.diffViewerProvider?.right,
            test.name,
            nodeId,
            executionId
          )
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onTestFailed(test: SMTestProxy) {
        onTestFailed(test, null)
      }

      override fun onTestIgnored(test: @NotNull SMTestProxy,
                                 nodeId: @Nullable String?) {
        doLaterWithClientId {
          val executionId = test.root.executionId
          val sb = StringBuilder()
          val p = object : Printer {
            override fun print(text: String?, contentType: ConsoleViewContentType?) {
              sb.append(text)
            }

            override fun onNewAvailable(printable: Printable) {}

            override fun printHyperlink(text: String?, info: HyperlinkInfo?) {}
            override fun mark() {}
          }
          test.printOn(p)
          val testsModel = TestIgnore(
            sb.toString(),
            test.stacktrace ?: "",
            test.name,
            nodeId,
            executionId
          )
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onTestIgnored(test: SMTestProxy) {
        onTestIgnored(test, null)
      }

      override fun onSuiteFinished(suite: @NotNull SMTestProxy,
                                   nodeId: @Nullable String?) {
        doLaterWithClientId {
          val executionId = suite.root.executionId
          val testsModel = SuiteFinished(suite.name, nodeId, executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onSuiteFinished(suite: SMTestProxy) {
        onSuiteFinished(suite, null)
      }

      override fun onTestingFinished(testsRoot: SMTestProxy.SMRootTestProxy) {
        doLaterWithClientId {
          val testsModel = Finished(testsRoot.name, testsRoot.executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onTestOutput(proxy: SMTestProxy,
                                event: @NotNull TestOutputEvent) {
        doLaterWithClientId {
          val testsModel = TestOutput(event.text, event.outputType == ProcessOutputTypes.STDOUT, proxy.name, event.id,
                                      proxy.root.executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }

      override fun onUncapturedOutput(proxy: SMTestProxy, text: String, type: Key<*>) {
        doLaterWithClientId {
          val contentTypeModel = ConsoleViewContentTypeModel(ConsoleViewContentType.getConsoleViewType(type).attributesKey?.externalName)
          val testsModel = Uncapture(CodeWithMeOutputMessage(text, contentTypeModel), proxy.root.executionId)
          testsRunContentModel.testEvent.fire(testsModel)
        }
      }
    })

    testsRunContentModel.navigate.advise(lifetime) {
      val locationUrl = it.selectedLocation ?: return@advise
      val proxy = getProxy(it.executionId, locationUrl) ?: return@advise
      proxy.getLocation(project, proxy.root.testConsoleProperties.scope)?.navigatable?.navigate(it.requestFocus)
    }
  }
}