package com.jetbrains.rdserver.toolWindow

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.components.service
import com.intellij.openapi.project.Project
import com.intellij.openapi.wm.RegisterToolWindowTask
import com.intellij.openapi.wm.ToolWindow
import com.intellij.openapi.wm.ToolWindowManager
import com.intellij.openapi.wm.impl.ToolWindowEventSource
import com.intellij.openapi.wm.impl.ToolWindowManagerImpl
import com.intellij.ui.viewModel.extraction.ToolWindowExtractorMode
import com.jetbrains.ide.model.uiautomation.BeToolWindowState
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.clientId

class BackendServerToolWindowManager(project: Project) : ToolWindowManagerImpl(project) {
  companion object {
    fun getInstance(project: Project): BackendServerToolWindowManager = project.service<ToolWindowManager>() as BackendServerToolWindowManager
  }

  init {
    invokeLater {
      if (project.isDisposed) return@invokeLater
      BackendToolwindowMirrorModeTracker(project)
    }
  }

  fun getHostToolWindow(id: String?): ToolWindow? = super.getToolWindow(id)

  override fun getToolWindow(id: String?): ToolWindow? {
    if (ClientId.isCurrentlyUnderLocalId) return super.getToolWindow(id)

    when (getBackendToolWindowMode(id)) {
      ToolWindowExtractorMode.DISABLE -> {
        return null
      }
      ToolWindowExtractorMode.FALLBACK -> {
        return super.getToolWindow(id)
      }
      ToolWindowExtractorMode.MIRROR -> {
        return super.getToolWindow(id)
      }
      ToolWindowExtractorMode.PER_CLIENT -> {
        return clientHost().getToolWindow(id)
      }
    }
  }

  override fun registerToolWindow(task: RegisterToolWindowTask): ToolWindow {
    if (ClientId.isCurrentlyUnderLocalId) return super.registerToolWindow(task)

    when (getBackendToolWindowMode(task.id)) {
      ToolWindowExtractorMode.DISABLE -> {
        error("Can't register client toolwindow ${task.id}")
      }
      ToolWindowExtractorMode.FALLBACK -> {
        return super.registerToolWindow(task)
      }
      ToolWindowExtractorMode.MIRROR -> {
        return super.registerToolWindow(task)
      }
      ToolWindowExtractorMode.PER_CLIENT -> {
        return clientHost().registerClientToolWindow(task)
      }
    }
  }

  override fun unregisterToolWindow(id: String) {
    if (ClientId.isCurrentlyUnderLocalId) {
      super.unregisterToolWindow(id)
      return
    }

    when (getBackendToolWindowMode(id)) {
      ToolWindowExtractorMode.DISABLE -> {
        error("Can't unregister client toolwindow ${id}")
      }
      ToolWindowExtractorMode.FALLBACK -> {
        super.unregisterToolWindow(id)
      }
      ToolWindowExtractorMode.MIRROR -> {
        super.unregisterToolWindow(id)
      }
      ToolWindowExtractorMode.PER_CLIENT -> {
        clientHost().unregisterToolWindow(id)
      }
    }
  }

  override fun shouldUpdateToolWindowContent(toolWindow: ToolWindow): Boolean {
    when (getBackendToolWindowMode(toolWindow.id)) {
      ToolWindowExtractorMode.DISABLE -> {
        return super.shouldUpdateToolWindowContent(toolWindow)
      }
      ToolWindowExtractorMode.FALLBACK -> {
        return super.shouldUpdateToolWindowContent(toolWindow)
      }
      ToolWindowExtractorMode.MIRROR -> {
        return super.shouldUpdateToolWindowContent(toolWindow) ||
               hasClientToolWindows(toolWindow.id)
      }
      ToolWindowExtractorMode.PER_CLIENT -> {
        return super.shouldUpdateToolWindowContent(toolWindow) ||
               toolWindow is BackendToolWindow
      }
    }
  }

  private fun hasClientToolWindows(id: String) = UserManager.getInstance(project).users.any {
    BackendToolWindowHost.getInstance(project, it.clientId).getToolWindow(id) != null
  }

  fun isActiveToolWindow(id: String): Boolean {
    if (ClientId.isCurrentlyUnderLocalId) {
      return activeToolWindowId == id
    }

    when (getBackendToolWindowMode(id)) {
      ToolWindowExtractorMode.DISABLE -> {
        return false
      }
      ToolWindowExtractorMode.FALLBACK -> {
        return activeToolWindowId == id
      }
      ToolWindowExtractorMode.MIRROR -> {
        return activeToolWindowId == id ||
               clientHost().getToolWindowState(id) == BeToolWindowState.ActiveAndFocused
      }
      ToolWindowExtractorMode.PER_CLIENT -> {
        return clientHost().getToolWindowState(id) == BeToolWindowState.ActiveAndFocused
      }
    }
  }

  override fun activateToolWindow(id: String, runnable: Runnable?, autoFocusContents: Boolean, source: ToolWindowEventSource?) {
    val toolWindow = getToolWindow(id) ?: return
    if (toolWindow is BackendToolWindow) {
      clientHost().setToolWindowState(id, BeToolWindowState.ActiveAndFocused)
    }
    else {
      super.activateToolWindow(id, runnable, autoFocusContents, source)
    }
  }

  override fun showToolWindow(id: String) {
    val toolWindow = getToolWindow(id) ?: return
    if (toolWindow is BackendToolWindow) {
      clientHost().setToolWindowState(id, BeToolWindowState.Active)
    }
    else {
      super.showToolWindow(id)
    }
  }

  override fun hideToolWindow(id: String, hideSide: Boolean, moveFocus: Boolean, source: ToolWindowEventSource?) {
    val toolWindow = getToolWindow(id) ?: return
    if (toolWindow is BackendToolWindow) {
      clientHost().setToolWindowState(id, BeToolWindowState.Hidden)
    }
    else {
      super.hideToolWindow(id, hideSide, moveFocus, source)
    }
  }

  private fun clientHost() = BackendToolWindowHost.getInstance(project, ClientId.current)
}
