package com.jetbrains.rdserver.toolWindow

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.impl.ContentManagerWatcher
import com.intellij.openapi.Disposable
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.ActionCallback
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.wm.ToolWindowAnchor
import com.intellij.openapi.wm.ToolWindowContentUiType
import com.intellij.openapi.wm.ToolWindowManager
import com.intellij.openapi.wm.ToolWindowType
import com.intellij.openapi.wm.ex.ToolWindowEx
import com.intellij.openapi.wm.impl.InternalDecorator
import com.intellij.ui.UIBundle
import com.intellij.ui.content.Content
import com.intellij.ui.content.ContentManager
import com.intellij.ui.content.ContentManagerListener
import com.intellij.ui.content.ContentUI
import com.intellij.ui.content.impl.ContentManagerImpl
import com.jetbrains.ide.model.uiautomation.BeToolWindow
import com.jetbrains.ide.model.uiautomation.BeToolWindowState
import com.jetbrains.rd.ide.toolWindows.toIdeaAnchor
import com.jetbrains.rd.ui.icons.toIdeaIcon
import com.jetbrains.rd.util.reactive.valueOrDefault
import java.awt.Rectangle
import java.awt.event.InputEvent
import javax.swing.Icon
import javax.swing.JComponent
import javax.swing.JPanel

class BackendToolWindow(private val _id: String,
                        val clientId: ClientId,
                        private val _project: Project,
                        private val _model: BeToolWindow) : ToolWindowEx {
  private val toolWindowManager get() = ToolWindowManager.getInstance(project) as BackendServerToolWindowManager

  private val _disposable: Disposable = Disposer.newDisposable()

  private val _panel = JPanel()
  private val _contentManager = BackendContentManager(_project)
  private val _decorator = BackendToolwindowHeaderDecorator()

  init {
    Disposer.register(_disposable, contentManager)
  }

  override fun getId(): String = _id
  override fun getDisposable(): Disposable = _disposable
  override fun getProject(): Project = _project
  override fun isDisposed(): Boolean = Disposer.isDisposed(_disposable)

  override fun getComponent(): JComponent = _panel
  override fun getContentManager(): ContentManager = _contentManager
  override fun getContentManagerIfCreated(): ContentManager = _contentManager
  override fun getDecorator(): InternalDecorator = _decorator

  override fun getAnchor(): ToolWindowAnchor = _model.position.toIdeaAnchor()
  override fun setAnchor(anchor: ToolWindowAnchor, runnable: Runnable?) = Unit

  override fun isSplitMode(): Boolean = _model.sideTool
  override fun setSplitMode(split: Boolean, runnable: Runnable?) = Unit

  override fun getReady(requestor: Any): ActionCallback = ActionCallback.DONE
  override fun isVisible(): Boolean = _model.state.valueOrDefault(BeToolWindowState.Hidden) != BeToolWindowState.Hidden
  override fun isActive(): Boolean {
    return isVisible &&
           ClientId.withClientId(clientId) {
             toolWindowManager.isActiveToolWindow(id)
           }
  }

  override fun activate(runnable: Runnable?) = activate(runnable, true)
  override fun activate(runnable: Runnable?, autoFocusContents: Boolean) = activate(runnable, autoFocusContents, true)
  override fun activate(runnable: Runnable?, autoFocusContents: Boolean, forced: Boolean) {
    ClientId.withClientId(clientId) {
      toolWindowManager.activateToolWindow(id, runnable, false)
    }
  }

  override fun show(runnable: Runnable?) {
    ClientId.withClientId(clientId) {
      toolWindowManager.showToolWindow(id)
    }
    callLater(runnable)
  }

  override fun hide(runnable: Runnable?) {
    ClientId.withClientId(clientId) {
      toolWindowManager.hideToolWindow(id, false)
    }
    callLater(runnable)
  }

  override fun remove() {
    ClientId.withClientId(clientId) {
      toolWindowManager.unregisterToolWindow(id)
    }
  }

  override fun isVisibleOnLargeStripe(): Boolean = true
  override fun setVisibleOnLargeStripe(visible: Boolean): Unit = Unit

  override fun getLargeStripeAnchor(): ToolWindowAnchor = anchor
  override fun setLargeStripeAnchor(anchor: ToolWindowAnchor) = Unit

  override fun isAutoHide(): Boolean = false
  override fun setAutoHide(value: Boolean) = Unit

  override fun getType(): ToolWindowType = ToolWindowType.DOCKED
  override fun setType(type: ToolWindowType, runnable: Runnable?) = Unit
  override fun getInternalType(): ToolWindowType = ToolWindowType.DOCKED

  override fun getIcon(): Icon? = _model.icon?.toIdeaIcon(project)
  override fun setIcon(icon: Icon) = Unit

  override fun getTitle(): String? = contentManager.selectedContent?.displayName
  override fun setTitle(title: String?) = Unit

  override fun getStripeTitle(): String = _model.title.text.value
  override fun setStripeTitle(title: String) = Unit

  override fun isAvailable(): Boolean = true
  override fun setAvailable(value: Boolean) = setAvailable(value, null)
  override fun setAvailable(value: Boolean, runnable: Runnable?) {
    callLater(runnable)
  }

  override fun setContentUiType(type: ToolWindowContentUiType, runnable: Runnable?) = Unit
  override fun setDefaultContentUiType(type: ToolWindowContentUiType) = Unit
  override fun getContentUiType(): ToolWindowContentUiType = ToolWindowContentUiType.TABBED

  override fun installWatcher(contentManager: ContentManager) = ContentManagerWatcher.watchContentManager(this, contentManager)
  override fun addContentManagerListener(listener: ContentManagerListener) = _contentManager.addContentManagerListener(listener)

  override fun setDefaultState(anchor: ToolWindowAnchor?, type: ToolWindowType?, floatingBounds: Rectangle?) = Unit
  override fun setToHideOnEmptyContent(hideOnEmpty: Boolean) = Unit

  override fun isShowStripeButton(): Boolean = true
  override fun setShowStripeButton(value: Boolean) = Unit

  override fun showContentPopup(inputEvent: InputEvent) = Unit
  override fun setTitleActions(actions: MutableList<AnAction>) = Unit

  override fun stretchWidth(value: Int) = Unit
  override fun stretchHeight(value: Int) = Unit

  override fun setAdditionalGearActions(additionalGearActions: ActionGroup?) = Unit
  override fun setTabActions(vararg actions: AnAction) = Unit
  override fun setTabDoubleClickActions(actions: MutableList<AnAction>) = Unit

  private fun callLater(runnable: Runnable?) {
    if (runnable != null) {
      ApplicationManager.getApplication().invokeLater(runnable)
    }
  }
}

private class BackendContentManager(project: Project) : ContentManagerImpl(BackendContentUI(), true, project) {
  override fun setSelectedContent(content: Content, requestFocus: Boolean, forcedFocus: Boolean, implicit: Boolean): ActionCallback =
    super.setSelectedContent(content, false, false, implicit)

  override fun removeContent(content: Content, dispose: Boolean, requestFocus: Boolean, forcedFocus: Boolean): ActionCallback =
    super.removeContent(content, dispose, false, false)

  override fun requestFocus(content: Content?, forced: Boolean): ActionCallback = ActionCallback.REJECTED
}

private class BackendContentUI : ContentUI {
  private val panel = JPanel()

  override fun getComponent(): JComponent = panel
  override fun setManager(manager: ContentManager) = Unit

  override fun isSingleSelection(): Boolean = true
  override fun isToSelectAddedContent(): Boolean = false
  override fun canBeEmptySelection(): Boolean = false
  override fun canChangeSelectionTo(content: Content, implicit: Boolean): Boolean = true

  override fun getCloseActionName(): String = UIBundle.message("tabbed.pane.close.tab.action.name")
  override fun getCloseAllButThisActionName(): String = UIBundle.message("tabbed.pane.close.all.tabs.but.this.action.name")
  override fun getPreviousContentActionName(): String = UIBundle.message("tabbed.pane.select.previous.tab")
  override fun getNextContentActionName(): String = UIBundle.message("tabbed.pane.select.next.tab")
}

private class BackendToolwindowHeaderDecorator : InternalDecorator() {
  private val toolbar = ActionManager.getInstance().createActionToolbar(ActionPlaces.UNKNOWN, DefaultActionGroup(), true)

  override fun getHeaderToolbar(): ActionToolbar = toolbar
  override fun getHeaderHeight(): Int = 0
  override fun setHeaderVisible(value: Boolean) = Unit
  override fun isHeaderVisible(): Boolean = false
}