package com.jetbrains.rdserver.toolWindow

import com.intellij.analysis.problemsView.toolWindow.ProblemsView
import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.application.invokeLater
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.extensions.ExtensionPointListener
import com.intellij.openapi.extensions.PluginDescriptor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.util.NlsContexts
import com.intellij.openapi.util.registry.Registry
import com.intellij.openapi.wm.*
import com.intellij.openapi.wm.ex.ToolWindowEx
import com.intellij.openapi.wm.ex.ToolWindowManagerListener
import com.intellij.openapi.wm.impl.findIconFromBean
import com.intellij.openapi.wm.impl.getStripeTitleSupplier
import com.intellij.ui.SimpleTextAttributes
import com.intellij.ui.content.Content
import com.intellij.ui.content.ContentManagerEvent
import com.intellij.ui.content.ContentManagerListener
import com.intellij.ui.content.impl.ContentImpl
import com.intellij.ui.viewModel.extraction.ToolWindowExtractorMode
import com.intellij.ui.viewModel.extraction.ToolWindowViewModelExtractor
import com.intellij.ui.viewModel.extraction.ToolWindowViewModelListener
import com.intellij.ui.viewModel.extraction.ViewModelToolWindowFactory
import com.intellij.util.PlatformUtils
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.ide.toolWindows.toModelAnchor
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.ui.bedsl.dsl.label
import com.jetbrains.rd.ui.bedsl.dsl.setText
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.LifetimeDefinition
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.reactive.*
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.permissions.ToolWindowAccessLevel
import com.jetbrains.rdserver.permissions.impl.PermissionManagerImpl
import com.jetbrains.rdserver.richText
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rdserver.ui.converters.TrackOptions
import com.jetbrains.rdserver.ui.converters.toBeModel
import javax.swing.Icon

@Suppress("NonDefaultConstructor")
class BackendToolWindowHost(
  val clientId: ClientId,
  val clientLifetime: Lifetime,
  val project: Project,
  solution: Solution) {

  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendToolWindowHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendToolWindowHost::class.java)
    }
  }

  private val toolWindowManager get() = BackendServerToolWindowManager.getInstance(project)

  private val toolWindowsModel = solution.toolWindowModel
  private val idToEntry = mutableMapOf<String, RemoteToolWindowEntry>()

  private var processingContentChange: Boolean = false

  init {
    // avoid cyclic dependency on BackendServerToolWindowManager
    invokeLater {
      if (clientLifetime.isNotAlive) return@invokeLater
      if (ApplicationManager.getApplication().isHeadlessEnvironment) return@invokeLater
      registerFactories(clientLifetime)
      registerListeners(clientLifetime)
    }

    clientLifetime.onTerminationIfAlive {
      for (entry in idToEntry.values) {
        entry.destroy()
      }
      idToEntry.clear()
    }
  }

  private fun registerFactories(lifetime: Lifetime) {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(ToolWindowManagerListener.TOPIC, object : ToolWindowManagerListener {
      override fun toolWindowsRegistered(ids: List<String>, toolWindowManager: ToolWindowManager) {
        for (id in ids) {
          if (getBackendToolWindowMode(id) == ToolWindowExtractorMode.MIRROR) {
            registerLazyMirrorToolWindow(id)
          }
        }
      }

      override fun toolWindowUnregistered(id: String, toolWindow: ToolWindow) {
        if (getBackendToolWindowMode(id) == ToolWindowExtractorMode.MIRROR) {
          unregisterToolWindow(id)
        }
      }
    })

    for (id in toolWindowManager.toolWindowIds) {
      if (getBackendToolWindowMode(id) == ToolWindowExtractorMode.MIRROR) {
        registerLazyMirrorToolWindow(id)
      }
    }


    ToolWindowEP.EP_NAME.point.addExtensionPointListener(object : ExtensionPointListener<ToolWindowEP> {
      override fun extensionAdded(extension: ToolWindowEP, pluginDescriptor: PluginDescriptor) {
        if (getBackendToolWindowMode(extension.id) == ToolWindowExtractorMode.PER_CLIENT) {
          registerLazyClientToolWindow(extension)
        }
      }

      override fun extensionRemoved(extension: ToolWindowEP, pluginDescriptor: PluginDescriptor) {
        if (getBackendToolWindowMode(extension.id) == ToolWindowExtractorMode.PER_CLIENT) {
          unregisterToolWindow(extension.id)
        }
      }
    }, true, lifetime.createNestedDisposable())
  }

  private fun registerListeners(lifetime: Lifetime) {
    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(ToolWindowViewModelListener.TOPIC, object : ToolWindowViewModelListener {
      override fun toolWindowChanged(id: String) {
        invokeLater {
          syncContents(id, force = true)
        }
      }

      override fun toolWindowsChanged() {
        invokeLater {
          for (id in idToEntry.keys) {
            syncContents(id, force = true)
          }
        }
      }
    })

    PermissionManagerImpl.getInstanceImpl(project).permissionModelChanged.advise(lifetime) {
      for (id in idToEntry.keys) {
        syncContents(id, force = true)
      }
    }
  }


  fun getToolWindow(id: String?): ToolWindow? = idToEntry[id]?.toolWindow

  fun registerClientToolWindow(task: RegisterToolWindowTask): ToolWindow {
    val id = task.id

    val windowModel = createToolWindowModel(
      anchor = task.anchor,
      sideTool = task.sideTool,
      icon = task.icon,
      displayName = task.stripeTitle?.get() ?: task.id)
    registerWindowModel(id, windowModel)

    val beToolWindow = windowModel.toolWindow
    val entry = idToEntry[id] ?: error("Entry not found for $id for user ${clientId}")
    val window = BackendToolWindow(id, clientId, project, beToolWindow)
    entry.toolWindow = window

    task.contentFactory?.let { factory ->
      ClientId.withClientId(clientId) {
        factory.init(window)
        factory.createToolWindowContent(project, window)
      }
    }
    task.component?.let { component ->
      val content = ContentImpl(component, "", false)
      window.contentManager.addContent(content)
    }
    trackToolWindowContents(id, window, windowModel)

    return window
  }

  private fun registerLazyClientToolWindow(extension: ToolWindowEP) {
    val id = extension.id
    val factory = extension.getToolWindowFactory(extension.pluginDescriptor) ?: error("Factory for toolwindow $id not found")

    val windowModel = createToolWindowModel(
      anchor = ToolWindowAnchor.fromText(extension.anchor),
      sideTool = extension.secondary,
      icon = findIconFromBean(extension, factory, extension.pluginDescriptor),
      displayName = getStripeTitleSupplier(id, extension.pluginDescriptor)?.get() ?: id)
    registerWindowModel(id, windowModel)

    val beToolWindow = windowModel.toolWindow
    beToolWindow.state.adviseUntil(clientLifetime) {
      if (it == BeToolWindowState.Hidden) return@adviseUntil false

      val entry = idToEntry[id]
      if (entry == null || entry.toolWindow != null) return@adviseUntil true

      val window = BackendToolWindow(id, clientId, project, beToolWindow)
      entry.toolWindow = window

      ClientId.withClientId(clientId) {
        factory.init(window)
        factory.createToolWindowContent(project, window)
      }
      trackToolWindowContents(id, window, windowModel)

      return@adviseUntil true
    }
  }

  fun registerLazyMirrorToolWindow(id: String) {
    val hostWindow = toolWindowManager.getHostToolWindow(id) as ToolWindowEx? ?: error("Host tool window with id $id could not be found")

    val windowModel = createToolWindowModel(
      anchor = hostWindow.anchor,
      sideTool = hostWindow.isSplitMode,
      icon = hostWindow.decoration?.icon,
      displayName = hostWindow.stripeTitle)
    registerWindowModel(id, windowModel)

    val beToolWindow = windowModel.toolWindow
    beToolWindow.state.adviseUntil(clientLifetime) {
      if (it == BeToolWindowState.Hidden) return@adviseUntil false

      val entry = idToEntry[id]
      if (entry == null || entry.toolWindow != null) return@adviseUntil true

      entry.toolWindow = hostWindow
      trackToolWindowContents(id, hostWindow, windowModel)
      return@adviseUntil true
    }
  }

  private fun registerWindowModel(id: String, windowModel: BeIdeaToolWindow) {
    if (idToEntry[id] != null) error("Toolwindow model is already registered: ${id} for user ${clientId}")

    idToEntry[id] = RemoteToolWindowEntry(windowModel)
    toolWindowsModel.availableToolWindows[id] = windowModel
  }

  fun unregisterToolWindow(id: String) {
    val oldEntry = idToEntry.remove(id)
    oldEntry?.destroy()

    toolWindowsModel.availableToolWindows.remove(id)
  }

  private fun trackToolWindowContents(id: String,
                                      toolWindow: ToolWindowEx,
                                      windowModel: BeIdeaToolWindow) {
    val beToolWindow = windowModel.toolWindow
    val isMirror = getBackendToolWindowMode(id) == ToolWindowExtractorMode.MIRROR

    beToolWindow.state.advise(clientLifetime) {
      project.messageBus.syncPublisher(ToolWindowManagerListener.TOPIC).stateChanged(toolWindowManager)
    }

    beToolWindow.content.tabs.adviseAddRemove(clientLifetime) { addRemove, _, content ->
      noReenter(::processingContentChange) {
        if (addRemove == AddRemove.Remove && !isMirror) {
          removeContent(id, content)
        }
      }
    }
    beToolWindow.content.selectedIndex.change.advise(clientLifetime) { index ->
      noReenter(::processingContentChange) {
        val content = toolWindow.contentManager.getContent(index)
        if (content != null && !isMirror) {
          toolWindow.contentManager.setSelectedContent(content)
        }
      }
    }

    toolWindow.contentManager.addContentManagerListener(object : ContentManagerListener {
      override fun contentAdded(event: ContentManagerEvent) {
        invokeLater { syncContents(id) }
      }

      override fun contentRemoved(event: ContentManagerEvent) {
        invokeLater { syncContents(id) }
      }

      override fun selectionChanged(event: ContentManagerEvent) {
        noReenter(::processingContentChange) {
          val content = beToolWindow.content
          if (event.index >= 0 && event.index < content.tabs.size) {
            content.selectedIndex.set(event.index)
          }
        }
      }
    })
    syncContents(id)
  }

  private fun syncContents(id: String, force: Boolean = false) {
    val entry = idToEntry[id] ?: return
    val window = entry.toolWindow ?: return
    val beToolWindow = entry.model.toolWindow

    val isMirror = getBackendToolWindowMode(id) == ToolWindowExtractorMode.MIRROR

    val accessLevel = when {
      BackendToolWindowModeProvider.isSafeToolWindow(id) -> ToolWindowAccessLevel.FullAccess
      else -> PermissionManager.getInstance(project).getToolWindowAccess(clientId)
    }
    entry.model.accessLevel.set(accessLevel.toRdAccessLevel())

    noReenter(::processingContentChange) {
      val contents = window.contentManager.contents.toSet()
      val toRemove = entry.contentToModel.filter { (content, tabEntry) ->
        force ||
        accessLevel == ToolWindowAccessLevel.Disabled ||
        !contents.contains(content)
      }

      for ((content, tabEntry) in toRemove) {
        entry.contentToModel.remove(content)
        beToolWindow.content.tabs.remove(tabEntry.tab)
        tabEntry.lifetimeDefinition.terminate()
      }

      if (accessLevel != ToolWindowAccessLevel.Disabled) {
        for (content in contents) {
          if (entry.contentToModel.contains(content)) continue

          val lifetime = content.defineNestedLifetime()
          val isReadOnly = accessLevel == ToolWindowAccessLevel.Readonly
          val contentModel = createContentModel(beToolWindow, content, isMirror, isReadOnly, lifetime)
          beToolWindow.content.tabs.add(contentModel)
          entry.contentToModel[content] = RemoteToolWindowEntry.TabEntry(contentModel, lifetime)
        }
      }
      else {
        val beModel = createStatusPanel(CodeWithMePluginBundle.message("toolwindow.message.access.denied.label"))
        val contentModel = BeTab(tabName = label(""),
                                 content = beModel,
                                 isCloseable = false)
        beToolWindow.content.tabs.add(contentModel)
        entry.contentToModel[null] = RemoteToolWindowEntry.TabEntry(contentModel, LifetimeDefinition.Terminated)
      }
    }
  }

  private fun removeContent(id: String, contentModel: BeTab) {
    val entry = idToEntry[id] ?: return
    val window = entry.toolWindow ?: return
    val (content, tabEntry) = entry.contentToModel.entries.find { it.value.tab == contentModel } ?: return
    if (content != null) {
      entry.contentToModel.remove(content)
      window.contentManager.removeContent(content, true)
      tabEntry.lifetimeDefinition.terminate()
    }
  }

  private fun createContentModel(beToolWindow: BeToolWindow,
                                 content: Content,
                                 isMirror: Boolean,
                                 isReadOnly: Boolean,
                                 lifetime: Lifetime): BeTab {
    val tabHolder = LazyTabContentHolder(content, lifetime,
                                         TrackOptions(canBeShared = isMirror,
                                                      clientId = clientId))

    val hostModel = tabHolder.beControl
    val beControl = if (isReadOnly) {
      val clientModel = hostModel.deepClone() as BeControl
      synchronizeTo(lifetime, hostModel, clientModel)
      clientModel
    }
    else {
      hostModel
    }

    val tab = BeTab(tabName = label(content.displayName.orEmpty()),
                    content = beControl,
                    isCloseable = content.isCloseable && !isMirror)

    content.addPropertyChangeListener {
      if (it.propertyName == Content.PROP_DISPLAY_NAME) {
        tab.tabName.setText(content.displayName.orEmpty())
      }
    }

    tabHolder.registerListeners(beToolWindow, tab)

    return tab
  }

  private fun createStatusPanel(text: String): BeControl {
    val label = BeRichText(false).also {
      it.text.set(richText {
        append(text,
               SimpleTextAttributes.GRAYED_ATTRIBUTES)
      })
    }
    val grid = BeGrid(GridOrientation.Horizontal).also {
      it.items.set(listOf(
        BeGridElement(BeSpacer(), BeUnitSize(BeSizingType.Fill)),
        BeGridElement(label, BeUnitSize(BeSizingType.Fit), BeAlignment.Center),
        BeGridElement(BeSpacer(), BeUnitSize(BeSizingType.Fill))))
    }
    return BeStyleControl(grid).also {
      it.backgroundColor.set(BeBackgroundColorModifier(ThemeColorId(ColorId.ListBackground)))
    }
  }

  private fun createToolWindowModel(displayName: String,
                                    icon: Icon?,
                                    anchor: ToolWindowAnchor,
                                    sideTool: Boolean): BeIdeaToolWindow {
    val beToolWindow = BeToolWindow(
      title = label(displayName),
      icon = icon?.toProtocolModel(),
      position = anchor.toModelAnchor(),
      sideTool = sideTool
    )
    return BeIdeaToolWindow(beToolWindow)
  }

  fun getToolWindowState(id: String): BeToolWindowState {
    val entry = idToEntry[id] ?: return BeToolWindowState.Hidden
    return entry.model.toolWindow.state.valueOrDefault(BeToolWindowState.Hidden)
  }

  fun setToolWindowState(id: String, state: BeToolWindowState) {
    val entry = idToEntry[id] ?: return
    entry.model.toolWindow.state.set(state)
  }

  fun setSharedAccess(id: String, isSharedAccess: Boolean) {
    val entry = idToEntry[id] ?: return
    entry.model.isSharedAccess.set(isSharedAccess)
  }
}

fun getBackendToolWindowMode(id: String?): ToolWindowExtractorMode {
  return BackendToolWindowModeProvider.getBackendToolWindowMode(id)
}

private fun findViewModelExtractor(id: String): ToolWindowViewModelExtractor? {
  return ToolWindowViewModelExtractor.EP_NAME.findFirstSafe { extractor -> extractor.isApplicable(id) }
}

private fun findToolWindowFactory(id: String): ToolWindowFactory? {
  val bean = ToolWindowEP.EP_NAME.findFirstSafe { it.id == id } ?: return null
  return bean.getToolWindowFactory(bean.pluginDescriptor)
}

private class RemoteToolWindowEntry(val model: BeIdeaToolWindow) {
  var toolWindow: ToolWindowEx? = null

  val contentToModel = mutableMapOf<Content?, TabEntry>()

  fun destroy() {
    val window = toolWindow
    if (window is BackendToolWindow) {
      Disposer.dispose(window.disposable)
    }
    contentToModel.clear()
  }

  class TabEntry(val tab: BeTab, val lifetimeDefinition: LifetimeDefinition)
}

object BackendToolWindowModeProvider {
  fun getBackendToolWindowMode(id: String?): ToolWindowExtractorMode {
    if (id == null) return ToolWindowExtractorMode.DISABLE

    val forcedMode = getForcedMode(id)
    if (forcedMode != null) return forcedMode

    val extractor = findViewModelExtractor(id)
    if (extractor != null) return extractor.mode

    val factory = findToolWindowFactory(id)
    if (factory is ViewModelToolWindowFactory) return ToolWindowExtractorMode.PER_CLIENT

    return getDefaultMode(id)
  }

  /**
   * Keep in sync with [getSupportedToolWindowsList].
   */
  private fun getDefaultMode(id: String): ToolWindowExtractorMode {
    if (ToolWindowId.ANT_BUILD == id) return ToolWindowExtractorMode.MIRROR
    if ("Gradle" == id) return ToolWindowExtractorMode.MIRROR
    if ("Maven" == id) return ToolWindowExtractorMode.MIRROR

    if (ToolWindowId.MESSAGES_WINDOW == id) return ToolWindowExtractorMode.MIRROR
    if ("CMake" == id) return ToolWindowExtractorMode.MIRROR

    if (ProblemsView.ID == id) return ToolWindowExtractorMode.PER_CLIENT

    //if ("Services" == id) return ToolWindowExtractorMode.MIRROR

    return ToolWindowExtractorMode.FALLBACK
  }

  fun isSafeToolWindow(id: String): Boolean {
    if (ProblemsView.ID == id) return true

    return false
  }

  private fun getForcedMode(id: String): ToolWindowExtractorMode? {
    val value = Registry.stringValue("codewithme.toolwindow.force.extractor.mode")
    if (value.isEmpty()) return null

    for (pattern in value.split(';')) {
      val list = pattern.split('=', limit = 2)
      if (list.size != 2) continue

      val patternId = list[0].trim()
      val patternValue = list[1].trim()
      if (id == patternId) {
        try {
          return ToolWindowExtractorMode.valueOf(patternValue)
        }
        catch (e: IllegalArgumentException) {
          logger<BackendToolWindowModeProvider>().warn(e)
        }
      }
    }
    return null
  }

  fun getSupportedToolWindowsList(): @NlsContexts.Tooltip String {
    if (PlatformUtils.isCLion() || PlatformUtils.isAppCode()) {
      return CodeWithMePluginBundle.message("permission.dialog.toolWindow.access.tooltip.clion")
    }
    else {
      return CodeWithMePluginBundle.message("permission.dialog.toolWindow.access.tooltip")
    }
  }
}

private class LazyTabContentHolder(val content: Content,
                                   val lifetime: Lifetime,
                                   val trackOptions: TrackOptions) {
  val beControl: BeDynamicControlContainer = BeDynamicControlContainer()

  private val isActive = Property(false)

  init {
    isActive.view(lifetime) { beControlLifetime, active ->
      beControl.content.set(if (active) content.component.toBeModel(trackOptions, beControlLifetime) else BeSpacer())
    }
  }

  fun registerListeners(beToolWindow: BeToolWindow, tab: BeTab) {
    fun isContentVisible(): Boolean {
      val isWindowShown = beToolWindow.state.valueOrDefault(BeToolWindowState.Hidden) != BeToolWindowState.Hidden
      val selectedIndex = beToolWindow.content.selectedIndex.valueOrDefault(0)
      val isTabSelected = beToolWindow.content.tabs.getOrNull(selectedIndex) == tab
      return isWindowShown && isTabSelected
    }

    beToolWindow.state.flowInto(lifetime, isActive) { isContentVisible() }
    beToolWindow.content.selectedIndex.flowInto(lifetime, isActive) { isContentVisible() }
    beToolWindow.content.tabs.flowInto(lifetime, isActive) { isContentVisible() }
    isActive.set(isContentVisible())
  }
}

private fun ToolWindowAccessLevel.toRdAccessLevel(): RdToolWindowAccessLevel = when (this) {
  ToolWindowAccessLevel.Disabled -> RdToolWindowAccessLevel.Disabled
  ToolWindowAccessLevel.Readonly -> RdToolWindowAccessLevel.Readonly
  ToolWindowAccessLevel.FullAccess -> RdToolWindowAccessLevel.FullAccess
}