// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.ui

import com.intellij.codeWithMe.ClientId
import com.intellij.ide.IdeEventQueue
import com.intellij.openapi.Disposable
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.openapi.ui.DialogWrapper
import com.intellij.openapi.ui.DialogWrapperPeer
import com.intellij.openapi.ui.impl.DialogWrapperPeerFactoryImpl
import com.intellij.openapi.util.ActionCallback
import com.intellij.openapi.util.Disposer
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.IViewableList
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rd.util.reactive.adviseUntil
import com.jetbrains.rd.util.reactive.hasValue
import com.jetbrains.rd.util.threading.SpinWait
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.ui.converters.toBeModel
import com.jetbrains.rdserver.ui.converters.verticalGridOf
import com.jetbrains.rdserver.ui.modalDialogs.tryProcessModalOnBackend
import com.jetbrains.rdserver.unattendedHost.UnattendedHostManager
import java.awt.*
import java.awt.event.KeyListener
import java.awt.event.MouseListener
import java.awt.event.MouseMotionListener
import javax.swing.JButton
import javax.swing.JComponent
import javax.swing.JPanel
import javax.swing.JRootPane

private val logger = getLogger<BackendDialogWrapperPeerFactory>()

internal class BackendDialogWrapperPeerFactory : DialogWrapperPeerFactoryImpl() {
  override fun createPeer(
    wrapper: DialogWrapper,
    project: Project?,
    canBeParent: Boolean,
    ideModalityType: DialogWrapper.IdeModalityType
  ): DialogWrapperPeer {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.createPeer(wrapper, project, canBeParent, ideModalityType)
    }

    if (project == null) {
      logger.error("Attempting to show dialog $wrapper under ClientId ${ClientId.current} without specifying a project")
      return super.createPeer(wrapper, project as Project?, canBeParent, ideModalityType)
    }
//  return BeDialogWrapperPeer(ClientId.current, project, RdServerPerClientComponents.getService(project, UIAutomationInteractionModel::class.java), wrapper, ideModalityType)
    return RdDialogWrapperPeer(ClientId.current, project, RdServerPerClientComponents.getService(project, UIAutomationInteractionModel::class.java), wrapper, ideModalityType)
  }
}

class RdDialogWrapperPeer(
  private val clientId: ClientId,
  private val project: Project,
  private val uiInteractionModel: UIAutomationInteractionModel,
  private val dialogWrapper: DialogWrapper,
  private val modalityType: DialogWrapper.IdeModalityType
) : DialogWrapperPeer() {
  companion object {
    private val logger = getLogger<RdDialogWrapperPeer>()
  }

  private val pane = JRootPane().apply { putClientProperty("DIALOG_ROOT_PANE", true) }

  private var title: String? = null
  private var content: JComponent? = null

  override fun getWindow(): Window? = null

  override fun getRootPane(): JRootPane = pane

  override fun isHeadless(): Boolean = true

  override fun getTitle(): String? = title

  override fun setTitle(title: String?) {
    this.title = title
  }

  override fun setContentPane(content: JComponent?) {
    this.content = content
  }

  override fun show(): ActionCallback {
    val content = content ?: return ActionCallback.DONE
    val lifetime = dialogWrapper.disposable.defineNestedLifetime()

    val unattendedHostManager = UnattendedHostManager.getInstance()
    if (isModal && (!unattendedHostManager.isUnattendedMode || !unattendedHostManager.isUnderControllerClientId()))
      return dialogWrapper.tryProcessModalOnBackend(project, clientId, lifetime, logger) ?: ActionCallback.DONE

    val centerPane = content.extractCenterPane().toBeModel(lifetime = lifetime.lifetime)
    val southSection = content.extractSouthPane()
    val dialogButtons = mutableListOf<BeButton>()
    val errorText = southSection?.extractCenterPane()
    val beErrorText = errorText?.toBeModel(lifetime = lifetime.lifetime)

    val dialogContent = if (beErrorText != null) {
      verticalGridOf(centerPane to true, beErrorText to false)
    }
    else {
      centerPane
    }

    val southPane = southSection?.extractSouthPane()
    southPane?.let { convertButtons(it, dialogButtons, lifetime.lifetime) }

    val actionCallback = ActionCallback()

    val dialog = BeDialog(DialogModality.MODAL, title, BeControlSizeProportional(BeControlSizeType.FIT_TO_CONTENT, 1.0f))
    dialog.content.set(dialogContent)
    dialog.buttons.set(dialogButtons)

    uiInteractionModel.activeDialogs.add(dialog)
    dialog.result.adviseNotNull(lifetime.lifetime) {
      if (dialog.result.changing) return@adviseNotNull
      dialogWrapper.close(when (it) {
                            DialogResult.OK -> DialogWrapper.OK_EXIT_CODE
                            DialogResult.CANCEL -> DialogWrapper.CANCEL_EXIT_CODE
                            DialogResult.CLOSE -> DialogWrapper.CLOSE_EXIT_CODE
                            DialogResult.NONE -> DialogWrapper.CLOSE_EXIT_CODE
                          })
    }

    uiInteractionModel.activeDialogs.adviseUntil(lifetime) {
      if (it is IViewableList.Event.Remove && it.oldValue == dialog) {
        lifetime.terminate()
        actionCallback.setDone()
        true
      }
      else {
        false
      }
    }

    Disposer.register(dialogWrapper.disposable, Disposable {
      if (!dialog.result.hasValue)
        dialog.result.set(DialogResult.CLOSE)
      actionCallback.setDone()
    })

    SpinWait.spinUntil {
      IdeEventQueue.getInstance().flushQueue()
      actionCallback.isProcessed
    }


    return actionCallback
  }

  override fun setAppIcons() {
    TODO("not implemented")
  }

  override fun isModal(): Boolean {
    return modalityType != DialogWrapper.IdeModalityType.MODELESS
  }

  override fun getSize(): Dimension {
    return Dimension(100, 100)
  }

  override fun setLocation(p0: Point) {
    TODO("not implemented")
  }

  override fun setLocation(p0: Int, p1: Int) {
    TODO("not implemented")
  }

  override fun addKeyListener(p0: KeyListener?) {
    TODO("not implemented")
  }

  override fun pack() {
    TODO("not implemented")
  }

  override fun centerInParent() {
    TODO("not implemented")
  }

  override fun setResizable(p0: Boolean) {
    TODO("not implemented")
  }

  override fun isResizable() {
    TODO("not implemented")
  }

  override fun repaint() {
  }

  override fun setUndecorated(p0: Boolean) {
    TODO("not implemented")
  }

  override fun toFront() {
    TODO("not implemented")
  }

  override fun getOwner(): Window {
    TODO("not implemented")
  }

  override fun setModal(p0: Boolean) {
    TODO("not implemented")
  }

  override fun getLocation(): Point {
    TODO("not implemented")
  }

  override fun isShowing(): Boolean {
    TODO("not implemented")
  }

  override fun setSize(p0: Int, p1: Int) {
    TODO("not implemented")
  }

  override fun addMouseListener(p0: MouseListener?) {
    TODO("not implemented")
  }

  override fun addMouseListener(p0: MouseMotionListener?) {
    TODO("not implemented")
  }

  override fun getPreferredSize(): Dimension {
    TODO("not implemented")
  }

  override fun validate() {
    TODO("not implemented")
  }

  override fun toBack() {
    TODO("not implemented")
  }

  override fun isVisible(): Boolean {
    TODO("not implemented")
  }

  override fun dispose() {
  }

  override fun getContentPane(): Container? {
    return content
  }
}

internal fun JComponent.extractCenterPane(): JComponent {
  val borderLayout = layout as? BorderLayout ?: return this
  return borderLayout.getLayoutComponent(this, BorderLayout.CENTER) as? JComponent ?: this
}

internal fun JComponent.extractSouthPane(): JComponent? {
  val borderLayout = layout as? BorderLayout ?: return null
  return borderLayout.getLayoutComponent(this, BorderLayout.SOUTH) as? JComponent
}

internal fun convertButtons(panel: JComponent, result: MutableList<BeButton>, lifetime: Lifetime) {
  for (component in panel.components) {
    if (component is JPanel) {
      convertButtons(component, result, lifetime)
    }
    else if (component is JButton) {
      val button = component.toBeModel(lifetime = lifetime) as? BeButton
      button?.let { result.add(it) }
    }
  }
}
