package com.jetbrains.rdserver.ui

import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.runAndLogException
import com.intellij.openapi.util.IconLoader
import com.intellij.openapi.util.io.FileUtilRt
import com.intellij.ui.DeferredIcon
import com.intellij.ui.DeferredIconImpl
import com.intellij.ui.LayeredIcon
import com.intellij.ui.RetrievableIcon
import com.intellij.ui.icons.RowIcon
import com.intellij.util.ExceptionUtil
import com.intellij.util.ui.ColorIcon
import com.intellij.util.ui.ColorizeProxyIcon
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.idea.LifetimedApplicationService
import com.jetbrains.rdserver.toRgbaColor
import java.io.IOException
import java.net.URL
import java.util.concurrent.ConcurrentHashMap
import java.util.concurrent.atomic.AtomicLong
import javax.swing.Icon

class BackendIconHost : LifetimedApplicationService() {
  companion object {
    fun getInstance(): BackendIconHost = service()

    private val LOG = logger<BackendIconHost>()
  }

  private val imageIconCache = ConcurrentHashMap<Icon, IconModel?>()
  private val imageIconCounterLong = AtomicLong()

  fun createModel(icon: Icon): IconModel? = when (icon) {
    is IconLoader.CachedImageIcon -> createImageIconModel(icon)
    is ColorIcon -> ColorIconModel(icon.iconColor.toRgbaColor()!!)
    is ColorizeProxyIcon -> createColorizedModel(icon)
    is RowIcon -> RowIconModel(icon.allIcons.mapNotNull { createModel(it) })
    is LayeredIcon -> CompositionIconModel(icon.allLayers.mapNotNull { createModel(it) })
    is DeferredIconImpl<*> -> createDeferredModel(icon)
    is DeferredIcon -> createModel(icon.evaluate())
    is RetrievableIcon -> createModel(icon.retrieveIcon())
    else -> null
  }

  private fun createImageIconModel(icon: IconLoader.CachedImageIcon): IconModel? {
    return imageIconCache.getOrPut(icon) {
      val path = icon.originalPath
      val whiteIcon = icon.getDarkIcon(false) as IconLoader.CachedImageIcon
      val darkIcon = icon.getDarkIcon(true) as IconLoader.CachedImageIcon
      val name = path?.let { FileUtilRt.getNameWithoutExtension(it) } ?: "protocol_icon_no_path"
      val ext = path?.let { FileUtilRt.getExtension(it) } ?: "png"

      val sources = listOf(
        IconSource("${name}.$ext", readIcon(whiteIcon.url, false, false)),
        IconSource("${name}@2x.$ext", readIcon(whiteIcon.url, false, true)),
        IconSource("${name}_dark.$ext", readIcon(darkIcon.url, true, false)),
        IconSource("${name}@2x_dark.$ext", readIcon(darkIcon.url, true, true))
      )

      if (sources.all { it.body == null }) {
        LOG.error("Failed to read icon:'$path'")
      }

      ImageIconModel(ImageIconInternWrapper(imageIconCounterLong.incrementAndGet(), path ?: "${name}.$ext", sources))
    }
  }

  private fun readIcon(url: URL?, dark: Boolean, retina: Boolean): ByteArray? {
    val path = url?.path ?: return null
    val nameWithWithoutExtension = path.substringBeforeLast(".")
    val extension = path.substringAfterLast(".")

    val normalizedName = nameWithWithoutExtension.removeSuffix("@2x_dark").removeSuffix("@2x").removeSuffix("_dark")
    val newFile =  normalizedName + (if (retina) "@2x" else "") + (if (dark) "_dark" else "") + "." + extension
    val newUrl = URL(url.protocol, url.host, url.port, newFile)
    val bytes = try { newUrl.readBytes() } catch (ex: IOException) { null }

    if (bytes == null && dark && retina) { // IJ allows putting modifiers in this order
      val fallbackPath = normalizedName + "_dark@2x" + "." + extension
      val fallbackUrl = URL(url.protocol, url.host, url.port, fallbackPath)
      return try { fallbackUrl.readBytes() } catch (ex: IOException) { null }
    }
    return bytes
  }

  private fun createDeferredModel(icon: DeferredIconImpl<*>): IconModel? {
    if (icon.isDone) return createModel(icon.retrieveIcon())

    if (icon.isNeedReadAction && !application.isReadAccessAllowed) {
      val message = "Calculating deferred icon without a read lock! icon: $icon"
      val extendedMessage = message.let { if (LOG.isTraceEnabled) it + "stacktrace: ${ExceptionUtil.currentStackTrace()}" else it }
      LOG.warn(extendedMessage)
    }
    return LOG.runAndLogException { createModel(icon.evaluate()) }
  }

  private fun createColorizedModel(icon: ColorizeProxyIcon): IconModel? {
    val rgbaColor = icon.color.toRgbaColor()
    val baseModel = createModel(icon.baseIcon)
    return when {
      rgbaColor != null && baseModel != null -> ColorizedIconModel(rgbaColor, baseModel)
      rgbaColor != null -> ColorIconModel(rgbaColor)
      baseModel != null -> baseModel
      else -> null
    }
  }
}
