package com.jetbrains.rdserver.ui

import com.intellij.notification.EventLog
import com.intellij.notification.Notification
import com.intellij.notification.NotificationType
import com.intellij.notification.Notifications
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.application.ModalityStateListener
import com.intellij.openapi.application.impl.LaterInvocator
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.progress.ProgressIndicator
import com.intellij.openapi.progress.util.ProgressWindow
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.ui.BalloonImpl
import com.jetbrains.codeWithMe.model.HostModality
import com.jetbrains.codeWithMe.model.codeWithMeSolutionExtModel
import com.jetbrains.ide.model.notifications.NotificationHyperlink
import com.jetbrains.ide.model.notifications.NotificationModel
import com.jetbrains.ide.model.notifications.NotificationsModel
import com.jetbrains.ide.model.notifications.RdNotificationEntryType
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.getHost
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.CodeWithMePluginBundle.message
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.unattendedHost.UnattendedHostManager
import java.awt.Dialog

@Suppress("NonDefaultConstructor")
class BackendNotificationsHost(
  lifetime: Lifetime,
  protocol: Protocol,
  solution: Solution,
  project: Project
) {
  companion object {
    private val LOG = logger<BackendNotificationsHost>()

    fun getInstance(project: Project): BackendNotificationsHost = RdServerPerClientComponents.getService(project)
  }

  private val knownLinks = HashMap<String, Pair<AnAction, Notification>>()
  private var linkCounter = 0
  private val codeWithMeSolutionExtModel = solution.codeWithMeSolutionExtModel

  val notificationsModel = NotificationsModel.create(lifetime, protocol)

  private var modalityDepth = 0

  init {
    if (UnattendedHostManager.getInstance().isUnattendedMode) {
      EventLog.getNotifications(project).forEach {
        notificationsModel.notification.fire(convertNotification(it))
      }
    }

    notificationsModel.executeHyperlink.advise(lifetime) { id ->
      val pair = knownLinks[id] ?: error("cannot find notification action by id $id")

      if (pair.second.isExpired)
        return@advise

      val dataContext = DataContext { dataId ->
        when (dataId) {
          CommonDataKeys.PROJECT.name -> project
          "Notification" -> pair.second
          else -> null
        }
      }

      pair.first.actionPerformed(AnActionEvent.createFromDataContext("", null, dataContext))
    }

    val sendNotifications = object : Notifications {
      override fun notify(notification: Notification) {
        if (UnattendedHostManager.getInstance().isUnattendedMode) {
          notificationsModel.notification.fire(convertNotification(notification))
        }
      }
    }
    application.messageBus.connect(lifetime.createNestedDisposable()).subscribe(Notifications.TOPIC, sendNotifications)
    project.messageBus.connect(lifetime.createNestedDisposable()).subscribe(Notifications.TOPIC, sendNotifications)

    if (LaterInvocator.getCurrentModalEntities().isNotEmpty()) {
      LOG.error("Modality stack is not empty. ${LaterInvocator.getCurrentModalEntities().joinToString { it.toString() }}")
    }
    LaterInvocator.addModalityStateListener(object : ModalityStateListener {
      override fun beforeModalityStateChanged(entering: Boolean, modalEntity: Any) {
        modalityDepth += if (entering) 1 else -1
        if (modalityDepth < 0) {
          LOG.error("Modality state was exited more times than entered")
          modalityDepth = 0
        }

        val hostName = UserManager.getInstance(project).getHost()?.name ?: message("host.defaultName")
        val modality = when (modalEntity) {
          is Dialog -> modalEntity.title?.let { message("host.modality.type.dialog", hostName, it) }
          is ProgressWindow -> modalEntity.title?.let { message("host.modality.type.progress", hostName, it) }
          is ProgressIndicator -> modalEntity.text?.let { message("host.modality.type.progress", hostName, it) }
          is BalloonImpl -> modalEntity.title?.let { message("host.modality.type.balloon", hostName, it) }
          else -> null
        }
        val modalityMessage = if (entering && modality != null) modality else message("host.modality.type.unknown", hostName)
        val title = message("host.modality.message", modalityMessage)
        codeWithMeSolutionExtModel.hostUnderModality.set(HostModality(modalityDepth, title))
      }
    }, lifetime.createNestedDisposable())
  }

  private fun convertNotification(notification: Notification): NotificationModel {
    return NotificationModel(notification.title, notification.content, true, notification.type.toProtocol(),
                             notification.actions.map { convertAction(it, notification) })
  }

  private fun convertAction(action: AnAction, notification: Notification): NotificationHyperlink {
    val currentId = (linkCounter++).toString()
    knownLinks[currentId] = action to notification
    return NotificationHyperlink(action.templateText ?: "null", currentId)
  }

  fun notifyClient(notification: Notification) {
    notificationsModel.notification.fire(convertNotification(notification))
  }

}

private fun NotificationType.toProtocol() = when (this) {
  NotificationType.INFORMATION, NotificationType.IDE_UPDATE -> RdNotificationEntryType.INFO
  NotificationType.WARNING -> RdNotificationEntryType.WARN
  NotificationType.ERROR -> RdNotificationEntryType.ERROR
}
