package com.jetbrains.rdserver.ui.converters

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.EditorKind
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.event.DocumentListener
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.impl.EditorComponentImpl
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.editor.impl.event.MarkupModelListener
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ProjectManager
import com.intellij.openapi.rd.createNestedDisposable
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ide.model.RdLanguage
import com.jetbrains.rd.ide.model.RichAttributedStringModel
import com.jetbrains.rd.ide.model.RichAttributedTextModel
import com.jetbrains.rd.ui.bedsl.extensions.UpdateFlag
import com.jetbrains.rd.ui.bedsl.extensions.ignoreOnReentrancy
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.BackendServerFileEditorManager.Companion.editorClientId
import com.jetbrains.rdserver.editors.EditorToEditorSynchronizer
import com.jetbrains.rdserver.editors.clientId

class EditorComponentImplConverter : RdConverter<EditorComponentImpl> {
  override fun convert(component: EditorComponentImpl, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
    return convertEditor(component.editor, trackOptions, lifetime)
  }
}

private fun convertEditor(editor: EditorEx, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
  val project: Project? = editor.project ?: ProjectManager.getInstance().openProjects.firstOrNull()

  val clientId = ClientId.current
  if (project != null && clientId != ClientId.localId) {
    return EditorImplHandler(trackOptions, lifetime).bind(project, editor, clientId)
  }
  else {
    logger<EditorComponentImplConverter>().warn(
      "Cannot convert an editor to a full-featured beModel, use a stub beModel instead. " +
      "project=$project, clientId=$clientId")
    return BeEditorHandler(trackOptions, lifetime).bind(editor)
  }
}

// it is expected that this class will not be called at all, so, for now it is a Plan B for unexpected cases (See usages)
private class BeEditorHandler(val trackOptions: TrackOptions, val lifetime: Lifetime) {
  private val reentrancyPreventer = UpdateFlag()

  fun bind(editor: EditorEx): BeControl {
    val richAttributedTextModel = createRichAttributedTextModel(editor)

    val beRichAttributedText = BeRichAttributedText(true)
    beRichAttributedText.text.set(richAttributedTextModel)

    val language: RdLanguage? = null

    val beEditor = BeEditor(beRichAttributedText, language)
    beEditor.settings.set(editor.toBeSettings())


    editor.document.addDocumentListener(object : DocumentListener {
      override fun documentChanged(event: DocumentEvent) {
        reentrancyPreventer.ignoreOnReentrancy {
          beEditor.text.text.set(createRichAttributedTextModel(editor))
        }
      }
    }, lifetime.createNestedDisposable())


    return beEditor
  }

  private fun createRichAttributedTextModel(editor: Editor): RichAttributedTextModel {
    val parts = mutableListOf<RichAttributedStringModel>()
    // ignoring highlightings because it's a plan B (see BeEditorHandler)
    parts.add(RichAttributedStringModel(editor.document.text, null))
    return RichAttributedTextModel(parts)
  }
}

private class EditorImplHandler(val trackOptions: TrackOptions, val lifetime: Lifetime) {
  fun bind(project: Project, editor: EditorEx, clientId: ClientId): BeControl {
    ClientId.withClientId(clientId) {
      val editorToBind: EditorEx
      if (trackOptions.canBeShared) {
        val hostEditor = editor
        editorToBind = EditorFactory.getInstance().createEditor(hostEditor.document, project, EditorKind.PREVIEW) as EditorImpl
        editorToBind.isViewer = hostEditor.isViewer
        lifetime.onTermination {
          EditorFactory.getInstance().releaseEditor(editorToBind)
        }

        copySettings(hostEditor, editorToBind)
        EditorMarkupSynchronizer(hostEditor, editorToBind, lifetime)
        EditorToEditorSynchronizer(hostEditor, editorToBind, lifetime).install()
      }
      else {
        editorToBind = editor
      }

      if (editorToBind.clientId() == ClientId.localId) {
        editorToBind.editorClientId = clientId
      }

      val textControlId = BackendEditorHost.getInstance(project).bindEditor(editorToBind)

      val beEditor = BeEditorImpl(textControlId)
      beEditor.settings.set(editorToBind.toBeSettings())
      return beEditor
    }
  }
}

private fun copySettings(fromEditor: EditorEx, toEditor: EditorEx) {
  val from = fromEditor.settings
  val to = toEditor.settings
  to.additionalColumnsCount = from.additionalColumnsCount
  to.additionalLinesCount = from.additionalLinesCount
  to.isCaretRowShown = from.isCaretRowShown
  to.isFoldingOutlineShown = from.isFoldingOutlineShown
  to.isIndentGuidesShown = from.isIndentGuidesShown
  to.isLineMarkerAreaShown = from.isLineMarkerAreaShown
  to.isLineNumbersShown = from.isLineNumbersShown
  to.isUseSoftWraps = from.isUseSoftWraps
  to.isWhitespacesShown = from.isWhitespacesShown

  toEditor.gutterComponentEx.isPaintBackground = fromEditor.gutterComponentEx.isPaintBackground
}

private fun EditorEx.toBeSettings(): BeIdeaEditorSettings {
  val settings = this.settings
  return BeIdeaEditorSettings(
    settings.additionalColumnsCount,
    settings.additionalLinesCount,
    settings.isCaretRowShown,
    settings.isFoldingOutlineShown,
    settings.isIndentGuidesShown,
    settings.isLineMarkerAreaShown,
    settings.isLineNumbersShown,
    settings.isUseSoftWraps,
    settings.isWhitespacesShown,
    gutterComponentEx.isPaintBackground
  )
}

private class EditorMarkupSynchronizer(val from: EditorEx, val to: EditorEx, val lifetime: Lifetime) : MarkupModelListener {
  private val map = mutableMapOf<RangeHighlighterEx, RangeHighlighterEx>()

  init {
    from.markupModel.addMarkupModelListener(lifetime.createNestedDisposable(), this)
    from.markupModel.processRangeHighlightersOverlappingWith(0, from.document.textLength) {
      copy(it)
      true
    }
  }

  override fun afterAdded(highlighter: RangeHighlighterEx) {
    copy(highlighter)
  }

  override fun beforeRemoved(highlighter: RangeHighlighterEx) {
    @Suppress("SSBasedInspection")
    map.remove(highlighter)?.dispose()
  }

  override fun attributesChanged(highlighter: RangeHighlighterEx,
                                 renderersChanged: Boolean,
                                 fontStyleChanged: Boolean,
                                 foregroundColorChanged: Boolean) {
    beforeRemoved(highlighter)
    afterAdded(highlighter)
  }

  private fun copy(highlighter: RangeHighlighterEx) {
    if (!highlighter.isValid) return

    val copy = to.markupModel.addRangeHighlighterAndChangeAttributes(
      highlighter.textAttributesKey, highlighter.startOffset, highlighter.endOffset,
      highlighter.layer, highlighter.targetArea, false) { copy: RangeHighlighterEx -> copy.copyFrom(highlighter) }

    map[highlighter] = copy
  }
}
