package com.jetbrains.rdserver.ui.converters

import com.intellij.openapi.application.invokeLater
import com.intellij.openapi.ui.VerticalFlowLayout
import com.intellij.uiDesigner.core.GridLayoutManager
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.ui.bedsl.dsl.verticalGrid
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rd.util.lifetime.isNotAlive
import com.jetbrains.rd.util.lifetime.onTermination
import net.miginfocom.swing.MigLayout
import java.awt.*
import java.awt.event.ComponentAdapter
import java.awt.event.ComponentEvent
import java.awt.event.ContainerEvent
import java.awt.event.ContainerListener
import javax.swing.*
import kotlin.math.roundToInt

/**
 * @author Alex Plate
 */
class BoxConverter : RdConverter<Box> {
  override fun convert(component: Box, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
    return createDynamicContainer(component, trackOptions, lifetime, ::convertPanel)
  }

  private fun convertPanel(component: Box, trackOptions: TrackOptions, lifetime: Lifetime) =
    BeJPanel(trackOptions, lifetime).buildControl(component.components.asList(), component.layout)
}


class JPanelConverter : RdConverter<JPanel> {
  override fun convert(component: JPanel, trackOptions: TrackOptions, lifetime: Lifetime): BeControl =
    createDynamicContainer(component, trackOptions, lifetime, ::convertPanel)

  private fun convertPanel(component: JPanel, trackOptions: TrackOptions, lifetime: Lifetime) =
    BeJPanel(trackOptions, lifetime).buildControl(component)
}

class BeJPanel(val trackOptions: TrackOptions, val lifetime: Lifetime) {
  fun buildControl(panel: JPanel): BeControl {
    val layout = panel.layout
    val components = panel.components.asList()
    return buildControl(components, layout)
  }

  fun buildControl(components: List<Component>, layout: LayoutManager): BeControl {
    return when (layout) {
      is BorderLayout -> borderLayoutToGrid(components, layout)
      is GridBagLayout -> gridBagToGrid(components, layout)
      is BoxLayout -> boxToGrid(components, layout)
      is CardLayout -> cardLayoutToGrid(components, layout)
      is VerticalFlowLayout -> verticalFlowLayoutToGrid(components, layout)
      is FlowLayout -> flowLayoutToGrid(components, layout)
      is MigLayout -> migLayoutToBeControl(components, layout)
      is GridLayoutManager -> gridLayoutManagerToBeControl(components, layout)
      is GridLayout -> gridLayoutToBeControl(components, layout)
      else -> {
        LOG.info("Improperly converting panel with unknown layout $layout")
        unknownGridToBeControl(components, layout)
      }
    }
  }

  private fun unknownGridToBeControl(components: List<Component>, layout: LayoutManager): BeControl {
    val grid = BeGrid(GridOrientation.Vertical)
    val noticeLabel = JLabel("Unsupported layout: " + layout.javaClass.simpleName) // NON-NLS
    grid.items.set((listOf(noticeLabel) + components)
                     .map { it.toGridElement() })
    return grid
  }

  private fun migLayoutToBeControl(components: List<Component>, layout: MigLayout): BeControl {
    fun Any?.asStringConstraint(): String {
      if (this == null)
        return ""

      if (this is String)
        return this

      LOG.info("Unknown MigLayout constraint: $this (${this.javaClass.name})")
      return ""
    }

    val beControl = BeMigLayout(
      layout.layoutConstraints.asStringConstraint(),
      layout.columnConstraints.asStringConstraint(),
      layout.rowConstraints.asStringConstraint()
    )

    val items = components.map {
      BeMigLayoutItem(
        it.toBeModel(),
        layout.getComponentConstraints(it).asStringConstraint()
      )
    }
    beControl.items.set(items)

    return beControl
  }

  /**
   * TODO stub implementation - implement correctly
   */
  private fun verticalFlowLayoutToGrid(components: List<Component>, layout: VerticalFlowLayout): BeControl {
    val gridElements = components.map { it.toGridElement() }
    val beGrid = BeGrid(GridOrientation.Vertical)

    beGrid.items.set(gridElements)

    return beGrid
  }

  private fun flowLayoutToGrid(components: List<Component>, layout: FlowLayout): BeControl {
    val gridElements = components.map { it.toGridElement() }
    val beGrid = BeGrid(GridOrientation.Horizontal)
    if (gridElements.isEmpty()) return BeSpacer()
    if (gridElements.count() == 1) return gridElements.first().content

    beGrid.items.set(gridElements)
    return beGrid
  }

  private fun borderLayoutToGrid(components: List<Component>, layout: BorderLayout): BeControl {
    val sides = BORDER_LAYOUT_SIDES.mapNotNull { side ->
      (layout.getLayoutComponent(side) as? JComponent)?.let { side to it }
    }.toMap()

    return borderLayoutToGrid(sides)
  }

  fun borderLayoutToGrid(sides: Map<String, JComponent>): BeControl {
    val visibleSides = sides.filter { it.value.isVisible }

    if (visibleSides.size == 0) return BeSpacer()
    if (visibleSides.size == 1) {
      return visibleSides.values.first().toBeModel()
    }

    val west = visibleSides[BorderLayout.WEST]?.toBeModel() to false
    val center = visibleSides[BorderLayout.CENTER]?.toBeModel() to true
    val east = visibleSides[BorderLayout.EAST]?.toBeModel() to false

    val middle = horizontalGridOf(west, center, east) to true

    val north = visibleSides[BorderLayout.NORTH]?.toBeModel() to false
    val south = visibleSides[BorderLayout.SOUTH]?.toBeModel() to false

    return verticalGridOf(north, middle, south)
  }

  private fun cardLayoutToGrid(components: List<Component>, layout: CardLayout): BeControl {
    val visibleComponent = components.find { it.isVisible } ?: components.firstOrNull()
    if (visibleComponent == null) return BeSpacer()
    return visibleComponent.toBeModel()
  }

  private fun gridBagToGrid(components: List<Component>, layout: GridBagLayout): BeControl {
    val converter = GridBagLayoutConverter(components, layout)
    if (converter.grid.size == 1) {
      return convertGrid(converter.getRow(0), GridOrientation.Horizontal)
    }
    if (converter.columnCount == 1) {
      return convertGrid(converter.getColumn(0), GridOrientation.Vertical)
    }
    return verticalGrid((0 until converter.columnCount).joinToString { if (converter.getColumnWeight(it) > 0) "*" else "auto" }) {
      for (currentRow in (0 until converter.rowCount)) {

        val row = converter.getRow(currentRow)
        val span = converter.columnCount - row.count()

        row.forEachIndexed { index, (component, constraints) ->
          val unitSize = if (constraints.weightx > 0) BeSizingType.Fill else BeSizingType.Fit
          val columnSpan = if (span > 0) if (constraints.weightx > 0) span * constraints.weightx.roundToInt() + 1 else 1 else 1
          if (index == 0)
            newRowChild(unitSize, columnSpan = columnSpan, expandGrid = false) { component.toBeModel() }
          else {
            existingRowChild(unitSize, columnSpan = columnSpan, expandGrid = false) { component.toBeModel() }
          }
        }
      }
    }
  }

  private fun convertGrid(components: List<Pair<JComponent, GridBagConstraints>>,
                          orientation: GridOrientation): BeGrid {
    val grid = BeGrid(orientation)
    val gridElements = mutableListOf<BeGridElement>()
    for ((component, constraints) in components) {
      val unitSize = BeUnitSize(if (constraints.weightx > 0) BeSizingType.Fill else BeSizingType.Fit)
      gridElements.add(BeGridElement(component.toBeModel(), unitSize))
    }
    grid.items.set(gridElements)
    return grid
  }

  private fun boxToGrid(components: List<Component>, layout: BoxLayout): BeControl {
    val orientation = when (layout.axis) {
      BoxLayout.X_AXIS, BoxLayout.LINE_AXIS -> GridOrientation.Horizontal
      else -> GridOrientation.Vertical
    }

    val grid = BeGrid(orientation)
    val gridElements = mutableListOf<BeGridElement>()
    for (component in components) {
      gridElements.add(component.toGridElement())
    }
    grid.items.set(gridElements)
    return grid
  }

  private fun gridLayoutManagerToBeControl(components: List<Component>, layout: GridLayoutManager): BeControl {
    var currentRow = -1
    return verticalGrid((1..layout.columnCount).joinToString { "*" }) {
      components.forEach { component ->
        if (!component.isVisible || component !is JComponent) return@forEach
        val constraints = layout.getConstraintsForComponent(component)
        if (constraints.row > currentRow) {
          newRowChild(columnSpan = constraints.colSpan, expandGrid = false) { component.toBeModel() }
        }
        else {
          existingRowChild(columnSpan = constraints.colSpan) { component.toBeModel() }
        }
        currentRow = constraints.row
      }
    }
  }

  private fun gridLayoutToBeControl(components: List<Component>, layout: GridLayout): BeControl {
    var currentCol = 0
    return verticalGrid((1..layout.columns).joinToString { "*" }) {
      components.forEach { component ->
        if (!component.isVisible || component !is JComponent) return@forEach

        if (currentCol == layout.columns) {
          currentCol = 0
          newRowChild(columnSpan = layout.vgap, expandGrid = false) { component.toBeModel() }
        }
        else {
          currentCol++
          existingRowChild(columnSpan = layout.vgap) { component.toBeModel() }
        }
      }
    }
  }

  private fun Component.toGridElement(): BeGridElement =
    BeGridElement(this.toBeModel(), BeUnitSize(BeSizingType.Fit))

  private fun Component.toBeModel(): BeControl = this.toBeModel(trackOptions, lifetime)

  companion object {
    private val LOG = getLogger<BeJPanel>()
  }
}


val BORDER_LAYOUT_SIDES = arrayOf(
  BorderLayout.CENTER,
  BorderLayout.NORTH,
  BorderLayout.SOUTH,
  BorderLayout.EAST,
  BorderLayout.WEST
)

/**
 * Create vertical [BeGrid] of [BeControl]s if it's not null.
 * [controls] - pair of [BeControl] to autoResize
 * [BeSizingType.Fill] if true, [BeSizingType.Fit] otherwise
 */
fun verticalGridOf(vararg controls: Pair<BeControl?, Boolean>): BeControl = gridOf(GridOrientation.Vertical, *controls)

/**
 * Create horizontal [BeGrid] of [BeControl]s if it's not null.
 * [controls] - pair of [BeControl] to autoResize
 * [BeSizingType.Fill] if true, [BeSizingType.Fit] otherwise
 */
fun horizontalGridOf(vararg controls: Pair<BeControl?, Boolean>): BeControl = gridOf(GridOrientation.Horizontal, *controls)

private fun gridOf(orientation: GridOrientation, vararg controls: Pair<BeControl?, Boolean>): BeControl {
  val elements = controls.filter { (c, _) -> c != null }
  if (elements.count() == 1) return elements.first().first!!
  return BeGrid(orientation).apply {
    items.set(
      controls.mapNotNull { (control, autoSize) ->
        control?.let { BeGridElement(it, BeUnitSize(if (autoSize) BeSizingType.Fill else BeSizingType.Fit)) }
      }
    )
  }
}

class GridBagLayoutConverter(components: List<Component>, private val layout: GridBagLayout) {
  val grid = mutableListOf<MutableList<JComponent?>>()

  init {
    var prevX = -1
    var prevY = -1

    for (component in components) {
      if (!component.isVisible || component !is JComponent) continue
      val constraints = layout.getConstraints(component)

      var curX = constraints.gridx
      var curY = constraints.gridy
      if (curX < 0) {
        curX = prevX + 1
      }
      if (curY < 0) {
        curY = prevY + 1
      }
      storeComponent(curX, curY, component)
      prevX = curX
      prevY = curY
    }
  }

  private fun storeComponent(x: Int, y: Int, component: JComponent) {
    while (grid.size <= y) {
      grid.add(mutableListOf())
    }
    val row = grid[y]
    while (row.size <= x) {
      row.add(null)
    }
    row[x] = component
  }

  val columnCount
    get() = grid.map { it.size }.maxOrNull() ?: 0

  val rowCount
    get() = grid.count()

  fun getRow(row: Int): List<Pair<JComponent, GridBagConstraints>> {
    return grid[row].filterNotNull().map { it to layout.getConstraints(it) }
  }

  fun getColumn(column: Int): List<Pair<JComponent, GridBagConstraints>> {
    return grid.mapNotNull { it.getOrNull(column) }.map { it to layout.getConstraints(it) }
  }

  fun getColumnWeight(column: Int): Double {
    return grid.mapNotNull { it.getOrNull(column) }.map { layout.getConstraints(it).weightx }.minOrNull() ?: 1.0
  }
}

fun <T : Container> createDynamicContainer(
  panel: T, trackOptions: TrackOptions, lifetime: Lifetime,
  builder: (panel: T, trackOptions: TrackOptions, lifetime: Lifetime) -> BeControl): BeDynamicControlContainer {
  return DynamicContainerHandler(panel, trackOptions, lifetime, builder).beControl
}

private class DynamicContainerHandler<T : Container>(
  component: T, val trackOptions: TrackOptions, val lifetime: Lifetime,
  builder: (component: T, trackOptions: TrackOptions, lifetime: Lifetime) -> BeControl) {
  val beControl: BeDynamicControlContainer = BeDynamicControlContainer()

  private var updateScheduled: Boolean = false

  init {
    val control = builder(component, trackOptions, lifetime)
    beControl.content.set(control)

    if (trackOptions.componentChangeTrack) {
      trackChildrenChanges(component, builder)
    }
  }

  private fun trackChildrenChanges(component: T, builder: (component: T, trackOptions: TrackOptions, lifetime: Lifetime) -> BeControl) {
    val componentListener = object : ComponentAdapter() {
      override fun componentShown(e: ComponentEvent?) {
        scheduleRebuild(component, builder)
      }

      override fun componentHidden(e: ComponentEvent?) {
        scheduleRebuild(component, builder)
      }
    }

    val containerListener = object : ContainerListener {
      override fun componentAdded(e: ContainerEvent) {
        (e.source as? Component)?.addComponentListener(componentListener)
        scheduleRebuild(component, builder)
      }

      override fun componentRemoved(e: ContainerEvent) {
        (e.source as? Component)?.removeComponentListener(componentListener)
        scheduleRebuild(component, builder)
      }
    }


    component.addContainerListener(containerListener)
    component.components.forEach { it.addComponentListener(componentListener) }

    lifetime.onTermination {
      component.removeContainerListener(containerListener)
      component.components.forEach { it.removeComponentListener(componentListener) }
    }
  }

  private fun scheduleRebuild(component: T, builder: (component: T, trackOptions: TrackOptions, lifetime: Lifetime) -> BeControl) {
    if (lifetime.isNotAlive) return

    if (updateScheduled) return
    updateScheduled = true
    invokeLater {
      updateScheduled = false
      if (lifetime.isAlive) {
        val control = builder(component, trackOptions, lifetime)
        beControl.content.set(control)
      }
    }
  }
}
