package com.jetbrains.rdserver.ui.converters

import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.ui.JBTabsPaneImpl
import com.intellij.ui.TabbedPane
import com.intellij.ui.TabbedPaneWrapper
import com.intellij.ui.content.Content
import com.intellij.ui.content.ContentManagerEvent
import com.intellij.ui.content.ContentManagerListener
import com.intellij.ui.content.TabbedPaneContentUI
import com.intellij.ui.tabs.JBTabs
import com.intellij.ui.tabs.TabInfo
import com.intellij.ui.tabs.TabsListener
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.ui.bedsl.dsl.label
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.toProtocolModel
import java.beans.PropertyChangeEvent

class TabbedPaneConverter : RdConverter<TabbedPaneWrapper.TabbedPaneHolder> {
  override fun convert(component: TabbedPaneWrapper.TabbedPaneHolder, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
    val wrapper = component.tabbedPaneWrapper
    val tabbedPane = wrapper.tabbedPane
    return when {
      wrapper is TabbedPaneContentUI.MyTabbedPaneWrapper -> return BeTabbedPaneContentUIHandler(wrapper, trackOptions, lifetime).control
      tabbedPane is JBTabsPaneImpl -> return JBTabsConverter().convert(tabbedPane.tabs, trackOptions, lifetime)
      else -> BeTabbedPaneWrapperHandler(tabbedPane, trackOptions, lifetime).control
    }
  }
}

class JBTabsConverter : RdConverter<JBTabs> {
  override fun convert(component: JBTabs, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
    if (component.presentation.isHideTabs) {
      return BeHiddenJBTabsHandler(component, trackOptions, lifetime).control
    }
    else {
      return BeJBTabsHandler(component, trackOptions, lifetime).control
    }
  }
}


class BeTabbedPaneContentUIHandler(tabs: TabbedPaneContentUI.MyTabbedPaneWrapper, val trackOptions: TrackOptions, val lifetime: Lifetime) {
  val control = BeTabbedControl()

  private var processingContents: Boolean = false

  init {
    updateTabs(tabs)

    if (trackOptions.modelChangeTrack) {
      val contentManagerListener = object : ContentManagerListener {
        override fun contentAdded(event: ContentManagerEvent) {
          noReenter(::processingContents) {
            val tabControl = createTabModel(tabs, event.index)
            control.tabs.add(event.index, tabControl)
          }
        }

        override fun contentRemoved(event: ContentManagerEvent) {
          noReenter(::processingContents) {
            control.tabs.removeAt(event.index)
          }
        }

        override fun selectionChanged(event: ContentManagerEvent) {
          noReenter(::processingContents) {
            updateSelection(tabs)
          }
        }
      }
      val contentManager = tabs.contentManager
      contentManager.addContentManagerListener(contentManagerListener)
      lifetime.onTermination { contentManager.removeContentManagerListener(contentManagerListener) }

      control.tabs.view(lifetime) { tabLifetime, index, _ ->
        val contentChangeListener: (evt: PropertyChangeEvent) -> Unit = { e ->
          if (e.propertyName == Content.PROP_COMPONENT) {
            noReenter(::processingContents) {
              val tabControl = createTabModel(tabs, index)
              control.tabs.set(index, tabControl)
            }
          }
        }
        val content = contentManager.getContent(index) ?: return@view
        content.addPropertyChangeListener(contentChangeListener)
        tabLifetime.onTermination { content.removePropertyChangeListener(contentChangeListener) }
      }

      control.selectedIndex.advise(lifetime) { index ->
        noReenter(::processingContents) {
          val content = contentManager.getContent(index) ?: return@noReenter
          contentManager.setSelectedContent(content, false)
        }
      }
    }
  }

  private fun updateTabs(tabs: TabbedPaneContentUI.MyTabbedPaneWrapper) {
    control.tabs.clear()
    for (i in 0 until tabs.tabCount) {
      val tabControl = createTabModel(tabs, i)
      control.tabs.add(tabControl)
    }
    updateSelection(tabs)
  }

  private fun updateSelection(tabs: TabbedPaneContentUI.MyTabbedPaneWrapper) {
    control.selectedIndex.set(tabs.selectedIndex.coerceAtLeast(0))
  }

  private fun createTabModel(tabs: TabbedPaneContentUI.MyTabbedPaneWrapper, index: Int): BeTab {
    val title = tabs.getTitleAt(index)
    val contentModel = tabs.getComponentAt(index).toBeModel(trackOptions, lifetime)
    return BeTab(tabName = label(title),
                 content = contentModel)
  }
}

class BeJBTabsHandler(tabs: JBTabs, val trackOptions: TrackOptions, val lifetime: Lifetime) {
  val control = BeTabbedControl()

  private var processingContents: Boolean = false

  init {
    updateTabs(tabs)

    if (trackOptions.modelChangeTrack) {
      tabs.addListener(object : TabsListener {
        override fun selectionChanged(oldSelection: TabInfo?, newSelection: TabInfo?) {
          noReenter(::processingContents) {
            updateSelection(tabs)
          }
        }
      }, lifetime.createNestedDisposable())

      control.selectedIndex.advise(lifetime) { index ->
        noReenter(::processingContents) {
          if (index >= 0 && index < tabs.tabCount) {
            tabs.select(tabs.getTabAt(index), false)
          }
        }
      }
    }
  }

  private fun updateTabs(tabs: JBTabs) {
    control.tabs.clear()
    for (i in 0 until tabs.tabCount) {
      val tabControl = createTabModel(tabs, i)
      control.tabs.add(tabControl)
    }
    updateSelection(tabs)
  }

  private fun updateSelection(tabs: JBTabs) {
    val selectedInfo = tabs.selectedInfo
    val selectedIndex = tabs.getIndexOf(selectedInfo)
    control.selectedIndex.set(selectedIndex.coerceAtLeast(0))
  }

  private fun createTabModel(tabs: JBTabs, index: Int): BeTab {
    val tab = tabs.getTabAt(index)
    val title = tab.text
    val icon = tab.icon
    val contentModel = tab.component.toBeModel(trackOptions, lifetime)
    return BeTab(tabName = label(title, icon?.toProtocolModel()),
                 content = contentModel)
  }
}

class BeHiddenJBTabsHandler(tabs: JBTabs, val trackOptions: TrackOptions, val lifetime: Lifetime) {
  val control: BeDynamicControlContainer = BeDynamicControlContainer()

  private var processingContents: Boolean = false

  init {
    updateTabs(tabs)

    if (trackOptions.modelChangeTrack) {
      tabs.addListener(object : TabsListener {
        override fun selectionChanged(oldSelection: TabInfo?, newSelection: TabInfo?) {
          noReenter(::processingContents) {
            updateTabs(tabs)
          }
        }
      }, lifetime.createNestedDisposable())
    }
  }

  private fun updateTabs(tabs: JBTabs) {
    val tab = tabs.selectedInfo
    if (tab != null) {
      control.content.set(tab.component.toBeModel(trackOptions, lifetime))
    }
    else {
      control.content.set(BeSpacer())
    }
  }
}

class BeTabbedPaneWrapperHandler(tabs: TabbedPane, val trackOptions: TrackOptions, val lifetime: Lifetime) {
  val control = BeTabbedControl()

  init {
    updateTabs(tabs)
  }

  private fun updateTabs(tabs: TabbedPane) {
    control.tabs.clear()
    for (i in 0 until tabs.tabCount) {
      val tabControl = createTabModel(tabs, i)
      control.tabs.add(tabControl)
    }
    updateSelection(tabs)
  }

  private fun updateSelection(tabs: TabbedPane) {
    control.selectedIndex.set(tabs.selectedIndex.coerceAtLeast(0))
  }

  private fun createTabModel(tabs: TabbedPane, index: Int): BeTab {
    val title = tabs.getTitleAt(index)
    val contentModel = tabs.getComponentAt(index).toBeModel(trackOptions, lifetime)
    return BeTab(tabName = label(title),
                 content = contentModel)
  }
}