package com.jetbrains.rdserver.ui.converters

import com.intellij.codeWithMe.ClientId
import java.awt.Component

/**
 * @author Alex Plate
 *
 * This class allows to configure converting from swing components to be controls and backwards
 *
 * [componentChangeTrack] and [modelChangeTrack] - `true` if the state of the swing component (or model) should be mapped
 * to the model (or component). Basically it means that by [componentChangeTrack] eq `false` all listeners are skipped
 * and by [modelChangeTrack] eq `false` all `advise`-methods are disabled.
 *
 * [beforeComponentChange] and [afterComponentChange] are actions that will be performed before and after mapping by component change.
 * `null` if no actions are required. There actions are disabled if [componentChangeTrack] is `false`
 *
 * [beforeModelChange] and [afterModelChange] are actions that will be performed before and after mapping by model change.
 * `null` if no actions are required. There actions are disabled if [modelChangeTrack] is `false`.
 * E.g. enable editing mode for tables with [beforeModelChange] and exiting it with [afterModelChange]
 *
 * [modelChangeComponentProvider] - code that provides a component to be changed if it's not the same as original component.
 * `null` should be returned if original component should be changed during mapping.
 * E.g. Using JTable, the original component in cell should not be changed, but the one, that will be returned by `getTableCellEditorComponent`
 *
 * [canBeShared] - whether same swing components can be mapped to controls multiple times independently
 * [clientId] - provides ClientId that should be used for created components
 *
 * [TrackOptions.instance] allows you to get instance of this class with enabled tracking for model and component and without any
 * actions before and after mapping.
 */
class TrackOptions(
  val componentChangeTrack: Boolean = true,
  val beforeComponentChange: (() -> Unit)? = null,
  val afterComponentChange: (() -> Unit)? = null,

  val modelChangeTrack: Boolean = true,
  val beforeModelChange: (() -> Unit)? = null,
  val modelChangeComponentProvider: (() -> Component)? = null,
  val afterModelChange: (() -> Unit)? = null,

  val canBeShared: Boolean = false,
  val clientId: ClientId? = null
) {
  companion object {
    val instance = TrackOptions()
    val trackModelChangeOnly = TrackOptions(componentChangeTrack = false)
    val trackComponentChangeOnly = TrackOptions(modelChangeTrack = false)
    val doNotTrack = TrackOptions(componentChangeTrack = false,
                                  modelChangeTrack = false)

    fun TrackOptions.pinClientId(): TrackOptions = when {
      clientId != null -> this
      else -> copy(ClientId.current)
    }

    fun TrackOptions.copy(clientId: ClientId?): TrackOptions {
      return TrackOptions(componentChangeTrack = componentChangeTrack,
                          beforeComponentChange = beforeComponentChange,
                          afterComponentChange = afterComponentChange,
                          modelChangeTrack = modelChangeTrack,
                          beforeModelChange = beforeModelChange,
                          modelChangeComponentProvider = modelChangeComponentProvider,
                          afterModelChange = afterModelChange,
                          canBeShared = canBeShared,
                          clientId = clientId)
    }
  }
}