package com.jetbrains.rdserver.ui.converters.toTreeGrid

import com.intellij.ui.ExpandedItemListCellRendererWrapper
import com.intellij.ui.components.JBViewport
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.ui.bedsl.extensions.UpdateFlag
import com.jetbrains.rd.ui.bedsl.extensions.ignoreOnReentrancy
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.ui.converters.RdConverter
import com.jetbrains.rdserver.ui.converters.TrackOptions
import com.jetbrains.rdserver.ui.converters.toBeModel
import com.jetbrains.rdserver.ui.converters.toTreeGrid.JListConverter.Companion.loadRow
import javax.swing.JList
import javax.swing.ListSelectionModel
import javax.swing.SwingUtilities
import javax.swing.event.ListDataEvent
import javax.swing.event.ListDataListener
import kotlin.math.min

class JListConverter : RdConverter<JList<Any?>> {
  private val selectionUpdate = UpdateFlag()
  override fun convert(component: JList<Any?>, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {

    val selectionModel = component.selectionModel
    val selectionKind = if (selectionModel.selectionMode == ListSelectionModel.SINGLE_SELECTION)
      BeTreeSelection.Single
    else
      BeTreeSelection.Multi

    val model = component.getModel()
    val modelRows = model?.size ?: 0
    var visibleRows: Int = component.visibleRowCount
    val forceVisibleRowCount = UIUtil.getClientProperty(component, JBViewport.FORCE_VISIBLE_ROW_COUNT_KEY)
    if (forceVisibleRowCount != null && !forceVisibleRowCount && visibleRows > 0) {
      visibleRows = min(modelRows, visibleRows)
    }

    val beTreeConfiguration = BeTreeConfiguration(
      listOf(BeColumnDefinition("", BeUnitSize(BeSizingType.Fit))),
      selectionKind,
      false,
      ChildOffsetSize.DEFAULT,
      true,
      false,
      visibleRowsCount = visibleRows,
      showBorders = BeShowBorders.None
    )

    val root = BeTreeGridLine(emptyList())
    val table = BeTreeGrid(root, beTreeConfiguration)

    fillContent(component, table, lifetime)
    component.addPropertyChangeListener("model") {
      if (it.newValue !is List<Any?>) return@addPropertyChangeListener
      SwingUtilities.invokeLater {
        table.fakeRoot.children.clear()
        fillContent(component, table, lifetime)
      }
    }

    return table
  }

  private fun fillContent(component: JList<Any?>,
                          table: BeTreeGrid,
                          lifetime: Lifetime) {
    val root = table.fakeRoot
    for (idx in 0 until component.model.size) {
      root.children.add(loadRow(idx, component, selectionUpdate))
    }
    component.model.addListDataListener(
      MyListModelListener(table, component, lifetime, selectionUpdate))
    component.addListSelectionListener {
      selectionUpdate.ignoreOnReentrancy {
        for (i in 0 until component.model.size) {
          val line = root.children[i] as BeTreeGridLine
          line.isSelected.set(component.isSelectedIndex(i))
        }
      }
    }
    if(component.selectedIndex < 0)
      component.selectedIndex = 0
  }

  companion object {
    internal fun loadRow(idx: Int, component: JList<Any?>, selectionUpdate: UpdateFlag): BeTreeGridLine {
      val listModel = component.model
      val data = listModel.getElementAt(idx)
      val renderer = component.cellRenderer as ExpandedItemListCellRendererWrapper

      val cellComponent = renderer.wrappee.getListCellRendererComponent(component, data, idx, false, true)


      val rowModel = BeTreeGridLine(listOf(cellComponent.toBeModel(TrackOptions.doNotTrack)))
      rowModel.isSelected.advise(Lifetime.Eternal) {
        selectionUpdate.ignoreOnReentrancy {
          if (it) {
            component.setSelectedValue(data, false)
          }
        }
      }

      if (component.selectionModel.isSelectedIndex(idx)) {
        rowModel.isSelected.set(true)
      }
      return rowModel
    }
  }
}

private class MyListModelListener(val table: BeTreeGrid,
                                  val component: JList<Any?>,
                                  val lifetime: Lifetime, private val selectionUpdate:UpdateFlag) : ListDataListener {
  override fun contentsChanged(e: ListDataEvent?) {
    e ?: return
    TODO()
  }

  override fun intervalRemoved(e: ListDataEvent?) {
    e ?: return
    for (i in e.index1 downTo e.index0) {
      table.fakeRoot.children.removeAt(i)
    }
  }

  override fun intervalAdded(e: ListDataEvent?) {
    e ?: return
    val root = table.fakeRoot
    val oldSize = root.children.size
    for (i in e.index0..e.index1) {
      val element = loadRow(i, component, selectionUpdate)
      if (e.index0 >= oldSize)
        root.children += element
      else
        root.children.add(i, element)
    }
  }
}