package com.jetbrains.rdserver.ui.converters.toTreeGrid

import com.intellij.usageView.UsageViewBundle
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.ui.converters.RdConverter
import com.jetbrains.rdserver.ui.converters.TrackOptions
import com.jetbrains.rdserver.ui.converters.toBeModel
import javax.swing.JTable
import javax.swing.ListSelectionModel
import javax.swing.SwingUtilities
import javax.swing.event.CellEditorListener
import javax.swing.event.ChangeEvent
import javax.swing.event.TableModelEvent
import javax.swing.event.TableModelListener
import javax.swing.table.TableModel

class JTableConverter : RdConverter<JTable> {

  override fun convert(component: JTable, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
    val columns = if (isShowUsagesTable(component)) {
      // tmp hack to fix ShowUsagesTable with dynamic columns count, creates max number of columns
      MutableList(4) { BeColumnDefinition("", BeUnitSize(BeSizingType.Fit)) }
    } else {
      component.columnModel.columns.toList().map { column ->
        BeColumnDefinition(column.headerValue?.toString().orEmpty(), BeUnitSize(BeSizingType.Fit))
      }
    }

    val selectionModel = component.selectionModel
    val selectionKind = if (selectionModel.selectionMode == ListSelectionModel.SINGLE_SELECTION)
      BeTreeSelection.Single
    else
      BeTreeSelection.Multi

    val beTreeConfiguration = BeTreeConfiguration(
      columns,
      selectionKind,
      true,
      ChildOffsetSize.DEFAULT,
      true,
      false,
      visibleRowsCount = component.rowCount
    )

    val root = BeTreeGridLine(emptyList())

    val table = BeTreeGrid(root, beTreeConfiguration)
    fillContent(component, root, lifetime, table)

    component.addPropertyChangeListener("model") {
      if (it.newValue !is TableModel) return@addPropertyChangeListener

      SwingUtilities.invokeLater {
        root.children.clear()
        fillContent(component, root, lifetime, table)
      }
    }
    return table
  }

  private fun isShowUsagesTable(component: JTable): Boolean {
    val model = component.model
    return model.rowCount > 0 && model.columnCount > 0 &&
           model.getValueAt(0, 0).toString() == UsageViewBundle.message("progress.searching")
  }

  private fun fillContent(component: JTable,
                          root: BeTreeGridLine,
                          lifetime: Lifetime,
                          table: BeTreeGrid) {
    for (row in 0 until component.rowCount) {
      root.children += loadRow(row, component, lifetime)
    }
    component.model.addTableModelListener(RdTableModelListener(table, component, lifetime){loadRow(it, component, lifetime)})
  }

  companion object {
    internal fun loadRow(row: Int, component: JTable, lifetime: Lifetime): BeTreeGridLine {
      val controls = List(component.columnCount) { column ->
        val data = component.getValueAt(row, column)

        if (component.isCellEditable(row, column)) {
          val cellEditor = component.getCellEditor(row, column)
          val editorComponent = cellEditor.getTableCellEditorComponent(component, data, false, row, column)

          // todo don't use row and column into closure, they can be changed
          val componentTrackOptions = TrackOptions(
            modelChangeTrack = true,
            modelChangeComponentProvider = {
              val componentData = component.getValueAt(row, column)

              val componentCellEditor = component.getCellEditor(row, column)
              componentCellEditor.getTableCellEditorComponent(component, componentData, false, row, column)
            },
            afterModelChange = {
              val componentCellEditor = component.getCellEditor(row, column)
              componentCellEditor.addCellEditorListener(RdTableCellEditorListener(row, column, component))
              componentCellEditor.stopCellEditing()
            }
          )

          editorComponent.toBeModel(componentTrackOptions, lifetime)
        } else {
          val cellRenderer = component.getCellRenderer(row, column)
          val cellComponent = cellRenderer.getTableCellRendererComponent(component, data, false, true, row, column)
          cellComponent.toBeModel(TrackOptions.doNotTrack, lifetime = lifetime)
        }
      }
      val rowModel = BeTreeGridLine(controls)
      rowModel.isSelected.advise(lifetime) {
        if (it) {
          component.setRowSelectionInterval(row, row)
        }
      }

      if (component.selectionModel.isSelectedIndex(row)) {
        rowModel.isSelected.set(true)
      }
      return rowModel
    }
  }
}

internal class RdTableModelListener(val table: BeTreeGrid,
                                    val component: JTable,
                                    val lifetime: Lifetime,
                                    val loadRow: (Int) -> BeTreeGridLine) : TableModelListener {
  override fun tableChanged(e: TableModelEvent) {
    if (e.column != TableModelEvent.ALL_COLUMNS) {
      // TODO
      return
    }

    val root = table.fakeRoot
    when (e.type) {
      TableModelEvent.UPDATE -> {
        if (e.lastRow == Int.MAX_VALUE) {
          root.children.clear()
          for (i in 0 until component.rowCount) {
            root.children.add(loadRow(i))
          }
        }
        else {
          for (i in e.firstRow..e.lastRow) {
            root.children[i] = loadRow(i)
          }
        }
      }
      TableModelEvent.INSERT -> {
        val oldSize = root.children.size
        require(e.firstRow <= oldSize) { "first row ${e.firstRow} <= old size ${oldSize}" }
        val append = e.firstRow == oldSize
        for (i in e.firstRow..e.lastRow) {
          val new = loadRow(i)
          if (append) {
            root.children += new
          } else {
            root.children.add(i, new)
          }
        }
      }
      TableModelEvent.DELETE -> {
        for (i in e.lastRow downTo e.firstRow) {
          root.children.removeAt(i)
        }
      }
    }
  }
}

internal class RdTableCellEditorListener(
  private val row: Int,
  private val column: Int,
  private val component: JTable
) : CellEditorListener {
  override fun editingStopped(e: ChangeEvent?) {
    val cellEditor = component.getCellEditor(row, column)
    val cellEditorValue = cellEditor.cellEditorValue
    component.setValueAt(cellEditorValue, row, column)

    cellEditor.removeCellEditorListener(this)
  }

  override fun editingCanceled(e: ChangeEvent?) {
    component.editingCanceled(e)
  }
}

