package com.jetbrains.rdserver.ui.converters.toTreeGrid

import com.intellij.ui.ClickListener
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.ui.bedsl.dsl.spacer
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rdserver.ui.converters.RdConverter
import com.jetbrains.rdserver.ui.converters.TrackOptions
import com.jetbrains.rdserver.ui.converters.toBeModel
import java.awt.KeyboardFocusManager
import java.awt.event.KeyEvent
import javax.swing.JTree
import javax.swing.event.*
import javax.swing.tree.TreePath
import kotlin.math.roundToInt

/**
 * @author Alex Plate
 */
class JTreeConverter : RdConverter<JTree> {
  /**
   * [trackOptions] are supported partially
   */
  override fun convert(component: JTree, trackOptions: TrackOptions, lifetime: Lifetime): BeControl =
    BeJTree(component, lifetime, trackOptions).beControl
}

class BeJTree(val tree: JTree, val lifetime: Lifetime, val trackOptions: TrackOptions) {
  val beControl: BeTreeGrid
  private val loadAsync = true
  private var processingSelectionChange = false
  private var processingExpandChange = false
  private val treeMap = mutableMapOf<TreePath, BeTreeGridLine>()

  private val rootGridLine: BeTreeGridLine

  init {
    rootGridLine = BeTreeGridLine(mutableListOf()).apply { isExpanded.value = true }
    beControl = BeTreeGrid(
      rootGridLine,
      BeTreeConfiguration(
        listOf(BeColumnDefinition("", BeUnitSize(BeSizingType.Fill))),
        BeTreeSelection.Single, false, ChildOffsetSize.DEFAULT, false, false, visibleRowsCount = tree.visibleRowCount
      )
    )

    beControl.enter.advise(lifetime) {
      val inputEvent = KeyEvent(tree, KeyEvent.KEY_PRESSED, System.currentTimeMillis(), 0, KeyEvent.VK_ENTER, '\r')
      KeyboardFocusManager.getCurrentKeyboardFocusManager().redispatchEvent(tree, inputEvent)
    }

    resetTreeStructure()

    if (trackOptions.componentChangeTrack) {
      tree.addTreeSelectionListener { event -> passSelectionToModel(event) }
      tree.addTreeExpansionListener(object : TreeExpansionListener {
        override fun treeExpanded(event: TreeExpansionEvent) {
          passExpansionToModel(event, true)
        }

        override fun treeCollapsed(event: TreeExpansionEvent) {
          passExpansionToModel(event, false)
        }
      })
    }

    if (trackOptions.modelChangeTrack) {
      tree.model.addTreeModelListener(MyTreeModelListener())
    }
  }

  private fun convertTreeModelLevel(node: Any, treePath: TreePath, loadChildren: Boolean = true): BeTreeGridLine {
    val gridLine = createGridLine(node, treePath)
    treeMap[treePath] = gridLine

    if (trackOptions.componentChangeTrack) {
      gridLine.isSelected.change.adviseNotNull(lifetime) { selected -> passSelectionToSwing(treePath, selected) }
      gridLine.isExpanded.change.adviseNotNull(lifetime) { expanded -> passExpansionToSwing(treePath, expanded) }
    }

    if (loadChildren) {
      loadNodeChildren(node, gridLine, treePath)
    }

    return gridLine
  }

  private fun loadNodeChildren(node: Any, gridLine: BeTreeGridLine, treePath: TreePath) {
    if (tree.model.isLeaf(node)) return
    if (loadAsync && !gridLine.isExpanded.value) {
      gridLine.children.add(createLoadingLine())

      gridLine.isExpanded.adviseNotNull(lifetime) { expanded ->
        if (expanded) {
          ensureChildrenLoaded(gridLine, treePath)
        }
      }
    }
    else {
      reloadNodeChildren(treePath, gridLine)
    }
  }

  private fun createGridLine(node: Any, treePath: TreePath): BeTreeGridLine {
    val leaf = tree.model.isLeaf(node)
    val renderComponent = tree.cellRenderer.getTreeCellRendererComponent(tree, node, false, true, leaf, 0, false)
    val columnElements = listOf(renderComponent.toBeModel(TrackOptions.doNotTrack, lifetime))
    val gridLine = BeTreeGridLine(columnElements)
    gridLine.isSelected.set(tree.isPathSelected(treePath))
    gridLine.isExpanded.set(tree.isExpanded(treePath))
    gridLine.signal.advise(lifetime) {
      val bounds = tree.getPathBounds(treePath) ?: return@advise
      tree.selectionPath = treePath
      val event = ClickListener.SyntheticClickEvent(tree, System.currentTimeMillis(), 0,
                                                    bounds.centerX.roundToInt(), bounds.centerY.roundToInt(),
                                                    2, false, 1)
      tree.dispatchEvent(event)
    }
    return gridLine
  }

  private fun createLoadingLine() = BeTreeGridLine(mutableListOf(spacer()))

  private fun isNotLoaded(gridLine: BeTreeGridLine): Boolean {
    val singleChild = gridLine.children.singleOrNull() as? BeTreeGridLine ?: return false
    return singleChild.columnElements.singleOrNull() is BeSpacer
  }

  private fun ensureChildrenLoaded(gridLine: BeTreeGridLine, treePath: TreePath) {
    if (isNotLoaded(gridLine)) {
      modifyModel {
        reloadNodeChildren(treePath, gridLine)
      }
    }
  }

  private fun resetTreeStructure() {
    rootGridLine.children.clear()
    treeMap.clear()

    val root = tree.model.root ?: return
    val rootPath = TreePath(root)

    if (tree.isRootVisible) {
      rootGridLine.children.add(convertTreeModelLevel(root, rootPath))
    }
    else {
      treeMap[rootPath] = rootGridLine
      reloadNodeChildren(rootPath, rootGridLine)
    }
  }

  private fun reloadNodeChildren(treePath: TreePath, gridLine: BeTreeGridLine) {
    gridLine.children.clear()

    val node = treePath.lastPathComponent
    for (childIndex in 0 until tree.model.getChildCount(node)) {
      val child = tree.model.getChild(node, childIndex)
      gridLine.children.add(convertTreeModelLevel(child, treePath.pathByAddingChild(child)))
    }
  }


  private fun passSelectionToModel(event: TreeSelectionEvent) {
    noReenter(::processingSelectionChange) {
      for ((index, path) in event.paths.withIndex()) {
        val gridLine = treeMap[path] ?: run {
          LOG.info("Could not find BeTreeGridLine for path $path")
          return@noReenter
        }
        gridLine.isSelected.set(event.isAddedPath(index))
      }
    }
  }

  private fun passExpansionToModel(event: TreeExpansionEvent, expanded: Boolean) {
    noReenter(::processingExpandChange) {
      val path: TreePath = event.path
      val gridLine = treeMap[path] ?: run {
        LOG.info("Could not find BeTreeGridLine for path $path")
        return@noReenter
      }
      gridLine.isExpanded.set(expanded)
    }
  }

  private fun passSelectionToSwing(treePath: TreePath, selected: Boolean) {
    noReenter(::processingSelectionChange) {
      if (selected) {
        tree.selectionModel.addSelectionPath(treePath)
      }
      else {
        tree.selectionModel.removeSelectionPath(treePath)
      }
    }
  }

  private fun passExpansionToSwing(treePath: TreePath, expanded: Boolean) {
    noReenter(::processingExpandChange) {
      if (expanded) {
        tree.expandPath(treePath)
      }
      else {
        tree.collapsePath(treePath)
      }
    }
  }


  private inner class MyTreeModelListener : TreeModelListener {
    override fun treeNodesChanged(e: TreeModelEvent) {
      modifyModel {
        if (e.childIndices == null) {
          if (tree.isRootVisible) {
            val node = tree.model.root
            val treePath = TreePath(node)
            updateNodePresentation(node, treePath)
          }
        }
        else {
          val treePath = e.treePath
          for (index in e.childIndices) {
            val child = tree.model.getChild(treePath.lastPathComponent, index)
            updateNodePresentation(child, treePath.pathByAddingChild(child))
          }
        }
      }
    }

    private fun updateNodePresentation(node: Any, treePath: TreePath) {
      val oldGridLine = treeMap[treePath] ?: return
      val parentGridLine = treeMap[treePath.parentPath] ?: return

      val newGridLine: BeTreeGridLine
      if (isNotLoaded(oldGridLine)) {
        newGridLine = convertTreeModelLevel(node, treePath)
      }
      else {
        newGridLine = convertTreeModelLevel(node, treePath, loadChildren = false)
        newGridLine.children.addAll(oldGridLine.children)
      }

      val oldIndex = parentGridLine.children.indexOf(oldGridLine)
      if (oldIndex != -1) parentGridLine.children[oldIndex] = newGridLine
    }


    override fun treeNodesInserted(e: TreeModelEvent) {
      modifyModel {
        val treePath = e.treePath
        val gridLine = treeMap[treePath] ?: return
        if (isNotLoaded(gridLine)) return

        val node = treePath.lastPathComponent
        for (index in e.childIndices) {
          val child = tree.model.getChild(node, index)
          gridLine.children.add(index, convertTreeModelLevel(child, treePath.pathByAddingChild(child)))
        }
      }
    }

    override fun treeNodesRemoved(e: TreeModelEvent) {
      modifyModel {
        val treePath = e.treePath
        val gridLine = treeMap[treePath] ?: return
        if (isNotLoaded(gridLine)) return

        for (child in e.children) {
          removeNodeMappingsUnder(child, treePath.pathByAddingChild(child))
        }

        for (index in e.childIndices.sortedDescending()) {
          gridLine.children.removeAt(index)
        }
      }
    }

    override fun treeStructureChanged(e: TreeModelEvent) {
      modifyModel {
        val treePath = e.treePath
        if (treePath == null || treePath.pathCount == 1) {
          resetTreeStructure()
        }
        else {
          val node = treePath.lastPathComponent
          val gridLine = treeMap[treePath] ?: return
          if (isNotLoaded(gridLine)) return

          for (childIndex in 0 until tree.model.getChildCount(node)) {
            val child = tree.model.getChild(node, childIndex)
            removeNodeMappingsUnder(child, treePath.pathByAddingChild(child))
          }

          reloadNodeChildren(treePath, gridLine)
        }
      }
    }

    private fun removeNodeMappingsUnder(node: Any, treePath: TreePath) {
      treeMap.remove(treePath)

      for (childIndex in 0 until tree.model.getChildCount(node)) {
        val child = tree.model.getChild(node, childIndex)
        removeNodeMappingsUnder(child, treePath.pathByAddingChild(child))
      }
    }
  }

  private inline fun modifyModel(task: () -> Unit) {
    trackOptions.beforeModelChange?.invoke()
    task()
    trackOptions.afterModelChange?.invoke()
  }

  companion object {
    private val LOG = getLogger<BeJTree>()
  }
}
