package com.jetbrains.rdserver.ui.converters.toTreeGrid.custom

import com.intellij.find.actions.ShowUsagesTable
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.AnActionEvent
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.actionSystem.ex.ActionUtil
import com.intellij.openapi.actionSystem.impl.ActionToolbarImpl
import com.intellij.openapi.project.Project
import com.intellij.usageView.UsageViewBundle
import com.intellij.usages.Usage
import com.intellij.usages.impl.UsageNode
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.uiautomation.*
import com.jetbrains.rd.framework.impl.RdProperty
import com.jetbrains.rd.ui.bedsl.dsl.withColor
import com.jetbrains.rd.ui.util.getStringPresentation
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.IMutableViewableList
import com.jetbrains.rdserver.ui.converters.RdConverter
import com.jetbrains.rdserver.ui.converters.TrackOptions
import com.jetbrains.rdserver.ui.converters.toBeModel
import com.jetbrains.rdserver.ui.converters.toTreeGrid.RdTableCellEditorListener
import com.jetbrains.rdserver.ui.converters.toTreeGrid.RdTableModelListener
import org.jetbrains.annotations.NotNull
import javax.swing.JComponent
import javax.swing.JTable
import javax.swing.ListSelectionModel
import javax.swing.SwingUtilities
import javax.swing.table.TableModel

class ShowUsagesTableConverter : RdConverter<ShowUsagesTable> {

  override fun convert(component: ShowUsagesTable, trackOptions: TrackOptions, lifetime: Lifetime): BeControl {
    val columns = if (isShowUsagesTable(component)) {
      // tmp hack to fix ShowUsagesTable with dynamic columns count, creates max number of columns
      MutableList(4) {
        if (it != 3) BeColumnDefinition("", BeUnitSize(BeSizingType.Fit))
        else BeColumnDefinition("", BeUnitSize(BeSizingType.Fill))
      }
      //.apply { this.add(BeColumnDefinition("", BeUnitSize(BeSizingType.Fill)))}
    }
    else {
      component.columnModel.columns.toList().map { column ->
        BeColumnDefinition(column.headerValue.toString(), BeUnitSize(BeSizingType.Fill))

      }.toMutableList() //.apply { this.add(BeColumnDefinition("", BeUnitSize(BeSizingType.Fill))) }
    }
    val selectionModel = component.selectionModel
    val selectionKind = if (selectionModel.selectionMode == ListSelectionModel.SINGLE_SELECTION)
      BeTreeSelection.Single
    else
      BeTreeSelection.Multi

    val beTreeConfiguration = BeTreeConfiguration(
      columns,
      selectionKind,
      true,
      ChildOffsetSize.DEFAULT,
      true,
      false,
      visibleRowsCount = 30
    ).apply {
      infiniteCycleScroll.set(true)
    }

    val root = BeTreeGridLine(emptyList())

    val table = BeTreeGrid(root, beTreeConfiguration)
    fillContent(component, root, lifetime, table)

    val behaviorControl = BeBehaviorControl(table)
    val actionManager = ActionManager.getInstance()
    ActionUtil.getActions(component).forEach { action ->
      val actionId = actionManager?.getId(action)
      var listener: BeActionListener? = null
      if (!actionId.isNullOrEmpty()) {
        listener = BeActionIdListener(actionId)
      }
      else {
        if (action.shortcutSet.shortcuts.isNotEmpty())
          listener = BeShortcutListener(action.shortcutSet.getStringPresentation(), false)
      }

      if (listener == null) return@forEach
      behaviorControl.behavior.add(listener.apply {
        this.reaction.advise(lifetime) {
          action.actionPerformed(createEvent(component, action, null))
        }
      })
    }
    component.addPropertyChangeListener("model") {
      if (it.newValue !is TableModel) return@addPropertyChangeListener

      SwingUtilities.invokeLater {
        root.children.clear()
        fillContent(component, root, lifetime, table)
      }
    }

    // TODO maintain multiple items selection
    component.selectionModel.addListSelectionListener {
      for ((i, row) in (root.children as IMutableViewableList<BeTreeGridLine>).withIndex()) {
        val toBeSelected = component.isRowSelected(i)
        if (row.isSelected.value != toBeSelected)
          row.isSelected.set(toBeSelected)
      }
    }

    return behaviorControl
  }

  private fun isShowUsagesTable(component: JTable): Boolean {
    val model = component.model
    return model.getValueAt(0, 0).toString() == UsageViewBundle.message("progress.searching")
  }

  private fun createEvent(component: JComponent,
                          action: AnAction,
                          project: Project?): @NotNull AnActionEvent {
    val event = AnActionEvent.createFromDataContext(if (component is ActionToolbarImpl) component.place else "",
                                                    action.templatePresentation,
                                                    createDataContext(component, project))
    return event
  }

  private fun createDataContext(component: JComponent, project: Project?): DataContext {
    return DataContext {
      /*when (it) {
        CommonDataKeys.PROJECT.name -> project
        *//*CommonDataKeys.EDITOR.name -> editor
        CommonDataKeys.PSI_FILE.name -> file
        CommonDataKeys.VIRTUAL_FILE.name -> file.virtualFile*//*
        else -> component.contedatactoolbarDataContext.getData(it)
      }*/
    }
  }

  private fun fillContent(component: ShowUsagesTable,
                          root: BeTreeGridLine,
                          lifetime: Lifetime,
                          table: BeTreeGrid) {
    for (row in 0 until component.rowCount) {
      root.children += loadRow(row, component, root, lifetime)
    }
    component.model.addTableModelListener(RdTableModelListener(table, component, lifetime) { loadRow(it, component, root, lifetime) })
  }

  companion object {
    internal fun loadRow(row: Int, component: ShowUsagesTable,
                         root: BeTreeGridLine, lifetime: Lifetime): BeTreeGridLine {
      val controls: List<BeControl?> = List(component.columnCount) { column ->
        val tableModel = component.model
        val data = tableModel.getValueAt(row, column) as? UsageNode

        if (tableModel.isCellEditable(row, column)) {
          val cellEditor = component.getCellEditor(row, column)
          val editorComponent = cellEditor.getTableCellEditorComponent(component, data, false, row, column)

          // todo don't use row and column into closure, they can be changed
          val componentTrackOptions = TrackOptions(
            modelChangeTrack = true,
            modelChangeComponentProvider = {
              val componentData = tableModel.getValueAt(row, column)

              val componentCellEditor = component.getCellEditor(row, column)
              componentCellEditor.getTableCellEditorComponent(component, componentData, false, row, column)
            },
            afterModelChange = {
              val componentCellEditor = component.getCellEditor(row, column)
              componentCellEditor.addCellEditorListener(RdTableCellEditorListener(row, column, component))
              componentCellEditor.stopCellEditing()
            }
          )

          editorComponent.toBeModel(componentTrackOptions, lifetime)
        }
        else {

          val cellRenderer = component.getCellRenderer(row, column)
          val cellComponent = cellRenderer.getTableCellRendererComponent(component, data, false, true, row, column)
          if (component.isSeparatorNode(data?.userObject as? Usage) && column != 1) {
            BeSpacer()
          }
          else {
            val control = cellComponent.toBeModel(TrackOptions.doNotTrack, lifetime = lifetime)
            if (cellComponent.background != component.background && cellComponent.isBackgroundSet &&
                cellComponent.background != UIUtil.getListSelectionBackground(true))
              control.withColor(cellComponent.background)
            else {
              control
            }
          }
        }
      }
      val rowModel = BeTreeGridLine(controls.filterNotNull())
      rowModel.isSelected.advise(lifetime) {
        if ((rowModel.isSelected as RdProperty<Boolean>).isLocalChange) return@advise
        if (it) {
          for ((i, beTreeGridLine) in (root.children as IMutableViewableList<BeTreeGridLine>).withIndex()) {
            if (beTreeGridLine.isSelected.value && !component.isRowSelected(i)) {
              // TODO maintain multiple items selection
              component.setRowSelectionInterval(i, i)
            }
          }
        }
      }

      if (component.selectionModel.isSelectedIndex(row)) {
        rowModel.isSelected.set(true)
      }
      return rowModel
    }
  }
}
