// Copyright 2000-2018 JetBrains s.r.o. Use of this source code is governed by the Apache 2.0 license that can be found in the LICENSE file.
package com.jetbrains.rdserver.ui.intentions

import com.intellij.codeInsight.daemon.impl.ShowIntentionsPass
import com.intellij.codeInsight.hint.HintManager
import com.intellij.codeInsight.intention.IntentionAction
import com.intellij.codeInsight.intention.impl.CachedIntentions
import com.intellij.codeInsight.intention.impl.IntentionActionWithTextCaching
import com.intellij.codeInsight.intention.impl.ShowIntentionActionsHandler
import com.intellij.codeInspection.actions.RunInspectionIntention
import com.intellij.openapi.actionSystem.ActionManager
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.actionSystem.ShortcutProvider
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diagnostic.trace
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.editor.impl.EditorImpl
import com.intellij.openapi.progress.ProcessCanceledException
import com.intellij.openapi.project.DumbService
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createLifetime
import com.intellij.openapi.util.Iconable
import com.intellij.psi.PsiDocumentManager
import com.intellij.psi.PsiFile
import com.intellij.psi.util.PsiUtilBase
import com.jetbrains.ide.model.popups.*
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.ide.model.Solution
import com.jetbrains.rd.ide.model.altEnterModel
import com.jetbrains.rd.ide.model.codeWithMeFocusesModel
import com.jetbrains.rd.platform.util.getLogger
import com.jetbrains.rd.ui.util.getStringPresentation
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.reactive.viewNotNull
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.toProtocolModel
import com.jetbrains.rdserver.toRichText
import com.jetbrains.rdserver.ui.feelLuckUi.BackendFeelLuckyUiHelper
import java.util.concurrent.atomic.AtomicInteger
import javax.swing.Icon

@Suppress("NonDefaultConstructor")
class BackendIntentionActionsHost(
  lifetime: Lifetime,
  project: Project,
  solution: Solution
) {
  val idCounter = AtomicInteger()

  companion object {
    private val logger = logger<BackendIntentionActionsHost>()
  }

  init {
    val altEnterModel = solution.altEnterModel
    altEnterModel.altEnterPopupRequested.view(lifetime) l@{ _, value ->
      if (!value) return@l
      val focusedTextControlId = solution.documentsOperationModel.codeWithMeFocusesModel.lastFocusedTextControl.value
      val editor = focusedTextControlId?.let { BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(it) }
      val psiFile = editor?.let { PsiDocumentManager.getInstance(project).getPsiFile(it.document) }
      logger.trace { "Received intentions popup request for ${focusedTextControlId?.documentId} ${psiFile?.name}" }
      if (editor != null && psiFile != null) {
        val actionsToShow = try {
          ShowIntentionsPass.getActionsToShow(editor, psiFile, true)
          .cacheAndFilterUnsupportedIntentions(project, psiFile, editor)
        } catch (pce: ProcessCanceledException) {
          altEnterModel.altEnterPopup.set(RdEmptyPopupMenuModel(emptyList()))
          return@l
        }
        val model = AltEnterMenu(this, psiFile.project, editor).show(
          (editor as EditorImpl).disposable.createLifetime(), actionsToShow)
        logger.trace { "Sending popup with {${model.items.valueOrNull?.value?.joinToString(", ") { it.text.parts.joinToString("") { it.text } } ?: "null"}} root items" }
        altEnterModel.altEnterPopup.set(model)
      }
      else {
        logger.trace { "Sending empty response" }
        altEnterModel.altEnterPopup.set(RdEmptyPopupMenuModel(emptyList()))
      }
    }
  }

  private fun ShowIntentionsPass.IntentionsInfo.cacheAndFilterUnsupportedIntentions(project: Project, psiFile: PsiFile, editor: Editor): CachedIntentions {
    intentionsToShow.retainAll { BackendIntentionActionFilter.acceptIntention(it.action) }
    errorFixesToShow.retainAll { BackendIntentionActionFilter.acceptErrorFix(it.action) }
    inspectionFixesToShow.retainAll { BackendIntentionActionFilter.acceptInspection(it.action) }
    notificationActionsToShow.retainAll { BackendIntentionActionFilter.acceptNotificationAction(it.action) }

    val cachedIntentions = CachedIntentions.create(project, psiFile, editor, this)
    cachedIntentions.wrapAndUpdateGutters()
    cachedIntentions.gutters.retainAll { BackendIntentionActionFilter.acceptGutterAction(it.action) }
    return cachedIntentions
  }
}

class AltEnterMenu(
  intentionActionsHost: BackendIntentionActionsHost,
  private val project: Project,
  private val editor: Editor,
) {
  private val itemsById = mutableMapOf<Int, IntentionActionWithTextCaching>()
  private val actionsById = mutableMapOf<Int, IntentionAction>()
  private val idCounter = intentionActionsHost.idCounter

  fun show(lifetime: Lifetime, cachedIntentions: CachedIntentions): RdPopupMenuModel {
    val model = RdPopupMenuModel(emptyList(), RdDefaultPopupWindowContext(), null, true, true)
    val items = convertItems(cachedIntentions)
    model.items.set(RdPopupMenuModelItems(items, 0, 0))
    model.executeItem.set(handler = ::executeItem)
    model.expandSubmenu.viewNotNull(lifetime) { _, expandRequest ->
      val item = itemsById[expandRequest.expandedItem]
      if (item != null) {
        LOG.debug("Processing expand item request for item ID ${expandRequest.expandedItem}")
        val submenuModel = RdPopupSubmenuModel(null, true, true)
        submenuModel.items.set(RdPopupMenuModelItems(convertItems(item.optionActions), 0, 0))
        submenuModel.executeItem.set(handler = ::executeItem)
        submenuModel.isPrepared.set(true)
        expandRequest.submenu.set(submenuModel)
      }
      else {
        LOG.debug("Received expand item request for unknown item ${expandRequest.expandedItem}")
      }
    }

    model.isPrepared.set(true)
    return model
  }

  private fun executeItem(lifetime: Lifetime, id: Int): RdTask<Boolean> {
    val action = actionsById[id] ?: return RdTask.fromResult(true)

    HintManager.getInstance().hideAllHints()
    if (project.isDisposed || editor.isDisposed) return RdTask.fromResult(true)

    if (DumbService.isDumb(project) && !DumbService.isDumbAware(action)) return RdTask.fromResult(false)

    PsiDocumentManager.getInstance(project).commitAllDocuments()

    val file = PsiUtilBase.getPsiFileInEditor(editor, project) ?: return RdTask.fromResult(false)

    return try {
      val feelLuckyUiHelper = RdServerPerClientComponents.getService<BackendFeelLuckyUiHelper>(project)
      feelLuckyUiHelper.executeWithUiInteractionMarker(IntentionActionFeelLuckyUiMarker.createToken(action)) {
        val result = ShowIntentionActionsHandler.chooseActionAndInvoke(file, editor, action, action.text)
        return@executeWithUiInteractionMarker RdTask.fromResult(result)
      }
    }
    catch (t: Throwable) {
      LOG.error(t)
      RdTask.faulted(t)
    }
  }

  private fun convertItems(cachedIntentions: CachedIntentions): List<RdPopupMenuModelItem> {

    fun MutableList<RdPopupMenuModelItem>.addActions(actions: Set<IntentionActionWithTextCaching>,
                                                     kind: RdPopupMenuItemKind) {
      for (action in actions) {
        add(convertItem(action, cachedIntentions.getIcon(action), kind))
      }
    }

    val result = mutableListOf<RdPopupMenuModelItem>()
    result.addActions(cachedIntentions.errorFixes, RdPopupMenuItemKind.ErrorFix)
    result.addActions(cachedIntentions.inspectionFixes, RdPopupMenuItemKind.InspectionFix)
    result.addActions(cachedIntentions.intentions, RdPopupMenuItemKind.Intention)
    result.addActions(cachedIntentions.gutters, RdPopupMenuItemKind.Gutter)
    result.addActions(cachedIntentions.notifications, RdPopupMenuItemKind.Notification)
    return result
  }

  private fun convertItem(
    item: IntentionActionWithTextCaching,
    icon: Icon,
    kind: RdPopupMenuItemKind
  ): RdPopupMenuModelItem {
    val intention = item.action
    val displayName = intention.text
    val iconModel = icon.toProtocolModel()
    val itemId = idCounter.incrementAndGet()
    itemsById[itemId] = item
    actionsById[itemId] = intention
    return RdPopupMenuModelItem(
      itemId,
      iconModel,
      null,
      displayName.toRichText(),
      item.shortcut?.getStringPresentation()?.firstOrNull()?.toRichText(),
      tryGetActionId(intention),
      null,
      isSeparator = false,
      isEnabled = true,
      canExpand = item.optionActions.isNotEmpty(),
      isGroupHeader = false,
      subItems = convertItems(item.optionActions),
      kind = kind
    )
  }

  private fun tryGetActionId(intention: IntentionAction): String? {
    val actionId = if (intention is AnAction) ActionManager.getInstance().getId(intention) else null
    return actionId
  }

  /**
   * CWM-672 
   * Some intention actions may cause an undesirable effect when performed on a guest.
   * E.g. they can open a modal dialog on a host or throw some exception. This method is used
   * to hide such intention actions from Alt+Enter popup on a guest.
   * 
   * @return `true` if action must be removed from guest's intention actions 
   */
  private fun filterOutIntentionAction(action: IntentionAction): Boolean {
    // RunInspectionIntention opens a modal dialog on a host
    return action is RunInspectionIntention
  }
  
  private fun convertItems(intentions: List<IntentionAction>): List<RdPopupMenuModelItem> {
    val result = mutableListOf<RdPopupMenuModelItem>()
    for (intention in intentions) {
      if (filterOutIntentionAction(intention))
        continue
      
      result.add(convertItem(intention))
    }
    return result
  }

  private fun convertItem(intention: IntentionAction): RdPopupMenuModelItem {
    val iconModel = (intention as? Iconable)?.getIcon(0)?.toProtocolModel()
    val itemId = idCounter.incrementAndGet()
    actionsById[itemId] = intention
    return RdPopupMenuModelItem(
      itemId,
      iconModel,
      null,
      intention.text.toRichText(),
      (intention as? ShortcutProvider)?.shortcut?.getStringPresentation()?.firstOrNull()?.toRichText(),
      tryGetActionId(intention),
      null,
      isSeparator = false,
      isEnabled = true,
      canExpand = false,
      isGroupHeader = false,
      subItems = null
    )
  }

  companion object {
    private val LOG = getLogger<AltEnterMenu>()
  }
}
