package com.jetbrains.rdserver.ui.modalDialogs

import com.intellij.codeWithMe.ClientId
import com.intellij.icons.AllIcons
import com.intellij.openapi.components.service
import com.intellij.openapi.project.Project
import com.intellij.ui.HyperlinkLabel
import com.intellij.ui.RoundedLineBorder
import com.intellij.util.ui.JBUI
import com.jetbrains.rd.platform.codeWithMe.CodeWithMePlatformBundle
import com.jetbrains.rd.platform.codeWithMe.ui.CWMBaseAcceptUserDialog
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.lifetime.SequentialLifetimes
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.connection.BackendGuestLimitNotificator
import com.jetbrains.rdserver.permissions.impl.PermissionManagerImpl
import com.jetbrains.rdserver.permissions.ui.PermissionsDialog
import com.jetbrains.rdserver.permissions.ui.present
import net.miginfocom.swing.MigLayout
import javax.swing.JComponent
import javax.swing.JLabel
import javax.swing.JPanel
import javax.swing.SwingUtilities

class CWMHostAcceptUserDialog(project: Project,
                              userName: String,
                              clientId: ClientId) : CWMBaseAcceptUserDialog(project, userName, clientId) {
  private val manager = PermissionManagerImpl.getInstanceImpl(project)
  private val permissionLabel = JLabel()
  private val permissionIcon = JLabel()
  private var lifetimes = SequentialLifetimes(project.lifetime)

  private val userCantJoin = JPanel(MigLayout("novisualpadding, ins 2 5 2 5, gap 0, fillx")).apply {
    add(JLabel(CodeWithMePluginBundle.message("connection.dialog.limit.reached")).apply {
      icon = AllIcons.General.Warning
    }, "growx")
    add(HyperlinkLabel(CodeWithMePluginBundle.message("license.upgrade")).apply {
      setHyperlinkTarget(CodeWithMePlatformBundle.message("license.buy.link"))
    })

    background = JBUI.CurrentTheme.Validator.warningBackgroundColor()
    border = RoundedLineBorder(JBUI.CurrentTheme.Validator.warningBorderColor(), 1)

    isVisible = false
  }

  init {
    setResizable(false)
  }

  private var limited: Boolean = false
    set(value) {
      if (value == field) return
      field = value

      userCantJoin.isVisible = value
      isOKActionEnabled = !value

      SwingUtilities.getWindowAncestor(userCantJoin)?.let {
        SwingUtilities.invokeLater { it.pack() }
      }
    }

  override fun show() {
    val def = lifetimes.next()
    manager.permissionModelChanged.advise(def) {
      updatePermissions()
    }

    project.service<BackendGuestLimitNotificator>().guestCount.advise(def) {
      application.invokeLater {
        checkLimits()
      }
    }

    checkLimits()
    updatePermissions()
    super.show()
  }

  private fun checkLimits() {
    limited = project.service<BackendGuestLimitNotificator>().isAtConcurrentGuestLimit
  }

  override fun onClose() {
    lifetimes.terminateCurrent()
  }

  private fun updatePermissions() {
    val model = manager.getModel(clientId)

    val mode = PermissionsDialog.Mode.values().singleOrNull {
      it.fileAccess == model.fileAccess &&
      it.terminalAccess == model.terminalAccess &&
      it.runAccess == model.runAccess
    } ?: PermissionsDialog.Mode.Custom

    permissionLabel.text = mode.title
    permissionIcon.icon = mode.icon

    val tooltip = CodeWithMePlatformBundle.message("connection.dialog.change.permissions.tooltip",
                                                   model.fileAccess.present(),
                                                   model.terminalAccess.present(),
                                                   model.runAccess.present())

    permissionLabel.toolTipText = tooltip
    permissionIcon.toolTipText = tooltip
  }

  override fun createRightPane(): JComponent {
    return super.createRightPane().apply {
      add(userCantJoin, "gaptop 15, growx")
    }
  }

  override fun JPanel.addExtraContent() {
    add(JPanel(MigLayout("ins 0, novisualpadding, gap 2")).apply {
      add(JLabel(CodeWithMePlatformBundle.message("connection.dialog.permissions.message")))
      add(permissionIcon)
      add(permissionLabel, "gapafter 3")
      add(HyperlinkLabel(CodeWithMePlatformBundle.message("connection.dialog.change.permissions")).apply {
        addHyperlinkListener { PermissionsDialog(project, null, false).show() }
      })
    })
  }
}