package com.jetbrains.rdserver.ui.popup

import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.actionSystem.DataProvider
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.defineNestedLifetime
import com.intellij.openapi.ui.popup.JBPopup
import com.intellij.openapi.ui.popup.JBPopupListener
import com.intellij.openapi.ui.popup.LightweightWindowEvent
import com.intellij.openapi.util.Computable
import com.intellij.openapi.util.Disposer
import com.intellij.ui.awt.RelativePoint
import com.intellij.util.ui.UIUtil
import com.jetbrains.ide.model.popups.*
import com.jetbrains.ide.model.uiautomation.BeRichText
import com.jetbrains.rd.util.reactive.adviseNotNull
import com.jetbrains.rdserver.htmlToRichText
import java.awt.Component
import java.awt.Dimension
import java.awt.Point
import java.awt.event.InputEvent
import java.awt.event.KeyEvent
import java.util.concurrent.CopyOnWriteArrayList
import java.util.concurrent.atomic.AtomicInteger
import javax.swing.JComponent
import javax.swing.JPanel
import com.jetbrains.rdserver.ui.popup.bePopups.impl.popupContextKey

abstract class RdJBPopupBase(
  private val popupInteractionModel: PopupInteractionModel,
  private val title: String?,
  private val cancelCallback: Computable<Boolean>?
) : JBPopup {
  companion object {
    private val idCounter = AtomicInteger()
  }
  private val listeners: MutableList<JBPopupListener> = CopyOnWriteArrayList()
  private var finalRunnable: Runnable? = null
  private val popupId: Int = idCounter.incrementAndGet()
  private var adText: String? = null
  private var okStatus = false
  private var isShown = false

  //region show
  private fun show(context: RdPopupWindowContextBase = RdDefaultPopupWindowContext()) {
    require(!isShown)
    val popupMenuModel = RdPopupMenuModel(emptyList(), context, adText, false, false)
    if (title != null) {
      updateTitle(popupMenuModel, title)
    }
    popupMenuModel.items.set(RdPopupMenuModelItems(createPopupModelItems(), 0, 0))
    popupMenuModel.executeItem.set(handler = ::doExecuteItem)
    val lifetime = defineNestedLifetime()
    popupMenuModel.selectedItem.adviseNotNull(lifetime) {
      val index = it.itemId ?: return@adviseNotNull
      handleSelectedItem(index)
    }

    listeners.forEach { listener -> listener.beforeShown(LightweightWindowEvent(this)) }
    popupInteractionModel.popups[popupId] = popupMenuModel
    isShown = true
  }

  override fun showUnderneathOf(componentUnder: Component) {
    show()
  }

  override fun showInBestPositionFor(dataContext: DataContext) {
    show()
  }

  override fun showInBestPositionFor(editor: Editor) {
    show()
  }

  override fun showInCenterOf(component: Component) {
    show()
  }

  override fun show(point: RelativePoint) {
    val context = UIUtil.getClientProperty(point.component, popupContextKey)
                  ?: RdDefaultPopupWindowContext()
    show(context)
  }

  override fun show(owner: Component) {
    show()
  }

  override fun showInScreenCoordinates(owner: Component, point: Point) {
    show()
  }

  override fun showInFocusCenter() {
    show()
  }

  override fun showCenteredInCurrentWindow(project: Project) {
    show()
  }
  //endregion show

  private fun doExecuteItem(index: Int): Boolean {
    try {
      handleExecuteItem(index)
    }
    catch (t1: Throwable) {
      try {
        finalRunnable?.run()
      }
      catch (t2: Throwable) {
        t1.addSuppressed(t2)
      }
      throw t1
    }
    finalRunnable?.run()
    return true
  }

  private fun updateTitle(popupMenuModel: RdPopupMenuModel, title: String) {
    popupMenuModel.toolbar.caption.set(BeRichText(false).apply { text.set(title.htmlToRichText()) })
  }

  protected fun updateItems() {
    if (isShown)
      popupInteractionModel.popups[popupId]!!.items.set(RdPopupMenuModelItems(createPopupModelItems(), 0, 0))
  }

  protected abstract fun createPopupModelItems(): List<RdPopupMenuModelItem>

  protected abstract fun handleExecuteItem(index: Int)

  protected abstract fun handleSelectedItem(index: Int)

  override fun setFinalRunnable(runnable: Runnable?) {
    finalRunnable = runnable
  }

  override fun isDisposed(): Boolean {
    return isShown && popupInteractionModel.popups[popupId] == null
  }

  override fun setCaption(title: String) {
    if (isShown)
      updateTitle(popupInteractionModel.popups[popupId]!!, title)
  }

  override fun setAdText(s: String?, alignment: Int) {
    adText = s
  }

  override fun canClose(): Boolean {
    return cancelCallback == null || cancelCallback.compute()
  }

  override fun cancel() {
    if (!canClose())
      return

    listeners.forEach { listener -> listener.onClosed(LightweightWindowEvent(this, okStatus)) }
    Disposer.dispose(this)
  }

  override fun closeOk(e: InputEvent?) {
    setOk(true)
    TODO()
  }

  private fun setOk(ok: Boolean) {
    okStatus = ok
  }

  override fun isPersistent(): Boolean {
    TODO()
    //return !myCancelOnClickOutside && !myCancelOnWindow
  }

  override fun setUiVisible(visible: Boolean) {
    TODO()
  }

  override fun addListener(listener: JBPopupListener) {
    listeners.add(0, listener)
  }

  override fun removeListener(listener: JBPopupListener) {
    listeners.remove(listener)
  }

  override fun <T : Any?> getUserData(userDataClass: Class<T>): T? {
    TODO()
  }

  override fun setDataProvider(dataProvider: DataProvider) {
    TODO()
  }

  override fun dispose() {

  }

  //region UI stub methods

  private val stub by lazy { JPanel() }

  override fun getContent(): JComponent {
    return stub // hack because of BackgroundUpdaterTaskBase.java:126
  }

  override fun moveToFitScreen() {
    // Cannot execute this operation for remote popup
  }

  override fun setMinimumSize(size: Dimension?) {
    // Cannot execute this operation for remote popup
  }

  override fun setSize(size: Dimension) {
    // Cannot execute this operation for remote popup
  }

  override fun setRequestFocus(b: Boolean) {
    // Cannot execute this operation for remote popup
  }

  override fun pack(width: Boolean, height: Boolean) {
    // Cannot execute this operation for remote popup
  }

  override fun isCancelKeyEnabled(): Boolean {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun getBestPositionFor(dataContext: DataContext): Point {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun cancel(e: InputEvent?) {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun getSize(): Dimension {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun getLocationOnScreen(): Point {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun isModalContext(): Boolean {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun setLocation(screenPoint: Point) {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun dispatchKeyEvent(e: KeyEvent): Boolean {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun getOwner(): Component {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun isFocused(): Boolean {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun isVisible(): Boolean {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  override fun isNativePopup(): Boolean {
    throw UnsupportedOperationException("Cannot execute this operation for remote popup")
  }

  //endregion
}