package com.jetbrains.rdserver.ui.popup.bePopups.impl

import com.intellij.CommonBundle
import com.intellij.codeWithMe.ClientId
import com.intellij.codeWithMe.ClientId.Companion.isValid
import com.intellij.openapi.actionSystem.*
import com.intellij.openapi.actionSystem.ex.ActionManagerEx
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.editor.Editor
import com.intellij.openapi.project.Project
import com.intellij.openapi.ui.MessageType
import com.intellij.openapi.ui.popup.*
import com.intellij.openapi.util.Condition
import com.intellij.openapi.util.Key
import com.intellij.openapi.wm.IdeFrame
import com.intellij.ui.CollectionListModel
import com.intellij.ui.ComponentUtil
import com.intellij.ui.awt.RelativePoint
import com.intellij.ui.components.JBList
import com.intellij.ui.popup.PopupFactoryImpl
import com.jetbrains.codeWithMe.model.PopupsModel
import com.jetbrains.ide.model.popups.PopupInteractionModel
import com.jetbrains.ide.model.popups.RdPopupWindowContextBase
import com.jetbrains.ide.model.uiautomation.UIAutomationInteractionModel
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.followMe.FollowMeManager
import com.jetbrains.rdserver.ui.feelLuckUi.BackendFeelLuckyUiHelper
import com.jetbrains.rdserver.ui.popup.BackendPopupModelHost
import com.jetbrains.rdserver.ui.popup.RdActionGroupPopup
import com.jetbrains.rdserver.ui.popup.RdBalloonBuilder
import java.awt.Color
import java.awt.Component
import java.awt.Point
import java.awt.event.InputEvent
import javax.swing.*
import javax.swing.event.HyperlinkListener
import javax.swing.event.ListSelectionListener

private val logger = logger<BackendBePopupFactory>()

internal class BackendBePopupFactory : PopupFactoryImpl() {
  private val popupInteractionModel: PopupInteractionModel
    get() = RdServerPerClientComponents.getServiceInRandomProject(BackendPopupModelHost::class.java).popupInteractionModel

  private val popupsModel: PopupsModel
    get() = RdServerPerClientComponents.getServiceInRandomProject(BackendPopupModelHost::class.java).popupsModel

  private fun tryGuessFocusedProject(): Project? {
    val focusOwner = FocusManager.getCurrentManager().focusOwner
    val frame = ComponentUtil.getParentOfType(IdeFrame::class.java as Class<out IdeFrame?>, focusOwner)
    return frame?.project
  }

  override fun createActionGroupPopup(
    title: String?,
    actionGroup: ActionGroup,
    dataContext: DataContext,
    showNumbers: Boolean,
    showDisabledActions: Boolean,
    honorActionMnemonics: Boolean,
    disposeCallback: Runnable?,
    maxRowCount: Int,
    preselectActionCondition: Condition<in AnAction>?
  ): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId) {
      return super.createActionGroupPopup(title, actionGroup, dataContext, showNumbers, showDisabledActions, honorActionMnemonics, disposeCallback, maxRowCount, preselectActionCondition)
    }
    return RdActionGroupPopup(popupInteractionModel, title, actionGroup, dataContext, showDisabledActions)
  }

  override fun createActionGroupPopup(
    title: String?,
    actionGroup: ActionGroup,
    dataContext: DataContext,
    aid: ActionSelectionAid?,
    showDisabledActions: Boolean,
    disposeCallback: Runnable?,
    maxRowCount: Int,
    preselectActionCondition: Condition<in AnAction>?,
    actionPlace: String?
  ): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionGroupPopup(title, actionGroup, dataContext, aid, showDisabledActions, disposeCallback, maxRowCount, preselectActionCondition, actionPlace)
    return RdActionGroupPopup(popupInteractionModel, title, actionGroup, dataContext,
                              showDisabledActions)
  }

  override fun guessBestPopupLocation(component: JComponent): RelativePoint {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.guessBestPopupLocation(component)
    return RelativePoint(component, Point(0, 0))
  }

  override fun guessBestPopupLocation(dataContext: DataContext): RelativePoint {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.guessBestPopupLocation(dataContext)
    PlatformDataKeys.CONTEXT_COMPONENT.getData(dataContext)?.let {
      return RelativePoint(it, Point(0, 0))
    }
    PlatformDataKeys.EDITOR.getData(dataContext)?.let {
      return guessBestPopupLocation(it)
    }
    throw UnsupportedOperationException("Can't determine popup location")
  }

  override fun guessBestPopupLocation(editor: Editor): RelativePoint {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.guessBestPopupLocation(editor)
    return RelativePoint(editor.contentComponent, Point(0, 0))
  }

  override fun createDialogBalloonBuilder(content: JComponent, title: String?): BalloonBuilder =
    if (ClientId.isCurrentlyUnderLocalId)
      super.createDialogBalloonBuilder(content, title)
    else
      RdBalloonBuilder(popupInteractionModel.protocol, content, title, RdServerPerClientComponents.getServiceInRandomProject(UIAutomationInteractionModel::class.java))


  override fun createComponentPopupBuilder(content: JComponent, preferableFocusComponent: JComponent?): ComponentPopupBuilder {
    if (ClientId.isCurrentlyUnderLocalId || ClientId.currentOrNull?.isValid != true) {
      val originalBuilder = super.createComponentPopupBuilder(content, preferableFocusComponent)

      val focusedProject = tryGuessFocusedProject()
      val actionId = (ActionManager.getInstance() as ActionManagerEx).lastPreformedActionId

      if (focusedProject != null) {
        val followMeManager = FollowMeManager.getInstance(focusedProject)
        if (followMeManager.getFollowers(ClientId.current).isNotEmpty() &&
            followMeManager.canReflectPopupToFollowers(actionId)) {
          return BackendComponentBePopupBuilder(content, originalBuilder, focusedProject)
        }
      }

      return originalBuilder
    }

    val feelLuckyUiHelper = RdServerPerClientComponents.getServiceInRandomProject(BackendFeelLuckyUiHelper::class.java)
    if(feelLuckyUiHelper.mayShowPopup())
      return BackendComponentBePopupBuilder(content, popupsModel)

    logger.debug("Remote ui isn't implemented yet.")
    return super.createComponentPopupBuilder(content, preferableFocusComponent)
  }

  override fun getParentBalloonFor(c: Component?): Balloon? {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.getParentBalloonFor(c)
    logger.debug("This feature might access UI.")
    return null
  }

  override fun isChildPopupFocused(parent: Component?): Boolean {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.isChildPopupFocused(parent)
    logger.debug("This feature might access UI.")
    return false
  }

  override fun getChildPopups(component: Component): MutableList<JBPopup> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.getChildPopups(component)
    logger.debug("This feature might access UI.")
    return mutableListOf()
  }

  override fun createActionGroupPopup(title: String?,
                                      actionGroup: ActionGroup,
                                      dataContext: DataContext,
                                      selectionAidMethod: ActionSelectionAid?,
                                      showDisabledActions: Boolean): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionGroupPopup(title, actionGroup, dataContext, selectionAidMethod, showDisabledActions)
    return RdActionGroupPopup(popupInteractionModel, title, actionGroup, dataContext,
                              showDisabledActions)
  }

  override fun getChildFocusedPopup(parent: Component?): JBPopup? {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.getChildFocusedPopup(parent)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createActionsStep(actionGroup: ActionGroup,
                                 dataContext: DataContext,
                                 showNumbers: Boolean,
                                 showDisabledActions: Boolean,
                                 title: String?,
                                 component: Component?,
                                 honorActionMnemonics: Boolean): ListPopupStep<*> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionsStep(actionGroup, dataContext, showNumbers, showDisabledActions, title, component, honorActionMnemonics)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createActionsStep(actionGroup: ActionGroup,
                                 dataContext: DataContext,
                                 showNumbers: Boolean,
                                 showDisabledActions: Boolean,
                                 title: String?,
                                 component: Component?,
                                 honorActionMnemonics: Boolean,
                                 defaultOptionIndex: Int,
                                 autoSelectionEnabled: Boolean): ListPopupStep<*> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionsStep(actionGroup, dataContext, showNumbers, showDisabledActions, title, component, honorActionMnemonics,
                                     defaultOptionIndex, autoSelectionEnabled)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createActionGroupPopup(title: String?,
                                      actionGroup: ActionGroup,
                                      dataContext: DataContext,
                                      selectionAidMethod: ActionSelectionAid?,
                                      showDisabledActions: Boolean,
                                      actionPlace: String?): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionGroupPopup(title, actionGroup, dataContext, selectionAidMethod, showDisabledActions, actionPlace)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createPopupChooserBuilder(table: JTable): PopupChooserBuilder<*> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createPopupChooserBuilder(table)
    val feelLuckyUiHelper = RdServerPerClientComponents.getServiceInRandomProject(BackendFeelLuckyUiHelper::class.java)
    if(feelLuckyUiHelper.mayShowPopup())
      return BackendBePopupChooserBuilder<Any>(popupsModel, table)

    logger.debug("Remote ui isn't implemented yet.")
    return super.createPopupChooserBuilder(table)
  }

  override fun <T : Any?> createListPopupBuilder(list: JList<T>): PopupChooserBuilder<T> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createListPopupBuilder(list)
    return BackendBePopupChooserBuilder(popupsModel, list)
  }

  override fun <T : Any?> createPopupChooserBuilder(list: MutableList<out T>): IPopupChooserBuilder<T> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createPopupChooserBuilder(list)
    return BackendBePopupChooserBuilder(popupsModel, JBList(CollectionListModel(list)))
  }

  override fun createActionGroupPopup(title: String?,
                                      actionGroup: ActionGroup,
                                      dataContext: DataContext,
                                      selectionAidMethod: ActionSelectionAid?,
                                      showDisabledActions: Boolean,
                                      disposeCallback: Runnable?,
                                      maxRowCount: Int): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionGroupPopup(title, actionGroup, dataContext, selectionAidMethod, showDisabledActions, disposeCallback,
                                          maxRowCount)
    throw UnsupportedOperationException("This feature is not supported yet.")

  }

  override fun createActionGroupPopup(title: String?,
                                      actionGroup: ActionGroup,
                                      dataContext: DataContext,
                                      showDisabledActions: Boolean,
                                      disposeCallback: Runnable?,
                                      maxRowCount: Int): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionGroupPopup(title, actionGroup, dataContext, showDisabledActions, disposeCallback, maxRowCount)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createHtmlTextBalloonBuilder(htmlContent: String,
                                            icon: Icon?,
                                            fillColor: Color?,
                                            listener: HyperlinkListener?): BalloonBuilder {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createHtmlTextBalloonBuilder(htmlContent, icon, fillColor, listener)
    
    return createHtmlTextBalloonBuilder(htmlContent, icon, null, fillColor, listener)
  }

  override fun createConfirmation(title: String?, onYes: Runnable?, defaultOptionIndex: Int): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createConfirmation(title, onYes, defaultOptionIndex)
    return createConfirmation(title, CommonBundle.getYesButtonText(), CommonBundle.getNoButtonText(), onYes, defaultOptionIndex)
  }

  override fun createConfirmation(title: String?, yesText: String?, noText: String?, onYes: Runnable?, defaultOptionIndex: Int): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createConfirmation(title, yesText, noText, onYes, defaultOptionIndex)
    return createConfirmation(title, yesText, noText, onYes, null, defaultOptionIndex)
  }

  override fun createMessage(text: String?): JBPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createMessage(text)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createConfirmation(title: String?,
                                  yesText: String?,
                                  noText: String?,
                                  onYes: Runnable?,
                                  onNo: Runnable?,
                                  defaultOptionIndex: Int): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createConfirmation(title, yesText, noText, onYes, onNo, defaultOptionIndex)

    val popup = createPopupChooserBuilder(mutableListOf(yesText, noText))
      .setTitle(title ?: "")
      .setMovable(false)
      .setResizable(false)
      .setRequestFocus(true)
      .setItemChosenCallback { item ->
        val runnable = if (item == yesText) onYes else onNo
        runnable?.run()
      }
      .createPopup()

    return object : ListPopup, JBPopup by popup {
      override fun getListStep(): ListPopupStep<*> {
        TODO("Not yet implemented")
      }

      override fun handleSelect(handleFinalChoices: Boolean) {
        TODO("Not yet implemented")
      }

      override fun handleSelect(handleFinalChoices: Boolean, e: InputEvent?) {
        TODO("Not yet implemented")
      }

      override fun setHandleAutoSelectionBeforeShow(autoHandle: Boolean) {
        TODO("Not yet implemented")
      }

      override fun addListSelectionListener(listSelectionListener: ListSelectionListener?) {
        TODO("Not yet implemented")
      }
    }
  }

  override fun createActionsStep(actionGroup: ActionGroup,
                                 dataContext: DataContext,
                                 actionPlace: String?,
                                 showNumbers: Boolean,
                                 showDisabledActions: Boolean,
                                 title: String?,
                                 component: Component?,
                                 honorActionMnemonics: Boolean,
                                 defaultOptionIndex: Int,
                                 autoSelectionEnabled: Boolean): ListPopupStep<ActionItem> {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createActionsStep(actionGroup, dataContext, actionPlace, showNumbers, showDisabledActions, title, component,
                                     honorActionMnemonics, defaultOptionIndex, autoSelectionEnabled)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createListPopup(step: ListPopupStep<*>): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createListPopup(step)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createListPopup(step: ListPopupStep<*>, maxRowCount: Int): ListPopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createListPopup(step, maxRowCount)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createTree(parent: JBPopup?, aStep: TreePopupStep<*>, parentValue: Any?): TreePopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createTree(parent, aStep, parentValue)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createTree(aStep: TreePopupStep<*>): TreePopup {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createTree(aStep)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun isBestPopupLocationVisible(editor: Editor): Boolean {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.isBestPopupLocationVisible(editor)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun getCenterOf(container: JComponent?, content: JComponent?): Point {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.getCenterOf(container, content)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun isPopupActive(): Boolean {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.isPopupActive()
    throw UnsupportedOperationException("This feature is not supported yet.")
  }

  override fun createBalloonBuilder(content: JComponent): BalloonBuilder {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createBalloonBuilder(content)
    return RdBalloonBuilder(popupInteractionModel.protocol, content, null, RdServerPerClientComponents.getServiceInRandomProject(UIAutomationInteractionModel::class.java))
  }

  override fun createHtmlTextBalloonBuilder(htmlContent: String,
                                            icon: Icon?,
                                            textColor: Color?,
                                            fillColor: Color?,
                                            listener: HyperlinkListener?): BalloonBuilder {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createHtmlTextBalloonBuilder(htmlContent, icon, textColor, fillColor, listener)
    
    return createBalloonBuilder(JLabel(htmlContent))
  }

  override fun createHtmlTextBalloonBuilder(htmlContent: String, messageType: MessageType?, listener: HyperlinkListener?): BalloonBuilder {
    if (ClientId.isCurrentlyUnderLocalId)
      return super.createHtmlTextBalloonBuilder(htmlContent, messageType, listener)
    throw UnsupportedOperationException("This feature is not supported yet.")
  }
}

internal val popupContextKey = Key.create<RdPopupWindowContextBase>("rdserver.popupContext")