package com.jetbrains.rdserver.unattendedHost

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.Logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ProjectManager
import com.intellij.openapi.project.ProjectManagerListener
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Key
import com.intellij.util.io.isDirectory
import com.jetbrains.codeWithMe.model.*
import com.jetbrains.rd.framework.*
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.framework.util.asCompletableFuture
import com.jetbrains.rd.framework.util.setSuspend
import com.jetbrains.rd.platform.codeWithMe.protocol.ConnectionState
import com.jetbrains.rd.platform.codeWithMe.protocol.SecureWireOverStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.SocketStreamTransport
import com.jetbrains.rd.platform.codeWithMe.protocol.WireConnectStatus
import com.jetbrains.rd.platform.util.SelfSignedCertificate
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.getOrCreateUserData
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.util.addUnique
import com.jetbrains.rd.util.lifetime.*
import com.jetbrains.rd.util.reactive.viewNotNull
import com.jetbrains.rdserver.connection.CwmClientConnectionGroup
import com.jetbrains.rdserver.joinLinks.RdServerJoinLinkManager
import com.jetbrains.rdserver.startup.EdtScheduler
import com.jetbrains.rdserver.startup.RdServerConnectionManager
import com.jetbrains.rdserver.startup.SecureClientData
import kotlinx.coroutines.future.await
import org.apache.commons.lang.SystemUtils
import java.io.File
import java.net.InetSocketAddress
import java.nio.channels.AsynchronousCloseException
import java.nio.channels.ServerSocketChannel
import java.nio.file.Files
import java.nio.file.Path
import java.util.concurrent.CompletableFuture
import kotlin.concurrent.thread
import kotlin.streams.toList

@Service
class UnattendedHostControlSocketManager {
  companion object {
    private val logger = Logger.getInstance(UnattendedHostControlSocketManager::class.java)
    private val projectInfosKey = Key.create<MutableList<RemoteControlProjectInfo>>("cwm.remoteControlProjectInfo")

    fun getInstance() = service<UnattendedHostControlSocketManager>()
  }

  private val controlSocketLifetimes = SequentialLifetimes(Lifetime.Eternal)

  var currentControlSocketPort: Int = 0
    private set

  fun shouldAcceptGuest(project: Project, data: SecureClientData): CompletableFuture<CwmClientConnectionGroup.AcceptState> {
    val projectService = UnattendedHostProjectService.getInstance(project)
    if (projectService.hasClientUsedValidTokenPreviously(data)) // auto-accept clients that have previously used a valid token once
      return CompletableFuture.completedFuture(CwmClientConnectionGroup.AcceptState.Accepted)

    if (!data.singleUseToken.isNullOrEmpty()) {
      logger.debug("Guest ${data.userName} is joining with token ${data.singleUseToken}")
      if (projectService.verifySingleUseConnectionToken(data)) {
        logger.debug("Accepting ${data.userName} because of valid token")
        return CompletableFuture.completedFuture(CwmClientConnectionGroup.AcceptState.Accepted)
      }
    }

    val projectInfos = project.getOrCreateUserData(projectInfosKey) { ArrayList() }

    if (projectInfos.isEmpty()) {
      if(UnattendedHostManager.getInstance().hasPassword) {
        logger.debug("No remote controllers connected, requesting password ${data.userName}")
        return CompletableFuture.completedFuture(CwmClientConnectionGroup.AcceptState.Password)
      } else {
        logger.warn("No remote controllers connected, no password set, declining ${data.userName}")
        return CompletableFuture.completedFuture(CwmClientConnectionGroup.AcceptState.Declined)
      }
    }

    val randomSession = projectInfos[0]
    // todo: do something with allowGuests property
    //if (randomSession.allowGuests.valueOrNull != true)
    //  return CompletableFuture.completedFuture(false)

    logger.debug("Asking remote controller to accept guest ${data.userName}")

    val fingerprintString = RdServerConnectionManager.getInstance(project).getPinString(data.clientFingerprint)

    val task = randomSession.acceptGuest.start(RemoteControlGuestJoinRequest(data.userName, data.clientId.value, fingerprintString))
    return task.asCompletableFuture().thenApply { if (it) CwmClientConnectionGroup.AcceptState.Accepted else CwmClientConnectionGroup.AcceptState.Declined }
  }

  fun startListening(): Int {
    val currentLifetime = controlSocketLifetimes.next()

    val serverSocketChannel = ServerSocketChannel.open().bind(InetSocketAddress("127.0.0.1", 0))
    val localPort = (serverSocketChannel.localAddress as InetSocketAddress).port

    val listeningThread = thread(name = "UnattendedHostControlSocketAccept($localPort)", isDaemon = true) {
      logger.info("Control socket running on port $localPort")

      while (serverSocketChannel.isOpen) {
        try {
          val socketChannel = serverSocketChannel.accept()
          socketChannel.socket().tcpNoDelay = true

          logger.info("Accepted connection from ${socketChannel.remoteAddress}")

          application.invokeLater {
            val id = "cs:tcp:${socketChannel.remoteAddress}"

            val transport = SocketStreamTransport(socketChannel, id)
            val wireLifetime = currentLifetime.createNested()
            val wire = SecureWireOverStreamTransport(wireLifetime, transport, EdtScheduler, id, true, { SelfSignedCertificate.permanentCertificateData }, {
              // todo: validate certificates/store client fingerprint?
            }, {
              // todo: validate certificates/store client fingerprint?
              WireConnectStatus.ACCEPTED
            })

            wire.connectionState.view(wireLifetime) { stateLt, state ->
              if(state != ConnectionState.CONNECTED) return@view

              EdtScheduler.queue {
                createControlModelOnWire(stateLt, wire)
                wire.startDeliveringMessages()
              }
            }
          }
        }
        catch (_: InterruptedException) {
          break
        }
        catch (_: AsynchronousCloseException) {
          break
        }
        catch (t: Throwable) {
          logger.warn("Error accepting connection at port $localPort: ${t.message}", t)
        }
      }

      logger.info("Not listening on control socket on port $localPort anymore")
    }

    currentControlSocketPort = localPort
    currentLifetime.onTermination {
      logger.info("Control socket on port $currentControlSocketPort was closed")
      currentControlSocketPort = 0 // todo: make it nullable instead of using magic zero?
      serverSocketChannel.close()
      listeningThread.interrupt()
    }

    return localPort
  }

  fun stopListening() {
    controlSocketLifetimes.terminateCurrent()
  }

  private fun createControlModelOnWire(lifetime: Lifetime, wire: SecureWireOverStreamTransport) {
    val protocol = Protocol("UnattendedHostControl-${wire.id}", Serializers(), Identities(IdKind.Server), wire.scheduler, wire, lifetime)

    createControlModelOnProtocol(lifetime, protocol)
  }

  fun createControlModelOnProtocol(lifetime: Lifetime, protocol: Protocol) {
    val controlModel = CodeWithMeUnattendedHostControlModel.create(lifetime, protocol)

    val session = RemoteControlSession()

    val projectManager = ProjectManager.getInstance()
    val projectListener = object : ProjectManagerListener {
      override fun projectOpened(project: Project) {
        handleProjectAdded(project, session, lifetime)
      }
    }

    projectManager.openProjects.forEach {
      if (!it.isDefault)
        handleProjectAdded(it, session, lifetime)
    }

    application.messageBus.connect(lifetime.createNestedDisposable()).subscribe(ProjectManager.TOPIC, projectListener)

    session.openNewProject.setSuspend(handlerScheduler = EdtScheduler.backgroundScheduler) { lt, projectPath ->
      val joinLink = UnattendedHostManager.openProjectAndGetJoinLink(projectPath).await() ?: throw RuntimeException("No join link provided")
      joinLink.second
    }

    session.listFsEntries.set(handlerScheduler = EdtScheduler.backgroundScheduler) { _, pathString ->
      val path = Path.of(pathString)
      RdTask.fromResult(
        if (path.isDirectory())
          Files.list(path).use { it.map { RemoteControlFsEntry(it.fileName.toString(), it.toString(), it.isDirectory()) }.toList() }
        else
          emptyList()
      )
    }

    session.getFsRoots.set(handlerScheduler = EdtScheduler.backgroundScheduler) { _, _ ->
      RdTask.fromResult(
        if (SystemUtils.IS_OS_WINDOWS)
          File.listRoots().map { RemoteControlFsEntry(it.name, it.path, true) }
        else
          listOf(RemoteControlFsEntry("/", "/", true))
      )
    }

    session.setControllerClientId.set { it ->
      UnattendedHostManager.getInstance().setControllerClientId(lifetime, ClientId(it))
    }

    controlModel.session.set(session)
  }

  private fun handleProjectAdded(project: Project, session: RemoteControlSession, sessisonLifetime: Lifetime) {
    val intersection = project.lifetime.intersect(sessisonLifetime)

    val projectInfo = RemoteControlProjectInfo(project.name, project.basePath ?: "")
    session.openProjects.addUnique(intersection, projectInfo)

    val instance = RdServerJoinLinkManager.getInstance(project)
    instance.getExistingJoinLink()?.let { projectInfo.joinLink.set(it.toString()) }
    instance.joinLinksUpdated.advise(intersection) {
      instance.getExistingJoinLink()?.let { projectInfo.joinLink.set(it.toString()) }
    }

    projectInfo.createSingleUseJoinToken.set { _, _ ->
      RdTask.fromResult(UnattendedHostProjectService.getInstance(project).createSingleUseConnectionToken())
    }

    project.getOrCreateUserData(projectInfosKey) { ArrayList() }.addUnique(intersection, projectInfo)

    // todo: do something with allowGuests property
  }
}