package com.jetbrains.rdserver.unattendedHost

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.PathManager
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.project.Project
import com.intellij.openapi.project.ex.ProjectManagerEx
import com.intellij.platform.PlatformProjectOpenProcessor
import com.intellij.util.io.exists
import com.intellij.util.io.isDirectory
import com.jetbrains.cwm.unattended.common.UnattendedHostPasswordUtil
import com.jetbrains.rd.platform.util.withFragment
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rdserver.connection.CwmClientConnectionGroup
import com.jetbrains.rdserver.joinLinks.RdServerJoinLinkManager
import com.jetbrains.rdserver.startup.SecureClientData
import kotlinx.serialization.Serializable
import kotlinx.serialization.json.Json
import java.nio.file.Path
import java.util.concurrent.CompletableFuture
import java.util.concurrent.CompletionStage

@Service
class UnattendedHostManager {
  companion object {
    private val logger = logger<UnattendedHostManager>()

    private val mapper = Json {
      ignoreUnknownKeys = true
      prettyPrint = true
    }

    fun getInstance(): UnattendedHostManager = service()

    fun openProjectAndGetJoinLink(projectPath: String): CompletionStage<Pair<Project, String>?> {
      val projectManager = ProjectManagerEx.getInstanceEx()
      val projectPathPath = Path.of(projectPath)
      if (!projectPathPath.isDirectory()) {
        logger.error("Project path is not a directory")
        return CompletableFuture.completedFuture(null)
      }
      val options = PlatformProjectOpenProcessor.createOptionsToOpenDotIdeaOrCreateNewIfNotExists(projectPathPath,
                                                                                                  null).withForceOpenInNewFrame(true)
      val project = projectManager.openProject(projectPathPath, options)
      if (project == null) return CompletableFuture.completedFuture(null)
      return RdServerJoinLinkManager.getInstance(project).getJoinLink().thenApply {
        val unattendedService = UnattendedHostProjectService.getInstance(project)
        project to it.withFragment(it.fragment + "&jt=" + unattendedService.createOrGetSingleUseConnectionToken()).toString()
      }
    }
  }

  var isUnattendedMode = false
    private set

  private var storedPassword: StoredPasswordData? = null

  val hasPassword: Boolean
    get() = storedPassword != null

  fun setUnattendedMode() {
    logger.info("Host switching to unattended mode, join requests will be denied/accepted programmatically")
    if (!isUnattendedMode) {
      UnattendedHostControlSocketManager.getInstance().startListening()
    }
    isUnattendedMode = true

    // stored in a separate file for easier CLI interop
    val pathToStoredPassword = PathManager.getConfigDir().resolve("cwm-passwd")
    val passwordInEnv = System.getenv("CWM_HOST_PASSWORD")
    if(!passwordInEnv.isNullOrEmpty()) {
      val (hash, salt) = UnattendedHostPasswordUtil.hashPasswordCreateSalt(passwordInEnv)

      val data = StoredPasswordData(1, hash, salt)
      storedPassword = data
      pathToStoredPassword.toFile().writeText(mapper.encodeToString(StoredPasswordData.serializer(), data))
    } else {
      if (pathToStoredPassword.exists()) {
        val storedData = mapper.decodeFromString(StoredPasswordData.serializer(), pathToStoredPassword.toFile().readText())
        if(storedData.version == 1) {
          storedPassword = storedData
        }
      }
    }
  }

  fun exitUnattendedMode() {
    logger.info("Host exiting unattended mode")
    if (isUnattendedMode)
      UnattendedHostControlSocketManager.getInstance().stopListening()
    isUnattendedMode = false
  }

  private var controllerClientId: ClientId? = null
  private var lastClientJoinedViaPassword: ClientId? = null

  fun setClientJoinedViaPassword(clientId: ClientId) {
    lastClientJoinedViaPassword = clientId
  }

  fun wasClientJoinedViaPassword(clientId: ClientId) = clientId == lastClientJoinedViaPassword

  fun setControllerClientId(lifetime: Lifetime, clientId: ClientId) {
    require(controllerClientId == null)

    lifetime.bracket(
      {
        logger.info("Unattended controller client is now ${clientId.value}")
        controllerClientId = clientId
      }, {
        logger.info("Unattended controller client has left")
        controllerClientId = null
      })
  }

  fun isControllerClientId(clientId: ClientId): Boolean {
    return clientId == controllerClientId
  }

  fun isUnderControllerClientId() = isControllerClientId(ClientId.current)

  fun shouldAutomaticallyAcceptGuest(project: Project, data: SecureClientData): CompletableFuture<CwmClientConnectionGroup.AcceptState> {
    return UnattendedHostControlSocketManager.getInstance().shouldAcceptGuest(project, data)
  }

  fun createPasswordRequest(): PasswordRequest? {
    return storedPassword?.let { tph ->
      val (targetHash, newSalt) = UnattendedHostPasswordUtil.hashPasswordCreateSalt(tph.hash)
      PasswordRequest(targetHash, tph.salt, newSalt)
    }
  }

  class PasswordRequest(private val targetResponse: ByteArray, val baseSalt: ByteArray, val tokenSalt: ByteArray) {
    fun checkResponse(response: ByteArray): Boolean {
      return targetResponse.contentEquals(response)
    }
  }

  @Serializable
  private data class StoredPasswordData(val version: Int, val hash: ByteArray, val salt: ByteArray) {

    override fun equals(other: Any?): Boolean {
      if (this === other) return true
      if (javaClass != other?.javaClass) return false

      other as StoredPasswordData

      if (version != other.version) return false
      if (!hash.contentEquals(other.hash)) return false
      if (!salt.contentEquals(other.salt)) return false

      return true
    }

    override fun hashCode(): Int {
      var result = version
      result = 31 * result + hash.contentHashCode()
      result = 31 * result + salt.contentHashCode()
      return result
    }
  }
}