package com.jetbrains.rdserver.vcs

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ApplicationManager
import com.intellij.openapi.components.Service
import com.intellij.openapi.components.service
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.event.DocumentListener
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.project.Project
import com.intellij.openapi.vcs.FilePath
import com.intellij.openapi.vcs.ProjectLevelVcsManager
import com.intellij.openapi.vcs.changes.*
import com.intellij.openapi.vcs.impl.ActiveChangeListTrackerImpl
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.VirtualFileManager
import com.intellij.openapi.vfs.newvfs.BulkFileListener
import com.intellij.openapi.vfs.newvfs.events.*
import com.intellij.util.containers.MultiMap
import com.intellij.vcsUtil.VcsUtil
import com.jetbrains.codeWithMe.model.RdUserModel
import com.jetbrains.rd.platform.codeWithMe.users.UserManager
import com.jetbrains.rd.platform.codeWithMe.users.clientId
import com.jetbrains.rd.platform.util.lifetime
import com.jetbrains.rd.platform.util.userData
import com.jetbrains.rd.util.reactive.AddRemove
import com.jetbrains.rdserver.CodeWithMePluginBundle

@Suppress("NonDefaultConstructor")
@Service
class BackendChangeListManager(val project: Project) {
  companion object {
    fun getInstance(project: Project): BackendChangeListManager = project.service()
  }

  init {
    UserManager.getInstance(project).users.adviseAddRemove(project.lifetime) { addRemove, _, userModel ->
      if (addRemove == AddRemove.Remove) {
        val changeListManager = ChangeListManager.getInstance(project)
        val listName = getChangeListName(userModel)
        val changeList = changeListManager.findChangeList(listName) ?: return@adviseAddRemove
        changeListManager.scheduleAutomaticEmptyChangeListDeletion(changeList, true)
      }
    }
  }

  fun createChangeListFor(clientId: ClientId): LocalChangeList? {
    val changeListManager = ChangeListManager.getInstance(project)
    val listName = getChangeListName(clientId) ?: return null
    return changeListManager.findChangeList(listName)
           ?: changeListManager.addChangeList(listName, null)
  }

  fun createChangeListFor(clientId: ClientId, gate: ChangeListManagerGate): LocalChangeList? {
    val listName = getChangeListName(clientId) ?: return null
    return gate.findChangeList(listName)
           ?: gate.addChangeList(listName, null)
  }

  private fun getChangeListName(clientId: ClientId): String? {
    val userModel = UserManager.getInstance(project).users.find { it.clientId == clientId } ?: return null
    return getChangeListName(userModel)
  }

  private fun getChangeListName(userModel: RdUserModel): String {
    return CodeWithMePluginBundle.message("cwm.client.changelist.name", userModel.name)
  }
}

@Suppress("NonDefaultConstructor")
class BackendActiveChangeListTracker(project: Project) : ActiveChangeListTrackerImpl(project) {
  override fun getActiveChangeListId(): String? {
    val activeId = super.getActiveChangeListId()
    if (activeId != null) return activeId

    if (ClientId.isCurrentlyUnderLocalId) return null

    return BackendChangeListManager.getInstance(project).createChangeListFor(ClientId.current)?.id
  }
}

@Suppress("NonDefaultConstructor")
class BackendChangeListChangeAssigner(val project: Project) : ChangeListChangeAssigner {
  private val changeListManager = ChangeListManager.getInstance(project)
  private val vcsManager = ProjectLevelVcsManager.getInstance(project)

  private val pendingFilesToMove = MultiMap.createSet<ClientId, FilePath>()
  private val activeFilesToMove = MultiMap.createSet<ClientId, FilePath>() // accessed from CLM thread

  init {
    EditorFactory.getInstance().eventMulticaster.addDocumentListener(MyDocumentListener(), project)
    ApplicationManager.getApplication().messageBus.connect(project).subscribe(VirtualFileManager.VFS_CHANGES, MyVirtualFileListener())
  }

  override fun beforeChangesProcessing(dirtyScope: VcsDirtyScope?) {
    synchronized(pendingFilesToMove) {
      // Ensure we won't lose files that were modified during CLM refresh
      assert(activeFilesToMove.isEmpty)
      activeFilesToMove.putAllValues(pendingFilesToMove)
      pendingFilesToMove.clear()
    }
  }

  override fun markChangesProcessed(dirtyScope: VcsDirtyScope?) {
    if (dirtyScope == null) {
      activeFilesToMove.clear()
      return
    }

    val toKeep = MultiMap.createSet<ClientId, FilePath>()
    for ((clientId, filePaths) in activeFilesToMove.entrySet()) {
      toKeep.putValues(clientId, filePaths.filterNot { dirtyScope.belongsTo(it) })
    }
    activeFilesToMove.clear()

    synchronized(pendingFilesToMove) {
      pendingFilesToMove.putAllValues(toKeep)
    }
  }

  override fun getChangeListIdFor(change: Change, gate: ChangeListManagerGate): String? {
    val clientId = getClientIdFor(change) ?: return null
    return BackendChangeListManager.getInstance(project).createChangeListFor(clientId, gate)?.id
  }

  private fun getClientIdFor(change: Change): ClientId? {
    for ((clientId, filePaths) in activeFilesToMove.entrySet()) {
      val bPath = ChangesUtil.getBeforePath(change)
      val aPath = ChangesUtil.getAfterPath(change)
      if (bPath != null && filePaths.contains(bPath) ||
          aPath != null && filePaths.contains(aPath)) {
        return clientId
      }
    }

    if (change.type == Change.Type.NEW) {
      val virtualFile = change.virtualFile
      val clientId = virtualFile?.vcsClientId
      if (clientId != null) {
        virtualFile.vcsClientId = null
        return clientId
      }
    }

    return null
  }

  private fun markClientModifiedFile(filePath: FilePath) {
    if (vcsManager.getVcsFor(filePath) != null) {
      synchronized(pendingFilesToMove) {
        pendingFilesToMove.putValue(ClientId.current, filePath)
      }
    }
  }

  private inner class MyDocumentListener : DocumentListener {
    override fun documentChanged(event: DocumentEvent) {
      if (!ApplicationManager.getApplication().isDispatchThread) return // disable for documents forUseInNonAWTThread
      if (ClientId.isCurrentlyUnderLocalId) return
      if (!changeListManager.areChangeListsEnabled()) return

      val document = event.document
      val virtualFile = FileDocumentManager.getInstance().getFile(document) ?: return

      val changeList = changeListManager.getChangeList(virtualFile)
      if (changeList == null) {
        markClientModifiedFile(VcsUtil.getFilePath(virtualFile))
      }
    }
  }

  private inner class MyVirtualFileListener : BulkFileListener {
    override fun before(events: List<VFileEvent>) {
      if (ClientId.isCurrentlyUnderLocalId) return
      if (!changeListManager.areChangeListsEnabled()) return

      for (event in events) {
        if (event.file?.isInLocalFileSystem == false) continue
        when {
          event is VFilePropertyChangeEvent && event.isRename -> handleModifiedFile(event.path)
          event is VFileDeleteEvent -> handleModifiedFile(event.path)
          event is VFileMoveEvent -> handleModifiedFile(event.path)
          event is VFileContentChangeEvent -> handleModifiedFile(event.path)
        }
      }
    }

    override fun after(events: List<VFileEvent>) {
      if (ClientId.isCurrentlyUnderLocalId) return
      if (!changeListManager.areChangeListsEnabled()) return

      for (event in events) {
        if (event.file?.isInLocalFileSystem == false) continue
        when {
          event is VFilePropertyChangeEvent && event.isRename -> handleNewFile(event.file, event.path)
          event is VFileCreateEvent -> handleNewFile(event.file, event.path)
          event is VFileMoveEvent -> handleNewFile(event.file, event.path)
          event is VFileCopyEvent -> handleNewFile(event.findCreatedFile(), event.path)
        }
      }
    }

    private fun handleModifiedFile(path: String) {
      markClientModifiedFile(VcsUtil.getFilePath(path))
    }

    private fun handleNewFile(file: VirtualFile?, path: String) {
      markClientModifiedFile(VcsUtil.getFilePath(path))
      if (file != null) file.vcsClientId = ClientId.current // leave marker for later, as file might be created as unversioned
    }
  }

  companion object {
    private var VirtualFile.vcsClientId by userData<ClientId>()
  }
}
