package com.jetbrains.rdserver.vcs

import com.intellij.codeWithMe.ClientId
import com.intellij.diff.DiffDialogHints
import com.intellij.diff.DiffManagerImpl
import com.intellij.diff.chains.DiffRequestChain
import com.intellij.diff.chains.SimpleDiffRequestChain
import com.intellij.diff.editor.ChainDiffVirtualFile
import com.intellij.diff.editor.DiffRequestProcessorEditor
import com.intellij.diff.editor.DiffVirtualFile
import com.intellij.diff.editor.SimpleDiffVirtualFile
import com.intellij.diff.impl.CacheDiffRequestChainProcessor
import com.intellij.diff.impl.DiffRequestProcessor
import com.intellij.diff.tools.util.side.OnesideTextDiffViewer
import com.intellij.diff.tools.util.side.TwosideTextDiffViewer
import com.intellij.diff.util.DiffUserDataKeysEx
import com.intellij.diff.util.FileEditorBase
import com.intellij.openapi.diff.DiffBundle
import com.intellij.openapi.fileEditor.*
import com.intellij.openapi.project.DumbAware
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Disposer
import com.intellij.openapi.vcs.changes.ChainBackedDiffPreviewProvider
import com.intellij.openapi.vcs.changes.ChangeViewDiffRequestProcessor
import com.intellij.openapi.vcs.changes.PreviewDiffVirtualFile
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.pom.Navigatable
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.noReenter
import com.jetbrains.rd.platform.util.onTerminationWith
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.editors.BackendFileEditorHandler
import com.jetbrains.rdserver.editors.BackendServerFileEditorManager
import com.jetbrains.rdserver.editors.clientId
import com.jetbrains.rdserver.followMe.FollowMeManager
import javax.swing.JComponent
import javax.swing.JPanel
import kotlin.streams.toList

class BackendServerDiffManager : DiffManagerImpl() {
  override fun showDiff(project: Project?, chain: DiffRequestChain, hints: DiffDialogHints) {
    if (ClientId.isCurrentlyUnderLocalId) {
      super.showDiff(project, chain, hints)
      return
    }

    showRemoteDiff(project, chain)
  }

  override fun showDiffBuiltin(project: Project?, chain: DiffRequestChain, hints: DiffDialogHints) {
    if (ClientId.isCurrentlyUnderLocalId) {
      super.showDiffBuiltin(project, chain, hints)
      return
    }

    showRemoteDiff(project, chain)
  }

  private fun showRemoteDiff(project: Project?, chain: DiffRequestChain) {
    if (project == null) error("Can't show diff: project is null - $chain")

    val request = ChainDiffVirtualFile(chain, DiffBundle.message("label.default.diff.editor.tab.name"))
    FileEditorManager.getInstance(project).openFile(request, true)
  }
}

class DiffBackendFileEditorHandler : BackendFileEditorHandler {
  override fun isAvailable(project: Project, file: VirtualFile, provider: FileEditorProvider): Boolean =
    provider is BackendDiffFileEditorProvider && canExtractRequestChain(file)

  override fun createFollowNavigatable(project: Project, editor: FileEditor): FileEditorNavigatable? {
    val file = editor.file as? DiffVirtualFile ?: return null
    return BackendDiffFileEditorNavigatable.create(project, file, editor)
  }
}

class BackendDiffFileEditorProvider : FileEditorProvider, DumbAware {
  override fun accept(project: Project, file: VirtualFile): Boolean {
    return file is DiffVirtualFile
  }

  override fun createEditor(project: Project, file: VirtualFile): FileEditor {
    file as DiffVirtualFile
    if (ClientId.isCurrentlyUnderLocalId) {
      val processor = file.createProcessor(project)
      processor.putContextUserData(DiffUserDataKeysEx.DIFF_IN_EDITOR, true)
      return BackendDiffRequestProcessorEditor(file, processor)
    }
    else {
      val chain = extractRequestChain(file) ?: error("Can't get request chain for $file")
      return DiffRequestRemoteEditor(project, ClientId.current, file, chain)
    }
  }

  override fun disposeEditor(editor: FileEditor) = Disposer.dispose(editor)
  override fun getEditorTypeId(): String = "BackendDiffFileEditorProvider"
  override fun getPolicy(): FileEditorPolicy = FileEditorPolicy.HIDE_DEFAULT_EDITOR
}

private class DiffRequestRemoteEditor(
  val project: Project,
  val clientId: ClientId,
  val file: DiffVirtualFile,
  chain: DiffRequestChain
) : FileEditorBase(), NavigatableFileEditor {
  private val model: DiffRequestChainModel = BackendDiffViewerHost.getInstance(project, clientId).registerRemoteEditor(chain, this)

  override fun canNavigateTo(navigatable: Navigatable): Boolean = navigatable is BackendDiffFileEditorNavigatable

  override fun navigateTo(navigatable: Navigatable) {
    if (navigatable is BackendDiffFileEditorNavigatable) {
      val handler = BackendDiffViewerHost.getInstance(project, clientId).findChainHandler(model)
      handler?.navigateTo(navigatable.chainIndex, navigatable.state)
    }
  }

  override fun dispose() {
    BackendDiffViewerHost.getInstance(project, clientId).releaseRemoteEditor(model)
  }

  override fun getComponent(): JComponent = JPanel()
  override fun getPreferredFocusedComponent(): JComponent? = null

  override fun isValid(): Boolean = true
  override fun getFile(): VirtualFile = file
  override fun getName(): String = DiffBundle.message("diff.file.editor.name")
}

private class BackendDiffRequestProcessorEditor(file: DiffVirtualFile, processor: DiffRequestProcessor)
  : DiffRequestProcessorEditor(file, processor), NavigatableFileEditor {
  override fun canNavigateTo(navigatable: Navigatable): Boolean = navigatable is BackendDiffFileEditorNavigatable

  override fun navigateTo(navigatable: Navigatable) {
    if (navigatable is BackendDiffFileEditorNavigatable) {
      if (processor is CacheDiffRequestChainProcessor) {
        processor.setCurrentRequest(navigatable.chainIndex)
      }
    }
  }
}

@Suppress("NonDefaultConstructor")
class BackendDiffViewerHost(
  val clientId: ClientId,
  lifetime: Lifetime,
  val project: Project,
  protocol: Protocol) {
  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendDiffViewerHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendDiffViewerHost::class.java)
    }
  }

  private val diffModel = DiffViewersModel.create(lifetime, protocol)

  private val activeEditors = mutableMapOf<DiffRequestChainModel, RemoteChainEditor>()
  private val activeHandlers = mutableMapOf<DiffRequestChainModel, DiffRequestChainModelHandler>()

  init {
    diffModel.chains.view(lifetime) { chainLifetime, _, chainModel ->
      chainLifetime.onTerminationWith(clientId) {
        val removedEditor = activeEditors.remove(chainModel) ?: error("Editor not found: $chainModel")
        activeHandlers.remove(chainModel) ?: error("Handler not found: $chainModel")

        closeEditor(removedEditor)
      }

      val chain = activeEditors[chainModel]!!.chain
      activeHandlers[chainModel] = DiffRequestChainModelHandler(project, chainLifetime, clientId, chain, chainModel)
    }

    val connection = project.messageBus.connect(lifetime.createNestedDisposable())
    connection.subscribe(FileEditorManagerListener.FILE_EDITOR_MANAGER, object : FileEditorManagerListener {
      override fun selectionChanged(event: FileEditorManagerEvent) {
        val file = event.newFile as? DiffVirtualFile ?: return
        val editorClientId = event.newEditor.clientId()

        if (editorClientId != clientId && FollowMeManager.getInstance(project).isFollowing(clientId, editorClientId)) {
          val navigatable = BackendDiffFileEditorNavigatable.create(project, file, event.newEditor)
                            ?: OpenFileDescriptor(project, file)
          ClientId.withClientId(clientId) {
            FileEditorManager.getInstance(project).openFileEditor(navigatable, true)
          }
        }

        if (editorClientId == clientId && FollowMeManager.getInstance(project).isHostFollowingClient(clientId)) {
          val navigatable = BackendDiffFileEditorNavigatable.create(project, file, event.newEditor)
                            ?: OpenFileDescriptor(project, file)
          ClientId.withClientId(ClientId.localId) {
            FileEditorManager.getInstance(project).openFileEditor(navigatable, true)
          }
        }

        if (editorClientId == clientId) {
          val model = findChainModel(file)
          if (model != null) {
            model.requestFocus.start(Unit)
          }
        }
      }
    })
  }

  fun registerRemoteEditor(chain: DiffRequestChain, editor: FileEditor): DiffRequestChainModel {
    val chainModel = DiffRequestChainModel()
    activeEditors[chainModel] = RemoteChainEditor(chain, editor)
    diffModel.chains.add(chainModel)
    return chainModel
  }

  fun releaseRemoteEditor(chainModel: DiffRequestChainModel) {
    diffModel.chains.remove(chainModel)
  }

  fun findChainModel(file: VirtualFile): DiffRequestChainModel? {
    return activeEditors.entries.find { it.value.editor.file == file }?.key
  }

  fun findChainHandler(chainModel: DiffRequestChainModel): DiffRequestChainModelHandler? {
    return activeHandlers[chainModel]
  }

  private fun closeEditor(removedEditor: RemoteChainEditor) {
    require(ClientId.current == clientId) {
      "ClientId mismatch. Current: ${ClientId.current}, Host: $clientId"
    }

    val file = removedEditor.editor.file ?: error("Can't get file for editor $removedEditor")
    val fileEditorManager = FileEditorManager.getInstance(project) as BackendServerFileEditorManager
    noReenter(file) {
      fileEditorManager.enableClosingRemoteEditor {
        FileEditorManager.getInstance(project).closeFile(file)
      }
    }
  }

  private data class RemoteChainEditor(val chain: DiffRequestChain, val editor: FileEditor)
}

private class BackendDiffFileEditorNavigatable(val project: Project,
                                               val file: DiffVirtualFile,
                                               val chainIndex: Int,
                                               val state: Any?) : FileEditorNavigatable {
  companion object {
    fun create(project: Project, file: DiffVirtualFile, editor: FileEditor): BackendDiffFileEditorNavigatable? {
      val chainIndex: Int
      val state: Any?
      when (editor) {
        is DiffRequestRemoteEditor -> {
          val chainModel = BackendDiffViewerHost.getInstance(project, editor.clientId).findChainModel(editor.file) ?: return null
          chainIndex = chainModel.index.valueOrNull ?: return null

          val viewerModel = chainModel.viewer.valueOrNull ?: return null
          state = when (viewerModel) {
            is OneSideDiffViewerModel -> viewerModel.state.valueOrNull
            is TwoSideDiffViewerModel -> viewerModel.state.valueOrNull
            else -> null
          }
        }
        is DiffRequestProcessorEditor -> {
          val processor = editor.processor as? ChangeViewDiffRequestProcessor ?: return null
          chainIndex = processor.allChanges.toList().indexOf(processor.currentChange)

          val viewer = processor.activeViewer ?: return null
          state = when (viewer) {
            is OnesideTextDiffViewer -> OneSideState(viewer.editor.caretModel.offset)
            is TwosideTextDiffViewer -> TwoSideState(viewer.currentSide.index, viewer.currentEditor.caretModel.offset)
            else -> null
          }
        }
        else -> return null
      }

      return BackendDiffFileEditorNavigatable(project, file, chainIndex, state)
    }
  }

  override fun navigate(requestFocus: Boolean) {
    FileEditorManager.getInstance(project).openFile(file, true)
  }

  override fun canNavigate(): Boolean = file.isValid
  override fun canNavigateToSource(): Boolean = canNavigate()
  override fun getFile(): VirtualFile = file
}

private fun canExtractRequestChain(file: VirtualFile): Boolean {
  return file is SimpleDiffVirtualFile ||
         file is ChainDiffVirtualFile ||
         file is PreviewDiffVirtualFile
}

private fun extractRequestChain(file: VirtualFile): DiffRequestChain? {
  return when (file) {
    is SimpleDiffVirtualFile -> SimpleDiffRequestChain(file.request)
    is ChainDiffVirtualFile -> file.chain
    is PreviewDiffVirtualFile -> (file.provider as? ChainBackedDiffPreviewProvider)?.createDiffRequestChain() // TODO: we need to use wrapper instead to handle new modified files.
    else -> null
  }
}
