package com.jetbrains.rdserver.vcs

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.actionSystem.CommonDataKeys
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.actionSystem.DataProvider
import com.intellij.openapi.project.Project
import com.intellij.openapi.vcs.FilePath
import com.intellij.openapi.vcs.VcsDataKeys
import com.intellij.openapi.vcs.changes.Change
import com.intellij.openapi.vcs.changes.ChangesUtil
import com.intellij.openapi.vcs.changes.ui.ChangesBrowserNode
import com.intellij.openapi.vcs.changes.ui.ChangesGroupingSupport
import com.intellij.openapi.vfs.VirtualFile
import com.jetbrains.rd.ide.model.ActionTimestampSetModel
import com.jetbrains.rd.ide.model.LocalChangesTimestampModel
import com.jetbrains.rdserver.actions.*

data class LocalChangesSelection(private val windowId: Int, private val ids: List<Int>) {
  fun getHandler(project: Project): BackendLocalChangesWindowHandler? =
    BackendLocalChangesHost.getInstance(project, ClientId.current).getHandler(windowId)

  fun isValid(project: Project): Boolean {
    val host = getHandler(project) ?: return false
    return ids.all { host.getNodeById(it) != null }
  }

  fun getNodes(project: Project): List<ChangesBrowserNode<*>>? {
    val host = getHandler(project) ?: return null
    return ids.mapNotNull { host.getNodeById(it) }
  }
}

class LocalChangesTimestamp(windowId: Int, selectedNodeIds: List<Int>, private val project: Project)
  : UpgradableTimestamp<LocalChangesTimestampModel>(NAME, LocalChangesTimestampModel(windowId, selectedNodeIds)) {
  companion object {
    val NAME = ActionTimestampKey<LocalChangesTimestamp, LocalChangesTimestampModel>(BackendDataKeys.LOCAL_CHANGES_SELECTION.name)
  }

  override fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult {
    if (!getData().isValid(project)) {
      return TimestampUpgradeResult.Failure("LocalChanges was already changed, operation can not be performed")
    }

    return TimestampUpgradeResult.Success(this)
  }

  override fun getData(): LocalChangesSelection {
    return LocalChangesSelection(model.windowId, model.selectedNodeIds)
  }
}

class BackendLocalChangesTimestampProvider(project: Project) : BackendActionTimestampProvider(project) {
  override fun createSet(actionId: String, dataContext: DataContext): ActionTimestampSet {
    return ActionTimestampSet.EMPTY
  }

  override fun createSet(actionId: String, setModel: ActionTimestampSetModel): ActionTimestampSet {
    val timestampModel = setModel.getTimestamp(LocalChangesTimestamp.NAME) ?: return ActionTimestampSet.EMPTY
    return ActionTimestampSet(LocalChangesTimestamp(timestampModel.windowId, timestampModel.selectedNodeIds, project))
  }
}

class BackendLocalChangesDataProvider : BackendActionDataProvider {
  override fun getData(dataId: String, dataProvider: DataProvider): Any? {
    val project = CommonDataKeys.PROJECT.getData(dataProvider) ?: return null
    val localChangesSelection = BackendDataKeys.LOCAL_CHANGES_SELECTION.getData(dataProvider) ?: return null
    val handler = localChangesSelection.getHandler(project) ?: return null

    if (ChangesGroupingSupport.KEY.`is`(dataId)) {
      return handler.groupingSupport
    }

    val nodes = localChangesSelection.getNodes(project)
    if (!nodes.isNullOrEmpty()) {
      if (VcsDataKeys.CHANGES.`is`(dataId)) {
        return nodes.mapNotNull { it.userObject as? Change }.toTypedArray()
      }
      if (CommonDataKeys.NAVIGATABLE_ARRAY.`is`(dataId)) {
        return ChangesUtil.getNavigatableArray(project, nodes.mapNotNull {
          when (val data = it.userObject) {
            is Change -> data.virtualFile
            is VirtualFile -> data
            is FilePath -> data.virtualFile
            else -> null
          }
        }.stream())
      }
    }

    return null
  }
}