package com.jetbrains.rdserver.vcs

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.invokeLater
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.vcs.ProjectLevelVcsManager
import com.intellij.openapi.vcs.VcsListener
import com.intellij.openapi.vcs.changes.*
import com.intellij.openapi.vcs.changes.actions.diff.ShowDiffAction
import com.intellij.openapi.vcs.changes.ui.*
import com.intellij.ui.SimpleTextAttributes
import com.intellij.util.containers.JBTreeTraverser
import com.intellij.util.containers.MultiMap
import com.intellij.util.ui.tree.TreeUtil
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.rd.framework.IProtocol
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.util.toModel
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.toProtocolModel
import org.jetbrains.annotations.Nls
import javax.swing.tree.DefaultMutableTreeNode
import javax.swing.tree.TreeNode

@Suppress("NonDefaultConstructor")
class BackendLocalChangesHost(
  clientId: ClientId,
  lifetime: Lifetime,
  val project: Project,
  protocol: IProtocol
) {
  companion object {
    fun getInstance(project: Project, clientId: ClientId): BackendLocalChangesHost {
      return RdServerPerClientComponents.getService(project, clientId, BackendLocalChangesHost::class.java)
    }
  }

  private val model = LocalChangesModel.create(lifetime, protocol)
  private val handlers = mutableMapOf<Int, BackendLocalChangesWindowHandler>()

  init {
    model.windows.view(lifetime) { windowLifetime, _, windowModel ->
      val handler = BackendLocalChangesWindowHandler(project, windowLifetime, clientId, windowModel)
      handlers[windowModel.windowId] = handler
      windowLifetime.onTermination { handlers.remove(windowModel.windowId) ?: error("Handler not found: ${windowModel.windowId}") }
    }

    project.messageBus.connect(lifetime.createNestedDisposable()).subscribe(ProjectLevelVcsManager.VCS_CONFIGURATION_CHANGED,
                                                                            VcsListener { invokeLater { updateAvailability() } })
    updateAvailability()
  }

  private fun updateAvailability() {
    if (project.isDisposed) return
    model.isAvailable.set(ProjectLevelVcsManager.getInstance(project).hasActiveVcss())
  }

  fun getHandler(windowId: Int): BackendLocalChangesWindowHandler? = handlers[windowId]
}


class BackendLocalChangesWindowHandler(val project: Project,
                                       lifetime: Lifetime,
                                       val clientId: ClientId,
                                       val model: LocalChangesWindowModel) {
  private val disposable = lifetime.createNestedDisposable()

  private val queue = MergingUpdateQueue("BackendLocalChangesHost", 300, true, null, disposable)
  val groupingSupport = ChangesGroupingSupport(project, this, false)

  private var lastNodeId: Int = 0
  private var idToDataMap: Map<Int, ChangesBrowserNode<*>> = emptyMap()
  private var roots = listOf<Node>()

  init {
    model.showDiff.set { nodeId ->
      val node = idToDataMap[nodeId] ?: return@set
      val change = node.userObject as? Change ?: return@set
      ShowDiffAction.showDiffForChange(project, listOf(change))
    }

    groupingSupport.setGroupingKeysOrSkip(ChangesTree.DEFAULT_GROUPING_KEYS.toSet())
    groupingSupport.addPropertyChangeListener { rebuildTree() }

    ChangeListManager.getInstance(project).addChangeListListener(object : ChangeListAdapter() {
      override fun changeListsChanged() {
        scheduleUpdate()
      }
    }, disposable)
    rebuildTree()
  }

  fun getNodeById(id: Int): ChangesBrowserNode<*>? = idToDataMap[id]

  private fun scheduleUpdate() {
    queue.queue(Update.create("update") {
      rebuildTree()
    })
  }

  private fun rebuildTree() {
    val changeListManager: ChangeListManagerImpl = ChangeListManagerImpl.getInstanceImpl(project)

    val treeModel = TreeModelBuilder(project, groupingSupport.grouping)
      .setChangeLists(changeListManager.changeListsCopy, false, null)
      .setLocallyDeletedPaths(changeListManager.deletedFiles)
      .setModifiedWithoutEditing(changeListManager.modifiedWithoutEditing)
      .setSwitchedFiles(changeListManager.switchedFilesMap)
      .setSwitchedRoots(changeListManager.switchedRoots)
      .setLockedFolders(changeListManager.lockedFolders)
      .setLogicallyLockedFiles(changeListManager.logicallyLockedFolders)
      .setUnversioned(changeListManager.unversionedFilesPaths)
      .build()

    val newRoots = mutableListOf<Node>()
    val nodeMap = mutableMapOf<ChangesBrowserNode<*>, Node>()
    val idToNodeMap = mutableMapOf<Int, Node>()
    val nodeModelRenderer = NodeModelRenderer(project)

    TreeUtil.treeNodeTraverser(treeModel.root as DefaultMutableTreeNode).preOrderDfsTraversal().forEach { node ->
      node as ChangesBrowserNode<*>

      if (node != treeModel.root) {
        val nodeId = ++lastNodeId
        val nodePresentation = node.toPresentationModel(nodeModelRenderer)

        val newNode = Node(nodeId, nodePresentation)
        nodeMap[node] = newNode
        idToNodeMap[nodeId] = newNode

        val parent = nodeMap[node.parent]
        if (parent == null) {
          newRoots += newNode
        }
        else {
          parent.children += newNode
        }
      }
    }

    tryReuseOldIds(roots, newRoots)
    tryReuseOldLeafNodes(roots, newRoots)

    roots = newRoots
    idToDataMap = mutableMapOf<Int, ChangesBrowserNode<*>>().also {
      nodeMap.forEach { (vcsNode, node) ->
        if (!node.presentation.isHelperNode) {
          val oldNode = it.put(node.nodeId, vcsNode)
          assert(oldNode == null) { "Duplicated ids in tree: $node - $oldNode" }
        }
      }
    }

    val allNodesIds = nodeMap.values.mapTo(mutableSetOf()) { node -> node.nodeId }
    val oldNodesIds = model.tree.nodes.keys
    val newNodes = allNodesIds - oldNodesIds
    val removedNodes = oldNodesIds - allNodesIds

    for (nodeId in removedNodes) {
      model.tree.nodes.remove(nodeId)
    }
    for (nodeId in newNodes) {
      val node = idToNodeMap[nodeId]!! // new nodes have their ids unchanged, so we can use map built before 'tryReuseOldIds' call
      model.tree.nodes[node.nodeId] = RdLocalChangesItem(node.children.map { it.nodeId },
                                                         node.presentation.text,
                                                         node.presentation.icon,
                                                         node.presentation.isHelperNode,
                                                         node.presentation.backgroundColor)
    }
    model.tree.roots.set(newRoots.map { it.nodeId })
  }
}

private fun tryReuseOldIds(oldChildren: List<Node>, newChildren: List<Node>) {
  var lastIndex = 0
  for (newNode in newChildren) {
    val index = oldChildren.indexOfFirst(lastIndex) {
      it.presentation.text.firstOrNull() == newNode.presentation.text.firstOrNull()
    }
    if (index != -1) {
      val oldNode = oldChildren[index]
      lastIndex = index + 1

      tryReuseOldIds(oldNode.children, newNode.children)

      if (newNode.presentation == oldNode.presentation &&
          newNode.children.map { it.nodeId } == oldNode.children.map { it.nodeId }) {
        newNode.nodeId = oldNode.nodeId
      }
    }
  }
}

private fun tryReuseOldLeafNodes(oldRoots: List<Node>, newRoots: List<Node>) {
  val oldLeafs = MultiMap<NodePresentationModel, Node>()
  val newLeafs = MultiMap<NodePresentationModel, Node>()
  val oldLeafsIds = mutableSetOf<Int>()
  val newLeafsIds = mutableSetOf<Int>()

  traverseNodes(oldRoots).forEach {
    if (it.children.isEmpty()) {
      oldLeafs.putValue(it.presentation, it)
      oldLeafsIds += it.nodeId
    }
  }
  traverseNodes(newRoots).forEach {
    if (it.children.isEmpty()) {
      newLeafs.putValue(it.presentation, it)
      newLeafsIds += it.nodeId
    }
  }

  for (presentation in newLeafs.keySet()) {
    // ensure no duplicated ids in resulted tree
    val oldNodes = oldLeafs[presentation].filter { !newLeafsIds.contains(it.nodeId) }.iterator()
    val newNodes = newLeafs[presentation].filter { !oldLeafsIds.contains(it.nodeId) }.iterator()
    while (oldNodes.hasNext() && newNodes.hasNext()) {
      newNodes.next().nodeId = oldNodes.next().nodeId
    }
  }
}

private fun traverseNodes(roots: List<Node>) = JBTreeTraverser.from<Node> { it.children }.withRoots(roots).postOrderDfsTraversal()

private class Node(var nodeId: Int, val presentation: NodePresentationModel) {
  val children = mutableListOf<Node>()
}

private data class NodePresentationModel(val text: List<RdTextChunk>,
                                         val icon: IconModel?,
                                         val isHelperNode: Boolean,
                                         val backgroundColor: Int?)

private fun TreeNode.toPresentationModel(renderer: NodeModelRenderer): NodePresentationModel {
  this as ChangesBrowserNode<*>
  renderer.clear()
  render(renderer, false, false, false)
  return NodePresentationModel(renderer.textFragments.toList(), renderer.icon?.toProtocolModel(), !isMeaningfulNode,
                               getBackgroundColor(renderer.project)?.rgb)
}

private class NodeModelRenderer(val project: Project) : ChangesBrowserNodeRenderer(project, { false }, false) {
  val textFragments = mutableListOf<RdTextChunk>()

  override fun clear() {
    super.clear()
    textFragments.clear()
  }

  override fun append(@Nls fragment: String, attributes: SimpleTextAttributes, isMainText: Boolean) {
    super.append(fragment, attributes, isMainText)
    textFragments += RdTextChunk(fragment, attributes.toModel())
  }
}

private inline fun <T> List<T>.indexOfFirst(startIndex: Int, predicate: (T) -> Boolean): Int {
  var index = startIndex
  val len = size
  while (index < len) {
    val value: T = get(index)
    if (predicate(value)) return index
    index++
  }
  return -1
}
