package com.jetbrains.rdserver.vcs

import com.intellij.diff.util.DiffUtil
import com.intellij.openapi.actionSystem.DataContext
import com.intellij.openapi.editor.Document
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.markup.HighlighterTargetArea
import com.intellij.openapi.project.Project
import com.intellij.openapi.util.TextRange
import com.intellij.openapi.vcs.ex.LineStatusActionBase
import com.intellij.openapi.vcs.ex.LineStatusMarkerRenderer
import com.intellij.openapi.vcs.ex.LineStatusTracker
import com.intellij.openapi.vcs.ex.Range
import com.intellij.openapi.vcs.impl.LineStatusTrackerManager
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.actions.*
import com.jetbrains.rdserver.actions.timestamps.OffsetTimestamp
import com.jetbrains.rdserver.daemon.*

class VcsGutterMarkHighlighterModelCreator(val project: Project) : BackendHighlighterModelCreator {
  override fun accept(highlighter: RangeHighlighterEx): Boolean {
    return highlighter.getUserData(LineStatusMarkerRenderer.TOOLTIP_KEY) != null ||
           highlighter.getUserData(LineStatusMarkerRenderer.MAIN_KEY) != null
  }

  override fun createModel(highlighter: RangeHighlighterEx, documentVersion: RdDocumentVersion): HighlighterModel? {
    val data = highlighter.getUserData(LineStatusMarkerRenderer.TOOLTIP_KEY) ?: return null
    return VcsGutterMarkHighlighterModel(
      data.type,
      highlighter.layer,
      highlighter.targetArea == HighlighterTargetArea.EXACT_RANGE,
      documentVersion,
      highlighter.isGreedyToLeft,
      highlighter.isGreedyToRight,
      highlighter.isThinErrorStripeMark,
      null,
      highlighter.id,
      BackendHighlighterRegistrationsHost.VCS_MARK_ID,
      highlighter.startOffset,
      highlighter.endOffset
    )
  }
}

class BackendVcsGutterMarkExtensionHandler(val project: Project) : BackendProtocolMarkupModelExtensionHandler {
  override fun consume(lifetime: Lifetime, extensions: List<MarkupModelExtension>, contributor: BackendMarkupModelContributor) {
    if (contributor is BackendEditorMarkupModelContributor) return
    val vcsMarkersModel = extensions.find { it.key == "VCS_GUTTER_PROVIDER_MODEL_KEY" } as? VcsGutterMarkersModel ?: return
    EditorMarkersContributor(lifetime, contributor.document, contributor, vcsMarkersModel)
  }

  private inner class EditorMarkersContributor(lifetime: Lifetime,
                                               val document: Document,
                                               val contributor: BackendMarkupModelContributor,
                                               vcsGutterMarkersModel: VcsGutterMarkersModel) {
    init {
      val lstm = LineStatusTrackerManager.getInstance(project)
      lstm.requestTrackerFor(document, this)
      lifetime.onTermination { lstm.releaseTrackerFor(document, this) }

      vcsGutterMarkersModel.getOldContent.set { id -> getRangeOldContent(id) }
    }

    private fun getRangeOldContent(highlighterId: Long): String? {
      return performRangeOperation(highlighterId) { tracker, range ->
        val textRange = DiffUtil.getLinesRange(tracker.vcsDocument, range.vcsLine1, range.vcsLine2)
        textRange.subSequence(tracker.vcsDocument.immutableCharSequence).toString()
      }
    }

    private fun <T> performRangeOperation(highlighterId: Long, task: (LineStatusTracker<*>, Range) -> T): T? {
      val tracker = LineStatusTrackerManager.getInstance(project).getLineStatusTracker(document) ?: return null

      val highlighter = contributor.getRangeMarker(highlighterId) ?: return null
      val textRange = TextRange(highlighter.startOffset, highlighter.endOffset)

      val line = document.getLineNumber(textRange.startOffset)
      val range = tracker.getRangeForLine(line) ?: return null

      return task(tracker, range)
    }
  }
}

class VcsSelectedLineTimestamp(private val offsetTimestamp: OffsetTimestamp)
  : UpgradableTimestamp<VcsSelectedLineTimestampModel>(NAME, VcsSelectedLineTimestampModel(offsetTimestamp.model)) {
  companion object {
    val NAME = ActionTimestampKey<VcsSelectedLineTimestamp, VcsSelectedLineTimestampModel>(LineStatusActionBase.SELECTED_OFFSET_KEY.name)
  }

  override fun upgrade(actualSet: ActionTimestampSet): TimestampUpgradeResult {
    val offsetUpgradeResult = offsetTimestamp.upgrade(actualSet)
    if (offsetUpgradeResult !is TimestampUpgradeResult.Success<*>) return offsetUpgradeResult
    val upgradedOffset = offsetUpgradeResult.timestamp as OffsetTimestamp

    return TimestampUpgradeResult.Success(VcsSelectedLineTimestamp(upgradedOffset))
  }

  override fun getData(): Any? = model.offset.offset
}

class VcsSelectedLineTimestampModelProvider(project: Project) : BackendActionTimestampProvider(project) {
  override fun createSet(actionId: String, dataContext: DataContext): ActionTimestampSet {
    return ActionTimestampSet.EMPTY
  }

  override fun createSet(actionId: String, setModel: ActionTimestampSetModel): ActionTimestampSet {
    val timestampModel = setModel.getTimestamp(VcsSelectedLineTimestamp.NAME) ?: return ActionTimestampSet.EMPTY

    val offsetModel = timestampModel.offset
    val offsetTimestamp = OffsetTimestamp(offsetModel.offset, offsetModel.isStickingToRight, offsetModel.document)

    return ActionTimestampSet(VcsSelectedLineTimestamp(offsetTimestamp))
  }
}