package com.jetbrains.rdserver.vcs

import com.intellij.codeWithMe.ClientId
import com.intellij.diff.DiffContext
import com.intellij.diff.chains.AsyncDiffRequestChain
import com.intellij.diff.chains.DiffRequestChain
import com.intellij.diff.contents.DocumentContent
import com.intellij.diff.contents.EmptyContent
import com.intellij.diff.requests.ContentDiffRequest
import com.intellij.diff.requests.DiffRequest
import com.intellij.diff.tools.util.base.TextDiffViewerUtil
import com.intellij.diff.util.DiffUtil
import com.intellij.openapi.ListSelection
import com.intellij.openapi.application.invokeLater
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.diff.DiffBundle
import com.intellij.openapi.fileEditor.FileDocumentManager
import com.intellij.openapi.progress.runBackgroundableTask
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.Key
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.vcs.toRdModel
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.CodeWithMePluginBundle
import com.jetbrains.rdserver.editors.BackendEditorHost
import com.jetbrains.rdserver.editors.BackendServerFileEditorManager.Companion.editorClientId
import com.jetbrains.rdserver.permissions.PermissionManager
import kotlin.math.max
import kotlin.math.min

class DiffRequestChainModelHandler(val project: Project,
                                   val lifetime: Lifetime,
                                   val clientId: ClientId,
                                   val chain: DiffRequestChain,
                                   val chainModel: DiffRequestChainModel) {
  private var index: Int = 0

  private val context = DumbDiffContext(project, chain)
  private var activeRequest: DiffRequest? = null
  private var pendingViewerState: Any? = null

  init {
    lifetime.onTermination {
      activeRequest?.onAssigned(false)
      activeRequest = null
    }

    chainModel.viewer.view(lifetime) { viewerLifetime, viewerModel ->
      if (viewerModel is OneSideDiffViewerModel) {
        OneSideRemoteDiffViewer(project, viewerLifetime, clientId, viewerModel)
      }
      if (viewerModel is TwoSideDiffViewerModel) {
        TwoSideRemoteDiffViewer(project, viewerLifetime, clientId, viewerModel)
      }
    }
    chainModel.goPrev.set { _ ->
      index = max(index - 1, 0)
      updateRequest()
    }
    chainModel.goNext.set { _ ->
      index = min(index + 1, chain.requests.size)
      updateRequest()
    }


    if (chain is AsyncDiffRequestChain) {
      chain.addListener(
        {
          index = chain.index
          updateRequest()
        }, lifetime.createNestedDisposable())

      chain.onAssigned(true)
      lifetime.onTermination { chain.onAssigned(false) }
    }

    index = chain.index
    chainModel.index.set(index)
    updateRequest()
  }

  private fun updateRequest() {
    val listSelection = ListSelection.createAt(chain.requests, index)

    chainModel.hasNext.set(index < listSelection.list.size - 1)
    chainModel.hasPrev.set(index > 0)
    chainModel.index.set(index)

    val selectedProducer = listSelection.list[listSelection.selectedIndex]
    runBackgroundableTask(DiffBundle.message("progress.title.loading.requests")) { indicator ->
      val request = selectedProducer.process(context, indicator)
      invokeLater {
        activeRequest?.onAssigned(false)
        activeRequest = request
        activeRequest?.onAssigned(true)

        val viewerModel = createViewerModel(project, context, request)
                          ?: ErrorDiffViewerModel(DiffBundle.message("error.message.cannot.show.diff"), null)
        chainModel.viewer.set(viewerModel)
        updateViewerState()
      }
    }
  }

  fun navigateTo(chainIndex: Int, viewerState: Any?) {
    pendingViewerState = viewerState

    if (chainIndex != index) {
      index = chainIndex
      updateRequest()
    }
    else {
      updateViewerState()
    }
  }

  private fun updateViewerState() {
    val model = chainModel.viewer.valueOrNull ?: return

    val newState = pendingViewerState
    pendingViewerState = null

    if (model is TwoSideDiffViewerModel && newState is TwoSideState) {
      model.state.set(newState)
    }
    if (model is OneSideDiffViewerModel && newState is OneSideState) {
      model.state.set(newState)
    }
  }

  companion object {
    private val LOG = logger<DiffRequestChainModelHandler>()

    private fun createViewerModel(project: Project, context: DiffContext, request: DiffRequest): DiffViewerModel? {
      try {
        if (request !is ContentDiffRequest || request.contents.size != 2) {
          LOG.warn("Request is not supported by backend: $request")
          return null
        }
        val content1 = request.contents[0]
        val content2 = request.contents[1]

        val textSettings = TextDiffViewerUtil.getTextSettings(context)
        val readOnlyFlags = TextDiffViewerUtil.checkForceReadOnly(context, request)

        if (content1 is DocumentContent && content2 is DocumentContent) {
          if (!canOpenDocument(project, content1)) {
            return ErrorDiffViewerModel(CodeWithMePluginBundle.message("diff.error.permission.cant.open.file"), null)
          }
          if (!canOpenDocument(project, content2)) {
            return ErrorDiffViewerModel(CodeWithMePluginBundle.message("diff.error.permission.cant.open.file"), null)
          }

          val contentModel1 = createDiffContentModel(project, content1, request.contentTitles[0], readOnlyFlags[0])
          val contentModel2 = createDiffContentModel(project, content2, request.contentTitles[1], readOnlyFlags[1])
          return TwoSideDiffViewerModel(contentModel1, contentModel2, request.title).apply {
            settings.set(textSettings.toRdModel())
          }
        }

        if (content1 is DocumentContent && content2 is EmptyContent) {
          if (!canOpenDocument(project, content1)) {
            return ErrorDiffViewerModel(CodeWithMePluginBundle.message("diff.error.permission.cant.open.file"), null)
          }

          val contentModel = createDiffContentModel(project, content1, request.contentTitles[0], readOnlyFlags[0])
          return OneSideDiffViewerModel(contentModel, true, request.title)
        }

        if (content1 is EmptyContent && content2 is DocumentContent) {
          if (!canOpenDocument(project, content2)) {
            return ErrorDiffViewerModel(CodeWithMePluginBundle.message("diff.error.permission.cant.open.file"), null)
          }

          val contentModel = createDiffContentModel(project, content2, request.contentTitles[1], readOnlyFlags[1])
          return OneSideDiffViewerModel(contentModel, false, request.title)
        }

        LOG.warn("Request is not supported by backend: $request")
        return null
      }
      catch (e: Throwable) {
        LOG.error(e)
        return null
      }
    }

    private fun createDiffContentModel(project: Project,
                                       documentContent: DocumentContent,
                                       title: String?,
                                       readOnly: Boolean): DiffViewerContentModel {
      val document = documentContent.document
      val editor = DiffUtil.createEditor(document, project, readOnly || !document.isWritable, true)
      DiffUtil.configureEditor(editor, documentContent, project)
      editor.editorClientId = ClientId.current

      val textControlId = BackendEditorHost.getInstance(project).bindEditor(editor)
      return DiffViewerContentModel(textControlId, title)
    }

    private fun canOpenDocument(project: Project, content: DocumentContent): Boolean {
      val file = FileDocumentManager.getInstance().getFile(content.document)
      if (file == null) return true
      return PermissionManager.getInstance(project).isVisible(ClientId.current, file)
    }
  }
}

private class DumbDiffContext(val _project: Project?, val chain: DiffRequestChain) : DiffContext() {
  override fun getProject(): Project? = _project
  override fun isWindowFocused(): Boolean = false
  override fun isFocusedInWindow(): Boolean = false
  override fun requestFocusInWindow() = Unit

  override fun <T : Any?> getUserData(key: Key<T>): T? {
    return chain.getUserData(key)
  }

  override fun <T : Any?> putUserData(key: Key<T>, value: T?) {
    chain.putUserData(key, value)
  }
}
