package com.jetbrains.rdserver.vcs

import com.intellij.codeWithMe.ClientId
import com.intellij.diff.comparison.ComparisonManager
import com.intellij.diff.fragments.DiffFragment
import com.intellij.diff.fragments.LineFragment
import com.intellij.diff.fragments.LineFragmentImpl
import com.intellij.diff.tools.util.FoldingModelSupport
import com.intellij.diff.tools.util.base.HighlightPolicy
import com.intellij.diff.tools.util.base.IgnorePolicy
import com.intellij.diff.tools.util.text.LineOffsetsUtil
import com.intellij.diff.tools.util.text.SimpleTextDiffProvider
import com.intellij.diff.util.DiffUtil
import com.intellij.diff.util.Side
import com.intellij.openapi.actionSystem.AnAction
import com.intellij.openapi.application.runReadAction
import com.intellij.openapi.diff.DiffBundle
import com.intellij.openapi.editor.EditorFactory
import com.intellij.openapi.editor.event.DocumentEvent
import com.intellij.openapi.editor.event.DocumentListener
import com.intellij.openapi.editor.ex.EditorEx
import com.intellij.openapi.editor.ex.MarkupModelEx
import com.intellij.openapi.editor.ex.RangeHighlighterEx
import com.intellij.openapi.editor.markup.GutterIconRenderer
import com.intellij.openapi.editor.markup.HighlighterLayer
import com.intellij.openapi.editor.markup.HighlighterTargetArea
import com.intellij.openapi.editor.markup.RangeHighlighter
import com.intellij.openapi.progress.DumbProgressIndicator
import com.intellij.openapi.project.DumbAwareAction
import com.intellij.openapi.project.Project
import com.intellij.openapi.rd.createNestedDisposable
import com.intellij.openapi.util.TextRange
import com.intellij.util.containers.PeekableIterator
import com.intellij.util.containers.PeekableIteratorWrapper
import com.intellij.util.ui.update.MergingUpdateQueue
import com.intellij.util.ui.update.Update
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.codeWithMe.vcs.toPolicy
import com.jetbrains.rd.platform.util.userData
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.onTermination
import com.jetbrains.rdserver.daemon.BackendHighlighterModelCreator
import com.jetbrains.rdserver.daemon.BackendHighlighterRegistrationsHost
import com.jetbrains.rdserver.daemon.markupModelContributor
import com.jetbrains.rdserver.editors.BackendEditorHost
import javax.swing.Icon

abstract class RemoteDiffViewer(val project: Project,
                                val lifetime: Lifetime,
                                val clientId: ClientId,
                                open val model: DiffViewerModel)

class OneSideRemoteDiffViewer(project: Project,
                              lifetime: Lifetime,
                              clientId: ClientId,
                              override val model: OneSideDiffViewerModel)
  : RemoteDiffViewer(project, lifetime, clientId, model) {
  init {
    val editor = BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(model.content.editorId)
    lifetime.onTermination {
      if (editor != null) EditorFactory.getInstance().releaseEditor(editor)
    }
  }
}

class TwoSideRemoteDiffViewer(project: Project,
                              lifetime: Lifetime,
                              clientId: ClientId,
                              override val model: TwoSideDiffViewerModel)
  : RemoteDiffViewer(project, lifetime, clientId, model) {
  private val editor1: EditorEx
  private val editor2: EditorEx
  private val editors: List<EditorEx>

  private val disposable = lifetime.createNestedDisposable()
  private val queue = MergingUpdateQueue("TwoSideRemoteDiffViewer", 300, true, null, disposable)

  private val highlighters = mutableMapOf<Any, MutableList<RangeHighlighter>>()
  private var lastChangeId: Long = 0
  private var documentStateWasChanged: Boolean = false

  init {
    val editor1 = BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(model.content1.editorId) as? EditorEx
    val editor2 = BackendEditorHost.getInstance(project).tryGetOwnRemoteEditor(model.content2.editorId) as? EditorEx
    lifetime.onTermination {
      if (editor1 != null) EditorFactory.getInstance().releaseEditor(editor1)
      if (editor2 != null) EditorFactory.getInstance().releaseEditor(editor2)
    }
    this.editor1 = editor1 ?: error("Editor not found: ${model.content1.editorId}")
    this.editor2 = editor2 ?: error("Editor not found: ${model.content2.editorId}")
    editors = listOf(editor1, editor2)

    model.getFoldingDescription.set { request ->
      // TODO: which thread is it?
      runReadAction {
        val document = listOf(editor1, editor2)[request.editorIndex].document
        val result = FoldingModelSupport.getLineSeparatorDescription(project, document, request.lineNumber) ?: return@runReadAction null
        DiffFoldingDescriptionReply(result.description, result.anchorLine)
      }
    }


    val documentListener = object : DocumentListener {
      override fun documentChanged(event: DocumentEvent) {
        documentStateWasChanged = true
        scheduleRediff()
      }
    }
    editor1.document.addDocumentListener(documentListener, disposable)
    editor2.document.addDocumentListener(documentListener, disposable)

    model.settings.advise(lifetime) {
      scheduleRediff()
    }

    scheduleRediff()
  }

  private fun scheduleRediff() {
    queue.queue(Update.create("rediff") { rediff() })
  }

  private fun rediff() {
    documentStateWasChanged = false

    val text1 = editor1.document.immutableCharSequence
    val text2 = editor2.document.immutableCharSequence
    val lineOffsets1 = LineOffsetsUtil.create(editor1.document)
    val lineOffsets2 = LineOffsetsUtil.create(editor2.document)
    val highlightPolicy = model.settings.valueOrNull?.highlightPolicy?.toPolicy() ?: HighlightPolicy.BY_WORD
    val ignorePolicy = model.settings.valueOrNull?.ignorePolicy?.toPolicy() ?: IgnorePolicy.DEFAULT

    val rawFragments = SimpleTextDiffProvider.compareRange(
      null, text1, text2, lineOffsets1, lineOffsets2, null,
      ignorePolicy, highlightPolicy, DumbProgressIndicator.INSTANCE
    ).single() // linesRanges == null
    val fragments = ComparisonManager.getInstance().processBlocks(
      rawFragments, text1, text2,
      ignorePolicy.comparisonPolicy, highlightPolicy.isShouldSquash, ignorePolicy.isShouldTrimChunks)

    Lifetime.using { lifetime ->
      editor1.document.markupModelContributor?.suppressor?.suppress(lifetime, "diff highlighting")
      editor2.document.markupModelContributor?.suppressor?.suppress(lifetime, "diff highlighting")

      updateBlockHighlighters(fragments)
      updateInnerHighlighters(fragments, Side.LEFT)
      updateInnerHighlighters(fragments, Side.RIGHT)
      updateGutterOperationActions(fragments, Side.LEFT)
      updateGutterOperationActions(fragments, Side.RIGHT)
    }
  }

  private fun updateBlockHighlighters(fragments: List<LineFragment>) {
    val highlighters1 = highlighters.computeIfAbsent(BlockType(Side.LEFT)) { mutableListOf() }
    val highlighters2 = highlighters.computeIfAbsent(BlockType(Side.RIGHT)) { mutableListOf() }
    val markupModel1 = editor1.markupModel
    val markupModel2 = editor2.markupModel

    val highlighterIt1: PeekableIterator<RangeHighlighter> = PeekableIteratorWrapper(highlighters1.iterator())
    val newHighlighters1 = mutableListOf<RangeHighlighter>()
    val oldHighlighters1 = mutableListOf<RangeHighlighter>()
    val highlighterIt2: PeekableIterator<RangeHighlighter> = PeekableIteratorWrapper(highlighters2.iterator())
    val newHighlighters2 = mutableListOf<RangeHighlighter>()
    val oldHighlighters2 = mutableListOf<RangeHighlighter>()

    for (fragment in fragments) {
      val type = fragment.lineFragmentType
      val isIgnored = fragment.innerFragments != null
      val isEmptyRange1 = fragment.startLine1 == fragment.endLine1
      val isEmptyRange2 = fragment.startLine2 == fragment.endLine2
      val isEOF1 = fragment.endLine1 == DiffUtil.getLineCount(editor1.document)
      val isEOF2 = fragment.endLine2 == DiffUtil.getLineCount(editor2.document)
      val textRange1 = TextRange(fragment.startOffset1, fragment.endOffset1)
      val textRange2 = TextRange(fragment.startOffset2, fragment.endOffset2)

      while (highlighterIt1.hasNext() &&
             highlighterIt1.peek().startOffset < textRange1.startOffset) {
        oldHighlighters1.add(highlighterIt1.next())
      }
      while (highlighterIt2.hasNext() &&
             highlighterIt2.peek().startOffset < textRange2.startOffset) {
        oldHighlighters2.add(highlighterIt2.next())
      }

      val oldHighlighter1 = if (highlighterIt1.hasNext()) highlighterIt1.peek() else null
      val oldHighlighter2 = if (highlighterIt2.hasNext()) highlighterIt2.peek() else null
      val model1 = oldHighlighter1?.blockModel
      val model2 = oldHighlighter2?.blockModel
      if (model1 != null &&
          model2 != null &&
          model1.changeId == model2.changeId &&
          oldHighlighter1.startOffset == textRange1.startOffset &&
          oldHighlighter1.endOffset == textRange1.endOffset &&
          oldHighlighter2.startOffset == textRange2.startOffset &&
          oldHighlighter2.endOffset == textRange2.endOffset &&
          model1.type == type && model1.isIgnored == isIgnored && model1.isEmptyRange == isEmptyRange1 && model1.isEOF == isEOF1 &&
          model2.type == type && model2.isIgnored == isIgnored && model2.isEmptyRange == isEmptyRange2 && model2.isEOF == isEOF2) {
        newHighlighters1.add(oldHighlighter1)
        newHighlighters2.add(oldHighlighter2)
        highlighterIt1.next()
        highlighterIt2.next()
      }
      else {
        val changeId = lastChangeId++
        newHighlighters1.add(markupModel1.addRangeHighlighterAndChangeAttributes(
          textRange1.startOffset, textRange1.endOffset,
          HighlighterLayer.LAST, null, HighlighterTargetArea.EXACT_RANGE, false
        ) {
          it.blockModel = BlockModel(type, changeId, isIgnored, isEmptyRange1, isEOF1)
        })
        newHighlighters2.add(markupModel2.addRangeHighlighterAndChangeAttributes(
          textRange2.startOffset, textRange2.endOffset,
          HighlighterLayer.LAST, null, HighlighterTargetArea.EXACT_RANGE, false
        ) {
          it.blockModel = BlockModel(type, changeId, isIgnored, isEmptyRange2, isEOF2)
        })
      }
    }

    while (highlighterIt1.hasNext()) {
      oldHighlighters1.add(highlighterIt1.next())
    }
    while (highlighterIt2.hasNext()) {
      oldHighlighters2.add(highlighterIt2.next())
    }

    for (highlighter in oldHighlighters1) {
      highlighter.dispose()
    }
    for (highlighter in oldHighlighters2) {
      highlighter.dispose()
    }

    highlighters1.clear()
    highlighters1.addAll(newHighlighters1)
    highlighters2.clear()
    highlighters2.addAll(newHighlighters2)
  }

  data class BlockType(val side: Side)

  private fun updateInnerHighlighters(fragments: List<LineFragment>, side: Side) {
    val sequence = fragments.asSequence().flatMap { lineFragment ->
      val innerFragments = lineFragment.innerFragments.orEmpty().asSequence()
      innerFragments.map { innerFragment ->
        val textRange = TextRange(side.getStartOffset(lineFragment) + side.getStartOffset(innerFragment),
                                  side.getStartOffset(lineFragment) + side.getEndOffset(innerFragment))
        Pair(textRange, innerFragment)
      }
    }
    updateHighlighters(sequence, InnerType(side))
  }

  private class InnerType(side: Side) : DiffHighlighterType<DiffFragment>(side) {
    override fun isSameHighlighter(highlighter: RangeHighlighter, fragment: DiffFragment): Boolean {
      val model = highlighter.innerModel
      return model != null && model.type == fragment.innerFragmentType
    }

    override fun createHighlighter(markupModel: MarkupModelEx, textRange: TextRange, fragment: DiffFragment): RangeHighlighter {
      return markupModel.addRangeHighlighterAndChangeAttributes(
        textRange.startOffset, textRange.endOffset,
        HighlighterLayer.LAST, null, HighlighterTargetArea.EXACT_RANGE, false
      ) {
        it.innerModel = InnerModel(fragment.innerFragmentType)
      }
    }
  }

  private fun updateGutterOperationActions(fragments: List<LineFragment>, source: Side) {
    val targetEditor = source.other().selectNotNull(editors)
    if (targetEditor.isViewer || !targetEditor.document.isWritable) return

    val side = source
    val sequence = fragments.asSequence().map { lineFragment ->
      val textRange = TextRange(side.getStartOffset(lineFragment), side.getEndOffset(lineFragment))
      Pair(textRange, lineFragment)
    }
    updateHighlighters(sequence, GutterOperationType(side))
  }

  private inner class GutterOperationType(side: Side) : DiffHighlighterType<LineFragment>(side) {
    override fun isSameHighlighter(highlighter: RangeHighlighter, fragment: LineFragment): Boolean {
      val renderer = highlighter.gutterIconRenderer as? MyApplyGutterAction
      return renderer != null &&
             renderer.fragment.startLine1 == fragment.startLine1 &&
             renderer.fragment.endLine1 == fragment.endLine1 &&
             renderer.fragment.startLine2 == fragment.startLine2 &&
             renderer.fragment.endLine2 == fragment.endLine2
    }

    override fun createHighlighter(markupModel: MarkupModelEx, textRange: TextRange, fragment: LineFragment): RangeHighlighter {
      return markupModel.addRangeHighlighter(textRange.startOffset, textRange.endOffset,
                                             HighlighterLayer.LAST, null, HighlighterTargetArea.EXACT_RANGE).apply {
        gutterIconRenderer = MyApplyGutterAction(LineFragmentImpl(fragment, null), side)
      }
    }
  }

  private fun <T> updateHighlighters(iterable: Sequence<Pair<TextRange, T>>, type: DiffHighlighterType<T>) {
    val markupModel = type.side.select(editors).markupModel
    val highlighters = highlighters.computeIfAbsent(type) { mutableListOf() }

    val highlighterIt: PeekableIterator<RangeHighlighter> = PeekableIteratorWrapper(highlighters.iterator())
    val newHighlighters = mutableListOf<RangeHighlighter>()
    val oldHighlighters = mutableListOf<RangeHighlighter>()

    for ((textRange, fragment) in iterable) {
      while (highlighterIt.hasNext() &&
             highlighterIt.peek().startOffset < textRange.startOffset) {
        oldHighlighters.add(highlighterIt.next())
      }
      val oldHighlighter = if (highlighterIt.hasNext()) highlighterIt.peek() else null
      if (oldHighlighter != null &&
          oldHighlighter.startOffset == textRange.startOffset &&
          oldHighlighter.endOffset == textRange.endOffset &&
          type.isSameHighlighter(oldHighlighter, fragment)) {
        // reuse existing highlighter if possible
        newHighlighters.add(oldHighlighter)
        highlighterIt.next()
      }
      else {
        newHighlighters.add(type.createHighlighter(markupModel, textRange, fragment))
      }
    }

    while (highlighterIt.hasNext()) {
      oldHighlighters.add(highlighterIt.next())
    }

    for (highlighter in oldHighlighters) {
      highlighter.dispose()
    }

    highlighters.clear()
    highlighters.addAll(newHighlighters)
  }

  private abstract class DiffHighlighterType<T>(val side: Side) {
    abstract fun isSameHighlighter(highlighter: RangeHighlighter, fragment: T): Boolean
    abstract fun createHighlighter(markupModel: MarkupModelEx, textRange: TextRange, fragment: T): RangeHighlighter

    override fun equals(other: Any?): Boolean {
      return other != null &&
             other.javaClass == javaClass &&
             (other as DiffHighlighterType<*>).side == side
    }

    override fun hashCode(): Int {
      return javaClass.hashCode() * side.hashCode()
    }
  }

  private inner class MyApplyGutterAction(val fragment: LineFragment, val source: Side) : GutterIconRenderer() {
    override fun getIcon(): Icon = DiffUtil.getArrowIcon(source)
    override fun equals(other: Any?): Boolean = this === other
    override fun hashCode(): Int = System.identityHashCode(this)

    override fun getTooltipText(): String? = DiffBundle.message("action.presentation.diff.accept.text")
    override fun getClickAction(): AnAction? = DumbAwareAction.create {
      if (documentStateWasChanged) return@create
      val target = source.other()
      val sourceDocument = source.selectNotNull(editors).document
      val targetDocument = target.selectNotNull(editors).document
      DiffUtil.executeWriteCommand(targetDocument, project, DiffBundle.message("message.replace.change.command")) {
        DiffUtil.applyModification(targetDocument, target.getStartLine(fragment), target.getEndLine(fragment),
                                   sourceDocument, source.getStartLine(fragment), source.getEndLine(fragment))
      }
    }
  }
}

class DiffMarkHighlighterModelCreator(val project: Project) : BackendHighlighterModelCreator {
  override fun accept(highlighter: RangeHighlighterEx): Boolean {
    return highlighter.blockModel != null || highlighter.innerModel != null
  }

  override fun createModel(highlighter: RangeHighlighterEx, documentVersion: RdDocumentVersion): HighlighterModel? {
    val block = highlighter.blockModel
    if (block != null) {
      return DiffBlockMarkHighlighterModel(
        block.changeId,
        block.isIgnored,
        block.isEmptyRange,
        block.isEOF,
        block.type,
        highlighter.layer,
        highlighter.targetArea == HighlighterTargetArea.EXACT_RANGE,
        documentVersion,
        highlighter.isGreedyToLeft,
        highlighter.isGreedyToRight,
        highlighter.isThinErrorStripeMark,
        null,
        highlighter.id,
        BackendHighlighterRegistrationsHost.DIFF_MARK_ID,
        highlighter.startOffset,
        highlighter.endOffset
      )
    }

    val inner = highlighter.innerModel
    if (inner != null) {
      return DiffInnerMarkHighlighterModel(
        inner.type,
        highlighter.layer,
        highlighter.targetArea == HighlighterTargetArea.EXACT_RANGE,
        documentVersion,
        highlighter.isGreedyToLeft,
        highlighter.isGreedyToRight,
        highlighter.isThinErrorStripeMark,
        null,
        highlighter.id,
        BackendHighlighterRegistrationsHost.DIFF_MARK_ID,
        highlighter.startOffset,
        highlighter.endOffset)
    }

    return null
  }
}

private val LineFragment.lineFragmentType: DiffMarkerType
  get() = when {
    startLine1 == endLine1 -> DiffMarkerType.INSERTED
    startLine2 == endLine2 -> DiffMarkerType.DELETED
    else -> DiffMarkerType.MODIFIED
  }

private val DiffFragment.innerFragmentType: DiffMarkerType
  get() = when {
    startOffset1 == endOffset1 -> DiffMarkerType.INSERTED
    startOffset2 == endOffset2 -> DiffMarkerType.DELETED
    else -> DiffMarkerType.MODIFIED
  }

private data class BlockModel(val type: DiffMarkerType,
                              val changeId: Long,
                              val isIgnored: Boolean,
                              val isEmptyRange: Boolean,
                              val isEOF: Boolean)

private data class InnerModel(val type: DiffMarkerType)

private var RangeHighlighter.blockModel: BlockModel? by userData()
private var RangeHighlighter.innerModel: InnerModel? by userData()
