package com.jetbrains.rdserver.vfs

import com.intellij.codeWithMe.ClientId
import com.intellij.openapi.application.ReadAction
import com.intellij.openapi.diagnostic.logger
import com.intellij.openapi.fileTypes.BinaryFileTypeDecompilers
import com.intellij.openapi.project.Project
import com.intellij.openapi.vfs.LocalFileSystem
import com.intellij.openapi.vfs.VirtualFile
import com.intellij.openapi.vfs.VirtualFileManager
import com.intellij.openapi.vfs.newvfs.BulkFileListener
import com.intellij.openapi.vfs.newvfs.NewVirtualFile
import com.intellij.openapi.vfs.newvfs.events.VFileDeleteEvent
import com.intellij.openapi.vfs.newvfs.events.VFileEvent
import com.intellij.openapi.vfs.newvfs.events.VFileMoveEvent
import com.intellij.openapi.vfs.newvfs.events.VFilePropertyChangeEvent
import com.intellij.testFramework.LightVirtualFileBase
import com.intellij.util.IconUtil
import com.intellij.util.concurrency.AppExecutorUtil
import com.intellij.util.containers.ContainerUtil
import com.jetbrains.codeWithMe.model.RdFileProperties
import com.jetbrains.codeWithMe.model.RdFileSnapshot
import com.jetbrains.codeWithMe.model.VfsModel
import com.jetbrains.rd.framework.Protocol
import com.jetbrains.rd.framework.impl.RdTask
import com.jetbrains.rd.ide.model.*
import com.jetbrains.rd.platform.util.application
import com.jetbrains.rd.platform.util.subscribe
import com.jetbrains.rd.platform.util.userData
import com.jetbrains.rd.util.lifetime.Lifetime
import com.jetbrains.rd.util.lifetime.isAlive
import com.jetbrains.rdserver.RdServerPerClientComponents
import com.jetbrains.rdserver.permissions.PermissionManager
import com.jetbrains.rdserver.permissions.impl.PermissionManagerImpl
import com.jetbrains.rdserver.toProtocolModel
import java.util.*

@Suppress("NonDefaultConstructor")
class BackendVirtualFileSystemHost(
  private val clientId: ClientId,
  lifetime: Lifetime,
  private val project: Project,
  protocol: Protocol
) {
  companion object {
    fun getInstance(project: Project): BackendVirtualFileSystemHost = RdServerPerClientComponents.getService(project)
    private val LOG = logger<BackendVirtualFileSystemHost>()
  }

  private val vfsModel: VfsModel = VfsModel.create(lifetime, protocol)
  private val lightVirtualFiles = ContainerUtil.createConcurrentSoftValueMap<RdFileId, VirtualFile>()

  init {
    vfsModel.findFile.set { params ->
      val fileId = params.id
      if (fileId != null) {
        val file = findVirtualFile(fileId)
        if (file != null && !file.isDirectory) {
          val value = RdFileSnapshot(fileId, file.toModel(params.sendContent))
          vfsModel.updates.fire(value)
          return@set value
        }

        vfsModel.updates.fire(RdFileSnapshot(fileId, null))
        return@set null
      }

      val path = params.path ?: return@set null
      val fileByPath = LocalFileSystem.getInstance().findFileByPath(path) ?: return@set null
      if (!PermissionManager.getInstance(project).isVisible(clientId, fileByPath)) return@set null

      val id = bindFile(fileByPath) ?: return@set null
      return@set RdFileSnapshot(id, fileByPath.toModel(params.sendContent))
    }

    vfsModel.findBadFiles.set { fileIds ->
      fileIds.filter { findVirtualFile(it) == null }
    }

    application.messageBus.subscribe(lifetime, VirtualFileManager.VFS_CHANGES, object : BulkFileListener {
      override fun after(events: List<VFileEvent>) {
        for (event in events) {
          if (event is VFilePropertyChangeEvent &&
              (event.propertyName == VirtualFile.PROP_WRITABLE || event.propertyName == VirtualFile.PROP_NAME) ||
              event is VFileMoveEvent ||
              event is VFileDeleteEvent
          ) {
            val file = event.file ?: continue
            val fileId = createId(file)
            if (fileId in vfsModel.files) {
              if (PermissionManager.getInstance(project).isVisible(clientId, file)) {
                vfsModel.updates.fire(RdFileSnapshot(fileId, file.toModel(sendContent = false)))
              }
              else {
                vfsModel.files.remove(fileId)
              }
            }
          }
        }
      }
    })

    vfsModel.getFileIcon.set { lt, param ->
      LOG.info("getFileIcon request for $param...")
      val result = RdTask<IconModel?>()
      ReadAction.nonBlocking {
        val file = findVirtualFile(param.fileId)
        val icon = file?.let { IconUtil.getIcon(file, param.flags, project) }
        result.set(icon?.toProtocolModel())
      }.expireWhen { !lt.isAlive }.submit(AppExecutorUtil.getAppExecutorService())

      return@set result
    }
    
    vfsModel.getContent.set { fileId ->
      val file = findVirtualFile(fileId)
      file?.contentsToByteArrayDecompileIfNeeded() ?: byteArrayOf()
    }

    val permissionManager = PermissionManagerImpl.getInstanceImpl(project)
    permissionManager.permissionModelChanged.advise(lifetime) {
      vfsModel.files
        .filter { findVirtualFile(it.key) == null }
        .forEach {
          vfsModel.files.remove(it.key)
        }
    }
  }

  fun findVirtualFile(fileId: RdFileId): VirtualFile? {
    val file = when (fileId) {
      is RdFileIdNew -> {
        VirtualFileManager.getInstance().findFileById(fileId.id) ?: return null
      }
      is RdFileIdOld -> {
        val fileSystem = VirtualFileManager.getInstance().getFileSystem(fileId.protocol)
        if (fileSystem == null) {
          LOG.error("file system is not found for $fileId")
          return null
        }
        fileSystem.findFileByPath(fileId.path) ?: return null
      }
      is RdFileIdInMem -> {
        return lightVirtualFiles[fileId]
      }
      else -> return null
    }

    if (!PermissionManager.getInstance(project).isVisible(ClientId.current, file))
      return null
    
    return file
  }

  private fun VirtualFile.toModel(sendContent: Boolean) = RdFileProperties(
    name,
    path,
    timeStamp,
    modificationStamp,
    length,
    if (sendContent) contentsToByteArray() else null,
    isWritable
  )

  private fun VirtualFile.contentsToByteArrayDecompileIfNeeded(): ByteArray {
    if (!this.fileType.isBinary)
      return contentsToByteArray()

    try {
      val decompiler = BinaryFileTypeDecompilers.getInstance().forFileType(fileType)
      if (decompiler != null) {
        LOG.info("decompiling file $this...")
        val text = decompiler.decompile(this)

        return text.toString().toByteArray()
      }
    }
    catch (t: Throwable) {
      LOG.warn("error while decompiling file $this", t)
    }

    return contentsToByteArray()
  }

  /**
   * @param sendProperties specifies whether to send initial data about the file (name, isWritable, etc.)
   */
  fun bindFile(file: VirtualFile, sendProperties: Boolean = true, sendContent: Boolean = false): RdFileId? {
    if (!PermissionManager.getInstance(project).isVisible(clientId, file)) return null

    val fileId = createId(file)
    if (!sendProperties) return fileId

    application.assertIsDispatchThread()

    LOG.info("publishing file $file")
    vfsModel.files[fileId] = file.toModel(sendContent)
    return fileId
  }

  private fun createId(file: VirtualFile): RdFileId {
    return if (file is NewVirtualFile) {
      RdFileIdNew(file.id)
    }
    else if (file is LightVirtualFileBase) {
      val existingId = file.fileId
      if (existingId != null) return existingId
      file.fileId = RdFileIdInMem(UUID.randomUUID().toString())
      lightVirtualFiles[file.fileId] = file
      return file.fileId!!
    } else {
      val protocol = file.fileSystem.protocol
      RdFileIdOld(protocol, file.path)
    }
  }
}

private var LightVirtualFileBase.fileId: RdFileId? by userData()

/**
 * Temp solution for code that uses RdDocumentId while meaning RdFileId
 */
fun createDocumentIdFromFile(project: Project, file: VirtualFile?): RdDocumentId? {
  if (file == null) return null
  val fileId = BackendVirtualFileSystemHost.getInstance(project).bindFile(file, false) ?: return null
  return RdDocumentIdWithFileId(fileId)
}
